from ml4h.TensorMap import TensorMap

field_id_102_pulse_rate_automated_reading_0_0 = TensorMap('102_Pulse-rate-automated-reading_0_0', loss='logcosh', channel_map={'no_102_Pulse-rate-automated-reading_0_0': 0})
field_id_102_pulse_rate_automated_reading_0_1 = TensorMap('102_Pulse-rate-automated-reading_0_1', loss='logcosh', channel_map={'no_102_Pulse-rate-automated-reading_0_1': 0})
field_id_1050_time_spend_outdoors_in_summer_0_0 = TensorMap('1050_Time-spend-outdoors-in-summer_0_0', loss='logcosh', channel_map={'no_1050_Time-spend-outdoors-in-summer_0_0': 0})
field_id_1060_time_spent_outdoors_in_winter_0_0 = TensorMap('1060_Time-spent-outdoors-in-winter_0_0', loss='logcosh', channel_map={'no_1060_Time-spent-outdoors-in-winter_0_0': 0})
field_id_1070_time_spent_watching_television_tv_0_0 = TensorMap('1070_Time-spent-watching-television-TV_0_0', loss='logcosh', channel_map={'no_1070_Time-spent-watching-television-TV_0_0': 0})
field_id_1080_time_spent_using_computer_0_0 = TensorMap('1080_Time-spent-using-computer_0_0', loss='logcosh', channel_map={'no_1080_Time-spent-using-computer_0_0': 0})
field_id_1090_time_spent_driving_0_0 = TensorMap('1090_Time-spent-driving_0_0', loss='logcosh', channel_map={'no_1090_Time-spent-driving_0_0': 0})
field_id_1269_exposure_to_tobacco_smoke_at_home_0_0 = TensorMap('1269_Exposure-to-tobacco-smoke-at-home_0_0', loss='logcosh', channel_map={'no_1269_Exposure-to-tobacco-smoke-at-home_0_0': 0})
field_id_1279_exposure_to_tobacco_smoke_outside_home_0_0 = TensorMap('1279_Exposure-to-tobacco-smoke-outside-home_0_0', loss='logcosh', channel_map={'no_1279_Exposure-to-tobacco-smoke-outside-home_0_0': 0})
field_id_1289_cooked_vegetable_intake_0_0 = TensorMap('1289_Cooked-vegetable-intake_0_0', loss='logcosh', channel_map={'no_1289_Cooked-vegetable-intake_0_0': 0})
field_id_1299_salad_raw_vegetable_intake_0_0 = TensorMap('1299_Salad-raw-vegetable-intake_0_0', loss='logcosh', channel_map={'no_1299_Salad-raw-vegetable-intake_0_0': 0})
field_id_129_place_of_birth_in_uk_north_coordinate_0_0 = TensorMap('129_Place-of-birth-in-UK-north-coordinate_0_0', loss='logcosh', channel_map={'no_129_Place-of-birth-in-UK-north-coordinate_0_0': 0})
field_id_1309_fresh_fruit_intake_0_0 = TensorMap('1309_Fresh-fruit-intake_0_0', loss='logcosh', channel_map={'no_1309_Fresh-fruit-intake_0_0': 0})
field_id_130_place_of_birth_in_uk_east_coordinate_0_0 = TensorMap('130_Place-of-birth-in-UK-east-coordinate_0_0', loss='logcosh', channel_map={'no_130_Place-of-birth-in-UK-east-coordinate_0_0': 0})
field_id_1319_dried_fruit_intake_0_0 = TensorMap('1319_Dried-fruit-intake_0_0', loss='logcosh', channel_map={'no_1319_Dried-fruit-intake_0_0': 0})
field_id_134_number_of_selfreported_cancers_0_0 = TensorMap('134_Number-of-selfreported-cancers_0_0', loss='logcosh', channel_map={'no_134_Number-of-selfreported-cancers_0_0': 0})
field_id_135_number_of_selfreported_noncancer_illnesses_0_0 = TensorMap('135_Number-of-selfreported-noncancer-illnesses_0_0', loss='logcosh', channel_map={'no_135_Number-of-selfreported-noncancer-illnesses_0_0': 0})
field_id_136_number_of_operations_selfreported_0_0 = TensorMap('136_Number-of-operations-selfreported_0_0', loss='logcosh', channel_map={'no_136_Number-of-operations-selfreported_0_0': 0})
field_id_137_number_of_treatmentsmedications_taken_0_0 = TensorMap('137_Number-of-treatmentsmedications-taken_0_0', loss='logcosh', channel_map={'no_137_Number-of-treatmentsmedications-taken_0_0': 0})
field_id_1438_bread_intake_0_0 = TensorMap('1438_Bread-intake_0_0', loss='logcosh', channel_map={'no_1438_Bread-intake_0_0': 0})
field_id_1458_cereal_intake_0_0 = TensorMap('1458_Cereal-intake_0_0', loss='logcosh', channel_map={'no_1458_Cereal-intake_0_0': 0})
field_id_1488_tea_intake_0_0 = TensorMap('1488_Tea-intake_0_0', loss='logcosh', channel_map={'no_1488_Tea-intake_0_0': 0})
field_id_1498_coffee_intake_0_0 = TensorMap('1498_Coffee-intake_0_0', loss='logcosh', channel_map={'no_1498_Coffee-intake_0_0': 0})
field_id_1528_water_intake_0_0 = TensorMap('1528_Water-intake_0_0', loss='logcosh', channel_map={'no_1528_Water-intake_0_0': 0})
field_id_1568_average_weekly_red_wine_intake_0_0 = TensorMap('1568_Average-weekly-red-wine-intake_0_0', loss='logcosh', channel_map={'no_1568_Average-weekly-red-wine-intake_0_0': 0})
field_id_1578_average_weekly_champagne_plus_white_wine_intake_0_0 = TensorMap('1578_Average-weekly-champagne-plus-white-wine-intake_0_0', loss='logcosh', channel_map={'no_1578_Average-weekly-champagne-plus-white-wine-intake_0_0': 0})
field_id_1588_average_weekly_beer_plus_cider_intake_0_0 = TensorMap('1588_Average-weekly-beer-plus-cider-intake_0_0', loss='logcosh', channel_map={'no_1588_Average-weekly-beer-plus-cider-intake_0_0': 0})
field_id_1598_average_weekly_spirits_intake_0_0 = TensorMap('1598_Average-weekly-spirits-intake_0_0', loss='logcosh', channel_map={'no_1598_Average-weekly-spirits-intake_0_0': 0})
field_id_1608_average_weekly_fortified_wine_intake_0_0 = TensorMap('1608_Average-weekly-fortified-wine-intake_0_0', loss='logcosh', channel_map={'no_1608_Average-weekly-fortified-wine-intake_0_0': 0})
field_id_1737_childhood_sunburn_occasions_0_0 = TensorMap('1737_Childhood-sunburn-occasions_0_0', loss='logcosh', channel_map={'no_1737_Childhood-sunburn-occasions_0_0': 0})
field_id_1845_mothers_age_0_0 = TensorMap('1845_Mothers-age_0_0', loss='logcosh', channel_map={'no_1845_Mothers-age_0_0': 0})
field_id_1873_number_of_full_brothers_0_0 = TensorMap('1873_Number-of-full-brothers_0_0', loss='logcosh', channel_map={'no_1873_Number-of-full-brothers_0_0': 0})
field_id_1883_number_of_full_sisters_0_0 = TensorMap('1883_Number-of-full-sisters_0_0', loss='logcosh', channel_map={'no_1883_Number-of-full-sisters_0_0': 0})
field_id_189_townsend_deprivation_index_at_recruitment_0_0 = TensorMap('189_Townsend-deprivation-index-at-recruitment_0_0', loss='logcosh', channel_map={'no_189_Townsend-deprivation-index-at-recruitment_0_0': 0})
field_id_20015_sitting_height_0_0 = TensorMap('20015_Sitting-height_0_0', loss='logcosh', channel_map={'no_20015_Sitting-height_0_0': 0})
field_id_20023_mean_time_to_correctly_identify_matches_0_0 = TensorMap('20023_Mean-time-to-correctly-identify-matches_0_0', loss='logcosh', channel_map={'no_20023_Mean-time-to-correctly-identify-matches_0_0': 0})
field_id_20159_number_of_symbol_digit_matches_made_correctly_0_0 = TensorMap('20159_Number-of-symbol-digit-matches-made-correctly_0_0', loss='logcosh', channel_map={'no_20159_Number-of-symbol-digit-matches-made-correctly_0_0': 0})
field_id_20191_fluid_intelligence_score_0_0 = TensorMap('20191_Fluid-intelligence-score_0_0', loss='logcosh', channel_map={'no_20191_Fluid-intelligence-score_0_0': 0})
field_id_21001_body_mass_index_bmi_0_0 = TensorMap('21001_Body-mass-index-BMI_0_0', loss='logcosh', channel_map={'no_21001_Body-mass-index-BMI_0_0': 0})
field_id_21002_weight_0_0 = TensorMap('21002_Weight_0_0', loss='logcosh', channel_map={'no_21002_Weight_0_0': 0})
field_id_21003_age_when_attended_assessment_centre_0_0 = TensorMap('21003_Age-when-attended-assessment-centre_0_0', loss='logcosh', channel_map={'no_21003_Age-when-attended-assessment-centre_0_0': 0})
field_id_22003_heterozygosity_0_0 = TensorMap('22003_Heterozygosity_0_0', loss='logcosh', channel_map={'no_22003_Heterozygosity_0_0': 0})
field_id_22004_heterozygosity_pca_corrected_0_0 = TensorMap('22004_Heterozygosity-PCA-corrected_0_0', loss='logcosh', channel_map={'no_22004_Heterozygosity-PCA-corrected_0_0': 0})
field_id_22005_missingness_0_0 = TensorMap('22005_Missingness_0_0', loss='logcosh', channel_map={'no_22005_Missingness_0_0': 0})
field_id_22009_genetic_principal_components_0_1 = TensorMap('22009_Genetic-principal-components_0_1', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_1': 0})
field_id_22009_genetic_principal_components_0_10 = TensorMap('22009_Genetic-principal-components_0_10', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_10': 0})
field_id_22009_genetic_principal_components_0_11 = TensorMap('22009_Genetic-principal-components_0_11', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_11': 0})
field_id_22009_genetic_principal_components_0_12 = TensorMap('22009_Genetic-principal-components_0_12', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_12': 0})
field_id_22009_genetic_principal_components_0_13 = TensorMap('22009_Genetic-principal-components_0_13', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_13': 0})
field_id_22009_genetic_principal_components_0_14 = TensorMap('22009_Genetic-principal-components_0_14', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_14': 0})
field_id_22009_genetic_principal_components_0_15 = TensorMap('22009_Genetic-principal-components_0_15', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_15': 0})
field_id_22009_genetic_principal_components_0_16 = TensorMap('22009_Genetic-principal-components_0_16', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_16': 0})
field_id_22009_genetic_principal_components_0_17 = TensorMap('22009_Genetic-principal-components_0_17', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_17': 0})
field_id_22009_genetic_principal_components_0_18 = TensorMap('22009_Genetic-principal-components_0_18', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_18': 0})
field_id_22009_genetic_principal_components_0_19 = TensorMap('22009_Genetic-principal-components_0_19', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_19': 0})
field_id_22009_genetic_principal_components_0_2 = TensorMap('22009_Genetic-principal-components_0_2', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_2': 0})
field_id_22009_genetic_principal_components_0_20 = TensorMap('22009_Genetic-principal-components_0_20', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_20': 0})
field_id_22009_genetic_principal_components_0_21 = TensorMap('22009_Genetic-principal-components_0_21', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_21': 0})
field_id_22009_genetic_principal_components_0_22 = TensorMap('22009_Genetic-principal-components_0_22', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_22': 0})
field_id_22009_genetic_principal_components_0_23 = TensorMap('22009_Genetic-principal-components_0_23', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_23': 0})
field_id_22009_genetic_principal_components_0_24 = TensorMap('22009_Genetic-principal-components_0_24', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_24': 0})
field_id_22009_genetic_principal_components_0_25 = TensorMap('22009_Genetic-principal-components_0_25', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_25': 0})
field_id_22009_genetic_principal_components_0_26 = TensorMap('22009_Genetic-principal-components_0_26', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_26': 0})
field_id_22009_genetic_principal_components_0_27 = TensorMap('22009_Genetic-principal-components_0_27', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_27': 0})
field_id_22009_genetic_principal_components_0_28 = TensorMap('22009_Genetic-principal-components_0_28', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_28': 0})
field_id_22009_genetic_principal_components_0_29 = TensorMap('22009_Genetic-principal-components_0_29', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_29': 0})
field_id_22009_genetic_principal_components_0_3 = TensorMap('22009_Genetic-principal-components_0_3', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_3': 0})
field_id_22009_genetic_principal_components_0_30 = TensorMap('22009_Genetic-principal-components_0_30', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_30': 0})
field_id_22009_genetic_principal_components_0_31 = TensorMap('22009_Genetic-principal-components_0_31', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_31': 0})
field_id_22009_genetic_principal_components_0_32 = TensorMap('22009_Genetic-principal-components_0_32', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_32': 0})
field_id_22009_genetic_principal_components_0_33 = TensorMap('22009_Genetic-principal-components_0_33', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_33': 0})
field_id_22009_genetic_principal_components_0_34 = TensorMap('22009_Genetic-principal-components_0_34', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_34': 0})
field_id_22009_genetic_principal_components_0_35 = TensorMap('22009_Genetic-principal-components_0_35', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_35': 0})
field_id_22009_genetic_principal_components_0_36 = TensorMap('22009_Genetic-principal-components_0_36', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_36': 0})
field_id_22009_genetic_principal_components_0_37 = TensorMap('22009_Genetic-principal-components_0_37', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_37': 0})
field_id_22009_genetic_principal_components_0_38 = TensorMap('22009_Genetic-principal-components_0_38', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_38': 0})
field_id_22009_genetic_principal_components_0_39 = TensorMap('22009_Genetic-principal-components_0_39', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_39': 0})
field_id_22009_genetic_principal_components_0_4 = TensorMap('22009_Genetic-principal-components_0_4', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_4': 0})
field_id_22009_genetic_principal_components_0_40 = TensorMap('22009_Genetic-principal-components_0_40', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_40': 0})
field_id_22009_genetic_principal_components_0_5 = TensorMap('22009_Genetic-principal-components_0_5', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_5': 0})
field_id_22009_genetic_principal_components_0_6 = TensorMap('22009_Genetic-principal-components_0_6', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_6': 0})
field_id_22009_genetic_principal_components_0_7 = TensorMap('22009_Genetic-principal-components_0_7', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_7': 0})
field_id_22009_genetic_principal_components_0_8 = TensorMap('22009_Genetic-principal-components_0_8', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_8': 0})
field_id_22009_genetic_principal_components_0_9 = TensorMap('22009_Genetic-principal-components_0_9', loss='logcosh', channel_map={'no_22009_Genetic-principal-components_0_9': 0})
field_id_22200_year_of_birth_0_0 = TensorMap('22200_Year-of-birth_0_0', loss='logcosh', channel_map={'no_22200_Year-of-birth_0_0': 0})
field_id_22599_number_of_jobs_held_0_0 = TensorMap('22599_Number-of-jobs-held_0_0', loss='logcosh', channel_map={'no_22599_Number-of-jobs-held_0_0': 0})
field_id_22602_year_job_started_0_0 = TensorMap('22602_Year-job-started_0_0', loss='logcosh', channel_map={'no_22602_Year-job-started_0_0': 0})
field_id_22602_year_job_started_0_1 = TensorMap('22602_Year-job-started_0_1', loss='logcosh', channel_map={'no_22602_Year-job-started_0_1': 0})
field_id_22603_year_job_ended_0_0 = TensorMap('22603_Year-job-ended_0_0', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_0': 0})
field_id_22603_year_job_ended_0_1 = TensorMap('22603_Year-job-ended_0_1', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_1': 0})
field_id_22605_work_hours_per_week_exact_value_0_1 = TensorMap('22605_Work-hours-per-week-exact-value_0_1', loss='logcosh', channel_map={'no_22605_Work-hours-per-week-exact-value_0_1': 0})
field_id_22661_number_of_gap_periods_0_0 = TensorMap('22661_Number-of-gap-periods_0_0', loss='logcosh', channel_map={'no_22661_Number-of-gap-periods_0_0': 0})
field_id_2277_frequency_of_solariumsunlamp_use_0_0 = TensorMap('2277_Frequency-of-solariumsunlamp-use_0_0', loss='logcosh', channel_map={'no_2277_Frequency-of-solariumsunlamp-use_0_0': 0})
field_id_23098_weight_0_0 = TensorMap('23098_Weight_0_0', loss='logcosh', channel_map={'no_23098_Weight_0_0': 0})
field_id_23099_body_fat_percentage_0_0 = TensorMap('23099_Body-fat-percentage_0_0', loss='logcosh', channel_map={'no_23099_Body-fat-percentage_0_0': 0})
field_id_23100_whole_body_fat_mass_0_0 = TensorMap('23100_Whole-body-fat-mass_0_0', loss='logcosh', channel_map={'no_23100_Whole-body-fat-mass_0_0': 0})
field_id_23101_whole_body_fatfree_mass_0_0 = TensorMap('23101_Whole-body-fatfree-mass_0_0', loss='logcosh', channel_map={'no_23101_Whole-body-fatfree-mass_0_0': 0})
field_id_23102_whole_body_water_mass_0_0 = TensorMap('23102_Whole-body-water-mass_0_0', loss='logcosh', channel_map={'no_23102_Whole-body-water-mass_0_0': 0})
field_id_23104_body_mass_index_bmi_0_0 = TensorMap('23104_Body-mass-index-BMI_0_0', loss='logcosh', channel_map={'no_23104_Body-mass-index-BMI_0_0': 0})
field_id_23105_basal_metabolic_rate_0_0 = TensorMap('23105_Basal-metabolic-rate_0_0', loss='logcosh', channel_map={'no_23105_Basal-metabolic-rate_0_0': 0})
field_id_23106_impedance_of_whole_body_0_0 = TensorMap('23106_Impedance-of-whole-body_0_0', loss='logcosh', channel_map={'no_23106_Impedance-of-whole-body_0_0': 0})
field_id_23107_impedance_of_leg_right_0_0 = TensorMap('23107_Impedance-of-leg-right_0_0', loss='logcosh', channel_map={'no_23107_Impedance-of-leg-right_0_0': 0})
field_id_23108_impedance_of_leg_left_0_0 = TensorMap('23108_Impedance-of-leg-left_0_0', loss='logcosh', channel_map={'no_23108_Impedance-of-leg-left_0_0': 0})
field_id_23109_impedance_of_arm_right_0_0 = TensorMap('23109_Impedance-of-arm-right_0_0', loss='logcosh', channel_map={'no_23109_Impedance-of-arm-right_0_0': 0})
field_id_23110_impedance_of_arm_left_0_0 = TensorMap('23110_Impedance-of-arm-left_0_0', loss='logcosh', channel_map={'no_23110_Impedance-of-arm-left_0_0': 0})
field_id_23111_leg_fat_percentage_right_0_0 = TensorMap('23111_Leg-fat-percentage-right_0_0', loss='logcosh', channel_map={'no_23111_Leg-fat-percentage-right_0_0': 0})
field_id_23112_leg_fat_mass_right_0_0 = TensorMap('23112_Leg-fat-mass-right_0_0', loss='logcosh', channel_map={'no_23112_Leg-fat-mass-right_0_0': 0})
field_id_23113_leg_fatfree_mass_right_0_0 = TensorMap('23113_Leg-fatfree-mass-right_0_0', loss='logcosh', channel_map={'no_23113_Leg-fatfree-mass-right_0_0': 0})
field_id_23114_leg_predicted_mass_right_0_0 = TensorMap('23114_Leg-predicted-mass-right_0_0', loss='logcosh', channel_map={'no_23114_Leg-predicted-mass-right_0_0': 0})
field_id_23115_leg_fat_percentage_left_0_0 = TensorMap('23115_Leg-fat-percentage-left_0_0', loss='logcosh', channel_map={'no_23115_Leg-fat-percentage-left_0_0': 0})
field_id_23116_leg_fat_mass_left_0_0 = TensorMap('23116_Leg-fat-mass-left_0_0', loss='logcosh', channel_map={'no_23116_Leg-fat-mass-left_0_0': 0})
field_id_23117_leg_fatfree_mass_left_0_0 = TensorMap('23117_Leg-fatfree-mass-left_0_0', loss='logcosh', channel_map={'no_23117_Leg-fatfree-mass-left_0_0': 0})
field_id_23118_leg_predicted_mass_left_0_0 = TensorMap('23118_Leg-predicted-mass-left_0_0', loss='logcosh', channel_map={'no_23118_Leg-predicted-mass-left_0_0': 0})
field_id_23119_arm_fat_percentage_right_0_0 = TensorMap('23119_Arm-fat-percentage-right_0_0', loss='logcosh', channel_map={'no_23119_Arm-fat-percentage-right_0_0': 0})
field_id_23120_arm_fat_mass_right_0_0 = TensorMap('23120_Arm-fat-mass-right_0_0', loss='logcosh', channel_map={'no_23120_Arm-fat-mass-right_0_0': 0})
field_id_23121_arm_fatfree_mass_right_0_0 = TensorMap('23121_Arm-fatfree-mass-right_0_0', loss='logcosh', channel_map={'no_23121_Arm-fatfree-mass-right_0_0': 0})
field_id_23122_arm_predicted_mass_right_0_0 = TensorMap('23122_Arm-predicted-mass-right_0_0', loss='logcosh', channel_map={'no_23122_Arm-predicted-mass-right_0_0': 0})
field_id_23123_arm_fat_percentage_left_0_0 = TensorMap('23123_Arm-fat-percentage-left_0_0', loss='logcosh', channel_map={'no_23123_Arm-fat-percentage-left_0_0': 0})
field_id_23124_arm_fat_mass_left_0_0 = TensorMap('23124_Arm-fat-mass-left_0_0', loss='logcosh', channel_map={'no_23124_Arm-fat-mass-left_0_0': 0})
field_id_23125_arm_fatfree_mass_left_0_0 = TensorMap('23125_Arm-fatfree-mass-left_0_0', loss='logcosh', channel_map={'no_23125_Arm-fatfree-mass-left_0_0': 0})
field_id_23126_arm_predicted_mass_left_0_0 = TensorMap('23126_Arm-predicted-mass-left_0_0', loss='logcosh', channel_map={'no_23126_Arm-predicted-mass-left_0_0': 0})
field_id_23127_trunk_fat_percentage_0_0 = TensorMap('23127_Trunk-fat-percentage_0_0', loss='logcosh', channel_map={'no_23127_Trunk-fat-percentage_0_0': 0})
field_id_23128_trunk_fat_mass_0_0 = TensorMap('23128_Trunk-fat-mass_0_0', loss='logcosh', channel_map={'no_23128_Trunk-fat-mass_0_0': 0})
field_id_23129_trunk_fatfree_mass_0_0 = TensorMap('23129_Trunk-fatfree-mass_0_0', loss='logcosh', channel_map={'no_23129_Trunk-fatfree-mass_0_0': 0})
field_id_23130_trunk_predicted_mass_0_0 = TensorMap('23130_Trunk-predicted-mass_0_0', loss='logcosh', channel_map={'no_23130_Trunk-predicted-mass_0_0': 0})
field_id_2405_number_of_children_fathered_0_0 = TensorMap('2405_Number-of-children-fathered_0_0', loss='logcosh', channel_map={'no_2405_Number-of-children-fathered_0_0': 0})
field_id_30000_white_blood_cell_leukocyte_count_0_0 = TensorMap('30000_White-blood-cell-leukocyte-count_0_0', loss='logcosh', channel_map={'no_30000_White-blood-cell-leukocyte-count_0_0': 0})
field_id_30010_red_blood_cell_erythrocyte_count_0_0 = TensorMap('30010_Red-blood-cell-erythrocyte-count_0_0', loss='logcosh', channel_map={'no_30010_Red-blood-cell-erythrocyte-count_0_0': 0})
field_id_30020_haemoglobin_concentration_0_0 = TensorMap('30020_Haemoglobin-concentration_0_0', loss='logcosh', channel_map={'no_30020_Haemoglobin-concentration_0_0': 0})
field_id_30030_haematocrit_percentage_0_0 = TensorMap('30030_Haematocrit-percentage_0_0', loss='logcosh', channel_map={'no_30030_Haematocrit-percentage_0_0': 0})
field_id_30040_mean_corpuscular_volume_0_0 = TensorMap('30040_Mean-corpuscular-volume_0_0', loss='logcosh', channel_map={'no_30040_Mean-corpuscular-volume_0_0': 0})
field_id_30050_mean_corpuscular_haemoglobin_0_0 = TensorMap('30050_Mean-corpuscular-haemoglobin_0_0', loss='logcosh', channel_map={'no_30050_Mean-corpuscular-haemoglobin_0_0': 0})
field_id_30060_mean_corpuscular_haemoglobin_concentration_0_0 = TensorMap('30060_Mean-corpuscular-haemoglobin-concentration_0_0', loss='logcosh', channel_map={'no_30060_Mean-corpuscular-haemoglobin-concentration_0_0': 0})
field_id_30070_red_blood_cell_erythrocyte_distribution_width_0_0 = TensorMap('30070_Red-blood-cell-erythrocyte-distribution-width_0_0', loss='logcosh', channel_map={'no_30070_Red-blood-cell-erythrocyte-distribution-width_0_0': 0})
field_id_30080_platelet_count_0_0 = TensorMap('30080_Platelet-count_0_0', loss='logcosh', channel_map={'no_30080_Platelet-count_0_0': 0})
field_id_30090_platelet_crit_0_0 = TensorMap('30090_Platelet-crit_0_0', loss='logcosh', channel_map={'no_30090_Platelet-crit_0_0': 0})
field_id_30100_mean_platelet_thrombocyte_volume_0_0 = TensorMap('30100_Mean-platelet-thrombocyte-volume_0_0', loss='logcosh', channel_map={'no_30100_Mean-platelet-thrombocyte-volume_0_0': 0})
field_id_30110_platelet_distribution_width_0_0 = TensorMap('30110_Platelet-distribution-width_0_0', loss='logcosh', channel_map={'no_30110_Platelet-distribution-width_0_0': 0})
field_id_30120_lymphocyte_count_0_0 = TensorMap('30120_Lymphocyte-count_0_0', loss='logcosh', channel_map={'no_30120_Lymphocyte-count_0_0': 0})
field_id_30130_monocyte_count_0_0 = TensorMap('30130_Monocyte-count_0_0', loss='logcosh', channel_map={'no_30130_Monocyte-count_0_0': 0})
field_id_30140_neutrophill_count_0_0 = TensorMap('30140_Neutrophill-count_0_0', loss='logcosh', channel_map={'no_30140_Neutrophill-count_0_0': 0})
field_id_30150_eosinophill_count_0_0 = TensorMap('30150_Eosinophill-count_0_0', loss='logcosh', channel_map={'no_30150_Eosinophill-count_0_0': 0})
field_id_30160_basophill_count_0_0 = TensorMap('30160_Basophill-count_0_0', loss='logcosh', channel_map={'no_30160_Basophill-count_0_0': 0})
field_id_30170_nucleated_red_blood_cell_count_0_0 = TensorMap('30170_Nucleated-red-blood-cell-count_0_0', loss='logcosh', channel_map={'no_30170_Nucleated-red-blood-cell-count_0_0': 0})
field_id_30180_lymphocyte_percentage_0_0 = TensorMap('30180_Lymphocyte-percentage_0_0', loss='logcosh', channel_map={'no_30180_Lymphocyte-percentage_0_0': 0})
field_id_30190_monocyte_percentage_0_0 = TensorMap('30190_Monocyte-percentage_0_0', loss='logcosh', channel_map={'no_30190_Monocyte-percentage_0_0': 0})
field_id_30200_neutrophill_percentage_0_0 = TensorMap('30200_Neutrophill-percentage_0_0', loss='logcosh', channel_map={'no_30200_Neutrophill-percentage_0_0': 0})
field_id_30210_eosinophill_percentage_0_0 = TensorMap('30210_Eosinophill-percentage_0_0', loss='logcosh', channel_map={'no_30210_Eosinophill-percentage_0_0': 0})
field_id_30220_basophill_percentage_0_0 = TensorMap('30220_Basophill-percentage_0_0', loss='logcosh', channel_map={'no_30220_Basophill-percentage_0_0': 0})
field_id_30230_nucleated_red_blood_cell_percentage_0_0 = TensorMap('30230_Nucleated-red-blood-cell-percentage_0_0', loss='logcosh', channel_map={'no_30230_Nucleated-red-blood-cell-percentage_0_0': 0})
field_id_30240_reticulocyte_percentage_0_0 = TensorMap('30240_Reticulocyte-percentage_0_0', loss='logcosh', channel_map={'no_30240_Reticulocyte-percentage_0_0': 0})
field_id_30250_reticulocyte_count_0_0 = TensorMap('30250_Reticulocyte-count_0_0', loss='logcosh', channel_map={'no_30250_Reticulocyte-count_0_0': 0})
field_id_30260_mean_reticulocyte_volume_0_0 = TensorMap('30260_Mean-reticulocyte-volume_0_0', loss='logcosh', channel_map={'no_30260_Mean-reticulocyte-volume_0_0': 0})
field_id_30270_mean_sphered_cell_volume_0_0 = TensorMap('30270_Mean-sphered-cell-volume_0_0', loss='logcosh', channel_map={'no_30270_Mean-sphered-cell-volume_0_0': 0})
field_id_30280_immature_reticulocyte_fraction_0_0 = TensorMap('30280_Immature-reticulocyte-fraction_0_0', loss='logcosh', channel_map={'no_30280_Immature-reticulocyte-fraction_0_0': 0})
field_id_30290_high_light_scatter_reticulocyte_percentage_0_0 = TensorMap('30290_High-light-scatter-reticulocyte-percentage_0_0', loss='logcosh', channel_map={'no_30290_High-light-scatter-reticulocyte-percentage_0_0': 0})
field_id_30300_high_light_scatter_reticulocyte_count_0_0 = TensorMap('30300_High-light-scatter-reticulocyte-count_0_0', loss='logcosh', channel_map={'no_30300_High-light-scatter-reticulocyte-count_0_0': 0})
field_id_30314_volume_of_edta1_plasma_held_by_ukb_0_0 = TensorMap('30314_Volume-of-EDTA1-plasma-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30314_Volume-of-EDTA1-plasma-held-by-UKB_0_0': 0})
field_id_30324_volume_of_edta1_red_cells_held_by_ukb_0_0 = TensorMap('30324_Volume-of-EDTA1-red-cells-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30324_Volume-of-EDTA1-red-cells-held-by-UKB_0_0': 0})
field_id_30334_volume_of_edta1_buffy_held_by_ukb_0_0 = TensorMap('30334_Volume-of-EDTA1-buffy-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30334_Volume-of-EDTA1-buffy-held-by-UKB_0_0': 0})
field_id_30344_volume_of_edta2_plasma_held_by_ukb_0_0 = TensorMap('30344_Volume-of-EDTA2-plasma-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30344_Volume-of-EDTA2-plasma-held-by-UKB_0_0': 0})
field_id_30354_volume_of_edta2_buffy_held_by_ukb_0_0 = TensorMap('30354_Volume-of-EDTA2-buffy-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30354_Volume-of-EDTA2-buffy-held-by-UKB_0_0': 0})
field_id_30364_volume_of_edta2_red_cells_held_by_ukb_0_0 = TensorMap('30364_Volume-of-EDTA2-red-cells-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30364_Volume-of-EDTA2-red-cells-held-by-UKB_0_0': 0})
field_id_30374_volume_of_lihep_plasma_held_by_ukb_0_0 = TensorMap('30374_Volume-of-LiHep-plasma-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30374_Volume-of-LiHep-plasma-held-by-UKB_0_0': 0})
field_id_30384_volume_of_serum_held_by_ukb_0_0 = TensorMap('30384_Volume-of-serum-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30384_Volume-of-serum-held-by-UKB_0_0': 0})
field_id_30404_volume_of_acd_held_by_ukb_0_0 = TensorMap('30404_Volume-of-ACD-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30404_Volume-of-ACD-held-by-UKB_0_0': 0})
field_id_30510_creatinine_enzymatic_in_urine_0_0 = TensorMap('30510_Creatinine-enzymatic-in-urine_0_0', loss='logcosh', channel_map={'no_30510_Creatinine-enzymatic-in-urine_0_0': 0})
field_id_30520_potassium_in_urine_0_0 = TensorMap('30520_Potassium-in-urine_0_0', loss='logcosh', channel_map={'no_30520_Potassium-in-urine_0_0': 0})
field_id_30530_sodium_in_urine_0_0 = TensorMap('30530_Sodium-in-urine_0_0', loss='logcosh', channel_map={'no_30530_Sodium-in-urine_0_0': 0})
field_id_3062_forced_vital_capacity_fvc_0_0 = TensorMap('3062_Forced-vital-capacity-FVC_0_0', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_0_0': 0})
field_id_3062_forced_vital_capacity_fvc_0_1 = TensorMap('3062_Forced-vital-capacity-FVC_0_1', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_0_1': 0})
field_id_3062_forced_vital_capacity_fvc_0_2 = TensorMap('3062_Forced-vital-capacity-FVC_0_2', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_0_2': 0})
field_id_30630_apolipoprotein_a_0_0 = TensorMap('30630_Apolipoprotein-A_0_0', loss='logcosh', channel_map={'no_30630_Apolipoprotein-A_0_0': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_0_0 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_0_0', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_0_0': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_0_1 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_0_1', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_0_1': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_0_2 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_0_2', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_0_2': 0})
field_id_30640_apolipoprotein_b_0_0 = TensorMap('30640_Apolipoprotein-B_0_0', loss='logcosh', channel_map={'no_30640_Apolipoprotein-B_0_0': 0})
field_id_3064_peak_expiratory_flow_pef_0_0 = TensorMap('3064_Peak-expiratory-flow-PEF_0_0', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_0_0': 0})
field_id_3064_peak_expiratory_flow_pef_0_1 = TensorMap('3064_Peak-expiratory-flow-PEF_0_1', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_0_1': 0})
field_id_3064_peak_expiratory_flow_pef_0_2 = TensorMap('3064_Peak-expiratory-flow-PEF_0_2', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_0_2': 0})
field_id_30660_direct_bilirubin_0_0 = TensorMap('30660_Direct-bilirubin_0_0', loss='logcosh', channel_map={'no_30660_Direct-bilirubin_0_0': 0})
field_id_30670_urea_0_0 = TensorMap('30670_Urea_0_0', loss='logcosh', channel_map={'no_30670_Urea_0_0': 0})
field_id_30680_calcium_0_0 = TensorMap('30680_Calcium_0_0', loss='logcosh', channel_map={'no_30680_Calcium_0_0': 0})
field_id_30690_cholesterol_0_0 = TensorMap('30690_Cholesterol_0_0', loss='logcosh', channel_map={'no_30690_Cholesterol_0_0': 0})
field_id_30710_creactive_protein_0_0 = TensorMap('30710_Creactive-protein_0_0', loss='logcosh', channel_map={'no_30710_Creactive-protein_0_0': 0})
field_id_30740_glucose_0_0 = TensorMap('30740_Glucose_0_0', loss='logcosh', channel_map={'no_30740_Glucose_0_0': 0})
field_id_30760_hdl_cholesterol_0_0 = TensorMap('30760_HDL-cholesterol_0_0', loss='logcosh', channel_map={'no_30760_HDL-cholesterol_0_0': 0})
field_id_30770_igf1_0_0 = TensorMap('30770_IGF1_0_0', loss='logcosh', channel_map={'no_30770_IGF1_0_0': 0})
field_id_30790_lipoprotein_a_0_0 = TensorMap('30790_Lipoprotein-A_0_0', loss='logcosh', channel_map={'no_30790_Lipoprotein-A_0_0': 0})
field_id_30810_phosphate_0_0 = TensorMap('30810_Phosphate_0_0', loss='logcosh', channel_map={'no_30810_Phosphate_0_0': 0})
field_id_30840_total_bilirubin_0_0 = TensorMap('30840_Total-bilirubin_0_0', loss='logcosh', channel_map={'no_30840_Total-bilirubin_0_0': 0})
field_id_30850_testosterone_0_0 = TensorMap('30850_Testosterone_0_0', loss='logcosh', channel_map={'no_30850_Testosterone_0_0': 0})
field_id_30860_total_protein_0_0 = TensorMap('30860_Total-protein_0_0', loss='logcosh', channel_map={'no_30860_Total-protein_0_0': 0})
field_id_30870_triglycerides_0_0 = TensorMap('30870_Triglycerides_0_0', loss='logcosh', channel_map={'no_30870_Triglycerides_0_0': 0})
field_id_30880_urate_0_0 = TensorMap('30880_Urate_0_0', loss='logcosh', channel_map={'no_30880_Urate_0_0': 0})
field_id_30890_vitamin_d_0_0 = TensorMap('30890_Vitamin-D_0_0', loss='logcosh', channel_map={'no_30890_Vitamin-D_0_0': 0})
field_id_3143_ankle_spacing_width_0_0 = TensorMap('3143_Ankle-spacing-width_0_0', loss='logcosh', channel_map={'no_3143_Ankle-spacing-width_0_0': 0})
field_id_3144_heel_broadband_ultrasound_attenuation_direct_entry_0_0 = TensorMap('3144_Heel-Broadband-ultrasound-attenuation-direct-entry_0_0', loss='logcosh', channel_map={'no_3144_Heel-Broadband-ultrasound-attenuation-direct-entry_0_0': 0})
field_id_3146_speed_of_sound_through_heel_0_0 = TensorMap('3146_Speed-of-sound-through-heel_0_0', loss='logcosh', channel_map={'no_3146_Speed-of-sound-through-heel_0_0': 0})
field_id_3147_heel_quantitative_ultrasound_index_qui_direct_entry_0_0 = TensorMap('3147_Heel-quantitative-ultrasound-index-QUI-direct-entry_0_0', loss='logcosh', channel_map={'no_3147_Heel-quantitative-ultrasound-index-QUI-direct-entry_0_0': 0})
field_id_3148_heel_bone_mineral_density_bmd_0_0 = TensorMap('3148_Heel-bone-mineral-density-BMD_0_0', loss='logcosh', channel_map={'no_3148_Heel-bone-mineral-density-BMD_0_0': 0})
field_id_34_year_of_birth_0_0 = TensorMap('34_Year-of-birth_0_0', loss='logcosh', channel_map={'no_34_Year-of-birth_0_0': 0})
field_id_396_number_of_columns_displayed_in_round_0_1 = TensorMap('396_Number-of-columns-displayed-in-round_0_1', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_0_1': 0})
field_id_396_number_of_columns_displayed_in_round_0_2 = TensorMap('396_Number-of-columns-displayed-in-round_0_2', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_0_2': 0})
field_id_397_number_of_rows_displayed_in_round_0_1 = TensorMap('397_Number-of-rows-displayed-in-round_0_1', loss='logcosh', channel_map={'no_397_Number-of-rows-displayed-in-round_0_1': 0})
field_id_397_number_of_rows_displayed_in_round_0_2 = TensorMap('397_Number-of-rows-displayed-in-round_0_2', loss='logcosh', channel_map={'no_397_Number-of-rows-displayed-in-round_0_2': 0})
field_id_398_number_of_correct_matches_in_round_0_1 = TensorMap('398_Number-of-correct-matches-in-round_0_1', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_0_1': 0})
field_id_398_number_of_correct_matches_in_round_0_2 = TensorMap('398_Number-of-correct-matches-in-round_0_2', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_0_2': 0})
field_id_399_number_of_incorrect_matches_in_round_0_1 = TensorMap('399_Number-of-incorrect-matches-in-round_0_1', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_0_1': 0})
field_id_399_number_of_incorrect_matches_in_round_0_2 = TensorMap('399_Number-of-incorrect-matches-in-round_0_2', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_0_2': 0})
field_id_400_time_to_complete_round_0_1 = TensorMap('400_Time-to-complete-round_0_1', loss='logcosh', channel_map={'no_400_Time-to-complete-round_0_1': 0})
field_id_400_time_to_complete_round_0_2 = TensorMap('400_Time-to-complete-round_0_2', loss='logcosh', channel_map={'no_400_Time-to-complete-round_0_2': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_1 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_1', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_1': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_10 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_10', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_10': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_11 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_11', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_11': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_4 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_4', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_4': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_5 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_5', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_5': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_7 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_7', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_7': 0})
field_id_4079_diastolic_blood_pressure_automated_reading_0_0 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_0_0', loss='logcosh', channel_map={'no_4079_Diastolic-blood-pressure-automated-reading_0_0': 0})
field_id_4079_diastolic_blood_pressure_automated_reading_0_1 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_0_1', loss='logcosh', channel_map={'no_4079_Diastolic-blood-pressure-automated-reading_0_1': 0})
field_id_4080_systolic_blood_pressure_automated_reading_0_0 = TensorMap('4080_Systolic-blood-pressure-automated-reading_0_0', loss='logcosh', channel_map={'no_4080_Systolic-blood-pressure-automated-reading_0_0': 0})
field_id_4080_systolic_blood_pressure_automated_reading_0_1 = TensorMap('4080_Systolic-blood-pressure-automated-reading_0_1', loss='logcosh', channel_map={'no_4080_Systolic-blood-pressure-automated-reading_0_1': 0})
field_id_46_hand_grip_strength_left_0_0 = TensorMap('46_Hand-grip-strength-left_0_0', loss='logcosh', channel_map={'no_46_Hand-grip-strength-left_0_0': 0})
field_id_47_hand_grip_strength_right_0_0 = TensorMap('47_Hand-grip-strength-right_0_0', loss='logcosh', channel_map={'no_47_Hand-grip-strength-right_0_0': 0})
field_id_48_waist_circumference_0_0 = TensorMap('48_Waist-circumference_0_0', loss='logcosh', channel_map={'no_48_Waist-circumference_0_0': 0})
field_id_49_hip_circumference_0_0 = TensorMap('49_Hip-circumference_0_0', loss='logcosh', channel_map={'no_49_Hip-circumference_0_0': 0})
field_id_50_standing_height_0_0 = TensorMap('50_Standing-height_0_0', loss='logcosh', channel_map={'no_50_Standing-height_0_0': 0})
field_id_51_seated_height_0_0 = TensorMap('51_Seated-height_0_0', loss='logcosh', channel_map={'no_51_Seated-height_0_0': 0})
field_id_699_length_of_time_at_current_address_0_0 = TensorMap('699_Length-of-time-at-current-address_0_0', loss='logcosh', channel_map={'no_699_Length-of-time-at-current-address_0_0': 0})
field_id_709_number_in_household_0_0 = TensorMap('709_Number-in-household_0_0', loss='logcosh', channel_map={'no_709_Number-in-household_0_0': 0})
field_id_845_age_completed_full_time_education_0_0 = TensorMap('845_Age-completed-full-time-education_0_0', loss='logcosh', channel_map={'no_845_Age-completed-full-time-education_0_0': 0})
field_id_864_number_of_daysweek_walked_10_minutes_0_0 = TensorMap('864_Number-of-daysweek-walked-10-minutes_0_0', loss='logcosh', channel_map={'no_864_Number-of-daysweek-walked-10-minutes_0_0': 0})
field_id_874_duration_of_walks_0_0 = TensorMap('874_Duration-of-walks_0_0', loss='logcosh', channel_map={'no_874_Duration-of-walks_0_0': 0})
field_id_884_number_of_daysweek_of_moderate_physical_activity_10_minutes_0_0 = TensorMap('884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_0_0', loss='logcosh', channel_map={'no_884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_0_0': 0})
field_id_894_duration_of_moderate_activity_0_0 = TensorMap('894_Duration-of-moderate-activity_0_0', loss='logcosh', channel_map={'no_894_Duration-of-moderate-activity_0_0': 0})
field_id_904_number_of_daysweek_of_vigorous_physical_activity_10_minutes_0_0 = TensorMap('904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_0_0', loss='logcosh', channel_map={'no_904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_0_0': 0})
field_id_914_duration_of_vigorous_activity_0_0 = TensorMap('914_Duration-of-vigorous-activity_0_0', loss='logcosh', channel_map={'no_914_Duration-of-vigorous-activity_0_0': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_0_0 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_0', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_0_0': 0})
field_id_20022_birth_weight_0_0 = TensorMap('20022_Birth-weight_0_0', loss='logcosh', channel_map={'no_20022_Birth-weight_0_0': 0})
field_id_20150_forced_expiratory_volume_in_1second_fev1_best_measure_0_0 = TensorMap('20150_Forced-expiratory-volume-in-1second-FEV1-Best-measure_0_0', loss='logcosh', channel_map={'no_20150_Forced-expiratory-volume-in-1second-FEV1-Best-measure_0_0': 0})
field_id_20153_forced_expiratory_volume_in_1second_fev1_predicted_0_0 = TensorMap('20153_Forced-expiratory-volume-in-1second-FEV1-predicted_0_0', loss='logcosh', channel_map={'no_20153_Forced-expiratory-volume-in-1second-FEV1-predicted_0_0': 0})
field_id_2217_age_started_wearing_glasses_or_contact_lenses_0_0 = TensorMap('2217_Age-started-wearing-glasses-or-contact-lenses_0_0', loss='logcosh', channel_map={'no_2217_Age-started-wearing-glasses-or-contact-lenses_0_0': 0})
field_id_22602_year_job_started_0_2 = TensorMap('22602_Year-job-started_0_2', loss='logcosh', channel_map={'no_22602_Year-job-started_0_2': 0})
field_id_22602_year_job_started_0_3 = TensorMap('22602_Year-job-started_0_3', loss='logcosh', channel_map={'no_22602_Year-job-started_0_3': 0})
field_id_22603_year_job_ended_0_2 = TensorMap('22603_Year-job-ended_0_2', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_2': 0})
field_id_22603_year_job_ended_0_3 = TensorMap('22603_Year-job-ended_0_3', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_3': 0})
field_id_22605_work_hours_per_week_exact_value_0_0 = TensorMap('22605_Work-hours-per-week-exact-value_0_0', loss='logcosh', channel_map={'no_22605_Work-hours-per-week-exact-value_0_0': 0})
field_id_22605_work_hours_per_week_exact_value_0_2 = TensorMap('22605_Work-hours-per-week-exact-value_0_2', loss='logcosh', channel_map={'no_22605_Work-hours-per-week-exact-value_0_2': 0})
field_id_22605_work_hours_per_week_exact_value_0_3 = TensorMap('22605_Work-hours-per-week-exact-value_0_3', loss='logcosh', channel_map={'no_22605_Work-hours-per-week-exact-value_0_3': 0})
field_id_22663_year_gap_started_0_0 = TensorMap('22663_Year-gap-started_0_0', loss='logcosh', channel_map={'no_22663_Year-gap-started_0_0': 0})
field_id_22663_year_gap_started_0_1 = TensorMap('22663_Year-gap-started_0_1', loss='logcosh', channel_map={'no_22663_Year-gap-started_0_1': 0})
field_id_22664_year_gap_ended_0_0 = TensorMap('22664_Year-gap-ended_0_0', loss='logcosh', channel_map={'no_22664_Year-gap-ended_0_0': 0})
field_id_22664_year_gap_ended_0_1 = TensorMap('22664_Year-gap-ended_0_1', loss='logcosh', channel_map={'no_22664_Year-gap-ended_0_1': 0})
field_id_2946_fathers_age_0_0 = TensorMap('2946_Fathers-age_0_0', loss='logcosh', channel_map={'no_2946_Fathers-age_0_0': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_2 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_2', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_2': 0})
field_id_92_operation_yearage_first_occurred_0_0 = TensorMap('92_Operation-yearage-first-occurred_0_0', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_0_0': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_0_1 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_1', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_0_1': 0})
field_id_20016_fluid_intelligence_score_0_0 = TensorMap('20016_Fluid-intelligence-score_0_0', loss='logcosh', channel_map={'no_20016_Fluid-intelligence-score_0_0': 0})
field_id_20019_speechreceptionthreshold_srt_estimate_left_0_0 = TensorMap('20019_Speechreceptionthreshold-SRT-estimate-left_0_0', loss='logcosh', channel_map={'no_20019_Speechreceptionthreshold-SRT-estimate-left_0_0': 0})
field_id_20021_speechreceptionthreshold_srt_estimate_right_0_0 = TensorMap('20021_Speechreceptionthreshold-SRT-estimate-right_0_0', loss='logcosh', channel_map={'no_20021_Speechreceptionthreshold-SRT-estimate-right_0_0': 0})
field_id_20128_number_of_fluid_intelligence_questions_attempted_within_time_limit_0_0 = TensorMap('20128_Number-of-fluid-intelligence-questions-attempted-within-time-limit_0_0', loss='logcosh', channel_map={'no_20128_Number-of-fluid-intelligence-questions-attempted-within-time-limit_0_0': 0})
field_id_20161_pack_years_of_smoking_0_0 = TensorMap('20161_Pack-years-of-smoking_0_0', loss='logcosh', channel_map={'no_20161_Pack-years-of-smoking_0_0': 0})
field_id_20162_pack_years_adult_smoking_as_proportion_of_life_span_exposed_to_smoking_0_0 = TensorMap('20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_0_0', loss='logcosh', channel_map={'no_20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_0_0': 0})
field_id_21021_pulse_wave_arterial_stiffness_index_0_0 = TensorMap('21021_Pulse-wave-Arterial-Stiffness-index_0_0', loss='logcosh', channel_map={'no_21021_Pulse-wave-Arterial-Stiffness-index_0_0': 0})
field_id_2867_age_started_smoking_in_former_smokers_0_0 = TensorMap('2867_Age-started-smoking-in-former-smokers_0_0', loss='logcosh', channel_map={'no_2867_Age-started-smoking-in-former-smokers_0_0': 0})
field_id_2887_number_of_cigarettes_previously_smoked_daily_0_0 = TensorMap('2887_Number-of-cigarettes-previously-smoked-daily_0_0', loss='logcosh', channel_map={'no_2887_Number-of-cigarettes-previously-smoked-daily_0_0': 0})
field_id_2897_age_stopped_smoking_0_0 = TensorMap('2897_Age-stopped-smoking_0_0', loss='logcosh', channel_map={'no_2897_Age-stopped-smoking_0_0': 0})
field_id_2926_number_of_unsuccessful_stopsmoking_attempts_0_0 = TensorMap('2926_Number-of-unsuccessful-stopsmoking-attempts_0_0', loss='logcosh', channel_map={'no_2926_Number-of-unsuccessful-stopsmoking-attempts_0_0': 0})
field_id_30414_volume_of_rna_held_by_ukb_0_0 = TensorMap('30414_Volume-of-RNA-held-by-UKB_0_0', loss='logcosh', channel_map={'no_30414_Volume-of-RNA-held-by-UKB_0_0': 0})
field_id_30500_microalbumin_in_urine_0_0 = TensorMap('30500_Microalbumin-in-urine_0_0', loss='logcosh', channel_map={'no_30500_Microalbumin-in-urine_0_0': 0})
field_id_3809_time_since_last_prostate_specific_antigen_psa_test_0_0 = TensorMap('3809_Time-since-last-prostate-specific-antigen-PSA-test_0_0', loss='logcosh', channel_map={'no_3809_Time-since-last-prostate-specific-antigen-PSA-test_0_0': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_3 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_3', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_3': 0})
field_id_4100_ankle_spacing_width_left_0_0 = TensorMap('4100_Ankle-spacing-width-left_0_0', loss='logcosh', channel_map={'no_4100_Ankle-spacing-width-left_0_0': 0})
field_id_4101_heel_broadband_ultrasound_attenuation_left_0_0 = TensorMap('4101_Heel-broadband-ultrasound-attenuation-left_0_0', loss='logcosh', channel_map={'no_4101_Heel-broadband-ultrasound-attenuation-left_0_0': 0})
field_id_4103_speed_of_sound_through_heel_left_0_0 = TensorMap('4103_Speed-of-sound-through-heel-left_0_0', loss='logcosh', channel_map={'no_4103_Speed-of-sound-through-heel-left_0_0': 0})
field_id_4104_heel_quantitative_ultrasound_index_qui_direct_entry_left_0_0 = TensorMap('4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_0_0', loss='logcosh', channel_map={'no_4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_0_0': 0})
field_id_4105_heel_bone_mineral_density_bmd_left_0_0 = TensorMap('4105_Heel-bone-mineral-density-BMD-left_0_0', loss='logcosh', channel_map={'no_4105_Heel-bone-mineral-density-BMD-left_0_0': 0})
field_id_4106_heel_bone_mineral_density_bmd_tscore_automated_left_0_0 = TensorMap('4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_0_0', loss='logcosh', channel_map={'no_4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_0_0': 0})
field_id_4119_ankle_spacing_width_right_0_0 = TensorMap('4119_Ankle-spacing-width-right_0_0', loss='logcosh', channel_map={'no_4119_Ankle-spacing-width-right_0_0': 0})
field_id_4120_heel_broadband_ultrasound_attenuation_right_0_0 = TensorMap('4120_Heel-broadband-ultrasound-attenuation-right_0_0', loss='logcosh', channel_map={'no_4120_Heel-broadband-ultrasound-attenuation-right_0_0': 0})
field_id_4122_speed_of_sound_through_heel_right_0_0 = TensorMap('4122_Speed-of-sound-through-heel-right_0_0', loss='logcosh', channel_map={'no_4122_Speed-of-sound-through-heel-right_0_0': 0})
field_id_4123_heel_quantitative_ultrasound_index_qui_direct_entry_right_0_0 = TensorMap('4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_0_0', loss='logcosh', channel_map={'no_4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_0_0': 0})
field_id_4124_heel_bone_mineral_density_bmd_right_0_0 = TensorMap('4124_Heel-bone-mineral-density-BMD-right_0_0', loss='logcosh', channel_map={'no_4124_Heel-bone-mineral-density-BMD-right_0_0': 0})
field_id_4125_heel_bone_mineral_density_bmd_tscore_automated_right_0_0 = TensorMap('4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_0_0', loss='logcosh', channel_map={'no_4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_0_0': 0})
field_id_4194_pulse_rate_0_0 = TensorMap('4194_Pulse-rate_0_0', loss='logcosh', channel_map={'no_4194_Pulse-rate_0_0': 0})
field_id_4195_pulse_wave_reflection_index_0_0 = TensorMap('4195_Pulse-wave-reflection-index_0_0', loss='logcosh', channel_map={'no_4195_Pulse-wave-reflection-index_0_0': 0})
field_id_4196_pulse_wave_peak_to_peak_time_0_0 = TensorMap('4196_Pulse-wave-peak-to-peak-time_0_0', loss='logcosh', channel_map={'no_4196_Pulse-wave-peak-to-peak-time_0_0': 0})
field_id_4198_position_of_the_pulse_wave_peak_0_0 = TensorMap('4198_Position-of-the-pulse-wave-peak_0_0', loss='logcosh', channel_map={'no_4198_Position-of-the-pulse-wave-peak_0_0': 0})
field_id_4199_position_of_pulse_wave_notch_0_0 = TensorMap('4199_Position-of-pulse-wave-notch_0_0', loss='logcosh', channel_map={'no_4199_Position-of-pulse-wave-notch_0_0': 0})
field_id_4200_position_of_the_shoulder_on_the_pulse_waveform_0_0 = TensorMap('4200_Position-of-the-shoulder-on-the-pulse-waveform_0_0', loss='logcosh', channel_map={'no_4200_Position-of-the-shoulder-on-the-pulse-waveform_0_0': 0})
field_id_4269_number_of_triplets_attempted_left_0_0 = TensorMap('4269_Number-of-triplets-attempted-left_0_0', loss='logcosh', channel_map={'no_4269_Number-of-triplets-attempted-left_0_0': 0})
field_id_4272_duration_of_hearing_test_left_0_0 = TensorMap('4272_Duration-of-hearing-test-left_0_0', loss='logcosh', channel_map={'no_4272_Duration-of-hearing-test-left_0_0': 0})
field_id_4276_number_of_triplets_attempted_right_0_0 = TensorMap('4276_Number-of-triplets-attempted-right_0_0', loss='logcosh', channel_map={'no_4276_Number-of-triplets-attempted-right_0_0': 0})
field_id_4279_duration_of_hearing_test_right_0_0 = TensorMap('4279_Duration-of-hearing-test-right_0_0', loss='logcosh', channel_map={'no_4279_Duration-of-hearing-test-right_0_0': 0})
field_id_4288_time_to_answer_0_0 = TensorMap('4288_Time-to-answer_0_0', loss='logcosh', channel_map={'no_4288_Time-to-answer_0_0': 0})
field_id_4290_duration_screen_displayed_0_0 = TensorMap('4290_Duration-screen-displayed_0_0', loss='logcosh', channel_map={'no_4290_Duration-screen-displayed_0_0': 0})
field_id_4291_number_of_attempts_0_0 = TensorMap('4291_Number-of-attempts_0_0', loss='logcosh', channel_map={'no_4291_Number-of-attempts_0_0': 0})
field_id_5057_number_of_older_siblings_0_0 = TensorMap('5057_Number-of-older-siblings_0_0', loss='logcosh', channel_map={'no_5057_Number-of-older-siblings_0_0': 0})
field_id_5074_number_of_letters_shown_in_round_left_0_0 = TensorMap('5074_Number-of-letters-shown-in-round-left_0_0', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_0_0': 0})
field_id_5074_number_of_letters_shown_in_round_left_0_1 = TensorMap('5074_Number-of-letters-shown-in-round-left_0_1', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_0_1': 0})
field_id_5075_number_of_letters_shown_in_round_right_0_0 = TensorMap('5075_Number-of-letters-shown-in-round-right_0_0', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_0_0': 0})
field_id_5076_number_of_letters_correct_in_round_right_0_0 = TensorMap('5076_Number-of-letters-correct-in-round-right_0_0', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_0_0': 0})
field_id_5077_number_of_letters_correct_in_round_left_0_0 = TensorMap('5077_Number-of-letters-correct-in-round-left_0_0', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_0_0': 0})
field_id_5077_number_of_letters_correct_in_round_left_0_1 = TensorMap('5077_Number-of-letters-correct-in-round-left_0_1', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_0_1': 0})
field_id_5078_logmar_in_round_left_0_0 = TensorMap('5078_logMAR-in-round-left_0_0', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_0_0': 0})
field_id_5078_logmar_in_round_left_0_1 = TensorMap('5078_logMAR-in-round-left_0_1', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_0_1': 0})
field_id_5079_logmar_in_round_right_0_0 = TensorMap('5079_logMAR-in-round-right_0_0', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_0_0': 0})
field_id_5084_spherical_power_right_0_0 = TensorMap('5084_Spherical-power-right_0_0', loss='logcosh', channel_map={'no_5084_Spherical-power-right_0_0': 0})
field_id_5084_spherical_power_right_0_1 = TensorMap('5084_Spherical-power-right_0_1', loss='logcosh', channel_map={'no_5084_Spherical-power-right_0_1': 0})
field_id_5084_spherical_power_right_0_2 = TensorMap('5084_Spherical-power-right_0_2', loss='logcosh', channel_map={'no_5084_Spherical-power-right_0_2': 0})
field_id_5085_spherical_power_left_0_0 = TensorMap('5085_Spherical-power-left_0_0', loss='logcosh', channel_map={'no_5085_Spherical-power-left_0_0': 0})
field_id_5085_spherical_power_left_0_1 = TensorMap('5085_Spherical-power-left_0_1', loss='logcosh', channel_map={'no_5085_Spherical-power-left_0_1': 0})
field_id_5085_spherical_power_left_0_2 = TensorMap('5085_Spherical-power-left_0_2', loss='logcosh', channel_map={'no_5085_Spherical-power-left_0_2': 0})
field_id_5086_cylindrical_power_left_0_0 = TensorMap('5086_Cylindrical-power-left_0_0', loss='logcosh', channel_map={'no_5086_Cylindrical-power-left_0_0': 0})
field_id_5086_cylindrical_power_left_0_1 = TensorMap('5086_Cylindrical-power-left_0_1', loss='logcosh', channel_map={'no_5086_Cylindrical-power-left_0_1': 0})
field_id_5086_cylindrical_power_left_0_2 = TensorMap('5086_Cylindrical-power-left_0_2', loss='logcosh', channel_map={'no_5086_Cylindrical-power-left_0_2': 0})
field_id_5087_cylindrical_power_right_0_0 = TensorMap('5087_Cylindrical-power-right_0_0', loss='logcosh', channel_map={'no_5087_Cylindrical-power-right_0_0': 0})
field_id_5087_cylindrical_power_right_0_1 = TensorMap('5087_Cylindrical-power-right_0_1', loss='logcosh', channel_map={'no_5087_Cylindrical-power-right_0_1': 0})
field_id_5087_cylindrical_power_right_0_2 = TensorMap('5087_Cylindrical-power-right_0_2', loss='logcosh', channel_map={'no_5087_Cylindrical-power-right_0_2': 0})
field_id_5088_astigmatism_angle_right_0_0 = TensorMap('5088_Astigmatism-angle-right_0_0', loss='logcosh', channel_map={'no_5088_Astigmatism-angle-right_0_0': 0})
field_id_5088_astigmatism_angle_right_0_1 = TensorMap('5088_Astigmatism-angle-right_0_1', loss='logcosh', channel_map={'no_5088_Astigmatism-angle-right_0_1': 0})
field_id_5088_astigmatism_angle_right_0_2 = TensorMap('5088_Astigmatism-angle-right_0_2', loss='logcosh', channel_map={'no_5088_Astigmatism-angle-right_0_2': 0})
field_id_5089_astigmatism_angle_left_0_0 = TensorMap('5089_Astigmatism-angle-left_0_0', loss='logcosh', channel_map={'no_5089_Astigmatism-angle-left_0_0': 0})
field_id_5089_astigmatism_angle_left_0_1 = TensorMap('5089_Astigmatism-angle-left_0_1', loss='logcosh', channel_map={'no_5089_Astigmatism-angle-left_0_1': 0})
field_id_5089_astigmatism_angle_left_0_2 = TensorMap('5089_Astigmatism-angle-left_0_2', loss='logcosh', channel_map={'no_5089_Astigmatism-angle-left_0_2': 0})
field_id_5096_3mm_weak_meridian_left_0_0 = TensorMap('5096_3mm-weak-meridian-left_0_0', loss='logcosh', channel_map={'no_5096_3mm-weak-meridian-left_0_0': 0})
field_id_5096_3mm_weak_meridian_left_0_1 = TensorMap('5096_3mm-weak-meridian-left_0_1', loss='logcosh', channel_map={'no_5096_3mm-weak-meridian-left_0_1': 0})
field_id_5096_3mm_weak_meridian_left_0_2 = TensorMap('5096_3mm-weak-meridian-left_0_2', loss='logcosh', channel_map={'no_5096_3mm-weak-meridian-left_0_2': 0})
field_id_5097_6mm_weak_meridian_left_0_0 = TensorMap('5097_6mm-weak-meridian-left_0_0', loss='logcosh', channel_map={'no_5097_6mm-weak-meridian-left_0_0': 0})
field_id_5097_6mm_weak_meridian_left_0_1 = TensorMap('5097_6mm-weak-meridian-left_0_1', loss='logcosh', channel_map={'no_5097_6mm-weak-meridian-left_0_1': 0})
field_id_5097_6mm_weak_meridian_left_0_2 = TensorMap('5097_6mm-weak-meridian-left_0_2', loss='logcosh', channel_map={'no_5097_6mm-weak-meridian-left_0_2': 0})
field_id_5098_6mm_weak_meridian_right_0_0 = TensorMap('5098_6mm-weak-meridian-right_0_0', loss='logcosh', channel_map={'no_5098_6mm-weak-meridian-right_0_0': 0})
field_id_5098_6mm_weak_meridian_right_0_1 = TensorMap('5098_6mm-weak-meridian-right_0_1', loss='logcosh', channel_map={'no_5098_6mm-weak-meridian-right_0_1': 0})
field_id_5099_3mm_weak_meridian_right_0_0 = TensorMap('5099_3mm-weak-meridian-right_0_0', loss='logcosh', channel_map={'no_5099_3mm-weak-meridian-right_0_0': 0})
field_id_5099_3mm_weak_meridian_right_0_1 = TensorMap('5099_3mm-weak-meridian-right_0_1', loss='logcosh', channel_map={'no_5099_3mm-weak-meridian-right_0_1': 0})
field_id_5100_3mm_weak_meridian_angle_right_0_0 = TensorMap('5100_3mm-weak-meridian-angle-right_0_0', loss='logcosh', channel_map={'no_5100_3mm-weak-meridian-angle-right_0_0': 0})
field_id_5100_3mm_weak_meridian_angle_right_0_1 = TensorMap('5100_3mm-weak-meridian-angle-right_0_1', loss='logcosh', channel_map={'no_5100_3mm-weak-meridian-angle-right_0_1': 0})
field_id_5101_6mm_weak_meridian_angle_right_0_0 = TensorMap('5101_6mm-weak-meridian-angle-right_0_0', loss='logcosh', channel_map={'no_5101_6mm-weak-meridian-angle-right_0_0': 0})
field_id_5101_6mm_weak_meridian_angle_right_0_1 = TensorMap('5101_6mm-weak-meridian-angle-right_0_1', loss='logcosh', channel_map={'no_5101_6mm-weak-meridian-angle-right_0_1': 0})
field_id_5102_6mm_weak_meridian_angle_left_0_0 = TensorMap('5102_6mm-weak-meridian-angle-left_0_0', loss='logcosh', channel_map={'no_5102_6mm-weak-meridian-angle-left_0_0': 0})
field_id_5102_6mm_weak_meridian_angle_left_0_1 = TensorMap('5102_6mm-weak-meridian-angle-left_0_1', loss='logcosh', channel_map={'no_5102_6mm-weak-meridian-angle-left_0_1': 0})
field_id_5102_6mm_weak_meridian_angle_left_0_2 = TensorMap('5102_6mm-weak-meridian-angle-left_0_2', loss='logcosh', channel_map={'no_5102_6mm-weak-meridian-angle-left_0_2': 0})
field_id_5103_3mm_weak_meridian_angle_left_0_0 = TensorMap('5103_3mm-weak-meridian-angle-left_0_0', loss='logcosh', channel_map={'no_5103_3mm-weak-meridian-angle-left_0_0': 0})
field_id_5103_3mm_weak_meridian_angle_left_0_1 = TensorMap('5103_3mm-weak-meridian-angle-left_0_1', loss='logcosh', channel_map={'no_5103_3mm-weak-meridian-angle-left_0_1': 0})
field_id_5103_3mm_weak_meridian_angle_left_0_2 = TensorMap('5103_3mm-weak-meridian-angle-left_0_2', loss='logcosh', channel_map={'no_5103_3mm-weak-meridian-angle-left_0_2': 0})
field_id_5104_3mm_strong_meridian_angle_left_0_0 = TensorMap('5104_3mm-strong-meridian-angle-left_0_0', loss='logcosh', channel_map={'no_5104_3mm-strong-meridian-angle-left_0_0': 0})
field_id_5104_3mm_strong_meridian_angle_left_0_1 = TensorMap('5104_3mm-strong-meridian-angle-left_0_1', loss='logcosh', channel_map={'no_5104_3mm-strong-meridian-angle-left_0_1': 0})
field_id_5104_3mm_strong_meridian_angle_left_0_2 = TensorMap('5104_3mm-strong-meridian-angle-left_0_2', loss='logcosh', channel_map={'no_5104_3mm-strong-meridian-angle-left_0_2': 0})
field_id_5105_6mm_strong_meridian_angle_left_0_0 = TensorMap('5105_6mm-strong-meridian-angle-left_0_0', loss='logcosh', channel_map={'no_5105_6mm-strong-meridian-angle-left_0_0': 0})
field_id_5105_6mm_strong_meridian_angle_left_0_1 = TensorMap('5105_6mm-strong-meridian-angle-left_0_1', loss='logcosh', channel_map={'no_5105_6mm-strong-meridian-angle-left_0_1': 0})
field_id_5105_6mm_strong_meridian_angle_left_0_2 = TensorMap('5105_6mm-strong-meridian-angle-left_0_2', loss='logcosh', channel_map={'no_5105_6mm-strong-meridian-angle-left_0_2': 0})
field_id_5106_6mm_strong_meridian_angle_right_0_0 = TensorMap('5106_6mm-strong-meridian-angle-right_0_0', loss='logcosh', channel_map={'no_5106_6mm-strong-meridian-angle-right_0_0': 0})
field_id_5106_6mm_strong_meridian_angle_right_0_1 = TensorMap('5106_6mm-strong-meridian-angle-right_0_1', loss='logcosh', channel_map={'no_5106_6mm-strong-meridian-angle-right_0_1': 0})
field_id_5107_3mm_strong_meridian_angle_right_0_0 = TensorMap('5107_3mm-strong-meridian-angle-right_0_0', loss='logcosh', channel_map={'no_5107_3mm-strong-meridian-angle-right_0_0': 0})
field_id_5107_3mm_strong_meridian_angle_right_0_1 = TensorMap('5107_3mm-strong-meridian-angle-right_0_1', loss='logcosh', channel_map={'no_5107_3mm-strong-meridian-angle-right_0_1': 0})
field_id_5108_3mm_asymmetry_angle_right_0_0 = TensorMap('5108_3mm-asymmetry-angle-right_0_0', loss='logcosh', channel_map={'no_5108_3mm-asymmetry-angle-right_0_0': 0})
field_id_5108_3mm_asymmetry_angle_right_0_1 = TensorMap('5108_3mm-asymmetry-angle-right_0_1', loss='logcosh', channel_map={'no_5108_3mm-asymmetry-angle-right_0_1': 0})
field_id_5109_6mm_asymmetry_angle_right_0_0 = TensorMap('5109_6mm-asymmetry-angle-right_0_0', loss='logcosh', channel_map={'no_5109_6mm-asymmetry-angle-right_0_0': 0})
field_id_5109_6mm_asymmetry_angle_right_0_1 = TensorMap('5109_6mm-asymmetry-angle-right_0_1', loss='logcosh', channel_map={'no_5109_6mm-asymmetry-angle-right_0_1': 0})
field_id_5110_6mm_asymmetry_angle_left_0_2 = TensorMap('5110_6mm-asymmetry-angle-left_0_2', loss='logcosh', channel_map={'no_5110_6mm-asymmetry-angle-left_0_2': 0})
field_id_5111_3mm_asymmetry_angle_left_0_0 = TensorMap('5111_3mm-asymmetry-angle-left_0_0', loss='logcosh', channel_map={'no_5111_3mm-asymmetry-angle-left_0_0': 0})
field_id_5111_3mm_asymmetry_angle_left_0_1 = TensorMap('5111_3mm-asymmetry-angle-left_0_1', loss='logcosh', channel_map={'no_5111_3mm-asymmetry-angle-left_0_1': 0})
field_id_5111_3mm_asymmetry_angle_left_0_2 = TensorMap('5111_3mm-asymmetry-angle-left_0_2', loss='logcosh', channel_map={'no_5111_3mm-asymmetry-angle-left_0_2': 0})
field_id_5112_3mm_cylindrical_power_angle_left_0_0 = TensorMap('5112_3mm-cylindrical-power-angle-left_0_0', loss='logcosh', channel_map={'no_5112_3mm-cylindrical-power-angle-left_0_0': 0})
field_id_5112_3mm_cylindrical_power_angle_left_0_1 = TensorMap('5112_3mm-cylindrical-power-angle-left_0_1', loss='logcosh', channel_map={'no_5112_3mm-cylindrical-power-angle-left_0_1': 0})
field_id_5112_3mm_cylindrical_power_angle_left_0_2 = TensorMap('5112_3mm-cylindrical-power-angle-left_0_2', loss='logcosh', channel_map={'no_5112_3mm-cylindrical-power-angle-left_0_2': 0})
field_id_5113_6mm_cylindrical_power_angle_left_0_0 = TensorMap('5113_6mm-cylindrical-power-angle-left_0_0', loss='logcosh', channel_map={'no_5113_6mm-cylindrical-power-angle-left_0_0': 0})
field_id_5113_6mm_cylindrical_power_angle_left_0_1 = TensorMap('5113_6mm-cylindrical-power-angle-left_0_1', loss='logcosh', channel_map={'no_5113_6mm-cylindrical-power-angle-left_0_1': 0})
field_id_5113_6mm_cylindrical_power_angle_left_0_2 = TensorMap('5113_6mm-cylindrical-power-angle-left_0_2', loss='logcosh', channel_map={'no_5113_6mm-cylindrical-power-angle-left_0_2': 0})
field_id_5114_6mm_cylindrical_power_angle_right_0_0 = TensorMap('5114_6mm-cylindrical-power-angle-right_0_0', loss='logcosh', channel_map={'no_5114_6mm-cylindrical-power-angle-right_0_0': 0})
field_id_5114_6mm_cylindrical_power_angle_right_0_1 = TensorMap('5114_6mm-cylindrical-power-angle-right_0_1', loss='logcosh', channel_map={'no_5114_6mm-cylindrical-power-angle-right_0_1': 0})
field_id_5115_3mm_cylindrical_power_angle_right_0_0 = TensorMap('5115_3mm-cylindrical-power-angle-right_0_0', loss='logcosh', channel_map={'no_5115_3mm-cylindrical-power-angle-right_0_0': 0})
field_id_5115_3mm_cylindrical_power_angle_right_0_1 = TensorMap('5115_3mm-cylindrical-power-angle-right_0_1', loss='logcosh', channel_map={'no_5115_3mm-cylindrical-power-angle-right_0_1': 0})
field_id_5116_3mm_cylindrical_power_right_0_0 = TensorMap('5116_3mm-cylindrical-power-right_0_0', loss='logcosh', channel_map={'no_5116_3mm-cylindrical-power-right_0_0': 0})
field_id_5116_3mm_cylindrical_power_right_0_1 = TensorMap('5116_3mm-cylindrical-power-right_0_1', loss='logcosh', channel_map={'no_5116_3mm-cylindrical-power-right_0_1': 0})
field_id_5117_6mm_cylindrical_power_right_0_0 = TensorMap('5117_6mm-cylindrical-power-right_0_0', loss='logcosh', channel_map={'no_5117_6mm-cylindrical-power-right_0_0': 0})
field_id_5117_6mm_cylindrical_power_right_0_1 = TensorMap('5117_6mm-cylindrical-power-right_0_1', loss='logcosh', channel_map={'no_5117_6mm-cylindrical-power-right_0_1': 0})
field_id_5118_6mm_cylindrical_power_left_0_0 = TensorMap('5118_6mm-cylindrical-power-left_0_0', loss='logcosh', channel_map={'no_5118_6mm-cylindrical-power-left_0_0': 0})
field_id_5118_6mm_cylindrical_power_left_0_1 = TensorMap('5118_6mm-cylindrical-power-left_0_1', loss='logcosh', channel_map={'no_5118_6mm-cylindrical-power-left_0_1': 0})
field_id_5118_6mm_cylindrical_power_left_0_2 = TensorMap('5118_6mm-cylindrical-power-left_0_2', loss='logcosh', channel_map={'no_5118_6mm-cylindrical-power-left_0_2': 0})
field_id_5119_3mm_cylindrical_power_left_0_0 = TensorMap('5119_3mm-cylindrical-power-left_0_0', loss='logcosh', channel_map={'no_5119_3mm-cylindrical-power-left_0_0': 0})
field_id_5119_3mm_cylindrical_power_left_0_1 = TensorMap('5119_3mm-cylindrical-power-left_0_1', loss='logcosh', channel_map={'no_5119_3mm-cylindrical-power-left_0_1': 0})
field_id_5119_3mm_cylindrical_power_left_0_2 = TensorMap('5119_3mm-cylindrical-power-left_0_2', loss='logcosh', channel_map={'no_5119_3mm-cylindrical-power-left_0_2': 0})
field_id_5132_3mm_strong_meridian_right_0_0 = TensorMap('5132_3mm-strong-meridian-right_0_0', loss='logcosh', channel_map={'no_5132_3mm-strong-meridian-right_0_0': 0})
field_id_5132_3mm_strong_meridian_right_0_1 = TensorMap('5132_3mm-strong-meridian-right_0_1', loss='logcosh', channel_map={'no_5132_3mm-strong-meridian-right_0_1': 0})
field_id_5133_6mm_strong_meridian_right_0_0 = TensorMap('5133_6mm-strong-meridian-right_0_0', loss='logcosh', channel_map={'no_5133_6mm-strong-meridian-right_0_0': 0})
field_id_5133_6mm_strong_meridian_right_0_1 = TensorMap('5133_6mm-strong-meridian-right_0_1', loss='logcosh', channel_map={'no_5133_6mm-strong-meridian-right_0_1': 0})
field_id_5134_6mm_strong_meridian_left_0_0 = TensorMap('5134_6mm-strong-meridian-left_0_0', loss='logcosh', channel_map={'no_5134_6mm-strong-meridian-left_0_0': 0})
field_id_5134_6mm_strong_meridian_left_0_1 = TensorMap('5134_6mm-strong-meridian-left_0_1', loss='logcosh', channel_map={'no_5134_6mm-strong-meridian-left_0_1': 0})
field_id_5134_6mm_strong_meridian_left_0_2 = TensorMap('5134_6mm-strong-meridian-left_0_2', loss='logcosh', channel_map={'no_5134_6mm-strong-meridian-left_0_2': 0})
field_id_5135_3mm_strong_meridian_left_0_0 = TensorMap('5135_3mm-strong-meridian-left_0_0', loss='logcosh', channel_map={'no_5135_3mm-strong-meridian-left_0_0': 0})
field_id_5135_3mm_strong_meridian_left_0_1 = TensorMap('5135_3mm-strong-meridian-left_0_1', loss='logcosh', channel_map={'no_5135_3mm-strong-meridian-left_0_1': 0})
field_id_5135_3mm_strong_meridian_left_0_2 = TensorMap('5135_3mm-strong-meridian-left_0_2', loss='logcosh', channel_map={'no_5135_3mm-strong-meridian-left_0_2': 0})
field_id_5156_3mm_asymmetry_index_left_0_0 = TensorMap('5156_3mm-asymmetry-index-left_0_0', loss='logcosh', channel_map={'no_5156_3mm-asymmetry-index-left_0_0': 0})
field_id_5156_3mm_asymmetry_index_left_0_1 = TensorMap('5156_3mm-asymmetry-index-left_0_1', loss='logcosh', channel_map={'no_5156_3mm-asymmetry-index-left_0_1': 0})
field_id_5156_3mm_asymmetry_index_left_0_2 = TensorMap('5156_3mm-asymmetry-index-left_0_2', loss='logcosh', channel_map={'no_5156_3mm-asymmetry-index-left_0_2': 0})
field_id_5157_6mm_asymmetry_index_left_0_2 = TensorMap('5157_6mm-asymmetry-index-left_0_2', loss='logcosh', channel_map={'no_5157_6mm-asymmetry-index-left_0_2': 0})
field_id_5158_6mm_asymmetry_index_right_0_0 = TensorMap('5158_6mm-asymmetry-index-right_0_0', loss='logcosh', channel_map={'no_5158_6mm-asymmetry-index-right_0_0': 0})
field_id_5158_6mm_asymmetry_index_right_0_1 = TensorMap('5158_6mm-asymmetry-index-right_0_1', loss='logcosh', channel_map={'no_5158_6mm-asymmetry-index-right_0_1': 0})
field_id_5159_3mm_asymmetry_index_right_0_0 = TensorMap('5159_3mm-asymmetry-index-right_0_0', loss='logcosh', channel_map={'no_5159_3mm-asymmetry-index-right_0_0': 0})
field_id_5159_3mm_asymmetry_index_right_0_1 = TensorMap('5159_3mm-asymmetry-index-right_0_1', loss='logcosh', channel_map={'no_5159_3mm-asymmetry-index-right_0_1': 0})
field_id_5160_3mm_regularity_index_right_0_0 = TensorMap('5160_3mm-regularity-index-right_0_0', loss='logcosh', channel_map={'no_5160_3mm-regularity-index-right_0_0': 0})
field_id_5160_3mm_regularity_index_right_0_1 = TensorMap('5160_3mm-regularity-index-right_0_1', loss='logcosh', channel_map={'no_5160_3mm-regularity-index-right_0_1': 0})
field_id_5161_6mm_regularity_index_right_0_0 = TensorMap('5161_6mm-regularity-index-right_0_0', loss='logcosh', channel_map={'no_5161_6mm-regularity-index-right_0_0': 0})
field_id_5161_6mm_regularity_index_right_0_1 = TensorMap('5161_6mm-regularity-index-right_0_1', loss='logcosh', channel_map={'no_5161_6mm-regularity-index-right_0_1': 0})
field_id_5162_6mm_regularity_index_left_0_2 = TensorMap('5162_6mm-regularity-index-left_0_2', loss='logcosh', channel_map={'no_5162_6mm-regularity-index-left_0_2': 0})
field_id_5163_3mm_regularity_index_left_0_0 = TensorMap('5163_3mm-regularity-index-left_0_0', loss='logcosh', channel_map={'no_5163_3mm-regularity-index-left_0_0': 0})
field_id_5163_3mm_regularity_index_left_0_1 = TensorMap('5163_3mm-regularity-index-left_0_1', loss='logcosh', channel_map={'no_5163_3mm-regularity-index-left_0_1': 0})
field_id_5163_3mm_regularity_index_left_0_2 = TensorMap('5163_3mm-regularity-index-left_0_2', loss='logcosh', channel_map={'no_5163_3mm-regularity-index-left_0_2': 0})
field_id_5186_duration_visualacuity_screen_displayed_right_0_0 = TensorMap('5186_Duration-visualacuity-screen-displayed-right_0_0', loss='logcosh', channel_map={'no_5186_Duration-visualacuity-screen-displayed-right_0_0': 0})
field_id_5188_duration_visualacuity_screen_displayed_left_0_0 = TensorMap('5188_Duration-visualacuity-screen-displayed-left_0_0', loss='logcosh', channel_map={'no_5188_Duration-visualacuity-screen-displayed-left_0_0': 0})
field_id_5190_duration_at_which_refractometer_first_shown_right_0_0 = TensorMap('5190_Duration-at-which-refractometer-first-shown-right_0_0', loss='logcosh', channel_map={'no_5190_Duration-at-which-refractometer-first-shown-right_0_0': 0})
field_id_5193_duration_at_which_refractometer_first_shown_left_0_0 = TensorMap('5193_Duration-at-which-refractometer-first-shown-left_0_0', loss='logcosh', channel_map={'no_5193_Duration-at-which-refractometer-first-shown-left_0_0': 0})
field_id_5199_logmar_initial_right_0_0 = TensorMap('5199_logMAR-initial-right_0_0', loss='logcosh', channel_map={'no_5199_logMAR-initial-right_0_0': 0})
field_id_5200_final_number_of_letters_displayed_right_0_0 = TensorMap('5200_Final-number-of-letters-displayed-right_0_0', loss='logcosh', channel_map={'no_5200_Final-number-of-letters-displayed-right_0_0': 0})
field_id_5201_logmar_final_right_0_0 = TensorMap('5201_logMAR-final-right_0_0', loss='logcosh', channel_map={'no_5201_logMAR-final-right_0_0': 0})
field_id_5202_number_of_rounds_to_result_right_0_0 = TensorMap('5202_Number-of-rounds-to-result-right_0_0', loss='logcosh', channel_map={'no_5202_Number-of-rounds-to-result-right_0_0': 0})
field_id_5204_distance_of_viewer_to_screen_right_0_0 = TensorMap('5204_Distance-of-viewer-to-screen-right_0_0', loss='logcosh', channel_map={'no_5204_Distance-of-viewer-to-screen-right_0_0': 0})
field_id_5206_logmar_initial_left_0_0 = TensorMap('5206_logMAR-initial-left_0_0', loss='logcosh', channel_map={'no_5206_logMAR-initial-left_0_0': 0})
field_id_5207_final_number_of_letters_displayed_left_0_0 = TensorMap('5207_Final-number-of-letters-displayed-left_0_0', loss='logcosh', channel_map={'no_5207_Final-number-of-letters-displayed-left_0_0': 0})
field_id_5208_logmar_final_left_0_0 = TensorMap('5208_logMAR-final-left_0_0', loss='logcosh', channel_map={'no_5208_logMAR-final-left_0_0': 0})
field_id_5209_number_of_rounds_to_result_left_0_0 = TensorMap('5209_Number-of-rounds-to-result-left_0_0', loss='logcosh', channel_map={'no_5209_Number-of-rounds-to-result-left_0_0': 0})
field_id_5211_distance_of_viewer_to_screen_left_0_0 = TensorMap('5211_Distance-of-viewer-to-screen-left_0_0', loss='logcosh', channel_map={'no_5211_Distance-of-viewer-to-screen-left_0_0': 0})
field_id_5215_vertex_distance_right_0_0 = TensorMap('5215_Vertex-distance-right_0_0', loss='logcosh', channel_map={'no_5215_Vertex-distance-right_0_0': 0})
field_id_5221_index_of_best_refractometry_result_right_0_0 = TensorMap('5221_Index-of-best-refractometry-result-right_0_0', loss='logcosh', channel_map={'no_5221_Index-of-best-refractometry-result-right_0_0': 0})
field_id_5237_3mm_index_of_best_keratometry_results_right_0_0 = TensorMap('5237_3mm-index-of-best-keratometry-results-right_0_0', loss='logcosh', channel_map={'no_5237_3mm-index-of-best-keratometry-results-right_0_0': 0})
field_id_5251_6mm_index_of_best_keratometry_results_right_0_0 = TensorMap('5251_6mm-index-of-best-keratometry-results-right_0_0', loss='logcosh', channel_map={'no_5251_6mm-index-of-best-keratometry-results-right_0_0': 0})
field_id_5254_intraocular_pressure_cornealcompensated_right_0_0 = TensorMap('5254_Intraocular-pressure-cornealcompensated-right_0_0', loss='logcosh', channel_map={'no_5254_Intraocular-pressure-cornealcompensated-right_0_0': 0})
field_id_5255_intraocular_pressure_goldmanncorrelated_right_0_0 = TensorMap('5255_Intraocular-pressure-Goldmanncorrelated-right_0_0', loss='logcosh', channel_map={'no_5255_Intraocular-pressure-Goldmanncorrelated-right_0_0': 0})
field_id_5256_corneal_hysteresis_right_0_0 = TensorMap('5256_Corneal-hysteresis-right_0_0', loss='logcosh', channel_map={'no_5256_Corneal-hysteresis-right_0_0': 0})
field_id_5257_corneal_resistance_factor_right_0_0 = TensorMap('5257_Corneal-resistance-factor-right_0_0', loss='logcosh', channel_map={'no_5257_Corneal-resistance-factor-right_0_0': 0})
field_id_5262_intraocular_pressure_cornealcompensated_left_0_0 = TensorMap('5262_Intraocular-pressure-cornealcompensated-left_0_0', loss='logcosh', channel_map={'no_5262_Intraocular-pressure-cornealcompensated-left_0_0': 0})
field_id_5263_intraocular_pressure_goldmanncorrelated_left_0_0 = TensorMap('5263_Intraocular-pressure-Goldmanncorrelated-left_0_0', loss='logcosh', channel_map={'no_5263_Intraocular-pressure-Goldmanncorrelated-left_0_0': 0})
field_id_5264_corneal_hysteresis_left_0_0 = TensorMap('5264_Corneal-hysteresis-left_0_0', loss='logcosh', channel_map={'no_5264_Corneal-hysteresis-left_0_0': 0})
field_id_5265_corneal_resistance_factor_left_0_0 = TensorMap('5265_Corneal-resistance-factor-left_0_0', loss='logcosh', channel_map={'no_5265_Corneal-resistance-factor-left_0_0': 0})
field_id_5274_vertex_distance_left_0_0 = TensorMap('5274_Vertex-distance-left_0_0', loss='logcosh', channel_map={'no_5274_Vertex-distance-left_0_0': 0})
field_id_5276_index_of_best_refractometry_result_left_0_0 = TensorMap('5276_Index-of-best-refractometry-result-left_0_0', loss='logcosh', channel_map={'no_5276_Index-of-best-refractometry-result-left_0_0': 0})
field_id_5292_3mm_index_of_best_keratometry_results_left_0_0 = TensorMap('5292_3mm-index-of-best-keratometry-results-left_0_0', loss='logcosh', channel_map={'no_5292_3mm-index-of-best-keratometry-results-left_0_0': 0})
field_id_5306_6mm_index_of_best_keratometry_results_left_0_0 = TensorMap('5306_6mm-index-of-best-keratometry-results-left_0_0', loss='logcosh', channel_map={'no_5306_6mm-index-of-best-keratometry-results-left_0_0': 0})
field_id_5364_average_weekly_intake_of_other_alcoholic_drinks_0_0 = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_0_0', loss='logcosh', channel_map={'no_5364_Average-weekly-intake-of-other-alcoholic-drinks_0_0': 0})
field_id_6032_maximum_workload_during_fitness_test_0_0 = TensorMap('6032_Maximum-workload-during-fitness-test_0_0', loss='logcosh', channel_map={'no_6032_Maximum-workload-during-fitness-test_0_0': 0})
field_id_6033_maximum_heart_rate_during_fitness_test_0_0 = TensorMap('6033_Maximum-heart-rate-during-fitness-test_0_0', loss='logcosh', channel_map={'no_6033_Maximum-heart-rate-during-fitness-test_0_0': 0})
field_id_6038_number_of_trend_entries_0_0 = TensorMap('6038_Number-of-trend-entries_0_0', loss='logcosh', channel_map={'no_6038_Number-of-trend-entries_0_0': 0})
field_id_6039_duration_of_fitness_test_0_0 = TensorMap('6039_Duration-of-fitness-test_0_0', loss='logcosh', channel_map={'no_6039_Duration-of-fitness-test_0_0': 0})
field_id_6071_duration_at_which_oct_screen_shown_right_0_0 = TensorMap('6071_Duration-at-which-OCT-screen-shown-right_0_0', loss='logcosh', channel_map={'no_6071_Duration-at-which-OCT-screen-shown-right_0_0': 0})
field_id_6073_duration_at_which_oct_screen_shown_left_0_0 = TensorMap('6073_Duration-at-which-OCT-screen-shown-left_0_0', loss='logcosh', channel_map={'no_6073_Duration-at-which-OCT-screen-shown-left_0_0': 0})
field_id_92_operation_yearage_first_occurred_0_1 = TensorMap('92_Operation-yearage-first-occurred_0_1', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_0_1': 0})
field_id_102_pulse_rate_automated_reading_2_0 = TensorMap('102_Pulse-rate-automated-reading_2_0', loss='logcosh', channel_map={'no_102_Pulse-rate-automated-reading_2_0': 0})
field_id_102_pulse_rate_automated_reading_2_1 = TensorMap('102_Pulse-rate-automated-reading_2_1', loss='logcosh', channel_map={'no_102_Pulse-rate-automated-reading_2_1': 0})
field_id_1050_time_spend_outdoors_in_summer_2_0 = TensorMap('1050_Time-spend-outdoors-in-summer_2_0', loss='logcosh', channel_map={'no_1050_Time-spend-outdoors-in-summer_2_0': 0})
field_id_1060_time_spent_outdoors_in_winter_2_0 = TensorMap('1060_Time-spent-outdoors-in-winter_2_0', loss='logcosh', channel_map={'no_1060_Time-spent-outdoors-in-winter_2_0': 0})
field_id_1070_time_spent_watching_television_tv_2_0 = TensorMap('1070_Time-spent-watching-television-TV_2_0', loss='logcosh', channel_map={'no_1070_Time-spent-watching-television-TV_2_0': 0})
field_id_1080_time_spent_using_computer_2_0 = TensorMap('1080_Time-spent-using-computer_2_0', loss='logcosh', channel_map={'no_1080_Time-spent-using-computer_2_0': 0})
field_id_1090_time_spent_driving_2_0 = TensorMap('1090_Time-spent-driving_2_0', loss='logcosh', channel_map={'no_1090_Time-spent-driving_2_0': 0})
field_id_1269_exposure_to_tobacco_smoke_at_home_2_0 = TensorMap('1269_Exposure-to-tobacco-smoke-at-home_2_0', loss='logcosh', channel_map={'no_1269_Exposure-to-tobacco-smoke-at-home_2_0': 0})
field_id_1279_exposure_to_tobacco_smoke_outside_home_2_0 = TensorMap('1279_Exposure-to-tobacco-smoke-outside-home_2_0', loss='logcosh', channel_map={'no_1279_Exposure-to-tobacco-smoke-outside-home_2_0': 0})
field_id_1289_cooked_vegetable_intake_2_0 = TensorMap('1289_Cooked-vegetable-intake_2_0', loss='logcosh', channel_map={'no_1289_Cooked-vegetable-intake_2_0': 0})
field_id_1299_salad_raw_vegetable_intake_2_0 = TensorMap('1299_Salad-raw-vegetable-intake_2_0', loss='logcosh', channel_map={'no_1299_Salad-raw-vegetable-intake_2_0': 0})
field_id_1309_fresh_fruit_intake_2_0 = TensorMap('1309_Fresh-fruit-intake_2_0', loss='logcosh', channel_map={'no_1309_Fresh-fruit-intake_2_0': 0})
field_id_1319_dried_fruit_intake_2_0 = TensorMap('1319_Dried-fruit-intake_2_0', loss='logcosh', channel_map={'no_1319_Dried-fruit-intake_2_0': 0})
field_id_134_number_of_selfreported_cancers_2_0 = TensorMap('134_Number-of-selfreported-cancers_2_0', loss='logcosh', channel_map={'no_134_Number-of-selfreported-cancers_2_0': 0})
field_id_135_number_of_selfreported_noncancer_illnesses_2_0 = TensorMap('135_Number-of-selfreported-noncancer-illnesses_2_0', loss='logcosh', channel_map={'no_135_Number-of-selfreported-noncancer-illnesses_2_0': 0})
field_id_136_number_of_operations_selfreported_2_0 = TensorMap('136_Number-of-operations-selfreported_2_0', loss='logcosh', channel_map={'no_136_Number-of-operations-selfreported_2_0': 0})
field_id_137_number_of_treatmentsmedications_taken_2_0 = TensorMap('137_Number-of-treatmentsmedications-taken_2_0', loss='logcosh', channel_map={'no_137_Number-of-treatmentsmedications-taken_2_0': 0})
field_id_1438_bread_intake_2_0 = TensorMap('1438_Bread-intake_2_0', loss='logcosh', channel_map={'no_1438_Bread-intake_2_0': 0})
field_id_1458_cereal_intake_2_0 = TensorMap('1458_Cereal-intake_2_0', loss='logcosh', channel_map={'no_1458_Cereal-intake_2_0': 0})
field_id_1488_tea_intake_2_0 = TensorMap('1488_Tea-intake_2_0', loss='logcosh', channel_map={'no_1488_Tea-intake_2_0': 0})
field_id_1498_coffee_intake_2_0 = TensorMap('1498_Coffee-intake_2_0', loss='logcosh', channel_map={'no_1498_Coffee-intake_2_0': 0})
field_id_1528_water_intake_2_0 = TensorMap('1528_Water-intake_2_0', loss='logcosh', channel_map={'no_1528_Water-intake_2_0': 0})
field_id_1568_average_weekly_red_wine_intake_2_0 = TensorMap('1568_Average-weekly-red-wine-intake_2_0', loss='logcosh', channel_map={'no_1568_Average-weekly-red-wine-intake_2_0': 0})
field_id_1578_average_weekly_champagne_plus_white_wine_intake_2_0 = TensorMap('1578_Average-weekly-champagne-plus-white-wine-intake_2_0', loss='logcosh', channel_map={'no_1578_Average-weekly-champagne-plus-white-wine-intake_2_0': 0})
field_id_1588_average_weekly_beer_plus_cider_intake_2_0 = TensorMap('1588_Average-weekly-beer-plus-cider-intake_2_0', loss='logcosh', channel_map={'no_1588_Average-weekly-beer-plus-cider-intake_2_0': 0})
field_id_1598_average_weekly_spirits_intake_2_0 = TensorMap('1598_Average-weekly-spirits-intake_2_0', loss='logcosh', channel_map={'no_1598_Average-weekly-spirits-intake_2_0': 0})
field_id_1608_average_weekly_fortified_wine_intake_2_0 = TensorMap('1608_Average-weekly-fortified-wine-intake_2_0', loss='logcosh', channel_map={'no_1608_Average-weekly-fortified-wine-intake_2_0': 0})
field_id_1807_fathers_age_at_death_0_0 = TensorMap('1807_Fathers-age-at-death_0_0', loss='logcosh', channel_map={'no_1807_Fathers-age-at-death_0_0': 0})
field_id_1807_fathers_age_at_death_2_0 = TensorMap('1807_Fathers-age-at-death_2_0', loss='logcosh', channel_map={'no_1807_Fathers-age-at-death_2_0': 0})
field_id_1873_number_of_full_brothers_2_0 = TensorMap('1873_Number-of-full-brothers_2_0', loss='logcosh', channel_map={'no_1873_Number-of-full-brothers_2_0': 0})
field_id_1883_number_of_full_sisters_2_0 = TensorMap('1883_Number-of-full-sisters_2_0', loss='logcosh', channel_map={'no_1883_Number-of-full-sisters_2_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_0 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_0', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_2_0 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_0', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_2_1 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_1', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_1': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_2_2 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_2', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_2': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_2_3 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_3', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_3': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_2_4 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_4', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_4': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_2_0 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_2_0', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_2_0': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_2_1 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_2_1', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_2_1': 0})
field_id_20015_sitting_height_2_0 = TensorMap('20015_Sitting-height_2_0', loss='logcosh', channel_map={'no_20015_Sitting-height_2_0': 0})
field_id_20016_fluid_intelligence_score_2_0 = TensorMap('20016_Fluid-intelligence-score_2_0', loss='logcosh', channel_map={'no_20016_Fluid-intelligence-score_2_0': 0})
field_id_20019_speechreceptionthreshold_srt_estimate_left_2_0 = TensorMap('20019_Speechreceptionthreshold-SRT-estimate-left_2_0', loss='logcosh', channel_map={'no_20019_Speechreceptionthreshold-SRT-estimate-left_2_0': 0})
field_id_20021_speechreceptionthreshold_srt_estimate_right_2_0 = TensorMap('20021_Speechreceptionthreshold-SRT-estimate-right_2_0', loss='logcosh', channel_map={'no_20021_Speechreceptionthreshold-SRT-estimate-right_2_0': 0})
field_id_20023_mean_time_to_correctly_identify_matches_2_0 = TensorMap('20023_Mean-time-to-correctly-identify-matches_2_0', loss='logcosh', channel_map={'no_20023_Mean-time-to-correctly-identify-matches_2_0': 0})
field_id_20128_number_of_fluid_intelligence_questions_attempted_within_time_limit_2_0 = TensorMap('20128_Number-of-fluid-intelligence-questions-attempted-within-time-limit_2_0', loss='logcosh', channel_map={'no_20128_Number-of-fluid-intelligence-questions-attempted-within-time-limit_2_0': 0})
field_id_21001_body_mass_index_bmi_2_0 = TensorMap('21001_Body-mass-index-BMI_2_0', loss='logcosh', channel_map={'no_21001_Body-mass-index-BMI_2_0': 0})
field_id_21002_weight_2_0 = TensorMap('21002_Weight_2_0', loss='logcosh', channel_map={'no_21002_Weight_2_0': 0})
field_id_21003_age_when_attended_assessment_centre_2_0 = TensorMap('21003_Age-when-attended-assessment-centre_2_0', loss='logcosh', channel_map={'no_21003_Age-when-attended-assessment-centre_2_0': 0})
field_id_21021_pulse_wave_arterial_stiffness_index_2_0 = TensorMap('21021_Pulse-wave-Arterial-Stiffness-index_2_0', loss='logcosh', channel_map={'no_21021_Pulse-wave-Arterial-Stiffness-index_2_0': 0})
field_id_2217_age_started_wearing_glasses_or_contact_lenses_2_0 = TensorMap('2217_Age-started-wearing-glasses-or-contact-lenses_2_0', loss='logcosh', channel_map={'no_2217_Age-started-wearing-glasses-or-contact-lenses_2_0': 0})
field_id_22602_year_job_started_0_4 = TensorMap('22602_Year-job-started_0_4', loss='logcosh', channel_map={'no_22602_Year-job-started_0_4': 0})
field_id_22602_year_job_started_0_5 = TensorMap('22602_Year-job-started_0_5', loss='logcosh', channel_map={'no_22602_Year-job-started_0_5': 0})
field_id_22603_year_job_ended_0_4 = TensorMap('22603_Year-job-ended_0_4', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_4': 0})
field_id_22603_year_job_ended_0_5 = TensorMap('22603_Year-job-ended_0_5', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_5': 0})
field_id_2277_frequency_of_solariumsunlamp_use_2_0 = TensorMap('2277_Frequency-of-solariumsunlamp-use_2_0', loss='logcosh', channel_map={'no_2277_Frequency-of-solariumsunlamp-use_2_0': 0})
field_id_23098_weight_2_0 = TensorMap('23098_Weight_2_0', loss='logcosh', channel_map={'no_23098_Weight_2_0': 0})
field_id_23099_body_fat_percentage_2_0 = TensorMap('23099_Body-fat-percentage_2_0', loss='logcosh', channel_map={'no_23099_Body-fat-percentage_2_0': 0})
field_id_23100_whole_body_fat_mass_2_0 = TensorMap('23100_Whole-body-fat-mass_2_0', loss='logcosh', channel_map={'no_23100_Whole-body-fat-mass_2_0': 0})
field_id_23101_whole_body_fatfree_mass_2_0 = TensorMap('23101_Whole-body-fatfree-mass_2_0', loss='logcosh', channel_map={'no_23101_Whole-body-fatfree-mass_2_0': 0})
field_id_23102_whole_body_water_mass_2_0 = TensorMap('23102_Whole-body-water-mass_2_0', loss='logcosh', channel_map={'no_23102_Whole-body-water-mass_2_0': 0})
field_id_23104_body_mass_index_bmi_2_0 = TensorMap('23104_Body-mass-index-BMI_2_0', loss='logcosh', channel_map={'no_23104_Body-mass-index-BMI_2_0': 0})
field_id_23105_basal_metabolic_rate_2_0 = TensorMap('23105_Basal-metabolic-rate_2_0', loss='logcosh', channel_map={'no_23105_Basal-metabolic-rate_2_0': 0})
field_id_23106_impedance_of_whole_body_2_0 = TensorMap('23106_Impedance-of-whole-body_2_0', loss='logcosh', channel_map={'no_23106_Impedance-of-whole-body_2_0': 0})
field_id_23107_impedance_of_leg_right_2_0 = TensorMap('23107_Impedance-of-leg-right_2_0', loss='logcosh', channel_map={'no_23107_Impedance-of-leg-right_2_0': 0})
field_id_23108_impedance_of_leg_left_2_0 = TensorMap('23108_Impedance-of-leg-left_2_0', loss='logcosh', channel_map={'no_23108_Impedance-of-leg-left_2_0': 0})
field_id_23109_impedance_of_arm_right_2_0 = TensorMap('23109_Impedance-of-arm-right_2_0', loss='logcosh', channel_map={'no_23109_Impedance-of-arm-right_2_0': 0})
field_id_23110_impedance_of_arm_left_2_0 = TensorMap('23110_Impedance-of-arm-left_2_0', loss='logcosh', channel_map={'no_23110_Impedance-of-arm-left_2_0': 0})
field_id_23111_leg_fat_percentage_right_2_0 = TensorMap('23111_Leg-fat-percentage-right_2_0', loss='logcosh', channel_map={'no_23111_Leg-fat-percentage-right_2_0': 0})
field_id_23112_leg_fat_mass_right_2_0 = TensorMap('23112_Leg-fat-mass-right_2_0', loss='logcosh', channel_map={'no_23112_Leg-fat-mass-right_2_0': 0})
field_id_23113_leg_fatfree_mass_right_2_0 = TensorMap('23113_Leg-fatfree-mass-right_2_0', loss='logcosh', channel_map={'no_23113_Leg-fatfree-mass-right_2_0': 0})
field_id_23114_leg_predicted_mass_right_2_0 = TensorMap('23114_Leg-predicted-mass-right_2_0', loss='logcosh', channel_map={'no_23114_Leg-predicted-mass-right_2_0': 0})
field_id_23115_leg_fat_percentage_left_2_0 = TensorMap('23115_Leg-fat-percentage-left_2_0', loss='logcosh', channel_map={'no_23115_Leg-fat-percentage-left_2_0': 0})
field_id_23116_leg_fat_mass_left_2_0 = TensorMap('23116_Leg-fat-mass-left_2_0', loss='logcosh', channel_map={'no_23116_Leg-fat-mass-left_2_0': 0})
field_id_23117_leg_fatfree_mass_left_2_0 = TensorMap('23117_Leg-fatfree-mass-left_2_0', loss='logcosh', channel_map={'no_23117_Leg-fatfree-mass-left_2_0': 0})
field_id_23118_leg_predicted_mass_left_2_0 = TensorMap('23118_Leg-predicted-mass-left_2_0', loss='logcosh', channel_map={'no_23118_Leg-predicted-mass-left_2_0': 0})
field_id_23119_arm_fat_percentage_right_2_0 = TensorMap('23119_Arm-fat-percentage-right_2_0', loss='logcosh', channel_map={'no_23119_Arm-fat-percentage-right_2_0': 0})
field_id_23120_arm_fat_mass_right_2_0 = TensorMap('23120_Arm-fat-mass-right_2_0', loss='logcosh', channel_map={'no_23120_Arm-fat-mass-right_2_0': 0})
field_id_23121_arm_fatfree_mass_right_2_0 = TensorMap('23121_Arm-fatfree-mass-right_2_0', loss='logcosh', channel_map={'no_23121_Arm-fatfree-mass-right_2_0': 0})
field_id_23122_arm_predicted_mass_right_2_0 = TensorMap('23122_Arm-predicted-mass-right_2_0', loss='logcosh', channel_map={'no_23122_Arm-predicted-mass-right_2_0': 0})
field_id_23123_arm_fat_percentage_left_2_0 = TensorMap('23123_Arm-fat-percentage-left_2_0', loss='logcosh', channel_map={'no_23123_Arm-fat-percentage-left_2_0': 0})
field_id_23124_arm_fat_mass_left_2_0 = TensorMap('23124_Arm-fat-mass-left_2_0', loss='logcosh', channel_map={'no_23124_Arm-fat-mass-left_2_0': 0})
field_id_23125_arm_fatfree_mass_left_2_0 = TensorMap('23125_Arm-fatfree-mass-left_2_0', loss='logcosh', channel_map={'no_23125_Arm-fatfree-mass-left_2_0': 0})
field_id_23126_arm_predicted_mass_left_2_0 = TensorMap('23126_Arm-predicted-mass-left_2_0', loss='logcosh', channel_map={'no_23126_Arm-predicted-mass-left_2_0': 0})
field_id_23127_trunk_fat_percentage_2_0 = TensorMap('23127_Trunk-fat-percentage_2_0', loss='logcosh', channel_map={'no_23127_Trunk-fat-percentage_2_0': 0})
field_id_23128_trunk_fat_mass_2_0 = TensorMap('23128_Trunk-fat-mass_2_0', loss='logcosh', channel_map={'no_23128_Trunk-fat-mass_2_0': 0})
field_id_23129_trunk_fatfree_mass_2_0 = TensorMap('23129_Trunk-fatfree-mass_2_0', loss='logcosh', channel_map={'no_23129_Trunk-fatfree-mass_2_0': 0})
field_id_23130_trunk_predicted_mass_2_0 = TensorMap('23130_Trunk-predicted-mass_2_0', loss='logcosh', channel_map={'no_23130_Trunk-predicted-mass_2_0': 0})
field_id_2355_most_recent_bowel_cancer_screening_0_0 = TensorMap('2355_Most-recent-bowel-cancer-screening_0_0', loss='logcosh', channel_map={'no_2355_Most-recent-bowel-cancer-screening_0_0': 0})
field_id_2355_most_recent_bowel_cancer_screening_2_0 = TensorMap('2355_Most-recent-bowel-cancer-screening_2_0', loss='logcosh', channel_map={'no_2355_Most-recent-bowel-cancer-screening_2_0': 0})
field_id_25000_volumetric_scaling_from_t1_head_image_to_standard_space_2_0 = TensorMap('25000_Volumetric-scaling-from-T1-head-image-to-standard-space_2_0', loss='logcosh', channel_map={'no_25000_Volumetric-scaling-from-T1-head-image-to-standard-space_2_0': 0})
field_id_25001_volume_of_peripheral_cortical_grey_matter_normalised_for_head_size_2_0 = TensorMap('25001_Volume-of-peripheral-cortical-grey-matter-normalised-for-head-size_2_0', loss='logcosh', channel_map={'no_25001_Volume-of-peripheral-cortical-grey-matter-normalised-for-head-size_2_0': 0})
field_id_25002_volume_of_peripheral_cortical_grey_matter_2_0 = TensorMap('25002_Volume-of-peripheral-cortical-grey-matter_2_0', loss='logcosh', channel_map={'no_25002_Volume-of-peripheral-cortical-grey-matter_2_0': 0})
field_id_25003_volume_of_ventricular_cerebrospinal_fluid_normalised_for_head_size_2_0 = TensorMap('25003_Volume-of-ventricular-cerebrospinal-fluid-normalised-for-head-size_2_0', loss='logcosh', channel_map={'no_25003_Volume-of-ventricular-cerebrospinal-fluid-normalised-for-head-size_2_0': 0})
field_id_25004_volume_of_ventricular_cerebrospinal_fluid_2_0 = TensorMap('25004_Volume-of-ventricular-cerebrospinal-fluid_2_0', loss='logcosh', channel_map={'no_25004_Volume-of-ventricular-cerebrospinal-fluid_2_0': 0})
field_id_25005_volume_of_grey_matter_normalised_for_head_size_2_0 = TensorMap('25005_Volume-of-grey-matter-normalised-for-head-size_2_0', loss='logcosh', channel_map={'no_25005_Volume-of-grey-matter-normalised-for-head-size_2_0': 0})
field_id_25006_volume_of_grey_matter_2_0 = TensorMap('25006_Volume-of-grey-matter_2_0', loss='logcosh', channel_map={'no_25006_Volume-of-grey-matter_2_0': 0})
field_id_25007_volume_of_white_matter_normalised_for_head_size_2_0 = TensorMap('25007_Volume-of-white-matter-normalised-for-head-size_2_0', loss='logcosh', channel_map={'no_25007_Volume-of-white-matter-normalised-for-head-size_2_0': 0})
field_id_25008_volume_of_white_matter_2_0 = TensorMap('25008_Volume-of-white-matter_2_0', loss='logcosh', channel_map={'no_25008_Volume-of-white-matter_2_0': 0})
field_id_25009_volume_of_brain_greywhite_matter_normalised_for_head_size_2_0 = TensorMap('25009_Volume-of-brain-greywhite-matter-normalised-for-head-size_2_0', loss='logcosh', channel_map={'no_25009_Volume-of-brain-greywhite-matter-normalised-for-head-size_2_0': 0})
field_id_25010_volume_of_brain_greywhite_matter_2_0 = TensorMap('25010_Volume-of-brain-greywhite-matter_2_0', loss='logcosh', channel_map={'no_25010_Volume-of-brain-greywhite-matter_2_0': 0})
field_id_25011_volume_of_thalamus_left_2_0 = TensorMap('25011_Volume-of-thalamus-left_2_0', loss='logcosh', channel_map={'no_25011_Volume-of-thalamus-left_2_0': 0})
field_id_25012_volume_of_thalamus_right_2_0 = TensorMap('25012_Volume-of-thalamus-right_2_0', loss='logcosh', channel_map={'no_25012_Volume-of-thalamus-right_2_0': 0})
field_id_25013_volume_of_caudate_left_2_0 = TensorMap('25013_Volume-of-caudate-left_2_0', loss='logcosh', channel_map={'no_25013_Volume-of-caudate-left_2_0': 0})
field_id_25014_volume_of_caudate_right_2_0 = TensorMap('25014_Volume-of-caudate-right_2_0', loss='logcosh', channel_map={'no_25014_Volume-of-caudate-right_2_0': 0})
field_id_25015_volume_of_putamen_left_2_0 = TensorMap('25015_Volume-of-putamen-left_2_0', loss='logcosh', channel_map={'no_25015_Volume-of-putamen-left_2_0': 0})
field_id_25016_volume_of_putamen_right_2_0 = TensorMap('25016_Volume-of-putamen-right_2_0', loss='logcosh', channel_map={'no_25016_Volume-of-putamen-right_2_0': 0})
field_id_25017_volume_of_pallidum_left_2_0 = TensorMap('25017_Volume-of-pallidum-left_2_0', loss='logcosh', channel_map={'no_25017_Volume-of-pallidum-left_2_0': 0})
field_id_25018_volume_of_pallidum_right_2_0 = TensorMap('25018_Volume-of-pallidum-right_2_0', loss='logcosh', channel_map={'no_25018_Volume-of-pallidum-right_2_0': 0})
field_id_25019_volume_of_hippocampus_left_2_0 = TensorMap('25019_Volume-of-hippocampus-left_2_0', loss='logcosh', channel_map={'no_25019_Volume-of-hippocampus-left_2_0': 0})
field_id_25020_volume_of_hippocampus_right_2_0 = TensorMap('25020_Volume-of-hippocampus-right_2_0', loss='logcosh', channel_map={'no_25020_Volume-of-hippocampus-right_2_0': 0})
field_id_25021_volume_of_amygdala_left_2_0 = TensorMap('25021_Volume-of-amygdala-left_2_0', loss='logcosh', channel_map={'no_25021_Volume-of-amygdala-left_2_0': 0})
field_id_25022_volume_of_amygdala_right_2_0 = TensorMap('25022_Volume-of-amygdala-right_2_0', loss='logcosh', channel_map={'no_25022_Volume-of-amygdala-right_2_0': 0})
field_id_25023_volume_of_accumbens_left_2_0 = TensorMap('25023_Volume-of-accumbens-left_2_0', loss='logcosh', channel_map={'no_25023_Volume-of-accumbens-left_2_0': 0})
field_id_25024_volume_of_accumbens_right_2_0 = TensorMap('25024_Volume-of-accumbens-right_2_0', loss='logcosh', channel_map={'no_25024_Volume-of-accumbens-right_2_0': 0})
field_id_25025_volume_of_brain_stem_4th_ventricle_2_0 = TensorMap('25025_Volume-of-brain-stem-4th-ventricle_2_0', loss='logcosh', channel_map={'no_25025_Volume-of-brain-stem-4th-ventricle_2_0': 0})
field_id_25026_median_t2star_in_thalamus_left_2_0 = TensorMap('25026_Median-T2star-in-thalamus-left_2_0', loss='logcosh', channel_map={'no_25026_Median-T2star-in-thalamus-left_2_0': 0})
field_id_25027_median_t2star_in_thalamus_right_2_0 = TensorMap('25027_Median-T2star-in-thalamus-right_2_0', loss='logcosh', channel_map={'no_25027_Median-T2star-in-thalamus-right_2_0': 0})
field_id_25028_median_t2star_in_caudate_left_2_0 = TensorMap('25028_Median-T2star-in-caudate-left_2_0', loss='logcosh', channel_map={'no_25028_Median-T2star-in-caudate-left_2_0': 0})
field_id_25029_median_t2star_in_caudate_right_2_0 = TensorMap('25029_Median-T2star-in-caudate-right_2_0', loss='logcosh', channel_map={'no_25029_Median-T2star-in-caudate-right_2_0': 0})
field_id_25030_median_t2star_in_putamen_left_2_0 = TensorMap('25030_Median-T2star-in-putamen-left_2_0', loss='logcosh', channel_map={'no_25030_Median-T2star-in-putamen-left_2_0': 0})
field_id_25031_median_t2star_in_putamen_right_2_0 = TensorMap('25031_Median-T2star-in-putamen-right_2_0', loss='logcosh', channel_map={'no_25031_Median-T2star-in-putamen-right_2_0': 0})
field_id_25032_median_t2star_in_pallidum_left_2_0 = TensorMap('25032_Median-T2star-in-pallidum-left_2_0', loss='logcosh', channel_map={'no_25032_Median-T2star-in-pallidum-left_2_0': 0})
field_id_25033_median_t2star_in_pallidum_right_2_0 = TensorMap('25033_Median-T2star-in-pallidum-right_2_0', loss='logcosh', channel_map={'no_25033_Median-T2star-in-pallidum-right_2_0': 0})
field_id_25034_median_t2star_in_hippocampus_left_2_0 = TensorMap('25034_Median-T2star-in-hippocampus-left_2_0', loss='logcosh', channel_map={'no_25034_Median-T2star-in-hippocampus-left_2_0': 0})
field_id_25035_median_t2star_in_hippocampus_right_2_0 = TensorMap('25035_Median-T2star-in-hippocampus-right_2_0', loss='logcosh', channel_map={'no_25035_Median-T2star-in-hippocampus-right_2_0': 0})
field_id_25036_median_t2star_in_amygdala_left_2_0 = TensorMap('25036_Median-T2star-in-amygdala-left_2_0', loss='logcosh', channel_map={'no_25036_Median-T2star-in-amygdala-left_2_0': 0})
field_id_25037_median_t2star_in_amygdala_right_2_0 = TensorMap('25037_Median-T2star-in-amygdala-right_2_0', loss='logcosh', channel_map={'no_25037_Median-T2star-in-amygdala-right_2_0': 0})
field_id_25038_median_t2star_in_accumbens_left_2_0 = TensorMap('25038_Median-T2star-in-accumbens-left_2_0', loss='logcosh', channel_map={'no_25038_Median-T2star-in-accumbens-left_2_0': 0})
field_id_25039_median_t2star_in_accumbens_right_2_0 = TensorMap('25039_Median-T2star-in-accumbens-right_2_0', loss='logcosh', channel_map={'no_25039_Median-T2star-in-accumbens-right_2_0': 0})
field_id_25056_mean_fa_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25056_Mean-FA-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25056_Mean-FA-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25057_mean_fa_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25057_Mean-FA-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25057_Mean-FA-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25058_mean_fa_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25058_Mean-FA-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25058_Mean-FA-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25059_mean_fa_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25059_Mean-FA-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25059_Mean-FA-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25060_mean_fa_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25060_Mean-FA-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25060_Mean-FA-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25061_mean_fa_in_fornix_on_fa_skeleton_2_0 = TensorMap('25061_Mean-FA-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25061_Mean-FA-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25062_mean_fa_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25062_Mean-FA-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25062_Mean-FA-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25063_mean_fa_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25063_Mean-FA-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25063_Mean-FA-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25064_mean_fa_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25064_Mean-FA-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25064_Mean-FA-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25065_mean_fa_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25065_Mean-FA-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25065_Mean-FA-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25066_mean_fa_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25066_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25066_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25067_mean_fa_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25067_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25067_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25068_mean_fa_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25068_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25068_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25069_mean_fa_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25069_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25069_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25070_mean_fa_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25070_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25070_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25071_mean_fa_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25071_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25071_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25072_mean_fa_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25072_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25072_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25073_mean_fa_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25073_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25073_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25074_mean_fa_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25074_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25074_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25075_mean_fa_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25075_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25075_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25076_mean_fa_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25076_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25076_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25077_mean_fa_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25077_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25077_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25078_mean_fa_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25078_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25078_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25079_mean_fa_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25079_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25079_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25080_mean_fa_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25080_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25080_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25081_mean_fa_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25081_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25081_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25082_mean_fa_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25082_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25082_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25083_mean_fa_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25083_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25083_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25084_mean_fa_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25084_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25084_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25085_mean_fa_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25085_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25085_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25086_mean_fa_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25086_Mean-FA-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25086_Mean-FA-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25087_mean_fa_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25087_Mean-FA-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25087_Mean-FA-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25088_mean_fa_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25088_Mean-FA-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25088_Mean-FA-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25089_mean_fa_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25089_Mean-FA-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25089_Mean-FA-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25090_mean_fa_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25090_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25090_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25091_mean_fa_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25091_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25091_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25092_mean_fa_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25092_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25092_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25093_mean_fa_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25093_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25093_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25094_mean_fa_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25094_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25094_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25095_mean_fa_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25095_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25095_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25096_mean_fa_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25096_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25096_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25097_mean_fa_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25097_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25097_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25098_mean_fa_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25098_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25098_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25099_mean_fa_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25099_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25099_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25100_mean_fa_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25100_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25100_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25101_mean_fa_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25101_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25101_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25102_mean_fa_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25102_Mean-FA-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25102_Mean-FA-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25103_mean_fa_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25103_Mean-FA-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25103_Mean-FA-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25104_mean_md_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25104_Mean-MD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25104_Mean-MD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25105_mean_md_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25105_Mean-MD-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25105_Mean-MD-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25106_mean_md_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25106_Mean-MD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25106_Mean-MD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25107_mean_md_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25107_Mean-MD-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25107_Mean-MD-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25108_mean_md_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25108_Mean-MD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25108_Mean-MD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25109_mean_md_in_fornix_on_fa_skeleton_2_0 = TensorMap('25109_Mean-MD-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25109_Mean-MD-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25110_mean_md_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25110_Mean-MD-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25110_Mean-MD-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25111_mean_md_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25111_Mean-MD-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25111_Mean-MD-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25112_mean_md_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25112_Mean-MD-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25112_Mean-MD-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25113_mean_md_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25113_Mean-MD-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25113_Mean-MD-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25114_mean_md_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25114_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25114_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25115_mean_md_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25115_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25115_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25116_mean_md_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25116_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25116_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25117_mean_md_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25117_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25117_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25118_mean_md_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25118_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25118_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25119_mean_md_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25119_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25119_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25120_mean_md_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25120_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25120_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25121_mean_md_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25121_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25121_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25122_mean_md_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25122_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25122_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25123_mean_md_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25123_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25123_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25124_mean_md_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25124_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25124_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25125_mean_md_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25125_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25125_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25126_mean_md_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25126_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25126_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25127_mean_md_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25127_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25127_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25128_mean_md_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25128_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25128_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25129_mean_md_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25129_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25129_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25130_mean_md_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25130_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25130_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25131_mean_md_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25131_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25131_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25132_mean_md_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25132_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25132_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25133_mean_md_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25133_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25133_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25134_mean_md_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25134_Mean-MD-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25134_Mean-MD-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25135_mean_md_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25135_Mean-MD-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25135_Mean-MD-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25136_mean_md_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25136_Mean-MD-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25136_Mean-MD-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25137_mean_md_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25137_Mean-MD-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25137_Mean-MD-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25138_mean_md_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25138_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25138_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25139_mean_md_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25139_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25139_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25140_mean_md_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25140_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25140_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25141_mean_md_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25141_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25141_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25142_mean_md_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25142_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25142_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25143_mean_md_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25143_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25143_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25144_mean_md_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25144_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25144_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25145_mean_md_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25145_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25145_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25146_mean_md_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25146_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25146_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25147_mean_md_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25147_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25147_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25148_mean_md_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25148_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25148_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25149_mean_md_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25149_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25149_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25150_mean_md_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25150_Mean-MD-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25150_Mean-MD-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25151_mean_md_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25151_Mean-MD-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25151_Mean-MD-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25152_mean_mo_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25152_Mean-MO-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25152_Mean-MO-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25153_mean_mo_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25153_Mean-MO-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25153_Mean-MO-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25154_mean_mo_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25154_Mean-MO-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25154_Mean-MO-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25155_mean_mo_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25155_Mean-MO-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25155_Mean-MO-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25156_mean_mo_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25156_Mean-MO-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25156_Mean-MO-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25157_mean_mo_in_fornix_on_fa_skeleton_2_0 = TensorMap('25157_Mean-MO-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25157_Mean-MO-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25158_mean_mo_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25158_Mean-MO-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25158_Mean-MO-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25159_mean_mo_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25159_Mean-MO-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25159_Mean-MO-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25160_mean_mo_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25160_Mean-MO-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25160_Mean-MO-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25161_mean_mo_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25161_Mean-MO-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25161_Mean-MO-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25162_mean_mo_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25162_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25162_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25163_mean_mo_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25163_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25163_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25164_mean_mo_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25164_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25164_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25165_mean_mo_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25165_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25165_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25166_mean_mo_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25166_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25166_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25167_mean_mo_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25167_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25167_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25168_mean_mo_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25168_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25168_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25169_mean_mo_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25169_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25169_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25170_mean_mo_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25170_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25170_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25171_mean_mo_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25171_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25171_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25172_mean_mo_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25172_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25172_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25173_mean_mo_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25173_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25173_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25174_mean_mo_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25174_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25174_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25175_mean_mo_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25175_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25175_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25176_mean_mo_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25176_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25176_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25177_mean_mo_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25177_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25177_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25178_mean_mo_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25178_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25178_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25179_mean_mo_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25179_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25179_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25180_mean_mo_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25180_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25180_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25181_mean_mo_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25181_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25181_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25182_mean_mo_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25182_Mean-MO-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25182_Mean-MO-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25183_mean_mo_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25183_Mean-MO-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25183_Mean-MO-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25184_mean_mo_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25184_Mean-MO-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25184_Mean-MO-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25185_mean_mo_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25185_Mean-MO-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25185_Mean-MO-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25186_mean_mo_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25186_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25186_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25187_mean_mo_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25187_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25187_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25188_mean_mo_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25188_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25188_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25189_mean_mo_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25189_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25189_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25190_mean_mo_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25190_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25190_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25191_mean_mo_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25191_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25191_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25192_mean_mo_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25192_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25192_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25193_mean_mo_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25193_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25193_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25194_mean_mo_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25194_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25194_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25195_mean_mo_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25195_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25195_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25196_mean_mo_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25196_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25196_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25197_mean_mo_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25197_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25197_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25198_mean_mo_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25198_Mean-MO-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25198_Mean-MO-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25199_mean_mo_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25199_Mean-MO-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25199_Mean-MO-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25200_mean_l1_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25200_Mean-L1-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25200_Mean-L1-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25201_mean_l1_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25201_Mean-L1-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25201_Mean-L1-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25202_mean_l1_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25202_Mean-L1-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25202_Mean-L1-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25203_mean_l1_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25203_Mean-L1-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25203_Mean-L1-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25204_mean_l1_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25204_Mean-L1-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25204_Mean-L1-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25205_mean_l1_in_fornix_on_fa_skeleton_2_0 = TensorMap('25205_Mean-L1-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25205_Mean-L1-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25206_mean_l1_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25206_Mean-L1-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25206_Mean-L1-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25207_mean_l1_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25207_Mean-L1-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25207_Mean-L1-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25208_mean_l1_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25208_Mean-L1-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25208_Mean-L1-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25209_mean_l1_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25209_Mean-L1-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25209_Mean-L1-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25210_mean_l1_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25210_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25210_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25211_mean_l1_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25211_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25211_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25212_mean_l1_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25212_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25212_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25213_mean_l1_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25213_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25213_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25214_mean_l1_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25214_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25214_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25215_mean_l1_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25215_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25215_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25216_mean_l1_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25216_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25216_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25217_mean_l1_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25217_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25217_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25218_mean_l1_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25218_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25218_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25219_mean_l1_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25219_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25219_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25220_mean_l1_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25220_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25220_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25221_mean_l1_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25221_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25221_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25222_mean_l1_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25222_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25222_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25223_mean_l1_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25223_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25223_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25224_mean_l1_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25224_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25224_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25225_mean_l1_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25225_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25225_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25226_mean_l1_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25226_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25226_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25227_mean_l1_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25227_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25227_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25228_mean_l1_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25228_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25228_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25229_mean_l1_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25229_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25229_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25230_mean_l1_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25230_Mean-L1-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25230_Mean-L1-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25231_mean_l1_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25231_Mean-L1-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25231_Mean-L1-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25232_mean_l1_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25232_Mean-L1-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25232_Mean-L1-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25233_mean_l1_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25233_Mean-L1-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25233_Mean-L1-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25234_mean_l1_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25234_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25234_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25235_mean_l1_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25235_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25235_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25236_mean_l1_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25236_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25236_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25237_mean_l1_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25237_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25237_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25238_mean_l1_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25238_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25238_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25239_mean_l1_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25239_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25239_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25240_mean_l1_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25240_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25240_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25241_mean_l1_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25241_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25241_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25242_mean_l1_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25242_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25242_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25243_mean_l1_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25243_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25243_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25244_mean_l1_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25244_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25244_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25245_mean_l1_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25245_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25245_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25246_mean_l1_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25246_Mean-L1-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25246_Mean-L1-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25247_mean_l1_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25247_Mean-L1-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25247_Mean-L1-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25248_mean_l2_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25248_Mean-L2-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25248_Mean-L2-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25249_mean_l2_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25249_Mean-L2-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25249_Mean-L2-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25250_mean_l2_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25250_Mean-L2-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25250_Mean-L2-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25251_mean_l2_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25251_Mean-L2-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25251_Mean-L2-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25252_mean_l2_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25252_Mean-L2-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25252_Mean-L2-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25253_mean_l2_in_fornix_on_fa_skeleton_2_0 = TensorMap('25253_Mean-L2-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25253_Mean-L2-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25254_mean_l2_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25254_Mean-L2-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25254_Mean-L2-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25255_mean_l2_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25255_Mean-L2-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25255_Mean-L2-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25256_mean_l2_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25256_Mean-L2-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25256_Mean-L2-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25257_mean_l2_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25257_Mean-L2-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25257_Mean-L2-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25258_mean_l2_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25258_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25258_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25259_mean_l2_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25259_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25259_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25260_mean_l2_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25260_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25260_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25261_mean_l2_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25261_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25261_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25262_mean_l2_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25262_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25262_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25263_mean_l2_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25263_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25263_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25264_mean_l2_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25264_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25264_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25265_mean_l2_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25265_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25265_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25266_mean_l2_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25266_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25266_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25267_mean_l2_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25267_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25267_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25268_mean_l2_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25268_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25268_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25269_mean_l2_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25269_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25269_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25270_mean_l2_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25270_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25270_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25271_mean_l2_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25271_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25271_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25272_mean_l2_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25272_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25272_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25273_mean_l2_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25273_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25273_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25274_mean_l2_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25274_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25274_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25275_mean_l2_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25275_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25275_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25276_mean_l2_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25276_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25276_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25277_mean_l2_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25277_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25277_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25278_mean_l2_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25278_Mean-L2-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25278_Mean-L2-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25279_mean_l2_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25279_Mean-L2-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25279_Mean-L2-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25280_mean_l2_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25280_Mean-L2-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25280_Mean-L2-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25281_mean_l2_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25281_Mean-L2-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25281_Mean-L2-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25282_mean_l2_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25282_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25282_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25283_mean_l2_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25283_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25283_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25284_mean_l2_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25284_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25284_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25285_mean_l2_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25285_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25285_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25286_mean_l2_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25286_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25286_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25287_mean_l2_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25287_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25287_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25288_mean_l2_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25288_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25288_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25289_mean_l2_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25289_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25289_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25290_mean_l2_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25290_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25290_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25291_mean_l2_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25291_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25291_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25292_mean_l2_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25292_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25292_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25293_mean_l2_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25293_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25293_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25294_mean_l2_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25294_Mean-L2-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25294_Mean-L2-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25295_mean_l2_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25295_Mean-L2-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25295_Mean-L2-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25296_mean_l3_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25296_Mean-L3-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25296_Mean-L3-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25297_mean_l3_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25297_Mean-L3-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25297_Mean-L3-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25298_mean_l3_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25298_Mean-L3-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25298_Mean-L3-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25299_mean_l3_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25299_Mean-L3-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25299_Mean-L3-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25300_mean_l3_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25300_Mean-L3-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25300_Mean-L3-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25301_mean_l3_in_fornix_on_fa_skeleton_2_0 = TensorMap('25301_Mean-L3-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25301_Mean-L3-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25302_mean_l3_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25302_Mean-L3-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25302_Mean-L3-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25303_mean_l3_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25303_Mean-L3-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25303_Mean-L3-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25304_mean_l3_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25304_Mean-L3-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25304_Mean-L3-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25305_mean_l3_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25305_Mean-L3-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25305_Mean-L3-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25306_mean_l3_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25306_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25306_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25307_mean_l3_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25307_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25307_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25308_mean_l3_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25308_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25308_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25309_mean_l3_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25309_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25309_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25310_mean_l3_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25310_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25310_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25311_mean_l3_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25311_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25311_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25312_mean_l3_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25312_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25312_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25313_mean_l3_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25313_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25313_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25314_mean_l3_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25314_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25314_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25315_mean_l3_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25315_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25315_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25316_mean_l3_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25316_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25316_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25317_mean_l3_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25317_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25317_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25318_mean_l3_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25318_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25318_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25319_mean_l3_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25319_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25319_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25320_mean_l3_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25320_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25320_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25321_mean_l3_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25321_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25321_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25322_mean_l3_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25322_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25322_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25323_mean_l3_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25323_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25323_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25324_mean_l3_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25324_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25324_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25325_mean_l3_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25325_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25325_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25326_mean_l3_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25326_Mean-L3-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25326_Mean-L3-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25327_mean_l3_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25327_Mean-L3-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25327_Mean-L3-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25328_mean_l3_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25328_Mean-L3-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25328_Mean-L3-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25329_mean_l3_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25329_Mean-L3-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25329_Mean-L3-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25330_mean_l3_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25330_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25330_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25331_mean_l3_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25331_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25331_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25332_mean_l3_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25332_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25332_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25333_mean_l3_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25333_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25333_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25334_mean_l3_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25334_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25334_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25335_mean_l3_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25335_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25335_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25336_mean_l3_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25336_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25336_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25337_mean_l3_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25337_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25337_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25338_mean_l3_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25338_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25338_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25339_mean_l3_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25339_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25339_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25340_mean_l3_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25340_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25340_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25341_mean_l3_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25341_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25341_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25342_mean_l3_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25342_Mean-L3-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25342_Mean-L3-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25343_mean_l3_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25343_Mean-L3-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25343_Mean-L3-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25344_mean_icvf_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25344_Mean-ICVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25344_Mean-ICVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25345_mean_icvf_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25345_Mean-ICVF-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25345_Mean-ICVF-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25346_mean_icvf_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25346_Mean-ICVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25346_Mean-ICVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25347_mean_icvf_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25347_Mean-ICVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25347_Mean-ICVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25348_mean_icvf_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25348_Mean-ICVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25348_Mean-ICVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25349_mean_icvf_in_fornix_on_fa_skeleton_2_0 = TensorMap('25349_Mean-ICVF-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25349_Mean-ICVF-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25350_mean_icvf_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25350_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25350_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25351_mean_icvf_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25351_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25351_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25352_mean_icvf_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25352_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25352_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25353_mean_icvf_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25353_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25353_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25354_mean_icvf_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25354_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25354_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25355_mean_icvf_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25355_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25355_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25356_mean_icvf_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25356_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25356_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25357_mean_icvf_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25357_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25357_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25358_mean_icvf_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25358_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25358_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25359_mean_icvf_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25359_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25359_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25360_mean_icvf_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25360_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25360_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25361_mean_icvf_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25361_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25361_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25362_mean_icvf_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25362_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25362_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25363_mean_icvf_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25363_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25363_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25364_mean_icvf_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25364_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25364_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25365_mean_icvf_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25365_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25365_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25366_mean_icvf_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25366_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25366_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25367_mean_icvf_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25367_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25367_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25368_mean_icvf_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25368_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25368_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25369_mean_icvf_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25369_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25369_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25370_mean_icvf_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25370_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25370_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25371_mean_icvf_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25371_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25371_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25372_mean_icvf_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25372_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25372_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25373_mean_icvf_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25373_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25373_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25374_mean_icvf_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25374_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25374_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25375_mean_icvf_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25375_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25375_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25376_mean_icvf_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25376_Mean-ICVF-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25376_Mean-ICVF-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25377_mean_icvf_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25377_Mean-ICVF-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25377_Mean-ICVF-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25378_mean_icvf_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25378_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25378_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25379_mean_icvf_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25379_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25379_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25380_mean_icvf_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25380_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25380_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25381_mean_icvf_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25381_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25381_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25382_mean_icvf_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25382_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25382_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25383_mean_icvf_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25383_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25383_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25384_mean_icvf_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25384_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25384_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25385_mean_icvf_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25385_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25385_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25386_mean_icvf_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25386_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25386_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25387_mean_icvf_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25387_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25387_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25388_mean_icvf_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25388_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25388_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25389_mean_icvf_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25389_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25389_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25390_mean_icvf_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25390_Mean-ICVF-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25390_Mean-ICVF-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25391_mean_icvf_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25391_Mean-ICVF-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25391_Mean-ICVF-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25392_mean_od_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25392_Mean-OD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25392_Mean-OD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25393_mean_od_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25393_Mean-OD-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25393_Mean-OD-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25394_mean_od_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25394_Mean-OD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25394_Mean-OD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25395_mean_od_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25395_Mean-OD-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25395_Mean-OD-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25396_mean_od_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25396_Mean-OD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25396_Mean-OD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25397_mean_od_in_fornix_on_fa_skeleton_2_0 = TensorMap('25397_Mean-OD-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25397_Mean-OD-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25398_mean_od_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25398_Mean-OD-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25398_Mean-OD-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25399_mean_od_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25399_Mean-OD-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25399_Mean-OD-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25400_mean_od_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25400_Mean-OD-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25400_Mean-OD-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25401_mean_od_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25401_Mean-OD-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25401_Mean-OD-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25402_mean_od_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25402_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25402_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25403_mean_od_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25403_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25403_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25404_mean_od_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25404_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25404_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25405_mean_od_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25405_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25405_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25406_mean_od_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25406_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25406_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25407_mean_od_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25407_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25407_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25408_mean_od_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25408_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25408_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25409_mean_od_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25409_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25409_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25410_mean_od_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25410_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25410_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25411_mean_od_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25411_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25411_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25412_mean_od_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25412_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25412_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25413_mean_od_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25413_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25413_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25414_mean_od_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25414_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25414_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25415_mean_od_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25415_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25415_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25416_mean_od_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25416_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25416_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25417_mean_od_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25417_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25417_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25418_mean_od_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25418_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25418_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25419_mean_od_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25419_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25419_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25420_mean_od_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25420_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25420_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25421_mean_od_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25421_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25421_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25422_mean_od_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25422_Mean-OD-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25422_Mean-OD-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25423_mean_od_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25423_Mean-OD-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25423_Mean-OD-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25424_mean_od_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25424_Mean-OD-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25424_Mean-OD-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25425_mean_od_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25425_Mean-OD-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25425_Mean-OD-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25426_mean_od_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25426_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25426_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25427_mean_od_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25427_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25427_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25428_mean_od_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25428_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25428_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25429_mean_od_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25429_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25429_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25430_mean_od_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25430_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25430_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25431_mean_od_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25431_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25431_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25432_mean_od_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25432_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25432_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25433_mean_od_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25433_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25433_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25434_mean_od_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25434_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25434_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25435_mean_od_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25435_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25435_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25436_mean_od_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25436_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25436_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25437_mean_od_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25437_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25437_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25438_mean_od_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25438_Mean-OD-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25438_Mean-OD-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25439_mean_od_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25439_Mean-OD-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25439_Mean-OD-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25440_mean_isovf_in_middle_cerebellar_peduncle_on_fa_skeleton_2_0 = TensorMap('25440_Mean-ISOVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25440_Mean-ISOVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0})
field_id_25441_mean_isovf_in_pontine_crossing_tract_on_fa_skeleton_2_0 = TensorMap('25441_Mean-ISOVF-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25441_Mean-ISOVF-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0})
field_id_25442_mean_isovf_in_genu_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25442_Mean-ISOVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25442_Mean-ISOVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25443_mean_isovf_in_body_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25443_Mean-ISOVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25443_Mean-ISOVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25444_mean_isovf_in_splenium_of_corpus_callosum_on_fa_skeleton_2_0 = TensorMap('25444_Mean-ISOVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25444_Mean-ISOVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0})
field_id_25445_mean_isovf_in_fornix_on_fa_skeleton_2_0 = TensorMap('25445_Mean-ISOVF-in-fornix-on-FA-skeleton_2_0', loss='logcosh', channel_map={'no_25445_Mean-ISOVF-in-fornix-on-FA-skeleton_2_0': 0})
field_id_25446_mean_isovf_in_corticospinal_tract_on_fa_skeleton_right_2_0 = TensorMap('25446_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25446_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0})
field_id_25447_mean_isovf_in_corticospinal_tract_on_fa_skeleton_left_2_0 = TensorMap('25447_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25447_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0})
field_id_25448_mean_isovf_in_medial_lemniscus_on_fa_skeleton_right_2_0 = TensorMap('25448_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25448_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0})
field_id_25449_mean_isovf_in_medial_lemniscus_on_fa_skeleton_left_2_0 = TensorMap('25449_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25449_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0})
field_id_25450_mean_isovf_in_inferior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25450_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25450_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25451_mean_isovf_in_inferior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25451_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25451_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25452_mean_isovf_in_superior_cerebellar_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25452_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25452_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25453_mean_isovf_in_superior_cerebellar_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25453_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25453_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25454_mean_isovf_in_cerebral_peduncle_on_fa_skeleton_right_2_0 = TensorMap('25454_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25454_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0})
field_id_25455_mean_isovf_in_cerebral_peduncle_on_fa_skeleton_left_2_0 = TensorMap('25455_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25455_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0})
field_id_25456_mean_isovf_in_anterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25456_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25456_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25457_mean_isovf_in_anterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25457_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25457_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25458_mean_isovf_in_posterior_limb_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25458_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25458_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25459_mean_isovf_in_posterior_limb_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25459_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25459_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25460_mean_isovf_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_right_2_0 = TensorMap('25460_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25460_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25461_mean_isovf_in_retrolenticular_part_of_internal_capsule_on_fa_skeleton_left_2_0 = TensorMap('25461_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25461_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25462_mean_isovf_in_anterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25462_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25462_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25463_mean_isovf_in_anterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25463_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25463_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25464_mean_isovf_in_superior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25464_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25464_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25465_mean_isovf_in_superior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25465_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25465_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25466_mean_isovf_in_posterior_corona_radiata_on_fa_skeleton_right_2_0 = TensorMap('25466_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25466_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0})
field_id_25467_mean_isovf_in_posterior_corona_radiata_on_fa_skeleton_left_2_0 = TensorMap('25467_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25467_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0})
field_id_25468_mean_isovf_in_posterior_thalamic_radiation_on_fa_skeleton_right_2_0 = TensorMap('25468_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25468_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0})
field_id_25469_mean_isovf_in_posterior_thalamic_radiation_on_fa_skeleton_left_2_0 = TensorMap('25469_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25469_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0})
field_id_25470_mean_isovf_in_sagittal_stratum_on_fa_skeleton_right_2_0 = TensorMap('25470_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25470_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0})
field_id_25471_mean_isovf_in_sagittal_stratum_on_fa_skeleton_left_2_0 = TensorMap('25471_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25471_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0})
field_id_25472_mean_isovf_in_external_capsule_on_fa_skeleton_right_2_0 = TensorMap('25472_Mean-ISOVF-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25472_Mean-ISOVF-in-external-capsule-on-FA-skeleton-right_2_0': 0})
field_id_25473_mean_isovf_in_external_capsule_on_fa_skeleton_left_2_0 = TensorMap('25473_Mean-ISOVF-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25473_Mean-ISOVF-in-external-capsule-on-FA-skeleton-left_2_0': 0})
field_id_25474_mean_isovf_in_cingulum_cingulate_gyrus_on_fa_skeleton_right_2_0 = TensorMap('25474_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25474_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0})
field_id_25475_mean_isovf_in_cingulum_cingulate_gyrus_on_fa_skeleton_left_2_0 = TensorMap('25475_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25475_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0})
field_id_25476_mean_isovf_in_cingulum_hippocampus_on_fa_skeleton_right_2_0 = TensorMap('25476_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25476_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0})
field_id_25477_mean_isovf_in_cingulum_hippocampus_on_fa_skeleton_left_2_0 = TensorMap('25477_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25477_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0})
field_id_25478_mean_isovf_in_fornix_cresstria_terminalis_on_fa_skeleton_right_2_0 = TensorMap('25478_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25478_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0})
field_id_25479_mean_isovf_in_fornix_cresstria_terminalis_on_fa_skeleton_left_2_0 = TensorMap('25479_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25479_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0})
field_id_25480_mean_isovf_in_superior_longitudinal_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25480_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25480_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25481_mean_isovf_in_superior_longitudinal_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25481_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25481_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25482_mean_isovf_in_superior_frontooccipital_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25482_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25482_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25483_mean_isovf_in_superior_frontooccipital_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25483_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25483_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25484_mean_isovf_in_uncinate_fasciculus_on_fa_skeleton_right_2_0 = TensorMap('25484_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25484_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0})
field_id_25485_mean_isovf_in_uncinate_fasciculus_on_fa_skeleton_left_2_0 = TensorMap('25485_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25485_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0})
field_id_25486_mean_isovf_in_tapetum_on_fa_skeleton_right_2_0 = TensorMap('25486_Mean-ISOVF-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', channel_map={'no_25486_Mean-ISOVF-in-tapetum-on-FA-skeleton-right_2_0': 0})
field_id_25487_mean_isovf_in_tapetum_on_fa_skeleton_left_2_0 = TensorMap('25487_Mean-ISOVF-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', channel_map={'no_25487_Mean-ISOVF-in-tapetum-on-FA-skeleton-left_2_0': 0})
field_id_25488_weightedmean_fa_in_tract_acoustic_radiation_left_2_0 = TensorMap('25488_Weightedmean-FA-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25488_Weightedmean-FA-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25489_weightedmean_fa_in_tract_acoustic_radiation_right_2_0 = TensorMap('25489_Weightedmean-FA-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25489_Weightedmean-FA-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25490_weightedmean_fa_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25490_Weightedmean-FA-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25490_Weightedmean-FA-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25491_weightedmean_fa_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25491_Weightedmean-FA-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25491_Weightedmean-FA-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25492_weightedmean_fa_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25492_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25492_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25493_weightedmean_fa_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25493_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25493_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25494_weightedmean_fa_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25494_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25494_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25495_weightedmean_fa_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25495_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25495_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25496_weightedmean_fa_in_tract_corticospinal_tract_left_2_0 = TensorMap('25496_Weightedmean-FA-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25496_Weightedmean-FA-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25497_weightedmean_fa_in_tract_corticospinal_tract_right_2_0 = TensorMap('25497_Weightedmean-FA-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25497_Weightedmean-FA-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25498_weightedmean_fa_in_tract_forceps_major_2_0 = TensorMap('25498_Weightedmean-FA-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25498_Weightedmean-FA-in-tract-forceps-major_2_0': 0})
field_id_25499_weightedmean_fa_in_tract_forceps_minor_2_0 = TensorMap('25499_Weightedmean-FA-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25499_Weightedmean-FA-in-tract-forceps-minor_2_0': 0})
field_id_25500_weightedmean_fa_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25500_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25500_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25501_weightedmean_fa_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25501_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25501_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25502_weightedmean_fa_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25502_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25502_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25503_weightedmean_fa_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25503_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25503_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25504_weightedmean_fa_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25504_Weightedmean-FA-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25504_Weightedmean-FA-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25505_weightedmean_fa_in_tract_medial_lemniscus_left_2_0 = TensorMap('25505_Weightedmean-FA-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25505_Weightedmean-FA-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25506_weightedmean_fa_in_tract_medial_lemniscus_right_2_0 = TensorMap('25506_Weightedmean-FA-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25506_Weightedmean-FA-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25507_weightedmean_fa_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25507_Weightedmean-FA-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25507_Weightedmean-FA-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25508_weightedmean_fa_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25508_Weightedmean-FA-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25508_Weightedmean-FA-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25509_weightedmean_fa_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25509_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25509_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25510_weightedmean_fa_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25510_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25510_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25511_weightedmean_fa_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25511_Weightedmean-FA-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25511_Weightedmean-FA-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25512_weightedmean_fa_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25512_Weightedmean-FA-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25512_Weightedmean-FA-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25513_weightedmean_fa_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25513_Weightedmean-FA-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25513_Weightedmean-FA-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25514_weightedmean_fa_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25514_Weightedmean-FA-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25514_Weightedmean-FA-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25515_weightedmean_md_in_tract_acoustic_radiation_left_2_0 = TensorMap('25515_Weightedmean-MD-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25515_Weightedmean-MD-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25516_weightedmean_md_in_tract_acoustic_radiation_right_2_0 = TensorMap('25516_Weightedmean-MD-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25516_Weightedmean-MD-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25517_weightedmean_md_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25517_Weightedmean-MD-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25517_Weightedmean-MD-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25518_weightedmean_md_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25518_Weightedmean-MD-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25518_Weightedmean-MD-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25519_weightedmean_md_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25519_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25519_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25520_weightedmean_md_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25520_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25520_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25521_weightedmean_md_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25521_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25521_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25522_weightedmean_md_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25522_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25522_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25523_weightedmean_md_in_tract_corticospinal_tract_left_2_0 = TensorMap('25523_Weightedmean-MD-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25523_Weightedmean-MD-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25524_weightedmean_md_in_tract_corticospinal_tract_right_2_0 = TensorMap('25524_Weightedmean-MD-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25524_Weightedmean-MD-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25525_weightedmean_md_in_tract_forceps_major_2_0 = TensorMap('25525_Weightedmean-MD-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25525_Weightedmean-MD-in-tract-forceps-major_2_0': 0})
field_id_25526_weightedmean_md_in_tract_forceps_minor_2_0 = TensorMap('25526_Weightedmean-MD-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25526_Weightedmean-MD-in-tract-forceps-minor_2_0': 0})
field_id_25527_weightedmean_md_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25527_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25527_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25528_weightedmean_md_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25528_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25528_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25529_weightedmean_md_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25529_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25529_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25530_weightedmean_md_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25530_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25530_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25531_weightedmean_md_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25531_Weightedmean-MD-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25531_Weightedmean-MD-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25532_weightedmean_md_in_tract_medial_lemniscus_left_2_0 = TensorMap('25532_Weightedmean-MD-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25532_Weightedmean-MD-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25533_weightedmean_md_in_tract_medial_lemniscus_right_2_0 = TensorMap('25533_Weightedmean-MD-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25533_Weightedmean-MD-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25534_weightedmean_md_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25534_Weightedmean-MD-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25534_Weightedmean-MD-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25535_weightedmean_md_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25535_Weightedmean-MD-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25535_Weightedmean-MD-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25536_weightedmean_md_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25536_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25536_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25537_weightedmean_md_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25537_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25537_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25538_weightedmean_md_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25538_Weightedmean-MD-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25538_Weightedmean-MD-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25539_weightedmean_md_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25539_Weightedmean-MD-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25539_Weightedmean-MD-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25540_weightedmean_md_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25540_Weightedmean-MD-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25540_Weightedmean-MD-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25541_weightedmean_md_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25541_Weightedmean-MD-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25541_Weightedmean-MD-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25542_weightedmean_mo_in_tract_acoustic_radiation_left_2_0 = TensorMap('25542_Weightedmean-MO-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25542_Weightedmean-MO-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25543_weightedmean_mo_in_tract_acoustic_radiation_right_2_0 = TensorMap('25543_Weightedmean-MO-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25543_Weightedmean-MO-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25544_weightedmean_mo_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25544_Weightedmean-MO-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25544_Weightedmean-MO-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25545_weightedmean_mo_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25545_Weightedmean-MO-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25545_Weightedmean-MO-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25546_weightedmean_mo_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25546_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25546_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25547_weightedmean_mo_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25547_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25547_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25548_weightedmean_mo_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25548_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25548_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25549_weightedmean_mo_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25549_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25549_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25550_weightedmean_mo_in_tract_corticospinal_tract_left_2_0 = TensorMap('25550_Weightedmean-MO-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25550_Weightedmean-MO-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25551_weightedmean_mo_in_tract_corticospinal_tract_right_2_0 = TensorMap('25551_Weightedmean-MO-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25551_Weightedmean-MO-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25552_weightedmean_mo_in_tract_forceps_major_2_0 = TensorMap('25552_Weightedmean-MO-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25552_Weightedmean-MO-in-tract-forceps-major_2_0': 0})
field_id_25553_weightedmean_mo_in_tract_forceps_minor_2_0 = TensorMap('25553_Weightedmean-MO-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25553_Weightedmean-MO-in-tract-forceps-minor_2_0': 0})
field_id_25554_weightedmean_mo_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25554_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25554_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25555_weightedmean_mo_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25555_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25555_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25556_weightedmean_mo_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25556_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25556_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25557_weightedmean_mo_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25557_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25557_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25558_weightedmean_mo_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25558_Weightedmean-MO-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25558_Weightedmean-MO-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25559_weightedmean_mo_in_tract_medial_lemniscus_left_2_0 = TensorMap('25559_Weightedmean-MO-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25559_Weightedmean-MO-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25560_weightedmean_mo_in_tract_medial_lemniscus_right_2_0 = TensorMap('25560_Weightedmean-MO-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25560_Weightedmean-MO-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25561_weightedmean_mo_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25561_Weightedmean-MO-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25561_Weightedmean-MO-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25562_weightedmean_mo_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25562_Weightedmean-MO-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25562_Weightedmean-MO-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25563_weightedmean_mo_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25563_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25563_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25564_weightedmean_mo_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25564_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25564_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25565_weightedmean_mo_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25565_Weightedmean-MO-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25565_Weightedmean-MO-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25566_weightedmean_mo_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25566_Weightedmean-MO-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25566_Weightedmean-MO-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25567_weightedmean_mo_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25567_Weightedmean-MO-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25567_Weightedmean-MO-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25568_weightedmean_mo_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25568_Weightedmean-MO-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25568_Weightedmean-MO-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25569_weightedmean_l1_in_tract_acoustic_radiation_left_2_0 = TensorMap('25569_Weightedmean-L1-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25569_Weightedmean-L1-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25570_weightedmean_l1_in_tract_acoustic_radiation_right_2_0 = TensorMap('25570_Weightedmean-L1-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25570_Weightedmean-L1-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25571_weightedmean_l1_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25571_Weightedmean-L1-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25571_Weightedmean-L1-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25572_weightedmean_l1_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25572_Weightedmean-L1-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25572_Weightedmean-L1-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25573_weightedmean_l1_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25573_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25573_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25574_weightedmean_l1_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25574_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25574_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25575_weightedmean_l1_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25575_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25575_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25576_weightedmean_l1_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25576_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25576_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25577_weightedmean_l1_in_tract_corticospinal_tract_left_2_0 = TensorMap('25577_Weightedmean-L1-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25577_Weightedmean-L1-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25578_weightedmean_l1_in_tract_corticospinal_tract_right_2_0 = TensorMap('25578_Weightedmean-L1-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25578_Weightedmean-L1-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25579_weightedmean_l1_in_tract_forceps_major_2_0 = TensorMap('25579_Weightedmean-L1-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25579_Weightedmean-L1-in-tract-forceps-major_2_0': 0})
field_id_25580_weightedmean_l1_in_tract_forceps_minor_2_0 = TensorMap('25580_Weightedmean-L1-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25580_Weightedmean-L1-in-tract-forceps-minor_2_0': 0})
field_id_25581_weightedmean_l1_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25581_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25581_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25582_weightedmean_l1_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25582_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25582_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25583_weightedmean_l1_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25583_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25583_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25584_weightedmean_l1_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25584_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25584_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25585_weightedmean_l1_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25585_Weightedmean-L1-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25585_Weightedmean-L1-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25586_weightedmean_l1_in_tract_medial_lemniscus_left_2_0 = TensorMap('25586_Weightedmean-L1-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25586_Weightedmean-L1-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25587_weightedmean_l1_in_tract_medial_lemniscus_right_2_0 = TensorMap('25587_Weightedmean-L1-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25587_Weightedmean-L1-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25588_weightedmean_l1_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25588_Weightedmean-L1-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25588_Weightedmean-L1-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25589_weightedmean_l1_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25589_Weightedmean-L1-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25589_Weightedmean-L1-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25590_weightedmean_l1_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25590_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25590_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25591_weightedmean_l1_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25591_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25591_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25592_weightedmean_l1_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25592_Weightedmean-L1-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25592_Weightedmean-L1-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25593_weightedmean_l1_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25593_Weightedmean-L1-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25593_Weightedmean-L1-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25594_weightedmean_l1_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25594_Weightedmean-L1-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25594_Weightedmean-L1-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25595_weightedmean_l1_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25595_Weightedmean-L1-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25595_Weightedmean-L1-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25596_weightedmean_l2_in_tract_acoustic_radiation_left_2_0 = TensorMap('25596_Weightedmean-L2-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25596_Weightedmean-L2-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25597_weightedmean_l2_in_tract_acoustic_radiation_right_2_0 = TensorMap('25597_Weightedmean-L2-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25597_Weightedmean-L2-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25598_weightedmean_l2_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25598_Weightedmean-L2-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25598_Weightedmean-L2-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25599_weightedmean_l2_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25599_Weightedmean-L2-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25599_Weightedmean-L2-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25600_weightedmean_l2_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25600_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25600_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25601_weightedmean_l2_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25601_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25601_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25602_weightedmean_l2_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25602_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25602_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25603_weightedmean_l2_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25603_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25603_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25604_weightedmean_l2_in_tract_corticospinal_tract_left_2_0 = TensorMap('25604_Weightedmean-L2-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25604_Weightedmean-L2-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25605_weightedmean_l2_in_tract_corticospinal_tract_right_2_0 = TensorMap('25605_Weightedmean-L2-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25605_Weightedmean-L2-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25606_weightedmean_l2_in_tract_forceps_major_2_0 = TensorMap('25606_Weightedmean-L2-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25606_Weightedmean-L2-in-tract-forceps-major_2_0': 0})
field_id_25607_weightedmean_l2_in_tract_forceps_minor_2_0 = TensorMap('25607_Weightedmean-L2-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25607_Weightedmean-L2-in-tract-forceps-minor_2_0': 0})
field_id_25608_weightedmean_l2_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25608_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25608_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25609_weightedmean_l2_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25609_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25609_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25610_weightedmean_l2_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25610_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25610_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25611_weightedmean_l2_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25611_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25611_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25612_weightedmean_l2_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25612_Weightedmean-L2-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25612_Weightedmean-L2-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25613_weightedmean_l2_in_tract_medial_lemniscus_left_2_0 = TensorMap('25613_Weightedmean-L2-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25613_Weightedmean-L2-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25614_weightedmean_l2_in_tract_medial_lemniscus_right_2_0 = TensorMap('25614_Weightedmean-L2-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25614_Weightedmean-L2-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25615_weightedmean_l2_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25615_Weightedmean-L2-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25615_Weightedmean-L2-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25616_weightedmean_l2_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25616_Weightedmean-L2-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25616_Weightedmean-L2-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25617_weightedmean_l2_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25617_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25617_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25618_weightedmean_l2_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25618_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25618_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25619_weightedmean_l2_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25619_Weightedmean-L2-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25619_Weightedmean-L2-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25620_weightedmean_l2_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25620_Weightedmean-L2-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25620_Weightedmean-L2-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25621_weightedmean_l2_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25621_Weightedmean-L2-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25621_Weightedmean-L2-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25622_weightedmean_l2_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25622_Weightedmean-L2-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25622_Weightedmean-L2-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25623_weightedmean_l3_in_tract_acoustic_radiation_left_2_0 = TensorMap('25623_Weightedmean-L3-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25623_Weightedmean-L3-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25624_weightedmean_l3_in_tract_acoustic_radiation_right_2_0 = TensorMap('25624_Weightedmean-L3-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25624_Weightedmean-L3-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25625_weightedmean_l3_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25625_Weightedmean-L3-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25625_Weightedmean-L3-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25626_weightedmean_l3_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25626_Weightedmean-L3-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25626_Weightedmean-L3-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25627_weightedmean_l3_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25627_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25627_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25628_weightedmean_l3_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25628_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25628_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25629_weightedmean_l3_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25629_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25629_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25630_weightedmean_l3_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25630_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25630_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25631_weightedmean_l3_in_tract_corticospinal_tract_left_2_0 = TensorMap('25631_Weightedmean-L3-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25631_Weightedmean-L3-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25632_weightedmean_l3_in_tract_corticospinal_tract_right_2_0 = TensorMap('25632_Weightedmean-L3-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25632_Weightedmean-L3-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25633_weightedmean_l3_in_tract_forceps_major_2_0 = TensorMap('25633_Weightedmean-L3-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25633_Weightedmean-L3-in-tract-forceps-major_2_0': 0})
field_id_25634_weightedmean_l3_in_tract_forceps_minor_2_0 = TensorMap('25634_Weightedmean-L3-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25634_Weightedmean-L3-in-tract-forceps-minor_2_0': 0})
field_id_25635_weightedmean_l3_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25635_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25635_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25636_weightedmean_l3_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25636_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25636_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25637_weightedmean_l3_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25637_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25637_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25638_weightedmean_l3_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25638_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25638_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25639_weightedmean_l3_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25639_Weightedmean-L3-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25639_Weightedmean-L3-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25640_weightedmean_l3_in_tract_medial_lemniscus_left_2_0 = TensorMap('25640_Weightedmean-L3-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25640_Weightedmean-L3-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25641_weightedmean_l3_in_tract_medial_lemniscus_right_2_0 = TensorMap('25641_Weightedmean-L3-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25641_Weightedmean-L3-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25642_weightedmean_l3_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25642_Weightedmean-L3-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25642_Weightedmean-L3-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25643_weightedmean_l3_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25643_Weightedmean-L3-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25643_Weightedmean-L3-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25644_weightedmean_l3_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25644_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25644_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25645_weightedmean_l3_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25645_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25645_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25646_weightedmean_l3_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25646_Weightedmean-L3-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25646_Weightedmean-L3-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25647_weightedmean_l3_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25647_Weightedmean-L3-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25647_Weightedmean-L3-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25648_weightedmean_l3_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25648_Weightedmean-L3-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25648_Weightedmean-L3-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25649_weightedmean_l3_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25649_Weightedmean-L3-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25649_Weightedmean-L3-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25650_weightedmean_icvf_in_tract_acoustic_radiation_left_2_0 = TensorMap('25650_Weightedmean-ICVF-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25650_Weightedmean-ICVF-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25651_weightedmean_icvf_in_tract_acoustic_radiation_right_2_0 = TensorMap('25651_Weightedmean-ICVF-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25651_Weightedmean-ICVF-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25652_weightedmean_icvf_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25652_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25652_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25653_weightedmean_icvf_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25653_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25653_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25654_weightedmean_icvf_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25654_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25654_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25655_weightedmean_icvf_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25655_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25655_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25656_weightedmean_icvf_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25656_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25656_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25657_weightedmean_icvf_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25657_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25657_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25658_weightedmean_icvf_in_tract_corticospinal_tract_left_2_0 = TensorMap('25658_Weightedmean-ICVF-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25658_Weightedmean-ICVF-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25659_weightedmean_icvf_in_tract_corticospinal_tract_right_2_0 = TensorMap('25659_Weightedmean-ICVF-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25659_Weightedmean-ICVF-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25660_weightedmean_icvf_in_tract_forceps_major_2_0 = TensorMap('25660_Weightedmean-ICVF-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25660_Weightedmean-ICVF-in-tract-forceps-major_2_0': 0})
field_id_25661_weightedmean_icvf_in_tract_forceps_minor_2_0 = TensorMap('25661_Weightedmean-ICVF-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25661_Weightedmean-ICVF-in-tract-forceps-minor_2_0': 0})
field_id_25662_weightedmean_icvf_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25662_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25662_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25663_weightedmean_icvf_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25663_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25663_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25664_weightedmean_icvf_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25664_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25664_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25665_weightedmean_icvf_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25665_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25665_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25666_weightedmean_icvf_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25666_Weightedmean-ICVF-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25666_Weightedmean-ICVF-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25667_weightedmean_icvf_in_tract_medial_lemniscus_left_2_0 = TensorMap('25667_Weightedmean-ICVF-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25667_Weightedmean-ICVF-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25668_weightedmean_icvf_in_tract_medial_lemniscus_right_2_0 = TensorMap('25668_Weightedmean-ICVF-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25668_Weightedmean-ICVF-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25669_weightedmean_icvf_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25669_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25669_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25670_weightedmean_icvf_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25670_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25670_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25671_weightedmean_icvf_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25671_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25671_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25672_weightedmean_icvf_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25672_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25672_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25673_weightedmean_icvf_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25673_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25673_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25674_weightedmean_icvf_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25674_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25674_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25675_weightedmean_icvf_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25675_Weightedmean-ICVF-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25675_Weightedmean-ICVF-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25676_weightedmean_icvf_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25676_Weightedmean-ICVF-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25676_Weightedmean-ICVF-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25677_weightedmean_od_in_tract_acoustic_radiation_left_2_0 = TensorMap('25677_Weightedmean-OD-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25677_Weightedmean-OD-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25678_weightedmean_od_in_tract_acoustic_radiation_right_2_0 = TensorMap('25678_Weightedmean-OD-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25678_Weightedmean-OD-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25679_weightedmean_od_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25679_Weightedmean-OD-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25679_Weightedmean-OD-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25680_weightedmean_od_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25680_Weightedmean-OD-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25680_Weightedmean-OD-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25681_weightedmean_od_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25681_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25681_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25682_weightedmean_od_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25682_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25682_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25683_weightedmean_od_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25683_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25683_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25684_weightedmean_od_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25684_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25684_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25685_weightedmean_od_in_tract_corticospinal_tract_left_2_0 = TensorMap('25685_Weightedmean-OD-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25685_Weightedmean-OD-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25686_weightedmean_od_in_tract_corticospinal_tract_right_2_0 = TensorMap('25686_Weightedmean-OD-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25686_Weightedmean-OD-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25687_weightedmean_od_in_tract_forceps_major_2_0 = TensorMap('25687_Weightedmean-OD-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25687_Weightedmean-OD-in-tract-forceps-major_2_0': 0})
field_id_25688_weightedmean_od_in_tract_forceps_minor_2_0 = TensorMap('25688_Weightedmean-OD-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25688_Weightedmean-OD-in-tract-forceps-minor_2_0': 0})
field_id_25689_weightedmean_od_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25689_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25689_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25690_weightedmean_od_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25690_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25690_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25691_weightedmean_od_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25691_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25691_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25692_weightedmean_od_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25692_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25692_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25693_weightedmean_od_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25693_Weightedmean-OD-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25693_Weightedmean-OD-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25694_weightedmean_od_in_tract_medial_lemniscus_left_2_0 = TensorMap('25694_Weightedmean-OD-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25694_Weightedmean-OD-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25695_weightedmean_od_in_tract_medial_lemniscus_right_2_0 = TensorMap('25695_Weightedmean-OD-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25695_Weightedmean-OD-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25696_weightedmean_od_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25696_Weightedmean-OD-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25696_Weightedmean-OD-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25697_weightedmean_od_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25697_Weightedmean-OD-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25697_Weightedmean-OD-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25698_weightedmean_od_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25698_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25698_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25699_weightedmean_od_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25699_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25699_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25700_weightedmean_od_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25700_Weightedmean-OD-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25700_Weightedmean-OD-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25701_weightedmean_od_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25701_Weightedmean-OD-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25701_Weightedmean-OD-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25702_weightedmean_od_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25702_Weightedmean-OD-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25702_Weightedmean-OD-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25703_weightedmean_od_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25703_Weightedmean-OD-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25703_Weightedmean-OD-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25704_weightedmean_isovf_in_tract_acoustic_radiation_left_2_0 = TensorMap('25704_Weightedmean-ISOVF-in-tract-acoustic-radiation-left_2_0', loss='logcosh', channel_map={'no_25704_Weightedmean-ISOVF-in-tract-acoustic-radiation-left_2_0': 0})
field_id_25705_weightedmean_isovf_in_tract_acoustic_radiation_right_2_0 = TensorMap('25705_Weightedmean-ISOVF-in-tract-acoustic-radiation-right_2_0', loss='logcosh', channel_map={'no_25705_Weightedmean-ISOVF-in-tract-acoustic-radiation-right_2_0': 0})
field_id_25706_weightedmean_isovf_in_tract_anterior_thalamic_radiation_left_2_0 = TensorMap('25706_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25706_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-left_2_0': 0})
field_id_25707_weightedmean_isovf_in_tract_anterior_thalamic_radiation_right_2_0 = TensorMap('25707_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25707_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-right_2_0': 0})
field_id_25708_weightedmean_isovf_in_tract_cingulate_gyrus_part_of_cingulum_left_2_0 = TensorMap('25708_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25708_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0})
field_id_25709_weightedmean_isovf_in_tract_cingulate_gyrus_part_of_cingulum_right_2_0 = TensorMap('25709_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25709_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0})
field_id_25710_weightedmean_isovf_in_tract_parahippocampal_part_of_cingulum_left_2_0 = TensorMap('25710_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', channel_map={'no_25710_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0})
field_id_25711_weightedmean_isovf_in_tract_parahippocampal_part_of_cingulum_right_2_0 = TensorMap('25711_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', channel_map={'no_25711_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0})
field_id_25712_weightedmean_isovf_in_tract_corticospinal_tract_left_2_0 = TensorMap('25712_Weightedmean-ISOVF-in-tract-corticospinal-tract-left_2_0', loss='logcosh', channel_map={'no_25712_Weightedmean-ISOVF-in-tract-corticospinal-tract-left_2_0': 0})
field_id_25713_weightedmean_isovf_in_tract_corticospinal_tract_right_2_0 = TensorMap('25713_Weightedmean-ISOVF-in-tract-corticospinal-tract-right_2_0', loss='logcosh', channel_map={'no_25713_Weightedmean-ISOVF-in-tract-corticospinal-tract-right_2_0': 0})
field_id_25714_weightedmean_isovf_in_tract_forceps_major_2_0 = TensorMap('25714_Weightedmean-ISOVF-in-tract-forceps-major_2_0', loss='logcosh', channel_map={'no_25714_Weightedmean-ISOVF-in-tract-forceps-major_2_0': 0})
field_id_25715_weightedmean_isovf_in_tract_forceps_minor_2_0 = TensorMap('25715_Weightedmean-ISOVF-in-tract-forceps-minor_2_0', loss='logcosh', channel_map={'no_25715_Weightedmean-ISOVF-in-tract-forceps-minor_2_0': 0})
field_id_25716_weightedmean_isovf_in_tract_inferior_frontooccipital_fasciculus_left_2_0 = TensorMap('25716_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25716_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0})
field_id_25717_weightedmean_isovf_in_tract_inferior_frontooccipital_fasciculus_right_2_0 = TensorMap('25717_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25717_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0})
field_id_25718_weightedmean_isovf_in_tract_inferior_longitudinal_fasciculus_left_2_0 = TensorMap('25718_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25718_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0})
field_id_25719_weightedmean_isovf_in_tract_inferior_longitudinal_fasciculus_right_2_0 = TensorMap('25719_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25719_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0})
field_id_25720_weightedmean_isovf_in_tract_middle_cerebellar_peduncle_2_0 = TensorMap('25720_Weightedmean-ISOVF-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', channel_map={'no_25720_Weightedmean-ISOVF-in-tract-middle-cerebellar-peduncle_2_0': 0})
field_id_25721_weightedmean_isovf_in_tract_medial_lemniscus_left_2_0 = TensorMap('25721_Weightedmean-ISOVF-in-tract-medial-lemniscus-left_2_0', loss='logcosh', channel_map={'no_25721_Weightedmean-ISOVF-in-tract-medial-lemniscus-left_2_0': 0})
field_id_25722_weightedmean_isovf_in_tract_medial_lemniscus_right_2_0 = TensorMap('25722_Weightedmean-ISOVF-in-tract-medial-lemniscus-right_2_0', loss='logcosh', channel_map={'no_25722_Weightedmean-ISOVF-in-tract-medial-lemniscus-right_2_0': 0})
field_id_25723_weightedmean_isovf_in_tract_posterior_thalamic_radiation_left_2_0 = TensorMap('25723_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25723_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-left_2_0': 0})
field_id_25724_weightedmean_isovf_in_tract_posterior_thalamic_radiation_right_2_0 = TensorMap('25724_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25724_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-right_2_0': 0})
field_id_25725_weightedmean_isovf_in_tract_superior_longitudinal_fasciculus_left_2_0 = TensorMap('25725_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25725_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-left_2_0': 0})
field_id_25726_weightedmean_isovf_in_tract_superior_longitudinal_fasciculus_right_2_0 = TensorMap('25726_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25726_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-right_2_0': 0})
field_id_25727_weightedmean_isovf_in_tract_superior_thalamic_radiation_left_2_0 = TensorMap('25727_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', channel_map={'no_25727_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-left_2_0': 0})
field_id_25728_weightedmean_isovf_in_tract_superior_thalamic_radiation_right_2_0 = TensorMap('25728_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', channel_map={'no_25728_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-right_2_0': 0})
field_id_25729_weightedmean_isovf_in_tract_uncinate_fasciculus_left_2_0 = TensorMap('25729_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', channel_map={'no_25729_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-left_2_0': 0})
field_id_25730_weightedmean_isovf_in_tract_uncinate_fasciculus_right_2_0 = TensorMap('25730_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', channel_map={'no_25730_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-right_2_0': 0})
field_id_25731_discrepancy_between_t1_brain_image_and_standardspace_brain_template_linearlyaligned_2_0 = TensorMap('25731_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-linearlyaligned_2_0', loss='logcosh', channel_map={'no_25731_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-linearlyaligned_2_0': 0})
field_id_25732_discrepancy_between_t1_brain_image_and_standardspace_brain_template_nonlinearlyaligned_2_0 = TensorMap('25732_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-nonlinearlyaligned_2_0', loss='logcosh', channel_map={'no_25732_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-nonlinearlyaligned_2_0': 0})
field_id_25734_inverted_signaltonoise_ratio_in_t1_2_0 = TensorMap('25734_Inverted-signaltonoise-ratio-in-T1_2_0', loss='logcosh', channel_map={'no_25734_Inverted-signaltonoise-ratio-in-T1_2_0': 0})
field_id_25735_inverted_contrasttonoise_ratio_in_t1_2_0 = TensorMap('25735_Inverted-contrasttonoise-ratio-in-T1_2_0', loss='logcosh', channel_map={'no_25735_Inverted-contrasttonoise-ratio-in-T1_2_0': 0})
field_id_25736_discrepancy_between_t2_flair_brain_image_and_t1_brain_image_2_0 = TensorMap('25736_Discrepancy-between-T2-FLAIR-brain-image-and-T1-brain-image_2_0', loss='logcosh', channel_map={'no_25736_Discrepancy-between-T2-FLAIR-brain-image-and-T1-brain-image_2_0': 0})
field_id_25737_discrepancy_between_dmri_brain_image_and_t1_brain_image_2_0 = TensorMap('25737_Discrepancy-between-dMRI-brain-image-and-T1-brain-image_2_0', loss='logcosh', channel_map={'no_25737_Discrepancy-between-dMRI-brain-image-and-T1-brain-image_2_0': 0})
field_id_25738_discrepancy_between_swi_brain_image_and_t1_brain_image_2_0 = TensorMap('25738_Discrepancy-between-SWI-brain-image-and-T1-brain-image_2_0', loss='logcosh', channel_map={'no_25738_Discrepancy-between-SWI-brain-image-and-T1-brain-image_2_0': 0})
field_id_25739_discrepancy_between_rfmri_brain_image_and_t1_brain_image_2_0 = TensorMap('25739_Discrepancy-between-rfMRI-brain-image-and-T1-brain-image_2_0', loss='logcosh', channel_map={'no_25739_Discrepancy-between-rfMRI-brain-image-and-T1-brain-image_2_0': 0})
field_id_25781_total_volume_of_white_matter_hyperintensities_from_t1_and_t2flair_images_2_0 = TensorMap('25781_Total-volume-of-white-matter-hyperintensities-from-T1-and-T2FLAIR-images_2_0', loss='logcosh', channel_map={'no_25781_Total-volume-of-white-matter-hyperintensities-from-T1-and-T2FLAIR-images_2_0': 0})
field_id_25782_volume_of_grey_matter_in_frontal_pole_left_2_0 = TensorMap('25782_Volume-of-grey-matter-in-Frontal-Pole-left_2_0', loss='logcosh', channel_map={'no_25782_Volume-of-grey-matter-in-Frontal-Pole-left_2_0': 0})
field_id_25783_volume_of_grey_matter_in_frontal_pole_right_2_0 = TensorMap('25783_Volume-of-grey-matter-in-Frontal-Pole-right_2_0', loss='logcosh', channel_map={'no_25783_Volume-of-grey-matter-in-Frontal-Pole-right_2_0': 0})
field_id_25784_volume_of_grey_matter_in_insular_cortex_left_2_0 = TensorMap('25784_Volume-of-grey-matter-in-Insular-Cortex-left_2_0', loss='logcosh', channel_map={'no_25784_Volume-of-grey-matter-in-Insular-Cortex-left_2_0': 0})
field_id_25785_volume_of_grey_matter_in_insular_cortex_right_2_0 = TensorMap('25785_Volume-of-grey-matter-in-Insular-Cortex-right_2_0', loss='logcosh', channel_map={'no_25785_Volume-of-grey-matter-in-Insular-Cortex-right_2_0': 0})
field_id_25786_volume_of_grey_matter_in_superior_frontal_gyrus_left_2_0 = TensorMap('25786_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25786_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-left_2_0': 0})
field_id_25787_volume_of_grey_matter_in_superior_frontal_gyrus_right_2_0 = TensorMap('25787_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25787_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-right_2_0': 0})
field_id_25788_volume_of_grey_matter_in_middle_frontal_gyrus_left_2_0 = TensorMap('25788_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25788_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-left_2_0': 0})
field_id_25789_volume_of_grey_matter_in_middle_frontal_gyrus_right_2_0 = TensorMap('25789_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25789_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-right_2_0': 0})
field_id_25790_volume_of_grey_matter_in_inferior_frontal_gyrus_pars_triangularis_left_2_0 = TensorMap('25790_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-left_2_0', loss='logcosh', channel_map={'no_25790_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-left_2_0': 0})
field_id_25791_volume_of_grey_matter_in_inferior_frontal_gyrus_pars_triangularis_right_2_0 = TensorMap('25791_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-right_2_0', loss='logcosh', channel_map={'no_25791_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-right_2_0': 0})
field_id_25792_volume_of_grey_matter_in_inferior_frontal_gyrus_pars_opercularis_left_2_0 = TensorMap('25792_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-left_2_0', loss='logcosh', channel_map={'no_25792_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-left_2_0': 0})
field_id_25793_volume_of_grey_matter_in_inferior_frontal_gyrus_pars_opercularis_right_2_0 = TensorMap('25793_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-right_2_0', loss='logcosh', channel_map={'no_25793_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-right_2_0': 0})
field_id_25794_volume_of_grey_matter_in_precentral_gyrus_left_2_0 = TensorMap('25794_Volume-of-grey-matter-in-Precentral-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25794_Volume-of-grey-matter-in-Precentral-Gyrus-left_2_0': 0})
field_id_25795_volume_of_grey_matter_in_precentral_gyrus_right_2_0 = TensorMap('25795_Volume-of-grey-matter-in-Precentral-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25795_Volume-of-grey-matter-in-Precentral-Gyrus-right_2_0': 0})
field_id_25796_volume_of_grey_matter_in_temporal_pole_left_2_0 = TensorMap('25796_Volume-of-grey-matter-in-Temporal-Pole-left_2_0', loss='logcosh', channel_map={'no_25796_Volume-of-grey-matter-in-Temporal-Pole-left_2_0': 0})
field_id_25797_volume_of_grey_matter_in_temporal_pole_right_2_0 = TensorMap('25797_Volume-of-grey-matter-in-Temporal-Pole-right_2_0', loss='logcosh', channel_map={'no_25797_Volume-of-grey-matter-in-Temporal-Pole-right_2_0': 0})
field_id_25798_volume_of_grey_matter_in_superior_temporal_gyrus_anterior_division_left_2_0 = TensorMap('25798_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25798_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-left_2_0': 0})
field_id_25799_volume_of_grey_matter_in_superior_temporal_gyrus_anterior_division_right_2_0 = TensorMap('25799_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25799_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-right_2_0': 0})
field_id_25800_volume_of_grey_matter_in_superior_temporal_gyrus_posterior_division_left_2_0 = TensorMap('25800_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25800_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-left_2_0': 0})
field_id_25801_volume_of_grey_matter_in_superior_temporal_gyrus_posterior_division_right_2_0 = TensorMap('25801_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25801_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-right_2_0': 0})
field_id_25802_volume_of_grey_matter_in_middle_temporal_gyrus_anterior_division_left_2_0 = TensorMap('25802_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25802_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-left_2_0': 0})
field_id_25803_volume_of_grey_matter_in_middle_temporal_gyrus_anterior_division_right_2_0 = TensorMap('25803_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25803_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-right_2_0': 0})
field_id_25804_volume_of_grey_matter_in_middle_temporal_gyrus_posterior_division_left_2_0 = TensorMap('25804_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25804_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-left_2_0': 0})
field_id_25805_volume_of_grey_matter_in_middle_temporal_gyrus_posterior_division_right_2_0 = TensorMap('25805_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25805_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-right_2_0': 0})
field_id_25806_volume_of_grey_matter_in_middle_temporal_gyrus_temporooccipital_part_left_2_0 = TensorMap('25806_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-left_2_0', loss='logcosh', channel_map={'no_25806_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-left_2_0': 0})
field_id_25807_volume_of_grey_matter_in_middle_temporal_gyrus_temporooccipital_part_right_2_0 = TensorMap('25807_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-right_2_0', loss='logcosh', channel_map={'no_25807_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-right_2_0': 0})
field_id_25808_volume_of_grey_matter_in_inferior_temporal_gyrus_anterior_division_left_2_0 = TensorMap('25808_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25808_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-left_2_0': 0})
field_id_25809_volume_of_grey_matter_in_inferior_temporal_gyrus_anterior_division_right_2_0 = TensorMap('25809_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25809_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-right_2_0': 0})
field_id_25810_volume_of_grey_matter_in_inferior_temporal_gyrus_posterior_division_left_2_0 = TensorMap('25810_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25810_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-left_2_0': 0})
field_id_25811_volume_of_grey_matter_in_inferior_temporal_gyrus_posterior_division_right_2_0 = TensorMap('25811_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25811_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-right_2_0': 0})
field_id_25812_volume_of_grey_matter_in_inferior_temporal_gyrus_temporooccipital_part_left_2_0 = TensorMap('25812_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-left_2_0', loss='logcosh', channel_map={'no_25812_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-left_2_0': 0})
field_id_25813_volume_of_grey_matter_in_inferior_temporal_gyrus_temporooccipital_part_right_2_0 = TensorMap('25813_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-right_2_0', loss='logcosh', channel_map={'no_25813_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-right_2_0': 0})
field_id_25814_volume_of_grey_matter_in_postcentral_gyrus_left_2_0 = TensorMap('25814_Volume-of-grey-matter-in-Postcentral-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25814_Volume-of-grey-matter-in-Postcentral-Gyrus-left_2_0': 0})
field_id_25815_volume_of_grey_matter_in_postcentral_gyrus_right_2_0 = TensorMap('25815_Volume-of-grey-matter-in-Postcentral-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25815_Volume-of-grey-matter-in-Postcentral-Gyrus-right_2_0': 0})
field_id_25816_volume_of_grey_matter_in_superior_parietal_lobule_left_2_0 = TensorMap('25816_Volume-of-grey-matter-in-Superior-Parietal-Lobule-left_2_0', loss='logcosh', channel_map={'no_25816_Volume-of-grey-matter-in-Superior-Parietal-Lobule-left_2_0': 0})
field_id_25817_volume_of_grey_matter_in_superior_parietal_lobule_right_2_0 = TensorMap('25817_Volume-of-grey-matter-in-Superior-Parietal-Lobule-right_2_0', loss='logcosh', channel_map={'no_25817_Volume-of-grey-matter-in-Superior-Parietal-Lobule-right_2_0': 0})
field_id_25818_volume_of_grey_matter_in_supramarginal_gyrus_anterior_division_left_2_0 = TensorMap('25818_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25818_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-left_2_0': 0})
field_id_25819_volume_of_grey_matter_in_supramarginal_gyrus_anterior_division_right_2_0 = TensorMap('25819_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25819_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-right_2_0': 0})
field_id_25820_volume_of_grey_matter_in_supramarginal_gyrus_posterior_division_left_2_0 = TensorMap('25820_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25820_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-left_2_0': 0})
field_id_25821_volume_of_grey_matter_in_supramarginal_gyrus_posterior_division_right_2_0 = TensorMap('25821_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25821_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-right_2_0': 0})
field_id_25822_volume_of_grey_matter_in_angular_gyrus_left_2_0 = TensorMap('25822_Volume-of-grey-matter-in-Angular-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25822_Volume-of-grey-matter-in-Angular-Gyrus-left_2_0': 0})
field_id_25823_volume_of_grey_matter_in_angular_gyrus_right_2_0 = TensorMap('25823_Volume-of-grey-matter-in-Angular-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25823_Volume-of-grey-matter-in-Angular-Gyrus-right_2_0': 0})
field_id_25824_volume_of_grey_matter_in_lateral_occipital_cortex_superior_division_left_2_0 = TensorMap('25824_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-left_2_0', loss='logcosh', channel_map={'no_25824_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-left_2_0': 0})
field_id_25825_volume_of_grey_matter_in_lateral_occipital_cortex_superior_division_right_2_0 = TensorMap('25825_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-right_2_0', loss='logcosh', channel_map={'no_25825_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-right_2_0': 0})
field_id_25826_volume_of_grey_matter_in_lateral_occipital_cortex_inferior_division_left_2_0 = TensorMap('25826_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-left_2_0', loss='logcosh', channel_map={'no_25826_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-left_2_0': 0})
field_id_25827_volume_of_grey_matter_in_lateral_occipital_cortex_inferior_division_right_2_0 = TensorMap('25827_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-right_2_0', loss='logcosh', channel_map={'no_25827_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-right_2_0': 0})
field_id_25828_volume_of_grey_matter_in_intracalcarine_cortex_left_2_0 = TensorMap('25828_Volume-of-grey-matter-in-Intracalcarine-Cortex-left_2_0', loss='logcosh', channel_map={'no_25828_Volume-of-grey-matter-in-Intracalcarine-Cortex-left_2_0': 0})
field_id_25829_volume_of_grey_matter_in_intracalcarine_cortex_right_2_0 = TensorMap('25829_Volume-of-grey-matter-in-Intracalcarine-Cortex-right_2_0', loss='logcosh', channel_map={'no_25829_Volume-of-grey-matter-in-Intracalcarine-Cortex-right_2_0': 0})
field_id_25830_volume_of_grey_matter_in_frontal_medial_cortex_left_2_0 = TensorMap('25830_Volume-of-grey-matter-in-Frontal-Medial-Cortex-left_2_0', loss='logcosh', channel_map={'no_25830_Volume-of-grey-matter-in-Frontal-Medial-Cortex-left_2_0': 0})
field_id_25831_volume_of_grey_matter_in_frontal_medial_cortex_right_2_0 = TensorMap('25831_Volume-of-grey-matter-in-Frontal-Medial-Cortex-right_2_0', loss='logcosh', channel_map={'no_25831_Volume-of-grey-matter-in-Frontal-Medial-Cortex-right_2_0': 0})
field_id_25832_volume_of_grey_matter_in_juxtapositional_lobule_cortex_formerly_supplementary_motor_cortex_left_2_0 = TensorMap('25832_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-left_2_0', loss='logcosh', channel_map={'no_25832_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-left_2_0': 0})
field_id_25833_volume_of_grey_matter_in_juxtapositional_lobule_cortex_formerly_supplementary_motor_cortex_right_2_0 = TensorMap('25833_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-right_2_0', loss='logcosh', channel_map={'no_25833_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-right_2_0': 0})
field_id_25834_volume_of_grey_matter_in_subcallosal_cortex_left_2_0 = TensorMap('25834_Volume-of-grey-matter-in-Subcallosal-Cortex-left_2_0', loss='logcosh', channel_map={'no_25834_Volume-of-grey-matter-in-Subcallosal-Cortex-left_2_0': 0})
field_id_25835_volume_of_grey_matter_in_subcallosal_cortex_right_2_0 = TensorMap('25835_Volume-of-grey-matter-in-Subcallosal-Cortex-right_2_0', loss='logcosh', channel_map={'no_25835_Volume-of-grey-matter-in-Subcallosal-Cortex-right_2_0': 0})
field_id_25836_volume_of_grey_matter_in_paracingulate_gyrus_left_2_0 = TensorMap('25836_Volume-of-grey-matter-in-Paracingulate-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25836_Volume-of-grey-matter-in-Paracingulate-Gyrus-left_2_0': 0})
field_id_25837_volume_of_grey_matter_in_paracingulate_gyrus_right_2_0 = TensorMap('25837_Volume-of-grey-matter-in-Paracingulate-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25837_Volume-of-grey-matter-in-Paracingulate-Gyrus-right_2_0': 0})
field_id_25838_volume_of_grey_matter_in_cingulate_gyrus_anterior_division_left_2_0 = TensorMap('25838_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25838_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-left_2_0': 0})
field_id_25839_volume_of_grey_matter_in_cingulate_gyrus_anterior_division_right_2_0 = TensorMap('25839_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25839_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-right_2_0': 0})
field_id_25840_volume_of_grey_matter_in_cingulate_gyrus_posterior_division_left_2_0 = TensorMap('25840_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25840_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-left_2_0': 0})
field_id_25841_volume_of_grey_matter_in_cingulate_gyrus_posterior_division_right_2_0 = TensorMap('25841_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25841_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-right_2_0': 0})
field_id_25842_volume_of_grey_matter_in_precuneous_cortex_left_2_0 = TensorMap('25842_Volume-of-grey-matter-in-Precuneous-Cortex-left_2_0', loss='logcosh', channel_map={'no_25842_Volume-of-grey-matter-in-Precuneous-Cortex-left_2_0': 0})
field_id_25843_volume_of_grey_matter_in_precuneous_cortex_right_2_0 = TensorMap('25843_Volume-of-grey-matter-in-Precuneous-Cortex-right_2_0', loss='logcosh', channel_map={'no_25843_Volume-of-grey-matter-in-Precuneous-Cortex-right_2_0': 0})
field_id_25844_volume_of_grey_matter_in_cuneal_cortex_left_2_0 = TensorMap('25844_Volume-of-grey-matter-in-Cuneal-Cortex-left_2_0', loss='logcosh', channel_map={'no_25844_Volume-of-grey-matter-in-Cuneal-Cortex-left_2_0': 0})
field_id_25845_volume_of_grey_matter_in_cuneal_cortex_right_2_0 = TensorMap('25845_Volume-of-grey-matter-in-Cuneal-Cortex-right_2_0', loss='logcosh', channel_map={'no_25845_Volume-of-grey-matter-in-Cuneal-Cortex-right_2_0': 0})
field_id_25846_volume_of_grey_matter_in_frontal_orbital_cortex_left_2_0 = TensorMap('25846_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-left_2_0', loss='logcosh', channel_map={'no_25846_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-left_2_0': 0})
field_id_25847_volume_of_grey_matter_in_frontal_orbital_cortex_right_2_0 = TensorMap('25847_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-right_2_0', loss='logcosh', channel_map={'no_25847_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-right_2_0': 0})
field_id_25848_volume_of_grey_matter_in_parahippocampal_gyrus_anterior_division_left_2_0 = TensorMap('25848_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25848_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-left_2_0': 0})
field_id_25849_volume_of_grey_matter_in_parahippocampal_gyrus_anterior_division_right_2_0 = TensorMap('25849_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25849_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-right_2_0': 0})
field_id_25850_volume_of_grey_matter_in_parahippocampal_gyrus_posterior_division_left_2_0 = TensorMap('25850_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25850_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-left_2_0': 0})
field_id_25851_volume_of_grey_matter_in_parahippocampal_gyrus_posterior_division_right_2_0 = TensorMap('25851_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25851_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-right_2_0': 0})
field_id_25852_volume_of_grey_matter_in_lingual_gyrus_left_2_0 = TensorMap('25852_Volume-of-grey-matter-in-Lingual-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25852_Volume-of-grey-matter-in-Lingual-Gyrus-left_2_0': 0})
field_id_25853_volume_of_grey_matter_in_lingual_gyrus_right_2_0 = TensorMap('25853_Volume-of-grey-matter-in-Lingual-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25853_Volume-of-grey-matter-in-Lingual-Gyrus-right_2_0': 0})
field_id_25854_volume_of_grey_matter_in_temporal_fusiform_cortex_anterior_division_left_2_0 = TensorMap('25854_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-left_2_0', loss='logcosh', channel_map={'no_25854_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-left_2_0': 0})
field_id_25855_volume_of_grey_matter_in_temporal_fusiform_cortex_anterior_division_right_2_0 = TensorMap('25855_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-right_2_0', loss='logcosh', channel_map={'no_25855_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-right_2_0': 0})
field_id_25856_volume_of_grey_matter_in_temporal_fusiform_cortex_posterior_division_left_2_0 = TensorMap('25856_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-left_2_0', loss='logcosh', channel_map={'no_25856_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-left_2_0': 0})
field_id_25857_volume_of_grey_matter_in_temporal_fusiform_cortex_posterior_division_right_2_0 = TensorMap('25857_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-right_2_0', loss='logcosh', channel_map={'no_25857_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-right_2_0': 0})
field_id_25858_volume_of_grey_matter_in_temporal_occipital_fusiform_cortex_left_2_0 = TensorMap('25858_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-left_2_0', loss='logcosh', channel_map={'no_25858_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-left_2_0': 0})
field_id_25859_volume_of_grey_matter_in_temporal_occipital_fusiform_cortex_right_2_0 = TensorMap('25859_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-right_2_0', loss='logcosh', channel_map={'no_25859_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-right_2_0': 0})
field_id_25860_volume_of_grey_matter_in_occipital_fusiform_gyrus_left_2_0 = TensorMap('25860_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-left_2_0', loss='logcosh', channel_map={'no_25860_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-left_2_0': 0})
field_id_25861_volume_of_grey_matter_in_occipital_fusiform_gyrus_right_2_0 = TensorMap('25861_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-right_2_0', loss='logcosh', channel_map={'no_25861_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-right_2_0': 0})
field_id_25862_volume_of_grey_matter_in_frontal_operculum_cortex_left_2_0 = TensorMap('25862_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-left_2_0', loss='logcosh', channel_map={'no_25862_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-left_2_0': 0})
field_id_25863_volume_of_grey_matter_in_frontal_operculum_cortex_right_2_0 = TensorMap('25863_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-right_2_0', loss='logcosh', channel_map={'no_25863_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-right_2_0': 0})
field_id_25864_volume_of_grey_matter_in_central_opercular_cortex_left_2_0 = TensorMap('25864_Volume-of-grey-matter-in-Central-Opercular-Cortex-left_2_0', loss='logcosh', channel_map={'no_25864_Volume-of-grey-matter-in-Central-Opercular-Cortex-left_2_0': 0})
field_id_25865_volume_of_grey_matter_in_central_opercular_cortex_right_2_0 = TensorMap('25865_Volume-of-grey-matter-in-Central-Opercular-Cortex-right_2_0', loss='logcosh', channel_map={'no_25865_Volume-of-grey-matter-in-Central-Opercular-Cortex-right_2_0': 0})
field_id_25866_volume_of_grey_matter_in_parietal_operculum_cortex_left_2_0 = TensorMap('25866_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-left_2_0', loss='logcosh', channel_map={'no_25866_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-left_2_0': 0})
field_id_25867_volume_of_grey_matter_in_parietal_operculum_cortex_right_2_0 = TensorMap('25867_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-right_2_0', loss='logcosh', channel_map={'no_25867_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-right_2_0': 0})
field_id_25868_volume_of_grey_matter_in_planum_polare_left_2_0 = TensorMap('25868_Volume-of-grey-matter-in-Planum-Polare-left_2_0', loss='logcosh', channel_map={'no_25868_Volume-of-grey-matter-in-Planum-Polare-left_2_0': 0})
field_id_25869_volume_of_grey_matter_in_planum_polare_right_2_0 = TensorMap('25869_Volume-of-grey-matter-in-Planum-Polare-right_2_0', loss='logcosh', channel_map={'no_25869_Volume-of-grey-matter-in-Planum-Polare-right_2_0': 0})
field_id_25870_volume_of_grey_matter_in_heschls_gyrus_includes_h1_and_h2_left_2_0 = TensorMap('25870_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-left_2_0', loss='logcosh', channel_map={'no_25870_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-left_2_0': 0})
field_id_25871_volume_of_grey_matter_in_heschls_gyrus_includes_h1_and_h2_right_2_0 = TensorMap('25871_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-right_2_0', loss='logcosh', channel_map={'no_25871_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-right_2_0': 0})
field_id_25872_volume_of_grey_matter_in_planum_temporale_left_2_0 = TensorMap('25872_Volume-of-grey-matter-in-Planum-Temporale-left_2_0', loss='logcosh', channel_map={'no_25872_Volume-of-grey-matter-in-Planum-Temporale-left_2_0': 0})
field_id_25873_volume_of_grey_matter_in_planum_temporale_right_2_0 = TensorMap('25873_Volume-of-grey-matter-in-Planum-Temporale-right_2_0', loss='logcosh', channel_map={'no_25873_Volume-of-grey-matter-in-Planum-Temporale-right_2_0': 0})
field_id_25874_volume_of_grey_matter_in_supracalcarine_cortex_left_2_0 = TensorMap('25874_Volume-of-grey-matter-in-Supracalcarine-Cortex-left_2_0', loss='logcosh', channel_map={'no_25874_Volume-of-grey-matter-in-Supracalcarine-Cortex-left_2_0': 0})
field_id_25875_volume_of_grey_matter_in_supracalcarine_cortex_right_2_0 = TensorMap('25875_Volume-of-grey-matter-in-Supracalcarine-Cortex-right_2_0', loss='logcosh', channel_map={'no_25875_Volume-of-grey-matter-in-Supracalcarine-Cortex-right_2_0': 0})
field_id_25876_volume_of_grey_matter_in_occipital_pole_left_2_0 = TensorMap('25876_Volume-of-grey-matter-in-Occipital-Pole-left_2_0', loss='logcosh', channel_map={'no_25876_Volume-of-grey-matter-in-Occipital-Pole-left_2_0': 0})
field_id_25877_volume_of_grey_matter_in_occipital_pole_right_2_0 = TensorMap('25877_Volume-of-grey-matter-in-Occipital-Pole-right_2_0', loss='logcosh', channel_map={'no_25877_Volume-of-grey-matter-in-Occipital-Pole-right_2_0': 0})
field_id_25878_volume_of_grey_matter_in_thalamus_left_2_0 = TensorMap('25878_Volume-of-grey-matter-in-Thalamus-left_2_0', loss='logcosh', channel_map={'no_25878_Volume-of-grey-matter-in-Thalamus-left_2_0': 0})
field_id_25879_volume_of_grey_matter_in_thalamus_right_2_0 = TensorMap('25879_Volume-of-grey-matter-in-Thalamus-right_2_0', loss='logcosh', channel_map={'no_25879_Volume-of-grey-matter-in-Thalamus-right_2_0': 0})
field_id_25880_volume_of_grey_matter_in_caudate_left_2_0 = TensorMap('25880_Volume-of-grey-matter-in-Caudate-left_2_0', loss='logcosh', channel_map={'no_25880_Volume-of-grey-matter-in-Caudate-left_2_0': 0})
field_id_25881_volume_of_grey_matter_in_caudate_right_2_0 = TensorMap('25881_Volume-of-grey-matter-in-Caudate-right_2_0', loss='logcosh', channel_map={'no_25881_Volume-of-grey-matter-in-Caudate-right_2_0': 0})
field_id_25882_volume_of_grey_matter_in_putamen_left_2_0 = TensorMap('25882_Volume-of-grey-matter-in-Putamen-left_2_0', loss='logcosh', channel_map={'no_25882_Volume-of-grey-matter-in-Putamen-left_2_0': 0})
field_id_25883_volume_of_grey_matter_in_putamen_right_2_0 = TensorMap('25883_Volume-of-grey-matter-in-Putamen-right_2_0', loss='logcosh', channel_map={'no_25883_Volume-of-grey-matter-in-Putamen-right_2_0': 0})
field_id_25884_volume_of_grey_matter_in_pallidum_left_2_0 = TensorMap('25884_Volume-of-grey-matter-in-Pallidum-left_2_0', loss='logcosh', channel_map={'no_25884_Volume-of-grey-matter-in-Pallidum-left_2_0': 0})
field_id_25885_volume_of_grey_matter_in_pallidum_right_2_0 = TensorMap('25885_Volume-of-grey-matter-in-Pallidum-right_2_0', loss='logcosh', channel_map={'no_25885_Volume-of-grey-matter-in-Pallidum-right_2_0': 0})
field_id_25886_volume_of_grey_matter_in_hippocampus_left_2_0 = TensorMap('25886_Volume-of-grey-matter-in-Hippocampus-left_2_0', loss='logcosh', channel_map={'no_25886_Volume-of-grey-matter-in-Hippocampus-left_2_0': 0})
field_id_25887_volume_of_grey_matter_in_hippocampus_right_2_0 = TensorMap('25887_Volume-of-grey-matter-in-Hippocampus-right_2_0', loss='logcosh', channel_map={'no_25887_Volume-of-grey-matter-in-Hippocampus-right_2_0': 0})
field_id_25888_volume_of_grey_matter_in_amygdala_left_2_0 = TensorMap('25888_Volume-of-grey-matter-in-Amygdala-left_2_0', loss='logcosh', channel_map={'no_25888_Volume-of-grey-matter-in-Amygdala-left_2_0': 0})
field_id_25889_volume_of_grey_matter_in_amygdala_right_2_0 = TensorMap('25889_Volume-of-grey-matter-in-Amygdala-right_2_0', loss='logcosh', channel_map={'no_25889_Volume-of-grey-matter-in-Amygdala-right_2_0': 0})
field_id_25890_volume_of_grey_matter_in_ventral_striatum_left_2_0 = TensorMap('25890_Volume-of-grey-matter-in-Ventral-Striatum-left_2_0', loss='logcosh', channel_map={'no_25890_Volume-of-grey-matter-in-Ventral-Striatum-left_2_0': 0})
field_id_25891_volume_of_grey_matter_in_ventral_striatum_right_2_0 = TensorMap('25891_Volume-of-grey-matter-in-Ventral-Striatum-right_2_0', loss='logcosh', channel_map={'no_25891_Volume-of-grey-matter-in-Ventral-Striatum-right_2_0': 0})
field_id_25892_volume_of_grey_matter_in_brainstem_2_0 = TensorMap('25892_Volume-of-grey-matter-in-BrainStem_2_0', loss='logcosh', channel_map={'no_25892_Volume-of-grey-matter-in-BrainStem_2_0': 0})
field_id_25893_volume_of_grey_matter_in_iiv_cerebellum_left_2_0 = TensorMap('25893_Volume-of-grey-matter-in-IIV-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25893_Volume-of-grey-matter-in-IIV-Cerebellum-left_2_0': 0})
field_id_25894_volume_of_grey_matter_in_iiv_cerebellum_right_2_0 = TensorMap('25894_Volume-of-grey-matter-in-IIV-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25894_Volume-of-grey-matter-in-IIV-Cerebellum-right_2_0': 0})
field_id_25895_volume_of_grey_matter_in_v_cerebellum_left_2_0 = TensorMap('25895_Volume-of-grey-matter-in-V-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25895_Volume-of-grey-matter-in-V-Cerebellum-left_2_0': 0})
field_id_25896_volume_of_grey_matter_in_v_cerebellum_right_2_0 = TensorMap('25896_Volume-of-grey-matter-in-V-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25896_Volume-of-grey-matter-in-V-Cerebellum-right_2_0': 0})
field_id_25897_volume_of_grey_matter_in_vi_cerebellum_left_2_0 = TensorMap('25897_Volume-of-grey-matter-in-VI-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25897_Volume-of-grey-matter-in-VI-Cerebellum-left_2_0': 0})
field_id_25898_volume_of_grey_matter_in_vi_cerebellum_vermis_2_0 = TensorMap('25898_Volume-of-grey-matter-in-VI-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25898_Volume-of-grey-matter-in-VI-Cerebellum-vermis_2_0': 0})
field_id_25899_volume_of_grey_matter_in_vi_cerebellum_right_2_0 = TensorMap('25899_Volume-of-grey-matter-in-VI-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25899_Volume-of-grey-matter-in-VI-Cerebellum-right_2_0': 0})
field_id_25900_volume_of_grey_matter_in_crus_i_cerebellum_left_2_0 = TensorMap('25900_Volume-of-grey-matter-in-Crus-I-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25900_Volume-of-grey-matter-in-Crus-I-Cerebellum-left_2_0': 0})
field_id_25901_volume_of_grey_matter_in_crus_i_cerebellum_vermis_2_0 = TensorMap('25901_Volume-of-grey-matter-in-Crus-I-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25901_Volume-of-grey-matter-in-Crus-I-Cerebellum-vermis_2_0': 0})
field_id_25902_volume_of_grey_matter_in_crus_i_cerebellum_right_2_0 = TensorMap('25902_Volume-of-grey-matter-in-Crus-I-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25902_Volume-of-grey-matter-in-Crus-I-Cerebellum-right_2_0': 0})
field_id_25903_volume_of_grey_matter_in_crus_ii_cerebellum_left_2_0 = TensorMap('25903_Volume-of-grey-matter-in-Crus-II-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25903_Volume-of-grey-matter-in-Crus-II-Cerebellum-left_2_0': 0})
field_id_25904_volume_of_grey_matter_in_crus_ii_cerebellum_vermis_2_0 = TensorMap('25904_Volume-of-grey-matter-in-Crus-II-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25904_Volume-of-grey-matter-in-Crus-II-Cerebellum-vermis_2_0': 0})
field_id_25905_volume_of_grey_matter_in_crus_ii_cerebellum_right_2_0 = TensorMap('25905_Volume-of-grey-matter-in-Crus-II-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25905_Volume-of-grey-matter-in-Crus-II-Cerebellum-right_2_0': 0})
field_id_25906_volume_of_grey_matter_in_viib_cerebellum_left_2_0 = TensorMap('25906_Volume-of-grey-matter-in-VIIb-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25906_Volume-of-grey-matter-in-VIIb-Cerebellum-left_2_0': 0})
field_id_25907_volume_of_grey_matter_in_viib_cerebellum_vermis_2_0 = TensorMap('25907_Volume-of-grey-matter-in-VIIb-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25907_Volume-of-grey-matter-in-VIIb-Cerebellum-vermis_2_0': 0})
field_id_25908_volume_of_grey_matter_in_viib_cerebellum_right_2_0 = TensorMap('25908_Volume-of-grey-matter-in-VIIb-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25908_Volume-of-grey-matter-in-VIIb-Cerebellum-right_2_0': 0})
field_id_25909_volume_of_grey_matter_in_viiia_cerebellum_left_2_0 = TensorMap('25909_Volume-of-grey-matter-in-VIIIa-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25909_Volume-of-grey-matter-in-VIIIa-Cerebellum-left_2_0': 0})
field_id_25910_volume_of_grey_matter_in_viiia_cerebellum_vermis_2_0 = TensorMap('25910_Volume-of-grey-matter-in-VIIIa-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25910_Volume-of-grey-matter-in-VIIIa-Cerebellum-vermis_2_0': 0})
field_id_25911_volume_of_grey_matter_in_viiia_cerebellum_right_2_0 = TensorMap('25911_Volume-of-grey-matter-in-VIIIa-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25911_Volume-of-grey-matter-in-VIIIa-Cerebellum-right_2_0': 0})
field_id_25912_volume_of_grey_matter_in_viiib_cerebellum_left_2_0 = TensorMap('25912_Volume-of-grey-matter-in-VIIIb-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25912_Volume-of-grey-matter-in-VIIIb-Cerebellum-left_2_0': 0})
field_id_25913_volume_of_grey_matter_in_viiib_cerebellum_vermis_2_0 = TensorMap('25913_Volume-of-grey-matter-in-VIIIb-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25913_Volume-of-grey-matter-in-VIIIb-Cerebellum-vermis_2_0': 0})
field_id_25914_volume_of_grey_matter_in_viiib_cerebellum_right_2_0 = TensorMap('25914_Volume-of-grey-matter-in-VIIIb-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25914_Volume-of-grey-matter-in-VIIIb-Cerebellum-right_2_0': 0})
field_id_25915_volume_of_grey_matter_in_ix_cerebellum_left_2_0 = TensorMap('25915_Volume-of-grey-matter-in-IX-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25915_Volume-of-grey-matter-in-IX-Cerebellum-left_2_0': 0})
field_id_25916_volume_of_grey_matter_in_ix_cerebellum_vermis_2_0 = TensorMap('25916_Volume-of-grey-matter-in-IX-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25916_Volume-of-grey-matter-in-IX-Cerebellum-vermis_2_0': 0})
field_id_25917_volume_of_grey_matter_in_ix_cerebellum_right_2_0 = TensorMap('25917_Volume-of-grey-matter-in-IX-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25917_Volume-of-grey-matter-in-IX-Cerebellum-right_2_0': 0})
field_id_25918_volume_of_grey_matter_in_x_cerebellum_left_2_0 = TensorMap('25918_Volume-of-grey-matter-in-X-Cerebellum-left_2_0', loss='logcosh', channel_map={'no_25918_Volume-of-grey-matter-in-X-Cerebellum-left_2_0': 0})
field_id_25919_volume_of_grey_matter_in_x_cerebellum_vermis_2_0 = TensorMap('25919_Volume-of-grey-matter-in-X-Cerebellum-vermis_2_0', loss='logcosh', channel_map={'no_25919_Volume-of-grey-matter-in-X-Cerebellum-vermis_2_0': 0})
field_id_25920_volume_of_grey_matter_in_x_cerebellum_right_2_0 = TensorMap('25920_Volume-of-grey-matter-in-X-Cerebellum-right_2_0', loss='logcosh', channel_map={'no_25920_Volume-of-grey-matter-in-X-Cerebellum-right_2_0': 0})
field_id_2684_years_since_last_breast_cancer_screening_mammogram_0_0 = TensorMap('2684_Years-since-last-breast-cancer-screening-mammogram_0_0', loss='logcosh', channel_map={'no_2684_Years-since-last-breast-cancer-screening-mammogram_0_0': 0})
field_id_2684_years_since_last_breast_cancer_screening_mammogram_2_0 = TensorMap('2684_Years-since-last-breast-cancer-screening-mammogram_2_0', loss='logcosh', channel_map={'no_2684_Years-since-last-breast-cancer-screening-mammogram_2_0': 0})
field_id_2704_years_since_last_cervical_smear_test_0_0 = TensorMap('2704_Years-since-last-cervical-smear-test_0_0', loss='logcosh', channel_map={'no_2704_Years-since-last-cervical-smear-test_0_0': 0})
field_id_2704_years_since_last_cervical_smear_test_2_0 = TensorMap('2704_Years-since-last-cervical-smear-test_2_0', loss='logcosh', channel_map={'no_2704_Years-since-last-cervical-smear-test_2_0': 0})
field_id_2714_age_when_periods_started_menarche_0_0 = TensorMap('2714_Age-when-periods-started-menarche_0_0', loss='logcosh', channel_map={'no_2714_Age-when-periods-started-menarche_0_0': 0})
field_id_2714_age_when_periods_started_menarche_2_0 = TensorMap('2714_Age-when-periods-started-menarche_2_0', loss='logcosh', channel_map={'no_2714_Age-when-periods-started-menarche_2_0': 0})
field_id_2734_number_of_live_births_0_0 = TensorMap('2734_Number-of-live-births_0_0', loss='logcosh', channel_map={'no_2734_Number-of-live-births_0_0': 0})
field_id_2734_number_of_live_births_2_0 = TensorMap('2734_Number-of-live-births_2_0', loss='logcosh', channel_map={'no_2734_Number-of-live-births_2_0': 0})
field_id_2744_birth_weight_of_first_child_0_0 = TensorMap('2744_Birth-weight-of-first-child_0_0', loss='logcosh', channel_map={'no_2744_Birth-weight-of-first-child_0_0': 0})
field_id_2744_birth_weight_of_first_child_2_0 = TensorMap('2744_Birth-weight-of-first-child_2_0', loss='logcosh', channel_map={'no_2744_Birth-weight-of-first-child_2_0': 0})
field_id_2754_age_at_first_live_birth_0_0 = TensorMap('2754_Age-at-first-live-birth_0_0', loss='logcosh', channel_map={'no_2754_Age-at-first-live-birth_0_0': 0})
field_id_2754_age_at_first_live_birth_2_0 = TensorMap('2754_Age-at-first-live-birth_2_0', loss='logcosh', channel_map={'no_2754_Age-at-first-live-birth_2_0': 0})
field_id_2764_age_at_last_live_birth_0_0 = TensorMap('2764_Age-at-last-live-birth_0_0', loss='logcosh', channel_map={'no_2764_Age-at-last-live-birth_0_0': 0})
field_id_2764_age_at_last_live_birth_2_0 = TensorMap('2764_Age-at-last-live-birth_2_0', loss='logcosh', channel_map={'no_2764_Age-at-last-live-birth_2_0': 0})
field_id_2794_age_started_oral_contraceptive_pill_0_0 = TensorMap('2794_Age-started-oral-contraceptive-pill_0_0', loss='logcosh', channel_map={'no_2794_Age-started-oral-contraceptive-pill_0_0': 0})
field_id_2794_age_started_oral_contraceptive_pill_2_0 = TensorMap('2794_Age-started-oral-contraceptive-pill_2_0', loss='logcosh', channel_map={'no_2794_Age-started-oral-contraceptive-pill_2_0': 0})
field_id_2804_age_when_last_used_oral_contraceptive_pill_0_0 = TensorMap('2804_Age-when-last-used-oral-contraceptive-pill_0_0', loss='logcosh', channel_map={'no_2804_Age-when-last-used-oral-contraceptive-pill_0_0': 0})
field_id_2804_age_when_last_used_oral_contraceptive_pill_2_0 = TensorMap('2804_Age-when-last-used-oral-contraceptive-pill_2_0', loss='logcosh', channel_map={'no_2804_Age-when-last-used-oral-contraceptive-pill_2_0': 0})
field_id_2966_age_high_blood_pressure_diagnosed_2_0 = TensorMap('2966_Age-high-blood-pressure-diagnosed_2_0', loss='logcosh', channel_map={'no_2966_Age-high-blood-pressure-diagnosed_2_0': 0})
field_id_3062_forced_vital_capacity_fvc_2_0 = TensorMap('3062_Forced-vital-capacity-FVC_2_0', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_2_0': 0})
field_id_3062_forced_vital_capacity_fvc_2_1 = TensorMap('3062_Forced-vital-capacity-FVC_2_1', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_2_1': 0})
field_id_3062_forced_vital_capacity_fvc_2_2 = TensorMap('3062_Forced-vital-capacity-FVC_2_2', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_2_2': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_2_0 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_2_0', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_2_0': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_2_1 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_2_1', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_2_1': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_2_2 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_2_2', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_2_2': 0})
field_id_3064_peak_expiratory_flow_pef_2_0 = TensorMap('3064_Peak-expiratory-flow-PEF_2_0', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_2_0': 0})
field_id_3064_peak_expiratory_flow_pef_2_1 = TensorMap('3064_Peak-expiratory-flow-PEF_2_1', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_2_1': 0})
field_id_3064_peak_expiratory_flow_pef_2_2 = TensorMap('3064_Peak-expiratory-flow-PEF_2_2', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_2_2': 0})
field_id_3526_mothers_age_at_death_0_0 = TensorMap('3526_Mothers-age-at-death_0_0', loss='logcosh', channel_map={'no_3526_Mothers-age-at-death_0_0': 0})
field_id_3526_mothers_age_at_death_2_0 = TensorMap('3526_Mothers-age-at-death_2_0', loss='logcosh', channel_map={'no_3526_Mothers-age-at-death_2_0': 0})
field_id_3581_age_at_menopause_last_menstrual_period_0_0 = TensorMap('3581_Age-at-menopause-last-menstrual-period_0_0', loss='logcosh', channel_map={'no_3581_Age-at-menopause-last-menstrual-period_0_0': 0})
field_id_3581_age_at_menopause_last_menstrual_period_2_0 = TensorMap('3581_Age-at-menopause-last-menstrual-period_2_0', loss='logcosh', channel_map={'no_3581_Age-at-menopause-last-menstrual-period_2_0': 0})
field_id_396_number_of_columns_displayed_in_round_2_1 = TensorMap('396_Number-of-columns-displayed-in-round_2_1', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_2_1': 0})
field_id_396_number_of_columns_displayed_in_round_2_2 = TensorMap('396_Number-of-columns-displayed-in-round_2_2', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_2_2': 0})
field_id_396_number_of_columns_displayed_in_round_2_3 = TensorMap('396_Number-of-columns-displayed-in-round_2_3', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_2_3': 0})
field_id_397_number_of_rows_displayed_in_round_2_1 = TensorMap('397_Number-of-rows-displayed-in-round_2_1', loss='logcosh', channel_map={'no_397_Number-of-rows-displayed-in-round_2_1': 0})
field_id_397_number_of_rows_displayed_in_round_2_2 = TensorMap('397_Number-of-rows-displayed-in-round_2_2', loss='logcosh', channel_map={'no_397_Number-of-rows-displayed-in-round_2_2': 0})
field_id_398_number_of_correct_matches_in_round_2_1 = TensorMap('398_Number-of-correct-matches-in-round_2_1', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_2_1': 0})
field_id_398_number_of_correct_matches_in_round_2_2 = TensorMap('398_Number-of-correct-matches-in-round_2_2', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_2_2': 0})
field_id_398_number_of_correct_matches_in_round_2_3 = TensorMap('398_Number-of-correct-matches-in-round_2_3', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_2_3': 0})
field_id_399_number_of_incorrect_matches_in_round_2_1 = TensorMap('399_Number-of-incorrect-matches-in-round_2_1', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_2_1': 0})
field_id_399_number_of_incorrect_matches_in_round_2_2 = TensorMap('399_Number-of-incorrect-matches-in-round_2_2', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_2_2': 0})
field_id_399_number_of_incorrect_matches_in_round_2_3 = TensorMap('399_Number-of-incorrect-matches-in-round_2_3', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_2_3': 0})
field_id_400_time_to_complete_round_2_1 = TensorMap('400_Time-to-complete-round_2_1', loss='logcosh', channel_map={'no_400_Time-to-complete-round_2_1': 0})
field_id_400_time_to_complete_round_2_2 = TensorMap('400_Time-to-complete-round_2_2', loss='logcosh', channel_map={'no_400_Time-to-complete-round_2_2': 0})
field_id_400_time_to_complete_round_2_3 = TensorMap('400_Time-to-complete-round_2_3', loss='logcosh', channel_map={'no_400_Time-to-complete-round_2_3': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_1 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_1', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_1': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_10 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_10', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_10': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_11 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_11', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_11': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_2 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_2', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_2': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_5 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_5', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_5': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_7 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_7', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_7': 0})
field_id_4079_diastolic_blood_pressure_automated_reading_2_0 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_2_0', loss='logcosh', channel_map={'no_4079_Diastolic-blood-pressure-automated-reading_2_0': 0})
field_id_4079_diastolic_blood_pressure_automated_reading_2_1 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_2_1', loss='logcosh', channel_map={'no_4079_Diastolic-blood-pressure-automated-reading_2_1': 0})
field_id_4080_systolic_blood_pressure_automated_reading_2_0 = TensorMap('4080_Systolic-blood-pressure-automated-reading_2_0', loss='logcosh', channel_map={'no_4080_Systolic-blood-pressure-automated-reading_2_0': 0})
field_id_4080_systolic_blood_pressure_automated_reading_2_1 = TensorMap('4080_Systolic-blood-pressure-automated-reading_2_1', loss='logcosh', channel_map={'no_4080_Systolic-blood-pressure-automated-reading_2_1': 0})
field_id_4194_pulse_rate_2_0 = TensorMap('4194_Pulse-rate_2_0', loss='logcosh', channel_map={'no_4194_Pulse-rate_2_0': 0})
field_id_4195_pulse_wave_reflection_index_2_0 = TensorMap('4195_Pulse-wave-reflection-index_2_0', loss='logcosh', channel_map={'no_4195_Pulse-wave-reflection-index_2_0': 0})
field_id_4196_pulse_wave_peak_to_peak_time_2_0 = TensorMap('4196_Pulse-wave-peak-to-peak-time_2_0', loss='logcosh', channel_map={'no_4196_Pulse-wave-peak-to-peak-time_2_0': 0})
field_id_4198_position_of_the_pulse_wave_peak_2_0 = TensorMap('4198_Position-of-the-pulse-wave-peak_2_0', loss='logcosh', channel_map={'no_4198_Position-of-the-pulse-wave-peak_2_0': 0})
field_id_4199_position_of_pulse_wave_notch_2_0 = TensorMap('4199_Position-of-pulse-wave-notch_2_0', loss='logcosh', channel_map={'no_4199_Position-of-pulse-wave-notch_2_0': 0})
field_id_4200_position_of_the_shoulder_on_the_pulse_waveform_2_0 = TensorMap('4200_Position-of-the-shoulder-on-the-pulse-waveform_2_0', loss='logcosh', channel_map={'no_4200_Position-of-the-shoulder-on-the-pulse-waveform_2_0': 0})
field_id_4256_time_elapsed_2_0 = TensorMap('4256_Time-elapsed_2_0', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_0': 0})
field_id_4256_time_elapsed_2_1 = TensorMap('4256_Time-elapsed_2_1', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_1': 0})
field_id_4256_time_elapsed_2_2 = TensorMap('4256_Time-elapsed_2_2', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_2': 0})
field_id_4256_time_elapsed_2_3 = TensorMap('4256_Time-elapsed_2_3', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_3': 0})
field_id_4256_time_elapsed_2_4 = TensorMap('4256_Time-elapsed_2_4', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_4': 0})
field_id_4256_time_elapsed_2_5 = TensorMap('4256_Time-elapsed_2_5', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_5': 0})
field_id_4256_time_elapsed_2_6 = TensorMap('4256_Time-elapsed_2_6', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_6': 0})
field_id_4260_round_of_numeric_memory_test_2_0 = TensorMap('4260_Round-of-numeric-memory-test_2_0', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_0': 0})
field_id_4260_round_of_numeric_memory_test_2_1 = TensorMap('4260_Round-of-numeric-memory-test_2_1', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_1': 0})
field_id_4260_round_of_numeric_memory_test_2_2 = TensorMap('4260_Round-of-numeric-memory-test_2_2', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_2': 0})
field_id_4260_round_of_numeric_memory_test_2_3 = TensorMap('4260_Round-of-numeric-memory-test_2_3', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_3': 0})
field_id_4260_round_of_numeric_memory_test_2_4 = TensorMap('4260_Round-of-numeric-memory-test_2_4', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_4': 0})
field_id_4260_round_of_numeric_memory_test_2_5 = TensorMap('4260_Round-of-numeric-memory-test_2_5', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_5': 0})
field_id_4260_round_of_numeric_memory_test_2_6 = TensorMap('4260_Round-of-numeric-memory-test_2_6', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_6': 0})
field_id_4269_number_of_triplets_attempted_left_2_0 = TensorMap('4269_Number-of-triplets-attempted-left_2_0', loss='logcosh', channel_map={'no_4269_Number-of-triplets-attempted-left_2_0': 0})
field_id_4272_duration_of_hearing_test_left_2_0 = TensorMap('4272_Duration-of-hearing-test-left_2_0', loss='logcosh', channel_map={'no_4272_Duration-of-hearing-test-left_2_0': 0})
field_id_4276_number_of_triplets_attempted_right_2_0 = TensorMap('4276_Number-of-triplets-attempted-right_2_0', loss='logcosh', channel_map={'no_4276_Number-of-triplets-attempted-right_2_0': 0})
field_id_4279_duration_of_hearing_test_right_2_0 = TensorMap('4279_Duration-of-hearing-test-right_2_0', loss='logcosh', channel_map={'no_4279_Duration-of-hearing-test-right_2_0': 0})
field_id_4282_maximum_digits_remembered_correctly_2_0 = TensorMap('4282_Maximum-digits-remembered-correctly_2_0', loss='logcosh', channel_map={'no_4282_Maximum-digits-remembered-correctly_2_0': 0})
field_id_4283_number_of_rounds_of_numeric_memory_test_performed_2_0 = TensorMap('4283_Number-of-rounds-of-numeric-memory-test-performed_2_0', loss='logcosh', channel_map={'no_4283_Number-of-rounds-of-numeric-memory-test-performed_2_0': 0})
field_id_4285_time_to_complete_test_2_0 = TensorMap('4285_Time-to-complete-test_2_0', loss='logcosh', channel_map={'no_4285_Time-to-complete-test_2_0': 0})
field_id_4288_time_to_answer_2_0 = TensorMap('4288_Time-to-answer_2_0', loss='logcosh', channel_map={'no_4288_Time-to-answer_2_0': 0})
field_id_4290_duration_screen_displayed_2_0 = TensorMap('4290_Duration-screen-displayed_2_0', loss='logcosh', channel_map={'no_4290_Duration-screen-displayed_2_0': 0})
field_id_4291_number_of_attempts_2_0 = TensorMap('4291_Number-of-attempts_2_0', loss='logcosh', channel_map={'no_4291_Number-of-attempts_2_0': 0})
field_id_46_hand_grip_strength_left_2_0 = TensorMap('46_Hand-grip-strength-left_2_0', loss='logcosh', channel_map={'no_46_Hand-grip-strength-left_2_0': 0})
field_id_47_hand_grip_strength_right_2_0 = TensorMap('47_Hand-grip-strength-right_2_0', loss='logcosh', channel_map={'no_47_Hand-grip-strength-right_2_0': 0})
field_id_48_waist_circumference_2_0 = TensorMap('48_Waist-circumference_2_0', loss='logcosh', channel_map={'no_48_Waist-circumference_2_0': 0})
field_id_49_hip_circumference_2_0 = TensorMap('49_Hip-circumference_2_0', loss='logcosh', channel_map={'no_49_Hip-circumference_2_0': 0})
field_id_5057_number_of_older_siblings_2_0 = TensorMap('5057_Number-of-older-siblings_2_0', loss='logcosh', channel_map={'no_5057_Number-of-older-siblings_2_0': 0})
field_id_5074_number_of_letters_shown_in_round_left_0_2 = TensorMap('5074_Number-of-letters-shown-in-round-left_0_2', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_0_2': 0})
field_id_5074_number_of_letters_shown_in_round_left_0_3 = TensorMap('5074_Number-of-letters-shown-in-round-left_0_3', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_0_3': 0})
field_id_5074_number_of_letters_shown_in_round_left_0_4 = TensorMap('5074_Number-of-letters-shown-in-round-left_0_4', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_0_4': 0})
field_id_5075_number_of_letters_shown_in_round_right_0_1 = TensorMap('5075_Number-of-letters-shown-in-round-right_0_1', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_0_1': 0})
field_id_5075_number_of_letters_shown_in_round_right_0_2 = TensorMap('5075_Number-of-letters-shown-in-round-right_0_2', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_0_2': 0})
field_id_5076_number_of_letters_correct_in_round_right_0_1 = TensorMap('5076_Number-of-letters-correct-in-round-right_0_1', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_0_1': 0})
field_id_5076_number_of_letters_correct_in_round_right_0_2 = TensorMap('5076_Number-of-letters-correct-in-round-right_0_2', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_0_2': 0})
field_id_5077_number_of_letters_correct_in_round_left_0_2 = TensorMap('5077_Number-of-letters-correct-in-round-left_0_2', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_0_2': 0})
field_id_5077_number_of_letters_correct_in_round_left_0_3 = TensorMap('5077_Number-of-letters-correct-in-round-left_0_3', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_0_3': 0})
field_id_5077_number_of_letters_correct_in_round_left_0_4 = TensorMap('5077_Number-of-letters-correct-in-round-left_0_4', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_0_4': 0})
field_id_5078_logmar_in_round_left_0_2 = TensorMap('5078_logMAR-in-round-left_0_2', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_0_2': 0})
field_id_5078_logmar_in_round_left_0_3 = TensorMap('5078_logMAR-in-round-left_0_3', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_0_3': 0})
field_id_5078_logmar_in_round_left_0_4 = TensorMap('5078_logMAR-in-round-left_0_4', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_0_4': 0})
field_id_5079_logmar_in_round_right_0_1 = TensorMap('5079_logMAR-in-round-right_0_1', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_0_1': 0})
field_id_5079_logmar_in_round_right_0_2 = TensorMap('5079_logMAR-in-round-right_0_2', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_0_2': 0})
field_id_5099_3mm_weak_meridian_right_0_2 = TensorMap('5099_3mm-weak-meridian-right_0_2', loss='logcosh', channel_map={'no_5099_3mm-weak-meridian-right_0_2': 0})
field_id_50_standing_height_2_0 = TensorMap('50_Standing-height_2_0', loss='logcosh', channel_map={'no_50_Standing-height_2_0': 0})
field_id_5100_3mm_weak_meridian_angle_right_0_2 = TensorMap('5100_3mm-weak-meridian-angle-right_0_2', loss='logcosh', channel_map={'no_5100_3mm-weak-meridian-angle-right_0_2': 0})
field_id_5107_3mm_strong_meridian_angle_right_0_2 = TensorMap('5107_3mm-strong-meridian-angle-right_0_2', loss='logcosh', channel_map={'no_5107_3mm-strong-meridian-angle-right_0_2': 0})
field_id_5108_3mm_asymmetry_angle_right_0_2 = TensorMap('5108_3mm-asymmetry-angle-right_0_2', loss='logcosh', channel_map={'no_5108_3mm-asymmetry-angle-right_0_2': 0})
field_id_5115_3mm_cylindrical_power_angle_right_0_2 = TensorMap('5115_3mm-cylindrical-power-angle-right_0_2', loss='logcosh', channel_map={'no_5115_3mm-cylindrical-power-angle-right_0_2': 0})
field_id_5116_3mm_cylindrical_power_right_0_2 = TensorMap('5116_3mm-cylindrical-power-right_0_2', loss='logcosh', channel_map={'no_5116_3mm-cylindrical-power-right_0_2': 0})
field_id_5132_3mm_strong_meridian_right_0_2 = TensorMap('5132_3mm-strong-meridian-right_0_2', loss='logcosh', channel_map={'no_5132_3mm-strong-meridian-right_0_2': 0})
field_id_5159_3mm_asymmetry_index_right_0_2 = TensorMap('5159_3mm-asymmetry-index-right_0_2', loss='logcosh', channel_map={'no_5159_3mm-asymmetry-index-right_0_2': 0})
field_id_5160_3mm_regularity_index_right_0_2 = TensorMap('5160_3mm-regularity-index-right_0_2', loss='logcosh', channel_map={'no_5160_3mm-regularity-index-right_0_2': 0})
field_id_51_seated_height_2_0 = TensorMap('51_Seated-height_2_0', loss='logcosh', channel_map={'no_51_Seated-height_2_0': 0})
field_id_5364_average_weekly_intake_of_other_alcoholic_drinks_2_0 = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_2_0', loss='logcosh', channel_map={'no_5364_Average-weekly-intake-of-other-alcoholic-drinks_2_0': 0})
field_id_699_length_of_time_at_current_address_2_0 = TensorMap('699_Length-of-time-at-current-address_2_0', loss='logcosh', channel_map={'no_699_Length-of-time-at-current-address_2_0': 0})
field_id_709_number_in_household_2_0 = TensorMap('709_Number-in-household_2_0', loss='logcosh', channel_map={'no_709_Number-in-household_2_0': 0})
field_id_864_number_of_daysweek_walked_10_minutes_2_0 = TensorMap('864_Number-of-daysweek-walked-10-minutes_2_0', loss='logcosh', channel_map={'no_864_Number-of-daysweek-walked-10-minutes_2_0': 0})
field_id_874_duration_of_walks_2_0 = TensorMap('874_Duration-of-walks_2_0', loss='logcosh', channel_map={'no_874_Duration-of-walks_2_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_0 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_0', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_2_0 = TensorMap('87_Noncancer-illness-yearage-first-occurred_2_0', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_2_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_2_1 = TensorMap('87_Noncancer-illness-yearage-first-occurred_2_1', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_2_1': 0})
field_id_87_noncancer_illness_yearage_first_occurred_2_2 = TensorMap('87_Noncancer-illness-yearage-first-occurred_2_2', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_2_2': 0})
field_id_87_noncancer_illness_yearage_first_occurred_2_3 = TensorMap('87_Noncancer-illness-yearage-first-occurred_2_3', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_2_3': 0})
field_id_87_noncancer_illness_yearage_first_occurred_2_4 = TensorMap('87_Noncancer-illness-yearage-first-occurred_2_4', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_2_4': 0})
field_id_884_number_of_daysweek_of_moderate_physical_activity_10_minutes_2_0 = TensorMap('884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_2_0', loss='logcosh', channel_map={'no_884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_2_0': 0})
field_id_90012_overall_acceleration_average_0_0 = TensorMap('90012_Overall-acceleration-average_0_0', loss='logcosh', channel_map={'no_90012_Overall-acceleration-average_0_0': 0})
field_id_90013_standard_deviation_of_acceleration_0_0 = TensorMap('90013_Standard-deviation-of-acceleration_0_0', loss='logcosh', channel_map={'no_90013_Standard-deviation-of-acceleration_0_0': 0})
field_id_904_number_of_daysweek_of_vigorous_physical_activity_10_minutes_2_0 = TensorMap('904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_2_0', loss='logcosh', channel_map={'no_904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_2_0': 0})
field_id_92_operation_yearage_first_occurred_2_0 = TensorMap('92_Operation-yearage-first-occurred_2_0', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_2_0': 0})
field_id_92_operation_yearage_first_occurred_2_1 = TensorMap('92_Operation-yearage-first-occurred_2_1', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_2_1': 0})
field_id_3659_year_immigrated_to_uk_united_kingdom_0_0 = TensorMap('3659_Year-immigrated-to-UK-United-Kingdom_0_0', loss='logcosh', channel_map={'no_3659_Year-immigrated-to-UK-United-Kingdom_0_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_1 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_1', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_1': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_2 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_2', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_2': 0})
field_id_3436_age_started_smoking_in_current_smokers_0_0 = TensorMap('3436_Age-started-smoking-in-current-smokers_0_0', loss='logcosh', channel_map={'no_3436_Age-started-smoking-in-current-smokers_0_0': 0})
field_id_3456_number_of_cigarettes_currently_smoked_daily_current_cigarette_smokers_0_0 = TensorMap('3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_0_0', loss='logcosh', channel_map={'no_3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_0_0': 0})
field_id_3536_age_started_hormonereplacement_therapy_hrt_0_0 = TensorMap('3536_Age-started-hormonereplacement-therapy-HRT_0_0', loss='logcosh', channel_map={'no_3536_Age-started-hormonereplacement-therapy-HRT_0_0': 0})
field_id_3546_age_last_used_hormonereplacement_therapy_hrt_0_0 = TensorMap('3546_Age-last-used-hormonereplacement-therapy-HRT_0_0', loss='logcosh', channel_map={'no_3546_Age-last-used-hormonereplacement-therapy-HRT_0_0': 0})
field_id_3992_age_emphysemachronic_bronchitis_diagnosed_0_0 = TensorMap('3992_Age-emphysemachronic-bronchitis-diagnosed_0_0', loss='logcosh', channel_map={'no_3992_Age-emphysemachronic-bronchitis-diagnosed_0_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_1 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_1', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_1': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_2 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_2', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_2': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_0_0 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_0_0', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_0_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_3 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_3', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_3': 0})
field_id_3761_age_hay_fever_rhinitis_or_eczema_diagnosed_0_0 = TensorMap('3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_0_0', loss='logcosh', channel_map={'no_3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_0_0': 0})
field_id_3786_age_asthma_diagnosed_0_0 = TensorMap('3786_Age-asthma-diagnosed_0_0', loss='logcosh', channel_map={'no_3786_Age-asthma-diagnosed_0_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_3 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_3', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_3': 0})
field_id_5075_number_of_letters_shown_in_round_right_0_3 = TensorMap('5075_Number-of-letters-shown-in-round-right_0_3', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_0_3': 0})
field_id_5075_number_of_letters_shown_in_round_right_0_4 = TensorMap('5075_Number-of-letters-shown-in-round-right_0_4', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_0_4': 0})
field_id_5076_number_of_letters_correct_in_round_right_0_3 = TensorMap('5076_Number-of-letters-correct-in-round-right_0_3', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_0_3': 0})
field_id_5076_number_of_letters_correct_in_round_right_0_4 = TensorMap('5076_Number-of-letters-correct-in-round-right_0_4', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_0_4': 0})
field_id_5079_logmar_in_round_right_0_3 = TensorMap('5079_logMAR-in-round-right_0_3', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_0_3': 0})
field_id_5079_logmar_in_round_right_0_4 = TensorMap('5079_logMAR-in-round-right_0_4', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_0_4': 0})
field_id_102_pulse_rate_automated_reading_1_0 = TensorMap('102_Pulse-rate-automated-reading_1_0', loss='logcosh', channel_map={'no_102_Pulse-rate-automated-reading_1_0': 0})
field_id_102_pulse_rate_automated_reading_1_1 = TensorMap('102_Pulse-rate-automated-reading_1_1', loss='logcosh', channel_map={'no_102_Pulse-rate-automated-reading_1_1': 0})
field_id_1050_time_spend_outdoors_in_summer_1_0 = TensorMap('1050_Time-spend-outdoors-in-summer_1_0', loss='logcosh', channel_map={'no_1050_Time-spend-outdoors-in-summer_1_0': 0})
field_id_1060_time_spent_outdoors_in_winter_1_0 = TensorMap('1060_Time-spent-outdoors-in-winter_1_0', loss='logcosh', channel_map={'no_1060_Time-spent-outdoors-in-winter_1_0': 0})
field_id_1070_time_spent_watching_television_tv_1_0 = TensorMap('1070_Time-spent-watching-television-TV_1_0', loss='logcosh', channel_map={'no_1070_Time-spent-watching-television-TV_1_0': 0})
field_id_1080_time_spent_using_computer_1_0 = TensorMap('1080_Time-spent-using-computer_1_0', loss='logcosh', channel_map={'no_1080_Time-spent-using-computer_1_0': 0})
field_id_1090_time_spent_driving_1_0 = TensorMap('1090_Time-spent-driving_1_0', loss='logcosh', channel_map={'no_1090_Time-spent-driving_1_0': 0})
field_id_1269_exposure_to_tobacco_smoke_at_home_1_0 = TensorMap('1269_Exposure-to-tobacco-smoke-at-home_1_0', loss='logcosh', channel_map={'no_1269_Exposure-to-tobacco-smoke-at-home_1_0': 0})
field_id_1279_exposure_to_tobacco_smoke_outside_home_1_0 = TensorMap('1279_Exposure-to-tobacco-smoke-outside-home_1_0', loss='logcosh', channel_map={'no_1279_Exposure-to-tobacco-smoke-outside-home_1_0': 0})
field_id_1289_cooked_vegetable_intake_1_0 = TensorMap('1289_Cooked-vegetable-intake_1_0', loss='logcosh', channel_map={'no_1289_Cooked-vegetable-intake_1_0': 0})
field_id_1299_salad_raw_vegetable_intake_1_0 = TensorMap('1299_Salad-raw-vegetable-intake_1_0', loss='logcosh', channel_map={'no_1299_Salad-raw-vegetable-intake_1_0': 0})
field_id_129_place_of_birth_in_uk_north_coordinate_1_0 = TensorMap('129_Place-of-birth-in-UK-north-coordinate_1_0', loss='logcosh', channel_map={'no_129_Place-of-birth-in-UK-north-coordinate_1_0': 0})
field_id_1309_fresh_fruit_intake_1_0 = TensorMap('1309_Fresh-fruit-intake_1_0', loss='logcosh', channel_map={'no_1309_Fresh-fruit-intake_1_0': 0})
field_id_130_place_of_birth_in_uk_east_coordinate_1_0 = TensorMap('130_Place-of-birth-in-UK-east-coordinate_1_0', loss='logcosh', channel_map={'no_130_Place-of-birth-in-UK-east-coordinate_1_0': 0})
field_id_1319_dried_fruit_intake_1_0 = TensorMap('1319_Dried-fruit-intake_1_0', loss='logcosh', channel_map={'no_1319_Dried-fruit-intake_1_0': 0})
field_id_134_number_of_selfreported_cancers_1_0 = TensorMap('134_Number-of-selfreported-cancers_1_0', loss='logcosh', channel_map={'no_134_Number-of-selfreported-cancers_1_0': 0})
field_id_135_number_of_selfreported_noncancer_illnesses_1_0 = TensorMap('135_Number-of-selfreported-noncancer-illnesses_1_0', loss='logcosh', channel_map={'no_135_Number-of-selfreported-noncancer-illnesses_1_0': 0})
field_id_136_number_of_operations_selfreported_1_0 = TensorMap('136_Number-of-operations-selfreported_1_0', loss='logcosh', channel_map={'no_136_Number-of-operations-selfreported_1_0': 0})
field_id_137_number_of_treatmentsmedications_taken_1_0 = TensorMap('137_Number-of-treatmentsmedications-taken_1_0', loss='logcosh', channel_map={'no_137_Number-of-treatmentsmedications-taken_1_0': 0})
field_id_1438_bread_intake_1_0 = TensorMap('1438_Bread-intake_1_0', loss='logcosh', channel_map={'no_1438_Bread-intake_1_0': 0})
field_id_1458_cereal_intake_1_0 = TensorMap('1458_Cereal-intake_1_0', loss='logcosh', channel_map={'no_1458_Cereal-intake_1_0': 0})
field_id_1488_tea_intake_1_0 = TensorMap('1488_Tea-intake_1_0', loss='logcosh', channel_map={'no_1488_Tea-intake_1_0': 0})
field_id_1498_coffee_intake_1_0 = TensorMap('1498_Coffee-intake_1_0', loss='logcosh', channel_map={'no_1498_Coffee-intake_1_0': 0})
field_id_1528_water_intake_1_0 = TensorMap('1528_Water-intake_1_0', loss='logcosh', channel_map={'no_1528_Water-intake_1_0': 0})
field_id_1568_average_weekly_red_wine_intake_1_0 = TensorMap('1568_Average-weekly-red-wine-intake_1_0', loss='logcosh', channel_map={'no_1568_Average-weekly-red-wine-intake_1_0': 0})
field_id_1578_average_weekly_champagne_plus_white_wine_intake_1_0 = TensorMap('1578_Average-weekly-champagne-plus-white-wine-intake_1_0', loss='logcosh', channel_map={'no_1578_Average-weekly-champagne-plus-white-wine-intake_1_0': 0})
field_id_1588_average_weekly_beer_plus_cider_intake_1_0 = TensorMap('1588_Average-weekly-beer-plus-cider-intake_1_0', loss='logcosh', channel_map={'no_1588_Average-weekly-beer-plus-cider-intake_1_0': 0})
field_id_1598_average_weekly_spirits_intake_1_0 = TensorMap('1598_Average-weekly-spirits-intake_1_0', loss='logcosh', channel_map={'no_1598_Average-weekly-spirits-intake_1_0': 0})
field_id_1608_average_weekly_fortified_wine_intake_1_0 = TensorMap('1608_Average-weekly-fortified-wine-intake_1_0', loss='logcosh', channel_map={'no_1608_Average-weekly-fortified-wine-intake_1_0': 0})
field_id_1737_childhood_sunburn_occasions_1_0 = TensorMap('1737_Childhood-sunburn-occasions_1_0', loss='logcosh', channel_map={'no_1737_Childhood-sunburn-occasions_1_0': 0})
field_id_1807_fathers_age_at_death_1_0 = TensorMap('1807_Fathers-age-at-death_1_0', loss='logcosh', channel_map={'no_1807_Fathers-age-at-death_1_0': 0})
field_id_1873_number_of_full_brothers_1_0 = TensorMap('1873_Number-of-full-brothers_1_0', loss='logcosh', channel_map={'no_1873_Number-of-full-brothers_1_0': 0})
field_id_1883_number_of_full_sisters_1_0 = TensorMap('1883_Number-of-full-sisters_1_0', loss='logcosh', channel_map={'no_1883_Number-of-full-sisters_1_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_4 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_4', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_4': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_1_0 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_0', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_1_1 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_1', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_1': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_1_2 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_2', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_2': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_1_0 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_1_0', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_1_0': 0})
field_id_20015_sitting_height_1_0 = TensorMap('20015_Sitting-height_1_0', loss='logcosh', channel_map={'no_20015_Sitting-height_1_0': 0})
field_id_20022_birth_weight_1_0 = TensorMap('20022_Birth-weight_1_0', loss='logcosh', channel_map={'no_20022_Birth-weight_1_0': 0})
field_id_20023_mean_time_to_correctly_identify_matches_1_0 = TensorMap('20023_Mean-time-to-correctly-identify-matches_1_0', loss='logcosh', channel_map={'no_20023_Mean-time-to-correctly-identify-matches_1_0': 0})
field_id_21001_body_mass_index_bmi_1_0 = TensorMap('21001_Body-mass-index-BMI_1_0', loss='logcosh', channel_map={'no_21001_Body-mass-index-BMI_1_0': 0})
field_id_21002_weight_1_0 = TensorMap('21002_Weight_1_0', loss='logcosh', channel_map={'no_21002_Weight_1_0': 0})
field_id_21003_age_when_attended_assessment_centre_1_0 = TensorMap('21003_Age-when-attended-assessment-centre_1_0', loss='logcosh', channel_map={'no_21003_Age-when-attended-assessment-centre_1_0': 0})
field_id_21021_pulse_wave_arterial_stiffness_index_1_0 = TensorMap('21021_Pulse-wave-Arterial-Stiffness-index_1_0', loss='logcosh', channel_map={'no_21021_Pulse-wave-Arterial-Stiffness-index_1_0': 0})
field_id_2217_age_started_wearing_glasses_or_contact_lenses_1_0 = TensorMap('2217_Age-started-wearing-glasses-or-contact-lenses_1_0', loss='logcosh', channel_map={'no_2217_Age-started-wearing-glasses-or-contact-lenses_1_0': 0})
field_id_2277_frequency_of_solariumsunlamp_use_1_0 = TensorMap('2277_Frequency-of-solariumsunlamp-use_1_0', loss='logcosh', channel_map={'no_2277_Frequency-of-solariumsunlamp-use_1_0': 0})
field_id_23098_weight_1_0 = TensorMap('23098_Weight_1_0', loss='logcosh', channel_map={'no_23098_Weight_1_0': 0})
field_id_23099_body_fat_percentage_1_0 = TensorMap('23099_Body-fat-percentage_1_0', loss='logcosh', channel_map={'no_23099_Body-fat-percentage_1_0': 0})
field_id_23100_whole_body_fat_mass_1_0 = TensorMap('23100_Whole-body-fat-mass_1_0', loss='logcosh', channel_map={'no_23100_Whole-body-fat-mass_1_0': 0})
field_id_23101_whole_body_fatfree_mass_1_0 = TensorMap('23101_Whole-body-fatfree-mass_1_0', loss='logcosh', channel_map={'no_23101_Whole-body-fatfree-mass_1_0': 0})
field_id_23102_whole_body_water_mass_1_0 = TensorMap('23102_Whole-body-water-mass_1_0', loss='logcosh', channel_map={'no_23102_Whole-body-water-mass_1_0': 0})
field_id_23104_body_mass_index_bmi_1_0 = TensorMap('23104_Body-mass-index-BMI_1_0', loss='logcosh', channel_map={'no_23104_Body-mass-index-BMI_1_0': 0})
field_id_23105_basal_metabolic_rate_1_0 = TensorMap('23105_Basal-metabolic-rate_1_0', loss='logcosh', channel_map={'no_23105_Basal-metabolic-rate_1_0': 0})
field_id_23106_impedance_of_whole_body_1_0 = TensorMap('23106_Impedance-of-whole-body_1_0', loss='logcosh', channel_map={'no_23106_Impedance-of-whole-body_1_0': 0})
field_id_23107_impedance_of_leg_right_1_0 = TensorMap('23107_Impedance-of-leg-right_1_0', loss='logcosh', channel_map={'no_23107_Impedance-of-leg-right_1_0': 0})
field_id_23108_impedance_of_leg_left_1_0 = TensorMap('23108_Impedance-of-leg-left_1_0', loss='logcosh', channel_map={'no_23108_Impedance-of-leg-left_1_0': 0})
field_id_23109_impedance_of_arm_right_1_0 = TensorMap('23109_Impedance-of-arm-right_1_0', loss='logcosh', channel_map={'no_23109_Impedance-of-arm-right_1_0': 0})
field_id_23110_impedance_of_arm_left_1_0 = TensorMap('23110_Impedance-of-arm-left_1_0', loss='logcosh', channel_map={'no_23110_Impedance-of-arm-left_1_0': 0})
field_id_23111_leg_fat_percentage_right_1_0 = TensorMap('23111_Leg-fat-percentage-right_1_0', loss='logcosh', channel_map={'no_23111_Leg-fat-percentage-right_1_0': 0})
field_id_23112_leg_fat_mass_right_1_0 = TensorMap('23112_Leg-fat-mass-right_1_0', loss='logcosh', channel_map={'no_23112_Leg-fat-mass-right_1_0': 0})
field_id_23113_leg_fatfree_mass_right_1_0 = TensorMap('23113_Leg-fatfree-mass-right_1_0', loss='logcosh', channel_map={'no_23113_Leg-fatfree-mass-right_1_0': 0})
field_id_23114_leg_predicted_mass_right_1_0 = TensorMap('23114_Leg-predicted-mass-right_1_0', loss='logcosh', channel_map={'no_23114_Leg-predicted-mass-right_1_0': 0})
field_id_23115_leg_fat_percentage_left_1_0 = TensorMap('23115_Leg-fat-percentage-left_1_0', loss='logcosh', channel_map={'no_23115_Leg-fat-percentage-left_1_0': 0})
field_id_23116_leg_fat_mass_left_1_0 = TensorMap('23116_Leg-fat-mass-left_1_0', loss='logcosh', channel_map={'no_23116_Leg-fat-mass-left_1_0': 0})
field_id_23117_leg_fatfree_mass_left_1_0 = TensorMap('23117_Leg-fatfree-mass-left_1_0', loss='logcosh', channel_map={'no_23117_Leg-fatfree-mass-left_1_0': 0})
field_id_23118_leg_predicted_mass_left_1_0 = TensorMap('23118_Leg-predicted-mass-left_1_0', loss='logcosh', channel_map={'no_23118_Leg-predicted-mass-left_1_0': 0})
field_id_23119_arm_fat_percentage_right_1_0 = TensorMap('23119_Arm-fat-percentage-right_1_0', loss='logcosh', channel_map={'no_23119_Arm-fat-percentage-right_1_0': 0})
field_id_23120_arm_fat_mass_right_1_0 = TensorMap('23120_Arm-fat-mass-right_1_0', loss='logcosh', channel_map={'no_23120_Arm-fat-mass-right_1_0': 0})
field_id_23121_arm_fatfree_mass_right_1_0 = TensorMap('23121_Arm-fatfree-mass-right_1_0', loss='logcosh', channel_map={'no_23121_Arm-fatfree-mass-right_1_0': 0})
field_id_23122_arm_predicted_mass_right_1_0 = TensorMap('23122_Arm-predicted-mass-right_1_0', loss='logcosh', channel_map={'no_23122_Arm-predicted-mass-right_1_0': 0})
field_id_23123_arm_fat_percentage_left_1_0 = TensorMap('23123_Arm-fat-percentage-left_1_0', loss='logcosh', channel_map={'no_23123_Arm-fat-percentage-left_1_0': 0})
field_id_23124_arm_fat_mass_left_1_0 = TensorMap('23124_Arm-fat-mass-left_1_0', loss='logcosh', channel_map={'no_23124_Arm-fat-mass-left_1_0': 0})
field_id_23125_arm_fatfree_mass_left_1_0 = TensorMap('23125_Arm-fatfree-mass-left_1_0', loss='logcosh', channel_map={'no_23125_Arm-fatfree-mass-left_1_0': 0})
field_id_23126_arm_predicted_mass_left_1_0 = TensorMap('23126_Arm-predicted-mass-left_1_0', loss='logcosh', channel_map={'no_23126_Arm-predicted-mass-left_1_0': 0})
field_id_23127_trunk_fat_percentage_1_0 = TensorMap('23127_Trunk-fat-percentage_1_0', loss='logcosh', channel_map={'no_23127_Trunk-fat-percentage_1_0': 0})
field_id_23128_trunk_fat_mass_1_0 = TensorMap('23128_Trunk-fat-mass_1_0', loss='logcosh', channel_map={'no_23128_Trunk-fat-mass_1_0': 0})
field_id_23129_trunk_fatfree_mass_1_0 = TensorMap('23129_Trunk-fatfree-mass_1_0', loss='logcosh', channel_map={'no_23129_Trunk-fatfree-mass_1_0': 0})
field_id_23130_trunk_predicted_mass_1_0 = TensorMap('23130_Trunk-predicted-mass_1_0', loss='logcosh', channel_map={'no_23130_Trunk-predicted-mass_1_0': 0})
field_id_2355_most_recent_bowel_cancer_screening_1_0 = TensorMap('2355_Most-recent-bowel-cancer-screening_1_0', loss='logcosh', channel_map={'no_2355_Most-recent-bowel-cancer-screening_1_0': 0})
field_id_2405_number_of_children_fathered_1_0 = TensorMap('2405_Number-of-children-fathered_1_0', loss='logcosh', channel_map={'no_2405_Number-of-children-fathered_1_0': 0})
field_id_2966_age_high_blood_pressure_diagnosed_0_0 = TensorMap('2966_Age-high-blood-pressure-diagnosed_0_0', loss='logcosh', channel_map={'no_2966_Age-high-blood-pressure-diagnosed_0_0': 0})
field_id_30000_white_blood_cell_leukocyte_count_1_0 = TensorMap('30000_White-blood-cell-leukocyte-count_1_0', loss='logcosh', channel_map={'no_30000_White-blood-cell-leukocyte-count_1_0': 0})
field_id_30010_red_blood_cell_erythrocyte_count_1_0 = TensorMap('30010_Red-blood-cell-erythrocyte-count_1_0', loss='logcosh', channel_map={'no_30010_Red-blood-cell-erythrocyte-count_1_0': 0})
field_id_30020_haemoglobin_concentration_1_0 = TensorMap('30020_Haemoglobin-concentration_1_0', loss='logcosh', channel_map={'no_30020_Haemoglobin-concentration_1_0': 0})
field_id_30030_haematocrit_percentage_1_0 = TensorMap('30030_Haematocrit-percentage_1_0', loss='logcosh', channel_map={'no_30030_Haematocrit-percentage_1_0': 0})
field_id_30040_mean_corpuscular_volume_1_0 = TensorMap('30040_Mean-corpuscular-volume_1_0', loss='logcosh', channel_map={'no_30040_Mean-corpuscular-volume_1_0': 0})
field_id_30050_mean_corpuscular_haemoglobin_1_0 = TensorMap('30050_Mean-corpuscular-haemoglobin_1_0', loss='logcosh', channel_map={'no_30050_Mean-corpuscular-haemoglobin_1_0': 0})
field_id_30060_mean_corpuscular_haemoglobin_concentration_1_0 = TensorMap('30060_Mean-corpuscular-haemoglobin-concentration_1_0', loss='logcosh', channel_map={'no_30060_Mean-corpuscular-haemoglobin-concentration_1_0': 0})
field_id_30070_red_blood_cell_erythrocyte_distribution_width_1_0 = TensorMap('30070_Red-blood-cell-erythrocyte-distribution-width_1_0', loss='logcosh', channel_map={'no_30070_Red-blood-cell-erythrocyte-distribution-width_1_0': 0})
field_id_30080_platelet_count_1_0 = TensorMap('30080_Platelet-count_1_0', loss='logcosh', channel_map={'no_30080_Platelet-count_1_0': 0})
field_id_30090_platelet_crit_1_0 = TensorMap('30090_Platelet-crit_1_0', loss='logcosh', channel_map={'no_30090_Platelet-crit_1_0': 0})
field_id_30100_mean_platelet_thrombocyte_volume_1_0 = TensorMap('30100_Mean-platelet-thrombocyte-volume_1_0', loss='logcosh', channel_map={'no_30100_Mean-platelet-thrombocyte-volume_1_0': 0})
field_id_30110_platelet_distribution_width_1_0 = TensorMap('30110_Platelet-distribution-width_1_0', loss='logcosh', channel_map={'no_30110_Platelet-distribution-width_1_0': 0})
field_id_30120_lymphocyte_count_1_0 = TensorMap('30120_Lymphocyte-count_1_0', loss='logcosh', channel_map={'no_30120_Lymphocyte-count_1_0': 0})
field_id_30130_monocyte_count_1_0 = TensorMap('30130_Monocyte-count_1_0', loss='logcosh', channel_map={'no_30130_Monocyte-count_1_0': 0})
field_id_30140_neutrophill_count_1_0 = TensorMap('30140_Neutrophill-count_1_0', loss='logcosh', channel_map={'no_30140_Neutrophill-count_1_0': 0})
field_id_30150_eosinophill_count_1_0 = TensorMap('30150_Eosinophill-count_1_0', loss='logcosh', channel_map={'no_30150_Eosinophill-count_1_0': 0})
field_id_30160_basophill_count_1_0 = TensorMap('30160_Basophill-count_1_0', loss='logcosh', channel_map={'no_30160_Basophill-count_1_0': 0})
field_id_30170_nucleated_red_blood_cell_count_1_0 = TensorMap('30170_Nucleated-red-blood-cell-count_1_0', loss='logcosh', channel_map={'no_30170_Nucleated-red-blood-cell-count_1_0': 0})
field_id_30180_lymphocyte_percentage_1_0 = TensorMap('30180_Lymphocyte-percentage_1_0', loss='logcosh', channel_map={'no_30180_Lymphocyte-percentage_1_0': 0})
field_id_30190_monocyte_percentage_1_0 = TensorMap('30190_Monocyte-percentage_1_0', loss='logcosh', channel_map={'no_30190_Monocyte-percentage_1_0': 0})
field_id_30200_neutrophill_percentage_1_0 = TensorMap('30200_Neutrophill-percentage_1_0', loss='logcosh', channel_map={'no_30200_Neutrophill-percentage_1_0': 0})
field_id_30210_eosinophill_percentage_1_0 = TensorMap('30210_Eosinophill-percentage_1_0', loss='logcosh', channel_map={'no_30210_Eosinophill-percentage_1_0': 0})
field_id_30220_basophill_percentage_1_0 = TensorMap('30220_Basophill-percentage_1_0', loss='logcosh', channel_map={'no_30220_Basophill-percentage_1_0': 0})
field_id_30230_nucleated_red_blood_cell_percentage_1_0 = TensorMap('30230_Nucleated-red-blood-cell-percentage_1_0', loss='logcosh', channel_map={'no_30230_Nucleated-red-blood-cell-percentage_1_0': 0})
field_id_30240_reticulocyte_percentage_1_0 = TensorMap('30240_Reticulocyte-percentage_1_0', loss='logcosh', channel_map={'no_30240_Reticulocyte-percentage_1_0': 0})
field_id_30250_reticulocyte_count_1_0 = TensorMap('30250_Reticulocyte-count_1_0', loss='logcosh', channel_map={'no_30250_Reticulocyte-count_1_0': 0})
field_id_30260_mean_reticulocyte_volume_1_0 = TensorMap('30260_Mean-reticulocyte-volume_1_0', loss='logcosh', channel_map={'no_30260_Mean-reticulocyte-volume_1_0': 0})
field_id_30270_mean_sphered_cell_volume_1_0 = TensorMap('30270_Mean-sphered-cell-volume_1_0', loss='logcosh', channel_map={'no_30270_Mean-sphered-cell-volume_1_0': 0})
field_id_30280_immature_reticulocyte_fraction_1_0 = TensorMap('30280_Immature-reticulocyte-fraction_1_0', loss='logcosh', channel_map={'no_30280_Immature-reticulocyte-fraction_1_0': 0})
field_id_30290_high_light_scatter_reticulocyte_percentage_1_0 = TensorMap('30290_High-light-scatter-reticulocyte-percentage_1_0', loss='logcosh', channel_map={'no_30290_High-light-scatter-reticulocyte-percentage_1_0': 0})
field_id_30300_high_light_scatter_reticulocyte_count_1_0 = TensorMap('30300_High-light-scatter-reticulocyte-count_1_0', loss='logcosh', channel_map={'no_30300_High-light-scatter-reticulocyte-count_1_0': 0})
field_id_30314_volume_of_edta1_plasma_held_by_ukb_1_0 = TensorMap('30314_Volume-of-EDTA1-plasma-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30314_Volume-of-EDTA1-plasma-held-by-UKB_1_0': 0})
field_id_30324_volume_of_edta1_red_cells_held_by_ukb_1_0 = TensorMap('30324_Volume-of-EDTA1-red-cells-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30324_Volume-of-EDTA1-red-cells-held-by-UKB_1_0': 0})
field_id_30334_volume_of_edta1_buffy_held_by_ukb_1_0 = TensorMap('30334_Volume-of-EDTA1-buffy-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30334_Volume-of-EDTA1-buffy-held-by-UKB_1_0': 0})
field_id_30344_volume_of_edta2_plasma_held_by_ukb_1_0 = TensorMap('30344_Volume-of-EDTA2-plasma-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30344_Volume-of-EDTA2-plasma-held-by-UKB_1_0': 0})
field_id_30354_volume_of_edta2_buffy_held_by_ukb_1_0 = TensorMap('30354_Volume-of-EDTA2-buffy-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30354_Volume-of-EDTA2-buffy-held-by-UKB_1_0': 0})
field_id_30364_volume_of_edta2_red_cells_held_by_ukb_1_0 = TensorMap('30364_Volume-of-EDTA2-red-cells-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30364_Volume-of-EDTA2-red-cells-held-by-UKB_1_0': 0})
field_id_30374_volume_of_lihep_plasma_held_by_ukb_1_0 = TensorMap('30374_Volume-of-LiHep-plasma-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30374_Volume-of-LiHep-plasma-held-by-UKB_1_0': 0})
field_id_30384_volume_of_serum_held_by_ukb_1_0 = TensorMap('30384_Volume-of-serum-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30384_Volume-of-serum-held-by-UKB_1_0': 0})
field_id_30404_volume_of_acd_held_by_ukb_1_0 = TensorMap('30404_Volume-of-ACD-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30404_Volume-of-ACD-held-by-UKB_1_0': 0})
field_id_30414_volume_of_rna_held_by_ukb_1_0 = TensorMap('30414_Volume-of-RNA-held-by-UKB_1_0', loss='logcosh', channel_map={'no_30414_Volume-of-RNA-held-by-UKB_1_0': 0})
field_id_30510_creatinine_enzymatic_in_urine_1_0 = TensorMap('30510_Creatinine-enzymatic-in-urine_1_0', loss='logcosh', channel_map={'no_30510_Creatinine-enzymatic-in-urine_1_0': 0})
field_id_30520_potassium_in_urine_1_0 = TensorMap('30520_Potassium-in-urine_1_0', loss='logcosh', channel_map={'no_30520_Potassium-in-urine_1_0': 0})
field_id_30530_sodium_in_urine_1_0 = TensorMap('30530_Sodium-in-urine_1_0', loss='logcosh', channel_map={'no_30530_Sodium-in-urine_1_0': 0})
field_id_3062_forced_vital_capacity_fvc_1_0 = TensorMap('3062_Forced-vital-capacity-FVC_1_0', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_1_0': 0})
field_id_3062_forced_vital_capacity_fvc_1_1 = TensorMap('3062_Forced-vital-capacity-FVC_1_1', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_1_1': 0})
field_id_30630_apolipoprotein_a_1_0 = TensorMap('30630_Apolipoprotein-A_1_0', loss='logcosh', channel_map={'no_30630_Apolipoprotein-A_1_0': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_1_0 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_1_0', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_1_0': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_1_1 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_1_1', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_1_1': 0})
field_id_30640_apolipoprotein_b_1_0 = TensorMap('30640_Apolipoprotein-B_1_0', loss='logcosh', channel_map={'no_30640_Apolipoprotein-B_1_0': 0})
field_id_3064_peak_expiratory_flow_pef_1_0 = TensorMap('3064_Peak-expiratory-flow-PEF_1_0', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_1_0': 0})
field_id_3064_peak_expiratory_flow_pef_1_1 = TensorMap('3064_Peak-expiratory-flow-PEF_1_1', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_1_1': 0})
field_id_30660_direct_bilirubin_1_0 = TensorMap('30660_Direct-bilirubin_1_0', loss='logcosh', channel_map={'no_30660_Direct-bilirubin_1_0': 0})
field_id_30670_urea_1_0 = TensorMap('30670_Urea_1_0', loss='logcosh', channel_map={'no_30670_Urea_1_0': 0})
field_id_30680_calcium_1_0 = TensorMap('30680_Calcium_1_0', loss='logcosh', channel_map={'no_30680_Calcium_1_0': 0})
field_id_30690_cholesterol_1_0 = TensorMap('30690_Cholesterol_1_0', loss='logcosh', channel_map={'no_30690_Cholesterol_1_0': 0})
field_id_30710_creactive_protein_1_0 = TensorMap('30710_Creactive-protein_1_0', loss='logcosh', channel_map={'no_30710_Creactive-protein_1_0': 0})
field_id_30740_glucose_1_0 = TensorMap('30740_Glucose_1_0', loss='logcosh', channel_map={'no_30740_Glucose_1_0': 0})
field_id_30760_hdl_cholesterol_1_0 = TensorMap('30760_HDL-cholesterol_1_0', loss='logcosh', channel_map={'no_30760_HDL-cholesterol_1_0': 0})
field_id_30770_igf1_1_0 = TensorMap('30770_IGF1_1_0', loss='logcosh', channel_map={'no_30770_IGF1_1_0': 0})
field_id_30790_lipoprotein_a_1_0 = TensorMap('30790_Lipoprotein-A_1_0', loss='logcosh', channel_map={'no_30790_Lipoprotein-A_1_0': 0})
field_id_30810_phosphate_1_0 = TensorMap('30810_Phosphate_1_0', loss='logcosh', channel_map={'no_30810_Phosphate_1_0': 0})
field_id_3083_heel_quantitative_ultrasound_index_qui_manual_entry_0_0 = TensorMap('3083_Heel-quantitative-ultrasound-index-QUI-manual-entry_0_0', loss='logcosh', channel_map={'no_3083_Heel-quantitative-ultrasound-index-QUI-manual-entry_0_0': 0})
field_id_30840_total_bilirubin_1_0 = TensorMap('30840_Total-bilirubin_1_0', loss='logcosh', channel_map={'no_30840_Total-bilirubin_1_0': 0})
field_id_3084_heel_bone_mineral_density_bmd_manual_entry_0_0 = TensorMap('3084_Heel-bone-mineral-density-BMD-manual-entry_0_0', loss='logcosh', channel_map={'no_3084_Heel-bone-mineral-density-BMD-manual-entry_0_0': 0})
field_id_30850_testosterone_1_0 = TensorMap('30850_Testosterone_1_0', loss='logcosh', channel_map={'no_30850_Testosterone_1_0': 0})
field_id_3085_heel_broadband_ultrasound_attenuation_bua_manual_entry_0_0 = TensorMap('3085_Heel-Broadband-ultrasound-attenuation-BUA-manual-entry_0_0', loss='logcosh', channel_map={'no_3085_Heel-Broadband-ultrasound-attenuation-BUA-manual-entry_0_0': 0})
field_id_30860_total_protein_1_0 = TensorMap('30860_Total-protein_1_0', loss='logcosh', channel_map={'no_30860_Total-protein_1_0': 0})
field_id_3086_speed_of_sound_through_heel_manual_entry_0_0 = TensorMap('3086_Speed-of-sound-through-heel-manual-entry_0_0', loss='logcosh', channel_map={'no_3086_Speed-of-sound-through-heel-manual-entry_0_0': 0})
field_id_30870_triglycerides_1_0 = TensorMap('30870_Triglycerides_1_0', loss='logcosh', channel_map={'no_30870_Triglycerides_1_0': 0})
field_id_30880_urate_1_0 = TensorMap('30880_Urate_1_0', loss='logcosh', channel_map={'no_30880_Urate_1_0': 0})
field_id_30890_vitamin_d_1_0 = TensorMap('30890_Vitamin-D_1_0', loss='logcosh', channel_map={'no_30890_Vitamin-D_1_0': 0})
field_id_3526_mothers_age_at_death_1_0 = TensorMap('3526_Mothers-age-at-death_1_0', loss='logcosh', channel_map={'no_3526_Mothers-age-at-death_1_0': 0})
field_id_3627_age_angina_diagnosed_0_0 = TensorMap('3627_Age-angina-diagnosed_0_0', loss='logcosh', channel_map={'no_3627_Age-angina-diagnosed_0_0': 0})
field_id_396_number_of_columns_displayed_in_round_1_1 = TensorMap('396_Number-of-columns-displayed-in-round_1_1', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_1_1': 0})
field_id_396_number_of_columns_displayed_in_round_1_2 = TensorMap('396_Number-of-columns-displayed-in-round_1_2', loss='logcosh', channel_map={'no_396_Number-of-columns-displayed-in-round_1_2': 0})
field_id_397_number_of_rows_displayed_in_round_1_1 = TensorMap('397_Number-of-rows-displayed-in-round_1_1', loss='logcosh', channel_map={'no_397_Number-of-rows-displayed-in-round_1_1': 0})
field_id_397_number_of_rows_displayed_in_round_1_2 = TensorMap('397_Number-of-rows-displayed-in-round_1_2', loss='logcosh', channel_map={'no_397_Number-of-rows-displayed-in-round_1_2': 0})
field_id_398_number_of_correct_matches_in_round_1_1 = TensorMap('398_Number-of-correct-matches-in-round_1_1', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_1_1': 0})
field_id_398_number_of_correct_matches_in_round_1_2 = TensorMap('398_Number-of-correct-matches-in-round_1_2', loss='logcosh', channel_map={'no_398_Number-of-correct-matches-in-round_1_2': 0})
field_id_399_number_of_incorrect_matches_in_round_1_1 = TensorMap('399_Number-of-incorrect-matches-in-round_1_1', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_1_1': 0})
field_id_399_number_of_incorrect_matches_in_round_1_2 = TensorMap('399_Number-of-incorrect-matches-in-round_1_2', loss='logcosh', channel_map={'no_399_Number-of-incorrect-matches-in-round_1_2': 0})
field_id_40008_age_at_cancer_diagnosis_0_0 = TensorMap('40008_Age-at-cancer-diagnosis_0_0', loss='logcosh', channel_map={'no_40008_Age-at-cancer-diagnosis_0_0': 0})
field_id_40009_reported_occurrences_of_cancer_0_0 = TensorMap('40009_Reported-occurrences-of-cancer_0_0', loss='logcosh', channel_map={'no_40009_Reported-occurrences-of-cancer_0_0': 0})
field_id_400_time_to_complete_round_1_1 = TensorMap('400_Time-to-complete-round_1_1', loss='logcosh', channel_map={'no_400_Time-to-complete-round_1_1': 0})
field_id_400_time_to_complete_round_1_2 = TensorMap('400_Time-to-complete-round_1_2', loss='logcosh', channel_map={'no_400_Time-to-complete-round_1_2': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_0 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_0', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_0': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_1 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_1', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_1': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_10 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_10', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_10': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_11 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_11', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_11': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_5 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_5', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_5': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_7 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_7', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_7': 0})
field_id_4079_diastolic_blood_pressure_automated_reading_1_0 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_1_0', loss='logcosh', channel_map={'no_4079_Diastolic-blood-pressure-automated-reading_1_0': 0})
field_id_4079_diastolic_blood_pressure_automated_reading_1_1 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_1_1', loss='logcosh', channel_map={'no_4079_Diastolic-blood-pressure-automated-reading_1_1': 0})
field_id_4080_systolic_blood_pressure_automated_reading_1_0 = TensorMap('4080_Systolic-blood-pressure-automated-reading_1_0', loss='logcosh', channel_map={'no_4080_Systolic-blood-pressure-automated-reading_1_0': 0})
field_id_4080_systolic_blood_pressure_automated_reading_1_1 = TensorMap('4080_Systolic-blood-pressure-automated-reading_1_1', loss='logcosh', channel_map={'no_4080_Systolic-blood-pressure-automated-reading_1_1': 0})
field_id_4100_ankle_spacing_width_left_1_0 = TensorMap('4100_Ankle-spacing-width-left_1_0', loss='logcosh', channel_map={'no_4100_Ankle-spacing-width-left_1_0': 0})
field_id_4101_heel_broadband_ultrasound_attenuation_left_1_0 = TensorMap('4101_Heel-broadband-ultrasound-attenuation-left_1_0', loss='logcosh', channel_map={'no_4101_Heel-broadband-ultrasound-attenuation-left_1_0': 0})
field_id_4103_speed_of_sound_through_heel_left_1_0 = TensorMap('4103_Speed-of-sound-through-heel-left_1_0', loss='logcosh', channel_map={'no_4103_Speed-of-sound-through-heel-left_1_0': 0})
field_id_4104_heel_quantitative_ultrasound_index_qui_direct_entry_left_1_0 = TensorMap('4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_1_0', loss='logcosh', channel_map={'no_4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_1_0': 0})
field_id_4105_heel_bone_mineral_density_bmd_left_1_0 = TensorMap('4105_Heel-bone-mineral-density-BMD-left_1_0', loss='logcosh', channel_map={'no_4105_Heel-bone-mineral-density-BMD-left_1_0': 0})
field_id_4106_heel_bone_mineral_density_bmd_tscore_automated_left_1_0 = TensorMap('4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_1_0', loss='logcosh', channel_map={'no_4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_1_0': 0})
field_id_4119_ankle_spacing_width_right_1_0 = TensorMap('4119_Ankle-spacing-width-right_1_0', loss='logcosh', channel_map={'no_4119_Ankle-spacing-width-right_1_0': 0})
field_id_4120_heel_broadband_ultrasound_attenuation_right_1_0 = TensorMap('4120_Heel-broadband-ultrasound-attenuation-right_1_0', loss='logcosh', channel_map={'no_4120_Heel-broadband-ultrasound-attenuation-right_1_0': 0})
field_id_4122_speed_of_sound_through_heel_right_1_0 = TensorMap('4122_Speed-of-sound-through-heel-right_1_0', loss='logcosh', channel_map={'no_4122_Speed-of-sound-through-heel-right_1_0': 0})
field_id_4123_heel_quantitative_ultrasound_index_qui_direct_entry_right_1_0 = TensorMap('4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_1_0', loss='logcosh', channel_map={'no_4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_1_0': 0})
field_id_4124_heel_bone_mineral_density_bmd_right_1_0 = TensorMap('4124_Heel-bone-mineral-density-BMD-right_1_0', loss='logcosh', channel_map={'no_4124_Heel-bone-mineral-density-BMD-right_1_0': 0})
field_id_4125_heel_bone_mineral_density_bmd_tscore_automated_right_1_0 = TensorMap('4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_1_0', loss='logcosh', channel_map={'no_4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_1_0': 0})
field_id_4194_pulse_rate_1_0 = TensorMap('4194_Pulse-rate_1_0', loss='logcosh', channel_map={'no_4194_Pulse-rate_1_0': 0})
field_id_4195_pulse_wave_reflection_index_1_0 = TensorMap('4195_Pulse-wave-reflection-index_1_0', loss='logcosh', channel_map={'no_4195_Pulse-wave-reflection-index_1_0': 0})
field_id_4196_pulse_wave_peak_to_peak_time_1_0 = TensorMap('4196_Pulse-wave-peak-to-peak-time_1_0', loss='logcosh', channel_map={'no_4196_Pulse-wave-peak-to-peak-time_1_0': 0})
field_id_4198_position_of_the_pulse_wave_peak_1_0 = TensorMap('4198_Position-of-the-pulse-wave-peak_1_0', loss='logcosh', channel_map={'no_4198_Position-of-the-pulse-wave-peak_1_0': 0})
field_id_4199_position_of_pulse_wave_notch_1_0 = TensorMap('4199_Position-of-pulse-wave-notch_1_0', loss='logcosh', channel_map={'no_4199_Position-of-pulse-wave-notch_1_0': 0})
field_id_4200_position_of_the_shoulder_on_the_pulse_waveform_1_0 = TensorMap('4200_Position-of-the-shoulder-on-the-pulse-waveform_1_0', loss='logcosh', channel_map={'no_4200_Position-of-the-shoulder-on-the-pulse-waveform_1_0': 0})
field_id_4288_time_to_answer_1_0 = TensorMap('4288_Time-to-answer_1_0', loss='logcosh', channel_map={'no_4288_Time-to-answer_1_0': 0})
field_id_4290_duration_screen_displayed_1_0 = TensorMap('4290_Duration-screen-displayed_1_0', loss='logcosh', channel_map={'no_4290_Duration-screen-displayed_1_0': 0})
field_id_4291_number_of_attempts_1_0 = TensorMap('4291_Number-of-attempts_1_0', loss='logcosh', channel_map={'no_4291_Number-of-attempts_1_0': 0})
field_id_46_hand_grip_strength_left_1_0 = TensorMap('46_Hand-grip-strength-left_1_0', loss='logcosh', channel_map={'no_46_Hand-grip-strength-left_1_0': 0})
field_id_47_hand_grip_strength_right_1_0 = TensorMap('47_Hand-grip-strength-right_1_0', loss='logcosh', channel_map={'no_47_Hand-grip-strength-right_1_0': 0})
field_id_48_waist_circumference_1_0 = TensorMap('48_Waist-circumference_1_0', loss='logcosh', channel_map={'no_48_Waist-circumference_1_0': 0})
field_id_49_hip_circumference_1_0 = TensorMap('49_Hip-circumference_1_0', loss='logcosh', channel_map={'no_49_Hip-circumference_1_0': 0})
field_id_5057_number_of_older_siblings_1_0 = TensorMap('5057_Number-of-older-siblings_1_0', loss='logcosh', channel_map={'no_5057_Number-of-older-siblings_1_0': 0})
field_id_5074_number_of_letters_shown_in_round_left_1_0 = TensorMap('5074_Number-of-letters-shown-in-round-left_1_0', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_1_0': 0})
field_id_5074_number_of_letters_shown_in_round_left_1_1 = TensorMap('5074_Number-of-letters-shown-in-round-left_1_1', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_1_1': 0})
field_id_5074_number_of_letters_shown_in_round_left_1_2 = TensorMap('5074_Number-of-letters-shown-in-round-left_1_2', loss='logcosh', channel_map={'no_5074_Number-of-letters-shown-in-round-left_1_2': 0})
field_id_5075_number_of_letters_shown_in_round_right_1_0 = TensorMap('5075_Number-of-letters-shown-in-round-right_1_0', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_1_0': 0})
field_id_5075_number_of_letters_shown_in_round_right_1_1 = TensorMap('5075_Number-of-letters-shown-in-round-right_1_1', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_1_1': 0})
field_id_5075_number_of_letters_shown_in_round_right_1_2 = TensorMap('5075_Number-of-letters-shown-in-round-right_1_2', loss='logcosh', channel_map={'no_5075_Number-of-letters-shown-in-round-right_1_2': 0})
field_id_5076_number_of_letters_correct_in_round_right_1_0 = TensorMap('5076_Number-of-letters-correct-in-round-right_1_0', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_1_0': 0})
field_id_5076_number_of_letters_correct_in_round_right_1_1 = TensorMap('5076_Number-of-letters-correct-in-round-right_1_1', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_1_1': 0})
field_id_5076_number_of_letters_correct_in_round_right_1_2 = TensorMap('5076_Number-of-letters-correct-in-round-right_1_2', loss='logcosh', channel_map={'no_5076_Number-of-letters-correct-in-round-right_1_2': 0})
field_id_5077_number_of_letters_correct_in_round_left_1_0 = TensorMap('5077_Number-of-letters-correct-in-round-left_1_0', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_1_0': 0})
field_id_5077_number_of_letters_correct_in_round_left_1_1 = TensorMap('5077_Number-of-letters-correct-in-round-left_1_1', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_1_1': 0})
field_id_5077_number_of_letters_correct_in_round_left_1_2 = TensorMap('5077_Number-of-letters-correct-in-round-left_1_2', loss='logcosh', channel_map={'no_5077_Number-of-letters-correct-in-round-left_1_2': 0})
field_id_5078_logmar_in_round_left_1_0 = TensorMap('5078_logMAR-in-round-left_1_0', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_1_0': 0})
field_id_5078_logmar_in_round_left_1_1 = TensorMap('5078_logMAR-in-round-left_1_1', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_1_1': 0})
field_id_5078_logmar_in_round_left_1_2 = TensorMap('5078_logMAR-in-round-left_1_2', loss='logcosh', channel_map={'no_5078_logMAR-in-round-left_1_2': 0})
field_id_5079_logmar_in_round_right_1_0 = TensorMap('5079_logMAR-in-round-right_1_0', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_1_0': 0})
field_id_5079_logmar_in_round_right_1_1 = TensorMap('5079_logMAR-in-round-right_1_1', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_1_1': 0})
field_id_5079_logmar_in_round_right_1_2 = TensorMap('5079_logMAR-in-round-right_1_2', loss='logcosh', channel_map={'no_5079_logMAR-in-round-right_1_2': 0})
field_id_5084_spherical_power_right_1_0 = TensorMap('5084_Spherical-power-right_1_0', loss='logcosh', channel_map={'no_5084_Spherical-power-right_1_0': 0})
field_id_5084_spherical_power_right_1_1 = TensorMap('5084_Spherical-power-right_1_1', loss='logcosh', channel_map={'no_5084_Spherical-power-right_1_1': 0})
field_id_5084_spherical_power_right_1_2 = TensorMap('5084_Spherical-power-right_1_2', loss='logcosh', channel_map={'no_5084_Spherical-power-right_1_2': 0})
field_id_5085_spherical_power_left_1_0 = TensorMap('5085_Spherical-power-left_1_0', loss='logcosh', channel_map={'no_5085_Spherical-power-left_1_0': 0})
field_id_5085_spherical_power_left_1_1 = TensorMap('5085_Spherical-power-left_1_1', loss='logcosh', channel_map={'no_5085_Spherical-power-left_1_1': 0})
field_id_5085_spherical_power_left_1_2 = TensorMap('5085_Spherical-power-left_1_2', loss='logcosh', channel_map={'no_5085_Spherical-power-left_1_2': 0})
field_id_5086_cylindrical_power_left_1_0 = TensorMap('5086_Cylindrical-power-left_1_0', loss='logcosh', channel_map={'no_5086_Cylindrical-power-left_1_0': 0})
field_id_5086_cylindrical_power_left_1_1 = TensorMap('5086_Cylindrical-power-left_1_1', loss='logcosh', channel_map={'no_5086_Cylindrical-power-left_1_1': 0})
field_id_5086_cylindrical_power_left_1_2 = TensorMap('5086_Cylindrical-power-left_1_2', loss='logcosh', channel_map={'no_5086_Cylindrical-power-left_1_2': 0})
field_id_5087_cylindrical_power_right_1_0 = TensorMap('5087_Cylindrical-power-right_1_0', loss='logcosh', channel_map={'no_5087_Cylindrical-power-right_1_0': 0})
field_id_5087_cylindrical_power_right_1_1 = TensorMap('5087_Cylindrical-power-right_1_1', loss='logcosh', channel_map={'no_5087_Cylindrical-power-right_1_1': 0})
field_id_5087_cylindrical_power_right_1_2 = TensorMap('5087_Cylindrical-power-right_1_2', loss='logcosh', channel_map={'no_5087_Cylindrical-power-right_1_2': 0})
field_id_5088_astigmatism_angle_right_1_0 = TensorMap('5088_Astigmatism-angle-right_1_0', loss='logcosh', channel_map={'no_5088_Astigmatism-angle-right_1_0': 0})
field_id_5088_astigmatism_angle_right_1_1 = TensorMap('5088_Astigmatism-angle-right_1_1', loss='logcosh', channel_map={'no_5088_Astigmatism-angle-right_1_1': 0})
field_id_5088_astigmatism_angle_right_1_2 = TensorMap('5088_Astigmatism-angle-right_1_2', loss='logcosh', channel_map={'no_5088_Astigmatism-angle-right_1_2': 0})
field_id_5089_astigmatism_angle_left_1_0 = TensorMap('5089_Astigmatism-angle-left_1_0', loss='logcosh', channel_map={'no_5089_Astigmatism-angle-left_1_0': 0})
field_id_5089_astigmatism_angle_left_1_1 = TensorMap('5089_Astigmatism-angle-left_1_1', loss='logcosh', channel_map={'no_5089_Astigmatism-angle-left_1_1': 0})
field_id_5089_astigmatism_angle_left_1_2 = TensorMap('5089_Astigmatism-angle-left_1_2', loss='logcosh', channel_map={'no_5089_Astigmatism-angle-left_1_2': 0})
field_id_5096_3mm_weak_meridian_left_1_0 = TensorMap('5096_3mm-weak-meridian-left_1_0', loss='logcosh', channel_map={'no_5096_3mm-weak-meridian-left_1_0': 0})
field_id_5096_3mm_weak_meridian_left_1_1 = TensorMap('5096_3mm-weak-meridian-left_1_1', loss='logcosh', channel_map={'no_5096_3mm-weak-meridian-left_1_1': 0})
field_id_5096_3mm_weak_meridian_left_1_2 = TensorMap('5096_3mm-weak-meridian-left_1_2', loss='logcosh', channel_map={'no_5096_3mm-weak-meridian-left_1_2': 0})
field_id_5097_6mm_weak_meridian_left_1_0 = TensorMap('5097_6mm-weak-meridian-left_1_0', loss='logcosh', channel_map={'no_5097_6mm-weak-meridian-left_1_0': 0})
field_id_5097_6mm_weak_meridian_left_1_1 = TensorMap('5097_6mm-weak-meridian-left_1_1', loss='logcosh', channel_map={'no_5097_6mm-weak-meridian-left_1_1': 0})
field_id_5097_6mm_weak_meridian_left_1_2 = TensorMap('5097_6mm-weak-meridian-left_1_2', loss='logcosh', channel_map={'no_5097_6mm-weak-meridian-left_1_2': 0})
field_id_5099_3mm_weak_meridian_right_1_0 = TensorMap('5099_3mm-weak-meridian-right_1_0', loss='logcosh', channel_map={'no_5099_3mm-weak-meridian-right_1_0': 0})
field_id_5099_3mm_weak_meridian_right_1_1 = TensorMap('5099_3mm-weak-meridian-right_1_1', loss='logcosh', channel_map={'no_5099_3mm-weak-meridian-right_1_1': 0})
field_id_50_standing_height_1_0 = TensorMap('50_Standing-height_1_0', loss='logcosh', channel_map={'no_50_Standing-height_1_0': 0})
field_id_5100_3mm_weak_meridian_angle_right_1_0 = TensorMap('5100_3mm-weak-meridian-angle-right_1_0', loss='logcosh', channel_map={'no_5100_3mm-weak-meridian-angle-right_1_0': 0})
field_id_5100_3mm_weak_meridian_angle_right_1_1 = TensorMap('5100_3mm-weak-meridian-angle-right_1_1', loss='logcosh', channel_map={'no_5100_3mm-weak-meridian-angle-right_1_1': 0})
field_id_5102_6mm_weak_meridian_angle_left_1_0 = TensorMap('5102_6mm-weak-meridian-angle-left_1_0', loss='logcosh', channel_map={'no_5102_6mm-weak-meridian-angle-left_1_0': 0})
field_id_5102_6mm_weak_meridian_angle_left_1_1 = TensorMap('5102_6mm-weak-meridian-angle-left_1_1', loss='logcosh', channel_map={'no_5102_6mm-weak-meridian-angle-left_1_1': 0})
field_id_5102_6mm_weak_meridian_angle_left_1_2 = TensorMap('5102_6mm-weak-meridian-angle-left_1_2', loss='logcosh', channel_map={'no_5102_6mm-weak-meridian-angle-left_1_2': 0})
field_id_5103_3mm_weak_meridian_angle_left_1_0 = TensorMap('5103_3mm-weak-meridian-angle-left_1_0', loss='logcosh', channel_map={'no_5103_3mm-weak-meridian-angle-left_1_0': 0})
field_id_5103_3mm_weak_meridian_angle_left_1_1 = TensorMap('5103_3mm-weak-meridian-angle-left_1_1', loss='logcosh', channel_map={'no_5103_3mm-weak-meridian-angle-left_1_1': 0})
field_id_5103_3mm_weak_meridian_angle_left_1_2 = TensorMap('5103_3mm-weak-meridian-angle-left_1_2', loss='logcosh', channel_map={'no_5103_3mm-weak-meridian-angle-left_1_2': 0})
field_id_5104_3mm_strong_meridian_angle_left_1_0 = TensorMap('5104_3mm-strong-meridian-angle-left_1_0', loss='logcosh', channel_map={'no_5104_3mm-strong-meridian-angle-left_1_0': 0})
field_id_5104_3mm_strong_meridian_angle_left_1_1 = TensorMap('5104_3mm-strong-meridian-angle-left_1_1', loss='logcosh', channel_map={'no_5104_3mm-strong-meridian-angle-left_1_1': 0})
field_id_5104_3mm_strong_meridian_angle_left_1_2 = TensorMap('5104_3mm-strong-meridian-angle-left_1_2', loss='logcosh', channel_map={'no_5104_3mm-strong-meridian-angle-left_1_2': 0})
field_id_5105_6mm_strong_meridian_angle_left_1_0 = TensorMap('5105_6mm-strong-meridian-angle-left_1_0', loss='logcosh', channel_map={'no_5105_6mm-strong-meridian-angle-left_1_0': 0})
field_id_5105_6mm_strong_meridian_angle_left_1_1 = TensorMap('5105_6mm-strong-meridian-angle-left_1_1', loss='logcosh', channel_map={'no_5105_6mm-strong-meridian-angle-left_1_1': 0})
field_id_5105_6mm_strong_meridian_angle_left_1_2 = TensorMap('5105_6mm-strong-meridian-angle-left_1_2', loss='logcosh', channel_map={'no_5105_6mm-strong-meridian-angle-left_1_2': 0})
field_id_5107_3mm_strong_meridian_angle_right_1_0 = TensorMap('5107_3mm-strong-meridian-angle-right_1_0', loss='logcosh', channel_map={'no_5107_3mm-strong-meridian-angle-right_1_0': 0})
field_id_5107_3mm_strong_meridian_angle_right_1_1 = TensorMap('5107_3mm-strong-meridian-angle-right_1_1', loss='logcosh', channel_map={'no_5107_3mm-strong-meridian-angle-right_1_1': 0})
field_id_5108_3mm_asymmetry_angle_right_1_0 = TensorMap('5108_3mm-asymmetry-angle-right_1_0', loss='logcosh', channel_map={'no_5108_3mm-asymmetry-angle-right_1_0': 0})
field_id_5108_3mm_asymmetry_angle_right_1_1 = TensorMap('5108_3mm-asymmetry-angle-right_1_1', loss='logcosh', channel_map={'no_5108_3mm-asymmetry-angle-right_1_1': 0})
field_id_5110_6mm_asymmetry_angle_left_1_1 = TensorMap('5110_6mm-asymmetry-angle-left_1_1', loss='logcosh', channel_map={'no_5110_6mm-asymmetry-angle-left_1_1': 0})
field_id_5110_6mm_asymmetry_angle_left_1_2 = TensorMap('5110_6mm-asymmetry-angle-left_1_2', loss='logcosh', channel_map={'no_5110_6mm-asymmetry-angle-left_1_2': 0})
field_id_5111_3mm_asymmetry_angle_left_1_0 = TensorMap('5111_3mm-asymmetry-angle-left_1_0', loss='logcosh', channel_map={'no_5111_3mm-asymmetry-angle-left_1_0': 0})
field_id_5111_3mm_asymmetry_angle_left_1_1 = TensorMap('5111_3mm-asymmetry-angle-left_1_1', loss='logcosh', channel_map={'no_5111_3mm-asymmetry-angle-left_1_1': 0})
field_id_5111_3mm_asymmetry_angle_left_1_2 = TensorMap('5111_3mm-asymmetry-angle-left_1_2', loss='logcosh', channel_map={'no_5111_3mm-asymmetry-angle-left_1_2': 0})
field_id_5112_3mm_cylindrical_power_angle_left_1_0 = TensorMap('5112_3mm-cylindrical-power-angle-left_1_0', loss='logcosh', channel_map={'no_5112_3mm-cylindrical-power-angle-left_1_0': 0})
field_id_5112_3mm_cylindrical_power_angle_left_1_1 = TensorMap('5112_3mm-cylindrical-power-angle-left_1_1', loss='logcosh', channel_map={'no_5112_3mm-cylindrical-power-angle-left_1_1': 0})
field_id_5112_3mm_cylindrical_power_angle_left_1_2 = TensorMap('5112_3mm-cylindrical-power-angle-left_1_2', loss='logcosh', channel_map={'no_5112_3mm-cylindrical-power-angle-left_1_2': 0})
field_id_5113_6mm_cylindrical_power_angle_left_1_0 = TensorMap('5113_6mm-cylindrical-power-angle-left_1_0', loss='logcosh', channel_map={'no_5113_6mm-cylindrical-power-angle-left_1_0': 0})
field_id_5113_6mm_cylindrical_power_angle_left_1_1 = TensorMap('5113_6mm-cylindrical-power-angle-left_1_1', loss='logcosh', channel_map={'no_5113_6mm-cylindrical-power-angle-left_1_1': 0})
field_id_5113_6mm_cylindrical_power_angle_left_1_2 = TensorMap('5113_6mm-cylindrical-power-angle-left_1_2', loss='logcosh', channel_map={'no_5113_6mm-cylindrical-power-angle-left_1_2': 0})
field_id_5115_3mm_cylindrical_power_angle_right_1_0 = TensorMap('5115_3mm-cylindrical-power-angle-right_1_0', loss='logcosh', channel_map={'no_5115_3mm-cylindrical-power-angle-right_1_0': 0})
field_id_5115_3mm_cylindrical_power_angle_right_1_1 = TensorMap('5115_3mm-cylindrical-power-angle-right_1_1', loss='logcosh', channel_map={'no_5115_3mm-cylindrical-power-angle-right_1_1': 0})
field_id_5116_3mm_cylindrical_power_right_1_0 = TensorMap('5116_3mm-cylindrical-power-right_1_0', loss='logcosh', channel_map={'no_5116_3mm-cylindrical-power-right_1_0': 0})
field_id_5116_3mm_cylindrical_power_right_1_1 = TensorMap('5116_3mm-cylindrical-power-right_1_1', loss='logcosh', channel_map={'no_5116_3mm-cylindrical-power-right_1_1': 0})
field_id_5118_6mm_cylindrical_power_left_1_0 = TensorMap('5118_6mm-cylindrical-power-left_1_0', loss='logcosh', channel_map={'no_5118_6mm-cylindrical-power-left_1_0': 0})
field_id_5118_6mm_cylindrical_power_left_1_1 = TensorMap('5118_6mm-cylindrical-power-left_1_1', loss='logcosh', channel_map={'no_5118_6mm-cylindrical-power-left_1_1': 0})
field_id_5118_6mm_cylindrical_power_left_1_2 = TensorMap('5118_6mm-cylindrical-power-left_1_2', loss='logcosh', channel_map={'no_5118_6mm-cylindrical-power-left_1_2': 0})
field_id_5119_3mm_cylindrical_power_left_1_0 = TensorMap('5119_3mm-cylindrical-power-left_1_0', loss='logcosh', channel_map={'no_5119_3mm-cylindrical-power-left_1_0': 0})
field_id_5119_3mm_cylindrical_power_left_1_1 = TensorMap('5119_3mm-cylindrical-power-left_1_1', loss='logcosh', channel_map={'no_5119_3mm-cylindrical-power-left_1_1': 0})
field_id_5119_3mm_cylindrical_power_left_1_2 = TensorMap('5119_3mm-cylindrical-power-left_1_2', loss='logcosh', channel_map={'no_5119_3mm-cylindrical-power-left_1_2': 0})
field_id_5132_3mm_strong_meridian_right_1_0 = TensorMap('5132_3mm-strong-meridian-right_1_0', loss='logcosh', channel_map={'no_5132_3mm-strong-meridian-right_1_0': 0})
field_id_5132_3mm_strong_meridian_right_1_1 = TensorMap('5132_3mm-strong-meridian-right_1_1', loss='logcosh', channel_map={'no_5132_3mm-strong-meridian-right_1_1': 0})
field_id_5134_6mm_strong_meridian_left_1_0 = TensorMap('5134_6mm-strong-meridian-left_1_0', loss='logcosh', channel_map={'no_5134_6mm-strong-meridian-left_1_0': 0})
field_id_5134_6mm_strong_meridian_left_1_1 = TensorMap('5134_6mm-strong-meridian-left_1_1', loss='logcosh', channel_map={'no_5134_6mm-strong-meridian-left_1_1': 0})
field_id_5134_6mm_strong_meridian_left_1_2 = TensorMap('5134_6mm-strong-meridian-left_1_2', loss='logcosh', channel_map={'no_5134_6mm-strong-meridian-left_1_2': 0})
field_id_5135_3mm_strong_meridian_left_1_0 = TensorMap('5135_3mm-strong-meridian-left_1_0', loss='logcosh', channel_map={'no_5135_3mm-strong-meridian-left_1_0': 0})
field_id_5135_3mm_strong_meridian_left_1_1 = TensorMap('5135_3mm-strong-meridian-left_1_1', loss='logcosh', channel_map={'no_5135_3mm-strong-meridian-left_1_1': 0})
field_id_5135_3mm_strong_meridian_left_1_2 = TensorMap('5135_3mm-strong-meridian-left_1_2', loss='logcosh', channel_map={'no_5135_3mm-strong-meridian-left_1_2': 0})
field_id_5156_3mm_asymmetry_index_left_1_0 = TensorMap('5156_3mm-asymmetry-index-left_1_0', loss='logcosh', channel_map={'no_5156_3mm-asymmetry-index-left_1_0': 0})
field_id_5156_3mm_asymmetry_index_left_1_1 = TensorMap('5156_3mm-asymmetry-index-left_1_1', loss='logcosh', channel_map={'no_5156_3mm-asymmetry-index-left_1_1': 0})
field_id_5156_3mm_asymmetry_index_left_1_2 = TensorMap('5156_3mm-asymmetry-index-left_1_2', loss='logcosh', channel_map={'no_5156_3mm-asymmetry-index-left_1_2': 0})
field_id_5157_6mm_asymmetry_index_left_1_1 = TensorMap('5157_6mm-asymmetry-index-left_1_1', loss='logcosh', channel_map={'no_5157_6mm-asymmetry-index-left_1_1': 0})
field_id_5157_6mm_asymmetry_index_left_1_2 = TensorMap('5157_6mm-asymmetry-index-left_1_2', loss='logcosh', channel_map={'no_5157_6mm-asymmetry-index-left_1_2': 0})
field_id_5159_3mm_asymmetry_index_right_1_0 = TensorMap('5159_3mm-asymmetry-index-right_1_0', loss='logcosh', channel_map={'no_5159_3mm-asymmetry-index-right_1_0': 0})
field_id_5159_3mm_asymmetry_index_right_1_1 = TensorMap('5159_3mm-asymmetry-index-right_1_1', loss='logcosh', channel_map={'no_5159_3mm-asymmetry-index-right_1_1': 0})
field_id_5160_3mm_regularity_index_right_1_0 = TensorMap('5160_3mm-regularity-index-right_1_0', loss='logcosh', channel_map={'no_5160_3mm-regularity-index-right_1_0': 0})
field_id_5160_3mm_regularity_index_right_1_1 = TensorMap('5160_3mm-regularity-index-right_1_1', loss='logcosh', channel_map={'no_5160_3mm-regularity-index-right_1_1': 0})
field_id_5162_6mm_regularity_index_left_1_1 = TensorMap('5162_6mm-regularity-index-left_1_1', loss='logcosh', channel_map={'no_5162_6mm-regularity-index-left_1_1': 0})
field_id_5162_6mm_regularity_index_left_1_2 = TensorMap('5162_6mm-regularity-index-left_1_2', loss='logcosh', channel_map={'no_5162_6mm-regularity-index-left_1_2': 0})
field_id_5163_3mm_regularity_index_left_1_0 = TensorMap('5163_3mm-regularity-index-left_1_0', loss='logcosh', channel_map={'no_5163_3mm-regularity-index-left_1_0': 0})
field_id_5163_3mm_regularity_index_left_1_1 = TensorMap('5163_3mm-regularity-index-left_1_1', loss='logcosh', channel_map={'no_5163_3mm-regularity-index-left_1_1': 0})
field_id_5163_3mm_regularity_index_left_1_2 = TensorMap('5163_3mm-regularity-index-left_1_2', loss='logcosh', channel_map={'no_5163_3mm-regularity-index-left_1_2': 0})
field_id_5186_duration_visualacuity_screen_displayed_right_1_0 = TensorMap('5186_Duration-visualacuity-screen-displayed-right_1_0', loss='logcosh', channel_map={'no_5186_Duration-visualacuity-screen-displayed-right_1_0': 0})
field_id_5188_duration_visualacuity_screen_displayed_left_1_0 = TensorMap('5188_Duration-visualacuity-screen-displayed-left_1_0', loss='logcosh', channel_map={'no_5188_Duration-visualacuity-screen-displayed-left_1_0': 0})
field_id_5190_duration_at_which_refractometer_first_shown_right_1_0 = TensorMap('5190_Duration-at-which-refractometer-first-shown-right_1_0', loss='logcosh', channel_map={'no_5190_Duration-at-which-refractometer-first-shown-right_1_0': 0})
field_id_5193_duration_at_which_refractometer_first_shown_left_1_0 = TensorMap('5193_Duration-at-which-refractometer-first-shown-left_1_0', loss='logcosh', channel_map={'no_5193_Duration-at-which-refractometer-first-shown-left_1_0': 0})
field_id_5199_logmar_initial_right_1_0 = TensorMap('5199_logMAR-initial-right_1_0', loss='logcosh', channel_map={'no_5199_logMAR-initial-right_1_0': 0})
field_id_51_seated_height_1_0 = TensorMap('51_Seated-height_1_0', loss='logcosh', channel_map={'no_51_Seated-height_1_0': 0})
field_id_5200_final_number_of_letters_displayed_right_1_0 = TensorMap('5200_Final-number-of-letters-displayed-right_1_0', loss='logcosh', channel_map={'no_5200_Final-number-of-letters-displayed-right_1_0': 0})
field_id_5201_logmar_final_right_1_0 = TensorMap('5201_logMAR-final-right_1_0', loss='logcosh', channel_map={'no_5201_logMAR-final-right_1_0': 0})
field_id_5202_number_of_rounds_to_result_right_1_0 = TensorMap('5202_Number-of-rounds-to-result-right_1_0', loss='logcosh', channel_map={'no_5202_Number-of-rounds-to-result-right_1_0': 0})
field_id_5204_distance_of_viewer_to_screen_right_1_0 = TensorMap('5204_Distance-of-viewer-to-screen-right_1_0', loss='logcosh', channel_map={'no_5204_Distance-of-viewer-to-screen-right_1_0': 0})
field_id_5206_logmar_initial_left_1_0 = TensorMap('5206_logMAR-initial-left_1_0', loss='logcosh', channel_map={'no_5206_logMAR-initial-left_1_0': 0})
field_id_5207_final_number_of_letters_displayed_left_1_0 = TensorMap('5207_Final-number-of-letters-displayed-left_1_0', loss='logcosh', channel_map={'no_5207_Final-number-of-letters-displayed-left_1_0': 0})
field_id_5208_logmar_final_left_1_0 = TensorMap('5208_logMAR-final-left_1_0', loss='logcosh', channel_map={'no_5208_logMAR-final-left_1_0': 0})
field_id_5209_number_of_rounds_to_result_left_1_0 = TensorMap('5209_Number-of-rounds-to-result-left_1_0', loss='logcosh', channel_map={'no_5209_Number-of-rounds-to-result-left_1_0': 0})
field_id_5211_distance_of_viewer_to_screen_left_1_0 = TensorMap('5211_Distance-of-viewer-to-screen-left_1_0', loss='logcosh', channel_map={'no_5211_Distance-of-viewer-to-screen-left_1_0': 0})
field_id_5215_vertex_distance_right_1_0 = TensorMap('5215_Vertex-distance-right_1_0', loss='logcosh', channel_map={'no_5215_Vertex-distance-right_1_0': 0})
field_id_5221_index_of_best_refractometry_result_right_1_0 = TensorMap('5221_Index-of-best-refractometry-result-right_1_0', loss='logcosh', channel_map={'no_5221_Index-of-best-refractometry-result-right_1_0': 0})
field_id_5237_3mm_index_of_best_keratometry_results_right_1_0 = TensorMap('5237_3mm-index-of-best-keratometry-results-right_1_0', loss='logcosh', channel_map={'no_5237_3mm-index-of-best-keratometry-results-right_1_0': 0})
field_id_5254_intraocular_pressure_cornealcompensated_right_1_0 = TensorMap('5254_Intraocular-pressure-cornealcompensated-right_1_0', loss='logcosh', channel_map={'no_5254_Intraocular-pressure-cornealcompensated-right_1_0': 0})
field_id_5255_intraocular_pressure_goldmanncorrelated_right_1_0 = TensorMap('5255_Intraocular-pressure-Goldmanncorrelated-right_1_0', loss='logcosh', channel_map={'no_5255_Intraocular-pressure-Goldmanncorrelated-right_1_0': 0})
field_id_5256_corneal_hysteresis_right_1_0 = TensorMap('5256_Corneal-hysteresis-right_1_0', loss='logcosh', channel_map={'no_5256_Corneal-hysteresis-right_1_0': 0})
field_id_5257_corneal_resistance_factor_right_1_0 = TensorMap('5257_Corneal-resistance-factor-right_1_0', loss='logcosh', channel_map={'no_5257_Corneal-resistance-factor-right_1_0': 0})
field_id_5262_intraocular_pressure_cornealcompensated_left_1_0 = TensorMap('5262_Intraocular-pressure-cornealcompensated-left_1_0', loss='logcosh', channel_map={'no_5262_Intraocular-pressure-cornealcompensated-left_1_0': 0})
field_id_5263_intraocular_pressure_goldmanncorrelated_left_1_0 = TensorMap('5263_Intraocular-pressure-Goldmanncorrelated-left_1_0', loss='logcosh', channel_map={'no_5263_Intraocular-pressure-Goldmanncorrelated-left_1_0': 0})
field_id_5264_corneal_hysteresis_left_1_0 = TensorMap('5264_Corneal-hysteresis-left_1_0', loss='logcosh', channel_map={'no_5264_Corneal-hysteresis-left_1_0': 0})
field_id_5265_corneal_resistance_factor_left_1_0 = TensorMap('5265_Corneal-resistance-factor-left_1_0', loss='logcosh', channel_map={'no_5265_Corneal-resistance-factor-left_1_0': 0})
field_id_5274_vertex_distance_left_1_0 = TensorMap('5274_Vertex-distance-left_1_0', loss='logcosh', channel_map={'no_5274_Vertex-distance-left_1_0': 0})
field_id_5276_index_of_best_refractometry_result_left_1_0 = TensorMap('5276_Index-of-best-refractometry-result-left_1_0', loss='logcosh', channel_map={'no_5276_Index-of-best-refractometry-result-left_1_0': 0})
field_id_5292_3mm_index_of_best_keratometry_results_left_1_0 = TensorMap('5292_3mm-index-of-best-keratometry-results-left_1_0', loss='logcosh', channel_map={'no_5292_3mm-index-of-best-keratometry-results-left_1_0': 0})
field_id_5306_6mm_index_of_best_keratometry_results_left_1_0 = TensorMap('5306_6mm-index-of-best-keratometry-results-left_1_0', loss='logcosh', channel_map={'no_5306_6mm-index-of-best-keratometry-results-left_1_0': 0})
field_id_5364_average_weekly_intake_of_other_alcoholic_drinks_1_0 = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_1_0', loss='logcosh', channel_map={'no_5364_Average-weekly-intake-of-other-alcoholic-drinks_1_0': 0})
field_id_6032_maximum_workload_during_fitness_test_1_0 = TensorMap('6032_Maximum-workload-during-fitness-test_1_0', loss='logcosh', channel_map={'no_6032_Maximum-workload-during-fitness-test_1_0': 0})
field_id_6033_maximum_heart_rate_during_fitness_test_1_0 = TensorMap('6033_Maximum-heart-rate-during-fitness-test_1_0', loss='logcosh', channel_map={'no_6033_Maximum-heart-rate-during-fitness-test_1_0': 0})
field_id_6038_number_of_trend_entries_1_0 = TensorMap('6038_Number-of-trend-entries_1_0', loss='logcosh', channel_map={'no_6038_Number-of-trend-entries_1_0': 0})
field_id_6039_duration_of_fitness_test_1_0 = TensorMap('6039_Duration-of-fitness-test_1_0', loss='logcosh', channel_map={'no_6039_Duration-of-fitness-test_1_0': 0})
field_id_6071_duration_at_which_oct_screen_shown_right_1_0 = TensorMap('6071_Duration-at-which-OCT-screen-shown-right_1_0', loss='logcosh', channel_map={'no_6071_Duration-at-which-OCT-screen-shown-right_1_0': 0})
field_id_6073_duration_at_which_oct_screen_shown_left_1_0 = TensorMap('6073_Duration-at-which-OCT-screen-shown-left_1_0', loss='logcosh', channel_map={'no_6073_Duration-at-which-OCT-screen-shown-left_1_0': 0})
field_id_699_length_of_time_at_current_address_1_0 = TensorMap('699_Length-of-time-at-current-address_1_0', loss='logcosh', channel_map={'no_699_Length-of-time-at-current-address_1_0': 0})
field_id_709_number_in_household_1_0 = TensorMap('709_Number-in-household_1_0', loss='logcosh', channel_map={'no_709_Number-in-household_1_0': 0})
field_id_845_age_completed_full_time_education_1_0 = TensorMap('845_Age-completed-full-time-education_1_0', loss='logcosh', channel_map={'no_845_Age-completed-full-time-education_1_0': 0})
field_id_864_number_of_daysweek_walked_10_minutes_1_0 = TensorMap('864_Number-of-daysweek-walked-10-minutes_1_0', loss='logcosh', channel_map={'no_864_Number-of-daysweek-walked-10-minutes_1_0': 0})
field_id_874_duration_of_walks_1_0 = TensorMap('874_Duration-of-walks_1_0', loss='logcosh', channel_map={'no_874_Duration-of-walks_1_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_4 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_4', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_4': 0})
field_id_87_noncancer_illness_yearage_first_occurred_1_0 = TensorMap('87_Noncancer-illness-yearage-first-occurred_1_0', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_1_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_1_1 = TensorMap('87_Noncancer-illness-yearage-first-occurred_1_1', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_1_1': 0})
field_id_87_noncancer_illness_yearage_first_occurred_1_2 = TensorMap('87_Noncancer-illness-yearage-first-occurred_1_2', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_1_2': 0})
field_id_884_number_of_daysweek_of_moderate_physical_activity_10_minutes_1_0 = TensorMap('884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_1_0', loss='logcosh', channel_map={'no_884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_1_0': 0})
field_id_894_duration_of_moderate_activity_1_0 = TensorMap('894_Duration-of-moderate-activity_1_0', loss='logcosh', channel_map={'no_894_Duration-of-moderate-activity_1_0': 0})
field_id_904_number_of_daysweek_of_vigorous_physical_activity_10_minutes_1_0 = TensorMap('904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_1_0', loss='logcosh', channel_map={'no_904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_1_0': 0})
field_id_92_operation_yearage_first_occurred_1_0 = TensorMap('92_Operation-yearage-first-occurred_1_0', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_1_0': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_5 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_5', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_5': 0})
field_id_20009_interpolated_age_of_participant_when_noncancer_illness_first_diagnosed_0_6 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_6', loss='logcosh', channel_map={'no_20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_6': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_0_2 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_2', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_0_2': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_0_3 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_3', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_0_3': 0})
field_id_22605_work_hours_per_week_exact_value_0_4 = TensorMap('22605_Work-hours-per-week-exact-value_0_4', loss='logcosh', channel_map={'no_22605_Work-hours-per-week-exact-value_0_4': 0})
field_id_22642_usual_length_of_each_night_shift_during_mixed_shift_periods_0_0 = TensorMap('22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_0', loss='logcosh', channel_map={'no_22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_0': 0})
field_id_22643_number_of_night_shifts_worked_monthly_during_mixed_shift_periods_0_0 = TensorMap('22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_0', loss='logcosh', channel_map={'no_22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_0': 0})
field_id_22644_consecutive_night_shifts_during_mixed_shift_periods_0_0 = TensorMap('22644_Consecutive-night-shifts-during-mixed-shift-periods_0_0', loss='logcosh', channel_map={'no_22644_Consecutive-night-shifts-during-mixed-shift-periods_0_0': 0})
field_id_22645_rest_days_during_mixed_shift_periods_0_0 = TensorMap('22645_Rest-days-during-mixed-shift-periods_0_0', loss='logcosh', channel_map={'no_22645_Rest-days-during-mixed-shift-periods_0_0': 0})
field_id_4256_time_elapsed_0_0 = TensorMap('4256_Time-elapsed_0_0', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_0': 0})
field_id_4256_time_elapsed_0_1 = TensorMap('4256_Time-elapsed_0_1', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_1': 0})
field_id_4256_time_elapsed_0_2 = TensorMap('4256_Time-elapsed_0_2', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_2': 0})
field_id_4256_time_elapsed_0_3 = TensorMap('4256_Time-elapsed_0_3', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_3': 0})
field_id_4256_time_elapsed_0_4 = TensorMap('4256_Time-elapsed_0_4', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_4': 0})
field_id_4260_round_of_numeric_memory_test_0_0 = TensorMap('4260_Round-of-numeric-memory-test_0_0', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_0': 0})
field_id_4260_round_of_numeric_memory_test_0_1 = TensorMap('4260_Round-of-numeric-memory-test_0_1', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_1': 0})
field_id_4260_round_of_numeric_memory_test_0_2 = TensorMap('4260_Round-of-numeric-memory-test_0_2', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_2': 0})
field_id_4260_round_of_numeric_memory_test_0_3 = TensorMap('4260_Round-of-numeric-memory-test_0_3', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_3': 0})
field_id_4260_round_of_numeric_memory_test_0_4 = TensorMap('4260_Round-of-numeric-memory-test_0_4', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_4': 0})
field_id_4282_maximum_digits_remembered_correctly_0_0 = TensorMap('4282_Maximum-digits-remembered-correctly_0_0', loss='logcosh', channel_map={'no_4282_Maximum-digits-remembered-correctly_0_0': 0})
field_id_4283_number_of_rounds_of_numeric_memory_test_performed_0_0 = TensorMap('4283_Number-of-rounds-of-numeric-memory-test-performed_0_0', loss='logcosh', channel_map={'no_4283_Number-of-rounds-of-numeric-memory-test-performed_0_0': 0})
field_id_4285_time_to_complete_test_0_0 = TensorMap('4285_Time-to-complete-test_0_0', loss='logcosh', channel_map={'no_4285_Time-to-complete-test_0_0': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_5 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_5', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_5': 0})
field_id_87_noncancer_illness_yearage_first_occurred_0_6 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_6', loss='logcosh', channel_map={'no_87_Noncancer-illness-yearage-first-occurred_0_6': 0})
field_id_92_operation_yearage_first_occurred_0_2 = TensorMap('92_Operation-yearage-first-occurred_0_2', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_0_2': 0})
field_id_92_operation_yearage_first_occurred_0_3 = TensorMap('92_Operation-yearage-first-occurred_0_3', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_0_3': 0})
field_id_20007_interpolated_age_of_participant_when_cancer_first_diagnosed_0_0 = TensorMap('20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_0', loss='logcosh', channel_map={'no_20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_0': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_0_4 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_4', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_0_4': 0})
field_id_40008_age_at_cancer_diagnosis_1_0 = TensorMap('40008_Age-at-cancer-diagnosis_1_0', loss='logcosh', channel_map={'no_40008_Age-at-cancer-diagnosis_1_0': 0})
field_id_4256_time_elapsed_0_5 = TensorMap('4256_Time-elapsed_0_5', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_5': 0})
field_id_4256_time_elapsed_0_6 = TensorMap('4256_Time-elapsed_0_6', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_6': 0})
field_id_4256_time_elapsed_0_7 = TensorMap('4256_Time-elapsed_0_7', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_7': 0})
field_id_4256_time_elapsed_0_8 = TensorMap('4256_Time-elapsed_0_8', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_8': 0})
field_id_4260_round_of_numeric_memory_test_0_5 = TensorMap('4260_Round-of-numeric-memory-test_0_5', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_5': 0})
field_id_4260_round_of_numeric_memory_test_0_6 = TensorMap('4260_Round-of-numeric-memory-test_0_6', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_6': 0})
field_id_4260_round_of_numeric_memory_test_0_7 = TensorMap('4260_Round-of-numeric-memory-test_0_7', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_7': 0})
field_id_4260_round_of_numeric_memory_test_0_8 = TensorMap('4260_Round-of-numeric-memory-test_0_8', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_8': 0})
field_id_4700_age_cataract_diagnosed_0_0 = TensorMap('4700_Age-cataract-diagnosed_0_0', loss='logcosh', channel_map={'no_4700_Age-cataract-diagnosed_0_0': 0})
field_id_84_cancer_yearage_first_occurred_0_0 = TensorMap('84_Cancer-yearage-first-occurred_0_0', loss='logcosh', channel_map={'no_84_Cancer-yearage-first-occurred_0_0': 0})
field_id_92_operation_yearage_first_occurred_0_4 = TensorMap('92_Operation-yearage-first-occurred_0_4', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_0_4': 0})
field_id_2976_age_diabetes_diagnosed_0_0 = TensorMap('2976_Age-diabetes-diagnosed_0_0', loss='logcosh', channel_map={'no_2976_Age-diabetes-diagnosed_0_0': 0})
field_id_5098_6mm_weak_meridian_right_0_2 = TensorMap('5098_6mm-weak-meridian-right_0_2', loss='logcosh', channel_map={'no_5098_6mm-weak-meridian-right_0_2': 0})
field_id_5101_6mm_weak_meridian_angle_right_0_2 = TensorMap('5101_6mm-weak-meridian-angle-right_0_2', loss='logcosh', channel_map={'no_5101_6mm-weak-meridian-angle-right_0_2': 0})
field_id_5106_6mm_strong_meridian_angle_right_0_2 = TensorMap('5106_6mm-strong-meridian-angle-right_0_2', loss='logcosh', channel_map={'no_5106_6mm-strong-meridian-angle-right_0_2': 0})
field_id_5109_6mm_asymmetry_angle_right_0_2 = TensorMap('5109_6mm-asymmetry-angle-right_0_2', loss='logcosh', channel_map={'no_5109_6mm-asymmetry-angle-right_0_2': 0})
field_id_5110_6mm_asymmetry_angle_left_0_0 = TensorMap('5110_6mm-asymmetry-angle-left_0_0', loss='logcosh', channel_map={'no_5110_6mm-asymmetry-angle-left_0_0': 0})
field_id_5110_6mm_asymmetry_angle_left_0_1 = TensorMap('5110_6mm-asymmetry-angle-left_0_1', loss='logcosh', channel_map={'no_5110_6mm-asymmetry-angle-left_0_1': 0})
field_id_5114_6mm_cylindrical_power_angle_right_0_2 = TensorMap('5114_6mm-cylindrical-power-angle-right_0_2', loss='logcosh', channel_map={'no_5114_6mm-cylindrical-power-angle-right_0_2': 0})
field_id_5117_6mm_cylindrical_power_right_0_2 = TensorMap('5117_6mm-cylindrical-power-right_0_2', loss='logcosh', channel_map={'no_5117_6mm-cylindrical-power-right_0_2': 0})
field_id_5133_6mm_strong_meridian_right_0_2 = TensorMap('5133_6mm-strong-meridian-right_0_2', loss='logcosh', channel_map={'no_5133_6mm-strong-meridian-right_0_2': 0})
field_id_5157_6mm_asymmetry_index_left_0_0 = TensorMap('5157_6mm-asymmetry-index-left_0_0', loss='logcosh', channel_map={'no_5157_6mm-asymmetry-index-left_0_0': 0})
field_id_5157_6mm_asymmetry_index_left_0_1 = TensorMap('5157_6mm-asymmetry-index-left_0_1', loss='logcosh', channel_map={'no_5157_6mm-asymmetry-index-left_0_1': 0})
field_id_5158_6mm_asymmetry_index_right_0_2 = TensorMap('5158_6mm-asymmetry-index-right_0_2', loss='logcosh', channel_map={'no_5158_6mm-asymmetry-index-right_0_2': 0})
field_id_5161_6mm_regularity_index_right_0_2 = TensorMap('5161_6mm-regularity-index-right_0_2', loss='logcosh', channel_map={'no_5161_6mm-regularity-index-right_0_2': 0})
field_id_5162_6mm_regularity_index_left_0_0 = TensorMap('5162_6mm-regularity-index-left_0_0', loss='logcosh', channel_map={'no_5162_6mm-regularity-index-left_0_0': 0})
field_id_5162_6mm_regularity_index_left_0_1 = TensorMap('5162_6mm-regularity-index-left_0_1', loss='logcosh', channel_map={'no_5162_6mm-regularity-index-left_0_1': 0})
field_id_30800_oestradiol_0_0 = TensorMap('30800_Oestradiol_0_0', loss='logcosh', channel_map={'no_30800_Oestradiol_0_0': 0})
field_id_3700_time_since_last_menstrual_period_0_0 = TensorMap('3700_Time-since-last-menstrual-period_0_0', loss='logcosh', channel_map={'no_3700_Time-since-last-menstrual-period_0_0': 0})
field_id_3710_length_of_menstrual_cycle_0_0 = TensorMap('3710_Length-of-menstrual-cycle_0_0', loss='logcosh', channel_map={'no_3710_Length-of-menstrual-cycle_0_0': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_0_5 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_5', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_0_5': 0})
field_id_92_operation_yearage_first_occurred_0_5 = TensorMap('92_Operation-yearage-first-occurred_0_5', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_0_5': 0})
field_id_4256_time_elapsed_0_9 = TensorMap('4256_Time-elapsed_0_9', loss='logcosh', channel_map={'no_4256_Time-elapsed_0_9': 0})
field_id_4260_round_of_numeric_memory_test_0_9 = TensorMap('4260_Round-of-numeric-memory-test_0_9', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_0_9': 0})
field_id_12673_heart_rate_during_pwa_2_0 = TensorMap('12673_Heart-rate-during-PWA_2_0', loss='logcosh', channel_map={'no_12673_Heart-rate-during-PWA_2_0': 0})
field_id_12673_heart_rate_during_pwa_2_1 = TensorMap('12673_Heart-rate-during-PWA_2_1', loss='logcosh', channel_map={'no_12673_Heart-rate-during-PWA_2_1': 0})
field_id_12674_systolic_brachial_blood_pressure_during_pwa_2_0 = TensorMap('12674_Systolic-brachial-blood-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12674_Systolic-brachial-blood-pressure-during-PWA_2_0': 0})
field_id_12675_diastolic_brachial_blood_pressure_during_pwa_2_0 = TensorMap('12675_Diastolic-brachial-blood-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12675_Diastolic-brachial-blood-pressure-during-PWA_2_0': 0})
field_id_12676_peripheral_pulse_pressure_during_pwa_2_0 = TensorMap('12676_Peripheral-pulse-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12676_Peripheral-pulse-pressure-during-PWA_2_0': 0})
field_id_12677_central_systolic_blood_pressure_during_pwa_2_0 = TensorMap('12677_Central-systolic-blood-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12677_Central-systolic-blood-pressure-during-PWA_2_0': 0})
field_id_12678_central_pulse_pressure_during_pwa_2_0 = TensorMap('12678_Central-pulse-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12678_Central-pulse-pressure-during-PWA_2_0': 0})
field_id_12679_number_of_beats_in_waveform_average_for_pwa_2_0 = TensorMap('12679_Number-of-beats-in-waveform-average-for-PWA_2_0', loss='logcosh', channel_map={'no_12679_Number-of-beats-in-waveform-average-for-PWA_2_0': 0})
field_id_12679_number_of_beats_in_waveform_average_for_pwa_2_1 = TensorMap('12679_Number-of-beats-in-waveform-average-for-PWA_2_1', loss='logcosh', channel_map={'no_12679_Number-of-beats-in-waveform-average-for-PWA_2_1': 0})
field_id_12680_central_augmentation_pressure_during_pwa_2_0 = TensorMap('12680_Central-augmentation-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12680_Central-augmentation-pressure-during-PWA_2_0': 0})
field_id_12681_augmentation_index_for_pwa_2_0 = TensorMap('12681_Augmentation-index-for-PWA_2_0', loss='logcosh', channel_map={'no_12681_Augmentation-index-for-PWA_2_0': 0})
field_id_12681_augmentation_index_for_pwa_2_1 = TensorMap('12681_Augmentation-index-for-PWA_2_1', loss='logcosh', channel_map={'no_12681_Augmentation-index-for-PWA_2_1': 0})
field_id_12682_cardiac_output_during_pwa_2_0 = TensorMap('12682_Cardiac-output-during-PWA_2_0', loss='logcosh', channel_map={'no_12682_Cardiac-output-during-PWA_2_0': 0})
field_id_12683_end_systolic_pressure_during_pwa_2_0 = TensorMap('12683_End-systolic-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12683_End-systolic-pressure-during-PWA_2_0': 0})
field_id_12684_end_systolic_pressure_index_during_pwa_2_0 = TensorMap('12684_End-systolic-pressure-index-during-PWA_2_0', loss='logcosh', channel_map={'no_12684_End-systolic-pressure-index-during-PWA_2_0': 0})
field_id_12685_total_peripheral_resistance_during_pwa_2_0 = TensorMap('12685_Total-peripheral-resistance-during-PWA_2_0', loss='logcosh', channel_map={'no_12685_Total-peripheral-resistance-during-PWA_2_0': 0})
field_id_12686_stroke_volume_during_pwa_2_0 = TensorMap('12686_Stroke-volume-during-PWA_2_0', loss='logcosh', channel_map={'no_12686_Stroke-volume-during-PWA_2_0': 0})
field_id_12687_mean_arterial_pressure_during_pwa_2_0 = TensorMap('12687_Mean-arterial-pressure-during-PWA_2_0', loss='logcosh', channel_map={'no_12687_Mean-arterial-pressure-during-PWA_2_0': 0})
field_id_12697_systolic_brachial_blood_pressure_2_0 = TensorMap('12697_Systolic-brachial-blood-pressure_2_0', loss='logcosh', channel_map={'no_12697_Systolic-brachial-blood-pressure_2_0': 0})
field_id_12698_diastolic_brachial_blood_pressure_2_0 = TensorMap('12698_Diastolic-brachial-blood-pressure_2_0', loss='logcosh', channel_map={'no_12698_Diastolic-brachial-blood-pressure_2_0': 0})
field_id_12699_number_of_pwa_tests_performed_2_0 = TensorMap('12699_Number-of-PWA-tests-performed_2_0', loss='logcosh', channel_map={'no_12699_Number-of-PWA-tests-performed_2_0': 0})
field_id_12702_cardiac_index_during_pwa_2_0 = TensorMap('12702_Cardiac-index-during-PWA_2_0', loss='logcosh', channel_map={'no_12702_Cardiac-index-during-PWA_2_0': 0})
field_id_129_place_of_birth_in_uk_north_coordinate_2_0 = TensorMap('129_Place-of-birth-in-UK-north-coordinate_2_0', loss='logcosh', channel_map={'no_129_Place-of-birth-in-UK-north-coordinate_2_0': 0})
field_id_130_place_of_birth_in_uk_east_coordinate_2_0 = TensorMap('130_Place-of-birth-in-UK-east-coordinate_2_0', loss='logcosh', channel_map={'no_130_Place-of-birth-in-UK-east-coordinate_2_0': 0})
field_id_1737_childhood_sunburn_occasions_2_0 = TensorMap('1737_Childhood-sunburn-occasions_2_0', loss='logcosh', channel_map={'no_1737_Childhood-sunburn-occasions_2_0': 0})
field_id_2405_number_of_children_fathered_2_0 = TensorMap('2405_Number-of-children-fathered_2_0', loss='logcosh', channel_map={'no_2405_Number-of-children-fathered_2_0': 0})
field_id_3761_age_hay_fever_rhinitis_or_eczema_diagnosed_2_0 = TensorMap('3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_2_0', loss='logcosh', channel_map={'no_3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_2_0': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_0 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_0', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_0': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_3 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_3', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_3': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_2_4 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_2_4', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_2_4': 0})
field_id_4100_ankle_spacing_width_left_2_0 = TensorMap('4100_Ankle-spacing-width-left_2_0', loss='logcosh', channel_map={'no_4100_Ankle-spacing-width-left_2_0': 0})
field_id_4101_heel_broadband_ultrasound_attenuation_left_2_0 = TensorMap('4101_Heel-broadband-ultrasound-attenuation-left_2_0', loss='logcosh', channel_map={'no_4101_Heel-broadband-ultrasound-attenuation-left_2_0': 0})
field_id_4103_speed_of_sound_through_heel_left_2_0 = TensorMap('4103_Speed-of-sound-through-heel-left_2_0', loss='logcosh', channel_map={'no_4103_Speed-of-sound-through-heel-left_2_0': 0})
field_id_4104_heel_quantitative_ultrasound_index_qui_direct_entry_left_2_0 = TensorMap('4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_2_0', loss='logcosh', channel_map={'no_4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_2_0': 0})
field_id_4105_heel_bone_mineral_density_bmd_left_2_0 = TensorMap('4105_Heel-bone-mineral-density-BMD-left_2_0', loss='logcosh', channel_map={'no_4105_Heel-bone-mineral-density-BMD-left_2_0': 0})
field_id_4106_heel_bone_mineral_density_bmd_tscore_automated_left_2_0 = TensorMap('4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_2_0', loss='logcosh', channel_map={'no_4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_2_0': 0})
field_id_4119_ankle_spacing_width_right_2_0 = TensorMap('4119_Ankle-spacing-width-right_2_0', loss='logcosh', channel_map={'no_4119_Ankle-spacing-width-right_2_0': 0})
field_id_4120_heel_broadband_ultrasound_attenuation_right_2_0 = TensorMap('4120_Heel-broadband-ultrasound-attenuation-right_2_0', loss='logcosh', channel_map={'no_4120_Heel-broadband-ultrasound-attenuation-right_2_0': 0})
field_id_4122_speed_of_sound_through_heel_right_2_0 = TensorMap('4122_Speed-of-sound-through-heel-right_2_0', loss='logcosh', channel_map={'no_4122_Speed-of-sound-through-heel-right_2_0': 0})
field_id_4123_heel_quantitative_ultrasound_index_qui_direct_entry_right_2_0 = TensorMap('4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_2_0', loss='logcosh', channel_map={'no_4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_2_0': 0})
field_id_4124_heel_bone_mineral_density_bmd_right_2_0 = TensorMap('4124_Heel-bone-mineral-density-BMD-right_2_0', loss='logcosh', channel_map={'no_4124_Heel-bone-mineral-density-BMD-right_2_0': 0})
field_id_4125_heel_bone_mineral_density_bmd_tscore_automated_right_2_0 = TensorMap('4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_2_0', loss='logcosh', channel_map={'no_4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_2_0': 0})
field_id_4407_average_monthly_red_wine_intake_0_0 = TensorMap('4407_Average-monthly-red-wine-intake_0_0', loss='logcosh', channel_map={'no_4407_Average-monthly-red-wine-intake_0_0': 0})
field_id_4407_average_monthly_red_wine_intake_2_0 = TensorMap('4407_Average-monthly-red-wine-intake_2_0', loss='logcosh', channel_map={'no_4407_Average-monthly-red-wine-intake_2_0': 0})
field_id_4418_average_monthly_champagne_plus_white_wine_intake_0_0 = TensorMap('4418_Average-monthly-champagne-plus-white-wine-intake_0_0', loss='logcosh', channel_map={'no_4418_Average-monthly-champagne-plus-white-wine-intake_0_0': 0})
field_id_4418_average_monthly_champagne_plus_white_wine_intake_2_0 = TensorMap('4418_Average-monthly-champagne-plus-white-wine-intake_2_0', loss='logcosh', channel_map={'no_4418_Average-monthly-champagne-plus-white-wine-intake_2_0': 0})
field_id_4429_average_monthly_beer_plus_cider_intake_0_0 = TensorMap('4429_Average-monthly-beer-plus-cider-intake_0_0', loss='logcosh', channel_map={'no_4429_Average-monthly-beer-plus-cider-intake_0_0': 0})
field_id_4429_average_monthly_beer_plus_cider_intake_2_0 = TensorMap('4429_Average-monthly-beer-plus-cider-intake_2_0', loss='logcosh', channel_map={'no_4429_Average-monthly-beer-plus-cider-intake_2_0': 0})
field_id_4440_average_monthly_spirits_intake_0_0 = TensorMap('4440_Average-monthly-spirits-intake_0_0', loss='logcosh', channel_map={'no_4440_Average-monthly-spirits-intake_0_0': 0})
field_id_4440_average_monthly_spirits_intake_2_0 = TensorMap('4440_Average-monthly-spirits-intake_2_0', loss='logcosh', channel_map={'no_4440_Average-monthly-spirits-intake_2_0': 0})
field_id_4451_average_monthly_fortified_wine_intake_0_0 = TensorMap('4451_Average-monthly-fortified-wine-intake_0_0', loss='logcosh', channel_map={'no_4451_Average-monthly-fortified-wine-intake_0_0': 0})
field_id_4451_average_monthly_fortified_wine_intake_2_0 = TensorMap('4451_Average-monthly-fortified-wine-intake_2_0', loss='logcosh', channel_map={'no_4451_Average-monthly-fortified-wine-intake_2_0': 0})
field_id_4462_average_monthly_intake_of_other_alcoholic_drinks_0_0 = TensorMap('4462_Average-monthly-intake-of-other-alcoholic-drinks_0_0', loss='logcosh', channel_map={'no_4462_Average-monthly-intake-of-other-alcoholic-drinks_0_0': 0})
field_id_4462_average_monthly_intake_of_other_alcoholic_drinks_2_0 = TensorMap('4462_Average-monthly-intake-of-other-alcoholic-drinks_2_0', loss='logcosh', channel_map={'no_4462_Average-monthly-intake-of-other-alcoholic-drinks_2_0': 0})
field_id_894_duration_of_moderate_activity_2_0 = TensorMap('894_Duration-of-moderate-activity_2_0', loss='logcosh', channel_map={'no_894_Duration-of-moderate-activity_2_0': 0})
field_id_914_duration_of_vigorous_activity_2_0 = TensorMap('914_Duration-of-vigorous-activity_2_0', loss='logcosh', channel_map={'no_914_Duration-of-vigorous-activity_2_0': 0})
field_id_3829_number_of_stillbirths_0_0 = TensorMap('3829_Number-of-stillbirths_0_0', loss='logcosh', channel_map={'no_3829_Number-of-stillbirths_0_0': 0})
field_id_3839_number_of_spontaneous_miscarriages_0_0 = TensorMap('3839_Number-of-spontaneous-miscarriages_0_0', loss='logcosh', channel_map={'no_3839_Number-of-spontaneous-miscarriages_0_0': 0})
field_id_3849_number_of_pregnancy_terminations_0_0 = TensorMap('3849_Number-of-pregnancy-terminations_0_0', loss='logcosh', channel_map={'no_3849_Number-of-pregnancy-terminations_0_0': 0})
field_id_3872_age_of_primiparous_women_at_birth_of_child_0_0 = TensorMap('3872_Age-of-primiparous-women-at-birth-of-child_0_0', loss='logcosh', channel_map={'no_3872_Age-of-primiparous-women-at-birth-of-child_0_0': 0})
field_id_3972_number_of_adopted_brothers_0_0 = TensorMap('3972_Number-of-adopted-brothers_0_0', loss='logcosh', channel_map={'no_3972_Number-of-adopted-brothers_0_0': 0})
field_id_3982_number_of_adopted_sisters_0_0 = TensorMap('3982_Number-of-adopted-sisters_0_0', loss='logcosh', channel_map={'no_3982_Number-of-adopted-sisters_0_0': 0})
field_id_12651_duration_of_eprime_test_2_0 = TensorMap('12651_Duration-of-eprime-test_2_0', loss='logcosh', channel_map={'no_12651_Duration-of-eprime-test_2_0': 0})
field_id_1845_mothers_age_2_0 = TensorMap('1845_Mothers-age_2_0', loss='logcosh', channel_map={'no_1845_Mothers-age_2_0': 0})
field_id_25040_median_bold_effect_in_groupdefined_mask_for_shapes_activation_2_0 = TensorMap('25040_Median-BOLD-effect-in-groupdefined-mask-for-shapes-activation_2_0', loss='logcosh', channel_map={'no_25040_Median-BOLD-effect-in-groupdefined-mask-for-shapes-activation_2_0': 0})
field_id_25042_median_zstatistic_in_groupdefined_mask_for_shapes_activation_2_0 = TensorMap('25042_Median-zstatistic-in-groupdefined-mask-for-shapes-activation_2_0', loss='logcosh', channel_map={'no_25042_Median-zstatistic-in-groupdefined-mask-for-shapes-activation_2_0': 0})
field_id_25044_median_bold_effect_in_groupdefined_mask_for_faces_activation_2_0 = TensorMap('25044_Median-BOLD-effect-in-groupdefined-mask-for-faces-activation_2_0', loss='logcosh', channel_map={'no_25044_Median-BOLD-effect-in-groupdefined-mask-for-faces-activation_2_0': 0})
field_id_25046_median_zstatistic_in_groupdefined_mask_for_faces_activation_2_0 = TensorMap('25046_Median-zstatistic-in-groupdefined-mask-for-faces-activation_2_0', loss='logcosh', channel_map={'no_25046_Median-zstatistic-in-groupdefined-mask-for-faces-activation_2_0': 0})
field_id_25048_median_bold_effect_in_groupdefined_mask_for_facesshapes_contrast_2_0 = TensorMap('25048_Median-BOLD-effect-in-groupdefined-mask-for-facesshapes-contrast_2_0', loss='logcosh', channel_map={'no_25048_Median-BOLD-effect-in-groupdefined-mask-for-facesshapes-contrast_2_0': 0})
field_id_25050_median_zstatistic_in_groupdefined_mask_for_facesshapes_contrast_2_0 = TensorMap('25050_Median-zstatistic-in-groupdefined-mask-for-facesshapes-contrast_2_0', loss='logcosh', channel_map={'no_25050_Median-zstatistic-in-groupdefined-mask-for-facesshapes-contrast_2_0': 0})
field_id_25052_median_bold_effect_in_groupdefined_amygdala_activation_mask_for_facesshapes_contrast_2_0 = TensorMap('25052_Median-BOLD-effect-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0', loss='logcosh', channel_map={'no_25052_Median-BOLD-effect-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0': 0})
field_id_25054_median_zstatistic_in_groupdefined_amygdala_activation_mask_for_facesshapes_contrast_2_0 = TensorMap('25054_Median-zstatistic-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0', loss='logcosh', channel_map={'no_25054_Median-zstatistic-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0': 0})
field_id_25740_discrepancy_between_tfmri_brain_image_and_t1_brain_image_2_0 = TensorMap('25740_Discrepancy-between-tfMRI-brain-image-and-T1-brain-image_2_0', loss='logcosh', channel_map={'no_25740_Discrepancy-between-tfMRI-brain-image-and-T1-brain-image_2_0': 0})
field_id_4256_time_elapsed_2_7 = TensorMap('4256_Time-elapsed_2_7', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_7': 0})
field_id_4256_time_elapsed_2_8 = TensorMap('4256_Time-elapsed_2_8', loss='logcosh', channel_map={'no_4256_Time-elapsed_2_8': 0})
field_id_4260_round_of_numeric_memory_test_2_7 = TensorMap('4260_Round-of-numeric-memory-test_2_7', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_7': 0})
field_id_4260_round_of_numeric_memory_test_2_8 = TensorMap('4260_Round-of-numeric-memory-test_2_8', loss='logcosh', channel_map={'no_4260_Round-of-numeric-memory-test_2_8': 0})
field_id_22663_year_gap_started_0_2 = TensorMap('22663_Year-gap-started_0_2', loss='logcosh', channel_map={'no_22663_Year-gap-started_0_2': 0})
field_id_22663_year_gap_started_0_3 = TensorMap('22663_Year-gap-started_0_3', loss='logcosh', channel_map={'no_22663_Year-gap-started_0_3': 0})
field_id_22664_year_gap_ended_0_2 = TensorMap('22664_Year-gap-ended_0_2', loss='logcosh', channel_map={'no_22664_Year-gap-ended_0_2': 0})
field_id_22664_year_gap_ended_0_3 = TensorMap('22664_Year-gap-ended_0_3', loss='logcosh', channel_map={'no_22664_Year-gap-ended_0_3': 0})
field_id_2824_age_at_hysterectomy_0_0 = TensorMap('2824_Age-at-hysterectomy_0_0', loss='logcosh', channel_map={'no_2824_Age-at-hysterectomy_0_0': 0})
field_id_93_systolic_blood_pressure_manual_reading_0_1 = TensorMap('93_Systolic-blood-pressure-manual-reading_0_1', loss='logcosh', channel_map={'no_93_Systolic-blood-pressure-manual-reading_0_1': 0})
field_id_94_diastolic_blood_pressure_manual_reading_0_1 = TensorMap('94_Diastolic-blood-pressure-manual-reading_0_1', loss='logcosh', channel_map={'no_94_Diastolic-blood-pressure-manual-reading_0_1': 0})
field_id_95_pulse_rate_during_bloodpressure_measurement_0_1 = TensorMap('95_Pulse-rate-during-bloodpressure-measurement_0_1', loss='logcosh', channel_map={'no_95_Pulse-rate-during-bloodpressure-measurement_0_1': 0})
field_id_40007_age_at_death_0_0 = TensorMap('40007_Age-at-death_0_0', loss='logcosh', channel_map={'no_40007_Age-at-death_0_0': 0})
field_id_12674_systolic_brachial_blood_pressure_during_pwa_2_1 = TensorMap('12674_Systolic-brachial-blood-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12674_Systolic-brachial-blood-pressure-during-PWA_2_1': 0})
field_id_12675_diastolic_brachial_blood_pressure_during_pwa_2_1 = TensorMap('12675_Diastolic-brachial-blood-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12675_Diastolic-brachial-blood-pressure-during-PWA_2_1': 0})
field_id_12676_peripheral_pulse_pressure_during_pwa_2_1 = TensorMap('12676_Peripheral-pulse-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12676_Peripheral-pulse-pressure-during-PWA_2_1': 0})
field_id_12677_central_systolic_blood_pressure_during_pwa_2_1 = TensorMap('12677_Central-systolic-blood-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12677_Central-systolic-blood-pressure-during-PWA_2_1': 0})
field_id_12678_central_pulse_pressure_during_pwa_2_1 = TensorMap('12678_Central-pulse-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12678_Central-pulse-pressure-during-PWA_2_1': 0})
field_id_12680_central_augmentation_pressure_during_pwa_2_1 = TensorMap('12680_Central-augmentation-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12680_Central-augmentation-pressure-during-PWA_2_1': 0})
field_id_12682_cardiac_output_during_pwa_2_1 = TensorMap('12682_Cardiac-output-during-PWA_2_1', loss='logcosh', channel_map={'no_12682_Cardiac-output-during-PWA_2_1': 0})
field_id_12683_end_systolic_pressure_during_pwa_2_1 = TensorMap('12683_End-systolic-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12683_End-systolic-pressure-during-PWA_2_1': 0})
field_id_12684_end_systolic_pressure_index_during_pwa_2_1 = TensorMap('12684_End-systolic-pressure-index-during-PWA_2_1', loss='logcosh', channel_map={'no_12684_End-systolic-pressure-index-during-PWA_2_1': 0})
field_id_12685_total_peripheral_resistance_during_pwa_2_1 = TensorMap('12685_Total-peripheral-resistance-during-PWA_2_1', loss='logcosh', channel_map={'no_12685_Total-peripheral-resistance-during-PWA_2_1': 0})
field_id_12686_stroke_volume_during_pwa_2_1 = TensorMap('12686_Stroke-volume-during-PWA_2_1', loss='logcosh', channel_map={'no_12686_Stroke-volume-during-PWA_2_1': 0})
field_id_12687_mean_arterial_pressure_during_pwa_2_1 = TensorMap('12687_Mean-arterial-pressure-during-PWA_2_1', loss='logcosh', channel_map={'no_12687_Mean-arterial-pressure-during-PWA_2_1': 0})
field_id_12702_cardiac_index_during_pwa_2_1 = TensorMap('12702_Cardiac-index-during-PWA_2_1', loss='logcosh', channel_map={'no_12702_Cardiac-index-during-PWA_2_1': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_2_2 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_2_2', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_2_2': 0})
field_id_3536_age_started_hormonereplacement_therapy_hrt_2_0 = TensorMap('3536_Age-started-hormonereplacement-therapy-HRT_2_0', loss='logcosh', channel_map={'no_3536_Age-started-hormonereplacement-therapy-HRT_2_0': 0})
field_id_3546_age_last_used_hormonereplacement_therapy_hrt_2_0 = TensorMap('3546_Age-last-used-hormonereplacement-therapy-HRT_2_0', loss='logcosh', channel_map={'no_3546_Age-last-used-hormonereplacement-therapy-HRT_2_0': 0})
field_id_3882_age_at_bilateral_oophorectomy_both_ovaries_removed_0_0 = TensorMap('3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_0_0', loss='logcosh', channel_map={'no_3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_0_0': 0})
field_id_92_operation_yearage_first_occurred_2_2 = TensorMap('92_Operation-yearage-first-occurred_2_2', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_2_2': 0})
field_id_3894_age_heart_attack_diagnosed_0_0 = TensorMap('3894_Age-heart-attack-diagnosed_0_0', loss='logcosh', channel_map={'no_3894_Age-heart-attack-diagnosed_0_0': 0})
field_id_3809_time_since_last_prostate_specific_antigen_psa_test_2_0 = TensorMap('3809_Time-since-last-prostate-specific-antigen-PSA-test_2_0', loss='logcosh', channel_map={'no_3809_Time-since-last-prostate-specific-antigen-PSA-test_2_0': 0})
field_id_20161_pack_years_of_smoking_2_0 = TensorMap('20161_Pack-years-of-smoking_2_0', loss='logcosh', channel_map={'no_20161_Pack-years-of-smoking_2_0': 0})
field_id_20162_pack_years_adult_smoking_as_proportion_of_life_span_exposed_to_smoking_2_0 = TensorMap('20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_2_0', loss='logcosh', channel_map={'no_20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_2_0': 0})
field_id_2867_age_started_smoking_in_former_smokers_2_0 = TensorMap('2867_Age-started-smoking-in-former-smokers_2_0', loss='logcosh', channel_map={'no_2867_Age-started-smoking-in-former-smokers_2_0': 0})
field_id_2887_number_of_cigarettes_previously_smoked_daily_2_0 = TensorMap('2887_Number-of-cigarettes-previously-smoked-daily_2_0', loss='logcosh', channel_map={'no_2887_Number-of-cigarettes-previously-smoked-daily_2_0': 0})
field_id_2897_age_stopped_smoking_2_0 = TensorMap('2897_Age-stopped-smoking_2_0', loss='logcosh', channel_map={'no_2897_Age-stopped-smoking_2_0': 0})
field_id_2926_number_of_unsuccessful_stopsmoking_attempts_2_0 = TensorMap('2926_Number-of-unsuccessful-stopsmoking-attempts_2_0', loss='logcosh', channel_map={'no_2926_Number-of-unsuccessful-stopsmoking-attempts_2_0': 0})
field_id_93_systolic_blood_pressure_manual_reading_0_0 = TensorMap('93_Systolic-blood-pressure-manual-reading_0_0', loss='logcosh', channel_map={'no_93_Systolic-blood-pressure-manual-reading_0_0': 0})
field_id_94_diastolic_blood_pressure_manual_reading_0_0 = TensorMap('94_Diastolic-blood-pressure-manual-reading_0_0', loss='logcosh', channel_map={'no_94_Diastolic-blood-pressure-manual-reading_0_0': 0})
field_id_95_pulse_rate_during_bloodpressure_measurement_0_0 = TensorMap('95_Pulse-rate-during-bloodpressure-measurement_0_0', loss='logcosh', channel_map={'no_95_Pulse-rate-during-bloodpressure-measurement_0_0': 0})
field_id_22602_year_job_started_0_6 = TensorMap('22602_Year-job-started_0_6', loss='logcosh', channel_map={'no_22602_Year-job-started_0_6': 0})
field_id_22603_year_job_ended_0_6 = TensorMap('22603_Year-job-ended_0_6', loss='logcosh', channel_map={'no_22603_Year-job-ended_0_6': 0})
field_id_3829_number_of_stillbirths_2_0 = TensorMap('3829_Number-of-stillbirths_2_0', loss='logcosh', channel_map={'no_3829_Number-of-stillbirths_2_0': 0})
field_id_3839_number_of_spontaneous_miscarriages_2_0 = TensorMap('3839_Number-of-spontaneous-miscarriages_2_0', loss='logcosh', channel_map={'no_3839_Number-of-spontaneous-miscarriages_2_0': 0})
field_id_3849_number_of_pregnancy_terminations_2_0 = TensorMap('3849_Number-of-pregnancy-terminations_2_0', loss='logcosh', channel_map={'no_3849_Number-of-pregnancy-terminations_2_0': 0})
field_id_3680_age_when_last_ate_meat_0_0 = TensorMap('3680_Age-when-last-ate-meat_0_0', loss='logcosh', channel_map={'no_3680_Age-when-last-ate-meat_0_0': 0})
field_id_22642_usual_length_of_each_night_shift_during_mixed_shift_periods_0_1 = TensorMap('22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_1', loss='logcosh', channel_map={'no_22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_1': 0})
field_id_22643_number_of_night_shifts_worked_monthly_during_mixed_shift_periods_0_1 = TensorMap('22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_1', loss='logcosh', channel_map={'no_22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_1': 0})
field_id_22644_consecutive_night_shifts_during_mixed_shift_periods_0_1 = TensorMap('22644_Consecutive-night-shifts-during-mixed-shift-periods_0_1', loss='logcosh', channel_map={'no_22644_Consecutive-night-shifts-during-mixed-shift-periods_0_1': 0})
field_id_22645_rest_days_during_mixed_shift_periods_0_1 = TensorMap('22645_Rest-days-during-mixed-shift-periods_0_1', loss='logcosh', channel_map={'no_22645_Rest-days-during-mixed-shift-periods_0_1': 0})
field_id_20011_interpolated_age_of_participant_when_operation_took_place_1_1 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_1_1', loss='logcosh', channel_map={'no_20011_Interpolated-Age-of-participant-when-operation-took-place_1_1': 0})
field_id_20016_fluid_intelligence_score_1_0 = TensorMap('20016_Fluid-intelligence-score_1_0', loss='logcosh', channel_map={'no_20016_Fluid-intelligence-score_1_0': 0})
field_id_20019_speechreceptionthreshold_srt_estimate_left_1_0 = TensorMap('20019_Speechreceptionthreshold-SRT-estimate-left_1_0', loss='logcosh', channel_map={'no_20019_Speechreceptionthreshold-SRT-estimate-left_1_0': 0})
field_id_20021_speechreceptionthreshold_srt_estimate_right_1_0 = TensorMap('20021_Speechreceptionthreshold-SRT-estimate-right_1_0', loss='logcosh', channel_map={'no_20021_Speechreceptionthreshold-SRT-estimate-right_1_0': 0})
field_id_20128_number_of_fluid_intelligence_questions_attempted_within_time_limit_1_0 = TensorMap('20128_Number-of-fluid-intelligence-questions-attempted-within-time-limit_1_0', loss='logcosh', channel_map={'no_20128_Number-of-fluid-intelligence-questions-attempted-within-time-limit_1_0': 0})
field_id_3062_forced_vital_capacity_fvc_1_2 = TensorMap('3062_Forced-vital-capacity-FVC_1_2', loss='logcosh', channel_map={'no_3062_Forced-vital-capacity-FVC_1_2': 0})
field_id_3063_forced_expiratory_volume_in_1second_fev1_1_2 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_1_2', loss='logcosh', channel_map={'no_3063_Forced-expiratory-volume-in-1second-FEV1_1_2': 0})
field_id_3064_peak_expiratory_flow_pef_1_2 = TensorMap('3064_Peak-expiratory-flow-PEF_1_2', loss='logcosh', channel_map={'no_3064_Peak-expiratory-flow-PEF_1_2': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_2 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_2', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_2': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_4 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_4', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_4': 0})
field_id_4269_number_of_triplets_attempted_left_1_0 = TensorMap('4269_Number-of-triplets-attempted-left_1_0', loss='logcosh', channel_map={'no_4269_Number-of-triplets-attempted-left_1_0': 0})
field_id_4272_duration_of_hearing_test_left_1_0 = TensorMap('4272_Duration-of-hearing-test-left_1_0', loss='logcosh', channel_map={'no_4272_Duration-of-hearing-test-left_1_0': 0})
field_id_4276_number_of_triplets_attempted_right_1_0 = TensorMap('4276_Number-of-triplets-attempted-right_1_0', loss='logcosh', channel_map={'no_4276_Number-of-triplets-attempted-right_1_0': 0})
field_id_4279_duration_of_hearing_test_right_1_0 = TensorMap('4279_Duration-of-hearing-test-right_1_0', loss='logcosh', channel_map={'no_4279_Duration-of-hearing-test-right_1_0': 0})
field_id_5098_6mm_weak_meridian_right_1_0 = TensorMap('5098_6mm-weak-meridian-right_1_0', loss='logcosh', channel_map={'no_5098_6mm-weak-meridian-right_1_0': 0})
field_id_5098_6mm_weak_meridian_right_1_1 = TensorMap('5098_6mm-weak-meridian-right_1_1', loss='logcosh', channel_map={'no_5098_6mm-weak-meridian-right_1_1': 0})
field_id_5098_6mm_weak_meridian_right_1_2 = TensorMap('5098_6mm-weak-meridian-right_1_2', loss='logcosh', channel_map={'no_5098_6mm-weak-meridian-right_1_2': 0})
field_id_5099_3mm_weak_meridian_right_1_2 = TensorMap('5099_3mm-weak-meridian-right_1_2', loss='logcosh', channel_map={'no_5099_3mm-weak-meridian-right_1_2': 0})
field_id_5100_3mm_weak_meridian_angle_right_1_2 = TensorMap('5100_3mm-weak-meridian-angle-right_1_2', loss='logcosh', channel_map={'no_5100_3mm-weak-meridian-angle-right_1_2': 0})
field_id_5101_6mm_weak_meridian_angle_right_1_0 = TensorMap('5101_6mm-weak-meridian-angle-right_1_0', loss='logcosh', channel_map={'no_5101_6mm-weak-meridian-angle-right_1_0': 0})
field_id_5101_6mm_weak_meridian_angle_right_1_1 = TensorMap('5101_6mm-weak-meridian-angle-right_1_1', loss='logcosh', channel_map={'no_5101_6mm-weak-meridian-angle-right_1_1': 0})
field_id_5101_6mm_weak_meridian_angle_right_1_2 = TensorMap('5101_6mm-weak-meridian-angle-right_1_2', loss='logcosh', channel_map={'no_5101_6mm-weak-meridian-angle-right_1_2': 0})
field_id_5106_6mm_strong_meridian_angle_right_1_0 = TensorMap('5106_6mm-strong-meridian-angle-right_1_0', loss='logcosh', channel_map={'no_5106_6mm-strong-meridian-angle-right_1_0': 0})
field_id_5106_6mm_strong_meridian_angle_right_1_1 = TensorMap('5106_6mm-strong-meridian-angle-right_1_1', loss='logcosh', channel_map={'no_5106_6mm-strong-meridian-angle-right_1_1': 0})
field_id_5106_6mm_strong_meridian_angle_right_1_2 = TensorMap('5106_6mm-strong-meridian-angle-right_1_2', loss='logcosh', channel_map={'no_5106_6mm-strong-meridian-angle-right_1_2': 0})
field_id_5107_3mm_strong_meridian_angle_right_1_2 = TensorMap('5107_3mm-strong-meridian-angle-right_1_2', loss='logcosh', channel_map={'no_5107_3mm-strong-meridian-angle-right_1_2': 0})
field_id_5108_3mm_asymmetry_angle_right_1_2 = TensorMap('5108_3mm-asymmetry-angle-right_1_2', loss='logcosh', channel_map={'no_5108_3mm-asymmetry-angle-right_1_2': 0})
field_id_5109_6mm_asymmetry_angle_right_1_0 = TensorMap('5109_6mm-asymmetry-angle-right_1_0', loss='logcosh', channel_map={'no_5109_6mm-asymmetry-angle-right_1_0': 0})
field_id_5109_6mm_asymmetry_angle_right_1_1 = TensorMap('5109_6mm-asymmetry-angle-right_1_1', loss='logcosh', channel_map={'no_5109_6mm-asymmetry-angle-right_1_1': 0})
field_id_5109_6mm_asymmetry_angle_right_1_2 = TensorMap('5109_6mm-asymmetry-angle-right_1_2', loss='logcosh', channel_map={'no_5109_6mm-asymmetry-angle-right_1_2': 0})
field_id_5110_6mm_asymmetry_angle_left_1_0 = TensorMap('5110_6mm-asymmetry-angle-left_1_0', loss='logcosh', channel_map={'no_5110_6mm-asymmetry-angle-left_1_0': 0})
field_id_5114_6mm_cylindrical_power_angle_right_1_0 = TensorMap('5114_6mm-cylindrical-power-angle-right_1_0', loss='logcosh', channel_map={'no_5114_6mm-cylindrical-power-angle-right_1_0': 0})
field_id_5114_6mm_cylindrical_power_angle_right_1_1 = TensorMap('5114_6mm-cylindrical-power-angle-right_1_1', loss='logcosh', channel_map={'no_5114_6mm-cylindrical-power-angle-right_1_1': 0})
field_id_5114_6mm_cylindrical_power_angle_right_1_2 = TensorMap('5114_6mm-cylindrical-power-angle-right_1_2', loss='logcosh', channel_map={'no_5114_6mm-cylindrical-power-angle-right_1_2': 0})
field_id_5115_3mm_cylindrical_power_angle_right_1_2 = TensorMap('5115_3mm-cylindrical-power-angle-right_1_2', loss='logcosh', channel_map={'no_5115_3mm-cylindrical-power-angle-right_1_2': 0})
field_id_5116_3mm_cylindrical_power_right_1_2 = TensorMap('5116_3mm-cylindrical-power-right_1_2', loss='logcosh', channel_map={'no_5116_3mm-cylindrical-power-right_1_2': 0})
field_id_5117_6mm_cylindrical_power_right_1_0 = TensorMap('5117_6mm-cylindrical-power-right_1_0', loss='logcosh', channel_map={'no_5117_6mm-cylindrical-power-right_1_0': 0})
field_id_5117_6mm_cylindrical_power_right_1_1 = TensorMap('5117_6mm-cylindrical-power-right_1_1', loss='logcosh', channel_map={'no_5117_6mm-cylindrical-power-right_1_1': 0})
field_id_5117_6mm_cylindrical_power_right_1_2 = TensorMap('5117_6mm-cylindrical-power-right_1_2', loss='logcosh', channel_map={'no_5117_6mm-cylindrical-power-right_1_2': 0})
field_id_5132_3mm_strong_meridian_right_1_2 = TensorMap('5132_3mm-strong-meridian-right_1_2', loss='logcosh', channel_map={'no_5132_3mm-strong-meridian-right_1_2': 0})
field_id_5133_6mm_strong_meridian_right_1_0 = TensorMap('5133_6mm-strong-meridian-right_1_0', loss='logcosh', channel_map={'no_5133_6mm-strong-meridian-right_1_0': 0})
field_id_5133_6mm_strong_meridian_right_1_1 = TensorMap('5133_6mm-strong-meridian-right_1_1', loss='logcosh', channel_map={'no_5133_6mm-strong-meridian-right_1_1': 0})
field_id_5133_6mm_strong_meridian_right_1_2 = TensorMap('5133_6mm-strong-meridian-right_1_2', loss='logcosh', channel_map={'no_5133_6mm-strong-meridian-right_1_2': 0})
field_id_5157_6mm_asymmetry_index_left_1_0 = TensorMap('5157_6mm-asymmetry-index-left_1_0', loss='logcosh', channel_map={'no_5157_6mm-asymmetry-index-left_1_0': 0})
field_id_5158_6mm_asymmetry_index_right_1_0 = TensorMap('5158_6mm-asymmetry-index-right_1_0', loss='logcosh', channel_map={'no_5158_6mm-asymmetry-index-right_1_0': 0})
field_id_5158_6mm_asymmetry_index_right_1_1 = TensorMap('5158_6mm-asymmetry-index-right_1_1', loss='logcosh', channel_map={'no_5158_6mm-asymmetry-index-right_1_1': 0})
field_id_5158_6mm_asymmetry_index_right_1_2 = TensorMap('5158_6mm-asymmetry-index-right_1_2', loss='logcosh', channel_map={'no_5158_6mm-asymmetry-index-right_1_2': 0})
field_id_5159_3mm_asymmetry_index_right_1_2 = TensorMap('5159_3mm-asymmetry-index-right_1_2', loss='logcosh', channel_map={'no_5159_3mm-asymmetry-index-right_1_2': 0})
field_id_5160_3mm_regularity_index_right_1_2 = TensorMap('5160_3mm-regularity-index-right_1_2', loss='logcosh', channel_map={'no_5160_3mm-regularity-index-right_1_2': 0})
field_id_5161_6mm_regularity_index_right_1_0 = TensorMap('5161_6mm-regularity-index-right_1_0', loss='logcosh', channel_map={'no_5161_6mm-regularity-index-right_1_0': 0})
field_id_5161_6mm_regularity_index_right_1_1 = TensorMap('5161_6mm-regularity-index-right_1_1', loss='logcosh', channel_map={'no_5161_6mm-regularity-index-right_1_1': 0})
field_id_5161_6mm_regularity_index_right_1_2 = TensorMap('5161_6mm-regularity-index-right_1_2', loss='logcosh', channel_map={'no_5161_6mm-regularity-index-right_1_2': 0})
field_id_5162_6mm_regularity_index_left_1_0 = TensorMap('5162_6mm-regularity-index-left_1_0', loss='logcosh', channel_map={'no_5162_6mm-regularity-index-left_1_0': 0})
field_id_5251_6mm_index_of_best_keratometry_results_right_1_0 = TensorMap('5251_6mm-index-of-best-keratometry-results-right_1_0', loss='logcosh', channel_map={'no_5251_6mm-index-of-best-keratometry-results-right_1_0': 0})
field_id_914_duration_of_vigorous_activity_1_0 = TensorMap('914_Duration-of-vigorous-activity_1_0', loss='logcosh', channel_map={'no_914_Duration-of-vigorous-activity_1_0': 0})
field_id_92_operation_yearage_first_occurred_1_1 = TensorMap('92_Operation-yearage-first-occurred_1_1', loss='logcosh', channel_map={'no_92_Operation-yearage-first-occurred_1_1': 0})
field_id_3160_weight_manual_entry_0_0 = TensorMap('3160_Weight-manual-entry_0_0', loss='logcosh', channel_map={'no_3160_Weight-manual-entry_0_0': 0})
field_id_4012_age_deepvein_thrombosis_dvt_blood_clot_in_leg_diagnosed_0_0 = TensorMap('4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_0_0', loss='logcosh', channel_map={'no_4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_0_0': 0})
field_id_2684_years_since_last_breast_cancer_screening_mammogram_1_0 = TensorMap('2684_Years-since-last-breast-cancer-screening-mammogram_1_0', loss='logcosh', channel_map={'no_2684_Years-since-last-breast-cancer-screening-mammogram_1_0': 0})
field_id_2704_years_since_last_cervical_smear_test_1_0 = TensorMap('2704_Years-since-last-cervical-smear-test_1_0', loss='logcosh', channel_map={'no_2704_Years-since-last-cervical-smear-test_1_0': 0})
field_id_2714_age_when_periods_started_menarche_1_0 = TensorMap('2714_Age-when-periods-started-menarche_1_0', loss='logcosh', channel_map={'no_2714_Age-when-periods-started-menarche_1_0': 0})
field_id_2734_number_of_live_births_1_0 = TensorMap('2734_Number-of-live-births_1_0', loss='logcosh', channel_map={'no_2734_Number-of-live-births_1_0': 0})
field_id_2744_birth_weight_of_first_child_1_0 = TensorMap('2744_Birth-weight-of-first-child_1_0', loss='logcosh', channel_map={'no_2744_Birth-weight-of-first-child_1_0': 0})
field_id_2794_age_started_oral_contraceptive_pill_1_0 = TensorMap('2794_Age-started-oral-contraceptive-pill_1_0', loss='logcosh', channel_map={'no_2794_Age-started-oral-contraceptive-pill_1_0': 0})
field_id_2804_age_when_last_used_oral_contraceptive_pill_1_0 = TensorMap('2804_Age-when-last-used-oral-contraceptive-pill_1_0', loss='logcosh', channel_map={'no_2804_Age-when-last-used-oral-contraceptive-pill_1_0': 0})
field_id_3581_age_at_menopause_last_menstrual_period_1_0 = TensorMap('3581_Age-at-menopause-last-menstrual-period_1_0', loss='logcosh', channel_map={'no_3581_Age-at-menopause-last-menstrual-period_1_0': 0})
field_id_4056_age_stroke_diagnosed_0_0 = TensorMap('4056_Age-stroke-diagnosed_0_0', loss='logcosh', channel_map={'no_4056_Age-stroke-diagnosed_0_0': 0})
field_id_404_duration_to_first_press_of_snapbutton_in_each_round_1_3 = TensorMap('404_Duration-to-first-press-of-snapbutton-in-each-round_1_3', loss='logcosh', channel_map={'no_404_Duration-to-first-press-of-snapbutton-in-each-round_1_3': 0})
