# TensorMaps automatically generated by tensor_map_maker.py
# DO NOT EDIT

from ml4h.defines import StorageType
from ml4h.metrics import weighted_crossentropy
from ml4h.tensormap.ukb.demographics import prevalent_incident_tensor
from ml4h.TensorMap import TensorMap, Interpretation, make_range_validator



#  TensorMaps for MPG disease phenotypes
allergic_rhinitis = TensorMap('allergic_rhinitis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_allergic_rhinitis':0, 'allergic_rhinitis':1}, loss=weighted_crossentropy([1.0, 8], 'allergic_rhinitis'))
alzheimers_or_dementia = TensorMap('alzheimers_or_dementia', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_alzheimers_or_dementia':0, 'alzheimers_or_dementia':1}, loss=weighted_crossentropy([1.0, 658], 'alzheimers_or_dementia'))
amyotrophic_lateral_sclerosis = TensorMap('amyotrophic_lateral_sclerosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_amyotrophic_lateral_sclerosis':0, 'amyotrophic_lateral_sclerosis':1}, loss=weighted_crossentropy([1.0, 867], 'amyotrophic_lateral_sclerosis'))
anxiety = TensorMap('anxiety', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_anxiety':0, 'anxiety':1}, loss=weighted_crossentropy([1.0, 29], 'anxiety'))
aortic_regurgitation = TensorMap('aortic_regurgitation', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_aortic_regurgitation':0, 'aortic_regurgitation':1}, loss=weighted_crossentropy([1.0, 184], 'aortic_regurgitation'))
aortic_stenosis = TensorMap('aortic_stenosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_aortic_stenosis':0, 'aortic_stenosis':1}, loss=weighted_crossentropy([1.0, 107], 'aortic_stenosis'))
aortic_valve_disease = TensorMap('aortic_valve_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_aortic_valve_disease':0, 'aortic_valve_disease':1}, loss=weighted_crossentropy([1.0, 58], 'aortic_valve_disease'))
asthma = TensorMap('asthma', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_asthma':0, 'asthma':1}, loss=weighted_crossentropy([1.0, 3], 'asthma'))
atrial_fibrillation_or_flutter = TensorMap('atrial_fibrillation_or_flutter', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_atrial_fibrillation_or_flutter':0, 'atrial_fibrillation_or_flutter':1}, loss=weighted_crossentropy([1.0, 13], 'atrial_fibrillation_or_flutter'))
atrial_septal_defect_patent_foramen_ovale = TensorMap('atrial_septal_defect_patent_foramen_ovale', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_atrial_septal_defect_patent_foramen_ovale':0, 'atrial_septal_defect_patent_foramen_ovale':1}, loss=weighted_crossentropy([1.0, 425], 'atrial_septal_defect_patent_foramen_ovale'))
atrial_septal_defect_patent_foramen_ovale_v2 = TensorMap('atrial_septal_defect_patent_foramen_ovale_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_atrial_septal_defect_patent_foramen_ovale_v2':0, 'atrial_septal_defect_patent_foramen_ovale_v2':1}, loss=weighted_crossentropy([1.0, 425], 'atrial_septal_defect_patent_foramen_ovale_v2'))
back_pain = TensorMap('back_pain', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_back_pain':0, 'back_pain':1}, loss=weighted_crossentropy([1.0, 11], 'back_pain'))
bariatric_surgery = TensorMap('bariatric_surgery', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bariatric_surgery':0, 'bariatric_surgery':1}, loss=weighted_crossentropy([1.0, 330], 'bariatric_surgery'))
bariatric_surgery_v2 = TensorMap('bariatric_surgery_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bariatric_surgery_v2':0, 'bariatric_surgery_v2':1}, loss=weighted_crossentropy([1.0, 512], 'bariatric_surgery_v2'))
bipolar_disorder = TensorMap('bipolar_disorder', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bipolar_disorder':0, 'bipolar_disorder':1}, loss=weighted_crossentropy([1.0, 127], 'bipolar_disorder'))
bladder_cancer = TensorMap('bladder_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bladder_cancer':0, 'bladder_cancer':1}, loss=weighted_crossentropy([1.0, 89], 'bladder_cancer'))
bradyarrhythmia_av_block_or_distal_conduction_disease = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease':0, 'bradyarrhythmia_av_block_or_distal_conduction_disease':1}, loss=weighted_crossentropy([1.0, 40], 'bradyarrhythmia_av_block_or_distal_conduction_disease'))
bradyarrhythmia_av_block_or_distal_conduction_disease_hard = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_hard', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_hard':0, 'bradyarrhythmia_av_block_or_distal_conduction_disease_hard':1}, loss=weighted_crossentropy([1.0, 70], 'bradyarrhythmia_av_block_or_distal_conduction_disease_hard'))
bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2 = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2':0, 'bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2':1}, loss=weighted_crossentropy([1.0, 73], 'bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2'))
bradyarrhythmia_av_block_or_distal_conduction_disease_v2 = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_v2':0, 'bradyarrhythmia_av_block_or_distal_conduction_disease_v2':1}, loss=weighted_crossentropy([1.0, 61], 'bradyarrhythmia_av_block_or_distal_conduction_disease_v2'))
bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted':0, 'bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted':1}, loss=weighted_crossentropy([1.0, 266], 'bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted'))
bradyarrhythmia_general_inclusive_definition = TensorMap('bradyarrhythmia_general_inclusive_definition', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_general_inclusive_definition':0, 'bradyarrhythmia_general_inclusive_definition':1}, loss=weighted_crossentropy([1.0, 26], 'bradyarrhythmia_general_inclusive_definition'))
bradyarrhythmia_only = TensorMap('bradyarrhythmia_only', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_only':0, 'bradyarrhythmia_only':1}, loss=weighted_crossentropy([1.0, 48], 'bradyarrhythmia_only'))
bradyarrhythmia_pacemaker = TensorMap('bradyarrhythmia_pacemaker', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_pacemaker':0, 'bradyarrhythmia_pacemaker':1}, loss=weighted_crossentropy([1.0, 52], 'bradyarrhythmia_pacemaker'))
bradyarrhythmia_pacemaker_v2 = TensorMap('bradyarrhythmia_pacemaker_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_pacemaker_v2':0, 'bradyarrhythmia_pacemaker_v2':1}, loss=weighted_crossentropy([1.0, 51], 'bradyarrhythmia_pacemaker_v2'))
bradyarrhythmia_sinus_node_dysfunction = TensorMap('bradyarrhythmia_sinus_node_dysfunction', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_sinus_node_dysfunction':0, 'bradyarrhythmia_sinus_node_dysfunction':1}, loss=weighted_crossentropy([1.0, 504], 'bradyarrhythmia_sinus_node_dysfunction'))
bradyarrhythmia_sinus_node_dysfunction_hard = TensorMap('bradyarrhythmia_sinus_node_dysfunction_hard', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_hard':0, 'bradyarrhythmia_sinus_node_dysfunction_hard':1}, loss=weighted_crossentropy([1.0, 733], 'bradyarrhythmia_sinus_node_dysfunction_hard'))
bradyarrhythmia_sinus_node_dysfunction_hard_v2 = TensorMap('bradyarrhythmia_sinus_node_dysfunction_hard_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_hard_v2':0, 'bradyarrhythmia_sinus_node_dysfunction_hard_v2':1}, loss=weighted_crossentropy([1.0, 747], 'bradyarrhythmia_sinus_node_dysfunction_hard_v2'))
bradyarrhythmia_sinus_node_dysfunction_v2 = TensorMap('bradyarrhythmia_sinus_node_dysfunction_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_v2':0, 'bradyarrhythmia_sinus_node_dysfunction_v2':1}, loss=weighted_crossentropy([1.0, 639], 'bradyarrhythmia_sinus_node_dysfunction_v2'))
bradyarrhythmia_sinus_node_dysfunction_v2_restricted = TensorMap('bradyarrhythmia_sinus_node_dysfunction_v2_restricted', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_v2_restricted':0, 'bradyarrhythmia_sinus_node_dysfunction_v2_restricted':1}, loss=weighted_crossentropy([1.0, 828], 'bradyarrhythmia_sinus_node_dysfunction_v2_restricted'))
bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical = TensorMap('bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical':0, 'bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical':1}, loss=weighted_crossentropy([1.0, 48], 'bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical'))
bradyarrrhythmia_general_inclusive_definition_clinical = TensorMap('bradyarrrhythmia_general_inclusive_definition_clinical', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_bradyarrrhythmia_general_inclusive_definition_clinical':0, 'bradyarrrhythmia_general_inclusive_definition_clinical':1}, loss=weighted_crossentropy([1.0, 47], 'bradyarrrhythmia_general_inclusive_definition_clinical'))
breast_cancer = TensorMap('breast_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_breast_cancer':0, 'breast_cancer':1}, loss=weighted_crossentropy([1.0, 15], 'breast_cancer'))
cardiac_surgery = TensorMap('cardiac_surgery', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_cardiac_surgery':0, 'cardiac_surgery':1}, loss=weighted_crossentropy([1.0, 33], 'cardiac_surgery'))
cataract = TensorMap('cataract', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_cataract':0, 'cataract':1}, loss=weighted_crossentropy([1.0, 8], 'cataract'))
cervical_cancer = TensorMap('cervical_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_cervical_cancer':0, 'cervical_cancer':1}, loss=weighted_crossentropy([1.0, 117], 'cervical_cancer'))
cholelithiasis_or_gallstones = TensorMap('cholelithiasis_or_gallstones', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_cholelithiasis_or_gallstones':0, 'cholelithiasis_or_gallstones':1}, loss=weighted_crossentropy([1.0, 11], 'cholelithiasis_or_gallstones'))
chronic_kidney_disease = TensorMap('chronic_kidney_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_chronic_kidney_disease':0, 'chronic_kidney_disease':1}, loss=weighted_crossentropy([1.0, 36], 'chronic_kidney_disease'))
chronic_obstructive_pulmonary_disease = TensorMap('chronic_obstructive_pulmonary_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_chronic_obstructive_pulmonary_disease':0, 'chronic_obstructive_pulmonary_disease':1}, loss=weighted_crossentropy([1.0, 14], 'chronic_obstructive_pulmonary_disease'))
colorectal_cancer = TensorMap('colorectal_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_colorectal_cancer':0, 'colorectal_cancer':1}, loss=weighted_crossentropy([1.0, 50], 'colorectal_cancer'))
conduction_system_disease_only = TensorMap('conduction_system_disease_only', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_conduction_system_disease_only':0, 'conduction_system_disease_only':1}, loss=weighted_crossentropy([1.0, 49], 'conduction_system_disease_only'))
congenital_heart_disease = TensorMap('congenital_heart_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_congenital_heart_disease':0, 'congenital_heart_disease':1}, loss=weighted_crossentropy([1.0, 218], 'congenital_heart_disease'))
congenital_heart_disease_ebstein_anomaly = TensorMap('congenital_heart_disease_ebstein_anomaly', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_congenital_heart_disease_ebstein_anomaly':0, 'congenital_heart_disease_ebstein_anomaly':1}, loss=weighted_crossentropy([1.0, 14813], 'congenital_heart_disease_ebstein_anomaly'))
coronary_artery_disease = TensorMap('coronary_artery_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_coronary_artery_disease':0, 'coronary_artery_disease':1}, loss=weighted_crossentropy([1.0, 10], 'coronary_artery_disease'))
coronary_artery_disease_hard = TensorMap('coronary_artery_disease_hard', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_coronary_artery_disease_hard':0, 'coronary_artery_disease_hard':1}, loss=weighted_crossentropy([1.0, 9], 'coronary_artery_disease_hard'))
coronary_artery_disease_intermediate = TensorMap('coronary_artery_disease_intermediate', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_coronary_artery_disease_intermediate':0, 'coronary_artery_disease_intermediate':1}, loss=weighted_crossentropy([1.0, 7], 'coronary_artery_disease_intermediate'))
coronary_artery_disease_soft = TensorMap('coronary_artery_disease_soft', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_coronary_artery_disease_soft':0, 'coronary_artery_disease_soft':1}, loss=weighted_crossentropy([1.0, 5], 'coronary_artery_disease_soft'))
crohns_disease = TensorMap('crohns_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_crohns_disease':0, 'crohns_disease':1}, loss=weighted_crossentropy([1.0, 119], 'crohns_disease'))
death = TensorMap('death', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_death':0, 'death':1}, loss=weighted_crossentropy([1.0, 17], 'death'))
depression = TensorMap('depression', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_depression':0, 'depression':1}, loss=weighted_crossentropy([1.0, 6], 'depression'))
dermatitis = TensorMap('dermatitis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_dermatitis':0, 'dermatitis':1}, loss=weighted_crossentropy([1.0, 14], 'dermatitis'))
diabetes_all = TensorMap('diabetes_all', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_diabetes_all':0, 'diabetes_all':1}, loss=weighted_crossentropy([1.0, 7], 'diabetes_all'))
diabetes_type_1 = TensorMap('diabetes_type_1', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_diabetes_type_1':0, 'diabetes_type_1':1}, loss=weighted_crossentropy([1.0, 65], 'diabetes_type_1'))
diabetes_type_2 = TensorMap('diabetes_type_2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_diabetes_type_2':0, 'diabetes_type_2':1}, loss=weighted_crossentropy([1.0, 9], 'diabetes_type_2'))
diverticular_disease = TensorMap('diverticular_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_diverticular_disease':0, 'diverticular_disease':1}, loss=weighted_crossentropy([1.0, 7], 'diverticular_disease'))
enlarged_prostate = TensorMap('enlarged_prostate', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_enlarged_prostate':0, 'enlarged_prostate':1}, loss=weighted_crossentropy([1.0, 14], 'enlarged_prostate'))
epilepsy = TensorMap('epilepsy', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_epilepsy':0, 'epilepsy':1}, loss=weighted_crossentropy([1.0, 41], 'epilepsy'))
gastroesophageal_reflux_disease = TensorMap('gastroesophageal_reflux_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_gastroesophageal_reflux_disease':0, 'gastroesophageal_reflux_disease':1}, loss=weighted_crossentropy([1.0, 5], 'gastroesophageal_reflux_disease'))
glaucoma = TensorMap('glaucoma', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_glaucoma':0, 'glaucoma':1}, loss=weighted_crossentropy([1.0, 28], 'glaucoma'))
gout = TensorMap('gout', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_gout':0, 'gout':1}, loss=weighted_crossentropy([1.0, 27], 'gout'))
heart_failure = TensorMap('heart_failure', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_heart_failure':0, 'heart_failure':1}, loss=weighted_crossentropy([1.0, 32], 'heart_failure'))
heart_failure_v2 = TensorMap('heart_failure_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_heart_failure_v2':0, 'heart_failure_v2':1}, loss=weighted_crossentropy([1.0, 32], 'heart_failure_v2'))
hypercholesterolemia = TensorMap('hypercholesterolemia', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_hypercholesterolemia':0, 'hypercholesterolemia':1}, loss=weighted_crossentropy([1.0, 2], 'hypercholesterolemia'))
hypertension = TensorMap('hypertension', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_hypertension':0, 'hypertension':1}, loss=weighted_crossentropy([1.0, 1], 'hypertension'))
hyperthyroidism = TensorMap('hyperthyroidism', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_hyperthyroidism':0, 'hyperthyroidism':1}, loss=weighted_crossentropy([1.0, 44], 'hyperthyroidism'))
hypertrophic_cardiomyopathy = TensorMap('hypertrophic_cardiomyopathy', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_hypertrophic_cardiomyopathy':0, 'hypertrophic_cardiomyopathy':1}, loss=weighted_crossentropy([1.0, 679], 'hypertrophic_cardiomyopathy'))
hypothyroidism = TensorMap('hypothyroidism', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_hypothyroidism':0, 'hypothyroidism':1}, loss=weighted_crossentropy([1.0, 8], 'hypothyroidism'))
implantable_cardioverter_defibrillator = TensorMap('implantable_cardioverter_defibrillator', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_implantable_cardioverter_defibrillator':0, 'implantable_cardioverter_defibrillator':1}, loss=weighted_crossentropy([1.0, 328], 'implantable_cardioverter_defibrillator'))
inflammatory_bowel_disease = TensorMap('inflammatory_bowel_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_inflammatory_bowel_disease':0, 'inflammatory_bowel_disease':1}, loss=weighted_crossentropy([1.0, 37], 'inflammatory_bowel_disease'))
intervertebral_disc_displacement = TensorMap('intervertebral_disc_displacement', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_intervertebral_disc_displacement':0, 'intervertebral_disc_displacement':1}, loss=weighted_crossentropy([1.0, 21], 'intervertebral_disc_displacement'))
intracerebral_hemorrhage = TensorMap('intracerebral_hemorrhage', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_intracerebral_hemorrhage':0, 'intracerebral_hemorrhage':1}, loss=weighted_crossentropy([1.0, 234], 'intracerebral_hemorrhage'))
iron_deficiency_anemia = TensorMap('iron_deficiency_anemia', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_iron_deficiency_anemia':0, 'iron_deficiency_anemia':1}, loss=weighted_crossentropy([1.0, 20], 'iron_deficiency_anemia'))
irritable_bowel_syndrome = TensorMap('irritable_bowel_syndrome', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_irritable_bowel_syndrome':0, 'irritable_bowel_syndrome':1}, loss=weighted_crossentropy([1.0, 14], 'irritable_bowel_syndrome'))
ischemic_stroke = TensorMap('ischemic_stroke', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ischemic_stroke':0, 'ischemic_stroke':1}, loss=weighted_crossentropy([1.0, 50], 'ischemic_stroke'))
lung_cancer = TensorMap('lung_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_lung_cancer':0, 'lung_cancer':1}, loss=weighted_crossentropy([1.0, 87], 'lung_cancer'))
lupus_erythematosus = TensorMap('lupus_erythematosus', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_lupus_erythematosus':0, 'lupus_erythematosus':1}, loss=weighted_crossentropy([1.0, 268], 'lupus_erythematosus'))
migraine = TensorMap('migraine', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_migraine':0, 'migraine':1}, loss=weighted_crossentropy([1.0, 14], 'migraine'))
mitral_regurgitation = TensorMap('mitral_regurgitation', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_mitral_regurgitation':0, 'mitral_regurgitation':1}, loss=weighted_crossentropy([1.0, 89], 'mitral_regurgitation'))
mitral_stenosis = TensorMap('mitral_stenosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_mitral_stenosis':0, 'mitral_stenosis':1}, loss=weighted_crossentropy([1.0, 1269], 'mitral_stenosis'))
mitral_valve_disease = TensorMap('mitral_valve_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_mitral_valve_disease':0, 'mitral_valve_disease':1}, loss=weighted_crossentropy([1.0, 52], 'mitral_valve_disease'))
mitral_valve_prolapse = TensorMap('mitral_valve_prolapse', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_mitral_valve_prolapse':0, 'mitral_valve_prolapse':1}, loss=weighted_crossentropy([1.0, 373], 'mitral_valve_prolapse'))
multiple_sclerosis = TensorMap('multiple_sclerosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_multiple_sclerosis':0, 'multiple_sclerosis':1}, loss=weighted_crossentropy([1.0, 114], 'multiple_sclerosis'))
myocardial_infarction = TensorMap('myocardial_infarction', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_myocardial_infarction':0, 'myocardial_infarction':1}, loss=weighted_crossentropy([1.0, 13], 'myocardial_infarction'))
osteoarthritis = TensorMap('osteoarthritis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_osteoarthritis':0, 'osteoarthritis':1}, loss=weighted_crossentropy([1.0, 3], 'osteoarthritis'))
osteoporosis = TensorMap('osteoporosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_osteoporosis':0, 'osteoporosis':1}, loss=weighted_crossentropy([1.0, 18], 'osteoporosis'))
pancreatitis = TensorMap('pancreatitis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_pancreatitis':0, 'pancreatitis':1}, loss=weighted_crossentropy([1.0, 80], 'pancreatitis'))
parkinsons_disease = TensorMap('parkinsons_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_parkinsons_disease':0, 'parkinsons_disease':1}, loss=weighted_crossentropy([1.0, 159], 'parkinsons_disease'))
peripheral_vascular_disease = TensorMap('peripheral_vascular_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_peripheral_vascular_disease':0, 'peripheral_vascular_disease':1}, loss=weighted_crossentropy([1.0, 41], 'peripheral_vascular_disease'))
pneumonia = TensorMap('pneumonia', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_pneumonia':0, 'pneumonia':1}, loss=weighted_crossentropy([1.0, 13], 'pneumonia'))
posttraumatic_stress_disorder = TensorMap('posttraumatic_stress_disorder', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_posttraumatic_stress_disorder':0, 'posttraumatic_stress_disorder':1}, loss=weighted_crossentropy([1.0, 508], 'posttraumatic_stress_disorder'))
prostate_cancer = TensorMap('prostate_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_prostate_cancer':0, 'prostate_cancer':1}, loss=weighted_crossentropy([1.0, 32], 'prostate_cancer'))
psoriasis = TensorMap('psoriasis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_psoriasis':0, 'psoriasis':1}, loss=weighted_crossentropy([1.0, 34], 'psoriasis'))
pulmonary_hypertension = TensorMap('pulmonary_hypertension', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_pulmonary_hypertension':0, 'pulmonary_hypertension':1}, loss=weighted_crossentropy([1.0, 273], 'pulmonary_hypertension'))
pulmonary_regurgitation = TensorMap('pulmonary_regurgitation', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_pulmonary_regurgitation':0, 'pulmonary_regurgitation':1}, loss=weighted_crossentropy([1.0, 2234], 'pulmonary_regurgitation'))
pulmonary_stenosis = TensorMap('pulmonary_stenosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_pulmonary_stenosis':0, 'pulmonary_stenosis':1}, loss=weighted_crossentropy([1.0, 14813], 'pulmonary_stenosis'))
pulmonary_valve_disease = TensorMap('pulmonary_valve_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_pulmonary_valve_disease':0, 'pulmonary_valve_disease':1}, loss=weighted_crossentropy([1.0, 1746], 'pulmonary_valve_disease'))
rheumatoid_arthritis = TensorMap('rheumatoid_arthritis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_rheumatoid_arthritis':0, 'rheumatoid_arthritis':1}, loss=weighted_crossentropy([1.0, 29], 'rheumatoid_arthritis'))
sarcoidosis = TensorMap('sarcoidosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_sarcoidosis':0, 'sarcoidosis':1}, loss=weighted_crossentropy([1.0, 182], 'sarcoidosis'))
schizophrenia = TensorMap('schizophrenia', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_schizophrenia':0, 'schizophrenia':1}, loss=weighted_crossentropy([1.0, 236], 'schizophrenia'))
sciatica = TensorMap('sciatica', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_sciatica':0, 'sciatica':1}, loss=weighted_crossentropy([1.0, 32], 'sciatica'))
skin_cancer = TensorMap('skin_cancer', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_skin_cancer':0, 'skin_cancer':1}, loss=weighted_crossentropy([1.0, 11], 'skin_cancer'))
sleep_apnea = TensorMap('sleep_apnea', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_sleep_apnea':0, 'sleep_apnea':1}, loss=weighted_crossentropy([1.0, 39], 'sleep_apnea'))
stroke = TensorMap('stroke', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_stroke':0, 'stroke':1}, loss=weighted_crossentropy([1.0, 21], 'stroke'))
subarachnoid_hemorrhage = TensorMap('subarachnoid_hemorrhage', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_subarachnoid_hemorrhage':0, 'subarachnoid_hemorrhage':1}, loss=weighted_crossentropy([1.0, 186], 'subarachnoid_hemorrhage'))
supraventricular_arrhythmia_general_inclusive_definition = TensorMap('supraventricular_arrhythmia_general_inclusive_definition', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_supraventricular_arrhythmia_general_inclusive_definition':0, 'supraventricular_arrhythmia_general_inclusive_definition':1}, loss=weighted_crossentropy([1.0, 68], 'supraventricular_arrhythmia_general_inclusive_definition'))
supraventricular_arrhythmia_general_inclusive_definition_clinical = TensorMap('supraventricular_arrhythmia_general_inclusive_definition_clinical', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_supraventricular_arrhythmia_general_inclusive_definition_clinical':0, 'supraventricular_arrhythmia_general_inclusive_definition_clinical':1}, loss=weighted_crossentropy([1.0, 66], 'supraventricular_arrhythmia_general_inclusive_definition_clinical'))
supraventricular_arrhythmia_premature_atrial_contractions = TensorMap('supraventricular_arrhythmia_premature_atrial_contractions', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_supraventricular_arrhythmia_premature_atrial_contractions':0, 'supraventricular_arrhythmia_premature_atrial_contractions':1}, loss=weighted_crossentropy([1.0, 1659], 'supraventricular_arrhythmia_premature_atrial_contractions'))
supraventricular_arrhythmia_svt = TensorMap('supraventricular_arrhythmia_svt', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_supraventricular_arrhythmia_svt':0, 'supraventricular_arrhythmia_svt':1}, loss=weighted_crossentropy([1.0, 69], 'supraventricular_arrhythmia_svt'))
supraventricular_arrhythmia_wpw = TensorMap('supraventricular_arrhythmia_wpw', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_supraventricular_arrhythmia_wpw':0, 'supraventricular_arrhythmia_wpw':1}, loss=weighted_crossentropy([1.0, 603], 'supraventricular_arrhythmia_wpw'))
supraventricular_arrhythmia_wpw_v2 = TensorMap('supraventricular_arrhythmia_wpw_v2', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_supraventricular_arrhythmia_wpw_v2':0, 'supraventricular_arrhythmia_wpw_v2':1}, loss=weighted_crossentropy([1.0, 612], 'supraventricular_arrhythmia_wpw_v2'))
tricuspid_regurgitation = TensorMap('tricuspid_regurgitation', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_tricuspid_regurgitation':0, 'tricuspid_regurgitation':1}, loss=weighted_crossentropy([1.0, 493], 'tricuspid_regurgitation'))
tricuspid_stenosis = TensorMap('tricuspid_stenosis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_tricuspid_stenosis':0, 'tricuspid_stenosis':1}, loss=weighted_crossentropy([1.0, 26664], 'tricuspid_stenosis'))
tricuspid_valve_disease = TensorMap('tricuspid_valve_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_tricuspid_valve_disease':0, 'tricuspid_valve_disease':1}, loss=weighted_crossentropy([1.0, 139], 'tricuspid_valve_disease'))
ulcerative_colitis = TensorMap('ulcerative_colitis', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ulcerative_colitis':0, 'ulcerative_colitis':1}, loss=weighted_crossentropy([1.0, 53], 'ulcerative_colitis'))
uterine_fibroids = TensorMap('uterine_fibroids', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_uterine_fibroids':0, 'uterine_fibroids':1}, loss=weighted_crossentropy([1.0, 11], 'uterine_fibroids'))
valve_disease = TensorMap('valve_disease', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_valve_disease':0, 'valve_disease':1}, loss=weighted_crossentropy([1.0, 30], 'valve_disease'))
valvular_disease_unspecified = TensorMap('valvular_disease_unspecified', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_valvular_disease_unspecified':0, 'valvular_disease_unspecified':1}, loss=weighted_crossentropy([1.0, 65], 'valvular_disease_unspecified'))
venous_thromboembolism = TensorMap('venous_thromboembolism', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_venous_thromboembolism':0, 'venous_thromboembolism':1}, loss=weighted_crossentropy([1.0, 15], 'venous_thromboembolism'))
ventricular_arrhythmia_general_inclusive_definition = TensorMap('ventricular_arrhythmia_general_inclusive_definition', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ventricular_arrhythmia_general_inclusive_definition':0, 'ventricular_arrhythmia_general_inclusive_definition':1}, loss=weighted_crossentropy([1.0, 80], 'ventricular_arrhythmia_general_inclusive_definition'))
ventricular_arrhythmia_general_inclusive_definition_clinical = TensorMap('ventricular_arrhythmia_general_inclusive_definition_clinical', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ventricular_arrhythmia_general_inclusive_definition_clinical':0, 'ventricular_arrhythmia_general_inclusive_definition_clinical':1}, loss=weighted_crossentropy([1.0, 93], 'ventricular_arrhythmia_general_inclusive_definition_clinical'))
ventricular_arrhythmia_sudden_cardiac_death = TensorMap('ventricular_arrhythmia_sudden_cardiac_death', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ventricular_arrhythmia_sudden_cardiac_death':0, 'ventricular_arrhythmia_sudden_cardiac_death':1}, loss=weighted_crossentropy([1.0, 169], 'ventricular_arrhythmia_sudden_cardiac_death'))
ventricular_arrhythmia_ventricular_premature_depolarizations = TensorMap('ventricular_arrhythmia_ventricular_premature_depolarizations', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ventricular_arrhythmia_ventricular_premature_depolarizations':0, 'ventricular_arrhythmia_ventricular_premature_depolarizations':1}, loss=weighted_crossentropy([1.0, 484], 'ventricular_arrhythmia_ventricular_premature_depolarizations'))
ventricular_arrhythmia_vt = TensorMap('ventricular_arrhythmia_vt', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', channel_map={'no_ventricular_arrhythmia_vt':0, 'ventricular_arrhythmia_vt':1}, loss=weighted_crossentropy([1.0, 169], 'ventricular_arrhythmia_vt'))


#  TensorMaps for prevalent and incident MPG disease phenotypes
allergic_rhinitis_prevalent_incident = TensorMap('allergic_rhinitis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/allergic_rhinitis_date'), channel_map={'no_allergic_rhinitis':0, 'prevalent_allergic_rhinitis':1, 'incident_allergic_rhinitis':2}, loss=weighted_crossentropy([1.0, 5, 69], 'allergic_rhinitis_prevalent_incident'))
alzheimers_or_dementia_prevalent_incident = TensorMap('alzheimers_or_dementia', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/alzheimers_or_dementia_date'), channel_map={'no_alzheimers_or_dementia':0, 'prevalent_alzheimers_or_dementia':1, 'incident_alzheimers_or_dementia':2}, loss=weighted_crossentropy([1.0, 1339, 691], 'alzheimers_or_dementia_prevalent_incident'))
amyotrophic_lateral_sclerosis_prevalent_incident = TensorMap('amyotrophic_lateral_sclerosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/amyotrophic_lateral_sclerosis_date'), channel_map={'no_amyotrophic_lateral_sclerosis':0, 'prevalent_amyotrophic_lateral_sclerosis':1, 'incident_amyotrophic_lateral_sclerosis':2}, loss=weighted_crossentropy([1.0, 4053, 663], 'amyotrophic_lateral_sclerosis_prevalent_incident'))
anxiety_prevalent_incident = TensorMap('anxiety', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/anxiety_date'), channel_map={'no_anxiety':0, 'prevalent_anxiety':1, 'incident_anxiety':2}, loss=weighted_crossentropy([1.0, 56, 29], 'anxiety_prevalent_incident'))
aortic_regurgitation_prevalent_incident = TensorMap('aortic_regurgitation', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/aortic_regurgitation_date'), channel_map={'no_aortic_regurgitation':0, 'prevalent_aortic_regurgitation':1, 'incident_aortic_regurgitation':2}, loss=weighted_crossentropy([1.0, 291, 199], 'aortic_regurgitation_prevalent_incident'))
aortic_stenosis_prevalent_incident = TensorMap('aortic_stenosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/aortic_stenosis_date'), channel_map={'no_aortic_stenosis':0, 'prevalent_aortic_stenosis':1, 'incident_aortic_stenosis':2}, loss=weighted_crossentropy([1.0, 200, 106], 'aortic_stenosis_prevalent_incident'))
aortic_valve_disease_prevalent_incident = TensorMap('aortic_valve_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/aortic_valve_disease_date'), channel_map={'no_aortic_valve_disease':0, 'prevalent_aortic_valve_disease':1, 'incident_aortic_valve_disease':2}, loss=weighted_crossentropy([1.0, 93, 63], 'aortic_valve_disease_prevalent_incident'))
asthma_prevalent_incident = TensorMap('asthma', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/asthma_date'), channel_map={'no_asthma':0, 'prevalent_asthma':1, 'incident_asthma':2}, loss=weighted_crossentropy([1.0, 2, 32], 'asthma_prevalent_incident'))
atrial_fibrillation_or_flutter_prevalent_incident = TensorMap('atrial_fibrillation_or_flutter', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/atrial_fibrillation_or_flutter_date'), channel_map={'no_atrial_fibrillation_or_flutter':0, 'prevalent_atrial_fibrillation_or_flutter':1, 'incident_atrial_fibrillation_or_flutter':2}, loss=weighted_crossentropy([1.0, 21, 15], 'atrial_fibrillation_or_flutter_prevalent_incident'))
atrial_septal_defect_patent_foramen_ovale_prevalent_incident = TensorMap('atrial_septal_defect_patent_foramen_ovale', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/atrial_septal_defect_patent_foramen_ovale_date'), channel_map={'no_atrial_septal_defect_patent_foramen_ovale':0, 'prevalent_atrial_septal_defect_patent_foramen_ovale':1, 'incident_atrial_septal_defect_patent_foramen_ovale':2}, loss=weighted_crossentropy([1.0, 518, 524], 'atrial_septal_defect_patent_foramen_ovale_prevalent_incident'))
atrial_septal_defect_patent_foramen_ovale_v2_prevalent_incident = TensorMap('atrial_septal_defect_patent_foramen_ovale_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/atrial_septal_defect_patent_foramen_ovale_v2_date'), channel_map={'no_atrial_septal_defect_patent_foramen_ovale_v2':0, 'prevalent_atrial_septal_defect_patent_foramen_ovale_v2':1, 'incident_atrial_septal_defect_patent_foramen_ovale_v2':2}, loss=weighted_crossentropy([1.0, 518, 524], 'atrial_septal_defect_patent_foramen_ovale_v2_prevalent_incident'))
back_pain_prevalent_incident = TensorMap('back_pain', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/back_pain_date'), channel_map={'no_back_pain':0, 'prevalent_back_pain':1, 'incident_back_pain':2}, loss=weighted_crossentropy([1.0, 13, 17], 'back_pain_prevalent_incident'))
bariatric_surgery_prevalent_incident = TensorMap('bariatric_surgery', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bariatric_surgery_date'), channel_map={'no_bariatric_surgery':0, 'prevalent_bariatric_surgery':1, 'incident_bariatric_surgery':2}, loss=weighted_crossentropy([1.0, 630, 317], 'bariatric_surgery_prevalent_incident'))
bariatric_surgery_v2_prevalent_incident = TensorMap('bariatric_surgery_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bariatric_surgery_v2_date'), channel_map={'no_bariatric_surgery_v2':0, 'prevalent_bariatric_surgery_v2':1, 'incident_bariatric_surgery_v2':2}, loss=weighted_crossentropy([1.0, 1239, 441], 'bariatric_surgery_v2_prevalent_incident'))
bipolar_disorder_prevalent_incident = TensorMap('bipolar_disorder', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bipolar_disorder_date'), channel_map={'no_bipolar_disorder':0, 'prevalent_bipolar_disorder':1, 'incident_bipolar_disorder':2}, loss=weighted_crossentropy([1.0, 106, 445], 'bipolar_disorder_prevalent_incident'))
bladder_cancer_prevalent_incident = TensorMap('bladder_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bladder_cancer_date'), channel_map={'no_bladder_cancer':0, 'prevalent_bladder_cancer':1, 'incident_bladder_cancer':2}, loss=weighted_crossentropy([1.0, 122, 121], 'bladder_cancer_prevalent_incident'))
bradyarrhythmia_av_block_or_distal_conduction_disease_prevalent_incident = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_av_block_or_distal_conduction_disease_date'), channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease':0, 'prevalent_bradyarrhythmia_av_block_or_distal_conduction_disease':1, 'incident_bradyarrhythmia_av_block_or_distal_conduction_disease':2}, loss=weighted_crossentropy([1.0, 92, 39], 'bradyarrhythmia_av_block_or_distal_conduction_disease_prevalent_incident'))
bradyarrhythmia_av_block_or_distal_conduction_disease_hard_prevalent_incident = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_hard', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_av_block_or_distal_conduction_disease_hard_date'), channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_hard':0, 'prevalent_bradyarrhythmia_av_block_or_distal_conduction_disease_hard':1, 'incident_bradyarrhythmia_av_block_or_distal_conduction_disease_hard':2}, loss=weighted_crossentropy([1.0, 171, 66], 'bradyarrhythmia_av_block_or_distal_conduction_disease_hard_prevalent_incident'))
bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2_prevalent_incident = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2_date'), channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2':0, 'prevalent_bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2':1, 'incident_bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2':2}, loss=weighted_crossentropy([1.0, 180, 67], 'bradyarrhythmia_av_block_or_distal_conduction_disease_hard_v2_prevalent_incident'))
bradyarrhythmia_av_block_or_distal_conduction_disease_v2_prevalent_incident = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_av_block_or_distal_conduction_disease_v2_date'), channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_v2':0, 'prevalent_bradyarrhythmia_av_block_or_distal_conduction_disease_v2':1, 'incident_bradyarrhythmia_av_block_or_distal_conduction_disease_v2':2}, loss=weighted_crossentropy([1.0, 135, 59], 'bradyarrhythmia_av_block_or_distal_conduction_disease_v2_prevalent_incident'))
bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted_prevalent_incident = TensorMap('bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted_date'), channel_map={'no_bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted':0, 'prevalent_bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted':1, 'incident_bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted':2}, loss=weighted_crossentropy([1.0, 488, 280], 'bradyarrhythmia_av_block_or_distal_conduction_disease_v2_restricted_prevalent_incident'))
bradyarrhythmia_general_inclusive_definition_prevalent_incident = TensorMap('bradyarrhythmia_general_inclusive_definition', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_general_inclusive_definition_date'), channel_map={'no_bradyarrhythmia_general_inclusive_definition':0, 'prevalent_bradyarrhythmia_general_inclusive_definition':1, 'incident_bradyarrhythmia_general_inclusive_definition':2}, loss=weighted_crossentropy([1.0, 52, 27], 'bradyarrhythmia_general_inclusive_definition_prevalent_incident'))
bradyarrhythmia_only_prevalent_incident = TensorMap('bradyarrhythmia_only', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_only_date'), channel_map={'no_bradyarrhythmia_only':0, 'prevalent_bradyarrhythmia_only':1, 'incident_bradyarrhythmia_only':2}, loss=weighted_crossentropy([1.0, 83, 54], 'bradyarrhythmia_only_prevalent_incident'))
bradyarrhythmia_pacemaker_prevalent_incident = TensorMap('bradyarrhythmia_pacemaker', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_pacemaker_date'), channel_map={'no_bradyarrhythmia_pacemaker':0, 'prevalent_bradyarrhythmia_pacemaker':1, 'incident_bradyarrhythmia_pacemaker':2}, loss=weighted_crossentropy([1.0, 91, 58], 'bradyarrhythmia_pacemaker_prevalent_incident'))
bradyarrhythmia_pacemaker_v2_prevalent_incident = TensorMap('bradyarrhythmia_pacemaker_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_pacemaker_v2_date'), channel_map={'no_bradyarrhythmia_pacemaker_v2':0, 'prevalent_bradyarrhythmia_pacemaker_v2':1, 'incident_bradyarrhythmia_pacemaker_v2':2}, loss=weighted_crossentropy([1.0, 90, 56], 'bradyarrhythmia_pacemaker_v2_prevalent_incident'))
bradyarrhythmia_sinus_node_dysfunction_prevalent_incident = TensorMap('bradyarrhythmia_sinus_node_dysfunction', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_sinus_node_dysfunction_date'), channel_map={'no_bradyarrhythmia_sinus_node_dysfunction':0, 'prevalent_bradyarrhythmia_sinus_node_dysfunction':1, 'incident_bradyarrhythmia_sinus_node_dysfunction':2}, loss=weighted_crossentropy([1.0, 937, 535], 'bradyarrhythmia_sinus_node_dysfunction_prevalent_incident'))
bradyarrhythmia_sinus_node_dysfunction_hard_prevalent_incident = TensorMap('bradyarrhythmia_sinus_node_dysfunction_hard', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_sinus_node_dysfunction_hard_date'), channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_hard':0, 'prevalent_bradyarrhythmia_sinus_node_dysfunction_hard':1, 'incident_bradyarrhythmia_sinus_node_dysfunction_hard':2}, loss=weighted_crossentropy([1.0, 1304, 802], 'bradyarrhythmia_sinus_node_dysfunction_hard_prevalent_incident'))
bradyarrhythmia_sinus_node_dysfunction_hard_v2_prevalent_incident = TensorMap('bradyarrhythmia_sinus_node_dysfunction_hard_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_sinus_node_dysfunction_hard_v2_date'), channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_hard_v2':0, 'prevalent_bradyarrhythmia_sinus_node_dysfunction_hard_v2':1, 'incident_bradyarrhythmia_sinus_node_dysfunction_hard_v2':2}, loss=weighted_crossentropy([1.0, 1339, 842], 'bradyarrhythmia_sinus_node_dysfunction_hard_v2_prevalent_incident'))
bradyarrhythmia_sinus_node_dysfunction_v2_prevalent_incident = TensorMap('bradyarrhythmia_sinus_node_dysfunction_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_sinus_node_dysfunction_v2_date'), channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_v2':0, 'prevalent_bradyarrhythmia_sinus_node_dysfunction_v2':1, 'incident_bradyarrhythmia_sinus_node_dysfunction_v2':2}, loss=weighted_crossentropy([1.0, 1079, 731], 'bradyarrhythmia_sinus_node_dysfunction_v2_prevalent_incident'))
bradyarrhythmia_sinus_node_dysfunction_v2_restricted_prevalent_incident = TensorMap('bradyarrhythmia_sinus_node_dysfunction_v2_restricted', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrhythmia_sinus_node_dysfunction_v2_restricted_date'), channel_map={'no_bradyarrhythmia_sinus_node_dysfunction_v2_restricted':0, 'prevalent_bradyarrhythmia_sinus_node_dysfunction_v2_restricted':1, 'incident_bradyarrhythmia_sinus_node_dysfunction_v2_restricted':2}, loss=weighted_crossentropy([1.0, 1304, 973], 'bradyarrhythmia_sinus_node_dysfunction_v2_restricted_prevalent_incident'))
bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical_prevalent_incident = TensorMap('bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical_date'), channel_map={'no_bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical':0, 'prevalent_bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical':1, 'incident_bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical':2}, loss=weighted_crossentropy([1.0, 84, 54], 'bradyarrrhythmia_av_block_or_distal_conduction_disease_clinical_prevalent_incident'))
bradyarrrhythmia_general_inclusive_definition_clinical_prevalent_incident = TensorMap('bradyarrrhythmia_general_inclusive_definition_clinical', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/bradyarrrhythmia_general_inclusive_definition_clinical_date'), channel_map={'no_bradyarrrhythmia_general_inclusive_definition_clinical':0, 'prevalent_bradyarrrhythmia_general_inclusive_definition_clinical':1, 'incident_bradyarrrhythmia_general_inclusive_definition_clinical':2}, loss=weighted_crossentropy([1.0, 82, 52], 'bradyarrrhythmia_general_inclusive_definition_clinical_prevalent_incident'))
breast_cancer_prevalent_incident = TensorMap('breast_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/breast_cancer_date'), channel_map={'no_breast_cancer':0, 'prevalent_breast_cancer':1, 'incident_breast_cancer':2}, loss=weighted_crossentropy([1.0, 14, 32], 'breast_cancer_prevalent_incident'))
cardiac_surgery_prevalent_incident = TensorMap('cardiac_surgery', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/cardiac_surgery_date'), channel_map={'no_cardiac_surgery':0, 'prevalent_cardiac_surgery':1, 'incident_cardiac_surgery':2}, loss=weighted_crossentropy([1.0, 39, 49], 'cardiac_surgery_prevalent_incident'))
cataract_prevalent_incident = TensorMap('cataract', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/cataract_date'), channel_map={'no_cataract':0, 'prevalent_cataract':1, 'incident_cataract':2}, loss=weighted_crossentropy([1.0, 11, 10], 'cataract_prevalent_incident'))
cervical_cancer_prevalent_incident = TensorMap('cervical_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/cervical_cancer_date'), channel_map={'no_cervical_cancer':0, 'prevalent_cervical_cancer':1, 'incident_cervical_cancer':2}, loss=weighted_crossentropy([1.0, 78, 1414], 'cervical_cancer_prevalent_incident'))
cholelithiasis_or_gallstones_prevalent_incident = TensorMap('cholelithiasis_or_gallstones', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/cholelithiasis_or_gallstones_date'), channel_map={'no_cholelithiasis_or_gallstones':0, 'prevalent_cholelithiasis_or_gallstones':1, 'incident_cholelithiasis_or_gallstones':2}, loss=weighted_crossentropy([1.0, 11, 22], 'cholelithiasis_or_gallstones_prevalent_incident'))
chronic_kidney_disease_prevalent_incident = TensorMap('chronic_kidney_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/chronic_kidney_disease_date'), channel_map={'no_chronic_kidney_disease':0, 'prevalent_chronic_kidney_disease':1, 'incident_chronic_kidney_disease':2}, loss=weighted_crossentropy([1.0, 104, 31], 'chronic_kidney_disease_prevalent_incident'))
chronic_obstructive_pulmonary_disease_prevalent_incident = TensorMap('chronic_obstructive_pulmonary_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/chronic_obstructive_pulmonary_disease_date'), channel_map={'no_chronic_obstructive_pulmonary_disease':0, 'prevalent_chronic_obstructive_pulmonary_disease':1, 'incident_chronic_obstructive_pulmonary_disease':2}, loss=weighted_crossentropy([1.0, 15, 27], 'chronic_obstructive_pulmonary_disease_prevalent_incident'))
colorectal_cancer_prevalent_incident = TensorMap('colorectal_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/colorectal_cancer_date'), channel_map={'no_colorectal_cancer':0, 'prevalent_colorectal_cancer':1, 'incident_colorectal_cancer':2}, loss=weighted_crossentropy([1.0, 59, 71], 'colorectal_cancer_prevalent_incident'))
conduction_system_disease_only_prevalent_incident = TensorMap('conduction_system_disease_only', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/conduction_system_disease_only_date'), channel_map={'no_conduction_system_disease_only':0, 'prevalent_conduction_system_disease_only':1, 'incident_conduction_system_disease_only':2}, loss=weighted_crossentropy([1.0, 118, 46], 'conduction_system_disease_only_prevalent_incident'))
congenital_heart_disease_prevalent_incident = TensorMap('congenital_heart_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/congenital_heart_disease_date'), channel_map={'no_congenital_heart_disease':0, 'prevalent_congenital_heart_disease':1, 'incident_congenital_heart_disease':2}, loss=weighted_crossentropy([1.0, 298, 262], 'congenital_heart_disease_prevalent_incident'))
congenital_heart_disease_ebstein_anomaly_prevalent_incident = TensorMap('congenital_heart_disease_ebstein_anomaly', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/congenital_heart_disease_ebstein_anomaly_date'), channel_map={'no_congenital_heart_disease_ebstein_anomaly':0, 'prevalent_congenital_heart_disease_ebstein_anomaly':1, 'incident_congenital_heart_disease_ebstein_anomaly':2}, loss=weighted_crossentropy([1.0, 21427, 149989], 'congenital_heart_disease_ebstein_anomaly_prevalent_incident'))
coronary_artery_disease_prevalent_incident = TensorMap('coronary_artery_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/coronary_artery_disease_date'), channel_map={'no_coronary_artery_disease':0, 'prevalent_coronary_artery_disease':1, 'incident_coronary_artery_disease':2}, loss=weighted_crossentropy([1.0, 10, 19], 'coronary_artery_disease_prevalent_incident'))
coronary_artery_disease_hard_prevalent_incident = TensorMap('coronary_artery_disease_hard', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/coronary_artery_disease_hard_date'), channel_map={'no_coronary_artery_disease_hard':0, 'prevalent_coronary_artery_disease_hard':1, 'incident_coronary_artery_disease_hard':2}, loss=weighted_crossentropy([1.0, 9, 18], 'coronary_artery_disease_hard_prevalent_incident'))
coronary_artery_disease_intermediate_prevalent_incident = TensorMap('coronary_artery_disease_intermediate', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/coronary_artery_disease_intermediate_date'), channel_map={'no_coronary_artery_disease_intermediate':0, 'prevalent_coronary_artery_disease_intermediate':1, 'incident_coronary_artery_disease_intermediate':2}, loss=weighted_crossentropy([1.0, 8, 11], 'coronary_artery_disease_intermediate_prevalent_incident'))
coronary_artery_disease_soft_prevalent_incident = TensorMap('coronary_artery_disease_soft', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/coronary_artery_disease_soft_date'), channel_map={'no_coronary_artery_disease_soft':0, 'prevalent_coronary_artery_disease_soft':1, 'incident_coronary_artery_disease_soft':2}, loss=weighted_crossentropy([1.0, 5, 10], 'coronary_artery_disease_soft_prevalent_incident'))
crohns_disease_prevalent_incident = TensorMap('crohns_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/crohns_disease_date'), channel_map={'no_crohns_disease':0, 'prevalent_crohns_disease':1, 'incident_crohns_disease':2}, loss=weighted_crossentropy([1.0, 114, 245], 'crohns_disease_prevalent_incident'))
death_prevalent_incident = TensorMap('death', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/death_date'), channel_map={'no_death':0, 'prevalent_death':1, 'incident_death':2}, loss=weighted_crossentropy([1.0, 37497, 11], 'death_prevalent_incident'))
depression_prevalent_incident = TensorMap('depression', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/depression_date'), channel_map={'no_depression':0, 'prevalent_depression':1, 'incident_depression':2}, loss=weighted_crossentropy([1.0, 5, 22], 'depression_prevalent_incident'))
dermatitis_prevalent_incident = TensorMap('dermatitis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/dermatitis_date'), channel_map={'no_dermatitis':0, 'prevalent_dermatitis':1, 'incident_dermatitis':2}, loss=weighted_crossentropy([1.0, 12, 56], 'dermatitis_prevalent_incident'))
diabetes_all_prevalent_incident = TensorMap('diabetes_all', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/diabetes_all_date'), channel_map={'no_diabetes_all':0, 'prevalent_diabetes_all':1, 'incident_diabetes_all':2}, loss=weighted_crossentropy([1.0, 6, 23], 'diabetes_all_prevalent_incident'))
diabetes_type_1_prevalent_incident = TensorMap('diabetes_type_1', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/diabetes_type_1_date'), channel_map={'no_diabetes_type_1':0, 'prevalent_diabetes_type_1':1, 'incident_diabetes_type_1':2}, loss=weighted_crossentropy([1.0, 62, 121], 'diabetes_type_1_prevalent_incident'))
diabetes_type_2_prevalent_incident = TensorMap('diabetes_type_2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/diabetes_type_2_date'), channel_map={'no_diabetes_type_2':0, 'prevalent_diabetes_type_2':1, 'incident_diabetes_type_2':2}, loss=weighted_crossentropy([1.0, 14, 12], 'diabetes_type_2_prevalent_incident'))
diverticular_disease_prevalent_incident = TensorMap('diverticular_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/diverticular_disease_date'), channel_map={'no_diverticular_disease':0, 'prevalent_diverticular_disease':1, 'incident_diverticular_disease':2}, loss=weighted_crossentropy([1.0, 11, 8], 'diverticular_disease_prevalent_incident'))
enlarged_prostate_prevalent_incident = TensorMap('enlarged_prostate', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/enlarged_prostate_date'), channel_map={'no_enlarged_prostate':0, 'prevalent_enlarged_prostate':1, 'incident_enlarged_prostate':2}, loss=weighted_crossentropy([1.0, 16, 22], 'enlarged_prostate_prevalent_incident'))
epilepsy_prevalent_incident = TensorMap('epilepsy', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/epilepsy_date'), channel_map={'no_epilepsy':0, 'prevalent_epilepsy':1, 'incident_epilepsy':2}, loss=weighted_crossentropy([1.0, 34, 132], 'epilepsy_prevalent_incident'))
gastroesophageal_reflux_disease_prevalent_incident = TensorMap('gastroesophageal_reflux_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/gastroesophageal_reflux_disease_date'), channel_map={'no_gastroesophageal_reflux_disease':0, 'prevalent_gastroesophageal_reflux_disease':1, 'incident_gastroesophageal_reflux_disease':2}, loss=weighted_crossentropy([1.0, 5, 11], 'gastroesophageal_reflux_disease_prevalent_incident'))
glaucoma_prevalent_incident = TensorMap('glaucoma', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/glaucoma_date'), channel_map={'no_glaucoma':0, 'prevalent_glaucoma':1, 'incident_glaucoma':2}, loss=weighted_crossentropy([1.0, 28, 59], 'glaucoma_prevalent_incident'))
gout_prevalent_incident = TensorMap('gout', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/gout_date'), channel_map={'no_gout':0, 'prevalent_gout':1, 'incident_gout':2}, loss=weighted_crossentropy([1.0, 22, 97], 'gout_prevalent_incident'))
heart_failure_prevalent_incident = TensorMap('heart_failure', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/heart_failure_date'), channel_map={'no_heart_failure':0, 'prevalent_heart_failure':1, 'incident_heart_failure':2}, loss=weighted_crossentropy([1.0, 58, 33], 'heart_failure_prevalent_incident'))
heart_failure_v2_prevalent_incident = TensorMap('heart_failure_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/heart_failure_v2_date'), channel_map={'no_heart_failure_v2':0, 'prevalent_heart_failure_v2':1, 'incident_heart_failure_v2':2}, loss=weighted_crossentropy([1.0, 58, 33], 'heart_failure_v2_prevalent_incident'))
hypercholesterolemia_prevalent_incident = TensorMap('hypercholesterolemia', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/hypercholesterolemia_date'), channel_map={'no_hypercholesterolemia':0, 'prevalent_hypercholesterolemia':1, 'incident_hypercholesterolemia':2}, loss=weighted_crossentropy([1.0, 2, 9], 'hypercholesterolemia_prevalent_incident'))
hypertension_prevalent_incident = TensorMap('hypertension', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/hypertension_date'), channel_map={'no_hypertension':0, 'prevalent_hypertension':1, 'incident_hypertension':2}, loss=weighted_crossentropy([1.0, 1, 7], 'hypertension_prevalent_incident'))
hyperthyroidism_prevalent_incident = TensorMap('hyperthyroidism', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/hyperthyroidism_date'), channel_map={'no_hyperthyroidism':0, 'prevalent_hyperthyroidism':1, 'incident_hyperthyroidism':2}, loss=weighted_crossentropy([1.0, 38, 149], 'hyperthyroidism_prevalent_incident'))
hypertrophic_cardiomyopathy_prevalent_incident = TensorMap('hypertrophic_cardiomyopathy', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/hypertrophic_cardiomyopathy_date'), channel_map={'no_hypertrophic_cardiomyopathy':0, 'prevalent_hypertrophic_cardiomyopathy':1, 'incident_hypertrophic_cardiomyopathy':2}, loss=weighted_crossentropy([1.0, 789, 1127], 'hypertrophic_cardiomyopathy_prevalent_incident'))
hypothyroidism_prevalent_incident = TensorMap('hypothyroidism', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/hypothyroidism_date'), channel_map={'no_hypothyroidism':0, 'prevalent_hypothyroidism':1, 'incident_hypothyroidism':2}, loss=weighted_crossentropy([1.0, 6, 39], 'hypothyroidism_prevalent_incident'))
implantable_cardioverter_defibrillator_prevalent_incident = TensorMap('implantable_cardioverter_defibrillator', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/implantable_cardioverter_defibrillator_date'), channel_map={'no_implantable_cardioverter_defibrillator':0, 'prevalent_implantable_cardioverter_defibrillator':1, 'incident_implantable_cardioverter_defibrillator':2}, loss=weighted_crossentropy([1.0, 1376, 261], 'implantable_cardioverter_defibrillator_prevalent_incident'))
inflammatory_bowel_disease_prevalent_incident = TensorMap('inflammatory_bowel_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/inflammatory_bowel_disease_date'), channel_map={'no_inflammatory_bowel_disease':0, 'prevalent_inflammatory_bowel_disease':1, 'incident_inflammatory_bowel_disease':2}, loss=weighted_crossentropy([1.0, 31, 130], 'inflammatory_bowel_disease_prevalent_incident'))
intervertebral_disc_displacement_prevalent_incident = TensorMap('intervertebral_disc_displacement', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/intervertebral_disc_displacement_date'), channel_map={'no_intervertebral_disc_displacement':0, 'prevalent_intervertebral_disc_displacement':1, 'incident_intervertebral_disc_displacement':2}, loss=weighted_crossentropy([1.0, 17, 81], 'intervertebral_disc_displacement_prevalent_incident'))
intracerebral_hemorrhage_prevalent_incident = TensorMap('intracerebral_hemorrhage', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/intracerebral_hemorrhage_date'), channel_map={'no_intracerebral_hemorrhage':0, 'prevalent_intracerebral_hemorrhage':1, 'incident_intracerebral_hemorrhage':2}, loss=weighted_crossentropy([1.0, 351, 265], 'intracerebral_hemorrhage_prevalent_incident'))
iron_deficiency_anemia_prevalent_incident = TensorMap('iron_deficiency_anemia', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/iron_deficiency_anemia_date'), channel_map={'no_iron_deficiency_anemia':0, 'prevalent_iron_deficiency_anemia':1, 'incident_iron_deficiency_anemia':2}, loss=weighted_crossentropy([1.0, 26, 25], 'iron_deficiency_anemia_prevalent_incident'))
irritable_bowel_syndrome_prevalent_incident = TensorMap('irritable_bowel_syndrome', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/irritable_bowel_syndrome_date'), channel_map={'no_irritable_bowel_syndrome':0, 'prevalent_irritable_bowel_syndrome':1, 'incident_irritable_bowel_syndrome':2}, loss=weighted_crossentropy([1.0, 12, 55], 'irritable_bowel_syndrome_prevalent_incident'))
ischemic_stroke_prevalent_incident = TensorMap('ischemic_stroke', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ischemic_stroke_date'), channel_map={'no_ischemic_stroke':0, 'prevalent_ischemic_stroke':1, 'incident_ischemic_stroke':2}, loss=weighted_crossentropy([1.0, 87, 54], 'ischemic_stroke_prevalent_incident'))
lung_cancer_prevalent_incident = TensorMap('lung_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/lung_cancer_date'), channel_map={'no_lung_cancer':0, 'prevalent_lung_cancer':1, 'incident_lung_cancer':2}, loss=weighted_crossentropy([1.0, 313, 68], 'lung_cancer_prevalent_incident'))
lupus_erythematosus_prevalent_incident = TensorMap('lupus_erythematosus', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/lupus_erythematosus_date'), channel_map={'no_lupus_erythematosus':0, 'prevalent_lupus_erythematosus':1, 'incident_lupus_erythematosus':2}, loss=weighted_crossentropy([1.0, 235, 955], 'lupus_erythematosus_prevalent_incident'))
migraine_prevalent_incident = TensorMap('migraine', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/migraine_date'), channel_map={'no_migraine':0, 'prevalent_migraine':1, 'incident_migraine':2}, loss=weighted_crossentropy([1.0, 10, 77], 'migraine_prevalent_incident'))
mitral_regurgitation_prevalent_incident = TensorMap('mitral_regurgitation', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/mitral_regurgitation_date'), channel_map={'no_mitral_regurgitation':0, 'prevalent_mitral_regurgitation':1, 'incident_mitral_regurgitation':2}, loss=weighted_crossentropy([1.0, 184, 101], 'mitral_regurgitation_prevalent_incident'))
mitral_stenosis_prevalent_incident = TensorMap('mitral_stenosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/mitral_stenosis_date'), channel_map={'no_mitral_stenosis':0, 'prevalent_mitral_stenosis':1, 'incident_mitral_stenosis':2}, loss=weighted_crossentropy([1.0, 1764, 2884], 'mitral_stenosis_prevalent_incident'))
mitral_valve_disease_prevalent_incident = TensorMap('mitral_valve_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/mitral_valve_disease_date'), channel_map={'no_mitral_valve_disease':0, 'prevalent_mitral_valve_disease':1, 'incident_mitral_valve_disease':2}, loss=weighted_crossentropy([1.0, 104, 59], 'mitral_valve_disease_prevalent_incident'))
mitral_valve_prolapse_prevalent_incident = TensorMap('mitral_valve_prolapse', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/mitral_valve_prolapse_date'), channel_map={'no_mitral_valve_prolapse':0, 'prevalent_mitral_valve_prolapse':1, 'incident_mitral_valve_prolapse':2}, loss=weighted_crossentropy([1.0, 553, 513], 'mitral_valve_prolapse_prevalent_incident'))
multiple_sclerosis_prevalent_incident = TensorMap('multiple_sclerosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/multiple_sclerosis_date'), channel_map={'no_multiple_sclerosis':0, 'prevalent_multiple_sclerosis':1, 'incident_multiple_sclerosis':2}, loss=weighted_crossentropy([1.0, 82, 753], 'multiple_sclerosis_prevalent_incident'))
myocardial_infarction_prevalent_incident = TensorMap('myocardial_infarction', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/myocardial_infarction_date'), channel_map={'no_myocardial_infarction':0, 'prevalent_myocardial_infarction':1, 'incident_myocardial_infarction':2}, loss=weighted_crossentropy([1.0, 13, 27], 'myocardial_infarction_prevalent_incident'))
osteoarthritis_prevalent_incident = TensorMap('osteoarthritis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/osteoarthritis_date'), channel_map={'no_osteoarthritis':0, 'prevalent_osteoarthritis':1, 'incident_osteoarthritis':2}, loss=weighted_crossentropy([1.0, 3, 6], 'osteoarthritis_prevalent_incident'))
osteoporosis_prevalent_incident = TensorMap('osteoporosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/osteoporosis_date'), channel_map={'no_osteoporosis':0, 'prevalent_osteoporosis':1, 'incident_osteoporosis':2}, loss=weighted_crossentropy([1.0, 17, 41], 'osteoporosis_prevalent_incident'))
pancreatitis_prevalent_incident = TensorMap('pancreatitis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/pancreatitis_date'), channel_map={'no_pancreatitis':0, 'prevalent_pancreatitis':1, 'incident_pancreatitis':2}, loss=weighted_crossentropy([1.0, 86, 131], 'pancreatitis_prevalent_incident'))
parkinsons_disease_prevalent_incident = TensorMap('parkinsons_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/parkinsons_disease_date'), channel_map={'no_parkinsons_disease':0, 'prevalent_parkinsons_disease':1, 'incident_parkinsons_disease':2}, loss=weighted_crossentropy([1.0, 180, 235], 'parkinsons_disease_prevalent_incident'))
peripheral_vascular_disease_prevalent_incident = TensorMap('peripheral_vascular_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/peripheral_vascular_disease_date'), channel_map={'no_peripheral_vascular_disease':0, 'prevalent_peripheral_vascular_disease':1, 'incident_peripheral_vascular_disease':2}, loss=weighted_crossentropy([1.0, 47, 71], 'peripheral_vascular_disease_prevalent_incident'))
pneumonia_prevalent_incident = TensorMap('pneumonia', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/pneumonia_date'), channel_map={'no_pneumonia':0, 'prevalent_pneumonia':1, 'incident_pneumonia':2}, loss=weighted_crossentropy([1.0, 17, 18], 'pneumonia_prevalent_incident'))
posttraumatic_stress_disorder_prevalent_incident = TensorMap('posttraumatic_stress_disorder', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/posttraumatic_stress_disorder_date'), channel_map={'no_posttraumatic_stress_disorder':0, 'prevalent_posttraumatic_stress_disorder':1, 'incident_posttraumatic_stress_disorder':2}, loss=weighted_crossentropy([1.0, 461, 1704], 'posttraumatic_stress_disorder_prevalent_incident'))
prostate_cancer_prevalent_incident = TensorMap('prostate_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/prostate_cancer_date'), channel_map={'no_prostate_cancer':0, 'prevalent_prostate_cancer':1, 'incident_prostate_cancer':2}, loss=weighted_crossentropy([1.0, 45, 41], 'prostate_cancer_prevalent_incident'))
psoriasis_prevalent_incident = TensorMap('psoriasis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/psoriasis_date'), channel_map={'no_psoriasis':0, 'prevalent_psoriasis':1, 'incident_psoriasis':2}, loss=weighted_crossentropy([1.0, 26, 144], 'psoriasis_prevalent_incident'))
pulmonary_hypertension_prevalent_incident = TensorMap('pulmonary_hypertension', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/pulmonary_hypertension_date'), channel_map={'no_pulmonary_hypertension':0, 'prevalent_pulmonary_hypertension':1, 'incident_pulmonary_hypertension':2}, loss=weighted_crossentropy([1.0, 993, 241], 'pulmonary_hypertension_prevalent_incident'))
pulmonary_regurgitation_prevalent_incident = TensorMap('pulmonary_regurgitation', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/pulmonary_regurgitation_date'), channel_map={'no_pulmonary_regurgitation':0, 'prevalent_pulmonary_regurgitation':1, 'incident_pulmonary_regurgitation':2}, loss=weighted_crossentropy([1.0, 7894, 1973], 'pulmonary_regurgitation_prevalent_incident'))
pulmonary_stenosis_prevalent_incident = TensorMap('pulmonary_stenosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/pulmonary_stenosis_date'), channel_map={'no_pulmonary_stenosis':0, 'prevalent_pulmonary_stenosis':1, 'incident_pulmonary_stenosis':2}, loss=weighted_crossentropy([1.0, 9374, 14998], 'pulmonary_stenosis_prevalent_incident'))
pulmonary_valve_disease_prevalent_incident = TensorMap('pulmonary_valve_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/pulmonary_valve_disease_date'), channel_map={'no_pulmonary_valve_disease':0, 'prevalent_pulmonary_valve_disease':1, 'incident_pulmonary_valve_disease':2}, loss=weighted_crossentropy([1.0, 3488, 1648], 'pulmonary_valve_disease_prevalent_incident'))
rheumatoid_arthritis_prevalent_incident = TensorMap('rheumatoid_arthritis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/rheumatoid_arthritis_date'), channel_map={'no_rheumatoid_arthritis':0, 'prevalent_rheumatoid_arthritis':1, 'incident_rheumatoid_arthritis':2}, loss=weighted_crossentropy([1.0, 25, 86], 'rheumatoid_arthritis_prevalent_incident'))
sarcoidosis_prevalent_incident = TensorMap('sarcoidosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/sarcoidosis_date'), channel_map={'no_sarcoidosis':0, 'prevalent_sarcoidosis':1, 'incident_sarcoidosis':2}, loss=weighted_crossentropy([1.0, 149, 789], 'sarcoidosis_prevalent_incident'))
schizophrenia_prevalent_incident = TensorMap('schizophrenia', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/schizophrenia_date'), channel_map={'no_schizophrenia':0, 'prevalent_schizophrenia':1, 'incident_schizophrenia':2}, loss=weighted_crossentropy([1.0, 189, 887], 'schizophrenia_prevalent_incident'))
sciatica_prevalent_incident = TensorMap('sciatica', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/sciatica_date'), channel_map={'no_sciatica':0, 'prevalent_sciatica':1, 'incident_sciatica':2}, loss=weighted_crossentropy([1.0, 27, 100], 'sciatica_prevalent_incident'))
skin_cancer_prevalent_incident = TensorMap('skin_cancer', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/skin_cancer_date'), channel_map={'no_skin_cancer':0, 'prevalent_skin_cancer':1, 'incident_skin_cancer':2}, loss=weighted_crossentropy([1.0, 13, 20], 'skin_cancer_prevalent_incident'))
sleep_apnea_prevalent_incident = TensorMap('sleep_apnea', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/sleep_apnea_date'), channel_map={'no_sleep_apnea':0, 'prevalent_sleep_apnea':1, 'incident_sleep_apnea':2}, loss=weighted_crossentropy([1.0, 49, 56], 'sleep_apnea_prevalent_incident'))
stroke_prevalent_incident = TensorMap('stroke', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/stroke_date'), channel_map={'no_stroke':0, 'prevalent_stroke':1, 'incident_stroke':2}, loss=weighted_crossentropy([1.0, 21, 42], 'stroke_prevalent_incident'))
subarachnoid_hemorrhage_prevalent_incident = TensorMap('subarachnoid_hemorrhage', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/subarachnoid_hemorrhage_date'), channel_map={'no_subarachnoid_hemorrhage':0, 'prevalent_subarachnoid_hemorrhage':1, 'incident_subarachnoid_hemorrhage':2}, loss=weighted_crossentropy([1.0, 175, 374], 'subarachnoid_hemorrhage_prevalent_incident'))
supraventricular_arrhythmia_general_inclusive_definition_prevalent_incident = TensorMap('supraventricular_arrhythmia_general_inclusive_definition', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/supraventricular_arrhythmia_general_inclusive_definition_date'), channel_map={'no_supraventricular_arrhythmia_general_inclusive_definition':0, 'prevalent_supraventricular_arrhythmia_general_inclusive_definition':1, 'incident_supraventricular_arrhythmia_general_inclusive_definition':2}, loss=weighted_crossentropy([1.0, 84, 104], 'supraventricular_arrhythmia_general_inclusive_definition_prevalent_incident'))
supraventricular_arrhythmia_general_inclusive_definition_clinical_prevalent_incident = TensorMap('supraventricular_arrhythmia_general_inclusive_definition_clinical', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/supraventricular_arrhythmia_general_inclusive_definition_clinical_date'), channel_map={'no_supraventricular_arrhythmia_general_inclusive_definition_clinical':0, 'prevalent_supraventricular_arrhythmia_general_inclusive_definition_clinical':1, 'incident_supraventricular_arrhythmia_general_inclusive_definition_clinical':2}, loss=weighted_crossentropy([1.0, 78, 107], 'supraventricular_arrhythmia_general_inclusive_definition_clinical_prevalent_incident'))
supraventricular_arrhythmia_premature_atrial_contractions_prevalent_incident = TensorMap('supraventricular_arrhythmia_premature_atrial_contractions', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/supraventricular_arrhythmia_premature_atrial_contractions_date'), channel_map={'no_supraventricular_arrhythmia_premature_atrial_contractions':0, 'prevalent_supraventricular_arrhythmia_premature_atrial_contractions':1, 'incident_supraventricular_arrhythmia_premature_atrial_contractions':2}, loss=weighted_crossentropy([1.0, 2727, 1704], 'supraventricular_arrhythmia_premature_atrial_contractions_prevalent_incident'))
supraventricular_arrhythmia_svt_prevalent_incident = TensorMap('supraventricular_arrhythmia_svt', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/supraventricular_arrhythmia_svt_date'), channel_map={'no_supraventricular_arrhythmia_svt':0, 'prevalent_supraventricular_arrhythmia_svt':1, 'incident_supraventricular_arrhythmia_svt':2}, loss=weighted_crossentropy([1.0, 83, 113], 'supraventricular_arrhythmia_svt_prevalent_incident'))
supraventricular_arrhythmia_wpw_prevalent_incident = TensorMap('supraventricular_arrhythmia_wpw', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/supraventricular_arrhythmia_wpw_date'), channel_map={'no_supraventricular_arrhythmia_wpw':0, 'prevalent_supraventricular_arrhythmia_wpw':1, 'incident_supraventricular_arrhythmia_wpw':2}, loss=weighted_crossentropy([1.0, 710, 920], 'supraventricular_arrhythmia_wpw_prevalent_incident'))
supraventricular_arrhythmia_wpw_v2_prevalent_incident = TensorMap('supraventricular_arrhythmia_wpw_v2', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/supraventricular_arrhythmia_wpw_v2_date'), channel_map={'no_supraventricular_arrhythmia_wpw_v2':0, 'prevalent_supraventricular_arrhythmia_wpw_v2':1, 'incident_supraventricular_arrhythmia_wpw_v2':2}, loss=weighted_crossentropy([1.0, 721, 920], 'supraventricular_arrhythmia_wpw_v2_prevalent_incident'))
tricuspid_regurgitation_prevalent_incident = TensorMap('tricuspid_regurgitation', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/tricuspid_regurgitation_date'), channel_map={'no_tricuspid_regurgitation':0, 'prevalent_tricuspid_regurgitation':1, 'incident_tricuspid_regurgitation':2}, loss=weighted_crossentropy([1.0, 1829, 422], 'tricuspid_regurgitation_prevalent_incident'))
tricuspid_stenosis_prevalent_incident = TensorMap('tricuspid_stenosis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/tricuspid_stenosis_date'), channel_map={'no_tricuspid_stenosis':0, 'prevalent_tricuspid_stenosis':1, 'incident_tricuspid_stenosis':2}, loss=weighted_crossentropy([1.0, 37497, 37497], 'tricuspid_stenosis_prevalent_incident'))
tricuspid_valve_disease_prevalent_incident = TensorMap('tricuspid_valve_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/tricuspid_valve_disease_date'), channel_map={'no_tricuspid_valve_disease':0, 'prevalent_tricuspid_valve_disease':1, 'incident_tricuspid_valve_disease':2}, loss=weighted_crossentropy([1.0, 529, 119], 'tricuspid_valve_disease_prevalent_incident'))
ulcerative_colitis_prevalent_incident = TensorMap('ulcerative_colitis', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ulcerative_colitis_date'), channel_map={'no_ulcerative_colitis':0, 'prevalent_ulcerative_colitis':1, 'incident_ulcerative_colitis':2}, loss=weighted_crossentropy([1.0, 47, 165], 'ulcerative_colitis_prevalent_incident'))
uterine_fibroids_prevalent_incident = TensorMap('uterine_fibroids', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/uterine_fibroids_date'), channel_map={'no_uterine_fibroids':0, 'prevalent_uterine_fibroids':1, 'incident_uterine_fibroids':2}, loss=weighted_crossentropy([1.0, 9, 47], 'uterine_fibroids_prevalent_incident'))
valve_disease_prevalent_incident = TensorMap('valve_disease', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/valve_disease_date'), channel_map={'no_valve_disease':0, 'prevalent_valve_disease':1, 'incident_valve_disease':2}, loss=weighted_crossentropy([1.0, 53, 34], 'valve_disease_prevalent_incident'))
valvular_disease_unspecified_prevalent_incident = TensorMap('valvular_disease_unspecified', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/valvular_disease_unspecified_date'), channel_map={'no_valvular_disease_unspecified':0, 'prevalent_valvular_disease_unspecified':1, 'incident_valvular_disease_unspecified':2}, loss=weighted_crossentropy([1.0, 46, 789], 'valvular_disease_unspecified_prevalent_incident'))
venous_thromboembolism_prevalent_incident = TensorMap('venous_thromboembolism', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/venous_thromboembolism_date'), channel_map={'no_venous_thromboembolism':0, 'prevalent_venous_thromboembolism':1, 'incident_venous_thromboembolism':2}, loss=weighted_crossentropy([1.0, 12, 42], 'venous_thromboembolism_prevalent_incident'))
ventricular_arrhythmia_general_inclusive_definition_prevalent_incident = TensorMap('ventricular_arrhythmia_general_inclusive_definition', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ventricular_arrhythmia_general_inclusive_definition_date'), channel_map={'no_ventricular_arrhythmia_general_inclusive_definition':0, 'prevalent_ventricular_arrhythmia_general_inclusive_definition':1, 'incident_ventricular_arrhythmia_general_inclusive_definition':2}, loss=weighted_crossentropy([1.0, 159, 89], 'ventricular_arrhythmia_general_inclusive_definition_prevalent_incident'))
ventricular_arrhythmia_general_inclusive_definition_clinical_prevalent_incident = TensorMap('ventricular_arrhythmia_general_inclusive_definition_clinical', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ventricular_arrhythmia_general_inclusive_definition_clinical_date'), channel_map={'no_ventricular_arrhythmia_general_inclusive_definition_clinical':0, 'prevalent_ventricular_arrhythmia_general_inclusive_definition_clinical':1, 'incident_ventricular_arrhythmia_general_inclusive_definition_clinical':2}, loss=weighted_crossentropy([1.0, 180, 104], 'ventricular_arrhythmia_general_inclusive_definition_clinical_prevalent_incident'))
ventricular_arrhythmia_sudden_cardiac_death_prevalent_incident = TensorMap('ventricular_arrhythmia_sudden_cardiac_death', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ventricular_arrhythmia_sudden_cardiac_death_date'), channel_map={'no_ventricular_arrhythmia_sudden_cardiac_death':0, 'prevalent_ventricular_arrhythmia_sudden_cardiac_death':1, 'incident_ventricular_arrhythmia_sudden_cardiac_death':2}, loss=weighted_crossentropy([1.0, 402, 163], 'ventricular_arrhythmia_sudden_cardiac_death_prevalent_incident'))
ventricular_arrhythmia_ventricular_premature_depolarizations_prevalent_incident = TensorMap('ventricular_arrhythmia_ventricular_premature_depolarizations', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ventricular_arrhythmia_ventricular_premature_depolarizations_date'), channel_map={'no_ventricular_arrhythmia_ventricular_premature_depolarizations':0, 'prevalent_ventricular_arrhythmia_ventricular_premature_depolarizations':1, 'incident_ventricular_arrhythmia_ventricular_premature_depolarizations':2}, loss=weighted_crossentropy([1.0, 1271, 503], 'ventricular_arrhythmia_ventricular_premature_depolarizations_prevalent_incident'))
ventricular_arrhythmia_vt_prevalent_incident = TensorMap('ventricular_arrhythmia_vt', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/ventricular_arrhythmia_vt_date'), channel_map={'no_ventricular_arrhythmia_vt':0, 'prevalent_ventricular_arrhythmia_vt':1, 'incident_ventricular_arrhythmia_vt':2}, loss=weighted_crossentropy([1.0, 295, 208], 'ventricular_arrhythmia_vt_prevalent_incident'))


#  TensorMaps for Phecode disease phenotypes
abdominal_aortic_aneurysm_phe = TensorMap('phecode_442.11', Interpretation.CATEGORICAL, channel_map={'no_abdominal_aortic_aneurysm':0, 'abdominal_aortic_aneurysm':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 348], 'phecode_442_11'))
abdominal_hernia_phe = TensorMap('phecode_550', Interpretation.CATEGORICAL, channel_map={'no_abdominal_hernia':0, 'abdominal_hernia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7], 'phecode_550'))
abdominal_pain_phe = TensorMap('phecode_785', Interpretation.CATEGORICAL, channel_map={'no_abdominal_pain':0, 'abdominal_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8], 'phecode_785'))
abnormal_findings_examination_of_lungs_phe = TensorMap('phecode_514', Interpretation.CATEGORICAL, channel_map={'no_abnormal_findings_examination_of_lungs':0, 'abnormal_findings_examination_of_lungs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 123], 'phecode_514'))
abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area_phe = TensorMap('phecode_564.8', Interpretation.CATEGORICAL, channel_map={'no_abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area':0, 'abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 190], 'phecode_564_8'))
abnormal_findings_on_examination_of_urine_phe = TensorMap('phecode_598', Interpretation.CATEGORICAL, channel_map={'no_abnormal_findings_on_examination_of_urine':0, 'abnormal_findings_on_examination_of_urine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 109], 'phecode_598'))
abnormal_findings_on_study_of_brain_and_or_nervous_system_phe = TensorMap('phecode_346', Interpretation.CATEGORICAL, channel_map={'no_abnormal_findings_on_study_of_brain_and_or_nervous_system':0, 'abnormal_findings_on_study_of_brain_and_or_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_346'))
abnormal_function_study_of_cardiovascular_system_phe = TensorMap('phecode_429.2', Interpretation.CATEGORICAL, channel_map={'no_abnormal_function_study_of_cardiovascular_system':0, 'abnormal_function_study_of_cardiovascular_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 575], 'phecode_429_2'))
abnormal_granulation_tissue_phe = TensorMap('phecode_701.5', Interpretation.CATEGORICAL, channel_map={'no_abnormal_granulation_tissue':0, 'abnormal_granulation_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 801], 'phecode_701_5'))
abnormal_heart_sounds_phe = TensorMap('phecode_396', Interpretation.CATEGORICAL, channel_map={'no_abnormal_heart_sounds':0, 'abnormal_heart_sounds':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 296], 'phecode_396'))
abnormal_involuntary_movements_phe = TensorMap('phecode_350.1', Interpretation.CATEGORICAL, channel_map={'no_abnormal_involuntary_movements':0, 'abnormal_involuntary_movements':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 50], 'phecode_350_1'))
abnormal_mammogram_phe = TensorMap('phecode_611.1', Interpretation.CATEGORICAL, channel_map={'no_abnormal_mammogram':0, 'abnormal_mammogram':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6944], 'phecode_611_1'))
abnormal_posture_phe = TensorMap('phecode_781.2', Interpretation.CATEGORICAL, channel_map={'no_abnormal_posture':0, 'abnormal_posture':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_781_2'))
abnormal_reflex_phe = TensorMap('phecode_350.5', Interpretation.CATEGORICAL, channel_map={'no_abnormal_reflex':0, 'abnormal_reflex':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_350_5'))
abnormal_results_of_function_study_of_kidney_phe = TensorMap('phecode_589', Interpretation.CATEGORICAL, channel_map={'no_abnormal_results_of_function_study_of_kidney':0, 'abnormal_results_of_function_study_of_kidney':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2525], 'phecode_589'))
abnormal_results_of_function_study_of_liver_phe = TensorMap('phecode_573.7', Interpretation.CATEGORICAL, channel_map={'no_abnormal_results_of_function_study_of_liver':0, 'abnormal_results_of_function_study_of_liver':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 90], 'phecode_573_7'))
abnormal_results_of_function_study_of_pulmonary_system_phe = TensorMap('phecode_514.1', Interpretation.CATEGORICAL, channel_map={'no_abnormal_results_of_function_study_of_pulmonary_system':0, 'abnormal_results_of_function_study_of_pulmonary_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9433], 'phecode_514_1'))
abnormal_results_of_function_study_of_thyroid_phe = TensorMap('phecode_246.7', Interpretation.CATEGORICAL, channel_map={'no_abnormal_results_of_function_study_of_thyroid':0, 'abnormal_results_of_function_study_of_thyroid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8333], 'phecode_246_7'))
abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc_phe = TensorMap('phecode_794', Interpretation.CATEGORICAL, channel_map={'no_abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc':0, 'abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3067], 'phecode_794'))
abnormal_serum_enzyme_levels_phe = TensorMap('phecode_573.9', Interpretation.CATEGORICAL, channel_map={'no_abnormal_serum_enzyme_levels':0, 'abnormal_serum_enzyme_levels':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1453], 'phecode_573_9'))
abnormal_spermatozoa_phe = TensorMap('phecode_609.2', Interpretation.CATEGORICAL, channel_map={'no_abnormal_spermatozoa':0, 'abnormal_spermatozoa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2512], 'phecode_609_2'))
abnormal_sputum_phe = TensorMap('phecode_516', Interpretation.CATEGORICAL, channel_map={'no_abnormal_sputum':0, 'abnormal_sputum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3703], 'phecode_516'))
abnormal_weight_gain_phe = TensorMap('phecode_278.4', Interpretation.CATEGORICAL, channel_map={'no_abnormal_weight_gain':0, 'abnormal_weight_gain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5376], 'phecode_278_4'))
abnormality_of_gait_phe = TensorMap('phecode_350.2', Interpretation.CATEGORICAL, channel_map={'no_abnormality_of_gait':0, 'abnormality_of_gait':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 155], 'phecode_350_2'))
abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium_phe = TensorMap('phecode_654.1', Interpretation.CATEGORICAL, channel_map={'no_abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium':0, 'abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 306], 'phecode_654_1'))
abnormality_of_red_blood_cells_phe = TensorMap('phecode_289.9', Interpretation.CATEGORICAL, channel_map={'no_abnormality_of_red_blood_cells':0, 'abnormality_of_red_blood_cells':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3846], 'phecode_289_9'))
abnormality_of_secretion_of_glucagon_or_gastrin_phe = TensorMap('phecode_251.8', Interpretation.CATEGORICAL, channel_map={'no_abnormality_of_secretion_of_glucagon_or_gastrin':0, 'abnormality_of_secretion_of_glucagon_or_gastrin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100000], 'phecode_251_8'))
abscess_or_ulceration_of_vulva_phe = TensorMap('phecode_614.54', Interpretation.CATEGORICAL, channel_map={'no_abscess_or_ulceration_of_vulva':0, 'abscess_or_ulceration_of_vulva':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1428], 'phecode_614_54'))
absent_or_infrequent_menstruation_phe = TensorMap('phecode_626.11', Interpretation.CATEGORICAL, channel_map={'no_absent_or_infrequent_menstruation':0, 'absent_or_infrequent_menstruation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3355], 'phecode_626_11'))
acid_base_balance_disorder_phe = TensorMap('phecode_276.4', Interpretation.CATEGORICAL, channel_map={'no_acid_base_balance_disorder':0, 'acid_base_balance_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83333], 'phecode_276_4'))
acidosis_phe = TensorMap('phecode_276.41', Interpretation.CATEGORICAL, channel_map={'no_acidosis':0, 'acidosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 256], 'phecode_276_41'))
acne_phe = TensorMap('phecode_706.1', Interpretation.CATEGORICAL, channel_map={'no_acne':0, 'acne':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4716], 'phecode_706_1'))
acquired_absence_of_breast_phe = TensorMap('phecode_175', Interpretation.CATEGORICAL, channel_map={'no_acquired_absence_of_breast':0, 'acquired_absence_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 31], 'phecode_175'))
acquired_acanthosis_nigricans_phe = TensorMap('phecode_701.6', Interpretation.CATEGORICAL, channel_map={'no_acquired_acanthosis_nigricans':0, 'acquired_acanthosis_nigricans':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3311], 'phecode_701_6'))
acquired_coagulation_factor_deficiency_phe = TensorMap('phecode_286.4', Interpretation.CATEGORICAL, channel_map={'no_acquired_coagulation_factor_deficiency':0, 'acquired_coagulation_factor_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_286_4'))
acquired_deformities_of_finger_phe = TensorMap('phecode_736.2', Interpretation.CATEGORICAL, channel_map={'no_acquired_deformities_of_finger':0, 'acquired_deformities_of_finger':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1567], 'phecode_736_2'))
acquired_foot_deformities_phe = TensorMap('phecode_735', Interpretation.CATEGORICAL, channel_map={'no_acquired_foot_deformities':0, 'acquired_foot_deformities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2525], 'phecode_735'))
acquired_hemolytic_anemias_phe = TensorMap('phecode_283', Interpretation.CATEGORICAL, channel_map={'no_acquired_hemolytic_anemias':0, 'acquired_hemolytic_anemias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11904], 'phecode_283'))
acquired_hypothyroidism_phe = TensorMap('phecode_244.2', Interpretation.CATEGORICAL, channel_map={'no_acquired_hypothyroidism':0, 'acquired_hypothyroidism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2857], 'phecode_244_2'))
acquired_spondylolisthesis_phe = TensorMap('phecode_738.4', Interpretation.CATEGORICAL, channel_map={'no_acquired_spondylolisthesis':0, 'acquired_spondylolisthesis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1256], 'phecode_738_4'))
acquired_toe_deformities_phe = TensorMap('phecode_735.2', Interpretation.CATEGORICAL, channel_map={'no_acquired_toe_deformities':0, 'acquired_toe_deformities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 170], 'phecode_735_2'))
acromegaly_and_gigantism_phe = TensorMap('phecode_253.11', Interpretation.CATEGORICAL, channel_map={'no_acromegaly_and_gigantism':0, 'acromegaly_and_gigantism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3703], 'phecode_253_11'))
actinic_keratosis_phe = TensorMap('phecode_702.1', Interpretation.CATEGORICAL, channel_map={'no_actinic_keratosis':0, 'actinic_keratosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 129], 'phecode_702_1'))
acute_and_chronic_tonsillitis_phe = TensorMap('phecode_474', Interpretation.CATEGORICAL, channel_map={'no_acute_and_chronic_tonsillitis':0, 'acute_and_chronic_tonsillitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 888], 'phecode_474'))
acute_and_subacute_necrosis_of_liver_phe = TensorMap('phecode_573.4', Interpretation.CATEGORICAL, channel_map={'no_acute_and_subacute_necrosis_of_liver':0, 'acute_and_subacute_necrosis_of_liver':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_573_4'))
acute_appendicitis_phe = TensorMap('phecode_540.11', Interpretation.CATEGORICAL, channel_map={'no_acute_appendicitis':0, 'acute_appendicitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 175], 'phecode_540_11'))
acute_bronchitis_and_bronchiolitis_phe = TensorMap('phecode_483', Interpretation.CATEGORICAL, channel_map={'no_acute_bronchitis_and_bronchiolitis':0, 'acute_bronchitis_and_bronchiolitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1474], 'phecode_483'))
acute_cystitis_phe = TensorMap('phecode_592.11', Interpretation.CATEGORICAL, channel_map={'no_acute_cystitis':0, 'acute_cystitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2617], 'phecode_592_11'))
acute_dermatitis_due_to_solar_radiation_phe = TensorMap('phecode_938.1', Interpretation.CATEGORICAL, channel_map={'no_acute_dermatitis_due_to_solar_radiation':0, 'acute_dermatitis_due_to_solar_radiation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_938_1'))
acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease_phe = TensorMap('phecode_446.2', Interpretation.CATEGORICAL, channel_map={'no_acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease':0, 'acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_446_2'))
acute_gastritis_phe = TensorMap('phecode_535.1', Interpretation.CATEGORICAL, channel_map={'no_acute_gastritis':0, 'acute_gastritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 291], 'phecode_535_1'))
acute_inflammatory_pelvic_disease_phe = TensorMap('phecode_614.31', Interpretation.CATEGORICAL, channel_map={'no_acute_inflammatory_pelvic_disease':0, 'acute_inflammatory_pelvic_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7812], 'phecode_614_31'))
acute_laryngitis_and_tracheitis_phe = TensorMap('phecode_465.4', Interpretation.CATEGORICAL, channel_map={'no_acute_laryngitis_and_tracheitis':0, 'acute_laryngitis_and_tracheitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2314], 'phecode_465_4'))
acute_osteomyelitis_phe = TensorMap('phecode_710.11', Interpretation.CATEGORICAL, channel_map={'no_acute_osteomyelitis':0, 'acute_osteomyelitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38461], 'phecode_710_11'))
acute_pain_phe = TensorMap('phecode_338.1', Interpretation.CATEGORICAL, channel_map={'no_acute_pain':0, 'acute_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 826], 'phecode_338_1'))
acute_pancreatitis_phe = TensorMap('phecode_577.1', Interpretation.CATEGORICAL, channel_map={'no_acute_pancreatitis':0, 'acute_pancreatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 283], 'phecode_577_1'))
acute_pericarditis_phe = TensorMap('phecode_420.21', Interpretation.CATEGORICAL, channel_map={'no_acute_pericarditis':0, 'acute_pericarditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11], 'phecode_420_21'))
acute_periodontitis_phe = TensorMap('phecode_523.31', Interpretation.CATEGORICAL, channel_map={'no_acute_periodontitis':0, 'acute_periodontitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2525], 'phecode_523_31'))
acute_pharyngitis_phe = TensorMap('phecode_465.2', Interpretation.CATEGORICAL, channel_map={'no_acute_pharyngitis':0, 'acute_pharyngitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 393], 'phecode_465_2'))
acute_posthemorrhagic_anemia_phe = TensorMap('phecode_285.1', Interpretation.CATEGORICAL, channel_map={'no_acute_posthemorrhagic_anemia':0, 'acute_posthemorrhagic_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1355], 'phecode_285_1'))
acute_prostatitis_phe = TensorMap('phecode_601.11', Interpretation.CATEGORICAL, channel_map={'no_acute_prostatitis':0, 'acute_prostatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1440], 'phecode_601_11'))
acute_reaction_to_stress_phe = TensorMap('phecode_300.8', Interpretation.CATEGORICAL, channel_map={'no_acute_reaction_to_stress':0, 'acute_reaction_to_stress':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2659], 'phecode_300_8'))
acute_renal_failure_phe = TensorMap('phecode_585.1', Interpretation.CATEGORICAL, channel_map={'no_acute_renal_failure':0, 'acute_renal_failure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 53], 'phecode_585_1'))
acute_rheumatic_heart_disease_phe = TensorMap('phecode_394.4', Interpretation.CATEGORICAL, channel_map={'no_acute_rheumatic_heart_disease':0, 'acute_rheumatic_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7692], 'phecode_394_4'))
acute_sinusitis_phe = TensorMap('phecode_464', Interpretation.CATEGORICAL, channel_map={'no_acute_sinusitis':0, 'acute_sinusitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1838], 'phecode_464'))
acute_tonsillitis_phe = TensorMap('phecode_474.1', Interpretation.CATEGORICAL, channel_map={'no_acute_tonsillitis':0, 'acute_tonsillitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 637], 'phecode_474_1'))
acute_transverse_myelitis_phe = TensorMap('phecode_323.2', Interpretation.CATEGORICAL, channel_map={'no_acute_transverse_myelitis':0, 'acute_transverse_myelitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4672], 'phecode_323_2'))
acute_upper_respiratory_infections_of_multiple_or_unspecified_sites_phe = TensorMap('phecode_465', Interpretation.CATEGORICAL, channel_map={'no_acute_upper_respiratory_infections_of_multiple_or_unspecified_sites':0, 'acute_upper_respiratory_infections_of_multiple_or_unspecified_sites':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27], 'phecode_465'))
acute_vascular_insufficiency_of_intestine_phe = TensorMap('phecode_441.1', Interpretation.CATEGORICAL, channel_map={'no_acute_vascular_insufficiency_of_intestine':0, 'acute_vascular_insufficiency_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1231], 'phecode_441_1'))
adjustment_reaction_phe = TensorMap('phecode_304', Interpretation.CATEGORICAL, channel_map={'no_adjustment_reaction':0, 'adjustment_reaction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1201], 'phecode_304'))
adrenal_hyperfunction_phe = TensorMap('phecode_255.1', Interpretation.CATEGORICAL, channel_map={'no_adrenal_hyperfunction':0, 'adrenal_hyperfunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_255_1'))
adrenogenital_disorders_phe = TensorMap('phecode_255.3', Interpretation.CATEGORICAL, channel_map={'no_adrenogenital_disorders':0, 'adrenogenital_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20833], 'phecode_255_3'))
adverse_drug_events_and_drug_allergies_phe = TensorMap('phecode_979', Interpretation.CATEGORICAL, channel_map={'no_adverse_drug_events_and_drug_allergies':0, 'adverse_drug_events_and_drug_allergies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 827], 'phecode_979'))
adverse_effects_of_antibacterials_not_penicillins_phe = TensorMap('phecode_960.1', Interpretation.CATEGORICAL, channel_map={'no_adverse_effects_of_antibacterials_not_penicillins':0, 'adverse_effects_of_antibacterials_not_penicillins':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5050], 'phecode_960_1'))
adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics_phe = TensorMap('phecode_967', Interpretation.CATEGORICAL, channel_map={'no_adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics':0, 'adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 701], 'phecode_967'))
adverse_reaction_to_serum_or_vaccine_phe = TensorMap('phecode_941', Interpretation.CATEGORICAL, channel_map={'no_adverse_reaction_to_serum_or_vaccine':0, 'adverse_reaction_to_serum_or_vaccine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1543], 'phecode_941'))
agorophobia_social_phobia_and_panic_disorder_phe = TensorMap('phecode_300.12', Interpretation.CATEGORICAL, channel_map={'no_agorophobia_social_phobia_and_panic_disorder':0, 'agorophobia_social_phobia_and_panic_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 393], 'phecode_300_12'))
alcohol_related_disorders_phe = TensorMap('phecode_317', Interpretation.CATEGORICAL, channel_map={'no_alcohol_related_disorders':0, 'alcohol_related_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 63], 'phecode_317'))
alcoholic_liver_damage_phe = TensorMap('phecode_317.11', Interpretation.CATEGORICAL, channel_map={'no_alcoholic_liver_damage':0, 'alcoholic_liver_damage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 412], 'phecode_317_11'))
alcoholism_phe = TensorMap('phecode_317.1', Interpretation.CATEGORICAL, channel_map={'no_alcoholism':0, 'alcoholism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 40], 'phecode_317_1'))
alkalosis_phe = TensorMap('phecode_276.42', Interpretation.CATEGORICAL, channel_map={'no_alkalosis':0, 'alkalosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2118], 'phecode_276_42'))
allergic_purpura_phe = TensorMap('phecode_287.2', Interpretation.CATEGORICAL, channel_map={'no_allergic_purpura':0, 'allergic_purpura':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8064], 'phecode_287_2'))
allergic_reaction_to_food_phe = TensorMap('phecode_930', Interpretation.CATEGORICAL, channel_map={'no_allergic_reaction_to_food':0, 'allergic_reaction_to_food':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1811], 'phecode_930'))
allergic_rhinitis_phe = TensorMap('phecode_476', Interpretation.CATEGORICAL, channel_map={'no_allergic_rhinitis':0, 'allergic_rhinitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 264], 'phecode_476'))
allergies_other_phe = TensorMap('phecode_949', Interpretation.CATEGORICAL, channel_map={'no_allergies_other':0, 'allergies_other':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 818], 'phecode_949'))
allergy_adverse_effect_of_penicillin_phe = TensorMap('phecode_960.2', Interpretation.CATEGORICAL, channel_map={'no_allergy_adverse_effect_of_penicillin':0, 'allergy_adverse_effect_of_penicillin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20], 'phecode_960_2'))
alopecia_phe = TensorMap('phecode_704.1', Interpretation.CATEGORICAL, channel_map={'no_alopecia':0, 'alopecia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2392], 'phecode_704_1'))
alopecia_areata_phe = TensorMap('phecode_704.11', Interpretation.CATEGORICAL, channel_map={'no_alopecia_areata':0, 'alopecia_areata':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18518], 'phecode_704_11'))
alteration_of_consciousness_phe = TensorMap('phecode_291.8', Interpretation.CATEGORICAL, channel_map={'no_alteration_of_consciousness':0, 'alteration_of_consciousness':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 649], 'phecode_291_8'))
altered_mental_status_phe = TensorMap('phecode_292.4', Interpretation.CATEGORICAL, channel_map={'no_altered_mental_status':0, 'altered_mental_status':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 130], 'phecode_292_4'))
althetes_foot_phe = TensorMap('phecode_110.12', Interpretation.CATEGORICAL, channel_map={'no_althetes_foot':0, 'althetes_foot':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2427], 'phecode_110_12'))
alzheimers_disease_phe = TensorMap('phecode_290.11', Interpretation.CATEGORICAL, channel_map={'no_alzheimers_disease':0, 'alzheimers_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 253], 'phecode_290_11'))
amblyopia_phe = TensorMap('phecode_368.1', Interpretation.CATEGORICAL, channel_map={'no_amblyopia':0, 'amblyopia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 551], 'phecode_368_1'))
amyloidosis_phe = TensorMap('phecode_270.33', Interpretation.CATEGORICAL, channel_map={'no_amyloidosis':0, 'amyloidosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2475], 'phecode_270_33'))
anal_and_rectal_conditions_phe = TensorMap('phecode_565', Interpretation.CATEGORICAL, channel_map={'no_anal_and_rectal_conditions':0, 'anal_and_rectal_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 43], 'phecode_565'))
anal_and_rectal_polyp_phe = TensorMap('phecode_565.1', Interpretation.CATEGORICAL, channel_map={'no_anal_and_rectal_polyp':0, 'anal_and_rectal_polyp':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45], 'phecode_565_1'))
anaphylactic_shock_nos_phe = TensorMap('phecode_946', Interpretation.CATEGORICAL, channel_map={'no_anaphylactic_shock_nos':0, 'anaphylactic_shock_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 625], 'phecode_946'))
anemia_during_pregnancy_phe = TensorMap('phecode_644', Interpretation.CATEGORICAL, channel_map={'no_anemia_during_pregnancy':0, 'anemia_during_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 425], 'phecode_644'))
anemia_in_neoplastic_disease_phe = TensorMap('phecode_285.22', Interpretation.CATEGORICAL, channel_map={'no_anemia_in_neoplastic_disease':0, 'anemia_in_neoplastic_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 766], 'phecode_285_22'))
anemia_of_chronic_disease_phe = TensorMap('phecode_285.2', Interpretation.CATEGORICAL, channel_map={'no_anemia_of_chronic_disease':0, 'anemia_of_chronic_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1059], 'phecode_285_2'))
aneurysm_and_dissection_of_heart_phe = TensorMap('phecode_411.41', Interpretation.CATEGORICAL, channel_map={'no_aneurysm_and_dissection_of_heart':0, 'aneurysm_and_dissection_of_heart':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 475], 'phecode_411_41'))
aneurysm_of_artery_of_lower_extremity_phe = TensorMap('phecode_442.3', Interpretation.CATEGORICAL, channel_map={'no_aneurysm_of_artery_of_lower_extremity':0, 'aneurysm_of_artery_of_lower_extremity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1992], 'phecode_442_3'))
aneurysm_of_iliac_artery_phe = TensorMap('phecode_442.2', Interpretation.CATEGORICAL, channel_map={'no_aneurysm_of_iliac_artery':0, 'aneurysm_of_iliac_artery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3246], 'phecode_442_2'))
aneurysm_of_other_specified_artery_phe = TensorMap('phecode_442.8', Interpretation.CATEGORICAL, channel_map={'no_aneurysm_of_other_specified_artery':0, 'aneurysm_of_other_specified_artery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1538], 'phecode_442_8'))
angina_pectoris_phe = TensorMap('phecode_411.3', Interpretation.CATEGORICAL, channel_map={'no_angina_pectoris':0, 'angina_pectoris':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 22], 'phecode_411_3'))
anisometropia_phe = TensorMap('phecode_368.3', Interpretation.CATEGORICAL, channel_map={'no_anisometropia':0, 'anisometropia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2000], 'phecode_368_3'))
ankylosing_spondylitis_phe = TensorMap('phecode_715.2', Interpretation.CATEGORICAL, channel_map={'no_ankylosing_spondylitis':0, 'ankylosing_spondylitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 869], 'phecode_715_2'))
ankylosis_of_joint_phe = TensorMap('phecode_741.1', Interpretation.CATEGORICAL, channel_map={'no_ankylosis_of_joint':0, 'ankylosis_of_joint':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 26315], 'phecode_741_1'))
anomalies_of_endocrine_glands_congenital_phe = TensorMap('phecode_759.1', Interpretation.CATEGORICAL, channel_map={'no_anomalies_of_endocrine_glands_congenital':0, 'anomalies_of_endocrine_glands_congenital':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4098], 'phecode_759_1'))
anomalies_of_jaw_size_symmetry_phe = TensorMap('phecode_526.3', Interpretation.CATEGORICAL, channel_map={'no_anomalies_of_jaw_size_symmetry':0, 'anomalies_of_jaw_size_symmetry':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6172], 'phecode_526_3'))
anomalies_of_pupillary_function_phe = TensorMap('phecode_379.4', Interpretation.CATEGORICAL, channel_map={'no_anomalies_of_pupillary_function':0, 'anomalies_of_pupillary_function':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4065], 'phecode_379_4'))
anomalies_of_respiratory_system_congenital_phe = TensorMap('phecode_748', Interpretation.CATEGORICAL, channel_map={'no_anomalies_of_respiratory_system_congenital':0, 'anomalies_of_respiratory_system_congenital':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4807], 'phecode_748'))
anomalies_of_tooth_position_malocclusion_phe = TensorMap('phecode_524.3', Interpretation.CATEGORICAL, channel_map={'no_anomalies_of_tooth_position_malocclusion':0, 'anomalies_of_tooth_position_malocclusion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1272], 'phecode_524_3'))
anomalous_atrioventricular_excitation_phe = TensorMap('phecode_426.4', Interpretation.CATEGORICAL, channel_map={'no_anomalous_atrioventricular_excitation':0, 'anomalous_atrioventricular_excitation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1689], 'phecode_426_4'))
anorexia_phe = TensorMap('phecode_260.6', Interpretation.CATEGORICAL, channel_map={'no_anorexia':0, 'anorexia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 381], 'phecode_260_6'))
anorexia_nervosa_phe = TensorMap('phecode_305.21', Interpretation.CATEGORICAL, channel_map={'no_anorexia_nervosa':0, 'anorexia_nervosa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7575], 'phecode_305_21'))
antepartum_hemorrhage_abruptio_placentae_and_placenta_previa_phe = TensorMap('phecode_635.2', Interpretation.CATEGORICAL, channel_map={'no_antepartum_hemorrhage_abruptio_placentae_and_placenta_previa':0, 'antepartum_hemorrhage_abruptio_placentae_and_placenta_previa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 425], 'phecode_635_2'))
anterior_horn_cell_disease_phe = TensorMap('phecode_334.2', Interpretation.CATEGORICAL, channel_map={'no_anterior_horn_cell_disease':0, 'anterior_horn_cell_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1358], 'phecode_334_2'))
anticoagulants_causing_adverse_effects_phe = TensorMap('phecode_964.1', Interpretation.CATEGORICAL, channel_map={'no_anticoagulants_causing_adverse_effects':0, 'anticoagulants_causing_adverse_effects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1193], 'phecode_964_1'))
antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use_phe = TensorMap('phecode_972.2', Interpretation.CATEGORICAL, channel_map={'no_antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use':0, 'antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7352], 'phecode_972_2'))
antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects_phe = TensorMap('phecode_963.1', Interpretation.CATEGORICAL, channel_map={'no_antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects':0, 'antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_963_1'))
antirheumatics_causing_adverse_effects_in_therapeutic_use_phe = TensorMap('phecode_965.2', Interpretation.CATEGORICAL, channel_map={'no_antirheumatics_causing_adverse_effects_in_therapeutic_use':0, 'antirheumatics_causing_adverse_effects_in_therapeutic_use':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_965_2'))
antisocial_borderline_personality_disorder_phe = TensorMap('phecode_301.2', Interpretation.CATEGORICAL, channel_map={'no_antisocial_borderline_personality_disorder':0, 'antisocial_borderline_personality_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2403], 'phecode_301_2'))
anxiety_disorder_phe = TensorMap('phecode_300.1', Interpretation.CATEGORICAL, channel_map={'no_anxiety_disorder':0, 'anxiety_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 51], 'phecode_300_1'))
anxiety_disorders_phe = TensorMap('phecode_300', Interpretation.CATEGORICAL, channel_map={'no_anxiety_disorders':0, 'anxiety_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 978], 'phecode_300'))
aortic_aneurysm_phe = TensorMap('phecode_442.1', Interpretation.CATEGORICAL, channel_map={'no_aortic_aneurysm':0, 'aortic_aneurysm':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 537], 'phecode_442_1'))
aortic_valve_disease_phe = TensorMap('phecode_394.3', Interpretation.CATEGORICAL, channel_map={'no_aortic_valve_disease':0, 'aortic_valve_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 235], 'phecode_394_3'))
aphakia_and_other_disorders_of_lens_phe = TensorMap('phecode_379.3', Interpretation.CATEGORICAL, channel_map={'no_aphakia_and_other_disorders_of_lens':0, 'aphakia_and_other_disorders_of_lens':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 14], 'phecode_379_3'))
aphasia_speech_disturbance_phe = TensorMap('phecode_292.1', Interpretation.CATEGORICAL, channel_map={'no_aphasia_speech_disturbance':0, 'aphasia_speech_disturbance':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 208], 'phecode_292_1'))
aplastic_anemia_phe = TensorMap('phecode_284', Interpretation.CATEGORICAL, channel_map={'no_aplastic_anemia':0, 'aplastic_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25], 'phecode_284'))
apnea_phe = TensorMap('phecode_513.31', Interpretation.CATEGORICAL, channel_map={'no_apnea':0, 'apnea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55555], 'phecode_513_31'))
appendiceal_conditions_phe = TensorMap('phecode_540', Interpretation.CATEGORICAL, channel_map={'no_appendiceal_conditions':0, 'appendiceal_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 105], 'phecode_540'))
appendicitis_phe = TensorMap('phecode_540.1', Interpretation.CATEGORICAL, channel_map={'no_appendicitis':0, 'appendicitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 681], 'phecode_540_1'))
arrhythmia_cardiac_nos_phe = TensorMap('phecode_427.5', Interpretation.CATEGORICAL, channel_map={'no_arrhythmia_cardiac_nos':0, 'arrhythmia_cardiac_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 353], 'phecode_427_5'))
arterial_dissection_phe = TensorMap('phecode_442.4', Interpretation.CATEGORICAL, channel_map={'no_arterial_dissection':0, 'arterial_dissection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7462], 'phecode_442_4'))
arterial_embolism_and_thrombosis_phe = TensorMap('phecode_444', Interpretation.CATEGORICAL, channel_map={'no_arterial_embolism_and_thrombosis':0, 'arterial_embolism_and_thrombosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 814], 'phecode_444'))
arterial_embolism_and_thrombosis_of_lower_extremity_artery_phe = TensorMap('phecode_444.1', Interpretation.CATEGORICAL, channel_map={'no_arterial_embolism_and_thrombosis_of_lower_extremity_artery':0, 'arterial_embolism_and_thrombosis_of_lower_extremity_artery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 627], 'phecode_444_1'))
arteritis_nos_phe = TensorMap('phecode_446.9', Interpretation.CATEGORICAL, channel_map={'no_arteritis_nos':0, 'arteritis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1262], 'phecode_446_9'))
arthropathy_associated_with_infections_phe = TensorMap('phecode_711', Interpretation.CATEGORICAL, channel_map={'no_arthropathy_associated_with_infections':0, 'arthropathy_associated_with_infections':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35714], 'phecode_711'))
arthropathy_associated_with_neurological_disorders_phe = TensorMap('phecode_713.5', Interpretation.CATEGORICAL, channel_map={'no_arthropathy_associated_with_neurological_disorders':0, 'arthropathy_associated_with_neurological_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4672], 'phecode_713_5'))
arthropathy_associated_with_other_disorders_classified_elsewhere_phe = TensorMap('phecode_713', Interpretation.CATEGORICAL, channel_map={'no_arthropathy_associated_with_other_disorders_classified_elsewhere':0, 'arthropathy_associated_with_other_disorders_classified_elsewhere':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6172], 'phecode_713'))
arthropathy_nos_phe = TensorMap('phecode_716.9', Interpretation.CATEGORICAL, channel_map={'no_arthropathy_nos':0, 'arthropathy_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6], 'phecode_716_9'))
arthropod_borne_diseases_phe = TensorMap('phecode_133', Interpretation.CATEGORICAL, channel_map={'no_arthropod_borne_diseases':0, 'arthropod_borne_diseases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2415], 'phecode_133'))
articular_cartilage_disorder_phe = TensorMap('phecode_742.8', Interpretation.CATEGORICAL, channel_map={'no_articular_cartilage_disorder':0, 'articular_cartilage_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 446], 'phecode_742_8'))
ascites_non_malignant_phe = TensorMap('phecode_572', Interpretation.CATEGORICAL, channel_map={'no_ascites_non_malignant':0, 'ascites_non_malignant':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 202], 'phecode_572'))
ascvd_phe = TensorMap('phecode_414.2', Interpretation.CATEGORICAL, channel_map={'no_ascvd':0, 'ascvd':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4], 'phecode_414_2'))
aseptic_necrosis_of_bone_phe = TensorMap('phecode_733.4', Interpretation.CATEGORICAL, channel_map={'no_aseptic_necrosis_of_bone':0, 'aseptic_necrosis_of_bone':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4545], 'phecode_733_4'))
aspergillosis_phe = TensorMap('phecode_117.4', Interpretation.CATEGORICAL, channel_map={'no_aspergillosis':0, 'aspergillosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2000], 'phecode_117_4'))
asphyxia_and_hypoxemia_phe = TensorMap('phecode_1013', Interpretation.CATEGORICAL, channel_map={'no_asphyxia_and_hypoxemia':0, 'asphyxia_and_hypoxemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1228], 'phecode_1013'))
asthma_phe = TensorMap('phecode_495', Interpretation.CATEGORICAL, channel_map={'no_asthma':0, 'asthma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13], 'phecode_495'))
asthma_with_exacerbation_phe = TensorMap('phecode_495.2', Interpretation.CATEGORICAL, channel_map={'no_asthma_with_exacerbation':0, 'asthma_with_exacerbation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1305], 'phecode_495_2'))
astigmatism_phe = TensorMap('phecode_367.2', Interpretation.CATEGORICAL, channel_map={'no_astigmatism':0, 'astigmatism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1445], 'phecode_367_2'))
atherosclerosis_phe = TensorMap('phecode_440', Interpretation.CATEGORICAL, channel_map={'no_atherosclerosis':0, 'atherosclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1953], 'phecode_440'))
atherosclerosis_of_aorta_phe = TensorMap('phecode_440.9', Interpretation.CATEGORICAL, channel_map={'no_atherosclerosis_of_aorta':0, 'atherosclerosis_of_aorta':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1992], 'phecode_440_9'))
atherosclerosis_of_renal_artery_phe = TensorMap('phecode_440.1', Interpretation.CATEGORICAL, channel_map={'no_atherosclerosis_of_renal_artery':0, 'atherosclerosis_of_renal_artery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4504], 'phecode_440_1'))
atherosclerosis_of_the_extremities_phe = TensorMap('phecode_440.2', Interpretation.CATEGORICAL, channel_map={'no_atherosclerosis_of_the_extremities':0, 'atherosclerosis_of_the_extremities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 758], 'phecode_440_2'))
atopic_contact_dermatitis_due_to_other_or_unspecified_phe = TensorMap('phecode_939', Interpretation.CATEGORICAL, channel_map={'no_atopic_contact_dermatitis_due_to_other_or_unspecified':0, 'atopic_contact_dermatitis_due_to_other_or_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 148], 'phecode_939'))
atrial_fibrillation_and_flutter_phe = TensorMap('phecode_427.2', Interpretation.CATEGORICAL, channel_map={'no_atrial_fibrillation_and_flutter':0, 'atrial_fibrillation_and_flutter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 22], 'phecode_427_2'))
atrioventricular_av_block_phe = TensorMap('phecode_426.2', Interpretation.CATEGORICAL, channel_map={'no_atrioventricular_av_block':0, 'atrioventricular_av_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2512], 'phecode_426_2'))
atrioventricular_block_complete_phe = TensorMap('phecode_426.24', Interpretation.CATEGORICAL, channel_map={'no_atrioventricular_block_complete':0, 'atrioventricular_block_complete':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 510], 'phecode_426_24'))
atrophic_gastritis_phe = TensorMap('phecode_535.2', Interpretation.CATEGORICAL, channel_map={'no_atrophic_gastritis':0, 'atrophic_gastritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 905], 'phecode_535_2'))
atrophy_of_edentulous_alveolar_ridge_phe = TensorMap('phecode_525.2', Interpretation.CATEGORICAL, channel_map={'no_atrophy_of_edentulous_alveolar_ridge':0, 'atrophy_of_edentulous_alveolar_ridge':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5747], 'phecode_525_2'))
atrophy_of_female_genital_tract_phe = TensorMap('phecode_624.2', Interpretation.CATEGORICAL, channel_map={'no_atrophy_of_female_genital_tract':0, 'atrophy_of_female_genital_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 766], 'phecode_624_2'))
autism_phe = TensorMap('phecode_313.3', Interpretation.CATEGORICAL, channel_map={'no_autism':0, 'autism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4716], 'phecode_313_3'))
autoimmune_hemolytic_anemias_phe = TensorMap('phecode_283.1', Interpretation.CATEGORICAL, channel_map={'no_autoimmune_hemolytic_anemias':0, 'autoimmune_hemolytic_anemias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3048], 'phecode_283_1'))
back_pain_phe = TensorMap('phecode_760', Interpretation.CATEGORICAL, channel_map={'no_back_pain':0, 'back_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55], 'phecode_760'))
bacterial_enteritis_phe = TensorMap('phecode_8.5', Interpretation.CATEGORICAL, channel_map={'no_bacterial_enteritis':0, 'bacterial_enteritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 168], 'phecode_8_5'))
bacterial_infection_nos_phe = TensorMap('phecode_41', Interpretation.CATEGORICAL, channel_map={'no_bacterial_infection_nos':0, 'bacterial_infection_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 31], 'phecode_41'))
bacterial_pneumonia_phe = TensorMap('phecode_480.1', Interpretation.CATEGORICAL, channel_map={'no_bacterial_pneumonia':0, 'bacterial_pneumonia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 331], 'phecode_480_1'))
balanoposthitis_phe = TensorMap('phecode_601.4', Interpretation.CATEGORICAL, channel_map={'no_balanoposthitis':0, 'balanoposthitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1298], 'phecode_601_4'))
bariatric_surgery_phe = TensorMap('phecode_539', Interpretation.CATEGORICAL, channel_map={'no_bariatric_surgery':0, 'bariatric_surgery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8333], 'phecode_539'))
behcets_syndrome_phe = TensorMap('phecode_711.3', Interpretation.CATEGORICAL, channel_map={'no_behcets_syndrome':0, 'behcets_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11627], 'phecode_711_3'))
benign_neoplasm_of_adrenal_gland_phe = TensorMap('phecode_227.1', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_adrenal_gland':0, 'benign_neoplasm_of_adrenal_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1683], 'phecode_227_1'))
benign_neoplasm_of_bone_and_articular_cartilage_phe = TensorMap('phecode_213', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_bone_and_articular_cartilage':0, 'benign_neoplasm_of_bone_and_articular_cartilage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1204], 'phecode_213'))
benign_neoplasm_of_brain_and_other_parts_of_nervous_system_phe = TensorMap('phecode_225', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_brain_and_other_parts_of_nervous_system':0, 'benign_neoplasm_of_brain_and_other_parts_of_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_225'))
benign_neoplasm_of_brain_cranial_nerves_meninges_phe = TensorMap('phecode_225.1', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_brain_cranial_nerves_meninges':0, 'benign_neoplasm_of_brain_cranial_nerves_meninges':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 450], 'phecode_225_1'))
benign_neoplasm_of_breast_phe = TensorMap('phecode_610.4', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_breast':0, 'benign_neoplasm_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 262], 'phecode_610_4'))
benign_neoplasm_of_colon_phe = TensorMap('phecode_208', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_colon':0, 'benign_neoplasm_of_colon':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16], 'phecode_208'))
benign_neoplasm_of_eye_phe = TensorMap('phecode_224', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_eye':0, 'benign_neoplasm_of_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5681], 'phecode_224'))
benign_neoplasm_of_eye_uveal_phe = TensorMap('phecode_224.1', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_eye_uveal':0, 'benign_neoplasm_of_eye_uveal':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5747], 'phecode_224_1'))
benign_neoplasm_of_kidney_and_other_urinary_organs_phe = TensorMap('phecode_223', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_kidney_and_other_urinary_organs':0, 'benign_neoplasm_of_kidney_and_other_urinary_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1865], 'phecode_223'))
benign_neoplasm_of_lip_oral_cavity_and_pharynx_phe = TensorMap('phecode_210', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_lip_oral_cavity_and_pharynx':0, 'benign_neoplasm_of_lip_oral_cavity_and_pharynx':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 391], 'phecode_210'))
benign_neoplasm_of_lymph_nodes_phe = TensorMap('phecode_229.1', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_lymph_nodes':0, 'benign_neoplasm_of_lymph_nodes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8333], 'phecode_229_1'))
benign_neoplasm_of_male_genital_organs_phe = TensorMap('phecode_222', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_male_genital_organs':0, 'benign_neoplasm_of_male_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2392], 'phecode_222'))
benign_neoplasm_of_other_endocrine_glands_and_related_structures_phe = TensorMap('phecode_227', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_other_endocrine_glands_and_related_structures':0, 'benign_neoplasm_of_other_endocrine_glands_and_related_structures':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_227'))
benign_neoplasm_of_other_female_genital_organs_phe = TensorMap('phecode_221', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_other_female_genital_organs':0, 'benign_neoplasm_of_other_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1838], 'phecode_221'))
benign_neoplasm_of_other_parts_of_digestive_system_phe = TensorMap('phecode_211', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_other_parts_of_digestive_system':0, 'benign_neoplasm_of_other_parts_of_digestive_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 59], 'phecode_211'))
benign_neoplasm_of_ovary_phe = TensorMap('phecode_220', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_ovary':0, 'benign_neoplasm_of_ovary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 256], 'phecode_220'))
benign_neoplasm_of_parathyroid_gland_phe = TensorMap('phecode_227.2', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_parathyroid_gland':0, 'benign_neoplasm_of_parathyroid_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 888], 'phecode_227_2'))
benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch_phe = TensorMap('phecode_227.3', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch':0, 'benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1002], 'phecode_227_3'))
benign_neoplasm_of_respiratory_and_intrathoracic_organs_phe = TensorMap('phecode_212', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_respiratory_and_intrathoracic_organs':0, 'benign_neoplasm_of_respiratory_and_intrathoracic_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1002], 'phecode_212'))
benign_neoplasm_of_skin_phe = TensorMap('phecode_216', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_skin':0, 'benign_neoplasm_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45], 'phecode_216'))
benign_neoplasm_of_spinal_cord_meninges_phe = TensorMap('phecode_225.2', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_spinal_cord_meninges':0, 'benign_neoplasm_of_spinal_cord_meninges':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6756], 'phecode_225_2'))
benign_neoplasm_of_thyroid_glands_phe = TensorMap('phecode_226', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_thyroid_glands':0, 'benign_neoplasm_of_thyroid_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1340], 'phecode_226'))
benign_neoplasm_of_unspecified_sites_phe = TensorMap('phecode_229', Interpretation.CATEGORICAL, channel_map={'no_benign_neoplasm_of_unspecified_sites':0, 'benign_neoplasm_of_unspecified_sites':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7], 'phecode_229'))
bipolar_phe = TensorMap('phecode_296.1', Interpretation.CATEGORICAL, channel_map={'no_bipolar':0, 'bipolar':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 312], 'phecode_296_1'))
bladder_neck_obstruction_phe = TensorMap('phecode_596.1', Interpretation.CATEGORICAL, channel_map={'no_bladder_neck_obstruction':0, 'bladder_neck_obstruction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 193], 'phecode_596_1'))
blastomycotic_infection_phe = TensorMap('phecode_117.3', Interpretation.CATEGORICAL, channel_map={'no_blastomycotic_infection':0, 'blastomycotic_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_117_3'))
blindness_and_low_vision_phe = TensorMap('phecode_367.9', Interpretation.CATEGORICAL, channel_map={'no_blindness_and_low_vision':0, 'blindness_and_low_vision':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 448], 'phecode_367_9'))
blood_in_stool_phe = TensorMap('phecode_578.2', Interpretation.CATEGORICAL, channel_map={'no_blood_in_stool':0, 'blood_in_stool':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 131], 'phecode_578_2'))
bone_cancer_phe = TensorMap('phecode_170.1', Interpretation.CATEGORICAL, channel_map={'no_bone_cancer':0, 'bone_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2564], 'phecode_170_1'))
bone_marrow_or_stem_cell_transplant_phe = TensorMap('phecode_860', Interpretation.CATEGORICAL, channel_map={'no_bone_marrow_or_stem_cell_transplant':0, 'bone_marrow_or_stem_cell_transplant':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 765], 'phecode_860'))
breast_cancer_phe = TensorMap('phecode_174', Interpretation.CATEGORICAL, channel_map={'no_breast_cancer':0, 'breast_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3333], 'phecode_174'))
breast_cancer_female_phe = TensorMap('phecode_174.1', Interpretation.CATEGORICAL, channel_map={'no_breast_cancer_female':0, 'breast_cancer_female':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 176], 'phecode_174_1'))
breast_disorder_nos_phe = TensorMap('phecode_613.9', Interpretation.CATEGORICAL, channel_map={'no_breast_disorder_nos':0, 'breast_disorder_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 70], 'phecode_613_9'))
broken_tooth_phe = TensorMap('phecode_873', Interpretation.CATEGORICAL, channel_map={'no_broken_tooth':0, 'broken_tooth':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3816], 'phecode_873'))
bronchiectasis_phe = TensorMap('phecode_496.3', Interpretation.CATEGORICAL, channel_map={'no_bronchiectasis':0, 'bronchiectasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 158], 'phecode_496_3'))
bronchitis_phe = TensorMap('phecode_497', Interpretation.CATEGORICAL, channel_map={'no_bronchitis':0, 'bronchitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10], 'phecode_497'))
bronchopneumonia_and_lung_abscess_phe = TensorMap('phecode_480.5', Interpretation.CATEGORICAL, channel_map={'no_bronchopneumonia_and_lung_abscess':0, 'bronchopneumonia_and_lung_abscess':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 463], 'phecode_480_5'))
bullous_dermatoses_phe = TensorMap('phecode_695.2', Interpretation.CATEGORICAL, channel_map={'no_bullous_dermatoses':0, 'bullous_dermatoses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1872], 'phecode_695_2'))
bundle_branch_block_phe = TensorMap('phecode_426.3', Interpretation.CATEGORICAL, channel_map={'no_bundle_branch_block':0, 'bundle_branch_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1035], 'phecode_426_3'))
burns_phe = TensorMap('phecode_1000', Interpretation.CATEGORICAL, channel_map={'no_burns':0, 'burns':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 653], 'phecode_1000'))
bursitis_phe = TensorMap('phecode_726.3', Interpretation.CATEGORICAL, channel_map={'no_bursitis':0, 'bursitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 662], 'phecode_726_3'))
bursitis_disorders_phe = TensorMap('phecode_727.2', Interpretation.CATEGORICAL, channel_map={'no_bursitis_disorders':0, 'bursitis_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_727_2'))
cachexia_phe = TensorMap('phecode_260.1', Interpretation.CATEGORICAL, channel_map={'no_cachexia':0, 'cachexia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6250], 'phecode_260_1'))
calcaneal_spur_exostosis_nos_phe = TensorMap('phecode_726.4', Interpretation.CATEGORICAL, channel_map={'no_calcaneal_spur_exostosis_nos':0, 'calcaneal_spur_exostosis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9090], 'phecode_726_4'))
calculus_of_bile_duct_phe = TensorMap('phecode_574.2', Interpretation.CATEGORICAL, channel_map={'no_calculus_of_bile_duct':0, 'calculus_of_bile_duct':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 137], 'phecode_574_2'))
calculus_of_kidney_phe = TensorMap('phecode_594.1', Interpretation.CATEGORICAL, channel_map={'no_calculus_of_kidney':0, 'calculus_of_kidney':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 111], 'phecode_594_1'))
calculus_of_lower_urinary_tract_phe = TensorMap('phecode_594.2', Interpretation.CATEGORICAL, channel_map={'no_calculus_of_lower_urinary_tract':0, 'calculus_of_lower_urinary_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 449], 'phecode_594_2'))
calculus_of_ureter_phe = TensorMap('phecode_594.3', Interpretation.CATEGORICAL, channel_map={'no_calculus_of_ureter':0, 'calculus_of_ureter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 154], 'phecode_594_3'))
cancer_of_bladder_phe = TensorMap('phecode_189.2', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_bladder':0, 'cancer_of_bladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 457], 'phecode_189_2'))
cancer_of_brain_phe = TensorMap('phecode_191.11', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_brain':0, 'cancer_of_brain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 692], 'phecode_191_11'))
cancer_of_brain_and_nervous_system_phe = TensorMap('phecode_191.1', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_brain_and_nervous_system':0, 'cancer_of_brain_and_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6849], 'phecode_191_1'))
cancer_of_bronchus_lung_phe = TensorMap('phecode_165.1', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_bronchus_lung':0, 'cancer_of_bronchus_lung':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 147], 'phecode_165_1'))
cancer_of_connective_tissue_phe = TensorMap('phecode_170.2', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_connective_tissue':0, 'cancer_of_connective_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 468], 'phecode_170_2'))
cancer_of_esophagus_phe = TensorMap('phecode_150', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_esophagus':0, 'cancer_of_esophagus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 51], 'phecode_150'))
cancer_of_eye_phe = TensorMap('phecode_190', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_eye':0, 'cancer_of_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 490], 'phecode_190'))
cancer_of_hypopharynx_phe = TensorMap('phecode_149.3', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_hypopharynx':0, 'cancer_of_hypopharynx':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6666], 'phecode_149_3'))
cancer_of_intrathoracic_organs_phe = TensorMap('phecode_164', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_intrathoracic_organs':0, 'cancer_of_intrathoracic_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6097], 'phecode_164'))
cancer_of_larynx_phe = TensorMap('phecode_149.4', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_larynx':0, 'cancer_of_larynx':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1461], 'phecode_149_4'))
cancer_of_larynx_pharynx_nasal_cavities_phe = TensorMap('phecode_149', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_larynx_pharynx_nasal_cavities':0, 'cancer_of_larynx_pharynx_nasal_cavities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 30], 'phecode_149'))
cancer_of_lip_phe = TensorMap('phecode_145.1', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_lip':0, 'cancer_of_lip':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6756], 'phecode_145_1'))
cancer_of_liver_and_intrahepatic_bile_duct_phe = TensorMap('phecode_155', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_liver_and_intrahepatic_bile_duct':0, 'cancer_of_liver_and_intrahepatic_bile_duct':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1201], 'phecode_155'))
cancer_of_major_salivary_glands_phe = TensorMap('phecode_145.3', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_major_salivary_glands':0, 'cancer_of_major_salivary_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3267], 'phecode_145_3'))
cancer_of_mouth_phe = TensorMap('phecode_145', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_mouth':0, 'cancer_of_mouth':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2369], 'phecode_145'))
cancer_of_nasopharynx_phe = TensorMap('phecode_149.2', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_nasopharynx':0, 'cancer_of_nasopharynx':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_149_2'))
cancer_of_of_nasal_cavities_phe = TensorMap('phecode_149.9', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_of_nasal_cavities':0, 'cancer_of_of_nasal_cavities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6329], 'phecode_149_9'))
cancer_of_oropharynx_phe = TensorMap('phecode_149.1', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_oropharynx':0, 'cancer_of_oropharynx':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1416], 'phecode_149_1'))
cancer_of_other_endocrine_glands_phe = TensorMap('phecode_194', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_other_endocrine_glands':0, 'cancer_of_other_endocrine_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3623], 'phecode_194'))
cancer_of_other_female_genital_organs_phe = TensorMap('phecode_184', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_other_female_genital_organs':0, 'cancer_of_other_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2487], 'phecode_184'))
cancer_of_other_female_genital_organs_excluding_uterus_and_ovary_phe = TensorMap('phecode_184.2', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_other_female_genital_organs_excluding_uterus_and_ovary':0, 'cancer_of_other_female_genital_organs_excluding_uterus_and_ovary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1170], 'phecode_184_2'))
cancer_of_other_lymphoid_histiocytic_tissue_phe = TensorMap('phecode_202', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_other_lymphoid_histiocytic_tissue':0, 'cancer_of_other_lymphoid_histiocytic_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 290], 'phecode_202'))
cancer_of_other_male_genital_organs_phe = TensorMap('phecode_187', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_other_male_genital_organs':0, 'cancer_of_other_male_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 48], 'phecode_187'))
cancer_of_prostate_phe = TensorMap('phecode_185', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_prostate':0, 'cancer_of_prostate':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 40], 'phecode_185'))
cancer_of_stomach_phe = TensorMap('phecode_151', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_stomach':0, 'cancer_of_stomach':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 583], 'phecode_151'))
cancer_of_the_gums_phe = TensorMap('phecode_145.4', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_the_gums':0, 'cancer_of_the_gums':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10638], 'phecode_145_4'))
cancer_of_the_mouth_floor_phe = TensorMap('phecode_145.5', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_the_mouth_floor':0, 'cancer_of_the_mouth_floor':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6329], 'phecode_145_5'))
cancer_of_tongue_phe = TensorMap('phecode_145.2', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_tongue':0, 'cancer_of_tongue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1176], 'phecode_145_2'))
cancer_of_urinary_organs_incl_kidney_and_bladder_phe = TensorMap('phecode_189', Interpretation.CATEGORICAL, channel_map={'no_cancer_of_urinary_organs_incl_kidney_and_bladder':0, 'cancer_of_urinary_organs_incl_kidney_and_bladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 121], 'phecode_189'))
cancer_suspected_or_other_phe = TensorMap('phecode_195', Interpretation.CATEGORICAL, channel_map={'no_cancer_suspected_or_other':0, 'cancer_suspected_or_other':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 422], 'phecode_195'))
cancer_within_the_respiratory_system_phe = TensorMap('phecode_165', Interpretation.CATEGORICAL, channel_map={'no_cancer_within_the_respiratory_system':0, 'cancer_within_the_respiratory_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 111], 'phecode_165'))
candidiasis_phe = TensorMap('phecode_112', Interpretation.CATEGORICAL, channel_map={'no_candidiasis':0, 'candidiasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 140], 'phecode_112'))
candidiasis_of_skin_and_nails_phe = TensorMap('phecode_112.3', Interpretation.CATEGORICAL, channel_map={'no_candidiasis_of_skin_and_nails':0, 'candidiasis_of_skin_and_nails':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8064], 'phecode_112_3'))
carbuncle_and_furuncle_phe = TensorMap('phecode_686.1', Interpretation.CATEGORICAL, channel_map={'no_carbuncle_and_furuncle':0, 'carbuncle_and_furuncle':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 146], 'phecode_686_1'))
carcinoid_syndrome_phe = TensorMap('phecode_259.2', Interpretation.CATEGORICAL, channel_map={'no_carcinoid_syndrome':0, 'carcinoid_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8333], 'phecode_259_2'))
carcinoma_in_situ_of_skin_phe = TensorMap('phecode_172.3', Interpretation.CATEGORICAL, channel_map={'no_carcinoma_in_situ_of_skin':0, 'carcinoma_in_situ_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 484], 'phecode_172_3'))
cardiac_and_circulatory_congenital_anomalies_phe = TensorMap('phecode_747', Interpretation.CATEGORICAL, channel_map={'no_cardiac_and_circulatory_congenital_anomalies':0, 'cardiac_and_circulatory_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 288], 'phecode_747'))
cardiac_arrest_phe = TensorMap('phecode_427.42', Interpretation.CATEGORICAL, channel_map={'no_cardiac_arrest':0, 'cardiac_arrest':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 309], 'phecode_427_42'))
cardiac_arrest_and_ventricular_fibrillation_phe = TensorMap('phecode_427.4', Interpretation.CATEGORICAL, channel_map={'no_cardiac_arrest_and_ventricular_fibrillation':0, 'cardiac_arrest_and_ventricular_fibrillation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 968], 'phecode_427_4'))
cardiac_conduction_disorders_phe = TensorMap('phecode_426', Interpretation.CATEGORICAL, channel_map={'no_cardiac_conduction_disorders':0, 'cardiac_conduction_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2049], 'phecode_426'))
cardiac_congenital_anomalies_phe = TensorMap('phecode_747.1', Interpretation.CATEGORICAL, channel_map={'no_cardiac_congenital_anomalies':0, 'cardiac_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1661], 'phecode_747_1'))
cardiac_dysrhythmias_phe = TensorMap('phecode_427', Interpretation.CATEGORICAL, channel_map={'no_cardiac_dysrhythmias':0, 'cardiac_dysrhythmias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1388], 'phecode_427'))
cardiac_pacemaker_device_in_situ_phe = TensorMap('phecode_426.9', Interpretation.CATEGORICAL, channel_map={'no_cardiac_pacemaker_device_in_situ':0, 'cardiac_pacemaker_device_in_situ':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1483], 'phecode_426_9'))
cardiac_pacemaker_in_situ_phe = TensorMap('phecode_426.91', Interpretation.CATEGORICAL, channel_map={'no_cardiac_pacemaker_in_situ':0, 'cardiac_pacemaker_in_situ':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 134], 'phecode_426_91'))
cardiac_shunt_heart_septal_defect_phe = TensorMap('phecode_747.11', Interpretation.CATEGORICAL, channel_map={'no_cardiac_shunt_heart_septal_defect':0, 'cardiac_shunt_heart_septal_defect':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 621], 'phecode_747_11'))
cardiogenic_shock_phe = TensorMap('phecode_797.1', Interpretation.CATEGORICAL, channel_map={'no_cardiogenic_shock':0, 'cardiogenic_shock':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3048], 'phecode_797_1'))
cardiomegaly_phe = TensorMap('phecode_416', Interpretation.CATEGORICAL, channel_map={'no_cardiomegaly':0, 'cardiomegaly':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 113], 'phecode_416'))
cardiomyopathy_phe = TensorMap('phecode_425', Interpretation.CATEGORICAL, channel_map={'no_cardiomyopathy':0, 'cardiomyopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16666], 'phecode_425'))
carditis_phe = TensorMap('phecode_420', Interpretation.CATEGORICAL, channel_map={'no_carditis':0, 'carditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 31250], 'phecode_420'))
cataplexy_and_narcolepsy_phe = TensorMap('phecode_347', Interpretation.CATEGORICAL, channel_map={'no_cataplexy_and_narcolepsy':0, 'cataplexy_and_narcolepsy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7042], 'phecode_347'))
cataract_phe = TensorMap('phecode_366', Interpretation.CATEGORICAL, channel_map={'no_cataract':0, 'cataract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18], 'phecode_366'))
celiac_disease_phe = TensorMap('phecode_557.1', Interpretation.CATEGORICAL, channel_map={'no_celiac_disease':0, 'celiac_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 201], 'phecode_557_1'))
cellulitis_and_abscess_of_arm_hand_phe = TensorMap('phecode_681.3', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_arm_hand':0, 'cellulitis_and_abscess_of_arm_hand':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 61], 'phecode_681_3'))
cellulitis_and_abscess_of_face_neck_phe = TensorMap('phecode_681.2', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_face_neck':0, 'cellulitis_and_abscess_of_face_neck':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 703], 'phecode_681_2'))
cellulitis_and_abscess_of_fingers_toes_phe = TensorMap('phecode_681.1', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_fingers_toes':0, 'cellulitis_and_abscess_of_fingers_toes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 600], 'phecode_681_1'))
cellulitis_and_abscess_of_foot_toe_phe = TensorMap('phecode_681.6', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_foot_toe':0, 'cellulitis_and_abscess_of_foot_toe':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 61], 'phecode_681_6'))
cellulitis_and_abscess_of_leg_except_foot_phe = TensorMap('phecode_681.5', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_leg_except_foot':0, 'cellulitis_and_abscess_of_leg_except_foot':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 61], 'phecode_681_5'))
cellulitis_and_abscess_of_oral_soft_tissues_phe = TensorMap('phecode_528.3', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_oral_soft_tissues':0, 'cellulitis_and_abscess_of_oral_soft_tissues':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1953], 'phecode_528_3'))
cellulitis_and_abscess_of_trunk_phe = TensorMap('phecode_681.7', Interpretation.CATEGORICAL, channel_map={'no_cellulitis_and_abscess_of_trunk':0, 'cellulitis_and_abscess_of_trunk':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 570], 'phecode_681_7'))
central_origin_vertigo_phe = TensorMap('phecode_386.21', Interpretation.CATEGORICAL, channel_map={'no_central_origin_vertigo':0, 'central_origin_vertigo':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29411], 'phecode_386_21'))
cerebral_aneurysm_phe = TensorMap('phecode_433.5', Interpretation.CATEGORICAL, channel_map={'no_cerebral_aneurysm':0, 'cerebral_aneurysm':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 917], 'phecode_433_5'))
cerebral_artery_occlusion_with_cerebral_infarction_phe = TensorMap('phecode_433.21', Interpretation.CATEGORICAL, channel_map={'no_cerebral_artery_occlusion_with_cerebral_infarction':0, 'cerebral_artery_occlusion_with_cerebral_infarction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 182], 'phecode_433_21'))
cerebral_atherosclerosis_phe = TensorMap('phecode_433.12', Interpretation.CATEGORICAL, channel_map={'no_cerebral_atherosclerosis':0, 'cerebral_atherosclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1259], 'phecode_433_12'))
cerebral_cysts_phe = TensorMap('phecode_348.4', Interpretation.CATEGORICAL, channel_map={'no_cerebral_cysts':0, 'cerebral_cysts':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2427], 'phecode_348_4'))
cerebral_degeneration_unspecified_phe = TensorMap('phecode_331.9', Interpretation.CATEGORICAL, channel_map={'no_cerebral_degeneration_unspecified':0, 'cerebral_degeneration_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 680], 'phecode_331_9'))
cerebral_edema_and_compression_of_brain_phe = TensorMap('phecode_348.2', Interpretation.CATEGORICAL, channel_map={'no_cerebral_edema_and_compression_of_brain':0, 'cerebral_edema_and_compression_of_brain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1182], 'phecode_348_2'))
cerebral_ischemia_phe = TensorMap('phecode_433.3', Interpretation.CATEGORICAL, channel_map={'no_cerebral_ischemia':0, 'cerebral_ischemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 284], 'phecode_433_3'))
cerebrovascular_disease_phe = TensorMap('phecode_433', Interpretation.CATEGORICAL, channel_map={'no_cerebrovascular_disease':0, 'cerebrovascular_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38], 'phecode_433'))
certain_early_complications_of_trauma_or_procedure_phe = TensorMap('phecode_958', Interpretation.CATEGORICAL, channel_map={'no_certain_early_complications_of_trauma_or_procedure':0, 'certain_early_complications_of_trauma_or_procedure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 742], 'phecode_958'))
cervical_cancer_phe = TensorMap('phecode_180.1', Interpretation.CATEGORICAL, channel_map={'no_cervical_cancer':0, 'cervical_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 73], 'phecode_180_1'))
cervical_incompetence_phe = TensorMap('phecode_636.8', Interpretation.CATEGORICAL, channel_map={'no_cervical_incompetence':0, 'cervical_incompetence':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7692], 'phecode_636_8'))
cervical_intraepithelial_neoplasia_cin_cervical_dysplasia_phe = TensorMap('phecode_180.3', Interpretation.CATEGORICAL, channel_map={'no_cervical_intraepithelial_neoplasia_cin_cervical_dysplasia':0, 'cervical_intraepithelial_neoplasia_cin_cervical_dysplasia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 193], 'phecode_180_3'))
cervical_radiculitis_phe = TensorMap('phecode_765', Interpretation.CATEGORICAL, channel_map={'no_cervical_radiculitis':0, 'cervical_radiculitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 674], 'phecode_765'))
cervicalgia_phe = TensorMap('phecode_761', Interpretation.CATEGORICAL, channel_map={'no_cervicalgia':0, 'cervicalgia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 338], 'phecode_761'))
cervicitis_and_endocervicitis_phe = TensorMap('phecode_614.51', Interpretation.CATEGORICAL, channel_map={'no_cervicitis_and_endocervicitis':0, 'cervicitis_and_endocervicitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 348], 'phecode_614_51'))
changes_in_skin_texture_phe = TensorMap('phecode_687.3', Interpretation.CATEGORICAL, channel_map={'no_changes_in_skin_texture':0, 'changes_in_skin_texture':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3267], 'phecode_687_3'))
chemotherapy_phe = TensorMap('phecode_197', Interpretation.CATEGORICAL, channel_map={'no_chemotherapy':0, 'chemotherapy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18], 'phecode_197'))
cholangitis_phe = TensorMap('phecode_575.1', Interpretation.CATEGORICAL, channel_map={'no_cholangitis':0, 'cholangitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 606], 'phecode_575_1'))
cholecystitis_without_cholelithiasis_phe = TensorMap('phecode_574.3', Interpretation.CATEGORICAL, channel_map={'no_cholecystitis_without_cholelithiasis':0, 'cholecystitis_without_cholelithiasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 133], 'phecode_574_3'))
cholelithiasis_phe = TensorMap('phecode_574.1', Interpretation.CATEGORICAL, channel_map={'no_cholelithiasis':0, 'cholelithiasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41], 'phecode_574_1'))
cholelithiasis_with_acute_cholecystitis_phe = TensorMap('phecode_574.11', Interpretation.CATEGORICAL, channel_map={'no_cholelithiasis_with_acute_cholecystitis':0, 'cholelithiasis_with_acute_cholecystitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 223], 'phecode_574_11'))
cholelithiasis_with_other_cholecystitis_phe = TensorMap('phecode_574.12', Interpretation.CATEGORICAL, channel_map={'no_cholelithiasis_with_other_cholecystitis':0, 'cholelithiasis_with_other_cholecystitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 66], 'phecode_574_12'))
cholesteatoma_phe = TensorMap('phecode_385.3', Interpretation.CATEGORICAL, channel_map={'no_cholesteatoma':0, 'cholesteatoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 695], 'phecode_385_3'))
cholesterolosis_of_gallbladder_phe = TensorMap('phecode_575.6', Interpretation.CATEGORICAL, channel_map={'no_cholesterolosis_of_gallbladder':0, 'cholesterolosis_of_gallbladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 802], 'phecode_575_6'))
chondrocalcinosis_phe = TensorMap('phecode_274.21', Interpretation.CATEGORICAL, channel_map={'no_chondrocalcinosis':0, 'chondrocalcinosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1256], 'phecode_274_21'))
chondromalacia_phe = TensorMap('phecode_733.9', Interpretation.CATEGORICAL, channel_map={'no_chondromalacia':0, 'chondromalacia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 14705], 'phecode_733_9'))
chorioretinal_inflammations_scars_and_other_disorders_of_choroid_phe = TensorMap('phecode_363', Interpretation.CATEGORICAL, channel_map={'no_chorioretinal_inflammations_scars_and_other_disorders_of_choroid':0, 'chorioretinal_inflammations_scars_and_other_disorders_of_choroid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3787], 'phecode_363'))
chorioretinal_scars_phe = TensorMap('phecode_363.3', Interpretation.CATEGORICAL, channel_map={'no_chorioretinal_scars':0, 'chorioretinal_scars':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3968], 'phecode_363_3'))
choroidal_degenerations_phe = TensorMap('phecode_363.4', Interpretation.CATEGORICAL, channel_map={'no_choroidal_degenerations':0, 'choroidal_degenerations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11904], 'phecode_363_4'))
chromosomal_anomalies_phe = TensorMap('phecode_758.1', Interpretation.CATEGORICAL, channel_map={'no_chromosomal_anomalies':0, 'chromosomal_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3246], 'phecode_758_1'))
chromosomal_anomalies_and_genetic_disorders_phe = TensorMap('phecode_758', Interpretation.CATEGORICAL, channel_map={'no_chromosomal_anomalies_and_genetic_disorders':0, 'chromosomal_anomalies_and_genetic_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4201], 'phecode_758'))
chronic_airway_obstruction_phe = TensorMap('phecode_496', Interpretation.CATEGORICAL, channel_map={'no_chronic_airway_obstruction':0, 'chronic_airway_obstruction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41], 'phecode_496'))
chronic_bronchitis_phe = TensorMap('phecode_496.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_bronchitis':0, 'chronic_bronchitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1072], 'phecode_496_2'))
chronic_cystitis_phe = TensorMap('phecode_592.12', Interpretation.CATEGORICAL, channel_map={'no_chronic_cystitis':0, 'chronic_cystitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 407], 'phecode_592_12'))
chronic_dermatitis_due_to_solar_radiation_phe = TensorMap('phecode_938.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_dermatitis_due_to_solar_radiation':0, 'chronic_dermatitis_due_to_solar_radiation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1319], 'phecode_938_2'))
chronic_fatigue_syndrome_phe = TensorMap('phecode_798.1', Interpretation.CATEGORICAL, channel_map={'no_chronic_fatigue_syndrome':0, 'chronic_fatigue_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 577], 'phecode_798_1'))
chronic_glomerulonephritis_nos_phe = TensorMap('phecode_580.14', Interpretation.CATEGORICAL, channel_map={'no_chronic_glomerulonephritis_nos':0, 'chronic_glomerulonephritis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 469], 'phecode_580_14'))
chronic_hepatitis_phe = TensorMap('phecode_70.4', Interpretation.CATEGORICAL, channel_map={'no_chronic_hepatitis':0, 'chronic_hepatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1269], 'phecode_70_4'))
chronic_inflammatory_disorders_of_orbit_phe = TensorMap('phecode_371.9', Interpretation.CATEGORICAL, channel_map={'no_chronic_inflammatory_disorders_of_orbit':0, 'chronic_inflammatory_disorders_of_orbit':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_371_9'))
chronic_inflammatory_pelvic_disease_phe = TensorMap('phecode_614.32', Interpretation.CATEGORICAL, channel_map={'no_chronic_inflammatory_pelvic_disease':0, 'chronic_inflammatory_pelvic_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 905], 'phecode_614_32'))
chronic_interstitial_cystitis_phe = TensorMap('phecode_592.13', Interpretation.CATEGORICAL, channel_map={'no_chronic_interstitial_cystitis':0, 'chronic_interstitial_cystitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1369], 'phecode_592_13'))
chronic_laryngitis_phe = TensorMap('phecode_473.1', Interpretation.CATEGORICAL, channel_map={'no_chronic_laryngitis':0, 'chronic_laryngitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4273], 'phecode_473_1'))
chronic_liver_disease_and_cirrhosis_phe = TensorMap('phecode_571', Interpretation.CATEGORICAL, channel_map={'no_chronic_liver_disease_and_cirrhosis':0, 'chronic_liver_disease_and_cirrhosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 546], 'phecode_571'))
chronic_lymphocytic_thyroiditis_phe = TensorMap('phecode_245.21', Interpretation.CATEGORICAL, channel_map={'no_chronic_lymphocytic_thyroiditis':0, 'chronic_lymphocytic_thyroiditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1655], 'phecode_245_21'))
chronic_osteomyelitis_phe = TensorMap('phecode_710.12', Interpretation.CATEGORICAL, channel_map={'no_chronic_osteomyelitis':0, 'chronic_osteomyelitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9259], 'phecode_710_12'))
chronic_pain_phe = TensorMap('phecode_338.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_pain':0, 'chronic_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 810], 'phecode_338_2'))
chronic_pancreatitis_phe = TensorMap('phecode_577.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_pancreatitis':0, 'chronic_pancreatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 676], 'phecode_577_2'))
chronic_passive_congestion_of_liver_phe = TensorMap('phecode_573.1', Interpretation.CATEGORICAL, channel_map={'no_chronic_passive_congestion_of_liver':0, 'chronic_passive_congestion_of_liver':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 14285], 'phecode_573_1'))
chronic_pericarditis_phe = TensorMap('phecode_420.22', Interpretation.CATEGORICAL, channel_map={'no_chronic_pericarditis':0, 'chronic_pericarditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4950], 'phecode_420_22'))
chronic_periodontitis_phe = TensorMap('phecode_523.32', Interpretation.CATEGORICAL, channel_map={'no_chronic_periodontitis':0, 'chronic_periodontitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 323], 'phecode_523_32'))
chronic_pharyngitis_and_nasopharyngitis_phe = TensorMap('phecode_472', Interpretation.CATEGORICAL, channel_map={'no_chronic_pharyngitis_and_nasopharyngitis':0, 'chronic_pharyngitis_and_nasopharyngitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 403], 'phecode_472'))
chronic_prostatitis_phe = TensorMap('phecode_601.12', Interpretation.CATEGORICAL, channel_map={'no_chronic_prostatitis':0, 'chronic_prostatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 398], 'phecode_601_12'))
chronic_pulmonary_heart_disease_phe = TensorMap('phecode_415.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_pulmonary_heart_disease':0, 'chronic_pulmonary_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1730], 'phecode_415_2'))
chronic_renal_failure_ckd_phe = TensorMap('phecode_585.3', Interpretation.CATEGORICAL, channel_map={'no_chronic_renal_failure_ckd':0, 'chronic_renal_failure_ckd':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 103], 'phecode_585_3'))
chronic_sinusitis_phe = TensorMap('phecode_475', Interpretation.CATEGORICAL, channel_map={'no_chronic_sinusitis':0, 'chronic_sinusitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 147], 'phecode_475'))
chronic_thyroiditis_phe = TensorMap('phecode_245.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_thyroiditis':0, 'chronic_thyroiditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25000], 'phecode_245_2'))
chronic_tonsillitis_and_adenoiditis_phe = TensorMap('phecode_474.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_tonsillitis_and_adenoiditis':0, 'chronic_tonsillitis_and_adenoiditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 370], 'phecode_474_2'))
chronic_ulcer_of_skin_phe = TensorMap('phecode_707', Interpretation.CATEGORICAL, channel_map={'no_chronic_ulcer_of_skin':0, 'chronic_ulcer_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 249], 'phecode_707'))
chronic_vascular_insufficiency_of_intestine_phe = TensorMap('phecode_441.2', Interpretation.CATEGORICAL, channel_map={'no_chronic_vascular_insufficiency_of_intestine':0, 'chronic_vascular_insufficiency_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7042], 'phecode_441_2'))
chronic_venous_insufficiency_cvi_phe = TensorMap('phecode_456', Interpretation.CATEGORICAL, channel_map={'no_chronic_venous_insufficiency_cvi':0, 'chronic_venous_insufficiency_cvi':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1432], 'phecode_456'))
circadian_rhythm_sleep_disorder_phe = TensorMap('phecode_327.6', Interpretation.CATEGORICAL, channel_map={'no_circadian_rhythm_sleep_disorder':0, 'circadian_rhythm_sleep_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11627], 'phecode_327_6'))
circulatory_disease_nec_phe = TensorMap('phecode_459.9', Interpretation.CATEGORICAL, channel_map={'no_circulatory_disease_nec':0, 'circulatory_disease_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2], 'phecode_459_9'))
circumscribed_scleroderma_phe = TensorMap('phecode_701.3', Interpretation.CATEGORICAL, channel_map={'no_circumscribed_scleroderma':0, 'circumscribed_scleroderma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5882], 'phecode_701_3'))
cirrhosis_of_liver_without_mention_of_alcohol_phe = TensorMap('phecode_571.51', Interpretation.CATEGORICAL, channel_map={'no_cirrhosis_of_liver_without_mention_of_alcohol':0, 'cirrhosis_of_liver_without_mention_of_alcohol':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 463], 'phecode_571_51'))
cleft_palate_phe = TensorMap('phecode_749.1', Interpretation.CATEGORICAL, channel_map={'no_cleft_palate':0, 'cleft_palate':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11627], 'phecode_749_1'))
coagulation_defects_complicating_pregnancy_or_postpartum_phe = TensorMap('phecode_286.3', Interpretation.CATEGORICAL, channel_map={'no_coagulation_defects_complicating_pregnancy_or_postpartum':0, 'coagulation_defects_complicating_pregnancy_or_postpartum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6097], 'phecode_286_3'))
coccidioidomycosis_phe = TensorMap('phecode_117.2', Interpretation.CATEGORICAL, channel_map={'no_coccidioidomycosis':0, 'coccidioidomycosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_117_2'))
colon_cancer_phe = TensorMap('phecode_153.2', Interpretation.CATEGORICAL, channel_map={'no_colon_cancer':0, 'colon_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 113], 'phecode_153_2'))
color_vision_deficiencies_phe = TensorMap('phecode_368.5', Interpretation.CATEGORICAL, channel_map={'no_color_vision_deficiencies':0, 'color_vision_deficiencies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_368_5'))
colorectal_cancer_phe = TensorMap('phecode_153', Interpretation.CATEGORICAL, channel_map={'no_colorectal_cancer':0, 'colorectal_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13157], 'phecode_153'))
coma_phe = TensorMap('phecode_348.7', Interpretation.CATEGORICAL, channel_map={'no_coma':0, 'coma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 950], 'phecode_348_7'))
complex_regional_central_pain_syndrome_phe = TensorMap('phecode_355', Interpretation.CATEGORICAL, channel_map={'no_complex_regional_central_pain_syndrome':0, 'complex_regional_central_pain_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10204], 'phecode_355'))
complication_due_to_other_implant_and_internal_device_phe = TensorMap('phecode_859', Interpretation.CATEGORICAL, channel_map={'no_complication_due_to_other_implant_and_internal_device':0, 'complication_due_to_other_implant_and_internal_device':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 109], 'phecode_859'))
complication_of_amputation_stump_phe = TensorMap('phecode_874', Interpretation.CATEGORICAL, channel_map={'no_complication_of_amputation_stump':0, 'complication_of_amputation_stump':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2057], 'phecode_874'))
complication_of_colostomy_or_enterostomy_phe = TensorMap('phecode_853', Interpretation.CATEGORICAL, channel_map={'no_complication_of_colostomy_or_enterostomy':0, 'complication_of_colostomy_or_enterostomy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 665], 'phecode_853'))
complication_of_internal_orthopedic_device_phe = TensorMap('phecode_858', Interpretation.CATEGORICAL, channel_map={'no_complication_of_internal_orthopedic_device':0, 'complication_of_internal_orthopedic_device':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 97], 'phecode_858'))
complications_following_abortion_or_ectopic_and_molar_pregnancies_phe = TensorMap('phecode_639', Interpretation.CATEGORICAL, channel_map={'no_complications_following_abortion_or_ectopic_and_molar_pregnancies':0, 'complications_following_abortion_or_ectopic_and_molar_pregnancies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8196], 'phecode_639'))
complications_of_cardiac_vascular_device_implant_and_graft_phe = TensorMap('phecode_854', Interpretation.CATEGORICAL, channel_map={'no_complications_of_cardiac_vascular_device_implant_and_graft':0, 'complications_of_cardiac_vascular_device_implant_and_graft':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 131], 'phecode_854'))
complications_of_in_utero_procedures_phe = TensorMap('phecode_679', Interpretation.CATEGORICAL, channel_map={'no_complications_of_in_utero_procedures':0, 'complications_of_in_utero_procedures':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7142], 'phecode_679'))
complications_of_labor_and_delivery_nec_phe = TensorMap('phecode_669', Interpretation.CATEGORICAL, channel_map={'no_complications_of_labor_and_delivery_nec':0, 'complications_of_labor_and_delivery_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 39], 'phecode_669'))
complications_of_surgical_and_medical_procedures_phe = TensorMap('phecode_1011', Interpretation.CATEGORICAL, channel_map={'no_complications_of_surgical_and_medical_procedures':0, 'complications_of_surgical_and_medical_procedures':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_1011'))
complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery_phe = TensorMap('phecode_668', Interpretation.CATEGORICAL, channel_map={'no_complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery':0, 'complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_668'))
complications_of_transplants_and_reattached_limbs_phe = TensorMap('phecode_851', Interpretation.CATEGORICAL, channel_map={'no_complications_of_transplants_and_reattached_limbs':0, 'complications_of_transplants_and_reattached_limbs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17], 'phecode_851'))
concussion_phe = TensorMap('phecode_817', Interpretation.CATEGORICAL, channel_map={'no_concussion':0, 'concussion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6097], 'phecode_817'))
conditions_involving_the_integument_and_temperature_regulation_of_fetus_and_newborn_phe = TensorMap('phecode_656.7', Interpretation.CATEGORICAL, channel_map={'no_conditions_involving_the_integument_and_temperature_regulation_of_fetus_and_newborn':0, 'conditions_involving_the_integument_and_temperature_regulation_of_fetus_and_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_656_7'))
conduct_disorders_phe = TensorMap('phecode_312', Interpretation.CATEGORICAL, channel_map={'no_conduct_disorders':0, 'conduct_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29411], 'phecode_312'))
conductive_hearing_loss_phe = TensorMap('phecode_389.2', Interpretation.CATEGORICAL, channel_map={'no_conductive_hearing_loss':0, 'conductive_hearing_loss':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 880], 'phecode_389_2'))
congenital_anomalies_of_abdominal_wall_diaphram_phe = TensorMap('phecode_756.1', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_abdominal_wall_diaphram':0, 'congenital_anomalies_of_abdominal_wall_diaphram':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15151], 'phecode_756_1'))
congenital_anomalies_of_breast_phe = TensorMap('phecode_612.3', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_breast':0, 'congenital_anomalies_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4385], 'phecode_612_3'))
congenital_anomalies_of_esophagus_phe = TensorMap('phecode_750.14', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_esophagus':0, 'congenital_anomalies_of_esophagus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7575], 'phecode_750_14'))
congenital_anomalies_of_face_and_neck_phe = TensorMap('phecode_749', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_face_and_neck':0, 'congenital_anomalies_of_face_and_neck':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 994], 'phecode_749'))
congenital_anomalies_of_female_genital_organs_phe = TensorMap('phecode_751.11', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_female_genital_organs':0, 'congenital_anomalies_of_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 856], 'phecode_751_11'))
congenital_anomalies_of_genital_organs_phe = TensorMap('phecode_751.1', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_genital_organs':0, 'congenital_anomalies_of_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 548], 'phecode_751_1'))
congenital_anomalies_of_great_vessels_phe = TensorMap('phecode_747.13', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_great_vessels':0, 'congenital_anomalies_of_great_vessels':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 172], 'phecode_747_13'))
congenital_anomalies_of_intestine_phe = TensorMap('phecode_750.21', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_intestine':0, 'congenital_anomalies_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1655], 'phecode_750_21'))
congenital_anomalies_of_limbs_phe = TensorMap('phecode_755', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_limbs':0, 'congenital_anomalies_of_limbs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_755'))
congenital_anomalies_of_male_genital_organs_phe = TensorMap('phecode_751.12', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_male_genital_organs':0, 'congenital_anomalies_of_male_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1779], 'phecode_751_12'))
congenital_anomalies_of_mouth_tongue_phe = TensorMap('phecode_750.13', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_mouth_tongue':0, 'congenital_anomalies_of_mouth_tongue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1838], 'phecode_750_13'))
congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue_phe = TensorMap('phecode_756.3', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue':0, 'congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4347], 'phecode_756_3'))
congenital_anomalies_of_peripheral_vascular_system_phe = TensorMap('phecode_747.2', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_peripheral_vascular_system':0, 'congenital_anomalies_of_peripheral_vascular_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2136], 'phecode_747_2'))
congenital_anomalies_of_posterior_segment_of_eye_phe = TensorMap('phecode_753.2', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_posterior_segment_of_eye':0, 'congenital_anomalies_of_posterior_segment_of_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21739], 'phecode_753_2'))
congenital_anomalies_of_skin_phe = TensorMap('phecode_691', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_skin':0, 'congenital_anomalies_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4545], 'phecode_691'))
congenital_anomalies_of_skull_and_face_bones_phe = TensorMap('phecode_749.2', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_skull_and_face_bones':0, 'congenital_anomalies_of_skull_and_face_bones':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15151], 'phecode_749_2'))
congenital_anomalies_of_stomach_phe = TensorMap('phecode_750.15', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_stomach':0, 'congenital_anomalies_of_stomach':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7462], 'phecode_750_15'))
congenital_anomalies_of_the_eye_phe = TensorMap('phecode_753', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_the_eye':0, 'congenital_anomalies_of_the_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8474], 'phecode_753'))
congenital_anomalies_of_the_integument_phe = TensorMap('phecode_757', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_the_integument':0, 'congenital_anomalies_of_the_integument':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21739], 'phecode_757'))
congenital_anomalies_of_upper_limb_including_shoulder_girdle_phe = TensorMap('phecode_755.4', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_upper_limb_including_shoulder_girdle':0, 'congenital_anomalies_of_upper_limb_including_shoulder_girdle':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8064], 'phecode_755_4'))
congenital_anomalies_of_urinary_system_phe = TensorMap('phecode_751.2', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomalies_of_urinary_system':0, 'congenital_anomalies_of_urinary_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 407], 'phecode_751_2'))
congenital_anomaly_of_fingers_toes_phe = TensorMap('phecode_755.3', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomaly_of_fingers_toes':0, 'congenital_anomaly_of_fingers_toes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12500], 'phecode_755_3'))
congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas_phe = TensorMap('phecode_750.22', Interpretation.CATEGORICAL, channel_map={'no_congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas':0, 'congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2994], 'phecode_750_22'))
congenital_cataract_and_lens_anomalies_phe = TensorMap('phecode_753.1', Interpretation.CATEGORICAL, channel_map={'no_congenital_cataract_and_lens_anomalies':0, 'congenital_cataract_and_lens_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5208], 'phecode_753_1'))
congenital_coagulation_defects_phe = TensorMap('phecode_286.1', Interpretation.CATEGORICAL, channel_map={'no_congenital_coagulation_defects':0, 'congenital_coagulation_defects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9615], 'phecode_286_1'))
congenital_deficiency_of_other_clotting_factors_including_factor_vii_phe = TensorMap('phecode_286.12', Interpretation.CATEGORICAL, channel_map={'no_congenital_deficiency_of_other_clotting_factors_including_factor_vii':0, 'congenital_deficiency_of_other_clotting_factors_including_factor_vii':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1228], 'phecode_286_12'))
congenital_deformities_of_feet_phe = TensorMap('phecode_755.1', Interpretation.CATEGORICAL, channel_map={'no_congenital_deformities_of_feet':0, 'congenital_deformities_of_feet':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1694], 'phecode_755_1'))
congenital_factor_viii_disorder_phe = TensorMap('phecode_286.13', Interpretation.CATEGORICAL, channel_map={'no_congenital_factor_viii_disorder':0, 'congenital_factor_viii_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5263], 'phecode_286_13'))
congenital_hip_dysplasia_and_deformity_phe = TensorMap('phecode_755.61', Interpretation.CATEGORICAL, channel_map={'no_congenital_hip_dysplasia_and_deformity':0, 'congenital_hip_dysplasia_and_deformity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3355], 'phecode_755_61'))
congenital_hypertrophic_pyloric_stenosis_phe = TensorMap('phecode_750.5', Interpretation.CATEGORICAL, channel_map={'no_congenital_hypertrophic_pyloric_stenosis':0, 'congenital_hypertrophic_pyloric_stenosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_750_5'))
congenital_hypothyroidism_phe = TensorMap('phecode_244.5', Interpretation.CATEGORICAL, channel_map={'no_congenital_hypothyroidism':0, 'congenital_hypothyroidism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12195], 'phecode_244_5'))
congenital_musculoskeletal_deformities_of_spine_phe = TensorMap('phecode_754', Interpretation.CATEGORICAL, channel_map={'no_congenital_musculoskeletal_deformities_of_spine':0, 'congenital_musculoskeletal_deformities_of_spine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3448], 'phecode_754'))
congenital_osteodystrophies_phe = TensorMap('phecode_756.5', Interpretation.CATEGORICAL, channel_map={'no_congenital_osteodystrophies':0, 'congenital_osteodystrophies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2283], 'phecode_756_5'))
congenital_pigmentary_anomalies_of_skin_phe = TensorMap('phecode_691.3', Interpretation.CATEGORICAL, channel_map={'no_congenital_pigmentary_anomalies_of_skin':0, 'congenital_pigmentary_anomalies_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17241], 'phecode_691_3'))
congestive_heart_failure_chf_nos_phe = TensorMap('phecode_428.1', Interpretation.CATEGORICAL, channel_map={'no_congestive_heart_failure_chf_nos':0, 'congestive_heart_failure_chf_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 138], 'phecode_428_1'))
conjunctivitis_infectious_phe = TensorMap('phecode_369.5', Interpretation.CATEGORICAL, channel_map={'no_conjunctivitis_infectious':0, 'conjunctivitis_infectious':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1344], 'phecode_369_5'))
conjunctivitis_noninfectious_phe = TensorMap('phecode_371.2', Interpretation.CATEGORICAL, channel_map={'no_conjunctivitis_noninfectious':0, 'conjunctivitis_noninfectious':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11904], 'phecode_371_2'))
constipation_phe = TensorMap('phecode_563', Interpretation.CATEGORICAL, channel_map={'no_constipation':0, 'constipation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29], 'phecode_563'))
constitutional_aplastic_anemia_phe = TensorMap('phecode_284.2', Interpretation.CATEGORICAL, channel_map={'no_constitutional_aplastic_anemia':0, 'constitutional_aplastic_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 22727], 'phecode_284_2'))
contact_dermatitis_and_other_eczema_due_to_plants_except_food_phe = TensorMap('phecode_931', Interpretation.CATEGORICAL, channel_map={'no_contact_dermatitis_and_other_eczema_due_to_plants_except_food':0, 'contact_dermatitis_and_other_eczema_due_to_plants_except_food':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_931'))
contracture_of_joint_phe = TensorMap('phecode_739', Interpretation.CATEGORICAL, channel_map={'no_contracture_of_joint':0, 'contracture_of_joint':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4672], 'phecode_739'))
contracture_of_palmar_fascia_dupuytrens_disease_phe = TensorMap('phecode_728.71', Interpretation.CATEGORICAL, channel_map={'no_contracture_of_palmar_fascia_dupuytrens_disease':0, 'contracture_of_palmar_fascia_dupuytrens_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 141], 'phecode_728_71'))
contracture_of_tendon_sheath_phe = TensorMap('phecode_727.7', Interpretation.CATEGORICAL, channel_map={'no_contracture_of_tendon_sheath':0, 'contracture_of_tendon_sheath':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2906], 'phecode_727_7'))
contusion_phe = TensorMap('phecode_916', Interpretation.CATEGORICAL, channel_map={'no_contusion':0, 'contusion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 238], 'phecode_916'))
convulsions_phe = TensorMap('phecode_345.3', Interpretation.CATEGORICAL, channel_map={'no_convulsions':0, 'convulsions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 151], 'phecode_345_3'))
cornea_replaced_by_transplant_phe = TensorMap('phecode_364.9', Interpretation.CATEGORICAL, channel_map={'no_cornea_replaced_by_transplant':0, 'cornea_replaced_by_transplant':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2824], 'phecode_364_9'))
corneal_degenerations_phe = TensorMap('phecode_364.4', Interpretation.CATEGORICAL, channel_map={'no_corneal_degenerations':0, 'corneal_degenerations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2272], 'phecode_364_4'))
corneal_dystrophy_phe = TensorMap('phecode_364.5', Interpretation.CATEGORICAL, channel_map={'no_corneal_dystrophy':0, 'corneal_dystrophy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1440], 'phecode_364_5'))
corneal_edema_phe = TensorMap('phecode_364.2', Interpretation.CATEGORICAL, channel_map={'no_corneal_edema':0, 'corneal_edema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3649], 'phecode_364_2'))
corneal_opacity_phe = TensorMap('phecode_364.1', Interpretation.CATEGORICAL, channel_map={'no_corneal_opacity':0, 'corneal_opacity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3048], 'phecode_364_1'))
corneal_opacity_and_other_disorders_of_cornea_phe = TensorMap('phecode_364', Interpretation.CATEGORICAL, channel_map={'no_corneal_opacity_and_other_disorders_of_cornea':0, 'corneal_opacity_and_other_disorders_of_cornea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1362], 'phecode_364'))
corneal_ulcer_phe = TensorMap('phecode_370.1', Interpretation.CATEGORICAL, channel_map={'no_corneal_ulcer':0, 'corneal_ulcer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2325], 'phecode_370_1'))
corns_and_callosities_phe = TensorMap('phecode_700', Interpretation.CATEGORICAL, channel_map={'no_corns_and_callosities':0, 'corns_and_callosities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1201], 'phecode_700'))
coronary_atherosclerosis_phe = TensorMap('phecode_411.4', Interpretation.CATEGORICAL, channel_map={'no_coronary_atherosclerosis':0, 'coronary_atherosclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11], 'phecode_411_4'))
costochondritis_phe = TensorMap('phecode_733.6', Interpretation.CATEGORICAL, channel_map={'no_costochondritis':0, 'costochondritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2127], 'phecode_733_6'))
cough_phe = TensorMap('phecode_512.8', Interpretation.CATEGORICAL, channel_map={'no_cough':0, 'cough':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 108], 'phecode_512_8'))
crushing_injury_phe = TensorMap('phecode_1006', Interpretation.CATEGORICAL, channel_map={'no_crushing_injury':0, 'crushing_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1623], 'phecode_1006'))
crushing_or_internal_injury_to_organs_phe = TensorMap('phecode_1008', Interpretation.CATEGORICAL, channel_map={'no_crushing_or_internal_injury_to_organs':0, 'crushing_or_internal_injury_to_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1050], 'phecode_1008'))
crystal_arthropathies_phe = TensorMap('phecode_274.2', Interpretation.CATEGORICAL, channel_map={'no_crystal_arthropathies':0, 'crystal_arthropathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 22727], 'phecode_274_2'))
curvature_of_spine_phe = TensorMap('phecode_737', Interpretation.CATEGORICAL, channel_map={'no_curvature_of_spine':0, 'curvature_of_spine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 377], 'phecode_737'))
cushings_syndrome_phe = TensorMap('phecode_255.11', Interpretation.CATEGORICAL, channel_map={'no_cushings_syndrome':0, 'cushings_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3333], 'phecode_255_11'))
cutaneous_lupus_erythematosus_phe = TensorMap('phecode_695.41', Interpretation.CATEGORICAL, channel_map={'no_cutaneous_lupus_erythematosus':0, 'cutaneous_lupus_erythematosus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2164], 'phecode_695_41'))
cyst_and_pseudocyst_of_pancreas_phe = TensorMap('phecode_577.3', Interpretation.CATEGORICAL, channel_map={'no_cyst_and_pseudocyst_of_pancreas':0, 'cyst_and_pseudocyst_of_pancreas':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 871], 'phecode_577_3'))
cyst_of_bone_phe = TensorMap('phecode_733.2', Interpretation.CATEGORICAL, channel_map={'no_cyst_of_bone':0, 'cyst_of_bone':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9090], 'phecode_733_2'))
cyst_of_kidney_acquired_phe = TensorMap('phecode_586.2', Interpretation.CATEGORICAL, channel_map={'no_cyst_of_kidney_acquired':0, 'cyst_of_kidney_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 232], 'phecode_586_2'))
cyst_of_the_salivary_gland_phe = TensorMap('phecode_528.41', Interpretation.CATEGORICAL, channel_map={'no_cyst_of_the_salivary_gland':0, 'cyst_of_the_salivary_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3787], 'phecode_528_41'))
cyst_or_abscess_of_bartholins_gland_phe = TensorMap('phecode_614.53', Interpretation.CATEGORICAL, channel_map={'no_cyst_or_abscess_of_bartholins_gland':0, 'cyst_or_abscess_of_bartholins_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 535], 'phecode_614_53'))
cystic_fibrosis_phe = TensorMap('phecode_499', Interpretation.CATEGORICAL, channel_map={'no_cystic_fibrosis':0, 'cystic_fibrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19230], 'phecode_499'))
cystic_kidney_disease_phe = TensorMap('phecode_751.21', Interpretation.CATEGORICAL, channel_map={'no_cystic_kidney_disease':0, 'cystic_kidney_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 831], 'phecode_751_21'))
cystic_mastopathy_phe = TensorMap('phecode_610.1', Interpretation.CATEGORICAL, channel_map={'no_cystic_mastopathy':0, 'cystic_mastopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 490], 'phecode_610_1'))
cystitis_phe = TensorMap('phecode_592.1', Interpretation.CATEGORICAL, channel_map={'no_cystitis':0, 'cystitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 182], 'phecode_592_1'))
cysts_of_oral_soft_tissues_phe = TensorMap('phecode_528.4', Interpretation.CATEGORICAL, channel_map={'no_cysts_of_oral_soft_tissues':0, 'cysts_of_oral_soft_tissues':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13157], 'phecode_528_4'))
cysts_of_the_jaws_phe = TensorMap('phecode_526.1', Interpretation.CATEGORICAL, channel_map={'no_cysts_of_the_jaws':0, 'cysts_of_the_jaws':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1152], 'phecode_526_1'))
decreased_libido_phe = TensorMap('phecode_302.1', Interpretation.CATEGORICAL, channel_map={'no_decreased_libido':0, 'decreased_libido':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_302_1'))
decubitus_ulcer_phe = TensorMap('phecode_707.1', Interpretation.CATEGORICAL, channel_map={'no_decubitus_ulcer':0, 'decubitus_ulcer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1319], 'phecode_707_1'))
defibrination_syndrome_phe = TensorMap('phecode_286.6', Interpretation.CATEGORICAL, channel_map={'no_defibrination_syndrome':0, 'defibrination_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 121], 'phecode_286_6'))
deficiency_anemias_phe = TensorMap('phecode_281.9', Interpretation.CATEGORICAL, channel_map={'no_deficiency_anemias':0, 'deficiency_anemias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5952], 'phecode_281_9'))
deficiency_of_humoral_immunity_phe = TensorMap('phecode_279.11', Interpretation.CATEGORICAL, channel_map={'no_deficiency_of_humoral_immunity':0, 'deficiency_of_humoral_immunity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2314], 'phecode_279_11'))
degeneration_of_intervertebral_disc_phe = TensorMap('phecode_722.6', Interpretation.CATEGORICAL, channel_map={'no_degeneration_of_intervertebral_disc':0, 'degeneration_of_intervertebral_disc':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 115], 'phecode_722_6'))
degeneration_of_macula_and_posterior_pole_of_retina_phe = TensorMap('phecode_362.2', Interpretation.CATEGORICAL, channel_map={'no_degeneration_of_macula_and_posterior_pole_of_retina':0, 'degeneration_of_macula_and_posterior_pole_of_retina':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 126], 'phecode_362_2'))
degenerative_and_vascular_disorders_of_ear_phe = TensorMap('phecode_389.3', Interpretation.CATEGORICAL, channel_map={'no_degenerative_and_vascular_disorders_of_ear':0, 'degenerative_and_vascular_disorders_of_ear':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7936], 'phecode_389_3'))
degenerative_disease_of_the_spinal_cord_phe = TensorMap('phecode_334', Interpretation.CATEGORICAL, channel_map={'no_degenerative_disease_of_the_spinal_cord':0, 'degenerative_disease_of_the_spinal_cord':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 227], 'phecode_334'))
degenerative_skin_conditions_and_other_dermatoses_phe = TensorMap('phecode_702', Interpretation.CATEGORICAL, channel_map={'no_degenerative_skin_conditions_and_other_dermatoses':0, 'degenerative_skin_conditions_and_other_dermatoses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_702'))
degenerative_skin_disorders_phe = TensorMap('phecode_702.4', Interpretation.CATEGORICAL, channel_map={'no_degenerative_skin_disorders':0, 'degenerative_skin_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_702_4'))
delay_in_sexual_development_and_puberty_nec_phe = TensorMap('phecode_259.3', Interpretation.CATEGORICAL, channel_map={'no_delay_in_sexual_development_and_puberty_nec':0, 'delay_in_sexual_development_and_puberty_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_259_3'))
delirium_dementia_and_amnestic_and_other_cognitive_disorders_phe = TensorMap('phecode_290', Interpretation.CATEGORICAL, channel_map={'no_delirium_dementia_and_amnestic_and_other_cognitive_disorders':0, 'delirium_dementia_and_amnestic_and_other_cognitive_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 706], 'phecode_290'))
delirium_due_to_conditions_classified_elsewhere_phe = TensorMap('phecode_290.2', Interpretation.CATEGORICAL, channel_map={'no_delirium_due_to_conditions_classified_elsewhere':0, 'delirium_due_to_conditions_classified_elsewhere':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 311], 'phecode_290_2'))
dementia_with_cerebral_degenerations_phe = TensorMap('phecode_290.12', Interpretation.CATEGORICAL, channel_map={'no_dementia_with_cerebral_degenerations':0, 'dementia_with_cerebral_degenerations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4347], 'phecode_290_12'))
dementias_phe = TensorMap('phecode_290.1', Interpretation.CATEGORICAL, channel_map={'no_dementias':0, 'dementias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 312], 'phecode_290_1'))
dental_abrasion_erosion_and_attrition_phe = TensorMap('phecode_521.2', Interpretation.CATEGORICAL, channel_map={'no_dental_abrasion_erosion_and_attrition':0, 'dental_abrasion_erosion_and_attrition':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35714], 'phecode_521_2'))
dental_caries_phe = TensorMap('phecode_521.1', Interpretation.CATEGORICAL, channel_map={'no_dental_caries':0, 'dental_caries':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 104], 'phecode_521_1'))
dentofacial_anomalies_including_malocclusion_phe = TensorMap('phecode_524', Interpretation.CATEGORICAL, channel_map={'no_dentofacial_anomalies_including_malocclusion':0, 'dentofacial_anomalies_including_malocclusion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10204], 'phecode_524'))
dependence_on_respirator_ventilator_or_supplemental_oxygen_phe = TensorMap('phecode_509.8', Interpretation.CATEGORICAL, channel_map={'no_dependence_on_respirator_ventilator_or_supplemental_oxygen':0, 'dependence_on_respirator_ventilator_or_supplemental_oxygen':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 374], 'phecode_509_8'))
depression_phe = TensorMap('phecode_296.2', Interpretation.CATEGORICAL, channel_map={'no_depression':0, 'depression':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 26], 'phecode_296_2'))
derangement_of_joint_non_traumatic_phe = TensorMap('phecode_742', Interpretation.CATEGORICAL, channel_map={'no_derangement_of_joint_non_traumatic':0, 'derangement_of_joint_non_traumatic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9803], 'phecode_742'))
dermatitis_due_to_solar_radiation_phe = TensorMap('phecode_938', Interpretation.CATEGORICAL, channel_map={'no_dermatitis_due_to_solar_radiation':0, 'dermatitis_due_to_solar_radiation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 133], 'phecode_938'))
dermatitis_herpetiformis_phe = TensorMap('phecode_695.21', Interpretation.CATEGORICAL, channel_map={'no_dermatitis_herpetiformis':0, 'dermatitis_herpetiformis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6024], 'phecode_695_21'))
dermatomycoses_phe = TensorMap('phecode_110.2', Interpretation.CATEGORICAL, channel_map={'no_dermatomycoses':0, 'dermatomycoses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_110_2'))
dermatomyositis_phe = TensorMap('phecode_709.5', Interpretation.CATEGORICAL, channel_map={'no_dermatomyositis':0, 'dermatomyositis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7246], 'phecode_709_5'))
dermatophytosis_phe = TensorMap('phecode_110.1', Interpretation.CATEGORICAL, channel_map={'no_dermatophytosis':0, 'dermatophytosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5952], 'phecode_110_1'))
dermatophytosis_of_nail_phe = TensorMap('phecode_110.11', Interpretation.CATEGORICAL, channel_map={'no_dermatophytosis_of_nail':0, 'dermatophytosis_of_nail':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4132], 'phecode_110_11'))
dermatophytosis_of_the_body_phe = TensorMap('phecode_110.13', Interpretation.CATEGORICAL, channel_map={'no_dermatophytosis_of_the_body':0, 'dermatophytosis_of_the_body':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10869], 'phecode_110_13'))
develomental_delays_and_disorders_phe = TensorMap('phecode_315', Interpretation.CATEGORICAL, channel_map={'no_develomental_delays_and_disorders':0, 'develomental_delays_and_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 520], 'phecode_315'))
diabetes_insipidus_phe = TensorMap('phecode_253.3', Interpretation.CATEGORICAL, channel_map={'no_diabetes_insipidus':0, 'diabetes_insipidus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3703], 'phecode_253_3'))
diabetes_mellitus_phe = TensorMap('phecode_250', Interpretation.CATEGORICAL, channel_map={'no_diabetes_mellitus':0, 'diabetes_mellitus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15], 'phecode_250'))
diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy_phe = TensorMap('phecode_649.1', Interpretation.CATEGORICAL, channel_map={'no_diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy':0, 'diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1592], 'phecode_649_1'))
diabetic_retinopathy_phe = TensorMap('phecode_250.7', Interpretation.CATEGORICAL, channel_map={'no_diabetic_retinopathy':0, 'diabetic_retinopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 216], 'phecode_250_7'))
diaper_or_napkin_rash_phe = TensorMap('phecode_949.1', Interpretation.CATEGORICAL, channel_map={'no_diaper_or_napkin_rash':0, 'diaper_or_napkin_rash':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_949_1'))
diaphragmatic_hernia_phe = TensorMap('phecode_550.2', Interpretation.CATEGORICAL, channel_map={'no_diaphragmatic_hernia':0, 'diaphragmatic_hernia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_550_2'))
diarrhea_phe = TensorMap('phecode_561.1', Interpretation.CATEGORICAL, channel_map={'no_diarrhea':0, 'diarrhea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 635], 'phecode_561_1'))
difficulty_in_walking_phe = TensorMap('phecode_741.3', Interpretation.CATEGORICAL, channel_map={'no_difficulty_in_walking':0, 'difficulty_in_walking':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1672], 'phecode_741_3'))
digestive_congenital_anomalies_phe = TensorMap('phecode_750', Interpretation.CATEGORICAL, channel_map={'no_digestive_congenital_anomalies':0, 'digestive_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 518], 'phecode_750'))
diplopia_and_disorders_of_binocular_vision_phe = TensorMap('phecode_368.2', Interpretation.CATEGORICAL, channel_map={'no_diplopia_and_disorders_of_binocular_vision':0, 'diplopia_and_disorders_of_binocular_vision':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 462], 'phecode_368_2'))
disease_of_capillaries_phe = TensorMap('phecode_448', Interpretation.CATEGORICAL, channel_map={'no_disease_of_capillaries':0, 'disease_of_capillaries':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45], 'phecode_448'))
disease_of_tricuspid_valve_phe = TensorMap('phecode_394.7', Interpretation.CATEGORICAL, channel_map={'no_disease_of_tricuspid_valve':0, 'disease_of_tricuspid_valve':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 244], 'phecode_394_7'))
diseases_and_other_conditions_of_the_tongue_phe = TensorMap('phecode_529', Interpretation.CATEGORICAL, channel_map={'no_diseases_and_other_conditions_of_the_tongue':0, 'diseases_and_other_conditions_of_the_tongue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 374], 'phecode_529'))
diseases_due_to_other_mycobacteria_phe = TensorMap('phecode_31', Interpretation.CATEGORICAL, channel_map={'no_diseases_due_to_other_mycobacteria':0, 'diseases_due_to_other_mycobacteria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15625], 'phecode_31'))
diseases_of_esophagus_phe = TensorMap('phecode_530', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_esophagus':0, 'diseases_of_esophagus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 571], 'phecode_530'))
diseases_of_hair_and_hair_follicles_phe = TensorMap('phecode_704', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_hair_and_hair_follicles':0, 'diseases_of_hair_and_hair_follicles':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 75], 'phecode_704'))
diseases_of_hard_tissues_of_teeth_phe = TensorMap('phecode_521', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_hard_tissues_of_teeth':0, 'diseases_of_hard_tissues_of_teeth':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16666], 'phecode_521'))
diseases_of_lips_phe = TensorMap('phecode_528.5', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_lips':0, 'diseases_of_lips':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 521], 'phecode_528_5'))
diseases_of_nail_nos_phe = TensorMap('phecode_703', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_nail_nos':0, 'diseases_of_nail_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1068], 'phecode_703'))
diseases_of_pancreas_phe = TensorMap('phecode_577', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_pancreas':0, 'diseases_of_pancreas':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 552], 'phecode_577'))
diseases_of_pulp_and_periapical_tissues_phe = TensorMap('phecode_522', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_pulp_and_periapical_tissues':0, 'diseases_of_pulp_and_periapical_tissues':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 643], 'phecode_522'))
diseases_of_spleen_phe = TensorMap('phecode_289.5', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_spleen':0, 'diseases_of_spleen':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 642], 'phecode_289_5'))
diseases_of_the_jaws_phe = TensorMap('phecode_526', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_the_jaws':0, 'diseases_of_the_jaws':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_526'))
diseases_of_the_larynx_and_vocal_cords_phe = TensorMap('phecode_473', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_the_larynx_and_vocal_cords':0, 'diseases_of_the_larynx_and_vocal_cords':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 239], 'phecode_473'))
diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue_phe = TensorMap('phecode_528', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue':0, 'diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 177], 'phecode_528'))
diseases_of_the_salivary_glands_phe = TensorMap('phecode_527', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_the_salivary_glands':0, 'diseases_of_the_salivary_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3649], 'phecode_527'))
diseases_of_thymus_gland_phe = TensorMap('phecode_254', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_thymus_gland':0, 'diseases_of_thymus_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19230], 'phecode_254'))
diseases_of_white_blood_cells_phe = TensorMap('phecode_288', Interpretation.CATEGORICAL, channel_map={'no_diseases_of_white_blood_cells':0, 'diseases_of_white_blood_cells':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 545], 'phecode_288'))
dislocation_phe = TensorMap('phecode_830', Interpretation.CATEGORICAL, channel_map={'no_dislocation':0, 'dislocation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 185], 'phecode_830'))
disorder_of_skin_and_subcutaneous_tissue_nos_phe = TensorMap('phecode_689', Interpretation.CATEGORICAL, channel_map={'no_disorder_of_skin_and_subcutaneous_tissue_nos':0, 'disorder_of_skin_and_subcutaneous_tissue_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10], 'phecode_689'))
disorders_involving_the_immune_mechanism_phe = TensorMap('phecode_279', Interpretation.CATEGORICAL, channel_map={'no_disorders_involving_the_immune_mechanism':0, 'disorders_involving_the_immune_mechanism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9433], 'phecode_279'))
disorders_of_accommodation_phe = TensorMap('phecode_368.7', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_accommodation':0, 'disorders_of_accommodation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83333], 'phecode_368_7'))
disorders_of_acoustic_nerve_phe = TensorMap('phecode_389.5', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_acoustic_nerve':0, 'disorders_of_acoustic_nerve':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12195], 'phecode_389_5'))
disorders_of_adrenal_glands_phe = TensorMap('phecode_255', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_adrenal_glands':0, 'disorders_of_adrenal_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1976], 'phecode_255'))
disorders_of_amino_acid_metabolism_phe = TensorMap('phecode_270.2', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_amino_acid_metabolism':0, 'disorders_of_amino_acid_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_270_2'))
disorders_of_bilirubin_excretion_phe = TensorMap('phecode_277.4', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_bilirubin_excretion':0, 'disorders_of_bilirubin_excretion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 914], 'phecode_277_4'))
disorders_of_calcium_phosphorus_metabolism_phe = TensorMap('phecode_275.5', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_calcium_phosphorus_metabolism':0, 'disorders_of_calcium_phosphorus_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 267], 'phecode_275_5'))
disorders_of_conjunctiva_phe = TensorMap('phecode_372', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_conjunctiva':0, 'disorders_of_conjunctiva':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 301], 'phecode_372'))
disorders_of_copper_metabolism_phe = TensorMap('phecode_275.2', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_copper_metabolism':0, 'disorders_of_copper_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55555], 'phecode_275_2'))
disorders_of_diaphragm_phe = TensorMap('phecode_513.8', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_diaphragm':0, 'disorders_of_diaphragm':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2074], 'phecode_513_8'))
disorders_of_esophageal_motility_phe = TensorMap('phecode_530.5', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_esophageal_motility':0, 'disorders_of_esophageal_motility':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 535], 'phecode_530_5'))
disorders_of_external_ear_phe = TensorMap('phecode_380', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_external_ear':0, 'disorders_of_external_ear':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 254], 'phecode_380'))
disorders_of_function_of_stomach_phe = TensorMap('phecode_536', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_function_of_stomach':0, 'disorders_of_function_of_stomach':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_536'))
disorders_of_iris_and_ciliary_body_phe = TensorMap('phecode_379.5', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_iris_and_ciliary_body':0, 'disorders_of_iris_and_ciliary_body':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 884], 'phecode_379_5'))
disorders_of_iron_metabolism_phe = TensorMap('phecode_275.1', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_iron_metabolism':0, 'disorders_of_iron_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 513], 'phecode_275_1'))
disorders_of_lacrimal_system_phe = TensorMap('phecode_375', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_lacrimal_system':0, 'disorders_of_lacrimal_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 203], 'phecode_375'))
disorders_of_magnesium_metabolism_phe = TensorMap('phecode_275.3', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_magnesium_metabolism':0, 'disorders_of_magnesium_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 622], 'phecode_275_3'))
disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract_phe = TensorMap('phecode_626', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract':0, 'disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 240], 'phecode_626'))
disorders_of_mineral_metabolism_phe = TensorMap('phecode_275', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_mineral_metabolism':0, 'disorders_of_mineral_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29411], 'phecode_275'))
disorders_of_muscle_ligament_and_fascia_phe = TensorMap('phecode_728', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_muscle_ligament_and_fascia':0, 'disorders_of_muscle_ligament_and_fascia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7246], 'phecode_728'))
disorders_of_optic_nerve_and_visual_pathways_phe = TensorMap('phecode_377', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_optic_nerve_and_visual_pathways':0, 'disorders_of_optic_nerve_and_visual_pathways':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 786], 'phecode_377'))
disorders_of_other_cranial_nerves_phe = TensorMap('phecode_352', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_other_cranial_nerves':0, 'disorders_of_other_cranial_nerves':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4716], 'phecode_352'))
disorders_of_parathyroid_gland_phe = TensorMap('phecode_252', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_parathyroid_gland':0, 'disorders_of_parathyroid_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11904], 'phecode_252'))
disorders_of_penis_phe = TensorMap('phecode_604', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_penis':0, 'disorders_of_penis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 476], 'phecode_604'))
disorders_of_phosphorus_metabolism_phe = TensorMap('phecode_275.53', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_phosphorus_metabolism':0, 'disorders_of_phosphorus_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1228], 'phecode_275_53'))
disorders_of_porphyrin_metabolism_phe = TensorMap('phecode_277.1', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_porphyrin_metabolism':0, 'disorders_of_porphyrin_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10638], 'phecode_277_1'))
disorders_of_refraction_and_accommodation_blindness_and_low_vision_phe = TensorMap('phecode_367', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_refraction_and_accommodation_blindness_and_low_vision':0, 'disorders_of_refraction_and_accommodation_blindness_and_low_vision':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3472], 'phecode_367'))
disorders_of_sweat_glands_phe = TensorMap('phecode_705', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_sweat_glands':0, 'disorders_of_sweat_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16129], 'phecode_705'))
disorders_of_the_autonomic_nervous_system_phe = TensorMap('phecode_337', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_the_autonomic_nervous_system':0, 'disorders_of_the_autonomic_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1851], 'phecode_337'))
disorders_of_the_globe_phe = TensorMap('phecode_360', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_the_globe':0, 'disorders_of_the_globe':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4504], 'phecode_360'))
disorders_of_the_orbit_phe = TensorMap('phecode_376', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_the_orbit':0, 'disorders_of_the_orbit':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4504], 'phecode_376'))
disorders_of_the_pituitary_gland_and_its_hypothalamic_control_phe = TensorMap('phecode_253', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_the_pituitary_gland_and_its_hypothalamic_control':0, 'disorders_of_the_pituitary_gland_and_its_hypothalamic_control':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3597], 'phecode_253'))
disorders_of_tooth_development_phe = TensorMap('phecode_520', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_tooth_development':0, 'disorders_of_tooth_development':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10416], 'phecode_520'))
disorders_of_urea_cycle_metabolism_phe = TensorMap('phecode_270.21', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_urea_cycle_metabolism':0, 'disorders_of_urea_cycle_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_270_21'))
disorders_of_uterus_nec_phe = TensorMap('phecode_619.2', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_uterus_nec':0, 'disorders_of_uterus_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 111], 'phecode_619_2'))
disorders_of_vitreous_body_phe = TensorMap('phecode_379.2', Interpretation.CATEGORICAL, channel_map={'no_disorders_of_vitreous_body':0, 'disorders_of_vitreous_body':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 191], 'phecode_379_2'))
disorders_resulting_from_impaired_renal_function_phe = TensorMap('phecode_588', Interpretation.CATEGORICAL, channel_map={'no_disorders_resulting_from_impaired_renal_function':0, 'disorders_resulting_from_impaired_renal_function':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3623], 'phecode_588'))
disorders_secondary_to_childbirth_surgery_trauma_phe = TensorMap('phecode_617', Interpretation.CATEGORICAL, channel_map={'no_disorders_secondary_to_childbirth_surgery_trauma':0, 'disorders_secondary_to_childbirth_surgery_trauma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1213], 'phecode_617'))
displacement_of_intervertebral_disc_phe = TensorMap('phecode_722.1', Interpretation.CATEGORICAL, channel_map={'no_displacement_of_intervertebral_disc':0, 'displacement_of_intervertebral_disc':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 967], 'phecode_722_1'))
dissociative_disorder_phe = TensorMap('phecode_303.1', Interpretation.CATEGORICAL, channel_map={'no_dissociative_disorder':0, 'dissociative_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7936], 'phecode_303_1'))
disturbance_of_salivary_secretion_phe = TensorMap('phecode_527.7', Interpretation.CATEGORICAL, channel_map={'no_disturbance_of_salivary_secretion':0, 'disturbance_of_salivary_secretion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4854], 'phecode_527_7'))
disturbance_of_skin_sensation_phe = TensorMap('phecode_687.4', Interpretation.CATEGORICAL, channel_map={'no_disturbance_of_skin_sensation':0, 'disturbance_of_skin_sensation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 48], 'phecode_687_4'))
disturbances_in_tooth_eruption_phe = TensorMap('phecode_520.2', Interpretation.CATEGORICAL, channel_map={'no_disturbances_in_tooth_eruption':0, 'disturbances_in_tooth_eruption':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 195], 'phecode_520_2'))
disturbances_of_amino_acid_transport_phe = TensorMap('phecode_270.1', Interpretation.CATEGORICAL, channel_map={'no_disturbances_of_amino_acid_transport':0, 'disturbances_of_amino_acid_transport':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11363], 'phecode_270_1'))
disturbances_of_sensation_of_smell_and_taste_phe = TensorMap('phecode_350.6', Interpretation.CATEGORICAL, channel_map={'no_disturbances_of_sensation_of_smell_and_taste':0, 'disturbances_of_sensation_of_smell_and_taste':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7246], 'phecode_350_6'))
disturbances_of_sulphur_bearing_amino_acid_metabolism_phe = TensorMap('phecode_270.11', Interpretation.CATEGORICAL, channel_map={'no_disturbances_of_sulphur_bearing_amino_acid_metabolism':0, 'disturbances_of_sulphur_bearing_amino_acid_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_270_11'))
diverticulosis_phe = TensorMap('phecode_562.1', Interpretation.CATEGORICAL, channel_map={'no_diverticulosis':0, 'diverticulosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_562_1'))
diverticulum_of_esophagus_acquired_phe = TensorMap('phecode_530.6', Interpretation.CATEGORICAL, channel_map={'no_diverticulum_of_esophagus_acquired':0, 'diverticulum_of_esophagus_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1984], 'phecode_530_6'))
dizziness_and_giddiness_light_headedness_and_vertigo_phe = TensorMap('phecode_386.9', Interpretation.CATEGORICAL, channel_map={'no_dizziness_and_giddiness_light_headedness_and_vertigo':0, 'dizziness_and_giddiness_light_headedness_and_vertigo':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 69], 'phecode_386_9'))
duodenal_ulcer_phe = TensorMap('phecode_531.3', Interpretation.CATEGORICAL, channel_map={'no_duodenal_ulcer':0, 'duodenal_ulcer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125], 'phecode_531_3'))
duodenitis_phe = TensorMap('phecode_535.6', Interpretation.CATEGORICAL, channel_map={'no_duodenitis':0, 'duodenitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 46], 'phecode_535_6'))
dyshidrosis_phe = TensorMap('phecode_705.1', Interpretation.CATEGORICAL, channel_map={'no_dyshidrosis':0, 'dyshidrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 31250], 'phecode_705_1'))
dysmenorrhea_phe = TensorMap('phecode_626.2', Interpretation.CATEGORICAL, channel_map={'no_dysmenorrhea':0, 'dysmenorrhea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 259], 'phecode_626_2'))
dyspareunia_phe = TensorMap('phecode_625.1', Interpretation.CATEGORICAL, channel_map={'no_dyspareunia':0, 'dyspareunia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 383], 'phecode_625_1'))
dysphagia_phe = TensorMap('phecode_532', Interpretation.CATEGORICAL, channel_map={'no_dysphagia':0, 'dysphagia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 50], 'phecode_532'))
dysplasia_of_female_genital_organs_phe = TensorMap('phecode_620', Interpretation.CATEGORICAL, channel_map={'no_dysplasia_of_female_genital_organs':0, 'dysplasia_of_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4504], 'phecode_620'))
dysthymic_disorder_phe = TensorMap('phecode_300.4', Interpretation.CATEGORICAL, channel_map={'no_dysthymic_disorder':0, 'dysthymic_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6756], 'phecode_300_4'))
dystrophy_of_female_genital_tract_phe = TensorMap('phecode_624.1', Interpretation.CATEGORICAL, channel_map={'no_dystrophy_of_female_genital_tract':0, 'dystrophy_of_female_genital_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1312], 'phecode_624_1'))
dysuria_phe = TensorMap('phecode_599.3', Interpretation.CATEGORICAL, channel_map={'no_dysuria':0, 'dysuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 297], 'phecode_599_3'))
e_coli_phe = TensorMap('phecode_41.4', Interpretation.CATEGORICAL, channel_map={'no_e_coli':0, 'e_coli':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 109], 'phecode_41_4'))
early_onset_of_delivery_phe = TensorMap('phecode_636.2', Interpretation.CATEGORICAL, channel_map={'no_early_onset_of_delivery':0, 'early_onset_of_delivery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 578], 'phecode_636_2'))
early_or_threatened_labor_hemorrhage_in_early_pregnancy_phe = TensorMap('phecode_636', Interpretation.CATEGORICAL, channel_map={'no_early_or_threatened_labor_hemorrhage_in_early_pregnancy':0, 'early_or_threatened_labor_hemorrhage_in_early_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 368], 'phecode_636'))
eating_disorder_phe = TensorMap('phecode_305.2', Interpretation.CATEGORICAL, channel_map={'no_eating_disorder':0, 'eating_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5813], 'phecode_305_2'))
ectopic_pregnancy_phe = TensorMap('phecode_634.3', Interpretation.CATEGORICAL, channel_map={'no_ectopic_pregnancy':0, 'ectopic_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 130], 'phecode_634_3'))
ectropion_or_entropion_phe = TensorMap('phecode_374.1', Interpretation.CATEGORICAL, channel_map={'no_ectropion_or_entropion':0, 'ectropion_or_entropion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 316], 'phecode_374_1'))
edema_phe = TensorMap('phecode_782.3', Interpretation.CATEGORICAL, channel_map={'no_edema':0, 'edema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 169], 'phecode_782_3'))
effects_of_heat_cold_and_air_pressure_phe = TensorMap('phecode_1014', Interpretation.CATEGORICAL, channel_map={'no_effects_of_heat_cold_and_air_pressure':0, 'effects_of_heat_cold_and_air_pressure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1106], 'phecode_1014'))
effects_of_other_external_causes_phe = TensorMap('phecode_1015', Interpretation.CATEGORICAL, channel_map={'no_effects_of_other_external_causes':0, 'effects_of_other_external_causes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 65], 'phecode_1015'))
effects_radiation_nos_phe = TensorMap('phecode_990', Interpretation.CATEGORICAL, channel_map={'no_effects_radiation_nos':0, 'effects_radiation_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 67], 'phecode_990'))
electrolyte_imbalance_phe = TensorMap('phecode_276.1', Interpretation.CATEGORICAL, channel_map={'no_electrolyte_imbalance':0, 'electrolyte_imbalance':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2083], 'phecode_276_1'))
elevated_blood_pressure_reading_without_diagnosis_of_hypertension_phe = TensorMap('phecode_402', Interpretation.CATEGORICAL, channel_map={'no_elevated_blood_pressure_reading_without_diagnosis_of_hypertension':0, 'elevated_blood_pressure_reading_without_diagnosis_of_hypertension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 226], 'phecode_402'))
elevated_sedimentation_rate_phe = TensorMap('phecode_790.1', Interpretation.CATEGORICAL, channel_map={'no_elevated_sedimentation_rate':0, 'elevated_sedimentation_rate':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4310], 'phecode_790_1'))
embolism_and_thrombosis_of_abdominal_aorta_phe = TensorMap('phecode_444.2', Interpretation.CATEGORICAL, channel_map={'no_embolism_and_thrombosis_of_abdominal_aorta':0, 'embolism_and_thrombosis_of_abdominal_aorta':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10869], 'phecode_444_2'))
emphysema_phe = TensorMap('phecode_496.1', Interpretation.CATEGORICAL, channel_map={'no_emphysema':0, 'emphysema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 180], 'phecode_496_1'))
empyema_and_pneumothorax_phe = TensorMap('phecode_506', Interpretation.CATEGORICAL, channel_map={'no_empyema_and_pneumothorax':0, 'empyema_and_pneumothorax':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 292], 'phecode_506'))
encephalitis_phe = TensorMap('phecode_323', Interpretation.CATEGORICAL, channel_map={'no_encephalitis':0, 'encephalitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 109], 'phecode_323'))
encephalitis_non_infectious_phe = TensorMap('phecode_323.8', Interpretation.CATEGORICAL, channel_map={'no_encephalitis_non_infectious':0, 'encephalitis_non_infectious':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 124], 'phecode_323_8'))
encephalopathy_not_elsewhere_classified_phe = TensorMap('phecode_348.8', Interpretation.CATEGORICAL, channel_map={'no_encephalopathy_not_elsewhere_classified':0, 'encephalopathy_not_elsewhere_classified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2232], 'phecode_348_8'))
endocarditis_phe = TensorMap('phecode_420.3', Interpretation.CATEGORICAL, channel_map={'no_endocarditis':0, 'endocarditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 467], 'phecode_420_3'))
endometrial_hyperplasia_phe = TensorMap('phecode_621', Interpretation.CATEGORICAL, channel_map={'no_endometrial_hyperplasia':0, 'endometrial_hyperplasia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 337], 'phecode_621'))
endometriosis_phe = TensorMap('phecode_615', Interpretation.CATEGORICAL, channel_map={'no_endometriosis':0, 'endometriosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100], 'phecode_615'))
enthesopathy_phe = TensorMap('phecode_726.1', Interpretation.CATEGORICAL, channel_map={'no_enthesopathy':0, 'enthesopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 51], 'phecode_726_1'))
eosinophilia_phe = TensorMap('phecode_288.3', Interpretation.CATEGORICAL, channel_map={'no_eosinophilia':0, 'eosinophilia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4672], 'phecode_288_3'))
epilepsy_phe = TensorMap('phecode_345.1', Interpretation.CATEGORICAL, channel_map={'no_epilepsy':0, 'epilepsy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1082], 'phecode_345_1'))
epilepsy_recurrent_seizures_convulsions_phe = TensorMap('phecode_345', Interpretation.CATEGORICAL, channel_map={'no_epilepsy_recurrent_seizures_convulsions':0, 'epilepsy_recurrent_seizures_convulsions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 106], 'phecode_345'))
epiphora_phe = TensorMap('phecode_375.2', Interpretation.CATEGORICAL, channel_map={'no_epiphora':0, 'epiphora':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 386], 'phecode_375_2'))
epistaxis_or_throat_hemorrhage_phe = TensorMap('phecode_477', Interpretation.CATEGORICAL, channel_map={'no_epistaxis_or_throat_hemorrhage':0, 'epistaxis_or_throat_hemorrhage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 144], 'phecode_477'))
erectile_dysfunction_ed_phe = TensorMap('phecode_605', Interpretation.CATEGORICAL, channel_map={'no_erectile_dysfunction_ed':0, 'erectile_dysfunction_ed':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 650], 'phecode_605'))
erythema_nodosum_phe = TensorMap('phecode_695.81', Interpretation.CATEGORICAL, channel_map={'no_erythema_nodosum':0, 'erythema_nodosum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4587], 'phecode_695_81'))
erythematosquamous_dermatosis_phe = TensorMap('phecode_690', Interpretation.CATEGORICAL, channel_map={'no_erythematosquamous_dermatosis':0, 'erythematosquamous_dermatosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13888], 'phecode_690'))
erythematous_conditions_phe = TensorMap('phecode_695', Interpretation.CATEGORICAL, channel_map={'no_erythematous_conditions':0, 'erythematous_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 318], 'phecode_695'))
esophageal_atresia_tracheoesophageal_fistula_phe = TensorMap('phecode_750.11', Interpretation.CATEGORICAL, channel_map={'no_esophageal_atresia_tracheoesophageal_fistula':0, 'esophageal_atresia_tracheoesophageal_fistula':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_750_11'))
esophageal_bleeding_varices_hemorrhage_phe = TensorMap('phecode_530.2', Interpretation.CATEGORICAL, channel_map={'no_esophageal_bleeding_varices_hemorrhage':0, 'esophageal_bleeding_varices_hemorrhage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 192], 'phecode_530_2'))
esophagitis_gerd_and_related_diseases_phe = TensorMap('phecode_530.1', Interpretation.CATEGORICAL, channel_map={'no_esophagitis_gerd_and_related_diseases':0, 'esophagitis_gerd_and_related_diseases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38], 'phecode_530_1'))
essential_hypertension_phe = TensorMap('phecode_401.1', Interpretation.CATEGORICAL, channel_map={'no_essential_hypertension':0, 'essential_hypertension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4], 'phecode_401_1'))
essential_tremor_phe = TensorMap('phecode_333.1', Interpretation.CATEGORICAL, channel_map={'no_essential_tremor':0, 'essential_tremor':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1253], 'phecode_333_1'))
eustachian_tube_disorders_phe = TensorMap('phecode_381.2', Interpretation.CATEGORICAL, channel_map={'no_eustachian_tube_disorders':0, 'eustachian_tube_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1915], 'phecode_381_2'))
excessive_or_frequent_menstruation_phe = TensorMap('phecode_626.12', Interpretation.CATEGORICAL, channel_map={'no_excessive_or_frequent_menstruation':0, 'excessive_or_frequent_menstruation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 44], 'phecode_626_12'))
excessive_vomiting_in_pregnancy_phe = TensorMap('phecode_643', Interpretation.CATEGORICAL, channel_map={'no_excessive_vomiting_in_pregnancy':0, 'excessive_vomiting_in_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3521], 'phecode_643'))
exophthalmos_phe = TensorMap('phecode_242.3', Interpretation.CATEGORICAL, channel_map={'no_exophthalmos':0, 'exophthalmos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2824], 'phecode_242_3'))
extrapyramidal_disease_and_abnormal_movement_disorders_phe = TensorMap('phecode_333', Interpretation.CATEGORICAL, channel_map={'no_extrapyramidal_disease_and_abnormal_movement_disorders':0, 'extrapyramidal_disease_and_abnormal_movement_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 149], 'phecode_333'))
extrinsic_allergic_alveolitis_phe = TensorMap('phecode_500.1', Interpretation.CATEGORICAL, channel_map={'no_extrinsic_allergic_alveolitis':0, 'extrinsic_allergic_alveolitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4065], 'phecode_500_1'))
eye_infection_viral_phe = TensorMap('phecode_369.2', Interpretation.CATEGORICAL, channel_map={'no_eye_infection_viral':0, 'eye_infection_viral':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2577], 'phecode_369_2'))
facial_nerve_disorders_cn7_phe = TensorMap('phecode_352.2', Interpretation.CATEGORICAL, channel_map={'no_facial_nerve_disorders_cn7':0, 'facial_nerve_disorders_cn7':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 396], 'phecode_352_2'))
family_history_phe = TensorMap('phecode_1100', Interpretation.CATEGORICAL, channel_map={'no_family_history':0, 'family_history':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 141], 'phecode_1100'))
fasciitis_phe = TensorMap('phecode_728.7', Interpretation.CATEGORICAL, channel_map={'no_fasciitis':0, 'fasciitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1404], 'phecode_728_7'))
femoral_hernia_phe = TensorMap('phecode_550.3', Interpretation.CATEGORICAL, channel_map={'no_femoral_hernia':0, 'femoral_hernia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 570], 'phecode_550_3'))
fetal_distress_and_abnormal_forces_of_labor_phe = TensorMap('phecode_661', Interpretation.CATEGORICAL, channel_map={'no_fetal_distress_and_abnormal_forces_of_labor':0, 'fetal_distress_and_abnormal_forces_of_labor':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 173], 'phecode_661'))
fever_of_unknown_origin_phe = TensorMap('phecode_783', Interpretation.CATEGORICAL, channel_map={'no_fever_of_unknown_origin':0, 'fever_of_unknown_origin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83], 'phecode_783'))
fibroadenosis_of_breast_phe = TensorMap('phecode_610.2', Interpretation.CATEGORICAL, channel_map={'no_fibroadenosis_of_breast':0, 'fibroadenosis_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1385], 'phecode_610_2'))
fibrosclerosis_of_breast_phe = TensorMap('phecode_610.3', Interpretation.CATEGORICAL, channel_map={'no_fibrosclerosis_of_breast':0, 'fibrosclerosis_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1369], 'phecode_610_3'))
first_degree_av_block_phe = TensorMap('phecode_426.21', Interpretation.CATEGORICAL, channel_map={'no_first_degree_av_block':0, 'first_degree_av_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 281], 'phecode_426_21'))
flat_foot_phe = TensorMap('phecode_735.1', Interpretation.CATEGORICAL, channel_map={'no_flat_foot':0, 'flat_foot':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2145], 'phecode_735_1'))
flatulence_phe = TensorMap('phecode_561.2', Interpretation.CATEGORICAL, channel_map={'no_flatulence':0, 'flatulence':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 188], 'phecode_561_2'))
fluid_overload_phe = TensorMap('phecode_276.6', Interpretation.CATEGORICAL, channel_map={'no_fluid_overload':0, 'fluid_overload':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 506], 'phecode_276_6'))
folate_deficiency_anemia_phe = TensorMap('phecode_281.13', Interpretation.CATEGORICAL, channel_map={'no_folate_deficiency_anemia':0, 'folate_deficiency_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1612], 'phecode_281_13'))
foreign_body_injury_phe = TensorMap('phecode_1001', Interpretation.CATEGORICAL, channel_map={'no_foreign_body_injury':0, 'foreign_body_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 372], 'phecode_1001'))
fracture_of_ankle_and_foot_phe = TensorMap('phecode_801', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_ankle_and_foot':0, 'fracture_of_ankle_and_foot':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1082], 'phecode_801'))
fracture_of_clavicle_or_scapula_phe = TensorMap('phecode_803.3', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_clavicle_or_scapula':0, 'fracture_of_clavicle_or_scapula':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1470], 'phecode_803_3'))
fracture_of_foot_phe = TensorMap('phecode_801.1', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_foot':0, 'fracture_of_foot':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17857], 'phecode_801_1'))
fracture_of_hand_or_wrist_phe = TensorMap('phecode_804', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_hand_or_wrist':0, 'fracture_of_hand_or_wrist':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 693], 'phecode_804'))
fracture_of_humerus_phe = TensorMap('phecode_803.1', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_humerus':0, 'fracture_of_humerus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1612], 'phecode_803_1'))
fracture_of_lower_limb_phe = TensorMap('phecode_800', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_lower_limb':0, 'fracture_of_lower_limb':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 863], 'phecode_800'))
fracture_of_neck_of_femur_phe = TensorMap('phecode_800.1', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_neck_of_femur':0, 'fracture_of_neck_of_femur':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1760], 'phecode_800_1'))
fracture_of_patella_phe = TensorMap('phecode_800.4', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_patella':0, 'fracture_of_patella':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3048], 'phecode_800_4'))
fracture_of_pelvis_phe = TensorMap('phecode_802', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_pelvis':0, 'fracture_of_pelvis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3472], 'phecode_802'))
fracture_of_radius_and_ulna_phe = TensorMap('phecode_803.2', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_radius_and_ulna':0, 'fracture_of_radius_and_ulna':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 510], 'phecode_803_2'))
fracture_of_ribs_phe = TensorMap('phecode_807', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_ribs':0, 'fracture_of_ribs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1344], 'phecode_807'))
fracture_of_tibia_and_fibula_phe = TensorMap('phecode_800.3', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_tibia_and_fibula':0, 'fracture_of_tibia_and_fibula':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1039], 'phecode_800_3'))
fracture_of_unspecified_bones_phe = TensorMap('phecode_809', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_unspecified_bones':0, 'fracture_of_unspecified_bones':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3105], 'phecode_809'))
fracture_of_unspecified_part_of_femur_phe = TensorMap('phecode_800.2', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_unspecified_part_of_femur':0, 'fracture_of_unspecified_part_of_femur':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3787], 'phecode_800_2'))
fracture_of_upper_limb_phe = TensorMap('phecode_803', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_upper_limb':0, 'fracture_of_upper_limb':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1677], 'phecode_803'))
fracture_of_vertebral_column_without_mention_of_spinal_cord_injury_phe = TensorMap('phecode_805', Interpretation.CATEGORICAL, channel_map={'no_fracture_of_vertebral_column_without_mention_of_spinal_cord_injury':0, 'fracture_of_vertebral_column_without_mention_of_spinal_cord_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1041], 'phecode_805'))
frequency_of_urination_and_polyuria_phe = TensorMap('phecode_599.5', Interpretation.CATEGORICAL, channel_map={'no_frequency_of_urination_and_polyuria':0, 'frequency_of_urination_and_polyuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 94], 'phecode_599_5'))
functional_digestive_disorders_phe = TensorMap('phecode_564', Interpretation.CATEGORICAL, channel_map={'no_functional_digestive_disorders':0, 'functional_digestive_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 419], 'phecode_564'))
functional_disorders_of_bladder_phe = TensorMap('phecode_596.5', Interpretation.CATEGORICAL, channel_map={'no_functional_disorders_of_bladder':0, 'functional_disorders_of_bladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 239], 'phecode_596_5'))
galactorrhea_phe = TensorMap('phecode_612.1', Interpretation.CATEGORICAL, channel_map={'no_galactorrhea':0, 'galactorrhea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35714], 'phecode_612_1'))
ganglion_and_cyst_of_synovium_tendon_and_bursa_phe = TensorMap('phecode_727.4', Interpretation.CATEGORICAL, channel_map={'no_ganglion_and_cyst_of_synovium_tendon_and_bursa':0, 'ganglion_and_cyst_of_synovium_tendon_and_bursa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 128], 'phecode_727_4'))
gangrene_phe = TensorMap('phecode_791', Interpretation.CATEGORICAL, channel_map={'no_gangrene':0, 'gangrene':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 517], 'phecode_791'))
gastric_ulcer_phe = TensorMap('phecode_531.2', Interpretation.CATEGORICAL, channel_map={'no_gastric_ulcer':0, 'gastric_ulcer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 82], 'phecode_531_2'))
gastritis_and_duodenitis_phe = TensorMap('phecode_535', Interpretation.CATEGORICAL, channel_map={'no_gastritis_and_duodenitis':0, 'gastritis_and_duodenitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19], 'phecode_535'))
gastroesophageal_laceration_hemorrhage_syndrome_phe = TensorMap('phecode_530.7', Interpretation.CATEGORICAL, channel_map={'no_gastroesophageal_laceration_hemorrhage_syndrome':0, 'gastroesophageal_laceration_hemorrhage_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1126], 'phecode_530_7'))
gastrointestinal_complications_phe = TensorMap('phecode_569.2', Interpretation.CATEGORICAL, channel_map={'no_gastrointestinal_complications':0, 'gastrointestinal_complications':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 689], 'phecode_569_2'))
gastrojejunal_ulcer_phe = TensorMap('phecode_531.5', Interpretation.CATEGORICAL, channel_map={'no_gastrojejunal_ulcer':0, 'gastrojejunal_ulcer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7812], 'phecode_531_5'))
generalized_anxiety_disorder_phe = TensorMap('phecode_300.11', Interpretation.CATEGORICAL, channel_map={'no_generalized_anxiety_disorder':0, 'generalized_anxiety_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2222], 'phecode_300_11'))
generalized_convulsive_epilepsy_phe = TensorMap('phecode_345.11', Interpretation.CATEGORICAL, channel_map={'no_generalized_convulsive_epilepsy':0, 'generalized_convulsive_epilepsy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 811], 'phecode_345_11'))
genital_prolapse_phe = TensorMap('phecode_618', Interpretation.CATEGORICAL, channel_map={'no_genital_prolapse':0, 'genital_prolapse':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 883], 'phecode_618'))
gerd_phe = TensorMap('phecode_530.11', Interpretation.CATEGORICAL, channel_map={'no_gerd':0, 'gerd':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21], 'phecode_530_11'))
giant_cell_arteritis_phe = TensorMap('phecode_446.5', Interpretation.CATEGORICAL, channel_map={'no_giant_cell_arteritis':0, 'giant_cell_arteritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 810], 'phecode_446_5'))
gingival_and_periodontal_diseases_phe = TensorMap('phecode_523', Interpretation.CATEGORICAL, channel_map={'no_gingival_and_periodontal_diseases':0, 'gingival_and_periodontal_diseases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 740], 'phecode_523'))
gingivitis_phe = TensorMap('phecode_523.1', Interpretation.CATEGORICAL, channel_map={'no_gingivitis':0, 'gingivitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3125], 'phecode_523_1'))
glaucoma_phe = TensorMap('phecode_365', Interpretation.CATEGORICAL, channel_map={'no_glaucoma':0, 'glaucoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 66], 'phecode_365'))
glossitis_phe = TensorMap('phecode_529.1', Interpretation.CATEGORICAL, channel_map={'no_glossitis':0, 'glossitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1136], 'phecode_529_1'))
glossodynia_phe = TensorMap('phecode_529.6', Interpretation.CATEGORICAL, channel_map={'no_glossodynia':0, 'glossodynia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12820], 'phecode_529_6'))
glucocorticoid_deficiency_phe = TensorMap('phecode_255.21', Interpretation.CATEGORICAL, channel_map={'no_glucocorticoid_deficiency':0, 'glucocorticoid_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 907], 'phecode_255_21'))
glycosuria_or_acetonuria_phe = TensorMap('phecode_250.5', Interpretation.CATEGORICAL, channel_map={'no_glycosuria_or_acetonuria':0, 'glycosuria_or_acetonuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2976], 'phecode_250_5'))
gonococcal_infections_phe = TensorMap('phecode_90.2', Interpretation.CATEGORICAL, channel_map={'no_gonococcal_infections':0, 'gonococcal_infections':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55555], 'phecode_90_2'))
gout_phe = TensorMap('phecode_274.1', Interpretation.CATEGORICAL, channel_map={'no_gout':0, 'gout':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 155], 'phecode_274_1'))
gouty_arthropathy_phe = TensorMap('phecode_274.11', Interpretation.CATEGORICAL, channel_map={'no_gouty_arthropathy':0, 'gouty_arthropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11904], 'phecode_274_11'))
gram_negative_septicemia_phe = TensorMap('phecode_38.1', Interpretation.CATEGORICAL, channel_map={'no_gram_negative_septicemia':0, 'gram_negative_septicemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 358], 'phecode_38_1'))
gram_positive_septicemia_phe = TensorMap('phecode_38.2', Interpretation.CATEGORICAL, channel_map={'no_gram_positive_septicemia':0, 'gram_positive_septicemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 664], 'phecode_38_2'))
graves_disease_phe = TensorMap('phecode_242.1', Interpretation.CATEGORICAL, channel_map={'no_graves_disease':0, 'graves_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 732], 'phecode_242_1'))
hallucinations_phe = TensorMap('phecode_292.6', Interpretation.CATEGORICAL, channel_map={'no_hallucinations':0, 'hallucinations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 927], 'phecode_292_6'))
hallux_rigidus_phe = TensorMap('phecode_735.23', Interpretation.CATEGORICAL, channel_map={'no_hallux_rigidus':0, 'hallux_rigidus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 246], 'phecode_735_23'))
hallux_valgus_bunion_phe = TensorMap('phecode_735.3', Interpretation.CATEGORICAL, channel_map={'no_hallux_valgus_bunion':0, 'hallux_valgus_bunion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55], 'phecode_735_3'))
hammer_toe_acquired_phe = TensorMap('phecode_735.21', Interpretation.CATEGORICAL, channel_map={'no_hammer_toe_acquired':0, 'hammer_toe_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 187], 'phecode_735_21'))
hearing_loss_phe = TensorMap('phecode_389', Interpretation.CATEGORICAL, channel_map={'no_hearing_loss':0, 'hearing_loss':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 79], 'phecode_389'))
heart_failure_nos_phe = TensorMap('phecode_428.2', Interpretation.CATEGORICAL, channel_map={'no_heart_failure_nos':0, 'heart_failure_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 70], 'phecode_428_2'))
heart_transplant_surgery_phe = TensorMap('phecode_429.1', Interpretation.CATEGORICAL, channel_map={'no_heart_transplant_surgery':0, 'heart_transplant_surgery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6097], 'phecode_429_1'))
heart_valve_replaced_phe = TensorMap('phecode_395.6', Interpretation.CATEGORICAL, channel_map={'no_heart_valve_replaced':0, 'heart_valve_replaced':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 216], 'phecode_395_6'))
heartburn_phe = TensorMap('phecode_530.9', Interpretation.CATEGORICAL, channel_map={'no_heartburn':0, 'heartburn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 169], 'phecode_530_9'))
helminthiases_phe = TensorMap('phecode_134', Interpretation.CATEGORICAL, channel_map={'no_helminthiases':0, 'helminthiases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2415], 'phecode_134'))
hemangioma_and_lymphangioma_any_site_phe = TensorMap('phecode_228', Interpretation.CATEGORICAL, channel_map={'no_hemangioma_and_lymphangioma_any_site':0, 'hemangioma_and_lymphangioma_any_site':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 232], 'phecode_228'))
hemarthrosis_phe = TensorMap('phecode_741.5', Interpretation.CATEGORICAL, channel_map={'no_hemarthrosis':0, 'hemarthrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25000], 'phecode_741_5'))
hematemesis_phe = TensorMap('phecode_578.1', Interpretation.CATEGORICAL, channel_map={'no_hematemesis':0, 'hematemesis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 183], 'phecode_578_1'))
hematological_disorders_of_newborn_phe = TensorMap('phecode_656.5', Interpretation.CATEGORICAL, channel_map={'no_hematological_disorders_of_newborn':0, 'hematological_disorders_of_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_656_5'))
hematuria_phe = TensorMap('phecode_593', Interpretation.CATEGORICAL, channel_map={'no_hematuria':0, 'hematuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23], 'phecode_593'))
hemiplegia_phe = TensorMap('phecode_342', Interpretation.CATEGORICAL, channel_map={'no_hemiplegia':0, 'hemiplegia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 204], 'phecode_342'))
hemoglobinuria_phe = TensorMap('phecode_285.8', Interpretation.CATEGORICAL, channel_map={'no_hemoglobinuria':0, 'hemoglobinuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21739], 'phecode_285_8'))
hemolytic_uremic_syndrome_phe = TensorMap('phecode_283.21', Interpretation.CATEGORICAL, channel_map={'no_hemolytic_uremic_syndrome':0, 'hemolytic_uremic_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_283_21'))
hemoptysis_phe = TensorMap('phecode_516.1', Interpretation.CATEGORICAL, channel_map={'no_hemoptysis':0, 'hemoptysis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166], 'phecode_516_1'))
hemorrhage_during_pregnancy_childbirth_and_postpartum_phe = TensorMap('phecode_635', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_during_pregnancy_childbirth_and_postpartum':0, 'hemorrhage_during_pregnancy_childbirth_and_postpartum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83333], 'phecode_635'))
hemorrhage_from_gastrointestinal_ulcer_phe = TensorMap('phecode_531.1', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_from_gastrointestinal_ulcer':0, 'hemorrhage_from_gastrointestinal_ulcer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 547], 'phecode_531_1'))
hemorrhage_in_early_pregnancy_phe = TensorMap('phecode_636.3', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_in_early_pregnancy':0, 'hemorrhage_in_early_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 123], 'phecode_636_3'))
hemorrhage_nos_phe = TensorMap('phecode_459.1', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_nos':0, 'hemorrhage_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3731], 'phecode_459_1'))
hemorrhage_of_gastrointestinal_tract_phe = TensorMap('phecode_578.9', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_of_gastrointestinal_tract':0, 'hemorrhage_of_gastrointestinal_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55], 'phecode_578_9'))
hemorrhage_of_rectum_and_anus_phe = TensorMap('phecode_578.8', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_of_rectum_and_anus':0, 'hemorrhage_of_rectum_and_anus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 28], 'phecode_578_8'))
hemorrhage_or_hematoma_complicating_a_procedure_phe = TensorMap('phecode_850', Interpretation.CATEGORICAL, channel_map={'no_hemorrhage_or_hematoma_complicating_a_procedure':0, 'hemorrhage_or_hematoma_complicating_a_procedure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 66], 'phecode_850'))
hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants_phe = TensorMap('phecode_286.5', Interpretation.CATEGORICAL, channel_map={'no_hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants':0, 'hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4310], 'phecode_286_5'))
hemorrhoids_phe = TensorMap('phecode_455', Interpretation.CATEGORICAL, channel_map={'no_hemorrhoids':0, 'hemorrhoids':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15], 'phecode_455'))
hepatitis_nos_phe = TensorMap('phecode_70.9', Interpretation.CATEGORICAL, channel_map={'no_hepatitis_nos':0, 'hepatitis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 706], 'phecode_70_9'))
hepatomegaly_phe = TensorMap('phecode_573.3', Interpretation.CATEGORICAL, channel_map={'no_hepatomegaly':0, 'hepatomegaly':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1084], 'phecode_573_3'))
hereditary_and_idiopathic_peripheral_neuropathy_phe = TensorMap('phecode_356', Interpretation.CATEGORICAL, channel_map={'no_hereditary_and_idiopathic_peripheral_neuropathy':0, 'hereditary_and_idiopathic_peripheral_neuropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2304], 'phecode_356'))
hereditary_disturbances_in_tooth_structure_phe = TensorMap('phecode_520.1', Interpretation.CATEGORICAL, channel_map={'no_hereditary_disturbances_in_tooth_structure':0, 'hereditary_disturbances_in_tooth_structure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8620], 'phecode_520_1'))
hereditary_retinal_dystrophies_phe = TensorMap('phecode_362.7', Interpretation.CATEGORICAL, channel_map={'no_hereditary_retinal_dystrophies':0, 'hereditary_retinal_dystrophies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4000], 'phecode_362_7'))
herpes_simplex_phe = TensorMap('phecode_54', Interpretation.CATEGORICAL, channel_map={'no_herpes_simplex':0, 'herpes_simplex':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1457], 'phecode_54'))
herpes_zoster_phe = TensorMap('phecode_53', Interpretation.CATEGORICAL, channel_map={'no_herpes_zoster':0, 'herpes_zoster':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 943], 'phecode_53'))
herpes_zoster_with_nervous_system_complications_phe = TensorMap('phecode_53.1', Interpretation.CATEGORICAL, channel_map={'no_herpes_zoster_with_nervous_system_complications':0, 'herpes_zoster_with_nervous_system_complications':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3289], 'phecode_53_1'))
hidradenitis_phe = TensorMap('phecode_705.3', Interpretation.CATEGORICAL, channel_map={'no_hidradenitis':0, 'hidradenitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2016], 'phecode_705_3'))
hirsutism_phe = TensorMap('phecode_704.2', Interpretation.CATEGORICAL, channel_map={'no_hirsutism':0, 'hirsutism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12820], 'phecode_704_2'))
histoplasmosis_phe = TensorMap('phecode_117.1', Interpretation.CATEGORICAL, channel_map={'no_histoplasmosis':0, 'histoplasmosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_117_1'))
hiv_infection_symptomatic_phe = TensorMap('phecode_71.1', Interpretation.CATEGORICAL, channel_map={'no_hiv_infection_symptomatic':0, 'hiv_infection_symptomatic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2336], 'phecode_71_1'))
hodgkins_disease_phe = TensorMap('phecode_201', Interpretation.CATEGORICAL, channel_map={'no_hodgkins_disease':0, 'hodgkins_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1369], 'phecode_201'))
hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use_phe = TensorMap('phecode_962.3', Interpretation.CATEGORICAL, channel_map={'no_hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use':0, 'hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 941], 'phecode_962_3'))
human_immunodeficiency_virus_hiv_disease_phe = TensorMap('phecode_71', Interpretation.CATEGORICAL, channel_map={'no_human_immunodeficiency_virus_hiv_disease':0, 'human_immunodeficiency_virus_hiv_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4273], 'phecode_71'))
hydrocele_phe = TensorMap('phecode_603.1', Interpretation.CATEGORICAL, channel_map={'no_hydrocele':0, 'hydrocele':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 268], 'phecode_603_1'))
hydrocephalus_phe = TensorMap('phecode_331.1', Interpretation.CATEGORICAL, channel_map={'no_hydrocephalus':0, 'hydrocephalus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 653], 'phecode_331_1'))
hydronephrosis_phe = TensorMap('phecode_595', Interpretation.CATEGORICAL, channel_map={'no_hydronephrosis':0, 'hydronephrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 162], 'phecode_595'))
hyperaldosteronism_phe = TensorMap('phecode_255.12', Interpretation.CATEGORICAL, channel_map={'no_hyperaldosteronism':0, 'hyperaldosteronism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4201], 'phecode_255_12'))
hypercholesterolemia_phe = TensorMap('phecode_272.11', Interpretation.CATEGORICAL, channel_map={'no_hypercholesterolemia':0, 'hypercholesterolemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10], 'phecode_272_11'))
hyperchylomicronemia_phe = TensorMap('phecode_272.14', Interpretation.CATEGORICAL, channel_map={'no_hyperchylomicronemia':0, 'hyperchylomicronemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_272_14'))
hyperemesis_gravidarum_phe = TensorMap('phecode_643.1', Interpretation.CATEGORICAL, channel_map={'no_hyperemesis_gravidarum':0, 'hyperemesis_gravidarum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2500], 'phecode_643_1'))
hyperglyceridemia_phe = TensorMap('phecode_272.12', Interpretation.CATEGORICAL, channel_map={'no_hyperglyceridemia':0, 'hyperglyceridemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5000], 'phecode_272_12'))
hyperhidrosis_phe = TensorMap('phecode_705.8', Interpretation.CATEGORICAL, channel_map={'no_hyperhidrosis':0, 'hyperhidrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 554], 'phecode_705_8'))
hyperlipidemia_phe = TensorMap('phecode_272.1', Interpretation.CATEGORICAL, channel_map={'no_hyperlipidemia':0, 'hyperlipidemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 63], 'phecode_272_1'))
hypermetropia_phe = TensorMap('phecode_367.8', Interpretation.CATEGORICAL, channel_map={'no_hypermetropia':0, 'hypermetropia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1256], 'phecode_367_8'))
hyperosmolality_and_or_hypernatremia_phe = TensorMap('phecode_276.11', Interpretation.CATEGORICAL, channel_map={'no_hyperosmolality_and_or_hypernatremia':0, 'hyperosmolality_and_or_hypernatremia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 996], 'phecode_276_11'))
hyperparathyroidism_phe = TensorMap('phecode_252.1', Interpretation.CATEGORICAL, channel_map={'no_hyperparathyroidism':0, 'hyperparathyroidism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 403], 'phecode_252_1'))
hyperplasia_of_prostate_phe = TensorMap('phecode_600', Interpretation.CATEGORICAL, channel_map={'no_hyperplasia_of_prostate':0, 'hyperplasia_of_prostate':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 30], 'phecode_600'))
hyperpotassemia_phe = TensorMap('phecode_276.13', Interpretation.CATEGORICAL, channel_map={'no_hyperpotassemia':0, 'hyperpotassemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 280], 'phecode_276_13'))
hypersensitivity_angiitis_phe = TensorMap('phecode_446.3', Interpretation.CATEGORICAL, channel_map={'no_hypersensitivity_angiitis':0, 'hypersensitivity_angiitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12195], 'phecode_446_3'))
hypersomnia_phe = TensorMap('phecode_327.1', Interpretation.CATEGORICAL, channel_map={'no_hypersomnia':0, 'hypersomnia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_327_1'))
hypertension_phe = TensorMap('phecode_401', Interpretation.CATEGORICAL, channel_map={'no_hypertension':0, 'hypertension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4], 'phecode_401'))
hypertension_complicating_pregnancy_childbirth_and_the_puerperium_phe = TensorMap('phecode_642', Interpretation.CATEGORICAL, channel_map={'no_hypertension_complicating_pregnancy_childbirth_and_the_puerperium':0, 'hypertension_complicating_pregnancy_childbirth_and_the_puerperium':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 291], 'phecode_642'))
hypertensive_chronic_kidney_disease_phe = TensorMap('phecode_401.22', Interpretation.CATEGORICAL, channel_map={'no_hypertensive_chronic_kidney_disease':0, 'hypertensive_chronic_kidney_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 241], 'phecode_401_22'))
hypertensive_heart_and_or_renal_disease_phe = TensorMap('phecode_401.2', Interpretation.CATEGORICAL, channel_map={'no_hypertensive_heart_and_or_renal_disease':0, 'hypertensive_heart_and_or_renal_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8620], 'phecode_401_2'))
hypertensive_heart_disease_phe = TensorMap('phecode_401.21', Interpretation.CATEGORICAL, channel_map={'no_hypertensive_heart_disease':0, 'hypertensive_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1824], 'phecode_401_21'))
hypertrophic_obstructive_cardiomyopathy_phe = TensorMap('phecode_425.11', Interpretation.CATEGORICAL, channel_map={'no_hypertrophic_obstructive_cardiomyopathy':0, 'hypertrophic_obstructive_cardiomyopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2325], 'phecode_425_11'))
hypertrophy_of_breast_gynecomastia_phe = TensorMap('phecode_612.2', Interpretation.CATEGORICAL, channel_map={'no_hypertrophy_of_breast_gynecomastia':0, 'hypertrophy_of_breast_gynecomastia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 472], 'phecode_612_2'))
hypertrophy_of_female_genital_organs_phe = TensorMap('phecode_623', Interpretation.CATEGORICAL, channel_map={'no_hypertrophy_of_female_genital_organs':0, 'hypertrophy_of_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 303], 'phecode_623'))
hypertrophy_of_salivary_gland_phe = TensorMap('phecode_527.1', Interpretation.CATEGORICAL, channel_map={'no_hypertrophy_of_salivary_gland':0, 'hypertrophy_of_salivary_gland':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4950], 'phecode_527_1'))
hyperventilation_phe = TensorMap('phecode_513.4', Interpretation.CATEGORICAL, channel_map={'no_hyperventilation':0, 'hyperventilation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1369], 'phecode_513_4'))
hypoglycemia_phe = TensorMap('phecode_251.1', Interpretation.CATEGORICAL, channel_map={'no_hypoglycemia':0, 'hypoglycemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 304], 'phecode_251_1'))
hypoparathyroidism_phe = TensorMap('phecode_252.2', Interpretation.CATEGORICAL, channel_map={'no_hypoparathyroidism':0, 'hypoparathyroidism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2000], 'phecode_252_2'))
hypopotassemia_phe = TensorMap('phecode_276.14', Interpretation.CATEGORICAL, channel_map={'no_hypopotassemia':0, 'hypopotassemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 183], 'phecode_276_14'))
hyposmolality_and_or_hyponatremia_phe = TensorMap('phecode_276.12', Interpretation.CATEGORICAL, channel_map={'no_hyposmolality_and_or_hyponatremia':0, 'hyposmolality_and_or_hyponatremia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 143], 'phecode_276_12'))
hypotension_phe = TensorMap('phecode_458', Interpretation.CATEGORICAL, channel_map={'no_hypotension':0, 'hypotension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 287], 'phecode_458'))
hypotension_nos_phe = TensorMap('phecode_458.9', Interpretation.CATEGORICAL, channel_map={'no_hypotension_nos':0, 'hypotension_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 82], 'phecode_458_9'))
hypothermia_chills_phe = TensorMap('phecode_780', Interpretation.CATEGORICAL, channel_map={'no_hypothermia_chills':0, 'hypothermia_chills':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1602], 'phecode_780'))
hypothyroidism_nos_phe = TensorMap('phecode_244.4', Interpretation.CATEGORICAL, channel_map={'no_hypothyroidism_nos':0, 'hypothyroidism_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23], 'phecode_244_4'))
hypotony_of_eye_phe = TensorMap('phecode_360.3', Interpretation.CATEGORICAL, channel_map={'no_hypotony_of_eye':0, 'hypotony_of_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11363], 'phecode_360_3'))
hypoventilation_phe = TensorMap('phecode_513.3', Interpretation.CATEGORICAL, channel_map={'no_hypoventilation':0, 'hypoventilation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3105], 'phecode_513_3'))
hypovolemia_phe = TensorMap('phecode_276.5', Interpretation.CATEGORICAL, channel_map={'no_hypovolemia':0, 'hypovolemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100], 'phecode_276_5'))
iatrogenic_endocrine_disorders_phe = TensorMap('phecode_258', Interpretation.CATEGORICAL, channel_map={'no_iatrogenic_endocrine_disorders':0, 'iatrogenic_endocrine_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3225], 'phecode_258'))
iatrogenic_hypotension_phe = TensorMap('phecode_458.2', Interpretation.CATEGORICAL, channel_map={'no_iatrogenic_hypotension':0, 'iatrogenic_hypotension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1377], 'phecode_458_2'))
ichthyosis_congenita_phe = TensorMap('phecode_691.1', Interpretation.CATEGORICAL, channel_map={'no_ichthyosis_congenita':0, 'ichthyosis_congenita':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29411], 'phecode_691_1'))
ileostomy_status_phe = TensorMap('phecode_559', Interpretation.CATEGORICAL, channel_map={'no_ileostomy_status':0, 'ileostomy_status':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 215], 'phecode_559'))
immunity_deficiency_phe = TensorMap('phecode_279.1', Interpretation.CATEGORICAL, channel_map={'no_immunity_deficiency':0, 'immunity_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3597], 'phecode_279_1'))
impacted_cerumen_phe = TensorMap('phecode_380.4', Interpretation.CATEGORICAL, channel_map={'no_impacted_cerumen':0, 'impacted_cerumen':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1072], 'phecode_380_4'))
impaction_of_intestine_phe = TensorMap('phecode_560.2', Interpretation.CATEGORICAL, channel_map={'no_impaction_of_intestine':0, 'impaction_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1162], 'phecode_560_2'))
impaired_fasting_glucose_phe = TensorMap('phecode_250.41', Interpretation.CATEGORICAL, channel_map={'no_impaired_fasting_glucose':0, 'impaired_fasting_glucose':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1054], 'phecode_250_41'))
impetigo_phe = TensorMap('phecode_686.2', Interpretation.CATEGORICAL, channel_map={'no_impetigo':0, 'impetigo':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10204], 'phecode_686_2'))
impulse_control_disorder_phe = TensorMap('phecode_312.3', Interpretation.CATEGORICAL, channel_map={'no_impulse_control_disorder':0, 'impulse_control_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38461], 'phecode_312_3'))
infantile_cerebral_palsy_phe = TensorMap('phecode_343', Interpretation.CATEGORICAL, channel_map={'no_infantile_cerebral_palsy':0, 'infantile_cerebral_palsy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2475], 'phecode_343'))
infection_inflammation_of_internal_prosthetic_device_implant_and_graft_phe = TensorMap('phecode_81', Interpretation.CATEGORICAL, channel_map={'no_infection_inflammation_of_internal_prosthetic_device_implant_and_graft':0, 'infection_inflammation_of_internal_prosthetic_device_implant_and_graft':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 288], 'phecode_81'))
infection_of_the_eye_phe = TensorMap('phecode_369', Interpretation.CATEGORICAL, channel_map={'no_infection_of_the_eye':0, 'infection_of_the_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1319], 'phecode_369'))
infections_of_genitourinary_tract_during_pregnancy_phe = TensorMap('phecode_647.1', Interpretation.CATEGORICAL, channel_map={'no_infections_of_genitourinary_tract_during_pregnancy':0, 'infections_of_genitourinary_tract_during_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1319], 'phecode_647_1'))
infections_specific_to_the_perinatal_period_phe = TensorMap('phecode_657', Interpretation.CATEGORICAL, channel_map={'no_infections_specific_to_the_perinatal_period':0, 'infections_specific_to_the_perinatal_period':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83333], 'phecode_657'))
infectious_and_parasitic_complications_affecting_pregnancy_phe = TensorMap('phecode_647', Interpretation.CATEGORICAL, channel_map={'no_infectious_and_parasitic_complications_affecting_pregnancy':0, 'infectious_and_parasitic_complications_affecting_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4000], 'phecode_647'))
infectious_mononucleosis_phe = TensorMap('phecode_79.2', Interpretation.CATEGORICAL, channel_map={'no_infectious_mononucleosis':0, 'infectious_mononucleosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6410], 'phecode_79_2'))
infective_connective_tissue_disorders_phe = TensorMap('phecode_712', Interpretation.CATEGORICAL, channel_map={'no_infective_connective_tissue_disorders':0, 'infective_connective_tissue_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 14705], 'phecode_712'))
infertility_female_phe = TensorMap('phecode_626.8', Interpretation.CATEGORICAL, channel_map={'no_infertility_female':0, 'infertility_female':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 378], 'phecode_626_8'))
infertility_female_associated_with_anovulation_phe = TensorMap('phecode_626.15', Interpretation.CATEGORICAL, channel_map={'no_infertility_female_associated_with_anovulation':0, 'infertility_female_associated_with_anovulation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38461], 'phecode_626_15'))
infertility_male_phe = TensorMap('phecode_609.1', Interpretation.CATEGORICAL, channel_map={'no_infertility_male':0, 'infertility_male':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6578], 'phecode_609_1'))
infestation_lice_mites_phe = TensorMap('phecode_132', Interpretation.CATEGORICAL, channel_map={'no_infestation_lice_mites':0, 'infestation_lice_mites':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11627], 'phecode_132'))
inflammation_of_eyelids_phe = TensorMap('phecode_371.3', Interpretation.CATEGORICAL, channel_map={'no_inflammation_of_eyelids':0, 'inflammation_of_eyelids':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 154], 'phecode_371_3'))
inflammation_of_the_eye_phe = TensorMap('phecode_371', Interpretation.CATEGORICAL, channel_map={'no_inflammation_of_the_eye':0, 'inflammation_of_the_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3521], 'phecode_371'))
inflammatory_and_toxic_neuropathy_phe = TensorMap('phecode_357', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_and_toxic_neuropathy':0, 'inflammatory_and_toxic_neuropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 220], 'phecode_357'))
inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis_phe = TensorMap('phecode_555', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis':0, 'inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19], 'phecode_555'))
inflammatory_conditions_of_jaw_phe = TensorMap('phecode_526.5', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_conditions_of_jaw':0, 'inflammatory_conditions_of_jaw':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3048], 'phecode_526_5'))
inflammatory_disease_of_breast_phe = TensorMap('phecode_613.1', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_disease_of_breast':0, 'inflammatory_disease_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 545], 'phecode_613_1'))
inflammatory_disease_of_cervix_vagina_and_vulva_phe = TensorMap('phecode_614.5', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_disease_of_cervix_vagina_and_vulva':0, 'inflammatory_disease_of_cervix_vagina_and_vulva':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2392], 'phecode_614_5'))
inflammatory_diseases_of_female_pelvic_organs_phe = TensorMap('phecode_614', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_diseases_of_female_pelvic_organs':0, 'inflammatory_diseases_of_female_pelvic_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18518], 'phecode_614'))
inflammatory_diseases_of_uterus_except_cervix_phe = TensorMap('phecode_614.4', Interpretation.CATEGORICAL, channel_map={'no_inflammatory_diseases_of_uterus_except_cervix':0, 'inflammatory_diseases_of_uterus_except_cervix':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1315], 'phecode_614_4'))
influenza_phe = TensorMap('phecode_481', Interpretation.CATEGORICAL, channel_map={'no_influenza':0, 'influenza':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 30], 'phecode_481'))
infusion_and_transfusion_reaction_phe = TensorMap('phecode_942', Interpretation.CATEGORICAL, channel_map={'no_infusion_and_transfusion_reaction':0, 'infusion_and_transfusion_reaction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_942'))
ingrowing_nail_phe = TensorMap('phecode_703.1', Interpretation.CATEGORICAL, channel_map={'no_ingrowing_nail':0, 'ingrowing_nail':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 459], 'phecode_703_1'))
inguinal_hernia_phe = TensorMap('phecode_550.1', Interpretation.CATEGORICAL, channel_map={'no_inguinal_hernia':0, 'inguinal_hernia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23], 'phecode_550_1'))
injuries_to_the_nervous_system_phe = TensorMap('phecode_907', Interpretation.CATEGORICAL, channel_map={'no_injuries_to_the_nervous_system':0, 'injuries_to_the_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 203], 'phecode_907'))
injury_nos_phe = TensorMap('phecode_1009', Interpretation.CATEGORICAL, channel_map={'no_injury_nos':0, 'injury_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 28], 'phecode_1009'))
injury_to_blood_vessels_phe = TensorMap('phecode_1007', Interpretation.CATEGORICAL, channel_map={'no_injury_to_blood_vessels':0, 'injury_to_blood_vessels':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1369], 'phecode_1007'))
injury_to_other_and_unspecified_nerves_phe = TensorMap('phecode_957', Interpretation.CATEGORICAL, channel_map={'no_injury_to_other_and_unspecified_nerves':0, 'injury_to_other_and_unspecified_nerves':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1237], 'phecode_957'))
insomnia_phe = TensorMap('phecode_327.4', Interpretation.CATEGORICAL, channel_map={'no_insomnia':0, 'insomnia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12500], 'phecode_327_4'))
insulin_pump_user_phe = TensorMap('phecode_250.3', Interpretation.CATEGORICAL, channel_map={'no_insulin_pump_user':0, 'insulin_pump_user':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6944], 'phecode_250_3'))
internal_derangement_of_knee_phe = TensorMap('phecode_835', Interpretation.CATEGORICAL, channel_map={'no_internal_derangement_of_knee':0, 'internal_derangement_of_knee':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 96], 'phecode_835'))
interstitial_emphysema_and_related_conditions_of_newborn_phe = TensorMap('phecode_656.22', Interpretation.CATEGORICAL, channel_map={'no_interstitial_emphysema_and_related_conditions_of_newborn':0, 'interstitial_emphysema_and_related_conditions_of_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_656_22'))
intervertebral_disc_disorder_with_myelopathy_phe = TensorMap('phecode_722.7', Interpretation.CATEGORICAL, channel_map={'no_intervertebral_disc_disorder_with_myelopathy':0, 'intervertebral_disc_disorder_with_myelopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 860], 'phecode_722_7'))
intervertebral_disc_disorders_phe = TensorMap('phecode_722', Interpretation.CATEGORICAL, channel_map={'no_intervertebral_disc_disorders':0, 'intervertebral_disc_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 58], 'phecode_722'))
intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption_phe = TensorMap('phecode_271.3', Interpretation.CATEGORICAL, channel_map={'no_intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption':0, 'intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1984], 'phecode_271_3'))
intestinal_e_coli_phe = TensorMap('phecode_8.51', Interpretation.CATEGORICAL, channel_map={'no_intestinal_e_coli':0, 'intestinal_e_coli':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10638], 'phecode_8_51'))
intestinal_helminthiases_phe = TensorMap('phecode_134.1', Interpretation.CATEGORICAL, channel_map={'no_intestinal_helminthiases':0, 'intestinal_helminthiases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2958], 'phecode_134_1'))
intestinal_infection_phe = TensorMap('phecode_8', Interpretation.CATEGORICAL, channel_map={'no_intestinal_infection':0, 'intestinal_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 50], 'phecode_8'))
intestinal_infection_due_to_c_difficile_phe = TensorMap('phecode_8.52', Interpretation.CATEGORICAL, channel_map={'no_intestinal_infection_due_to_c_difficile':0, 'intestinal_infection_due_to_c_difficile':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 493], 'phecode_8_52'))
intestinal_infection_due_to_protozoa_phe = TensorMap('phecode_8.7', Interpretation.CATEGORICAL, channel_map={'no_intestinal_infection_due_to_protozoa':0, 'intestinal_infection_due_to_protozoa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6250], 'phecode_8_7'))
intestinal_malabsorption_non_celiac_phe = TensorMap('phecode_557', Interpretation.CATEGORICAL, channel_map={'no_intestinal_malabsorption_non_celiac':0, 'intestinal_malabsorption_non_celiac':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1165], 'phecode_557'))
intracerebral_hemorrhage_phe = TensorMap('phecode_430.2', Interpretation.CATEGORICAL, channel_map={'no_intracerebral_hemorrhage':0, 'intracerebral_hemorrhage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 442], 'phecode_430_2'))
intracranial_hemorrhage_phe = TensorMap('phecode_430', Interpretation.CATEGORICAL, channel_map={'no_intracranial_hemorrhage':0, 'intracranial_hemorrhage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2747], 'phecode_430'))
intracranial_hemorrhage_injury_phe = TensorMap('phecode_818', Interpretation.CATEGORICAL, channel_map={'no_intracranial_hemorrhage_injury':0, 'intracranial_hemorrhage_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2732], 'phecode_818'))
iron_deficiency_anemia_secondary_to_blood_loss_chronic_phe = TensorMap('phecode_280.2', Interpretation.CATEGORICAL, channel_map={'no_iron_deficiency_anemia_secondary_to_blood_loss_chronic':0, 'iron_deficiency_anemia_secondary_to_blood_loss_chronic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 684], 'phecode_280_2'))
iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss_phe = TensorMap('phecode_280.1', Interpretation.CATEGORICAL, channel_map={'no_iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss':0, 'iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 40], 'phecode_280_1'))
irregular_menstrual_bleeding_phe = TensorMap('phecode_626.14', Interpretation.CATEGORICAL, channel_map={'no_irregular_menstrual_bleeding':0, 'irregular_menstrual_bleeding':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 109], 'phecode_626_14'))
irregular_menstrual_cycle_phe = TensorMap('phecode_626.13', Interpretation.CATEGORICAL, channel_map={'no_irregular_menstrual_cycle':0, 'irregular_menstrual_cycle':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 236], 'phecode_626_13'))
irregular_menstrual_cycle_bleeding_phe = TensorMap('phecode_626.1', Interpretation.CATEGORICAL, channel_map={'no_irregular_menstrual_cycle_bleeding':0, 'irregular_menstrual_cycle_bleeding':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 235], 'phecode_626_1'))
irritable_bowel_syndrome_phe = TensorMap('phecode_564.1', Interpretation.CATEGORICAL, channel_map={'no_irritable_bowel_syndrome':0, 'irritable_bowel_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 60], 'phecode_564_1'))
ischemic_heart_disease_phe = TensorMap('phecode_411', Interpretation.CATEGORICAL, channel_map={'no_ischemic_heart_disease':0, 'ischemic_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11], 'phecode_411'))
jakob_creutzfeldt_disease_phe = TensorMap('phecode_324.1', Interpretation.CATEGORICAL, channel_map={'no_jakob_creutzfeldt_disease':0, 'jakob_creutzfeldt_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21739], 'phecode_324_1'))
jaundice_not_of_newborn_phe = TensorMap('phecode_573.5', Interpretation.CATEGORICAL, channel_map={'no_jaundice_not_of_newborn':0, 'jaundice_not_of_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 386], 'phecode_573_5'))
jaw_disease_nos_phe = TensorMap('phecode_526.9', Interpretation.CATEGORICAL, channel_map={'no_jaw_disease_nos':0, 'jaw_disease_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_526_9'))
joint_effusions_phe = TensorMap('phecode_741.4', Interpretation.CATEGORICAL, channel_map={'no_joint_effusions':0, 'joint_effusions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1457], 'phecode_741_4'))
joint_ligament_sprain_phe = TensorMap('phecode_840.3', Interpretation.CATEGORICAL, channel_map={'no_joint_ligament_sprain':0, 'joint_ligament_sprain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2049], 'phecode_840_3'))
juvenile_osteochondrosis_phe = TensorMap('phecode_732.1', Interpretation.CATEGORICAL, channel_map={'no_juvenile_osteochondrosis':0, 'juvenile_osteochondrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2136], 'phecode_732_1'))
juvenile_rheumatoid_arthritis_phe = TensorMap('phecode_714.2', Interpretation.CATEGORICAL, channel_map={'no_juvenile_rheumatoid_arthritis':0, 'juvenile_rheumatoid_arthritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_714_2'))
kaposis_sarcoma_phe = TensorMap('phecode_230', Interpretation.CATEGORICAL, channel_map={'no_kaposis_sarcoma':0, 'kaposis_sarcoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 33333], 'phecode_230'))
keloid_scar_phe = TensorMap('phecode_701.4', Interpretation.CATEGORICAL, channel_map={'no_keloid_scar':0, 'keloid_scar':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1385], 'phecode_701_4'))
keratitis_phe = TensorMap('phecode_370', Interpretation.CATEGORICAL, channel_map={'no_keratitis':0, 'keratitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2631], 'phecode_370'))
keratoconjunctivitis_phe = TensorMap('phecode_370.3', Interpretation.CATEGORICAL, channel_map={'no_keratoconjunctivitis':0, 'keratoconjunctivitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11627], 'phecode_370_3'))
keratoconus_phe = TensorMap('phecode_364.41', Interpretation.CATEGORICAL, channel_map={'no_keratoconus':0, 'keratoconus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3787], 'phecode_364_41'))
keratoderma_acquired_phe = TensorMap('phecode_701.1', Interpretation.CATEGORICAL, channel_map={'no_keratoderma_acquired':0, 'keratoderma_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5813], 'phecode_701_1'))
kidney_replaced_by_transpant_phe = TensorMap('phecode_587', Interpretation.CATEGORICAL, channel_map={'no_kidney_replaced_by_transpant':0, 'kidney_replaced_by_transpant':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 905], 'phecode_587'))
known_or_suspected_fetal_abnormality_affecting_management_of_mother_phe = TensorMap('phecode_655', Interpretation.CATEGORICAL, channel_map={'no_known_or_suspected_fetal_abnormality_affecting_management_of_mother':0, 'known_or_suspected_fetal_abnormality_affecting_management_of_mother':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 84], 'phecode_655'))
kyphoscoliosis_and_scoliosis_phe = TensorMap('phecode_737.3', Interpretation.CATEGORICAL, channel_map={'no_kyphoscoliosis_and_scoliosis':0, 'kyphoscoliosis_and_scoliosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 702], 'phecode_737_3'))
kyphosis_acquired_phe = TensorMap('phecode_737.1', Interpretation.CATEGORICAL, channel_map={'no_kyphosis_acquired':0, 'kyphosis_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5555], 'phecode_737_1'))
labyrinthitis_phe = TensorMap('phecode_386.3', Interpretation.CATEGORICAL, channel_map={'no_labyrinthitis':0, 'labyrinthitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 474], 'phecode_386_3'))
lack_of_coordination_phe = TensorMap('phecode_350.3', Interpretation.CATEGORICAL, channel_map={'no_lack_of_coordination':0, 'lack_of_coordination':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 936], 'phecode_350_3'))
lack_of_normal_physiological_development_unspecified_phe = TensorMap('phecode_264.9', Interpretation.CATEGORICAL, channel_map={'no_lack_of_normal_physiological_development_unspecified':0, 'lack_of_normal_physiological_development_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23809], 'phecode_264_9'))
lagophthalmos_phe = TensorMap('phecode_374.2', Interpretation.CATEGORICAL, channel_map={'no_lagophthalmos':0, 'lagophthalmos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8196], 'phecode_374_2'))
large_cell_lymphoma_phe = TensorMap('phecode_202.24', Interpretation.CATEGORICAL, channel_map={'no_large_cell_lymphoma':0, 'large_cell_lymphoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 572], 'phecode_202_24'))
late_effect_phe = TensorMap('phecode_1012', Interpretation.CATEGORICAL, channel_map={'no_late_effect':0, 'late_effect':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16666], 'phecode_1012'))
late_effects_of_cerebrovascular_disease_phe = TensorMap('phecode_433.8', Interpretation.CATEGORICAL, channel_map={'no_late_effects_of_cerebrovascular_disease':0, 'late_effects_of_cerebrovascular_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 247], 'phecode_433_8'))
late_effects_of_musculoskeletal_and_connective_tissue_injuries_phe = TensorMap('phecode_905', Interpretation.CATEGORICAL, channel_map={'no_late_effects_of_musculoskeletal_and_connective_tissue_injuries':0, 'late_effects_of_musculoskeletal_and_connective_tissue_injuries':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3105], 'phecode_905'))
late_pregnancy_and_failed_induction_phe = TensorMap('phecode_645', Interpretation.CATEGORICAL, channel_map={'no_late_pregnancy_and_failed_induction':0, 'late_pregnancy_and_failed_induction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 305], 'phecode_645'))
laxity_of_ligament_or_hypermobility_syndrome_phe = TensorMap('phecode_728.2', Interpretation.CATEGORICAL, channel_map={'no_laxity_of_ligament_or_hypermobility_syndrome':0, 'laxity_of_ligament_or_hypermobility_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2762], 'phecode_728_2'))
learning_disorder_phe = TensorMap('phecode_315.1', Interpretation.CATEGORICAL, channel_map={'no_learning_disorder':0, 'learning_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1824], 'phecode_315_1'))
left_bundle_branch_block_phe = TensorMap('phecode_426.32', Interpretation.CATEGORICAL, channel_map={'no_left_bundle_branch_block':0, 'left_bundle_branch_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 168], 'phecode_426_32'))
leukemia_phe = TensorMap('phecode_204', Interpretation.CATEGORICAL, channel_map={'no_leukemia':0, 'leukemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 757], 'phecode_204'))
leukoplakia_of_oral_mucosa_phe = TensorMap('phecode_528.6', Interpretation.CATEGORICAL, channel_map={'no_leukoplakia_of_oral_mucosa':0, 'leukoplakia_of_oral_mucosa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1101], 'phecode_528_6'))
lipoma_phe = TensorMap('phecode_214', Interpretation.CATEGORICAL, channel_map={'no_lipoma':0, 'lipoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 218], 'phecode_214'))
lipoma_of_skin_and_subcutaneous_tissue_phe = TensorMap('phecode_214.1', Interpretation.CATEGORICAL, channel_map={'no_lipoma_of_skin_and_subcutaneous_tissue':0, 'lipoma_of_skin_and_subcutaneous_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 74], 'phecode_214_1'))
lipoprotein_disorders_phe = TensorMap('phecode_277.51', Interpretation.CATEGORICAL, channel_map={'no_lipoprotein_disorders':0, 'lipoprotein_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5882], 'phecode_277_51'))
liver_abscess_and_sequelae_of_chronic_liver_disease_phe = TensorMap('phecode_571.8', Interpretation.CATEGORICAL, channel_map={'no_liver_abscess_and_sequelae_of_chronic_liver_disease':0, 'liver_abscess_and_sequelae_of_chronic_liver_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 554], 'phecode_571_8'))
liver_replaced_by_transplant_phe = TensorMap('phecode_573.2', Interpretation.CATEGORICAL, channel_map={'no_liver_replaced_by_transplant':0, 'liver_replaced_by_transplant':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3401], 'phecode_573_2'))
localized_adiposity_phe = TensorMap('phecode_278.3', Interpretation.CATEGORICAL, channel_map={'no_localized_adiposity':0, 'localized_adiposity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4854], 'phecode_278_3'))
localized_superficial_swelling_mass_or_lump_phe = TensorMap('phecode_687.2', Interpretation.CATEGORICAL, channel_map={'no_localized_superficial_swelling_mass_or_lump':0, 'localized_superficial_swelling_mass_or_lump':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1689], 'phecode_687_2'))
loose_body_in_joint_phe = TensorMap('phecode_742.1', Interpretation.CATEGORICAL, channel_map={'no_loose_body_in_joint':0, 'loose_body_in_joint':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7812], 'phecode_742_1'))
lordosis_acquired_phe = TensorMap('phecode_737.2', Interpretation.CATEGORICAL, channel_map={'no_lordosis_acquired':0, 'lordosis_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_737_2'))
lower_gastrointestinal_congenital_anomalies_phe = TensorMap('phecode_750.2', Interpretation.CATEGORICAL, channel_map={'no_lower_gastrointestinal_congenital_anomalies':0, 'lower_gastrointestinal_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25000], 'phecode_750_2'))
lump_or_mass_in_breast_phe = TensorMap('phecode_611.3', Interpretation.CATEGORICAL, channel_map={'no_lump_or_mass_in_breast':0, 'lump_or_mass_in_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 244], 'phecode_611_3'))
lung_disease_due_to_external_agents_phe = TensorMap('phecode_500', Interpretation.CATEGORICAL, channel_map={'no_lung_disease_due_to_external_agents':0, 'lung_disease_due_to_external_agents':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3030], 'phecode_500'))
lung_transplant_phe = TensorMap('phecode_510.2', Interpretation.CATEGORICAL, channel_map={'no_lung_transplant':0, 'lung_transplant':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_510_2'))
lyme_disease_phe = TensorMap('phecode_130.1', Interpretation.CATEGORICAL, channel_map={'no_lyme_disease':0, 'lyme_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11904], 'phecode_130_1'))
lymphadenitis_phe = TensorMap('phecode_289.4', Interpretation.CATEGORICAL, channel_map={'no_lymphadenitis':0, 'lymphadenitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125], 'phecode_289_4'))
lymphoid_leukemia_phe = TensorMap('phecode_204.1', Interpretation.CATEGORICAL, channel_map={'no_lymphoid_leukemia':0, 'lymphoid_leukemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6250], 'phecode_204_1'))
lymphoid_leukemia_acute_phe = TensorMap('phecode_204.11', Interpretation.CATEGORICAL, channel_map={'no_lymphoid_leukemia_acute':0, 'lymphoid_leukemia_acute':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5154], 'phecode_204_11'))
lymphoid_leukemia_chronic_phe = TensorMap('phecode_204.12', Interpretation.CATEGORICAL, channel_map={'no_lymphoid_leukemia_chronic':0, 'lymphoid_leukemia_chronic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 664], 'phecode_204_12'))
lymphosarcoma_phe = TensorMap('phecode_202.23', Interpretation.CATEGORICAL, channel_map={'no_lymphosarcoma':0, 'lymphosarcoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35714], 'phecode_202_23'))
macroglobulinemia_phe = TensorMap('phecode_270.35', Interpretation.CATEGORICAL, channel_map={'no_macroglobulinemia':0, 'macroglobulinemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4545], 'phecode_270_35'))
macular_degeneration_senile_of_retina_nos_phe = TensorMap('phecode_362.29', Interpretation.CATEGORICAL, channel_map={'no_macular_degeneration_senile_of_retina_nos':0, 'macular_degeneration_senile_of_retina_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 126], 'phecode_362_29'))
major_depressive_disorder_phe = TensorMap('phecode_296.22', Interpretation.CATEGORICAL, channel_map={'no_major_depressive_disorder':0, 'major_depressive_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 891], 'phecode_296_22'))
major_puerperal_infection_phe = TensorMap('phecode_647.3', Interpretation.CATEGORICAL, channel_map={'no_major_puerperal_infection':0, 'major_puerperal_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 616], 'phecode_647_3'))
malaise_and_fatigue_phe = TensorMap('phecode_798', Interpretation.CATEGORICAL, channel_map={'no_malaise_and_fatigue':0, 'malaise_and_fatigue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 114], 'phecode_798'))
malignant_neoplasm_of_bladder_phe = TensorMap('phecode_189.21', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_bladder':0, 'malignant_neoplasm_of_bladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 168], 'phecode_189_21'))
malignant_neoplasm_of_female_breast_phe = TensorMap('phecode_174.11', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_female_breast':0, 'malignant_neoplasm_of_female_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29], 'phecode_174_11'))
malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts_phe = TensorMap('phecode_159.3', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts':0, 'malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1612], 'phecode_159_3'))
malignant_neoplasm_of_head_face_and_neck_phe = TensorMap('phecode_195.3', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_head_face_and_neck':0, 'malignant_neoplasm_of_head_face_and_neck':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9090], 'phecode_195_3'))
malignant_neoplasm_of_kidney_except_pelvis_phe = TensorMap('phecode_189.11', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_kidney_except_pelvis':0, 'malignant_neoplasm_of_kidney_except_pelvis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 338], 'phecode_189_11'))
malignant_neoplasm_of_liver_primary_phe = TensorMap('phecode_155.1', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_liver_primary':0, 'malignant_neoplasm_of_liver_primary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2008], 'phecode_155_1'))
malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum_phe = TensorMap('phecode_159', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum':0, 'malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 66], 'phecode_159'))
malignant_neoplasm_of_other_urinary_organs_phe = TensorMap('phecode_189.4', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_other_urinary_organs':0, 'malignant_neoplasm_of_other_urinary_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 112], 'phecode_189_4'))
malignant_neoplasm_of_ovary_phe = TensorMap('phecode_184.11', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_ovary':0, 'malignant_neoplasm_of_ovary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 67], 'phecode_184_11'))
malignant_neoplasm_of_ovary_and_other_uterine_adnexa_phe = TensorMap('phecode_184.1', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_ovary_and_other_uterine_adnexa':0, 'malignant_neoplasm_of_ovary_and_other_uterine_adnexa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 139], 'phecode_184_1'))
malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus_phe = TensorMap('phecode_153.3', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus':0, 'malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 169], 'phecode_153_3'))
malignant_neoplasm_of_renal_pelvis_phe = TensorMap('phecode_189.12', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_renal_pelvis':0, 'malignant_neoplasm_of_renal_pelvis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4761], 'phecode_189_12'))
malignant_neoplasm_of_retroperitoneum_and_peritoneum_phe = TensorMap('phecode_159.4', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_retroperitoneum_and_peritoneum':0, 'malignant_neoplasm_of_retroperitoneum_and_peritoneum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1718], 'phecode_159_4'))
malignant_neoplasm_of_small_intestine_including_duodenum_phe = TensorMap('phecode_159.2', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_small_intestine_including_duodenum':0, 'malignant_neoplasm_of_small_intestine_including_duodenum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1838], 'phecode_159_2'))
malignant_neoplasm_of_testis_phe = TensorMap('phecode_187.2', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_testis':0, 'malignant_neoplasm_of_testis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 73], 'phecode_187_2'))
malignant_neoplasm_of_unspecified_male_genital_organ_phe = TensorMap('phecode_187.1', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_unspecified_male_genital_organ':0, 'malignant_neoplasm_of_unspecified_male_genital_organ':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 40], 'phecode_187_1'))
malignant_neoplasm_of_uterus_phe = TensorMap('phecode_182', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_of_uterus':0, 'malignant_neoplasm_of_uterus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 271], 'phecode_182'))
malignant_neoplasm_other_phe = TensorMap('phecode_195.1', Interpretation.CATEGORICAL, channel_map={'no_malignant_neoplasm_other':0, 'malignant_neoplasm_other':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6], 'phecode_195_1'))
malposition_and_malpresentation_of_fetus_or_obstruction_phe = TensorMap('phecode_652', Interpretation.CATEGORICAL, channel_map={'no_malposition_and_malpresentation_of_fetus_or_obstruction':0, 'malposition_and_malpresentation_of_fetus_or_obstruction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 219], 'phecode_652'))
malunion_and_nonunion_of_fracture_phe = TensorMap('phecode_733.8', Interpretation.CATEGORICAL, channel_map={'no_malunion_and_nonunion_of_fracture':0, 'malunion_and_nonunion_of_fracture':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2487], 'phecode_733_8'))
manlignant_and_unknown_neoplasms_of_brain_and_nervous_system_phe = TensorMap('phecode_191', Interpretation.CATEGORICAL, channel_map={'no_manlignant_and_unknown_neoplasms_of_brain_and_nervous_system':0, 'manlignant_and_unknown_neoplasms_of_brain_and_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1644], 'phecode_191'))
mastodynia_phe = TensorMap('phecode_613.5', Interpretation.CATEGORICAL, channel_map={'no_mastodynia':0, 'mastodynia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4761], 'phecode_613_5'))
mastoiditis_related_conditions_phe = TensorMap('phecode_381.3', Interpretation.CATEGORICAL, channel_map={'no_mastoiditis_related_conditions':0, 'mastoiditis_related_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2450], 'phecode_381_3'))
maternal_complication_of_pregnancy_affecting_fetus_or_newborn_phe = TensorMap('phecode_658', Interpretation.CATEGORICAL, channel_map={'no_maternal_complication_of_pregnancy_affecting_fetus_or_newborn':0, 'maternal_complication_of_pregnancy_affecting_fetus_or_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12820], 'phecode_658'))
mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft_phe = TensorMap('phecode_857', Interpretation.CATEGORICAL, channel_map={'no_mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft':0, 'mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 261], 'phecode_857'))
medulloadrenal_hyperfunction_phe = TensorMap('phecode_255.13', Interpretation.CATEGORICAL, channel_map={'no_medulloadrenal_hyperfunction':0, 'medulloadrenal_hyperfunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55555], 'phecode_255_13'))
megaloblastic_anemia_phe = TensorMap('phecode_281.1', Interpretation.CATEGORICAL, channel_map={'no_megaloblastic_anemia':0, 'megaloblastic_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29411], 'phecode_281_1'))
melanomas_of_skin_phe = TensorMap('phecode_172.11', Interpretation.CATEGORICAL, channel_map={'no_melanomas_of_skin':0, 'melanomas_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 26], 'phecode_172_11'))
memory_loss_phe = TensorMap('phecode_292.3', Interpretation.CATEGORICAL, channel_map={'no_memory_loss':0, 'memory_loss':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 463], 'phecode_292_3'))
menieres_disease_phe = TensorMap('phecode_386.1', Interpretation.CATEGORICAL, channel_map={'no_menieres_disease':0, 'menieres_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 616], 'phecode_386_1'))
meningitis_phe = TensorMap('phecode_320', Interpretation.CATEGORICAL, channel_map={'no_meningitis':0, 'meningitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 910], 'phecode_320'))
menopausal_and_postmenopausal_disorders_phe = TensorMap('phecode_627', Interpretation.CATEGORICAL, channel_map={'no_menopausal_and_postmenopausal_disorders':0, 'menopausal_and_postmenopausal_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 673], 'phecode_627'))
mental_retardation_phe = TensorMap('phecode_315.3', Interpretation.CATEGORICAL, channel_map={'no_mental_retardation':0, 'mental_retardation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4950], 'phecode_315_3'))
migrain_with_aura_phe = TensorMap('phecode_340.1', Interpretation.CATEGORICAL, channel_map={'no_migrain_with_aura':0, 'migrain_with_aura':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1506], 'phecode_340_1'))
migraine_phe = TensorMap('phecode_340', Interpretation.CATEGORICAL, channel_map={'no_migraine':0, 'migraine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 119], 'phecode_340'))
mild_cognitive_impairment_phe = TensorMap('phecode_292.2', Interpretation.CATEGORICAL, channel_map={'no_mild_cognitive_impairment':0, 'mild_cognitive_impairment':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1291], 'phecode_292_2'))
mineral_deficiency_nec_phe = TensorMap('phecode_262', Interpretation.CATEGORICAL, channel_map={'no_mineral_deficiency_nec':0, 'mineral_deficiency_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1091], 'phecode_262'))
miscarriage_stillbirth_phe = TensorMap('phecode_634', Interpretation.CATEGORICAL, channel_map={'no_miscarriage_stillbirth':0, 'miscarriage_stillbirth':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 87], 'phecode_634'))
missed_abortion_hydatidiform_mole_phe = TensorMap('phecode_634.1', Interpretation.CATEGORICAL, channel_map={'no_missed_abortion_hydatidiform_mole':0, 'missed_abortion_hydatidiform_mole':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 319], 'phecode_634_1'))
mitral_valve_disease_phe = TensorMap('phecode_394.2', Interpretation.CATEGORICAL, channel_map={'no_mitral_valve_disease':0, 'mitral_valve_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 74], 'phecode_394_2'))
mittelschmerz_phe = TensorMap('phecode_626.21', Interpretation.CATEGORICAL, channel_map={'no_mittelschmerz':0, 'mittelschmerz':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15625], 'phecode_626_21'))
mixed_hyperlipidemia_phe = TensorMap('phecode_272.13', Interpretation.CATEGORICAL, channel_map={'no_mixed_hyperlipidemia':0, 'mixed_hyperlipidemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2109], 'phecode_272_13'))
monocytic_leukemia_phe = TensorMap('phecode_204.3', Interpretation.CATEGORICAL, channel_map={'no_monocytic_leukemia':0, 'monocytic_leukemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8333], 'phecode_204_3'))
mood_disorders_phe = TensorMap('phecode_296', Interpretation.CATEGORICAL, channel_map={'no_mood_disorders':0, 'mood_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 24], 'phecode_296'))
moyamoya_disease_phe = TensorMap('phecode_433.32', Interpretation.CATEGORICAL, channel_map={'no_moyamoya_disease':0, 'moyamoya_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_433_32'))
mucous_polyp_of_cervix_phe = TensorMap('phecode_622.2', Interpretation.CATEGORICAL, channel_map={'no_mucous_polyp_of_cervix':0, 'mucous_polyp_of_cervix':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 122], 'phecode_622_2'))
multiple_gestation_phe = TensorMap('phecode_651', Interpretation.CATEGORICAL, channel_map={'no_multiple_gestation':0, 'multiple_gestation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1057], 'phecode_651'))
multiple_myeloma_phe = TensorMap('phecode_204.4', Interpretation.CATEGORICAL, channel_map={'no_multiple_myeloma':0, 'multiple_myeloma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 567], 'phecode_204_4'))
multiple_sclerosis_phe = TensorMap('phecode_335', Interpretation.CATEGORICAL, channel_map={'no_multiple_sclerosis':0, 'multiple_sclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 282], 'phecode_335'))
muscular_calcification_and_ossification_phe = TensorMap('phecode_728.1', Interpretation.CATEGORICAL, channel_map={'no_muscular_calcification_and_ossification':0, 'muscular_calcification_and_ossification':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 33333], 'phecode_728_1'))
muscular_dystrophies_phe = TensorMap('phecode_359.1', Interpretation.CATEGORICAL, channel_map={'no_muscular_dystrophies':0, 'muscular_dystrophies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2777], 'phecode_359_1'))
muscular_wasting_and_disuse_atrophy_phe = TensorMap('phecode_772.1', Interpretation.CATEGORICAL, channel_map={'no_muscular_wasting_and_disuse_atrophy':0, 'muscular_wasting_and_disuse_atrophy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15625], 'phecode_772_1'))
musculoskeletal_symptoms_referable_to_limbs_phe = TensorMap('phecode_771', Interpretation.CATEGORICAL, channel_map={'no_musculoskeletal_symptoms_referable_to_limbs':0, 'musculoskeletal_symptoms_referable_to_limbs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 428], 'phecode_771'))
myalgia_and_myositis_unspecified_phe = TensorMap('phecode_770', Interpretation.CATEGORICAL, channel_map={'no_myalgia_and_myositis_unspecified':0, 'myalgia_and_myositis_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2008], 'phecode_770'))
myasthenia_gravis_phe = TensorMap('phecode_358.1', Interpretation.CATEGORICAL, channel_map={'no_myasthenia_gravis':0, 'myasthenia_gravis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1872], 'phecode_358_1'))
mycoses_phe = TensorMap('phecode_117', Interpretation.CATEGORICAL, channel_map={'no_mycoses':0, 'mycoses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 111], 'phecode_117'))
myeloid_leukemia_phe = TensorMap('phecode_204.2', Interpretation.CATEGORICAL, channel_map={'no_myeloid_leukemia':0, 'myeloid_leukemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9433], 'phecode_204_2'))
myeloid_leukemia_acute_phe = TensorMap('phecode_204.21', Interpretation.CATEGORICAL, channel_map={'no_myeloid_leukemia_acute':0, 'myeloid_leukemia_acute':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1096], 'phecode_204_21'))
myeloid_leukemia_chronic_phe = TensorMap('phecode_204.22', Interpretation.CATEGORICAL, channel_map={'no_myeloid_leukemia_chronic':0, 'myeloid_leukemia_chronic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2336], 'phecode_204_22'))
myeloproliferative_disease_phe = TensorMap('phecode_200', Interpretation.CATEGORICAL, channel_map={'no_myeloproliferative_disease':0, 'myeloproliferative_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 476], 'phecode_200'))
myocardial_infarction_phe = TensorMap('phecode_411.2', Interpretation.CATEGORICAL, channel_map={'no_myocardial_infarction':0, 'myocardial_infarction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 28], 'phecode_411_2'))
myocarditis_phe = TensorMap('phecode_420.1', Interpretation.CATEGORICAL, channel_map={'no_myocarditis':0, 'myocarditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3012], 'phecode_420_1'))
myoclonus_phe = TensorMap('phecode_333.2', Interpretation.CATEGORICAL, channel_map={'no_myoclonus':0, 'myoclonus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3676], 'phecode_333_2'))
myoneural_disorders_phe = TensorMap('phecode_358', Interpretation.CATEGORICAL, channel_map={'no_myoneural_disorders':0, 'myoneural_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_358'))
myopathy_phe = TensorMap('phecode_359.2', Interpretation.CATEGORICAL, channel_map={'no_myopathy':0, 'myopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 673], 'phecode_359_2'))
myopia_phe = TensorMap('phecode_367.1', Interpretation.CATEGORICAL, channel_map={'no_myopia':0, 'myopia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 240], 'phecode_367_1'))
myringitis_phe = TensorMap('phecode_384.1', Interpretation.CATEGORICAL, channel_map={'no_myringitis':0, 'myringitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20833], 'phecode_384_1'))
nasal_polyps_phe = TensorMap('phecode_471', Interpretation.CATEGORICAL, channel_map={'no_nasal_polyps':0, 'nasal_polyps':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 121], 'phecode_471'))
nausea_and_vomiting_phe = TensorMap('phecode_789', Interpretation.CATEGORICAL, channel_map={'no_nausea_and_vomiting':0, 'nausea_and_vomiting':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 28], 'phecode_789'))
neoplasm_of_uncertain_behavior_phe = TensorMap('phecode_199', Interpretation.CATEGORICAL, channel_map={'no_neoplasm_of_uncertain_behavior':0, 'neoplasm_of_uncertain_behavior':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 307], 'phecode_199'))
neoplasm_of_uncertain_behavior_of_breast_phe = TensorMap('phecode_174.3', Interpretation.CATEGORICAL, channel_map={'no_neoplasm_of_uncertain_behavior_of_breast':0, 'neoplasm_of_uncertain_behavior_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3401], 'phecode_174_3'))
neoplasm_of_uncertain_behavior_of_male_genital_organs_phe = TensorMap('phecode_187.8', Interpretation.CATEGORICAL, channel_map={'no_neoplasm_of_uncertain_behavior_of_male_genital_organs':0, 'neoplasm_of_uncertain_behavior_of_male_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3787], 'phecode_187_8'))
neoplasm_of_uncertain_behavior_of_skin_phe = TensorMap('phecode_173', Interpretation.CATEGORICAL, channel_map={'no_neoplasm_of_uncertain_behavior_of_skin':0, 'neoplasm_of_uncertain_behavior_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2183], 'phecode_173'))
neoplasm_of_unspecified_nature_of_digestive_system_phe = TensorMap('phecode_158', Interpretation.CATEGORICAL, channel_map={'no_neoplasm_of_unspecified_nature_of_digestive_system':0, 'neoplasm_of_unspecified_nature_of_digestive_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 317], 'phecode_158'))
nephritis_and_nephropathy_with_pathological_lesion_phe = TensorMap('phecode_580.32', Interpretation.CATEGORICAL, channel_map={'no_nephritis_and_nephropathy_with_pathological_lesion':0, 'nephritis_and_nephropathy_with_pathological_lesion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16], 'phecode_580_32'))
nephritis_nephrosis_renal_sclerosis_phe = TensorMap('phecode_580', Interpretation.CATEGORICAL, channel_map={'no_nephritis_nephrosis_renal_sclerosis':0, 'nephritis_nephrosis_renal_sclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2127], 'phecode_580'))
nephrotic_syndrome_without_mention_of_glomerulonephritis_phe = TensorMap('phecode_580.2', Interpretation.CATEGORICAL, channel_map={'no_nephrotic_syndrome_without_mention_of_glomerulonephritis':0, 'nephrotic_syndrome_without_mention_of_glomerulonephritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 623], 'phecode_580_2'))
nerve_plexus_lesions_phe = TensorMap('phecode_353.1', Interpretation.CATEGORICAL, channel_map={'no_nerve_plexus_lesions':0, 'nerve_plexus_lesions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3012], 'phecode_353_1'))
nerve_root_and_plexus_disorders_phe = TensorMap('phecode_353', Interpretation.CATEGORICAL, channel_map={'no_nerve_root_and_plexus_disorders':0, 'nerve_root_and_plexus_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 315], 'phecode_353'))
nerve_root_lesions_phe = TensorMap('phecode_353.2', Interpretation.CATEGORICAL, channel_map={'no_nerve_root_lesions':0, 'nerve_root_lesions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6024], 'phecode_353_2'))
nervous_system_congenital_anomalies_phe = TensorMap('phecode_752', Interpretation.CATEGORICAL, channel_map={'no_nervous_system_congenital_anomalies':0, 'nervous_system_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 67], 'phecode_752'))
neural_tube_defects_phe = TensorMap('phecode_752.1', Interpretation.CATEGORICAL, channel_map={'no_neural_tube_defects':0, 'neural_tube_defects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5555], 'phecode_752_1'))
neuralgia_neuritis_and_radiculitis_nos_phe = TensorMap('phecode_766', Interpretation.CATEGORICAL, channel_map={'no_neuralgia_neuritis_and_radiculitis_nos':0, 'neuralgia_neuritis_and_radiculitis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1028], 'phecode_766'))
neurofibromatosis_phe = TensorMap('phecode_199.4', Interpretation.CATEGORICAL, channel_map={'no_neurofibromatosis':0, 'neurofibromatosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3875], 'phecode_199_4'))
neurological_disorders_phe = TensorMap('phecode_292', Interpretation.CATEGORICAL, channel_map={'no_neurological_disorders':0, 'neurological_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 597], 'phecode_292'))
neutropenia_phe = TensorMap('phecode_288.11', Interpretation.CATEGORICAL, channel_map={'no_neutropenia':0, 'neutropenia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 101], 'phecode_288_11'))
nevus_non_neoplastic_phe = TensorMap('phecode_217.1', Interpretation.CATEGORICAL, channel_map={'no_nevus_non_neoplastic':0, 'nevus_non_neoplastic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 628], 'phecode_217_1'))
nodular_lymphoma_phe = TensorMap('phecode_202.21', Interpretation.CATEGORICAL, channel_map={'no_nodular_lymphoma':0, 'nodular_lymphoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 948], 'phecode_202_21'))
non_autoimmune_hemolytic_anemias_phe = TensorMap('phecode_283.2', Interpretation.CATEGORICAL, channel_map={'no_non_autoimmune_hemolytic_anemias':0, 'non_autoimmune_hemolytic_anemias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25000], 'phecode_283_2'))
non_hodgkins_lymphoma_phe = TensorMap('phecode_202.2', Interpretation.CATEGORICAL, channel_map={'no_non_hodgkins_lymphoma':0, 'non_hodgkins_lymphoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 98], 'phecode_202_2'))
non_proliferative_glomerulonephritis_phe = TensorMap('phecode_580.12', Interpretation.CATEGORICAL, channel_map={'no_non_proliferative_glomerulonephritis':0, 'non_proliferative_glomerulonephritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2358], 'phecode_580_12'))
nonallopathic_lesions_nec_phe = TensorMap('phecode_769', Interpretation.CATEGORICAL, channel_map={'no_nonallopathic_lesions_nec':0, 'nonallopathic_lesions_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_769'))
noninfectious_dermatoses_of_eyelid_phe = TensorMap('phecode_371.33', Interpretation.CATEGORICAL, channel_map={'no_noninfectious_dermatoses_of_eyelid':0, 'noninfectious_dermatoses_of_eyelid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23809], 'phecode_371_33'))
noninfectious_disorders_of_lymphatic_channels_phe = TensorMap('phecode_450', Interpretation.CATEGORICAL, channel_map={'no_noninfectious_disorders_of_lymphatic_channels':0, 'noninfectious_disorders_of_lymphatic_channels':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 411], 'phecode_450'))
noninfectious_gastroenteritis_phe = TensorMap('phecode_558', Interpretation.CATEGORICAL, channel_map={'no_noninfectious_gastroenteritis':0, 'noninfectious_gastroenteritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19], 'phecode_558'))
noninflammatory_disorders_of_cervix_phe = TensorMap('phecode_619.3', Interpretation.CATEGORICAL, channel_map={'no_noninflammatory_disorders_of_cervix':0, 'noninflammatory_disorders_of_cervix':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 142], 'phecode_619_3'))
noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament_phe = TensorMap('phecode_619.1', Interpretation.CATEGORICAL, channel_map={'no_noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament':0, 'noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 502], 'phecode_619_1'))
noninflammatory_disorders_of_vagina_phe = TensorMap('phecode_619.4', Interpretation.CATEGORICAL, channel_map={'no_noninflammatory_disorders_of_vagina':0, 'noninflammatory_disorders_of_vagina':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 240], 'phecode_619_4'))
noninflammatory_disorders_of_vulva_and_perineum_phe = TensorMap('phecode_619.5', Interpretation.CATEGORICAL, channel_map={'no_noninflammatory_disorders_of_vulva_and_perineum':0, 'noninflammatory_disorders_of_vulva_and_perineum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 279], 'phecode_619_5'))
noninflammatory_female_genital_disorders_phe = TensorMap('phecode_619', Interpretation.CATEGORICAL, channel_map={'no_noninflammatory_female_genital_disorders':0, 'noninflammatory_female_genital_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7], 'phecode_619'))
nonrheumatic_aortic_valve_disorders_phe = TensorMap('phecode_395.2', Interpretation.CATEGORICAL, channel_map={'no_nonrheumatic_aortic_valve_disorders':0, 'nonrheumatic_aortic_valve_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1237], 'phecode_395_2'))
nonrheumatic_mitral_valve_disorders_phe = TensorMap('phecode_395.1', Interpretation.CATEGORICAL, channel_map={'no_nonrheumatic_mitral_valve_disorders':0, 'nonrheumatic_mitral_valve_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 104], 'phecode_395_1'))
nonrheumatic_pulmonary_valve_disorders_phe = TensorMap('phecode_395.4', Interpretation.CATEGORICAL, channel_map={'no_nonrheumatic_pulmonary_valve_disorders':0, 'nonrheumatic_pulmonary_valve_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2551], 'phecode_395_4'))
nonrheumatic_tricuspid_valve_disorders_phe = TensorMap('phecode_395.3', Interpretation.CATEGORICAL, channel_map={'no_nonrheumatic_tricuspid_valve_disorders':0, 'nonrheumatic_tricuspid_valve_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2525], 'phecode_395_3'))
nonsenile_cataract_phe = TensorMap('phecode_366.1', Interpretation.CATEGORICAL, channel_map={'no_nonsenile_cataract':0, 'nonsenile_cataract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4424], 'phecode_366_1'))
nonspecific_abnormal_findings_in_cerebrospinal_fluid_phe = TensorMap('phecode_346.3', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_findings_in_cerebrospinal_fluid':0, 'nonspecific_abnormal_findings_in_cerebrospinal_fluid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 186], 'phecode_346_3'))
nonspecific_abnormal_findings_in_stool_contents_phe = TensorMap('phecode_579.8', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_findings_in_stool_contents':0, 'nonspecific_abnormal_findings_in_stool_contents':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 173], 'phecode_579_8'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract_phe = TensorMap('phecode_575.9', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 892], 'phecode_575_9'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system_phe = TensorMap('phecode_793', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2283], 'phecode_793'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc_phe = TensorMap('phecode_793.2', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 548], 'phecode_793_2'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head_phe = TensorMap('phecode_346.1', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2232], 'phecode_346_1'))
nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system_phe = TensorMap('phecode_346.2', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system':0, 'nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16666], 'phecode_346_2'))
nonspecific_abnormal_results_of_other_endocrine_function_study_phe = TensorMap('phecode_259.1', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_abnormal_results_of_other_endocrine_function_study':0, 'nonspecific_abnormal_results_of_other_endocrine_function_study':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11627], 'phecode_259_1'))
nonspecific_chest_pain_phe = TensorMap('phecode_418', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_chest_pain':0, 'nonspecific_chest_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_418'))
nonspecific_findings_on_examination_of_blood_phe = TensorMap('phecode_790', Interpretation.CATEGORICAL, channel_map={'no_nonspecific_findings_on_examination_of_blood':0, 'nonspecific_findings_on_examination_of_blood':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1851], 'phecode_790'))
nontoxic_multinodular_goiter_phe = TensorMap('phecode_241.2', Interpretation.CATEGORICAL, channel_map={'no_nontoxic_multinodular_goiter':0, 'nontoxic_multinodular_goiter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 499], 'phecode_241_2'))
nontoxic_nodular_goiter_phe = TensorMap('phecode_241', Interpretation.CATEGORICAL, channel_map={'no_nontoxic_nodular_goiter':0, 'nontoxic_nodular_goiter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_241'))
nontoxic_uninodular_goiter_phe = TensorMap('phecode_241.1', Interpretation.CATEGORICAL, channel_map={'no_nontoxic_uninodular_goiter':0, 'nontoxic_uninodular_goiter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 677], 'phecode_241_1'))
normal_delivery_phe = TensorMap('phecode_650', Interpretation.CATEGORICAL, channel_map={'no_normal_delivery':0, 'normal_delivery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 189], 'phecode_650'))
nutritional_marasmus_phe = TensorMap('phecode_260.22', Interpretation.CATEGORICAL, channel_map={'no_nutritional_marasmus':0, 'nutritional_marasmus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_260_22'))
nystagmus_and_other_irregular_eye_movements_phe = TensorMap('phecode_378.2', Interpretation.CATEGORICAL, channel_map={'no_nystagmus_and_other_irregular_eye_movements':0, 'nystagmus_and_other_irregular_eye_movements':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1479], 'phecode_378_2'))
obesity_phe = TensorMap('phecode_278.1', Interpretation.CATEGORICAL, channel_map={'no_obesity':0, 'obesity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 26], 'phecode_278_1'))
obsessive_compulsive_disorders_phe = TensorMap('phecode_300.3', Interpretation.CATEGORICAL, channel_map={'no_obsessive_compulsive_disorders':0, 'obsessive_compulsive_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2212], 'phecode_300_3'))
obstetrical_birth_trauma_phe = TensorMap('phecode_665', Interpretation.CATEGORICAL, channel_map={'no_obstetrical_birth_trauma':0, 'obstetrical_birth_trauma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 72], 'phecode_665'))
obstruction_of_bile_duct_phe = TensorMap('phecode_575.2', Interpretation.CATEGORICAL, channel_map={'no_obstruction_of_bile_duct':0, 'obstruction_of_bile_duct':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 400], 'phecode_575_2'))
obstructive_chronic_bronchitis_phe = TensorMap('phecode_496.21', Interpretation.CATEGORICAL, channel_map={'no_obstructive_chronic_bronchitis':0, 'obstructive_chronic_bronchitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 112], 'phecode_496_21'))
obstructive_genitourinary_defect_phe = TensorMap('phecode_751.3', Interpretation.CATEGORICAL, channel_map={'no_obstructive_genitourinary_defect':0, 'obstructive_genitourinary_defect':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18518], 'phecode_751_3'))
occlusion_and_stenosis_of_precerebral_arteries_phe = TensorMap('phecode_433.1', Interpretation.CATEGORICAL, channel_map={'no_occlusion_and_stenosis_of_precerebral_arteries':0, 'occlusion_and_stenosis_of_precerebral_arteries':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 327], 'phecode_433_1'))
occlusion_of_cerebral_arteries_phe = TensorMap('phecode_433.2', Interpretation.CATEGORICAL, channel_map={'no_occlusion_of_cerebral_arteries':0, 'occlusion_of_cerebral_arteries':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 110], 'phecode_433_2'))
occlusion_of_cerebral_arteries_with_cerebral_infarction_phe = TensorMap('phecode_433.11', Interpretation.CATEGORICAL, channel_map={'no_occlusion_of_cerebral_arteries_with_cerebral_infarction':0, 'occlusion_of_cerebral_arteries_with_cerebral_infarction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2369], 'phecode_433_11'))
oliguria_and_anuria_phe = TensorMap('phecode_599.6', Interpretation.CATEGORICAL, channel_map={'no_oliguria_and_anuria':0, 'oliguria_and_anuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2808], 'phecode_599_6'))
open_wound_of_ear_phe = TensorMap('phecode_870.2', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_ear':0, 'open_wound_of_ear':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2049], 'phecode_870_2'))
open_wound_of_foot_except_toe_s_alone_phe = TensorMap('phecode_871.3', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_foot_except_toe_s_alone':0, 'open_wound_of_foot_except_toe_s_alone':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2074], 'phecode_871_3'))
open_wound_of_genital_organs_phe = TensorMap('phecode_870.8', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_genital_organs':0, 'open_wound_of_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6410], 'phecode_870_8'))
open_wound_of_lip_and_mouth_phe = TensorMap('phecode_870.5', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_lip_and_mouth':0, 'open_wound_of_lip_and_mouth':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 909], 'phecode_870_5'))
open_wound_of_neck_phe = TensorMap('phecode_870.6', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_neck':0, 'open_wound_of_neck':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2941], 'phecode_870_6'))
open_wound_of_nose_and_sinus_phe = TensorMap('phecode_870.4', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_nose_and_sinus':0, 'open_wound_of_nose_and_sinus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1385], 'phecode_870_4'))
open_wound_of_toe_s_phe = TensorMap('phecode_871.4', Interpretation.CATEGORICAL, channel_map={'no_open_wound_of_toe_s':0, 'open_wound_of_toe_s':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4464], 'phecode_871_4'))
open_wound_or_laceration_of_eye_or_eyelid_phe = TensorMap('phecode_870.1', Interpretation.CATEGORICAL, channel_map={'no_open_wound_or_laceration_of_eye_or_eyelid':0, 'open_wound_or_laceration_of_eye_or_eyelid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 666], 'phecode_870_1'))
open_wounds_of_extremities_phe = TensorMap('phecode_871', Interpretation.CATEGORICAL, channel_map={'no_open_wounds_of_extremities':0, 'open_wounds_of_extremities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 84], 'phecode_871'))
open_wounds_of_head_neck_and_trunk_phe = TensorMap('phecode_870', Interpretation.CATEGORICAL, channel_map={'no_open_wounds_of_head_neck_and_trunk':0, 'open_wounds_of_head_neck_and_trunk':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2000], 'phecode_870'))
opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use_phe = TensorMap('phecode_965.1', Interpretation.CATEGORICAL, channel_map={'no_opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use':0, 'opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 251], 'phecode_965_1'))
optic_atrophy_phe = TensorMap('phecode_377.1', Interpretation.CATEGORICAL, channel_map={'no_optic_atrophy':0, 'optic_atrophy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5102], 'phecode_377_1'))
optic_neuritis_neuropathy_phe = TensorMap('phecode_377.3', Interpretation.CATEGORICAL, channel_map={'no_optic_neuritis_neuropathy':0, 'optic_neuritis_neuropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1915], 'phecode_377_3'))
oral_aphthae_phe = TensorMap('phecode_528.12', Interpretation.CATEGORICAL, channel_map={'no_oral_aphthae':0, 'oral_aphthae':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4587], 'phecode_528_12'))
organic_or_persistent_insomnia_phe = TensorMap('phecode_327.41', Interpretation.CATEGORICAL, channel_map={'no_organic_or_persistent_insomnia':0, 'organic_or_persistent_insomnia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1945], 'phecode_327_41'))
orthostatic_hypotension_phe = TensorMap('phecode_458.1', Interpretation.CATEGORICAL, channel_map={'no_orthostatic_hypotension':0, 'orthostatic_hypotension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 205], 'phecode_458_1'))
osteitis_deformans_and_osteopathies_associated_with_other_disorders_classified_elsewhere_phe = TensorMap('phecode_731', Interpretation.CATEGORICAL, channel_map={'no_osteitis_deformans_and_osteopathies_associated_with_other_disorders_classified_elsewhere':0, 'osteitis_deformans_and_osteopathies_associated_with_other_disorders_classified_elsewhere':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_731'))
osteitis_deformans_pagets_disease_of_bone_phe = TensorMap('phecode_731.1', Interpretation.CATEGORICAL, channel_map={'no_osteitis_deformans_pagets_disease_of_bone':0, 'osteitis_deformans_pagets_disease_of_bone':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6329], 'phecode_731_1'))
osteoarthritis_localized_phe = TensorMap('phecode_740.1', Interpretation.CATEGORICAL, channel_map={'no_osteoarthritis_localized':0, 'osteoarthritis_localized':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35], 'phecode_740_1'))
osteoarthrosis_generalized_phe = TensorMap('phecode_740.2', Interpretation.CATEGORICAL, channel_map={'no_osteoarthrosis_generalized':0, 'osteoarthrosis_generalized':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 664], 'phecode_740_2'))
osteoarthrosis_localized_primary_phe = TensorMap('phecode_740.11', Interpretation.CATEGORICAL, channel_map={'no_osteoarthrosis_localized_primary':0, 'osteoarthrosis_localized_primary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 42], 'phecode_740_11'))
osteoarthrosis_localized_secondary_phe = TensorMap('phecode_740.12', Interpretation.CATEGORICAL, channel_map={'no_osteoarthrosis_localized_secondary':0, 'osteoarthrosis_localized_secondary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1347], 'phecode_740_12'))
osteoarthrosis_nos_phe = TensorMap('phecode_740.9', Interpretation.CATEGORICAL, channel_map={'no_osteoarthrosis_nos':0, 'osteoarthrosis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 59], 'phecode_740_9'))
osteochondritis_dissecans_phe = TensorMap('phecode_732.7', Interpretation.CATEGORICAL, channel_map={'no_osteochondritis_dissecans':0, 'osteochondritis_dissecans':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7042], 'phecode_732_7'))
osteochondropathies_phe = TensorMap('phecode_732', Interpretation.CATEGORICAL, channel_map={'no_osteochondropathies':0, 'osteochondropathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1628], 'phecode_732'))
osteomyelitis_periostitis_and_other_infections_involving_bone_phe = TensorMap('phecode_710', Interpretation.CATEGORICAL, channel_map={'no_osteomyelitis_periostitis_and_other_infections_involving_bone':0, 'osteomyelitis_periostitis_and_other_infections_involving_bone':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_710'))
osteopathy_resulting_from_poliomyelitis_phe = TensorMap('phecode_710.3', Interpretation.CATEGORICAL, channel_map={'no_osteopathy_resulting_from_poliomyelitis':0, 'osteopathy_resulting_from_poliomyelitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_710_3'))
osteopenia_or_other_disorder_of_bone_and_cartilage_phe = TensorMap('phecode_743.9', Interpretation.CATEGORICAL, channel_map={'no_osteopenia_or_other_disorder_of_bone_and_cartilage':0, 'osteopenia_or_other_disorder_of_bone_and_cartilage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 67], 'phecode_743_9'))
osteoporosis_phe = TensorMap('phecode_743.1', Interpretation.CATEGORICAL, channel_map={'no_osteoporosis':0, 'osteoporosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6024], 'phecode_743_1'))
osteoporosis_nos_phe = TensorMap('phecode_743.11', Interpretation.CATEGORICAL, channel_map={'no_osteoporosis_nos':0, 'osteoporosis_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 84], 'phecode_743_11'))
otalgia_phe = TensorMap('phecode_382', Interpretation.CATEGORICAL, channel_map={'no_otalgia':0, 'otalgia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1213], 'phecode_382'))
other_abnormal_blood_chemistry_phe = TensorMap('phecode_790.6', Interpretation.CATEGORICAL, channel_map={'no_other_abnormal_blood_chemistry':0, 'other_abnormal_blood_chemistry':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 49], 'phecode_790_6'))
other_abnormal_glucose_phe = TensorMap('phecode_250.42', Interpretation.CATEGORICAL, channel_map={'no_other_abnormal_glucose':0, 'other_abnormal_glucose':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 754], 'phecode_250_42'))
other_abnormality_of_urination_phe = TensorMap('phecode_599.9', Interpretation.CATEGORICAL, channel_map={'no_other_abnormality_of_urination':0, 'other_abnormality_of_urination':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 204], 'phecode_599_9'))
other_acquired_deformities_of_limbs_phe = TensorMap('phecode_736', Interpretation.CATEGORICAL, channel_map={'no_other_acquired_deformities_of_limbs':0, 'other_acquired_deformities_of_limbs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1275], 'phecode_736'))
other_acquired_musculoskeletal_deformity_phe = TensorMap('phecode_738', Interpretation.CATEGORICAL, channel_map={'no_other_acquired_musculoskeletal_deformity':0, 'other_acquired_musculoskeletal_deformity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 431], 'phecode_738'))
other_acute_and_subacute_forms_of_ischemic_heart_disease_phe = TensorMap('phecode_411.9', Interpretation.CATEGORICAL, channel_map={'no_other_acute_and_subacute_forms_of_ischemic_heart_disease':0, 'other_acute_and_subacute_forms_of_ischemic_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 247], 'phecode_411_9'))
other_allied_disorders_of_spine_phe = TensorMap('phecode_721.8', Interpretation.CATEGORICAL, channel_map={'no_other_allied_disorders_of_spine':0, 'other_allied_disorders_of_spine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19], 'phecode_721_8'))
other_alveolar_and_parietoalveolar_pneumonopathy_phe = TensorMap('phecode_504', Interpretation.CATEGORICAL, channel_map={'no_other_alveolar_and_parietoalveolar_pneumonopathy':0, 'other_alveolar_and_parietoalveolar_pneumonopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 929], 'phecode_504'))
other_and_unspecified_coagulation_defects_phe = TensorMap('phecode_286.7', Interpretation.CATEGORICAL, channel_map={'no_other_and_unspecified_coagulation_defects':0, 'other_and_unspecified_coagulation_defects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 875], 'phecode_286_7'))
other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother_phe = TensorMap('phecode_654', Interpretation.CATEGORICAL, channel_map={'no_other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother':0, 'other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2994], 'phecode_654'))
other_and_unspecified_congenital_anomalies_phe = TensorMap('phecode_759', Interpretation.CATEGORICAL, channel_map={'no_other_and_unspecified_congenital_anomalies':0, 'other_and_unspecified_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 551], 'phecode_759'))
other_and_unspecified_disc_disorder_phe = TensorMap('phecode_722.9', Interpretation.CATEGORICAL, channel_map={'no_other_and_unspecified_disc_disorder':0, 'other_and_unspecified_disc_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 86], 'phecode_722_9'))
other_and_unspecified_disorders_of_the_nervous_system_phe = TensorMap('phecode_349', Interpretation.CATEGORICAL, channel_map={'no_other_and_unspecified_disorders_of_the_nervous_system':0, 'other_and_unspecified_disorders_of_the_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8], 'phecode_349'))
other_anemias_phe = TensorMap('phecode_285', Interpretation.CATEGORICAL, channel_map={'no_other_anemias':0, 'other_anemias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27], 'phecode_285'))
other_aneurysm_phe = TensorMap('phecode_442', Interpretation.CATEGORICAL, channel_map={'no_other_aneurysm':0, 'other_aneurysm':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10869], 'phecode_442'))
other_arthropathies_phe = TensorMap('phecode_716', Interpretation.CATEGORICAL, channel_map={'no_other_arthropathies':0, 'other_arthropathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5102], 'phecode_716'))
other_benign_neoplasm_of_connective_and_other_soft_tissue_phe = TensorMap('phecode_215', Interpretation.CATEGORICAL, channel_map={'no_other_benign_neoplasm_of_connective_and_other_soft_tissue':0, 'other_benign_neoplasm_of_connective_and_other_soft_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 343], 'phecode_215'))
other_benign_neoplasm_of_uterus_phe = TensorMap('phecode_218.2', Interpretation.CATEGORICAL, channel_map={'no_other_benign_neoplasm_of_uterus':0, 'other_benign_neoplasm_of_uterus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1054], 'phecode_218_2'))
other_cardiac_conduction_disorders_phe = TensorMap('phecode_426.8', Interpretation.CATEGORICAL, channel_map={'no_other_cardiac_conduction_disorders':0, 'other_cardiac_conduction_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4950], 'phecode_426_8'))
other_cardiomyopathy_phe = TensorMap('phecode_425.8', Interpretation.CATEGORICAL, channel_map={'no_other_cardiomyopathy':0, 'other_cardiomyopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6097], 'phecode_425_8'))
other_cerebral_degenerations_phe = TensorMap('phecode_331', Interpretation.CATEGORICAL, channel_map={'no_other_cerebral_degenerations':0, 'other_cerebral_degenerations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1213], 'phecode_331'))
other_chronic_ischemic_heart_disease_unspecified_phe = TensorMap('phecode_411.8', Interpretation.CATEGORICAL, channel_map={'no_other_chronic_ischemic_heart_disease_unspecified':0, 'other_chronic_ischemic_heart_disease_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11], 'phecode_411_8'))
other_chronic_nonalcoholic_liver_disease_phe = TensorMap('phecode_571.5', Interpretation.CATEGORICAL, channel_map={'no_other_chronic_nonalcoholic_liver_disease':0, 'other_chronic_nonalcoholic_liver_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 174], 'phecode_571_5'))
other_cns_infection_and_poliomyelitis_phe = TensorMap('phecode_324', Interpretation.CATEGORICAL, channel_map={'no_other_cns_infection_and_poliomyelitis':0, 'other_cns_infection_and_poliomyelitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45], 'phecode_324'))
other_complications_of_pregnancy_nec_phe = TensorMap('phecode_646', Interpretation.CATEGORICAL, channel_map={'no_other_complications_of_pregnancy_nec':0, 'other_complications_of_pregnancy_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 156], 'phecode_646'))
other_complications_of_the_puerperium_nec_phe = TensorMap('phecode_674', Interpretation.CATEGORICAL, channel_map={'no_other_complications_of_the_puerperium_nec':0, 'other_complications_of_the_puerperium_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1377], 'phecode_674'))
other_conditions_of_brain_phe = TensorMap('phecode_348', Interpretation.CATEGORICAL, channel_map={'no_other_conditions_of_brain':0, 'other_conditions_of_brain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2173], 'phecode_348'))
other_conditions_of_brain_nos_phe = TensorMap('phecode_348.9', Interpretation.CATEGORICAL, channel_map={'no_other_conditions_of_brain_nos':0, 'other_conditions_of_brain_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 778], 'phecode_348_9'))
other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium_phe = TensorMap('phecode_649', Interpretation.CATEGORICAL, channel_map={'no_other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium':0, 'other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3225], 'phecode_649'))
other_congenital_anomalies_of_lower_limb_including_pelvic_girdle_phe = TensorMap('phecode_755.6', Interpretation.CATEGORICAL, channel_map={'no_other_congenital_anomalies_of_lower_limb_including_pelvic_girdle':0, 'other_congenital_anomalies_of_lower_limb_including_pelvic_girdle':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6578], 'phecode_755_6'))
other_congenital_musculoskeletal_anomalies_phe = TensorMap('phecode_756', Interpretation.CATEGORICAL, channel_map={'no_other_congenital_musculoskeletal_anomalies':0, 'other_congenital_musculoskeletal_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 442], 'phecode_756'))
other_deficiency_anemia_phe = TensorMap('phecode_281', Interpretation.CATEGORICAL, channel_map={'no_other_deficiency_anemia':0, 'other_deficiency_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12], 'phecode_281'))
other_degenerative_diseases_of_the_basal_ganglia_phe = TensorMap('phecode_333.8', Interpretation.CATEGORICAL, channel_map={'no_other_degenerative_diseases_of_the_basal_ganglia':0, 'other_degenerative_diseases_of_the_basal_ganglia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5882], 'phecode_333_8'))
other_demyelinating_diseases_of_central_nervous_system_phe = TensorMap('phecode_341', Interpretation.CATEGORICAL, channel_map={'no_other_demyelinating_diseases_of_central_nervous_system':0, 'other_demyelinating_diseases_of_central_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 245], 'phecode_341'))
other_derangement_of_joint_phe = TensorMap('phecode_742.9', Interpretation.CATEGORICAL, channel_map={'no_other_derangement_of_joint':0, 'other_derangement_of_joint':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19], 'phecode_742_9'))
other_diseases_of_blood_and_blood_forming_organs_phe = TensorMap('phecode_289', Interpretation.CATEGORICAL, channel_map={'no_other_diseases_of_blood_and_blood_forming_organs':0, 'other_diseases_of_blood_and_blood_forming_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 68], 'phecode_289'))
other_diseases_of_lung_phe = TensorMap('phecode_510', Interpretation.CATEGORICAL, channel_map={'no_other_diseases_of_lung':0, 'other_diseases_of_lung':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 339], 'phecode_510'))
other_diseases_of_respiratory_system_nec_phe = TensorMap('phecode_519.8', Interpretation.CATEGORICAL, channel_map={'no_other_diseases_of_respiratory_system_nec':0, 'other_diseases_of_respiratory_system_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 36], 'phecode_519_8'))
other_diseases_of_respiratory_system_not_elsewhere_classified_phe = TensorMap('phecode_519', Interpretation.CATEGORICAL, channel_map={'no_other_diseases_of_respiratory_system_not_elsewhere_classified':0, 'other_diseases_of_respiratory_system_not_elsewhere_classified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5], 'phecode_519'))
other_diseases_of_the_teeth_and_supporting_structures_phe = TensorMap('phecode_525', Interpretation.CATEGORICAL, channel_map={'no_other_diseases_of_the_teeth_and_supporting_structures':0, 'other_diseases_of_the_teeth_and_supporting_structures':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 131], 'phecode_525'))
other_disorders_of_arteries_and_arterioles_phe = TensorMap('phecode_447', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_arteries_and_arterioles':0, 'other_disorders_of_arteries_and_arterioles':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 603], 'phecode_447'))
other_disorders_of_biliary_tract_phe = TensorMap('phecode_575.8', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_biliary_tract':0, 'other_disorders_of_biliary_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 315], 'phecode_575_8'))
other_disorders_of_bladder_phe = TensorMap('phecode_596', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_bladder':0, 'other_disorders_of_bladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 47], 'phecode_596'))
other_disorders_of_bone_and_cartilage_phe = TensorMap('phecode_733', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_bone_and_cartilage':0, 'other_disorders_of_bone_and_cartilage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 420], 'phecode_733'))
other_disorders_of_carbohydrate_transport_and_metabolism_phe = TensorMap('phecode_271.9', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_carbohydrate_transport_and_metabolism':0, 'other_disorders_of_carbohydrate_transport_and_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13888], 'phecode_271_9'))
other_disorders_of_cervical_region_phe = TensorMap('phecode_723', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_cervical_region':0, 'other_disorders_of_cervical_region':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 833], 'phecode_723'))
other_disorders_of_circulatory_system_phe = TensorMap('phecode_459', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_circulatory_system':0, 'other_disorders_of_circulatory_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2450], 'phecode_459'))
other_disorders_of_ear_phe = TensorMap('phecode_388', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_ear':0, 'other_disorders_of_ear':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 34], 'phecode_388'))
other_disorders_of_eye_phe = TensorMap('phecode_379', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_eye':0, 'other_disorders_of_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9], 'phecode_379'))
other_disorders_of_eyelids_phe = TensorMap('phecode_374', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_eyelids':0, 'other_disorders_of_eyelids':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 119], 'phecode_374'))
other_disorders_of_gallbladder_phe = TensorMap('phecode_575.7', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_gallbladder':0, 'other_disorders_of_gallbladder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 246], 'phecode_575_7'))
other_disorders_of_intestine_phe = TensorMap('phecode_569', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_intestine':0, 'other_disorders_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2], 'phecode_569'))
other_disorders_of_lipoid_metabolism_phe = TensorMap('phecode_277.5', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_lipoid_metabolism':0, 'other_disorders_of_lipoid_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1612], 'phecode_277_5'))
other_disorders_of_liver_phe = TensorMap('phecode_573', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_liver':0, 'other_disorders_of_liver':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 61], 'phecode_573'))
other_disorders_of_male_genital_organs_phe = TensorMap('phecode_608', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_male_genital_organs':0, 'other_disorders_of_male_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18], 'phecode_608'))
other_disorders_of_metabolism_phe = TensorMap('phecode_277', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_metabolism':0, 'other_disorders_of_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7], 'phecode_277'))
other_disorders_of_middle_ear_and_mastoid_phe = TensorMap('phecode_385', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_middle_ear_and_mastoid':0, 'other_disorders_of_middle_ear_and_mastoid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2631], 'phecode_385'))
other_disorders_of_neurohypophysis_phe = TensorMap('phecode_253.7', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_neurohypophysis':0, 'other_disorders_of_neurohypophysis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1831], 'phecode_253_7'))
other_disorders_of_pancreatic_internal_secretion_phe = TensorMap('phecode_251', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_pancreatic_internal_secretion':0, 'other_disorders_of_pancreatic_internal_secretion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_251'))
other_disorders_of_peritoneum_phe = TensorMap('phecode_568', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_peritoneum':0, 'other_disorders_of_peritoneum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 82], 'phecode_568'))
other_disorders_of_prostate_phe = TensorMap('phecode_602', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_prostate':0, 'other_disorders_of_prostate':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 266], 'phecode_602'))
other_disorders_of_purine_and_pyrimidine_metabolism_phe = TensorMap('phecode_277.2', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_purine_and_pyrimidine_metabolism':0, 'other_disorders_of_purine_and_pyrimidine_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_277_2'))
other_disorders_of_soft_tissues_phe = TensorMap('phecode_729', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_soft_tissues':0, 'other_disorders_of_soft_tissues':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15], 'phecode_729'))
other_disorders_of_stomach_and_duodenum_phe = TensorMap('phecode_537', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_stomach_and_duodenum':0, 'other_disorders_of_stomach_and_duodenum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 97], 'phecode_537'))
other_disorders_of_synovium_tendon_and_bursa_phe = TensorMap('phecode_727', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_synovium_tendon_and_bursa':0, 'other_disorders_of_synovium_tendon_and_bursa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 57], 'phecode_727'))
other_disorders_of_testis_phe = TensorMap('phecode_603', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_testis':0, 'other_disorders_of_testis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 399], 'phecode_603'))
other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation_phe = TensorMap('phecode_676', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation':0, 'other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6172], 'phecode_676'))
other_disorders_of_the_kidney_and_ureters_phe = TensorMap('phecode_586', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_the_kidney_and_ureters':0, 'other_disorders_of_the_kidney_and_ureters':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 119], 'phecode_586'))
other_disorders_of_thyroid_phe = TensorMap('phecode_246', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_thyroid':0, 'other_disorders_of_thyroid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 451], 'phecode_246'))
other_disorders_of_tympanic_membrane_phe = TensorMap('phecode_384', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_tympanic_membrane':0, 'other_disorders_of_tympanic_membrane':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1028], 'phecode_384'))
other_disorders_of_urethra_and_urinary_tract_phe = TensorMap('phecode_597', Interpretation.CATEGORICAL, channel_map={'no_other_disorders_of_urethra_and_urinary_tract':0, 'other_disorders_of_urethra_and_urinary_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 465], 'phecode_597'))
other_dyschromia_phe = TensorMap('phecode_694.2', Interpretation.CATEGORICAL, channel_map={'no_other_dyschromia':0, 'other_dyschromia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 438], 'phecode_694_2'))
other_dyspnea_phe = TensorMap('phecode_512.9', Interpretation.CATEGORICAL, channel_map={'no_other_dyspnea':0, 'other_dyspnea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 378], 'phecode_512_9'))
other_endocrine_disorders_phe = TensorMap('phecode_259', Interpretation.CATEGORICAL, channel_map={'no_other_endocrine_disorders':0, 'other_endocrine_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2762], 'phecode_259'))
other_forms_of_chronic_heart_disease_phe = TensorMap('phecode_414', Interpretation.CATEGORICAL, channel_map={'no_other_forms_of_chronic_heart_disease':0, 'other_forms_of_chronic_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 154], 'phecode_414'))
other_headache_syndromes_phe = TensorMap('phecode_339', Interpretation.CATEGORICAL, channel_map={'no_other_headache_syndromes':0, 'other_headache_syndromes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 44], 'phecode_339'))
other_heart_block_phe = TensorMap('phecode_426.25', Interpretation.CATEGORICAL, channel_map={'no_other_heart_block':0, 'other_heart_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2702], 'phecode_426_25'))
other_hemoglobinopathies_phe = TensorMap('phecode_282.8', Interpretation.CATEGORICAL, channel_map={'no_other_hemoglobinopathies':0, 'other_hemoglobinopathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1362], 'phecode_282_8'))
other_hereditary_hemolytic_anemias_phe = TensorMap('phecode_282.9', Interpretation.CATEGORICAL, channel_map={'no_other_hereditary_hemolytic_anemias':0, 'other_hereditary_hemolytic_anemias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 432], 'phecode_282_9'))
other_high_risk_pregnancy_phe = TensorMap('phecode_638', Interpretation.CATEGORICAL, channel_map={'no_other_high_risk_pregnancy':0, 'other_high_risk_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5494], 'phecode_638'))
other_hypertensive_complications_phe = TensorMap('phecode_401.3', Interpretation.CATEGORICAL, channel_map={'no_other_hypertensive_complications':0, 'other_hypertensive_complications':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2403], 'phecode_401_3'))
other_hypertrophic_and_atrophic_conditions_of_skin_phe = TensorMap('phecode_701', Interpretation.CATEGORICAL, channel_map={'no_other_hypertrophic_and_atrophic_conditions_of_skin':0, 'other_hypertrophic_and_atrophic_conditions_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 218], 'phecode_701'))
other_hypertrophic_cardiomyopathy_phe = TensorMap('phecode_425.12', Interpretation.CATEGORICAL, channel_map={'no_other_hypertrophic_cardiomyopathy':0, 'other_hypertrophic_cardiomyopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1730], 'phecode_425_12'))
other_ill_defined_and_unknown_causes_of_morbidity_and_mortality_phe = TensorMap('phecode_1019', Interpretation.CATEGORICAL, channel_map={'no_other_ill_defined_and_unknown_causes_of_morbidity_and_mortality':0, 'other_ill_defined_and_unknown_causes_of_morbidity_and_mortality':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 24], 'phecode_1019'))
other_immunological_findings_phe = TensorMap('phecode_279.7', Interpretation.CATEGORICAL, channel_map={'no_other_immunological_findings':0, 'other_immunological_findings':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1436], 'phecode_279_7'))
other_infectious_and_parasitic_diseases_phe = TensorMap('phecode_136', Interpretation.CATEGORICAL, channel_map={'no_other_infectious_and_parasitic_diseases':0, 'other_infectious_and_parasitic_diseases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 628], 'phecode_136'))
other_inflammatory_disorders_of_male_genital_organs_phe = TensorMap('phecode_601.8', Interpretation.CATEGORICAL, channel_map={'no_other_inflammatory_disorders_of_male_genital_organs':0, 'other_inflammatory_disorders_of_male_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1278], 'phecode_601_8'))
other_inflammatory_spondylopathies_phe = TensorMap('phecode_715', Interpretation.CATEGORICAL, channel_map={'no_other_inflammatory_spondylopathies':0, 'other_inflammatory_spondylopathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 931], 'phecode_715'))
other_intestinal_obstruction_phe = TensorMap('phecode_560.4', Interpretation.CATEGORICAL, channel_map={'no_other_intestinal_obstruction':0, 'other_intestinal_obstruction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 101], 'phecode_560_4'))
other_local_infections_of_skin_and_subcutaneous_tissue_phe = TensorMap('phecode_686', Interpretation.CATEGORICAL, channel_map={'no_other_local_infections_of_skin_and_subcutaneous_tissue':0, 'other_local_infections_of_skin_and_subcutaneous_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 66], 'phecode_686'))
other_mental_disorder_phe = TensorMap('phecode_306', Interpretation.CATEGORICAL, channel_map={'no_other_mental_disorder':0, 'other_mental_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2], 'phecode_306'))
other_non_epithelial_cancer_of_skin_phe = TensorMap('phecode_172.2', Interpretation.CATEGORICAL, channel_map={'no_other_non_epithelial_cancer_of_skin':0, 'other_non_epithelial_cancer_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 31], 'phecode_172_2'))
other_nondiabetic_retinopathy_phe = TensorMap('phecode_362.3', Interpretation.CATEGORICAL, channel_map={'no_other_nondiabetic_retinopathy':0, 'other_nondiabetic_retinopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4629], 'phecode_362_3'))
other_nonmalignant_breast_conditions_phe = TensorMap('phecode_613', Interpretation.CATEGORICAL, channel_map={'no_other_nonmalignant_breast_conditions':0, 'other_nonmalignant_breast_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38461], 'phecode_613'))
other_nonspecific_findings_on_examination_of_urine_phe = TensorMap('phecode_598.9', Interpretation.CATEGORICAL, channel_map={'no_other_nonspecific_findings_on_examination_of_urine':0, 'other_nonspecific_findings_on_examination_of_urine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1098], 'phecode_598_9'))
other_nutritional_deficiency_phe = TensorMap('phecode_263', Interpretation.CATEGORICAL, channel_map={'no_other_nutritional_deficiency':0, 'other_nutritional_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 147], 'phecode_263'))
other_open_wound_of_head_and_face_phe = TensorMap('phecode_870.3', Interpretation.CATEGORICAL, channel_map={'no_other_open_wound_of_head_and_face':0, 'other_open_wound_of_head_and_face':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 117], 'phecode_870_3'))
other_paralytic_syndromes_phe = TensorMap('phecode_344', Interpretation.CATEGORICAL, channel_map={'no_other_paralytic_syndromes':0, 'other_paralytic_syndromes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 507], 'phecode_344'))
other_perinatal_conditions_of_fetus_or_newborn_phe = TensorMap('phecode_656', Interpretation.CATEGORICAL, channel_map={'no_other_perinatal_conditions_of_fetus_or_newborn':0, 'other_perinatal_conditions_of_fetus_or_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 691], 'phecode_656'))
other_peripheral_nerve_disorders_phe = TensorMap('phecode_351', Interpretation.CATEGORICAL, channel_map={'no_other_peripheral_nerve_disorders':0, 'other_peripheral_nerve_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 29], 'phecode_351'))
other_persistent_mental_disorders_due_to_conditions_classified_elsewhere_phe = TensorMap('phecode_290.3', Interpretation.CATEGORICAL, channel_map={'no_other_persistent_mental_disorders_due_to_conditions_classified_elsewhere':0, 'other_persistent_mental_disorders_due_to_conditions_classified_elsewhere':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3875], 'phecode_290_3'))
other_pulmonary_inflamation_or_edema_phe = TensorMap('phecode_505', Interpretation.CATEGORICAL, channel_map={'no_other_pulmonary_inflamation_or_edema':0, 'other_pulmonary_inflamation_or_edema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5555], 'phecode_505'))
other_retinal_disorders_phe = TensorMap('phecode_362', Interpretation.CATEGORICAL, channel_map={'no_other_retinal_disorders':0, 'other_retinal_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 338], 'phecode_362'))
other_signs_and_symptoms_in_breast_phe = TensorMap('phecode_613.7', Interpretation.CATEGORICAL, channel_map={'no_other_signs_and_symptoms_in_breast':0, 'other_signs_and_symptoms_in_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 801], 'phecode_613_7'))
other_signs_and_symptoms_involving_emotional_state_phe = TensorMap('phecode_1004', Interpretation.CATEGORICAL, channel_map={'no_other_signs_and_symptoms_involving_emotional_state':0, 'other_signs_and_symptoms_involving_emotional_state':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 707], 'phecode_1004'))
other_specified_benign_mammary_dysplasias_phe = TensorMap('phecode_610.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_benign_mammary_dysplasias':0, 'other_specified_benign_mammary_dysplasias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 576], 'phecode_610_8'))
other_specified_cardiac_dysrhythmias_phe = TensorMap('phecode_427.3', Interpretation.CATEGORICAL, channel_map={'no_other_specified_cardiac_dysrhythmias':0, 'other_specified_cardiac_dysrhythmias':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 98], 'phecode_427_3'))
other_specified_congenital_anomalies_of_kidney_phe = TensorMap('phecode_751.22', Interpretation.CATEGORICAL, channel_map={'no_other_specified_congenital_anomalies_of_kidney':0, 'other_specified_congenital_anomalies_of_kidney':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 988], 'phecode_751_22'))
other_specified_congenital_anomalies_of_nervous_system_phe = TensorMap('phecode_752.2', Interpretation.CATEGORICAL, channel_map={'no_other_specified_congenital_anomalies_of_nervous_system':0, 'other_specified_congenital_anomalies_of_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4716], 'phecode_752_2'))
other_specified_diffuse_diseases_of_connective_tissue_phe = TensorMap('phecode_709.6', Interpretation.CATEGORICAL, channel_map={'no_other_specified_diffuse_diseases_of_connective_tissue':0, 'other_specified_diffuse_diseases_of_connective_tissue':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5208], 'phecode_709_6'))
other_specified_diseases_of_hair_and_hair_follicles_phe = TensorMap('phecode_704.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_diseases_of_hair_and_hair_follicles':0, 'other_specified_diseases_of_hair_and_hair_follicles':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4807], 'phecode_704_8'))
other_specified_diseases_of_sebaceous_glands_phe = TensorMap('phecode_706.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_diseases_of_sebaceous_glands':0, 'other_specified_diseases_of_sebaceous_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_706_8'))
other_specified_diseases_of_the_salivary_glands_phe = TensorMap('phecode_527.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_diseases_of_the_salivary_glands':0, 'other_specified_diseases_of_the_salivary_glands':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2824], 'phecode_527_8'))
other_specified_disorders_involving_the_immune_mechanism_phe = TensorMap('phecode_279.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_disorders_involving_the_immune_mechanism':0, 'other_specified_disorders_involving_the_immune_mechanism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 14705], 'phecode_279_8'))
other_specified_disorders_of_breast_phe = TensorMap('phecode_613.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_disorders_of_breast':0, 'other_specified_disorders_of_breast':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1096], 'phecode_613_8'))
other_specified_erythematous_conditions_phe = TensorMap('phecode_695.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_erythematous_conditions':0, 'other_specified_erythematous_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1968], 'phecode_695_8'))
other_specified_gastritis_phe = TensorMap('phecode_535.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_gastritis':0, 'other_specified_gastritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 39], 'phecode_535_8'))
other_specified_osteoporosis_phe = TensorMap('phecode_743.13', Interpretation.CATEGORICAL, channel_map={'no_other_specified_osteoporosis':0, 'other_specified_osteoporosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2145], 'phecode_743_13'))
other_specified_peripheral_vascular_diseases_phe = TensorMap('phecode_443.8', Interpretation.CATEGORICAL, channel_map={'no_other_specified_peripheral_vascular_diseases':0, 'other_specified_peripheral_vascular_diseases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7352], 'phecode_443_8'))
other_sprains_and_strains_phe = TensorMap('phecode_842', Interpretation.CATEGORICAL, channel_map={'no_other_sprains_and_strains':0, 'other_sprains_and_strains':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_842'))
other_symptoms_phe = TensorMap('phecode_1005', Interpretation.CATEGORICAL, channel_map={'no_other_symptoms':0, 'other_symptoms':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 267], 'phecode_1005'))
other_symptoms_disorders_or_the_urinary_system_phe = TensorMap('phecode_599', Interpretation.CATEGORICAL, channel_map={'no_other_symptoms_disorders_or_the_urinary_system':0, 'other_symptoms_disorders_or_the_urinary_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3], 'phecode_599'))
other_symptoms_involving_abdomen_and_pelvis_phe = TensorMap('phecode_579', Interpretation.CATEGORICAL, channel_map={'no_other_symptoms_involving_abdomen_and_pelvis':0, 'other_symptoms_involving_abdomen_and_pelvis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 316], 'phecode_579'))
other_symptoms_involving_urinary_system_phe = TensorMap('phecode_599.8', Interpretation.CATEGORICAL, channel_map={'no_other_symptoms_involving_urinary_system':0, 'other_symptoms_involving_urinary_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6944], 'phecode_599_8'))
other_symptoms_of_respiratory_system_phe = TensorMap('phecode_512', Interpretation.CATEGORICAL, channel_map={'no_other_symptoms_of_respiratory_system':0, 'other_symptoms_of_respiratory_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 56], 'phecode_512'))
other_symptoms_referable_to_back_phe = TensorMap('phecode_724.8', Interpretation.CATEGORICAL, channel_map={'no_other_symptoms_referable_to_back':0, 'other_symptoms_referable_to_back':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23809], 'phecode_724_8'))
other_tests_phe = TensorMap('phecode_1010', Interpretation.CATEGORICAL, channel_map={'no_other_tests':0, 'other_tests':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2], 'phecode_1010'))
other_unspecified_back_disorders_phe = TensorMap('phecode_724.9', Interpretation.CATEGORICAL, channel_map={'no_other_unspecified_back_disorders':0, 'other_unspecified_back_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 727], 'phecode_724_9'))
other_upper_respiratory_disease_phe = TensorMap('phecode_479', Interpretation.CATEGORICAL, channel_map={'no_other_upper_respiratory_disease':0, 'other_upper_respiratory_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 24], 'phecode_479'))
other_venous_embolism_and_thrombosis_phe = TensorMap('phecode_452', Interpretation.CATEGORICAL, channel_map={'no_other_venous_embolism_and_thrombosis':0, 'other_venous_embolism_and_thrombosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 593], 'phecode_452'))
other_vitamin_b12_deficiency_anemia_phe = TensorMap('phecode_281.12', Interpretation.CATEGORICAL, channel_map={'no_other_vitamin_b12_deficiency_anemia':0, 'other_vitamin_b12_deficiency_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1689], 'phecode_281_12'))
otitis_externa_phe = TensorMap('phecode_380.1', Interpretation.CATEGORICAL, channel_map={'no_otitis_externa':0, 'otitis_externa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 703], 'phecode_380_1'))
otitis_media_phe = TensorMap('phecode_381.1', Interpretation.CATEGORICAL, channel_map={'no_otitis_media':0, 'otitis_media':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 116], 'phecode_381_1'))
otorrhea_phe = TensorMap('phecode_381.9', Interpretation.CATEGORICAL, channel_map={'no_otorrhea':0, 'otorrhea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1506], 'phecode_381_9'))
otosclerosis_phe = TensorMap('phecode_383', Interpretation.CATEGORICAL, channel_map={'no_otosclerosis':0, 'otosclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1275], 'phecode_383'))
ovarian_cyst_phe = TensorMap('phecode_628', Interpretation.CATEGORICAL, channel_map={'no_ovarian_cyst':0, 'ovarian_cyst':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 82], 'phecode_628'))
ovarian_dysfunction_phe = TensorMap('phecode_256', Interpretation.CATEGORICAL, channel_map={'no_ovarian_dysfunction':0, 'ovarian_dysfunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9433], 'phecode_256'))
overweight_obesity_and_other_hyperalimentation_phe = TensorMap('phecode_278', Interpretation.CATEGORICAL, channel_map={'no_overweight_obesity_and_other_hyperalimentation':0, 'overweight_obesity_and_other_hyperalimentation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100000], 'phecode_278'))
pain_and_other_symptoms_associated_with_female_genital_organs_phe = TensorMap('phecode_625', Interpretation.CATEGORICAL, channel_map={'no_pain_and_other_symptoms_associated_with_female_genital_organs':0, 'pain_and_other_symptoms_associated_with_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 207], 'phecode_625'))
pain_in_joint_phe = TensorMap('phecode_745', Interpretation.CATEGORICAL, channel_map={'no_pain_in_joint':0, 'pain_in_joint':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 227], 'phecode_745'))
pain_in_limb_phe = TensorMap('phecode_773', Interpretation.CATEGORICAL, channel_map={'no_pain_in_limb':0, 'pain_in_limb':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 271], 'phecode_773'))
pain_swelling_or_discharge_of_eye_phe = TensorMap('phecode_379.9', Interpretation.CATEGORICAL, channel_map={'no_pain_swelling_or_discharge_of_eye':0, 'pain_swelling_or_discharge_of_eye':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1886], 'phecode_379_9'))
painful_respiration_phe = TensorMap('phecode_512.2', Interpretation.CATEGORICAL, channel_map={'no_painful_respiration':0, 'painful_respiration':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1501], 'phecode_512_2'))
palindromic_rheumatism_phe = TensorMap('phecode_716.8', Interpretation.CATEGORICAL, channel_map={'no_palindromic_rheumatism':0, 'palindromic_rheumatism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21739], 'phecode_716_8'))
pallor_and_flushing_phe = TensorMap('phecode_782.6', Interpretation.CATEGORICAL, channel_map={'no_pallor_and_flushing':0, 'pallor_and_flushing':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1845], 'phecode_782_6'))
palpitations_phe = TensorMap('phecode_427.9', Interpretation.CATEGORICAL, channel_map={'no_palpitations':0, 'palpitations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 88], 'phecode_427_9'))
pancreatic_cancer_phe = TensorMap('phecode_157', Interpretation.CATEGORICAL, channel_map={'no_pancreatic_cancer':0, 'pancreatic_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 514], 'phecode_157'))
panniculitis_phe = TensorMap('phecode_729.3', Interpretation.CATEGORICAL, channel_map={'no_panniculitis':0, 'panniculitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8620], 'phecode_729_3'))
paralysis_spasm_of_vocal_cords_or_larynx_phe = TensorMap('phecode_473.3', Interpretation.CATEGORICAL, channel_map={'no_paralysis_spasm_of_vocal_cords_or_larynx':0, 'paralysis_spasm_of_vocal_cords_or_larynx':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1222], 'phecode_473_3'))
paralytic_ileus_phe = TensorMap('phecode_560.1', Interpretation.CATEGORICAL, channel_map={'no_paralytic_ileus':0, 'paralytic_ileus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 580], 'phecode_560_1'))
paralytic_strabismus_phe = TensorMap('phecode_378.5', Interpretation.CATEGORICAL, channel_map={'no_paralytic_strabismus':0, 'paralytic_strabismus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1213], 'phecode_378_5'))
paranoid_disorders_phe = TensorMap('phecode_295.2', Interpretation.CATEGORICAL, channel_map={'no_paranoid_disorders':0, 'paranoid_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1582], 'phecode_295_2'))
paraproteinemia_phe = TensorMap('phecode_270.32', Interpretation.CATEGORICAL, channel_map={'no_paraproteinemia':0, 'paraproteinemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1326], 'phecode_270_32'))
parapsoriasis_phe = TensorMap('phecode_696.2', Interpretation.CATEGORICAL, channel_map={'no_parapsoriasis':0, 'parapsoriasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19230], 'phecode_696_2'))
parasomnia_phe = TensorMap('phecode_327.5', Interpretation.CATEGORICAL, channel_map={'no_parasomnia':0, 'parasomnia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 19230], 'phecode_327_5'))
parkinsons_disease_phe = TensorMap('phecode_332', Interpretation.CATEGORICAL, channel_map={'no_parkinsons_disease':0, 'parkinsons_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 258], 'phecode_332'))
paroxysmal_supraventricular_tachycardia_phe = TensorMap('phecode_427.11', Interpretation.CATEGORICAL, channel_map={'no_paroxysmal_supraventricular_tachycardia':0, 'paroxysmal_supraventricular_tachycardia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 146], 'phecode_427_11'))
paroxysmal_tachycardia_unspecified_phe = TensorMap('phecode_427.1', Interpretation.CATEGORICAL, channel_map={'no_paroxysmal_tachycardia_unspecified':0, 'paroxysmal_tachycardia_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3184], 'phecode_427_1'))
paroxysmal_ventricular_tachycardia_phe = TensorMap('phecode_427.12', Interpretation.CATEGORICAL, channel_map={'no_paroxysmal_ventricular_tachycardia':0, 'paroxysmal_ventricular_tachycardia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 361], 'phecode_427_12'))
partial_epilepsy_phe = TensorMap('phecode_345.12', Interpretation.CATEGORICAL, channel_map={'no_partial_epilepsy':0, 'partial_epilepsy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1193], 'phecode_345_12'))
pathologic_fracture_phe = TensorMap('phecode_743.2', Interpretation.CATEGORICAL, channel_map={'no_pathologic_fracture':0, 'pathologic_fracture':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13888], 'phecode_743_2'))
pathologic_fracture_of_vertebrae_phe = TensorMap('phecode_743.21', Interpretation.CATEGORICAL, channel_map={'no_pathologic_fracture_of_vertebrae':0, 'pathologic_fracture_of_vertebrae':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9090], 'phecode_743_21'))
pathological_developmental_or_recurrent_dislocation_phe = TensorMap('phecode_742.2', Interpretation.CATEGORICAL, channel_map={'no_pathological_developmental_or_recurrent_dislocation':0, 'pathological_developmental_or_recurrent_dislocation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6024], 'phecode_742_2'))
pectus_and_other_congenital_anomalies_of_ribs_sternum_phe = TensorMap('phecode_756.2', Interpretation.CATEGORICAL, channel_map={'no_pectus_and_other_congenital_anomalies_of_ribs_sternum':0, 'pectus_and_other_congenital_anomalies_of_ribs_sternum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55555], 'phecode_756_2'))
pectus_carinatum_phe = TensorMap('phecode_756.22', Interpretation.CATEGORICAL, channel_map={'no_pectus_carinatum':0, 'pectus_carinatum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_756_22'))
pectus_excavatum_phe = TensorMap('phecode_756.21', Interpretation.CATEGORICAL, channel_map={'no_pectus_excavatum':0, 'pectus_excavatum':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21739], 'phecode_756_21'))
pediculosis_and_phthirus_infestation_phe = TensorMap('phecode_132.1', Interpretation.CATEGORICAL, channel_map={'no_pediculosis_and_phthirus_infestation':0, 'pediculosis_and_phthirus_infestation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9433], 'phecode_132_1'))
pelvic_inflammatory_disease_nos_phe = TensorMap('phecode_614.33', Interpretation.CATEGORICAL, channel_map={'no_pelvic_inflammatory_disease_nos':0, 'pelvic_inflammatory_disease_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 61], 'phecode_614_33'))
pelvic_inflammatory_disease_pid_phe = TensorMap('phecode_614.3', Interpretation.CATEGORICAL, channel_map={'no_pelvic_inflammatory_disease_pid':0, 'pelvic_inflammatory_disease_pid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2512], 'phecode_614_3'))
pelvic_peritoneal_adhesions_female_postoperative_postinfection_phe = TensorMap('phecode_614.1', Interpretation.CATEGORICAL, channel_map={'no_pelvic_peritoneal_adhesions_female_postoperative_postinfection':0, 'pelvic_peritoneal_adhesions_female_postoperative_postinfection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 143], 'phecode_614_1'))
pemphigus_and_pemphigoid_phe = TensorMap('phecode_695.22', Interpretation.CATEGORICAL, channel_map={'no_pemphigus_and_pemphigoid':0, 'pemphigus_and_pemphigoid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3546], 'phecode_695_22'))
peptic_ulcer_site_unspecified_phe = TensorMap('phecode_531.4', Interpretation.CATEGORICAL, channel_map={'no_peptic_ulcer_site_unspecified':0, 'peptic_ulcer_site_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 848], 'phecode_531_4'))
perforation_of_tympanic_membrane_phe = TensorMap('phecode_384.4', Interpretation.CATEGORICAL, channel_map={'no_perforation_of_tympanic_membrane':0, 'perforation_of_tympanic_membrane':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 378], 'phecode_384_4'))
periapical_abscess_phe = TensorMap('phecode_522.5', Interpretation.CATEGORICAL, channel_map={'no_periapical_abscess':0, 'periapical_abscess':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 291], 'phecode_522_5'))
pericarditis_phe = TensorMap('phecode_420.2', Interpretation.CATEGORICAL, channel_map={'no_pericarditis':0, 'pericarditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 292], 'phecode_420_2'))
periodontitis_acute_or_chronic_phe = TensorMap('phecode_523.3', Interpretation.CATEGORICAL, channel_map={'no_periodontitis_acute_or_chronic':0, 'periodontitis_acute_or_chronic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_523_3'))
peripheral_angiopathy_in_diseases_classified_elsewhere_phe = TensorMap('phecode_443.7', Interpretation.CATEGORICAL, channel_map={'no_peripheral_angiopathy_in_diseases_classified_elsewhere':0, 'peripheral_angiopathy_in_diseases_classified_elsewhere':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 801], 'phecode_443_7'))
peripheral_autonomic_neuropathy_phe = TensorMap('phecode_337.1', Interpretation.CATEGORICAL, channel_map={'no_peripheral_autonomic_neuropathy':0, 'peripheral_autonomic_neuropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5263], 'phecode_337_1'))
peripheral_enthesopathies_and_allied_syndromes_phe = TensorMap('phecode_726', Interpretation.CATEGORICAL, channel_map={'no_peripheral_enthesopathies_and_allied_syndromes':0, 'peripheral_enthesopathies_and_allied_syndromes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41], 'phecode_726'))
peripheral_or_central_vertigo_phe = TensorMap('phecode_386.2', Interpretation.CATEGORICAL, channel_map={'no_peripheral_or_central_vertigo':0, 'peripheral_or_central_vertigo':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 768], 'phecode_386_2'))
peripheral_retinal_degenerations_phe = TensorMap('phecode_362.6', Interpretation.CATEGORICAL, channel_map={'no_peripheral_retinal_degenerations':0, 'peripheral_retinal_degenerations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2032], 'phecode_362_6'))
peripheral_vascular_disease_phe = TensorMap('phecode_443', Interpretation.CATEGORICAL, channel_map={'no_peripheral_vascular_disease':0, 'peripheral_vascular_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_443'))
peripheral_vascular_disease_unspecified_phe = TensorMap('phecode_443.9', Interpretation.CATEGORICAL, channel_map={'no_peripheral_vascular_disease_unspecified':0, 'peripheral_vascular_disease_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 126], 'phecode_443_9'))
peritoneal_adhesions_postoperative_postinfection_phe = TensorMap('phecode_568.1', Interpretation.CATEGORICAL, channel_map={'no_peritoneal_adhesions_postoperative_postinfection':0, 'peritoneal_adhesions_postoperative_postinfection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100], 'phecode_568_1'))
peritoneal_or_intestinal_adhesions_phe = TensorMap('phecode_560.3', Interpretation.CATEGORICAL, channel_map={'no_peritoneal_or_intestinal_adhesions':0, 'peritoneal_or_intestinal_adhesions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 385], 'phecode_560_3'))
peritonitis_and_retroperitoneal_infections_phe = TensorMap('phecode_567', Interpretation.CATEGORICAL, channel_map={'no_peritonitis_and_retroperitoneal_infections':0, 'peritonitis_and_retroperitoneal_infections':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 386], 'phecode_567'))
pernicious_anemia_phe = TensorMap('phecode_281.11', Interpretation.CATEGORICAL, channel_map={'no_pernicious_anemia':0, 'pernicious_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 459], 'phecode_281_11'))
personal_history_of_diseases_of_blood_and_blood_forming_organs_phe = TensorMap('phecode_289.3', Interpretation.CATEGORICAL, channel_map={'no_personal_history_of_diseases_of_blood_and_blood_forming_organs':0, 'personal_history_of_diseases_of_blood_and_blood_forming_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 968], 'phecode_289_3'))
personal_history_of_diseases_of_digestive_system_phe = TensorMap('phecode_564.9', Interpretation.CATEGORICAL, channel_map={'no_personal_history_of_diseases_of_digestive_system':0, 'personal_history_of_diseases_of_digestive_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21], 'phecode_564_9'))
personality_disorders_phe = TensorMap('phecode_301', Interpretation.CATEGORICAL, channel_map={'no_personality_disorders':0, 'personality_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 809], 'phecode_301'))
pervasive_developmental_disorders_phe = TensorMap('phecode_313', Interpretation.CATEGORICAL, channel_map={'no_pervasive_developmental_disorders':0, 'pervasive_developmental_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7692], 'phecode_313'))
phenylketonuria_pku_phe = TensorMap('phecode_270.12', Interpretation.CATEGORICAL, channel_map={'no_phenylketonuria_pku':0, 'phenylketonuria_pku':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_270_12'))
phlebitis_and_thrombophlebitis_phe = TensorMap('phecode_451', Interpretation.CATEGORICAL, channel_map={'no_phlebitis_and_thrombophlebitis':0, 'phlebitis_and_thrombophlebitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 954], 'phecode_451'))
phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses_phe = TensorMap('phecode_325', Interpretation.CATEGORICAL, channel_map={'no_phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses':0, 'phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7692], 'phecode_325'))
phlebitis_and_thrombophlebitis_of_lower_extremities_phe = TensorMap('phecode_451.2', Interpretation.CATEGORICAL, channel_map={'no_phlebitis_and_thrombophlebitis_of_lower_extremities':0, 'phlebitis_and_thrombophlebitis_of_lower_extremities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 98], 'phecode_451_2'))
phobia_phe = TensorMap('phecode_300.13', Interpretation.CATEGORICAL, channel_map={'no_phobia':0, 'phobia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 490], 'phecode_300_13'))
pilonidal_cyst_phe = TensorMap('phecode_686.3', Interpretation.CATEGORICAL, channel_map={'no_pilonidal_cyst':0, 'pilonidal_cyst':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 770], 'phecode_686_3'))
pituitary_hyperfunction_phe = TensorMap('phecode_253.1', Interpretation.CATEGORICAL, channel_map={'no_pituitary_hyperfunction':0, 'pituitary_hyperfunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3571], 'phecode_253_1'))
pituitary_hypofunction_phe = TensorMap('phecode_253.2', Interpretation.CATEGORICAL, channel_map={'no_pituitary_hypofunction':0, 'pituitary_hypofunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1228], 'phecode_253_2'))
pityriasis_phe = TensorMap('phecode_696.3', Interpretation.CATEGORICAL, channel_map={'no_pityriasis':0, 'pityriasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7462], 'phecode_696_3'))
placenta_previa_and_abruptio_placenta_phe = TensorMap('phecode_635.3', Interpretation.CATEGORICAL, channel_map={'no_placenta_previa_and_abruptio_placenta':0, 'placenta_previa_and_abruptio_placenta':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 295], 'phecode_635_3'))
pleurisy_pleural_effusion_phe = TensorMap('phecode_507', Interpretation.CATEGORICAL, channel_map={'no_pleurisy_pleural_effusion':0, 'pleurisy_pleural_effusion':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 44], 'phecode_507'))
pneumococcal_pneumonia_phe = TensorMap('phecode_480.11', Interpretation.CATEGORICAL, channel_map={'no_pneumococcal_pneumonia':0, 'pneumococcal_pneumonia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 50], 'phecode_480_11'))
pneumoconiosis_phe = TensorMap('phecode_500.2', Interpretation.CATEGORICAL, channel_map={'no_pneumoconiosis':0, 'pneumoconiosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 284], 'phecode_500_2'))
pneumonia_phe = TensorMap('phecode_480', Interpretation.CATEGORICAL, channel_map={'no_pneumonia':0, 'pneumonia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 77], 'phecode_480'))
pneumonia_due_to_fungus_mycoses_phe = TensorMap('phecode_480.3', Interpretation.CATEGORICAL, channel_map={'no_pneumonia_due_to_fungus_mycoses':0, 'pneumonia_due_to_fungus_mycoses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2762], 'phecode_480_3'))
pneumonitis_due_to_inhalation_of_food_or_vomitus_phe = TensorMap('phecode_501', Interpretation.CATEGORICAL, channel_map={'no_pneumonitis_due_to_inhalation_of_food_or_vomitus':0, 'pneumonitis_due_to_inhalation_of_food_or_vomitus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 448], 'phecode_501'))
poisoning_allergy_of_sulfonamides_phe = TensorMap('phecode_961.1', Interpretation.CATEGORICAL, channel_map={'no_poisoning_allergy_of_sulfonamides':0, 'poisoning_allergy_of_sulfonamides':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 401], 'phecode_961_1'))
poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system_phe = TensorMap('phecode_975', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system':0, 'poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6849], 'phecode_975'))
poisoning_by_agents_primarily_affecting_blood_constituents_phe = TensorMap('phecode_964', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_agents_primarily_affecting_blood_constituents':0, 'poisoning_by_agents_primarily_affecting_blood_constituents':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25000], 'phecode_964'))
poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs_phe = TensorMap('phecode_976', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs':0, 'poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7042], 'phecode_976'))
poisoning_by_agents_primarily_affecting_the_cardiovascular_system_phe = TensorMap('phecode_972', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_agents_primarily_affecting_the_cardiovascular_system':0, 'poisoning_by_agents_primarily_affecting_the_cardiovascular_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1538], 'phecode_972'))
poisoning_by_agents_primarily_affecting_the_gastrointestinal_system_phe = TensorMap('phecode_973', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_agents_primarily_affecting_the_gastrointestinal_system':0, 'poisoning_by_agents_primarily_affecting_the_gastrointestinal_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3378], 'phecode_973'))
poisoning_by_analgesics_antipyretics_and_antirheumatics_phe = TensorMap('phecode_965', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_analgesics_antipyretics_and_antirheumatics':0, 'poisoning_by_analgesics_antipyretics_and_antirheumatics':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 56], 'phecode_965'))
poisoning_by_antibiotics_phe = TensorMap('phecode_960', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_antibiotics':0, 'poisoning_by_antibiotics':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 91], 'phecode_960'))
poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs_phe = TensorMap('phecode_966', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs':0, 'poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 724], 'phecode_966'))
poisoning_by_antifungal_antibiotics_phe = TensorMap('phecode_960.3', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_antifungal_antibiotics':0, 'poisoning_by_antifungal_antibiotics':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23809], 'phecode_960_3'))
poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system_phe = TensorMap('phecode_971', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system':0, 'poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2000], 'phecode_971'))
poisoning_by_hormones_and_synthetic_substitutes_phe = TensorMap('phecode_962', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_hormones_and_synthetic_substitutes':0, 'poisoning_by_hormones_and_synthetic_substitutes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2202], 'phecode_962'))
poisoning_by_other_anti_infectives_phe = TensorMap('phecode_961', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_other_anti_infectives':0, 'poisoning_by_other_anti_infectives':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 758], 'phecode_961'))
poisoning_by_primarily_systemic_agents_phe = TensorMap('phecode_963', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_primarily_systemic_agents':0, 'poisoning_by_primarily_systemic_agents':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2057], 'phecode_963'))
poisoning_by_psychotropic_agents_phe = TensorMap('phecode_969', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_psychotropic_agents':0, 'poisoning_by_psychotropic_agents':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 204], 'phecode_969'))
poisoning_by_water_mineral_and_uric_acid_metabolism_drugs_phe = TensorMap('phecode_974', Interpretation.CATEGORICAL, channel_map={'no_poisoning_by_water_mineral_and_uric_acid_metabolism_drugs':0, 'poisoning_by_water_mineral_and_uric_acid_metabolism_drugs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2840], 'phecode_974'))
polyarteritis_nodosa_phe = TensorMap('phecode_446.6', Interpretation.CATEGORICAL, channel_map={'no_polyarteritis_nodosa':0, 'polyarteritis_nodosa':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9259], 'phecode_446_6'))
polyarteritis_nodosa_and_allied_conditions_phe = TensorMap('phecode_446', Interpretation.CATEGORICAL, channel_map={'no_polyarteritis_nodosa_and_allied_conditions':0, 'polyarteritis_nodosa_and_allied_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_446'))
polyclonal_hypergammaglobulinemia_phe = TensorMap('phecode_270.31', Interpretation.CATEGORICAL, channel_map={'no_polyclonal_hypergammaglobulinemia':0, 'polyclonal_hypergammaglobulinemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18518], 'phecode_270_31'))
polycystic_ovaries_phe = TensorMap('phecode_256.4', Interpretation.CATEGORICAL, channel_map={'no_polycystic_ovaries':0, 'polycystic_ovaries':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1562], 'phecode_256_4'))
polycythemia_secondary_phe = TensorMap('phecode_289.8', Interpretation.CATEGORICAL, channel_map={'no_polycythemia_secondary':0, 'polycythemia_secondary':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1075], 'phecode_289_8'))
polycythemia_vera_phe = TensorMap('phecode_200.1', Interpretation.CATEGORICAL, channel_map={'no_polycythemia_vera':0, 'polycythemia_vera':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 905], 'phecode_200_1'))
polydipsia_phe = TensorMap('phecode_276.8', Interpretation.CATEGORICAL, channel_map={'no_polydipsia':0, 'polydipsia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2976], 'phecode_276_8'))
polymyalgia_rheumatica_phe = TensorMap('phecode_717', Interpretation.CATEGORICAL, channel_map={'no_polymyalgia_rheumatica':0, 'polymyalgia_rheumatica':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 255], 'phecode_717'))
polymyositis_phe = TensorMap('phecode_709.4', Interpretation.CATEGORICAL, channel_map={'no_polymyositis':0, 'polymyositis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8064], 'phecode_709_4'))
polyneuropathy_due_to_drugs_phe = TensorMap('phecode_316.1', Interpretation.CATEGORICAL, channel_map={'no_polyneuropathy_due_to_drugs':0, 'polyneuropathy_due_to_drugs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4385], 'phecode_316_1'))
polyneuropathy_in_diabetes_phe = TensorMap('phecode_250.6', Interpretation.CATEGORICAL, channel_map={'no_polyneuropathy_in_diabetes':0, 'polyneuropathy_in_diabetes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 742], 'phecode_250_6'))
polyp_of_corpus_uteri_phe = TensorMap('phecode_622.1', Interpretation.CATEGORICAL, channel_map={'no_polyp_of_corpus_uteri':0, 'polyp_of_corpus_uteri':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 47], 'phecode_622_1'))
polyp_of_female_genital_organs_phe = TensorMap('phecode_622', Interpretation.CATEGORICAL, channel_map={'no_polyp_of_female_genital_organs':0, 'polyp_of_female_genital_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1766], 'phecode_622'))
polyphagia_phe = TensorMap('phecode_260.7', Interpretation.CATEGORICAL, channel_map={'no_polyphagia':0, 'polyphagia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100000], 'phecode_260_7'))
portal_hypertension_phe = TensorMap('phecode_571.81', Interpretation.CATEGORICAL, channel_map={'no_portal_hypertension':0, 'portal_hypertension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 566], 'phecode_571_81'))
postablative_ovarian_failure_phe = TensorMap('phecode_258.1', Interpretation.CATEGORICAL, channel_map={'no_postablative_ovarian_failure':0, 'postablative_ovarian_failure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17241], 'phecode_258_1'))
postinflammatory_pulmonary_fibrosis_phe = TensorMap('phecode_502', Interpretation.CATEGORICAL, channel_map={'no_postinflammatory_pulmonary_fibrosis':0, 'postinflammatory_pulmonary_fibrosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 332], 'phecode_502'))
postlaminectomy_syndrome_phe = TensorMap('phecode_722.8', Interpretation.CATEGORICAL, channel_map={'no_postlaminectomy_syndrome':0, 'postlaminectomy_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5555], 'phecode_722_8'))
postmenopausal_atrophic_vaginitis_phe = TensorMap('phecode_627.3', Interpretation.CATEGORICAL, channel_map={'no_postmenopausal_atrophic_vaginitis':0, 'postmenopausal_atrophic_vaginitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 321], 'phecode_627_3'))
postmenopausal_bleeding_phe = TensorMap('phecode_627.1', Interpretation.CATEGORICAL, channel_map={'no_postmenopausal_bleeding':0, 'postmenopausal_bleeding':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 42], 'phecode_627_1'))
postoperative_infection_phe = TensorMap('phecode_80', Interpretation.CATEGORICAL, channel_map={'no_postoperative_infection':0, 'postoperative_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 80], 'phecode_80'))
postoperative_shock_phe = TensorMap('phecode_958.1', Interpretation.CATEGORICAL, channel_map={'no_postoperative_shock':0, 'postoperative_shock':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6849], 'phecode_958_1'))
postphlebitic_syndrome_phe = TensorMap('phecode_452.8', Interpretation.CATEGORICAL, channel_map={'no_postphlebitic_syndrome':0, 'postphlebitic_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 25000], 'phecode_452_8'))
posttraumatic_stress_disorder_phe = TensorMap('phecode_300.9', Interpretation.CATEGORICAL, channel_map={'no_posttraumatic_stress_disorder':0, 'posttraumatic_stress_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2212], 'phecode_300_9'))
precordial_pain_phe = TensorMap('phecode_418.1', Interpretation.CATEGORICAL, channel_map={'no_precordial_pain':0, 'precordial_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 87], 'phecode_418_1'))
preeclampsia_and_eclampsia_phe = TensorMap('phecode_642.1', Interpretation.CATEGORICAL, channel_map={'no_preeclampsia_and_eclampsia':0, 'preeclampsia_and_eclampsia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1136], 'phecode_642_1'))
premature_beats_phe = TensorMap('phecode_427.6', Interpretation.CATEGORICAL, channel_map={'no_premature_beats':0, 'premature_beats':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 475], 'phecode_427_6'))
premature_menopause_and_other_ovarian_failure_phe = TensorMap('phecode_627.5', Interpretation.CATEGORICAL, channel_map={'no_premature_menopause_and_other_ovarian_failure':0, 'premature_menopause_and_other_ovarian_failure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8771], 'phecode_627_5'))
premenopausal_menorrhagia_phe = TensorMap('phecode_627.4', Interpretation.CATEGORICAL, channel_map={'no_premenopausal_menorrhagia':0, 'premenopausal_menorrhagia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1291], 'phecode_627_4'))
premenstrual_tension_syndromes_phe = TensorMap('phecode_626.4', Interpretation.CATEGORICAL, channel_map={'no_premenstrual_tension_syndromes':0, 'premenstrual_tension_syndromes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2857], 'phecode_626_4'))
presbyopia_phe = TensorMap('phecode_367.4', Interpretation.CATEGORICAL, channel_map={'no_presbyopia':0, 'presbyopia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15625], 'phecode_367_4'))
primary_angle_closure_glaucoma_phe = TensorMap('phecode_365.2', Interpretation.CATEGORICAL, channel_map={'no_primary_angle_closure_glaucoma':0, 'primary_angle_closure_glaucoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 456], 'phecode_365_2'))
primary_biliary_cirrhosis_phe = TensorMap('phecode_571.6', Interpretation.CATEGORICAL, channel_map={'no_primary_biliary_cirrhosis':0, 'primary_biliary_cirrhosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1366], 'phecode_571_6'))
primary_intrinsic_cardiomyopathies_phe = TensorMap('phecode_425.1', Interpretation.CATEGORICAL, channel_map={'no_primary_intrinsic_cardiomyopathies':0, 'primary_intrinsic_cardiomyopathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 309], 'phecode_425_1'))
primary_open_angle_glaucoma_phe = TensorMap('phecode_365.11', Interpretation.CATEGORICAL, channel_map={'no_primary_open_angle_glaucoma':0, 'primary_open_angle_glaucoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 267], 'phecode_365_11'))
primary_pulmonary_hypertension_phe = TensorMap('phecode_415.21', Interpretation.CATEGORICAL, channel_map={'no_primary_pulmonary_hypertension':0, 'primary_pulmonary_hypertension':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 782], 'phecode_415_21'))
primary_thrombocytopenia_phe = TensorMap('phecode_287.31', Interpretation.CATEGORICAL, channel_map={'no_primary_thrombocytopenia':0, 'primary_thrombocytopenia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 819], 'phecode_287_31'))
problems_associated_with_amniotic_cavity_and_membranes_phe = TensorMap('phecode_653', Interpretation.CATEGORICAL, channel_map={'no_problems_associated_with_amniotic_cavity_and_membranes':0, 'problems_associated_with_amniotic_cavity_and_membranes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 242], 'phecode_653'))
progressive_myopia_phe = TensorMap('phecode_360.2', Interpretation.CATEGORICAL, channel_map={'no_progressive_myopia':0, 'progressive_myopia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1497], 'phecode_360_2'))
prolapse_of_vaginal_vault_after_hysterectomy_phe = TensorMap('phecode_618.5', Interpretation.CATEGORICAL, channel_map={'no_prolapse_of_vaginal_vault_after_hysterectomy':0, 'prolapse_of_vaginal_vault_after_hysterectomy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 776], 'phecode_618_5'))
prolapse_of_vaginal_walls_phe = TensorMap('phecode_618.1', Interpretation.CATEGORICAL, channel_map={'no_prolapse_of_vaginal_walls':0, 'prolapse_of_vaginal_walls':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 51], 'phecode_618_1'))
proliferative_glomerulonephritis_phe = TensorMap('phecode_580.11', Interpretation.CATEGORICAL, channel_map={'no_proliferative_glomerulonephritis':0, 'proliferative_glomerulonephritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4950], 'phecode_580_11'))
prostatitis_phe = TensorMap('phecode_601.1', Interpretation.CATEGORICAL, channel_map={'no_prostatitis':0, 'prostatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 567], 'phecode_601_1'))
protein_calorie_malnutrition_phe = TensorMap('phecode_260', Interpretation.CATEGORICAL, channel_map={'no_protein_calorie_malnutrition':0, 'protein_calorie_malnutrition':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1754], 'phecode_260'))
proteinuria_phe = TensorMap('phecode_269', Interpretation.CATEGORICAL, channel_map={'no_proteinuria':0, 'proteinuria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 442], 'phecode_269'))
protozoan_infection_phe = TensorMap('phecode_131', Interpretation.CATEGORICAL, channel_map={'no_protozoan_infection':0, 'protozoan_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17857], 'phecode_131'))
prurigo_and_lichen_phe = TensorMap('phecode_695.7', Interpretation.CATEGORICAL, channel_map={'no_prurigo_and_lichen':0, 'prurigo_and_lichen':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 451], 'phecode_695_7'))
pruritus_and_related_conditions_phe = TensorMap('phecode_698', Interpretation.CATEGORICAL, channel_map={'no_pruritus_and_related_conditions':0, 'pruritus_and_related_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 438], 'phecode_698'))
pseudomonal_pneumonia_phe = TensorMap('phecode_480.12', Interpretation.CATEGORICAL, channel_map={'no_pseudomonal_pneumonia':0, 'pseudomonal_pneumonia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3472], 'phecode_480_12'))
psoriasis_phe = TensorMap('phecode_696.4', Interpretation.CATEGORICAL, channel_map={'no_psoriasis':0, 'psoriasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 207], 'phecode_696_4'))
psoriasis_vulgaris_phe = TensorMap('phecode_696.41', Interpretation.CATEGORICAL, channel_map={'no_psoriasis_vulgaris':0, 'psoriasis_vulgaris':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 192], 'phecode_696_41'))
psoriatic_arthropathy_phe = TensorMap('phecode_696.42', Interpretation.CATEGORICAL, channel_map={'no_psoriatic_arthropathy':0, 'psoriatic_arthropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 484], 'phecode_696_42'))
psychogenic_and_somatoform_disorders_phe = TensorMap('phecode_303', Interpretation.CATEGORICAL, channel_map={'no_psychogenic_and_somatoform_disorders':0, 'psychogenic_and_somatoform_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_303'))
psychogenic_disorder_phe = TensorMap('phecode_303.3', Interpretation.CATEGORICAL, channel_map={'no_psychogenic_disorder':0, 'psychogenic_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 423], 'phecode_303_3'))
psychosis_phe = TensorMap('phecode_295.3', Interpretation.CATEGORICAL, channel_map={'no_psychosis':0, 'psychosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1718], 'phecode_295_3'))
ptosis_of_eyelid_phe = TensorMap('phecode_374.3', Interpretation.CATEGORICAL, channel_map={'no_ptosis_of_eyelid':0, 'ptosis_of_eyelid':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 186], 'phecode_374_3'))
pulmonary_collapse_interstitial_and_compensatory_emphysema_phe = TensorMap('phecode_508', Interpretation.CATEGORICAL, channel_map={'no_pulmonary_collapse_interstitial_and_compensatory_emphysema':0, 'pulmonary_collapse_interstitial_and_compensatory_emphysema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 139], 'phecode_508'))
pulmonary_congestion_and_hypostasis_phe = TensorMap('phecode_503', Interpretation.CATEGORICAL, channel_map={'no_pulmonary_congestion_and_hypostasis':0, 'pulmonary_congestion_and_hypostasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 639], 'phecode_503'))
pulmonary_embolism_and_infarction_acute_phe = TensorMap('phecode_415.11', Interpretation.CATEGORICAL, channel_map={'no_pulmonary_embolism_and_infarction_acute':0, 'pulmonary_embolism_and_infarction_acute':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7352], 'phecode_415_11'))
pulmonary_heart_disease_phe = TensorMap('phecode_415', Interpretation.CATEGORICAL, channel_map={'no_pulmonary_heart_disease':0, 'pulmonary_heart_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 92], 'phecode_415'))
pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery_phe = TensorMap('phecode_509.3', Interpretation.CATEGORICAL, channel_map={'no_pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery':0, 'pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 38461], 'phecode_509_3'))
pulpitis_and_necrosis_of_tooth_pulp_phe = TensorMap('phecode_522.1', Interpretation.CATEGORICAL, channel_map={'no_pulpitis_and_necrosis_of_tooth_pulp':0, 'pulpitis_and_necrosis_of_tooth_pulp':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6756], 'phecode_522_1'))
purpura_and_other_hemorrhagic_conditions_phe = TensorMap('phecode_287', Interpretation.CATEGORICAL, channel_map={'no_purpura_and_other_hemorrhagic_conditions':0, 'purpura_and_other_hemorrhagic_conditions':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3355], 'phecode_287'))
pyelonephritis_phe = TensorMap('phecode_590', Interpretation.CATEGORICAL, channel_map={'no_pyelonephritis':0, 'pyelonephritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 85], 'phecode_590'))
pyoderma_phe = TensorMap('phecode_686.5', Interpretation.CATEGORICAL, channel_map={'no_pyoderma':0, 'pyoderma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7246], 'phecode_686_5'))
pyogenic_arthritis_phe = TensorMap('phecode_711.1', Interpretation.CATEGORICAL, channel_map={'no_pyogenic_arthritis':0, 'pyogenic_arthritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5747], 'phecode_711_1'))
pyogenic_granuloma_phe = TensorMap('phecode_686.4', Interpretation.CATEGORICAL, channel_map={'no_pyogenic_granuloma':0, 'pyogenic_granuloma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1016], 'phecode_686_4'))
qualitative_platelet_defects_phe = TensorMap('phecode_287.4', Interpretation.CATEGORICAL, channel_map={'no_qualitative_platelet_defects':0, 'qualitative_platelet_defects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_287_4'))
radiotherapy_phe = TensorMap('phecode_196', Interpretation.CATEGORICAL, channel_map={'no_radiotherapy':0, 'radiotherapy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83333], 'phecode_196'))
rash_and_other_nonspecific_skin_eruption_phe = TensorMap('phecode_687.1', Interpretation.CATEGORICAL, channel_map={'no_rash_and_other_nonspecific_skin_eruption':0, 'rash_and_other_nonspecific_skin_eruption':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 152], 'phecode_687_1'))
raynauds_syndrome_phe = TensorMap('phecode_443.1', Interpretation.CATEGORICAL, channel_map={'no_raynauds_syndrome':0, 'raynauds_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 279], 'phecode_443_1'))
redundant_prepuce_and_phimosis_bxo_phe = TensorMap('phecode_604.1', Interpretation.CATEGORICAL, channel_map={'no_redundant_prepuce_and_phimosis_bxo':0, 'redundant_prepuce_and_phimosis_bxo':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 136], 'phecode_604_1'))
reflux_esophagitis_phe = TensorMap('phecode_530.14', Interpretation.CATEGORICAL, channel_map={'no_reflux_esophagitis':0, 'reflux_esophagitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 33], 'phecode_530_14'))
regional_enteritis_phe = TensorMap('phecode_555.1', Interpretation.CATEGORICAL, channel_map={'no_regional_enteritis':0, 'regional_enteritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 209], 'phecode_555_1'))
reiters_disease_phe = TensorMap('phecode_711.2', Interpretation.CATEGORICAL, channel_map={'no_reiters_disease':0, 'reiters_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17857], 'phecode_711_2'))
renal_colic_phe = TensorMap('phecode_594.8', Interpretation.CATEGORICAL, channel_map={'no_renal_colic':0, 'renal_colic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 202], 'phecode_594_8'))
renal_dialysis_phe = TensorMap('phecode_585.31', Interpretation.CATEGORICAL, channel_map={'no_renal_dialysis':0, 'renal_dialysis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 598], 'phecode_585_31'))
renal_failure_phe = TensorMap('phecode_585', Interpretation.CATEGORICAL, channel_map={'no_renal_failure':0, 'renal_failure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6756], 'phecode_585'))
renal_failure_nos_phe = TensorMap('phecode_585.2', Interpretation.CATEGORICAL, channel_map={'no_renal_failure_nos':0, 'renal_failure_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 30], 'phecode_585_2'))
renal_osteodystrophy_phe = TensorMap('phecode_588.1', Interpretation.CATEGORICAL, channel_map={'no_renal_osteodystrophy':0, 'renal_osteodystrophy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_588_1'))
respiratory_arrest_phe = TensorMap('phecode_509.5', Interpretation.CATEGORICAL, channel_map={'no_respiratory_arrest':0, 'respiratory_arrest':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2487], 'phecode_509_5'))
respiratory_complications_phe = TensorMap('phecode_519.2', Interpretation.CATEGORICAL, channel_map={'no_respiratory_complications':0, 'respiratory_complications':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1201], 'phecode_519_2'))
respiratory_failure_phe = TensorMap('phecode_509.1', Interpretation.CATEGORICAL, channel_map={'no_respiratory_failure':0, 'respiratory_failure':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 161], 'phecode_509_1'))
respiratory_insufficiency_phe = TensorMap('phecode_509.2', Interpretation.CATEGORICAL, channel_map={'no_respiratory_insufficiency':0, 'respiratory_insufficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 102], 'phecode_509_2'))
retention_of_urine_phe = TensorMap('phecode_599.2', Interpretation.CATEGORICAL, channel_map={'no_retention_of_urine':0, 'retention_of_urine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 46], 'phecode_599_2'))
retinal_detachment_with_retinal_defect_phe = TensorMap('phecode_361.1', Interpretation.CATEGORICAL, channel_map={'no_retinal_detachment_with_retinal_defect':0, 'retinal_detachment_with_retinal_defect':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 253], 'phecode_361_1'))
retinal_detachments_and_defects_phe = TensorMap('phecode_361', Interpretation.CATEGORICAL, channel_map={'no_retinal_detachments_and_defects':0, 'retinal_detachments_and_defects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 177], 'phecode_361'))
retinal_hemorrhage_ischemia_phe = TensorMap('phecode_362.8', Interpretation.CATEGORICAL, channel_map={'no_retinal_hemorrhage_ischemia':0, 'retinal_hemorrhage_ischemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3205], 'phecode_362_8'))
retinal_vascular_changes_and_abnomalities_phe = TensorMap('phecode_362.4', Interpretation.CATEGORICAL, channel_map={'no_retinal_vascular_changes_and_abnomalities':0, 'retinal_vascular_changes_and_abnomalities':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 376], 'phecode_362_4'))
retinopathy_of_prematurity_phe = TensorMap('phecode_362.1', Interpretation.CATEGORICAL, channel_map={'no_retinopathy_of_prematurity':0, 'retinopathy_of_prematurity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 83333], 'phecode_362_1'))
retinoschisis_and_retinal_cysts_phe = TensorMap('phecode_361.2', Interpretation.CATEGORICAL, channel_map={'no_retinoschisis_and_retinal_cysts':0, 'retinoschisis_and_retinal_cysts':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4464], 'phecode_361_2'))
rhabdomyolysis_phe = TensorMap('phecode_772.4', Interpretation.CATEGORICAL, channel_map={'no_rhabdomyolysis':0, 'rhabdomyolysis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 62500], 'phecode_772_4'))
rhesus_isoimmunization_in_pregnancy_phe = TensorMap('phecode_654.2', Interpretation.CATEGORICAL, channel_map={'no_rhesus_isoimmunization_in_pregnancy':0, 'rhesus_isoimmunization_in_pregnancy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1366], 'phecode_654_2'))
rheumatic_disease_of_the_heart_valves_phe = TensorMap('phecode_394', Interpretation.CATEGORICAL, channel_map={'no_rheumatic_disease_of_the_heart_valves':0, 'rheumatic_disease_of_the_heart_valves':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 298], 'phecode_394'))
rheumatic_fever_chorea_phe = TensorMap('phecode_41.21', Interpretation.CATEGORICAL, channel_map={'no_rheumatic_fever_chorea':0, 'rheumatic_fever_chorea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 18518], 'phecode_41_21'))
rheumatism_unspecified_and_fibrositis_phe = TensorMap('phecode_729.1', Interpretation.CATEGORICAL, channel_map={'no_rheumatism_unspecified_and_fibrositis':0, 'rheumatism_unspecified_and_fibrositis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1063], 'phecode_729_1'))
rheumatoid_arthritis_phe = TensorMap('phecode_714.1', Interpretation.CATEGORICAL, channel_map={'no_rheumatoid_arthritis':0, 'rheumatoid_arthritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 124], 'phecode_714_1'))
rheumatoid_arthritis_and_other_inflammatory_polyarthropathies_phe = TensorMap('phecode_714', Interpretation.CATEGORICAL, channel_map={'no_rheumatoid_arthritis_and_other_inflammatory_polyarthropathies':0, 'rheumatoid_arthritis_and_other_inflammatory_polyarthropathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41], 'phecode_714'))
rickets_or_osteomalacia_phe = TensorMap('phecode_261.41', Interpretation.CATEGORICAL, channel_map={'no_rickets_or_osteomalacia':0, 'rickets_or_osteomalacia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13513], 'phecode_261_41'))
right_bundle_branch_block_phe = TensorMap('phecode_426.31', Interpretation.CATEGORICAL, channel_map={'no_right_bundle_branch_block':0, 'right_bundle_branch_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 190], 'phecode_426_31'))
rosacea_phe = TensorMap('phecode_695.3', Interpretation.CATEGORICAL, channel_map={'no_rosacea':0, 'rosacea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1098], 'phecode_695_3'))
rupture_of_synovium_phe = TensorMap('phecode_727.5', Interpretation.CATEGORICAL, channel_map={'no_rupture_of_synovium':0, 'rupture_of_synovium':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 892], 'phecode_727_5'))
rupture_of_tendon_nontraumatic_phe = TensorMap('phecode_727.6', Interpretation.CATEGORICAL, channel_map={'no_rupture_of_tendon_nontraumatic':0, 'rupture_of_tendon_nontraumatic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4032], 'phecode_727_6'))
sacroiliitis_nec_phe = TensorMap('phecode_715.1', Interpretation.CATEGORICAL, channel_map={'no_sacroiliitis_nec':0, 'sacroiliitis_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3846], 'phecode_715_1'))
salicylates_causing_adverse_effects_in_therapeutic_use_phe = TensorMap('phecode_965.3', Interpretation.CATEGORICAL, channel_map={'no_salicylates_causing_adverse_effects_in_therapeutic_use':0, 'salicylates_causing_adverse_effects_in_therapeutic_use':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 996], 'phecode_965_3'))
sarcoidosis_phe = TensorMap('phecode_697', Interpretation.CATEGORICAL, channel_map={'no_sarcoidosis':0, 'sarcoidosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 305], 'phecode_697'))
scar_conditions_and_fibrosis_of_skin_phe = TensorMap('phecode_701.2', Interpretation.CATEGORICAL, channel_map={'no_scar_conditions_and_fibrosis_of_skin':0, 'scar_conditions_and_fibrosis_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 159], 'phecode_701_2'))
schizoid_personality_disorder_phe = TensorMap('phecode_301.1', Interpretation.CATEGORICAL, channel_map={'no_schizoid_personality_disorder':0, 'schizoid_personality_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9803], 'phecode_301_1'))
schizophrenia_phe = TensorMap('phecode_295.1', Interpretation.CATEGORICAL, channel_map={'no_schizophrenia':0, 'schizophrenia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 511], 'phecode_295_1'))
schizophrenia_and_other_psychotic_disorders_phe = TensorMap('phecode_295', Interpretation.CATEGORICAL, channel_map={'no_schizophrenia_and_other_psychotic_disorders':0, 'schizophrenia_and_other_psychotic_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 26315], 'phecode_295'))
schmorls_nodes_phe = TensorMap('phecode_722.3', Interpretation.CATEGORICAL, channel_map={'no_schmorls_nodes':0, 'schmorls_nodes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 9090], 'phecode_722_3'))
sciatica_phe = TensorMap('phecode_764', Interpretation.CATEGORICAL, channel_map={'no_sciatica':0, 'sciatica':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 251], 'phecode_764'))
scleritis_and_episcleritis_phe = TensorMap('phecode_379.1', Interpretation.CATEGORICAL, channel_map={'no_scleritis_and_episcleritis':0, 'scleritis_and_episcleritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7812], 'phecode_379_1'))
sebaceous_cyst_phe = TensorMap('phecode_706.2', Interpretation.CATEGORICAL, channel_map={'no_sebaceous_cyst':0, 'sebaceous_cyst':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45], 'phecode_706_2'))
seborrheic_dermatitis_phe = TensorMap('phecode_690.1', Interpretation.CATEGORICAL, channel_map={'no_seborrheic_dermatitis':0, 'seborrheic_dermatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5494], 'phecode_690_1'))
seborrheic_keratosis_phe = TensorMap('phecode_702.2', Interpretation.CATEGORICAL, channel_map={'no_seborrheic_keratosis':0, 'seborrheic_keratosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 116], 'phecode_702_2'))
second_degree_av_block_phe = TensorMap('phecode_426.23', Interpretation.CATEGORICAL, channel_map={'no_second_degree_av_block':0, 'second_degree_av_block':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 668], 'phecode_426_23'))
secondary_diabetes_mellitus_phe = TensorMap('phecode_249', Interpretation.CATEGORICAL, channel_map={'no_secondary_diabetes_mellitus':0, 'secondary_diabetes_mellitus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 100000], 'phecode_249'))
secondary_extrinsic_cardiomyopathies_phe = TensorMap('phecode_425.2', Interpretation.CATEGORICAL, channel_map={'no_secondary_extrinsic_cardiomyopathies':0, 'secondary_extrinsic_cardiomyopathies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55555], 'phecode_425_2'))
secondary_hypothyroidism_phe = TensorMap('phecode_244.1', Interpretation.CATEGORICAL, channel_map={'no_secondary_hypothyroidism':0, 'secondary_hypothyroidism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 281], 'phecode_244_1'))
secondary_malignancy_of_bone_phe = TensorMap('phecode_198.6', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignancy_of_bone':0, 'secondary_malignancy_of_bone':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 140], 'phecode_198_6'))
secondary_malignancy_of_brain_spine_phe = TensorMap('phecode_198.5', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignancy_of_brain_spine':0, 'secondary_malignancy_of_brain_spine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 380], 'phecode_198_5'))
secondary_malignancy_of_lymph_nodes_phe = TensorMap('phecode_198.1', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignancy_of_lymph_nodes':0, 'secondary_malignancy_of_lymph_nodes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 61], 'phecode_198_1'))
secondary_malignancy_of_respiratory_organs_phe = TensorMap('phecode_198.2', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignancy_of_respiratory_organs':0, 'secondary_malignancy_of_respiratory_organs':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 136], 'phecode_198_2'))
secondary_malignant_neoplasm_phe = TensorMap('phecode_198', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignant_neoplasm':0, 'secondary_malignant_neoplasm':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 280], 'phecode_198'))
secondary_malignant_neoplasm_of_digestive_systems_phe = TensorMap('phecode_198.3', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignant_neoplasm_of_digestive_systems':0, 'secondary_malignant_neoplasm_of_digestive_systems':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 199], 'phecode_198_3'))
secondary_malignant_neoplasm_of_liver_phe = TensorMap('phecode_198.4', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignant_neoplasm_of_liver':0, 'secondary_malignant_neoplasm_of_liver':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 116], 'phecode_198_4'))
secondary_malignant_neoplasm_of_skin_phe = TensorMap('phecode_198.7', Interpretation.CATEGORICAL, channel_map={'no_secondary_malignant_neoplasm_of_skin':0, 'secondary_malignant_neoplasm_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1683], 'phecode_198_7'))
secondary_thrombocytopenia_phe = TensorMap('phecode_287.32', Interpretation.CATEGORICAL, channel_map={'no_secondary_thrombocytopenia':0, 'secondary_thrombocytopenia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3787], 'phecode_287_32'))
senile_cataract_phe = TensorMap('phecode_366.2', Interpretation.CATEGORICAL, channel_map={'no_senile_cataract':0, 'senile_cataract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 33], 'phecode_366_2'))
senile_osteoporosis_phe = TensorMap('phecode_743.12', Interpretation.CATEGORICAL, channel_map={'no_senile_osteoporosis':0, 'senile_osteoporosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 12820], 'phecode_743_12'))
sensorineural_hearing_loss_phe = TensorMap('phecode_389.1', Interpretation.CATEGORICAL, channel_map={'no_sensorineural_hearing_loss':0, 'sensorineural_hearing_loss':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 655], 'phecode_389_1'))
separation_of_retinal_layers_phe = TensorMap('phecode_362.31', Interpretation.CATEGORICAL, channel_map={'no_separation_of_retinal_layers':0, 'separation_of_retinal_layers':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3521], 'phecode_362_31'))
sepsis_phe = TensorMap('phecode_994.2', Interpretation.CATEGORICAL, channel_map={'no_sepsis':0, 'sepsis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 98], 'phecode_994_2'))
septal_deviations_turbinate_hypertrophy_phe = TensorMap('phecode_470', Interpretation.CATEGORICAL, channel_map={'no_septal_deviations_turbinate_hypertrophy':0, 'septal_deviations_turbinate_hypertrophy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 82], 'phecode_470'))
septicemia_phe = TensorMap('phecode_38', Interpretation.CATEGORICAL, channel_map={'no_septicemia':0, 'septicemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 93], 'phecode_38'))
severe_protein_calorie_malnutrition_phe = TensorMap('phecode_260.2', Interpretation.CATEGORICAL, channel_map={'no_severe_protein_calorie_malnutrition':0, 'severe_protein_calorie_malnutrition':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_260_2'))
sexual_and_gender_identity_disorders_phe = TensorMap('phecode_302', Interpretation.CATEGORICAL, channel_map={'no_sexual_and_gender_identity_disorders':0, 'sexual_and_gender_identity_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1037], 'phecode_302'))
sexually_transmitted_infections_not_hiv_or_hepatitis_phe = TensorMap('phecode_90', Interpretation.CATEGORICAL, channel_map={'no_sexually_transmitted_infections_not_hiv_or_hepatitis':0, 'sexually_transmitted_infections_not_hiv_or_hepatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1333], 'phecode_90'))
shock_phe = TensorMap('phecode_797', Interpretation.CATEGORICAL, channel_map={'no_shock':0, 'shock':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6172], 'phecode_797'))
short_gestation_low_birth_weight_and_fetal_growth_retardation_phe = TensorMap('phecode_637', Interpretation.CATEGORICAL, channel_map={'no_short_gestation_low_birth_weight_and_fetal_growth_retardation':0, 'short_gestation_low_birth_weight_and_fetal_growth_retardation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_637'))
shortness_of_breath_phe = TensorMap('phecode_512.7', Interpretation.CATEGORICAL, channel_map={'no_shortness_of_breath':0, 'shortness_of_breath':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 53], 'phecode_512_7'))
sialoadenitis_phe = TensorMap('phecode_527.2', Interpretation.CATEGORICAL, channel_map={'no_sialoadenitis':0, 'sialoadenitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1000], 'phecode_527_2'))
sialolithiasis_phe = TensorMap('phecode_528.7', Interpretation.CATEGORICAL, channel_map={'no_sialolithiasis':0, 'sialolithiasis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1312], 'phecode_528_7'))
sicca_syndrome_phe = TensorMap('phecode_709.2', Interpretation.CATEGORICAL, channel_map={'no_sicca_syndrome':0, 'sicca_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 616], 'phecode_709_2'))
sickle_cell_anemia_phe = TensorMap('phecode_282.5', Interpretation.CATEGORICAL, channel_map={'no_sickle_cell_anemia':0, 'sickle_cell_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1033], 'phecode_282_5'))
sideroblastic_anemia_phe = TensorMap('phecode_285.3', Interpretation.CATEGORICAL, channel_map={'no_sideroblastic_anemia':0, 'sideroblastic_anemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_285_3'))
simple_and_unspecified_goiter_phe = TensorMap('phecode_240', Interpretation.CATEGORICAL, channel_map={'no_simple_and_unspecified_goiter':0, 'simple_and_unspecified_goiter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 586], 'phecode_240'))
sinoatrial_node_dysfunction_bradycardia_phe = TensorMap('phecode_427.8', Interpretation.CATEGORICAL, channel_map={'no_sinoatrial_node_dysfunction_bradycardia':0, 'sinoatrial_node_dysfunction_bradycardia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 717], 'phecode_427_8'))
skull_and_face_fracture_and_other_intercranial_injury_phe = TensorMap('phecode_819', Interpretation.CATEGORICAL, channel_map={'no_skull_and_face_fracture_and_other_intercranial_injury':0, 'skull_and_face_fracture_and_other_intercranial_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 358], 'phecode_819'))
sleep_apnea_phe = TensorMap('phecode_327.3', Interpretation.CATEGORICAL, channel_map={'no_sleep_apnea':0, 'sleep_apnea':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 72], 'phecode_327_3'))
sleep_disorders_phe = TensorMap('phecode_327', Interpretation.CATEGORICAL, channel_map={'no_sleep_disorders':0, 'sleep_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 374], 'phecode_327'))
small_kidney_phe = TensorMap('phecode_586.11', Interpretation.CATEGORICAL, channel_map={'no_small_kidney':0, 'small_kidney':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1650], 'phecode_586_11'))
somatoform_disorder_phe = TensorMap('phecode_303.4', Interpretation.CATEGORICAL, channel_map={'no_somatoform_disorder':0, 'somatoform_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2824], 'phecode_303_4'))
specific_nonpsychotic_mental_disorders_due_to_brain_damage_phe = TensorMap('phecode_291.4', Interpretation.CATEGORICAL, channel_map={'no_specific_nonpsychotic_mental_disorders_due_to_brain_damage':0, 'specific_nonpsychotic_mental_disorders_due_to_brain_damage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5882], 'phecode_291_4'))
speech_and_language_disorder_phe = TensorMap('phecode_315.2', Interpretation.CATEGORICAL, channel_map={'no_speech_and_language_disorder':0, 'speech_and_language_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3472], 'phecode_315_2'))
spermatocele_phe = TensorMap('phecode_603.2', Interpretation.CATEGORICAL, channel_map={'no_spermatocele':0, 'spermatocele':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13157], 'phecode_603_2'))
spina_bifida_phe = TensorMap('phecode_752.11', Interpretation.CATEGORICAL, channel_map={'no_spina_bifida':0, 'spina_bifida':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1760], 'phecode_752_11'))
spinal_cord_injury_without_evidence_of_spinal_bone_injury_phe = TensorMap('phecode_952', Interpretation.CATEGORICAL, channel_map={'no_spinal_cord_injury_without_evidence_of_spinal_bone_injury':0, 'spinal_cord_injury_without_evidence_of_spinal_bone_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3906], 'phecode_952'))
spinal_enthesopathy_phe = TensorMap('phecode_715.3', Interpretation.CATEGORICAL, channel_map={'no_spinal_enthesopathy':0, 'spinal_enthesopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_715_3'))
spinal_stenosis_phe = TensorMap('phecode_720', Interpretation.CATEGORICAL, channel_map={'no_spinal_stenosis':0, 'spinal_stenosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 446], 'phecode_720'))
spinocerebellar_disease_phe = TensorMap('phecode_334.1', Interpretation.CATEGORICAL, channel_map={'no_spinocerebellar_disease':0, 'spinocerebellar_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2475], 'phecode_334_1'))
spirochetal_infection_phe = TensorMap('phecode_130', Interpretation.CATEGORICAL, channel_map={'no_spirochetal_infection':0, 'spirochetal_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4672], 'phecode_130'))
splenomegaly_phe = TensorMap('phecode_579.2', Interpretation.CATEGORICAL, channel_map={'no_splenomegaly':0, 'splenomegaly':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 741], 'phecode_579_2'))
spondylolisthesis_congenital_phe = TensorMap('phecode_754.2', Interpretation.CATEGORICAL, channel_map={'no_spondylolisthesis_congenital':0, 'spondylolisthesis_congenital':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 125000], 'phecode_754_2'))
spondylosis_and_allied_disorders_phe = TensorMap('phecode_721', Interpretation.CATEGORICAL, channel_map={'no_spondylosis_and_allied_disorders':0, 'spondylosis_and_allied_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 305], 'phecode_721'))
spondylosis_with_myelopathy_phe = TensorMap('phecode_721.2', Interpretation.CATEGORICAL, channel_map={'no_spondylosis_with_myelopathy':0, 'spondylosis_with_myelopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10000], 'phecode_721_2'))
spondylosis_without_myelopathy_phe = TensorMap('phecode_721.1', Interpretation.CATEGORICAL, channel_map={'no_spondylosis_without_myelopathy':0, 'spondylosis_without_myelopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 346], 'phecode_721_1'))
spontaneous_ecchymoses_phe = TensorMap('phecode_287.1', Interpretation.CATEGORICAL, channel_map={'no_spontaneous_ecchymoses':0, 'spontaneous_ecchymoses':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2732], 'phecode_287_1'))
sprains_and_strains_phe = TensorMap('phecode_840', Interpretation.CATEGORICAL, channel_map={'no_sprains_and_strains':0, 'sprains_and_strains':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1061], 'phecode_840'))
staphylococcus_infections_phe = TensorMap('phecode_41.1', Interpretation.CATEGORICAL, channel_map={'no_staphylococcus_infections':0, 'staphylococcus_infections':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 108], 'phecode_41_1'))
stiffness_of_joint_phe = TensorMap('phecode_741.2', Interpretation.CATEGORICAL, channel_map={'no_stiffness_of_joint':0, 'stiffness_of_joint':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3496], 'phecode_741_2'))
stomatitis_and_mucositis_ulcerative_phe = TensorMap('phecode_528.11', Interpretation.CATEGORICAL, channel_map={'no_stomatitis_and_mucositis_ulcerative':0, 'stomatitis_and_mucositis_ulcerative':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 826], 'phecode_528_11'))
strabismus_and_other_disorders_of_binocular_eye_movements_phe = TensorMap('phecode_378', Interpretation.CATEGORICAL, channel_map={'no_strabismus_and_other_disorders_of_binocular_eye_movements':0, 'strabismus_and_other_disorders_of_binocular_eye_movements':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5434], 'phecode_378'))
strabismus_not_specified_as_paralytic_phe = TensorMap('phecode_378.1', Interpretation.CATEGORICAL, channel_map={'no_strabismus_not_specified_as_paralytic':0, 'strabismus_not_specified_as_paralytic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 405], 'phecode_378_1'))
streptococcus_infection_phe = TensorMap('phecode_41.2', Interpretation.CATEGORICAL, channel_map={'no_streptococcus_infection':0, 'streptococcus_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 205], 'phecode_41_2'))
stress_fracture_phe = TensorMap('phecode_743.4', Interpretation.CATEGORICAL, channel_map={'no_stress_fracture':0, 'stress_fracture':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_743_4'))
stress_incontinence_female_phe = TensorMap('phecode_624.9', Interpretation.CATEGORICAL, channel_map={'no_stress_incontinence_female':0, 'stress_incontinence_female':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 66], 'phecode_624_9'))
stricture_and_stenosis_of_esophagus_phe = TensorMap('phecode_530.3', Interpretation.CATEGORICAL, channel_map={'no_stricture_and_stenosis_of_esophagus':0, 'stricture_and_stenosis_of_esophagus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 175], 'phecode_530_3'))
stricture_obstruction_of_ureter_phe = TensorMap('phecode_586.4', Interpretation.CATEGORICAL, channel_map={'no_stricture_obstruction_of_ureter':0, 'stricture_obstruction_of_ureter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 369], 'phecode_586_4'))
stricture_of_artery_phe = TensorMap('phecode_447.1', Interpretation.CATEGORICAL, channel_map={'no_stricture_of_artery':0, 'stricture_of_artery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 407], 'phecode_447_1'))
subarachnoid_hemorrhage_phe = TensorMap('phecode_430.1', Interpretation.CATEGORICAL, channel_map={'no_subarachnoid_hemorrhage':0, 'subarachnoid_hemorrhage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 458], 'phecode_430_1'))
subdural_hemorrhage_phe = TensorMap('phecode_430.3', Interpretation.CATEGORICAL, channel_map={'no_subdural_hemorrhage':0, 'subdural_hemorrhage':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1116], 'phecode_430_3'))
subjective_visual_disturbances_phe = TensorMap('phecode_368.9', Interpretation.CATEGORICAL, channel_map={'no_subjective_visual_disturbances':0, 'subjective_visual_disturbances':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 499], 'phecode_368_9'))
substance_addiction_and_disorders_phe = TensorMap('phecode_316', Interpretation.CATEGORICAL, channel_map={'no_substance_addiction_and_disorders':0, 'substance_addiction_and_disorders':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 833], 'phecode_316'))
suicidal_ideation_or_attempt_phe = TensorMap('phecode_297', Interpretation.CATEGORICAL, channel_map={'no_suicidal_ideation_or_attempt':0, 'suicidal_ideation_or_attempt':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15151], 'phecode_297'))
suicide_or_self_inflicted_injury_phe = TensorMap('phecode_297.2', Interpretation.CATEGORICAL, channel_map={'no_suicide_or_self_inflicted_injury':0, 'suicide_or_self_inflicted_injury':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 141], 'phecode_297_2'))
superficial_cellulitis_and_abscess_phe = TensorMap('phecode_681', Interpretation.CATEGORICAL, channel_map={'no_superficial_cellulitis_and_abscess':0, 'superficial_cellulitis_and_abscess':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 683], 'phecode_681'))
superficial_injury_without_mention_of_infection_phe = TensorMap('phecode_915', Interpretation.CATEGORICAL, channel_map={'no_superficial_injury_without_mention_of_infection':0, 'superficial_injury_without_mention_of_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 74], 'phecode_915'))
suppurative_and_unspecified_otitis_media_phe = TensorMap('phecode_381.11', Interpretation.CATEGORICAL, channel_map={'no_suppurative_and_unspecified_otitis_media':0, 'suppurative_and_unspecified_otitis_media':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 476], 'phecode_381_11'))
supraventricular_premature_beats_phe = TensorMap('phecode_427.61', Interpretation.CATEGORICAL, channel_map={'no_supraventricular_premature_beats':0, 'supraventricular_premature_beats':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2551], 'phecode_427_61'))
swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos_phe = TensorMap('phecode_293.1', Interpretation.CATEGORICAL, channel_map={'no_swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos':0, 'swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 401], 'phecode_293_1'))
swelling_of_limb_phe = TensorMap('phecode_771.1', Interpretation.CATEGORICAL, channel_map={'no_swelling_of_limb':0, 'swelling_of_limb':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 426], 'phecode_771_1'))
symbolic_dysfunction_phe = TensorMap('phecode_292.12', Interpretation.CATEGORICAL, channel_map={'no_symbolic_dysfunction':0, 'symbolic_dysfunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 14705], 'phecode_292_12'))
symptomatic_menopause_phe = TensorMap('phecode_627.2', Interpretation.CATEGORICAL, channel_map={'no_symptomatic_menopause':0, 'symptomatic_menopause':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2127], 'phecode_627_2'))
symptoms_affecting_skin_phe = TensorMap('phecode_687', Interpretation.CATEGORICAL, channel_map={'no_symptoms_affecting_skin':0, 'symptoms_affecting_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1369], 'phecode_687'))
symptoms_and_disorders_of_the_joints_phe = TensorMap('phecode_741', Interpretation.CATEGORICAL, channel_map={'no_symptoms_and_disorders_of_the_joints':0, 'symptoms_and_disorders_of_the_joints':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 22], 'phecode_741'))
symptoms_concerning_nutrition_metabolism_and_development_phe = TensorMap('phecode_1002', Interpretation.CATEGORICAL, channel_map={'no_symptoms_concerning_nutrition_metabolism_and_development':0, 'symptoms_concerning_nutrition_metabolism_and_development':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 55], 'phecode_1002'))
symptoms_involving_cardiovascular_system_phe = TensorMap('phecode_429.3', Interpretation.CATEGORICAL, channel_map={'no_symptoms_involving_cardiovascular_system':0, 'symptoms_involving_cardiovascular_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1661], 'phecode_429_3'))
symptoms_involving_digestive_system_phe = TensorMap('phecode_561', Interpretation.CATEGORICAL, channel_map={'no_symptoms_involving_digestive_system':0, 'symptoms_involving_digestive_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23], 'phecode_561'))
symptoms_involving_female_genital_tract_phe = TensorMap('phecode_624', Interpretation.CATEGORICAL, channel_map={'no_symptoms_involving_female_genital_tract':0, 'symptoms_involving_female_genital_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_624'))
symptoms_involving_head_and_neck_phe = TensorMap('phecode_293', Interpretation.CATEGORICAL, channel_map={'no_symptoms_involving_head_and_neck':0, 'symptoms_involving_head_and_neck':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 178], 'phecode_293'))
symptoms_involving_nervous_and_musculoskeletal_systems_phe = TensorMap('phecode_781', Interpretation.CATEGORICAL, channel_map={'no_symptoms_involving_nervous_and_musculoskeletal_systems':0, 'symptoms_involving_nervous_and_musculoskeletal_systems':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 24], 'phecode_781'))
symptoms_involving_respiratory_system_and_other_chest_symptoms_phe = TensorMap('phecode_519.9', Interpretation.CATEGORICAL, channel_map={'no_symptoms_involving_respiratory_system_and_other_chest_symptoms':0, 'symptoms_involving_respiratory_system_and_other_chest_symptoms':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1173], 'phecode_519_9'))
symptoms_of_the_muscles_phe = TensorMap('phecode_772', Interpretation.CATEGORICAL, channel_map={'no_symptoms_of_the_muscles':0, 'symptoms_of_the_muscles':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1612], 'phecode_772'))
syncope_and_collapse_phe = TensorMap('phecode_788', Interpretation.CATEGORICAL, channel_map={'no_syncope_and_collapse':0, 'syncope_and_collapse':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 37], 'phecode_788'))
synoviopathy_phe = TensorMap('phecode_726.2', Interpretation.CATEGORICAL, channel_map={'no_synoviopathy':0, 'synoviopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1028], 'phecode_726_2'))
synovitis_and_tenosynovitis_phe = TensorMap('phecode_727.1', Interpretation.CATEGORICAL, channel_map={'no_synovitis_and_tenosynovitis':0, 'synovitis_and_tenosynovitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 147], 'phecode_727_1'))
systemic_lupus_erythematosus_phe = TensorMap('phecode_695.42', Interpretation.CATEGORICAL, channel_map={'no_systemic_lupus_erythematosus':0, 'systemic_lupus_erythematosus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 875], 'phecode_695_42'))
systemic_sclerosis_phe = TensorMap('phecode_709.3', Interpretation.CATEGORICAL, channel_map={'no_systemic_sclerosis':0, 'systemic_sclerosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1766], 'phecode_709_3'))
tachycardia_nos_phe = TensorMap('phecode_427.7', Interpretation.CATEGORICAL, channel_map={'no_tachycardia_nos':0, 'tachycardia_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 138], 'phecode_427_7'))
takayasus_disease_phe = TensorMap('phecode_446.7', Interpretation.CATEGORICAL, channel_map={'no_takayasus_disease':0, 'takayasus_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41666], 'phecode_446_7'))
telogen_effluvium_phe = TensorMap('phecode_704.12', Interpretation.CATEGORICAL, channel_map={'no_telogen_effluvium':0, 'telogen_effluvium':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 250000], 'phecode_704_12'))
temporomandibular_joint_disorder_unspecified_phe = TensorMap('phecode_526.41', Interpretation.CATEGORICAL, channel_map={'no_temporomandibular_joint_disorder_unspecified':0, 'temporomandibular_joint_disorder_unspecified':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1519], 'phecode_526_41'))
tension_headache_phe = TensorMap('phecode_306.9', Interpretation.CATEGORICAL, channel_map={'no_tension_headache':0, 'tension_headache':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1082], 'phecode_306_9'))
testicular_dysfunction_phe = TensorMap('phecode_257', Interpretation.CATEGORICAL, channel_map={'no_testicular_dysfunction':0, 'testicular_dysfunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 22727], 'phecode_257'))
testicular_hypofunction_phe = TensorMap('phecode_257.1', Interpretation.CATEGORICAL, channel_map={'no_testicular_hypofunction':0, 'testicular_hypofunction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2092], 'phecode_257_1'))
throat_pain_phe = TensorMap('phecode_478', Interpretation.CATEGORICAL, channel_map={'no_throat_pain':0, 'throat_pain':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 959], 'phecode_478'))
thromboangiitis_obliterans_phe = TensorMap('phecode_446.1', Interpretation.CATEGORICAL, channel_map={'no_thromboangiitis_obliterans':0, 'thromboangiitis_obliterans':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27777], 'phecode_446_1'))
thrombocytopenia_phe = TensorMap('phecode_287.3', Interpretation.CATEGORICAL, channel_map={'no_thrombocytopenia':0, 'thrombocytopenia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 249], 'phecode_287_3'))
thrombotic_microangiopathy_phe = TensorMap('phecode_446.8', Interpretation.CATEGORICAL, channel_map={'no_thrombotic_microangiopathy':0, 'thrombotic_microangiopathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20000], 'phecode_446_8'))
thyroid_cancer_phe = TensorMap('phecode_193', Interpretation.CATEGORICAL, channel_map={'no_thyroid_cancer':0, 'thyroid_cancer':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 754], 'phecode_193'))
thyroiditis_phe = TensorMap('phecode_245', Interpretation.CATEGORICAL, channel_map={'no_thyroiditis':0, 'thyroiditis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5813], 'phecode_245'))
thyroiditis_acute_and_subacute_phe = TensorMap('phecode_245.1', Interpretation.CATEGORICAL, channel_map={'no_thyroiditis_acute_and_subacute':0, 'thyroiditis_acute_and_subacute':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 23809], 'phecode_245_1'))
thyrotoxicosis_with_or_without_goiter_phe = TensorMap('phecode_242', Interpretation.CATEGORICAL, channel_map={'no_thyrotoxicosis_with_or_without_goiter':0, 'thyrotoxicosis_with_or_without_goiter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 234], 'phecode_242'))
tics_and_choreas_phe = TensorMap('phecode_333.3', Interpretation.CATEGORICAL, channel_map={'no_tics_and_choreas':0, 'tics_and_choreas':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 800], 'phecode_333_3'))
tics_and_stuttering_phe = TensorMap('phecode_313.2', Interpretation.CATEGORICAL, channel_map={'no_tics_and_stuttering':0, 'tics_and_stuttering':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15151], 'phecode_313_2'))
tinnitus_phe = TensorMap('phecode_389.4', Interpretation.CATEGORICAL, channel_map={'no_tinnitus':0, 'tinnitus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 592], 'phecode_389_4'))
tobacco_use_disorder_phe = TensorMap('phecode_318', Interpretation.CATEGORICAL, channel_map={'no_tobacco_use_disorder':0, 'tobacco_use_disorder':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 17], 'phecode_318'))
tooth_complications_likely_association_with_other_diseases_phe = TensorMap('phecode_521.4', Interpretation.CATEGORICAL, channel_map={'no_tooth_complications_likely_association_with_other_diseases':0, 'tooth_complications_likely_association_with_other_diseases':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15625], 'phecode_521_4'))
torsion_dystonia_phe = TensorMap('phecode_333.4', Interpretation.CATEGORICAL, channel_map={'no_torsion_dystonia':0, 'torsion_dystonia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1269], 'phecode_333_4'))
torticollis_phe = TensorMap('phecode_723.1', Interpretation.CATEGORICAL, channel_map={'no_torticollis':0, 'torticollis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4545], 'phecode_723_1'))
toxic_effect_of_carbon_monoxide_phe = TensorMap('phecode_986', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_carbon_monoxide':0, 'toxic_effect_of_carbon_monoxide':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 7352], 'phecode_986'))
toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis_phe = TensorMap('phecode_983', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis':0, 'toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16666], 'phecode_983'))
toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents_phe = TensorMap('phecode_981', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents':0, 'toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 278], 'phecode_981'))
toxic_effect_of_noxious_substances_eaten_as_food_phe = TensorMap('phecode_988', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_noxious_substances_eaten_as_food':0, 'toxic_effect_of_noxious_substances_eaten_as_food':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11111], 'phecode_988'))
toxic_effect_of_other_gases_fumes_or_vapors_phe = TensorMap('phecode_987', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_other_gases_fumes_or_vapors':0, 'toxic_effect_of_other_gases_fumes_or_vapors':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3623], 'phecode_987'))
toxic_effect_of_other_metals_phe = TensorMap('phecode_985', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_other_metals':0, 'toxic_effect_of_other_metals':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8928], 'phecode_985'))
toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source_phe = TensorMap('phecode_989', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source':0, 'toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 278], 'phecode_989'))
toxic_effect_of_venom_phe = TensorMap('phecode_913', Interpretation.CATEGORICAL, channel_map={'no_toxic_effect_of_venom':0, 'toxic_effect_of_venom':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3067], 'phecode_913'))
toxic_erythema_phe = TensorMap('phecode_695.1', Interpretation.CATEGORICAL, channel_map={'no_toxic_erythema':0, 'toxic_erythema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3333], 'phecode_695_1'))
toxic_gastroenteritis_and_colitis_phe = TensorMap('phecode_569.1', Interpretation.CATEGORICAL, channel_map={'no_toxic_gastroenteritis_and_colitis':0, 'toxic_gastroenteritis_and_colitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1582], 'phecode_569_1'))
toxic_multinodular_goiter_phe = TensorMap('phecode_242.2', Interpretation.CATEGORICAL, channel_map={'no_toxic_multinodular_goiter':0, 'toxic_multinodular_goiter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3546], 'phecode_242_2'))
transient_cerebral_ischemia_phe = TensorMap('phecode_433.31', Interpretation.CATEGORICAL, channel_map={'no_transient_cerebral_ischemia':0, 'transient_cerebral_ischemia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 164], 'phecode_433_31'))
transient_mental_disorders_due_to_conditions_classified_elsewhere_phe = TensorMap('phecode_291.1', Interpretation.CATEGORICAL, channel_map={'no_transient_mental_disorders_due_to_conditions_classified_elsewhere':0, 'transient_mental_disorders_due_to_conditions_classified_elsewhere':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2551], 'phecode_291_1'))
transitory_tachypnea_or_apnea_of_newborn_phe = TensorMap('phecode_656.26', Interpretation.CATEGORICAL, channel_map={'no_transitory_tachypnea_or_apnea_of_newborn':0, 'transitory_tachypnea_or_apnea_of_newborn':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 166666], 'phecode_656_26'))
traumatic_amputation_phe = TensorMap('phecode_872', Interpretation.CATEGORICAL, channel_map={'no_traumatic_amputation':0, 'traumatic_amputation':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 838], 'phecode_872'))
traumatic_and_surgical_subcutaneous_emphysema_phe = TensorMap('phecode_958.2', Interpretation.CATEGORICAL, channel_map={'no_traumatic_and_surgical_subcutaneous_emphysema':0, 'traumatic_and_surgical_subcutaneous_emphysema':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4273], 'phecode_958_2'))
traumatic_arthropathy_phe = TensorMap('phecode_836', Interpretation.CATEGORICAL, channel_map={'no_traumatic_arthropathy':0, 'traumatic_arthropathy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1529], 'phecode_836'))
traumatic_cataract_phe = TensorMap('phecode_366.3', Interpretation.CATEGORICAL, channel_map={'no_traumatic_cataract':0, 'traumatic_cataract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3521], 'phecode_366_3'))
trigeminal_nerve_disorders_cn5_phe = TensorMap('phecode_352.1', Interpretation.CATEGORICAL, channel_map={'no_trigeminal_nerve_disorders_cn5':0, 'trigeminal_nerve_disorders_cn5':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 20], 'phecode_352_1'))
tuberculosis_phe = TensorMap('phecode_10', Interpretation.CATEGORICAL, channel_map={'no_tuberculosis':0, 'tuberculosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1457], 'phecode_10'))
tympanosclerosis_and_middle_ear_disease_related_to_otitis_media_phe = TensorMap('phecode_385.5', Interpretation.CATEGORICAL, channel_map={'no_tympanosclerosis_and_middle_ear_disease_related_to_otitis_media':0, 'tympanosclerosis_and_middle_ear_disease_related_to_otitis_media':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1694], 'phecode_385_5'))
type_1_diabetes_phe = TensorMap('phecode_250.1', Interpretation.CATEGORICAL, channel_map={'no_type_1_diabetes':0, 'type_1_diabetes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 130], 'phecode_250_1'))
type_1_diabetes_with_ketoacidosis_phe = TensorMap('phecode_250.11', Interpretation.CATEGORICAL, channel_map={'no_type_1_diabetes_with_ketoacidosis':0, 'type_1_diabetes_with_ketoacidosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1329], 'phecode_250_11'))
type_1_diabetes_with_neurological_manifestations_phe = TensorMap('phecode_250.14', Interpretation.CATEGORICAL, channel_map={'no_type_1_diabetes_with_neurological_manifestations':0, 'type_1_diabetes_with_neurological_manifestations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1968], 'phecode_250_14'))
type_1_diabetes_with_ophthalmic_manifestations_phe = TensorMap('phecode_250.13', Interpretation.CATEGORICAL, channel_map={'no_type_1_diabetes_with_ophthalmic_manifestations':0, 'type_1_diabetes_with_ophthalmic_manifestations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 702], 'phecode_250_13'))
type_1_diabetes_with_renal_manifestations_phe = TensorMap('phecode_250.12', Interpretation.CATEGORICAL, channel_map={'no_type_1_diabetes_with_renal_manifestations':0, 'type_1_diabetes_with_renal_manifestations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3816], 'phecode_250_12'))
type_2_diabetes_phe = TensorMap('phecode_250.2', Interpretation.CATEGORICAL, channel_map={'no_type_2_diabetes':0, 'type_2_diabetes':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 16], 'phecode_250_2'))
type_2_diabetes_with_ketoacidosis_phe = TensorMap('phecode_250.21', Interpretation.CATEGORICAL, channel_map={'no_type_2_diabetes_with_ketoacidosis':0, 'type_2_diabetes_with_ketoacidosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2192], 'phecode_250_21'))
type_2_diabetes_with_neurological_manifestations_phe = TensorMap('phecode_250.24', Interpretation.CATEGORICAL, channel_map={'no_type_2_diabetes_with_neurological_manifestations':0, 'type_2_diabetes_with_neurological_manifestations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 502], 'phecode_250_24'))
type_2_diabetes_with_ophthalmic_manifestations_phe = TensorMap('phecode_250.23', Interpretation.CATEGORICAL, channel_map={'no_type_2_diabetes_with_ophthalmic_manifestations':0, 'type_2_diabetes_with_ophthalmic_manifestations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 219], 'phecode_250_23'))
type_2_diabetes_with_renal_manifestations_phe = TensorMap('phecode_250.22', Interpretation.CATEGORICAL, channel_map={'no_type_2_diabetes_with_renal_manifestations':0, 'type_2_diabetes_with_renal_manifestations':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1272], 'phecode_250_22'))
ulcer_of_esophagus_phe = TensorMap('phecode_530.12', Interpretation.CATEGORICAL, channel_map={'no_ulcer_of_esophagus':0, 'ulcer_of_esophagus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 73], 'phecode_530_12'))
ulceration_of_intestine_phe = TensorMap('phecode_556.1', Interpretation.CATEGORICAL, channel_map={'no_ulceration_of_intestine':0, 'ulceration_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 456], 'phecode_556_1'))
ulceration_of_the_lower_gi_tract_phe = TensorMap('phecode_556', Interpretation.CATEGORICAL, channel_map={'no_ulceration_of_the_lower_gi_tract':0, 'ulceration_of_the_lower_gi_tract':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 815], 'phecode_556'))
ulcerative_colitis_phe = TensorMap('phecode_555.2', Interpretation.CATEGORICAL, channel_map={'no_ulcerative_colitis':0, 'ulcerative_colitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 122], 'phecode_555_2'))
ulcerative_colitis_chronic_phe = TensorMap('phecode_555.21', Interpretation.CATEGORICAL, channel_map={'no_ulcerative_colitis_chronic':0, 'ulcerative_colitis_chronic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 623], 'phecode_555_21'))
umbilical_cord_complications_during_labor_and_delivery_phe = TensorMap('phecode_663', Interpretation.CATEGORICAL, channel_map={'no_umbilical_cord_complications_during_labor_and_delivery':0, 'umbilical_cord_complications_during_labor_and_delivery':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 621], 'phecode_663'))
umbilical_hernia_phe = TensorMap('phecode_550.4', Interpretation.CATEGORICAL, channel_map={'no_umbilical_hernia':0, 'umbilical_hernia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 92], 'phecode_550_4'))
unspecified_diffuse_connective_tissue_disease_phe = TensorMap('phecode_709.7', Interpretation.CATEGORICAL, channel_map={'no_unspecified_diffuse_connective_tissue_disease':0, 'unspecified_diffuse_connective_tissue_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4], 'phecode_709_7'))
unspecified_disorder_of_lipoid_metabolism_phe = TensorMap('phecode_272.9', Interpretation.CATEGORICAL, channel_map={'no_unspecified_disorder_of_lipoid_metabolism':0, 'unspecified_disorder_of_lipoid_metabolism':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1237], 'phecode_272_9'))
unspecified_erythematous_condition_phe = TensorMap('phecode_695.9', Interpretation.CATEGORICAL, channel_map={'no_unspecified_erythematous_condition':0, 'unspecified_erythematous_condition':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 683], 'phecode_695_9'))
unspecified_monoarthritis_phe = TensorMap('phecode_716.2', Interpretation.CATEGORICAL, channel_map={'no_unspecified_monoarthritis':0, 'unspecified_monoarthritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 21], 'phecode_716_2'))
unspecified_osteomyelitis_phe = TensorMap('phecode_710.19', Interpretation.CATEGORICAL, channel_map={'no_unspecified_osteomyelitis':0, 'unspecified_osteomyelitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3597], 'phecode_710_19'))
unspecified_polyarthropathy_or_polyarthritis_phe = TensorMap('phecode_716.1', Interpretation.CATEGORICAL, channel_map={'no_unspecified_polyarthropathy_or_polyarthritis':0, 'unspecified_polyarthropathy_or_polyarthritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 85], 'phecode_716_1'))
unstable_angina_intermediate_coronary_syndrome_phe = TensorMap('phecode_411.1', Interpretation.CATEGORICAL, channel_map={'no_unstable_angina_intermediate_coronary_syndrome':0, 'unstable_angina_intermediate_coronary_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 72], 'phecode_411_1'))
upper_gastrointestinal_congenital_anomalies_phe = TensorMap('phecode_750.1', Interpretation.CATEGORICAL, channel_map={'no_upper_gastrointestinal_congenital_anomalies':0, 'upper_gastrointestinal_congenital_anomalies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 15151], 'phecode_750_1'))
urethral_discharge_phe = TensorMap('phecode_599.7', Interpretation.CATEGORICAL, channel_map={'no_urethral_discharge':0, 'urethral_discharge':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 8620], 'phecode_599_7'))
urethral_stricture_due_to_infecton_phe = TensorMap('phecode_592.3', Interpretation.CATEGORICAL, channel_map={'no_urethral_stricture_due_to_infecton':0, 'urethral_stricture_due_to_infecton':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_592_3'))
urethral_stricture_not_specified_as_infectious_phe = TensorMap('phecode_597.1', Interpretation.CATEGORICAL, channel_map={'no_urethral_stricture_not_specified_as_infectious':0, 'urethral_stricture_not_specified_as_infectious':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 106], 'phecode_597_1'))
urethral_syndrome_phe = TensorMap('phecode_592.21', Interpretation.CATEGORICAL, channel_map={'no_urethral_syndrome':0, 'urethral_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 10638], 'phecode_592_21'))
urethritis_and_urethral_syndrome_phe = TensorMap('phecode_592.2', Interpretation.CATEGORICAL, channel_map={'no_urethritis_and_urethral_syndrome':0, 'urethritis_and_urethral_syndrome':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3225], 'phecode_592_2'))
urinary_calculus_phe = TensorMap('phecode_594', Interpretation.CATEGORICAL, channel_map={'no_urinary_calculus':0, 'urinary_calculus':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 56], 'phecode_594'))
urinary_complications_nec_phe = TensorMap('phecode_597.2', Interpretation.CATEGORICAL, channel_map={'no_urinary_complications_nec':0, 'urinary_complications_nec':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4424], 'phecode_597_2'))
urinary_incontinence_phe = TensorMap('phecode_599.4', Interpretation.CATEGORICAL, channel_map={'no_urinary_incontinence':0, 'urinary_incontinence':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 41], 'phecode_599_4'))
urinary_obstruction_phe = TensorMap('phecode_599.1', Interpretation.CATEGORICAL, channel_map={'no_urinary_obstruction':0, 'urinary_obstruction':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2074], 'phecode_599_1'))
urinary_tract_infection_phe = TensorMap('phecode_591', Interpretation.CATEGORICAL, channel_map={'no_urinary_tract_infection':0, 'urinary_tract_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 27], 'phecode_591'))
urticaria_phe = TensorMap('phecode_947', Interpretation.CATEGORICAL, channel_map={'no_urticaria':0, 'urticaria':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 592], 'phecode_947'))
uterine_leiomyoma_phe = TensorMap('phecode_218.1', Interpretation.CATEGORICAL, channel_map={'no_uterine_leiomyoma':0, 'uterine_leiomyoma':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35], 'phecode_218_1'))
uterine_uterovaginal_prolapse_phe = TensorMap('phecode_618.2', Interpretation.CATEGORICAL, channel_map={'no_uterine_uterovaginal_prolapse':0, 'uterine_uterovaginal_prolapse':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 70], 'phecode_618_2'))
uveitis_noninfectious_or_nos_phe = TensorMap('phecode_371.1', Interpretation.CATEGORICAL, channel_map={'no_uveitis_noninfectious_or_nos':0, 'uveitis_noninfectious_or_nos':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 679], 'phecode_371_1'))
vaginal_enterocele_congenital_or_acquired_phe = TensorMap('phecode_618.6', Interpretation.CATEGORICAL, channel_map={'no_vaginal_enterocele_congenital_or_acquired':0, 'vaginal_enterocele_congenital_or_acquired':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 567], 'phecode_618_6'))
vaginitis_and_vulvovaginitis_phe = TensorMap('phecode_614.52', Interpretation.CATEGORICAL, channel_map={'no_vaginitis_and_vulvovaginitis':0, 'vaginitis_and_vulvovaginitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 963], 'phecode_614_52'))
valvular_heart_disease_heart_chambers_phe = TensorMap('phecode_747.12', Interpretation.CATEGORICAL, channel_map={'no_valvular_heart_disease_heart_chambers':0, 'valvular_heart_disease_heart_chambers':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1340], 'phecode_747_12'))
varicella_infection_phe = TensorMap('phecode_79.1', Interpretation.CATEGORICAL, channel_map={'no_varicella_infection':0, 'varicella_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5882], 'phecode_79_1'))
varicose_veins_phe = TensorMap('phecode_454', Interpretation.CATEGORICAL, channel_map={'no_varicose_veins':0, 'varicose_veins':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 721], 'phecode_454'))
varicose_veins_of_lower_extremity_phe = TensorMap('phecode_454.1', Interpretation.CATEGORICAL, channel_map={'no_varicose_veins_of_lower_extremity':0, 'varicose_veins_of_lower_extremity':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 37], 'phecode_454_1'))
varicose_veins_of_lower_extremity_symptomtic_phe = TensorMap('phecode_454.11', Interpretation.CATEGORICAL, channel_map={'no_varicose_veins_of_lower_extremity_symptomtic':0, 'varicose_veins_of_lower_extremity_symptomtic':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 472], 'phecode_454_11'))
vascular_dementia_phe = TensorMap('phecode_290.16', Interpretation.CATEGORICAL, channel_map={'no_vascular_dementia':0, 'vascular_dementia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1000], 'phecode_290_16'))
vascular_disorders_of_kidney_hypertrophy_phe = TensorMap('phecode_586.3', Interpretation.CATEGORICAL, channel_map={'no_vascular_disorders_of_kidney_hypertrophy':0, 'vascular_disorders_of_kidney_hypertrophy':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4032], 'phecode_586_3'))
vascular_disorders_of_skin_phe = TensorMap('phecode_694.3', Interpretation.CATEGORICAL, channel_map={'no_vascular_disorders_of_skin':0, 'vascular_disorders_of_skin':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2577], 'phecode_694_3'))
vascular_hamartomas_and_non_neoplastic_nevi_phe = TensorMap('phecode_217', Interpretation.CATEGORICAL, channel_map={'no_vascular_hamartomas_and_non_neoplastic_nevi':0, 'vascular_hamartomas_and_non_neoplastic_nevi':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 6849], 'phecode_217'))
vascular_insufficiency_of_intestine_phe = TensorMap('phecode_441', Interpretation.CATEGORICAL, channel_map={'no_vascular_insufficiency_of_intestine':0, 'vascular_insufficiency_of_intestine':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1026], 'phecode_441'))
venereal_diseases_due_to_chlamydia_trachomatis_phe = TensorMap('phecode_90.3', Interpretation.CATEGORICAL, channel_map={'no_venereal_diseases_due_to_chlamydia_trachomatis':0, 'venereal_diseases_due_to_chlamydia_trachomatis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 45454], 'phecode_90_3'))
venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium_phe = TensorMap('phecode_671', Interpretation.CATEGORICAL, channel_map={'no_venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium':0, 'venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1886], 'phecode_671'))
ventral_hernia_phe = TensorMap('phecode_550.5', Interpretation.CATEGORICAL, channel_map={'no_ventral_hernia':0, 'ventral_hernia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 101], 'phecode_550_5'))
ventricular_fibrillation_and_flutter_phe = TensorMap('phecode_427.41', Interpretation.CATEGORICAL, channel_map={'no_ventricular_fibrillation_and_flutter':0, 'ventricular_fibrillation_and_flutter':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 968], 'phecode_427_41'))
vertiginous_syndromes_and_other_disorders_of_vestibular_system_phe = TensorMap('phecode_386', Interpretation.CATEGORICAL, channel_map={'no_vertiginous_syndromes_and_other_disorders_of_vestibular_system':0, 'vertiginous_syndromes_and_other_disorders_of_vestibular_system':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 165], 'phecode_386'))
vesicoureteral_reflux_phe = TensorMap('phecode_586.12', Interpretation.CATEGORICAL, channel_map={'no_vesicoureteral_reflux':0, 'vesicoureteral_reflux':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 11363], 'phecode_586_12'))
villonodular_synovitis_phe = TensorMap('phecode_741.6', Interpretation.CATEGORICAL, channel_map={'no_villonodular_synovitis':0, 'villonodular_synovitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 35714], 'phecode_741_6'))
viral_enteritis_phe = TensorMap('phecode_8.6', Interpretation.CATEGORICAL, channel_map={'no_viral_enteritis':0, 'viral_enteritis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 380], 'phecode_8_6'))
viral_hepatitis_phe = TensorMap('phecode_70', Interpretation.CATEGORICAL, channel_map={'no_viral_hepatitis':0, 'viral_hepatitis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 441], 'phecode_70'))
viral_hepatitis_a_phe = TensorMap('phecode_70.1', Interpretation.CATEGORICAL, channel_map={'no_viral_hepatitis_a':0, 'viral_hepatitis_a':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 5102], 'phecode_70_1'))
viral_hepatitis_b_phe = TensorMap('phecode_70.2', Interpretation.CATEGORICAL, channel_map={'no_viral_hepatitis_b':0, 'viral_hepatitis_b':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1623], 'phecode_70_2'))
viral_hepatitis_c_phe = TensorMap('phecode_70.3', Interpretation.CATEGORICAL, channel_map={'no_viral_hepatitis_c':0, 'viral_hepatitis_c':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 994], 'phecode_70_3'))
viral_infection_phe = TensorMap('phecode_79', Interpretation.CATEGORICAL, channel_map={'no_viral_infection':0, 'viral_infection':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 88], 'phecode_79'))
viral_pneumonia_phe = TensorMap('phecode_480.2', Interpretation.CATEGORICAL, channel_map={'no_viral_pneumonia':0, 'viral_pneumonia':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4201], 'phecode_480_2'))
viral_warts_hpv_phe = TensorMap('phecode_78', Interpretation.CATEGORICAL, channel_map={'no_viral_warts_hpv':0, 'viral_warts_hpv':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 402], 'phecode_78'))
visual_disturbances_phe = TensorMap('phecode_368', Interpretation.CATEGORICAL, channel_map={'no_visual_disturbances':0, 'visual_disturbances':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 303], 'phecode_368'))
visual_field_defects_phe = TensorMap('phecode_368.4', Interpretation.CATEGORICAL, channel_map={'no_visual_field_defects':0, 'visual_field_defects':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 946], 'phecode_368_4'))
vitamin_a_deficiency_phe = TensorMap('phecode_261.1', Interpretation.CATEGORICAL, channel_map={'no_vitamin_a_deficiency':0, 'vitamin_a_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 71428], 'phecode_261_1'))
vitamin_b_complex_deficiencies_phe = TensorMap('phecode_261.2', Interpretation.CATEGORICAL, channel_map={'no_vitamin_b_complex_deficiencies':0, 'vitamin_b_complex_deficiencies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 364], 'phecode_261_2'))
vitamin_c_deficiencies_phe = TensorMap('phecode_261.3', Interpretation.CATEGORICAL, channel_map={'no_vitamin_c_deficiencies':0, 'vitamin_c_deficiencies':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 26315], 'phecode_261_3'))
vitamin_d_deficiency_phe = TensorMap('phecode_261.4', Interpretation.CATEGORICAL, channel_map={'no_vitamin_d_deficiency':0, 'vitamin_d_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 362], 'phecode_261_4'))
vitamin_deficiency_phe = TensorMap('phecode_261', Interpretation.CATEGORICAL, channel_map={'no_vitamin_deficiency':0, 'vitamin_deficiency':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 13513], 'phecode_261'))
vitiligo_phe = TensorMap('phecode_694.1', Interpretation.CATEGORICAL, channel_map={'no_vitiligo':0, 'vitiligo':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 4901], 'phecode_694_1'))
voice_disturbance_phe = TensorMap('phecode_473.4', Interpretation.CATEGORICAL, channel_map={'no_voice_disturbance':0, 'voice_disturbance':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 355], 'phecode_473_4'))
von_willebrands_disease_phe = TensorMap('phecode_286.11', Interpretation.CATEGORICAL, channel_map={'no_von_willebrands_disease':0, 'von_willebrands_disease':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 3267], 'phecode_286_11'))
wegeners_granulomatosis_phe = TensorMap('phecode_446.4', Interpretation.CATEGORICAL, channel_map={'no_wegeners_granulomatosis':0, 'wegeners_granulomatosis':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 2008], 'phecode_446_4'))
wheezing_phe = TensorMap('phecode_512.1', Interpretation.CATEGORICAL, channel_map={'no_wheezing':0, 'wheezing':1}, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, 1225], 'phecode_512_1'))


#  TensorMaps for prevalent and incident Phecode disease phenotypes
abdominal_aortic_aneurysm_phe_pi = TensorMap('phecode_442.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442.11_date'), channel_map={'no_abdominal_aortic_aneurysm':0, 'abdominal_aortic_aneurysm_prevalent':1, 'abdominal_aortic_aneurysm_incident':2}, loss=weighted_crossentropy([1.0, 1724, 435], 'abdominal_aortic_aneurysm_pi'))
abdominal_hernia_phe_pi = TensorMap('phecode_550',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_550_date'), channel_map={'no_abdominal_hernia':0, 'abdominal_hernia_prevalent':1, 'abdominal_hernia_incident':2}, loss=weighted_crossentropy([1.0, 16, 14], 'abdominal_hernia_pi'))
abdominal_pain_phe_pi = TensorMap('phecode_785',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_785_date'), channel_map={'no_abdominal_pain':0, 'abdominal_pain_prevalent':1, 'abdominal_pain_incident':2}, loss=weighted_crossentropy([1.0, 16, 19], 'abdominal_pain_pi'))
abnormal_findings_examination_of_lungs_phe_pi = TensorMap('phecode_514',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_514_date'), channel_map={'no_abnormal_findings_examination_of_lungs':0, 'abnormal_findings_examination_of_lungs_prevalent':1, 'abnormal_findings_examination_of_lungs_incident':2}, loss=weighted_crossentropy([1.0, 492, 164], 'abnormal_findings_examination_of_lungs_pi'))
abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area_phe_pi = TensorMap('phecode_564.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_564.8_date'), channel_map={'no_abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area':0, 'abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area_prevalent':1, 'abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area_incident':2}, loss=weighted_crossentropy([1.0, 608, 276], 'abnormal_findings_on_exam_of_gastrointestinal_tract_abdominal_area_pi'))
abnormal_findings_on_examination_of_urine_phe_pi = TensorMap('phecode_598',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_598_date'), channel_map={'no_abnormal_findings_on_examination_of_urine':0, 'abnormal_findings_on_examination_of_urine_prevalent':1, 'abnormal_findings_on_examination_of_urine_incident':2}, loss=weighted_crossentropy([1.0, 256, 189], 'abnormal_findings_on_examination_of_urine_pi'))
abnormal_findings_on_study_of_brain_and_or_nervous_system_phe_pi = TensorMap('phecode_346',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_346_date'), channel_map={'no_abnormal_findings_on_study_of_brain_and_or_nervous_system':0, 'abnormal_findings_on_study_of_brain_and_or_nervous_system_prevalent':1, 'abnormal_findings_on_study_of_brain_and_or_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 71428, 250000], 'abnormal_findings_on_study_of_brain_and_or_nervous_system_pi'))
abnormal_function_study_of_cardiovascular_system_phe_pi = TensorMap('phecode_429.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_429.2_date'), channel_map={'no_abnormal_function_study_of_cardiovascular_system':0, 'abnormal_function_study_of_cardiovascular_system_prevalent':1, 'abnormal_function_study_of_cardiovascular_system_incident':2}, loss=weighted_crossentropy([1.0, 1689, 871], 'abnormal_function_study_of_cardiovascular_system_pi'))
abnormal_granulation_tissue_phe_pi = TensorMap('phecode_701.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701.5_date'), channel_map={'no_abnormal_granulation_tissue':0, 'abnormal_granulation_tissue_prevalent':1, 'abnormal_granulation_tissue_incident':2}, loss=weighted_crossentropy([1.0, 1524, 1683], 'abnormal_granulation_tissue_pi'))
abnormal_heart_sounds_phe_pi = TensorMap('phecode_396',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_396_date'), channel_map={'no_abnormal_heart_sounds':0, 'abnormal_heart_sounds_prevalent':1, 'abnormal_heart_sounds_incident':2}, loss=weighted_crossentropy([1.0, 1039, 414], 'abnormal_heart_sounds_pi'))
abnormal_involuntary_movements_phe_pi = TensorMap('phecode_350.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_350.1_date'), channel_map={'no_abnormal_involuntary_movements':0, 'abnormal_involuntary_movements_prevalent':1, 'abnormal_involuntary_movements_incident':2}, loss=weighted_crossentropy([1.0, 294, 60], 'abnormal_involuntary_movements_pi'))
abnormal_mammogram_phe_pi = TensorMap('phecode_611.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_611.1_date'), channel_map={'no_abnormal_mammogram':0, 'abnormal_mammogram_prevalent':1, 'abnormal_mammogram_incident':2}, loss=weighted_crossentropy([1.0, 10638, 19230], 'abnormal_mammogram_pi'))
abnormal_posture_phe_pi = TensorMap('phecode_781.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_781.2_date'), channel_map={'no_abnormal_posture':0, 'abnormal_posture_prevalent':1, 'abnormal_posture_incident':2}, loss=weighted_crossentropy([1.0, 250000, 71428], 'abnormal_posture_pi'))
abnormal_reflex_phe_pi = TensorMap('phecode_350.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_350.5_date'), channel_map={'no_abnormal_reflex':0, 'abnormal_reflex_prevalent':1, 'abnormal_reflex_incident':2}, loss=weighted_crossentropy([1.0, 71428, 26315], 'abnormal_reflex_pi'))
abnormal_results_of_function_study_of_kidney_phe_pi = TensorMap('phecode_589',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_589_date'), channel_map={'no_abnormal_results_of_function_study_of_kidney':0, 'abnormal_results_of_function_study_of_kidney_prevalent':1, 'abnormal_results_of_function_study_of_kidney_incident':2}, loss=weighted_crossentropy([1.0, 8620, 3546], 'abnormal_results_of_function_study_of_kidney_pi'))
abnormal_results_of_function_study_of_liver_phe_pi = TensorMap('phecode_573.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.7_date'), channel_map={'no_abnormal_results_of_function_study_of_liver':0, 'abnormal_results_of_function_study_of_liver_prevalent':1, 'abnormal_results_of_function_study_of_liver_incident':2}, loss=weighted_crossentropy([1.0, 299, 130], 'abnormal_results_of_function_study_of_liver_pi'))
abnormal_results_of_function_study_of_pulmonary_system_phe_pi = TensorMap('phecode_514.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_514.1_date'), channel_map={'no_abnormal_results_of_function_study_of_pulmonary_system':0, 'abnormal_results_of_function_study_of_pulmonary_system_prevalent':1, 'abnormal_results_of_function_study_of_pulmonary_system_incident':2}, loss=weighted_crossentropy([1.0, 26315, 14285], 'abnormal_results_of_function_study_of_pulmonary_system_pi'))
abnormal_results_of_function_study_of_thyroid_phe_pi = TensorMap('phecode_246.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_246.7_date'), channel_map={'no_abnormal_results_of_function_study_of_thyroid':0, 'abnormal_results_of_function_study_of_thyroid_prevalent':1, 'abnormal_results_of_function_study_of_thyroid_incident':2}, loss=weighted_crossentropy([1.0, 20000, 13888], 'abnormal_results_of_function_study_of_thyroid_pi'))
abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc_phe_pi = TensorMap('phecode_794',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_794_date'), channel_map={'no_abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc':0, 'abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc_prevalent':1, 'abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc_incident':2}, loss=weighted_crossentropy([1.0, 3703, 17241], 'abnormal_results_of_other_function_studies_bladder_pancreas_placenta_spleen_etc_pi'))
abnormal_serum_enzyme_levels_phe_pi = TensorMap('phecode_573.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.9_date'), channel_map={'no_abnormal_serum_enzyme_levels':0, 'abnormal_serum_enzyme_levels_prevalent':1, 'abnormal_serum_enzyme_levels_incident':2}, loss=weighted_crossentropy([1.0, 3496, 2475], 'abnormal_serum_enzyme_levels_pi'))
abnormal_spermatozoa_phe_pi = TensorMap('phecode_609.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_609.2_date'), channel_map={'no_abnormal_spermatozoa':0, 'abnormal_spermatozoa_prevalent':1, 'abnormal_spermatozoa_incident':2}, loss=weighted_crossentropy([1.0, 2840, 20833], 'abnormal_spermatozoa_pi'))
abnormal_sputum_phe_pi = TensorMap('phecode_516',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_516_date'), channel_map={'no_abnormal_sputum':0, 'abnormal_sputum_prevalent':1, 'abnormal_sputum_incident':2}, loss=weighted_crossentropy([1.0, 16666, 4716], 'abnormal_sputum_pi'))
abnormal_weight_gain_phe_pi = TensorMap('phecode_278.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_278.4_date'), channel_map={'no_abnormal_weight_gain':0, 'abnormal_weight_gain_prevalent':1, 'abnormal_weight_gain_incident':2}, loss=weighted_crossentropy([1.0, 9433, 12195], 'abnormal_weight_gain_pi'))
abnormality_of_gait_phe_pi = TensorMap('phecode_350.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_350.2_date'), channel_map={'no_abnormality_of_gait':0, 'abnormality_of_gait_prevalent':1, 'abnormality_of_gait_incident':2}, loss=weighted_crossentropy([1.0, 1160, 179], 'abnormality_of_gait_pi'))
abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium_phe_pi = TensorMap('phecode_654.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_654.1_date'), channel_map={'no_abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium':0, 'abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium_prevalent':1, 'abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium_incident':2}, loss=weighted_crossentropy([1.0, 321, 6410], 'abnormality_of_organs_and_soft_tissues_of_pelvis_complicating_pregnancy_childbirth_or_the_puerperium_pi'))
abnormality_of_red_blood_cells_phe_pi = TensorMap('phecode_289.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_289.9_date'), channel_map={'no_abnormality_of_red_blood_cells':0, 'abnormality_of_red_blood_cells_prevalent':1, 'abnormality_of_red_blood_cells_incident':2}, loss=weighted_crossentropy([1.0, 11111, 5813], 'abnormality_of_red_blood_cells_pi'))
abnormality_of_secretion_of_glucagon_or_gastrin_phe_pi = TensorMap('phecode_251.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_251.8_date'), channel_map={'no_abnormality_of_secretion_of_glucagon_or_gastrin':0, 'abnormality_of_secretion_of_glucagon_or_gastrin_prevalent':1, 'abnormality_of_secretion_of_glucagon_or_gastrin_incident':2}, loss=weighted_crossentropy([1.0, 166666, 166666], 'abnormality_of_secretion_of_glucagon_or_gastrin_pi'))
abscess_or_ulceration_of_vulva_phe_pi = TensorMap('phecode_614.54',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.54_date'), channel_map={'no_abscess_or_ulceration_of_vulva':0, 'abscess_or_ulceration_of_vulva_prevalent':1, 'abscess_or_ulceration_of_vulva_incident':2}, loss=weighted_crossentropy([1.0, 2145, 4237], 'abscess_or_ulceration_of_vulva_pi'))
absent_or_infrequent_menstruation_phe_pi = TensorMap('phecode_626.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.11_date'), channel_map={'no_absent_or_infrequent_menstruation':0, 'absent_or_infrequent_menstruation_prevalent':1, 'absent_or_infrequent_menstruation_incident':2}, loss=weighted_crossentropy([1.0, 4132, 17241], 'absent_or_infrequent_menstruation_pi'))
acid_base_balance_disorder_phe_pi = TensorMap('phecode_276.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.4_date'), channel_map={'no_acid_base_balance_disorder':0, 'acid_base_balance_disorder_prevalent':1, 'acid_base_balance_disorder_incident':2}, loss=weighted_crossentropy([1.0, 166666, 125000], 'acid_base_balance_disorder_pi'))
acidosis_phe_pi = TensorMap('phecode_276.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.41_date'), channel_map={'no_acidosis':0, 'acidosis_prevalent':1, 'acidosis_incident':2}, loss=weighted_crossentropy([1.0, 2617, 284], 'acidosis_pi'))
acne_phe_pi = TensorMap('phecode_706.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_706.1_date'), channel_map={'no_acne':0, 'acne_prevalent':1, 'acne_incident':2}, loss=weighted_crossentropy([1.0, 10869, 8196], 'acne_pi'))
acquired_absence_of_breast_phe_pi = TensorMap('phecode_175',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_175_date'), channel_map={'no_acquired_absence_of_breast':0, 'acquired_absence_of_breast_prevalent':1, 'acquired_absence_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 114, 43], 'acquired_absence_of_breast_pi'))
acquired_acanthosis_nigricans_phe_pi = TensorMap('phecode_701.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701.6_date'), channel_map={'no_acquired_acanthosis_nigricans':0, 'acquired_acanthosis_nigricans_prevalent':1, 'acquired_acanthosis_nigricans_incident':2}, loss=weighted_crossentropy([1.0, 8333, 5434], 'acquired_acanthosis_nigricans_pi'))
acquired_coagulation_factor_deficiency_phe_pi = TensorMap('phecode_286.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.4_date'), channel_map={'no_acquired_coagulation_factor_deficiency':0, 'acquired_coagulation_factor_deficiency_prevalent':1, 'acquired_coagulation_factor_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 50000, 55555], 'acquired_coagulation_factor_deficiency_pi'))
acquired_deformities_of_finger_phe_pi = TensorMap('phecode_736.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_736.2_date'), channel_map={'no_acquired_deformities_of_finger':0, 'acquired_deformities_of_finger_prevalent':1, 'acquired_deformities_of_finger_incident':2}, loss=weighted_crossentropy([1.0, 2777, 3571], 'acquired_deformities_of_finger_pi'))
acquired_foot_deformities_phe_pi = TensorMap('phecode_735',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_735_date'), channel_map={'no_acquired_foot_deformities':0, 'acquired_foot_deformities_prevalent':1, 'acquired_foot_deformities_incident':2}, loss=weighted_crossentropy([1.0, 7692, 3731], 'acquired_foot_deformities_pi'))
acquired_hemolytic_anemias_phe_pi = TensorMap('phecode_283',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_283_date'), channel_map={'no_acquired_hemolytic_anemias':0, 'acquired_hemolytic_anemias_prevalent':1, 'acquired_hemolytic_anemias_incident':2}, loss=weighted_crossentropy([1.0, 23809, 22727], 'acquired_hemolytic_anemias_pi'))
acquired_hypothyroidism_phe_pi = TensorMap('phecode_244.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_244.2_date'), channel_map={'no_acquired_hypothyroidism':0, 'acquired_hypothyroidism_prevalent':1, 'acquired_hypothyroidism_incident':2}, loss=weighted_crossentropy([1.0, 10869, 3846], 'acquired_hypothyroidism_pi'))
acquired_spondylolisthesis_phe_pi = TensorMap('phecode_738.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_738.4_date'), channel_map={'no_acquired_spondylolisthesis':0, 'acquired_spondylolisthesis_prevalent':1, 'acquired_spondylolisthesis_incident':2}, loss=weighted_crossentropy([1.0, 4000, 1824], 'acquired_spondylolisthesis_pi'))
acquired_toe_deformities_phe_pi = TensorMap('phecode_735.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_735.2_date'), channel_map={'no_acquired_toe_deformities':0, 'acquired_toe_deformities_prevalent':1, 'acquired_toe_deformities_incident':2}, loss=weighted_crossentropy([1.0, 407, 293], 'acquired_toe_deformities_pi'))
acromegaly_and_gigantism_phe_pi = TensorMap('phecode_253.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_253.11_date'), channel_map={'no_acromegaly_and_gigantism':0, 'acromegaly_and_gigantism_prevalent':1, 'acromegaly_and_gigantism_incident':2}, loss=weighted_crossentropy([1.0, 5494, 11111], 'acromegaly_and_gigantism_pi'))
actinic_keratosis_phe_pi = TensorMap('phecode_702.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_702.1_date'), channel_map={'no_actinic_keratosis':0, 'actinic_keratosis_prevalent':1, 'actinic_keratosis_incident':2}, loss=weighted_crossentropy([1.0, 411, 189], 'actinic_keratosis_pi'))
acute_and_chronic_tonsillitis_phe_pi = TensorMap('phecode_474',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_474_date'), channel_map={'no_acute_and_chronic_tonsillitis':0, 'acute_and_chronic_tonsillitis_prevalent':1, 'acute_and_chronic_tonsillitis_incident':2}, loss=weighted_crossentropy([1.0, 1373, 2500], 'acute_and_chronic_tonsillitis_pi'))
acute_and_subacute_necrosis_of_liver_phe_pi = TensorMap('phecode_573.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.4_date'), channel_map={'no_acute_and_subacute_necrosis_of_liver':0, 'acute_and_subacute_necrosis_of_liver_prevalent':1, 'acute_and_subacute_necrosis_of_liver_incident':2}, loss=weighted_crossentropy([1.0, 166666, 250000], 'acute_and_subacute_necrosis_of_liver_pi'))
acute_appendicitis_phe_pi = TensorMap('phecode_540.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_540.11_date'), channel_map={'no_acute_appendicitis':0, 'acute_appendicitis_prevalent':1, 'acute_appendicitis_incident':2}, loss=weighted_crossentropy([1.0, 222, 827], 'acute_appendicitis_pi'))
acute_bronchitis_and_bronchiolitis_phe_pi = TensorMap('phecode_483',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_483_date'), channel_map={'no_acute_bronchitis_and_bronchiolitis':0, 'acute_bronchitis_and_bronchiolitis_prevalent':1, 'acute_bronchitis_and_bronchiolitis_incident':2}, loss=weighted_crossentropy([1.0, 4424, 2202], 'acute_bronchitis_and_bronchiolitis_pi'))
acute_cystitis_phe_pi = TensorMap('phecode_592.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_592.11_date'), channel_map={'no_acute_cystitis':0, 'acute_cystitis_prevalent':1, 'acute_cystitis_incident':2}, loss=weighted_crossentropy([1.0, 5000, 5434], 'acute_cystitis_pi'))
acute_dermatitis_due_to_solar_radiation_phe_pi = TensorMap('phecode_938.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_938.1_date'), channel_map={'no_acute_dermatitis_due_to_solar_radiation':0, 'acute_dermatitis_due_to_solar_radiation_prevalent':1, 'acute_dermatitis_due_to_solar_radiation_incident':2}, loss=weighted_crossentropy([1.0, 45454, 33333], 'acute_dermatitis_due_to_solar_radiation_pi'))
acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease_phe_pi = TensorMap('phecode_446.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.2_date'), channel_map={'no_acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease':0, 'acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease_prevalent':1, 'acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease_incident':2}, loss=weighted_crossentropy([1.0, 250000, 166666], 'acute_febrile_mucocutaneous_lymph_node_syndrome_kawasaki_disease_pi'))
acute_gastritis_phe_pi = TensorMap('phecode_535.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_535.1_date'), channel_map={'no_acute_gastritis':0, 'acute_gastritis_prevalent':1, 'acute_gastritis_incident':2}, loss=weighted_crossentropy([1.0, 596, 569], 'acute_gastritis_pi'))
acute_inflammatory_pelvic_disease_phe_pi = TensorMap('phecode_614.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.31_date'), channel_map={'no_acute_inflammatory_pelvic_disease':0, 'acute_inflammatory_pelvic_disease_prevalent':1, 'acute_inflammatory_pelvic_disease_incident':2}, loss=weighted_crossentropy([1.0, 9803, 35714], 'acute_inflammatory_pelvic_disease_pi'))
acute_laryngitis_and_tracheitis_phe_pi = TensorMap('phecode_465.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_465.4_date'), channel_map={'no_acute_laryngitis_and_tracheitis':0, 'acute_laryngitis_and_tracheitis_prevalent':1, 'acute_laryngitis_and_tracheitis_incident':2}, loss=weighted_crossentropy([1.0, 5813, 3816], 'acute_laryngitis_and_tracheitis_pi'))
acute_osteomyelitis_phe_pi = TensorMap('phecode_710.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_710.11_date'), channel_map={'no_acute_osteomyelitis':0, 'acute_osteomyelitis_prevalent':1, 'acute_osteomyelitis_incident':2}, loss=weighted_crossentropy([1.0, 45454, 166666], 'acute_osteomyelitis_pi'))
acute_pain_phe_pi = TensorMap('phecode_338.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_338.1_date'), channel_map={'no_acute_pain':0, 'acute_pain_prevalent':1, 'acute_pain_incident':2}, loss=weighted_crossentropy([1.0, 1805, 1519], 'acute_pain_pi'))
acute_pancreatitis_phe_pi = TensorMap('phecode_577.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_577.1_date'), channel_map={'no_acute_pancreatitis':0, 'acute_pancreatitis_prevalent':1, 'acute_pancreatitis_incident':2}, loss=weighted_crossentropy([1.0, 409, 924], 'acute_pancreatitis_pi'))
acute_pericarditis_phe_pi = TensorMap('phecode_420.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_420.21_date'), channel_map={'no_acute_pericarditis':0, 'acute_pericarditis_prevalent':1, 'acute_pericarditis_incident':2}, loss=weighted_crossentropy([1.0, 36, 16], 'acute_pericarditis_pi'))
acute_periodontitis_phe_pi = TensorMap('phecode_523.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_523.31_date'), channel_map={'no_acute_periodontitis':0, 'acute_periodontitis_prevalent':1, 'acute_periodontitis_incident':2}, loss=weighted_crossentropy([1.0, 4347, 5952], 'acute_periodontitis_pi'))
acute_pharyngitis_phe_pi = TensorMap('phecode_465.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_465.2_date'), channel_map={'no_acute_pharyngitis':0, 'acute_pharyngitis_prevalent':1, 'acute_pharyngitis_incident':2}, loss=weighted_crossentropy([1.0, 1028, 635], 'acute_pharyngitis_pi'))
acute_posthemorrhagic_anemia_phe_pi = TensorMap('phecode_285.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_285.1_date'), channel_map={'no_acute_posthemorrhagic_anemia':0, 'acute_posthemorrhagic_anemia_prevalent':1, 'acute_posthemorrhagic_anemia_incident':2}, loss=weighted_crossentropy([1.0, 2040, 4000], 'acute_posthemorrhagic_anemia_pi'))
acute_prostatitis_phe_pi = TensorMap('phecode_601.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_601.11_date'), channel_map={'no_acute_prostatitis':0, 'acute_prostatitis_prevalent':1, 'acute_prostatitis_incident':2}, loss=weighted_crossentropy([1.0, 5882, 1901], 'acute_prostatitis_pi'))
acute_reaction_to_stress_phe_pi = TensorMap('phecode_300.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.8_date'), channel_map={'no_acute_reaction_to_stress':0, 'acute_reaction_to_stress_prevalent':1, 'acute_reaction_to_stress_incident':2}, loss=weighted_crossentropy([1.0, 4807, 5882], 'acute_reaction_to_stress_pi'))
acute_renal_failure_phe_pi = TensorMap('phecode_585.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_585.1_date'), channel_map={'no_acute_renal_failure':0, 'acute_renal_failure_prevalent':1, 'acute_renal_failure_incident':2}, loss=weighted_crossentropy([1.0, 676, 58], 'acute_renal_failure_pi'))
acute_rheumatic_heart_disease_phe_pi = TensorMap('phecode_394.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_394.4_date'), channel_map={'no_acute_rheumatic_heart_disease':0, 'acute_rheumatic_heart_disease_prevalent':1, 'acute_rheumatic_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 15625, 14705], 'acute_rheumatic_heart_disease_pi'))
acute_sinusitis_phe_pi = TensorMap('phecode_464',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_464_date'), channel_map={'no_acute_sinusitis':0, 'acute_sinusitis_prevalent':1, 'acute_sinusitis_incident':2}, loss=weighted_crossentropy([1.0, 3424, 3937], 'acute_sinusitis_pi'))
acute_tonsillitis_phe_pi = TensorMap('phecode_474.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_474.1_date'), channel_map={'no_acute_tonsillitis':0, 'acute_tonsillitis_prevalent':1, 'acute_tonsillitis_incident':2}, loss=weighted_crossentropy([1.0, 1030, 1666], 'acute_tonsillitis_pi'))
acute_transverse_myelitis_phe_pi = TensorMap('phecode_323.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_323.2_date'), channel_map={'no_acute_transverse_myelitis':0, 'acute_transverse_myelitis_prevalent':1, 'acute_transverse_myelitis_incident':2}, loss=weighted_crossentropy([1.0, 8196, 10638], 'acute_transverse_myelitis_pi'))
acute_upper_respiratory_infections_of_multiple_or_unspecified_sites_phe_pi = TensorMap('phecode_465',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_465_date'), channel_map={'no_acute_upper_respiratory_infections_of_multiple_or_unspecified_sites':0, 'acute_upper_respiratory_infections_of_multiple_or_unspecified_sites_prevalent':1, 'acute_upper_respiratory_infections_of_multiple_or_unspecified_sites_incident':2}, loss=weighted_crossentropy([1.0, 78, 43], 'acute_upper_respiratory_infections_of_multiple_or_unspecified_sites_pi'))
acute_vascular_insufficiency_of_intestine_phe_pi = TensorMap('phecode_441.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_441.1_date'), channel_map={'no_acute_vascular_insufficiency_of_intestine':0, 'acute_vascular_insufficiency_of_intestine_prevalent':1, 'acute_vascular_insufficiency_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 5102, 1618], 'acute_vascular_insufficiency_of_intestine_pi'))
adjustment_reaction_phe_pi = TensorMap('phecode_304',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_304_date'), channel_map={'no_adjustment_reaction':0, 'adjustment_reaction_prevalent':1, 'adjustment_reaction_incident':2}, loss=weighted_crossentropy([1.0, 1953, 3105], 'adjustment_reaction_pi'))
adrenogenital_disorders_phe_pi = TensorMap('phecode_255.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_255.3_date'), channel_map={'no_adrenogenital_disorders':0, 'adrenogenital_disorders_prevalent':1, 'adrenogenital_disorders_incident':2}, loss=weighted_crossentropy([1.0, 31250, 55555], 'adrenogenital_disorders_pi'))
adverse_drug_events_and_drug_allergies_phe_pi = TensorMap('phecode_979',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_979_date'), channel_map={'no_adverse_drug_events_and_drug_allergies':0, 'adverse_drug_events_and_drug_allergies_prevalent':1, 'adverse_drug_events_and_drug_allergies_incident':2}, loss=weighted_crossentropy([1.0, 1336, 2164], 'adverse_drug_events_and_drug_allergies_pi'))
adverse_effects_of_antibacterials_not_penicillins_phe_pi = TensorMap('phecode_960.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_960.1_date'), channel_map={'no_adverse_effects_of_antibacterials_not_penicillins':0, 'adverse_effects_of_antibacterials_not_penicillins_prevalent':1, 'adverse_effects_of_antibacterials_not_penicillins_incident':2}, loss=weighted_crossentropy([1.0, 8928, 11363], 'adverse_effects_of_antibacterials_not_penicillins_pi'))
adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics_phe_pi = TensorMap('phecode_967',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_967_date'), channel_map={'no_adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics':0, 'adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics_prevalent':1, 'adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics_incident':2}, loss=weighted_crossentropy([1.0, 1818, 1138], 'adverse_effects_of_sedatives_or_other_central_nervous_system_depressants_and_anesthetics_pi'))
adverse_reaction_to_serum_or_vaccine_phe_pi = TensorMap('phecode_941',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_941_date'), channel_map={'no_adverse_reaction_to_serum_or_vaccine':0, 'adverse_reaction_to_serum_or_vaccine_prevalent':1, 'adverse_reaction_to_serum_or_vaccine_incident':2}, loss=weighted_crossentropy([1.0, 4629, 2304], 'adverse_reaction_to_serum_or_vaccine_pi'))
agorophobia_social_phobia_and_panic_disorder_phe_pi = TensorMap('phecode_300.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.12_date'), channel_map={'no_agorophobia_social_phobia_and_panic_disorder':0, 'agorophobia_social_phobia_and_panic_disorder_prevalent':1, 'agorophobia_social_phobia_and_panic_disorder_incident':2}, loss=weighted_crossentropy([1.0, 1639, 517], 'agorophobia_social_phobia_and_panic_disorder_pi'))
alcohol_related_disorders_phe_pi = TensorMap('phecode_317',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_317_date'), channel_map={'no_alcohol_related_disorders':0, 'alcohol_related_disorders_prevalent':1, 'alcohol_related_disorders_incident':2}, loss=weighted_crossentropy([1.0, 147, 110], 'alcohol_related_disorders_pi'))
alcoholic_liver_damage_phe_pi = TensorMap('phecode_317.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_317.11_date'), channel_map={'no_alcoholic_liver_damage':0, 'alcoholic_liver_damage_prevalent':1, 'alcoholic_liver_damage_incident':2}, loss=weighted_crossentropy([1.0, 1018, 692], 'alcoholic_liver_damage_pi'))
alcoholism_phe_pi = TensorMap('phecode_317.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_317.1_date'), channel_map={'no_alcoholism':0, 'alcoholism_prevalent':1, 'alcoholism_incident':2}, loss=weighted_crossentropy([1.0, 215, 49], 'alcoholism_pi'))
alkalosis_phe_pi = TensorMap('phecode_276.42',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.42_date'), channel_map={'no_alkalosis':0, 'alkalosis_prevalent':1, 'alkalosis_incident':2}, loss=weighted_crossentropy([1.0, 31250, 2262], 'alkalosis_pi'))
allergic_purpura_phe_pi = TensorMap('phecode_287.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287.2_date'), channel_map={'no_allergic_purpura':0, 'allergic_purpura_prevalent':1, 'allergic_purpura_incident':2}, loss=weighted_crossentropy([1.0, 15151, 16666], 'allergic_purpura_pi'))
allergic_reaction_to_food_phe_pi = TensorMap('phecode_930',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_930_date'), channel_map={'no_allergic_reaction_to_food':0, 'allergic_reaction_to_food_prevalent':1, 'allergic_reaction_to_food_incident':2}, loss=weighted_crossentropy([1.0, 3448, 3787], 'allergic_reaction_to_food_pi'))
allergic_rhinitis_phe_pi = TensorMap('phecode_476',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_476_date'), channel_map={'no_allergic_rhinitis':0, 'allergic_rhinitis_prevalent':1, 'allergic_rhinitis_incident':2}, loss=weighted_crossentropy([1.0, 952, 365], 'allergic_rhinitis_pi'))
allergies_other_phe_pi = TensorMap('phecode_949',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_949_date'), channel_map={'no_allergies_other':0, 'allergies_other_prevalent':1, 'allergies_other_incident':2}, loss=weighted_crossentropy([1.0, 1628, 1639], 'allergies_other_pi'))
allergy_adverse_effect_of_penicillin_phe_pi = TensorMap('phecode_960.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_960.2_date'), channel_map={'no_allergy_adverse_effect_of_penicillin':0, 'allergy_adverse_effect_of_penicillin_prevalent':1, 'allergy_adverse_effect_of_penicillin_incident':2}, loss=weighted_crossentropy([1.0, 57, 33], 'allergy_adverse_effect_of_penicillin_pi'))
alopecia_phe_pi = TensorMap('phecode_704.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_704.1_date'), channel_map={'no_alopecia':0, 'alopecia_prevalent':1, 'alopecia_incident':2}, loss=weighted_crossentropy([1.0, 7692, 3448], 'alopecia_pi'))
alopecia_areata_phe_pi = TensorMap('phecode_704.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_704.11_date'), channel_map={'no_alopecia_areata':0, 'alopecia_areata_prevalent':1, 'alopecia_areata_incident':2}, loss=weighted_crossentropy([1.0, 62500, 25000], 'alopecia_areata_pi'))
alteration_of_consciousness_phe_pi = TensorMap('phecode_291.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_291.8_date'), channel_map={'no_alteration_of_consciousness':0, 'alteration_of_consciousness_prevalent':1, 'alteration_of_consciousness_incident':2}, loss=weighted_crossentropy([1.0, 2604, 863], 'alteration_of_consciousness_pi'))
altered_mental_status_phe_pi = TensorMap('phecode_292.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292.4_date'), channel_map={'no_altered_mental_status':0, 'altered_mental_status_prevalent':1, 'altered_mental_status_incident':2}, loss=weighted_crossentropy([1.0, 866, 154], 'altered_mental_status_pi'))
althetes_foot_phe_pi = TensorMap('phecode_110.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_110.12_date'), channel_map={'no_althetes_foot':0, 'althetes_foot_prevalent':1, 'althetes_foot_incident':2}, loss=weighted_crossentropy([1.0, 4901, 4761], 'althetes_foot_pi'))
alzheimers_disease_phe_pi = TensorMap('phecode_290.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290.11_date'), channel_map={'no_alzheimers_disease':0, 'alzheimers_disease_prevalent':1, 'alzheimers_disease_incident':2}, loss=weighted_crossentropy([1.0, 2702, 279], 'alzheimers_disease_pi'))
amblyopia_phe_pi = TensorMap('phecode_368.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.1_date'), channel_map={'no_amblyopia':0, 'amblyopia_prevalent':1, 'amblyopia_incident':2}, loss=weighted_crossentropy([1.0, 2906, 679], 'amblyopia_pi'))
amyloidosis_phe_pi = TensorMap('phecode_270.33',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.33_date'), channel_map={'no_amyloidosis':0, 'amyloidosis_prevalent':1, 'amyloidosis_incident':2}, loss=weighted_crossentropy([1.0, 12820, 3048], 'amyloidosis_pi'))
anal_and_rectal_conditions_phe_pi = TensorMap('phecode_565',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_565_date'), channel_map={'no_anal_and_rectal_conditions':0, 'anal_and_rectal_conditions_prevalent':1, 'anal_and_rectal_conditions_incident':2}, loss=weighted_crossentropy([1.0, 83, 89], 'anal_and_rectal_conditions_pi'))
anal_and_rectal_polyp_phe_pi = TensorMap('phecode_565.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_565.1_date'), channel_map={'no_anal_and_rectal_polyp':0, 'anal_and_rectal_polyp_prevalent':1, 'anal_and_rectal_polyp_incident':2}, loss=weighted_crossentropy([1.0, 130, 69], 'anal_and_rectal_polyp_pi'))
anaphylactic_shock_nos_phe_pi = TensorMap('phecode_946',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_946_date'), channel_map={'no_anaphylactic_shock_nos':0, 'anaphylactic_shock_nos_prevalent':1, 'anaphylactic_shock_nos_incident':2}, loss=weighted_crossentropy([1.0, 1519, 1059], 'anaphylactic_shock_nos_pi'))
anemia_during_pregnancy_phe_pi = TensorMap('phecode_644',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_644_date'), channel_map={'no_anemia_during_pregnancy':0, 'anemia_during_pregnancy_prevalent':1, 'anemia_during_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 437, 14705], 'anemia_during_pregnancy_pi'))
anemia_in_neoplastic_disease_phe_pi = TensorMap('phecode_285.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_285.22_date'), channel_map={'no_anemia_in_neoplastic_disease':0, 'anemia_in_neoplastic_disease_prevalent':1, 'anemia_in_neoplastic_disease_incident':2}, loss=weighted_crossentropy([1.0, 3333, 994], 'anemia_in_neoplastic_disease_pi'))
anemia_of_chronic_disease_phe_pi = TensorMap('phecode_285.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_285.2_date'), channel_map={'no_anemia_of_chronic_disease':0, 'anemia_of_chronic_disease_prevalent':1, 'anemia_of_chronic_disease_incident':2}, loss=weighted_crossentropy([1.0, 3676, 1483], 'anemia_of_chronic_disease_pi'))
aneurysm_and_dissection_of_heart_phe_pi = TensorMap('phecode_411.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.41_date'), channel_map={'no_aneurysm_and_dissection_of_heart':0, 'aneurysm_and_dissection_of_heart_prevalent':1, 'aneurysm_and_dissection_of_heart_incident':2}, loss=weighted_crossentropy([1.0, 1246, 766], 'aneurysm_and_dissection_of_heart_pi'))
aneurysm_of_artery_of_lower_extremity_phe_pi = TensorMap('phecode_442.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442.3_date'), channel_map={'no_aneurysm_of_artery_of_lower_extremity':0, 'aneurysm_of_artery_of_lower_extremity_prevalent':1, 'aneurysm_of_artery_of_lower_extremity_incident':2}, loss=weighted_crossentropy([1.0, 6172, 2923], 'aneurysm_of_artery_of_lower_extremity_pi'))
aneurysm_of_iliac_artery_phe_pi = TensorMap('phecode_442.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442.2_date'), channel_map={'no_aneurysm_of_iliac_artery':0, 'aneurysm_of_iliac_artery_prevalent':1, 'aneurysm_of_iliac_artery_incident':2}, loss=weighted_crossentropy([1.0, 19230, 3875], 'aneurysm_of_iliac_artery_pi'))
aneurysm_of_other_specified_artery_phe_pi = TensorMap('phecode_442.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442.8_date'), channel_map={'no_aneurysm_of_other_specified_artery':0, 'aneurysm_of_other_specified_artery_prevalent':1, 'aneurysm_of_other_specified_artery_incident':2}, loss=weighted_crossentropy([1.0, 5102, 2192], 'aneurysm_of_other_specified_artery_pi'))
angina_pectoris_phe_pi = TensorMap('phecode_411.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.3_date'), channel_map={'no_angina_pectoris':0, 'angina_pectoris_prevalent':1, 'angina_pectoris_incident':2}, loss=weighted_crossentropy([1.0, 48, 40], 'angina_pectoris_pi'))
anisometropia_phe_pi = TensorMap('phecode_368.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.3_date'), channel_map={'no_anisometropia':0, 'anisometropia_prevalent':1, 'anisometropia_incident':2}, loss=weighted_crossentropy([1.0, 10000, 2487], 'anisometropia_pi'))
ankylosing_spondylitis_phe_pi = TensorMap('phecode_715.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_715.2_date'), channel_map={'no_ankylosing_spondylitis':0, 'ankylosing_spondylitis_prevalent':1, 'ankylosing_spondylitis_incident':2}, loss=weighted_crossentropy([1.0, 2212, 1428], 'ankylosing_spondylitis_pi'))
ankylosis_of_joint_phe_pi = TensorMap('phecode_741.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741.1_date'), channel_map={'no_ankylosis_of_joint':0, 'ankylosis_of_joint_prevalent':1, 'ankylosis_of_joint_incident':2}, loss=weighted_crossentropy([1.0, 41666, 62500], 'ankylosis_of_joint_pi'))
anomalies_of_endocrine_glands_congenital_phe_pi = TensorMap('phecode_759.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_759.1_date'), channel_map={'no_anomalies_of_endocrine_glands_congenital':0, 'anomalies_of_endocrine_glands_congenital_prevalent':1, 'anomalies_of_endocrine_glands_congenital_incident':2}, loss=weighted_crossentropy([1.0, 7692, 8620], 'anomalies_of_endocrine_glands_congenital_pi'))
anomalies_of_jaw_size_symmetry_phe_pi = TensorMap('phecode_526.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_526.3_date'), channel_map={'no_anomalies_of_jaw_size_symmetry':0, 'anomalies_of_jaw_size_symmetry_prevalent':1, 'anomalies_of_jaw_size_symmetry_incident':2}, loss=weighted_crossentropy([1.0, 9803, 16129], 'anomalies_of_jaw_size_symmetry_pi'))
anomalies_of_pupillary_function_phe_pi = TensorMap('phecode_379.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379.4_date'), channel_map={'no_anomalies_of_pupillary_function':0, 'anomalies_of_pupillary_function_prevalent':1, 'anomalies_of_pupillary_function_incident':2}, loss=weighted_crossentropy([1.0, 13513, 5747], 'anomalies_of_pupillary_function_pi'))
anomalies_of_respiratory_system_congenital_phe_pi = TensorMap('phecode_748',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_748_date'), channel_map={'no_anomalies_of_respiratory_system_congenital':0, 'anomalies_of_respiratory_system_congenital_prevalent':1, 'anomalies_of_respiratory_system_congenital_incident':2}, loss=weighted_crossentropy([1.0, 9615, 9433], 'anomalies_of_respiratory_system_congenital_pi'))
anomalies_of_tooth_position_malocclusion_phe_pi = TensorMap('phecode_524.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_524.3_date'), channel_map={'no_anomalies_of_tooth_position_malocclusion':0, 'anomalies_of_tooth_position_malocclusion_prevalent':1, 'anomalies_of_tooth_position_malocclusion_incident':2}, loss=weighted_crossentropy([1.0, 2155, 3086], 'anomalies_of_tooth_position_malocclusion_pi'))
anomalous_atrioventricular_excitation_phe_pi = TensorMap('phecode_426.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.4_date'), channel_map={'no_anomalous_atrioventricular_excitation':0, 'anomalous_atrioventricular_excitation_prevalent':1, 'anomalous_atrioventricular_excitation_incident':2}, loss=weighted_crossentropy([1.0, 3012, 3816], 'anomalous_atrioventricular_excitation_pi'))
anorexia_phe_pi = TensorMap('phecode_260.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_260.6_date'), channel_map={'no_anorexia':0, 'anorexia_prevalent':1, 'anorexia_incident':2}, loss=weighted_crossentropy([1.0, 1572, 503], 'anorexia_pi'))
anorexia_nervosa_phe_pi = TensorMap('phecode_305.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_305.21_date'), channel_map={'no_anorexia_nervosa':0, 'anorexia_nervosa_prevalent':1, 'anorexia_nervosa_incident':2}, loss=weighted_crossentropy([1.0, 15151, 14705], 'anorexia_nervosa_pi'))
antepartum_hemorrhage_abruptio_placentae_and_placenta_previa_phe_pi = TensorMap('phecode_635.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_635.2_date'), channel_map={'no_antepartum_hemorrhage_abruptio_placentae_and_placenta_previa':0, 'antepartum_hemorrhage_abruptio_placentae_and_placenta_previa_prevalent':1, 'antepartum_hemorrhage_abruptio_placentae_and_placenta_previa_incident':2}, loss=weighted_crossentropy([1.0, 442, 10638], 'antepartum_hemorrhage_abruptio_placentae_and_placenta_previa_pi'))
anterior_horn_cell_disease_phe_pi = TensorMap('phecode_334.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_334.2_date'), channel_map={'no_anterior_horn_cell_disease':0, 'anterior_horn_cell_disease_prevalent':1, 'anterior_horn_cell_disease_incident':2}, loss=weighted_crossentropy([1.0, 12195, 1524], 'anterior_horn_cell_disease_pi'))
anticoagulants_causing_adverse_effects_phe_pi = TensorMap('phecode_964.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_964.1_date'), channel_map={'no_anticoagulants_causing_adverse_effects':0, 'anticoagulants_causing_adverse_effects_prevalent':1, 'anticoagulants_causing_adverse_effects_incident':2}, loss=weighted_crossentropy([1.0, 3472, 1811], 'anticoagulants_causing_adverse_effects_pi'))
antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use_phe_pi = TensorMap('phecode_972.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_972.2_date'), channel_map={'no_antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use':0, 'antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use_prevalent':1, 'antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use_incident':2}, loss=weighted_crossentropy([1.0, 29411, 9615], 'antilipemic_and_antiarteriosclerotic_drugs_causing_adverse_effects_in_therapeutic_use_pi'))
antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects_phe_pi = TensorMap('phecode_963.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_963.1_date'), channel_map={'no_antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects':0, 'antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects_prevalent':1, 'antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects_incident':2}, loss=weighted_crossentropy([1.0, 23809, 13513], 'antineoplastic_and_immunosuppressive_drugs_causing_adverse_effects_pi'))
antirheumatics_causing_adverse_effects_in_therapeutic_use_phe_pi = TensorMap('phecode_965.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_965.2_date'), channel_map={'no_antirheumatics_causing_adverse_effects_in_therapeutic_use':0, 'antirheumatics_causing_adverse_effects_in_therapeutic_use_prevalent':1, 'antirheumatics_causing_adverse_effects_in_therapeutic_use_incident':2}, loss=weighted_crossentropy([1.0, 125000, 62500], 'antirheumatics_causing_adverse_effects_in_therapeutic_use_pi'))
antisocial_borderline_personality_disorder_phe_pi = TensorMap('phecode_301.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_301.2_date'), channel_map={'no_antisocial_borderline_personality_disorder':0, 'antisocial_borderline_personality_disorder_prevalent':1, 'antisocial_borderline_personality_disorder_incident':2}, loss=weighted_crossentropy([1.0, 4385, 5263], 'antisocial_borderline_personality_disorder_pi'))
anxiety_disorder_phe_pi = TensorMap('phecode_300.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.1_date'), channel_map={'no_anxiety_disorder':0, 'anxiety_disorder_prevalent':1, 'anxiety_disorder_incident':2}, loss=weighted_crossentropy([1.0, 383, 58], 'anxiety_disorder_pi'))
anxiety_disorders_phe_pi = TensorMap('phecode_300',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300_date'), channel_map={'no_anxiety_disorders':0, 'anxiety_disorders_prevalent':1, 'anxiety_disorders_incident':2}, loss=weighted_crossentropy([1.0, 2717, 1524], 'anxiety_disorders_pi'))
aortic_aneurysm_phe_pi = TensorMap('phecode_442.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442.1_date'), channel_map={'no_aortic_aneurysm':0, 'aortic_aneurysm_prevalent':1, 'aortic_aneurysm_incident':2}, loss=weighted_crossentropy([1.0, 2439, 688], 'aortic_aneurysm_pi'))
aortic_valve_disease_phe_pi = TensorMap('phecode_394.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_394.3_date'), channel_map={'no_aortic_valve_disease':0, 'aortic_valve_disease_prevalent':1, 'aortic_valve_disease_incident':2}, loss=weighted_crossentropy([1.0, 844, 326], 'aortic_valve_disease_pi'))
aphakia_and_other_disorders_of_lens_phe_pi = TensorMap('phecode_379.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379.3_date'), channel_map={'no_aphakia_and_other_disorders_of_lens':0, 'aphakia_and_other_disorders_of_lens_prevalent':1, 'aphakia_and_other_disorders_of_lens_incident':2}, loss=weighted_crossentropy([1.0, 55, 19], 'aphakia_and_other_disorders_of_lens_pi'))
aphasia_speech_disturbance_phe_pi = TensorMap('phecode_292.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292.1_date'), channel_map={'no_aphasia_speech_disturbance':0, 'aphasia_speech_disturbance_prevalent':1, 'aphasia_speech_disturbance_incident':2}, loss=weighted_crossentropy([1.0, 773, 284], 'aphasia_speech_disturbance_pi'))
aplastic_anemia_phe_pi = TensorMap('phecode_284',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_284_date'), channel_map={'no_aplastic_anemia':0, 'aplastic_anemia_prevalent':1, 'aplastic_anemia_incident':2}, loss=weighted_crossentropy([1.0, 75, 39], 'aplastic_anemia_pi'))
apnea_phe_pi = TensorMap('phecode_513.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_513.31_date'), channel_map={'no_apnea':0, 'apnea_prevalent':1, 'apnea_incident':2}, loss=weighted_crossentropy([1.0, 250000, 62500], 'apnea_pi'))
appendiceal_conditions_phe_pi = TensorMap('phecode_540',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_540_date'), channel_map={'no_appendiceal_conditions':0, 'appendiceal_conditions_prevalent':1, 'appendiceal_conditions_incident':2}, loss=weighted_crossentropy([1.0, 180, 255], 'appendiceal_conditions_pi'))
appendicitis_phe_pi = TensorMap('phecode_540.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_540.1_date'), channel_map={'no_appendicitis':0, 'appendicitis_prevalent':1, 'appendicitis_incident':2}, loss=weighted_crossentropy([1.0, 1146, 1672], 'appendicitis_pi'))
arrhythmia_cardiac_nos_phe_pi = TensorMap('phecode_427.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.5_date'), channel_map={'no_arrhythmia_cardiac_nos':0, 'arrhythmia_cardiac_nos_prevalent':1, 'arrhythmia_cardiac_nos_incident':2}, loss=weighted_crossentropy([1.0, 1182, 503], 'arrhythmia_cardiac_nos_pi'))
arterial_dissection_phe_pi = TensorMap('phecode_442.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442.4_date'), channel_map={'no_arterial_dissection':0, 'arterial_dissection_prevalent':1, 'arterial_dissection_incident':2}, loss=weighted_crossentropy([1.0, 10869, 22727], 'arterial_dissection_pi'))
arterial_embolism_and_thrombosis_phe_pi = TensorMap('phecode_444',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_444_date'), channel_map={'no_arterial_embolism_and_thrombosis':0, 'arterial_embolism_and_thrombosis_prevalent':1, 'arterial_embolism_and_thrombosis_incident':2}, loss=weighted_crossentropy([1.0, 2127, 1315], 'arterial_embolism_and_thrombosis_pi'))
arterial_embolism_and_thrombosis_of_lower_extremity_artery_phe_pi = TensorMap('phecode_444.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_444.1_date'), channel_map={'no_arterial_embolism_and_thrombosis_of_lower_extremity_artery':0, 'arterial_embolism_and_thrombosis_of_lower_extremity_artery_prevalent':1, 'arterial_embolism_and_thrombosis_of_lower_extremity_artery_incident':2}, loss=weighted_crossentropy([1.0, 1901, 934], 'arterial_embolism_and_thrombosis_of_lower_extremity_artery_pi'))
arteritis_nos_phe_pi = TensorMap('phecode_446.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.9_date'), channel_map={'no_arteritis_nos':0, 'arteritis_nos_prevalent':1, 'arteritis_nos_incident':2}, loss=weighted_crossentropy([1.0, 3649, 1923], 'arteritis_nos_pi'))
arthropathy_associated_with_infections_phe_pi = TensorMap('phecode_711',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_711_date'), channel_map={'no_arthropathy_associated_with_infections':0, 'arthropathy_associated_with_infections_prevalent':1, 'arthropathy_associated_with_infections_incident':2}, loss=weighted_crossentropy([1.0, 71428, 62500], 'arthropathy_associated_with_infections_pi'))
arthropathy_associated_with_neurological_disorders_phe_pi = TensorMap('phecode_713.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_713.5_date'), channel_map={'no_arthropathy_associated_with_neurological_disorders':0, 'arthropathy_associated_with_neurological_disorders_prevalent':1, 'arthropathy_associated_with_neurological_disorders_incident':2}, loss=weighted_crossentropy([1.0, 11363, 7812], 'arthropathy_associated_with_neurological_disorders_pi'))
arthropathy_associated_with_other_disorders_classified_elsewhere_phe_pi = TensorMap('phecode_713',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_713_date'), channel_map={'no_arthropathy_associated_with_other_disorders_classified_elsewhere':0, 'arthropathy_associated_with_other_disorders_classified_elsewhere_prevalent':1, 'arthropathy_associated_with_other_disorders_classified_elsewhere_incident':2}, loss=weighted_crossentropy([1.0, 13157, 11363], 'arthropathy_associated_with_other_disorders_classified_elsewhere_pi'))
arthropathy_nos_phe_pi = TensorMap('phecode_716.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_716.9_date'), channel_map={'no_arthropathy_nos':0, 'arthropathy_nos_prevalent':1, 'arthropathy_nos_incident':2}, loss=weighted_crossentropy([1.0, 17, 10], 'arthropathy_nos_pi'))
arthropod_borne_diseases_phe_pi = TensorMap('phecode_133',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_133_date'), channel_map={'no_arthropod_borne_diseases':0, 'arthropod_borne_diseases_prevalent':1, 'arthropod_borne_diseases_incident':2}, loss=weighted_crossentropy([1.0, 3676, 6944], 'arthropod_borne_diseases_pi'))
articular_cartilage_disorder_phe_pi = TensorMap('phecode_742.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_742.8_date'), channel_map={'no_articular_cartilage_disorder':0, 'articular_cartilage_disorder_prevalent':1, 'articular_cartilage_disorder_incident':2}, loss=weighted_crossentropy([1.0, 1182, 715], 'articular_cartilage_disorder_pi'))
ascites_non_malignant_phe_pi = TensorMap('phecode_572',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_572_date'), channel_map={'no_ascites_non_malignant':0, 'ascites_non_malignant_prevalent':1, 'ascites_non_malignant_incident':2}, loss=weighted_crossentropy([1.0, 1121, 246], 'ascites_non_malignant_pi'))
ascvd_phe_pi = TensorMap('phecode_414.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_414.2_date'), channel_map={'no_ascvd':0, 'ascvd_prevalent':1, 'ascvd_incident':2}, loss=weighted_crossentropy([1.0, 12, 6], 'ascvd_pi'))
aseptic_necrosis_of_bone_phe_pi = TensorMap('phecode_733.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_733.4_date'), channel_map={'no_aseptic_necrosis_of_bone':0, 'aseptic_necrosis_of_bone_prevalent':1, 'aseptic_necrosis_of_bone_incident':2}, loss=weighted_crossentropy([1.0, 6493, 14705], 'aseptic_necrosis_of_bone_pi'))
aspergillosis_phe_pi = TensorMap('phecode_117.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_117.4_date'), channel_map={'no_aspergillosis':0, 'aspergillosis_prevalent':1, 'aspergillosis_incident':2}, loss=weighted_crossentropy([1.0, 7692, 2688], 'aspergillosis_pi'))
asphyxia_and_hypoxemia_phe_pi = TensorMap('phecode_1013',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1013_date'), channel_map={'no_asphyxia_and_hypoxemia':0, 'asphyxia_and_hypoxemia_prevalent':1, 'asphyxia_and_hypoxemia_incident':2}, loss=weighted_crossentropy([1.0, 8474, 1432], 'asphyxia_and_hypoxemia_pi'))
asthma_phe_pi = TensorMap('phecode_495',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_495_date'), channel_map={'no_asthma':0, 'asthma_prevalent':1, 'asthma_incident':2}, loss=weighted_crossentropy([1.0, 31, 23], 'asthma_pi'))
asthma_with_exacerbation_phe_pi = TensorMap('phecode_495.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_495.2_date'), channel_map={'no_asthma_with_exacerbation':0, 'asthma_with_exacerbation_prevalent':1, 'asthma_with_exacerbation_incident':2}, loss=weighted_crossentropy([1.0, 1858, 4347], 'asthma_with_exacerbation_pi'))
astigmatism_phe_pi = TensorMap('phecode_367.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_367.2_date'), channel_map={'no_astigmatism':0, 'astigmatism_prevalent':1, 'astigmatism_incident':2}, loss=weighted_crossentropy([1.0, 4854, 2049], 'astigmatism_pi'))
atherosclerosis_phe_pi = TensorMap('phecode_440',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_440_date'), channel_map={'no_atherosclerosis':0, 'atherosclerosis_prevalent':1, 'atherosclerosis_incident':2}, loss=weighted_crossentropy([1.0, 4629, 3355], 'atherosclerosis_pi'))
atherosclerosis_of_aorta_phe_pi = TensorMap('phecode_440.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_440.9_date'), channel_map={'no_atherosclerosis_of_aorta':0, 'atherosclerosis_of_aorta_prevalent':1, 'atherosclerosis_of_aorta_incident':2}, loss=weighted_crossentropy([1.0, 4424, 3597], 'atherosclerosis_of_aorta_pi'))
atherosclerosis_of_renal_artery_phe_pi = TensorMap('phecode_440.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_440.1_date'), channel_map={'no_atherosclerosis_of_renal_artery':0, 'atherosclerosis_of_renal_artery_prevalent':1, 'atherosclerosis_of_renal_artery_incident':2}, loss=weighted_crossentropy([1.0, 6756, 13157], 'atherosclerosis_of_renal_artery_pi'))
atherosclerosis_of_the_extremities_phe_pi = TensorMap('phecode_440.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_440.2_date'), channel_map={'no_atherosclerosis_of_the_extremities':0, 'atherosclerosis_of_the_extremities_prevalent':1, 'atherosclerosis_of_the_extremities_incident':2}, loss=weighted_crossentropy([1.0, 1340, 1742], 'atherosclerosis_of_the_extremities_pi'))
atopic_contact_dermatitis_due_to_other_or_unspecified_phe_pi = TensorMap('phecode_939',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_939_date'), channel_map={'no_atopic_contact_dermatitis_due_to_other_or_unspecified':0, 'atopic_contact_dermatitis_due_to_other_or_unspecified_prevalent':1, 'atopic_contact_dermatitis_due_to_other_or_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 471, 217], 'atopic_contact_dermatitis_due_to_other_or_unspecified_pi'))
atrial_fibrillation_and_flutter_phe_pi = TensorMap('phecode_427.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.2_date'), channel_map={'no_atrial_fibrillation_and_flutter':0, 'atrial_fibrillation_and_flutter_prevalent':1, 'atrial_fibrillation_and_flutter_incident':2}, loss=weighted_crossentropy([1.0, 78, 32], 'atrial_fibrillation_and_flutter_pi'))
atrioventricular_av_block_phe_pi = TensorMap('phecode_426.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.2_date'), channel_map={'no_atrioventricular_av_block':0, 'atrioventricular_av_block_prevalent':1, 'atrioventricular_av_block_incident':2}, loss=weighted_crossentropy([1.0, 13513, 3067], 'atrioventricular_av_block_pi'))
atrioventricular_block_complete_phe_pi = TensorMap('phecode_426.24',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.24_date'), channel_map={'no_atrioventricular_block_complete':0, 'atrioventricular_block_complete_prevalent':1, 'atrioventricular_block_complete_incident':2}, loss=weighted_crossentropy([1.0, 2074, 676], 'atrioventricular_block_complete_pi'))
atrophic_gastritis_phe_pi = TensorMap('phecode_535.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_535.2_date'), channel_map={'no_atrophic_gastritis':0, 'atrophic_gastritis_prevalent':1, 'atrophic_gastritis_incident':2}, loss=weighted_crossentropy([1.0, 2380, 1457], 'atrophic_gastritis_pi'))
atrophy_of_edentulous_alveolar_ridge_phe_pi = TensorMap('phecode_525.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_525.2_date'), channel_map={'no_atrophy_of_edentulous_alveolar_ridge':0, 'atrophy_of_edentulous_alveolar_ridge_prevalent':1, 'atrophy_of_edentulous_alveolar_ridge_incident':2}, loss=weighted_crossentropy([1.0, 11627, 11111], 'atrophy_of_edentulous_alveolar_ridge_pi'))
atrophy_of_female_genital_tract_phe_pi = TensorMap('phecode_624.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_624.2_date'), channel_map={'no_atrophy_of_female_genital_tract':0, 'atrophy_of_female_genital_tract_prevalent':1, 'atrophy_of_female_genital_tract_incident':2}, loss=weighted_crossentropy([1.0, 1915, 1275], 'atrophy_of_female_genital_tract_pi'))
autism_phe_pi = TensorMap('phecode_313.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_313.3_date'), channel_map={'no_autism':0, 'autism_prevalent':1, 'autism_incident':2}, loss=weighted_crossentropy([1.0, 41666, 5263], 'autism_pi'))
autoimmune_hemolytic_anemias_phe_pi = TensorMap('phecode_283.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_283.1_date'), channel_map={'no_autoimmune_hemolytic_anemias':0, 'autoimmune_hemolytic_anemias_prevalent':1, 'autoimmune_hemolytic_anemias_incident':2}, loss=weighted_crossentropy([1.0, 8928, 4587], 'autoimmune_hemolytic_anemias_pi'))
back_pain_phe_pi = TensorMap('phecode_760',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_760_date'), channel_map={'no_back_pain':0, 'back_pain_prevalent':1, 'back_pain_incident':2}, loss=weighted_crossentropy([1.0, 127, 98], 'back_pain_pi'))
bacterial_enteritis_phe_pi = TensorMap('phecode_8.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_8.5_date'), channel_map={'no_bacterial_enteritis':0, 'bacterial_enteritis_prevalent':1, 'bacterial_enteritis_incident':2}, loss=weighted_crossentropy([1.0, 308, 369], 'bacterial_enteritis_pi'))
bacterial_infection_nos_phe_pi = TensorMap('phecode_41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_41_date'), channel_map={'no_bacterial_infection_nos':0, 'bacterial_infection_nos_prevalent':1, 'bacterial_infection_nos_incident':2}, loss=weighted_crossentropy([1.0, 83, 49], 'bacterial_infection_nos_pi'))
bacterial_pneumonia_phe_pi = TensorMap('phecode_480.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480.1_date'), channel_map={'no_bacterial_pneumonia':0, 'bacterial_pneumonia_prevalent':1, 'bacterial_pneumonia_incident':2}, loss=weighted_crossentropy([1.0, 1259, 449], 'bacterial_pneumonia_pi'))
balanoposthitis_phe_pi = TensorMap('phecode_601.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_601.4_date'), channel_map={'no_balanoposthitis':0, 'balanoposthitis_prevalent':1, 'balanoposthitis_incident':2}, loss=weighted_crossentropy([1.0, 2890, 2347], 'balanoposthitis_pi'))
bariatric_surgery_phe_pi = TensorMap('phecode_539',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_539_date'), channel_map={'no_bariatric_surgery':0, 'bariatric_surgery_prevalent':1, 'bariatric_surgery_incident':2}, loss=weighted_crossentropy([1.0, 26315, 11904], 'bariatric_surgery_pi'))
behcets_syndrome_phe_pi = TensorMap('phecode_711.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_711.3_date'), channel_map={'no_behcets_syndrome':0, 'behcets_syndrome_prevalent':1, 'behcets_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 21739, 23809], 'behcets_syndrome_pi'))
benign_neoplasm_of_adrenal_gland_phe_pi = TensorMap('phecode_227.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_227.1_date'), channel_map={'no_benign_neoplasm_of_adrenal_gland':0, 'benign_neoplasm_of_adrenal_gland_prevalent':1, 'benign_neoplasm_of_adrenal_gland_incident':2}, loss=weighted_crossentropy([1.0, 6097, 2314], 'benign_neoplasm_of_adrenal_gland_pi'))
benign_neoplasm_of_bone_and_articular_cartilage_phe_pi = TensorMap('phecode_213',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_213_date'), channel_map={'no_benign_neoplasm_of_bone_and_articular_cartilage':0, 'benign_neoplasm_of_bone_and_articular_cartilage_prevalent':1, 'benign_neoplasm_of_bone_and_articular_cartilage_incident':2}, loss=weighted_crossentropy([1.0, 2092, 2824], 'benign_neoplasm_of_bone_and_articular_cartilage_pi'))
benign_neoplasm_of_brain_cranial_nerves_meninges_phe_pi = TensorMap('phecode_225.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_225.1_date'), channel_map={'no_benign_neoplasm_of_brain_cranial_nerves_meninges':0, 'benign_neoplasm_of_brain_cranial_nerves_meninges_prevalent':1, 'benign_neoplasm_of_brain_cranial_nerves_meninges_incident':2}, loss=weighted_crossentropy([1.0, 961, 846], 'benign_neoplasm_of_brain_cranial_nerves_meninges_pi'))
benign_neoplasm_of_breast_phe_pi = TensorMap('phecode_610.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_610.4_date'), channel_map={'no_benign_neoplasm_of_breast':0, 'benign_neoplasm_of_breast_prevalent':1, 'benign_neoplasm_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 381, 843], 'benign_neoplasm_of_breast_pi'))
benign_neoplasm_of_colon_phe_pi = TensorMap('phecode_208',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_208_date'), channel_map={'no_benign_neoplasm_of_colon':0, 'benign_neoplasm_of_colon_prevalent':1, 'benign_neoplasm_of_colon_incident':2}, loss=weighted_crossentropy([1.0, 57, 22], 'benign_neoplasm_of_colon_pi'))
benign_neoplasm_of_eye_phe_pi = TensorMap('phecode_224',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_224_date'), channel_map={'no_benign_neoplasm_of_eye':0, 'benign_neoplasm_of_eye_prevalent':1, 'benign_neoplasm_of_eye_incident':2}, loss=weighted_crossentropy([1.0, 11111, 11363], 'benign_neoplasm_of_eye_pi'))
benign_neoplasm_of_eye_uveal_phe_pi = TensorMap('phecode_224.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_224.1_date'), channel_map={'no_benign_neoplasm_of_eye_uveal':0, 'benign_neoplasm_of_eye_uveal_prevalent':1, 'benign_neoplasm_of_eye_uveal_incident':2}, loss=weighted_crossentropy([1.0, 29411, 7042], 'benign_neoplasm_of_eye_uveal_pi'))
benign_neoplasm_of_kidney_and_other_urinary_organs_phe_pi = TensorMap('phecode_223',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_223_date'), channel_map={'no_benign_neoplasm_of_kidney_and_other_urinary_organs':0, 'benign_neoplasm_of_kidney_and_other_urinary_organs_prevalent':1, 'benign_neoplasm_of_kidney_and_other_urinary_organs_incident':2}, loss=weighted_crossentropy([1.0, 5617, 2777], 'benign_neoplasm_of_kidney_and_other_urinary_organs_pi'))
benign_neoplasm_of_lip_oral_cavity_and_pharynx_phe_pi = TensorMap('phecode_210',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_210_date'), channel_map={'no_benign_neoplasm_of_lip_oral_cavity_and_pharynx':0, 'benign_neoplasm_of_lip_oral_cavity_and_pharynx_prevalent':1, 'benign_neoplasm_of_lip_oral_cavity_and_pharynx_incident':2}, loss=weighted_crossentropy([1.0, 793, 772], 'benign_neoplasm_of_lip_oral_cavity_and_pharynx_pi'))
benign_neoplasm_of_lymph_nodes_phe_pi = TensorMap('phecode_229.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_229.1_date'), channel_map={'no_benign_neoplasm_of_lymph_nodes':0, 'benign_neoplasm_of_lymph_nodes_prevalent':1, 'benign_neoplasm_of_lymph_nodes_incident':2}, loss=weighted_crossentropy([1.0, 16666, 16129], 'benign_neoplasm_of_lymph_nodes_pi'))
benign_neoplasm_of_male_genital_organs_phe_pi = TensorMap('phecode_222',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_222_date'), channel_map={'no_benign_neoplasm_of_male_genital_organs':0, 'benign_neoplasm_of_male_genital_organs_prevalent':1, 'benign_neoplasm_of_male_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 5050, 4504], 'benign_neoplasm_of_male_genital_organs_pi'))
benign_neoplasm_of_other_endocrine_glands_and_related_structures_phe_pi = TensorMap('phecode_227',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_227_date'), channel_map={'no_benign_neoplasm_of_other_endocrine_glands_and_related_structures':0, 'benign_neoplasm_of_other_endocrine_glands_and_related_structures_prevalent':1, 'benign_neoplasm_of_other_endocrine_glands_and_related_structures_incident':2}, loss=weighted_crossentropy([1.0, 50000, 55555], 'benign_neoplasm_of_other_endocrine_glands_and_related_structures_pi'))
benign_neoplasm_of_other_female_genital_organs_phe_pi = TensorMap('phecode_221',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_221_date'), channel_map={'no_benign_neoplasm_of_other_female_genital_organs':0, 'benign_neoplasm_of_other_female_genital_organs_prevalent':1, 'benign_neoplasm_of_other_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 2808, 5263], 'benign_neoplasm_of_other_female_genital_organs_pi'))
benign_neoplasm_of_other_parts_of_digestive_system_phe_pi = TensorMap('phecode_211',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_211_date'), channel_map={'no_benign_neoplasm_of_other_parts_of_digestive_system':0, 'benign_neoplasm_of_other_parts_of_digestive_system_prevalent':1, 'benign_neoplasm_of_other_parts_of_digestive_system_incident':2}, loss=weighted_crossentropy([1.0, 226, 80], 'benign_neoplasm_of_other_parts_of_digestive_system_pi'))
benign_neoplasm_of_ovary_phe_pi = TensorMap('phecode_220',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_220_date'), channel_map={'no_benign_neoplasm_of_ovary':0, 'benign_neoplasm_of_ovary_prevalent':1, 'benign_neoplasm_of_ovary_incident':2}, loss=weighted_crossentropy([1.0, 494, 533], 'benign_neoplasm_of_ovary_pi'))
benign_neoplasm_of_parathyroid_gland_phe_pi = TensorMap('phecode_227.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_227.2_date'), channel_map={'no_benign_neoplasm_of_parathyroid_gland':0, 'benign_neoplasm_of_parathyroid_gland_prevalent':1, 'benign_neoplasm_of_parathyroid_gland_incident':2}, loss=weighted_crossentropy([1.0, 3030, 1253], 'benign_neoplasm_of_parathyroid_gland_pi'))
benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch_phe_pi = TensorMap('phecode_227.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_227.3_date'), channel_map={'no_benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch':0, 'benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch_prevalent':1, 'benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch_incident':2}, loss=weighted_crossentropy([1.0, 2192, 1838], 'benign_neoplasm_of_pituitary_gland_and_craniopharyngeal_duct_pouch_pi'))
benign_neoplasm_of_respiratory_and_intrathoracic_organs_phe_pi = TensorMap('phecode_212',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_212_date'), channel_map={'no_benign_neoplasm_of_respiratory_and_intrathoracic_organs':0, 'benign_neoplasm_of_respiratory_and_intrathoracic_organs_prevalent':1, 'benign_neoplasm_of_respiratory_and_intrathoracic_organs_incident':2}, loss=weighted_crossentropy([1.0, 2016, 1984], 'benign_neoplasm_of_respiratory_and_intrathoracic_organs_pi'))
benign_neoplasm_of_skin_phe_pi = TensorMap('phecode_216',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_216_date'), channel_map={'no_benign_neoplasm_of_skin':0, 'benign_neoplasm_of_skin_prevalent':1, 'benign_neoplasm_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 72, 118], 'benign_neoplasm_of_skin_pi'))
benign_neoplasm_of_spinal_cord_meninges_phe_pi = TensorMap('phecode_225.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_225.2_date'), channel_map={'no_benign_neoplasm_of_spinal_cord_meninges':0, 'benign_neoplasm_of_spinal_cord_meninges_prevalent':1, 'benign_neoplasm_of_spinal_cord_meninges_incident':2}, loss=weighted_crossentropy([1.0, 13513, 13157], 'benign_neoplasm_of_spinal_cord_meninges_pi'))
benign_neoplasm_of_thyroid_glands_phe_pi = TensorMap('phecode_226',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_226_date'), channel_map={'no_benign_neoplasm_of_thyroid_glands':0, 'benign_neoplasm_of_thyroid_glands_prevalent':1, 'benign_neoplasm_of_thyroid_glands_incident':2}, loss=weighted_crossentropy([1.0, 2450, 2941], 'benign_neoplasm_of_thyroid_glands_pi'))
benign_neoplasm_of_unspecified_sites_phe_pi = TensorMap('phecode_229',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_229_date'), channel_map={'no_benign_neoplasm_of_unspecified_sites':0, 'benign_neoplasm_of_unspecified_sites_prevalent':1, 'benign_neoplasm_of_unspecified_sites_incident':2}, loss=weighted_crossentropy([1.0, 14, 17], 'benign_neoplasm_of_unspecified_sites_pi'))
bipolar_phe_pi = TensorMap('phecode_296.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_296.1_date'), channel_map={'no_bipolar':0, 'bipolar_prevalent':1, 'bipolar_incident':2}, loss=weighted_crossentropy([1.0, 686, 573], 'bipolar_pi'))
bladder_neck_obstruction_phe_pi = TensorMap('phecode_596.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_596.1_date'), channel_map={'no_bladder_neck_obstruction':0, 'bladder_neck_obstruction_prevalent':1, 'bladder_neck_obstruction_incident':2}, loss=weighted_crossentropy([1.0, 346, 437], 'bladder_neck_obstruction_pi'))
blindness_and_low_vision_phe_pi = TensorMap('phecode_367.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_367.9_date'), channel_map={'no_blindness_and_low_vision':0, 'blindness_and_low_vision_prevalent':1, 'blindness_and_low_vision_incident':2}, loss=weighted_crossentropy([1.0, 1488, 641], 'blindness_and_low_vision_pi'))
blood_in_stool_phe_pi = TensorMap('phecode_578.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_578.2_date'), channel_map={'no_blood_in_stool':0, 'blood_in_stool_prevalent':1, 'blood_in_stool_incident':2}, loss=weighted_crossentropy([1.0, 327, 219], 'blood_in_stool_pi'))
bone_cancer_phe_pi = TensorMap('phecode_170.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_170.1_date'), channel_map={'no_bone_cancer':0, 'bone_cancer_prevalent':1, 'bone_cancer_incident':2}, loss=weighted_crossentropy([1.0, 7042, 4000], 'bone_cancer_pi'))
bone_marrow_or_stem_cell_transplant_phe_pi = TensorMap('phecode_860',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_860_date'), channel_map={'no_bone_marrow_or_stem_cell_transplant':0, 'bone_marrow_or_stem_cell_transplant_prevalent':1, 'bone_marrow_or_stem_cell_transplant_incident':2}, loss=weighted_crossentropy([1.0, 2688, 1068], 'bone_marrow_or_stem_cell_transplant_pi'))
breast_cancer_phe_pi = TensorMap('phecode_174',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_174_date'), channel_map={'no_breast_cancer':0, 'breast_cancer_prevalent':1, 'breast_cancer_incident':2}, loss=weighted_crossentropy([1.0, 6329, 6944], 'breast_cancer_pi'))
breast_cancer_female_phe_pi = TensorMap('phecode_174.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_174.1_date'), channel_map={'no_breast_cancer_female':0, 'breast_cancer_female_prevalent':1, 'breast_cancer_female_incident':2}, loss=weighted_crossentropy([1.0, 353, 353], 'breast_cancer_female_pi'))
breast_disorder_nos_phe_pi = TensorMap('phecode_613.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_613.9_date'), channel_map={'no_breast_disorder_nos':0, 'breast_disorder_nos_prevalent':1, 'breast_disorder_nos_incident':2}, loss=weighted_crossentropy([1.0, 98, 245], 'breast_disorder_nos_pi'))
broken_tooth_phe_pi = TensorMap('phecode_873',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_873_date'), channel_map={'no_broken_tooth':0, 'broken_tooth_prevalent':1, 'broken_tooth_incident':2}, loss=weighted_crossentropy([1.0, 5882, 10638], 'broken_tooth_pi'))
bronchiectasis_phe_pi = TensorMap('phecode_496.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_496.3_date'), channel_map={'no_bronchiectasis':0, 'bronchiectasis_prevalent':1, 'bronchiectasis_incident':2}, loss=weighted_crossentropy([1.0, 802, 197], 'bronchiectasis_pi'))
bronchitis_phe_pi = TensorMap('phecode_497',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_497_date'), channel_map={'no_bronchitis':0, 'bronchitis_prevalent':1, 'bronchitis_incident':2}, loss=weighted_crossentropy([1.0, 27, 16], 'bronchitis_pi'))
bronchopneumonia_and_lung_abscess_phe_pi = TensorMap('phecode_480.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480.5_date'), channel_map={'no_bronchopneumonia_and_lung_abscess':0, 'bronchopneumonia_and_lung_abscess_prevalent':1, 'bronchopneumonia_and_lung_abscess_incident':2}, loss=weighted_crossentropy([1.0, 1607, 650], 'bronchopneumonia_and_lung_abscess_pi'))
bullous_dermatoses_phe_pi = TensorMap('phecode_695.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.2_date'), channel_map={'no_bullous_dermatoses':0, 'bullous_dermatoses_prevalent':1, 'bullous_dermatoses_incident':2}, loss=weighted_crossentropy([1.0, 4854, 3030], 'bullous_dermatoses_pi'))
bundle_branch_block_phe_pi = TensorMap('phecode_426.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.3_date'), channel_map={'no_bundle_branch_block':0, 'bundle_branch_block_prevalent':1, 'bundle_branch_block_incident':2}, loss=weighted_crossentropy([1.0, 4098, 1381], 'bundle_branch_block_pi'))
burns_phe_pi = TensorMap('phecode_1000',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1000_date'), channel_map={'no_burns':0, 'burns_prevalent':1, 'burns_incident':2}, loss=weighted_crossentropy([1.0, 1262, 1351], 'burns_pi'))
bursitis_phe_pi = TensorMap('phecode_726.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_726.3_date'), channel_map={'no_bursitis':0, 'bursitis_prevalent':1, 'bursitis_incident':2}, loss=weighted_crossentropy([1.0, 1449, 1216], 'bursitis_pi'))
bursitis_disorders_phe_pi = TensorMap('phecode_727.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727.2_date'), channel_map={'no_bursitis_disorders':0, 'bursitis_disorders_prevalent':1, 'bursitis_disorders_incident':2}, loss=weighted_crossentropy([1.0, 29411, 55555], 'bursitis_disorders_pi'))
cachexia_phe_pi = TensorMap('phecode_260.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_260.1_date'), channel_map={'no_cachexia':0, 'cachexia_prevalent':1, 'cachexia_incident':2}, loss=weighted_crossentropy([1.0, 41666, 7246], 'cachexia_pi'))
calcaneal_spur_exostosis_nos_phe_pi = TensorMap('phecode_726.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_726.4_date'), channel_map={'no_calcaneal_spur_exostosis_nos':0, 'calcaneal_spur_exostosis_nos_prevalent':1, 'calcaneal_spur_exostosis_nos_incident':2}, loss=weighted_crossentropy([1.0, 25000, 13888], 'calcaneal_spur_exostosis_nos_pi'))
calculus_of_bile_duct_phe_pi = TensorMap('phecode_574.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_574.2_date'), channel_map={'no_calculus_of_bile_duct':0, 'calculus_of_bile_duct_prevalent':1, 'calculus_of_bile_duct_incident':2}, loss=weighted_crossentropy([1.0, 312, 244], 'calculus_of_bile_duct_pi'))
calculus_of_kidney_phe_pi = TensorMap('phecode_594.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_594.1_date'), channel_map={'no_calculus_of_kidney':0, 'calculus_of_kidney_prevalent':1, 'calculus_of_kidney_incident':2}, loss=weighted_crossentropy([1.0, 238, 208], 'calculus_of_kidney_pi'))
calculus_of_lower_urinary_tract_phe_pi = TensorMap('phecode_594.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_594.2_date'), channel_map={'no_calculus_of_lower_urinary_tract':0, 'calculus_of_lower_urinary_tract_prevalent':1, 'calculus_of_lower_urinary_tract_incident':2}, loss=weighted_crossentropy([1.0, 1216, 711], 'calculus_of_lower_urinary_tract_pi'))
calculus_of_ureter_phe_pi = TensorMap('phecode_594.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_594.3_date'), channel_map={'no_calculus_of_ureter':0, 'calculus_of_ureter_prevalent':1, 'calculus_of_ureter_incident':2}, loss=weighted_crossentropy([1.0, 283, 339], 'calculus_of_ureter_pi'))
cancer_of_bladder_phe_pi = TensorMap('phecode_189.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_189.2_date'), channel_map={'no_cancer_of_bladder':0, 'cancer_of_bladder_prevalent':1, 'cancer_of_bladder_incident':2}, loss=weighted_crossentropy([1.0, 1216, 733], 'cancer_of_bladder_pi'))
cancer_of_brain_phe_pi = TensorMap('phecode_191.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_191.11_date'), channel_map={'no_cancer_of_brain':0, 'cancer_of_brain_prevalent':1, 'cancer_of_brain_incident':2}, loss=weighted_crossentropy([1.0, 4901, 805], 'cancer_of_brain_pi'))
cancer_of_brain_and_nervous_system_phe_pi = TensorMap('phecode_191.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_191.1_date'), channel_map={'no_cancer_of_brain_and_nervous_system':0, 'cancer_of_brain_and_nervous_system_prevalent':1, 'cancer_of_brain_and_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 20000, 10204], 'cancer_of_brain_and_nervous_system_pi'))
cancer_of_bronchus_lung_phe_pi = TensorMap('phecode_165.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_165.1_date'), channel_map={'no_cancer_of_bronchus_lung':0, 'cancer_of_bronchus_lung_prevalent':1, 'cancer_of_bronchus_lung_incident':2}, loss=weighted_crossentropy([1.0, 1483, 163], 'cancer_of_bronchus_lung_pi'))
cancer_of_connective_tissue_phe_pi = TensorMap('phecode_170.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_170.2_date'), channel_map={'no_cancer_of_connective_tissue':0, 'cancer_of_connective_tissue_prevalent':1, 'cancer_of_connective_tissue_incident':2}, loss=weighted_crossentropy([1.0, 2074, 604], 'cancer_of_connective_tissue_pi'))
cancer_of_esophagus_phe_pi = TensorMap('phecode_150',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_150_date'), channel_map={'no_cancer_of_esophagus':0, 'cancer_of_esophagus_prevalent':1, 'cancer_of_esophagus_incident':2}, loss=weighted_crossentropy([1.0, 189, 70], 'cancer_of_esophagus_pi'))
cancer_of_eye_phe_pi = TensorMap('phecode_190',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_190_date'), channel_map={'no_cancer_of_eye':0, 'cancer_of_eye_prevalent':1, 'cancer_of_eye_incident':2}, loss=weighted_crossentropy([1.0, 1984, 650], 'cancer_of_eye_pi'))
cancer_of_hypopharynx_phe_pi = TensorMap('phecode_149.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_149.3_date'), channel_map={'no_cancer_of_hypopharynx':0, 'cancer_of_hypopharynx_prevalent':1, 'cancer_of_hypopharynx_incident':2}, loss=weighted_crossentropy([1.0, 18518, 10204], 'cancer_of_hypopharynx_pi'))
cancer_of_intrathoracic_organs_phe_pi = TensorMap('phecode_164',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_164_date'), channel_map={'no_cancer_of_intrathoracic_organs':0, 'cancer_of_intrathoracic_organs_prevalent':1, 'cancer_of_intrathoracic_organs_incident':2}, loss=weighted_crossentropy([1.0, 15151, 10000], 'cancer_of_intrathoracic_organs_pi'))
cancer_of_larynx_phe_pi = TensorMap('phecode_149.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_149.4_date'), channel_map={'no_cancer_of_larynx':0, 'cancer_of_larynx_prevalent':1, 'cancer_of_larynx_incident':2}, loss=weighted_crossentropy([1.0, 3012, 2824], 'cancer_of_larynx_pi'))
cancer_of_larynx_pharynx_nasal_cavities_phe_pi = TensorMap('phecode_149',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_149_date'), channel_map={'no_cancer_of_larynx_pharynx_nasal_cavities':0, 'cancer_of_larynx_pharynx_nasal_cavities_prevalent':1, 'cancer_of_larynx_pharynx_nasal_cavities_incident':2}, loss=weighted_crossentropy([1.0, 100, 43], 'cancer_of_larynx_pharynx_nasal_cavities_pi'))
cancer_of_lip_phe_pi = TensorMap('phecode_145.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_145.1_date'), channel_map={'no_cancer_of_lip':0, 'cancer_of_lip_prevalent':1, 'cancer_of_lip_incident':2}, loss=weighted_crossentropy([1.0, 16666, 11111], 'cancer_of_lip_pi'))
cancer_of_liver_and_intrahepatic_bile_duct_phe_pi = TensorMap('phecode_155',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_155_date'), channel_map={'no_cancer_of_liver_and_intrahepatic_bile_duct':0, 'cancer_of_liver_and_intrahepatic_bile_duct_prevalent':1, 'cancer_of_liver_and_intrahepatic_bile_duct_incident':2}, loss=weighted_crossentropy([1.0, 15151, 1302], 'cancer_of_liver_and_intrahepatic_bile_duct_pi'))
cancer_of_major_salivary_glands_phe_pi = TensorMap('phecode_145.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_145.3_date'), channel_map={'no_cancer_of_major_salivary_glands':0, 'cancer_of_major_salivary_glands_prevalent':1, 'cancer_of_major_salivary_glands_incident':2}, loss=weighted_crossentropy([1.0, 6756, 6250], 'cancer_of_major_salivary_glands_pi'))
cancer_of_mouth_phe_pi = TensorMap('phecode_145',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_145_date'), channel_map={'no_cancer_of_mouth':0, 'cancer_of_mouth_prevalent':1, 'cancer_of_mouth_incident':2}, loss=weighted_crossentropy([1.0, 6024, 3875], 'cancer_of_mouth_pi'))
cancer_of_nasopharynx_phe_pi = TensorMap('phecode_149.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_149.2_date'), channel_map={'no_cancer_of_nasopharynx':0, 'cancer_of_nasopharynx_prevalent':1, 'cancer_of_nasopharynx_incident':2}, loss=weighted_crossentropy([1.0, 19230, 15625], 'cancer_of_nasopharynx_pi'))
cancer_of_of_nasal_cavities_phe_pi = TensorMap('phecode_149.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_149.9_date'), channel_map={'no_cancer_of_of_nasal_cavities':0, 'cancer_of_of_nasal_cavities_prevalent':1, 'cancer_of_of_nasal_cavities_incident':2}, loss=weighted_crossentropy([1.0, 20000, 9090], 'cancer_of_of_nasal_cavities_pi'))
cancer_of_oropharynx_phe_pi = TensorMap('phecode_149.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_149.1_date'), channel_map={'no_cancer_of_oropharynx':0, 'cancer_of_oropharynx_prevalent':1, 'cancer_of_oropharynx_incident':2}, loss=weighted_crossentropy([1.0, 4032, 2173], 'cancer_of_oropharynx_pi'))
cancer_of_other_endocrine_glands_phe_pi = TensorMap('phecode_194',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_194_date'), channel_map={'no_cancer_of_other_endocrine_glands':0, 'cancer_of_other_endocrine_glands_prevalent':1, 'cancer_of_other_endocrine_glands_incident':2}, loss=weighted_crossentropy([1.0, 9615, 5747], 'cancer_of_other_endocrine_glands_pi'))
cancer_of_other_female_genital_organs_phe_pi = TensorMap('phecode_184',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_184_date'), channel_map={'no_cancer_of_other_female_genital_organs':0, 'cancer_of_other_female_genital_organs_prevalent':1, 'cancer_of_other_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 4065, 6329], 'cancer_of_other_female_genital_organs_pi'))
cancer_of_other_female_genital_organs_excluding_uterus_and_ovary_phe_pi = TensorMap('phecode_184.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_184.2_date'), channel_map={'no_cancer_of_other_female_genital_organs_excluding_uterus_and_ovary':0, 'cancer_of_other_female_genital_organs_excluding_uterus_and_ovary_prevalent':1, 'cancer_of_other_female_genital_organs_excluding_uterus_and_ovary_incident':2}, loss=weighted_crossentropy([1.0, 2212, 2475], 'cancer_of_other_female_genital_organs_excluding_uterus_and_ovary_pi'))
cancer_of_other_lymphoid_histiocytic_tissue_phe_pi = TensorMap('phecode_202',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_202_date'), channel_map={'no_cancer_of_other_lymphoid_histiocytic_tissue':0, 'cancer_of_other_lymphoid_histiocytic_tissue_prevalent':1, 'cancer_of_other_lymphoid_histiocytic_tissue_incident':2}, loss=weighted_crossentropy([1.0, 1084, 395], 'cancer_of_other_lymphoid_histiocytic_tissue_pi'))
cancer_of_other_male_genital_organs_phe_pi = TensorMap('phecode_187',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_187_date'), channel_map={'no_cancer_of_other_male_genital_organs':0, 'cancer_of_other_male_genital_organs_prevalent':1, 'cancer_of_other_male_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 114, 84], 'cancer_of_other_male_genital_organs_pi'))
cancer_of_prostate_phe_pi = TensorMap('phecode_185',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_185_date'), channel_map={'no_cancer_of_prostate':0, 'cancer_of_prostate_prevalent':1, 'cancer_of_prostate_incident':2}, loss=weighted_crossentropy([1.0, 140, 57], 'cancer_of_prostate_pi'))
cancer_of_stomach_phe_pi = TensorMap('phecode_151',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_151_date'), channel_map={'no_cancer_of_stomach':0, 'cancer_of_stomach_prevalent':1, 'cancer_of_stomach_incident':2}, loss=weighted_crossentropy([1.0, 3048, 720], 'cancer_of_stomach_pi'))
cancer_of_the_gums_phe_pi = TensorMap('phecode_145.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_145.4_date'), channel_map={'no_cancer_of_the_gums':0, 'cancer_of_the_gums_prevalent':1, 'cancer_of_the_gums_incident':2}, loss=weighted_crossentropy([1.0, 41666, 13888], 'cancer_of_the_gums_pi'))
cancer_of_the_mouth_floor_phe_pi = TensorMap('phecode_145.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_145.5_date'), channel_map={'no_cancer_of_the_mouth_floor':0, 'cancer_of_the_mouth_floor_prevalent':1, 'cancer_of_the_mouth_floor_incident':2}, loss=weighted_crossentropy([1.0, 12820, 12195], 'cancer_of_the_mouth_floor_pi'))
cancer_of_tongue_phe_pi = TensorMap('phecode_145.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_145.2_date'), channel_map={'no_cancer_of_tongue':0, 'cancer_of_tongue_prevalent':1, 'cancer_of_tongue_incident':2}, loss=weighted_crossentropy([1.0, 3105, 1886], 'cancer_of_tongue_pi'))
cancer_of_urinary_organs_incl_kidney_and_bladder_phe_pi = TensorMap('phecode_189',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_189_date'), channel_map={'no_cancer_of_urinary_organs_incl_kidney_and_bladder':0, 'cancer_of_urinary_organs_incl_kidney_and_bladder_prevalent':1, 'cancer_of_urinary_organs_incl_kidney_and_bladder_incident':2}, loss=weighted_crossentropy([1.0, 326, 194], 'cancer_of_urinary_organs_incl_kidney_and_bladder_pi'))
cancer_suspected_or_other_phe_pi = TensorMap('phecode_195',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_195_date'), channel_map={'no_cancer_suspected_or_other':0, 'cancer_suspected_or_other_prevalent':1, 'cancer_suspected_or_other_incident':2}, loss=weighted_crossentropy([1.0, 823, 866], 'cancer_suspected_or_other_pi'))
cancer_within_the_respiratory_system_phe_pi = TensorMap('phecode_165',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_165_date'), channel_map={'no_cancer_within_the_respiratory_system':0, 'cancer_within_the_respiratory_system_prevalent':1, 'cancer_within_the_respiratory_system_incident':2}, loss=weighted_crossentropy([1.0, 751, 130], 'cancer_within_the_respiratory_system_pi'))
candidiasis_phe_pi = TensorMap('phecode_112',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_112_date'), channel_map={'no_candidiasis':0, 'candidiasis_prevalent':1, 'candidiasis_incident':2}, loss=weighted_crossentropy([1.0, 699, 175], 'candidiasis_pi'))
candidiasis_of_skin_and_nails_phe_pi = TensorMap('phecode_112.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_112.3_date'), channel_map={'no_candidiasis_of_skin_and_nails':0, 'candidiasis_of_skin_and_nails_prevalent':1, 'candidiasis_of_skin_and_nails_incident':2}, loss=weighted_crossentropy([1.0, 35714, 10204], 'candidiasis_of_skin_and_nails_pi'))
carbuncle_and_furuncle_phe_pi = TensorMap('phecode_686.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_686.1_date'), channel_map={'no_carbuncle_and_furuncle':0, 'carbuncle_and_furuncle_prevalent':1, 'carbuncle_and_furuncle_incident':2}, loss=weighted_crossentropy([1.0, 284, 300], 'carbuncle_and_furuncle_pi'))
carcinoid_syndrome_phe_pi = TensorMap('phecode_259.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_259.2_date'), channel_map={'no_carcinoid_syndrome':0, 'carcinoid_syndrome_prevalent':1, 'carcinoid_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 35714, 10638], 'carcinoid_syndrome_pi'))
carcinoma_in_situ_of_skin_phe_pi = TensorMap('phecode_172.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_172.3_date'), channel_map={'no_carcinoma_in_situ_of_skin':0, 'carcinoma_in_situ_of_skin_prevalent':1, 'carcinoma_in_situ_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 1805, 662], 'carcinoma_in_situ_of_skin_pi'))
cardiac_and_circulatory_congenital_anomalies_phe_pi = TensorMap('phecode_747',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_747_date'), channel_map={'no_cardiac_and_circulatory_congenital_anomalies':0, 'cardiac_and_circulatory_congenital_anomalies_prevalent':1, 'cardiac_and_circulatory_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 699, 489], 'cardiac_and_circulatory_congenital_anomalies_pi'))
cardiac_arrest_phe_pi = TensorMap('phecode_427.42',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.42_date'), channel_map={'no_cardiac_arrest':0, 'cardiac_arrest_prevalent':1, 'cardiac_arrest_incident':2}, loss=weighted_crossentropy([1.0, 2016, 365], 'cardiac_arrest_pi'))
cardiac_arrest_and_ventricular_fibrillation_phe_pi = TensorMap('phecode_427.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.4_date'), channel_map={'no_cardiac_arrest_and_ventricular_fibrillation':0, 'cardiac_arrest_and_ventricular_fibrillation_prevalent':1, 'cardiac_arrest_and_ventricular_fibrillation_incident':2}, loss=weighted_crossentropy([1.0, 3787, 1298], 'cardiac_arrest_and_ventricular_fibrillation_pi'))
cardiac_conduction_disorders_phe_pi = TensorMap('phecode_426',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426_date'), channel_map={'no_cardiac_conduction_disorders':0, 'cardiac_conduction_disorders_prevalent':1, 'cardiac_conduction_disorders_incident':2}, loss=weighted_crossentropy([1.0, 5494, 3246], 'cardiac_conduction_disorders_pi'))
cardiac_congenital_anomalies_phe_pi = TensorMap('phecode_747.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_747.1_date'), channel_map={'no_cardiac_congenital_anomalies':0, 'cardiac_congenital_anomalies_prevalent':1, 'cardiac_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 3937, 2857], 'cardiac_congenital_anomalies_pi'))
cardiac_dysrhythmias_phe_pi = TensorMap('phecode_427',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427_date'), channel_map={'no_cardiac_dysrhythmias':0, 'cardiac_dysrhythmias_prevalent':1, 'cardiac_dysrhythmias_incident':2}, loss=weighted_crossentropy([1.0, 5747, 1824], 'cardiac_dysrhythmias_pi'))
cardiac_pacemaker_device_in_situ_phe_pi = TensorMap('phecode_426.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.9_date'), channel_map={'no_cardiac_pacemaker_device_in_situ':0, 'cardiac_pacemaker_device_in_situ_prevalent':1, 'cardiac_pacemaker_device_in_situ_incident':2}, loss=weighted_crossentropy([1.0, 5555, 2016], 'cardiac_pacemaker_device_in_situ_pi'))
cardiac_pacemaker_in_situ_phe_pi = TensorMap('phecode_426.91',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.91_date'), channel_map={'no_cardiac_pacemaker_in_situ':0, 'cardiac_pacemaker_in_situ_prevalent':1, 'cardiac_pacemaker_in_situ_incident':2}, loss=weighted_crossentropy([1.0, 640, 170], 'cardiac_pacemaker_in_situ_pi'))
cardiac_shunt_heart_septal_defect_phe_pi = TensorMap('phecode_747.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_747.11_date'), channel_map={'no_cardiac_shunt_heart_septal_defect':0, 'cardiac_shunt_heart_septal_defect_prevalent':1, 'cardiac_shunt_heart_septal_defect_incident':2}, loss=weighted_crossentropy([1.0, 1366, 1138], 'cardiac_shunt_heart_septal_defect_pi'))
cardiogenic_shock_phe_pi = TensorMap('phecode_797.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_797.1_date'), channel_map={'no_cardiogenic_shock':0, 'cardiogenic_shock_prevalent':1, 'cardiogenic_shock_incident':2}, loss=weighted_crossentropy([1.0, 41666, 3267], 'cardiogenic_shock_pi'))
cardiomegaly_phe_pi = TensorMap('phecode_416',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_416_date'), channel_map={'no_cardiomegaly':0, 'cardiomegaly_prevalent':1, 'cardiomegaly_incident':2}, loss=weighted_crossentropy([1.0, 591, 140], 'cardiomegaly_pi'))
cardiomyopathy_phe_pi = TensorMap('phecode_425',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_425_date'), channel_map={'no_cardiomyopathy':0, 'cardiomyopathy_prevalent':1, 'cardiomyopathy_incident':2}, loss=weighted_crossentropy([1.0, 100000, 19230], 'cardiomyopathy_pi'))
carditis_phe_pi = TensorMap('phecode_420',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_420_date'), channel_map={'no_carditis':0, 'carditis_prevalent':1, 'carditis_incident':2}, loss=weighted_crossentropy([1.0, 50000, 71428], 'carditis_pi'))
cataplexy_and_narcolepsy_phe_pi = TensorMap('phecode_347',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_347_date'), channel_map={'no_cataplexy_and_narcolepsy':0, 'cataplexy_and_narcolepsy_prevalent':1, 'cataplexy_and_narcolepsy_incident':2}, loss=weighted_crossentropy([1.0, 25000, 9615], 'cataplexy_and_narcolepsy_pi'))
cataract_phe_pi = TensorMap('phecode_366',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_366_date'), channel_map={'no_cataract':0, 'cataract_prevalent':1, 'cataract_incident':2}, loss=weighted_crossentropy([1.0, 71, 24], 'cataract_pi'))
celiac_disease_phe_pi = TensorMap('phecode_557.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_557.1_date'), channel_map={'no_celiac_disease':0, 'celiac_disease_prevalent':1, 'celiac_disease_incident':2}, loss=weighted_crossentropy([1.0, 464, 354], 'celiac_disease_pi'))
cellulitis_and_abscess_of_arm_hand_phe_pi = TensorMap('phecode_681.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681.3_date'), channel_map={'no_cellulitis_and_abscess_of_arm_hand':0, 'cellulitis_and_abscess_of_arm_hand_prevalent':1, 'cellulitis_and_abscess_of_arm_hand_incident':2}, loss=weighted_crossentropy([1.0, 145, 106], 'cellulitis_and_abscess_of_arm_hand_pi'))
cellulitis_and_abscess_of_face_neck_phe_pi = TensorMap('phecode_681.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681.2_date'), channel_map={'no_cellulitis_and_abscess_of_face_neck':0, 'cellulitis_and_abscess_of_face_neck_prevalent':1, 'cellulitis_and_abscess_of_face_neck_incident':2}, loss=weighted_crossentropy([1.0, 1672, 1210], 'cellulitis_and_abscess_of_face_neck_pi'))
cellulitis_and_abscess_of_fingers_toes_phe_pi = TensorMap('phecode_681.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681.1_date'), channel_map={'no_cellulitis_and_abscess_of_fingers_toes':0, 'cellulitis_and_abscess_of_fingers_toes_prevalent':1, 'cellulitis_and_abscess_of_fingers_toes_incident':2}, loss=weighted_crossentropy([1.0, 1506, 996], 'cellulitis_and_abscess_of_fingers_toes_pi'))
cellulitis_and_abscess_of_foot_toe_phe_pi = TensorMap('phecode_681.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681.6_date'), channel_map={'no_cellulitis_and_abscess_of_foot_toe':0, 'cellulitis_and_abscess_of_foot_toe_prevalent':1, 'cellulitis_and_abscess_of_foot_toe_incident':2}, loss=weighted_crossentropy([1.0, 145, 106], 'cellulitis_and_abscess_of_foot_toe_pi'))
cellulitis_and_abscess_of_leg_except_foot_phe_pi = TensorMap('phecode_681.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681.5_date'), channel_map={'no_cellulitis_and_abscess_of_leg_except_foot':0, 'cellulitis_and_abscess_of_leg_except_foot_prevalent':1, 'cellulitis_and_abscess_of_leg_except_foot_incident':2}, loss=weighted_crossentropy([1.0, 145, 106], 'cellulitis_and_abscess_of_leg_except_foot_pi'))
cellulitis_and_abscess_of_oral_soft_tissues_phe_pi = TensorMap('phecode_528.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.3_date'), channel_map={'no_cellulitis_and_abscess_of_oral_soft_tissues':0, 'cellulitis_and_abscess_of_oral_soft_tissues_prevalent':1, 'cellulitis_and_abscess_of_oral_soft_tissues_incident':2}, loss=weighted_crossentropy([1.0, 4201, 3623], 'cellulitis_and_abscess_of_oral_soft_tissues_pi'))
cellulitis_and_abscess_of_trunk_phe_pi = TensorMap('phecode_681.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681.7_date'), channel_map={'no_cellulitis_and_abscess_of_trunk':0, 'cellulitis_and_abscess_of_trunk_prevalent':1, 'cellulitis_and_abscess_of_trunk_incident':2}, loss=weighted_crossentropy([1.0, 2092, 782], 'cellulitis_and_abscess_of_trunk_pi'))
central_origin_vertigo_phe_pi = TensorMap('phecode_386.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_386.21_date'), channel_map={'no_central_origin_vertigo':0, 'central_origin_vertigo_prevalent':1, 'central_origin_vertigo_incident':2}, loss=weighted_crossentropy([1.0, 83333, 41666], 'central_origin_vertigo_pi'))
cerebral_aneurysm_phe_pi = TensorMap('phecode_433.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.5_date'), channel_map={'no_cerebral_aneurysm':0, 'cerebral_aneurysm_prevalent':1, 'cerebral_aneurysm_incident':2}, loss=weighted_crossentropy([1.0, 2183, 1577], 'cerebral_aneurysm_pi'))
cerebral_artery_occlusion_with_cerebral_infarction_phe_pi = TensorMap('phecode_433.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.21_date'), channel_map={'no_cerebral_artery_occlusion_with_cerebral_infarction':0, 'cerebral_artery_occlusion_with_cerebral_infarction_prevalent':1, 'cerebral_artery_occlusion_with_cerebral_infarction_incident':2}, loss=weighted_crossentropy([1.0, 530, 278], 'cerebral_artery_occlusion_with_cerebral_infarction_pi'))
cerebral_atherosclerosis_phe_pi = TensorMap('phecode_433.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.12_date'), channel_map={'no_cerebral_atherosclerosis':0, 'cerebral_atherosclerosis_prevalent':1, 'cerebral_atherosclerosis_incident':2}, loss=weighted_crossentropy([1.0, 12820, 1392], 'cerebral_atherosclerosis_pi'))
cerebral_cysts_phe_pi = TensorMap('phecode_348.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_348.4_date'), channel_map={'no_cerebral_cysts':0, 'cerebral_cysts_prevalent':1, 'cerebral_cysts_incident':2}, loss=weighted_crossentropy([1.0, 6250, 3937], 'cerebral_cysts_pi'))
cerebral_degeneration_unspecified_phe_pi = TensorMap('phecode_331.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_331.9_date'), channel_map={'no_cerebral_degeneration_unspecified':0, 'cerebral_degeneration_unspecified_prevalent':1, 'cerebral_degeneration_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 4464, 801], 'cerebral_degeneration_unspecified_pi'))
cerebral_edema_and_compression_of_brain_phe_pi = TensorMap('phecode_348.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_348.2_date'), channel_map={'no_cerebral_edema_and_compression_of_brain':0, 'cerebral_edema_and_compression_of_brain_prevalent':1, 'cerebral_edema_and_compression_of_brain_incident':2}, loss=weighted_crossentropy([1.0, 5102, 1533], 'cerebral_edema_and_compression_of_brain_pi'))
cerebral_ischemia_phe_pi = TensorMap('phecode_433.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.3_date'), channel_map={'no_cerebral_ischemia':0, 'cerebral_ischemia_prevalent':1, 'cerebral_ischemia_incident':2}, loss=weighted_crossentropy([1.0, 2551, 320], 'cerebral_ischemia_pi'))
cerebrovascular_disease_phe_pi = TensorMap('phecode_433',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433_date'), channel_map={'no_cerebrovascular_disease':0, 'cerebrovascular_disease_prevalent':1, 'cerebrovascular_disease_incident':2}, loss=weighted_crossentropy([1.0, 124, 55], 'cerebrovascular_disease_pi'))
certain_early_complications_of_trauma_or_procedure_phe_pi = TensorMap('phecode_958',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_958_date'), channel_map={'no_certain_early_complications_of_trauma_or_procedure':0, 'certain_early_complications_of_trauma_or_procedure_prevalent':1, 'certain_early_complications_of_trauma_or_procedure_incident':2}, loss=weighted_crossentropy([1.0, 1381, 1602], 'certain_early_complications_of_trauma_or_procedure_pi'))
cervical_cancer_phe_pi = TensorMap('phecode_180.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_180.1_date'), channel_map={'no_cervical_cancer':0, 'cervical_cancer_prevalent':1, 'cervical_cancer_incident':2}, loss=weighted_crossentropy([1.0, 271, 101], 'cervical_cancer_pi'))
cervical_incompetence_phe_pi = TensorMap('phecode_636.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_636.8_date'), channel_map={'no_cervical_incompetence':0, 'cervical_incompetence_prevalent':1, 'cervical_incompetence_incident':2}, loss=weighted_crossentropy([1.0, 8064, 125000], 'cervical_incompetence_pi'))
cervical_intraepithelial_neoplasia_cin_cervical_dysplasia_phe_pi = TensorMap('phecode_180.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_180.3_date'), channel_map={'no_cervical_intraepithelial_neoplasia_cin_cervical_dysplasia':0, 'cervical_intraepithelial_neoplasia_cin_cervical_dysplasia_prevalent':1, 'cervical_intraepithelial_neoplasia_cin_cervical_dysplasia_incident':2}, loss=weighted_crossentropy([1.0, 225, 1385], 'cervical_intraepithelial_neoplasia_cin_cervical_dysplasia_pi'))
cervical_radiculitis_phe_pi = TensorMap('phecode_765',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_765_date'), channel_map={'no_cervical_radiculitis':0, 'cervical_radiculitis_prevalent':1, 'cervical_radiculitis_incident':2}, loss=weighted_crossentropy([1.0, 1893, 1046], 'cervical_radiculitis_pi'))
cervicalgia_phe_pi = TensorMap('phecode_761',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_761_date'), channel_map={'no_cervicalgia':0, 'cervicalgia_prevalent':1, 'cervicalgia_incident':2}, loss=weighted_crossentropy([1.0, 813, 578], 'cervicalgia_pi'))
cervicitis_and_endocervicitis_phe_pi = TensorMap('phecode_614.51',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.51_date'), channel_map={'no_cervicitis_and_endocervicitis':0, 'cervicitis_and_endocervicitis_prevalent':1, 'cervicitis_and_endocervicitis_incident':2}, loss=weighted_crossentropy([1.0, 437, 1700], 'cervicitis_and_endocervicitis_pi'))
changes_in_skin_texture_phe_pi = TensorMap('phecode_687.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_687.3_date'), channel_map={'no_changes_in_skin_texture':0, 'changes_in_skin_texture_prevalent':1, 'changes_in_skin_texture_incident':2}, loss=weighted_crossentropy([1.0, 8771, 5154], 'changes_in_skin_texture_pi'))
chemotherapy_phe_pi = TensorMap('phecode_197',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_197_date'), channel_map={'no_chemotherapy':0, 'chemotherapy_prevalent':1, 'chemotherapy_incident':2}, loss=weighted_crossentropy([1.0, 85, 23], 'chemotherapy_pi'))
cholangitis_phe_pi = TensorMap('phecode_575.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_575.1_date'), channel_map={'no_cholangitis':0, 'cholangitis_prevalent':1, 'cholangitis_incident':2}, loss=weighted_crossentropy([1.0, 2906, 765], 'cholangitis_pi'))
cholecystitis_without_cholelithiasis_phe_pi = TensorMap('phecode_574.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_574.3_date'), channel_map={'no_cholecystitis_without_cholelithiasis':0, 'cholecystitis_without_cholelithiasis_prevalent':1, 'cholecystitis_without_cholelithiasis_incident':2}, loss=weighted_crossentropy([1.0, 268, 265], 'cholecystitis_without_cholelithiasis_pi'))
cholelithiasis_phe_pi = TensorMap('phecode_574.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_574.1_date'), channel_map={'no_cholelithiasis':0, 'cholelithiasis_prevalent':1, 'cholelithiasis_incident':2}, loss=weighted_crossentropy([1.0, 84, 80], 'cholelithiasis_pi'))
cholelithiasis_with_acute_cholecystitis_phe_pi = TensorMap('phecode_574.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_574.11_date'), channel_map={'no_cholelithiasis_with_acute_cholecystitis':0, 'cholelithiasis_with_acute_cholecystitis_prevalent':1, 'cholelithiasis_with_acute_cholecystitis_incident':2}, loss=weighted_crossentropy([1.0, 550, 376], 'cholelithiasis_with_acute_cholecystitis_pi'))
cholelithiasis_with_other_cholecystitis_phe_pi = TensorMap('phecode_574.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_574.12_date'), channel_map={'no_cholelithiasis_with_other_cholecystitis':0, 'cholelithiasis_with_other_cholecystitis_prevalent':1, 'cholelithiasis_with_other_cholecystitis_incident':2}, loss=weighted_crossentropy([1.0, 147, 121], 'cholelithiasis_with_other_cholecystitis_pi'))
cholesteatoma_phe_pi = TensorMap('phecode_385.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_385.3_date'), channel_map={'no_cholesteatoma':0, 'cholesteatoma_prevalent':1, 'cholesteatoma_incident':2}, loss=weighted_crossentropy([1.0, 1126, 1811], 'cholesteatoma_pi'))
cholesterolosis_of_gallbladder_phe_pi = TensorMap('phecode_575.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_575.6_date'), channel_map={'no_cholesterolosis_of_gallbladder':0, 'cholesterolosis_of_gallbladder_prevalent':1, 'cholesterolosis_of_gallbladder_incident':2}, loss=weighted_crossentropy([1.0, 1901, 1385], 'cholesterolosis_of_gallbladder_pi'))
chondrocalcinosis_phe_pi = TensorMap('phecode_274.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_274.21_date'), channel_map={'no_chondrocalcinosis':0, 'chondrocalcinosis_prevalent':1, 'chondrocalcinosis_incident':2}, loss=weighted_crossentropy([1.0, 3030, 2136], 'chondrocalcinosis_pi'))
chondromalacia_phe_pi = TensorMap('phecode_733.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_733.9_date'), channel_map={'no_chondromalacia':0, 'chondromalacia_prevalent':1, 'chondromalacia_incident':2}, loss=weighted_crossentropy([1.0, 16666, 100000], 'chondromalacia_pi'))
chorioretinal_inflammations_scars_and_other_disorders_of_choroid_phe_pi = TensorMap('phecode_363',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_363_date'), channel_map={'no_chorioretinal_inflammations_scars_and_other_disorders_of_choroid':0, 'chorioretinal_inflammations_scars_and_other_disorders_of_choroid_prevalent':1, 'chorioretinal_inflammations_scars_and_other_disorders_of_choroid_incident':2}, loss=weighted_crossentropy([1.0, 11363, 5617], 'chorioretinal_inflammations_scars_and_other_disorders_of_choroid_pi'))
chorioretinal_scars_phe_pi = TensorMap('phecode_363.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_363.3_date'), channel_map={'no_chorioretinal_scars':0, 'chorioretinal_scars_prevalent':1, 'chorioretinal_scars_incident':2}, loss=weighted_crossentropy([1.0, 17241, 5102], 'chorioretinal_scars_pi'))
choroidal_degenerations_phe_pi = TensorMap('phecode_363.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_363.4_date'), channel_map={'no_choroidal_degenerations':0, 'choroidal_degenerations_prevalent':1, 'choroidal_degenerations_incident':2}, loss=weighted_crossentropy([1.0, 27777, 20000], 'choroidal_degenerations_pi'))
chromosomal_anomalies_phe_pi = TensorMap('phecode_758.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_758.1_date'), channel_map={'no_chromosomal_anomalies':0, 'chromosomal_anomalies_prevalent':1, 'chromosomal_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 7246, 5813], 'chromosomal_anomalies_pi'))
chromosomal_anomalies_and_genetic_disorders_phe_pi = TensorMap('phecode_758',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_758_date'), channel_map={'no_chromosomal_anomalies_and_genetic_disorders':0, 'chromosomal_anomalies_and_genetic_disorders_prevalent':1, 'chromosomal_anomalies_and_genetic_disorders_incident':2}, loss=weighted_crossentropy([1.0, 7352, 9615], 'chromosomal_anomalies_and_genetic_disorders_pi'))
chronic_airway_obstruction_phe_pi = TensorMap('phecode_496',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_496_date'), channel_map={'no_chronic_airway_obstruction':0, 'chronic_airway_obstruction_prevalent':1, 'chronic_airway_obstruction_incident':2}, loss=weighted_crossentropy([1.0, 219, 50], 'chronic_airway_obstruction_pi'))
chronic_bronchitis_phe_pi = TensorMap('phecode_496.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_496.2_date'), channel_map={'no_chronic_bronchitis':0, 'chronic_bronchitis_prevalent':1, 'chronic_bronchitis_incident':2}, loss=weighted_crossentropy([1.0, 3164, 1618], 'chronic_bronchitis_pi'))
chronic_cystitis_phe_pi = TensorMap('phecode_592.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_592.12_date'), channel_map={'no_chronic_cystitis':0, 'chronic_cystitis_prevalent':1, 'chronic_cystitis_incident':2}, loss=weighted_crossentropy([1.0, 956, 708], 'chronic_cystitis_pi'))
chronic_dermatitis_due_to_solar_radiation_phe_pi = TensorMap('phecode_938.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_938.2_date'), channel_map={'no_chronic_dermatitis_due_to_solar_radiation':0, 'chronic_dermatitis_due_to_solar_radiation_prevalent':1, 'chronic_dermatitis_due_to_solar_radiation_incident':2}, loss=weighted_crossentropy([1.0, 3937, 1976], 'chronic_dermatitis_due_to_solar_radiation_pi'))
chronic_fatigue_syndrome_phe_pi = TensorMap('phecode_798.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_798.1_date'), channel_map={'no_chronic_fatigue_syndrome':0, 'chronic_fatigue_syndrome_prevalent':1, 'chronic_fatigue_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 1865, 834], 'chronic_fatigue_syndrome_pi'))
chronic_glomerulonephritis_nos_phe_pi = TensorMap('phecode_580.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_580.14_date'), channel_map={'no_chronic_glomerulonephritis_nos':0, 'chronic_glomerulonephritis_nos_prevalent':1, 'chronic_glomerulonephritis_nos_incident':2}, loss=weighted_crossentropy([1.0, 3906, 533], 'chronic_glomerulonephritis_nos_pi'))
chronic_hepatitis_phe_pi = TensorMap('phecode_70.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_70.4_date'), channel_map={'no_chronic_hepatitis':0, 'chronic_hepatitis_prevalent':1, 'chronic_hepatitis_incident':2}, loss=weighted_crossentropy([1.0, 2958, 2212], 'chronic_hepatitis_pi'))
chronic_inflammatory_disorders_of_orbit_phe_pi = TensorMap('phecode_371.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_371.9_date'), channel_map={'no_chronic_inflammatory_disorders_of_orbit':0, 'chronic_inflammatory_disorders_of_orbit_prevalent':1, 'chronic_inflammatory_disorders_of_orbit_incident':2}, loss=weighted_crossentropy([1.0, 62500, 45454], 'chronic_inflammatory_disorders_of_orbit_pi'))
chronic_inflammatory_pelvic_disease_phe_pi = TensorMap('phecode_614.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.32_date'), channel_map={'no_chronic_inflammatory_pelvic_disease':0, 'chronic_inflammatory_pelvic_disease_prevalent':1, 'chronic_inflammatory_pelvic_disease_incident':2}, loss=weighted_crossentropy([1.0, 1278, 3086], 'chronic_inflammatory_pelvic_disease_pi'))
chronic_interstitial_cystitis_phe_pi = TensorMap('phecode_592.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_592.13_date'), channel_map={'no_chronic_interstitial_cystitis':0, 'chronic_interstitial_cystitis_prevalent':1, 'chronic_interstitial_cystitis_incident':2}, loss=weighted_crossentropy([1.0, 2590, 2890], 'chronic_interstitial_cystitis_pi'))
chronic_laryngitis_phe_pi = TensorMap('phecode_473.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_473.1_date'), channel_map={'no_chronic_laryngitis':0, 'chronic_laryngitis_prevalent':1, 'chronic_laryngitis_incident':2}, loss=weighted_crossentropy([1.0, 6097, 13888], 'chronic_laryngitis_pi'))
chronic_liver_disease_and_cirrhosis_phe_pi = TensorMap('phecode_571',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_571_date'), channel_map={'no_chronic_liver_disease_and_cirrhosis':0, 'chronic_liver_disease_and_cirrhosis_prevalent':1, 'chronic_liver_disease_and_cirrhosis_incident':2}, loss=weighted_crossentropy([1.0, 2604, 690], 'chronic_liver_disease_and_cirrhosis_pi'))
chronic_lymphocytic_thyroiditis_phe_pi = TensorMap('phecode_245.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_245.21_date'), channel_map={'no_chronic_lymphocytic_thyroiditis':0, 'chronic_lymphocytic_thyroiditis_prevalent':1, 'chronic_lymphocytic_thyroiditis_incident':2}, loss=weighted_crossentropy([1.0, 3424, 3184], 'chronic_lymphocytic_thyroiditis_pi'))
chronic_osteomyelitis_phe_pi = TensorMap('phecode_710.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_710.12_date'), channel_map={'no_chronic_osteomyelitis':0, 'chronic_osteomyelitis_prevalent':1, 'chronic_osteomyelitis_incident':2}, loss=weighted_crossentropy([1.0, 12500, 33333], 'chronic_osteomyelitis_pi'))
chronic_pain_phe_pi = TensorMap('phecode_338.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_338.2_date'), channel_map={'no_chronic_pain':0, 'chronic_pain_prevalent':1, 'chronic_pain_incident':2}, loss=weighted_crossentropy([1.0, 3597, 1043], 'chronic_pain_pi'))
chronic_pancreatitis_phe_pi = TensorMap('phecode_577.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_577.2_date'), channel_map={'no_chronic_pancreatitis':0, 'chronic_pancreatitis_prevalent':1, 'chronic_pancreatitis_incident':2}, loss=weighted_crossentropy([1.0, 1474, 1246], 'chronic_pancreatitis_pi'))
chronic_passive_congestion_of_liver_phe_pi = TensorMap('phecode_573.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.1_date'), channel_map={'no_chronic_passive_congestion_of_liver':0, 'chronic_passive_congestion_of_liver_prevalent':1, 'chronic_passive_congestion_of_liver_incident':2}, loss=weighted_crossentropy([1.0, 166666, 15151], 'chronic_passive_congestion_of_liver_pi'))
chronic_pericarditis_phe_pi = TensorMap('phecode_420.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_420.22_date'), channel_map={'no_chronic_pericarditis':0, 'chronic_pericarditis_prevalent':1, 'chronic_pericarditis_incident':2}, loss=weighted_crossentropy([1.0, 23809, 6172], 'chronic_pericarditis_pi'))
chronic_periodontitis_phe_pi = TensorMap('phecode_523.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_523.32_date'), channel_map={'no_chronic_periodontitis':0, 'chronic_periodontitis_prevalent':1, 'chronic_periodontitis_incident':2}, loss=weighted_crossentropy([1.0, 572, 742], 'chronic_periodontitis_pi'))
chronic_pharyngitis_and_nasopharyngitis_phe_pi = TensorMap('phecode_472',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_472_date'), channel_map={'no_chronic_pharyngitis_and_nasopharyngitis':0, 'chronic_pharyngitis_and_nasopharyngitis_prevalent':1, 'chronic_pharyngitis_and_nasopharyngitis_incident':2}, loss=weighted_crossentropy([1.0, 826, 786], 'chronic_pharyngitis_and_nasopharyngitis_pi'))
chronic_prostatitis_phe_pi = TensorMap('phecode_601.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_601.12_date'), channel_map={'no_chronic_prostatitis':0, 'chronic_prostatitis_prevalent':1, 'chronic_prostatitis_incident':2}, loss=weighted_crossentropy([1.0, 1179, 601], 'chronic_prostatitis_pi'))
chronic_pulmonary_heart_disease_phe_pi = TensorMap('phecode_415.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_415.2_date'), channel_map={'no_chronic_pulmonary_heart_disease':0, 'chronic_pulmonary_heart_disease_prevalent':1, 'chronic_pulmonary_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 9803, 2092], 'chronic_pulmonary_heart_disease_pi'))
chronic_renal_failure_ckd_phe_pi = TensorMap('phecode_585.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_585.3_date'), channel_map={'no_chronic_renal_failure_ckd':0, 'chronic_renal_failure_ckd_prevalent':1, 'chronic_renal_failure_ckd_incident':2}, loss=weighted_crossentropy([1.0, 651, 123], 'chronic_renal_failure_ckd_pi'))
chronic_sinusitis_phe_pi = TensorMap('phecode_475',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_475_date'), channel_map={'no_chronic_sinusitis':0, 'chronic_sinusitis_prevalent':1, 'chronic_sinusitis_incident':2}, loss=weighted_crossentropy([1.0, 266, 331], 'chronic_sinusitis_pi'))
chronic_thyroiditis_phe_pi = TensorMap('phecode_245.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_245.2_date'), channel_map={'no_chronic_thyroiditis':0, 'chronic_thyroiditis_prevalent':1, 'chronic_thyroiditis_incident':2}, loss=weighted_crossentropy([1.0, 33333, 83333], 'chronic_thyroiditis_pi'))
chronic_tonsillitis_and_adenoiditis_phe_pi = TensorMap('phecode_474.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_474.2_date'), channel_map={'no_chronic_tonsillitis_and_adenoiditis':0, 'chronic_tonsillitis_and_adenoiditis_prevalent':1, 'chronic_tonsillitis_and_adenoiditis_incident':2}, loss=weighted_crossentropy([1.0, 548, 1138], 'chronic_tonsillitis_and_adenoiditis_pi'))
chronic_ulcer_of_skin_phe_pi = TensorMap('phecode_707',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_707_date'), channel_map={'no_chronic_ulcer_of_skin':0, 'chronic_ulcer_of_skin_prevalent':1, 'chronic_ulcer_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 924, 340], 'chronic_ulcer_of_skin_pi'))
chronic_vascular_insufficiency_of_intestine_phe_pi = TensorMap('phecode_441.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_441.2_date'), channel_map={'no_chronic_vascular_insufficiency_of_intestine':0, 'chronic_vascular_insufficiency_of_intestine_prevalent':1, 'chronic_vascular_insufficiency_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 25000, 9615], 'chronic_vascular_insufficiency_of_intestine_pi'))
chronic_venous_insufficiency_cvi_phe_pi = TensorMap('phecode_456',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_456_date'), channel_map={'no_chronic_venous_insufficiency_cvi':0, 'chronic_venous_insufficiency_cvi_prevalent':1, 'chronic_venous_insufficiency_cvi_incident':2}, loss=weighted_crossentropy([1.0, 7462, 1766], 'chronic_venous_insufficiency_cvi_pi'))
circadian_rhythm_sleep_disorder_phe_pi = TensorMap('phecode_327.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327.6_date'), channel_map={'no_circadian_rhythm_sleep_disorder':0, 'circadian_rhythm_sleep_disorder_prevalent':1, 'circadian_rhythm_sleep_disorder_incident':2}, loss=weighted_crossentropy([1.0, 20000, 26315], 'circadian_rhythm_sleep_disorder_pi'))
circulatory_disease_nec_phe_pi = TensorMap('phecode_459.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_459.9_date'), channel_map={'no_circulatory_disease_nec':0, 'circulatory_disease_nec_prevalent':1, 'circulatory_disease_nec_incident':2}, loss=weighted_crossentropy([1.0, 6, 5], 'circulatory_disease_nec_pi'))
circumscribed_scleroderma_phe_pi = TensorMap('phecode_701.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701.3_date'), channel_map={'no_circumscribed_scleroderma':0, 'circumscribed_scleroderma_prevalent':1, 'circumscribed_scleroderma_incident':2}, loss=weighted_crossentropy([1.0, 13513, 10204], 'circumscribed_scleroderma_pi'))
cirrhosis_of_liver_without_mention_of_alcohol_phe_pi = TensorMap('phecode_571.51',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_571.51_date'), channel_map={'no_cirrhosis_of_liver_without_mention_of_alcohol':0, 'cirrhosis_of_liver_without_mention_of_alcohol_prevalent':1, 'cirrhosis_of_liver_without_mention_of_alcohol_incident':2}, loss=weighted_crossentropy([1.0, 2100, 594], 'cirrhosis_of_liver_without_mention_of_alcohol_pi'))
cleft_palate_phe_pi = TensorMap('phecode_749.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_749.1_date'), channel_map={'no_cleft_palate':0, 'cleft_palate_prevalent':1, 'cleft_palate_incident':2}, loss=weighted_crossentropy([1.0, 19230, 27777], 'cleft_palate_pi'))
coagulation_defects_complicating_pregnancy_or_postpartum_phe_pi = TensorMap('phecode_286.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.3_date'), channel_map={'no_coagulation_defects_complicating_pregnancy_or_postpartum':0, 'coagulation_defects_complicating_pregnancy_or_postpartum_prevalent':1, 'coagulation_defects_complicating_pregnancy_or_postpartum_incident':2}, loss=weighted_crossentropy([1.0, 6756, 55555], 'coagulation_defects_complicating_pregnancy_or_postpartum_pi'))
colon_cancer_phe_pi = TensorMap('phecode_153.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_153.2_date'), channel_map={'no_colon_cancer':0, 'colon_cancer_prevalent':1, 'colon_cancer_incident':2}, loss=weighted_crossentropy([1.0, 368, 164], 'colon_cancer_pi'))
color_vision_deficiencies_phe_pi = TensorMap('phecode_368.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.5_date'), channel_map={'no_color_vision_deficiencies':0, 'color_vision_deficiencies_prevalent':1, 'color_vision_deficiencies_incident':2}, loss=weighted_crossentropy([1.0, 100000, 71428], 'color_vision_deficiencies_pi'))
colorectal_cancer_phe_pi = TensorMap('phecode_153',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_153_date'), channel_map={'no_colorectal_cancer':0, 'colorectal_cancer_prevalent':1, 'colorectal_cancer_incident':2}, loss=weighted_crossentropy([1.0, 35714, 20000], 'colorectal_cancer_pi'))
coma_phe_pi = TensorMap('phecode_348.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_348.7_date'), channel_map={'no_coma':0, 'coma_prevalent':1, 'coma_incident':2}, loss=weighted_crossentropy([1.0, 2994, 1388], 'coma_pi'))
complex_regional_central_pain_syndrome_phe_pi = TensorMap('phecode_355',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_355_date'), channel_map={'no_complex_regional_central_pain_syndrome':0, 'complex_regional_central_pain_syndrome_prevalent':1, 'complex_regional_central_pain_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 21739, 18518], 'complex_regional_central_pain_syndrome_pi'))
complication_due_to_other_implant_and_internal_device_phe_pi = TensorMap('phecode_859',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_859_date'), channel_map={'no_complication_due_to_other_implant_and_internal_device':0, 'complication_due_to_other_implant_and_internal_device_prevalent':1, 'complication_due_to_other_implant_and_internal_device_incident':2}, loss=weighted_crossentropy([1.0, 332, 162], 'complication_due_to_other_implant_and_internal_device_pi'))
complication_of_amputation_stump_phe_pi = TensorMap('phecode_874',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_874_date'), channel_map={'no_complication_of_amputation_stump':0, 'complication_of_amputation_stump_prevalent':1, 'complication_of_amputation_stump_incident':2}, loss=weighted_crossentropy([1.0, 4464, 3787], 'complication_of_amputation_stump_pi'))
complication_of_colostomy_or_enterostomy_phe_pi = TensorMap('phecode_853',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_853_date'), channel_map={'no_complication_of_colostomy_or_enterostomy':0, 'complication_of_colostomy_or_enterostomy_prevalent':1, 'complication_of_colostomy_or_enterostomy_incident':2}, loss=weighted_crossentropy([1.0, 1886, 1026], 'complication_of_colostomy_or_enterostomy_pi'))
complication_of_internal_orthopedic_device_phe_pi = TensorMap('phecode_858',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_858_date'), channel_map={'no_complication_of_internal_orthopedic_device':0, 'complication_of_internal_orthopedic_device_prevalent':1, 'complication_of_internal_orthopedic_device_incident':2}, loss=weighted_crossentropy([1.0, 272, 151], 'complication_of_internal_orthopedic_device_pi'))
complications_following_abortion_or_ectopic_and_molar_pregnancies_phe_pi = TensorMap('phecode_639',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_639_date'), channel_map={'no_complications_following_abortion_or_ectopic_and_molar_pregnancies':0, 'complications_following_abortion_or_ectopic_and_molar_pregnancies_prevalent':1, 'complications_following_abortion_or_ectopic_and_molar_pregnancies_incident':2}, loss=weighted_crossentropy([1.0, 9259, 62500], 'complications_following_abortion_or_ectopic_and_molar_pregnancies_pi'))
complications_of_cardiac_vascular_device_implant_and_graft_phe_pi = TensorMap('phecode_854',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_854_date'), channel_map={'no_complications_of_cardiac_vascular_device_implant_and_graft':0, 'complications_of_cardiac_vascular_device_implant_and_graft_prevalent':1, 'complications_of_cardiac_vascular_device_implant_and_graft_incident':2}, loss=weighted_crossentropy([1.0, 369, 204], 'complications_of_cardiac_vascular_device_implant_and_graft_pi'))
complications_of_in_utero_procedures_phe_pi = TensorMap('phecode_679',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_679_date'), channel_map={'no_complications_of_in_utero_procedures':0, 'complications_of_in_utero_procedures_prevalent':1, 'complications_of_in_utero_procedures_incident':2}, loss=weighted_crossentropy([1.0, 7352, 166666], 'complications_of_in_utero_procedures_pi'))
complications_of_labor_and_delivery_nec_phe_pi = TensorMap('phecode_669',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_669_date'), channel_map={'no_complications_of_labor_and_delivery_nec':0, 'complications_of_labor_and_delivery_nec_prevalent':1, 'complications_of_labor_and_delivery_nec_incident':2}, loss=weighted_crossentropy([1.0, 40, 1831], 'complications_of_labor_and_delivery_nec_pi'))
complications_of_surgical_and_medical_procedures_phe_pi = TensorMap('phecode_1011',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1011_date'), channel_map={'no_complications_of_surgical_and_medical_procedures':0, 'complications_of_surgical_and_medical_procedures_prevalent':1, 'complications_of_surgical_and_medical_procedures_incident':2}, loss=weighted_crossentropy([1.0, 32, 20], 'complications_of_surgical_and_medical_procedures_pi'))
complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery_phe_pi = TensorMap('phecode_668',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_668_date'), channel_map={'no_complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery':0, 'complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery_prevalent':1, 'complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery_incident':2}, loss=weighted_crossentropy([1.0, 8928, 250000], 'complications_of_the_administration_of_anesthetic_or_other_sedation_in_labor_and_delivery_pi'))
complications_of_transplants_and_reattached_limbs_phe_pi = TensorMap('phecode_851',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_851_date'), channel_map={'no_complications_of_transplants_and_reattached_limbs':0, 'complications_of_transplants_and_reattached_limbs_prevalent':1, 'complications_of_transplants_and_reattached_limbs_incident':2}, loss=weighted_crossentropy([1.0, 45, 29], 'complications_of_transplants_and_reattached_limbs_pi'))
concussion_phe_pi = TensorMap('phecode_817',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_817_date'), channel_map={'no_concussion':0, 'concussion_prevalent':1, 'concussion_incident':2}, loss=weighted_crossentropy([1.0, 7575, 29411], 'concussion_pi'))
conduct_disorders_phe_pi = TensorMap('phecode_312',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_312_date'), channel_map={'no_conduct_disorders':0, 'conduct_disorders_prevalent':1, 'conduct_disorders_incident':2}, loss=weighted_crossentropy([1.0, 55555, 55555], 'conduct_disorders_pi'))
conductive_hearing_loss_phe_pi = TensorMap('phecode_389.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_389.2_date'), channel_map={'no_conductive_hearing_loss':0, 'conductive_hearing_loss_prevalent':1, 'conductive_hearing_loss_incident':2}, loss=weighted_crossentropy([1.0, 1845, 1677], 'conductive_hearing_loss_pi'))
congenital_anomalies_of_abdominal_wall_diaphram_phe_pi = TensorMap('phecode_756.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_756.1_date'), channel_map={'no_congenital_anomalies_of_abdominal_wall_diaphram':0, 'congenital_anomalies_of_abdominal_wall_diaphram_prevalent':1, 'congenital_anomalies_of_abdominal_wall_diaphram_incident':2}, loss=weighted_crossentropy([1.0, 50000, 20833], 'congenital_anomalies_of_abdominal_wall_diaphram_pi'))
congenital_anomalies_of_breast_phe_pi = TensorMap('phecode_612.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_612.3_date'), channel_map={'no_congenital_anomalies_of_breast':0, 'congenital_anomalies_of_breast_prevalent':1, 'congenital_anomalies_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 5952, 16129], 'congenital_anomalies_of_breast_pi'))
congenital_anomalies_of_esophagus_phe_pi = TensorMap('phecode_750.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.14_date'), channel_map={'no_congenital_anomalies_of_esophagus':0, 'congenital_anomalies_of_esophagus_prevalent':1, 'congenital_anomalies_of_esophagus_incident':2}, loss=weighted_crossentropy([1.0, 55555, 8620], 'congenital_anomalies_of_esophagus_pi'))
congenital_anomalies_of_face_and_neck_phe_pi = TensorMap('phecode_749',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_749_date'), channel_map={'no_congenital_anomalies_of_face_and_neck':0, 'congenital_anomalies_of_face_and_neck_prevalent':1, 'congenital_anomalies_of_face_and_neck_incident':2}, loss=weighted_crossentropy([1.0, 1572, 2688], 'congenital_anomalies_of_face_and_neck_pi'))
congenital_anomalies_of_female_genital_organs_phe_pi = TensorMap('phecode_751.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.11_date'), channel_map={'no_congenital_anomalies_of_female_genital_organs':0, 'congenital_anomalies_of_female_genital_organs_prevalent':1, 'congenital_anomalies_of_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 1366, 2283], 'congenital_anomalies_of_female_genital_organs_pi'))
congenital_anomalies_of_genital_organs_phe_pi = TensorMap('phecode_751.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.1_date'), channel_map={'no_congenital_anomalies_of_genital_organs':0, 'congenital_anomalies_of_genital_organs_prevalent':1, 'congenital_anomalies_of_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 910, 1377], 'congenital_anomalies_of_genital_organs_pi'))
congenital_anomalies_of_great_vessels_phe_pi = TensorMap('phecode_747.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_747.13_date'), channel_map={'no_congenital_anomalies_of_great_vessels':0, 'congenital_anomalies_of_great_vessels_prevalent':1, 'congenital_anomalies_of_great_vessels_incident':2}, loss=weighted_crossentropy([1.0, 721, 226], 'congenital_anomalies_of_great_vessels_pi'))
congenital_anomalies_of_intestine_phe_pi = TensorMap('phecode_750.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.21_date'), channel_map={'no_congenital_anomalies_of_intestine':0, 'congenital_anomalies_of_intestine_prevalent':1, 'congenital_anomalies_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 3875, 2873], 'congenital_anomalies_of_intestine_pi'))
congenital_anomalies_of_limbs_phe_pi = TensorMap('phecode_755',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_755_date'), channel_map={'no_congenital_anomalies_of_limbs':0, 'congenital_anomalies_of_limbs_prevalent':1, 'congenital_anomalies_of_limbs_incident':2}, loss=weighted_crossentropy([1.0, 100000, 125000], 'congenital_anomalies_of_limbs_pi'))
congenital_anomalies_of_male_genital_organs_phe_pi = TensorMap('phecode_751.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.12_date'), channel_map={'no_congenital_anomalies_of_male_genital_organs':0, 'congenital_anomalies_of_male_genital_organs_prevalent':1, 'congenital_anomalies_of_male_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 3267, 3875], 'congenital_anomalies_of_male_genital_organs_pi'))
congenital_anomalies_of_mouth_tongue_phe_pi = TensorMap('phecode_750.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.13_date'), channel_map={'no_congenital_anomalies_of_mouth_tongue':0, 'congenital_anomalies_of_mouth_tongue_prevalent':1, 'congenital_anomalies_of_mouth_tongue_incident':2}, loss=weighted_crossentropy([1.0, 4854, 2941], 'congenital_anomalies_of_mouth_tongue_pi'))
congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue_phe_pi = TensorMap('phecode_756.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_756.3_date'), channel_map={'no_congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue':0, 'congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue_prevalent':1, 'congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue_incident':2}, loss=weighted_crossentropy([1.0, 13157, 6410], 'congenital_anomalies_of_muscle_tendon_fascia_and_connective_tissue_pi'))
congenital_anomalies_of_peripheral_vascular_system_phe_pi = TensorMap('phecode_747.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_747.2_date'), channel_map={'no_congenital_anomalies_of_peripheral_vascular_system':0, 'congenital_anomalies_of_peripheral_vascular_system_prevalent':1, 'congenital_anomalies_of_peripheral_vascular_system_incident':2}, loss=weighted_crossentropy([1.0, 4273, 4237], 'congenital_anomalies_of_peripheral_vascular_system_pi'))
congenital_anomalies_of_posterior_segment_of_eye_phe_pi = TensorMap('phecode_753.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_753.2_date'), channel_map={'no_congenital_anomalies_of_posterior_segment_of_eye':0, 'congenital_anomalies_of_posterior_segment_of_eye_prevalent':1, 'congenital_anomalies_of_posterior_segment_of_eye_incident':2}, loss=weighted_crossentropy([1.0, 38461, 45454], 'congenital_anomalies_of_posterior_segment_of_eye_pi'))
congenital_anomalies_of_skin_phe_pi = TensorMap('phecode_691',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_691_date'), channel_map={'no_congenital_anomalies_of_skin':0, 'congenital_anomalies_of_skin_prevalent':1, 'congenital_anomalies_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 12500, 7042], 'congenital_anomalies_of_skin_pi'))
congenital_anomalies_of_skull_and_face_bones_phe_pi = TensorMap('phecode_749.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_749.2_date'), channel_map={'no_congenital_anomalies_of_skull_and_face_bones':0, 'congenital_anomalies_of_skull_and_face_bones_prevalent':1, 'congenital_anomalies_of_skull_and_face_bones_incident':2}, loss=weighted_crossentropy([1.0, 27777, 31250], 'congenital_anomalies_of_skull_and_face_bones_pi'))
congenital_anomalies_of_stomach_phe_pi = TensorMap('phecode_750.15',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.15_date'), channel_map={'no_congenital_anomalies_of_stomach':0, 'congenital_anomalies_of_stomach_prevalent':1, 'congenital_anomalies_of_stomach_incident':2}, loss=weighted_crossentropy([1.0, 55555, 8474], 'congenital_anomalies_of_stomach_pi'))
congenital_anomalies_of_the_eye_phe_pi = TensorMap('phecode_753',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_753_date'), channel_map={'no_congenital_anomalies_of_the_eye':0, 'congenital_anomalies_of_the_eye_prevalent':1, 'congenital_anomalies_of_the_eye_incident':2}, loss=weighted_crossentropy([1.0, 13513, 21739], 'congenital_anomalies_of_the_eye_pi'))
congenital_anomalies_of_the_integument_phe_pi = TensorMap('phecode_757',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_757_date'), channel_map={'no_congenital_anomalies_of_the_integument':0, 'congenital_anomalies_of_the_integument_prevalent':1, 'congenital_anomalies_of_the_integument_incident':2}, loss=weighted_crossentropy([1.0, 33333, 55555], 'congenital_anomalies_of_the_integument_pi'))
congenital_anomalies_of_upper_limb_including_shoulder_girdle_phe_pi = TensorMap('phecode_755.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_755.4_date'), channel_map={'no_congenital_anomalies_of_upper_limb_including_shoulder_girdle':0, 'congenital_anomalies_of_upper_limb_including_shoulder_girdle_prevalent':1, 'congenital_anomalies_of_upper_limb_including_shoulder_girdle_incident':2}, loss=weighted_crossentropy([1.0, 14705, 17241], 'congenital_anomalies_of_upper_limb_including_shoulder_girdle_pi'))
congenital_anomalies_of_urinary_system_phe_pi = TensorMap('phecode_751.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.2_date'), channel_map={'no_congenital_anomalies_of_urinary_system':0, 'congenital_anomalies_of_urinary_system_prevalent':1, 'congenital_anomalies_of_urinary_system_incident':2}, loss=weighted_crossentropy([1.0, 936, 721], 'congenital_anomalies_of_urinary_system_pi'))
congenital_anomaly_of_fingers_toes_phe_pi = TensorMap('phecode_755.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_755.3_date'), channel_map={'no_congenital_anomaly_of_fingers_toes':0, 'congenital_anomaly_of_fingers_toes_prevalent':1, 'congenital_anomaly_of_fingers_toes_incident':2}, loss=weighted_crossentropy([1.0, 20833, 29411], 'congenital_anomaly_of_fingers_toes_pi'))
congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas_phe_pi = TensorMap('phecode_750.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.22_date'), channel_map={'no_congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas':0, 'congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas_prevalent':1, 'congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas_incident':2}, loss=weighted_crossentropy([1.0, 7246, 5050], 'congenital_anomaly_of_gallbladder_bile_ducts_liver_pancreas_pi'))
congenital_cataract_and_lens_anomalies_phe_pi = TensorMap('phecode_753.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_753.1_date'), channel_map={'no_congenital_cataract_and_lens_anomalies':0, 'congenital_cataract_and_lens_anomalies_prevalent':1, 'congenital_cataract_and_lens_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 10000, 10638], 'congenital_cataract_and_lens_anomalies_pi'))
congenital_coagulation_defects_phe_pi = TensorMap('phecode_286.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.1_date'), channel_map={'no_congenital_coagulation_defects':0, 'congenital_coagulation_defects_prevalent':1, 'congenital_coagulation_defects_incident':2}, loss=weighted_crossentropy([1.0, 15625, 23809], 'congenital_coagulation_defects_pi'))
congenital_deficiency_of_other_clotting_factors_including_factor_vii_phe_pi = TensorMap('phecode_286.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.12_date'), channel_map={'no_congenital_deficiency_of_other_clotting_factors_including_factor_vii':0, 'congenital_deficiency_of_other_clotting_factors_including_factor_vii_prevalent':1, 'congenital_deficiency_of_other_clotting_factors_including_factor_vii_incident':2}, loss=weighted_crossentropy([1.0, 2747, 2212], 'congenital_deficiency_of_other_clotting_factors_including_factor_vii_pi'))
congenital_deformities_of_feet_phe_pi = TensorMap('phecode_755.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_755.1_date'), channel_map={'no_congenital_deformities_of_feet':0, 'congenital_deformities_of_feet_prevalent':1, 'congenital_deformities_of_feet_incident':2}, loss=weighted_crossentropy([1.0, 4950, 2564], 'congenital_deformities_of_feet_pi'))
congenital_factor_viii_disorder_phe_pi = TensorMap('phecode_286.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.13_date'), channel_map={'no_congenital_factor_viii_disorder':0, 'congenital_factor_viii_disorder_prevalent':1, 'congenital_factor_viii_disorder_incident':2}, loss=weighted_crossentropy([1.0, 10416, 10416], 'congenital_factor_viii_disorder_pi'))
congenital_hip_dysplasia_and_deformity_phe_pi = TensorMap('phecode_755.61',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_755.61_date'), channel_map={'no_congenital_hip_dysplasia_and_deformity':0, 'congenital_hip_dysplasia_and_deformity_prevalent':1, 'congenital_hip_dysplasia_and_deformity_incident':2}, loss=weighted_crossentropy([1.0, 6578, 6756], 'congenital_hip_dysplasia_and_deformity_pi'))
congenital_hypothyroidism_phe_pi = TensorMap('phecode_244.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_244.5_date'), channel_map={'no_congenital_hypothyroidism':0, 'congenital_hypothyroidism_prevalent':1, 'congenital_hypothyroidism_incident':2}, loss=weighted_crossentropy([1.0, 83333, 13888], 'congenital_hypothyroidism_pi'))
congenital_musculoskeletal_deformities_of_spine_phe_pi = TensorMap('phecode_754',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_754_date'), channel_map={'no_congenital_musculoskeletal_deformities_of_spine':0, 'congenital_musculoskeletal_deformities_of_spine_prevalent':1, 'congenital_musculoskeletal_deformities_of_spine_incident':2}, loss=weighted_crossentropy([1.0, 5617, 8771], 'congenital_musculoskeletal_deformities_of_spine_pi'))
congenital_osteodystrophies_phe_pi = TensorMap('phecode_756.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_756.5_date'), channel_map={'no_congenital_osteodystrophies':0, 'congenital_osteodystrophies_prevalent':1, 'congenital_osteodystrophies_incident':2}, loss=weighted_crossentropy([1.0, 4347, 4761], 'congenital_osteodystrophies_pi'))
congenital_pigmentary_anomalies_of_skin_phe_pi = TensorMap('phecode_691.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_691.3_date'), channel_map={'no_congenital_pigmentary_anomalies_of_skin':0, 'congenital_pigmentary_anomalies_of_skin_prevalent':1, 'congenital_pigmentary_anomalies_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 33333, 33333], 'congenital_pigmentary_anomalies_of_skin_pi'))
congestive_heart_failure_chf_nos_phe_pi = TensorMap('phecode_428.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_428.1_date'), channel_map={'no_congestive_heart_failure_chf_nos':0, 'congestive_heart_failure_chf_nos_prevalent':1, 'congestive_heart_failure_chf_nos_incident':2}, loss=weighted_crossentropy([1.0, 773, 169], 'congestive_heart_failure_chf_nos_pi'))
conjunctivitis_infectious_phe_pi = TensorMap('phecode_369.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_369.5_date'), channel_map={'no_conjunctivitis_infectious':0, 'conjunctivitis_infectious_prevalent':1, 'conjunctivitis_infectious_incident':2}, loss=weighted_crossentropy([1.0, 4504, 1908], 'conjunctivitis_infectious_pi'))
conjunctivitis_noninfectious_phe_pi = TensorMap('phecode_371.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_371.2_date'), channel_map={'no_conjunctivitis_noninfectious':0, 'conjunctivitis_noninfectious_prevalent':1, 'conjunctivitis_noninfectious_incident':2}, loss=weighted_crossentropy([1.0, 23809, 22727], 'conjunctivitis_noninfectious_pi'))
constipation_phe_pi = TensorMap('phecode_563',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_563_date'), channel_map={'no_constipation':0, 'constipation_prevalent':1, 'constipation_incident':2}, loss=weighted_crossentropy([1.0, 104, 41], 'constipation_pi'))
constitutional_aplastic_anemia_phe_pi = TensorMap('phecode_284.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_284.2_date'), channel_map={'no_constitutional_aplastic_anemia':0, 'constitutional_aplastic_anemia_prevalent':1, 'constitutional_aplastic_anemia_incident':2}, loss=weighted_crossentropy([1.0, 62500, 33333], 'constitutional_aplastic_anemia_pi'))
contact_dermatitis_and_other_eczema_due_to_plants_except_food_phe_pi = TensorMap('phecode_931',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_931_date'), channel_map={'no_contact_dermatitis_and_other_eczema_due_to_plants_except_food':0, 'contact_dermatitis_and_other_eczema_due_to_plants_except_food_prevalent':1, 'contact_dermatitis_and_other_eczema_due_to_plants_except_food_incident':2}, loss=weighted_crossentropy([1.0, 125000, 125000], 'contact_dermatitis_and_other_eczema_due_to_plants_except_food_pi'))
contracture_of_joint_phe_pi = TensorMap('phecode_739',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_739_date'), channel_map={'no_contracture_of_joint':0, 'contracture_of_joint_prevalent':1, 'contracture_of_joint_incident':2}, loss=weighted_crossentropy([1.0, 6493, 16129], 'contracture_of_joint_pi'))
contracture_of_palmar_fascia_dupuytrens_disease_phe_pi = TensorMap('phecode_728.71',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_728.71_date'), channel_map={'no_contracture_of_palmar_fascia_dupuytrens_disease':0, 'contracture_of_palmar_fascia_dupuytrens_disease_prevalent':1, 'contracture_of_palmar_fascia_dupuytrens_disease_incident':2}, loss=weighted_crossentropy([1.0, 510, 195], 'contracture_of_palmar_fascia_dupuytrens_disease_pi'))
contracture_of_tendon_sheath_phe_pi = TensorMap('phecode_727.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727.7_date'), channel_map={'no_contracture_of_tendon_sheath':0, 'contracture_of_tendon_sheath_prevalent':1, 'contracture_of_tendon_sheath_incident':2}, loss=weighted_crossentropy([1.0, 7575, 4672], 'contracture_of_tendon_sheath_pi'))
contusion_phe_pi = TensorMap('phecode_916',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_916_date'), channel_map={'no_contusion':0, 'contusion_prevalent':1, 'contusion_incident':2}, loss=weighted_crossentropy([1.0, 701, 361], 'contusion_pi'))
convulsions_phe_pi = TensorMap('phecode_345.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_345.3_date'), channel_map={'no_convulsions':0, 'convulsions_prevalent':1, 'convulsions_incident':2}, loss=weighted_crossentropy([1.0, 394, 246], 'convulsions_pi'))
cornea_replaced_by_transplant_phe_pi = TensorMap('phecode_364.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364.9_date'), channel_map={'no_cornea_replaced_by_transplant':0, 'cornea_replaced_by_transplant_prevalent':1, 'cornea_replaced_by_transplant_incident':2}, loss=weighted_crossentropy([1.0, 9259, 4032], 'cornea_replaced_by_transplant_pi'))
corneal_degenerations_phe_pi = TensorMap('phecode_364.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364.4_date'), channel_map={'no_corneal_degenerations':0, 'corneal_degenerations_prevalent':1, 'corneal_degenerations_incident':2}, loss=weighted_crossentropy([1.0, 11627, 2808], 'corneal_degenerations_pi'))
corneal_dystrophy_phe_pi = TensorMap('phecode_364.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364.5_date'), channel_map={'no_corneal_dystrophy':0, 'corneal_dystrophy_prevalent':1, 'corneal_dystrophy_incident':2}, loss=weighted_crossentropy([1.0, 6250, 1865], 'corneal_dystrophy_pi'))
corneal_edema_phe_pi = TensorMap('phecode_364.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364.2_date'), channel_map={'no_corneal_edema':0, 'corneal_edema_prevalent':1, 'corneal_edema_incident':2}, loss=weighted_crossentropy([1.0, 10204, 5617], 'corneal_edema_pi'))
corneal_opacity_phe_pi = TensorMap('phecode_364.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364.1_date'), channel_map={'no_corneal_opacity':0, 'corneal_opacity_prevalent':1, 'corneal_opacity_incident':2}, loss=weighted_crossentropy([1.0, 10869, 4201], 'corneal_opacity_pi'))
corneal_opacity_and_other_disorders_of_cornea_phe_pi = TensorMap('phecode_364',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364_date'), channel_map={'no_corneal_opacity_and_other_disorders_of_cornea':0, 'corneal_opacity_and_other_disorders_of_cornea_prevalent':1, 'corneal_opacity_and_other_disorders_of_cornea_incident':2}, loss=weighted_crossentropy([1.0, 4854, 1886], 'corneal_opacity_and_other_disorders_of_cornea_pi'))
corneal_ulcer_phe_pi = TensorMap('phecode_370.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_370.1_date'), channel_map={'no_corneal_ulcer':0, 'corneal_ulcer_prevalent':1, 'corneal_ulcer_incident':2}, loss=weighted_crossentropy([1.0, 5102, 4237], 'corneal_ulcer_pi'))
corns_and_callosities_phe_pi = TensorMap('phecode_700',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_700_date'), channel_map={'no_corns_and_callosities':0, 'corns_and_callosities_prevalent':1, 'corns_and_callosities_incident':2}, loss=weighted_crossentropy([1.0, 2840, 2074], 'corns_and_callosities_pi'))
coronary_atherosclerosis_phe_pi = TensorMap('phecode_411.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.4_date'), channel_map={'no_coronary_atherosclerosis':0, 'coronary_atherosclerosis_prevalent':1, 'coronary_atherosclerosis_incident':2}, loss=weighted_crossentropy([1.0, 24, 21], 'coronary_atherosclerosis_pi'))
costochondritis_phe_pi = TensorMap('phecode_733.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_733.6_date'), channel_map={'no_costochondritis':0, 'costochondritis_prevalent':1, 'costochondritis_incident':2}, loss=weighted_crossentropy([1.0, 33333, 2262], 'costochondritis_pi'))
cough_phe_pi = TensorMap('phecode_512.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_512.8_date'), channel_map={'no_cough':0, 'cough_prevalent':1, 'cough_incident':2}, loss=weighted_crossentropy([1.0, 358, 154], 'cough_pi'))
crushing_injury_phe_pi = TensorMap('phecode_1006',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1006_date'), channel_map={'no_crushing_injury':0, 'crushing_injury_prevalent':1, 'crushing_injury_incident':2}, loss=weighted_crossentropy([1.0, 2304, 5434], 'crushing_injury_pi'))
crushing_or_internal_injury_to_organs_phe_pi = TensorMap('phecode_1008',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1008_date'), channel_map={'no_crushing_or_internal_injury_to_organs':0, 'crushing_or_internal_injury_to_organs_prevalent':1, 'crushing_or_internal_injury_to_organs_incident':2}, loss=weighted_crossentropy([1.0, 1644, 2890], 'crushing_or_internal_injury_to_organs_pi'))
crystal_arthropathies_phe_pi = TensorMap('phecode_274.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_274.2_date'), channel_map={'no_crystal_arthropathies':0, 'crystal_arthropathies_prevalent':1, 'crystal_arthropathies_incident':2}, loss=weighted_crossentropy([1.0, 83333, 29411], 'crystal_arthropathies_pi'))
curvature_of_spine_phe_pi = TensorMap('phecode_737',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_737_date'), channel_map={'no_curvature_of_spine':0, 'curvature_of_spine_prevalent':1, 'curvature_of_spine_incident':2}, loss=weighted_crossentropy([1.0, 1644, 489], 'curvature_of_spine_pi'))
cushings_syndrome_phe_pi = TensorMap('phecode_255.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_255.11_date'), channel_map={'no_cushings_syndrome':0, 'cushings_syndrome_prevalent':1, 'cushings_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 6172, 7142], 'cushings_syndrome_pi'))
cutaneous_lupus_erythematosus_phe_pi = TensorMap('phecode_695.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.41_date'), channel_map={'no_cutaneous_lupus_erythematosus':0, 'cutaneous_lupus_erythematosus_prevalent':1, 'cutaneous_lupus_erythematosus_incident':2}, loss=weighted_crossentropy([1.0, 6250, 3289], 'cutaneous_lupus_erythematosus_pi'))
cyst_and_pseudocyst_of_pancreas_phe_pi = TensorMap('phecode_577.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_577.3_date'), channel_map={'no_cyst_and_pseudocyst_of_pancreas':0, 'cyst_and_pseudocyst_of_pancreas_prevalent':1, 'cyst_and_pseudocyst_of_pancreas_incident':2}, loss=weighted_crossentropy([1.0, 3521, 1154], 'cyst_and_pseudocyst_of_pancreas_pi'))
cyst_of_bone_phe_pi = TensorMap('phecode_733.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_733.2_date'), channel_map={'no_cyst_of_bone':0, 'cyst_of_bone_prevalent':1, 'cyst_of_bone_incident':2}, loss=weighted_crossentropy([1.0, 12195, 33333], 'cyst_of_bone_pi'))
cyst_of_kidney_acquired_phe_pi = TensorMap('phecode_586.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_586.2_date'), channel_map={'no_cyst_of_kidney_acquired':0, 'cyst_of_kidney_acquired_prevalent':1, 'cyst_of_kidney_acquired_incident':2}, loss=weighted_crossentropy([1.0, 1012, 302], 'cyst_of_kidney_acquired_pi'))
cyst_of_the_salivary_gland_phe_pi = TensorMap('phecode_528.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.41_date'), channel_map={'no_cyst_of_the_salivary_gland':0, 'cyst_of_the_salivary_gland_prevalent':1, 'cyst_of_the_salivary_gland_incident':2}, loss=weighted_crossentropy([1.0, 6756, 8474], 'cyst_of_the_salivary_gland_pi'))
cyst_or_abscess_of_bartholins_gland_phe_pi = TensorMap('phecode_614.53',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.53_date'), channel_map={'no_cyst_or_abscess_of_bartholins_gland':0, 'cyst_or_abscess_of_bartholins_gland_prevalent':1, 'cyst_or_abscess_of_bartholins_gland_incident':2}, loss=weighted_crossentropy([1.0, 677, 2551], 'cyst_or_abscess_of_bartholins_gland_pi'))
cystic_fibrosis_phe_pi = TensorMap('phecode_499',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_499_date'), channel_map={'no_cystic_fibrosis':0, 'cystic_fibrosis_prevalent':1, 'cystic_fibrosis_incident':2}, loss=weighted_crossentropy([1.0, 38461, 35714], 'cystic_fibrosis_pi'))
cystic_kidney_disease_phe_pi = TensorMap('phecode_751.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.21_date'), channel_map={'no_cystic_kidney_disease':0, 'cystic_kidney_disease_prevalent':1, 'cystic_kidney_disease_incident':2}, loss=weighted_crossentropy([1.0, 1851, 1506], 'cystic_kidney_disease_pi'))
cystic_mastopathy_phe_pi = TensorMap('phecode_610.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_610.1_date'), channel_map={'no_cystic_mastopathy':0, 'cystic_mastopathy_prevalent':1, 'cystic_mastopathy_incident':2}, loss=weighted_crossentropy([1.0, 631, 2183], 'cystic_mastopathy_pi'))
cystitis_phe_pi = TensorMap('phecode_592.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_592.1_date'), channel_map={'no_cystitis':0, 'cystitis_prevalent':1, 'cystitis_incident':2}, loss=weighted_crossentropy([1.0, 457, 303], 'cystitis_pi'))
cysts_of_oral_soft_tissues_phe_pi = TensorMap('phecode_528.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.4_date'), channel_map={'no_cysts_of_oral_soft_tissues':0, 'cysts_of_oral_soft_tissues_prevalent':1, 'cysts_of_oral_soft_tissues_incident':2}, loss=weighted_crossentropy([1.0, 16129, 62500], 'cysts_of_oral_soft_tissues_pi'))
cysts_of_the_jaws_phe_pi = TensorMap('phecode_526.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_526.1_date'), channel_map={'no_cysts_of_the_jaws':0, 'cysts_of_the_jaws_prevalent':1, 'cysts_of_the_jaws_incident':2}, loss=weighted_crossentropy([1.0, 2183, 2427], 'cysts_of_the_jaws_pi'))
decreased_libido_phe_pi = TensorMap('phecode_302.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_302.1_date'), channel_map={'no_decreased_libido':0, 'decreased_libido_prevalent':1, 'decreased_libido_incident':2}, loss=weighted_crossentropy([1.0, 31250, 50000], 'decreased_libido_pi'))
decubitus_ulcer_phe_pi = TensorMap('phecode_707.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_707.1_date'), channel_map={'no_decubitus_ulcer':0, 'decubitus_ulcer_prevalent':1, 'decubitus_ulcer_incident':2}, loss=weighted_crossentropy([1.0, 3875, 1992], 'decubitus_ulcer_pi'))
defibrination_syndrome_phe_pi = TensorMap('phecode_286.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.6_date'), channel_map={'no_defibrination_syndrome':0, 'defibrination_syndrome_prevalent':1, 'defibrination_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 390, 176], 'defibrination_syndrome_pi'))
deficiency_anemias_phe_pi = TensorMap('phecode_281.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_281.9_date'), channel_map={'no_deficiency_anemias':0, 'deficiency_anemias_prevalent':1, 'deficiency_anemias_incident':2}, loss=weighted_crossentropy([1.0, 20000, 8333], 'deficiency_anemias_pi'))
deficiency_of_humoral_immunity_phe_pi = TensorMap('phecode_279.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_279.11_date'), channel_map={'no_deficiency_of_humoral_immunity':0, 'deficiency_of_humoral_immunity_prevalent':1, 'deficiency_of_humoral_immunity_incident':2}, loss=weighted_crossentropy([1.0, 9803, 3012], 'deficiency_of_humoral_immunity_pi'))
degeneration_of_intervertebral_disc_phe_pi = TensorMap('phecode_722.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722.6_date'), channel_map={'no_degeneration_of_intervertebral_disc':0, 'degeneration_of_intervertebral_disc_prevalent':1, 'degeneration_of_intervertebral_disc_incident':2}, loss=weighted_crossentropy([1.0, 367, 168], 'degeneration_of_intervertebral_disc_pi'))
degeneration_of_macula_and_posterior_pole_of_retina_phe_pi = TensorMap('phecode_362.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.2_date'), channel_map={'no_degeneration_of_macula_and_posterior_pole_of_retina':0, 'degeneration_of_macula_and_posterior_pole_of_retina_prevalent':1, 'degeneration_of_macula_and_posterior_pole_of_retina_incident':2}, loss=weighted_crossentropy([1.0, 727, 153], 'degeneration_of_macula_and_posterior_pole_of_retina_pi'))
degenerative_and_vascular_disorders_of_ear_phe_pi = TensorMap('phecode_389.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_389.3_date'), channel_map={'no_degenerative_and_vascular_disorders_of_ear':0, 'degenerative_and_vascular_disorders_of_ear_prevalent':1, 'degenerative_and_vascular_disorders_of_ear_incident':2}, loss=weighted_crossentropy([1.0, 35714, 10000], 'degenerative_and_vascular_disorders_of_ear_pi'))
degenerative_disease_of_the_spinal_cord_phe_pi = TensorMap('phecode_334',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_334_date'), channel_map={'no_degenerative_disease_of_the_spinal_cord':0, 'degenerative_disease_of_the_spinal_cord_prevalent':1, 'degenerative_disease_of_the_spinal_cord_incident':2}, loss=weighted_crossentropy([1.0, 716, 332], 'degenerative_disease_of_the_spinal_cord_pi'))
degenerative_skin_conditions_and_other_dermatoses_phe_pi = TensorMap('phecode_702',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_702_date'), channel_map={'no_degenerative_skin_conditions_and_other_dermatoses':0, 'degenerative_skin_conditions_and_other_dermatoses_prevalent':1, 'degenerative_skin_conditions_and_other_dermatoses_incident':2}, loss=weighted_crossentropy([1.0, 166666, 50000], 'degenerative_skin_conditions_and_other_dermatoses_pi'))
degenerative_skin_disorders_phe_pi = TensorMap('phecode_702.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_702.4_date'), channel_map={'no_degenerative_skin_disorders':0, 'degenerative_skin_disorders_prevalent':1, 'degenerative_skin_disorders_incident':2}, loss=weighted_crossentropy([1.0, 100000, 62500], 'degenerative_skin_disorders_pi'))
delirium_dementia_and_amnestic_and_other_cognitive_disorders_phe_pi = TensorMap('phecode_290',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290_date'), channel_map={'no_delirium_dementia_and_amnestic_and_other_cognitive_disorders':0, 'delirium_dementia_and_amnestic_and_other_cognitive_disorders_prevalent':1, 'delirium_dementia_and_amnestic_and_other_cognitive_disorders_incident':2}, loss=weighted_crossentropy([1.0, 4385, 840], 'delirium_dementia_and_amnestic_and_other_cognitive_disorders_pi'))
delirium_due_to_conditions_classified_elsewhere_phe_pi = TensorMap('phecode_290.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290.2_date'), channel_map={'no_delirium_due_to_conditions_classified_elsewhere':0, 'delirium_due_to_conditions_classified_elsewhere_prevalent':1, 'delirium_due_to_conditions_classified_elsewhere_incident':2}, loss=weighted_crossentropy([1.0, 3496, 341], 'delirium_due_to_conditions_classified_elsewhere_pi'))
dementia_with_cerebral_degenerations_phe_pi = TensorMap('phecode_290.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290.12_date'), channel_map={'no_dementia_with_cerebral_degenerations':0, 'dementia_with_cerebral_degenerations_prevalent':1, 'dementia_with_cerebral_degenerations_incident':2}, loss=weighted_crossentropy([1.0, 71428, 4587], 'dementia_with_cerebral_degenerations_pi'))
dementias_phe_pi = TensorMap('phecode_290.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290.1_date'), channel_map={'no_dementias':0, 'dementias_prevalent':1, 'dementias_incident':2}, loss=weighted_crossentropy([1.0, 9259, 322], 'dementias_pi'))
dental_abrasion_erosion_and_attrition_phe_pi = TensorMap('phecode_521.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_521.2_date'), channel_map={'no_dental_abrasion_erosion_and_attrition':0, 'dental_abrasion_erosion_and_attrition_prevalent':1, 'dental_abrasion_erosion_and_attrition_incident':2}, loss=weighted_crossentropy([1.0, 100000, 50000], 'dental_abrasion_erosion_and_attrition_pi'))
dental_caries_phe_pi = TensorMap('phecode_521.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_521.1_date'), channel_map={'no_dental_caries':0, 'dental_caries_prevalent':1, 'dental_caries_incident':2}, loss=weighted_crossentropy([1.0, 244, 183], 'dental_caries_pi'))
dentofacial_anomalies_including_malocclusion_phe_pi = TensorMap('phecode_524',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_524_date'), channel_map={'no_dentofacial_anomalies_including_malocclusion':0, 'dentofacial_anomalies_including_malocclusion_prevalent':1, 'dentofacial_anomalies_including_malocclusion_incident':2}, loss=weighted_crossentropy([1.0, 14285, 33333], 'dentofacial_anomalies_including_malocclusion_pi'))
dependence_on_respirator_ventilator_or_supplemental_oxygen_phe_pi = TensorMap('phecode_509.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_509.8_date'), channel_map={'no_dependence_on_respirator_ventilator_or_supplemental_oxygen':0, 'dependence_on_respirator_ventilator_or_supplemental_oxygen_prevalent':1, 'dependence_on_respirator_ventilator_or_supplemental_oxygen_incident':2}, loss=weighted_crossentropy([1.0, 1984, 461], 'dependence_on_respirator_ventilator_or_supplemental_oxygen_pi'))
depression_phe_pi = TensorMap('phecode_296.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_296.2_date'), channel_map={'no_depression':0, 'depression_prevalent':1, 'depression_incident':2}, loss=weighted_crossentropy([1.0, 108, 34], 'depression_pi'))
derangement_of_joint_non_traumatic_phe_pi = TensorMap('phecode_742',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_742_date'), channel_map={'no_derangement_of_joint_non_traumatic':0, 'derangement_of_joint_non_traumatic_prevalent':1, 'derangement_of_joint_non_traumatic_incident':2}, loss=weighted_crossentropy([1.0, 33333, 13513], 'derangement_of_joint_non_traumatic_pi'))
dermatitis_due_to_solar_radiation_phe_pi = TensorMap('phecode_938',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_938_date'), channel_map={'no_dermatitis_due_to_solar_radiation':0, 'dermatitis_due_to_solar_radiation_prevalent':1, 'dermatitis_due_to_solar_radiation_incident':2}, loss=weighted_crossentropy([1.0, 463, 187], 'dermatitis_due_to_solar_radiation_pi'))
dermatitis_herpetiformis_phe_pi = TensorMap('phecode_695.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.21_date'), channel_map={'no_dermatitis_herpetiformis':0, 'dermatitis_herpetiformis_prevalent':1, 'dermatitis_herpetiformis_incident':2}, loss=weighted_crossentropy([1.0, 12195, 11627], 'dermatitis_herpetiformis_pi'))
dermatomycoses_phe_pi = TensorMap('phecode_110.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_110.2_date'), channel_map={'no_dermatomycoses':0, 'dermatomycoses_prevalent':1, 'dermatomycoses_incident':2}, loss=weighted_crossentropy([1.0, 100000, 62500], 'dermatomycoses_pi'))
dermatomyositis_phe_pi = TensorMap('phecode_709.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_709.5_date'), channel_map={'no_dermatomyositis':0, 'dermatomyositis_prevalent':1, 'dermatomyositis_incident':2}, loss=weighted_crossentropy([1.0, 17241, 12195], 'dermatomyositis_pi'))
dermatophytosis_phe_pi = TensorMap('phecode_110.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_110.1_date'), channel_map={'no_dermatophytosis':0, 'dermatophytosis_prevalent':1, 'dermatophytosis_incident':2}, loss=weighted_crossentropy([1.0, 20000, 8333], 'dermatophytosis_pi'))
dermatophytosis_of_nail_phe_pi = TensorMap('phecode_110.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_110.11_date'), channel_map={'no_dermatophytosis_of_nail':0, 'dermatophytosis_of_nail_prevalent':1, 'dermatophytosis_of_nail_incident':2}, loss=weighted_crossentropy([1.0, 14285, 5747], 'dermatophytosis_of_nail_pi'))
dermatophytosis_of_the_body_phe_pi = TensorMap('phecode_110.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_110.13_date'), channel_map={'no_dermatophytosis_of_the_body':0, 'dermatophytosis_of_the_body_prevalent':1, 'dermatophytosis_of_the_body_incident':2}, loss=weighted_crossentropy([1.0, 45454, 13888], 'dermatophytosis_of_the_body_pi'))
develomental_delays_and_disorders_phe_pi = TensorMap('phecode_315',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_315_date'), channel_map={'no_develomental_delays_and_disorders':0, 'develomental_delays_and_disorders_prevalent':1, 'develomental_delays_and_disorders_incident':2}, loss=weighted_crossentropy([1.0, 3067, 625], 'develomental_delays_and_disorders_pi'))
diabetes_insipidus_phe_pi = TensorMap('phecode_253.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_253.3_date'), channel_map={'no_diabetes_insipidus':0, 'diabetes_insipidus_prevalent':1, 'diabetes_insipidus_incident':2}, loss=weighted_crossentropy([1.0, 7142, 7575], 'diabetes_insipidus_pi'))
diabetes_mellitus_phe_pi = TensorMap('phecode_250',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250_date'), channel_map={'no_diabetes_mellitus':0, 'diabetes_mellitus_prevalent':1, 'diabetes_mellitus_incident':2}, loss=weighted_crossentropy([1.0, 45, 24], 'diabetes_mellitus_pi'))
diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy_phe_pi = TensorMap('phecode_649.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_649.1_date'), channel_map={'no_diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy':0, 'diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy_prevalent':1, 'diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 1724, 20000], 'diabetes_or_abnormal_glucose_tolerance_complicating_pregnancy_pi'))
diabetic_retinopathy_phe_pi = TensorMap('phecode_250.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.7_date'), channel_map={'no_diabetic_retinopathy':0, 'diabetic_retinopathy_prevalent':1, 'diabetic_retinopathy_incident':2}, loss=weighted_crossentropy([1.0, 616, 332], 'diabetic_retinopathy_pi'))
diaper_or_napkin_rash_phe_pi = TensorMap('phecode_949.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_949.1_date'), channel_map={'no_diaper_or_napkin_rash':0, 'diaper_or_napkin_rash_prevalent':1, 'diaper_or_napkin_rash_incident':2}, loss=weighted_crossentropy([1.0, 250000, 71428], 'diaper_or_napkin_rash_pi'))
diaphragmatic_hernia_phe_pi = TensorMap('phecode_550.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_550.2_date'), channel_map={'no_diaphragmatic_hernia':0, 'diaphragmatic_hernia_prevalent':1, 'diaphragmatic_hernia_incident':2}, loss=weighted_crossentropy([1.0, 30, 21], 'diaphragmatic_hernia_pi'))
diarrhea_phe_pi = TensorMap('phecode_561.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_561.1_date'), channel_map={'no_diarrhea':0, 'diarrhea_prevalent':1, 'diarrhea_incident':2}, loss=weighted_crossentropy([1.0, 2857, 815], 'diarrhea_pi'))
difficulty_in_walking_phe_pi = TensorMap('phecode_741.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741.3_date'), channel_map={'no_difficulty_in_walking':0, 'difficulty_in_walking_prevalent':1, 'difficulty_in_walking_incident':2}, loss=weighted_crossentropy([1.0, 6250, 2272], 'difficulty_in_walking_pi'))
digestive_congenital_anomalies_phe_pi = TensorMap('phecode_750',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750_date'), channel_map={'no_digestive_congenital_anomalies':0, 'digestive_congenital_anomalies_prevalent':1, 'digestive_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 1351, 838], 'digestive_congenital_anomalies_pi'))
diplopia_and_disorders_of_binocular_vision_phe_pi = TensorMap('phecode_368.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.2_date'), channel_map={'no_diplopia_and_disorders_of_binocular_vision':0, 'diplopia_and_disorders_of_binocular_vision_prevalent':1, 'diplopia_and_disorders_of_binocular_vision_incident':2}, loss=weighted_crossentropy([1.0, 1265, 727], 'diplopia_and_disorders_of_binocular_vision_pi'))
disease_of_capillaries_phe_pi = TensorMap('phecode_448',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_448_date'), channel_map={'no_disease_of_capillaries':0, 'disease_of_capillaries_prevalent':1, 'disease_of_capillaries_incident':2}, loss=weighted_crossentropy([1.0, 132, 68], 'disease_of_capillaries_pi'))
disease_of_tricuspid_valve_phe_pi = TensorMap('phecode_394.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_394.7_date'), channel_map={'no_disease_of_tricuspid_valve':0, 'disease_of_tricuspid_valve_prevalent':1, 'disease_of_tricuspid_valve_incident':2}, loss=weighted_crossentropy([1.0, 2024, 277], 'disease_of_tricuspid_valve_pi'))
diseases_and_other_conditions_of_the_tongue_phe_pi = TensorMap('phecode_529',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_529_date'), channel_map={'no_diseases_and_other_conditions_of_the_tongue':0, 'diseases_and_other_conditions_of_the_tongue_prevalent':1, 'diseases_and_other_conditions_of_the_tongue_incident':2}, loss=weighted_crossentropy([1.0, 996, 600], 'diseases_and_other_conditions_of_the_tongue_pi'))
diseases_due_to_other_mycobacteria_phe_pi = TensorMap('phecode_31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_31_date'), channel_map={'no_diseases_due_to_other_mycobacteria':0, 'diseases_due_to_other_mycobacteria_prevalent':1, 'diseases_due_to_other_mycobacteria_incident':2}, loss=weighted_crossentropy([1.0, 83333, 18518], 'diseases_due_to_other_mycobacteria_pi'))
diseases_of_esophagus_phe_pi = TensorMap('phecode_530',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530_date'), channel_map={'no_diseases_of_esophagus':0, 'diseases_of_esophagus_prevalent':1, 'diseases_of_esophagus_incident':2}, loss=weighted_crossentropy([1.0, 1243, 1054], 'diseases_of_esophagus_pi'))
diseases_of_hair_and_hair_follicles_phe_pi = TensorMap('phecode_704',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_704_date'), channel_map={'no_diseases_of_hair_and_hair_follicles':0, 'diseases_of_hair_and_hair_follicles_prevalent':1, 'diseases_of_hair_and_hair_follicles_incident':2}, loss=weighted_crossentropy([1.0, 110, 237], 'diseases_of_hair_and_hair_follicles_pi'))
diseases_of_hard_tissues_of_teeth_phe_pi = TensorMap('phecode_521',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_521_date'), channel_map={'no_diseases_of_hard_tissues_of_teeth':0, 'diseases_of_hard_tissues_of_teeth_prevalent':1, 'diseases_of_hard_tissues_of_teeth_incident':2}, loss=weighted_crossentropy([1.0, 33333, 31250], 'diseases_of_hard_tissues_of_teeth_pi'))
diseases_of_lips_phe_pi = TensorMap('phecode_528.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.5_date'), channel_map={'no_diseases_of_lips':0, 'diseases_of_lips_prevalent':1, 'diseases_of_lips_incident':2}, loss=weighted_crossentropy([1.0, 1246, 894], 'diseases_of_lips_pi'))
diseases_of_nail_nos_phe_pi = TensorMap('phecode_703',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_703_date'), channel_map={'no_diseases_of_nail_nos':0, 'diseases_of_nail_nos_prevalent':1, 'diseases_of_nail_nos_incident':2}, loss=weighted_crossentropy([1.0, 2380, 1930], 'diseases_of_nail_nos_pi'))
diseases_of_pancreas_phe_pi = TensorMap('phecode_577',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_577_date'), channel_map={'no_diseases_of_pancreas':0, 'diseases_of_pancreas_prevalent':1, 'diseases_of_pancreas_incident':2}, loss=weighted_crossentropy([1.0, 3333, 661], 'diseases_of_pancreas_pi'))
diseases_of_pulp_and_periapical_tissues_phe_pi = TensorMap('phecode_522',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_522_date'), channel_map={'no_diseases_of_pulp_and_periapical_tissues':0, 'diseases_of_pulp_and_periapical_tissues_prevalent':1, 'diseases_of_pulp_and_periapical_tissues_incident':2}, loss=weighted_crossentropy([1.0, 976, 1879], 'diseases_of_pulp_and_periapical_tissues_pi'))
diseases_of_spleen_phe_pi = TensorMap('phecode_289.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_289.5_date'), channel_map={'no_diseases_of_spleen':0, 'diseases_of_spleen_prevalent':1, 'diseases_of_spleen_incident':2}, loss=weighted_crossentropy([1.0, 2202, 905], 'diseases_of_spleen_pi'))
diseases_of_the_jaws_phe_pi = TensorMap('phecode_526',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_526_date'), channel_map={'no_diseases_of_the_jaws':0, 'diseases_of_the_jaws_prevalent':1, 'diseases_of_the_jaws_incident':2}, loss=weighted_crossentropy([1.0, 26, 22], 'diseases_of_the_jaws_pi'))
diseases_of_the_larynx_and_vocal_cords_phe_pi = TensorMap('phecode_473',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_473_date'), channel_map={'no_diseases_of_the_larynx_and_vocal_cords':0, 'diseases_of_the_larynx_and_vocal_cords_prevalent':1, 'diseases_of_the_larynx_and_vocal_cords_incident':2}, loss=weighted_crossentropy([1.0, 457, 504], 'diseases_of_the_larynx_and_vocal_cords_pi'))
diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue_phe_pi = TensorMap('phecode_528',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528_date'), channel_map={'no_diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue':0, 'diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue_prevalent':1, 'diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue_incident':2}, loss=weighted_crossentropy([1.0, 453, 292], 'diseases_of_the_oral_soft_tissues_excluding_lesions_specific_for_gingiva_and_tongue_pi'))
diseases_of_the_salivary_glands_phe_pi = TensorMap('phecode_527',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_527_date'), channel_map={'no_diseases_of_the_salivary_glands':0, 'diseases_of_the_salivary_glands_prevalent':1, 'diseases_of_the_salivary_glands_incident':2}, loss=weighted_crossentropy([1.0, 5882, 9433], 'diseases_of_the_salivary_glands_pi'))
diseases_of_thymus_gland_phe_pi = TensorMap('phecode_254',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_254_date'), channel_map={'no_diseases_of_thymus_gland':0, 'diseases_of_thymus_gland_prevalent':1, 'diseases_of_thymus_gland_incident':2}, loss=weighted_crossentropy([1.0, 45454, 31250], 'diseases_of_thymus_gland_pi'))
diseases_of_white_blood_cells_phe_pi = TensorMap('phecode_288',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_288_date'), channel_map={'no_diseases_of_white_blood_cells':0, 'diseases_of_white_blood_cells_prevalent':1, 'diseases_of_white_blood_cells_incident':2}, loss=weighted_crossentropy([1.0, 1666, 810], 'diseases_of_white_blood_cells_pi'))
dislocation_phe_pi = TensorMap('phecode_830',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_830_date'), channel_map={'no_dislocation':0, 'dislocation_prevalent':1, 'dislocation_incident':2}, loss=weighted_crossentropy([1.0, 335, 415], 'dislocation_pi'))
disorder_of_skin_and_subcutaneous_tissue_nos_phe_pi = TensorMap('phecode_689',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_689_date'), channel_map={'no_disorder_of_skin_and_subcutaneous_tissue_nos':0, 'disorder_of_skin_and_subcutaneous_tissue_nos_prevalent':1, 'disorder_of_skin_and_subcutaneous_tissue_nos_incident':2}, loss=weighted_crossentropy([1.0, 21, 20], 'disorder_of_skin_and_subcutaneous_tissue_nos_pi'))
disorders_involving_the_immune_mechanism_phe_pi = TensorMap('phecode_279',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_279_date'), channel_map={'no_disorders_involving_the_immune_mechanism':0, 'disorders_involving_the_immune_mechanism_prevalent':1, 'disorders_involving_the_immune_mechanism_incident':2}, loss=weighted_crossentropy([1.0, 125000, 10000], 'disorders_involving_the_immune_mechanism_pi'))
disorders_of_accommodation_phe_pi = TensorMap('phecode_368.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.7_date'), channel_map={'no_disorders_of_accommodation':0, 'disorders_of_accommodation_prevalent':1, 'disorders_of_accommodation_incident':2}, loss=weighted_crossentropy([1.0, 166666, 125000], 'disorders_of_accommodation_pi'))
disorders_of_acoustic_nerve_phe_pi = TensorMap('phecode_389.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_389.5_date'), channel_map={'no_disorders_of_acoustic_nerve':0, 'disorders_of_acoustic_nerve_prevalent':1, 'disorders_of_acoustic_nerve_incident':2}, loss=weighted_crossentropy([1.0, 45454, 16129], 'disorders_of_acoustic_nerve_pi'))
disorders_of_adrenal_glands_phe_pi = TensorMap('phecode_255',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_255_date'), channel_map={'no_disorders_of_adrenal_glands':0, 'disorders_of_adrenal_glands_prevalent':1, 'disorders_of_adrenal_glands_incident':2}, loss=weighted_crossentropy([1.0, 9433, 2487], 'disorders_of_adrenal_glands_pi'))
disorders_of_amino_acid_metabolism_phe_pi = TensorMap('phecode_270.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.2_date'), channel_map={'no_disorders_of_amino_acid_metabolism':0, 'disorders_of_amino_acid_metabolism_prevalent':1, 'disorders_of_amino_acid_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 125000, 125000], 'disorders_of_amino_acid_metabolism_pi'))
disorders_of_bilirubin_excretion_phe_pi = TensorMap('phecode_277.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_277.4_date'), channel_map={'no_disorders_of_bilirubin_excretion':0, 'disorders_of_bilirubin_excretion_prevalent':1, 'disorders_of_bilirubin_excretion_incident':2}, loss=weighted_crossentropy([1.0, 3289, 1262], 'disorders_of_bilirubin_excretion_pi'))
disorders_of_calcium_phosphorus_metabolism_phe_pi = TensorMap('phecode_275.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_275.5_date'), channel_map={'no_disorders_of_calcium_phosphorus_metabolism':0, 'disorders_of_calcium_phosphorus_metabolism_prevalent':1, 'disorders_of_calcium_phosphorus_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 1661, 318], 'disorders_of_calcium_phosphorus_metabolism_pi'))
disorders_of_conjunctiva_phe_pi = TensorMap('phecode_372',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_372_date'), channel_map={'no_disorders_of_conjunctiva':0, 'disorders_of_conjunctiva_prevalent':1, 'disorders_of_conjunctiva_incident':2}, loss=weighted_crossentropy([1.0, 732, 512], 'disorders_of_conjunctiva_pi'))
disorders_of_copper_metabolism_phe_pi = TensorMap('phecode_275.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_275.2_date'), channel_map={'no_disorders_of_copper_metabolism':0, 'disorders_of_copper_metabolism_prevalent':1, 'disorders_of_copper_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 250000, 62500], 'disorders_of_copper_metabolism_pi'))
disorders_of_diaphragm_phe_pi = TensorMap('phecode_513.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_513.8_date'), channel_map={'no_disorders_of_diaphragm':0, 'disorders_of_diaphragm_prevalent':1, 'disorders_of_diaphragm_incident':2}, loss=weighted_crossentropy([1.0, 8620, 2717], 'disorders_of_diaphragm_pi'))
disorders_of_esophageal_motility_phe_pi = TensorMap('phecode_530.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.5_date'), channel_map={'no_disorders_of_esophageal_motility':0, 'disorders_of_esophageal_motility_prevalent':1, 'disorders_of_esophageal_motility_incident':2}, loss=weighted_crossentropy([1.0, 1362, 880], 'disorders_of_esophageal_motility_pi'))
disorders_of_external_ear_phe_pi = TensorMap('phecode_380',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_380_date'), channel_map={'no_disorders_of_external_ear':0, 'disorders_of_external_ear_prevalent':1, 'disorders_of_external_ear_incident':2}, loss=weighted_crossentropy([1.0, 531, 486], 'disorders_of_external_ear_pi'))
disorders_of_function_of_stomach_phe_pi = TensorMap('phecode_536',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_536_date'), channel_map={'no_disorders_of_function_of_stomach':0, 'disorders_of_function_of_stomach_prevalent':1, 'disorders_of_function_of_stomach_incident':2}, loss=weighted_crossentropy([1.0, 71428, 41666], 'disorders_of_function_of_stomach_pi'))
disorders_of_iris_and_ciliary_body_phe_pi = TensorMap('phecode_379.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379.5_date'), channel_map={'no_disorders_of_iris_and_ciliary_body':0, 'disorders_of_iris_and_ciliary_body_prevalent':1, 'disorders_of_iris_and_ciliary_body_incident':2}, loss=weighted_crossentropy([1.0, 3311, 1204], 'disorders_of_iris_and_ciliary_body_pi'))
disorders_of_iron_metabolism_phe_pi = TensorMap('phecode_275.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_275.1_date'), channel_map={'no_disorders_of_iron_metabolism':0, 'disorders_of_iron_metabolism_prevalent':1, 'disorders_of_iron_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 1466, 788], 'disorders_of_iron_metabolism_pi'))
disorders_of_lacrimal_system_phe_pi = TensorMap('phecode_375',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_375_date'), channel_map={'no_disorders_of_lacrimal_system':0, 'disorders_of_lacrimal_system_prevalent':1, 'disorders_of_lacrimal_system_incident':2}, loss=weighted_crossentropy([1.0, 540, 325], 'disorders_of_lacrimal_system_pi'))
disorders_of_magnesium_metabolism_phe_pi = TensorMap('phecode_275.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_275.3_date'), channel_map={'no_disorders_of_magnesium_metabolism':0, 'disorders_of_magnesium_metabolism_prevalent':1, 'disorders_of_magnesium_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 15151, 648], 'disorders_of_magnesium_metabolism_pi'))
disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract_phe_pi = TensorMap('phecode_626',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626_date'), channel_map={'no_disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract':0, 'disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract_prevalent':1, 'disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract_incident':2}, loss=weighted_crossentropy([1.0, 329, 884], 'disorders_of_menstruation_and_other_abnormal_bleeding_from_female_genital_tract_pi'))
disorders_of_mineral_metabolism_phe_pi = TensorMap('phecode_275',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_275_date'), channel_map={'no_disorders_of_mineral_metabolism':0, 'disorders_of_mineral_metabolism_prevalent':1, 'disorders_of_mineral_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 71428, 45454], 'disorders_of_mineral_metabolism_pi'))
disorders_of_muscle_ligament_and_fascia_phe_pi = TensorMap('phecode_728',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_728_date'), channel_map={'no_disorders_of_muscle_ligament_and_fascia':0, 'disorders_of_muscle_ligament_and_fascia_prevalent':1, 'disorders_of_muscle_ligament_and_fascia_incident':2}, loss=weighted_crossentropy([1.0, 14705, 13888], 'disorders_of_muscle_ligament_and_fascia_pi'))
disorders_of_optic_nerve_and_visual_pathways_phe_pi = TensorMap('phecode_377',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_377_date'), channel_map={'no_disorders_of_optic_nerve_and_visual_pathways':0, 'disorders_of_optic_nerve_and_visual_pathways_prevalent':1, 'disorders_of_optic_nerve_and_visual_pathways_incident':2}, loss=weighted_crossentropy([1.0, 2127, 1243], 'disorders_of_optic_nerve_and_visual_pathways_pi'))
disorders_of_other_cranial_nerves_phe_pi = TensorMap('phecode_352',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_352_date'), channel_map={'no_disorders_of_other_cranial_nerves':0, 'disorders_of_other_cranial_nerves_prevalent':1, 'disorders_of_other_cranial_nerves_incident':2}, loss=weighted_crossentropy([1.0, 12820, 7352], 'disorders_of_other_cranial_nerves_pi'))
disorders_of_parathyroid_gland_phe_pi = TensorMap('phecode_252',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_252_date'), channel_map={'no_disorders_of_parathyroid_gland':0, 'disorders_of_parathyroid_gland_prevalent':1, 'disorders_of_parathyroid_gland_incident':2}, loss=weighted_crossentropy([1.0, 33333, 17857], 'disorders_of_parathyroid_gland_pi'))
disorders_of_penis_phe_pi = TensorMap('phecode_604',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_604_date'), channel_map={'no_disorders_of_penis':0, 'disorders_of_penis_prevalent':1, 'disorders_of_penis_incident':2}, loss=weighted_crossentropy([1.0, 1355, 733], 'disorders_of_penis_pi'))
disorders_of_phosphorus_metabolism_phe_pi = TensorMap('phecode_275.53',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_275.53_date'), channel_map={'no_disorders_of_phosphorus_metabolism':0, 'disorders_of_phosphorus_metabolism_prevalent':1, 'disorders_of_phosphorus_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 35714, 1269], 'disorders_of_phosphorus_metabolism_pi'))
disorders_of_porphyrin_metabolism_phe_pi = TensorMap('phecode_277.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_277.1_date'), channel_map={'no_disorders_of_porphyrin_metabolism':0, 'disorders_of_porphyrin_metabolism_prevalent':1, 'disorders_of_porphyrin_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 20833, 20833], 'disorders_of_porphyrin_metabolism_pi'))
disorders_of_refraction_and_accommodation_blindness_and_low_vision_phe_pi = TensorMap('phecode_367',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_367_date'), channel_map={'no_disorders_of_refraction_and_accommodation_blindness_and_low_vision':0, 'disorders_of_refraction_and_accommodation_blindness_and_low_vision_prevalent':1, 'disorders_of_refraction_and_accommodation_blindness_and_low_vision_incident':2}, loss=weighted_crossentropy([1.0, 20000, 4166], 'disorders_of_refraction_and_accommodation_blindness_and_low_vision_pi'))
disorders_of_sweat_glands_phe_pi = TensorMap('phecode_705',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_705_date'), channel_map={'no_disorders_of_sweat_glands':0, 'disorders_of_sweat_glands_prevalent':1, 'disorders_of_sweat_glands_incident':2}, loss=weighted_crossentropy([1.0, 38461, 26315], 'disorders_of_sweat_glands_pi'))
disorders_of_the_autonomic_nervous_system_phe_pi = TensorMap('phecode_337',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_337_date'), channel_map={'no_disorders_of_the_autonomic_nervous_system':0, 'disorders_of_the_autonomic_nervous_system_prevalent':1, 'disorders_of_the_autonomic_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 5376, 2808], 'disorders_of_the_autonomic_nervous_system_pi'))
disorders_of_the_globe_phe_pi = TensorMap('phecode_360',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_360_date'), channel_map={'no_disorders_of_the_globe':0, 'disorders_of_the_globe_prevalent':1, 'disorders_of_the_globe_incident':2}, loss=weighted_crossentropy([1.0, 11363, 7352], 'disorders_of_the_globe_pi'))
disorders_of_the_orbit_phe_pi = TensorMap('phecode_376',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_376_date'), channel_map={'no_disorders_of_the_orbit':0, 'disorders_of_the_orbit_prevalent':1, 'disorders_of_the_orbit_incident':2}, loss=weighted_crossentropy([1.0, 10000, 8064], 'disorders_of_the_orbit_pi'))
disorders_of_the_pituitary_gland_and_its_hypothalamic_control_phe_pi = TensorMap('phecode_253',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_253_date'), channel_map={'no_disorders_of_the_pituitary_gland_and_its_hypothalamic_control':0, 'disorders_of_the_pituitary_gland_and_its_hypothalamic_control_prevalent':1, 'disorders_of_the_pituitary_gland_and_its_hypothalamic_control_incident':2}, loss=weighted_crossentropy([1.0, 8474, 6172], 'disorders_of_the_pituitary_gland_and_its_hypothalamic_control_pi'))
disorders_of_tooth_development_phe_pi = TensorMap('phecode_520',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_520_date'), channel_map={'no_disorders_of_tooth_development':0, 'disorders_of_tooth_development_prevalent':1, 'disorders_of_tooth_development_incident':2}, loss=weighted_crossentropy([1.0, 15625, 29411], 'disorders_of_tooth_development_pi'))
disorders_of_urea_cycle_metabolism_phe_pi = TensorMap('phecode_270.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.21_date'), channel_map={'no_disorders_of_urea_cycle_metabolism':0, 'disorders_of_urea_cycle_metabolism_prevalent':1, 'disorders_of_urea_cycle_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 250000, 71428], 'disorders_of_urea_cycle_metabolism_pi'))
disorders_of_uterus_nec_phe_pi = TensorMap('phecode_619.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_619.2_date'), channel_map={'no_disorders_of_uterus_nec':0, 'disorders_of_uterus_nec_prevalent':1, 'disorders_of_uterus_nec_incident':2}, loss=weighted_crossentropy([1.0, 198, 256], 'disorders_of_uterus_nec_pi'))
disorders_of_vitreous_body_phe_pi = TensorMap('phecode_379.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379.2_date'), channel_map={'no_disorders_of_vitreous_body':0, 'disorders_of_vitreous_body_prevalent':1, 'disorders_of_vitreous_body_incident':2}, loss=weighted_crossentropy([1.0, 549, 292], 'disorders_of_vitreous_body_pi'))
disorders_resulting_from_impaired_renal_function_phe_pi = TensorMap('phecode_588',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_588_date'), channel_map={'no_disorders_resulting_from_impaired_renal_function':0, 'disorders_resulting_from_impaired_renal_function_prevalent':1, 'disorders_resulting_from_impaired_renal_function_incident':2}, loss=weighted_crossentropy([1.0, 8620, 6172], 'disorders_resulting_from_impaired_renal_function_pi'))
disorders_secondary_to_childbirth_surgery_trauma_phe_pi = TensorMap('phecode_617',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_617_date'), channel_map={'no_disorders_secondary_to_childbirth_surgery_trauma':0, 'disorders_secondary_to_childbirth_surgery_trauma_prevalent':1, 'disorders_secondary_to_childbirth_surgery_trauma_incident':2}, loss=weighted_crossentropy([1.0, 2192, 2702], 'disorders_secondary_to_childbirth_surgery_trauma_pi'))
displacement_of_intervertebral_disc_phe_pi = TensorMap('phecode_722.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722.1_date'), channel_map={'no_displacement_of_intervertebral_disc':0, 'displacement_of_intervertebral_disc_prevalent':1, 'displacement_of_intervertebral_disc_incident':2}, loss=weighted_crossentropy([1.0, 2551, 1552], 'displacement_of_intervertebral_disc_pi'))
dissociative_disorder_phe_pi = TensorMap('phecode_303.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_303.1_date'), channel_map={'no_dissociative_disorder':0, 'dissociative_disorder_prevalent':1, 'dissociative_disorder_incident':2}, loss=weighted_crossentropy([1.0, 15151, 16129], 'dissociative_disorder_pi'))
disturbance_of_salivary_secretion_phe_pi = TensorMap('phecode_527.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_527.7_date'), channel_map={'no_disturbance_of_salivary_secretion':0, 'disturbance_of_salivary_secretion_prevalent':1, 'disturbance_of_salivary_secretion_incident':2}, loss=weighted_crossentropy([1.0, 18518, 6493], 'disturbance_of_salivary_secretion_pi'))
disturbance_of_skin_sensation_phe_pi = TensorMap('phecode_687.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_687.4_date'), channel_map={'no_disturbance_of_skin_sensation':0, 'disturbance_of_skin_sensation_prevalent':1, 'disturbance_of_skin_sensation_incident':2}, loss=weighted_crossentropy([1.0, 118, 80], 'disturbance_of_skin_sensation_pi'))
disturbances_in_tooth_eruption_phe_pi = TensorMap('phecode_520.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_520.2_date'), channel_map={'no_disturbances_in_tooth_eruption':0, 'disturbances_in_tooth_eruption_prevalent':1, 'disturbances_in_tooth_eruption_incident':2}, loss=weighted_crossentropy([1.0, 273, 686], 'disturbances_in_tooth_eruption_pi'))
disturbances_of_amino_acid_transport_phe_pi = TensorMap('phecode_270.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.1_date'), channel_map={'no_disturbances_of_amino_acid_transport':0, 'disturbances_of_amino_acid_transport_prevalent':1, 'disturbances_of_amino_acid_transport_incident':2}, loss=weighted_crossentropy([1.0, 29411, 17857], 'disturbances_of_amino_acid_transport_pi'))
disturbances_of_sensation_of_smell_and_taste_phe_pi = TensorMap('phecode_350.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_350.6_date'), channel_map={'no_disturbances_of_sensation_of_smell_and_taste':0, 'disturbances_of_sensation_of_smell_and_taste_prevalent':1, 'disturbances_of_sensation_of_smell_and_taste_incident':2}, loss=weighted_crossentropy([1.0, 18518, 11627], 'disturbances_of_sensation_of_smell_and_taste_pi'))
disturbances_of_sulphur_bearing_amino_acid_metabolism_phe_pi = TensorMap('phecode_270.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.11_date'), channel_map={'no_disturbances_of_sulphur_bearing_amino_acid_metabolism':0, 'disturbances_of_sulphur_bearing_amino_acid_metabolism_prevalent':1, 'disturbances_of_sulphur_bearing_amino_acid_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 62500, 100000], 'disturbances_of_sulphur_bearing_amino_acid_metabolism_pi'))
diverticulosis_phe_pi = TensorMap('phecode_562.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_562.1_date'), channel_map={'no_diverticulosis':0, 'diverticulosis_prevalent':1, 'diverticulosis_incident':2}, loss=weighted_crossentropy([1.0, 44, 16], 'diverticulosis_pi'))
diverticulum_of_esophagus_acquired_phe_pi = TensorMap('phecode_530.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.6_date'), channel_map={'no_diverticulum_of_esophagus_acquired':0, 'diverticulum_of_esophagus_acquired_prevalent':1, 'diverticulum_of_esophagus_acquired_incident':2}, loss=weighted_crossentropy([1.0, 6578, 2824], 'diverticulum_of_esophagus_acquired_pi'))
dizziness_and_giddiness_light_headedness_and_vertigo_phe_pi = TensorMap('phecode_386.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_386.9_date'), channel_map={'no_dizziness_and_giddiness_light_headedness_and_vertigo':0, 'dizziness_and_giddiness_light_headedness_and_vertigo_prevalent':1, 'dizziness_and_giddiness_light_headedness_and_vertigo_incident':2}, loss=weighted_crossentropy([1.0, 212, 102], 'dizziness_and_giddiness_light_headedness_and_vertigo_pi'))
duodenal_ulcer_phe_pi = TensorMap('phecode_531.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_531.3_date'), channel_map={'no_duodenal_ulcer':0, 'duodenal_ulcer_prevalent':1, 'duodenal_ulcer_incident':2}, loss=weighted_crossentropy([1.0, 242, 260], 'duodenal_ulcer_pi'))
duodenitis_phe_pi = TensorMap('phecode_535.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_535.6_date'), channel_map={'no_duodenitis':0, 'duodenitis_prevalent':1, 'duodenitis_incident':2}, loss=weighted_crossentropy([1.0, 93, 92], 'duodenitis_pi'))
dyshidrosis_phe_pi = TensorMap('phecode_705.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_705.1_date'), channel_map={'no_dyshidrosis':0, 'dyshidrosis_prevalent':1, 'dyshidrosis_incident':2}, loss=weighted_crossentropy([1.0, 50000, 71428], 'dyshidrosis_pi'))
dysmenorrhea_phe_pi = TensorMap('phecode_626.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.2_date'), channel_map={'no_dysmenorrhea':0, 'dysmenorrhea_prevalent':1, 'dysmenorrhea_incident':2}, loss=weighted_crossentropy([1.0, 306, 1661], 'dysmenorrhea_pi'))
dyspareunia_phe_pi = TensorMap('phecode_625.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_625.1_date'), channel_map={'no_dyspareunia':0, 'dyspareunia_prevalent':1, 'dyspareunia_incident':2}, loss=weighted_crossentropy([1.0, 523, 1432], 'dyspareunia_pi'))
dysphagia_phe_pi = TensorMap('phecode_532',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_532_date'), channel_map={'no_dysphagia':0, 'dysphagia_prevalent':1, 'dysphagia_incident':2}, loss=weighted_crossentropy([1.0, 158, 72], 'dysphagia_pi'))
dysplasia_of_female_genital_organs_phe_pi = TensorMap('phecode_620',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_620_date'), channel_map={'no_dysplasia_of_female_genital_organs':0, 'dysplasia_of_female_genital_organs_prevalent':1, 'dysplasia_of_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 6944, 12500], 'dysplasia_of_female_genital_organs_pi'))
dysthymic_disorder_phe_pi = TensorMap('phecode_300.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.4_date'), channel_map={'no_dysthymic_disorder':0, 'dysthymic_disorder_prevalent':1, 'dysthymic_disorder_incident':2}, loss=weighted_crossentropy([1.0, 12195, 14705], 'dysthymic_disorder_pi'))
dystrophy_of_female_genital_tract_phe_pi = TensorMap('phecode_624.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_624.1_date'), channel_map={'no_dystrophy_of_female_genital_tract':0, 'dystrophy_of_female_genital_tract_prevalent':1, 'dystrophy_of_female_genital_tract_incident':2}, loss=weighted_crossentropy([1.0, 5000, 1773], 'dystrophy_of_female_genital_tract_pi'))
dysuria_phe_pi = TensorMap('phecode_599.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.3_date'), channel_map={'no_dysuria':0, 'dysuria_prevalent':1, 'dysuria_incident':2}, loss=weighted_crossentropy([1.0, 641, 554], 'dysuria_pi'))
e_coli_phe_pi = TensorMap('phecode_41.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_41.4_date'), channel_map={'no_e_coli':0, 'e_coli_prevalent':1, 'e_coli_incident':2}, loss=weighted_crossentropy([1.0, 547, 136], 'e_coli_pi'))
early_onset_of_delivery_phe_pi = TensorMap('phecode_636.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_636.2_date'), channel_map={'no_early_onset_of_delivery':0, 'early_onset_of_delivery_prevalent':1, 'early_onset_of_delivery_incident':2}, loss=weighted_crossentropy([1.0, 601, 14705], 'early_onset_of_delivery_pi'))
early_or_threatened_labor_hemorrhage_in_early_pregnancy_phe_pi = TensorMap('phecode_636',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_636_date'), channel_map={'no_early_or_threatened_labor_hemorrhage_in_early_pregnancy':0, 'early_or_threatened_labor_hemorrhage_in_early_pregnancy_prevalent':1, 'early_or_threatened_labor_hemorrhage_in_early_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 369, 100000], 'early_or_threatened_labor_hemorrhage_in_early_pregnancy_pi'))
eating_disorder_phe_pi = TensorMap('phecode_305.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_305.2_date'), channel_map={'no_eating_disorder':0, 'eating_disorder_prevalent':1, 'eating_disorder_incident':2}, loss=weighted_crossentropy([1.0, 16129, 8928], 'eating_disorder_pi'))
ectopic_pregnancy_phe_pi = TensorMap('phecode_634.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_634.3_date'), channel_map={'no_ectopic_pregnancy':0, 'ectopic_pregnancy_prevalent':1, 'ectopic_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 134, 4385], 'ectopic_pregnancy_pi'))
ectropion_or_entropion_phe_pi = TensorMap('phecode_374.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_374.1_date'), channel_map={'no_ectropion_or_entropion':0, 'ectropion_or_entropion_prevalent':1, 'ectropion_or_entropion_incident':2}, loss=weighted_crossentropy([1.0, 1168, 433], 'ectropion_or_entropion_pi'))
edema_phe_pi = TensorMap('phecode_782.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_782.3_date'), channel_map={'no_edema':0, 'edema_prevalent':1, 'edema_incident':2}, loss=weighted_crossentropy([1.0, 1024, 202], 'edema_pi'))
effects_of_heat_cold_and_air_pressure_phe_pi = TensorMap('phecode_1014',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1014_date'), channel_map={'no_effects_of_heat_cold_and_air_pressure':0, 'effects_of_heat_cold_and_air_pressure_prevalent':1, 'effects_of_heat_cold_and_air_pressure_incident':2}, loss=weighted_crossentropy([1.0, 2415, 2032], 'effects_of_heat_cold_and_air_pressure_pi'))
effects_of_other_external_causes_phe_pi = TensorMap('phecode_1015',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1015_date'), channel_map={'no_effects_of_other_external_causes':0, 'effects_of_other_external_causes_prevalent':1, 'effects_of_other_external_causes_incident':2}, loss=weighted_crossentropy([1.0, 165, 108], 'effects_of_other_external_causes_pi'))
effects_radiation_nos_phe_pi = TensorMap('phecode_990',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_990_date'), channel_map={'no_effects_radiation_nos':0, 'effects_radiation_nos_prevalent':1, 'effects_radiation_nos_incident':2}, loss=weighted_crossentropy([1.0, 363, 82], 'effects_radiation_nos_pi'))
electrolyte_imbalance_phe_pi = TensorMap('phecode_276.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.1_date'), channel_map={'no_electrolyte_imbalance':0, 'electrolyte_imbalance_prevalent':1, 'electrolyte_imbalance_incident':2}, loss=weighted_crossentropy([1.0, 18518, 2336], 'electrolyte_imbalance_pi'))
elevated_blood_pressure_reading_without_diagnosis_of_hypertension_phe_pi = TensorMap('phecode_402',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_402_date'), channel_map={'no_elevated_blood_pressure_reading_without_diagnosis_of_hypertension':0, 'elevated_blood_pressure_reading_without_diagnosis_of_hypertension_prevalent':1, 'elevated_blood_pressure_reading_without_diagnosis_of_hypertension_incident':2}, loss=weighted_crossentropy([1.0, 524, 399], 'elevated_blood_pressure_reading_without_diagnosis_of_hypertension_pi'))
elevated_sedimentation_rate_phe_pi = TensorMap('phecode_790.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_790.1_date'), channel_map={'no_elevated_sedimentation_rate':0, 'elevated_sedimentation_rate_prevalent':1, 'elevated_sedimentation_rate_incident':2}, loss=weighted_crossentropy([1.0, 6493, 12500], 'elevated_sedimentation_rate_pi'))
embolism_and_thrombosis_of_abdominal_aorta_phe_pi = TensorMap('phecode_444.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_444.2_date'), channel_map={'no_embolism_and_thrombosis_of_abdominal_aorta':0, 'embolism_and_thrombosis_of_abdominal_aorta_prevalent':1, 'embolism_and_thrombosis_of_abdominal_aorta_incident':2}, loss=weighted_crossentropy([1.0, 45454, 13888], 'embolism_and_thrombosis_of_abdominal_aorta_pi'))
emphysema_phe_pi = TensorMap('phecode_496.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_496.1_date'), channel_map={'no_emphysema':0, 'emphysema_prevalent':1, 'emphysema_incident':2}, loss=weighted_crossentropy([1.0, 934, 223], 'emphysema_pi'))
empyema_and_pneumothorax_phe_pi = TensorMap('phecode_506',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_506_date'), channel_map={'no_empyema_and_pneumothorax':0, 'empyema_and_pneumothorax_prevalent':1, 'empyema_and_pneumothorax_incident':2}, loss=weighted_crossentropy([1.0, 819, 453], 'empyema_and_pneumothorax_pi'))
encephalitis_phe_pi = TensorMap('phecode_323',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_323_date'), channel_map={'no_encephalitis':0, 'encephalitis_prevalent':1, 'encephalitis_incident':2}, loss=weighted_crossentropy([1.0, 277, 179], 'encephalitis_pi'))
encephalitis_non_infectious_phe_pi = TensorMap('phecode_323.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_323.8_date'), channel_map={'no_encephalitis_non_infectious':0, 'encephalitis_non_infectious_prevalent':1, 'encephalitis_non_infectious_incident':2}, loss=weighted_crossentropy([1.0, 326, 200], 'encephalitis_non_infectious_pi'))
encephalopathy_not_elsewhere_classified_phe_pi = TensorMap('phecode_348.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_348.8_date'), channel_map={'no_encephalopathy_not_elsewhere_classified':0, 'encephalopathy_not_elsewhere_classified_prevalent':1, 'encephalopathy_not_elsewhere_classified_incident':2}, loss=weighted_crossentropy([1.0, 14705, 2617], 'encephalopathy_not_elsewhere_classified_pi'))
endocarditis_phe_pi = TensorMap('phecode_420.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_420.3_date'), channel_map={'no_endocarditis':0, 'endocarditis_prevalent':1, 'endocarditis_incident':2}, loss=weighted_crossentropy([1.0, 1672, 648], 'endocarditis_pi'))
endometrial_hyperplasia_phe_pi = TensorMap('phecode_621',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_621_date'), channel_map={'no_endometrial_hyperplasia':0, 'endometrial_hyperplasia_prevalent':1, 'endometrial_hyperplasia_incident':2}, loss=weighted_crossentropy([1.0, 543, 886], 'endometrial_hyperplasia_pi'))
endometriosis_phe_pi = TensorMap('phecode_615',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_615_date'), channel_map={'no_endometriosis':0, 'endometriosis_prevalent':1, 'endometriosis_incident':2}, loss=weighted_crossentropy([1.0, 133, 398], 'endometriosis_pi'))
enthesopathy_phe_pi = TensorMap('phecode_726.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_726.1_date'), channel_map={'no_enthesopathy':0, 'enthesopathy_prevalent':1, 'enthesopathy_incident':2}, loss=weighted_crossentropy([1.0, 157, 76], 'enthesopathy_pi'))
eosinophilia_phe_pi = TensorMap('phecode_288.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_288.3_date'), channel_map={'no_eosinophilia':0, 'eosinophilia_prevalent':1, 'eosinophilia_incident':2}, loss=weighted_crossentropy([1.0, 17241, 6329], 'eosinophilia_pi'))
epilepsy_phe_pi = TensorMap('phecode_345.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_345.1_date'), channel_map={'no_epilepsy':0, 'epilepsy_prevalent':1, 'epilepsy_incident':2}, loss=weighted_crossentropy([1.0, 3571, 1547], 'epilepsy_pi'))
epilepsy_recurrent_seizures_convulsions_phe_pi = TensorMap('phecode_345',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_345_date'), channel_map={'no_epilepsy_recurrent_seizures_convulsions':0, 'epilepsy_recurrent_seizures_convulsions_prevalent':1, 'epilepsy_recurrent_seizures_convulsions_incident':2}, loss=weighted_crossentropy([1.0, 233, 196], 'epilepsy_recurrent_seizures_convulsions_pi'))
epiphora_phe_pi = TensorMap('phecode_375.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_375.2_date'), channel_map={'no_epiphora':0, 'epiphora_prevalent':1, 'epiphora_incident':2}, loss=weighted_crossentropy([1.0, 915, 666], 'epiphora_pi'))
epistaxis_or_throat_hemorrhage_phe_pi = TensorMap('phecode_477',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_477_date'), channel_map={'no_epistaxis_or_throat_hemorrhage':0, 'epistaxis_or_throat_hemorrhage_prevalent':1, 'epistaxis_or_throat_hemorrhage_incident':2}, loss=weighted_crossentropy([1.0, 363, 239], 'epistaxis_or_throat_hemorrhage_pi'))
erectile_dysfunction_ed_phe_pi = TensorMap('phecode_605',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_605_date'), channel_map={'no_erectile_dysfunction_ed':0, 'erectile_dysfunction_ed_prevalent':1, 'erectile_dysfunction_ed_incident':2}, loss=weighted_crossentropy([1.0, 1385, 1222], 'erectile_dysfunction_ed_pi'))
erythema_nodosum_phe_pi = TensorMap('phecode_695.81',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.81_date'), channel_map={'no_erythema_nodosum':0, 'erythema_nodosum_prevalent':1, 'erythema_nodosum_incident':2}, loss=weighted_crossentropy([1.0, 6410, 15625], 'erythema_nodosum_pi'))
erythematosquamous_dermatosis_phe_pi = TensorMap('phecode_690',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_690_date'), channel_map={'no_erythematosquamous_dermatosis':0, 'erythematosquamous_dermatosis_prevalent':1, 'erythematosquamous_dermatosis_incident':2}, loss=weighted_crossentropy([1.0, 31250, 23809], 'erythematosquamous_dermatosis_pi'))
erythematous_conditions_phe_pi = TensorMap('phecode_695',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695_date'), channel_map={'no_erythematous_conditions':0, 'erythematous_conditions_prevalent':1, 'erythematous_conditions_incident':2}, loss=weighted_crossentropy([1.0, 868, 503], 'erythematous_conditions_pi'))
esophageal_atresia_tracheoesophageal_fistula_phe_pi = TensorMap('phecode_750.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.11_date'), channel_map={'no_esophageal_atresia_tracheoesophageal_fistula':0, 'esophageal_atresia_tracheoesophageal_fistula_prevalent':1, 'esophageal_atresia_tracheoesophageal_fistula_incident':2}, loss=weighted_crossentropy([1.0, 14705, 20833], 'esophageal_atresia_tracheoesophageal_fistula_pi'))
esophageal_bleeding_varices_hemorrhage_phe_pi = TensorMap('phecode_530.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.2_date'), channel_map={'no_esophageal_bleeding_varices_hemorrhage':0, 'esophageal_bleeding_varices_hemorrhage_prevalent':1, 'esophageal_bleeding_varices_hemorrhage_incident':2}, loss=weighted_crossentropy([1.0, 574, 288], 'esophageal_bleeding_varices_hemorrhage_pi'))
esophagitis_gerd_and_related_diseases_phe_pi = TensorMap('phecode_530.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.1_date'), channel_map={'no_esophagitis_gerd_and_related_diseases':0, 'esophagitis_gerd_and_related_diseases_prevalent':1, 'esophagitis_gerd_and_related_diseases_incident':2}, loss=weighted_crossentropy([1.0, 83, 71], 'esophagitis_gerd_and_related_diseases_pi'))
essential_hypertension_phe_pi = TensorMap('phecode_401.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_401.1_date'), channel_map={'no_essential_hypertension':0, 'essential_hypertension_prevalent':1, 'essential_hypertension_incident':2}, loss=weighted_crossentropy([1.0, 12, 6], 'essential_hypertension_pi'))
essential_tremor_phe_pi = TensorMap('phecode_333.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_333.1_date'), channel_map={'no_essential_tremor':0, 'essential_tremor_prevalent':1, 'essential_tremor_incident':2}, loss=weighted_crossentropy([1.0, 12820, 1385], 'essential_tremor_pi'))
eustachian_tube_disorders_phe_pi = TensorMap('phecode_381.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_381.2_date'), channel_map={'no_eustachian_tube_disorders':0, 'eustachian_tube_disorders_prevalent':1, 'eustachian_tube_disorders_incident':2}, loss=weighted_crossentropy([1.0, 3759, 3875], 'eustachian_tube_disorders_pi'))
excessive_or_frequent_menstruation_phe_pi = TensorMap('phecode_626.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.12_date'), channel_map={'no_excessive_or_frequent_menstruation':0, 'excessive_or_frequent_menstruation_prevalent':1, 'excessive_or_frequent_menstruation_incident':2}, loss=weighted_crossentropy([1.0, 57, 194], 'excessive_or_frequent_menstruation_pi'))
excessive_vomiting_in_pregnancy_phe_pi = TensorMap('phecode_643',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_643_date'), channel_map={'no_excessive_vomiting_in_pregnancy':0, 'excessive_vomiting_in_pregnancy_prevalent':1, 'excessive_vomiting_in_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 3571, 166666], 'excessive_vomiting_in_pregnancy_pi'))
exophthalmos_phe_pi = TensorMap('phecode_242.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_242.3_date'), channel_map={'no_exophthalmos':0, 'exophthalmos_prevalent':1, 'exophthalmos_incident':2}, loss=weighted_crossentropy([1.0, 6493, 4950], 'exophthalmos_pi'))
extrapyramidal_disease_and_abnormal_movement_disorders_phe_pi = TensorMap('phecode_333',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_333_date'), channel_map={'no_extrapyramidal_disease_and_abnormal_movement_disorders':0, 'extrapyramidal_disease_and_abnormal_movement_disorders_prevalent':1, 'extrapyramidal_disease_and_abnormal_movement_disorders_incident':2}, loss=weighted_crossentropy([1.0, 768, 186], 'extrapyramidal_disease_and_abnormal_movement_disorders_pi'))
extrinsic_allergic_alveolitis_phe_pi = TensorMap('phecode_500.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_500.1_date'), channel_map={'no_extrinsic_allergic_alveolitis':0, 'extrinsic_allergic_alveolitis_prevalent':1, 'extrinsic_allergic_alveolitis_incident':2}, loss=weighted_crossentropy([1.0, 13888, 5681], 'extrinsic_allergic_alveolitis_pi'))
eye_infection_viral_phe_pi = TensorMap('phecode_369.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_369.2_date'), channel_map={'no_eye_infection_viral':0, 'eye_infection_viral_prevalent':1, 'eye_infection_viral_incident':2}, loss=weighted_crossentropy([1.0, 8620, 3649], 'eye_infection_viral_pi'))
facial_nerve_disorders_cn7_phe_pi = TensorMap('phecode_352.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_352.2_date'), channel_map={'no_facial_nerve_disorders_cn7':0, 'facial_nerve_disorders_cn7_prevalent':1, 'facial_nerve_disorders_cn7_incident':2}, loss=weighted_crossentropy([1.0, 994, 658], 'facial_nerve_disorders_cn7_pi'))
family_history_phe_pi = TensorMap('phecode_1100',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1100_date'), channel_map={'no_family_history':0, 'family_history_prevalent':1, 'family_history_incident':2}, loss=weighted_crossentropy([1.0, 546, 190], 'family_history_pi'))
fasciitis_phe_pi = TensorMap('phecode_728.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_728.7_date'), channel_map={'no_fasciitis':0, 'fasciitis_prevalent':1, 'fasciitis_incident':2}, loss=weighted_crossentropy([1.0, 4629, 2008], 'fasciitis_pi'))
femoral_hernia_phe_pi = TensorMap('phecode_550.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_550.3_date'), channel_map={'no_femoral_hernia':0, 'femoral_hernia_prevalent':1, 'femoral_hernia_incident':2}, loss=weighted_crossentropy([1.0, 1126, 1154], 'femoral_hernia_pi'))
fetal_distress_and_abnormal_forces_of_labor_phe_pi = TensorMap('phecode_661',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_661_date'), channel_map={'no_fetal_distress_and_abnormal_forces_of_labor':0, 'fetal_distress_and_abnormal_forces_of_labor_prevalent':1, 'fetal_distress_and_abnormal_forces_of_labor_incident':2}, loss=weighted_crossentropy([1.0, 178, 6578], 'fetal_distress_and_abnormal_forces_of_labor_pi'))
fever_of_unknown_origin_phe_pi = TensorMap('phecode_783',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_783_date'), channel_map={'no_fever_of_unknown_origin':0, 'fever_of_unknown_origin_prevalent':1, 'fever_of_unknown_origin_incident':2}, loss=weighted_crossentropy([1.0, 234, 130], 'fever_of_unknown_origin_pi'))
fibroadenosis_of_breast_phe_pi = TensorMap('phecode_610.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_610.2_date'), channel_map={'no_fibroadenosis_of_breast':0, 'fibroadenosis_of_breast_prevalent':1, 'fibroadenosis_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 1915, 4950], 'fibroadenosis_of_breast_pi'))
fibrosclerosis_of_breast_phe_pi = TensorMap('phecode_610.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_610.3_date'), channel_map={'no_fibrosclerosis_of_breast':0, 'fibrosclerosis_of_breast_prevalent':1, 'fibrosclerosis_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 1937, 4629], 'fibrosclerosis_of_breast_pi'))
first_degree_av_block_phe_pi = TensorMap('phecode_426.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.21_date'), channel_map={'no_first_degree_av_block':0, 'first_degree_av_block_prevalent':1, 'first_degree_av_block_incident':2}, loss=weighted_crossentropy([1.0, 1945, 328], 'first_degree_av_block_pi'))
flat_foot_phe_pi = TensorMap('phecode_735.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_735.1_date'), channel_map={'no_flat_foot':0, 'flat_foot_prevalent':1, 'flat_foot_incident':2}, loss=weighted_crossentropy([1.0, 12195, 2590], 'flat_foot_pi'))
flatulence_phe_pi = TensorMap('phecode_561.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_561.2_date'), channel_map={'no_flatulence':0, 'flatulence_prevalent':1, 'flatulence_incident':2}, loss=weighted_crossentropy([1.0, 544, 288], 'flatulence_pi'))
fluid_overload_phe_pi = TensorMap('phecode_276.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.6_date'), channel_map={'no_fluid_overload':0, 'fluid_overload_prevalent':1, 'fluid_overload_incident':2}, loss=weighted_crossentropy([1.0, 5681, 554], 'fluid_overload_pi'))
folate_deficiency_anemia_phe_pi = TensorMap('phecode_281.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_281.13_date'), channel_map={'no_folate_deficiency_anemia':0, 'folate_deficiency_anemia_prevalent':1, 'folate_deficiency_anemia_incident':2}, loss=weighted_crossentropy([1.0, 6172, 2173], 'folate_deficiency_anemia_pi'))
foreign_body_injury_phe_pi = TensorMap('phecode_1001',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1001_date'), channel_map={'no_foreign_body_injury':0, 'foreign_body_injury_prevalent':1, 'foreign_body_injury_incident':2}, loss=weighted_crossentropy([1.0, 796, 699], 'foreign_body_injury_pi'))
fracture_of_ankle_and_foot_phe_pi = TensorMap('phecode_801',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_801_date'), channel_map={'no_fracture_of_ankle_and_foot':0, 'fracture_of_ankle_and_foot_prevalent':1, 'fracture_of_ankle_and_foot_incident':2}, loss=weighted_crossentropy([1.0, 1322, 5882], 'fracture_of_ankle_and_foot_pi'))
fracture_of_clavicle_or_scapula_phe_pi = TensorMap('phecode_803.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_803.3_date'), channel_map={'no_fracture_of_clavicle_or_scapula':0, 'fracture_of_clavicle_or_scapula_prevalent':1, 'fracture_of_clavicle_or_scapula_incident':2}, loss=weighted_crossentropy([1.0, 2049, 5154], 'fracture_of_clavicle_or_scapula_pi'))
fracture_of_foot_phe_pi = TensorMap('phecode_801.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_801.1_date'), channel_map={'no_fracture_of_foot':0, 'fracture_of_foot_prevalent':1, 'fracture_of_foot_incident':2}, loss=weighted_crossentropy([1.0, 23809, 62500], 'fracture_of_foot_pi'))
fracture_of_hand_or_wrist_phe_pi = TensorMap('phecode_804',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_804_date'), channel_map={'no_fracture_of_hand_or_wrist':0, 'fracture_of_hand_or_wrist_prevalent':1, 'fracture_of_hand_or_wrist_incident':2}, loss=weighted_crossentropy([1.0, 924, 2762], 'fracture_of_hand_or_wrist_pi'))
fracture_of_humerus_phe_pi = TensorMap('phecode_803.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_803.1_date'), channel_map={'no_fracture_of_humerus':0, 'fracture_of_humerus_prevalent':1, 'fracture_of_humerus_incident':2}, loss=weighted_crossentropy([1.0, 2358, 5050], 'fracture_of_humerus_pi'))
fracture_of_lower_limb_phe_pi = TensorMap('phecode_800',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_800_date'), channel_map={'no_fracture_of_lower_limb':0, 'fracture_of_lower_limb_prevalent':1, 'fracture_of_lower_limb_incident':2}, loss=weighted_crossentropy([1.0, 1063, 4545], 'fracture_of_lower_limb_pi'))
fracture_of_neck_of_femur_phe_pi = TensorMap('phecode_800.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_800.1_date'), channel_map={'no_fracture_of_neck_of_femur':0, 'fracture_of_neck_of_femur_prevalent':1, 'fracture_of_neck_of_femur_incident':2}, loss=weighted_crossentropy([1.0, 2777, 4761], 'fracture_of_neck_of_femur_pi'))
fracture_of_patella_phe_pi = TensorMap('phecode_800.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_800.4_date'), channel_map={'no_fracture_of_patella':0, 'fracture_of_patella_prevalent':1, 'fracture_of_patella_incident':2}, loss=weighted_crossentropy([1.0, 3448, 25000], 'fracture_of_patella_pi'))
fracture_of_pelvis_phe_pi = TensorMap('phecode_802',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_802_date'), channel_map={'no_fracture_of_pelvis':0, 'fracture_of_pelvis_prevalent':1, 'fracture_of_pelvis_incident':2}, loss=weighted_crossentropy([1.0, 5319, 9803], 'fracture_of_pelvis_pi'))
fracture_of_radius_and_ulna_phe_pi = TensorMap('phecode_803.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_803.2_date'), channel_map={'no_fracture_of_radius_and_ulna':0, 'fracture_of_radius_and_ulna_prevalent':1, 'fracture_of_radius_and_ulna_incident':2}, loss=weighted_crossentropy([1.0, 679, 2049], 'fracture_of_radius_and_ulna_pi'))
fracture_of_ribs_phe_pi = TensorMap('phecode_807',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_807_date'), channel_map={'no_fracture_of_ribs':0, 'fracture_of_ribs_prevalent':1, 'fracture_of_ribs_incident':2}, loss=weighted_crossentropy([1.0, 1754, 5681], 'fracture_of_ribs_pi'))
fracture_of_tibia_and_fibula_phe_pi = TensorMap('phecode_800.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_800.3_date'), channel_map={'no_fracture_of_tibia_and_fibula':0, 'fracture_of_tibia_and_fibula_prevalent':1, 'fracture_of_tibia_and_fibula_incident':2}, loss=weighted_crossentropy([1.0, 1285, 5376], 'fracture_of_tibia_and_fibula_pi'))
fracture_of_unspecified_bones_phe_pi = TensorMap('phecode_809',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_809_date'), channel_map={'no_fracture_of_unspecified_bones':0, 'fracture_of_unspecified_bones_prevalent':1, 'fracture_of_unspecified_bones_incident':2}, loss=weighted_crossentropy([1.0, 4545, 9615], 'fracture_of_unspecified_bones_pi'))
fracture_of_unspecified_part_of_femur_phe_pi = TensorMap('phecode_800.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_800.2_date'), channel_map={'no_fracture_of_unspecified_part_of_femur':0, 'fracture_of_unspecified_part_of_femur_prevalent':1, 'fracture_of_unspecified_part_of_femur_incident':2}, loss=weighted_crossentropy([1.0, 5154, 13888], 'fracture_of_unspecified_part_of_femur_pi'))
fracture_of_upper_limb_phe_pi = TensorMap('phecode_803',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_803_date'), channel_map={'no_fracture_of_upper_limb':0, 'fracture_of_upper_limb_prevalent':1, 'fracture_of_upper_limb_incident':2}, loss=weighted_crossentropy([1.0, 2659, 4504], 'fracture_of_upper_limb_pi'))
fracture_of_vertebral_column_without_mention_of_spinal_cord_injury_phe_pi = TensorMap('phecode_805',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_805_date'), channel_map={'no_fracture_of_vertebral_column_without_mention_of_spinal_cord_injury':0, 'fracture_of_vertebral_column_without_mention_of_spinal_cord_injury_prevalent':1, 'fracture_of_vertebral_column_without_mention_of_spinal_cord_injury_incident':2}, loss=weighted_crossentropy([1.0, 1612, 2923], 'fracture_of_vertebral_column_without_mention_of_spinal_cord_injury_pi'))
frequency_of_urination_and_polyuria_phe_pi = TensorMap('phecode_599.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.5_date'), channel_map={'no_frequency_of_urination_and_polyuria':0, 'frequency_of_urination_and_polyuria_prevalent':1, 'frequency_of_urination_and_polyuria_incident':2}, loss=weighted_crossentropy([1.0, 177, 203], 'frequency_of_urination_and_polyuria_pi'))
functional_digestive_disorders_phe_pi = TensorMap('phecode_564',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_564_date'), channel_map={'no_functional_digestive_disorders':0, 'functional_digestive_disorders_prevalent':1, 'functional_digestive_disorders_incident':2}, loss=weighted_crossentropy([1.0, 927, 765], 'functional_digestive_disorders_pi'))
functional_disorders_of_bladder_phe_pi = TensorMap('phecode_596.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_596.5_date'), channel_map={'no_functional_disorders_of_bladder':0, 'functional_disorders_of_bladder_prevalent':1, 'functional_disorders_of_bladder_incident':2}, loss=weighted_crossentropy([1.0, 423, 552], 'functional_disorders_of_bladder_pi'))
galactorrhea_phe_pi = TensorMap('phecode_612.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_612.1_date'), channel_map={'no_galactorrhea':0, 'galactorrhea_prevalent':1, 'galactorrhea_incident':2}, loss=weighted_crossentropy([1.0, 38461, 250000], 'galactorrhea_pi'))
ganglion_and_cyst_of_synovium_tendon_and_bursa_phe_pi = TensorMap('phecode_727.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727.4_date'), channel_map={'no_ganglion_and_cyst_of_synovium_tendon_and_bursa':0, 'ganglion_and_cyst_of_synovium_tendon_and_bursa_prevalent':1, 'ganglion_and_cyst_of_synovium_tendon_and_bursa_incident':2}, loss=weighted_crossentropy([1.0, 243, 271], 'ganglion_and_cyst_of_synovium_tendon_and_bursa_pi'))
gangrene_phe_pi = TensorMap('phecode_791',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_791_date'), channel_map={'no_gangrene':0, 'gangrene_prevalent':1, 'gangrene_incident':2}, loss=weighted_crossentropy([1.0, 2369, 660], 'gangrene_pi'))
gastric_ulcer_phe_pi = TensorMap('phecode_531.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_531.2_date'), channel_map={'no_gastric_ulcer':0, 'gastric_ulcer_prevalent':1, 'gastric_ulcer_incident':2}, loss=weighted_crossentropy([1.0, 197, 141], 'gastric_ulcer_pi'))
gastritis_and_duodenitis_phe_pi = TensorMap('phecode_535',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_535_date'), channel_map={'no_gastritis_and_duodenitis':0, 'gastritis_and_duodenitis_prevalent':1, 'gastritis_and_duodenitis_incident':2}, loss=weighted_crossentropy([1.0, 50, 32], 'gastritis_and_duodenitis_pi'))
gastroesophageal_laceration_hemorrhage_syndrome_phe_pi = TensorMap('phecode_530.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.7_date'), channel_map={'no_gastroesophageal_laceration_hemorrhage_syndrome':0, 'gastroesophageal_laceration_hemorrhage_syndrome_prevalent':1, 'gastroesophageal_laceration_hemorrhage_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 2109, 2403], 'gastroesophageal_laceration_hemorrhage_syndrome_pi'))
gastrointestinal_complications_phe_pi = TensorMap('phecode_569.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_569.2_date'), channel_map={'no_gastrointestinal_complications':0, 'gastrointestinal_complications_prevalent':1, 'gastrointestinal_complications_incident':2}, loss=weighted_crossentropy([1.0, 2500, 950], 'gastrointestinal_complications_pi'))
gastrojejunal_ulcer_phe_pi = TensorMap('phecode_531.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_531.5_date'), channel_map={'no_gastrojejunal_ulcer':0, 'gastrojejunal_ulcer_prevalent':1, 'gastrojejunal_ulcer_incident':2}, loss=weighted_crossentropy([1.0, 27777, 10638], 'gastrojejunal_ulcer_pi'))
generalized_anxiety_disorder_phe_pi = TensorMap('phecode_300.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.11_date'), channel_map={'no_generalized_anxiety_disorder':0, 'generalized_anxiety_disorder_prevalent':1, 'generalized_anxiety_disorder_incident':2}, loss=weighted_crossentropy([1.0, 5434, 3731], 'generalized_anxiety_disorder_pi'))
generalized_convulsive_epilepsy_phe_pi = TensorMap('phecode_345.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_345.11_date'), channel_map={'no_generalized_convulsive_epilepsy':0, 'generalized_convulsive_epilepsy_prevalent':1, 'generalized_convulsive_epilepsy_incident':2}, loss=weighted_crossentropy([1.0, 2164, 1295], 'generalized_convulsive_epilepsy_pi'))
genital_prolapse_phe_pi = TensorMap('phecode_618',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_618_date'), channel_map={'no_genital_prolapse':0, 'genital_prolapse_prevalent':1, 'genital_prolapse_incident':2}, loss=weighted_crossentropy([1.0, 1831, 1700], 'genital_prolapse_pi'))
gerd_phe_pi = TensorMap('phecode_530.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.11_date'), channel_map={'no_gerd':0, 'gerd_prevalent':1, 'gerd_incident':2}, loss=weighted_crossentropy([1.0, 82, 28], 'gerd_pi'))
giant_cell_arteritis_phe_pi = TensorMap('phecode_446.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.5_date'), channel_map={'no_giant_cell_arteritis':0, 'giant_cell_arteritis_prevalent':1, 'giant_cell_arteritis_incident':2}, loss=weighted_crossentropy([1.0, 3597, 1043], 'giant_cell_arteritis_pi'))
gingival_and_periodontal_diseases_phe_pi = TensorMap('phecode_523',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_523_date'), channel_map={'no_gingival_and_periodontal_diseases':0, 'gingival_and_periodontal_diseases_prevalent':1, 'gingival_and_periodontal_diseases_incident':2}, loss=weighted_crossentropy([1.0, 1886, 1216], 'gingival_and_periodontal_diseases_pi'))
gingivitis_phe_pi = TensorMap('phecode_523.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_523.1_date'), channel_map={'no_gingivitis':0, 'gingivitis_prevalent':1, 'gingivitis_incident':2}, loss=weighted_crossentropy([1.0, 8771, 4807], 'gingivitis_pi'))
glaucoma_phe_pi = TensorMap('phecode_365',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_365_date'), channel_map={'no_glaucoma':0, 'glaucoma_prevalent':1, 'glaucoma_incident':2}, loss=weighted_crossentropy([1.0, 282, 87], 'glaucoma_pi'))
glossitis_phe_pi = TensorMap('phecode_529.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_529.1_date'), channel_map={'no_glossitis':0, 'glossitis_prevalent':1, 'glossitis_incident':2}, loss=weighted_crossentropy([1.0, 3597, 1655], 'glossitis_pi'))
glossodynia_phe_pi = TensorMap('phecode_529.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_529.6_date'), channel_map={'no_glossodynia':0, 'glossodynia_prevalent':1, 'glossodynia_incident':2}, loss=weighted_crossentropy([1.0, 29411, 21739], 'glossodynia_pi'))
glucocorticoid_deficiency_phe_pi = TensorMap('phecode_255.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_255.21_date'), channel_map={'no_glucocorticoid_deficiency':0, 'glucocorticoid_deficiency_prevalent':1, 'glucocorticoid_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 2325, 1483], 'glucocorticoid_deficiency_pi'))
glycosuria_or_acetonuria_phe_pi = TensorMap('phecode_250.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.5_date'), channel_map={'no_glycosuria_or_acetonuria':0, 'glycosuria_or_acetonuria_prevalent':1, 'glycosuria_or_acetonuria_incident':2}, loss=weighted_crossentropy([1.0, 3571, 17241], 'glycosuria_or_acetonuria_pi'))
gout_phe_pi = TensorMap('phecode_274.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_274.1_date'), channel_map={'no_gout':0, 'gout_prevalent':1, 'gout_incident':2}, loss=weighted_crossentropy([1.0, 1282, 177], 'gout_pi'))
gouty_arthropathy_phe_pi = TensorMap('phecode_274.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_274.11_date'), channel_map={'no_gouty_arthropathy':0, 'gouty_arthropathy_prevalent':1, 'gouty_arthropathy_incident':2}, loss=weighted_crossentropy([1.0, 41666, 16129], 'gouty_arthropathy_pi'))
gram_negative_septicemia_phe_pi = TensorMap('phecode_38.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_38.1_date'), channel_map={'no_gram_negative_septicemia':0, 'gram_negative_septicemia_prevalent':1, 'gram_negative_septicemia_incident':2}, loss=weighted_crossentropy([1.0, 1655, 456], 'gram_negative_septicemia_pi'))
gram_positive_septicemia_phe_pi = TensorMap('phecode_38.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_38.2_date'), channel_map={'no_gram_positive_septicemia':0, 'gram_positive_septicemia_prevalent':1, 'gram_positive_septicemia_incident':2}, loss=weighted_crossentropy([1.0, 2212, 946], 'gram_positive_septicemia_pi'))
graves_disease_phe_pi = TensorMap('phecode_242.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_242.1_date'), channel_map={'no_graves_disease':0, 'graves_disease_prevalent':1, 'graves_disease_incident':2}, loss=weighted_crossentropy([1.0, 1792, 1234], 'graves_disease_pi'))
hallucinations_phe_pi = TensorMap('phecode_292.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292.6_date'), channel_map={'no_hallucinations':0, 'hallucinations_prevalent':1, 'hallucinations_incident':2}, loss=weighted_crossentropy([1.0, 5882, 1098], 'hallucinations_pi'))
hallux_rigidus_phe_pi = TensorMap('phecode_735.23',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_735.23_date'), channel_map={'no_hallux_rigidus':0, 'hallux_rigidus_prevalent':1, 'hallux_rigidus_incident':2}, loss=weighted_crossentropy([1.0, 560, 438], 'hallux_rigidus_pi'))
hallux_valgus_bunion_phe_pi = TensorMap('phecode_735.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_735.3_date'), channel_map={'no_hallux_valgus_bunion':0, 'hallux_valgus_bunion_prevalent':1, 'hallux_valgus_bunion_incident':2}, loss=weighted_crossentropy([1.0, 115, 107], 'hallux_valgus_bunion_pi'))
hammer_toe_acquired_phe_pi = TensorMap('phecode_735.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_735.21_date'), channel_map={'no_hammer_toe_acquired':0, 'hammer_toe_acquired_prevalent':1, 'hammer_toe_acquired_incident':2}, loss=weighted_crossentropy([1.0, 488, 303], 'hammer_toe_acquired_pi'))
hearing_loss_phe_pi = TensorMap('phecode_389',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_389_date'), channel_map={'no_hearing_loss':0, 'hearing_loss_prevalent':1, 'hearing_loss_incident':2}, loss=weighted_crossentropy([1.0, 454, 95], 'hearing_loss_pi'))
heart_failure_nos_phe_pi = TensorMap('phecode_428.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_428.2_date'), channel_map={'no_heart_failure_nos':0, 'heart_failure_nos_prevalent':1, 'heart_failure_nos_incident':2}, loss=weighted_crossentropy([1.0, 277, 95], 'heart_failure_nos_pi'))
heart_transplant_surgery_phe_pi = TensorMap('phecode_429.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_429.1_date'), channel_map={'no_heart_transplant_surgery':0, 'heart_transplant_surgery_prevalent':1, 'heart_transplant_surgery_incident':2}, loss=weighted_crossentropy([1.0, 8771, 19230], 'heart_transplant_surgery_pi'))
heart_valve_replaced_phe_pi = TensorMap('phecode_395.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_395.6_date'), channel_map={'no_heart_valve_replaced':0, 'heart_valve_replaced_prevalent':1, 'heart_valve_replaced_incident':2}, loss=weighted_crossentropy([1.0, 673, 319], 'heart_valve_replaced_pi'))
heartburn_phe_pi = TensorMap('phecode_530.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.9_date'), channel_map={'no_heartburn':0, 'heartburn_prevalent':1, 'heartburn_incident':2}, loss=weighted_crossentropy([1.0, 359, 321], 'heartburn_pi'))
helminthiases_phe_pi = TensorMap('phecode_134',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_134_date'), channel_map={'no_helminthiases':0, 'helminthiases_prevalent':1, 'helminthiases_incident':2}, loss=weighted_crossentropy([1.0, 6849, 3703], 'helminthiases_pi'))
hemangioma_and_lymphangioma_any_site_phe_pi = TensorMap('phecode_228',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_228_date'), channel_map={'no_hemangioma_and_lymphangioma_any_site':0, 'hemangioma_and_lymphangioma_any_site_prevalent':1, 'hemangioma_and_lymphangioma_any_site_incident':2}, loss=weighted_crossentropy([1.0, 484, 446], 'hemangioma_and_lymphangioma_any_site_pi'))
hemarthrosis_phe_pi = TensorMap('phecode_741.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741.5_date'), channel_map={'no_hemarthrosis':0, 'hemarthrosis_prevalent':1, 'hemarthrosis_incident':2}, loss=weighted_crossentropy([1.0, 35714, 71428], 'hemarthrosis_pi'))
hematemesis_phe_pi = TensorMap('phecode_578.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_578.1_date'), channel_map={'no_hematemesis':0, 'hematemesis_prevalent':1, 'hematemesis_incident':2}, loss=weighted_crossentropy([1.0, 394, 342], 'hematemesis_pi'))
hematuria_phe_pi = TensorMap('phecode_593',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_593_date'), channel_map={'no_hematuria':0, 'hematuria_prevalent':1, 'hematuria_incident':2}, loss=weighted_crossentropy([1.0, 47, 45], 'hematuria_pi'))
hemiplegia_phe_pi = TensorMap('phecode_342',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_342_date'), channel_map={'no_hemiplegia':0, 'hemiplegia_prevalent':1, 'hemiplegia_incident':2}, loss=weighted_crossentropy([1.0, 556, 322], 'hemiplegia_pi'))
hemoglobinuria_phe_pi = TensorMap('phecode_285.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_285.8_date'), channel_map={'no_hemoglobinuria':0, 'hemoglobinuria_prevalent':1, 'hemoglobinuria_incident':2}, loss=weighted_crossentropy([1.0, 41666, 41666], 'hemoglobinuria_pi'))
hemolytic_uremic_syndrome_phe_pi = TensorMap('phecode_283.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_283.21_date'), channel_map={'no_hemolytic_uremic_syndrome':0, 'hemolytic_uremic_syndrome_prevalent':1, 'hemolytic_uremic_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 83333, 83333], 'hemolytic_uremic_syndrome_pi'))
hemoptysis_phe_pi = TensorMap('phecode_516.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_516.1_date'), channel_map={'no_hemoptysis':0, 'hemoptysis_prevalent':1, 'hemoptysis_incident':2}, loss=weighted_crossentropy([1.0, 380, 295], 'hemoptysis_pi'))
hemorrhage_from_gastrointestinal_ulcer_phe_pi = TensorMap('phecode_531.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_531.1_date'), channel_map={'no_hemorrhage_from_gastrointestinal_ulcer':0, 'hemorrhage_from_gastrointestinal_ulcer_prevalent':1, 'hemorrhage_from_gastrointestinal_ulcer_incident':2}, loss=weighted_crossentropy([1.0, 1213, 994], 'hemorrhage_from_gastrointestinal_ulcer_pi'))
hemorrhage_in_early_pregnancy_phe_pi = TensorMap('phecode_636.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_636.3_date'), channel_map={'no_hemorrhage_in_early_pregnancy':0, 'hemorrhage_in_early_pregnancy_prevalent':1, 'hemorrhage_in_early_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 127, 3968], 'hemorrhage_in_early_pregnancy_pi'))
hemorrhage_nos_phe_pi = TensorMap('phecode_459.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_459.1_date'), channel_map={'no_hemorrhage_nos':0, 'hemorrhage_nos_prevalent':1, 'hemorrhage_nos_incident':2}, loss=weighted_crossentropy([1.0, 11627, 5434], 'hemorrhage_nos_pi'))
hemorrhage_of_gastrointestinal_tract_phe_pi = TensorMap('phecode_578.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_578.9_date'), channel_map={'no_hemorrhage_of_gastrointestinal_tract':0, 'hemorrhage_of_gastrointestinal_tract_prevalent':1, 'hemorrhage_of_gastrointestinal_tract_incident':2}, loss=weighted_crossentropy([1.0, 523, 61], 'hemorrhage_of_gastrointestinal_tract_pi'))
hemorrhage_of_rectum_and_anus_phe_pi = TensorMap('phecode_578.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_578.8_date'), channel_map={'no_hemorrhage_of_rectum_and_anus':0, 'hemorrhage_of_rectum_and_anus_prevalent':1, 'hemorrhage_of_rectum_and_anus_incident':2}, loss=weighted_crossentropy([1.0, 46, 73], 'hemorrhage_of_rectum_and_anus_pi'))
hemorrhage_or_hematoma_complicating_a_procedure_phe_pi = TensorMap('phecode_850',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_850_date'), channel_map={'no_hemorrhage_or_hematoma_complicating_a_procedure':0, 'hemorrhage_or_hematoma_complicating_a_procedure_prevalent':1, 'hemorrhage_or_hematoma_complicating_a_procedure_incident':2}, loss=weighted_crossentropy([1.0, 147, 122], 'hemorrhage_or_hematoma_complicating_a_procedure_pi'))
hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants_phe_pi = TensorMap('phecode_286.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.5_date'), channel_map={'no_hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants':0, 'hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants_prevalent':1, 'hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants_incident':2}, loss=weighted_crossentropy([1.0, 7142, 10638], 'hemorrhagic_disorder_due_to_intrinsic_circulating_anticoagulants_pi'))
hemorrhoids_phe_pi = TensorMap('phecode_455',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_455_date'), channel_map={'no_hemorrhoids':0, 'hemorrhoids_prevalent':1, 'hemorrhoids_incident':2}, loss=weighted_crossentropy([1.0, 34, 28], 'hemorrhoids_pi'))
hepatitis_nos_phe_pi = TensorMap('phecode_70.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_70.9_date'), channel_map={'no_hepatitis_nos':0, 'hepatitis_nos_prevalent':1, 'hepatitis_nos_incident':2}, loss=weighted_crossentropy([1.0, 2538, 976], 'hepatitis_nos_pi'))
hepatomegaly_phe_pi = TensorMap('phecode_573.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.3_date'), channel_map={'no_hepatomegaly':0, 'hepatomegaly_prevalent':1, 'hepatomegaly_incident':2}, loss=weighted_crossentropy([1.0, 3355, 1597], 'hepatomegaly_pi'))
hereditary_and_idiopathic_peripheral_neuropathy_phe_pi = TensorMap('phecode_356',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_356_date'), channel_map={'no_hereditary_and_idiopathic_peripheral_neuropathy':0, 'hereditary_and_idiopathic_peripheral_neuropathy_prevalent':1, 'hereditary_and_idiopathic_peripheral_neuropathy_incident':2}, loss=weighted_crossentropy([1.0, 5747, 3816], 'hereditary_and_idiopathic_peripheral_neuropathy_pi'))
hereditary_disturbances_in_tooth_structure_phe_pi = TensorMap('phecode_520.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_520.1_date'), channel_map={'no_hereditary_disturbances_in_tooth_structure':0, 'hereditary_disturbances_in_tooth_structure_prevalent':1, 'hereditary_disturbances_in_tooth_structure_incident':2}, loss=weighted_crossentropy([1.0, 14285, 20833], 'hereditary_disturbances_in_tooth_structure_pi'))
hereditary_retinal_dystrophies_phe_pi = TensorMap('phecode_362.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.7_date'), channel_map={'no_hereditary_retinal_dystrophies':0, 'hereditary_retinal_dystrophies_prevalent':1, 'hereditary_retinal_dystrophies_incident':2}, loss=weighted_crossentropy([1.0, 10204, 6493], 'hereditary_retinal_dystrophies_pi'))
herpes_simplex_phe_pi = TensorMap('phecode_54',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_54_date'), channel_map={'no_herpes_simplex':0, 'herpes_simplex_prevalent':1, 'herpes_simplex_incident':2}, loss=weighted_crossentropy([1.0, 3496, 2487], 'herpes_simplex_pi'))
herpes_zoster_phe_pi = TensorMap('phecode_53',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_53_date'), channel_map={'no_herpes_zoster':0, 'herpes_zoster_prevalent':1, 'herpes_zoster_incident':2}, loss=weighted_crossentropy([1.0, 2976, 1377], 'herpes_zoster_pi'))
herpes_zoster_with_nervous_system_complications_phe_pi = TensorMap('phecode_53.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_53.1_date'), channel_map={'no_herpes_zoster_with_nervous_system_complications':0, 'herpes_zoster_with_nervous_system_complications_prevalent':1, 'herpes_zoster_with_nervous_system_complications_incident':2}, loss=weighted_crossentropy([1.0, 8771, 5208], 'herpes_zoster_with_nervous_system_complications_pi'))
hidradenitis_phe_pi = TensorMap('phecode_705.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_705.3_date'), channel_map={'no_hidradenitis':0, 'hidradenitis_prevalent':1, 'hidradenitis_incident':2}, loss=weighted_crossentropy([1.0, 2857, 6756], 'hidradenitis_pi'))
hirsutism_phe_pi = TensorMap('phecode_704.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_704.2_date'), channel_map={'no_hirsutism':0, 'hirsutism_prevalent':1, 'hirsutism_incident':2}, loss=weighted_crossentropy([1.0, 17857, 41666], 'hirsutism_pi'))
histoplasmosis_phe_pi = TensorMap('phecode_117.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_117.1_date'), channel_map={'no_histoplasmosis':0, 'histoplasmosis_prevalent':1, 'histoplasmosis_incident':2}, loss=weighted_crossentropy([1.0, 166666, 100000], 'histoplasmosis_pi'))
hiv_infection_symptomatic_phe_pi = TensorMap('phecode_71.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_71.1_date'), channel_map={'no_hiv_infection_symptomatic':0, 'hiv_infection_symptomatic_prevalent':1, 'hiv_infection_symptomatic_incident':2}, loss=weighted_crossentropy([1.0, 2439, 50000], 'hiv_infection_symptomatic_pi'))
hodgkins_disease_phe_pi = TensorMap('phecode_201',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_201_date'), channel_map={'no_hodgkins_disease':0, 'hodgkins_disease_prevalent':1, 'hodgkins_disease_incident':2}, loss=weighted_crossentropy([1.0, 2717, 2747], 'hodgkins_disease_pi'))
hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use_phe_pi = TensorMap('phecode_962.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_962.3_date'), channel_map={'no_hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use':0, 'hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use_prevalent':1, 'hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use_incident':2}, loss=weighted_crossentropy([1.0, 2525, 1497], 'hormones_and_synthetic_substitutes_causing_adverse_effects_in_therapeutic_use_pi'))
human_immunodeficiency_virus_hiv_disease_phe_pi = TensorMap('phecode_71',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_71_date'), channel_map={'no_human_immunodeficiency_virus_hiv_disease':0, 'human_immunodeficiency_virus_hiv_disease_prevalent':1, 'human_immunodeficiency_virus_hiv_disease_incident':2}, loss=weighted_crossentropy([1.0, 4545, 62500], 'human_immunodeficiency_virus_hiv_disease_pi'))
hydrocele_phe_pi = TensorMap('phecode_603.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_603.1_date'), channel_map={'no_hydrocele':0, 'hydrocele_prevalent':1, 'hydrocele_incident':2}, loss=weighted_crossentropy([1.0, 570, 506], 'hydrocele_pi'))
hydrocephalus_phe_pi = TensorMap('phecode_331.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_331.1_date'), channel_map={'no_hydrocephalus':0, 'hydrocephalus_prevalent':1, 'hydrocephalus_incident':2}, loss=weighted_crossentropy([1.0, 2252, 919], 'hydrocephalus_pi'))
hydronephrosis_phe_pi = TensorMap('phecode_595',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_595_date'), channel_map={'no_hydronephrosis':0, 'hydronephrosis_prevalent':1, 'hydronephrosis_incident':2}, loss=weighted_crossentropy([1.0, 668, 214], 'hydronephrosis_pi'))
hyperaldosteronism_phe_pi = TensorMap('phecode_255.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_255.12_date'), channel_map={'no_hyperaldosteronism':0, 'hyperaldosteronism_prevalent':1, 'hyperaldosteronism_incident':2}, loss=weighted_crossentropy([1.0, 8064, 8620], 'hyperaldosteronism_pi'))
hypercholesterolemia_phe_pi = TensorMap('phecode_272.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_272.11_date'), channel_map={'no_hypercholesterolemia':0, 'hypercholesterolemia_prevalent':1, 'hypercholesterolemia_incident':2}, loss=weighted_crossentropy([1.0, 31, 14], 'hypercholesterolemia_pi'))
hyperchylomicronemia_phe_pi = TensorMap('phecode_272.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_272.14_date'), channel_map={'no_hyperchylomicronemia':0, 'hyperchylomicronemia_prevalent':1, 'hyperchylomicronemia_incident':2}, loss=weighted_crossentropy([1.0, 83333, 83333], 'hyperchylomicronemia_pi'))
hyperemesis_gravidarum_phe_pi = TensorMap('phecode_643.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_643.1_date'), channel_map={'no_hyperemesis_gravidarum':0, 'hyperemesis_gravidarum_prevalent':1, 'hyperemesis_gravidarum_incident':2}, loss=weighted_crossentropy([1.0, 2590, 62500], 'hyperemesis_gravidarum_pi'))
hyperglyceridemia_phe_pi = TensorMap('phecode_272.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_272.12_date'), channel_map={'no_hyperglyceridemia':0, 'hyperglyceridemia_prevalent':1, 'hyperglyceridemia_incident':2}, loss=weighted_crossentropy([1.0, 14285, 7575], 'hyperglyceridemia_pi'))
hyperhidrosis_phe_pi = TensorMap('phecode_705.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_705.8_date'), channel_map={'no_hyperhidrosis':0, 'hyperhidrosis_prevalent':1, 'hyperhidrosis_incident':2}, loss=weighted_crossentropy([1.0, 1623, 840], 'hyperhidrosis_pi'))
hyperlipidemia_phe_pi = TensorMap('phecode_272.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_272.1_date'), channel_map={'no_hyperlipidemia':0, 'hyperlipidemia_prevalent':1, 'hyperlipidemia_incident':2}, loss=weighted_crossentropy([1.0, 200, 93], 'hyperlipidemia_pi'))
hypermetropia_phe_pi = TensorMap('phecode_367.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_367.8_date'), channel_map={'no_hypermetropia':0, 'hypermetropia_prevalent':1, 'hypermetropia_incident':2}, loss=weighted_crossentropy([1.0, 5050, 1666], 'hypermetropia_pi'))
hyperosmolality_and_or_hypernatremia_phe_pi = TensorMap('phecode_276.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.11_date'), channel_map={'no_hyperosmolality_and_or_hypernatremia':0, 'hyperosmolality_and_or_hypernatremia_prevalent':1, 'hyperosmolality_and_or_hypernatremia_incident':2}, loss=weighted_crossentropy([1.0, 8474, 1126], 'hyperosmolality_and_or_hypernatremia_pi'))
hyperparathyroidism_phe_pi = TensorMap('phecode_252.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_252.1_date'), channel_map={'no_hyperparathyroidism':0, 'hyperparathyroidism_prevalent':1, 'hyperparathyroidism_incident':2}, loss=weighted_crossentropy([1.0, 1547, 545], 'hyperparathyroidism_pi'))
hyperplasia_of_prostate_phe_pi = TensorMap('phecode_600',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_600_date'), channel_map={'no_hyperplasia_of_prostate':0, 'hyperplasia_of_prostate_prevalent':1, 'hyperplasia_of_prostate_incident':2}, loss=weighted_crossentropy([1.0, 95, 44], 'hyperplasia_of_prostate_pi'))
hyperpotassemia_phe_pi = TensorMap('phecode_276.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.13_date'), channel_map={'no_hyperpotassemia':0, 'hyperpotassemia_prevalent':1, 'hyperpotassemia_incident':2}, loss=weighted_crossentropy([1.0, 2024, 324], 'hyperpotassemia_pi'))
hypersensitivity_angiitis_phe_pi = TensorMap('phecode_446.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.3_date'), channel_map={'no_hypersensitivity_angiitis':0, 'hypersensitivity_angiitis_prevalent':1, 'hypersensitivity_angiitis_incident':2}, loss=weighted_crossentropy([1.0, 35714, 17857], 'hypersensitivity_angiitis_pi'))
hypersomnia_phe_pi = TensorMap('phecode_327.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327.1_date'), channel_map={'no_hypersomnia':0, 'hypersomnia_prevalent':1, 'hypersomnia_incident':2}, loss=weighted_crossentropy([1.0, 250000, 250000], 'hypersomnia_pi'))
hypertension_phe_pi = TensorMap('phecode_401',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_401_date'), channel_map={'no_hypertension':0, 'hypertension_prevalent':1, 'hypertension_incident':2}, loss=weighted_crossentropy([1.0, 12, 6], 'hypertension_pi'))
hypertension_complicating_pregnancy_childbirth_and_the_puerperium_phe_pi = TensorMap('phecode_642',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_642_date'), channel_map={'no_hypertension_complicating_pregnancy_childbirth_and_the_puerperium':0, 'hypertension_complicating_pregnancy_childbirth_and_the_puerperium_prevalent':1, 'hypertension_complicating_pregnancy_childbirth_and_the_puerperium_incident':2}, loss=weighted_crossentropy([1.0, 304, 6410], 'hypertension_complicating_pregnancy_childbirth_and_the_puerperium_pi'))
hypertensive_chronic_kidney_disease_phe_pi = TensorMap('phecode_401.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_401.22_date'), channel_map={'no_hypertensive_chronic_kidney_disease':0, 'hypertensive_chronic_kidney_disease_prevalent':1, 'hypertensive_chronic_kidney_disease_incident':2}, loss=weighted_crossentropy([1.0, 630, 390], 'hypertensive_chronic_kidney_disease_pi'))
hypertensive_heart_and_or_renal_disease_phe_pi = TensorMap('phecode_401.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_401.2_date'), channel_map={'no_hypertensive_heart_and_or_renal_disease':0, 'hypertensive_heart_and_or_renal_disease_prevalent':1, 'hypertensive_heart_and_or_renal_disease_incident':2}, loss=weighted_crossentropy([1.0, 41666, 10638], 'hypertensive_heart_and_or_renal_disease_pi'))
hypertensive_heart_disease_phe_pi = TensorMap('phecode_401.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_401.21_date'), channel_map={'no_hypertensive_heart_disease':0, 'hypertensive_heart_disease_prevalent':1, 'hypertensive_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 4464, 3067], 'hypertensive_heart_disease_pi'))
hypertrophic_obstructive_cardiomyopathy_phe_pi = TensorMap('phecode_425.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_425.11_date'), channel_map={'no_hypertrophic_obstructive_cardiomyopathy':0, 'hypertrophic_obstructive_cardiomyopathy_prevalent':1, 'hypertrophic_obstructive_cardiomyopathy_incident':2}, loss=weighted_crossentropy([1.0, 8333, 3205], 'hypertrophic_obstructive_cardiomyopathy_pi'))
hypertrophy_of_breast_gynecomastia_phe_pi = TensorMap('phecode_612.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_612.2_date'), channel_map={'no_hypertrophy_of_breast_gynecomastia':0, 'hypertrophy_of_breast_gynecomastia_prevalent':1, 'hypertrophy_of_breast_gynecomastia_incident':2}, loss=weighted_crossentropy([1.0, 624, 1937], 'hypertrophy_of_breast_gynecomastia_pi'))
hypertrophy_of_female_genital_organs_phe_pi = TensorMap('phecode_623',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_623_date'), channel_map={'no_hypertrophy_of_female_genital_organs':0, 'hypertrophy_of_female_genital_organs_prevalent':1, 'hypertrophy_of_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 463, 877], 'hypertrophy_of_female_genital_organs_pi'))
hypertrophy_of_salivary_gland_phe_pi = TensorMap('phecode_527.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_527.1_date'), channel_map={'no_hypertrophy_of_salivary_gland':0, 'hypertrophy_of_salivary_gland_prevalent':1, 'hypertrophy_of_salivary_gland_incident':2}, loss=weighted_crossentropy([1.0, 6944, 16666], 'hypertrophy_of_salivary_gland_pi'))
hyperventilation_phe_pi = TensorMap('phecode_513.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_513.4_date'), channel_map={'no_hyperventilation':0, 'hyperventilation_prevalent':1, 'hyperventilation_incident':2}, loss=weighted_crossentropy([1.0, 2262, 3448], 'hyperventilation_pi'))
hypoglycemia_phe_pi = TensorMap('phecode_251.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_251.1_date'), channel_map={'no_hypoglycemia':0, 'hypoglycemia_prevalent':1, 'hypoglycemia_incident':2}, loss=weighted_crossentropy([1.0, 1298, 396], 'hypoglycemia_pi'))
hypoparathyroidism_phe_pi = TensorMap('phecode_252.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_252.2_date'), channel_map={'no_hypoparathyroidism':0, 'hypoparathyroidism_prevalent':1, 'hypoparathyroidism_incident':2}, loss=weighted_crossentropy([1.0, 8928, 2564], 'hypoparathyroidism_pi'))
hypopotassemia_phe_pi = TensorMap('phecode_276.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.14_date'), channel_map={'no_hypopotassemia':0, 'hypopotassemia_prevalent':1, 'hypopotassemia_incident':2}, loss=weighted_crossentropy([1.0, 1315, 213], 'hypopotassemia_pi'))
hyposmolality_and_or_hyponatremia_phe_pi = TensorMap('phecode_276.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.12_date'), channel_map={'no_hyposmolality_and_or_hyponatremia':0, 'hyposmolality_and_or_hyponatremia_prevalent':1, 'hyposmolality_and_or_hyponatremia_incident':2}, loss=weighted_crossentropy([1.0, 1201, 162], 'hyposmolality_and_or_hyponatremia_pi'))
hypotension_phe_pi = TensorMap('phecode_458',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_458_date'), channel_map={'no_hypotension':0, 'hypotension_prevalent':1, 'hypotension_incident':2}, loss=weighted_crossentropy([1.0, 1488, 356], 'hypotension_pi'))
hypotension_nos_phe_pi = TensorMap('phecode_458.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_458.9_date'), channel_map={'no_hypotension_nos':0, 'hypotension_nos_prevalent':1, 'hypotension_nos_incident':2}, loss=weighted_crossentropy([1.0, 458, 100], 'hypotension_nos_pi'))
hypothermia_chills_phe_pi = TensorMap('phecode_780',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_780_date'), channel_map={'no_hypothermia_chills':0, 'hypothermia_chills_prevalent':1, 'hypothermia_chills_incident':2}, loss=weighted_crossentropy([1.0, 13513, 1811], 'hypothermia_chills_pi'))
hypothyroidism_nos_phe_pi = TensorMap('phecode_244.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_244.4_date'), channel_map={'no_hypothyroidism_nos':0, 'hypothyroidism_nos_prevalent':1, 'hypothyroidism_nos_incident':2}, loss=weighted_crossentropy([1.0, 74, 35], 'hypothyroidism_nos_pi'))
hypotony_of_eye_phe_pi = TensorMap('phecode_360.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_360.3_date'), channel_map={'no_hypotony_of_eye':0, 'hypotony_of_eye_prevalent':1, 'hypotony_of_eye_incident':2}, loss=weighted_crossentropy([1.0, 41666, 15151], 'hypotony_of_eye_pi'))
hypoventilation_phe_pi = TensorMap('phecode_513.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_513.3_date'), channel_map={'no_hypoventilation':0, 'hypoventilation_prevalent':1, 'hypoventilation_incident':2}, loss=weighted_crossentropy([1.0, 18518, 3703], 'hypoventilation_pi'))
hypovolemia_phe_pi = TensorMap('phecode_276.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.5_date'), channel_map={'no_hypovolemia':0, 'hypovolemia_prevalent':1, 'hypovolemia_incident':2}, loss=weighted_crossentropy([1.0, 691, 117], 'hypovolemia_pi'))
iatrogenic_endocrine_disorders_phe_pi = TensorMap('phecode_258',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_258_date'), channel_map={'no_iatrogenic_endocrine_disorders':0, 'iatrogenic_endocrine_disorders_prevalent':1, 'iatrogenic_endocrine_disorders_incident':2}, loss=weighted_crossentropy([1.0, 7692, 5494], 'iatrogenic_endocrine_disorders_pi'))
iatrogenic_hypotension_phe_pi = TensorMap('phecode_458.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_458.2_date'), channel_map={'no_iatrogenic_hypotension':0, 'iatrogenic_hypotension_prevalent':1, 'iatrogenic_hypotension_incident':2}, loss=weighted_crossentropy([1.0, 5747, 1805], 'iatrogenic_hypotension_pi'))
ichthyosis_congenita_phe_pi = TensorMap('phecode_691.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_691.1_date'), channel_map={'no_ichthyosis_congenita':0, 'ichthyosis_congenita_prevalent':1, 'ichthyosis_congenita_incident':2}, loss=weighted_crossentropy([1.0, 100000, 38461], 'ichthyosis_congenita_pi'))
ileostomy_status_phe_pi = TensorMap('phecode_559',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_559_date'), channel_map={'no_ileostomy_status':0, 'ileostomy_status_prevalent':1, 'ileostomy_status_incident':2}, loss=weighted_crossentropy([1.0, 532, 361], 'ileostomy_status_pi'))
immunity_deficiency_phe_pi = TensorMap('phecode_279.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_279.1_date'), channel_map={'no_immunity_deficiency':0, 'immunity_deficiency_prevalent':1, 'immunity_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 11904, 5102], 'immunity_deficiency_pi'))
impacted_cerumen_phe_pi = TensorMap('phecode_380.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_380.4_date'), channel_map={'no_impacted_cerumen':0, 'impacted_cerumen_prevalent':1, 'impacted_cerumen_incident':2}, loss=weighted_crossentropy([1.0, 2475, 1886], 'impacted_cerumen_pi'))
impaction_of_intestine_phe_pi = TensorMap('phecode_560.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_560.2_date'), channel_map={'no_impaction_of_intestine':0, 'impaction_of_intestine_prevalent':1, 'impaction_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 4587, 1552], 'impaction_of_intestine_pi'))
impaired_fasting_glucose_phe_pi = TensorMap('phecode_250.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.41_date'), channel_map={'no_impaired_fasting_glucose':0, 'impaired_fasting_glucose_prevalent':1, 'impaired_fasting_glucose_incident':2}, loss=weighted_crossentropy([1.0, 2808, 1683], 'impaired_fasting_glucose_pi'))
impetigo_phe_pi = TensorMap('phecode_686.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_686.2_date'), channel_map={'no_impetigo':0, 'impetigo_prevalent':1, 'impetigo_incident':2}, loss=weighted_crossentropy([1.0, 20000, 20000], 'impetigo_pi'))
impulse_control_disorder_phe_pi = TensorMap('phecode_312.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_312.3_date'), channel_map={'no_impulse_control_disorder':0, 'impulse_control_disorder_prevalent':1, 'impulse_control_disorder_incident':2}, loss=weighted_crossentropy([1.0, 71428, 71428], 'impulse_control_disorder_pi'))
infantile_cerebral_palsy_phe_pi = TensorMap('phecode_343',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_343_date'), channel_map={'no_infantile_cerebral_palsy':0, 'infantile_cerebral_palsy_prevalent':1, 'infantile_cerebral_palsy_incident':2}, loss=weighted_crossentropy([1.0, 4587, 5319], 'infantile_cerebral_palsy_pi'))
infection_inflammation_of_internal_prosthetic_device_implant_and_graft_phe_pi = TensorMap('phecode_81',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_81_date'), channel_map={'no_infection_inflammation_of_internal_prosthetic_device_implant_and_graft':0, 'infection_inflammation_of_internal_prosthetic_device_implant_and_graft_prevalent':1, 'infection_inflammation_of_internal_prosthetic_device_implant_and_graft_incident':2}, loss=weighted_crossentropy([1.0, 664, 510], 'infection_inflammation_of_internal_prosthetic_device_implant_and_graft_pi'))
infection_of_the_eye_phe_pi = TensorMap('phecode_369',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_369_date'), channel_map={'no_infection_of_the_eye':0, 'infection_of_the_eye_prevalent':1, 'infection_of_the_eye_incident':2}, loss=weighted_crossentropy([1.0, 2906, 2403], 'infection_of_the_eye_pi'))
infections_of_genitourinary_tract_during_pregnancy_phe_pi = TensorMap('phecode_647.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_647.1_date'), channel_map={'no_infections_of_genitourinary_tract_during_pregnancy':0, 'infections_of_genitourinary_tract_during_pregnancy_prevalent':1, 'infections_of_genitourinary_tract_during_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 1358, 41666], 'infections_of_genitourinary_tract_during_pregnancy_pi'))
infections_specific_to_the_perinatal_period_phe_pi = TensorMap('phecode_657',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_657_date'), channel_map={'no_infections_specific_to_the_perinatal_period':0, 'infections_specific_to_the_perinatal_period_prevalent':1, 'infections_specific_to_the_perinatal_period_incident':2}, loss=weighted_crossentropy([1.0, 125000, 166666], 'infections_specific_to_the_perinatal_period_pi'))
infectious_and_parasitic_complications_affecting_pregnancy_phe_pi = TensorMap('phecode_647',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_647_date'), channel_map={'no_infectious_and_parasitic_complications_affecting_pregnancy':0, 'infectious_and_parasitic_complications_affecting_pregnancy_prevalent':1, 'infectious_and_parasitic_complications_affecting_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 4098, 125000], 'infectious_and_parasitic_complications_affecting_pregnancy_pi'))
infectious_mononucleosis_phe_pi = TensorMap('phecode_79.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_79.2_date'), channel_map={'no_infectious_mononucleosis':0, 'infectious_mononucleosis_prevalent':1, 'infectious_mononucleosis_incident':2}, loss=weighted_crossentropy([1.0, 14285, 11363], 'infectious_mononucleosis_pi'))
infective_connective_tissue_disorders_phe_pi = TensorMap('phecode_712',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_712_date'), channel_map={'no_infective_connective_tissue_disorders':0, 'infective_connective_tissue_disorders_prevalent':1, 'infective_connective_tissue_disorders_incident':2}, loss=weighted_crossentropy([1.0, 23809, 35714], 'infective_connective_tissue_disorders_pi'))
infertility_female_phe_pi = TensorMap('phecode_626.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.8_date'), channel_map={'no_infertility_female':0, 'infertility_female_prevalent':1, 'infertility_female_incident':2}, loss=weighted_crossentropy([1.0, 388, 15151], 'infertility_female_pi'))
infertility_female_associated_with_anovulation_phe_pi = TensorMap('phecode_626.15',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.15_date'), channel_map={'no_infertility_female_associated_with_anovulation':0, 'infertility_female_associated_with_anovulation_prevalent':1, 'infertility_female_associated_with_anovulation_incident':2}, loss=weighted_crossentropy([1.0, 45454, 166666], 'infertility_female_associated_with_anovulation_pi'))
infertility_male_phe_pi = TensorMap('phecode_609.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_609.1_date'), channel_map={'no_infertility_male':0, 'infertility_male_prevalent':1, 'infertility_male_incident':2}, loss=weighted_crossentropy([1.0, 7575, 45454], 'infertility_male_pi'))
infestation_lice_mites_phe_pi = TensorMap('phecode_132',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_132_date'), channel_map={'no_infestation_lice_mites':0, 'infestation_lice_mites_prevalent':1, 'infestation_lice_mites_incident':2}, loss=weighted_crossentropy([1.0, 25000, 20833], 'infestation_lice_mites_pi'))
inflammation_of_eyelids_phe_pi = TensorMap('phecode_371.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_371.3_date'), channel_map={'no_inflammation_of_eyelids':0, 'inflammation_of_eyelids_prevalent':1, 'inflammation_of_eyelids_incident':2}, loss=weighted_crossentropy([1.0, 279, 345], 'inflammation_of_eyelids_pi'))
inflammation_of_the_eye_phe_pi = TensorMap('phecode_371',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_371_date'), channel_map={'no_inflammation_of_the_eye':0, 'inflammation_of_the_eye_prevalent':1, 'inflammation_of_the_eye_incident':2}, loss=weighted_crossentropy([1.0, 10869, 5154], 'inflammation_of_the_eye_pi'))
inflammatory_and_toxic_neuropathy_phe_pi = TensorMap('phecode_357',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_357_date'), channel_map={'no_inflammatory_and_toxic_neuropathy':0, 'inflammatory_and_toxic_neuropathy_prevalent':1, 'inflammatory_and_toxic_neuropathy_incident':2}, loss=weighted_crossentropy([1.0, 952, 286], 'inflammatory_and_toxic_neuropathy_pi'))
inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis_phe_pi = TensorMap('phecode_555',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_555_date'), channel_map={'no_inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis':0, 'inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis_prevalent':1, 'inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis_incident':2}, loss=weighted_crossentropy([1.0, 36, 42], 'inflammatory_bowel_disease_and_other_gastroenteritis_and_colitis_pi'))
inflammatory_conditions_of_jaw_phe_pi = TensorMap('phecode_526.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_526.5_date'), channel_map={'no_inflammatory_conditions_of_jaw':0, 'inflammatory_conditions_of_jaw_prevalent':1, 'inflammatory_conditions_of_jaw_incident':2}, loss=weighted_crossentropy([1.0, 7462, 5102], 'inflammatory_conditions_of_jaw_pi'))
inflammatory_disease_of_breast_phe_pi = TensorMap('phecode_613.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_613.1_date'), channel_map={'no_inflammatory_disease_of_breast':0, 'inflammatory_disease_of_breast_prevalent':1, 'inflammatory_disease_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 880, 1432], 'inflammatory_disease_of_breast_pi'))
inflammatory_disease_of_cervix_vagina_and_vulva_phe_pi = TensorMap('phecode_614.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.5_date'), channel_map={'no_inflammatory_disease_of_cervix_vagina_and_vulva':0, 'inflammatory_disease_of_cervix_vagina_and_vulva_prevalent':1, 'inflammatory_disease_of_cervix_vagina_and_vulva_incident':2}, loss=weighted_crossentropy([1.0, 4166, 5555], 'inflammatory_disease_of_cervix_vagina_and_vulva_pi'))
inflammatory_diseases_of_female_pelvic_organs_phe_pi = TensorMap('phecode_614',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614_date'), channel_map={'no_inflammatory_diseases_of_female_pelvic_organs':0, 'inflammatory_diseases_of_female_pelvic_organs_prevalent':1, 'inflammatory_diseases_of_female_pelvic_organs_incident':2}, loss=weighted_crossentropy([1.0, 45454, 29411], 'inflammatory_diseases_of_female_pelvic_organs_pi'))
inflammatory_diseases_of_uterus_except_cervix_phe_pi = TensorMap('phecode_614.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.4_date'), channel_map={'no_inflammatory_diseases_of_uterus_except_cervix':0, 'inflammatory_diseases_of_uterus_except_cervix_prevalent':1, 'inflammatory_diseases_of_uterus_except_cervix_incident':2}, loss=weighted_crossentropy([1.0, 1930, 4098], 'inflammatory_diseases_of_uterus_except_cervix_pi'))
influenza_phe_pi = TensorMap('phecode_481',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_481_date'), channel_map={'no_influenza':0, 'influenza_prevalent':1, 'influenza_incident':2}, loss=weighted_crossentropy([1.0, 131, 40], 'influenza_pi'))
infusion_and_transfusion_reaction_phe_pi = TensorMap('phecode_942',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_942_date'), channel_map={'no_infusion_and_transfusion_reaction':0, 'infusion_and_transfusion_reaction_prevalent':1, 'infusion_and_transfusion_reaction_incident':2}, loss=weighted_crossentropy([1.0, 83333, 38461], 'infusion_and_transfusion_reaction_pi'))
ingrowing_nail_phe_pi = TensorMap('phecode_703.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_703.1_date'), channel_map={'no_ingrowing_nail':0, 'ingrowing_nail_prevalent':1, 'ingrowing_nail_incident':2}, loss=weighted_crossentropy([1.0, 636, 1639], 'ingrowing_nail_pi'))
inguinal_hernia_phe_pi = TensorMap('phecode_550.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_550.1_date'), channel_map={'no_inguinal_hernia':0, 'inguinal_hernia_prevalent':1, 'inguinal_hernia_incident':2}, loss=weighted_crossentropy([1.0, 46, 48], 'inguinal_hernia_pi'))
injuries_to_the_nervous_system_phe_pi = TensorMap('phecode_907',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_907_date'), channel_map={'no_injuries_to_the_nervous_system':0, 'injuries_to_the_nervous_system_prevalent':1, 'injuries_to_the_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 394, 418], 'injuries_to_the_nervous_system_pi'))
injury_nos_phe_pi = TensorMap('phecode_1009',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1009_date'), channel_map={'no_injury_nos':0, 'injury_nos_prevalent':1, 'injury_nos_incident':2}, loss=weighted_crossentropy([1.0, 55, 60], 'injury_nos_pi'))
injury_to_blood_vessels_phe_pi = TensorMap('phecode_1007',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1007_date'), channel_map={'no_injury_to_blood_vessels':0, 'injury_to_blood_vessels_prevalent':1, 'injury_to_blood_vessels_incident':2}, loss=weighted_crossentropy([1.0, 2551, 2941], 'injury_to_blood_vessels_pi'))
injury_to_other_and_unspecified_nerves_phe_pi = TensorMap('phecode_957',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_957_date'), channel_map={'no_injury_to_other_and_unspecified_nerves':0, 'injury_to_other_and_unspecified_nerves_prevalent':1, 'injury_to_other_and_unspecified_nerves_incident':2}, loss=weighted_crossentropy([1.0, 5102, 1628], 'injury_to_other_and_unspecified_nerves_pi'))
insomnia_phe_pi = TensorMap('phecode_327.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327.4_date'), channel_map={'no_insomnia':0, 'insomnia_prevalent':1, 'insomnia_incident':2}, loss=weighted_crossentropy([1.0, 45454, 16666], 'insomnia_pi'))
insulin_pump_user_phe_pi = TensorMap('phecode_250.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.3_date'), channel_map={'no_insulin_pump_user':0, 'insulin_pump_user_prevalent':1, 'insulin_pump_user_incident':2}, loss=weighted_crossentropy([1.0, 41666, 8196], 'insulin_pump_user_pi'))
internal_derangement_of_knee_phe_pi = TensorMap('phecode_835',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_835_date'), channel_map={'no_internal_derangement_of_knee':0, 'internal_derangement_of_knee_prevalent':1, 'internal_derangement_of_knee_incident':2}, loss=weighted_crossentropy([1.0, 158, 246], 'internal_derangement_of_knee_pi'))
intervertebral_disc_disorder_with_myelopathy_phe_pi = TensorMap('phecode_722.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722.7_date'), channel_map={'no_intervertebral_disc_disorder_with_myelopathy':0, 'intervertebral_disc_disorder_with_myelopathy_prevalent':1, 'intervertebral_disc_disorder_with_myelopathy_incident':2}, loss=weighted_crossentropy([1.0, 2463, 1319], 'intervertebral_disc_disorder_with_myelopathy_pi'))
intervertebral_disc_disorders_phe_pi = TensorMap('phecode_722',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722_date'), channel_map={'no_intervertebral_disc_disorders':0, 'intervertebral_disc_disorders_prevalent':1, 'intervertebral_disc_disorders_incident':2}, loss=weighted_crossentropy([1.0, 150, 96], 'intervertebral_disc_disorders_pi'))
intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption_phe_pi = TensorMap('phecode_271.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_271.3_date'), channel_map={'no_intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption':0, 'intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption_prevalent':1, 'intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption_incident':2}, loss=weighted_crossentropy([1.0, 7812, 2645], 'intestinal_disaccharidase_deficiencies_and_disaccharide_malabsorption_pi'))
intestinal_e_coli_phe_pi = TensorMap('phecode_8.51',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_8.51_date'), channel_map={'no_intestinal_e_coli':0, 'intestinal_e_coli_prevalent':1, 'intestinal_e_coli_incident':2}, loss=weighted_crossentropy([1.0, 22727, 19230], 'intestinal_e_coli_pi'))
intestinal_helminthiases_phe_pi = TensorMap('phecode_134.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_134.1_date'), channel_map={'no_intestinal_helminthiases':0, 'intestinal_helminthiases_prevalent':1, 'intestinal_helminthiases_incident':2}, loss=weighted_crossentropy([1.0, 8928, 4385], 'intestinal_helminthiases_pi'))
intestinal_infection_phe_pi = TensorMap('phecode_8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_8_date'), channel_map={'no_intestinal_infection':0, 'intestinal_infection_prevalent':1, 'intestinal_infection_incident':2}, loss=weighted_crossentropy([1.0, 170, 71], 'intestinal_infection_pi'))
intestinal_infection_due_to_c_difficile_phe_pi = TensorMap('phecode_8.52',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_8.52_date'), channel_map={'no_intestinal_infection_due_to_c_difficile':0, 'intestinal_infection_due_to_c_difficile_prevalent':1, 'intestinal_infection_due_to_c_difficile_incident':2}, loss=weighted_crossentropy([1.0, 1466, 742], 'intestinal_infection_due_to_c_difficile_pi'))
intestinal_infection_due_to_protozoa_phe_pi = TensorMap('phecode_8.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_8.7_date'), channel_map={'no_intestinal_infection_due_to_protozoa':0, 'intestinal_infection_due_to_protozoa_prevalent':1, 'intestinal_infection_due_to_protozoa_incident':2}, loss=weighted_crossentropy([1.0, 14705, 10638], 'intestinal_infection_due_to_protozoa_pi'))
intestinal_malabsorption_non_celiac_phe_pi = TensorMap('phecode_557',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_557_date'), channel_map={'no_intestinal_malabsorption_non_celiac':0, 'intestinal_malabsorption_non_celiac_prevalent':1, 'intestinal_malabsorption_non_celiac_incident':2}, loss=weighted_crossentropy([1.0, 2392, 2262], 'intestinal_malabsorption_non_celiac_pi'))
intracerebral_hemorrhage_phe_pi = TensorMap('phecode_430.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_430.2_date'), channel_map={'no_intracerebral_hemorrhage':0, 'intracerebral_hemorrhage_prevalent':1, 'intracerebral_hemorrhage_incident':2}, loss=weighted_crossentropy([1.0, 1628, 606], 'intracerebral_hemorrhage_pi'))
intracranial_hemorrhage_phe_pi = TensorMap('phecode_430',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_430_date'), channel_map={'no_intracranial_hemorrhage':0, 'intracranial_hemorrhage_prevalent':1, 'intracranial_hemorrhage_incident':2}, loss=weighted_crossentropy([1.0, 6944, 4504], 'intracranial_hemorrhage_pi'))
intracranial_hemorrhage_injury_phe_pi = TensorMap('phecode_818',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_818_date'), channel_map={'no_intracranial_hemorrhage_injury':0, 'intracranial_hemorrhage_injury_prevalent':1, 'intracranial_hemorrhage_injury_incident':2}, loss=weighted_crossentropy([1.0, 5050, 5882], 'intracranial_hemorrhage_injury_pi'))
iron_deficiency_anemia_secondary_to_blood_loss_chronic_phe_pi = TensorMap('phecode_280.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_280.2_date'), channel_map={'no_iron_deficiency_anemia_secondary_to_blood_loss_chronic':0, 'iron_deficiency_anemia_secondary_to_blood_loss_chronic_prevalent':1, 'iron_deficiency_anemia_secondary_to_blood_loss_chronic_incident':2}, loss=weighted_crossentropy([1.0, 1291, 1453], 'iron_deficiency_anemia_secondary_to_blood_loss_chronic_pi'))
iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss_phe_pi = TensorMap('phecode_280.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_280.1_date'), channel_map={'no_iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss':0, 'iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss_prevalent':1, 'iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss_incident':2}, loss=weighted_crossentropy([1.0, 153, 54], 'iron_deficiency_anemias_unspecified_or_not_due_to_blood_loss_pi'))
irregular_menstrual_bleeding_phe_pi = TensorMap('phecode_626.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.14_date'), channel_map={'no_irregular_menstrual_bleeding':0, 'irregular_menstrual_bleeding_prevalent':1, 'irregular_menstrual_bleeding_incident':2}, loss=weighted_crossentropy([1.0, 138, 527], 'irregular_menstrual_bleeding_pi'))
irregular_menstrual_cycle_phe_pi = TensorMap('phecode_626.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.13_date'), channel_map={'no_irregular_menstrual_cycle':0, 'irregular_menstrual_cycle_prevalent':1, 'irregular_menstrual_cycle_incident':2}, loss=weighted_crossentropy([1.0, 284, 1388], 'irregular_menstrual_cycle_pi'))
irregular_menstrual_cycle_bleeding_phe_pi = TensorMap('phecode_626.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.1_date'), channel_map={'no_irregular_menstrual_cycle_bleeding':0, 'irregular_menstrual_cycle_bleeding_prevalent':1, 'irregular_menstrual_cycle_bleeding_incident':2}, loss=weighted_crossentropy([1.0, 276, 1557], 'irregular_menstrual_cycle_bleeding_pi'))
irritable_bowel_syndrome_phe_pi = TensorMap('phecode_564.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_564.1_date'), channel_map={'no_irritable_bowel_syndrome':0, 'irritable_bowel_syndrome_prevalent':1, 'irritable_bowel_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 158, 98], 'irritable_bowel_syndrome_pi'))
ischemic_heart_disease_phe_pi = TensorMap('phecode_411',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411_date'), channel_map={'no_ischemic_heart_disease':0, 'ischemic_heart_disease_prevalent':1, 'ischemic_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 24, 21], 'ischemic_heart_disease_pi'))
jaundice_not_of_newborn_phe_pi = TensorMap('phecode_573.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.5_date'), channel_map={'no_jaundice_not_of_newborn':0, 'jaundice_not_of_newborn_prevalent':1, 'jaundice_not_of_newborn_incident':2}, loss=weighted_crossentropy([1.0, 1033, 615], 'jaundice_not_of_newborn_pi'))
jaw_disease_nos_phe_pi = TensorMap('phecode_526.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_526.9_date'), channel_map={'no_jaw_disease_nos':0, 'jaw_disease_nos_prevalent':1, 'jaw_disease_nos_incident':2}, loss=weighted_crossentropy([1.0, 26, 22], 'jaw_disease_nos_pi'))
joint_effusions_phe_pi = TensorMap('phecode_741.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741.4_date'), channel_map={'no_joint_effusions':0, 'joint_effusions_prevalent':1, 'joint_effusions_incident':2}, loss=weighted_crossentropy([1.0, 2551, 3378], 'joint_effusions_pi'))
joint_ligament_sprain_phe_pi = TensorMap('phecode_840.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_840.3_date'), channel_map={'no_joint_ligament_sprain':0, 'joint_ligament_sprain_prevalent':1, 'joint_ligament_sprain_incident':2}, loss=weighted_crossentropy([1.0, 3676, 4587], 'joint_ligament_sprain_pi'))
juvenile_osteochondrosis_phe_pi = TensorMap('phecode_732.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_732.1_date'), channel_map={'no_juvenile_osteochondrosis':0, 'juvenile_osteochondrosis_prevalent':1, 'juvenile_osteochondrosis_incident':2}, loss=weighted_crossentropy([1.0, 7575, 2958], 'juvenile_osteochondrosis_pi'))
juvenile_rheumatoid_arthritis_phe_pi = TensorMap('phecode_714.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_714.2_date'), channel_map={'no_juvenile_rheumatoid_arthritis':0, 'juvenile_rheumatoid_arthritis_prevalent':1, 'juvenile_rheumatoid_arthritis_incident':2}, loss=weighted_crossentropy([1.0, 38461, 38461], 'juvenile_rheumatoid_arthritis_pi'))
kaposis_sarcoma_phe_pi = TensorMap('phecode_230',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_230_date'), channel_map={'no_kaposis_sarcoma':0, 'kaposis_sarcoma_prevalent':1, 'kaposis_sarcoma_incident':2}, loss=weighted_crossentropy([1.0, 41666, 125000], 'kaposis_sarcoma_pi'))
keloid_scar_phe_pi = TensorMap('phecode_701.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701.4_date'), channel_map={'no_keloid_scar':0, 'keloid_scar_prevalent':1, 'keloid_scar_incident':2}, loss=weighted_crossentropy([1.0, 2538, 3030], 'keloid_scar_pi'))
keratitis_phe_pi = TensorMap('phecode_370',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_370_date'), channel_map={'no_keratitis':0, 'keratitis_prevalent':1, 'keratitis_incident':2}, loss=weighted_crossentropy([1.0, 6666, 4310], 'keratitis_pi'))
keratoconjunctivitis_phe_pi = TensorMap('phecode_370.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_370.3_date'), channel_map={'no_keratoconjunctivitis':0, 'keratoconjunctivitis_prevalent':1, 'keratoconjunctivitis_incident':2}, loss=weighted_crossentropy([1.0, 26315, 20000], 'keratoconjunctivitis_pi'))
keratoconus_phe_pi = TensorMap('phecode_364.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_364.41_date'), channel_map={'no_keratoconus':0, 'keratoconus_prevalent':1, 'keratoconus_incident':2}, loss=weighted_crossentropy([1.0, 8064, 7042], 'keratoconus_pi'))
keratoderma_acquired_phe_pi = TensorMap('phecode_701.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701.1_date'), channel_map={'no_keratoderma_acquired':0, 'keratoderma_acquired_prevalent':1, 'keratoderma_acquired_incident':2}, loss=weighted_crossentropy([1.0, 13888, 9803], 'keratoderma_acquired_pi'))
kidney_replaced_by_transpant_phe_pi = TensorMap('phecode_587',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_587_date'), channel_map={'no_kidney_replaced_by_transpant':0, 'kidney_replaced_by_transpant_prevalent':1, 'kidney_replaced_by_transpant_incident':2}, loss=weighted_crossentropy([1.0, 1587, 2100], 'kidney_replaced_by_transpant_pi'))
known_or_suspected_fetal_abnormality_affecting_management_of_mother_phe_pi = TensorMap('phecode_655',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_655_date'), channel_map={'no_known_or_suspected_fetal_abnormality_affecting_management_of_mother':0, 'known_or_suspected_fetal_abnormality_affecting_management_of_mother_prevalent':1, 'known_or_suspected_fetal_abnormality_affecting_management_of_mother_incident':2}, loss=weighted_crossentropy([1.0, 87, 2717], 'known_or_suspected_fetal_abnormality_affecting_management_of_mother_pi'))
kyphoscoliosis_and_scoliosis_phe_pi = TensorMap('phecode_737.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_737.3_date'), channel_map={'no_kyphoscoliosis_and_scoliosis':0, 'kyphoscoliosis_and_scoliosis_prevalent':1, 'kyphoscoliosis_and_scoliosis_incident':2}, loss=weighted_crossentropy([1.0, 4464, 831], 'kyphoscoliosis_and_scoliosis_pi'))
kyphosis_acquired_phe_pi = TensorMap('phecode_737.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_737.1_date'), channel_map={'no_kyphosis_acquired':0, 'kyphosis_acquired_prevalent':1, 'kyphosis_acquired_incident':2}, loss=weighted_crossentropy([1.0, 27777, 6849], 'kyphosis_acquired_pi'))
labyrinthitis_phe_pi = TensorMap('phecode_386.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_386.3_date'), channel_map={'no_labyrinthitis':0, 'labyrinthitis_prevalent':1, 'labyrinthitis_incident':2}, loss=weighted_crossentropy([1.0, 1123, 819], 'labyrinthitis_pi'))
lack_of_coordination_phe_pi = TensorMap('phecode_350.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_350.3_date'), channel_map={'no_lack_of_coordination':0, 'lack_of_coordination_prevalent':1, 'lack_of_coordination_incident':2}, loss=weighted_crossentropy([1.0, 2976, 1362], 'lack_of_coordination_pi'))
lack_of_normal_physiological_development_unspecified_phe_pi = TensorMap('phecode_264.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_264.9_date'), channel_map={'no_lack_of_normal_physiological_development_unspecified':0, 'lack_of_normal_physiological_development_unspecified_prevalent':1, 'lack_of_normal_physiological_development_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 166666, 26315], 'lack_of_normal_physiological_development_unspecified_pi'))
lagophthalmos_phe_pi = TensorMap('phecode_374.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_374.2_date'), channel_map={'no_lagophthalmos':0, 'lagophthalmos_prevalent':1, 'lagophthalmos_incident':2}, loss=weighted_crossentropy([1.0, 17857, 14705], 'lagophthalmos_pi'))
large_cell_lymphoma_phe_pi = TensorMap('phecode_202.24',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_202.24_date'), channel_map={'no_large_cell_lymphoma':0, 'large_cell_lymphoma_prevalent':1, 'large_cell_lymphoma_incident':2}, loss=weighted_crossentropy([1.0, 2747, 722], 'large_cell_lymphoma_pi'))
late_effect_phe_pi = TensorMap('phecode_1012',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1012_date'), channel_map={'no_late_effect':0, 'late_effect_prevalent':1, 'late_effect_incident':2}, loss=weighted_crossentropy([1.0, 25000, 45454], 'late_effect_pi'))
late_effects_of_cerebrovascular_disease_phe_pi = TensorMap('phecode_433.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.8_date'), channel_map={'no_late_effects_of_cerebrovascular_disease':0, 'late_effects_of_cerebrovascular_disease_prevalent':1, 'late_effects_of_cerebrovascular_disease_incident':2}, loss=weighted_crossentropy([1.0, 816, 355], 'late_effects_of_cerebrovascular_disease_pi'))
late_effects_of_musculoskeletal_and_connective_tissue_injuries_phe_pi = TensorMap('phecode_905',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_905_date'), channel_map={'no_late_effects_of_musculoskeletal_and_connective_tissue_injuries':0, 'late_effects_of_musculoskeletal_and_connective_tissue_injuries_prevalent':1, 'late_effects_of_musculoskeletal_and_connective_tissue_injuries_incident':2}, loss=weighted_crossentropy([1.0, 5617, 6849], 'late_effects_of_musculoskeletal_and_connective_tissue_injuries_pi'))
late_pregnancy_and_failed_induction_phe_pi = TensorMap('phecode_645',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_645_date'), channel_map={'no_late_pregnancy_and_failed_induction':0, 'late_pregnancy_and_failed_induction_prevalent':1, 'late_pregnancy_and_failed_induction_incident':2}, loss=weighted_crossentropy([1.0, 313, 11111], 'late_pregnancy_and_failed_induction_pi'))
laxity_of_ligament_or_hypermobility_syndrome_phe_pi = TensorMap('phecode_728.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_728.2_date'), channel_map={'no_laxity_of_ligament_or_hypermobility_syndrome':0, 'laxity_of_ligament_or_hypermobility_syndrome_prevalent':1, 'laxity_of_ligament_or_hypermobility_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 6666, 4672], 'laxity_of_ligament_or_hypermobility_syndrome_pi'))
learning_disorder_phe_pi = TensorMap('phecode_315.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_315.1_date'), channel_map={'no_learning_disorder':0, 'learning_disorder_prevalent':1, 'learning_disorder_incident':2}, loss=weighted_crossentropy([1.0, 6578, 2512], 'learning_disorder_pi'))
left_bundle_branch_block_phe_pi = TensorMap('phecode_426.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.32_date'), channel_map={'no_left_bundle_branch_block':0, 'left_bundle_branch_block_prevalent':1, 'left_bundle_branch_block_incident':2}, loss=weighted_crossentropy([1.0, 818, 211], 'left_bundle_branch_block_pi'))
leukemia_phe_pi = TensorMap('phecode_204',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204_date'), channel_map={'no_leukemia':0, 'leukemia_prevalent':1, 'leukemia_incident':2}, loss=weighted_crossentropy([1.0, 2631, 1061], 'leukemia_pi'))
leukoplakia_of_oral_mucosa_phe_pi = TensorMap('phecode_528.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.6_date'), channel_map={'no_leukoplakia_of_oral_mucosa':0, 'leukoplakia_of_oral_mucosa_prevalent':1, 'leukoplakia_of_oral_mucosa_incident':2}, loss=weighted_crossentropy([1.0, 2793, 1811], 'leukoplakia_of_oral_mucosa_pi'))
lipoma_phe_pi = TensorMap('phecode_214',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_214_date'), channel_map={'no_lipoma':0, 'lipoma_prevalent':1, 'lipoma_incident':2}, loss=weighted_crossentropy([1.0, 498, 390], 'lipoma_pi'))
lipoma_of_skin_and_subcutaneous_tissue_phe_pi = TensorMap('phecode_214.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_214.1_date'), channel_map={'no_lipoma_of_skin_and_subcutaneous_tissue':0, 'lipoma_of_skin_and_subcutaneous_tissue_prevalent':1, 'lipoma_of_skin_and_subcutaneous_tissue_incident':2}, loss=weighted_crossentropy([1.0, 109, 234], 'lipoma_of_skin_and_subcutaneous_tissue_pi'))
lipoprotein_disorders_phe_pi = TensorMap('phecode_277.51',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_277.51_date'), channel_map={'no_lipoprotein_disorders':0, 'lipoprotein_disorders_prevalent':1, 'lipoprotein_disorders_incident':2}, loss=weighted_crossentropy([1.0, 11904, 11363], 'lipoprotein_disorders_pi'))
liver_abscess_and_sequelae_of_chronic_liver_disease_phe_pi = TensorMap('phecode_571.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_571.8_date'), channel_map={'no_liver_abscess_and_sequelae_of_chronic_liver_disease':0, 'liver_abscess_and_sequelae_of_chronic_liver_disease_prevalent':1, 'liver_abscess_and_sequelae_of_chronic_liver_disease_incident':2}, loss=weighted_crossentropy([1.0, 2732, 694], 'liver_abscess_and_sequelae_of_chronic_liver_disease_pi'))
liver_replaced_by_transplant_phe_pi = TensorMap('phecode_573.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573.2_date'), channel_map={'no_liver_replaced_by_transplant':0, 'liver_replaced_by_transplant_prevalent':1, 'liver_replaced_by_transplant_incident':2}, loss=weighted_crossentropy([1.0, 6172, 7462], 'liver_replaced_by_transplant_pi'))
localized_adiposity_phe_pi = TensorMap('phecode_278.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_278.3_date'), channel_map={'no_localized_adiposity':0, 'localized_adiposity_prevalent':1, 'localized_adiposity_incident':2}, loss=weighted_crossentropy([1.0, 5813, 27777], 'localized_adiposity_pi'))
localized_superficial_swelling_mass_or_lump_phe_pi = TensorMap('phecode_687.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_687.2_date'), channel_map={'no_localized_superficial_swelling_mass_or_lump':0, 'localized_superficial_swelling_mass_or_lump_prevalent':1, 'localized_superficial_swelling_mass_or_lump_incident':2}, loss=weighted_crossentropy([1.0, 2336, 6024], 'localized_superficial_swelling_mass_or_lump_pi'))
loose_body_in_joint_phe_pi = TensorMap('phecode_742.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_742.1_date'), channel_map={'no_loose_body_in_joint':0, 'loose_body_in_joint_prevalent':1, 'loose_body_in_joint_incident':2}, loss=weighted_crossentropy([1.0, 9803, 35714], 'loose_body_in_joint_pi'))
lordosis_acquired_phe_pi = TensorMap('phecode_737.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_737.2_date'), channel_map={'no_lordosis_acquired':0, 'lordosis_acquired_prevalent':1, 'lordosis_acquired_incident':2}, loss=weighted_crossentropy([1.0, 250000, 50000], 'lordosis_acquired_pi'))
lower_gastrointestinal_congenital_anomalies_phe_pi = TensorMap('phecode_750.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.2_date'), channel_map={'no_lower_gastrointestinal_congenital_anomalies':0, 'lower_gastrointestinal_congenital_anomalies_prevalent':1, 'lower_gastrointestinal_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 55555, 41666], 'lower_gastrointestinal_congenital_anomalies_pi'))
lump_or_mass_in_breast_phe_pi = TensorMap('phecode_611.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_611.3_date'), channel_map={'no_lump_or_mass_in_breast':0, 'lump_or_mass_in_breast_prevalent':1, 'lump_or_mass_in_breast_incident':2}, loss=weighted_crossentropy([1.0, 309, 1157], 'lump_or_mass_in_breast_pi'))
lung_disease_due_to_external_agents_phe_pi = TensorMap('phecode_500',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_500_date'), channel_map={'no_lung_disease_due_to_external_agents':0, 'lung_disease_due_to_external_agents_prevalent':1, 'lung_disease_due_to_external_agents_incident':2}, loss=weighted_crossentropy([1.0, 12820, 3937], 'lung_disease_due_to_external_agents_pi'))
lung_transplant_phe_pi = TensorMap('phecode_510.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_510.2_date'), channel_map={'no_lung_transplant':0, 'lung_transplant_prevalent':1, 'lung_transplant_incident':2}, loss=weighted_crossentropy([1.0, 16666, 17857], 'lung_transplant_pi'))
lyme_disease_phe_pi = TensorMap('phecode_130.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_130.1_date'), channel_map={'no_lyme_disease':0, 'lyme_disease_prevalent':1, 'lyme_disease_incident':2}, loss=weighted_crossentropy([1.0, 22727, 23809], 'lyme_disease_pi'))
lymphadenitis_phe_pi = TensorMap('phecode_289.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_289.4_date'), channel_map={'no_lymphadenitis':0, 'lymphadenitis_prevalent':1, 'lymphadenitis_incident':2}, loss=weighted_crossentropy([1.0, 351, 194], 'lymphadenitis_pi'))
lymphoid_leukemia_phe_pi = TensorMap('phecode_204.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.1_date'), channel_map={'no_lymphoid_leukemia':0, 'lymphoid_leukemia_prevalent':1, 'lymphoid_leukemia_incident':2}, loss=weighted_crossentropy([1.0, 22727, 8474], 'lymphoid_leukemia_pi'))
lymphoid_leukemia_acute_phe_pi = TensorMap('phecode_204.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.11_date'), channel_map={'no_lymphoid_leukemia_acute':0, 'lymphoid_leukemia_acute_prevalent':1, 'lymphoid_leukemia_acute_incident':2}, loss=weighted_crossentropy([1.0, 17857, 7142], 'lymphoid_leukemia_acute_pi'))
lymphoid_leukemia_chronic_phe_pi = TensorMap('phecode_204.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.12_date'), channel_map={'no_lymphoid_leukemia_chronic':0, 'lymphoid_leukemia_chronic_prevalent':1, 'lymphoid_leukemia_chronic_incident':2}, loss=weighted_crossentropy([1.0, 2369, 922], 'lymphoid_leukemia_chronic_pi'))
lymphosarcoma_phe_pi = TensorMap('phecode_202.23',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_202.23_date'), channel_map={'no_lymphosarcoma':0, 'lymphosarcoma_prevalent':1, 'lymphosarcoma_incident':2}, loss=weighted_crossentropy([1.0, 50000, 100000], 'lymphosarcoma_pi'))
macroglobulinemia_phe_pi = TensorMap('phecode_270.35',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.35_date'), channel_map={'no_macroglobulinemia':0, 'macroglobulinemia_prevalent':1, 'macroglobulinemia_incident':2}, loss=weighted_crossentropy([1.0, 18518, 5952], 'macroglobulinemia_pi'))
macular_degeneration_senile_of_retina_nos_phe_pi = TensorMap('phecode_362.29',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.29_date'), channel_map={'no_macular_degeneration_senile_of_retina_nos':0, 'macular_degeneration_senile_of_retina_nos_prevalent':1, 'macular_degeneration_senile_of_retina_nos_incident':2}, loss=weighted_crossentropy([1.0, 727, 153], 'macular_degeneration_senile_of_retina_nos_pi'))
major_depressive_disorder_phe_pi = TensorMap('phecode_296.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_296.22_date'), channel_map={'no_major_depressive_disorder':0, 'major_depressive_disorder_prevalent':1, 'major_depressive_disorder_incident':2}, loss=weighted_crossentropy([1.0, 1492, 2202], 'major_depressive_disorder_pi'))
major_puerperal_infection_phe_pi = TensorMap('phecode_647.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_647.3_date'), channel_map={'no_major_puerperal_infection':0, 'major_puerperal_infection_prevalent':1, 'major_puerperal_infection_incident':2}, loss=weighted_crossentropy([1.0, 649, 11904], 'major_puerperal_infection_pi'))
malaise_and_fatigue_phe_pi = TensorMap('phecode_798',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_798_date'), channel_map={'no_malaise_and_fatigue':0, 'malaise_and_fatigue_prevalent':1, 'malaise_and_fatigue_incident':2}, loss=weighted_crossentropy([1.0, 377, 164], 'malaise_and_fatigue_pi'))
malignant_neoplasm_of_bladder_phe_pi = TensorMap('phecode_189.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_189.21_date'), channel_map={'no_malignant_neoplasm_of_bladder':0, 'malignant_neoplasm_of_bladder_prevalent':1, 'malignant_neoplasm_of_bladder_incident':2}, loss=weighted_crossentropy([1.0, 463, 264], 'malignant_neoplasm_of_bladder_pi'))
malignant_neoplasm_of_female_breast_phe_pi = TensorMap('phecode_174.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_174.11_date'), channel_map={'no_malignant_neoplasm_of_female_breast':0, 'malignant_neoplasm_of_female_breast_prevalent':1, 'malignant_neoplasm_of_female_breast_incident':2}, loss=weighted_crossentropy([1.0, 59, 59], 'malignant_neoplasm_of_female_breast_pi'))
malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts_phe_pi = TensorMap('phecode_159.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_159.3_date'), channel_map={'no_malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts':0, 'malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts_prevalent':1, 'malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts_incident':2}, loss=weighted_crossentropy([1.0, 13157, 1831], 'malignant_neoplasm_of_gallbladder_and_extrahepatic_bile_ducts_pi'))
malignant_neoplasm_of_head_face_and_neck_phe_pi = TensorMap('phecode_195.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_195.3_date'), channel_map={'no_malignant_neoplasm_of_head_face_and_neck':0, 'malignant_neoplasm_of_head_face_and_neck_prevalent':1, 'malignant_neoplasm_of_head_face_and_neck_incident':2}, loss=weighted_crossentropy([1.0, 20000, 16129], 'malignant_neoplasm_of_head_face_and_neck_pi'))
malignant_neoplasm_of_kidney_except_pelvis_phe_pi = TensorMap('phecode_189.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_189.11_date'), channel_map={'no_malignant_neoplasm_of_kidney_except_pelvis':0, 'malignant_neoplasm_of_kidney_except_pelvis_prevalent':1, 'malignant_neoplasm_of_kidney_except_pelvis_incident':2}, loss=weighted_crossentropy([1.0, 1190, 473], 'malignant_neoplasm_of_kidney_except_pelvis_pi'))
malignant_neoplasm_of_liver_primary_phe_pi = TensorMap('phecode_155.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_155.1_date'), channel_map={'no_malignant_neoplasm_of_liver_primary':0, 'malignant_neoplasm_of_liver_primary_prevalent':1, 'malignant_neoplasm_of_liver_primary_incident':2}, loss=weighted_crossentropy([1.0, 19230, 2232], 'malignant_neoplasm_of_liver_primary_pi'))
malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum_phe_pi = TensorMap('phecode_159',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_159_date'), channel_map={'no_malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum':0, 'malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum_prevalent':1, 'malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum_incident':2}, loss=weighted_crossentropy([1.0, 200, 99], 'malignant_neoplasm_of_other_and_ill_defined_sites_within_the_digestive_organs_and_peritoneum_pi'))
malignant_neoplasm_of_other_urinary_organs_phe_pi = TensorMap('phecode_189.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_189.4_date'), channel_map={'no_malignant_neoplasm_of_other_urinary_organs':0, 'malignant_neoplasm_of_other_urinary_organs_prevalent':1, 'malignant_neoplasm_of_other_urinary_organs_incident':2}, loss=weighted_crossentropy([1.0, 328, 170], 'malignant_neoplasm_of_other_urinary_organs_pi'))
malignant_neoplasm_of_ovary_phe_pi = TensorMap('phecode_184.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_184.11_date'), channel_map={'no_malignant_neoplasm_of_ovary':0, 'malignant_neoplasm_of_ovary_prevalent':1, 'malignant_neoplasm_of_ovary_incident':2}, loss=weighted_crossentropy([1.0, 245, 92], 'malignant_neoplasm_of_ovary_pi'))
malignant_neoplasm_of_ovary_and_other_uterine_adnexa_phe_pi = TensorMap('phecode_184.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_184.1_date'), channel_map={'no_malignant_neoplasm_of_ovary_and_other_uterine_adnexa':0, 'malignant_neoplasm_of_ovary_and_other_uterine_adnexa_prevalent':1, 'malignant_neoplasm_of_ovary_and_other_uterine_adnexa_incident':2}, loss=weighted_crossentropy([1.0, 345, 234], 'malignant_neoplasm_of_ovary_and_other_uterine_adnexa_pi'))
malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus_phe_pi = TensorMap('phecode_153.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_153.3_date'), channel_map={'no_malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus':0, 'malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus_prevalent':1, 'malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus_incident':2}, loss=weighted_crossentropy([1.0, 468, 264], 'malignant_neoplasm_of_rectum_rectosigmoid_junction_and_anus_pi'))
malignant_neoplasm_of_renal_pelvis_phe_pi = TensorMap('phecode_189.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_189.12_date'), channel_map={'no_malignant_neoplasm_of_renal_pelvis':0, 'malignant_neoplasm_of_renal_pelvis_prevalent':1, 'malignant_neoplasm_of_renal_pelvis_incident':2}, loss=weighted_crossentropy([1.0, 23809, 5882], 'malignant_neoplasm_of_renal_pelvis_pi'))
malignant_neoplasm_of_retroperitoneum_and_peritoneum_phe_pi = TensorMap('phecode_159.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_159.4_date'), channel_map={'no_malignant_neoplasm_of_retroperitoneum_and_peritoneum':0, 'malignant_neoplasm_of_retroperitoneum_and_peritoneum_prevalent':1, 'malignant_neoplasm_of_retroperitoneum_and_peritoneum_incident':2}, loss=weighted_crossentropy([1.0, 11363, 2016], 'malignant_neoplasm_of_retroperitoneum_and_peritoneum_pi'))
malignant_neoplasm_of_small_intestine_including_duodenum_phe_pi = TensorMap('phecode_159.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_159.2_date'), channel_map={'no_malignant_neoplasm_of_small_intestine_including_duodenum':0, 'malignant_neoplasm_of_small_intestine_including_duodenum_prevalent':1, 'malignant_neoplasm_of_small_intestine_including_duodenum_incident':2}, loss=weighted_crossentropy([1.0, 7936, 2380], 'malignant_neoplasm_of_small_intestine_including_duodenum_pi'))
malignant_neoplasm_of_testis_phe_pi = TensorMap('phecode_187.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_187.2_date'), channel_map={'no_malignant_neoplasm_of_testis':0, 'malignant_neoplasm_of_testis_prevalent':1, 'malignant_neoplasm_of_testis_incident':2}, loss=weighted_crossentropy([1.0, 267, 101], 'malignant_neoplasm_of_testis_pi'))
malignant_neoplasm_of_unspecified_male_genital_organ_phe_pi = TensorMap('phecode_187.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_187.1_date'), channel_map={'no_malignant_neoplasm_of_unspecified_male_genital_organ':0, 'malignant_neoplasm_of_unspecified_male_genital_organ_prevalent':1, 'malignant_neoplasm_of_unspecified_male_genital_organ_incident':2}, loss=weighted_crossentropy([1.0, 135, 58], 'malignant_neoplasm_of_unspecified_male_genital_organ_pi'))
malignant_neoplasm_of_uterus_phe_pi = TensorMap('phecode_182',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_182_date'), channel_map={'no_malignant_neoplasm_of_uterus':0, 'malignant_neoplasm_of_uterus_prevalent':1, 'malignant_neoplasm_of_uterus_incident':2}, loss=weighted_crossentropy([1.0, 684, 448], 'malignant_neoplasm_of_uterus_pi'))
malignant_neoplasm_other_phe_pi = TensorMap('phecode_195.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_195.1_date'), channel_map={'no_malignant_neoplasm_other':0, 'malignant_neoplasm_other_prevalent':1, 'malignant_neoplasm_other_incident':2}, loss=weighted_crossentropy([1.0, 18, 10], 'malignant_neoplasm_other_pi'))
malposition_and_malpresentation_of_fetus_or_obstruction_phe_pi = TensorMap('phecode_652',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_652_date'), channel_map={'no_malposition_and_malpresentation_of_fetus_or_obstruction':0, 'malposition_and_malpresentation_of_fetus_or_obstruction_prevalent':1, 'malposition_and_malpresentation_of_fetus_or_obstruction_incident':2}, loss=weighted_crossentropy([1.0, 226, 6493], 'malposition_and_malpresentation_of_fetus_or_obstruction_pi'))
malunion_and_nonunion_of_fracture_phe_pi = TensorMap('phecode_733.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_733.8_date'), channel_map={'no_malunion_and_nonunion_of_fracture':0, 'malunion_and_nonunion_of_fracture_prevalent':1, 'malunion_and_nonunion_of_fracture_incident':2}, loss=weighted_crossentropy([1.0, 3267, 10204], 'malunion_and_nonunion_of_fracture_pi'))
manlignant_and_unknown_neoplasms_of_brain_and_nervous_system_phe_pi = TensorMap('phecode_191',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_191_date'), channel_map={'no_manlignant_and_unknown_neoplasms_of_brain_and_nervous_system':0, 'manlignant_and_unknown_neoplasms_of_brain_and_nervous_system_prevalent':1, 'manlignant_and_unknown_neoplasms_of_brain_and_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 5681, 2304], 'manlignant_and_unknown_neoplasms_of_brain_and_nervous_system_pi'))
mastodynia_phe_pi = TensorMap('phecode_613.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_613.5_date'), channel_map={'no_mastodynia':0, 'mastodynia_prevalent':1, 'mastodynia_incident':2}, loss=weighted_crossentropy([1.0, 8620, 10416], 'mastodynia_pi'))
mastoiditis_related_conditions_phe_pi = TensorMap('phecode_381.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_381.3_date'), channel_map={'no_mastoiditis_related_conditions':0, 'mastoiditis_related_conditions_prevalent':1, 'mastoiditis_related_conditions_incident':2}, loss=weighted_crossentropy([1.0, 4310, 5617], 'mastoiditis_related_conditions_pi'))
maternal_complication_of_pregnancy_affecting_fetus_or_newborn_phe_pi = TensorMap('phecode_658',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_658_date'), channel_map={'no_maternal_complication_of_pregnancy_affecting_fetus_or_newborn':0, 'maternal_complication_of_pregnancy_affecting_fetus_or_newborn_prevalent':1, 'maternal_complication_of_pregnancy_affecting_fetus_or_newborn_incident':2}, loss=weighted_crossentropy([1.0, 16129, 55555], 'maternal_complication_of_pregnancy_affecting_fetus_or_newborn_pi'))
mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft_phe_pi = TensorMap('phecode_857',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_857_date'), channel_map={'no_mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft':0, 'mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft_prevalent':1, 'mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft_incident':2}, loss=weighted_crossentropy([1.0, 878, 372], 'mechanical_complication_of_unspecified_genitourinary_device_implant_and_graft_pi'))
medulloadrenal_hyperfunction_phe_pi = TensorMap('phecode_255.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_255.13_date'), channel_map={'no_medulloadrenal_hyperfunction':0, 'medulloadrenal_hyperfunction_prevalent':1, 'medulloadrenal_hyperfunction_incident':2}, loss=weighted_crossentropy([1.0, 166666, 71428], 'medulloadrenal_hyperfunction_pi'))
megaloblastic_anemia_phe_pi = TensorMap('phecode_281.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_281.1_date'), channel_map={'no_megaloblastic_anemia':0, 'megaloblastic_anemia_prevalent':1, 'megaloblastic_anemia_incident':2}, loss=weighted_crossentropy([1.0, 62500, 50000], 'megaloblastic_anemia_pi'))
melanomas_of_skin_phe_pi = TensorMap('phecode_172.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_172.11_date'), channel_map={'no_melanomas_of_skin':0, 'melanomas_of_skin_prevalent':1, 'melanomas_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 79, 39], 'melanomas_of_skin_pi'))
memory_loss_phe_pi = TensorMap('phecode_292.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292.3_date'), channel_map={'no_memory_loss':0, 'memory_loss_prevalent':1, 'memory_loss_incident':2}, loss=weighted_crossentropy([1.0, 1492, 672], 'memory_loss_pi'))
menieres_disease_phe_pi = TensorMap('phecode_386.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_386.1_date'), channel_map={'no_menieres_disease':0, 'menieres_disease_prevalent':1, 'menieres_disease_incident':2}, loss=weighted_crossentropy([1.0, 1700, 965], 'menieres_disease_pi'))
meningitis_phe_pi = TensorMap('phecode_320',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_320_date'), channel_map={'no_meningitis':0, 'meningitis_prevalent':1, 'meningitis_incident':2}, loss=weighted_crossentropy([1.0, 1677, 1984], 'meningitis_pi'))
menopausal_and_postmenopausal_disorders_phe_pi = TensorMap('phecode_627',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_627_date'), channel_map={'no_menopausal_and_postmenopausal_disorders':0, 'menopausal_and_postmenopausal_disorders_prevalent':1, 'menopausal_and_postmenopausal_disorders_incident':2}, loss=weighted_crossentropy([1.0, 859, 3105], 'menopausal_and_postmenopausal_disorders_pi'))
mental_retardation_phe_pi = TensorMap('phecode_315.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_315.3_date'), channel_map={'no_mental_retardation':0, 'mental_retardation_prevalent':1, 'mental_retardation_incident':2}, loss=weighted_crossentropy([1.0, 20000, 6493], 'mental_retardation_pi'))
migrain_with_aura_phe_pi = TensorMap('phecode_340.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_340.1_date'), channel_map={'no_migrain_with_aura':0, 'migrain_with_aura_prevalent':1, 'migrain_with_aura_incident':2}, loss=weighted_crossentropy([1.0, 4629, 2222], 'migrain_with_aura_pi'))
migraine_phe_pi = TensorMap('phecode_340',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_340_date'), channel_map={'no_migraine':0, 'migraine_prevalent':1, 'migraine_incident':2}, loss=weighted_crossentropy([1.0, 340, 184], 'migraine_pi'))
mild_cognitive_impairment_phe_pi = TensorMap('phecode_292.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292.2_date'), channel_map={'no_mild_cognitive_impairment':0, 'mild_cognitive_impairment_prevalent':1, 'mild_cognitive_impairment_incident':2}, loss=weighted_crossentropy([1.0, 13513, 1424], 'mild_cognitive_impairment_pi'))
mineral_deficiency_nec_phe_pi = TensorMap('phecode_262',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_262_date'), channel_map={'no_mineral_deficiency_nec':0, 'mineral_deficiency_nec_prevalent':1, 'mineral_deficiency_nec_incident':2}, loss=weighted_crossentropy([1.0, 5952, 1333], 'mineral_deficiency_nec_pi'))
miscarriage_stillbirth_phe_pi = TensorMap('phecode_634',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_634_date'), channel_map={'no_miscarriage_stillbirth':0, 'miscarriage_stillbirth_prevalent':1, 'miscarriage_stillbirth_incident':2}, loss=weighted_crossentropy([1.0, 90, 2538], 'miscarriage_stillbirth_pi'))
missed_abortion_hydatidiform_mole_phe_pi = TensorMap('phecode_634.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_634.1_date'), channel_map={'no_missed_abortion_hydatidiform_mole':0, 'missed_abortion_hydatidiform_mole_prevalent':1, 'missed_abortion_hydatidiform_mole_incident':2}, loss=weighted_crossentropy([1.0, 342, 4761], 'missed_abortion_hydatidiform_mole_pi'))
mitral_valve_disease_phe_pi = TensorMap('phecode_394.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_394.2_date'), channel_map={'no_mitral_valve_disease':0, 'mitral_valve_disease_prevalent':1, 'mitral_valve_disease_incident':2}, loss=weighted_crossentropy([1.0, 236, 107], 'mitral_valve_disease_pi'))
mittelschmerz_phe_pi = TensorMap('phecode_626.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.21_date'), channel_map={'no_mittelschmerz':0, 'mittelschmerz_prevalent':1, 'mittelschmerz_incident':2}, loss=weighted_crossentropy([1.0, 19230, 71428], 'mittelschmerz_pi'))
mixed_hyperlipidemia_phe_pi = TensorMap('phecode_272.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_272.13_date'), channel_map={'no_mixed_hyperlipidemia':0, 'mixed_hyperlipidemia_prevalent':1, 'mixed_hyperlipidemia_incident':2}, loss=weighted_crossentropy([1.0, 5813, 3289], 'mixed_hyperlipidemia_pi'))
monocytic_leukemia_phe_pi = TensorMap('phecode_204.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.3_date'), channel_map={'no_monocytic_leukemia':0, 'monocytic_leukemia_prevalent':1, 'monocytic_leukemia_incident':2}, loss=weighted_crossentropy([1.0, 166666, 8620], 'monocytic_leukemia_pi'))
mood_disorders_phe_pi = TensorMap('phecode_296',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_296_date'), channel_map={'no_mood_disorders':0, 'mood_disorders_prevalent':1, 'mood_disorders_incident':2}, loss=weighted_crossentropy([1.0, 94, 33], 'mood_disorders_pi'))
moyamoya_disease_phe_pi = TensorMap('phecode_433.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.32_date'), channel_map={'no_moyamoya_disease':0, 'moyamoya_disease_prevalent':1, 'moyamoya_disease_incident':2}, loss=weighted_crossentropy([1.0, 250000, 166666], 'moyamoya_disease_pi'))
mucous_polyp_of_cervix_phe_pi = TensorMap('phecode_622.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_622.2_date'), channel_map={'no_mucous_polyp_of_cervix':0, 'mucous_polyp_of_cervix_prevalent':1, 'mucous_polyp_of_cervix_incident':2}, loss=weighted_crossentropy([1.0, 163, 484], 'mucous_polyp_of_cervix_pi'))
multiple_gestation_phe_pi = TensorMap('phecode_651',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_651_date'), channel_map={'no_multiple_gestation':0, 'multiple_gestation_prevalent':1, 'multiple_gestation_incident':2}, loss=weighted_crossentropy([1.0, 1160, 11627], 'multiple_gestation_pi'))
multiple_myeloma_phe_pi = TensorMap('phecode_204.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.4_date'), channel_map={'no_multiple_myeloma':0, 'multiple_myeloma_prevalent':1, 'multiple_myeloma_incident':2}, loss=weighted_crossentropy([1.0, 2242, 758], 'multiple_myeloma_pi'))
multiple_sclerosis_phe_pi = TensorMap('phecode_335',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_335_date'), channel_map={'no_multiple_sclerosis':0, 'multiple_sclerosis_prevalent':1, 'multiple_sclerosis_incident':2}, loss=weighted_crossentropy([1.0, 473, 703], 'multiple_sclerosis_pi'))
muscular_calcification_and_ossification_phe_pi = TensorMap('phecode_728.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_728.1_date'), channel_map={'no_muscular_calcification_and_ossification':0, 'muscular_calcification_and_ossification_prevalent':1, 'muscular_calcification_and_ossification_incident':2}, loss=weighted_crossentropy([1.0, 50000, 83333], 'muscular_calcification_and_ossification_pi'))
muscular_dystrophies_phe_pi = TensorMap('phecode_359.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_359.1_date'), channel_map={'no_muscular_dystrophies':0, 'muscular_dystrophies_prevalent':1, 'muscular_dystrophies_incident':2}, loss=weighted_crossentropy([1.0, 6410, 4854], 'muscular_dystrophies_pi'))
muscular_wasting_and_disuse_atrophy_phe_pi = TensorMap('phecode_772.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_772.1_date'), channel_map={'no_muscular_wasting_and_disuse_atrophy':0, 'muscular_wasting_and_disuse_atrophy_prevalent':1, 'muscular_wasting_and_disuse_atrophy_incident':2}, loss=weighted_crossentropy([1.0, 35714, 26315], 'muscular_wasting_and_disuse_atrophy_pi'))
musculoskeletal_symptoms_referable_to_limbs_phe_pi = TensorMap('phecode_771',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_771_date'), channel_map={'no_musculoskeletal_symptoms_referable_to_limbs':0, 'musculoskeletal_symptoms_referable_to_limbs_prevalent':1, 'musculoskeletal_symptoms_referable_to_limbs_incident':2}, loss=weighted_crossentropy([1.0, 780, 946], 'musculoskeletal_symptoms_referable_to_limbs_pi'))
myalgia_and_myositis_unspecified_phe_pi = TensorMap('phecode_770',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_770_date'), channel_map={'no_myalgia_and_myositis_unspecified':0, 'myalgia_and_myositis_unspecified_prevalent':1, 'myalgia_and_myositis_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 4385, 3676], 'myalgia_and_myositis_unspecified_pi'))
myasthenia_gravis_phe_pi = TensorMap('phecode_358.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_358.1_date'), channel_map={'no_myasthenia_gravis':0, 'myasthenia_gravis_prevalent':1, 'myasthenia_gravis_incident':2}, loss=weighted_crossentropy([1.0, 4385, 3246], 'myasthenia_gravis_pi'))
mycoses_phe_pi = TensorMap('phecode_117',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_117_date'), channel_map={'no_mycoses':0, 'mycoses_prevalent':1, 'mycoses_incident':2}, loss=weighted_crossentropy([1.0, 499, 143], 'mycoses_pi'))
myeloid_leukemia_phe_pi = TensorMap('phecode_204.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.2_date'), channel_map={'no_myeloid_leukemia':0, 'myeloid_leukemia_prevalent':1, 'myeloid_leukemia_incident':2}, loss=weighted_crossentropy([1.0, 38461, 12195], 'myeloid_leukemia_pi'))
myeloid_leukemia_acute_phe_pi = TensorMap('phecode_204.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.21_date'), channel_map={'no_myeloid_leukemia_acute':0, 'myeloid_leukemia_acute_prevalent':1, 'myeloid_leukemia_acute_incident':2}, loss=weighted_crossentropy([1.0, 4545, 1440], 'myeloid_leukemia_acute_pi'))
myeloid_leukemia_chronic_phe_pi = TensorMap('phecode_204.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_204.22_date'), channel_map={'no_myeloid_leukemia_chronic':0, 'myeloid_leukemia_chronic_prevalent':1, 'myeloid_leukemia_chronic_incident':2}, loss=weighted_crossentropy([1.0, 4310, 5050], 'myeloid_leukemia_chronic_pi'))
myeloproliferative_disease_phe_pi = TensorMap('phecode_200',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_200_date'), channel_map={'no_myeloproliferative_disease':0, 'myeloproliferative_disease_prevalent':1, 'myeloproliferative_disease_incident':2}, loss=weighted_crossentropy([1.0, 1730, 656], 'myeloproliferative_disease_pi'))
myocardial_infarction_phe_pi = TensorMap('phecode_411.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.2_date'), channel_map={'no_myocardial_infarction':0, 'myocardial_infarction_prevalent':1, 'myocardial_infarction_incident':2}, loss=weighted_crossentropy([1.0, 74, 47], 'myocardial_infarction_pi'))
myocarditis_phe_pi = TensorMap('phecode_420.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_420.1_date'), channel_map={'no_myocarditis':0, 'myocarditis_prevalent':1, 'myocarditis_incident':2}, loss=weighted_crossentropy([1.0, 9803, 4310], 'myocarditis_pi'))
myoclonus_phe_pi = TensorMap('phecode_333.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_333.2_date'), channel_map={'no_myoclonus':0, 'myoclonus_prevalent':1, 'myoclonus_incident':2}, loss=weighted_crossentropy([1.0, 20833, 4424], 'myoclonus_pi'))
myoneural_disorders_phe_pi = TensorMap('phecode_358',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_358_date'), channel_map={'no_myoneural_disorders':0, 'myoneural_disorders_prevalent':1, 'myoneural_disorders_incident':2}, loss=weighted_crossentropy([1.0, 29411, 12195], 'myoneural_disorders_pi'))
myopathy_phe_pi = TensorMap('phecode_359.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_359.2_date'), channel_map={'no_myopathy':0, 'myopathy_prevalent':1, 'myopathy_incident':2}, loss=weighted_crossentropy([1.0, 1683, 1121], 'myopathy_pi'))
myopia_phe_pi = TensorMap('phecode_367.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_367.1_date'), channel_map={'no_myopia':0, 'myopia_prevalent':1, 'myopia_incident':2}, loss=weighted_crossentropy([1.0, 891, 329], 'myopia_pi'))
myringitis_phe_pi = TensorMap('phecode_384.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_384.1_date'), channel_map={'no_myringitis':0, 'myringitis_prevalent':1, 'myringitis_incident':2}, loss=weighted_crossentropy([1.0, 25000, 100000], 'myringitis_pi'))
nasal_polyps_phe_pi = TensorMap('phecode_471',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_471_date'), channel_map={'no_nasal_polyps':0, 'nasal_polyps_prevalent':1, 'nasal_polyps_incident':2}, loss=weighted_crossentropy([1.0, 194, 320], 'nasal_polyps_pi'))
nausea_and_vomiting_phe_pi = TensorMap('phecode_789',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_789_date'), channel_map={'no_nausea_and_vomiting':0, 'nausea_and_vomiting_prevalent':1, 'nausea_and_vomiting_incident':2}, loss=weighted_crossentropy([1.0, 79, 45], 'nausea_and_vomiting_pi'))
neoplasm_of_uncertain_behavior_phe_pi = TensorMap('phecode_199',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_199_date'), channel_map={'no_neoplasm_of_uncertain_behavior':0, 'neoplasm_of_uncertain_behavior_prevalent':1, 'neoplasm_of_uncertain_behavior_incident':2}, loss=weighted_crossentropy([1.0, 756, 517], 'neoplasm_of_uncertain_behavior_pi'))
neoplasm_of_uncertain_behavior_of_breast_phe_pi = TensorMap('phecode_174.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_174.3_date'), channel_map={'no_neoplasm_of_uncertain_behavior_of_breast':0, 'neoplasm_of_uncertain_behavior_of_breast_prevalent':1, 'neoplasm_of_uncertain_behavior_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 4761, 11627], 'neoplasm_of_uncertain_behavior_of_breast_pi'))
neoplasm_of_uncertain_behavior_of_male_genital_organs_phe_pi = TensorMap('phecode_187.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_187.8_date'), channel_map={'no_neoplasm_of_uncertain_behavior_of_male_genital_organs':0, 'neoplasm_of_uncertain_behavior_of_male_genital_organs_prevalent':1, 'neoplasm_of_uncertain_behavior_of_male_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 5813, 10638], 'neoplasm_of_uncertain_behavior_of_male_genital_organs_pi'))
neoplasm_of_uncertain_behavior_of_skin_phe_pi = TensorMap('phecode_173',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_173_date'), channel_map={'no_neoplasm_of_uncertain_behavior_of_skin':0, 'neoplasm_of_uncertain_behavior_of_skin_prevalent':1, 'neoplasm_of_uncertain_behavior_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 3067, 7462], 'neoplasm_of_uncertain_behavior_of_skin_pi'))
neoplasm_of_unspecified_nature_of_digestive_system_phe_pi = TensorMap('phecode_158',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_158_date'), channel_map={'no_neoplasm_of_unspecified_nature_of_digestive_system':0, 'neoplasm_of_unspecified_nature_of_digestive_system_prevalent':1, 'neoplasm_of_unspecified_nature_of_digestive_system_incident':2}, loss=weighted_crossentropy([1.0, 986, 467], 'neoplasm_of_unspecified_nature_of_digestive_system_pi'))
nephritis_and_nephropathy_with_pathological_lesion_phe_pi = TensorMap('phecode_580.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_580.32_date'), channel_map={'no_nephritis_and_nephropathy_with_pathological_lesion':0, 'nephritis_and_nephropathy_with_pathological_lesion_prevalent':1, 'nephritis_and_nephropathy_with_pathological_lesion_incident':2}, loss=weighted_crossentropy([1.0, 34, 30], 'nephritis_and_nephropathy_with_pathological_lesion_pi'))
nephritis_nephrosis_renal_sclerosis_phe_pi = TensorMap('phecode_580',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_580_date'), channel_map={'no_nephritis_nephrosis_renal_sclerosis':0, 'nephritis_nephrosis_renal_sclerosis_prevalent':1, 'nephritis_nephrosis_renal_sclerosis_incident':2}, loss=weighted_crossentropy([1.0, 5102, 3623], 'nephritis_nephrosis_renal_sclerosis_pi'))
nephrotic_syndrome_without_mention_of_glomerulonephritis_phe_pi = TensorMap('phecode_580.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_580.2_date'), channel_map={'no_nephrotic_syndrome_without_mention_of_glomerulonephritis':0, 'nephrotic_syndrome_without_mention_of_glomerulonephritis_prevalent':1, 'nephrotic_syndrome_without_mention_of_glomerulonephritis_incident':2}, loss=weighted_crossentropy([1.0, 1773, 959], 'nephrotic_syndrome_without_mention_of_glomerulonephritis_pi'))
nerve_plexus_lesions_phe_pi = TensorMap('phecode_353.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_353.1_date'), channel_map={'no_nerve_plexus_lesions':0, 'nerve_plexus_lesions_prevalent':1, 'nerve_plexus_lesions_incident':2}, loss=weighted_crossentropy([1.0, 6578, 5494], 'nerve_plexus_lesions_pi'))
nerve_root_and_plexus_disorders_phe_pi = TensorMap('phecode_353',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_353_date'), channel_map={'no_nerve_root_and_plexus_disorders':0, 'nerve_root_and_plexus_disorders_prevalent':1, 'nerve_root_and_plexus_disorders_incident':2}, loss=weighted_crossentropy([1.0, 1366, 410], 'nerve_root_and_plexus_disorders_pi'))
nerve_root_lesions_phe_pi = TensorMap('phecode_353.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_353.2_date'), channel_map={'no_nerve_root_lesions':0, 'nerve_root_lesions_prevalent':1, 'nerve_root_lesions_incident':2}, loss=weighted_crossentropy([1.0, 12195, 11627], 'nerve_root_lesions_pi'))
nervous_system_congenital_anomalies_phe_pi = TensorMap('phecode_752',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_752_date'), channel_map={'no_nervous_system_congenital_anomalies':0, 'nervous_system_congenital_anomalies_prevalent':1, 'nervous_system_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 142, 129], 'nervous_system_congenital_anomalies_pi'))
neural_tube_defects_phe_pi = TensorMap('phecode_752.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_752.1_date'), channel_map={'no_neural_tube_defects':0, 'neural_tube_defects_prevalent':1, 'neural_tube_defects_incident':2}, loss=weighted_crossentropy([1.0, 13888, 9090], 'neural_tube_defects_pi'))
neuralgia_neuritis_and_radiculitis_nos_phe_pi = TensorMap('phecode_766',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_766_date'), channel_map={'no_neuralgia_neuritis_and_radiculitis_nos':0, 'neuralgia_neuritis_and_radiculitis_nos_prevalent':1, 'neuralgia_neuritis_and_radiculitis_nos_incident':2}, loss=weighted_crossentropy([1.0, 2808, 1618], 'neuralgia_neuritis_and_radiculitis_nos_pi'))
neurofibromatosis_phe_pi = TensorMap('phecode_199.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_199.4_date'), channel_map={'no_neurofibromatosis':0, 'neurofibromatosis_prevalent':1, 'neurofibromatosis_incident':2}, loss=weighted_crossentropy([1.0, 6410, 9615], 'neurofibromatosis_pi'))
neurological_disorders_phe_pi = TensorMap('phecode_292',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292_date'), channel_map={'no_neurological_disorders':0, 'neurological_disorders_prevalent':1, 'neurological_disorders_incident':2}, loss=weighted_crossentropy([1.0, 4587, 685], 'neurological_disorders_pi'))
neutropenia_phe_pi = TensorMap('phecode_288.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_288.11_date'), channel_map={'no_neutropenia':0, 'neutropenia_prevalent':1, 'neutropenia_incident':2}, loss=weighted_crossentropy([1.0, 433, 132], 'neutropenia_pi'))
nevus_non_neoplastic_phe_pi = TensorMap('phecode_217.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_217.1_date'), channel_map={'no_nevus_non_neoplastic':0, 'nevus_non_neoplastic_prevalent':1, 'nevus_non_neoplastic_incident':2}, loss=weighted_crossentropy([1.0, 1529, 1063], 'nevus_non_neoplastic_pi'))
nodular_lymphoma_phe_pi = TensorMap('phecode_202.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_202.21_date'), channel_map={'no_nodular_lymphoma':0, 'nodular_lymphoma_prevalent':1, 'nodular_lymphoma_incident':2}, loss=weighted_crossentropy([1.0, 2252, 1633], 'nodular_lymphoma_pi'))
non_autoimmune_hemolytic_anemias_phe_pi = TensorMap('phecode_283.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_283.2_date'), channel_map={'no_non_autoimmune_hemolytic_anemias':0, 'non_autoimmune_hemolytic_anemias_prevalent':1, 'non_autoimmune_hemolytic_anemias_incident':2}, loss=weighted_crossentropy([1.0, 83333, 33333], 'non_autoimmune_hemolytic_anemias_pi'))
non_hodgkins_lymphoma_phe_pi = TensorMap('phecode_202.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_202.2_date'), channel_map={'no_non_hodgkins_lymphoma':0, 'non_hodgkins_lymphoma_prevalent':1, 'non_hodgkins_lymphoma_incident':2}, loss=weighted_crossentropy([1.0, 290, 149], 'non_hodgkins_lymphoma_pi'))
non_proliferative_glomerulonephritis_phe_pi = TensorMap('phecode_580.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_580.12_date'), channel_map={'no_non_proliferative_glomerulonephritis':0, 'non_proliferative_glomerulonephritis_prevalent':1, 'non_proliferative_glomerulonephritis_incident':2}, loss=weighted_crossentropy([1.0, 5555, 4065], 'non_proliferative_glomerulonephritis_pi'))
noninfectious_dermatoses_of_eyelid_phe_pi = TensorMap('phecode_371.33',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_371.33_date'), channel_map={'no_noninfectious_dermatoses_of_eyelid':0, 'noninfectious_dermatoses_of_eyelid_prevalent':1, 'noninfectious_dermatoses_of_eyelid_incident':2}, loss=weighted_crossentropy([1.0, 38461, 55555], 'noninfectious_dermatoses_of_eyelid_pi'))
noninfectious_disorders_of_lymphatic_channels_phe_pi = TensorMap('phecode_450',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_450_date'), channel_map={'no_noninfectious_disorders_of_lymphatic_channels':0, 'noninfectious_disorders_of_lymphatic_channels_prevalent':1, 'noninfectious_disorders_of_lymphatic_channels_incident':2}, loss=weighted_crossentropy([1.0, 1886, 526], 'noninfectious_disorders_of_lymphatic_channels_pi'))
noninfectious_gastroenteritis_phe_pi = TensorMap('phecode_558',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_558_date'), channel_map={'no_noninfectious_gastroenteritis':0, 'noninfectious_gastroenteritis_prevalent':1, 'noninfectious_gastroenteritis_incident':2}, loss=weighted_crossentropy([1.0, 36, 42], 'noninfectious_gastroenteritis_pi'))
noninflammatory_disorders_of_cervix_phe_pi = TensorMap('phecode_619.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_619.3_date'), channel_map={'no_noninflammatory_disorders_of_cervix':0, 'noninflammatory_disorders_of_cervix_prevalent':1, 'noninflammatory_disorders_of_cervix_incident':2}, loss=weighted_crossentropy([1.0, 208, 448], 'noninflammatory_disorders_of_cervix_pi'))
noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament_phe_pi = TensorMap('phecode_619.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_619.1_date'), channel_map={'no_noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament':0, 'noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament_prevalent':1, 'noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament_incident':2}, loss=weighted_crossentropy([1.0, 891, 1146], 'noninflammatory_disorders_of_ovary_fallopian_tube_and_broad_ligament_pi'))
noninflammatory_disorders_of_vagina_phe_pi = TensorMap('phecode_619.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_619.4_date'), channel_map={'no_noninflammatory_disorders_of_vagina':0, 'noninflammatory_disorders_of_vagina_prevalent':1, 'noninflammatory_disorders_of_vagina_incident':2}, loss=weighted_crossentropy([1.0, 362, 712], 'noninflammatory_disorders_of_vagina_pi'))
noninflammatory_disorders_of_vulva_and_perineum_phe_pi = TensorMap('phecode_619.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_619.5_date'), channel_map={'no_noninflammatory_disorders_of_vulva_and_perineum':0, 'noninflammatory_disorders_of_vulva_and_perineum_prevalent':1, 'noninflammatory_disorders_of_vulva_and_perineum_incident':2}, loss=weighted_crossentropy([1.0, 444, 754], 'noninflammatory_disorders_of_vulva_and_perineum_pi'))
noninflammatory_female_genital_disorders_phe_pi = TensorMap('phecode_619',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_619_date'), channel_map={'no_noninflammatory_female_genital_disorders':0, 'noninflammatory_female_genital_disorders_prevalent':1, 'noninflammatory_female_genital_disorders_incident':2}, loss=weighted_crossentropy([1.0, 11, 27], 'noninflammatory_female_genital_disorders_pi'))
nonrheumatic_aortic_valve_disorders_phe_pi = TensorMap('phecode_395.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_395.2_date'), channel_map={'no_nonrheumatic_aortic_valve_disorders':0, 'nonrheumatic_aortic_valve_disorders_prevalent':1, 'nonrheumatic_aortic_valve_disorders_incident':2}, loss=weighted_crossentropy([1.0, 4424, 1712], 'nonrheumatic_aortic_valve_disorders_pi'))
nonrheumatic_mitral_valve_disorders_phe_pi = TensorMap('phecode_395.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_395.1_date'), channel_map={'no_nonrheumatic_mitral_valve_disorders':0, 'nonrheumatic_mitral_valve_disorders_prevalent':1, 'nonrheumatic_mitral_valve_disorders_incident':2}, loss=weighted_crossentropy([1.0, 418, 140], 'nonrheumatic_mitral_valve_disorders_pi'))
nonrheumatic_pulmonary_valve_disorders_phe_pi = TensorMap('phecode_395.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_395.4_date'), channel_map={'no_nonrheumatic_pulmonary_valve_disorders':0, 'nonrheumatic_pulmonary_valve_disorders_prevalent':1, 'nonrheumatic_pulmonary_valve_disorders_incident':2}, loss=weighted_crossentropy([1.0, 20000, 2906], 'nonrheumatic_pulmonary_valve_disorders_pi'))
nonrheumatic_tricuspid_valve_disorders_phe_pi = TensorMap('phecode_395.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_395.3_date'), channel_map={'no_nonrheumatic_tricuspid_valve_disorders':0, 'nonrheumatic_tricuspid_valve_disorders_prevalent':1, 'nonrheumatic_tricuspid_valve_disorders_incident':2}, loss=weighted_crossentropy([1.0, 16666, 2958], 'nonrheumatic_tricuspid_valve_disorders_pi'))
nonsenile_cataract_phe_pi = TensorMap('phecode_366.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_366.1_date'), channel_map={'no_nonsenile_cataract':0, 'nonsenile_cataract_prevalent':1, 'nonsenile_cataract_incident':2}, loss=weighted_crossentropy([1.0, 4950, 38461], 'nonsenile_cataract_pi'))
nonspecific_abnormal_findings_in_cerebrospinal_fluid_phe_pi = TensorMap('phecode_346.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_346.3_date'), channel_map={'no_nonspecific_abnormal_findings_in_cerebrospinal_fluid':0, 'nonspecific_abnormal_findings_in_cerebrospinal_fluid_prevalent':1, 'nonspecific_abnormal_findings_in_cerebrospinal_fluid_incident':2}, loss=weighted_crossentropy([1.0, 241, 803], 'nonspecific_abnormal_findings_in_cerebrospinal_fluid_pi'))
nonspecific_abnormal_findings_in_stool_contents_phe_pi = TensorMap('phecode_579.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_579.8_date'), channel_map={'no_nonspecific_abnormal_findings_in_stool_contents':0, 'nonspecific_abnormal_findings_in_stool_contents_prevalent':1, 'nonspecific_abnormal_findings_in_stool_contents_incident':2}, loss=weighted_crossentropy([1.0, 1176, 203], 'nonspecific_abnormal_findings_in_stool_contents_pi'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract_phe_pi = TensorMap('phecode_575.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_575.9_date'), channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract_prevalent':1, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract_incident':2}, loss=weighted_crossentropy([1.0, 3355, 1213], 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_biliary_tract_pi'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system_phe_pi = TensorMap('phecode_793',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_793_date'), channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system_prevalent':1, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system_incident':2}, loss=weighted_crossentropy([1.0, 9433, 2994], 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_musculoskeletal_system_pi'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc_phe_pi = TensorMap('phecode_793.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_793.2_date'), channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc_prevalent':1, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc_incident':2}, loss=weighted_crossentropy([1.0, 2617, 692], 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_other_intrathoracic_organs_echocardiogram_etc_pi'))
nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head_phe_pi = TensorMap('phecode_346.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_346.1_date'), channel_map={'no_nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head':0, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head_prevalent':1, 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head_incident':2}, loss=weighted_crossentropy([1.0, 8928, 2958], 'nonspecific_abnormal_findings_on_radiological_and_other_examination_of_skull_and_head_pi'))
nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system_phe_pi = TensorMap('phecode_346.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_346.2_date'), channel_map={'no_nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system':0, 'nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system_prevalent':1, 'nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 38461, 27777], 'nonspecific_abnormal_results_of_function_study_of_brain_and_central_nervous_system_pi'))
nonspecific_abnormal_results_of_other_endocrine_function_study_phe_pi = TensorMap('phecode_259.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_259.1_date'), channel_map={'no_nonspecific_abnormal_results_of_other_endocrine_function_study':0, 'nonspecific_abnormal_results_of_other_endocrine_function_study_prevalent':1, 'nonspecific_abnormal_results_of_other_endocrine_function_study_incident':2}, loss=weighted_crossentropy([1.0, 33333, 17241], 'nonspecific_abnormal_results_of_other_endocrine_function_study_pi'))
nonspecific_chest_pain_phe_pi = TensorMap('phecode_418',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_418_date'), channel_map={'no_nonspecific_chest_pain':0, 'nonspecific_chest_pain_prevalent':1, 'nonspecific_chest_pain_incident':2}, loss=weighted_crossentropy([1.0, 22, 26], 'nonspecific_chest_pain_pi'))
nonspecific_findings_on_examination_of_blood_phe_pi = TensorMap('phecode_790',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_790_date'), channel_map={'no_nonspecific_findings_on_examination_of_blood':0, 'nonspecific_findings_on_examination_of_blood_prevalent':1, 'nonspecific_findings_on_examination_of_blood_incident':2}, loss=weighted_crossentropy([1.0, 6410, 2590], 'nonspecific_findings_on_examination_of_blood_pi'))
nontoxic_multinodular_goiter_phe_pi = TensorMap('phecode_241.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_241.2_date'), channel_map={'no_nontoxic_multinodular_goiter':0, 'nontoxic_multinodular_goiter_prevalent':1, 'nontoxic_multinodular_goiter_incident':2}, loss=weighted_crossentropy([1.0, 1061, 939], 'nontoxic_multinodular_goiter_pi'))
nontoxic_nodular_goiter_phe_pi = TensorMap('phecode_241',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_241_date'), channel_map={'no_nontoxic_nodular_goiter':0, 'nontoxic_nodular_goiter_prevalent':1, 'nontoxic_nodular_goiter_incident':2}, loss=weighted_crossentropy([1.0, 100000, 62500], 'nontoxic_nodular_goiter_pi'))
nontoxic_uninodular_goiter_phe_pi = TensorMap('phecode_241.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_241.1_date'), channel_map={'no_nontoxic_uninodular_goiter':0, 'nontoxic_uninodular_goiter_prevalent':1, 'nontoxic_uninodular_goiter_incident':2}, loss=weighted_crossentropy([1.0, 1886, 1054], 'nontoxic_uninodular_goiter_pi'))
normal_delivery_phe_pi = TensorMap('phecode_650',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_650_date'), channel_map={'no_normal_delivery':0, 'normal_delivery_prevalent':1, 'normal_delivery_incident':2}, loss=weighted_crossentropy([1.0, 190, 35714], 'normal_delivery_pi'))
nutritional_marasmus_phe_pi = TensorMap('phecode_260.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_260.22_date'), channel_map={'no_nutritional_marasmus':0, 'nutritional_marasmus_prevalent':1, 'nutritional_marasmus_incident':2}, loss=weighted_crossentropy([1.0, 166666, 55555], 'nutritional_marasmus_pi'))
nystagmus_and_other_irregular_eye_movements_phe_pi = TensorMap('phecode_378.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_378.2_date'), channel_map={'no_nystagmus_and_other_irregular_eye_movements':0, 'nystagmus_and_other_irregular_eye_movements_prevalent':1, 'nystagmus_and_other_irregular_eye_movements_incident':2}, loss=weighted_crossentropy([1.0, 4629, 2164], 'nystagmus_and_other_irregular_eye_movements_pi'))
obesity_phe_pi = TensorMap('phecode_278.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_278.1_date'), channel_map={'no_obesity':0, 'obesity_prevalent':1, 'obesity_incident':2}, loss=weighted_crossentropy([1.0, 148, 31], 'obesity_pi'))
obsessive_compulsive_disorders_phe_pi = TensorMap('phecode_300.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.3_date'), channel_map={'no_obsessive_compulsive_disorders':0, 'obsessive_compulsive_disorders_prevalent':1, 'obsessive_compulsive_disorders_incident':2}, loss=weighted_crossentropy([1.0, 6172, 3424], 'obsessive_compulsive_disorders_pi'))
obstetrical_birth_trauma_phe_pi = TensorMap('phecode_665',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_665_date'), channel_map={'no_obstetrical_birth_trauma':0, 'obstetrical_birth_trauma_prevalent':1, 'obstetrical_birth_trauma_incident':2}, loss=weighted_crossentropy([1.0, 73, 4237], 'obstetrical_birth_trauma_pi'))
obstruction_of_bile_duct_phe_pi = TensorMap('phecode_575.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_575.2_date'), channel_map={'no_obstruction_of_bile_duct':0, 'obstruction_of_bile_duct_prevalent':1, 'obstruction_of_bile_duct_incident':2}, loss=weighted_crossentropy([1.0, 1557, 538], 'obstruction_of_bile_duct_pi'))
obstructive_chronic_bronchitis_phe_pi = TensorMap('phecode_496.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_496.21_date'), channel_map={'no_obstructive_chronic_bronchitis':0, 'obstructive_chronic_bronchitis_prevalent':1, 'obstructive_chronic_bronchitis_incident':2}, loss=weighted_crossentropy([1.0, 530, 143], 'obstructive_chronic_bronchitis_pi'))
obstructive_genitourinary_defect_phe_pi = TensorMap('phecode_751.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.3_date'), channel_map={'no_obstructive_genitourinary_defect':0, 'obstructive_genitourinary_defect_prevalent':1, 'obstructive_genitourinary_defect_incident':2}, loss=weighted_crossentropy([1.0, 29411, 45454], 'obstructive_genitourinary_defect_pi'))
occlusion_and_stenosis_of_precerebral_arteries_phe_pi = TensorMap('phecode_433.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.1_date'), channel_map={'no_occlusion_and_stenosis_of_precerebral_arteries':0, 'occlusion_and_stenosis_of_precerebral_arteries_prevalent':1, 'occlusion_and_stenosis_of_precerebral_arteries_incident':2}, loss=weighted_crossentropy([1.0, 1089, 468], 'occlusion_and_stenosis_of_precerebral_arteries_pi'))
occlusion_of_cerebral_arteries_phe_pi = TensorMap('phecode_433.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.2_date'), channel_map={'no_occlusion_of_cerebral_arteries':0, 'occlusion_of_cerebral_arteries_prevalent':1, 'occlusion_of_cerebral_arteries_incident':2}, loss=weighted_crossentropy([1.0, 425, 149], 'occlusion_of_cerebral_arteries_pi'))
occlusion_of_cerebral_arteries_with_cerebral_infarction_phe_pi = TensorMap('phecode_433.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.11_date'), channel_map={'no_occlusion_of_cerebral_arteries_with_cerebral_infarction':0, 'occlusion_of_cerebral_arteries_with_cerebral_infarction_prevalent':1, 'occlusion_of_cerebral_arteries_with_cerebral_infarction_incident':2}, loss=weighted_crossentropy([1.0, 12500, 2906], 'occlusion_of_cerebral_arteries_with_cerebral_infarction_pi'))
oliguria_and_anuria_phe_pi = TensorMap('phecode_599.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.6_date'), channel_map={'no_oliguria_and_anuria':0, 'oliguria_and_anuria_prevalent':1, 'oliguria_and_anuria_incident':2}, loss=weighted_crossentropy([1.0, 17857, 3311], 'oliguria_and_anuria_pi'))
open_wound_of_ear_phe_pi = TensorMap('phecode_870.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.2_date'), channel_map={'no_open_wound_of_ear':0, 'open_wound_of_ear_prevalent':1, 'open_wound_of_ear_incident':2}, loss=weighted_crossentropy([1.0, 4000, 4166], 'open_wound_of_ear_pi'))
open_wound_of_foot_except_toe_s_alone_phe_pi = TensorMap('phecode_871.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_871.3_date'), channel_map={'no_open_wound_of_foot_except_toe_s_alone':0, 'open_wound_of_foot_except_toe_s_alone_prevalent':1, 'open_wound_of_foot_except_toe_s_alone_incident':2}, loss=weighted_crossentropy([1.0, 4166, 4098], 'open_wound_of_foot_except_toe_s_alone_pi'))
open_wound_of_genital_organs_phe_pi = TensorMap('phecode_870.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.8_date'), channel_map={'no_open_wound_of_genital_organs':0, 'open_wound_of_genital_organs_prevalent':1, 'open_wound_of_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 15151, 10869], 'open_wound_of_genital_organs_pi'))
open_wound_of_lip_and_mouth_phe_pi = TensorMap('phecode_870.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.5_date'), channel_map={'no_open_wound_of_lip_and_mouth':0, 'open_wound_of_lip_and_mouth_prevalent':1, 'open_wound_of_lip_and_mouth_incident':2}, loss=weighted_crossentropy([1.0, 2450, 1440], 'open_wound_of_lip_and_mouth_pi'))
open_wound_of_neck_phe_pi = TensorMap('phecode_870.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.6_date'), channel_map={'no_open_wound_of_neck':0, 'open_wound_of_neck_prevalent':1, 'open_wound_of_neck_incident':2}, loss=weighted_crossentropy([1.0, 5434, 6329], 'open_wound_of_neck_pi'))
open_wound_of_nose_and_sinus_phe_pi = TensorMap('phecode_870.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.4_date'), channel_map={'no_open_wound_of_nose_and_sinus':0, 'open_wound_of_nose_and_sinus_prevalent':1, 'open_wound_of_nose_and_sinus_incident':2}, loss=weighted_crossentropy([1.0, 4098, 2083], 'open_wound_of_nose_and_sinus_pi'))
open_wound_of_toe_s_phe_pi = TensorMap('phecode_871.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_871.4_date'), channel_map={'no_open_wound_of_toe_s':0, 'open_wound_of_toe_s_prevalent':1, 'open_wound_of_toe_s_incident':2}, loss=weighted_crossentropy([1.0, 10638, 7575], 'open_wound_of_toe_s_pi'))
open_wound_or_laceration_of_eye_or_eyelid_phe_pi = TensorMap('phecode_870.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.1_date'), channel_map={'no_open_wound_or_laceration_of_eye_or_eyelid':0, 'open_wound_or_laceration_of_eye_or_eyelid_prevalent':1, 'open_wound_or_laceration_of_eye_or_eyelid_incident':2}, loss=weighted_crossentropy([1.0, 1562, 1160], 'open_wound_or_laceration_of_eye_or_eyelid_pi'))
open_wounds_of_extremities_phe_pi = TensorMap('phecode_871',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_871_date'), channel_map={'no_open_wounds_of_extremities':0, 'open_wounds_of_extremities_prevalent':1, 'open_wounds_of_extremities_incident':2}, loss=weighted_crossentropy([1.0, 168, 169], 'open_wounds_of_extremities_pi'))
open_wounds_of_head_neck_and_trunk_phe_pi = TensorMap('phecode_870',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870_date'), channel_map={'no_open_wounds_of_head_neck_and_trunk':0, 'open_wounds_of_head_neck_and_trunk_prevalent':1, 'open_wounds_of_head_neck_and_trunk_incident':2}, loss=weighted_crossentropy([1.0, 3816, 4166], 'open_wounds_of_head_neck_and_trunk_pi'))
opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use_phe_pi = TensorMap('phecode_965.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_965.1_date'), channel_map={'no_opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use':0, 'opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use_prevalent':1, 'opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use_incident':2}, loss=weighted_crossentropy([1.0, 1805, 292], 'opiates_and_related_narcotics_causing_adverse_effects_in_therapeutic_use_pi'))
optic_atrophy_phe_pi = TensorMap('phecode_377.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_377.1_date'), channel_map={'no_optic_atrophy':0, 'optic_atrophy_prevalent':1, 'optic_atrophy_incident':2}, loss=weighted_crossentropy([1.0, 15151, 7575], 'optic_atrophy_pi'))
optic_neuritis_neuropathy_phe_pi = TensorMap('phecode_377.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_377.3_date'), channel_map={'no_optic_neuritis_neuropathy':0, 'optic_neuritis_neuropathy_prevalent':1, 'optic_neuritis_neuropathy_incident':2}, loss=weighted_crossentropy([1.0, 4385, 3378], 'optic_neuritis_neuropathy_pi'))
oral_aphthae_phe_pi = TensorMap('phecode_528.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.12_date'), channel_map={'no_oral_aphthae':0, 'oral_aphthae_prevalent':1, 'oral_aphthae_incident':2}, loss=weighted_crossentropy([1.0, 8196, 10204], 'oral_aphthae_pi'))
organic_or_persistent_insomnia_phe_pi = TensorMap('phecode_327.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327.41_date'), channel_map={'no_organic_or_persistent_insomnia':0, 'organic_or_persistent_insomnia_prevalent':1, 'organic_or_persistent_insomnia_incident':2}, loss=weighted_crossentropy([1.0, 10869, 2358], 'organic_or_persistent_insomnia_pi'))
orthostatic_hypotension_phe_pi = TensorMap('phecode_458.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_458.1_date'), channel_map={'no_orthostatic_hypotension':0, 'orthostatic_hypotension_prevalent':1, 'orthostatic_hypotension_incident':2}, loss=weighted_crossentropy([1.0, 1474, 238], 'orthostatic_hypotension_pi'))
osteitis_deformans_pagets_disease_of_bone_phe_pi = TensorMap('phecode_731.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_731.1_date'), channel_map={'no_osteitis_deformans_pagets_disease_of_bone':0, 'osteitis_deformans_pagets_disease_of_bone_prevalent':1, 'osteitis_deformans_pagets_disease_of_bone_incident':2}, loss=weighted_crossentropy([1.0, 14285, 11111], 'osteitis_deformans_pagets_disease_of_bone_pi'))
osteoarthritis_localized_phe_pi = TensorMap('phecode_740.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_740.1_date'), channel_map={'no_osteoarthritis_localized':0, 'osteoarthritis_localized_prevalent':1, 'osteoarthritis_localized_incident':2}, loss=weighted_crossentropy([1.0, 160, 45], 'osteoarthritis_localized_pi'))
osteoarthrosis_generalized_phe_pi = TensorMap('phecode_740.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_740.2_date'), channel_map={'no_osteoarthrosis_generalized':0, 'osteoarthrosis_generalized_prevalent':1, 'osteoarthrosis_generalized_incident':2}, loss=weighted_crossentropy([1.0, 1661, 1106], 'osteoarthrosis_generalized_pi'))
osteoarthrosis_localized_primary_phe_pi = TensorMap('phecode_740.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_740.11_date'), channel_map={'no_osteoarthrosis_localized_primary':0, 'osteoarthrosis_localized_primary_prevalent':1, 'osteoarthrosis_localized_primary_incident':2}, loss=weighted_crossentropy([1.0, 76, 95], 'osteoarthrosis_localized_primary_pi'))
osteoarthrosis_localized_secondary_phe_pi = TensorMap('phecode_740.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_740.12_date'), channel_map={'no_osteoarthrosis_localized_secondary':0, 'osteoarthrosis_localized_secondary_prevalent':1, 'osteoarthrosis_localized_secondary_incident':2}, loss=weighted_crossentropy([1.0, 3012, 2427], 'osteoarthrosis_localized_secondary_pi'))
osteoarthrosis_nos_phe_pi = TensorMap('phecode_740.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_740.9_date'), channel_map={'no_osteoarthrosis_nos':0, 'osteoarthrosis_nos_prevalent':1, 'osteoarthrosis_nos_incident':2}, loss=weighted_crossentropy([1.0, 397, 69], 'osteoarthrosis_nos_pi'))
osteochondritis_dissecans_phe_pi = TensorMap('phecode_732.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_732.7_date'), channel_map={'no_osteochondritis_dissecans':0, 'osteochondritis_dissecans_prevalent':1, 'osteochondritis_dissecans_incident':2}, loss=weighted_crossentropy([1.0, 16129, 12195], 'osteochondritis_dissecans_pi'))
osteochondropathies_phe_pi = TensorMap('phecode_732',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_732_date'), channel_map={'no_osteochondropathies':0, 'osteochondropathies_prevalent':1, 'osteochondropathies_incident':2}, loss=weighted_crossentropy([1.0, 4310, 2604], 'osteochondropathies_pi'))
osteomyelitis_periostitis_and_other_infections_involving_bone_phe_pi = TensorMap('phecode_710',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_710_date'), channel_map={'no_osteomyelitis_periostitis_and_other_infections_involving_bone':0, 'osteomyelitis_periostitis_and_other_infections_involving_bone_prevalent':1, 'osteomyelitis_periostitis_and_other_infections_involving_bone_incident':2}, loss=weighted_crossentropy([1.0, 100000, 125000], 'osteomyelitis_periostitis_and_other_infections_involving_bone_pi'))
osteopenia_or_other_disorder_of_bone_and_cartilage_phe_pi = TensorMap('phecode_743.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.9_date'), channel_map={'no_osteopenia_or_other_disorder_of_bone_and_cartilage':0, 'osteopenia_or_other_disorder_of_bone_and_cartilage_prevalent':1, 'osteopenia_or_other_disorder_of_bone_and_cartilage_incident':2}, loss=weighted_crossentropy([1.0, 406, 80], 'osteopenia_or_other_disorder_of_bone_and_cartilage_pi'))
osteoporosis_phe_pi = TensorMap('phecode_743.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.1_date'), channel_map={'no_osteoporosis':0, 'osteoporosis_prevalent':1, 'osteoporosis_incident':2}, loss=weighted_crossentropy([1.0, 22727, 8064], 'osteoporosis_pi'))
osteoporosis_nos_phe_pi = TensorMap('phecode_743.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.11_date'), channel_map={'no_osteoporosis_nos':0, 'osteoporosis_nos_prevalent':1, 'osteoporosis_nos_incident':2}, loss=weighted_crossentropy([1.0, 927, 93], 'osteoporosis_nos_pi'))
otalgia_phe_pi = TensorMap('phecode_382',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_382_date'), channel_map={'no_otalgia':0, 'otalgia_prevalent':1, 'otalgia_incident':2}, loss=weighted_crossentropy([1.0, 2590, 2272], 'otalgia_pi'))
other_abnormal_blood_chemistry_phe_pi = TensorMap('phecode_790.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_790.6_date'), channel_map={'no_other_abnormal_blood_chemistry':0, 'other_abnormal_blood_chemistry_prevalent':1, 'other_abnormal_blood_chemistry_incident':2}, loss=weighted_crossentropy([1.0, 187, 67], 'other_abnormal_blood_chemistry_pi'))
other_abnormal_glucose_phe_pi = TensorMap('phecode_250.42',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.42_date'), channel_map={'no_other_abnormal_glucose':0, 'other_abnormal_glucose_prevalent':1, 'other_abnormal_glucose_incident':2}, loss=weighted_crossentropy([1.0, 1543, 1470], 'other_abnormal_glucose_pi'))
other_abnormality_of_urination_phe_pi = TensorMap('phecode_599.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.9_date'), channel_map={'no_other_abnormality_of_urination':0, 'other_abnormality_of_urination_prevalent':1, 'other_abnormality_of_urination_incident':2}, loss=weighted_crossentropy([1.0, 370, 457], 'other_abnormality_of_urination_pi'))
other_acquired_deformities_of_limbs_phe_pi = TensorMap('phecode_736',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_736_date'), channel_map={'no_other_acquired_deformities_of_limbs':0, 'other_acquired_deformities_of_limbs_prevalent':1, 'other_acquired_deformities_of_limbs_incident':2}, loss=weighted_crossentropy([1.0, 2314, 2824], 'other_acquired_deformities_of_limbs_pi'))
other_acquired_musculoskeletal_deformity_phe_pi = TensorMap('phecode_738',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_738_date'), channel_map={'no_other_acquired_musculoskeletal_deformity':0, 'other_acquired_musculoskeletal_deformity_prevalent':1, 'other_acquired_musculoskeletal_deformity_incident':2}, loss=weighted_crossentropy([1.0, 881, 841], 'other_acquired_musculoskeletal_deformity_pi'))
other_acute_and_subacute_forms_of_ischemic_heart_disease_phe_pi = TensorMap('phecode_411.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.9_date'), channel_map={'no_other_acute_and_subacute_forms_of_ischemic_heart_disease':0, 'other_acute_and_subacute_forms_of_ischemic_heart_disease_prevalent':1, 'other_acute_and_subacute_forms_of_ischemic_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 1440, 298], 'other_acute_and_subacute_forms_of_ischemic_heart_disease_pi'))
other_allied_disorders_of_spine_phe_pi = TensorMap('phecode_721.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_721.8_date'), channel_map={'no_other_allied_disorders_of_spine':0, 'other_allied_disorders_of_spine_prevalent':1, 'other_allied_disorders_of_spine_incident':2}, loss=weighted_crossentropy([1.0, 47, 33], 'other_allied_disorders_of_spine_pi'))
other_alveolar_and_parietoalveolar_pneumonopathy_phe_pi = TensorMap('phecode_504',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_504_date'), channel_map={'no_other_alveolar_and_parietoalveolar_pneumonopathy':0, 'other_alveolar_and_parietoalveolar_pneumonopathy_prevalent':1, 'other_alveolar_and_parietoalveolar_pneumonopathy_incident':2}, loss=weighted_crossentropy([1.0, 5434, 1118], 'other_alveolar_and_parietoalveolar_pneumonopathy_pi'))
other_and_unspecified_coagulation_defects_phe_pi = TensorMap('phecode_286.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.7_date'), channel_map={'no_other_and_unspecified_coagulation_defects':0, 'other_and_unspecified_coagulation_defects_prevalent':1, 'other_and_unspecified_coagulation_defects_incident':2}, loss=weighted_crossentropy([1.0, 1838, 1666], 'other_and_unspecified_coagulation_defects_pi'))
other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother_phe_pi = TensorMap('phecode_654',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_654_date'), channel_map={'no_other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother':0, 'other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother_prevalent':1, 'other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother_incident':2}, loss=weighted_crossentropy([1.0, 3184, 45454], 'other_and_unspecified_complications_of_birth_puerperium_affecting_management_of_mother_pi'))
other_and_unspecified_congenital_anomalies_phe_pi = TensorMap('phecode_759',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_759_date'), channel_map={'no_other_and_unspecified_congenital_anomalies':0, 'other_and_unspecified_congenital_anomalies_prevalent':1, 'other_and_unspecified_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 992, 1237], 'other_and_unspecified_congenital_anomalies_pi'))
other_and_unspecified_disc_disorder_phe_pi = TensorMap('phecode_722.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722.9_date'), channel_map={'no_other_and_unspecified_disc_disorder':0, 'other_and_unspecified_disc_disorder_prevalent':1, 'other_and_unspecified_disc_disorder_incident':2}, loss=weighted_crossentropy([1.0, 244, 133], 'other_and_unspecified_disc_disorder_pi'))
other_and_unspecified_disorders_of_the_nervous_system_phe_pi = TensorMap('phecode_349',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_349_date'), channel_map={'no_other_and_unspecified_disorders_of_the_nervous_system':0, 'other_and_unspecified_disorders_of_the_nervous_system_prevalent':1, 'other_and_unspecified_disorders_of_the_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 20, 15], 'other_and_unspecified_disorders_of_the_nervous_system_pi'))
other_anemias_phe_pi = TensorMap('phecode_285',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_285_date'), channel_map={'no_other_anemias':0, 'other_anemias_prevalent':1, 'other_anemias_incident':2}, loss=weighted_crossentropy([1.0, 81, 41], 'other_anemias_pi'))
other_aneurysm_phe_pi = TensorMap('phecode_442',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_442_date'), channel_map={'no_other_aneurysm':0, 'other_aneurysm_prevalent':1, 'other_aneurysm_incident':2}, loss=weighted_crossentropy([1.0, 20833, 21739], 'other_aneurysm_pi'))
other_arthropathies_phe_pi = TensorMap('phecode_716',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_716_date'), channel_map={'no_other_arthropathies':0, 'other_arthropathies_prevalent':1, 'other_arthropathies_incident':2}, loss=weighted_crossentropy([1.0, 13513, 8064], 'other_arthropathies_pi'))
other_benign_neoplasm_of_connective_and_other_soft_tissue_phe_pi = TensorMap('phecode_215',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_215_date'), channel_map={'no_other_benign_neoplasm_of_connective_and_other_soft_tissue':0, 'other_benign_neoplasm_of_connective_and_other_soft_tissue_prevalent':1, 'other_benign_neoplasm_of_connective_and_other_soft_tissue_incident':2}, loss=weighted_crossentropy([1.0, 611, 780], 'other_benign_neoplasm_of_connective_and_other_soft_tissue_pi'))
other_benign_neoplasm_of_uterus_phe_pi = TensorMap('phecode_218.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_218.2_date'), channel_map={'no_other_benign_neoplasm_of_uterus':0, 'other_benign_neoplasm_of_uterus_prevalent':1, 'other_benign_neoplasm_of_uterus_incident':2}, loss=weighted_crossentropy([1.0, 1457, 3787], 'other_benign_neoplasm_of_uterus_pi'))
other_cardiac_conduction_disorders_phe_pi = TensorMap('phecode_426.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.8_date'), channel_map={'no_other_cardiac_conduction_disorders':0, 'other_cardiac_conduction_disorders_prevalent':1, 'other_cardiac_conduction_disorders_incident':2}, loss=weighted_crossentropy([1.0, 29411, 5882], 'other_cardiac_conduction_disorders_pi'))
other_cardiomyopathy_phe_pi = TensorMap('phecode_425.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_425.8_date'), channel_map={'no_other_cardiomyopathy':0, 'other_cardiomyopathy_prevalent':1, 'other_cardiomyopathy_incident':2}, loss=weighted_crossentropy([1.0, 13513, 10869], 'other_cardiomyopathy_pi'))
other_cerebral_degenerations_phe_pi = TensorMap('phecode_331',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_331_date'), channel_map={'no_other_cerebral_degenerations':0, 'other_cerebral_degenerations_prevalent':1, 'other_cerebral_degenerations_incident':2}, loss=weighted_crossentropy([1.0, 8474, 1412], 'other_cerebral_degenerations_pi'))
other_chronic_ischemic_heart_disease_unspecified_phe_pi = TensorMap('phecode_411.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.8_date'), channel_map={'no_other_chronic_ischemic_heart_disease_unspecified':0, 'other_chronic_ischemic_heart_disease_unspecified_prevalent':1, 'other_chronic_ischemic_heart_disease_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 24, 21], 'other_chronic_ischemic_heart_disease_unspecified_pi'))
other_chronic_nonalcoholic_liver_disease_phe_pi = TensorMap('phecode_571.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_571.5_date'), channel_map={'no_other_chronic_nonalcoholic_liver_disease':0, 'other_chronic_nonalcoholic_liver_disease_prevalent':1, 'other_chronic_nonalcoholic_liver_disease_incident':2}, loss=weighted_crossentropy([1.0, 888, 216], 'other_chronic_nonalcoholic_liver_disease_pi'))
other_cns_infection_and_poliomyelitis_phe_pi = TensorMap('phecode_324',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_324_date'), channel_map={'no_other_cns_infection_and_poliomyelitis':0, 'other_cns_infection_and_poliomyelitis_prevalent':1, 'other_cns_infection_and_poliomyelitis_incident':2}, loss=weighted_crossentropy([1.0, 293, 53], 'other_cns_infection_and_poliomyelitis_pi'))
other_complications_of_pregnancy_nec_phe_pi = TensorMap('phecode_646',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_646_date'), channel_map={'no_other_complications_of_pregnancy_nec':0, 'other_complications_of_pregnancy_nec_prevalent':1, 'other_complications_of_pregnancy_nec_incident':2}, loss=weighted_crossentropy([1.0, 161, 4950], 'other_complications_of_pregnancy_nec_pi'))
other_complications_of_the_puerperium_nec_phe_pi = TensorMap('phecode_674',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_674_date'), channel_map={'no_other_complications_of_the_puerperium_nec':0, 'other_complications_of_the_puerperium_nec_prevalent':1, 'other_complications_of_the_puerperium_nec_incident':2}, loss=weighted_crossentropy([1.0, 1497, 16666], 'other_complications_of_the_puerperium_nec_pi'))
other_conditions_of_brain_phe_pi = TensorMap('phecode_348',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_348_date'), channel_map={'no_other_conditions_of_brain':0, 'other_conditions_of_brain_prevalent':1, 'other_conditions_of_brain_incident':2}, loss=weighted_crossentropy([1.0, 25000, 2369], 'other_conditions_of_brain_pi'))
other_conditions_of_brain_nos_phe_pi = TensorMap('phecode_348.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_348.9_date'), channel_map={'no_other_conditions_of_brain_nos':0, 'other_conditions_of_brain_nos_prevalent':1, 'other_conditions_of_brain_nos_incident':2}, loss=weighted_crossentropy([1.0, 4065, 961], 'other_conditions_of_brain_nos_pi'))
other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium_phe_pi = TensorMap('phecode_649',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_649_date'), channel_map={'no_other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium':0, 'other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium_prevalent':1, 'other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium_incident':2}, loss=weighted_crossentropy([1.0, 3448, 45454], 'other_conditions_or_status_of_the_mother_complicating_pregnancy_childbirth_or_the_puerperium_pi'))
other_congenital_anomalies_of_lower_limb_including_pelvic_girdle_phe_pi = TensorMap('phecode_755.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_755.6_date'), channel_map={'no_other_congenital_anomalies_of_lower_limb_including_pelvic_girdle':0, 'other_congenital_anomalies_of_lower_limb_including_pelvic_girdle_prevalent':1, 'other_congenital_anomalies_of_lower_limb_including_pelvic_girdle_incident':2}, loss=weighted_crossentropy([1.0, 20000, 9615], 'other_congenital_anomalies_of_lower_limb_including_pelvic_girdle_pi'))
other_congenital_musculoskeletal_anomalies_phe_pi = TensorMap('phecode_756',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_756_date'), channel_map={'no_other_congenital_musculoskeletal_anomalies':0, 'other_congenital_musculoskeletal_anomalies_prevalent':1, 'other_congenital_musculoskeletal_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 968, 814], 'other_congenital_musculoskeletal_anomalies_pi'))
other_deficiency_anemia_phe_pi = TensorMap('phecode_281',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_281_date'), channel_map={'no_other_deficiency_anemia':0, 'other_deficiency_anemia_prevalent':1, 'other_deficiency_anemia_incident':2}, loss=weighted_crossentropy([1.0, 37, 19], 'other_deficiency_anemia_pi'))
other_degenerative_diseases_of_the_basal_ganglia_phe_pi = TensorMap('phecode_333.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_333.8_date'), channel_map={'no_other_degenerative_diseases_of_the_basal_ganglia':0, 'other_degenerative_diseases_of_the_basal_ganglia_prevalent':1, 'other_degenerative_diseases_of_the_basal_ganglia_incident':2}, loss=weighted_crossentropy([1.0, 50000, 6578], 'other_degenerative_diseases_of_the_basal_ganglia_pi'))
other_demyelinating_diseases_of_central_nervous_system_phe_pi = TensorMap('phecode_341',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_341_date'), channel_map={'no_other_demyelinating_diseases_of_central_nervous_system':0, 'other_demyelinating_diseases_of_central_nervous_system_prevalent':1, 'other_demyelinating_diseases_of_central_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 402, 632], 'other_demyelinating_diseases_of_central_nervous_system_pi'))
other_derangement_of_joint_phe_pi = TensorMap('phecode_742.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_742.9_date'), channel_map={'no_other_derangement_of_joint':0, 'other_derangement_of_joint_prevalent':1, 'other_derangement_of_joint_incident':2}, loss=weighted_crossentropy([1.0, 46, 33], 'other_derangement_of_joint_pi'))
other_diseases_of_blood_and_blood_forming_organs_phe_pi = TensorMap('phecode_289',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_289_date'), channel_map={'no_other_diseases_of_blood_and_blood_forming_organs':0, 'other_diseases_of_blood_and_blood_forming_organs_prevalent':1, 'other_diseases_of_blood_and_blood_forming_organs_incident':2}, loss=weighted_crossentropy([1.0, 248, 94], 'other_diseases_of_blood_and_blood_forming_organs_pi'))
other_diseases_of_lung_phe_pi = TensorMap('phecode_510',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_510_date'), channel_map={'no_other_diseases_of_lung':0, 'other_diseases_of_lung_prevalent':1, 'other_diseases_of_lung_incident':2}, loss=weighted_crossentropy([1.0, 2659, 389], 'other_diseases_of_lung_pi'))
other_diseases_of_respiratory_system_nec_phe_pi = TensorMap('phecode_519.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_519.8_date'), channel_map={'no_other_diseases_of_respiratory_system_nec':0, 'other_diseases_of_respiratory_system_nec_prevalent':1, 'other_diseases_of_respiratory_system_nec_incident':2}, loss=weighted_crossentropy([1.0, 109, 54], 'other_diseases_of_respiratory_system_nec_pi'))
other_diseases_of_respiratory_system_not_elsewhere_classified_phe_pi = TensorMap('phecode_519',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_519_date'), channel_map={'no_other_diseases_of_respiratory_system_not_elsewhere_classified':0, 'other_diseases_of_respiratory_system_not_elsewhere_classified_prevalent':1, 'other_diseases_of_respiratory_system_not_elsewhere_classified_incident':2}, loss=weighted_crossentropy([1.0, 13, 9], 'other_diseases_of_respiratory_system_not_elsewhere_classified_pi'))
other_diseases_of_the_teeth_and_supporting_structures_phe_pi = TensorMap('phecode_525',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_525_date'), channel_map={'no_other_diseases_of_the_teeth_and_supporting_structures':0, 'other_diseases_of_the_teeth_and_supporting_structures_prevalent':1, 'other_diseases_of_the_teeth_and_supporting_structures_incident':2}, loss=weighted_crossentropy([1.0, 303, 231], 'other_diseases_of_the_teeth_and_supporting_structures_pi'))
other_disorders_of_arteries_and_arterioles_phe_pi = TensorMap('phecode_447',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_447_date'), channel_map={'no_other_disorders_of_arteries_and_arterioles':0, 'other_disorders_of_arteries_and_arterioles_prevalent':1, 'other_disorders_of_arteries_and_arterioles_incident':2}, loss=weighted_crossentropy([1.0, 1666, 945], 'other_disorders_of_arteries_and_arterioles_pi'))
other_disorders_of_biliary_tract_phe_pi = TensorMap('phecode_575.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_575.8_date'), channel_map={'no_other_disorders_of_biliary_tract':0, 'other_disorders_of_biliary_tract_prevalent':1, 'other_disorders_of_biliary_tract_incident':2}, loss=weighted_crossentropy([1.0, 1008, 458], 'other_disorders_of_biliary_tract_pi'))
other_disorders_of_bladder_phe_pi = TensorMap('phecode_596',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_596_date'), channel_map={'no_other_disorders_of_bladder':0, 'other_disorders_of_bladder_prevalent':1, 'other_disorders_of_bladder_incident':2}, loss=weighted_crossentropy([1.0, 118, 79], 'other_disorders_of_bladder_pi'))
other_disorders_of_bone_and_cartilage_phe_pi = TensorMap('phecode_733',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_733_date'), channel_map={'no_other_disorders_of_bone_and_cartilage':0, 'other_disorders_of_bone_and_cartilage_prevalent':1, 'other_disorders_of_bone_and_cartilage_incident':2}, loss=weighted_crossentropy([1.0, 1937, 536], 'other_disorders_of_bone_and_cartilage_pi'))
other_disorders_of_carbohydrate_transport_and_metabolism_phe_pi = TensorMap('phecode_271.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_271.9_date'), channel_map={'no_other_disorders_of_carbohydrate_transport_and_metabolism':0, 'other_disorders_of_carbohydrate_transport_and_metabolism_prevalent':1, 'other_disorders_of_carbohydrate_transport_and_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 29411, 25000], 'other_disorders_of_carbohydrate_transport_and_metabolism_pi'))
other_disorders_of_cervical_region_phe_pi = TensorMap('phecode_723',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_723_date'), channel_map={'no_other_disorders_of_cervical_region':0, 'other_disorders_of_cervical_region_prevalent':1, 'other_disorders_of_cervical_region_incident':2}, loss=weighted_crossentropy([1.0, 2336, 1291], 'other_disorders_of_cervical_region_pi'))
other_disorders_of_circulatory_system_phe_pi = TensorMap('phecode_459',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_459_date'), channel_map={'no_other_disorders_of_circulatory_system':0, 'other_disorders_of_circulatory_system_prevalent':1, 'other_disorders_of_circulatory_system_incident':2}, loss=weighted_crossentropy([1.0, 9259, 3311], 'other_disorders_of_circulatory_system_pi'))
other_disorders_of_ear_phe_pi = TensorMap('phecode_388',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_388_date'), channel_map={'no_other_disorders_of_ear':0, 'other_disorders_of_ear_prevalent':1, 'other_disorders_of_ear_incident':2}, loss=weighted_crossentropy([1.0, 88, 56], 'other_disorders_of_ear_pi'))
other_disorders_of_eye_phe_pi = TensorMap('phecode_379',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379_date'), channel_map={'no_other_disorders_of_eye':0, 'other_disorders_of_eye_prevalent':1, 'other_disorders_of_eye_incident':2}, loss=weighted_crossentropy([1.0, 30, 14], 'other_disorders_of_eye_pi'))
other_disorders_of_eyelids_phe_pi = TensorMap('phecode_374',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_374_date'), channel_map={'no_other_disorders_of_eyelids':0, 'other_disorders_of_eyelids_prevalent':1, 'other_disorders_of_eyelids_incident':2}, loss=weighted_crossentropy([1.0, 207, 279], 'other_disorders_of_eyelids_pi'))
other_disorders_of_gallbladder_phe_pi = TensorMap('phecode_575.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_575.7_date'), channel_map={'no_other_disorders_of_gallbladder':0, 'other_disorders_of_gallbladder_prevalent':1, 'other_disorders_of_gallbladder_incident':2}, loss=weighted_crossentropy([1.0, 660, 393], 'other_disorders_of_gallbladder_pi'))
other_disorders_of_intestine_phe_pi = TensorMap('phecode_569',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_569_date'), channel_map={'no_other_disorders_of_intestine':0, 'other_disorders_of_intestine_prevalent':1, 'other_disorders_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 4, 5], 'other_disorders_of_intestine_pi'))
other_disorders_of_lipoid_metabolism_phe_pi = TensorMap('phecode_277.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_277.5_date'), channel_map={'no_other_disorders_of_lipoid_metabolism':0, 'other_disorders_of_lipoid_metabolism_prevalent':1, 'other_disorders_of_lipoid_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 5952, 2202], 'other_disorders_of_lipoid_metabolism_pi'))
other_disorders_of_liver_phe_pi = TensorMap('phecode_573',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_573_date'), channel_map={'no_other_disorders_of_liver':0, 'other_disorders_of_liver_prevalent':1, 'other_disorders_of_liver_incident':2}, loss=weighted_crossentropy([1.0, 222, 84], 'other_disorders_of_liver_pi'))
other_disorders_of_male_genital_organs_phe_pi = TensorMap('phecode_608',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_608_date'), channel_map={'no_other_disorders_of_male_genital_organs':0, 'other_disorders_of_male_genital_organs_prevalent':1, 'other_disorders_of_male_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 43, 31], 'other_disorders_of_male_genital_organs_pi'))
other_disorders_of_metabolism_phe_pi = TensorMap('phecode_277',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_277_date'), channel_map={'no_other_disorders_of_metabolism':0, 'other_disorders_of_metabolism_prevalent':1, 'other_disorders_of_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 24, 10], 'other_disorders_of_metabolism_pi'))
other_disorders_of_middle_ear_and_mastoid_phe_pi = TensorMap('phecode_385',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_385_date'), channel_map={'no_other_disorders_of_middle_ear_and_mastoid':0, 'other_disorders_of_middle_ear_and_mastoid_prevalent':1, 'other_disorders_of_middle_ear_and_mastoid_incident':2}, loss=weighted_crossentropy([1.0, 4098, 7246], 'other_disorders_of_middle_ear_and_mastoid_pi'))
other_disorders_of_neurohypophysis_phe_pi = TensorMap('phecode_253.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_253.7_date'), channel_map={'no_other_disorders_of_neurohypophysis':0, 'other_disorders_of_neurohypophysis_prevalent':1, 'other_disorders_of_neurohypophysis_incident':2}, loss=weighted_crossentropy([1.0, 12500, 2136], 'other_disorders_of_neurohypophysis_pi'))
other_disorders_of_pancreatic_internal_secretion_phe_pi = TensorMap('phecode_251',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_251_date'), channel_map={'no_other_disorders_of_pancreatic_internal_secretion':0, 'other_disorders_of_pancreatic_internal_secretion_prevalent':1, 'other_disorders_of_pancreatic_internal_secretion_incident':2}, loss=weighted_crossentropy([1.0, 166666, 50000], 'other_disorders_of_pancreatic_internal_secretion_pi'))
other_disorders_of_peritoneum_phe_pi = TensorMap('phecode_568',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_568_date'), channel_map={'no_other_disorders_of_peritoneum':0, 'other_disorders_of_peritoneum_prevalent':1, 'other_disorders_of_peritoneum_incident':2}, loss=weighted_crossentropy([1.0, 264, 120], 'other_disorders_of_peritoneum_pi'))
other_disorders_of_prostate_phe_pi = TensorMap('phecode_602',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_602_date'), channel_map={'no_other_disorders_of_prostate':0, 'other_disorders_of_prostate_prevalent':1, 'other_disorders_of_prostate_incident':2}, loss=weighted_crossentropy([1.0, 698, 431], 'other_disorders_of_prostate_pi'))
other_disorders_of_purine_and_pyrimidine_metabolism_phe_pi = TensorMap('phecode_277.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_277.2_date'), channel_map={'no_other_disorders_of_purine_and_pyrimidine_metabolism':0, 'other_disorders_of_purine_and_pyrimidine_metabolism_prevalent':1, 'other_disorders_of_purine_and_pyrimidine_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 166666, 83333], 'other_disorders_of_purine_and_pyrimidine_metabolism_pi'))
other_disorders_of_soft_tissues_phe_pi = TensorMap('phecode_729',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_729_date'), channel_map={'no_other_disorders_of_soft_tissues':0, 'other_disorders_of_soft_tissues_prevalent':1, 'other_disorders_of_soft_tissues_incident':2}, loss=weighted_crossentropy([1.0, 37, 26], 'other_disorders_of_soft_tissues_pi'))
other_disorders_of_stomach_and_duodenum_phe_pi = TensorMap('phecode_537',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_537_date'), channel_map={'no_other_disorders_of_stomach_and_duodenum':0, 'other_disorders_of_stomach_and_duodenum_prevalent':1, 'other_disorders_of_stomach_and_duodenum_incident':2}, loss=weighted_crossentropy([1.0, 249, 159], 'other_disorders_of_stomach_and_duodenum_pi'))
other_disorders_of_synovium_tendon_and_bursa_phe_pi = TensorMap('phecode_727',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727_date'), channel_map={'no_other_disorders_of_synovium_tendon_and_bursa':0, 'other_disorders_of_synovium_tendon_and_bursa_prevalent':1, 'other_disorders_of_synovium_tendon_and_bursa_incident':2}, loss=weighted_crossentropy([1.0, 125, 106], 'other_disorders_of_synovium_tendon_and_bursa_pi'))
other_disorders_of_testis_phe_pi = TensorMap('phecode_603',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_603_date'), channel_map={'no_other_disorders_of_testis':0, 'other_disorders_of_testis_prevalent':1, 'other_disorders_of_testis_incident':2}, loss=weighted_crossentropy([1.0, 739, 865], 'other_disorders_of_testis_pi'))
other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation_phe_pi = TensorMap('phecode_676',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_676_date'), channel_map={'no_other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation':0, 'other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation_prevalent':1, 'other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation_incident':2}, loss=weighted_crossentropy([1.0, 6578, 83333], 'other_disorders_of_the_breast_associated_with_childbirth_and_disorders_of_lactation_pi'))
other_disorders_of_the_kidney_and_ureters_phe_pi = TensorMap('phecode_586',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_586_date'), channel_map={'no_other_disorders_of_the_kidney_and_ureters':0, 'other_disorders_of_the_kidney_and_ureters_prevalent':1, 'other_disorders_of_the_kidney_and_ureters_incident':2}, loss=weighted_crossentropy([1.0, 423, 165], 'other_disorders_of_the_kidney_and_ureters_pi'))
other_disorders_of_thyroid_phe_pi = TensorMap('phecode_246',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_246_date'), channel_map={'no_other_disorders_of_thyroid':0, 'other_disorders_of_thyroid_prevalent':1, 'other_disorders_of_thyroid_incident':2}, loss=weighted_crossentropy([1.0, 1432, 657], 'other_disorders_of_thyroid_pi'))
other_disorders_of_tympanic_membrane_phe_pi = TensorMap('phecode_384',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_384_date'), channel_map={'no_other_disorders_of_tympanic_membrane':0, 'other_disorders_of_tympanic_membrane_prevalent':1, 'other_disorders_of_tympanic_membrane_incident':2}, loss=weighted_crossentropy([1.0, 1805, 2380], 'other_disorders_of_tympanic_membrane_pi'))
other_disorders_of_urethra_and_urinary_tract_phe_pi = TensorMap('phecode_597',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_597_date'), channel_map={'no_other_disorders_of_urethra_and_urinary_tract':0, 'other_disorders_of_urethra_and_urinary_tract_prevalent':1, 'other_disorders_of_urethra_and_urinary_tract_incident':2}, loss=weighted_crossentropy([1.0, 1002, 868], 'other_disorders_of_urethra_and_urinary_tract_pi'))
other_dyschromia_phe_pi = TensorMap('phecode_694.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_694.2_date'), channel_map={'no_other_dyschromia':0, 'other_dyschromia_prevalent':1, 'other_dyschromia_incident':2}, loss=weighted_crossentropy([1.0, 1075, 740], 'other_dyschromia_pi'))
other_dyspnea_phe_pi = TensorMap('phecode_512.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_512.9_date'), channel_map={'no_other_dyspnea':0, 'other_dyspnea_prevalent':1, 'other_dyspnea_incident':2}, loss=weighted_crossentropy([1.0, 974, 617], 'other_dyspnea_pi'))
other_endocrine_disorders_phe_pi = TensorMap('phecode_259',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_259_date'), channel_map={'no_other_endocrine_disorders':0, 'other_endocrine_disorders_prevalent':1, 'other_endocrine_disorders_incident':2}, loss=weighted_crossentropy([1.0, 4716, 6578], 'other_endocrine_disorders_pi'))
other_forms_of_chronic_heart_disease_phe_pi = TensorMap('phecode_414',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_414_date'), channel_map={'no_other_forms_of_chronic_heart_disease':0, 'other_forms_of_chronic_heart_disease_prevalent':1, 'other_forms_of_chronic_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 1329, 174], 'other_forms_of_chronic_heart_disease_pi'))
other_headache_syndromes_phe_pi = TensorMap('phecode_339',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_339_date'), channel_map={'no_other_headache_syndromes':0, 'other_headache_syndromes_prevalent':1, 'other_headache_syndromes_incident':2}, loss=weighted_crossentropy([1.0, 98, 80], 'other_headache_syndromes_pi'))
other_heart_block_phe_pi = TensorMap('phecode_426.25',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.25_date'), channel_map={'no_other_heart_block':0, 'other_heart_block_prevalent':1, 'other_heart_block_incident':2}, loss=weighted_crossentropy([1.0, 7936, 4065], 'other_heart_block_pi'))
other_hemoglobinopathies_phe_pi = TensorMap('phecode_282.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_282.8_date'), channel_map={'no_other_hemoglobinopathies':0, 'other_hemoglobinopathies_prevalent':1, 'other_hemoglobinopathies_incident':2}, loss=weighted_crossentropy([1.0, 3267, 2325], 'other_hemoglobinopathies_pi'))
other_hereditary_hemolytic_anemias_phe_pi = TensorMap('phecode_282.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_282.9_date'), channel_map={'no_other_hereditary_hemolytic_anemias':0, 'other_hereditary_hemolytic_anemias_prevalent':1, 'other_hereditary_hemolytic_anemias_incident':2}, loss=weighted_crossentropy([1.0, 970, 777], 'other_hereditary_hemolytic_anemias_pi'))
other_high_risk_pregnancy_phe_pi = TensorMap('phecode_638',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_638_date'), channel_map={'no_other_high_risk_pregnancy':0, 'other_high_risk_pregnancy_prevalent':1, 'other_high_risk_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 7692, 18518], 'other_high_risk_pregnancy_pi'))
other_hypertensive_complications_phe_pi = TensorMap('phecode_401.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_401.3_date'), channel_map={'no_other_hypertensive_complications':0, 'other_hypertensive_complications_prevalent':1, 'other_hypertensive_complications_incident':2}, loss=weighted_crossentropy([1.0, 4504, 5102], 'other_hypertensive_complications_pi'))
other_hypertrophic_and_atrophic_conditions_of_skin_phe_pi = TensorMap('phecode_701',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701_date'), channel_map={'no_other_hypertrophic_and_atrophic_conditions_of_skin':0, 'other_hypertrophic_and_atrophic_conditions_of_skin_prevalent':1, 'other_hypertrophic_and_atrophic_conditions_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 412, 465], 'other_hypertrophic_and_atrophic_conditions_of_skin_pi'))
other_hypertrophic_cardiomyopathy_phe_pi = TensorMap('phecode_425.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_425.12_date'), channel_map={'no_other_hypertrophic_cardiomyopathy':0, 'other_hypertrophic_cardiomyopathy_prevalent':1, 'other_hypertrophic_cardiomyopathy_incident':2}, loss=weighted_crossentropy([1.0, 6329, 2369], 'other_hypertrophic_cardiomyopathy_pi'))
other_ill_defined_and_unknown_causes_of_morbidity_and_mortality_phe_pi = TensorMap('phecode_1019',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1019_date'), channel_map={'no_other_ill_defined_and_unknown_causes_of_morbidity_and_mortality':0, 'other_ill_defined_and_unknown_causes_of_morbidity_and_mortality_prevalent':1, 'other_ill_defined_and_unknown_causes_of_morbidity_and_mortality_incident':2}, loss=weighted_crossentropy([1.0, 29, 143], 'other_ill_defined_and_unknown_causes_of_morbidity_and_mortality_pi'))
other_immunological_findings_phe_pi = TensorMap('phecode_279.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_279.7_date'), channel_map={'no_other_immunological_findings':0, 'other_immunological_findings_prevalent':1, 'other_immunological_findings_incident':2}, loss=weighted_crossentropy([1.0, 2016, 4950], 'other_immunological_findings_pi'))
other_infectious_and_parasitic_diseases_phe_pi = TensorMap('phecode_136',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_136_date'), channel_map={'no_other_infectious_and_parasitic_diseases':0, 'other_infectious_and_parasitic_diseases_prevalent':1, 'other_infectious_and_parasitic_diseases_incident':2}, loss=weighted_crossentropy([1.0, 1519, 1068], 'other_infectious_and_parasitic_diseases_pi'))
other_inflammatory_disorders_of_male_genital_organs_phe_pi = TensorMap('phecode_601.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_601.8_date'), channel_map={'no_other_inflammatory_disorders_of_male_genital_organs':0, 'other_inflammatory_disorders_of_male_genital_organs_prevalent':1, 'other_inflammatory_disorders_of_male_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 2732, 2392], 'other_inflammatory_disorders_of_male_genital_organs_pi'))
other_inflammatory_spondylopathies_phe_pi = TensorMap('phecode_715',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_715_date'), channel_map={'no_other_inflammatory_spondylopathies':0, 'other_inflammatory_spondylopathies_prevalent':1, 'other_inflammatory_spondylopathies_incident':2}, loss=weighted_crossentropy([1.0, 5882, 1103], 'other_inflammatory_spondylopathies_pi'))
other_intestinal_obstruction_phe_pi = TensorMap('phecode_560.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_560.4_date'), channel_map={'no_other_intestinal_obstruction':0, 'other_intestinal_obstruction_prevalent':1, 'other_intestinal_obstruction_incident':2}, loss=weighted_crossentropy([1.0, 308, 151], 'other_intestinal_obstruction_pi'))
other_local_infections_of_skin_and_subcutaneous_tissue_phe_pi = TensorMap('phecode_686',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_686_date'), channel_map={'no_other_local_infections_of_skin_and_subcutaneous_tissue':0, 'other_local_infections_of_skin_and_subcutaneous_tissue_prevalent':1, 'other_local_infections_of_skin_and_subcutaneous_tissue_incident':2}, loss=weighted_crossentropy([1.0, 158, 115], 'other_local_infections_of_skin_and_subcutaneous_tissue_pi'))
other_mental_disorder_phe_pi = TensorMap('phecode_306',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_306_date'), channel_map={'no_other_mental_disorder':0, 'other_mental_disorder_prevalent':1, 'other_mental_disorder_incident':2}, loss=weighted_crossentropy([1.0, 4, 4], 'other_mental_disorder_pi'))
other_non_epithelial_cancer_of_skin_phe_pi = TensorMap('phecode_172.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_172.2_date'), channel_map={'no_other_non_epithelial_cancer_of_skin':0, 'other_non_epithelial_cancer_of_skin_prevalent':1, 'other_non_epithelial_cancer_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 99, 45], 'other_non_epithelial_cancer_of_skin_pi'))
other_nondiabetic_retinopathy_phe_pi = TensorMap('phecode_362.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.3_date'), channel_map={'no_other_nondiabetic_retinopathy':0, 'other_nondiabetic_retinopathy_prevalent':1, 'other_nondiabetic_retinopathy_incident':2}, loss=weighted_crossentropy([1.0, 8620, 9803], 'other_nondiabetic_retinopathy_pi'))
other_nonspecific_findings_on_examination_of_urine_phe_pi = TensorMap('phecode_598.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_598.9_date'), channel_map={'no_other_nonspecific_findings_on_examination_of_urine':0, 'other_nonspecific_findings_on_examination_of_urine_prevalent':1, 'other_nonspecific_findings_on_examination_of_urine_incident':2}, loss=weighted_crossentropy([1.0, 3546, 1587], 'other_nonspecific_findings_on_examination_of_urine_pi'))
other_nutritional_deficiency_phe_pi = TensorMap('phecode_263',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_263_date'), channel_map={'no_other_nutritional_deficiency':0, 'other_nutritional_deficiency_prevalent':1, 'other_nutritional_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 1160, 168], 'other_nutritional_deficiency_pi'))
other_open_wound_of_head_and_face_phe_pi = TensorMap('phecode_870.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_870.3_date'), channel_map={'no_other_open_wound_of_head_and_face':0, 'other_open_wound_of_head_and_face_prevalent':1, 'other_open_wound_of_head_and_face_incident':2}, loss=weighted_crossentropy([1.0, 313, 186], 'other_open_wound_of_head_and_face_pi'))
other_paralytic_syndromes_phe_pi = TensorMap('phecode_344',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_344_date'), channel_map={'no_other_paralytic_syndromes':0, 'other_paralytic_syndromes_prevalent':1, 'other_paralytic_syndromes_incident':2}, loss=weighted_crossentropy([1.0, 1449, 780], 'other_paralytic_syndromes_pi'))
other_perinatal_conditions_of_fetus_or_newborn_phe_pi = TensorMap('phecode_656',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_656_date'), channel_map={'no_other_perinatal_conditions_of_fetus_or_newborn':0, 'other_perinatal_conditions_of_fetus_or_newborn_prevalent':1, 'other_perinatal_conditions_of_fetus_or_newborn_incident':2}, loss=weighted_crossentropy([1.0, 726, 13888], 'other_perinatal_conditions_of_fetus_or_newborn_pi'))
other_peripheral_nerve_disorders_phe_pi = TensorMap('phecode_351',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_351_date'), channel_map={'no_other_peripheral_nerve_disorders':0, 'other_peripheral_nerve_disorders_prevalent':1, 'other_peripheral_nerve_disorders_incident':2}, loss=weighted_crossentropy([1.0, 56, 61], 'other_peripheral_nerve_disorders_pi'))
other_persistent_mental_disorders_due_to_conditions_classified_elsewhere_phe_pi = TensorMap('phecode_290.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290.3_date'), channel_map={'no_other_persistent_mental_disorders_due_to_conditions_classified_elsewhere':0, 'other_persistent_mental_disorders_due_to_conditions_classified_elsewhere_prevalent':1, 'other_persistent_mental_disorders_due_to_conditions_classified_elsewhere_incident':2}, loss=weighted_crossentropy([1.0, 7936, 7462], 'other_persistent_mental_disorders_due_to_conditions_classified_elsewhere_pi'))
other_pulmonary_inflamation_or_edema_phe_pi = TensorMap('phecode_505',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_505_date'), channel_map={'no_other_pulmonary_inflamation_or_edema':0, 'other_pulmonary_inflamation_or_edema_prevalent':1, 'other_pulmonary_inflamation_or_edema_incident':2}, loss=weighted_crossentropy([1.0, 16666, 8196], 'other_pulmonary_inflamation_or_edema_pi'))
other_retinal_disorders_phe_pi = TensorMap('phecode_362',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362_date'), channel_map={'no_other_retinal_disorders':0, 'other_retinal_disorders_prevalent':1, 'other_retinal_disorders_incident':2}, loss=weighted_crossentropy([1.0, 1173, 475], 'other_retinal_disorders_pi'))
other_signs_and_symptoms_in_breast_phe_pi = TensorMap('phecode_613.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_613.7_date'), channel_map={'no_other_signs_and_symptoms_in_breast':0, 'other_signs_and_symptoms_in_breast_prevalent':1, 'other_signs_and_symptoms_in_breast_incident':2}, loss=weighted_crossentropy([1.0, 1246, 2232], 'other_signs_and_symptoms_in_breast_pi'))
other_signs_and_symptoms_involving_emotional_state_phe_pi = TensorMap('phecode_1004',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1004_date'), channel_map={'no_other_signs_and_symptoms_involving_emotional_state':0, 'other_signs_and_symptoms_involving_emotional_state_prevalent':1, 'other_signs_and_symptoms_involving_emotional_state_incident':2}, loss=weighted_crossentropy([1.0, 9803, 761], 'other_signs_and_symptoms_involving_emotional_state_pi'))
other_specified_benign_mammary_dysplasias_phe_pi = TensorMap('phecode_610.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_610.8_date'), channel_map={'no_other_specified_benign_mammary_dysplasias':0, 'other_specified_benign_mammary_dysplasias_prevalent':1, 'other_specified_benign_mammary_dysplasias_incident':2}, loss=weighted_crossentropy([1.0, 909, 1567], 'other_specified_benign_mammary_dysplasias_pi'))
other_specified_cardiac_dysrhythmias_phe_pi = TensorMap('phecode_427.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.3_date'), channel_map={'no_other_specified_cardiac_dysrhythmias':0, 'other_specified_cardiac_dysrhythmias_prevalent':1, 'other_specified_cardiac_dysrhythmias_incident':2}, loss=weighted_crossentropy([1.0, 358, 136], 'other_specified_cardiac_dysrhythmias_pi'))
other_specified_congenital_anomalies_of_kidney_phe_pi = TensorMap('phecode_751.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_751.22_date'), channel_map={'no_other_specified_congenital_anomalies_of_kidney':0, 'other_specified_congenital_anomalies_of_kidney_prevalent':1, 'other_specified_congenital_anomalies_of_kidney_incident':2}, loss=weighted_crossentropy([1.0, 2659, 1567], 'other_specified_congenital_anomalies_of_kidney_pi'))
other_specified_congenital_anomalies_of_nervous_system_phe_pi = TensorMap('phecode_752.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_752.2_date'), channel_map={'no_other_specified_congenital_anomalies_of_nervous_system':0, 'other_specified_congenital_anomalies_of_nervous_system_prevalent':1, 'other_specified_congenital_anomalies_of_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 11363, 7936], 'other_specified_congenital_anomalies_of_nervous_system_pi'))
other_specified_diffuse_diseases_of_connective_tissue_phe_pi = TensorMap('phecode_709.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_709.6_date'), channel_map={'no_other_specified_diffuse_diseases_of_connective_tissue':0, 'other_specified_diffuse_diseases_of_connective_tissue_prevalent':1, 'other_specified_diffuse_diseases_of_connective_tissue_incident':2}, loss=weighted_crossentropy([1.0, 17857, 7246], 'other_specified_diffuse_diseases_of_connective_tissue_pi'))
other_specified_diseases_of_hair_and_hair_follicles_phe_pi = TensorMap('phecode_704.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_704.8_date'), channel_map={'no_other_specified_diseases_of_hair_and_hair_follicles':0, 'other_specified_diseases_of_hair_and_hair_follicles_prevalent':1, 'other_specified_diseases_of_hair_and_hair_follicles_incident':2}, loss=weighted_crossentropy([1.0, 10416, 8771], 'other_specified_diseases_of_hair_and_hair_follicles_pi'))
other_specified_diseases_of_sebaceous_glands_phe_pi = TensorMap('phecode_706.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_706.8_date'), channel_map={'no_other_specified_diseases_of_sebaceous_glands':0, 'other_specified_diseases_of_sebaceous_glands_prevalent':1, 'other_specified_diseases_of_sebaceous_glands_incident':2}, loss=weighted_crossentropy([1.0, 166666, 100000], 'other_specified_diseases_of_sebaceous_glands_pi'))
other_specified_diseases_of_the_salivary_glands_phe_pi = TensorMap('phecode_527.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_527.8_date'), channel_map={'no_other_specified_diseases_of_the_salivary_glands':0, 'other_specified_diseases_of_the_salivary_glands_prevalent':1, 'other_specified_diseases_of_the_salivary_glands_incident':2}, loss=weighted_crossentropy([1.0, 5952, 5319], 'other_specified_diseases_of_the_salivary_glands_pi'))
other_specified_disorders_involving_the_immune_mechanism_phe_pi = TensorMap('phecode_279.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_279.8_date'), channel_map={'no_other_specified_disorders_involving_the_immune_mechanism':0, 'other_specified_disorders_involving_the_immune_mechanism_prevalent':1, 'other_specified_disorders_involving_the_immune_mechanism_incident':2}, loss=weighted_crossentropy([1.0, 50000, 20000], 'other_specified_disorders_involving_the_immune_mechanism_pi'))
other_specified_disorders_of_breast_phe_pi = TensorMap('phecode_613.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_613.8_date'), channel_map={'no_other_specified_disorders_of_breast':0, 'other_specified_disorders_of_breast_prevalent':1, 'other_specified_disorders_of_breast_incident':2}, loss=weighted_crossentropy([1.0, 2057, 2336], 'other_specified_disorders_of_breast_pi'))
other_specified_erythematous_conditions_phe_pi = TensorMap('phecode_695.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.8_date'), channel_map={'no_other_specified_erythematous_conditions':0, 'other_specified_erythematous_conditions_prevalent':1, 'other_specified_erythematous_conditions_incident':2}, loss=weighted_crossentropy([1.0, 4761, 3333], 'other_specified_erythematous_conditions_pi'))
other_specified_gastritis_phe_pi = TensorMap('phecode_535.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_535.8_date'), channel_map={'no_other_specified_gastritis':0, 'other_specified_gastritis_prevalent':1, 'other_specified_gastritis_incident':2}, loss=weighted_crossentropy([1.0, 96, 65], 'other_specified_gastritis_pi'))
other_specified_osteoporosis_phe_pi = TensorMap('phecode_743.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.13_date'), channel_map={'no_other_specified_osteoporosis':0, 'other_specified_osteoporosis_prevalent':1, 'other_specified_osteoporosis_incident':2}, loss=weighted_crossentropy([1.0, 2958, 7692], 'other_specified_osteoporosis_pi'))
other_specified_peripheral_vascular_diseases_phe_pi = TensorMap('phecode_443.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_443.8_date'), channel_map={'no_other_specified_peripheral_vascular_diseases':0, 'other_specified_peripheral_vascular_diseases_prevalent':1, 'other_specified_peripheral_vascular_diseases_incident':2}, loss=weighted_crossentropy([1.0, 21739, 10869], 'other_specified_peripheral_vascular_diseases_pi'))
other_sprains_and_strains_phe_pi = TensorMap('phecode_842',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_842_date'), channel_map={'no_other_sprains_and_strains':0, 'other_sprains_and_strains_prevalent':1, 'other_sprains_and_strains_incident':2}, loss=weighted_crossentropy([1.0, 166666, 100000], 'other_sprains_and_strains_pi'))
other_symptoms_phe_pi = TensorMap('phecode_1005',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1005_date'), channel_map={'no_other_symptoms':0, 'other_symptoms_prevalent':1, 'other_symptoms_incident':2}, loss=weighted_crossentropy([1.0, 1063, 357], 'other_symptoms_pi'))
other_symptoms_disorders_or_the_urinary_system_phe_pi = TensorMap('phecode_599',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599_date'), channel_map={'no_other_symptoms_disorders_or_the_urinary_system':0, 'other_symptoms_disorders_or_the_urinary_system_prevalent':1, 'other_symptoms_disorders_or_the_urinary_system_incident':2}, loss=weighted_crossentropy([1.0, 6, 8], 'other_symptoms_disorders_or_the_urinary_system_pi'))
other_symptoms_involving_abdomen_and_pelvis_phe_pi = TensorMap('phecode_579',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_579_date'), channel_map={'no_other_symptoms_involving_abdomen_and_pelvis':0, 'other_symptoms_involving_abdomen_and_pelvis_prevalent':1, 'other_symptoms_involving_abdomen_and_pelvis_incident':2}, loss=weighted_crossentropy([1.0, 655, 610], 'other_symptoms_involving_abdomen_and_pelvis_pi'))
other_symptoms_involving_urinary_system_phe_pi = TensorMap('phecode_599.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.8_date'), channel_map={'no_other_symptoms_involving_urinary_system':0, 'other_symptoms_involving_urinary_system_prevalent':1, 'other_symptoms_involving_urinary_system_incident':2}, loss=weighted_crossentropy([1.0, 18518, 10869], 'other_symptoms_involving_urinary_system_pi'))
other_symptoms_of_respiratory_system_phe_pi = TensorMap('phecode_512',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_512_date'), channel_map={'no_other_symptoms_of_respiratory_system':0, 'other_symptoms_of_respiratory_system_prevalent':1, 'other_symptoms_of_respiratory_system_incident':2}, loss=weighted_crossentropy([1.0, 316, 68], 'other_symptoms_of_respiratory_system_pi'))
other_symptoms_referable_to_back_phe_pi = TensorMap('phecode_724.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_724.8_date'), channel_map={'no_other_symptoms_referable_to_back':0, 'other_symptoms_referable_to_back_prevalent':1, 'other_symptoms_referable_to_back_incident':2}, loss=weighted_crossentropy([1.0, 166666, 26315], 'other_symptoms_referable_to_back_pi'))
other_tests_phe_pi = TensorMap('phecode_1010',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1010_date'), channel_map={'no_other_tests':0, 'other_tests_prevalent':1, 'other_tests_incident':2}, loss=weighted_crossentropy([1.0, 4, 4], 'other_tests_pi'))
other_unspecified_back_disorders_phe_pi = TensorMap('phecode_724.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_724.9_date'), channel_map={'no_other_unspecified_back_disorders':0, 'other_unspecified_back_disorders_prevalent':1, 'other_unspecified_back_disorders_incident':2}, loss=weighted_crossentropy([1.0, 2192, 1086], 'other_unspecified_back_disorders_pi'))
other_upper_respiratory_disease_phe_pi = TensorMap('phecode_479',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_479_date'), channel_map={'no_other_upper_respiratory_disease':0, 'other_upper_respiratory_disease_prevalent':1, 'other_upper_respiratory_disease_incident':2}, loss=weighted_crossentropy([1.0, 42, 59], 'other_upper_respiratory_disease_pi'))
other_venous_embolism_and_thrombosis_phe_pi = TensorMap('phecode_452',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_452_date'), channel_map={'no_other_venous_embolism_and_thrombosis':0, 'other_venous_embolism_and_thrombosis_prevalent':1, 'other_venous_embolism_and_thrombosis_incident':2}, loss=weighted_crossentropy([1.0, 2100, 825], 'other_venous_embolism_and_thrombosis_pi'))
other_vitamin_b12_deficiency_anemia_phe_pi = TensorMap('phecode_281.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_281.12_date'), channel_map={'no_other_vitamin_b12_deficiency_anemia':0, 'other_vitamin_b12_deficiency_anemia_prevalent':1, 'other_vitamin_b12_deficiency_anemia_incident':2}, loss=weighted_crossentropy([1.0, 5154, 2500], 'other_vitamin_b12_deficiency_anemia_pi'))
otitis_externa_phe_pi = TensorMap('phecode_380.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_380.1_date'), channel_map={'no_otitis_externa':0, 'otitis_externa_prevalent':1, 'otitis_externa_incident':2}, loss=weighted_crossentropy([1.0, 1291, 1538], 'otitis_externa_pi'))
otitis_media_phe_pi = TensorMap('phecode_381.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_381.1_date'), channel_map={'no_otitis_media':0, 'otitis_media_prevalent':1, 'otitis_media_incident':2}, loss=weighted_crossentropy([1.0, 195, 290], 'otitis_media_pi'))
otorrhea_phe_pi = TensorMap('phecode_381.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_381.9_date'), channel_map={'no_otorrhea':0, 'otorrhea_prevalent':1, 'otorrhea_incident':2}, loss=weighted_crossentropy([1.0, 2500, 3759], 'otorrhea_pi'))
otosclerosis_phe_pi = TensorMap('phecode_383',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_383_date'), channel_map={'no_otosclerosis':0, 'otosclerosis_prevalent':1, 'otosclerosis_incident':2}, loss=weighted_crossentropy([1.0, 1976, 3571], 'otosclerosis_pi'))
ovarian_cyst_phe_pi = TensorMap('phecode_628',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_628_date'), channel_map={'no_ovarian_cyst':0, 'ovarian_cyst_prevalent':1, 'ovarian_cyst_incident':2}, loss=weighted_crossentropy([1.0, 123, 249], 'ovarian_cyst_pi'))
ovarian_dysfunction_phe_pi = TensorMap('phecode_256',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_256_date'), channel_map={'no_ovarian_dysfunction':0, 'ovarian_dysfunction_prevalent':1, 'ovarian_dysfunction_incident':2}, loss=weighted_crossentropy([1.0, 11111, 55555], 'ovarian_dysfunction_pi'))
overweight_obesity_and_other_hyperalimentation_phe_pi = TensorMap('phecode_278',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_278_date'), channel_map={'no_overweight_obesity_and_other_hyperalimentation':0, 'overweight_obesity_and_other_hyperalimentation_prevalent':1, 'overweight_obesity_and_other_hyperalimentation_incident':2}, loss=weighted_crossentropy([1.0, 250000, 125000], 'overweight_obesity_and_other_hyperalimentation_pi'))
pain_and_other_symptoms_associated_with_female_genital_organs_phe_pi = TensorMap('phecode_625',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_625_date'), channel_map={'no_pain_and_other_symptoms_associated_with_female_genital_organs':0, 'pain_and_other_symptoms_associated_with_female_genital_organs_prevalent':1, 'pain_and_other_symptoms_associated_with_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 271, 884], 'pain_and_other_symptoms_associated_with_female_genital_organs_pi'))
pain_in_joint_phe_pi = TensorMap('phecode_745',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_745_date'), channel_map={'no_pain_in_joint':0, 'pain_in_joint_prevalent':1, 'pain_in_joint_incident':2}, loss=weighted_crossentropy([1.0, 324, 757], 'pain_in_joint_pi'))
pain_in_limb_phe_pi = TensorMap('phecode_773',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_773_date'), channel_map={'no_pain_in_limb':0, 'pain_in_limb_prevalent':1, 'pain_in_limb_incident':2}, loss=weighted_crossentropy([1.0, 414, 782], 'pain_in_limb_pi'))
pain_swelling_or_discharge_of_eye_phe_pi = TensorMap('phecode_379.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379.9_date'), channel_map={'no_pain_swelling_or_discharge_of_eye':0, 'pain_swelling_or_discharge_of_eye_prevalent':1, 'pain_swelling_or_discharge_of_eye_incident':2}, loss=weighted_crossentropy([1.0, 5154, 2958], 'pain_swelling_or_discharge_of_eye_pi'))
painful_respiration_phe_pi = TensorMap('phecode_512.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_512.2_date'), channel_map={'no_painful_respiration':0, 'painful_respiration_prevalent':1, 'painful_respiration_incident':2}, loss=weighted_crossentropy([1.0, 3875, 2439], 'painful_respiration_pi'))
palindromic_rheumatism_phe_pi = TensorMap('phecode_716.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_716.8_date'), channel_map={'no_palindromic_rheumatism':0, 'palindromic_rheumatism_prevalent':1, 'palindromic_rheumatism_incident':2}, loss=weighted_crossentropy([1.0, 166666, 23809], 'palindromic_rheumatism_pi'))
pallor_and_flushing_phe_pi = TensorMap('phecode_782.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_782.6_date'), channel_map={'no_pallor_and_flushing':0, 'pallor_and_flushing_prevalent':1, 'pallor_and_flushing_incident':2}, loss=weighted_crossentropy([1.0, 6849, 2512], 'pallor_and_flushing_pi'))
palpitations_phe_pi = TensorMap('phecode_427.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.9_date'), channel_map={'no_palpitations':0, 'palpitations_prevalent':1, 'palpitations_incident':2}, loss=weighted_crossentropy([1.0, 219, 149], 'palpitations_pi'))
pancreatic_cancer_phe_pi = TensorMap('phecode_157',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_157_date'), channel_map={'no_pancreatic_cancer':0, 'pancreatic_cancer_prevalent':1, 'pancreatic_cancer_incident':2}, loss=weighted_crossentropy([1.0, 6493, 558], 'pancreatic_cancer_pi'))
panniculitis_phe_pi = TensorMap('phecode_729.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_729.3_date'), channel_map={'no_panniculitis':0, 'panniculitis_prevalent':1, 'panniculitis_incident':2}, loss=weighted_crossentropy([1.0, 38461, 10869], 'panniculitis_pi'))
paralysis_spasm_of_vocal_cords_or_larynx_phe_pi = TensorMap('phecode_473.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_473.3_date'), channel_map={'no_paralysis_spasm_of_vocal_cords_or_larynx':0, 'paralysis_spasm_of_vocal_cords_or_larynx_prevalent':1, 'paralysis_spasm_of_vocal_cords_or_larynx_incident':2}, loss=weighted_crossentropy([1.0, 4273, 1706], 'paralysis_spasm_of_vocal_cords_or_larynx_pi'))
paralytic_ileus_phe_pi = TensorMap('phecode_560.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_560.1_date'), channel_map={'no_paralytic_ileus':0, 'paralytic_ileus_prevalent':1, 'paralytic_ileus_incident':2}, loss=weighted_crossentropy([1.0, 2840, 728], 'paralytic_ileus_pi'))
paralytic_strabismus_phe_pi = TensorMap('phecode_378.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_378.5_date'), channel_map={'no_paralytic_strabismus':0, 'paralytic_strabismus_prevalent':1, 'paralytic_strabismus_incident':2}, loss=weighted_crossentropy([1.0, 3164, 1960], 'paralytic_strabismus_pi'))
paranoid_disorders_phe_pi = TensorMap('phecode_295.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_295.2_date'), channel_map={'no_paranoid_disorders':0, 'paranoid_disorders_prevalent':1, 'paranoid_disorders_incident':2}, loss=weighted_crossentropy([1.0, 3875, 2659], 'paranoid_disorders_pi'))
paraproteinemia_phe_pi = TensorMap('phecode_270.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.32_date'), channel_map={'no_paraproteinemia':0, 'paraproteinemia_prevalent':1, 'paraproteinemia_incident':2}, loss=weighted_crossentropy([1.0, 10638, 1510], 'paraproteinemia_pi'))
parapsoriasis_phe_pi = TensorMap('phecode_696.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_696.2_date'), channel_map={'no_parapsoriasis':0, 'parapsoriasis_prevalent':1, 'parapsoriasis_incident':2}, loss=weighted_crossentropy([1.0, 35714, 38461], 'parapsoriasis_pi'))
parasomnia_phe_pi = TensorMap('phecode_327.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327.5_date'), channel_map={'no_parasomnia':0, 'parasomnia_prevalent':1, 'parasomnia_incident':2}, loss=weighted_crossentropy([1.0, 125000, 21739], 'parasomnia_pi'))
parkinsons_disease_phe_pi = TensorMap('phecode_332',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_332_date'), channel_map={'no_parkinsons_disease':0, 'parkinsons_disease_prevalent':1, 'parkinsons_disease_incident':2}, loss=weighted_crossentropy([1.0, 1319, 320], 'parkinsons_disease_pi'))
paroxysmal_supraventricular_tachycardia_phe_pi = TensorMap('phecode_427.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.11_date'), channel_map={'no_paroxysmal_supraventricular_tachycardia':0, 'paroxysmal_supraventricular_tachycardia_prevalent':1, 'paroxysmal_supraventricular_tachycardia_incident':2}, loss=weighted_crossentropy([1.0, 330, 261], 'paroxysmal_supraventricular_tachycardia_pi'))
paroxysmal_tachycardia_unspecified_phe_pi = TensorMap('phecode_427.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.1_date'), channel_map={'no_paroxysmal_tachycardia_unspecified':0, 'paroxysmal_tachycardia_unspecified_prevalent':1, 'paroxysmal_tachycardia_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 5434, 7575], 'paroxysmal_tachycardia_unspecified_pi'))
paroxysmal_ventricular_tachycardia_phe_pi = TensorMap('phecode_427.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.12_date'), channel_map={'no_paroxysmal_ventricular_tachycardia':0, 'paroxysmal_ventricular_tachycardia_prevalent':1, 'paroxysmal_ventricular_tachycardia_incident':2}, loss=weighted_crossentropy([1.0, 1048, 550], 'paroxysmal_ventricular_tachycardia_pi'))
partial_epilepsy_phe_pi = TensorMap('phecode_345.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_345.12_date'), channel_map={'no_partial_epilepsy':0, 'partial_epilepsy_prevalent':1, 'partial_epilepsy_incident':2}, loss=weighted_crossentropy([1.0, 2762, 2092], 'partial_epilepsy_pi'))
pathologic_fracture_phe_pi = TensorMap('phecode_743.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.2_date'), channel_map={'no_pathologic_fracture':0, 'pathologic_fracture_prevalent':1, 'pathologic_fracture_incident':2}, loss=weighted_crossentropy([1.0, 22727, 33333], 'pathologic_fracture_pi'))
pathologic_fracture_of_vertebrae_phe_pi = TensorMap('phecode_743.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.21_date'), channel_map={'no_pathologic_fracture_of_vertebrae':0, 'pathologic_fracture_of_vertebrae_prevalent':1, 'pathologic_fracture_of_vertebrae_incident':2}, loss=weighted_crossentropy([1.0, 35714, 11904], 'pathologic_fracture_of_vertebrae_pi'))
pathological_developmental_or_recurrent_dislocation_phe_pi = TensorMap('phecode_742.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_742.2_date'), channel_map={'no_pathological_developmental_or_recurrent_dislocation':0, 'pathological_developmental_or_recurrent_dislocation_prevalent':1, 'pathological_developmental_or_recurrent_dislocation_incident':2}, loss=weighted_crossentropy([1.0, 8771, 18518], 'pathological_developmental_or_recurrent_dislocation_pi'))
pectus_and_other_congenital_anomalies_of_ribs_sternum_phe_pi = TensorMap('phecode_756.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_756.2_date'), channel_map={'no_pectus_and_other_congenital_anomalies_of_ribs_sternum':0, 'pectus_and_other_congenital_anomalies_of_ribs_sternum_prevalent':1, 'pectus_and_other_congenital_anomalies_of_ribs_sternum_incident':2}, loss=weighted_crossentropy([1.0, 100000, 100000], 'pectus_and_other_congenital_anomalies_of_ribs_sternum_pi'))
pectus_excavatum_phe_pi = TensorMap('phecode_756.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_756.21_date'), channel_map={'no_pectus_excavatum':0, 'pectus_excavatum_prevalent':1, 'pectus_excavatum_incident':2}, loss=weighted_crossentropy([1.0, 62500, 31250], 'pectus_excavatum_pi'))
pediculosis_and_phthirus_infestation_phe_pi = TensorMap('phecode_132.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_132.1_date'), channel_map={'no_pediculosis_and_phthirus_infestation':0, 'pediculosis_and_phthirus_infestation_prevalent':1, 'pediculosis_and_phthirus_infestation_incident':2}, loss=weighted_crossentropy([1.0, 23809, 15151], 'pediculosis_and_phthirus_infestation_pi'))
pelvic_inflammatory_disease_nos_phe_pi = TensorMap('phecode_614.33',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.33_date'), channel_map={'no_pelvic_inflammatory_disease_nos':0, 'pelvic_inflammatory_disease_nos_prevalent':1, 'pelvic_inflammatory_disease_nos_incident':2}, loss=weighted_crossentropy([1.0, 90, 196], 'pelvic_inflammatory_disease_nos_pi'))
pelvic_inflammatory_disease_pid_phe_pi = TensorMap('phecode_614.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.3_date'), channel_map={'no_pelvic_inflammatory_disease_pid':0, 'pelvic_inflammatory_disease_pid_prevalent':1, 'pelvic_inflammatory_disease_pid_incident':2}, loss=weighted_crossentropy([1.0, 4273, 6024], 'pelvic_inflammatory_disease_pid_pi'))
pelvic_peritoneal_adhesions_female_postoperative_postinfection_phe_pi = TensorMap('phecode_614.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.1_date'), channel_map={'no_pelvic_peritoneal_adhesions_female_postoperative_postinfection':0, 'pelvic_peritoneal_adhesions_female_postoperative_postinfection_prevalent':1, 'pelvic_peritoneal_adhesions_female_postoperative_postinfection_incident':2}, loss=weighted_crossentropy([1.0, 228, 384], 'pelvic_peritoneal_adhesions_female_postoperative_postinfection_pi'))
pemphigus_and_pemphigoid_phe_pi = TensorMap('phecode_695.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.22_date'), channel_map={'no_pemphigus_and_pemphigoid':0, 'pemphigus_and_pemphigoid_prevalent':1, 'pemphigus_and_pemphigoid_incident':2}, loss=weighted_crossentropy([1.0, 11363, 5102], 'pemphigus_and_pemphigoid_pi'))
peptic_ulcer_site_unspecified_phe_pi = TensorMap('phecode_531.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_531.4_date'), channel_map={'no_peptic_ulcer_site_unspecified':0, 'peptic_ulcer_site_unspecified_prevalent':1, 'peptic_ulcer_site_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 1706, 1683], 'peptic_ulcer_site_unspecified_pi'))
perforation_of_tympanic_membrane_phe_pi = TensorMap('phecode_384.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_384.4_date'), channel_map={'no_perforation_of_tympanic_membrane':0, 'perforation_of_tympanic_membrane_prevalent':1, 'perforation_of_tympanic_membrane_incident':2}, loss=weighted_crossentropy([1.0, 620, 967], 'perforation_of_tympanic_membrane_pi'))
periapical_abscess_phe_pi = TensorMap('phecode_522.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_522.5_date'), channel_map={'no_periapical_abscess':0, 'periapical_abscess_prevalent':1, 'periapical_abscess_incident':2}, loss=weighted_crossentropy([1.0, 615, 553], 'periapical_abscess_pi'))
pericarditis_phe_pi = TensorMap('phecode_420.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_420.2_date'), channel_map={'no_pericarditis':0, 'pericarditis_prevalent':1, 'pericarditis_incident':2}, loss=weighted_crossentropy([1.0, 1028, 409], 'pericarditis_pi'))
periodontitis_acute_or_chronic_phe_pi = TensorMap('phecode_523.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_523.3_date'), channel_map={'no_periodontitis_acute_or_chronic':0, 'periodontitis_acute_or_chronic_prevalent':1, 'periodontitis_acute_or_chronic_incident':2}, loss=weighted_crossentropy([1.0, 55555, 125000], 'periodontitis_acute_or_chronic_pi'))
peripheral_angiopathy_in_diseases_classified_elsewhere_phe_pi = TensorMap('phecode_443.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_443.7_date'), channel_map={'no_peripheral_angiopathy_in_diseases_classified_elsewhere':0, 'peripheral_angiopathy_in_diseases_classified_elsewhere_prevalent':1, 'peripheral_angiopathy_in_diseases_classified_elsewhere_incident':2}, loss=weighted_crossentropy([1.0, 2415, 1196], 'peripheral_angiopathy_in_diseases_classified_elsewhere_pi'))
peripheral_autonomic_neuropathy_phe_pi = TensorMap('phecode_337.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_337.1_date'), channel_map={'no_peripheral_autonomic_neuropathy':0, 'peripheral_autonomic_neuropathy_prevalent':1, 'peripheral_autonomic_neuropathy_incident':2}, loss=weighted_crossentropy([1.0, 9433, 11627], 'peripheral_autonomic_neuropathy_pi'))
peripheral_enthesopathies_and_allied_syndromes_phe_pi = TensorMap('phecode_726',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_726_date'), channel_map={'no_peripheral_enthesopathies_and_allied_syndromes':0, 'peripheral_enthesopathies_and_allied_syndromes_prevalent':1, 'peripheral_enthesopathies_and_allied_syndromes_incident':2}, loss=weighted_crossentropy([1.0, 109, 66], 'peripheral_enthesopathies_and_allied_syndromes_pi'))
peripheral_or_central_vertigo_phe_pi = TensorMap('phecode_386.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_386.2_date'), channel_map={'no_peripheral_or_central_vertigo':0, 'peripheral_or_central_vertigo_prevalent':1, 'peripheral_or_central_vertigo_incident':2}, loss=weighted_crossentropy([1.0, 3597, 974], 'peripheral_or_central_vertigo_pi'))
peripheral_retinal_degenerations_phe_pi = TensorMap('phecode_362.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.6_date'), channel_map={'no_peripheral_retinal_degenerations':0, 'peripheral_retinal_degenerations_prevalent':1, 'peripheral_retinal_degenerations_incident':2}, loss=weighted_crossentropy([1.0, 3311, 5208], 'peripheral_retinal_degenerations_pi'))
peripheral_vascular_disease_unspecified_phe_pi = TensorMap('phecode_443.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_443.9_date'), channel_map={'no_peripheral_vascular_disease_unspecified':0, 'peripheral_vascular_disease_unspecified_prevalent':1, 'peripheral_vascular_disease_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 369, 191], 'peripheral_vascular_disease_unspecified_pi'))
peritoneal_adhesions_postoperative_postinfection_phe_pi = TensorMap('phecode_568.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_568.1_date'), channel_map={'no_peritoneal_adhesions_postoperative_postinfection':0, 'peritoneal_adhesions_postoperative_postinfection_prevalent':1, 'peritoneal_adhesions_postoperative_postinfection_incident':2}, loss=weighted_crossentropy([1.0, 324, 145], 'peritoneal_adhesions_postoperative_postinfection_pi'))
peritoneal_or_intestinal_adhesions_phe_pi = TensorMap('phecode_560.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_560.3_date'), channel_map={'no_peritoneal_or_intestinal_adhesions':0, 'peritoneal_or_intestinal_adhesions_prevalent':1, 'peritoneal_or_intestinal_adhesions_incident':2}, loss=weighted_crossentropy([1.0, 1089, 596], 'peritoneal_or_intestinal_adhesions_pi'))
peritonitis_and_retroperitoneal_infections_phe_pi = TensorMap('phecode_567',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_567_date'), channel_map={'no_peritonitis_and_retroperitoneal_infections':0, 'peritonitis_and_retroperitoneal_infections_prevalent':1, 'peritonitis_and_retroperitoneal_infections_incident':2}, loss=weighted_crossentropy([1.0, 1041, 614], 'peritonitis_and_retroperitoneal_infections_pi'))
pernicious_anemia_phe_pi = TensorMap('phecode_281.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_281.11_date'), channel_map={'no_pernicious_anemia':0, 'pernicious_anemia_prevalent':1, 'pernicious_anemia_incident':2}, loss=weighted_crossentropy([1.0, 1792, 616], 'pernicious_anemia_pi'))
personal_history_of_diseases_of_blood_and_blood_forming_organs_phe_pi = TensorMap('phecode_289.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_289.3_date'), channel_map={'no_personal_history_of_diseases_of_blood_and_blood_forming_organs':0, 'personal_history_of_diseases_of_blood_and_blood_forming_organs_prevalent':1, 'personal_history_of_diseases_of_blood_and_blood_forming_organs_incident':2}, loss=weighted_crossentropy([1.0, 3205, 1385], 'personal_history_of_diseases_of_blood_and_blood_forming_organs_pi'))
personal_history_of_diseases_of_digestive_system_phe_pi = TensorMap('phecode_564.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_564.9_date'), channel_map={'no_personal_history_of_diseases_of_digestive_system':0, 'personal_history_of_diseases_of_digestive_system_prevalent':1, 'personal_history_of_diseases_of_digestive_system_incident':2}, loss=weighted_crossentropy([1.0, 65, 30], 'personal_history_of_diseases_of_digestive_system_pi'))
personality_disorders_phe_pi = TensorMap('phecode_301',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_301_date'), channel_map={'no_personality_disorders':0, 'personality_disorders_prevalent':1, 'personality_disorders_incident':2}, loss=weighted_crossentropy([1.0, 1445, 1831], 'personality_disorders_pi'))
pervasive_developmental_disorders_phe_pi = TensorMap('phecode_313',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_313_date'), channel_map={'no_pervasive_developmental_disorders':0, 'pervasive_developmental_disorders_prevalent':1, 'pervasive_developmental_disorders_incident':2}, loss=weighted_crossentropy([1.0, 26315, 10638], 'pervasive_developmental_disorders_pi'))
phlebitis_and_thrombophlebitis_phe_pi = TensorMap('phecode_451',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_451_date'), channel_map={'no_phlebitis_and_thrombophlebitis':0, 'phlebitis_and_thrombophlebitis_prevalent':1, 'phlebitis_and_thrombophlebitis_incident':2}, loss=weighted_crossentropy([1.0, 2551, 1519], 'phlebitis_and_thrombophlebitis_pi'))
phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses_phe_pi = TensorMap('phecode_325',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_325_date'), channel_map={'no_phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses':0, 'phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses_prevalent':1, 'phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses_incident':2}, loss=weighted_crossentropy([1.0, 21739, 11627], 'phlebitis_and_thrombophlebitis_of_intracranial_venous_sinuses_pi'))
phlebitis_and_thrombophlebitis_of_lower_extremities_phe_pi = TensorMap('phecode_451.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_451.2_date'), channel_map={'no_phlebitis_and_thrombophlebitis_of_lower_extremities':0, 'phlebitis_and_thrombophlebitis_of_lower_extremities_prevalent':1, 'phlebitis_and_thrombophlebitis_of_lower_extremities_incident':2}, loss=weighted_crossentropy([1.0, 216, 182], 'phlebitis_and_thrombophlebitis_of_lower_extremities_pi'))
phobia_phe_pi = TensorMap('phecode_300.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.13_date'), channel_map={'no_phobia':0, 'phobia_prevalent':1, 'phobia_incident':2}, loss=weighted_crossentropy([1.0, 4132, 555], 'phobia_pi'))
pilonidal_cyst_phe_pi = TensorMap('phecode_686.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_686.3_date'), channel_map={'no_pilonidal_cyst':0, 'pilonidal_cyst_prevalent':1, 'pilonidal_cyst_incident':2}, loss=weighted_crossentropy([1.0, 1075, 2702], 'pilonidal_cyst_pi'))
pituitary_hyperfunction_phe_pi = TensorMap('phecode_253.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_253.1_date'), channel_map={'no_pituitary_hyperfunction':0, 'pituitary_hyperfunction_prevalent':1, 'pituitary_hyperfunction_incident':2}, loss=weighted_crossentropy([1.0, 5617, 9615], 'pituitary_hyperfunction_pi'))
pituitary_hypofunction_phe_pi = TensorMap('phecode_253.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_253.2_date'), channel_map={'no_pituitary_hypofunction':0, 'pituitary_hypofunction_prevalent':1, 'pituitary_hypofunction_incident':2}, loss=weighted_crossentropy([1.0, 2645, 2283], 'pituitary_hypofunction_pi'))
pityriasis_phe_pi = TensorMap('phecode_696.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_696.3_date'), channel_map={'no_pityriasis':0, 'pityriasis_prevalent':1, 'pityriasis_incident':2}, loss=weighted_crossentropy([1.0, 18518, 12195], 'pityriasis_pi'))
placenta_previa_and_abruptio_placenta_phe_pi = TensorMap('phecode_635.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_635.3_date'), channel_map={'no_placenta_previa_and_abruptio_placenta':0, 'placenta_previa_and_abruptio_placenta_prevalent':1, 'placenta_previa_and_abruptio_placenta_incident':2}, loss=weighted_crossentropy([1.0, 312, 5376], 'placenta_previa_and_abruptio_placenta_pi'))
pleurisy_pleural_effusion_phe_pi = TensorMap('phecode_507',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_507_date'), channel_map={'no_pleurisy_pleural_effusion':0, 'pleurisy_pleural_effusion_prevalent':1, 'pleurisy_pleural_effusion_incident':2}, loss=weighted_crossentropy([1.0, 189, 58], 'pleurisy_pleural_effusion_pi'))
pneumococcal_pneumonia_phe_pi = TensorMap('phecode_480.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480.11_date'), channel_map={'no_pneumococcal_pneumonia':0, 'pneumococcal_pneumonia_prevalent':1, 'pneumococcal_pneumonia_incident':2}, loss=weighted_crossentropy([1.0, 248, 62], 'pneumococcal_pneumonia_pi'))
pneumoconiosis_phe_pi = TensorMap('phecode_500.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_500.2_date'), channel_map={'no_pneumoconiosis':0, 'pneumoconiosis_prevalent':1, 'pneumoconiosis_incident':2}, loss=weighted_crossentropy([1.0, 2232, 325], 'pneumoconiosis_pi'))
pneumonia_phe_pi = TensorMap('phecode_480',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480_date'), channel_map={'no_pneumonia':0, 'pneumonia_prevalent':1, 'pneumonia_incident':2}, loss=weighted_crossentropy([1.0, 353, 98], 'pneumonia_pi'))
pneumonia_due_to_fungus_mycoses_phe_pi = TensorMap('phecode_480.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480.3_date'), channel_map={'no_pneumonia_due_to_fungus_mycoses':0, 'pneumonia_due_to_fungus_mycoses_prevalent':1, 'pneumonia_due_to_fungus_mycoses_incident':2}, loss=weighted_crossentropy([1.0, 11904, 3571], 'pneumonia_due_to_fungus_mycoses_pi'))
pneumonitis_due_to_inhalation_of_food_or_vomitus_phe_pi = TensorMap('phecode_501',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_501_date'), channel_map={'no_pneumonitis_due_to_inhalation_of_food_or_vomitus':0, 'pneumonitis_due_to_inhalation_of_food_or_vomitus_prevalent':1, 'pneumonitis_due_to_inhalation_of_food_or_vomitus_incident':2}, loss=weighted_crossentropy([1.0, 5813, 485], 'pneumonitis_due_to_inhalation_of_food_or_vomitus_pi'))
poisoning_allergy_of_sulfonamides_phe_pi = TensorMap('phecode_961.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_961.1_date'), channel_map={'no_poisoning_allergy_of_sulfonamides':0, 'poisoning_allergy_of_sulfonamides_prevalent':1, 'poisoning_allergy_of_sulfonamides_incident':2}, loss=weighted_crossentropy([1.0, 988, 674], 'poisoning_allergy_of_sulfonamides_pi'))
poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system_phe_pi = TensorMap('phecode_975',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_975_date'), channel_map={'no_poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system':0, 'poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system_prevalent':1, 'poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system_incident':2}, loss=weighted_crossentropy([1.0, 13888, 13157], 'poisoning_by_agents_primarily_acting_on_the_smooth_and_skeletal_muscles_and_respiratory_system_pi'))
poisoning_by_agents_primarily_affecting_blood_constituents_phe_pi = TensorMap('phecode_964',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_964_date'), channel_map={'no_poisoning_by_agents_primarily_affecting_blood_constituents':0, 'poisoning_by_agents_primarily_affecting_blood_constituents_prevalent':1, 'poisoning_by_agents_primarily_affecting_blood_constituents_incident':2}, loss=weighted_crossentropy([1.0, 38461, 62500], 'poisoning_by_agents_primarily_affecting_blood_constituents_pi'))
poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs_phe_pi = TensorMap('phecode_976',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_976_date'), channel_map={'no_poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs':0, 'poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs_prevalent':1, 'poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs_incident':2}, loss=weighted_crossentropy([1.0, 13888, 13888], 'poisoning_by_agents_primarily_affecting_skin_mucous_membrane_ophthalmological_otorhinolaryngological_dental_drugs_pi'))
poisoning_by_agents_primarily_affecting_the_cardiovascular_system_phe_pi = TensorMap('phecode_972',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_972_date'), channel_map={'no_poisoning_by_agents_primarily_affecting_the_cardiovascular_system':0, 'poisoning_by_agents_primarily_affecting_the_cardiovascular_system_prevalent':1, 'poisoning_by_agents_primarily_affecting_the_cardiovascular_system_incident':2}, loss=weighted_crossentropy([1.0, 4385, 2358], 'poisoning_by_agents_primarily_affecting_the_cardiovascular_system_pi'))
poisoning_by_agents_primarily_affecting_the_gastrointestinal_system_phe_pi = TensorMap('phecode_973',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_973_date'), channel_map={'no_poisoning_by_agents_primarily_affecting_the_gastrointestinal_system':0, 'poisoning_by_agents_primarily_affecting_the_gastrointestinal_system_prevalent':1, 'poisoning_by_agents_primarily_affecting_the_gastrointestinal_system_incident':2}, loss=weighted_crossentropy([1.0, 8928, 5376], 'poisoning_by_agents_primarily_affecting_the_gastrointestinal_system_pi'))
poisoning_by_analgesics_antipyretics_and_antirheumatics_phe_pi = TensorMap('phecode_965',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_965_date'), channel_map={'no_poisoning_by_analgesics_antipyretics_and_antirheumatics':0, 'poisoning_by_analgesics_antipyretics_and_antirheumatics_prevalent':1, 'poisoning_by_analgesics_antipyretics_and_antirheumatics_incident':2}, loss=weighted_crossentropy([1.0, 162, 85], 'poisoning_by_analgesics_antipyretics_and_antirheumatics_pi'))
poisoning_by_antibiotics_phe_pi = TensorMap('phecode_960',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_960_date'), channel_map={'no_poisoning_by_antibiotics':0, 'poisoning_by_antibiotics_prevalent':1, 'poisoning_by_antibiotics_incident':2}, loss=weighted_crossentropy([1.0, 328, 125], 'poisoning_by_antibiotics_pi'))
poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs_phe_pi = TensorMap('phecode_966',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_966_date'), channel_map={'no_poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs':0, 'poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs_prevalent':1, 'poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs_incident':2}, loss=weighted_crossentropy([1.0, 1228, 1760], 'poisoning_by_anticonvulsants_and_anti_parkinsonism_drugs_pi'))
poisoning_by_antifungal_antibiotics_phe_pi = TensorMap('phecode_960.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_960.3_date'), channel_map={'no_poisoning_by_antifungal_antibiotics':0, 'poisoning_by_antifungal_antibiotics_prevalent':1, 'poisoning_by_antifungal_antibiotics_incident':2}, loss=weighted_crossentropy([1.0, 100000, 29411], 'poisoning_by_antifungal_antibiotics_pi'))
poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system_phe_pi = TensorMap('phecode_971',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_971_date'), channel_map={'no_poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system':0, 'poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system_prevalent':1, 'poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system_incident':2}, loss=weighted_crossentropy([1.0, 3787, 4201], 'poisoning_by_drugs_primarily_affecting_the_autonomic_nervous_system_pi'))
poisoning_by_hormones_and_synthetic_substitutes_phe_pi = TensorMap('phecode_962',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_962_date'), channel_map={'no_poisoning_by_hormones_and_synthetic_substitutes':0, 'poisoning_by_hormones_and_synthetic_substitutes_prevalent':1, 'poisoning_by_hormones_and_synthetic_substitutes_incident':2}, loss=weighted_crossentropy([1.0, 4854, 4000], 'poisoning_by_hormones_and_synthetic_substitutes_pi'))
poisoning_by_other_anti_infectives_phe_pi = TensorMap('phecode_961',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_961_date'), channel_map={'no_poisoning_by_other_anti_infectives':0, 'poisoning_by_other_anti_infectives_prevalent':1, 'poisoning_by_other_anti_infectives_incident':2}, loss=weighted_crossentropy([1.0, 2358, 1116], 'poisoning_by_other_anti_infectives_pi'))
poisoning_by_primarily_systemic_agents_phe_pi = TensorMap('phecode_963',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_963_date'), channel_map={'no_poisoning_by_primarily_systemic_agents':0, 'poisoning_by_primarily_systemic_agents_prevalent':1, 'poisoning_by_primarily_systemic_agents_incident':2}, loss=weighted_crossentropy([1.0, 4310, 3906], 'poisoning_by_primarily_systemic_agents_pi'))
poisoning_by_psychotropic_agents_phe_pi = TensorMap('phecode_969',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_969_date'), channel_map={'no_poisoning_by_psychotropic_agents':0, 'poisoning_by_psychotropic_agents_prevalent':1, 'poisoning_by_psychotropic_agents_incident':2}, loss=weighted_crossentropy([1.0, 342, 504], 'poisoning_by_psychotropic_agents_pi'))
poisoning_by_water_mineral_and_uric_acid_metabolism_drugs_phe_pi = TensorMap('phecode_974',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_974_date'), channel_map={'no_poisoning_by_water_mineral_and_uric_acid_metabolism_drugs':0, 'poisoning_by_water_mineral_and_uric_acid_metabolism_drugs_prevalent':1, 'poisoning_by_water_mineral_and_uric_acid_metabolism_drugs_incident':2}, loss=weighted_crossentropy([1.0, 14285, 3521], 'poisoning_by_water_mineral_and_uric_acid_metabolism_drugs_pi'))
polyarteritis_nodosa_phe_pi = TensorMap('phecode_446.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.6_date'), channel_map={'no_polyarteritis_nodosa':0, 'polyarteritis_nodosa_prevalent':1, 'polyarteritis_nodosa_incident':2}, loss=weighted_crossentropy([1.0, 22727, 15151], 'polyarteritis_nodosa_pi'))
polyclonal_hypergammaglobulinemia_phe_pi = TensorMap('phecode_270.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_270.31_date'), channel_map={'no_polyclonal_hypergammaglobulinemia':0, 'polyclonal_hypergammaglobulinemia_prevalent':1, 'polyclonal_hypergammaglobulinemia_incident':2}, loss=weighted_crossentropy([1.0, 41666, 31250], 'polyclonal_hypergammaglobulinemia_pi'))
polycystic_ovaries_phe_pi = TensorMap('phecode_256.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_256.4_date'), channel_map={'no_polycystic_ovaries':0, 'polycystic_ovaries_prevalent':1, 'polycystic_ovaries_incident':2}, loss=weighted_crossentropy([1.0, 2631, 3816], 'polycystic_ovaries_pi'))
polycythemia_secondary_phe_pi = TensorMap('phecode_289.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_289.8_date'), channel_map={'no_polycythemia_secondary':0, 'polycythemia_secondary_prevalent':1, 'polycythemia_secondary_incident':2}, loss=weighted_crossentropy([1.0, 4098, 1453], 'polycythemia_secondary_pi'))
polycythemia_vera_phe_pi = TensorMap('phecode_200.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_200.1_date'), channel_map={'no_polycythemia_vera':0, 'polycythemia_vera_prevalent':1, 'polycythemia_vera_incident':2}, loss=weighted_crossentropy([1.0, 1650, 2000], 'polycythemia_vera_pi'))
polydipsia_phe_pi = TensorMap('phecode_276.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_276.8_date'), channel_map={'no_polydipsia':0, 'polydipsia_prevalent':1, 'polydipsia_incident':2}, loss=weighted_crossentropy([1.0, 5617, 6250], 'polydipsia_pi'))
polymyalgia_rheumatica_phe_pi = TensorMap('phecode_717',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_717_date'), channel_map={'no_polymyalgia_rheumatica':0, 'polymyalgia_rheumatica_prevalent':1, 'polymyalgia_rheumatica_incident':2}, loss=weighted_crossentropy([1.0, 1811, 296], 'polymyalgia_rheumatica_pi'))
polymyositis_phe_pi = TensorMap('phecode_709.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_709.4_date'), channel_map={'no_polymyositis':0, 'polymyositis_prevalent':1, 'polymyositis_incident':2}, loss=weighted_crossentropy([1.0, 15625, 16129], 'polymyositis_pi'))
polyneuropathy_due_to_drugs_phe_pi = TensorMap('phecode_316.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_316.1_date'), channel_map={'no_polyneuropathy_due_to_drugs':0, 'polyneuropathy_due_to_drugs_prevalent':1, 'polyneuropathy_due_to_drugs_incident':2}, loss=weighted_crossentropy([1.0, 38461, 4901], 'polyneuropathy_due_to_drugs_pi'))
polyneuropathy_in_diabetes_phe_pi = TensorMap('phecode_250.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.6_date'), channel_map={'no_polyneuropathy_in_diabetes':0, 'polyneuropathy_in_diabetes_prevalent':1, 'polyneuropathy_in_diabetes_incident':2}, loss=weighted_crossentropy([1.0, 4424, 891], 'polyneuropathy_in_diabetes_pi'))
polyp_of_corpus_uteri_phe_pi = TensorMap('phecode_622.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_622.1_date'), channel_map={'no_polyp_of_corpus_uteri':0, 'polyp_of_corpus_uteri_prevalent':1, 'polyp_of_corpus_uteri_incident':2}, loss=weighted_crossentropy([1.0, 77, 122], 'polyp_of_corpus_uteri_pi'))
polyp_of_female_genital_organs_phe_pi = TensorMap('phecode_622',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_622_date'), channel_map={'no_polyp_of_female_genital_organs':0, 'polyp_of_female_genital_organs_prevalent':1, 'polyp_of_female_genital_organs_incident':2}, loss=weighted_crossentropy([1.0, 2645, 5263], 'polyp_of_female_genital_organs_pi'))
polyphagia_phe_pi = TensorMap('phecode_260.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_260.7_date'), channel_map={'no_polyphagia':0, 'polyphagia_prevalent':1, 'polyphagia_incident':2}, loss=weighted_crossentropy([1.0, 166666, 166666], 'polyphagia_pi'))
portal_hypertension_phe_pi = TensorMap('phecode_571.81',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_571.81_date'), channel_map={'no_portal_hypertension':0, 'portal_hypertension_prevalent':1, 'portal_hypertension_incident':2}, loss=weighted_crossentropy([1.0, 2747, 712], 'portal_hypertension_pi'))
postablative_ovarian_failure_phe_pi = TensorMap('phecode_258.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_258.1_date'), channel_map={'no_postablative_ovarian_failure':0, 'postablative_ovarian_failure_prevalent':1, 'postablative_ovarian_failure_incident':2}, loss=weighted_crossentropy([1.0, 19230, 125000], 'postablative_ovarian_failure_pi'))
postinflammatory_pulmonary_fibrosis_phe_pi = TensorMap('phecode_502',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_502_date'), channel_map={'no_postinflammatory_pulmonary_fibrosis':0, 'postinflammatory_pulmonary_fibrosis_prevalent':1, 'postinflammatory_pulmonary_fibrosis_incident':2}, loss=weighted_crossentropy([1.0, 2024, 396], 'postinflammatory_pulmonary_fibrosis_pi'))
postlaminectomy_syndrome_phe_pi = TensorMap('phecode_722.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722.8_date'), channel_map={'no_postlaminectomy_syndrome':0, 'postlaminectomy_syndrome_prevalent':1, 'postlaminectomy_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 15151, 8620], 'postlaminectomy_syndrome_pi'))
postmenopausal_atrophic_vaginitis_phe_pi = TensorMap('phecode_627.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_627.3_date'), channel_map={'no_postmenopausal_atrophic_vaginitis':0, 'postmenopausal_atrophic_vaginitis_prevalent':1, 'postmenopausal_atrophic_vaginitis_incident':2}, loss=weighted_crossentropy([1.0, 794, 539], 'postmenopausal_atrophic_vaginitis_pi'))
postmenopausal_bleeding_phe_pi = TensorMap('phecode_627.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_627.1_date'), channel_map={'no_postmenopausal_bleeding':0, 'postmenopausal_bleeding_prevalent':1, 'postmenopausal_bleeding_incident':2}, loss=weighted_crossentropy([1.0, 64, 125], 'postmenopausal_bleeding_pi'))
postoperative_infection_phe_pi = TensorMap('phecode_80',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_80_date'), channel_map={'no_postoperative_infection':0, 'postoperative_infection_prevalent':1, 'postoperative_infection_incident':2}, loss=weighted_crossentropy([1.0, 169, 153], 'postoperative_infection_pi'))
postoperative_shock_phe_pi = TensorMap('phecode_958.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_958.1_date'), channel_map={'no_postoperative_shock':0, 'postoperative_shock_prevalent':1, 'postoperative_shock_incident':2}, loss=weighted_crossentropy([1.0, 15625, 11904], 'postoperative_shock_pi'))
postphlebitic_syndrome_phe_pi = TensorMap('phecode_452.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_452.8_date'), channel_map={'no_postphlebitic_syndrome':0, 'postphlebitic_syndrome_prevalent':1, 'postphlebitic_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 71428, 35714], 'postphlebitic_syndrome_pi'))
posttraumatic_stress_disorder_phe_pi = TensorMap('phecode_300.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_300.9_date'), channel_map={'no_posttraumatic_stress_disorder':0, 'posttraumatic_stress_disorder_prevalent':1, 'posttraumatic_stress_disorder_incident':2}, loss=weighted_crossentropy([1.0, 10416, 2793], 'posttraumatic_stress_disorder_pi'))
precordial_pain_phe_pi = TensorMap('phecode_418.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_418.1_date'), channel_map={'no_precordial_pain':0, 'precordial_pain_prevalent':1, 'precordial_pain_incident':2}, loss=weighted_crossentropy([1.0, 379, 113], 'precordial_pain_pi'))
preeclampsia_and_eclampsia_phe_pi = TensorMap('phecode_642.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_642.1_date'), channel_map={'no_preeclampsia_and_eclampsia':0, 'preeclampsia_and_eclampsia_prevalent':1, 'preeclampsia_and_eclampsia_incident':2}, loss=weighted_crossentropy([1.0, 1184, 26315], 'preeclampsia_and_eclampsia_pi'))
premature_beats_phe_pi = TensorMap('phecode_427.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.6_date'), channel_map={'no_premature_beats':0, 'premature_beats_prevalent':1, 'premature_beats_incident':2}, loss=weighted_crossentropy([1.0, 2232, 603], 'premature_beats_pi'))
premature_menopause_and_other_ovarian_failure_phe_pi = TensorMap('phecode_627.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_627.5_date'), channel_map={'no_premature_menopause_and_other_ovarian_failure':0, 'premature_menopause_and_other_ovarian_failure_prevalent':1, 'premature_menopause_and_other_ovarian_failure_incident':2}, loss=weighted_crossentropy([1.0, 11904, 31250], 'premature_menopause_and_other_ovarian_failure_pi'))
premenopausal_menorrhagia_phe_pi = TensorMap('phecode_627.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_627.4_date'), channel_map={'no_premenopausal_menorrhagia':0, 'premenopausal_menorrhagia_prevalent':1, 'premenopausal_menorrhagia_incident':2}, loss=weighted_crossentropy([1.0, 1718, 5154], 'premenopausal_menorrhagia_pi'))
premenstrual_tension_syndromes_phe_pi = TensorMap('phecode_626.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_626.4_date'), channel_map={'no_premenstrual_tension_syndromes':0, 'premenstrual_tension_syndromes_prevalent':1, 'premenstrual_tension_syndromes_incident':2}, loss=weighted_crossentropy([1.0, 3267, 21739], 'premenstrual_tension_syndromes_pi'))
presbyopia_phe_pi = TensorMap('phecode_367.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_367.4_date'), channel_map={'no_presbyopia':0, 'presbyopia_prevalent':1, 'presbyopia_incident':2}, loss=weighted_crossentropy([1.0, 62500, 20000], 'presbyopia_pi'))
primary_angle_closure_glaucoma_phe_pi = TensorMap('phecode_365.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_365.2_date'), channel_map={'no_primary_angle_closure_glaucoma':0, 'primary_angle_closure_glaucoma_prevalent':1, 'primary_angle_closure_glaucoma_incident':2}, loss=weighted_crossentropy([1.0, 1824, 608], 'primary_angle_closure_glaucoma_pi'))
primary_biliary_cirrhosis_phe_pi = TensorMap('phecode_571.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_571.6_date'), channel_map={'no_primary_biliary_cirrhosis':0, 'primary_biliary_cirrhosis_prevalent':1, 'primary_biliary_cirrhosis_incident':2}, loss=weighted_crossentropy([1.0, 3676, 2164], 'primary_biliary_cirrhosis_pi'))
primary_intrinsic_cardiomyopathies_phe_pi = TensorMap('phecode_425.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_425.1_date'), channel_map={'no_primary_intrinsic_cardiomyopathies':0, 'primary_intrinsic_cardiomyopathies_prevalent':1, 'primary_intrinsic_cardiomyopathies_incident':2}, loss=weighted_crossentropy([1.0, 1012, 444], 'primary_intrinsic_cardiomyopathies_pi'))
primary_open_angle_glaucoma_phe_pi = TensorMap('phecode_365.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_365.11_date'), channel_map={'no_primary_open_angle_glaucoma':0, 'primary_open_angle_glaucoma_prevalent':1, 'primary_open_angle_glaucoma_incident':2}, loss=weighted_crossentropy([1.0, 1225, 342], 'primary_open_angle_glaucoma_pi'))
primary_pulmonary_hypertension_phe_pi = TensorMap('phecode_415.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_415.21_date'), channel_map={'no_primary_pulmonary_hypertension':0, 'primary_pulmonary_hypertension_prevalent':1, 'primary_pulmonary_hypertension_incident':2}, loss=weighted_crossentropy([1.0, 2369, 1165], 'primary_pulmonary_hypertension_pi'))
primary_thrombocytopenia_phe_pi = TensorMap('phecode_287.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287.31_date'), channel_map={'no_primary_thrombocytopenia':0, 'primary_thrombocytopenia_prevalent':1, 'primary_thrombocytopenia_incident':2}, loss=weighted_crossentropy([1.0, 2109, 1336], 'primary_thrombocytopenia_pi'))
problems_associated_with_amniotic_cavity_and_membranes_phe_pi = TensorMap('phecode_653',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_653_date'), channel_map={'no_problems_associated_with_amniotic_cavity_and_membranes':0, 'problems_associated_with_amniotic_cavity_and_membranes_prevalent':1, 'problems_associated_with_amniotic_cavity_and_membranes_incident':2}, loss=weighted_crossentropy([1.0, 252, 5952], 'problems_associated_with_amniotic_cavity_and_membranes_pi'))
progressive_myopia_phe_pi = TensorMap('phecode_360.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_360.2_date'), channel_map={'no_progressive_myopia':0, 'progressive_myopia_prevalent':1, 'progressive_myopia_incident':2}, loss=weighted_crossentropy([1.0, 3937, 2403], 'progressive_myopia_pi'))
prolapse_of_vaginal_vault_after_hysterectomy_phe_pi = TensorMap('phecode_618.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_618.5_date'), channel_map={'no_prolapse_of_vaginal_vault_after_hysterectomy':0, 'prolapse_of_vaginal_vault_after_hysterectomy_prevalent':1, 'prolapse_of_vaginal_vault_after_hysterectomy_incident':2}, loss=weighted_crossentropy([1.0, 1901, 1308], 'prolapse_of_vaginal_vault_after_hysterectomy_pi'))
prolapse_of_vaginal_walls_phe_pi = TensorMap('phecode_618.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_618.1_date'), channel_map={'no_prolapse_of_vaginal_walls':0, 'prolapse_of_vaginal_walls_prevalent':1, 'prolapse_of_vaginal_walls_incident':2}, loss=weighted_crossentropy([1.0, 103, 102], 'prolapse_of_vaginal_walls_pi'))
proliferative_glomerulonephritis_phe_pi = TensorMap('phecode_580.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_580.11_date'), channel_map={'no_proliferative_glomerulonephritis':0, 'proliferative_glomerulonephritis_prevalent':1, 'proliferative_glomerulonephritis_incident':2}, loss=weighted_crossentropy([1.0, 10416, 9259], 'proliferative_glomerulonephritis_pi'))
prostatitis_phe_pi = TensorMap('phecode_601.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_601.1_date'), channel_map={'no_prostatitis':0, 'prostatitis_prevalent':1, 'prostatitis_incident':2}, loss=weighted_crossentropy([1.0, 1243, 1041], 'prostatitis_pi'))
protein_calorie_malnutrition_phe_pi = TensorMap('phecode_260',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_260_date'), channel_map={'no_protein_calorie_malnutrition':0, 'protein_calorie_malnutrition_prevalent':1, 'protein_calorie_malnutrition_incident':2}, loss=weighted_crossentropy([1.0, 10204, 2109], 'protein_calorie_malnutrition_pi'))
proteinuria_phe_pi = TensorMap('phecode_269',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_269_date'), channel_map={'no_proteinuria':0, 'proteinuria_prevalent':1, 'proteinuria_incident':2}, loss=weighted_crossentropy([1.0, 934, 840], 'proteinuria_pi'))
protozoan_infection_phe_pi = TensorMap('phecode_131',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_131_date'), channel_map={'no_protozoan_infection':0, 'protozoan_infection_prevalent':1, 'protozoan_infection_incident':2}, loss=weighted_crossentropy([1.0, 29411, 41666], 'protozoan_infection_pi'))
prurigo_and_lichen_phe_pi = TensorMap('phecode_695.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.7_date'), channel_map={'no_prurigo_and_lichen':0, 'prurigo_and_lichen_prevalent':1, 'prurigo_and_lichen_incident':2}, loss=weighted_crossentropy([1.0, 1050, 791], 'prurigo_and_lichen_pi'))
pruritus_and_related_conditions_phe_pi = TensorMap('phecode_698',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_698_date'), channel_map={'no_pruritus_and_related_conditions':0, 'pruritus_and_related_conditions_prevalent':1, 'pruritus_and_related_conditions_incident':2}, loss=weighted_crossentropy([1.0, 1030, 763], 'pruritus_and_related_conditions_pi'))
pseudomonal_pneumonia_phe_pi = TensorMap('phecode_480.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480.12_date'), channel_map={'no_pseudomonal_pneumonia':0, 'pseudomonal_pneumonia_prevalent':1, 'pseudomonal_pneumonia_incident':2}, loss=weighted_crossentropy([1.0, 27777, 3937], 'pseudomonal_pneumonia_pi'))
psoriasis_phe_pi = TensorMap('phecode_696.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_696.4_date'), channel_map={'no_psoriasis':0, 'psoriasis_prevalent':1, 'psoriasis_incident':2}, loss=weighted_crossentropy([1.0, 611, 314], 'psoriasis_pi'))
psoriasis_vulgaris_phe_pi = TensorMap('phecode_696.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_696.41_date'), channel_map={'no_psoriasis_vulgaris':0, 'psoriasis_vulgaris_prevalent':1, 'psoriasis_vulgaris_incident':2}, loss=weighted_crossentropy([1.0, 560, 293], 'psoriasis_vulgaris_pi'))
psoriatic_arthropathy_phe_pi = TensorMap('phecode_696.42',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_696.42_date'), channel_map={'no_psoriatic_arthropathy':0, 'psoriatic_arthropathy_prevalent':1, 'psoriatic_arthropathy_incident':2}, loss=weighted_crossentropy([1.0, 1256, 786], 'psoriatic_arthropathy_pi'))
psychogenic_and_somatoform_disorders_phe_pi = TensorMap('phecode_303',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_303_date'), channel_map={'no_psychogenic_and_somatoform_disorders':0, 'psychogenic_and_somatoform_disorders_prevalent':1, 'psychogenic_and_somatoform_disorders_incident':2}, loss=weighted_crossentropy([1.0, 100000, 166666], 'psychogenic_and_somatoform_disorders_pi'))
psychogenic_disorder_phe_pi = TensorMap('phecode_303.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_303.3_date'), channel_map={'no_psychogenic_disorder':0, 'psychogenic_disorder_prevalent':1, 'psychogenic_disorder_incident':2}, loss=weighted_crossentropy([1.0, 875, 816], 'psychogenic_disorder_pi'))
psychosis_phe_pi = TensorMap('phecode_295.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_295.3_date'), channel_map={'no_psychosis':0, 'psychosis_prevalent':1, 'psychosis_incident':2}, loss=weighted_crossentropy([1.0, 4424, 2793], 'psychosis_pi'))
ptosis_of_eyelid_phe_pi = TensorMap('phecode_374.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_374.3_date'), channel_map={'no_ptosis_of_eyelid':0, 'ptosis_of_eyelid_prevalent':1, 'ptosis_of_eyelid_incident':2}, loss=weighted_crossentropy([1.0, 481, 304], 'ptosis_of_eyelid_pi'))
pulmonary_collapse_interstitial_and_compensatory_emphysema_phe_pi = TensorMap('phecode_508',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_508_date'), channel_map={'no_pulmonary_collapse_interstitial_and_compensatory_emphysema':0, 'pulmonary_collapse_interstitial_and_compensatory_emphysema_prevalent':1, 'pulmonary_collapse_interstitial_and_compensatory_emphysema_incident':2}, loss=weighted_crossentropy([1.0, 963, 163], 'pulmonary_collapse_interstitial_and_compensatory_emphysema_pi'))
pulmonary_congestion_and_hypostasis_phe_pi = TensorMap('phecode_503',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_503_date'), channel_map={'no_pulmonary_congestion_and_hypostasis':0, 'pulmonary_congestion_and_hypostasis_prevalent':1, 'pulmonary_congestion_and_hypostasis_incident':2}, loss=weighted_crossentropy([1.0, 2109, 915], 'pulmonary_congestion_and_hypostasis_pi'))
pulmonary_embolism_and_infarction_acute_phe_pi = TensorMap('phecode_415.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_415.11_date'), channel_map={'no_pulmonary_embolism_and_infarction_acute':0, 'pulmonary_embolism_and_infarction_acute_prevalent':1, 'pulmonary_embolism_and_infarction_acute_incident':2}, loss=weighted_crossentropy([1.0, 29411, 9615], 'pulmonary_embolism_and_infarction_acute_pi'))
pulmonary_heart_disease_phe_pi = TensorMap('phecode_415',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_415_date'), channel_map={'no_pulmonary_heart_disease':0, 'pulmonary_heart_disease_prevalent':1, 'pulmonary_heart_disease_incident':2}, loss=weighted_crossentropy([1.0, 293, 134], 'pulmonary_heart_disease_pi'))
pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery_phe_pi = TensorMap('phecode_509.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_509.3_date'), channel_map={'no_pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery':0, 'pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery_prevalent':1, 'pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery_incident':2}, loss=weighted_crossentropy([1.0, 50000, 125000], 'pulmonary_insufficiency_or_respiratory_failure_following_trauma_and_surgery_pi'))
pulpitis_and_necrosis_of_tooth_pulp_phe_pi = TensorMap('phecode_522.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_522.1_date'), channel_map={'no_pulpitis_and_necrosis_of_tooth_pulp':0, 'pulpitis_and_necrosis_of_tooth_pulp_prevalent':1, 'pulpitis_and_necrosis_of_tooth_pulp_incident':2}, loss=weighted_crossentropy([1.0, 14705, 12195], 'pulpitis_and_necrosis_of_tooth_pulp_pi'))
purpura_and_other_hemorrhagic_conditions_phe_pi = TensorMap('phecode_287',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287_date'), channel_map={'no_purpura_and_other_hemorrhagic_conditions':0, 'purpura_and_other_hemorrhagic_conditions_prevalent':1, 'purpura_and_other_hemorrhagic_conditions_incident':2}, loss=weighted_crossentropy([1.0, 15625, 4237], 'purpura_and_other_hemorrhagic_conditions_pi'))
pyelonephritis_phe_pi = TensorMap('phecode_590',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_590_date'), channel_map={'no_pyelonephritis':0, 'pyelonephritis_prevalent':1, 'pyelonephritis_incident':2}, loss=weighted_crossentropy([1.0, 256, 128], 'pyelonephritis_pi'))
pyoderma_phe_pi = TensorMap('phecode_686.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_686.5_date'), channel_map={'no_pyoderma':0, 'pyoderma_prevalent':1, 'pyoderma_incident':2}, loss=weighted_crossentropy([1.0, 16666, 12500], 'pyoderma_pi'))
pyogenic_arthritis_phe_pi = TensorMap('phecode_711.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_711.1_date'), channel_map={'no_pyogenic_arthritis':0, 'pyogenic_arthritis_prevalent':1, 'pyogenic_arthritis_incident':2}, loss=weighted_crossentropy([1.0, 10638, 12195], 'pyogenic_arthritis_pi'))
pyogenic_granuloma_phe_pi = TensorMap('phecode_686.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_686.4_date'), channel_map={'no_pyogenic_granuloma':0, 'pyogenic_granuloma_prevalent':1, 'pyogenic_granuloma_incident':2}, loss=weighted_crossentropy([1.0, 1879, 2202], 'pyogenic_granuloma_pi'))
qualitative_platelet_defects_phe_pi = TensorMap('phecode_287.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287.4_date'), channel_map={'no_qualitative_platelet_defects':0, 'qualitative_platelet_defects_prevalent':1, 'qualitative_platelet_defects_incident':2}, loss=weighted_crossentropy([1.0, 55555, 50000], 'qualitative_platelet_defects_pi'))
radiotherapy_phe_pi = TensorMap('phecode_196',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_196_date'), channel_map={'no_radiotherapy':0, 'radiotherapy_prevalent':1, 'radiotherapy_incident':2}, loss=weighted_crossentropy([1.0, 166666, 125000], 'radiotherapy_pi'))
rash_and_other_nonspecific_skin_eruption_phe_pi = TensorMap('phecode_687.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_687.1_date'), channel_map={'no_rash_and_other_nonspecific_skin_eruption':0, 'rash_and_other_nonspecific_skin_eruption_prevalent':1, 'rash_and_other_nonspecific_skin_eruption_incident':2}, loss=weighted_crossentropy([1.0, 469, 226], 'rash_and_other_nonspecific_skin_eruption_pi'))
raynauds_syndrome_phe_pi = TensorMap('phecode_443.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_443.1_date'), channel_map={'no_raynauds_syndrome':0, 'raynauds_syndrome_prevalent':1, 'raynauds_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 1039, 381], 'raynauds_syndrome_pi'))
redundant_prepuce_and_phimosis_bxo_phe_pi = TensorMap('phecode_604.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_604.1_date'), channel_map={'no_redundant_prepuce_and_phimosis_bxo':0, 'redundant_prepuce_and_phimosis_bxo_prevalent':1, 'redundant_prepuce_and_phimosis_bxo_incident':2}, loss=weighted_crossentropy([1.0, 258, 289], 'redundant_prepuce_and_phimosis_bxo_pi'))
reflux_esophagitis_phe_pi = TensorMap('phecode_530.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.14_date'), channel_map={'no_reflux_esophagitis':0, 'reflux_esophagitis_prevalent':1, 'reflux_esophagitis_incident':2}, loss=weighted_crossentropy([1.0, 69, 65], 'reflux_esophagitis_pi'))
regional_enteritis_phe_pi = TensorMap('phecode_555.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_555.1_date'), channel_map={'no_regional_enteritis':0, 'regional_enteritis_prevalent':1, 'regional_enteritis_incident':2}, loss=weighted_crossentropy([1.0, 370, 484], 'regional_enteritis_pi'))
reiters_disease_phe_pi = TensorMap('phecode_711.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_711.2_date'), channel_map={'no_reiters_disease':0, 'reiters_disease_prevalent':1, 'reiters_disease_incident':2}, loss=weighted_crossentropy([1.0, 35714, 33333], 'reiters_disease_pi'))
renal_colic_phe_pi = TensorMap('phecode_594.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_594.8_date'), channel_map={'no_renal_colic':0, 'renal_colic_prevalent':1, 'renal_colic_incident':2}, loss=weighted_crossentropy([1.0, 282, 720], 'renal_colic_pi'))
renal_dialysis_phe_pi = TensorMap('phecode_585.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_585.31_date'), channel_map={'no_renal_dialysis':0, 'renal_dialysis_prevalent':1, 'renal_dialysis_incident':2}, loss=weighted_crossentropy([1.0, 1567, 965], 'renal_dialysis_pi'))
renal_failure_phe_pi = TensorMap('phecode_585',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_585_date'), channel_map={'no_renal_failure':0, 'renal_failure_prevalent':1, 'renal_failure_incident':2}, loss=weighted_crossentropy([1.0, 16129, 11363], 'renal_failure_pi'))
renal_failure_nos_phe_pi = TensorMap('phecode_585.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_585.2_date'), channel_map={'no_renal_failure_nos':0, 'renal_failure_nos_prevalent':1, 'renal_failure_nos_incident':2}, loss=weighted_crossentropy([1.0, 261, 34], 'renal_failure_nos_pi'))
renal_osteodystrophy_phe_pi = TensorMap('phecode_588.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_588.1_date'), channel_map={'no_renal_osteodystrophy':0, 'renal_osteodystrophy_prevalent':1, 'renal_osteodystrophy_incident':2}, loss=weighted_crossentropy([1.0, 100000, 71428], 'renal_osteodystrophy_pi'))
respiratory_arrest_phe_pi = TensorMap('phecode_509.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_509.5_date'), channel_map={'no_respiratory_arrest':0, 'respiratory_arrest_prevalent':1, 'respiratory_arrest_incident':2}, loss=weighted_crossentropy([1.0, 8928, 3424], 'respiratory_arrest_pi'))
respiratory_complications_phe_pi = TensorMap('phecode_519.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_519.2_date'), channel_map={'no_respiratory_complications':0, 'respiratory_complications_prevalent':1, 'respiratory_complications_incident':2}, loss=weighted_crossentropy([1.0, 3067, 1968], 'respiratory_complications_pi'))
respiratory_failure_phe_pi = TensorMap('phecode_509.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_509.1_date'), channel_map={'no_respiratory_failure':0, 'respiratory_failure_prevalent':1, 'respiratory_failure_incident':2}, loss=weighted_crossentropy([1.0, 1355, 183], 'respiratory_failure_pi'))
respiratory_insufficiency_phe_pi = TensorMap('phecode_509.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_509.2_date'), channel_map={'no_respiratory_insufficiency':0, 'respiratory_insufficiency_prevalent':1, 'respiratory_insufficiency_incident':2}, loss=weighted_crossentropy([1.0, 600, 122], 'respiratory_insufficiency_pi'))
retention_of_urine_phe_pi = TensorMap('phecode_599.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.2_date'), channel_map={'no_retention_of_urine':0, 'retention_of_urine_prevalent':1, 'retention_of_urine_incident':2}, loss=weighted_crossentropy([1.0, 164, 64], 'retention_of_urine_pi'))
retinal_detachment_with_retinal_defect_phe_pi = TensorMap('phecode_361.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_361.1_date'), channel_map={'no_retinal_detachment_with_retinal_defect':0, 'retinal_detachment_with_retinal_defect_prevalent':1, 'retinal_detachment_with_retinal_defect_incident':2}, loss=weighted_crossentropy([1.0, 672, 405], 'retinal_detachment_with_retinal_defect_pi'))
retinal_detachments_and_defects_phe_pi = TensorMap('phecode_361',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_361_date'), channel_map={'no_retinal_detachments_and_defects':0, 'retinal_detachments_and_defects_prevalent':1, 'retinal_detachments_and_defects_incident':2}, loss=weighted_crossentropy([1.0, 346, 364], 'retinal_detachments_and_defects_pi'))
retinal_hemorrhage_ischemia_phe_pi = TensorMap('phecode_362.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.8_date'), channel_map={'no_retinal_hemorrhage_ischemia':0, 'retinal_hemorrhage_ischemia_prevalent':1, 'retinal_hemorrhage_ischemia_incident':2}, loss=weighted_crossentropy([1.0, 8196, 5208], 'retinal_hemorrhage_ischemia_pi'))
retinal_vascular_changes_and_abnomalities_phe_pi = TensorMap('phecode_362.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.4_date'), channel_map={'no_retinal_vascular_changes_and_abnomalities':0, 'retinal_vascular_changes_and_abnomalities_prevalent':1, 'retinal_vascular_changes_and_abnomalities_incident':2}, loss=weighted_crossentropy([1.0, 1168, 554], 'retinal_vascular_changes_and_abnomalities_pi'))
retinopathy_of_prematurity_phe_pi = TensorMap('phecode_362.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.1_date'), channel_map={'no_retinopathy_of_prematurity':0, 'retinopathy_of_prematurity_prevalent':1, 'retinopathy_of_prematurity_incident':2}, loss=weighted_crossentropy([1.0, 125000, 166666], 'retinopathy_of_prematurity_pi'))
retinoschisis_and_retinal_cysts_phe_pi = TensorMap('phecode_361.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_361.2_date'), channel_map={'no_retinoschisis_and_retinal_cysts':0, 'retinoschisis_and_retinal_cysts_prevalent':1, 'retinoschisis_and_retinal_cysts_incident':2}, loss=weighted_crossentropy([1.0, 12500, 6849], 'retinoschisis_and_retinal_cysts_pi'))
rhabdomyolysis_phe_pi = TensorMap('phecode_772.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_772.4_date'), channel_map={'no_rhabdomyolysis':0, 'rhabdomyolysis_prevalent':1, 'rhabdomyolysis_incident':2}, loss=weighted_crossentropy([1.0, 83333, 166666], 'rhabdomyolysis_pi'))
rhesus_isoimmunization_in_pregnancy_phe_pi = TensorMap('phecode_654.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_654.2_date'), channel_map={'no_rhesus_isoimmunization_in_pregnancy':0, 'rhesus_isoimmunization_in_pregnancy_prevalent':1, 'rhesus_isoimmunization_in_pregnancy_incident':2}, loss=weighted_crossentropy([1.0, 1396, 55555], 'rhesus_isoimmunization_in_pregnancy_pi'))
rheumatic_disease_of_the_heart_valves_phe_pi = TensorMap('phecode_394',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_394_date'), channel_map={'no_rheumatic_disease_of_the_heart_valves':0, 'rheumatic_disease_of_the_heart_valves_prevalent':1, 'rheumatic_disease_of_the_heart_valves_incident':2}, loss=weighted_crossentropy([1.0, 1824, 356], 'rheumatic_disease_of_the_heart_valves_pi'))
rheumatic_fever_chorea_phe_pi = TensorMap('phecode_41.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_41.21_date'), channel_map={'no_rheumatic_fever_chorea':0, 'rheumatic_fever_chorea_prevalent':1, 'rheumatic_fever_chorea_incident':2}, loss=weighted_crossentropy([1.0, 50000, 27777], 'rheumatic_fever_chorea_pi'))
rheumatism_unspecified_and_fibrositis_phe_pi = TensorMap('phecode_729.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_729.1_date'), channel_map={'no_rheumatism_unspecified_and_fibrositis':0, 'rheumatism_unspecified_and_fibrositis_prevalent':1, 'rheumatism_unspecified_and_fibrositis_incident':2}, loss=weighted_crossentropy([1.0, 2564, 1811], 'rheumatism_unspecified_and_fibrositis_pi'))
rheumatoid_arthritis_phe_pi = TensorMap('phecode_714.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_714.1_date'), channel_map={'no_rheumatoid_arthritis':0, 'rheumatoid_arthritis_prevalent':1, 'rheumatoid_arthritis_incident':2}, loss=weighted_crossentropy([1.0, 557, 160], 'rheumatoid_arthritis_pi'))
rheumatoid_arthritis_and_other_inflammatory_polyarthropathies_phe_pi = TensorMap('phecode_714',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_714_date'), channel_map={'no_rheumatoid_arthritis_and_other_inflammatory_polyarthropathies':0, 'rheumatoid_arthritis_and_other_inflammatory_polyarthropathies_prevalent':1, 'rheumatoid_arthritis_and_other_inflammatory_polyarthropathies_incident':2}, loss=weighted_crossentropy([1.0, 282, 48], 'rheumatoid_arthritis_and_other_inflammatory_polyarthropathies_pi'))
rickets_or_osteomalacia_phe_pi = TensorMap('phecode_261.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_261.41_date'), channel_map={'no_rickets_or_osteomalacia':0, 'rickets_or_osteomalacia_prevalent':1, 'rickets_or_osteomalacia_incident':2}, loss=weighted_crossentropy([1.0, 22727, 31250], 'rickets_or_osteomalacia_pi'))
right_bundle_branch_block_phe_pi = TensorMap('phecode_426.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.31_date'), channel_map={'no_right_bundle_branch_block':0, 'right_bundle_branch_block_prevalent':1, 'right_bundle_branch_block_incident':2}, loss=weighted_crossentropy([1.0, 976, 236], 'right_bundle_branch_block_pi'))
rosacea_phe_pi = TensorMap('phecode_695.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.3_date'), channel_map={'no_rosacea':0, 'rosacea_prevalent':1, 'rosacea_incident':2}, loss=weighted_crossentropy([1.0, 3030, 1718], 'rosacea_pi'))
rupture_of_synovium_phe_pi = TensorMap('phecode_727.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727.5_date'), channel_map={'no_rupture_of_synovium':0, 'rupture_of_synovium_prevalent':1, 'rupture_of_synovium_incident':2}, loss=weighted_crossentropy([1.0, 2808, 1305], 'rupture_of_synovium_pi'))
rupture_of_tendon_nontraumatic_phe_pi = TensorMap('phecode_727.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727.6_date'), channel_map={'no_rupture_of_tendon_nontraumatic':0, 'rupture_of_tendon_nontraumatic_prevalent':1, 'rupture_of_tendon_nontraumatic_incident':2}, loss=weighted_crossentropy([1.0, 5434, 15151], 'rupture_of_tendon_nontraumatic_pi'))
sacroiliitis_nec_phe_pi = TensorMap('phecode_715.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_715.1_date'), channel_map={'no_sacroiliitis_nec':0, 'sacroiliitis_nec_prevalent':1, 'sacroiliitis_nec_incident':2}, loss=weighted_crossentropy([1.0, 20000, 4716], 'sacroiliitis_nec_pi'))
salicylates_causing_adverse_effects_in_therapeutic_use_phe_pi = TensorMap('phecode_965.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_965.3_date'), channel_map={'no_salicylates_causing_adverse_effects_in_therapeutic_use':0, 'salicylates_causing_adverse_effects_in_therapeutic_use_prevalent':1, 'salicylates_causing_adverse_effects_in_therapeutic_use_incident':2}, loss=weighted_crossentropy([1.0, 1872, 2118], 'salicylates_causing_adverse_effects_in_therapeutic_use_pi'))
sarcoidosis_phe_pi = TensorMap('phecode_697',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_697_date'), channel_map={'no_sarcoidosis':0, 'sarcoidosis_prevalent':1, 'sarcoidosis_incident':2}, loss=weighted_crossentropy([1.0, 844, 478], 'sarcoidosis_pi'))
scar_conditions_and_fibrosis_of_skin_phe_pi = TensorMap('phecode_701.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_701.2_date'), channel_map={'no_scar_conditions_and_fibrosis_of_skin':0, 'scar_conditions_and_fibrosis_of_skin_prevalent':1, 'scar_conditions_and_fibrosis_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 317, 320], 'scar_conditions_and_fibrosis_of_skin_pi'))
schizoid_personality_disorder_phe_pi = TensorMap('phecode_301.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_301.1_date'), channel_map={'no_schizoid_personality_disorder':0, 'schizoid_personality_disorder_prevalent':1, 'schizoid_personality_disorder_incident':2}, loss=weighted_crossentropy([1.0, 21739, 17241], 'schizoid_personality_disorder_pi'))
schizophrenia_phe_pi = TensorMap('phecode_295.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_295.1_date'), channel_map={'no_schizophrenia':0, 'schizophrenia_prevalent':1, 'schizophrenia_incident':2}, loss=weighted_crossentropy([1.0, 869, 1240], 'schizophrenia_pi'))
schizophrenia_and_other_psychotic_disorders_phe_pi = TensorMap('phecode_295',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_295_date'), channel_map={'no_schizophrenia_and_other_psychotic_disorders':0, 'schizophrenia_and_other_psychotic_disorders_prevalent':1, 'schizophrenia_and_other_psychotic_disorders_incident':2}, loss=weighted_crossentropy([1.0, 50000, 50000], 'schizophrenia_and_other_psychotic_disorders_pi'))
schmorls_nodes_phe_pi = TensorMap('phecode_722.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_722.3_date'), channel_map={'no_schmorls_nodes':0, 'schmorls_nodes_prevalent':1, 'schmorls_nodes_incident':2}, loss=weighted_crossentropy([1.0, 31250, 12500], 'schmorls_nodes_pi'))
sciatica_phe_pi = TensorMap('phecode_764',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_764_date'), channel_map={'no_sciatica':0, 'sciatica_prevalent':1, 'sciatica_incident':2}, loss=weighted_crossentropy([1.0, 641, 412], 'sciatica_pi'))
scleritis_and_episcleritis_phe_pi = TensorMap('phecode_379.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_379.1_date'), channel_map={'no_scleritis_and_episcleritis':0, 'scleritis_and_episcleritis_prevalent':1, 'scleritis_and_episcleritis_incident':2}, loss=weighted_crossentropy([1.0, 20833, 12195], 'scleritis_and_episcleritis_pi'))
sebaceous_cyst_phe_pi = TensorMap('phecode_706.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_706.2_date'), channel_map={'no_sebaceous_cyst':0, 'sebaceous_cyst_prevalent':1, 'sebaceous_cyst_incident':2}, loss=weighted_crossentropy([1.0, 70, 130], 'sebaceous_cyst_pi'))
seborrheic_dermatitis_phe_pi = TensorMap('phecode_690.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_690.1_date'), channel_map={'no_seborrheic_dermatitis':0, 'seborrheic_dermatitis_prevalent':1, 'seborrheic_dermatitis_incident':2}, loss=weighted_crossentropy([1.0, 15151, 8474], 'seborrheic_dermatitis_pi'))
seborrheic_keratosis_phe_pi = TensorMap('phecode_702.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_702.2_date'), channel_map={'no_seborrheic_keratosis':0, 'seborrheic_keratosis_prevalent':1, 'seborrheic_keratosis_incident':2}, loss=weighted_crossentropy([1.0, 263, 208], 'seborrheic_keratosis_pi'))
second_degree_av_block_phe_pi = TensorMap('phecode_426.23',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_426.23_date'), channel_map={'no_second_degree_av_block':0, 'second_degree_av_block_prevalent':1, 'second_degree_av_block_incident':2}, loss=weighted_crossentropy([1.0, 3067, 853], 'second_degree_av_block_pi'))
secondary_diabetes_mellitus_phe_pi = TensorMap('phecode_249',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_249_date'), channel_map={'no_secondary_diabetes_mellitus':0, 'secondary_diabetes_mellitus_prevalent':1, 'secondary_diabetes_mellitus_incident':2}, loss=weighted_crossentropy([1.0, 250000, 125000], 'secondary_diabetes_mellitus_pi'))
secondary_extrinsic_cardiomyopathies_phe_pi = TensorMap('phecode_425.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_425.2_date'), channel_map={'no_secondary_extrinsic_cardiomyopathies':0, 'secondary_extrinsic_cardiomyopathies_prevalent':1, 'secondary_extrinsic_cardiomyopathies_incident':2}, loss=weighted_crossentropy([1.0, 100000, 100000], 'secondary_extrinsic_cardiomyopathies_pi'))
secondary_hypothyroidism_phe_pi = TensorMap('phecode_244.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_244.1_date'), channel_map={'no_secondary_hypothyroidism':0, 'secondary_hypothyroidism_prevalent':1, 'secondary_hypothyroidism_incident':2}, loss=weighted_crossentropy([1.0, 976, 395], 'secondary_hypothyroidism_pi'))
secondary_malignancy_of_bone_phe_pi = TensorMap('phecode_198.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.6_date'), channel_map={'no_secondary_malignancy_of_bone':0, 'secondary_malignancy_of_bone_prevalent':1, 'secondary_malignancy_of_bone_incident':2}, loss=weighted_crossentropy([1.0, 2262, 150], 'secondary_malignancy_of_bone_pi'))
secondary_malignancy_of_brain_spine_phe_pi = TensorMap('phecode_198.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.5_date'), channel_map={'no_secondary_malignancy_of_brain_spine':0, 'secondary_malignancy_of_brain_spine_prevalent':1, 'secondary_malignancy_of_brain_spine_incident':2}, loss=weighted_crossentropy([1.0, 12500, 391], 'secondary_malignancy_of_brain_spine_pi'))
secondary_malignancy_of_lymph_nodes_phe_pi = TensorMap('phecode_198.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.1_date'), channel_map={'no_secondary_malignancy_of_lymph_nodes':0, 'secondary_malignancy_of_lymph_nodes_prevalent':1, 'secondary_malignancy_of_lymph_nodes_incident':2}, loss=weighted_crossentropy([1.0, 233, 83], 'secondary_malignancy_of_lymph_nodes_pi'))
secondary_malignancy_of_respiratory_organs_phe_pi = TensorMap('phecode_198.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.2_date'), channel_map={'no_secondary_malignancy_of_respiratory_organs':0, 'secondary_malignancy_of_respiratory_organs_prevalent':1, 'secondary_malignancy_of_respiratory_organs_incident':2}, loss=weighted_crossentropy([1.0, 2604, 143], 'secondary_malignancy_of_respiratory_organs_pi'))
secondary_malignant_neoplasm_phe_pi = TensorMap('phecode_198',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198_date'), channel_map={'no_secondary_malignant_neoplasm':0, 'secondary_malignant_neoplasm_prevalent':1, 'secondary_malignant_neoplasm_incident':2}, loss=weighted_crossentropy([1.0, 2551, 315], 'secondary_malignant_neoplasm_pi'))
secondary_malignant_neoplasm_of_digestive_systems_phe_pi = TensorMap('phecode_198.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.3_date'), channel_map={'no_secondary_malignant_neoplasm_of_digestive_systems':0, 'secondary_malignant_neoplasm_of_digestive_systems_prevalent':1, 'secondary_malignant_neoplasm_of_digestive_systems_incident':2}, loss=weighted_crossentropy([1.0, 2427, 217], 'secondary_malignant_neoplasm_of_digestive_systems_pi'))
secondary_malignant_neoplasm_of_liver_phe_pi = TensorMap('phecode_198.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.4_date'), channel_map={'no_secondary_malignant_neoplasm_of_liver':0, 'secondary_malignant_neoplasm_of_liver_prevalent':1, 'secondary_malignant_neoplasm_of_liver_incident':2}, loss=weighted_crossentropy([1.0, 1689, 124], 'secondary_malignant_neoplasm_of_liver_pi'))
secondary_malignant_neoplasm_of_skin_phe_pi = TensorMap('phecode_198.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_198.7_date'), channel_map={'no_secondary_malignant_neoplasm_of_skin':0, 'secondary_malignant_neoplasm_of_skin_prevalent':1, 'secondary_malignant_neoplasm_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 13157, 1923], 'secondary_malignant_neoplasm_of_skin_pi'))
secondary_thrombocytopenia_phe_pi = TensorMap('phecode_287.32',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287.32_date'), channel_map={'no_secondary_thrombocytopenia':0, 'secondary_thrombocytopenia_prevalent':1, 'secondary_thrombocytopenia_incident':2}, loss=weighted_crossentropy([1.0, 16666, 4854], 'secondary_thrombocytopenia_pi'))
senile_cataract_phe_pi = TensorMap('phecode_366.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_366.2_date'), channel_map={'no_senile_cataract':0, 'senile_cataract_prevalent':1, 'senile_cataract_incident':2}, loss=weighted_crossentropy([1.0, 171, 41], 'senile_cataract_pi'))
senile_osteoporosis_phe_pi = TensorMap('phecode_743.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_743.12_date'), channel_map={'no_senile_osteoporosis':0, 'senile_osteoporosis_prevalent':1, 'senile_osteoporosis_incident':2}, loss=weighted_crossentropy([1.0, 100000, 14285], 'senile_osteoporosis_pi'))
sensorineural_hearing_loss_phe_pi = TensorMap('phecode_389.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_389.1_date'), channel_map={'no_sensorineural_hearing_loss':0, 'sensorineural_hearing_loss_prevalent':1, 'sensorineural_hearing_loss_incident':2}, loss=weighted_crossentropy([1.0, 3571, 801], 'sensorineural_hearing_loss_pi'))
separation_of_retinal_layers_phe_pi = TensorMap('phecode_362.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_362.31_date'), channel_map={'no_separation_of_retinal_layers':0, 'separation_of_retinal_layers_prevalent':1, 'separation_of_retinal_layers_incident':2}, loss=weighted_crossentropy([1.0, 7936, 6250], 'separation_of_retinal_layers_pi'))
sepsis_phe_pi = TensorMap('phecode_994.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_994.2_date'), channel_map={'no_sepsis':0, 'sepsis_prevalent':1, 'sepsis_incident':2}, loss=weighted_crossentropy([1.0, 773, 112], 'sepsis_pi'))
septal_deviations_turbinate_hypertrophy_phe_pi = TensorMap('phecode_470',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_470_date'), channel_map={'no_septal_deviations_turbinate_hypertrophy':0, 'septal_deviations_turbinate_hypertrophy_prevalent':1, 'septal_deviations_turbinate_hypertrophy_incident':2}, loss=weighted_crossentropy([1.0, 120, 265], 'septal_deviations_turbinate_hypertrophy_pi'))
septicemia_phe_pi = TensorMap('phecode_38',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_38_date'), channel_map={'no_septicemia':0, 'septicemia_prevalent':1, 'septicemia_incident':2}, loss=weighted_crossentropy([1.0, 712, 107], 'septicemia_pi'))
sexual_and_gender_identity_disorders_phe_pi = TensorMap('phecode_302',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_302_date'), channel_map={'no_sexual_and_gender_identity_disorders':0, 'sexual_and_gender_identity_disorders_prevalent':1, 'sexual_and_gender_identity_disorders_incident':2}, loss=weighted_crossentropy([1.0, 2173, 1976], 'sexual_and_gender_identity_disorders_pi'))
sexually_transmitted_infections_not_hiv_or_hepatitis_phe_pi = TensorMap('phecode_90',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_90_date'), channel_map={'no_sexually_transmitted_infections_not_hiv_or_hepatitis':0, 'sexually_transmitted_infections_not_hiv_or_hepatitis_prevalent':1, 'sexually_transmitted_infections_not_hiv_or_hepatitis_incident':2}, loss=weighted_crossentropy([1.0, 1479, 13157], 'sexually_transmitted_infections_not_hiv_or_hepatitis_pi'))
shock_phe_pi = TensorMap('phecode_797',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_797_date'), channel_map={'no_shock':0, 'shock_prevalent':1, 'shock_incident':2}, loss=weighted_crossentropy([1.0, 35714, 7352], 'shock_pi'))
shortness_of_breath_phe_pi = TensorMap('phecode_512.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_512.7_date'), channel_map={'no_shortness_of_breath':0, 'shortness_of_breath_prevalent':1, 'shortness_of_breath_incident':2}, loss=weighted_crossentropy([1.0, 182, 75], 'shortness_of_breath_pi'))
sialoadenitis_phe_pi = TensorMap('phecode_527.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_527.2_date'), channel_map={'no_sialoadenitis':0, 'sialoadenitis_prevalent':1, 'sialoadenitis_incident':2}, loss=weighted_crossentropy([1.0, 2000, 1992], 'sialoadenitis_pi'))
sialolithiasis_phe_pi = TensorMap('phecode_528.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.7_date'), channel_map={'no_sialolithiasis':0, 'sialolithiasis_prevalent':1, 'sialolithiasis_incident':2}, loss=weighted_crossentropy([1.0, 2155, 3333], 'sialolithiasis_pi'))
sicca_syndrome_phe_pi = TensorMap('phecode_709.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_709.2_date'), channel_map={'no_sicca_syndrome':0, 'sicca_syndrome_prevalent':1, 'sicca_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 2032, 883], 'sicca_syndrome_pi'))
sickle_cell_anemia_phe_pi = TensorMap('phecode_282.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_282.5_date'), channel_map={'no_sickle_cell_anemia':0, 'sickle_cell_anemia_prevalent':1, 'sickle_cell_anemia_incident':2}, loss=weighted_crossentropy([1.0, 1976, 2155], 'sickle_cell_anemia_pi'))
sideroblastic_anemia_phe_pi = TensorMap('phecode_285.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_285.3_date'), channel_map={'no_sideroblastic_anemia':0, 'sideroblastic_anemia_prevalent':1, 'sideroblastic_anemia_incident':2}, loss=weighted_crossentropy([1.0, 62500, 27777], 'sideroblastic_anemia_pi'))
simple_and_unspecified_goiter_phe_pi = TensorMap('phecode_240',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_240_date'), channel_map={'no_simple_and_unspecified_goiter':0, 'simple_and_unspecified_goiter_prevalent':1, 'simple_and_unspecified_goiter_incident':2}, loss=weighted_crossentropy([1.0, 1061, 1305], 'simple_and_unspecified_goiter_pi'))
sinoatrial_node_dysfunction_bradycardia_phe_pi = TensorMap('phecode_427.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.8_date'), channel_map={'no_sinoatrial_node_dysfunction_bradycardia':0, 'sinoatrial_node_dysfunction_bradycardia_prevalent':1, 'sinoatrial_node_dysfunction_bradycardia_incident':2}, loss=weighted_crossentropy([1.0, 2762, 967], 'sinoatrial_node_dysfunction_bradycardia_pi'))
skull_and_face_fracture_and_other_intercranial_injury_phe_pi = TensorMap('phecode_819',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_819_date'), channel_map={'no_skull_and_face_fracture_and_other_intercranial_injury':0, 'skull_and_face_fracture_and_other_intercranial_injury_prevalent':1, 'skull_and_face_fracture_and_other_intercranial_injury_incident':2}, loss=weighted_crossentropy([1.0, 539, 1070], 'skull_and_face_fracture_and_other_intercranial_injury_pi'))
sleep_apnea_phe_pi = TensorMap('phecode_327.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327.3_date'), channel_map={'no_sleep_apnea':0, 'sleep_apnea_prevalent':1, 'sleep_apnea_incident':2}, loss=weighted_crossentropy([1.0, 201, 114], 'sleep_apnea_pi'))
sleep_disorders_phe_pi = TensorMap('phecode_327',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_327_date'), channel_map={'no_sleep_disorders':0, 'sleep_disorders_prevalent':1, 'sleep_disorders_incident':2}, loss=weighted_crossentropy([1.0, 701, 801], 'sleep_disorders_pi'))
small_kidney_phe_pi = TensorMap('phecode_586.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_586.11_date'), channel_map={'no_small_kidney':0, 'small_kidney_prevalent':1, 'small_kidney_incident':2}, loss=weighted_crossentropy([1.0, 5882, 2283], 'small_kidney_pi'))
somatoform_disorder_phe_pi = TensorMap('phecode_303.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_303.4_date'), channel_map={'no_somatoform_disorder':0, 'somatoform_disorder_prevalent':1, 'somatoform_disorder_incident':2}, loss=weighted_crossentropy([1.0, 6097, 5208], 'somatoform_disorder_pi'))
specific_nonpsychotic_mental_disorders_due_to_brain_damage_phe_pi = TensorMap('phecode_291.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_291.4_date'), channel_map={'no_specific_nonpsychotic_mental_disorders_due_to_brain_damage':0, 'specific_nonpsychotic_mental_disorders_due_to_brain_damage_prevalent':1, 'specific_nonpsychotic_mental_disorders_due_to_brain_damage_incident':2}, loss=weighted_crossentropy([1.0, 11363, 11904], 'specific_nonpsychotic_mental_disorders_due_to_brain_damage_pi'))
speech_and_language_disorder_phe_pi = TensorMap('phecode_315.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_315.2_date'), channel_map={'no_speech_and_language_disorder':0, 'speech_and_language_disorder_prevalent':1, 'speech_and_language_disorder_incident':2}, loss=weighted_crossentropy([1.0, 14285, 4545], 'speech_and_language_disorder_pi'))
spermatocele_phe_pi = TensorMap('phecode_603.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_603.2_date'), channel_map={'no_spermatocele':0, 'spermatocele_prevalent':1, 'spermatocele_incident':2}, loss=weighted_crossentropy([1.0, 20000, 35714], 'spermatocele_pi'))
spina_bifida_phe_pi = TensorMap('phecode_752.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_752.11_date'), channel_map={'no_spina_bifida':0, 'spina_bifida_prevalent':1, 'spina_bifida_incident':2}, loss=weighted_crossentropy([1.0, 3649, 3378], 'spina_bifida_pi'))
spinal_cord_injury_without_evidence_of_spinal_bone_injury_phe_pi = TensorMap('phecode_952',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_952_date'), channel_map={'no_spinal_cord_injury_without_evidence_of_spinal_bone_injury':0, 'spinal_cord_injury_without_evidence_of_spinal_bone_injury_prevalent':1, 'spinal_cord_injury_without_evidence_of_spinal_bone_injury_incident':2}, loss=weighted_crossentropy([1.0, 12820, 5555], 'spinal_cord_injury_without_evidence_of_spinal_bone_injury_pi'))
spinal_stenosis_phe_pi = TensorMap('phecode_720',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_720_date'), channel_map={'no_spinal_stenosis':0, 'spinal_stenosis_prevalent':1, 'spinal_stenosis_incident':2}, loss=weighted_crossentropy([1.0, 1618, 616], 'spinal_stenosis_pi'))
spinocerebellar_disease_phe_pi = TensorMap('phecode_334.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_334.1_date'), channel_map={'no_spinocerebellar_disease':0, 'spinocerebellar_disease_prevalent':1, 'spinocerebellar_disease_incident':2}, loss=weighted_crossentropy([1.0, 9803, 3289], 'spinocerebellar_disease_pi'))
spirochetal_infection_phe_pi = TensorMap('phecode_130',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_130_date'), channel_map={'no_spirochetal_infection':0, 'spirochetal_infection_prevalent':1, 'spirochetal_infection_incident':2}, loss=weighted_crossentropy([1.0, 9615, 8928], 'spirochetal_infection_pi'))
splenomegaly_phe_pi = TensorMap('phecode_579.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_579.2_date'), channel_map={'no_splenomegaly':0, 'splenomegaly_prevalent':1, 'splenomegaly_incident':2}, loss=weighted_crossentropy([1.0, 3144, 968], 'splenomegaly_pi'))
spondylolisthesis_congenital_phe_pi = TensorMap('phecode_754.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_754.2_date'), channel_map={'no_spondylolisthesis_congenital':0, 'spondylolisthesis_congenital_prevalent':1, 'spondylolisthesis_congenital_incident':2}, loss=weighted_crossentropy([1.0, 166666, 250000], 'spondylolisthesis_congenital_pi'))
spondylosis_and_allied_disorders_phe_pi = TensorMap('phecode_721',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_721_date'), channel_map={'no_spondylosis_and_allied_disorders':0, 'spondylosis_and_allied_disorders_prevalent':1, 'spondylosis_and_allied_disorders_incident':2}, loss=weighted_crossentropy([1.0, 1773, 368], 'spondylosis_and_allied_disorders_pi'))
spondylosis_with_myelopathy_phe_pi = TensorMap('phecode_721.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_721.2_date'), channel_map={'no_spondylosis_with_myelopathy':0, 'spondylosis_with_myelopathy_prevalent':1, 'spondylosis_with_myelopathy_incident':2}, loss=weighted_crossentropy([1.0, 17857, 21739], 'spondylosis_with_myelopathy_pi'))
spondylosis_without_myelopathy_phe_pi = TensorMap('phecode_721.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_721.1_date'), channel_map={'no_spondylosis_without_myelopathy':0, 'spondylosis_without_myelopathy_prevalent':1, 'spondylosis_without_myelopathy_incident':2}, loss=weighted_crossentropy([1.0, 744, 645], 'spondylosis_without_myelopathy_pi'))
spontaneous_ecchymoses_phe_pi = TensorMap('phecode_287.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287.1_date'), channel_map={'no_spontaneous_ecchymoses':0, 'spontaneous_ecchymoses_prevalent':1, 'spontaneous_ecchymoses_incident':2}, loss=weighted_crossentropy([1.0, 8064, 4098], 'spontaneous_ecchymoses_pi'))
sprains_and_strains_phe_pi = TensorMap('phecode_840',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_840_date'), channel_map={'no_sprains_and_strains':0, 'sprains_and_strains_prevalent':1, 'sprains_and_strains_incident':2}, loss=weighted_crossentropy([1.0, 1845, 2487], 'sprains_and_strains_pi'))
staphylococcus_infections_phe_pi = TensorMap('phecode_41.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_41.1_date'), channel_map={'no_staphylococcus_infections':0, 'staphylococcus_infections_prevalent':1, 'staphylococcus_infections_incident':2}, loss=weighted_crossentropy([1.0, 263, 183], 'staphylococcus_infections_pi'))
stiffness_of_joint_phe_pi = TensorMap('phecode_741.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741.2_date'), channel_map={'no_stiffness_of_joint':0, 'stiffness_of_joint_prevalent':1, 'stiffness_of_joint_incident':2}, loss=weighted_crossentropy([1.0, 4504, 15151], 'stiffness_of_joint_pi'))
stomatitis_and_mucositis_ulcerative_phe_pi = TensorMap('phecode_528.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_528.11_date'), channel_map={'no_stomatitis_and_mucositis_ulcerative':0, 'stomatitis_and_mucositis_ulcerative_prevalent':1, 'stomatitis_and_mucositis_ulcerative_incident':2}, loss=weighted_crossentropy([1.0, 3048, 1131], 'stomatitis_and_mucositis_ulcerative_pi'))
strabismus_and_other_disorders_of_binocular_eye_movements_phe_pi = TensorMap('phecode_378',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_378_date'), channel_map={'no_strabismus_and_other_disorders_of_binocular_eye_movements':0, 'strabismus_and_other_disorders_of_binocular_eye_movements_prevalent':1, 'strabismus_and_other_disorders_of_binocular_eye_movements_incident':2}, loss=weighted_crossentropy([1.0, 11627, 10000], 'strabismus_and_other_disorders_of_binocular_eye_movements_pi'))
strabismus_not_specified_as_paralytic_phe_pi = TensorMap('phecode_378.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_378.1_date'), channel_map={'no_strabismus_not_specified_as_paralytic':0, 'strabismus_not_specified_as_paralytic_prevalent':1, 'strabismus_not_specified_as_paralytic_incident':2}, loss=weighted_crossentropy([1.0, 775, 848], 'strabismus_not_specified_as_paralytic_pi'))
streptococcus_infection_phe_pi = TensorMap('phecode_41.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_41.2_date'), channel_map={'no_streptococcus_infection':0, 'streptococcus_infection_prevalent':1, 'streptococcus_infection_incident':2}, loss=weighted_crossentropy([1.0, 542, 331], 'streptococcus_infection_pi'))
stress_incontinence_female_phe_pi = TensorMap('phecode_624.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_624.9_date'), channel_map={'no_stress_incontinence_female':0, 'stress_incontinence_female_prevalent':1, 'stress_incontinence_female_incident':2}, loss=weighted_crossentropy([1.0, 109, 170], 'stress_incontinence_female_pi'))
stricture_and_stenosis_of_esophagus_phe_pi = TensorMap('phecode_530.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.3_date'), channel_map={'no_stricture_and_stenosis_of_esophagus':0, 'stricture_and_stenosis_of_esophagus_prevalent':1, 'stricture_and_stenosis_of_esophagus_incident':2}, loss=weighted_crossentropy([1.0, 553, 257], 'stricture_and_stenosis_of_esophagus_pi'))
stricture_obstruction_of_ureter_phe_pi = TensorMap('phecode_586.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_586.4_date'), channel_map={'no_stricture_obstruction_of_ureter':0, 'stricture_obstruction_of_ureter_prevalent':1, 'stricture_obstruction_of_ureter_incident':2}, loss=weighted_crossentropy([1.0, 954, 603], 'stricture_obstruction_of_ureter_pi'))
stricture_of_artery_phe_pi = TensorMap('phecode_447.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_447.1_date'), channel_map={'no_stricture_of_artery':0, 'stricture_of_artery_prevalent':1, 'stricture_of_artery_incident':2}, loss=weighted_crossentropy([1.0, 1228, 609], 'stricture_of_artery_pi'))
subarachnoid_hemorrhage_phe_pi = TensorMap('phecode_430.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_430.1_date'), channel_map={'no_subarachnoid_hemorrhage':0, 'subarachnoid_hemorrhage_prevalent':1, 'subarachnoid_hemorrhage_incident':2}, loss=weighted_crossentropy([1.0, 888, 946], 'subarachnoid_hemorrhage_pi'))
subdural_hemorrhage_phe_pi = TensorMap('phecode_430.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_430.3_date'), channel_map={'no_subdural_hemorrhage':0, 'subdural_hemorrhage_prevalent':1, 'subdural_hemorrhage_incident':2}, loss=weighted_crossentropy([1.0, 4950, 1436], 'subdural_hemorrhage_pi'))
subjective_visual_disturbances_phe_pi = TensorMap('phecode_368.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.9_date'), channel_map={'no_subjective_visual_disturbances':0, 'subjective_visual_disturbances_prevalent':1, 'subjective_visual_disturbances_incident':2}, loss=weighted_crossentropy([1.0, 1557, 734], 'subjective_visual_disturbances_pi'))
substance_addiction_and_disorders_phe_pi = TensorMap('phecode_316',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_316_date'), channel_map={'no_substance_addiction_and_disorders':0, 'substance_addiction_and_disorders_prevalent':1, 'substance_addiction_and_disorders_incident':2}, loss=weighted_crossentropy([1.0, 2066, 1392], 'substance_addiction_and_disorders_pi'))
suicidal_ideation_or_attempt_phe_pi = TensorMap('phecode_297',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_297_date'), channel_map={'no_suicidal_ideation_or_attempt':0, 'suicidal_ideation_or_attempt_prevalent':1, 'suicidal_ideation_or_attempt_incident':2}, loss=weighted_crossentropy([1.0, 22727, 41666], 'suicidal_ideation_or_attempt_pi'))
suicide_or_self_inflicted_injury_phe_pi = TensorMap('phecode_297.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_297.2_date'), channel_map={'no_suicide_or_self_inflicted_injury':0, 'suicide_or_self_inflicted_injury_prevalent':1, 'suicide_or_self_inflicted_injury_incident':2}, loss=weighted_crossentropy([1.0, 215, 414], 'suicide_or_self_inflicted_injury_pi'))
superficial_cellulitis_and_abscess_phe_pi = TensorMap('phecode_681',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_681_date'), channel_map={'no_superficial_cellulitis_and_abscess':0, 'superficial_cellulitis_and_abscess_prevalent':1, 'superficial_cellulitis_and_abscess_incident':2}, loss=weighted_crossentropy([1.0, 1404, 1326], 'superficial_cellulitis_and_abscess_pi'))
superficial_injury_without_mention_of_infection_phe_pi = TensorMap('phecode_915',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_915_date'), channel_map={'no_superficial_injury_without_mention_of_infection':0, 'superficial_injury_without_mention_of_infection_prevalent':1, 'superficial_injury_without_mention_of_infection_incident':2}, loss=weighted_crossentropy([1.0, 224, 112], 'superficial_injury_without_mention_of_infection_pi'))
suppurative_and_unspecified_otitis_media_phe_pi = TensorMap('phecode_381.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_381.11_date'), channel_map={'no_suppurative_and_unspecified_otitis_media':0, 'suppurative_and_unspecified_otitis_media_prevalent':1, 'suppurative_and_unspecified_otitis_media_incident':2}, loss=weighted_crossentropy([1.0, 793, 1190], 'suppurative_and_unspecified_otitis_media_pi'))
supraventricular_premature_beats_phe_pi = TensorMap('phecode_427.61',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.61_date'), channel_map={'no_supraventricular_premature_beats':0, 'supraventricular_premature_beats_prevalent':1, 'supraventricular_premature_beats_incident':2}, loss=weighted_crossentropy([1.0, 10638, 3333], 'supraventricular_premature_beats_pi'))
swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos_phe_pi = TensorMap('phecode_293.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_293.1_date'), channel_map={'no_swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos':0, 'swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos_prevalent':1, 'swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos_incident':2}, loss=weighted_crossentropy([1.0, 950, 695], 'swelling_mass_or_lump_in_head_and_neck_space_occupying_lesion_intracranial_nos_pi'))
swelling_of_limb_phe_pi = TensorMap('phecode_771.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_771.1_date'), channel_map={'no_swelling_of_limb':0, 'swelling_of_limb_prevalent':1, 'swelling_of_limb_incident':2}, loss=weighted_crossentropy([1.0, 678, 1149], 'swelling_of_limb_pi'))
symbolic_dysfunction_phe_pi = TensorMap('phecode_292.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_292.12_date'), channel_map={'no_symbolic_dysfunction':0, 'symbolic_dysfunction_prevalent':1, 'symbolic_dysfunction_incident':2}, loss=weighted_crossentropy([1.0, 55555, 19230], 'symbolic_dysfunction_pi'))
symptomatic_menopause_phe_pi = TensorMap('phecode_627.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_627.2_date'), channel_map={'no_symptomatic_menopause':0, 'symptomatic_menopause_prevalent':1, 'symptomatic_menopause_incident':2}, loss=weighted_crossentropy([1.0, 2500, 13888], 'symptomatic_menopause_pi'))
symptoms_affecting_skin_phe_pi = TensorMap('phecode_687',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_687_date'), channel_map={'no_symptoms_affecting_skin':0, 'symptoms_affecting_skin_prevalent':1, 'symptoms_affecting_skin_incident':2}, loss=weighted_crossentropy([1.0, 6849, 1706], 'symptoms_affecting_skin_pi'))
symptoms_and_disorders_of_the_joints_phe_pi = TensorMap('phecode_741',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741_date'), channel_map={'no_symptoms_and_disorders_of_the_joints':0, 'symptoms_and_disorders_of_the_joints_prevalent':1, 'symptoms_and_disorders_of_the_joints_incident':2}, loss=weighted_crossentropy([1.0, 41, 50], 'symptoms_and_disorders_of_the_joints_pi'))
symptoms_concerning_nutrition_metabolism_and_development_phe_pi = TensorMap('phecode_1002',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_1002_date'), channel_map={'no_symptoms_concerning_nutrition_metabolism_and_development':0, 'symptoms_concerning_nutrition_metabolism_and_development_prevalent':1, 'symptoms_concerning_nutrition_metabolism_and_development_incident':2}, loss=weighted_crossentropy([1.0, 207, 76], 'symptoms_concerning_nutrition_metabolism_and_development_pi'))
symptoms_involving_cardiovascular_system_phe_pi = TensorMap('phecode_429.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_429.3_date'), channel_map={'no_symptoms_involving_cardiovascular_system':0, 'symptoms_involving_cardiovascular_system_prevalent':1, 'symptoms_involving_cardiovascular_system_incident':2}, loss=weighted_crossentropy([1.0, 5102, 2450], 'symptoms_involving_cardiovascular_system_pi'))
symptoms_involving_digestive_system_phe_pi = TensorMap('phecode_561',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_561_date'), channel_map={'no_symptoms_involving_digestive_system':0, 'symptoms_involving_digestive_system_prevalent':1, 'symptoms_involving_digestive_system_incident':2}, loss=weighted_crossentropy([1.0, 66, 36], 'symptoms_involving_digestive_system_pi'))
symptoms_involving_female_genital_tract_phe_pi = TensorMap('phecode_624',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_624_date'), channel_map={'no_symptoms_involving_female_genital_tract':0, 'symptoms_involving_female_genital_tract_prevalent':1, 'symptoms_involving_female_genital_tract_incident':2}, loss=weighted_crossentropy([1.0, 33333, 45454], 'symptoms_involving_female_genital_tract_pi'))
symptoms_involving_head_and_neck_phe_pi = TensorMap('phecode_293',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_293_date'), channel_map={'no_symptoms_involving_head_and_neck':0, 'symptoms_involving_head_and_neck_prevalent':1, 'symptoms_involving_head_and_neck_incident':2}, loss=weighted_crossentropy([1.0, 249, 630], 'symptoms_involving_head_and_neck_pi'))
symptoms_involving_nervous_and_musculoskeletal_systems_phe_pi = TensorMap('phecode_781',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_781_date'), channel_map={'no_symptoms_involving_nervous_and_musculoskeletal_systems':0, 'symptoms_involving_nervous_and_musculoskeletal_systems_prevalent':1, 'symptoms_involving_nervous_and_musculoskeletal_systems_incident':2}, loss=weighted_crossentropy([1.0, 65, 39], 'symptoms_involving_nervous_and_musculoskeletal_systems_pi'))
symptoms_involving_respiratory_system_and_other_chest_symptoms_phe_pi = TensorMap('phecode_519.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_519.9_date'), channel_map={'no_symptoms_involving_respiratory_system_and_other_chest_symptoms':0, 'symptoms_involving_respiratory_system_and_other_chest_symptoms_prevalent':1, 'symptoms_involving_respiratory_system_and_other_chest_symptoms_incident':2}, loss=weighted_crossentropy([1.0, 2415, 2272], 'symptoms_involving_respiratory_system_and_other_chest_symptoms_pi'))
symptoms_of_the_muscles_phe_pi = TensorMap('phecode_772',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_772_date'), channel_map={'no_symptoms_of_the_muscles':0, 'symptoms_of_the_muscles_prevalent':1, 'symptoms_of_the_muscles_incident':2}, loss=weighted_crossentropy([1.0, 4000, 2688], 'symptoms_of_the_muscles_pi'))
syncope_and_collapse_phe_pi = TensorMap('phecode_788',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_788_date'), channel_map={'no_syncope_and_collapse':0, 'syncope_and_collapse_prevalent':1, 'syncope_and_collapse_incident':2}, loss=weighted_crossentropy([1.0, 104, 58], 'syncope_and_collapse_pi'))
synoviopathy_phe_pi = TensorMap('phecode_726.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_726.2_date'), channel_map={'no_synoviopathy':0, 'synoviopathy_prevalent':1, 'synoviopathy_incident':2}, loss=weighted_crossentropy([1.0, 3164, 1519], 'synoviopathy_pi'))
synovitis_and_tenosynovitis_phe_pi = TensorMap('phecode_727.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_727.1_date'), channel_map={'no_synovitis_and_tenosynovitis':0, 'synovitis_and_tenosynovitis_prevalent':1, 'synovitis_and_tenosynovitis_incident':2}, loss=weighted_crossentropy([1.0, 377, 243], 'synovitis_and_tenosynovitis_pi'))
systemic_lupus_erythematosus_phe_pi = TensorMap('phecode_695.42',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.42_date'), channel_map={'no_systemic_lupus_erythematosus':0, 'systemic_lupus_erythematosus_prevalent':1, 'systemic_lupus_erythematosus_incident':2}, loss=weighted_crossentropy([1.0, 1677, 1824], 'systemic_lupus_erythematosus_pi'))
systemic_sclerosis_phe_pi = TensorMap('phecode_709.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_709.3_date'), channel_map={'no_systemic_sclerosis':0, 'systemic_sclerosis_prevalent':1, 'systemic_sclerosis_incident':2}, loss=weighted_crossentropy([1.0, 4347, 2958], 'systemic_sclerosis_pi'))
tachycardia_nos_phe_pi = TensorMap('phecode_427.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.7_date'), channel_map={'no_tachycardia_nos':0, 'tachycardia_nos_prevalent':1, 'tachycardia_nos_incident':2}, loss=weighted_crossentropy([1.0, 801, 167], 'tachycardia_nos_pi'))
takayasus_disease_phe_pi = TensorMap('phecode_446.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.7_date'), channel_map={'no_takayasus_disease':0, 'takayasus_disease_prevalent':1, 'takayasus_disease_incident':2}, loss=weighted_crossentropy([1.0, 100000, 62500], 'takayasus_disease_pi'))
temporomandibular_joint_disorder_unspecified_phe_pi = TensorMap('phecode_526.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_526.41_date'), channel_map={'no_temporomandibular_joint_disorder_unspecified':0, 'temporomandibular_joint_disorder_unspecified_prevalent':1, 'temporomandibular_joint_disorder_unspecified_incident':2}, loss=weighted_crossentropy([1.0, 3205, 2873], 'temporomandibular_joint_disorder_unspecified_pi'))
tension_headache_phe_pi = TensorMap('phecode_306.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_306.9_date'), channel_map={'no_tension_headache':0, 'tension_headache_prevalent':1, 'tension_headache_incident':2}, loss=weighted_crossentropy([1.0, 2242, 2083], 'tension_headache_pi'))
testicular_dysfunction_phe_pi = TensorMap('phecode_257',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_257_date'), channel_map={'no_testicular_dysfunction':0, 'testicular_dysfunction_prevalent':1, 'testicular_dysfunction_incident':2}, loss=weighted_crossentropy([1.0, 38461, 50000], 'testicular_dysfunction_pi'))
testicular_hypofunction_phe_pi = TensorMap('phecode_257.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_257.1_date'), channel_map={'no_testicular_hypofunction':0, 'testicular_hypofunction_prevalent':1, 'testicular_hypofunction_incident':2}, loss=weighted_crossentropy([1.0, 5050, 3546], 'testicular_hypofunction_pi'))
throat_pain_phe_pi = TensorMap('phecode_478',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_478_date'), channel_map={'no_throat_pain':0, 'throat_pain_prevalent':1, 'throat_pain_incident':2}, loss=weighted_crossentropy([1.0, 1845, 1992], 'throat_pain_pi'))
thromboangiitis_obliterans_phe_pi = TensorMap('phecode_446.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.1_date'), channel_map={'no_thromboangiitis_obliterans':0, 'thromboangiitis_obliterans_prevalent':1, 'thromboangiitis_obliterans_incident':2}, loss=weighted_crossentropy([1.0, 83333, 38461], 'thromboangiitis_obliterans_pi'))
thrombocytopenia_phe_pi = TensorMap('phecode_287.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_287.3_date'), channel_map={'no_thrombocytopenia':0, 'thrombocytopenia_prevalent':1, 'thrombocytopenia_incident':2}, loss=weighted_crossentropy([1.0, 1282, 309], 'thrombocytopenia_pi'))
thrombotic_microangiopathy_phe_pi = TensorMap('phecode_446.8',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.8_date'), channel_map={'no_thrombotic_microangiopathy':0, 'thrombotic_microangiopathy_prevalent':1, 'thrombotic_microangiopathy_incident':2}, loss=weighted_crossentropy([1.0, 31250, 50000], 'thrombotic_microangiopathy_pi'))
thyroid_cancer_phe_pi = TensorMap('phecode_193',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_193_date'), channel_map={'no_thyroid_cancer':0, 'thyroid_cancer_prevalent':1, 'thyroid_cancer_incident':2}, loss=weighted_crossentropy([1.0, 1742, 1326], 'thyroid_cancer_pi'))
thyroiditis_phe_pi = TensorMap('phecode_245',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_245_date'), channel_map={'no_thyroiditis':0, 'thyroiditis_prevalent':1, 'thyroiditis_incident':2}, loss=weighted_crossentropy([1.0, 13157, 10204], 'thyroiditis_pi'))
thyroiditis_acute_and_subacute_phe_pi = TensorMap('phecode_245.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_245.1_date'), channel_map={'no_thyroiditis_acute_and_subacute':0, 'thyroiditis_acute_and_subacute_prevalent':1, 'thyroiditis_acute_and_subacute_incident':2}, loss=weighted_crossentropy([1.0, 45454, 45454], 'thyroiditis_acute_and_subacute_pi'))
thyrotoxicosis_with_or_without_goiter_phe_pi = TensorMap('phecode_242',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_242_date'), channel_map={'no_thyrotoxicosis_with_or_without_goiter':0, 'thyrotoxicosis_with_or_without_goiter_prevalent':1, 'thyrotoxicosis_with_or_without_goiter_incident':2}, loss=weighted_crossentropy([1.0, 605, 383], 'thyrotoxicosis_with_or_without_goiter_pi'))
tics_and_choreas_phe_pi = TensorMap('phecode_333.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_333.3_date'), channel_map={'no_tics_and_choreas':0, 'tics_and_choreas_prevalent':1, 'tics_and_choreas_incident':2}, loss=weighted_crossentropy([1.0, 5050, 948], 'tics_and_choreas_pi'))
tics_and_stuttering_phe_pi = TensorMap('phecode_313.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_313.2_date'), channel_map={'no_tics_and_stuttering':0, 'tics_and_stuttering_prevalent':1, 'tics_and_stuttering_incident':2}, loss=weighted_crossentropy([1.0, 62500, 19230], 'tics_and_stuttering_pi'))
tinnitus_phe_pi = TensorMap('phecode_389.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_389.4_date'), channel_map={'no_tinnitus':0, 'tinnitus_prevalent':1, 'tinnitus_incident':2}, loss=weighted_crossentropy([1.0, 2040, 833], 'tinnitus_pi'))
tobacco_use_disorder_phe_pi = TensorMap('phecode_318',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_318_date'), channel_map={'no_tobacco_use_disorder':0, 'tobacco_use_disorder_prevalent':1, 'tobacco_use_disorder_incident':2}, loss=weighted_crossentropy([1.0, 46, 28], 'tobacco_use_disorder_pi'))
tooth_complications_likely_association_with_other_diseases_phe_pi = TensorMap('phecode_521.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_521.4_date'), channel_map={'no_tooth_complications_likely_association_with_other_diseases':0, 'tooth_complications_likely_association_with_other_diseases_prevalent':1, 'tooth_complications_likely_association_with_other_diseases_incident':2}, loss=weighted_crossentropy([1.0, 38461, 25000], 'tooth_complications_likely_association_with_other_diseases_pi'))
torsion_dystonia_phe_pi = TensorMap('phecode_333.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_333.4_date'), channel_map={'no_torsion_dystonia':0, 'torsion_dystonia_prevalent':1, 'torsion_dystonia_incident':2}, loss=weighted_crossentropy([1.0, 3623, 1945], 'torsion_dystonia_pi'))
torticollis_phe_pi = TensorMap('phecode_723.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_723.1_date'), channel_map={'no_torticollis':0, 'torticollis_prevalent':1, 'torticollis_incident':2}, loss=weighted_crossentropy([1.0, 12820, 6944], 'torticollis_pi'))
toxic_effect_of_carbon_monoxide_phe_pi = TensorMap('phecode_986',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_986_date'), channel_map={'no_toxic_effect_of_carbon_monoxide':0, 'toxic_effect_of_carbon_monoxide_prevalent':1, 'toxic_effect_of_carbon_monoxide_incident':2}, loss=weighted_crossentropy([1.0, 11111, 20833], 'toxic_effect_of_carbon_monoxide_pi'))
toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis_phe_pi = TensorMap('phecode_983',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_983_date'), channel_map={'no_toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis':0, 'toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis_prevalent':1, 'toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis_incident':2}, loss=weighted_crossentropy([1.0, 38461, 27777], 'toxic_effect_of_corrosive_aromatics_acids_and_caustic_alkalis_pi'))
toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents_phe_pi = TensorMap('phecode_981',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_981_date'), channel_map={'no_toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents':0, 'toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents_prevalent':1, 'toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents_incident':2}, loss=weighted_crossentropy([1.0, 452, 724], 'toxic_effect_of_non_ethyl_alcohol_and_petroleum_and_other_solvents_pi'))
toxic_effect_of_noxious_substances_eaten_as_food_phe_pi = TensorMap('phecode_988',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_988_date'), channel_map={'no_toxic_effect_of_noxious_substances_eaten_as_food':0, 'toxic_effect_of_noxious_substances_eaten_as_food_prevalent':1, 'toxic_effect_of_noxious_substances_eaten_as_food_incident':2}, loss=weighted_crossentropy([1.0, 22727, 20833], 'toxic_effect_of_noxious_substances_eaten_as_food_pi'))
toxic_effect_of_other_gases_fumes_or_vapors_phe_pi = TensorMap('phecode_987',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_987_date'), channel_map={'no_toxic_effect_of_other_gases_fumes_or_vapors':0, 'toxic_effect_of_other_gases_fumes_or_vapors_prevalent':1, 'toxic_effect_of_other_gases_fumes_or_vapors_incident':2}, loss=weighted_crossentropy([1.0, 5555, 10204], 'toxic_effect_of_other_gases_fumes_or_vapors_pi'))
toxic_effect_of_other_metals_phe_pi = TensorMap('phecode_985',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_985_date'), channel_map={'no_toxic_effect_of_other_metals':0, 'toxic_effect_of_other_metals_prevalent':1, 'toxic_effect_of_other_metals_incident':2}, loss=weighted_crossentropy([1.0, 13513, 25000], 'toxic_effect_of_other_metals_pi'))
toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source_phe_pi = TensorMap('phecode_989',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_989_date'), channel_map={'no_toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source':0, 'toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source_prevalent':1, 'toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source_incident':2}, loss=weighted_crossentropy([1.0, 452, 724], 'toxic_effect_of_other_substances_chiefly_nonmedicinal_as_to_source_pi'))
toxic_effect_of_venom_phe_pi = TensorMap('phecode_913',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_913_date'), channel_map={'no_toxic_effect_of_venom':0, 'toxic_effect_of_venom_prevalent':1, 'toxic_effect_of_venom_incident':2}, loss=weighted_crossentropy([1.0, 5681, 6578], 'toxic_effect_of_venom_pi'))
toxic_erythema_phe_pi = TensorMap('phecode_695.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.1_date'), channel_map={'no_toxic_erythema':0, 'toxic_erythema_prevalent':1, 'toxic_erythema_incident':2}, loss=weighted_crossentropy([1.0, 6756, 6493], 'toxic_erythema_pi'))
toxic_gastroenteritis_and_colitis_phe_pi = TensorMap('phecode_569.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_569.1_date'), channel_map={'no_toxic_gastroenteritis_and_colitis':0, 'toxic_gastroenteritis_and_colitis_prevalent':1, 'toxic_gastroenteritis_and_colitis_incident':2}, loss=weighted_crossentropy([1.0, 20833, 1706], 'toxic_gastroenteritis_and_colitis_pi'))
toxic_multinodular_goiter_phe_pi = TensorMap('phecode_242.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_242.2_date'), channel_map={'no_toxic_multinodular_goiter':0, 'toxic_multinodular_goiter_prevalent':1, 'toxic_multinodular_goiter_incident':2}, loss=weighted_crossentropy([1.0, 9803, 5494], 'toxic_multinodular_goiter_pi'))
transient_cerebral_ischemia_phe_pi = TensorMap('phecode_433.31',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_433.31_date'), channel_map={'no_transient_cerebral_ischemia':0, 'transient_cerebral_ischemia_prevalent':1, 'transient_cerebral_ischemia_incident':2}, loss=weighted_crossentropy([1.0, 434, 264], 'transient_cerebral_ischemia_pi'))
transient_mental_disorders_due_to_conditions_classified_elsewhere_phe_pi = TensorMap('phecode_291.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_291.1_date'), channel_map={'no_transient_mental_disorders_due_to_conditions_classified_elsewhere':0, 'transient_mental_disorders_due_to_conditions_classified_elsewhere_prevalent':1, 'transient_mental_disorders_due_to_conditions_classified_elsewhere_incident':2}, loss=weighted_crossentropy([1.0, 4950, 5208], 'transient_mental_disorders_due_to_conditions_classified_elsewhere_pi'))
transitory_tachypnea_or_apnea_of_newborn_phe_pi = TensorMap('phecode_656.26',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_656.26_date'), channel_map={'no_transitory_tachypnea_or_apnea_of_newborn':0, 'transitory_tachypnea_or_apnea_of_newborn_prevalent':1, 'transitory_tachypnea_or_apnea_of_newborn_incident':2}, loss=weighted_crossentropy([1.0, 250000, 250000], 'transitory_tachypnea_or_apnea_of_newborn_pi'))
traumatic_amputation_phe_pi = TensorMap('phecode_872',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_872_date'), channel_map={'no_traumatic_amputation':0, 'traumatic_amputation_prevalent':1, 'traumatic_amputation_incident':2}, loss=weighted_crossentropy([1.0, 1420, 2040], 'traumatic_amputation_pi'))
traumatic_and_surgical_subcutaneous_emphysema_phe_pi = TensorMap('phecode_958.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_958.2_date'), channel_map={'no_traumatic_and_surgical_subcutaneous_emphysema':0, 'traumatic_and_surgical_subcutaneous_emphysema_prevalent':1, 'traumatic_and_surgical_subcutaneous_emphysema_incident':2}, loss=weighted_crossentropy([1.0, 17857, 5555], 'traumatic_and_surgical_subcutaneous_emphysema_pi'))
traumatic_arthropathy_phe_pi = TensorMap('phecode_836',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_836_date'), channel_map={'no_traumatic_arthropathy':0, 'traumatic_arthropathy_prevalent':1, 'traumatic_arthropathy_incident':2}, loss=weighted_crossentropy([1.0, 3355, 2793], 'traumatic_arthropathy_pi'))
traumatic_cataract_phe_pi = TensorMap('phecode_366.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_366.3_date'), channel_map={'no_traumatic_cataract':0, 'traumatic_cataract_prevalent':1, 'traumatic_cataract_incident':2}, loss=weighted_crossentropy([1.0, 7246, 6756], 'traumatic_cataract_pi'))
trigeminal_nerve_disorders_cn5_phe_pi = TensorMap('phecode_352.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_352.1_date'), channel_map={'no_trigeminal_nerve_disorders_cn5':0, 'trigeminal_nerve_disorders_cn5_prevalent':1, 'trigeminal_nerve_disorders_cn5_incident':2}, loss=weighted_crossentropy([1.0, 42, 37], 'trigeminal_nerve_disorders_cn5_pi'))
tuberculosis_phe_pi = TensorMap('phecode_10',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_10_date'), channel_map={'no_tuberculosis':0, 'tuberculosis_prevalent':1, 'tuberculosis_incident':2}, loss=weighted_crossentropy([1.0, 2958, 2857], 'tuberculosis_pi'))
tympanosclerosis_and_middle_ear_disease_related_to_otitis_media_phe_pi = TensorMap('phecode_385.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_385.5_date'), channel_map={'no_tympanosclerosis_and_middle_ear_disease_related_to_otitis_media':0, 'tympanosclerosis_and_middle_ear_disease_related_to_otitis_media_prevalent':1, 'tympanosclerosis_and_middle_ear_disease_related_to_otitis_media_incident':2}, loss=weighted_crossentropy([1.0, 2923, 4000], 'tympanosclerosis_and_middle_ear_disease_related_to_otitis_media_pi'))
type_1_diabetes_phe_pi = TensorMap('phecode_250.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.1_date'), channel_map={'no_type_1_diabetes':0, 'type_1_diabetes_prevalent':1, 'type_1_diabetes_incident':2}, loss=weighted_crossentropy([1.0, 246, 275], 'type_1_diabetes_pi'))
type_1_diabetes_with_ketoacidosis_phe_pi = TensorMap('phecode_250.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.11_date'), channel_map={'no_type_1_diabetes_with_ketoacidosis':0, 'type_1_diabetes_with_ketoacidosis_prevalent':1, 'type_1_diabetes_with_ketoacidosis_incident':2}, loss=weighted_crossentropy([1.0, 2500, 2824], 'type_1_diabetes_with_ketoacidosis_pi'))
type_1_diabetes_with_neurological_manifestations_phe_pi = TensorMap('phecode_250.14',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.14_date'), channel_map={'no_type_1_diabetes_with_neurological_manifestations':0, 'type_1_diabetes_with_neurological_manifestations_prevalent':1, 'type_1_diabetes_with_neurological_manifestations_incident':2}, loss=weighted_crossentropy([1.0, 5000, 3225], 'type_1_diabetes_with_neurological_manifestations_pi'))
type_1_diabetes_with_ophthalmic_manifestations_phe_pi = TensorMap('phecode_250.13',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.13_date'), channel_map={'no_type_1_diabetes_with_ophthalmic_manifestations':0, 'type_1_diabetes_with_ophthalmic_manifestations_prevalent':1, 'type_1_diabetes_with_ophthalmic_manifestations_incident':2}, loss=weighted_crossentropy([1.0, 1510, 1308], 'type_1_diabetes_with_ophthalmic_manifestations_pi'))
type_1_diabetes_with_renal_manifestations_phe_pi = TensorMap('phecode_250.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.12_date'), channel_map={'no_type_1_diabetes_with_renal_manifestations':0, 'type_1_diabetes_with_renal_manifestations_prevalent':1, 'type_1_diabetes_with_renal_manifestations_incident':2}, loss=weighted_crossentropy([1.0, 8196, 7042], 'type_1_diabetes_with_renal_manifestations_pi'))
type_2_diabetes_phe_pi = TensorMap('phecode_250.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.2_date'), channel_map={'no_type_2_diabetes':0, 'type_2_diabetes_prevalent':1, 'type_2_diabetes_incident':2}, loss=weighted_crossentropy([1.0, 51, 24], 'type_2_diabetes_pi'))
type_2_diabetes_with_ketoacidosis_phe_pi = TensorMap('phecode_250.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.21_date'), channel_map={'no_type_2_diabetes_with_ketoacidosis':0, 'type_2_diabetes_with_ketoacidosis_prevalent':1, 'type_2_diabetes_with_ketoacidosis_incident':2}, loss=weighted_crossentropy([1.0, 9259, 2857], 'type_2_diabetes_with_ketoacidosis_pi'))
type_2_diabetes_with_neurological_manifestations_phe_pi = TensorMap('phecode_250.24',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.24_date'), channel_map={'no_type_2_diabetes_with_neurological_manifestations':0, 'type_2_diabetes_with_neurological_manifestations_prevalent':1, 'type_2_diabetes_with_neurological_manifestations_incident':2}, loss=weighted_crossentropy([1.0, 1766, 700], 'type_2_diabetes_with_neurological_manifestations_pi'))
type_2_diabetes_with_ophthalmic_manifestations_phe_pi = TensorMap('phecode_250.23',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.23_date'), channel_map={'no_type_2_diabetes_with_ophthalmic_manifestations':0, 'type_2_diabetes_with_ophthalmic_manifestations_prevalent':1, 'type_2_diabetes_with_ophthalmic_manifestations_incident':2}, loss=weighted_crossentropy([1.0, 717, 315], 'type_2_diabetes_with_ophthalmic_manifestations_pi'))
type_2_diabetes_with_renal_manifestations_phe_pi = TensorMap('phecode_250.22',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_250.22_date'), channel_map={'no_type_2_diabetes_with_renal_manifestations':0, 'type_2_diabetes_with_renal_manifestations_prevalent':1, 'type_2_diabetes_with_renal_manifestations_incident':2}, loss=weighted_crossentropy([1.0, 5376, 1661], 'type_2_diabetes_with_renal_manifestations_pi'))
ulcer_of_esophagus_phe_pi = TensorMap('phecode_530.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_530.12_date'), channel_map={'no_ulcer_of_esophagus':0, 'ulcer_of_esophagus_prevalent':1, 'ulcer_of_esophagus_incident':2}, loss=weighted_crossentropy([1.0, 140, 154], 'ulcer_of_esophagus_pi'))
ulceration_of_intestine_phe_pi = TensorMap('phecode_556.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_556.1_date'), channel_map={'no_ulceration_of_intestine':0, 'ulceration_of_intestine_prevalent':1, 'ulceration_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 1838, 606], 'ulceration_of_intestine_pi'))
ulceration_of_the_lower_gi_tract_phe_pi = TensorMap('phecode_556',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_556_date'), channel_map={'no_ulceration_of_the_lower_gi_tract':0, 'ulceration_of_the_lower_gi_tract_prevalent':1, 'ulceration_of_the_lower_gi_tract_incident':2}, loss=weighted_crossentropy([1.0, 2155, 1308], 'ulceration_of_the_lower_gi_tract_pi'))
ulcerative_colitis_phe_pi = TensorMap('phecode_555.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_555.2_date'), channel_map={'no_ulcerative_colitis':0, 'ulcerative_colitis_prevalent':1, 'ulcerative_colitis_incident':2}, loss=weighted_crossentropy([1.0, 213, 284], 'ulcerative_colitis_pi'))
ulcerative_colitis_chronic_phe_pi = TensorMap('phecode_555.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_555.21_date'), channel_map={'no_ulcerative_colitis_chronic':0, 'ulcerative_colitis_chronic_prevalent':1, 'ulcerative_colitis_chronic_incident':2}, loss=weighted_crossentropy([1.0, 1766, 961], 'ulcerative_colitis_chronic_pi'))
umbilical_cord_complications_during_labor_and_delivery_phe_pi = TensorMap('phecode_663',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_663_date'), channel_map={'no_umbilical_cord_complications_during_labor_and_delivery':0, 'umbilical_cord_complications_during_labor_and_delivery_prevalent':1, 'umbilical_cord_complications_during_labor_and_delivery_incident':2}, loss=weighted_crossentropy([1.0, 624, 100000], 'umbilical_cord_complications_during_labor_and_delivery_pi'))
umbilical_hernia_phe_pi = TensorMap('phecode_550.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_550.4_date'), channel_map={'no_umbilical_hernia':0, 'umbilical_hernia_prevalent':1, 'umbilical_hernia_incident':2}, loss=weighted_crossentropy([1.0, 203, 170], 'umbilical_hernia_pi'))
unspecified_diffuse_connective_tissue_disease_phe_pi = TensorMap('phecode_709.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_709.7_date'), channel_map={'no_unspecified_diffuse_connective_tissue_disease':0, 'unspecified_diffuse_connective_tissue_disease_prevalent':1, 'unspecified_diffuse_connective_tissue_disease_incident':2}, loss=weighted_crossentropy([1.0, 10, 7], 'unspecified_diffuse_connective_tissue_disease_pi'))
unspecified_disorder_of_lipoid_metabolism_phe_pi = TensorMap('phecode_272.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_272.9_date'), channel_map={'no_unspecified_disorder_of_lipoid_metabolism':0, 'unspecified_disorder_of_lipoid_metabolism_prevalent':1, 'unspecified_disorder_of_lipoid_metabolism_incident':2}, loss=weighted_crossentropy([1.0, 7692, 1470], 'unspecified_disorder_of_lipoid_metabolism_pi'))
unspecified_erythematous_condition_phe_pi = TensorMap('phecode_695.9',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_695.9_date'), channel_map={'no_unspecified_erythematous_condition':0, 'unspecified_erythematous_condition_prevalent':1, 'unspecified_erythematous_condition_incident':2}, loss=weighted_crossentropy([1.0, 2380, 957], 'unspecified_erythematous_condition_pi'))
unspecified_monoarthritis_phe_pi = TensorMap('phecode_716.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_716.2_date'), channel_map={'no_unspecified_monoarthritis':0, 'unspecified_monoarthritis_prevalent':1, 'unspecified_monoarthritis_incident':2}, loss=weighted_crossentropy([1.0, 80, 28], 'unspecified_monoarthritis_pi'))
unspecified_osteomyelitis_phe_pi = TensorMap('phecode_710.19',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_710.19_date'), channel_map={'no_unspecified_osteomyelitis':0, 'unspecified_osteomyelitis_prevalent':1, 'unspecified_osteomyelitis_incident':2}, loss=weighted_crossentropy([1.0, 8474, 6172], 'unspecified_osteomyelitis_pi'))
unspecified_polyarthropathy_or_polyarthritis_phe_pi = TensorMap('phecode_716.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_716.1_date'), channel_map={'no_unspecified_polyarthropathy_or_polyarthritis':0, 'unspecified_polyarthropathy_or_polyarthritis_prevalent':1, 'unspecified_polyarthropathy_or_polyarthritis_incident':2}, loss=weighted_crossentropy([1.0, 536, 101], 'unspecified_polyarthropathy_or_polyarthritis_pi'))
unstable_angina_intermediate_coronary_syndrome_phe_pi = TensorMap('phecode_411.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_411.1_date'), channel_map={'no_unstable_angina_intermediate_coronary_syndrome':0, 'unstable_angina_intermediate_coronary_syndrome_prevalent':1, 'unstable_angina_intermediate_coronary_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 124, 174], 'unstable_angina_intermediate_coronary_syndrome_pi'))
upper_gastrointestinal_congenital_anomalies_phe_pi = TensorMap('phecode_750.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_750.1_date'), channel_map={'no_upper_gastrointestinal_congenital_anomalies':0, 'upper_gastrointestinal_congenital_anomalies_prevalent':1, 'upper_gastrointestinal_congenital_anomalies_incident':2}, loss=weighted_crossentropy([1.0, 35714, 25000], 'upper_gastrointestinal_congenital_anomalies_pi'))
urethral_discharge_phe_pi = TensorMap('phecode_599.7',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.7_date'), channel_map={'no_urethral_discharge':0, 'urethral_discharge_prevalent':1, 'urethral_discharge_incident':2}, loss=weighted_crossentropy([1.0, 20000, 14705], 'urethral_discharge_pi'))
urethral_stricture_not_specified_as_infectious_phe_pi = TensorMap('phecode_597.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_597.1_date'), channel_map={'no_urethral_stricture_not_specified_as_infectious':0, 'urethral_stricture_not_specified_as_infectious_prevalent':1, 'urethral_stricture_not_specified_as_infectious_incident':2}, loss=weighted_crossentropy([1.0, 231, 197], 'urethral_stricture_not_specified_as_infectious_pi'))
urethral_syndrome_phe_pi = TensorMap('phecode_592.21',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_592.21_date'), channel_map={'no_urethral_syndrome':0, 'urethral_syndrome_prevalent':1, 'urethral_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 13888, 41666], 'urethral_syndrome_pi'))
urethritis_and_urethral_syndrome_phe_pi = TensorMap('phecode_592.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_592.2_date'), channel_map={'no_urethritis_and_urethral_syndrome':0, 'urethritis_and_urethral_syndrome_prevalent':1, 'urethritis_and_urethral_syndrome_incident':2}, loss=weighted_crossentropy([1.0, 6493, 6329], 'urethritis_and_urethral_syndrome_pi'))
urinary_calculus_phe_pi = TensorMap('phecode_594',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_594_date'), channel_map={'no_urinary_calculus':0, 'urinary_calculus_prevalent':1, 'urinary_calculus_incident':2}, loss=weighted_crossentropy([1.0, 104, 121], 'urinary_calculus_pi'))
urinary_complications_nec_phe_pi = TensorMap('phecode_597.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_597.2_date'), channel_map={'no_urinary_complications_nec':0, 'urinary_complications_nec_prevalent':1, 'urinary_complications_nec_incident':2}, loss=weighted_crossentropy([1.0, 17857, 5813], 'urinary_complications_nec_pi'))
urinary_incontinence_phe_pi = TensorMap('phecode_599.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.4_date'), channel_map={'no_urinary_incontinence':0, 'urinary_incontinence_prevalent':1, 'urinary_incontinence_incident':2}, loss=weighted_crossentropy([1.0, 79, 88], 'urinary_incontinence_pi'))
urinary_obstruction_phe_pi = TensorMap('phecode_599.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_599.1_date'), channel_map={'no_urinary_obstruction':0, 'urinary_obstruction_prevalent':1, 'urinary_obstruction_incident':2}, loss=weighted_crossentropy([1.0, 5154, 3448], 'urinary_obstruction_pi'))
urinary_tract_infection_phe_pi = TensorMap('phecode_591',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_591_date'), channel_map={'no_urinary_tract_infection':0, 'urinary_tract_infection_prevalent':1, 'urinary_tract_infection_incident':2}, loss=weighted_crossentropy([1.0, 74, 42], 'urinary_tract_infection_pi'))
urticaria_phe_pi = TensorMap('phecode_947',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_947_date'), channel_map={'no_urticaria':0, 'urticaria_prevalent':1, 'urticaria_incident':2}, loss=weighted_crossentropy([1.0, 1612, 934], 'urticaria_pi'))
uterine_leiomyoma_phe_pi = TensorMap('phecode_218.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_218.1_date'), channel_map={'no_uterine_leiomyoma':0, 'uterine_leiomyoma_prevalent':1, 'uterine_leiomyoma_incident':2}, loss=weighted_crossentropy([1.0, 51, 120], 'uterine_leiomyoma_pi'))
uterine_uterovaginal_prolapse_phe_pi = TensorMap('phecode_618.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_618.2_date'), channel_map={'no_uterine_uterovaginal_prolapse':0, 'uterine_uterovaginal_prolapse_prevalent':1, 'uterine_uterovaginal_prolapse_incident':2}, loss=weighted_crossentropy([1.0, 140, 143], 'uterine_uterovaginal_prolapse_pi'))
uveitis_noninfectious_or_nos_phe_pi = TensorMap('phecode_371.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_371.1_date'), channel_map={'no_uveitis_noninfectious_or_nos':0, 'uveitis_noninfectious_or_nos_prevalent':1, 'uveitis_noninfectious_or_nos_incident':2}, loss=weighted_crossentropy([1.0, 1633, 1160], 'uveitis_noninfectious_or_nos_pi'))
vaginal_enterocele_congenital_or_acquired_phe_pi = TensorMap('phecode_618.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_618.6_date'), channel_map={'no_vaginal_enterocele_congenital_or_acquired':0, 'vaginal_enterocele_congenital_or_acquired_prevalent':1, 'vaginal_enterocele_congenital_or_acquired_incident':2}, loss=weighted_crossentropy([1.0, 1298, 1006], 'vaginal_enterocele_congenital_or_acquired_pi'))
vaginitis_and_vulvovaginitis_phe_pi = TensorMap('phecode_614.52',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_614.52_date'), channel_map={'no_vaginitis_and_vulvovaginitis':0, 'vaginitis_and_vulvovaginitis_prevalent':1, 'vaginitis_and_vulvovaginitis_incident':2}, loss=weighted_crossentropy([1.0, 1572, 2475], 'vaginitis_and_vulvovaginitis_pi'))
valvular_heart_disease_heart_chambers_phe_pi = TensorMap('phecode_747.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_747.12_date'), channel_map={'no_valvular_heart_disease_heart_chambers':0, 'valvular_heart_disease_heart_chambers_prevalent':1, 'valvular_heart_disease_heart_chambers_incident':2}, loss=weighted_crossentropy([1.0, 5000, 1824], 'valvular_heart_disease_heart_chambers_pi'))
varicella_infection_phe_pi = TensorMap('phecode_79.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_79.1_date'), channel_map={'no_varicella_infection':0, 'varicella_infection_prevalent':1, 'varicella_infection_incident':2}, loss=weighted_crossentropy([1.0, 10638, 12820], 'varicella_infection_pi'))
varicose_veins_phe_pi = TensorMap('phecode_454',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_454_date'), channel_map={'no_varicose_veins':0, 'varicose_veins_prevalent':1, 'varicose_veins_incident':2}, loss=weighted_crossentropy([1.0, 1445, 1436], 'varicose_veins_pi'))
varicose_veins_of_lower_extremity_phe_pi = TensorMap('phecode_454.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_454.1_date'), channel_map={'no_varicose_veins_of_lower_extremity':0, 'varicose_veins_of_lower_extremity_prevalent':1, 'varicose_veins_of_lower_extremity_incident':2}, loss=weighted_crossentropy([1.0, 52, 130], 'varicose_veins_of_lower_extremity_pi'))
varicose_veins_of_lower_extremity_symptomtic_phe_pi = TensorMap('phecode_454.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_454.11_date'), channel_map={'no_varicose_veins_of_lower_extremity_symptomtic':0, 'varicose_veins_of_lower_extremity_symptomtic_prevalent':1, 'varicose_veins_of_lower_extremity_symptomtic_incident':2}, loss=weighted_crossentropy([1.0, 1655, 660], 'varicose_veins_of_lower_extremity_symptomtic_pi'))
vascular_dementia_phe_pi = TensorMap('phecode_290.16',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_290.16_date'), channel_map={'no_vascular_dementia':0, 'vascular_dementia_prevalent':1, 'vascular_dementia_incident':2}, loss=weighted_crossentropy([1.0, 55555, 1016], 'vascular_dementia_pi'))
vascular_disorders_of_kidney_hypertrophy_phe_pi = TensorMap('phecode_586.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_586.3_date'), channel_map={'no_vascular_disorders_of_kidney_hypertrophy':0, 'vascular_disorders_of_kidney_hypertrophy_prevalent':1, 'vascular_disorders_of_kidney_hypertrophy_incident':2}, loss=weighted_crossentropy([1.0, 18518, 5102], 'vascular_disorders_of_kidney_hypertrophy_pi'))
vascular_disorders_of_skin_phe_pi = TensorMap('phecode_694.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_694.3_date'), channel_map={'no_vascular_disorders_of_skin':0, 'vascular_disorders_of_skin_prevalent':1, 'vascular_disorders_of_skin_incident':2}, loss=weighted_crossentropy([1.0, 6410, 4273], 'vascular_disorders_of_skin_pi'))
vascular_hamartomas_and_non_neoplastic_nevi_phe_pi = TensorMap('phecode_217',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_217_date'), channel_map={'no_vascular_hamartomas_and_non_neoplastic_nevi':0, 'vascular_hamartomas_and_non_neoplastic_nevi_prevalent':1, 'vascular_hamartomas_and_non_neoplastic_nevi_incident':2}, loss=weighted_crossentropy([1.0, 7812, 50000], 'vascular_hamartomas_and_non_neoplastic_nevi_pi'))
vascular_insufficiency_of_intestine_phe_pi = TensorMap('phecode_441',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_441_date'), channel_map={'no_vascular_insufficiency_of_intestine':0, 'vascular_insufficiency_of_intestine_prevalent':1, 'vascular_insufficiency_of_intestine_incident':2}, loss=weighted_crossentropy([1.0, 5434, 1262], 'vascular_insufficiency_of_intestine_pi'))
venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium_phe_pi = TensorMap('phecode_671',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_671_date'), channel_map={'no_venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium':0, 'venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium_prevalent':1, 'venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium_incident':2}, loss=weighted_crossentropy([1.0, 1953, 50000], 'venous_cerebrovascular_complications_embolism_in_pregnancy_and_the_puerperium_pi'))
ventral_hernia_phe_pi = TensorMap('phecode_550.5',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_550.5_date'), channel_map={'no_ventral_hernia':0, 'ventral_hernia_prevalent':1, 'ventral_hernia_incident':2}, loss=weighted_crossentropy([1.0, 241, 175], 'ventral_hernia_pi'))
ventricular_fibrillation_and_flutter_phe_pi = TensorMap('phecode_427.41',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_427.41_date'), channel_map={'no_ventricular_fibrillation_and_flutter':0, 'ventricular_fibrillation_and_flutter_prevalent':1, 'ventricular_fibrillation_and_flutter_incident':2}, loss=weighted_crossentropy([1.0, 3787, 1298], 'ventricular_fibrillation_and_flutter_pi'))
vertiginous_syndromes_and_other_disorders_of_vestibular_system_phe_pi = TensorMap('phecode_386',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_386_date'), channel_map={'no_vertiginous_syndromes_and_other_disorders_of_vestibular_system':0, 'vertiginous_syndromes_and_other_disorders_of_vestibular_system_prevalent':1, 'vertiginous_syndromes_and_other_disorders_of_vestibular_system_incident':2}, loss=weighted_crossentropy([1.0, 426, 271], 'vertiginous_syndromes_and_other_disorders_of_vestibular_system_pi'))
vesicoureteral_reflux_phe_pi = TensorMap('phecode_586.12',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_586.12_date'), channel_map={'no_vesicoureteral_reflux':0, 'vesicoureteral_reflux_prevalent':1, 'vesicoureteral_reflux_incident':2}, loss=weighted_crossentropy([1.0, 22727, 21739], 'vesicoureteral_reflux_pi'))
villonodular_synovitis_phe_pi = TensorMap('phecode_741.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_741.6_date'), channel_map={'no_villonodular_synovitis':0, 'villonodular_synovitis_prevalent':1, 'villonodular_synovitis_incident':2}, loss=weighted_crossentropy([1.0, 41666, 166666], 'villonodular_synovitis_pi'))
viral_enteritis_phe_pi = TensorMap('phecode_8.6',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_8.6_date'), channel_map={'no_viral_enteritis':0, 'viral_enteritis_prevalent':1, 'viral_enteritis_incident':2}, loss=weighted_crossentropy([1.0, 1470, 512], 'viral_enteritis_pi'))
viral_hepatitis_phe_pi = TensorMap('phecode_70',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_70_date'), channel_map={'no_viral_hepatitis':0, 'viral_hepatitis_prevalent':1, 'viral_hepatitis_incident':2}, loss=weighted_crossentropy([1.0, 823, 950], 'viral_hepatitis_pi'))
viral_hepatitis_a_phe_pi = TensorMap('phecode_70.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_70.1_date'), channel_map={'no_viral_hepatitis_a':0, 'viral_hepatitis_a_prevalent':1, 'viral_hepatitis_a_incident':2}, loss=weighted_crossentropy([1.0, 10416, 9803], 'viral_hepatitis_a_pi'))
viral_hepatitis_b_phe_pi = TensorMap('phecode_70.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_70.2_date'), channel_map={'no_viral_hepatitis_b':0, 'viral_hepatitis_b_prevalent':1, 'viral_hepatitis_b_incident':2}, loss=weighted_crossentropy([1.0, 3448, 3048], 'viral_hepatitis_b_pi'))
viral_hepatitis_c_phe_pi = TensorMap('phecode_70.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_70.3_date'), channel_map={'no_viral_hepatitis_c':0, 'viral_hepatitis_c_prevalent':1, 'viral_hepatitis_c_incident':2}, loss=weighted_crossentropy([1.0, 1557, 2732], 'viral_hepatitis_c_pi'))
viral_infection_phe_pi = TensorMap('phecode_79',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_79_date'), channel_map={'no_viral_infection':0, 'viral_infection_prevalent':1, 'viral_infection_incident':2}, loss=weighted_crossentropy([1.0, 186, 170], 'viral_infection_pi'))
viral_pneumonia_phe_pi = TensorMap('phecode_480.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_480.2_date'), channel_map={'no_viral_pneumonia':0, 'viral_pneumonia_prevalent':1, 'viral_pneumonia_incident':2}, loss=weighted_crossentropy([1.0, 14705, 5813], 'viral_pneumonia_pi'))
viral_warts_hpv_phe_pi = TensorMap('phecode_78',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_78_date'), channel_map={'no_viral_warts_hpv':0, 'viral_warts_hpv_prevalent':1, 'viral_warts_hpv_incident':2}, loss=weighted_crossentropy([1.0, 684, 976], 'viral_warts_hpv_pi'))
visual_disturbances_phe_pi = TensorMap('phecode_368',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368_date'), channel_map={'no_visual_disturbances':0, 'visual_disturbances_prevalent':1, 'visual_disturbances_incident':2}, loss=weighted_crossentropy([1.0, 886, 460], 'visual_disturbances_pi'))
visual_field_defects_phe_pi = TensorMap('phecode_368.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_368.4_date'), channel_map={'no_visual_field_defects':0, 'visual_field_defects_prevalent':1, 'visual_field_defects_incident':2}, loss=weighted_crossentropy([1.0, 2617, 1479], 'visual_field_defects_pi'))
vitamin_a_deficiency_phe_pi = TensorMap('phecode_261.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_261.1_date'), channel_map={'no_vitamin_a_deficiency':0, 'vitamin_a_deficiency_prevalent':1, 'vitamin_a_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 250000, 83333], 'vitamin_a_deficiency_pi'))
vitamin_b_complex_deficiencies_phe_pi = TensorMap('phecode_261.2',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_261.2_date'), channel_map={'no_vitamin_b_complex_deficiencies':0, 'vitamin_b_complex_deficiencies_prevalent':1, 'vitamin_b_complex_deficiencies_incident':2}, loss=weighted_crossentropy([1.0, 1845, 453], 'vitamin_b_complex_deficiencies_pi'))
vitamin_c_deficiencies_phe_pi = TensorMap('phecode_261.3',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_261.3_date'), channel_map={'no_vitamin_c_deficiencies':0, 'vitamin_c_deficiencies_prevalent':1, 'vitamin_c_deficiencies_incident':2}, loss=weighted_crossentropy([1.0, 71428, 38461], 'vitamin_c_deficiencies_pi'))
vitamin_d_deficiency_phe_pi = TensorMap('phecode_261.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_261.4_date'), channel_map={'no_vitamin_d_deficiency':0, 'vitamin_d_deficiency_prevalent':1, 'vitamin_d_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 19230, 368], 'vitamin_d_deficiency_pi'))
vitamin_deficiency_phe_pi = TensorMap('phecode_261',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_261_date'), channel_map={'no_vitamin_deficiency':0, 'vitamin_deficiency_prevalent':1, 'vitamin_deficiency_incident':2}, loss=weighted_crossentropy([1.0, 71428, 16129], 'vitamin_deficiency_pi'))
vitiligo_phe_pi = TensorMap('phecode_694.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_694.1_date'), channel_map={'no_vitiligo':0, 'vitiligo_prevalent':1, 'vitiligo_incident':2}, loss=weighted_crossentropy([1.0, 19230, 6493], 'vitiligo_pi'))
voice_disturbance_phe_pi = TensorMap('phecode_473.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_473.4_date'), channel_map={'no_voice_disturbance':0, 'voice_disturbance_prevalent':1, 'voice_disturbance_incident':2}, loss=weighted_crossentropy([1.0, 633, 806], 'voice_disturbance_pi'))
von_willebrands_disease_phe_pi = TensorMap('phecode_286.11',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_286.11_date'), channel_map={'no_von_willebrands_disease':0, 'von_willebrands_disease_prevalent':1, 'von_willebrands_disease_incident':2}, loss=weighted_crossentropy([1.0, 5555, 7812], 'von_willebrands_disease_pi'))
wegeners_granulomatosis_phe_pi = TensorMap('phecode_446.4',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_446.4_date'), channel_map={'no_wegeners_granulomatosis':0, 'wegeners_granulomatosis_prevalent':1, 'wegeners_granulomatosis_incident':2}, loss=weighted_crossentropy([1.0, 4000, 4000], 'wegeners_granulomatosis_pi'))
wheezing_phe_pi = TensorMap('phecode_512.1',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/phecode_512.1_date'), channel_map={'no_wheezing':0, 'wheezing_prevalent':1, 'wheezing_incident':2}, loss=weighted_crossentropy([1.0, 4629, 1661], 'wheezing_pi'))


#  Continuous tensor maps
ukb_34_0 = TensorMap('34_Year-of-birth_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1951.5428168330234, 'std': 8.116157117024178}, annotation_units=1, channel_map={'34_Year-of-birth_0_0': 0, })
ukb_46_1 = TensorMap('46_Hand-grip-strength-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.966782178217812, 'std': 10.631230226721195}, annotation_units=1, channel_map={'46_Hand-grip-strength-left_1_0': 0, })
ukb_46_2 = TensorMap('46_Hand-grip-strength-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.97282938238099, 'std': 10.523027242484078}, annotation_units=1, channel_map={'46_Hand-grip-strength-left_2_0': 0, })
ukb_46_0 = TensorMap('46_Hand-grip-strength-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.546084555803468, 'std': 11.327359082668965}, annotation_units=1, channel_map={'46_Hand-grip-strength-left_0_0': 0, })
ukb_47_1 = TensorMap('47_Hand-grip-strength-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.862725698738565, 'std': 10.591084586229796}, annotation_units=1, channel_map={'47_Hand-grip-strength-right_1_0': 0, })
ukb_47_2 = TensorMap('47_Hand-grip-strength-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.284588289640862, 'std': 10.527523528595319}, annotation_units=1, channel_map={'47_Hand-grip-strength-right_2_0': 0, })
ukb_47_0 = TensorMap('47_Hand-grip-strength-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.67895383148411, 'std': 11.282845062461273}, annotation_units=1, channel_map={'47_Hand-grip-strength-right_0_0': 0, })
ukb_48_1 = TensorMap('48_Waist-circumference_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.56421296068493, 'std': 13.231872790272519}, annotation_units=1, channel_map={'48_Waist-circumference_1_0': 0, })
ukb_48_0 = TensorMap('48_Waist-circumference_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.31081898412393, 'std': 13.485877124184546}, annotation_units=1, channel_map={'48_Waist-circumference_0_0': 0, })
ukb_48_2 = TensorMap('48_Waist-circumference_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 88.62862951158587, 'std': 12.81065223895848}, annotation_units=1, channel_map={'48_Waist-circumference_2_0': 0, })
ukb_49_1 = TensorMap('49_Hip-circumference_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 103.01338714440396, 'std': 9.2303431712558}, annotation_units=1, channel_map={'49_Hip-circumference_1_0': 0, })
ukb_49_0 = TensorMap('49_Hip-circumference_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 103.40158393578464, 'std': 9.249535675810918}, annotation_units=1, channel_map={'49_Hip-circumference_0_0': 0, })
ukb_49_2 = TensorMap('49_Hip-circumference_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 101.34906256280688, 'std': 8.886137120506753}, annotation_units=1, channel_map={'49_Hip-circumference_2_0': 0, })
ukb_50_1 = TensorMap('50_Standing-height_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 168.70145220045288, 'std': 9.140747494840934}, annotation_units=1, channel_map={'50_Standing-height_1_0': 0, })
ukb_50_0 = TensorMap('50_Standing-height_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 168.44354526127148, 'std': 9.279066076396694}, annotation_units=1, channel_map={'50_Standing-height_0_0': 0, })
ukb_50_2 = TensorMap('50_Standing-height_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 169.1791997240428, 'std': 9.262964511395763}, annotation_units=1, channel_map={'50_Standing-height_2_0': 0, })
ukb_68_1 = TensorMap('68_Number-of-blood-samples-taken_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.978659585976297, 'std': 0.38592072791869997}, annotation_units=1, channel_map={'68_Number-of-blood-samples-taken_1_0': 0, })
ukb_68_2 = TensorMap('68_Number-of-blood-samples-taken_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.961902876375259, 'std': 1.4707973350282677}, annotation_units=1, channel_map={'68_Number-of-blood-samples-taken_2_0': 0, })
ukb_68_0 = TensorMap('68_Number-of-blood-samples-taken_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.221057964856844, 'std': 0.5520992326996531}, annotation_units=1, channel_map={'68_Number-of-blood-samples-taken_0_0': 0, })
ukb_74_0 = TensorMap('74_Fasting-time_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.8270666642729054, 'std': 2.4659168405053093}, annotation_units=1, channel_map={'74_Fasting-time_0_0': 0, })
ukb_74_1 = TensorMap('74_Fasting-time_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.009539263411515, 'std': 1.7450073406005804}, annotation_units=1, channel_map={'74_Fasting-time_1_0': 0, })
ukb_74_2 = TensorMap('74_Fasting-time_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.4765636049899578, 'std': 2.3159634991049094}, annotation_units=1, channel_map={'74_Fasting-time_2_0': 0, })
ukb_77_0 = TensorMap('77_Heel-bone-ultrasound-Tscore-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.23921666393691945, 'std': 1.1765476827580195}, annotation_units=1, channel_map={'77_Heel-bone-ultrasound-Tscore-manual-entry_0_0': 0, })
ukb_78_0 = TensorMap('78_Heel-bone-mineral-density-BMD-Tscore-automated_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.33688782331446526, 'std': 1.2452428354839498}, annotation_units=1, channel_map={'78_Heel-bone-mineral-density-BMD-Tscore-automated_0_0': 0, })
ukb_84_1 = TensorMap('84_Cancer-yearage-first-occurred_1_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 1953.9798602548287, 'std': 306.9089125416748}, annotation_units=4, channel_map={'84_Cancer-yearage-first-occurred_1_0': 0, '84_Cancer-yearage-first-occurred_1_1': 1, '84_Cancer-yearage-first-occurred_1_2': 2, '84_Cancer-yearage-first-occurred_1_3': 3, })
ukb_84_2 = TensorMap('84_Cancer-yearage-first-occurred_2_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 1616.5678690080686, 'std': 782.2068017323489}, annotation_units=5, channel_map={'84_Cancer-yearage-first-occurred_2_0': 0, '84_Cancer-yearage-first-occurred_2_1': 1, '84_Cancer-yearage-first-occurred_2_2': 2, '84_Cancer-yearage-first-occurred_2_3': 3, '84_Cancer-yearage-first-occurred_2_4': 4, })
ukb_84_0 = TensorMap('84_Cancer-yearage-first-occurred_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 1754.5987946428575, 'std': 648.139663695507}, annotation_units=6, channel_map={'84_Cancer-yearage-first-occurred_0_0': 0, '84_Cancer-yearage-first-occurred_0_1': 1, '84_Cancer-yearage-first-occurred_0_2': 2, '84_Cancer-yearage-first-occurred_0_3': 3, '84_Cancer-yearage-first-occurred_0_4': 4, '84_Cancer-yearage-first-occurred_0_5': 5, })
ukb_87_2 = TensorMap('87_Noncancer-illness-yearage-first-occurred_2_33', loss='logcosh', path_prefix='continuous', normalization={'mean': 1047.2531665156168, 'std': 983.4917786440917}, annotation_units=34, channel_map={'87_Noncancer-illness-yearage-first-occurred_2_0': 0, '87_Noncancer-illness-yearage-first-occurred_2_1': 1, '87_Noncancer-illness-yearage-first-occurred_2_2': 2, '87_Noncancer-illness-yearage-first-occurred_2_3': 3, '87_Noncancer-illness-yearage-first-occurred_2_4': 4, '87_Noncancer-illness-yearage-first-occurred_2_5': 5, '87_Noncancer-illness-yearage-first-occurred_2_6': 6, '87_Noncancer-illness-yearage-first-occurred_2_7': 7, '87_Noncancer-illness-yearage-first-occurred_2_8': 8, '87_Noncancer-illness-yearage-first-occurred_2_9': 9, '87_Noncancer-illness-yearage-first-occurred_2_10': 10, '87_Noncancer-illness-yearage-first-occurred_2_11': 11, '87_Noncancer-illness-yearage-first-occurred_2_12': 12, '87_Noncancer-illness-yearage-first-occurred_2_13': 13, '87_Noncancer-illness-yearage-first-occurred_2_14': 14, '87_Noncancer-illness-yearage-first-occurred_2_15': 15, '87_Noncancer-illness-yearage-first-occurred_2_16': 16, '87_Noncancer-illness-yearage-first-occurred_2_17': 17, '87_Noncancer-illness-yearage-first-occurred_2_18': 18, '87_Noncancer-illness-yearage-first-occurred_2_19': 19, '87_Noncancer-illness-yearage-first-occurred_2_20': 20, '87_Noncancer-illness-yearage-first-occurred_2_21': 21, '87_Noncancer-illness-yearage-first-occurred_2_22': 22, '87_Noncancer-illness-yearage-first-occurred_2_23': 23, '87_Noncancer-illness-yearage-first-occurred_2_24': 24, '87_Noncancer-illness-yearage-first-occurred_2_25': 25, '87_Noncancer-illness-yearage-first-occurred_2_26': 26, '87_Noncancer-illness-yearage-first-occurred_2_27': 27, '87_Noncancer-illness-yearage-first-occurred_2_28': 28, '87_Noncancer-illness-yearage-first-occurred_2_29': 29, '87_Noncancer-illness-yearage-first-occurred_2_30': 30, '87_Noncancer-illness-yearage-first-occurred_2_31': 31, '87_Noncancer-illness-yearage-first-occurred_2_32': 32, '87_Noncancer-illness-yearage-first-occurred_2_33': 33, })
ukb_87_1 = TensorMap('87_Noncancer-illness-yearage-first-occurred_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 1402.0798865431568, 'std': 897.2801083950734}, annotation_units=16, channel_map={'87_Noncancer-illness-yearage-first-occurred_1_0': 0, '87_Noncancer-illness-yearage-first-occurred_1_1': 1, '87_Noncancer-illness-yearage-first-occurred_1_2': 2, '87_Noncancer-illness-yearage-first-occurred_1_3': 3, '87_Noncancer-illness-yearage-first-occurred_1_4': 4, '87_Noncancer-illness-yearage-first-occurred_1_5': 5, '87_Noncancer-illness-yearage-first-occurred_1_6': 6, '87_Noncancer-illness-yearage-first-occurred_1_7': 7, '87_Noncancer-illness-yearage-first-occurred_1_8': 8, '87_Noncancer-illness-yearage-first-occurred_1_9': 9, '87_Noncancer-illness-yearage-first-occurred_1_10': 10, '87_Noncancer-illness-yearage-first-occurred_1_11': 11, '87_Noncancer-illness-yearage-first-occurred_1_12': 12, '87_Noncancer-illness-yearage-first-occurred_1_13': 13, '87_Noncancer-illness-yearage-first-occurred_1_14': 14, '87_Noncancer-illness-yearage-first-occurred_1_15': 15, })
ukb_87_0 = TensorMap('87_Noncancer-illness-yearage-first-occurred_0_28', loss='logcosh', path_prefix='continuous', normalization={'mean': 1008.9611756727271, 'std': 980.2192163749811}, annotation_units=29, channel_map={'87_Noncancer-illness-yearage-first-occurred_0_0': 0, '87_Noncancer-illness-yearage-first-occurred_0_1': 1, '87_Noncancer-illness-yearage-first-occurred_0_2': 2, '87_Noncancer-illness-yearage-first-occurred_0_3': 3, '87_Noncancer-illness-yearage-first-occurred_0_4': 4, '87_Noncancer-illness-yearage-first-occurred_0_5': 5, '87_Noncancer-illness-yearage-first-occurred_0_6': 6, '87_Noncancer-illness-yearage-first-occurred_0_7': 7, '87_Noncancer-illness-yearage-first-occurred_0_8': 8, '87_Noncancer-illness-yearage-first-occurred_0_9': 9, '87_Noncancer-illness-yearage-first-occurred_0_10': 10, '87_Noncancer-illness-yearage-first-occurred_0_11': 11, '87_Noncancer-illness-yearage-first-occurred_0_12': 12, '87_Noncancer-illness-yearage-first-occurred_0_13': 13, '87_Noncancer-illness-yearage-first-occurred_0_14': 14, '87_Noncancer-illness-yearage-first-occurred_0_15': 15, '87_Noncancer-illness-yearage-first-occurred_0_16': 16, '87_Noncancer-illness-yearage-first-occurred_0_17': 17, '87_Noncancer-illness-yearage-first-occurred_0_18': 18, '87_Noncancer-illness-yearage-first-occurred_0_19': 19, '87_Noncancer-illness-yearage-first-occurred_0_20': 20, '87_Noncancer-illness-yearage-first-occurred_0_21': 21, '87_Noncancer-illness-yearage-first-occurred_0_22': 22, '87_Noncancer-illness-yearage-first-occurred_0_23': 23, '87_Noncancer-illness-yearage-first-occurred_0_24': 24, '87_Noncancer-illness-yearage-first-occurred_0_25': 25, '87_Noncancer-illness-yearage-first-occurred_0_26': 26, '87_Noncancer-illness-yearage-first-occurred_0_27': 27, '87_Noncancer-illness-yearage-first-occurred_0_28': 28, })
ukb_92_0 = TensorMap('92_Operation-yearage-first-occurred_0_31', loss='logcosh', path_prefix='continuous', normalization={'mean': 1261.0216460437098, 'std': 948.5629772860955}, annotation_units=32, channel_map={'92_Operation-yearage-first-occurred_0_0': 0, '92_Operation-yearage-first-occurred_0_1': 1, '92_Operation-yearage-first-occurred_0_2': 2, '92_Operation-yearage-first-occurred_0_3': 3, '92_Operation-yearage-first-occurred_0_4': 4, '92_Operation-yearage-first-occurred_0_5': 5, '92_Operation-yearage-first-occurred_0_6': 6, '92_Operation-yearage-first-occurred_0_7': 7, '92_Operation-yearage-first-occurred_0_8': 8, '92_Operation-yearage-first-occurred_0_9': 9, '92_Operation-yearage-first-occurred_0_10': 10, '92_Operation-yearage-first-occurred_0_11': 11, '92_Operation-yearage-first-occurred_0_12': 12, '92_Operation-yearage-first-occurred_0_13': 13, '92_Operation-yearage-first-occurred_0_14': 14, '92_Operation-yearage-first-occurred_0_15': 15, '92_Operation-yearage-first-occurred_0_16': 16, '92_Operation-yearage-first-occurred_0_17': 17, '92_Operation-yearage-first-occurred_0_18': 18, '92_Operation-yearage-first-occurred_0_19': 19, '92_Operation-yearage-first-occurred_0_20': 20, '92_Operation-yearage-first-occurred_0_21': 21, '92_Operation-yearage-first-occurred_0_22': 22, '92_Operation-yearage-first-occurred_0_23': 23, '92_Operation-yearage-first-occurred_0_24': 24, '92_Operation-yearage-first-occurred_0_25': 25, '92_Operation-yearage-first-occurred_0_26': 26, '92_Operation-yearage-first-occurred_0_27': 27, '92_Operation-yearage-first-occurred_0_28': 28, '92_Operation-yearage-first-occurred_0_29': 29, '92_Operation-yearage-first-occurred_0_30': 30, '92_Operation-yearage-first-occurred_0_31': 31, })
ukb_92_2 = TensorMap('92_Operation-yearage-first-occurred_2_17', loss='logcosh', path_prefix='continuous', normalization={'mean': 1181.1182519690067, 'std': 970.0452145354333}, annotation_units=18, channel_map={'92_Operation-yearage-first-occurred_2_0': 0, '92_Operation-yearage-first-occurred_2_1': 1, '92_Operation-yearage-first-occurred_2_2': 2, '92_Operation-yearage-first-occurred_2_3': 3, '92_Operation-yearage-first-occurred_2_4': 4, '92_Operation-yearage-first-occurred_2_5': 5, '92_Operation-yearage-first-occurred_2_6': 6, '92_Operation-yearage-first-occurred_2_7': 7, '92_Operation-yearage-first-occurred_2_8': 8, '92_Operation-yearage-first-occurred_2_9': 9, '92_Operation-yearage-first-occurred_2_10': 10, '92_Operation-yearage-first-occurred_2_11': 11, '92_Operation-yearage-first-occurred_2_12': 12, '92_Operation-yearage-first-occurred_2_13': 13, '92_Operation-yearage-first-occurred_2_14': 14, '92_Operation-yearage-first-occurred_2_15': 15, '92_Operation-yearage-first-occurred_2_16': 16, '92_Operation-yearage-first-occurred_2_17': 17, })
ukb_92_1 = TensorMap('92_Operation-yearage-first-occurred_1_14', loss='logcosh', path_prefix='continuous', normalization={'mean': 1654.0880419382263, 'std': 737.0434086543747}, annotation_units=15, channel_map={'92_Operation-yearage-first-occurred_1_0': 0, '92_Operation-yearage-first-occurred_1_1': 1, '92_Operation-yearage-first-occurred_1_2': 2, '92_Operation-yearage-first-occurred_1_3': 3, '92_Operation-yearage-first-occurred_1_4': 4, '92_Operation-yearage-first-occurred_1_5': 5, '92_Operation-yearage-first-occurred_1_6': 6, '92_Operation-yearage-first-occurred_1_7': 7, '92_Operation-yearage-first-occurred_1_8': 8, '92_Operation-yearage-first-occurred_1_9': 9, '92_Operation-yearage-first-occurred_1_10': 10, '92_Operation-yearage-first-occurred_1_11': 11, '92_Operation-yearage-first-occurred_1_12': 12, '92_Operation-yearage-first-occurred_1_13': 13, '92_Operation-yearage-first-occurred_1_14': 14, })
ukb_93_1 = TensorMap('93_Systolic-blood-pressure-manual-reading_1_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 141.57414448669203, 'std': 21.85765051858924}, annotation_units=2, channel_map={'93_Systolic-blood-pressure-manual-reading_1_0': 0, '93_Systolic-blood-pressure-manual-reading_1_1': 1, })
ukb_93_2 = TensorMap('93_Systolic-blood-pressure-manual-reading_2_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 138.73156379116065, 'std': 18.769958816296604}, annotation_units=2, channel_map={'93_Systolic-blood-pressure-manual-reading_2_0': 0, '93_Systolic-blood-pressure-manual-reading_2_1': 1, })
ukb_93_0 = TensorMap('93_Systolic-blood-pressure-manual-reading_0_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 138.69562699102232, 'std': 19.381706048164933}, annotation_units=2, channel_map={'93_Systolic-blood-pressure-manual-reading_0_0': 0, '93_Systolic-blood-pressure-manual-reading_0_1': 1, })
ukb_94_2 = TensorMap('94_Diastolic-blood-pressure-manual-reading_2_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.36997621134343, 'std': 10.621683025771384}, annotation_units=2, channel_map={'94_Diastolic-blood-pressure-manual-reading_2_0': 0, '94_Diastolic-blood-pressure-manual-reading_2_1': 1, })
ukb_94_1 = TensorMap('94_Diastolic-blood-pressure-manual-reading_1_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 80.99999999999999, 'std': 11.96878892774647}, annotation_units=2, channel_map={'94_Diastolic-blood-pressure-manual-reading_1_0': 0, '94_Diastolic-blood-pressure-manual-reading_1_1': 1, })
ukb_94_0 = TensorMap('94_Diastolic-blood-pressure-manual-reading_0_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.8377394398246, 'std': 10.709969832085603}, annotation_units=2, channel_map={'94_Diastolic-blood-pressure-manual-reading_0_0': 0, '94_Diastolic-blood-pressure-manual-reading_0_1': 1, })
ukb_95_0 = TensorMap('95_Pulse-rate-during-bloodpressure-measurement_0_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 70.49032587260217, 'std': 11.664236416568059}, annotation_units=2, channel_map={'95_Pulse-rate-during-bloodpressure-measurement_0_0': 0, '95_Pulse-rate-during-bloodpressure-measurement_0_1': 1, })
ukb_95_1 = TensorMap('95_Pulse-rate-during-bloodpressure-measurement_1_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 67.02661596958173, 'std': 11.226470506412381}, annotation_units=2, channel_map={'95_Pulse-rate-during-bloodpressure-measurement_1_0': 0, '95_Pulse-rate-during-bloodpressure-measurement_1_1': 1, })
ukb_95_2 = TensorMap('95_Pulse-rate-during-bloodpressure-measurement_2_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 70.16326530612243, 'std': 11.801997299209166}, annotation_units=2, channel_map={'95_Pulse-rate-during-bloodpressure-measurement_2_0': 0, '95_Pulse-rate-during-bloodpressure-measurement_2_1': 1, })
ukb_102_1 = TensorMap('102_Pulse-rate-automated-reading_1_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 66.47514728451905, 'std': 11.017439707935342}, annotation_units=2, channel_map={'102_Pulse-rate-automated-reading_1_0': 0, '102_Pulse-rate-automated-reading_1_1': 1, })
ukb_102_2 = TensorMap('102_Pulse-rate-automated-reading_2_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 69.31812811877346, 'std': 11.89438577532399}, annotation_units=2, channel_map={'102_Pulse-rate-automated-reading_2_0': 0, '102_Pulse-rate-automated-reading_2_1': 1, })
ukb_102_0 = TensorMap('102_Pulse-rate-automated-reading_0_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 69.33447813577959, 'std': 11.453019128287748}, annotation_units=2, channel_map={'102_Pulse-rate-automated-reading_0_0': 0, '102_Pulse-rate-automated-reading_0_1': 1, })
ukb_129_1 = TensorMap('129_Place-of-birth-in-UK-north-coordinate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 367860.24710284255, 'std': 112547.05092100817}, annotation_units=1, channel_map={'129_Place-of-birth-in-UK-north-coordinate_1_0': 0, })
ukb_129_0 = TensorMap('129_Place-of-birth-in-UK-north-coordinate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 354194.86608619534, 'std': 167068.55848801878}, annotation_units=1, channel_map={'129_Place-of-birth-in-UK-north-coordinate_0_0': 0, })
ukb_129_2 = TensorMap('129_Place-of-birth-in-UK-north-coordinate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 353256.53288925893, 'std': 119505.04056451505}, annotation_units=1, channel_map={'129_Place-of-birth-in-UK-north-coordinate_2_0': 0, })
ukb_130_2 = TensorMap('130_Place-of-birth-in-UK-east-coordinate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 400938.46461282263, 'std': 90302.33714323193}, annotation_units=1, channel_map={'130_Place-of-birth-in-UK-east-coordinate_2_0': 0, })
ukb_130_0 = TensorMap('130_Place-of-birth-in-UK-east-coordinate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 405710.36677603924, 'std': 94504.3898173737}, annotation_units=1, channel_map={'130_Place-of-birth-in-UK-east-coordinate_0_0': 0, })
ukb_130_1 = TensorMap('130_Place-of-birth-in-UK-east-coordinate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 398045.8210777946, 'std': 78599.13473303957}, annotation_units=1, channel_map={'130_Place-of-birth-in-UK-east-coordinate_1_0': 0, })
ukb_134_0 = TensorMap('134_Number-of-selfreported-cancers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08930102018442253, 'std': 0.3085774171849537}, annotation_units=1, channel_map={'134_Number-of-selfreported-cancers_0_0': 0, })
ukb_134_1 = TensorMap('134_Number-of-selfreported-cancers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.11964004720692367, 'std': 0.3490769566575782}, annotation_units=1, channel_map={'134_Number-of-selfreported-cancers_1_0': 0, })
ukb_134_2 = TensorMap('134_Number-of-selfreported-cancers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1293272771912595, 'std': 0.36466959499059515}, annotation_units=1, channel_map={'134_Number-of-selfreported-cancers_2_0': 0, })
ukb_135_1 = TensorMap('135_Number-of-selfreported-noncancer-illnesses_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7391817466561763, 'std': 2.3460687024760927}, annotation_units=1, channel_map={'135_Number-of-selfreported-noncancer-illnesses_1_0': 0, })
ukb_135_0 = TensorMap('135_Number-of-selfreported-noncancer-illnesses_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8634651187823157, 'std': 1.868687135297511}, annotation_units=1, channel_map={'135_Number-of-selfreported-noncancer-illnesses_0_0': 0, })
ukb_135_2 = TensorMap('135_Number-of-selfreported-noncancer-illnesses_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7767309108765033, 'std': 3.1514119492594235}, annotation_units=1, channel_map={'135_Number-of-selfreported-noncancer-illnesses_2_0': 0, })
ukb_136_2 = TensorMap('136_Number-of-operations-selfreported_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.924932482199853, 'std': 1.6436185394380622}, annotation_units=1, channel_map={'136_Number-of-operations-selfreported_2_0': 0, })
ukb_136_0 = TensorMap('136_Number-of-operations-selfreported_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7307813440600868, 'std': 1.551633635135569}, annotation_units=1, channel_map={'136_Number-of-operations-selfreported_0_0': 0, })
ukb_136_1 = TensorMap('136_Number-of-operations-selfreported_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7353461841070024, 'std': 1.4625836321005081}, annotation_units=1, channel_map={'136_Number-of-operations-selfreported_1_0': 0, })
ukb_137_2 = TensorMap('137_Number-of-treatmentsmedications-taken_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.997606187085686, 'std': 2.250111683162527}, annotation_units=1, channel_map={'137_Number-of-treatmentsmedications-taken_2_0': 0, })
ukb_137_1 = TensorMap('137_Number-of-treatmentsmedications-taken_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7368214004720692, 'std': 2.7159924712300003}, annotation_units=1, channel_map={'137_Number-of-treatmentsmedications-taken_1_0': 0, })
ukb_137_0 = TensorMap('137_Number-of-treatmentsmedications-taken_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.459684974704688, 'std': 2.6811321559778927}, annotation_units=1, channel_map={'137_Number-of-treatmentsmedications-taken_0_0': 0, })
ukb_189_0 = TensorMap('189_Townsend-deprivation-index-at-recruitment_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -1.2937350780782744, 'std': 3.0950584973258257}, annotation_units=1, channel_map={'189_Townsend-deprivation-index-at-recruitment_0_0': 0, })
ukb_864_1 = TensorMap('864_Number-of-daysweek-walked-10-minutes_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.208242520309093, 'std': 2.0061152415903893}, annotation_units=1, channel_map={'864_Number-of-daysweek-walked-10-minutes_1_0': 0, })
ukb_864_0 = TensorMap('864_Number-of-daysweek-walked-10-minutes_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.361930056113816, 'std': 1.9817371830771895}, annotation_units=1, channel_map={'864_Number-of-daysweek-walked-10-minutes_0_0': 0, })
ukb_864_2 = TensorMap('864_Number-of-daysweek-walked-10-minutes_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.645109020098359, 'std': 1.7748651078267916}, annotation_units=1, channel_map={'864_Number-of-daysweek-walked-10-minutes_2_0': 0, })
ukb_874_0 = TensorMap('874_Duration-of-walks_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 61.640558181937784, 'std': 78.79452297243806}, annotation_units=1, channel_map={'874_Duration-of-walks_0_0': 0, })
ukb_874_1 = TensorMap('874_Duration-of-walks_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.25258468983722, 'std': 56.45237206485051}, annotation_units=1, channel_map={'874_Duration-of-walks_1_0': 0, })
ukb_874_2 = TensorMap('874_Duration-of-walks_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.86918180406768, 'std': 64.64595251034733}, annotation_units=1, channel_map={'874_Duration-of-walks_2_0': 0, })
ukb_884_1 = TensorMap('884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.6479429178685283, 'std': 2.2591440247750225}, annotation_units=1, channel_map={'884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_1_0': 0, })
ukb_884_0 = TensorMap('884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.6259258012388917, 'std': 2.334379403427264}, annotation_units=1, channel_map={'884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_0_0': 0, })
ukb_884_2 = TensorMap('884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.9670806913923387, 'std': 2.2126090342350135}, annotation_units=1, channel_map={'884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_2_0': 0, })
ukb_894_1 = TensorMap('894_Duration-of-moderate-activity_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.42454143713299, 'std': 62.535299967562466}, annotation_units=1, channel_map={'894_Duration-of-moderate-activity_1_0': 0, })
ukb_894_0 = TensorMap('894_Duration-of-moderate-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 66.2859868738906, 'std': 77.28561471201621}, annotation_units=1, channel_map={'894_Duration-of-moderate-activity_0_0': 0, })
ukb_894_2 = TensorMap('894_Duration-of-moderate-activity_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 63.1332708904797, 'std': 64.77702829243636}, annotation_units=1, channel_map={'894_Duration-of-moderate-activity_2_0': 0, })
ukb_904_2 = TensorMap('904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.981525316090295, 'std': 1.8852519727580237}, annotation_units=1, channel_map={'904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_2_0': 0, })
ukb_904_1 = TensorMap('904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8243814121907063, 'std': 1.862695454321091}, annotation_units=1, channel_map={'904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_1_0': 0, })
ukb_904_0 = TensorMap('904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8387379989893886, 'std': 1.9593440684135033}, annotation_units=1, channel_map={'904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_0_0': 0, })
ukb_914_1 = TensorMap('914_Duration-of-vigorous-activity_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.88408516192186, 'std': 41.19315504545286}, annotation_units=1, channel_map={'914_Duration-of-vigorous-activity_1_0': 0, })
ukb_914_0 = TensorMap('914_Duration-of-vigorous-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.853302756719245, 'std': 48.15790408772608}, annotation_units=1, channel_map={'914_Duration-of-vigorous-activity_0_0': 0, })
ukb_914_2 = TensorMap('914_Duration-of-vigorous-activity_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.02596311203502, 'std': 40.60281990718074}, annotation_units=1, channel_map={'914_Duration-of-vigorous-activity_2_0': 0, })
ukb_1050_2 = TensorMap('1050_Time-spend-outdoors-in-summer_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.796896294016794, 'std': 2.199860319442461}, annotation_units=1, channel_map={'1050_Time-spend-outdoors-in-summer_2_0': 0, })
ukb_1050_0 = TensorMap('1050_Time-spend-outdoors-in-summer_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.795678700368706, 'std': 2.3994732770508778}, annotation_units=1, channel_map={'1050_Time-spend-outdoors-in-summer_0_0': 0, })
ukb_1050_1 = TensorMap('1050_Time-spend-outdoors-in-summer_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.561582297977205, 'std': 2.1957098997905944}, annotation_units=1, channel_map={'1050_Time-spend-outdoors-in-summer_1_0': 0, })
ukb_1060_1 = TensorMap('1060_Time-spent-outdoors-in-winter_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.6666155575999178, 'std': 1.505027494049304}, annotation_units=1, channel_map={'1060_Time-spent-outdoors-in-winter_1_0': 0, })
ukb_1060_2 = TensorMap('1060_Time-spent-outdoors-in-winter_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8363336992316137, 'std': 1.4812754174635858}, annotation_units=1, channel_map={'1060_Time-spent-outdoors-in-winter_2_0': 0, })
ukb_1060_0 = TensorMap('1060_Time-spent-outdoors-in-winter_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.946760893914342, 'std': 1.814279817665687}, annotation_units=1, channel_map={'1060_Time-spent-outdoors-in-winter_0_0': 0, })
ukb_1070_2 = TensorMap('1070_Time-spent-watching-television-TV_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.81251240184823, 'std': 1.552981588750078}, annotation_units=1, channel_map={'1070_Time-spent-watching-television-TV_2_0': 0, })
ukb_1070_0 = TensorMap('1070_Time-spent-watching-television-TV_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7988658696776016, 'std': 1.6783477182274054}, annotation_units=1, channel_map={'1070_Time-spent-watching-television-TV_0_0': 0, })
ukb_1070_1 = TensorMap('1070_Time-spent-watching-television-TV_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8348053485962406, 'std': 1.60588679933294}, annotation_units=1, channel_map={'1070_Time-spent-watching-television-TV_1_0': 0, })
ukb_1080_2 = TensorMap('1080_Time-spent-using-computer_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5754069536611646, 'std': 1.3814423616289675}, annotation_units=1, channel_map={'1080_Time-spent-using-computer_2_0': 0, })
ukb_1080_0 = TensorMap('1080_Time-spent-using-computer_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0828958547510272, 'std': 1.3865554107228948}, annotation_units=1, channel_map={'1080_Time-spent-using-computer_0_0': 0, })
ukb_1080_1 = TensorMap('1080_Time-spent-using-computer_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.448178856973645, 'std': 1.383982771397717}, annotation_units=1, channel_map={'1080_Time-spent-using-computer_1_0': 0, })
ukb_1090_0 = TensorMap('1090_Time-spent-driving_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9779771499100696, 'std': 1.2239094735413558}, annotation_units=1, channel_map={'1090_Time-spent-driving_0_0': 0, })
ukb_1090_1 = TensorMap('1090_Time-spent-driving_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8802739998014494, 'std': 0.9199593397384247}, annotation_units=1, channel_map={'1090_Time-spent-driving_1_0': 0, })
ukb_1090_2 = TensorMap('1090_Time-spent-driving_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9020505361624458, 'std': 0.94485221198409}, annotation_units=1, channel_map={'1090_Time-spent-driving_2_0': 0, })
ukb_1160_0 = TensorMap('1160_Sleep-duration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.154202119120256, 'std': 1.1113586587917186}, annotation_units=1, channel_map={'1160_Sleep-duration_0_0': 0, })
ukb_1160_1 = TensorMap('1160_Sleep-duration_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.23463852450933, 'std': 1.058378574473056}, annotation_units=1, channel_map={'1160_Sleep-duration_1_0': 0, })
ukb_1160_2 = TensorMap('1160_Sleep-duration_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.143990612951881, 'std': 1.0467499872189683}, annotation_units=1, channel_map={'1160_Sleep-duration_2_0': 0, })
ukb_1269_0 = TensorMap('1269_Exposure-to-tobacco-smoke-at-home_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.540621752243099, 'std': 4.5161248541653185}, annotation_units=1, channel_map={'1269_Exposure-to-tobacco-smoke-at-home_0_0': 0, })
ukb_1269_2 = TensorMap('1269_Exposure-to-tobacco-smoke-at-home_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15935192780968013, 'std': 2.407226704732319}, annotation_units=1, channel_map={'1269_Exposure-to-tobacco-smoke-at-home_2_0': 0, })
ukb_1269_1 = TensorMap('1269_Exposure-to-tobacco-smoke-at-home_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.24976835169360656, 'std': 3.2487607692446194}, annotation_units=1, channel_map={'1269_Exposure-to-tobacco-smoke-at-home_1_0': 0, })
ukb_1279_2 = TensorMap('1279_Exposure-to-tobacco-smoke-outside-home_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.21676264784197158, 'std': 1.2835272972915215}, annotation_units=1, channel_map={'1279_Exposure-to-tobacco-smoke-outside-home_2_0': 0, })
ukb_1279_1 = TensorMap('1279_Exposure-to-tobacco-smoke-outside-home_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23972936400541278, 'std': 1.3757051327432581}, annotation_units=1, channel_map={'1279_Exposure-to-tobacco-smoke-outside-home_1_0': 0, })
ukb_1279_0 = TensorMap('1279_Exposure-to-tobacco-smoke-outside-home_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5148142702661684, 'std': 2.552245700179311}, annotation_units=1, channel_map={'1279_Exposure-to-tobacco-smoke-outside-home_0_0': 0, })
ukb_1289_0 = TensorMap('1289_Cooked-vegetable-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.733310165635569, 'std': 1.9271647319384553}, annotation_units=1, channel_map={'1289_Cooked-vegetable-intake_0_0': 0, })
ukb_1289_2 = TensorMap('1289_Cooked-vegetable-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8271987711912607, 'std': 1.9227820873256185}, annotation_units=1, channel_map={'1289_Cooked-vegetable-intake_2_0': 0, })
ukb_1289_1 = TensorMap('1289_Cooked-vegetable-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.674958049550883, 'std': 1.8012764082166277}, annotation_units=1, channel_map={'1289_Cooked-vegetable-intake_1_0': 0, })
ukb_1299_2 = TensorMap('1299_Salad-raw-vegetable-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.251437272313297, 'std': 2.187847774174094}, annotation_units=1, channel_map={'1299_Salad-raw-vegetable-intake_2_0': 0, })
ukb_1299_0 = TensorMap('1299_Salad-raw-vegetable-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.189287995665954, 'std': 2.157119547120408}, annotation_units=1, channel_map={'1299_Salad-raw-vegetable-intake_0_0': 0, })
ukb_1299_1 = TensorMap('1299_Salad-raw-vegetable-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9719076116524064, 'std': 1.933399239838042}, annotation_units=1, channel_map={'1299_Salad-raw-vegetable-intake_1_0': 0, })
ukb_1309_0 = TensorMap('1309_Fresh-fruit-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.235155569352787, 'std': 1.613492103587494}, annotation_units=1, channel_map={'1309_Fresh-fruit-intake_0_0': 0, })
ukb_1309_1 = TensorMap('1309_Fresh-fruit-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.187629323086018, 'std': 1.5030665100018141}, annotation_units=1, channel_map={'1309_Fresh-fruit-intake_1_0': 0, })
ukb_1309_2 = TensorMap('1309_Fresh-fruit-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.268061967203828, 'std': 1.6501996463423703}, annotation_units=1, channel_map={'1309_Fresh-fruit-intake_2_0': 0, })
ukb_1319_1 = TensorMap('1319_Dried-fruit-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9151344404903125, 'std': 1.589242757682482}, annotation_units=1, channel_map={'1319_Dried-fruit-intake_1_0': 0, })
ukb_1319_2 = TensorMap('1319_Dried-fruit-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9016451667897936, 'std': 1.702338635747111}, annotation_units=1, channel_map={'1319_Dried-fruit-intake_2_0': 0, })
ukb_1319_0 = TensorMap('1319_Dried-fruit-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8517596660054848, 'std': 1.753491640401324}, annotation_units=1, channel_map={'1319_Dried-fruit-intake_0_0': 0, })
ukb_1438_2 = TensorMap('1438_Bread-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.247571298716059, 'std': 7.358849002858928}, annotation_units=1, channel_map={'1438_Bread-intake_2_0': 0, })
ukb_1438_1 = TensorMap('1438_Bread-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.031135621723214, 'std': 8.337381368665286}, annotation_units=1, channel_map={'1438_Bread-intake_1_0': 0, })
ukb_1438_0 = TensorMap('1438_Bread-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.268483269682061, 'std': 8.616980216186523}, annotation_units=1, channel_map={'1438_Bread-intake_0_0': 0, })
ukb_1458_0 = TensorMap('1458_Cereal-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.496601254582553, 'std': 2.7950692055756186}, annotation_units=1, channel_map={'1458_Cereal-intake_0_0': 0, })
ukb_1458_1 = TensorMap('1458_Cereal-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.8168774917556725, 'std': 2.7789088243747186}, annotation_units=1, channel_map={'1458_Cereal-intake_1_0': 0, })
ukb_1458_2 = TensorMap('1458_Cereal-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.469385155409614, 'std': 2.8375783862735853}, annotation_units=1, channel_map={'1458_Cereal-intake_2_0': 0, })
ukb_1488_2 = TensorMap('1488_Tea-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.334823702529855, 'std': 2.620958765560841}, annotation_units=1, channel_map={'1488_Tea-intake_2_0': 0, })
ukb_1488_0 = TensorMap('1488_Tea-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.4244916477129625, 'std': 2.882557329461547}, annotation_units=1, channel_map={'1488_Tea-intake_0_0': 0, })
ukb_1488_1 = TensorMap('1488_Tea-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5254700265774184, 'std': 2.7660770460707065}, annotation_units=1, channel_map={'1488_Tea-intake_1_0': 0, })
ukb_1498_0 = TensorMap('1498_Coffee-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.01511286927633, 'std': 2.091538483627113}, annotation_units=1, channel_map={'1498_Coffee-intake_0_0': 0, })
ukb_1498_1 = TensorMap('1498_Coffee-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.063071927580438, 'std': 1.9972076668177663}, annotation_units=1, channel_map={'1498_Coffee-intake_1_0': 0, })
ukb_1498_2 = TensorMap('1498_Coffee-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.04724376025808, 'std': 1.860703471238719}, annotation_units=1, channel_map={'1498_Coffee-intake_2_0': 0, })
ukb_1528_0 = TensorMap('1528_Water-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7322974802814244, 'std': 2.2620300811906597}, annotation_units=1, channel_map={'1528_Water-intake_0_0': 0, })
ukb_1528_2 = TensorMap('1528_Water-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5283953772325876, 'std': 2.0315310695537}, annotation_units=1, channel_map={'1528_Water-intake_2_0': 0, })
ukb_1528_1 = TensorMap('1528_Water-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.168763568186304, 'std': 1.886024281764599}, annotation_units=1, channel_map={'1528_Water-intake_1_0': 0, })
ukb_1568_0 = TensorMap('1568_Average-weekly-red-wine-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.9152579289595435, 'std': 5.582280031635246}, annotation_units=1, channel_map={'1568_Average-weekly-red-wine-intake_0_0': 0, })
ukb_1568_2 = TensorMap('1568_Average-weekly-red-wine-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.701974153650995, 'std': 5.225171377492647}, annotation_units=1, channel_map={'1568_Average-weekly-red-wine-intake_2_0': 0, })
ukb_1568_1 = TensorMap('1568_Average-weekly-red-wine-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.900138121546962, 'std': 5.370605251577174}, annotation_units=1, channel_map={'1568_Average-weekly-red-wine-intake_1_0': 0, })
ukb_1578_0 = TensorMap('1578_Average-weekly-champagne-plus-white-wine-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.6539523474601396, 'std': 4.711206710140075}, annotation_units=1, channel_map={'1578_Average-weekly-champagne-plus-white-wine-intake_0_0': 0, })
ukb_1578_1 = TensorMap('1578_Average-weekly-champagne-plus-white-wine-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.356249568161404, 'std': 4.190686794088415}, annotation_units=1, channel_map={'1578_Average-weekly-champagne-plus-white-wine-intake_1_0': 0, })
ukb_1578_2 = TensorMap('1578_Average-weekly-champagne-plus-white-wine-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.486797240504321, 'std': 4.355314029803978}, annotation_units=1, channel_map={'1578_Average-weekly-champagne-plus-white-wine-intake_2_0': 0, })
ukb_1588_1 = TensorMap('1588_Average-weekly-beer-plus-cider-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.4752830709748688, 'std': 4.489503445698575}, annotation_units=1, channel_map={'1588_Average-weekly-beer-plus-cider-intake_1_0': 0, })
ukb_1588_0 = TensorMap('1588_Average-weekly-beer-plus-cider-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.0129877280169217, 'std': 5.605161315256421}, annotation_units=1, channel_map={'1588_Average-weekly-beer-plus-cider-intake_0_0': 0, })
ukb_1588_2 = TensorMap('1588_Average-weekly-beer-plus-cider-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.3756731073804245, 'std': 4.163989516510834}, annotation_units=1, channel_map={'1588_Average-weekly-beer-plus-cider-intake_2_0': 0, })
ukb_1598_1 = TensorMap('1598_Average-weekly-spirits-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.4630793672722253, 'std': 3.9157992934050494}, annotation_units=1, channel_map={'1598_Average-weekly-spirits-intake_1_0': 0, })
ukb_1598_2 = TensorMap('1598_Average-weekly-spirits-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.6438943108783621, 'std': 4.028235965830109}, annotation_units=1, channel_map={'1598_Average-weekly-spirits-intake_2_0': 0, })
ukb_1598_0 = TensorMap('1598_Average-weekly-spirits-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.854245206180769, 'std': 5.341212242611891}, annotation_units=1, channel_map={'1598_Average-weekly-spirits-intake_0_0': 0, })
ukb_1608_0 = TensorMap('1608_Average-weekly-fortified-wine-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23854966812523384, 'std': 1.2726276064345976}, annotation_units=1, channel_map={'1608_Average-weekly-fortified-wine-intake_0_0': 0, })
ukb_1608_1 = TensorMap('1608_Average-weekly-fortified-wine-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23671497584541068, 'std': 1.023907319385953}, annotation_units=1, channel_map={'1608_Average-weekly-fortified-wine-intake_1_0': 0, })
ukb_1608_2 = TensorMap('1608_Average-weekly-fortified-wine-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1775471399144351, 'std': 1.007208376090604}, annotation_units=1, channel_map={'1608_Average-weekly-fortified-wine-intake_2_0': 0, })
ukb_1737_0 = TensorMap('1737_Childhood-sunburn-occasions_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.6145275628266378, 'std': 5.099119672044519}, annotation_units=1, channel_map={'1737_Childhood-sunburn-occasions_0_0': 0, })
ukb_1737_1 = TensorMap('1737_Childhood-sunburn-occasions_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.751155420573089, 'std': 3.6425922584314243}, annotation_units=1, channel_map={'1737_Childhood-sunburn-occasions_1_0': 0, })
ukb_1737_2 = TensorMap('1737_Childhood-sunburn-occasions_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9583380892592173, 'std': 5.046267590850488}, annotation_units=1, channel_map={'1737_Childhood-sunburn-occasions_2_0': 0, })
ukb_1807_0 = TensorMap('1807_Fathers-age-at-death_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 70.55713596848496, 'std': 13.016618246405548}, annotation_units=1, channel_map={'1807_Fathers-age-at-death_0_0': 0, })
ukb_1807_1 = TensorMap('1807_Fathers-age-at-death_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 72.24787969924813, 'std': 13.00164137801769}, annotation_units=1, channel_map={'1807_Fathers-age-at-death_1_0': 0, })
ukb_1807_2 = TensorMap('1807_Fathers-age-at-death_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 73.11696679892832, 'std': 13.113774847019958}, annotation_units=1, channel_map={'1807_Fathers-age-at-death_2_0': 0, })
ukb_1845_2 = TensorMap('1845_Mothers-age_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 83.37040151355897, 'std': 6.813201186172373}, annotation_units=1, channel_map={'1845_Mothers-age_2_0': 0, })
ukb_1845_0 = TensorMap('1845_Mothers-age_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.26739494692293, 'std': 8.147467930745506}, annotation_units=1, channel_map={'1845_Mothers-age_0_0': 0, })
ukb_1845_1 = TensorMap('1845_Mothers-age_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.15924657534248, 'std': 7.41909494506527}, annotation_units=1, channel_map={'1845_Mothers-age_1_0': 0, })
ukb_1873_0 = TensorMap('1873_Number-of-full-brothers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1572501901140682, 'std': 1.2897284410977443}, annotation_units=1, channel_map={'1873_Number-of-full-brothers_0_0': 0, })
ukb_1873_2 = TensorMap('1873_Number-of-full-brothers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0603260245666395, 'std': 1.1381762301646574}, annotation_units=1, channel_map={'1873_Number-of-full-brothers_2_0': 0, })
ukb_1873_1 = TensorMap('1873_Number-of-full-brothers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0022439413583326, 'std': 1.127950985881737}, annotation_units=1, channel_map={'1873_Number-of-full-brothers_1_0': 0, })
ukb_1883_1 = TensorMap('1883_Number-of-full-sisters_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9370730491149338, 'std': 1.0970323211429756}, annotation_units=1, channel_map={'1883_Number-of-full-sisters_1_0': 0, })
ukb_1883_2 = TensorMap('1883_Number-of-full-sisters_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9857093663911846, 'std': 1.1071511467949204}, annotation_units=1, channel_map={'1883_Number-of-full-sisters_2_0': 0, })
ukb_1883_0 = TensorMap('1883_Number-of-full-sisters_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0770222705227814, 'std': 1.2215502368418465}, annotation_units=1, channel_map={'1883_Number-of-full-sisters_0_0': 0, })
ukb_2217_0 = TensorMap('2217_Age-started-wearing-glasses-or-contact-lenses_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.791384009085775, 'std': 16.920436780693194}, annotation_units=1, channel_map={'2217_Age-started-wearing-glasses-or-contact-lenses_0_0': 0, })
ukb_2217_2 = TensorMap('2217_Age-started-wearing-glasses-or-contact-lenses_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.99161509900991, 'std': 17.492891704288652}, annotation_units=1, channel_map={'2217_Age-started-wearing-glasses-or-contact-lenses_2_0': 0, })
ukb_2217_1 = TensorMap('2217_Age-started-wearing-glasses-or-contact-lenses_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.221826424870468, 'std': 17.468234543919515}, annotation_units=1, channel_map={'2217_Age-started-wearing-glasses-or-contact-lenses_1_0': 0, })
ukb_2277_2 = TensorMap('2277_Frequency-of-solariumsunlamp-use_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3538984645652454, 'std': 3.7797743965865616}, annotation_units=1, channel_map={'2277_Frequency-of-solariumsunlamp-use_2_0': 0, })
ukb_2277_1 = TensorMap('2277_Frequency-of-solariumsunlamp-use_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3068630353266234, 'std': 2.7034177855967494}, annotation_units=1, channel_map={'2277_Frequency-of-solariumsunlamp-use_1_0': 0, })
ukb_2277_0 = TensorMap('2277_Frequency-of-solariumsunlamp-use_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5301349151674393, 'std': 4.513942902841888}, annotation_units=1, channel_map={'2277_Frequency-of-solariumsunlamp-use_0_0': 0, })
ukb_2355_0 = TensorMap('2355_Most-recent-bowel-cancer-screening_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.930222571630541, 'std': 4.311275554808345}, annotation_units=1, channel_map={'2355_Most-recent-bowel-cancer-screening_0_0': 0, })
ukb_2355_1 = TensorMap('2355_Most-recent-bowel-cancer-screening_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8676282563528845, 'std': 2.786216892450628}, annotation_units=1, channel_map={'2355_Most-recent-bowel-cancer-screening_1_0': 0, })
ukb_2355_2 = TensorMap('2355_Most-recent-bowel-cancer-screening_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.940024373259053, 'std': 2.8955998556238836}, annotation_units=1, channel_map={'2355_Most-recent-bowel-cancer-screening_2_0': 0, })
ukb_2405_0 = TensorMap('2405_Number-of-children-fathered_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.813973449556381, 'std': 1.3398570356329154}, annotation_units=1, channel_map={'2405_Number-of-children-fathered_0_0': 0, })
ukb_2405_2 = TensorMap('2405_Number-of-children-fathered_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7984409799554564, 'std': 1.2932362159813275}, annotation_units=1, channel_map={'2405_Number-of-children-fathered_2_0': 0, })
ukb_2405_1 = TensorMap('2405_Number-of-children-fathered_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.755589276681841, 'std': 1.2190110040732747}, annotation_units=1, channel_map={'2405_Number-of-children-fathered_1_0': 0, })
ukb_2684_1 = TensorMap('2684_Years-since-last-breast-cancer-screening-mammogram_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7559465623981756, 'std': 1.725378041416919}, annotation_units=1, channel_map={'2684_Years-since-last-breast-cancer-screening-mammogram_1_0': 0, })
ukb_2684_0 = TensorMap('2684_Years-since-last-breast-cancer-screening-mammogram_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8358682086301237, 'std': 2.176641748438276}, annotation_units=1, channel_map={'2684_Years-since-last-breast-cancer-screening-mammogram_0_0': 0, })
ukb_2684_2 = TensorMap('2684_Years-since-last-breast-cancer-screening-mammogram_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9784339353612164, 'std': 1.9821356448553542}, annotation_units=1, channel_map={'2684_Years-since-last-breast-cancer-screening-mammogram_2_0': 0, })
ukb_2704_1 = TensorMap('2704_Years-since-last-cervical-smear-test_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.48972483455242, 'std': 6.771486939720797}, annotation_units=1, channel_map={'2704_Years-since-last-cervical-smear-test_1_0': 0, })
ukb_2704_2 = TensorMap('2704_Years-since-last-cervical-smear-test_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.255318425881874, 'std': 7.524821380534116}, annotation_units=1, channel_map={'2704_Years-since-last-cervical-smear-test_2_0': 0, })
ukb_2704_0 = TensorMap('2704_Years-since-last-cervical-smear-test_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.168212935741287, 'std': 5.63150945242373}, annotation_units=1, channel_map={'2704_Years-since-last-cervical-smear-test_0_0': 0, })
ukb_2714_1 = TensorMap('2714_Age-when-periods-started-menarche_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.968131433095806, 'std': 1.6005584718160575}, annotation_units=1, channel_map={'2714_Age-when-periods-started-menarche_1_0': 0, })
ukb_2714_0 = TensorMap('2714_Age-when-periods-started-menarche_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.969831333681098, 'std': 1.6232172496143258}, annotation_units=1, channel_map={'2714_Age-when-periods-started-menarche_0_0': 0, })
ukb_2714_2 = TensorMap('2714_Age-when-periods-started-menarche_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.992641648270789, 'std': 1.5918080259501795}, annotation_units=1, channel_map={'2714_Age-when-periods-started-menarche_2_0': 0, })
ukb_2734_0 = TensorMap('2734_Number-of-live-births_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8217541722767847, 'std': 1.2016883451757863}, annotation_units=1, channel_map={'2734_Number-of-live-births_0_0': 0, })
ukb_2734_1 = TensorMap('2734_Number-of-live-births_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7523085802231626, 'std': 1.1645391689855076}, annotation_units=1, channel_map={'2734_Number-of-live-births_1_0': 0, })
ukb_2734_2 = TensorMap('2734_Number-of-live-births_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7348026822029237, 'std': 1.1633339821512452}, annotation_units=1, channel_map={'2734_Number-of-live-births_2_0': 0, })
ukb_2744_0 = TensorMap('2744_Birth-weight-of-first-child_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.942976502742263, 'std': 1.4074428513845638}, annotation_units=1, channel_map={'2744_Birth-weight-of-first-child_0_0': 0, })
ukb_2744_1 = TensorMap('2744_Birth-weight-of-first-child_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.033007334963324, 'std': 1.3212960335073876}, annotation_units=1, channel_map={'2744_Birth-weight-of-first-child_1_0': 0, })
ukb_2744_2 = TensorMap('2744_Birth-weight-of-first-child_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.034405280157281, 'std': 1.4128509622868657}, annotation_units=1, channel_map={'2744_Birth-weight-of-first-child_2_0': 0, })
ukb_2754_2 = TensorMap('2754_Age-at-first-live-birth_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.24698444388986, 'std': 4.62572297884837}, annotation_units=1, channel_map={'2754_Age-at-first-live-birth_2_0': 0, })
ukb_2754_0 = TensorMap('2754_Age-at-first-live-birth_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.267277615783694, 'std': 4.747366029144674}, annotation_units=1, channel_map={'2754_Age-at-first-live-birth_0_0': 0, })
ukb_2754_1 = TensorMap('2754_Age-at-first-live-birth_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.953271028037378, 'std': 4.456123181151487}, annotation_units=1, channel_map={'2754_Age-at-first-live-birth_1_0': 0, })
ukb_2764_1 = TensorMap('2764_Age-at-last-live-birth_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.511644623346747, 'std': 4.876004669437672}, annotation_units=1, channel_map={'2764_Age-at-last-live-birth_1_0': 0, })
ukb_2764_0 = TensorMap('2764_Age-at-last-live-birth_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.15137842984506, 'std': 5.280985327850116}, annotation_units=1, channel_map={'2764_Age-at-last-live-birth_0_0': 0, })
ukb_2764_2 = TensorMap('2764_Age-at-last-live-birth_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.824723400715413, 'std': 4.990237492683754}, annotation_units=1, channel_map={'2764_Age-at-last-live-birth_2_0': 0, })
ukb_2794_0 = TensorMap('2794_Age-started-oral-contraceptive-pill_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.535554780838968, 'std': 4.6719370796221416}, annotation_units=1, channel_map={'2794_Age-started-oral-contraceptive-pill_0_0': 0, })
ukb_2794_1 = TensorMap('2794_Age-started-oral-contraceptive-pill_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.656082071323894, 'std': 4.499793395537406}, annotation_units=1, channel_map={'2794_Age-started-oral-contraceptive-pill_1_0': 0, })
ukb_2794_2 = TensorMap('2794_Age-started-oral-contraceptive-pill_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.826958831341308, 'std': 4.252060938753426}, annotation_units=1, channel_map={'2794_Age-started-oral-contraceptive-pill_2_0': 0, })
ukb_2804_0 = TensorMap('2804_Age-when-last-used-oral-contraceptive-pill_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.823013855015322, 'std': 7.516648891125272}, annotation_units=1, channel_map={'2804_Age-when-last-used-oral-contraceptive-pill_0_0': 0, })
ukb_2804_1 = TensorMap('2804_Age-when-last-used-oral-contraceptive-pill_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.18562390631309, 'std': 7.741907587634704}, annotation_units=1, channel_map={'2804_Age-when-last-used-oral-contraceptive-pill_1_0': 0, })
ukb_2804_2 = TensorMap('2804_Age-when-last-used-oral-contraceptive-pill_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.37835668417979, 'std': 8.085081850932088}, annotation_units=1, channel_map={'2804_Age-when-last-used-oral-contraceptive-pill_2_0': 0, })
ukb_2824_0 = TensorMap('2824_Age-at-hysterectomy_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.92806996429574, 'std': 8.037896993292408}, annotation_units=1, channel_map={'2824_Age-at-hysterectomy_0_0': 0, })
ukb_2824_1 = TensorMap('2824_Age-at-hysterectomy_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.34810126582279, 'std': 8.551272471629076}, annotation_units=1, channel_map={'2824_Age-at-hysterectomy_1_0': 0, })
ukb_2824_2 = TensorMap('2824_Age-at-hysterectomy_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.68274350140934, 'std': 9.045513211687988}, annotation_units=1, channel_map={'2824_Age-at-hysterectomy_2_0': 0, })
ukb_2867_2 = TensorMap('2867_Age-started-smoking-in-former-smokers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.36760391198043, 'std': 3.55591486251228}, annotation_units=1, channel_map={'2867_Age-started-smoking-in-former-smokers_2_0': 0, })
ukb_2867_1 = TensorMap('2867_Age-started-smoking-in-former-smokers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.35707200952193, 'std': 3.568581772863722}, annotation_units=1, channel_map={'2867_Age-started-smoking-in-former-smokers_1_0': 0, })
ukb_2867_0 = TensorMap('2867_Age-started-smoking-in-former-smokers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.2822654367157, 'std': 3.7022878879460355}, annotation_units=1, channel_map={'2867_Age-started-smoking-in-former-smokers_0_0': 0, })
ukb_2887_0 = TensorMap('2887_Number-of-cigarettes-previously-smoked-daily_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.04508700236131, 'std': 10.516344971358313}, annotation_units=1, channel_map={'2887_Number-of-cigarettes-previously-smoked-daily_0_0': 0, })
ukb_2887_2 = TensorMap('2887_Number-of-cigarettes-previously-smoked-daily_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.437234453456934, 'std': 8.998409249322494}, annotation_units=1, channel_map={'2887_Number-of-cigarettes-previously-smoked-daily_2_0': 0, })
ukb_2887_1 = TensorMap('2887_Number-of-cigarettes-previously-smoked-daily_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 18.53008233058898, 'std': 9.71416842851607}, annotation_units=1, channel_map={'2887_Number-of-cigarettes-previously-smoked-daily_1_0': 0, })
ukb_2897_0 = TensorMap('2897_Age-stopped-smoking_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.85940162769237, 'std': 11.609253789489419}, annotation_units=1, channel_map={'2897_Age-stopped-smoking_0_0': 0, })
ukb_2897_2 = TensorMap('2897_Age-stopped-smoking_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.0122204570451, 'std': 12.027664975218007}, annotation_units=1, channel_map={'2897_Age-stopped-smoking_2_0': 0, })
ukb_2897_1 = TensorMap('2897_Age-stopped-smoking_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.096652802535154, 'std': 11.920171890814723}, annotation_units=1, channel_map={'2897_Age-stopped-smoking_1_0': 0, })
ukb_2926_1 = TensorMap('2926_Number-of-unsuccessful-stopsmoking-attempts_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.872848537005164, 'std': 5.900871346871442}, annotation_units=1, channel_map={'2926_Number-of-unsuccessful-stopsmoking-attempts_1_0': 0, })
ukb_2926_2 = TensorMap('2926_Number-of-unsuccessful-stopsmoking-attempts_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.027420850443767, 'std': 6.771297280674471}, annotation_units=1, channel_map={'2926_Number-of-unsuccessful-stopsmoking-attempts_2_0': 0, })
ukb_2926_0 = TensorMap('2926_Number-of-unsuccessful-stopsmoking-attempts_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.959226365237814, 'std': 7.502349093367745}, annotation_units=1, channel_map={'2926_Number-of-unsuccessful-stopsmoking-attempts_0_0': 0, })
ukb_2946_2 = TensorMap('2946_Fathers-age_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 83.17363214213609, 'std': 6.158033819701597}, annotation_units=1, channel_map={'2946_Fathers-age_2_0': 0, })
ukb_2946_1 = TensorMap('2946_Fathers-age_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 81.5430699481865, 'std': 6.67652991696039}, annotation_units=1, channel_map={'2946_Fathers-age_1_0': 0, })
ukb_2946_0 = TensorMap('2946_Fathers-age_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.86716383540293, 'std': 7.339379202289825}, annotation_units=1, channel_map={'2946_Fathers-age_0_0': 0, })
ukb_2966_0 = TensorMap('2966_Age-high-blood-pressure-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.62922015775057, 'std': 9.947561475383283}, annotation_units=1, channel_map={'2966_Age-high-blood-pressure-diagnosed_0_0': 0, })
ukb_2966_2 = TensorMap('2966_Age-high-blood-pressure-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.26281352235549, 'std': 10.247914306496828}, annotation_units=1, channel_map={'2966_Age-high-blood-pressure-diagnosed_2_0': 0, })
ukb_2966_1 = TensorMap('2966_Age-high-blood-pressure-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 52.75146475146474, 'std': 9.903448599750352}, annotation_units=1, channel_map={'2966_Age-high-blood-pressure-diagnosed_1_0': 0, })
ukb_2976_2 = TensorMap('2976_Age-diabetes-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 54.668490153172854, 'std': 12.207813433626267}, annotation_units=1, channel_map={'2976_Age-diabetes-diagnosed_2_0': 0, })
ukb_2976_1 = TensorMap('2976_Age-diabetes-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 54.25794732061762, 'std': 11.901441584223187}, annotation_units=1, channel_map={'2976_Age-diabetes-diagnosed_1_0': 0, })
ukb_2976_0 = TensorMap('2976_Age-diabetes-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.96104899369791, 'std': 12.596448570403021}, annotation_units=1, channel_map={'2976_Age-diabetes-diagnosed_0_0': 0, })
ukb_3062_1 = TensorMap('3062_Forced-vital-capacity-FVC_1_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.520976062274387, 'std': 1.0070632831303588}, annotation_units=3, channel_map={'3062_Forced-vital-capacity-FVC_1_0': 0, '3062_Forced-vital-capacity-FVC_1_1': 1, '3062_Forced-vital-capacity-FVC_1_2': 2, })
ukb_3062_0 = TensorMap('3062_Forced-vital-capacity-FVC_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5991826271920435, 'std': 1.0827952056506123}, annotation_units=3, channel_map={'3062_Forced-vital-capacity-FVC_0_0': 0, '3062_Forced-vital-capacity-FVC_0_1': 1, '3062_Forced-vital-capacity-FVC_0_2': 2, })
ukb_3062_2 = TensorMap('3062_Forced-vital-capacity-FVC_2_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.550082014554694, 'std': 1.0455056389708097}, annotation_units=3, channel_map={'3062_Forced-vital-capacity-FVC_2_0': 0, '3062_Forced-vital-capacity-FVC_2_1': 1, '3062_Forced-vital-capacity-FVC_2_2': 2, })
ukb_3063_0 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.723789321870705, 'std': 0.8269198582377145}, annotation_units=3, channel_map={'3063_Forced-expiratory-volume-in-1second-FEV1_0_0': 0, '3063_Forced-expiratory-volume-in-1second-FEV1_0_1': 1, '3063_Forced-expiratory-volume-in-1second-FEV1_0_2': 2, })
ukb_3063_2 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_2_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.639508952292943, 'std': 0.7953489292152817}, annotation_units=3, channel_map={'3063_Forced-expiratory-volume-in-1second-FEV1_2_0': 0, '3063_Forced-expiratory-volume-in-1second-FEV1_2_1': 1, '3063_Forced-expiratory-volume-in-1second-FEV1_2_2': 2, })
ukb_3063_1 = TensorMap('3063_Forced-expiratory-volume-in-1second-FEV1_1_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.621810151854316, 'std': 0.8085055840307454}, annotation_units=3, channel_map={'3063_Forced-expiratory-volume-in-1second-FEV1_1_0': 0, '3063_Forced-expiratory-volume-in-1second-FEV1_1_1': 1, '3063_Forced-expiratory-volume-in-1second-FEV1_1_2': 2, })
ukb_3064_0 = TensorMap('3064_Peak-expiratory-flow-PEF_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 365.8363272327348, 'std': 143.87922997760828}, annotation_units=3, channel_map={'3064_Peak-expiratory-flow-PEF_0_0': 0, '3064_Peak-expiratory-flow-PEF_0_1': 1, '3064_Peak-expiratory-flow-PEF_0_2': 2, })
ukb_3064_1 = TensorMap('3064_Peak-expiratory-flow-PEF_1_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 342.8814608667594, 'std': 145.34349840941277}, annotation_units=3, channel_map={'3064_Peak-expiratory-flow-PEF_1_0': 0, '3064_Peak-expiratory-flow-PEF_1_1': 1, '3064_Peak-expiratory-flow-PEF_1_2': 2, })
ukb_3064_2 = TensorMap('3064_Peak-expiratory-flow-PEF_2_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 342.0647568441724, 'std': 146.4449729557129}, annotation_units=3, channel_map={'3064_Peak-expiratory-flow-PEF_2_0': 0, '3064_Peak-expiratory-flow-PEF_2_1': 1, '3064_Peak-expiratory-flow-PEF_2_2': 2, })
ukb_3083_0 = TensorMap('3083_Heel-quantitative-ultrasound-index-QUI-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 97.68872542642548, 'std': 20.378981936847804}, annotation_units=1, channel_map={'3083_Heel-quantitative-ultrasound-index-QUI-manual-entry_0_0': 0, })
ukb_3084_0 = TensorMap('3084_Heel-bone-mineral-density-BMD-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5423222583588759, 'std': 0.12769083733767947}, annotation_units=1, channel_map={'3084_Heel-bone-mineral-density-BMD-manual-entry_0_0': 0, })
ukb_3085_0 = TensorMap('3085_Heel-Broadband-ultrasound-attenuation-BUA-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.18169716184278, 'std': 19.63987587039883}, annotation_units=1, channel_map={'3085_Heel-Broadband-ultrasound-attenuation-BUA-manual-entry_0_0': 0, })
ukb_3086_0 = TensorMap('3086_Speed-of-sound-through-heel-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1553.188703526054, 'std': 43.56541278705645}, annotation_units=1, channel_map={'3086_Speed-of-sound-through-heel-manual-entry_0_0': 0, })
ukb_3137_0 = TensorMap('3137_Number-of-measurements-made_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.735440488050005, 'std': 0.44248075066155873}, annotation_units=1, channel_map={'3137_Number-of-measurements-made_0_0': 0, })
ukb_3137_2 = TensorMap('3137_Number-of-measurements-made_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.735401921132456, 'std': 0.44191317330802976}, annotation_units=1, channel_map={'3137_Number-of-measurements-made_2_0': 0, })
ukb_3137_1 = TensorMap('3137_Number-of-measurements-made_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.738856669428334, 'std': 0.4393957464750786}, annotation_units=1, channel_map={'3137_Number-of-measurements-made_1_0': 0, })
ukb_3143_0 = TensorMap('3143_Ankle-spacing-width_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.8405259583004, 'std': 4.605420712700601}, annotation_units=1, channel_map={'3143_Ankle-spacing-width_0_0': 0, })
ukb_3144_0 = TensorMap('3144_Heel-Broadband-ultrasound-attenuation-direct-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.91299497354228, 'std': 19.602098137766433}, annotation_units=1, channel_map={'3144_Heel-Broadband-ultrasound-attenuation-direct-entry_0_0': 0, })
ukb_3146_0 = TensorMap('3146_Speed-of-sound-through-heel_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1553.774095256863, 'std': 35.749527179298255}, annotation_units=1, channel_map={'3146_Speed-of-sound-through-heel_0_0': 0, })
ukb_3147_0 = TensorMap('3147_Heel-quantitative-ultrasound-index-QUI-direct-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 97.98964639607354, 'std': 22.057125971374475}, annotation_units=1, channel_map={'3147_Heel-quantitative-ultrasound-index-QUI-direct-entry_0_0': 0, })
ukb_3148_0 = TensorMap('3148_Heel-bone-mineral-density-BMD_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5435519713261646, 'std': 0.13892501345714736}, annotation_units=1, channel_map={'3148_Heel-bone-mineral-density-BMD_0_0': 0, })
ukb_3160_1 = TensorMap('3160_Weight-manual-entry_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 80.87451737451735, 'std': 17.270999106432345}, annotation_units=1, channel_map={'3160_Weight-manual-entry_1_0': 0, })
ukb_3160_0 = TensorMap('3160_Weight-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.50371855880354, 'std': 18.71700229682232}, annotation_units=1, channel_map={'3160_Weight-manual-entry_0_0': 0, })
ukb_3160_2 = TensorMap('3160_Weight-manual-entry_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 75.77009102730821, 'std': 16.723385731670632}, annotation_units=1, channel_map={'3160_Weight-manual-entry_2_0': 0, })
ukb_3436_0 = TensorMap('3436_Age-started-smoking-in-current-smokers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.841919257061775, 'std': 5.793512220681287}, annotation_units=1, channel_map={'3436_Age-started-smoking-in-current-smokers_0_0': 0, })
ukb_3436_1 = TensorMap('3436_Age-started-smoking-in-current-smokers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 18.633385335413415, 'std': 7.202955725106682}, annotation_units=1, channel_map={'3436_Age-started-smoking-in-current-smokers_1_0': 0, })
ukb_3436_2 = TensorMap('3436_Age-started-smoking-in-current-smokers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 18.92792792792793, 'std': 7.404196007012701}, annotation_units=1, channel_map={'3436_Age-started-smoking-in-current-smokers_2_0': 0, })
ukb_3456_0 = TensorMap('3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.541285291596385, 'std': 8.449555990368303}, annotation_units=1, channel_map={'3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_0_0': 0, })
ukb_3456_1 = TensorMap('3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.849220103986134, 'std': 8.122957237763817}, annotation_units=1, channel_map={'3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_1_0': 0, })
ukb_3456_2 = TensorMap('3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.79266572637518, 'std': 7.916096190725274}, annotation_units=1, channel_map={'3456_Number-of-cigarettes-currently-smoked-daily-current-cigarette-smokers_2_0': 0, })
ukb_3526_2 = TensorMap('3526_Mothers-age-at-death_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.57056903535337, 'std': 13.033913807832029}, annotation_units=1, channel_map={'3526_Mothers-age-at-death_2_0': 0, })
ukb_3526_0 = TensorMap('3526_Mothers-age-at-death_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 74.12840796819283, 'std': 13.249138193853833}, annotation_units=1, channel_map={'3526_Mothers-age-at-death_0_0': 0, })
ukb_3526_1 = TensorMap('3526_Mothers-age-at-death_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.56108214738619, 'std': 13.041472584452073}, annotation_units=1, channel_map={'3526_Mothers-age-at-death_1_0': 0, })
ukb_3536_0 = TensorMap('3536_Age-started-hormonereplacement-therapy-HRT_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.37671894106794, 'std': 5.445443728155725}, annotation_units=1, channel_map={'3536_Age-started-hormonereplacement-therapy-HRT_0_0': 0, })
ukb_3536_1 = TensorMap('3536_Age-started-hormonereplacement-therapy-HRT_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.757452574525736, 'std': 5.37546407319619}, annotation_units=1, channel_map={'3536_Age-started-hormonereplacement-therapy-HRT_1_0': 0, })
ukb_3536_2 = TensorMap('3536_Age-started-hormonereplacement-therapy-HRT_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.888042398145075, 'std': 5.838366144867489}, annotation_units=1, channel_map={'3536_Age-started-hormonereplacement-therapy-HRT_2_0': 0, })
ukb_3546_0 = TensorMap('3546_Age-last-used-hormonereplacement-therapy-HRT_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.92169718601282, 'std': 5.941782727750997}, annotation_units=1, channel_map={'3546_Age-last-used-hormonereplacement-therapy-HRT_0_0': 0, })
ukb_3546_2 = TensorMap('3546_Age-last-used-hormonereplacement-therapy-HRT_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 54.02254722730043, 'std': 6.149473921559484}, annotation_units=1, channel_map={'3546_Age-last-used-hormonereplacement-therapy-HRT_2_0': 0, })
ukb_3546_1 = TensorMap('3546_Age-last-used-hormonereplacement-therapy-HRT_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 54.43424396442186, 'std': 5.86652155464712}, annotation_units=1, channel_map={'3546_Age-last-used-hormonereplacement-therapy-HRT_1_0': 0, })
ukb_3581_1 = TensorMap('3581_Age-at-menopause-last-menstrual-period_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.22321678321678, 'std': 4.913715507218971}, annotation_units=1, channel_map={'3581_Age-at-menopause-last-menstrual-period_1_0': 0, })
ukb_3581_0 = TensorMap('3581_Age-at-menopause-last-menstrual-period_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.664334262414876, 'std': 5.141535161191496}, annotation_units=1, channel_map={'3581_Age-at-menopause-last-menstrual-period_0_0': 0, })
ukb_3581_2 = TensorMap('3581_Age-at-menopause-last-menstrual-period_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.277397260273965, 'std': 4.867237622147741}, annotation_units=1, channel_map={'3581_Age-at-menopause-last-menstrual-period_2_0': 0, })
ukb_3627_0 = TensorMap('3627_Age-angina-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.528122084499536, 'std': 7.995373672649764}, annotation_units=1, channel_map={'3627_Age-angina-diagnosed_0_0': 0, })
ukb_3627_1 = TensorMap('3627_Age-angina-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 55.79288025889968, 'std': 8.636499444039991}, annotation_units=1, channel_map={'3627_Age-angina-diagnosed_1_0': 0, })
ukb_3627_2 = TensorMap('3627_Age-angina-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.20460048426151, 'std': 9.191987454046336}, annotation_units=1, channel_map={'3627_Age-angina-diagnosed_2_0': 0, })
ukb_3659_2 = TensorMap('3659_Year-immigrated-to-UK-United-Kingdom_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1972.8964941569282, 'std': 15.466955964860894}, annotation_units=1, channel_map={'3659_Year-immigrated-to-UK-United-Kingdom_2_0': 0, })
ukb_3659_0 = TensorMap('3659_Year-immigrated-to-UK-United-Kingdom_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1976.0454825251352, 'std': 16.240769055300316}, annotation_units=1, channel_map={'3659_Year-immigrated-to-UK-United-Kingdom_0_0': 0, })
ukb_3659_1 = TensorMap('3659_Year-immigrated-to-UK-United-Kingdom_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1971.7350705754614, 'std': 15.407697018486953}, annotation_units=1, channel_map={'3659_Year-immigrated-to-UK-United-Kingdom_1_0': 0, })
ukb_3680_2 = TensorMap('3680_Age-when-last-ate-meat_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.21325051759835, 'std': 14.874794921959493}, annotation_units=1, channel_map={'3680_Age-when-last-ate-meat_2_0': 0, })
ukb_3680_0 = TensorMap('3680_Age-when-last-ate-meat_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.73972739527685, 'std': 14.418396711020819}, annotation_units=1, channel_map={'3680_Age-when-last-ate-meat_0_0': 0, })
ukb_3680_1 = TensorMap('3680_Age-when-last-ate-meat_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.3978102189781, 'std': 13.195562930213656}, annotation_units=1, channel_map={'3680_Age-when-last-ate-meat_1_0': 0, })
ukb_3700_2 = TensorMap('3700_Time-since-last-menstrual-period_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.178142076502738, 'std': 30.221320072365057}, annotation_units=1, channel_map={'3700_Time-since-last-menstrual-period_2_0': 0, })
ukb_3700_1 = TensorMap('3700_Time-since-last-menstrual-period_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.30518018018018, 'std': 23.886293404258655}, annotation_units=1, channel_map={'3700_Time-since-last-menstrual-period_1_0': 0, })
ukb_3700_0 = TensorMap('3700_Time-since-last-menstrual-period_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.229588661058848, 'std': 22.325207002246323}, annotation_units=1, channel_map={'3700_Time-since-last-menstrual-period_0_0': 0, })
ukb_3710_0 = TensorMap('3710_Length-of-menstrual-cycle_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.76382935637382, 'std': 14.242673038653914}, annotation_units=1, channel_map={'3710_Length-of-menstrual-cycle_0_0': 0, })
ukb_3710_1 = TensorMap('3710_Length-of-menstrual-cycle_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 18.11971104231166, 'std': 15.790732241090202}, annotation_units=1, channel_map={'3710_Length-of-menstrual-cycle_1_0': 0, })
ukb_3710_2 = TensorMap('3710_Length-of-menstrual-cycle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.463825569871158, 'std': 20.263270432101148}, annotation_units=1, channel_map={'3710_Length-of-menstrual-cycle_2_0': 0, })
ukb_3761_1 = TensorMap('3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.24331926863572, 'std': 17.009930446255137}, annotation_units=1, channel_map={'3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_1_0': 0, })
ukb_3761_0 = TensorMap('3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.206855225311603, 'std': 16.127075100760692}, annotation_units=1, channel_map={'3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_0_0': 0, })
ukb_3761_2 = TensorMap('3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.293945831120542, 'std': 17.41338641051822}, annotation_units=1, channel_map={'3761_Age-hay-fever-rhinitis-or-eczema-diagnosed_2_0': 0, })
ukb_3786_0 = TensorMap('3786_Age-asthma-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.156387837152117, 'std': 18.68854991513305}, annotation_units=1, channel_map={'3786_Age-asthma-diagnosed_0_0': 0, })
ukb_3786_1 = TensorMap('3786_Age-asthma-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.44038461538461, 'std': 19.651076165563918}, annotation_units=1, channel_map={'3786_Age-asthma-diagnosed_1_0': 0, })
ukb_3786_2 = TensorMap('3786_Age-asthma-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.983220568335586, 'std': 19.833749733698397}, annotation_units=1, channel_map={'3786_Age-asthma-diagnosed_2_0': 0, })
ukb_3809_1 = TensorMap('3809_Time-since-last-prostate-specific-antigen-PSA-test_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.4097510373443987, 'std': 2.7454049978234814}, annotation_units=1, channel_map={'3809_Time-since-last-prostate-specific-antigen-PSA-test_1_0': 0, })
ukb_3809_2 = TensorMap('3809_Time-since-last-prostate-specific-antigen-PSA-test_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.6313194444444448, 'std': 3.282008172575194}, annotation_units=1, channel_map={'3809_Time-since-last-prostate-specific-antigen-PSA-test_2_0': 0, })
ukb_3809_0 = TensorMap('3809_Time-since-last-prostate-specific-antigen-PSA-test_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.259440635542495, 'std': 2.4138830716050705}, annotation_units=1, channel_map={'3809_Time-since-last-prostate-specific-antigen-PSA-test_0_0': 0, })
ukb_3829_0 = TensorMap('3829_Number-of-stillbirths_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09751554937285155, 'std': 0.3785595696632465}, annotation_units=1, channel_map={'3829_Number-of-stillbirths_0_0': 0, })
ukb_3829_1 = TensorMap('3829_Number-of-stillbirths_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07996083550913838, 'std': 0.3997263164680728}, annotation_units=1, channel_map={'3829_Number-of-stillbirths_1_0': 0, })
ukb_3829_2 = TensorMap('3829_Number-of-stillbirths_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07399425287356326, 'std': 0.3357385623063343}, annotation_units=1, channel_map={'3829_Number-of-stillbirths_2_0': 0, })
ukb_3839_0 = TensorMap('3839_Number-of-spontaneous-miscarriages_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9474939509160044, 'std': 1.0744542549213256}, annotation_units=1, channel_map={'3839_Number-of-spontaneous-miscarriages_0_0': 0, })
ukb_3839_1 = TensorMap('3839_Number-of-spontaneous-miscarriages_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9741914407056517, 'std': 1.0137874891427696}, annotation_units=1, channel_map={'3839_Number-of-spontaneous-miscarriages_1_0': 0, })
ukb_3839_2 = TensorMap('3839_Number-of-spontaneous-miscarriages_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.924901185770751, 'std': 0.9713421671520918}, annotation_units=1, channel_map={'3839_Number-of-spontaneous-miscarriages_2_0': 0, })
ukb_3849_0 = TensorMap('3849_Number-of-pregnancy-terminations_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5511111625806303, 'std': 0.7549118899564189}, annotation_units=1, channel_map={'3849_Number-of-pregnancy-terminations_0_0': 0, })
ukb_3849_2 = TensorMap('3849_Number-of-pregnancy-terminations_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5208295761947701, 'std': 0.690306541474882}, annotation_units=1, channel_map={'3849_Number-of-pregnancy-terminations_2_0': 0, })
ukb_3849_1 = TensorMap('3849_Number-of-pregnancy-terminations_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46017699115044247, 'std': 0.6518351155347414}, annotation_units=1, channel_map={'3849_Number-of-pregnancy-terminations_1_0': 0, })
ukb_3872_2 = TensorMap('3872_Age-of-primiparous-women-at-birth-of-child_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.895358649789024, 'std': 6.315339722161406}, annotation_units=1, channel_map={'3872_Age-of-primiparous-women-at-birth-of-child_2_0': 0, })
ukb_3872_0 = TensorMap('3872_Age-of-primiparous-women-at-birth-of-child_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.997146228356613, 'std': 6.417228931220436}, annotation_units=1, channel_map={'3872_Age-of-primiparous-women-at-birth-of-child_0_0': 0, })
ukb_3872_1 = TensorMap('3872_Age-of-primiparous-women-at-birth-of-child_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.290199081163852, 'std': 6.361462912081801}, annotation_units=1, channel_map={'3872_Age-of-primiparous-women-at-birth-of-child_1_0': 0, })
ukb_3882_2 = TensorMap('3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.09875776397516, 'std': 8.460976759732345}, annotation_units=1, channel_map={'3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_2_0': 0, })
ukb_3882_0 = TensorMap('3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.44636286442116, 'std': 7.773399828695975}, annotation_units=1, channel_map={'3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_0_0': 0, })
ukb_3882_1 = TensorMap('3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.05468749999999, 'std': 7.953237653684042}, annotation_units=1, channel_map={'3882_Age-at-bilateral-oophorectomy-both-ovaries-removed_1_0': 0, })
ukb_3894_1 = TensorMap('3894_Age-heart-attack-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 55.15717092337918, 'std': 9.016763295497277}, annotation_units=1, channel_map={'3894_Age-heart-attack-diagnosed_1_0': 0, })
ukb_3894_2 = TensorMap('3894_Age-heart-attack-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.33558863328822, 'std': 9.0439137943391}, annotation_units=1, channel_map={'3894_Age-heart-attack-diagnosed_2_0': 0, })
ukb_3894_0 = TensorMap('3894_Age-heart-attack-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 52.696820860868016, 'std': 8.419897728297665}, annotation_units=1, channel_map={'3894_Age-heart-attack-diagnosed_0_0': 0, })
ukb_3972_1 = TensorMap('3972_Number-of-adopted-brothers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.52991452991453, 'std': 0.9225142875258155}, annotation_units=1, channel_map={'3972_Number-of-adopted-brothers_1_0': 0, })
ukb_3972_2 = TensorMap('3972_Number-of-adopted-brothers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5532879818594105, 'std': 0.8215806998849338}, annotation_units=1, channel_map={'3972_Number-of-adopted-brothers_2_0': 0, })
ukb_3972_0 = TensorMap('3972_Number-of-adopted-brothers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6873114181356199, 'std': 1.1830089577600718}, annotation_units=1, channel_map={'3972_Number-of-adopted-brothers_0_0': 0, })
ukb_3982_0 = TensorMap('3982_Number-of-adopted-sisters_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6215491402429408, 'std': 1.0470997366739097}, annotation_units=1, channel_map={'3982_Number-of-adopted-sisters_0_0': 0, })
ukb_3982_1 = TensorMap('3982_Number-of-adopted-sisters_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4377682403433476, 'std': 0.7524318032019482}, annotation_units=1, channel_map={'3982_Number-of-adopted-sisters_1_0': 0, })
ukb_3982_2 = TensorMap('3982_Number-of-adopted-sisters_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5727272727272726, 'std': 1.0075712552143552}, annotation_units=1, channel_map={'3982_Number-of-adopted-sisters_2_0': 0, })
ukb_3992_2 = TensorMap('3992_Age-emphysemachronic-bronchitis-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.456193353474305, 'std': 20.910205917832027}, annotation_units=1, channel_map={'3992_Age-emphysemachronic-bronchitis-diagnosed_2_0': 0, })
ukb_3992_0 = TensorMap('3992_Age-emphysemachronic-bronchitis-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.862643080124855, 'std': 19.951267875285666}, annotation_units=1, channel_map={'3992_Age-emphysemachronic-bronchitis-diagnosed_0_0': 0, })
ukb_3992_1 = TensorMap('3992_Age-emphysemachronic-bronchitis-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.597285067873315, 'std': 19.029063700637305}, annotation_units=1, channel_map={'3992_Age-emphysemachronic-bronchitis-diagnosed_1_0': 0, })
ukb_4012_0 = TensorMap('4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.03936501126248, 'std': 13.764394088734916}, annotation_units=1, channel_map={'4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_0_0': 0, })
ukb_4012_2 = TensorMap('4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.268248175182485, 'std': 14.827316033747374}, annotation_units=1, channel_map={'4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_2_0': 0, })
ukb_4012_1 = TensorMap('4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.72606382978722, 'std': 14.723476453108763}, annotation_units=1, channel_map={'4012_Age-deepvein-thrombosis-DVT-blood-clot-in-leg-diagnosed_1_0': 0, })
ukb_4022_2 = TensorMap('4022_Age-pulmonary-embolism-blood-clot-in-lung-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.62307692307694, 'std': 13.91605577225847}, annotation_units=1, channel_map={'4022_Age-pulmonary-embolism-blood-clot-in-lung-diagnosed_2_0': 0, })
ukb_4022_0 = TensorMap('4022_Age-pulmonary-embolism-blood-clot-in-lung-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.80101137800253, 'std': 13.012234972007672}, annotation_units=1, channel_map={'4022_Age-pulmonary-embolism-blood-clot-in-lung-diagnosed_0_0': 0, })
ukb_4022_1 = TensorMap('4022_Age-pulmonary-embolism-blood-clot-in-lung-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 48.18238993710692, 'std': 13.50192454444731}, annotation_units=1, channel_map={'4022_Age-pulmonary-embolism-blood-clot-in-lung-diagnosed_1_0': 0, })
ukb_4056_1 = TensorMap('4056_Age-stroke-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.71386430678464, 'std': 9.787572853380512}, annotation_units=1, channel_map={'4056_Age-stroke-diagnosed_1_0': 0, })
ukb_4056_0 = TensorMap('4056_Age-stroke-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.57258174852922, 'std': 9.477761437965563}, annotation_units=1, channel_map={'4056_Age-stroke-diagnosed_0_0': 0, })
ukb_4056_2 = TensorMap('4056_Age-stroke-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.548514851485166, 'std': 9.614500187809355}, annotation_units=1, channel_map={'4056_Age-stroke-diagnosed_2_0': 0, })
ukb_4079_1 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_1_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 80.15698796970149, 'std': 10.085490427812452}, annotation_units=2, channel_map={'4079_Diastolic-blood-pressure-automated-reading_1_0': 0, '4079_Diastolic-blood-pressure-automated-reading_1_1': 1, })
ukb_4079_0 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_0_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.20657551284782, 'std': 10.496040770224477}, annotation_units=2, channel_map={'4079_Diastolic-blood-pressure-automated-reading_0_0': 0, '4079_Diastolic-blood-pressure-automated-reading_0_1': 1, })
ukb_4079_2 = TensorMap('4079_Diastolic-blood-pressure-automated-reading_2_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.62527110186377, 'std': 10.32255932496225}, annotation_units=2, channel_map={'4079_Diastolic-blood-pressure-automated-reading_2_0': 0, '4079_Diastolic-blood-pressure-automated-reading_2_1': 1, })
ukb_4080_0 = TensorMap('4080_Systolic-blood-pressure-automated-reading_0_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 137.79964191990322, 'std': 19.292863700283753}, annotation_units=2, channel_map={'4080_Systolic-blood-pressure-automated-reading_0_0': 0, '4080_Systolic-blood-pressure-automated-reading_0_1': 1, })
ukb_4080_1 = TensorMap('4080_Systolic-blood-pressure-automated-reading_1_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 138.87172885048653, 'std': 19.246995969612506}, annotation_units=2, channel_map={'4080_Systolic-blood-pressure-automated-reading_1_0': 0, '4080_Systolic-blood-pressure-automated-reading_1_1': 1, })
ukb_4080_2 = TensorMap('4080_Systolic-blood-pressure-automated-reading_2_1', loss='logcosh', path_prefix='continuous', normalization={'mean': 137.87120664421366, 'std': 18.92717078493836}, annotation_units=2, channel_map={'4080_Systolic-blood-pressure-automated-reading_2_0': 0, '4080_Systolic-blood-pressure-automated-reading_2_1': 1, })
ukb_4100_0 = TensorMap('4100_Ankle-spacing-width-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.02652529853431, 'std': 4.628201953878173}, annotation_units=1, channel_map={'4100_Ankle-spacing-width-left_0_0': 0, })
ukb_4100_1 = TensorMap('4100_Ankle-spacing-width-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.1427076604764, 'std': 4.511404183853649}, annotation_units=1, channel_map={'4100_Ankle-spacing-width-left_1_0': 0, })
ukb_4100_2 = TensorMap('4100_Ankle-spacing-width-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.32222672452967, 'std': 4.668582204260848}, annotation_units=1, channel_map={'4100_Ankle-spacing-width-left_2_0': 0, })
ukb_4101_1 = TensorMap('4101_Heel-broadband-ultrasound-attenuation-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.41956368591035, 'std': 19.68942582697869}, annotation_units=1, channel_map={'4101_Heel-broadband-ultrasound-attenuation-left_1_0': 0, })
ukb_4101_2 = TensorMap('4101_Heel-broadband-ultrasound-attenuation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.31187115922928, 'std': 19.88961985085298}, annotation_units=1, channel_map={'4101_Heel-broadband-ultrasound-attenuation-left_2_0': 0, })
ukb_4101_0 = TensorMap('4101_Heel-broadband-ultrasound-attenuation-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.92567923195446, 'std': 20.761681015550657}, annotation_units=1, channel_map={'4101_Heel-broadband-ultrasound-attenuation-left_0_0': 0, })
ukb_4103_2 = TensorMap('4103_Speed-of-sound-through-heel-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1556.1225674815955, 'std': 35.67769530674799}, annotation_units=1, channel_map={'4103_Speed-of-sound-through-heel-left_2_0': 0, })
ukb_4103_1 = TensorMap('4103_Speed-of-sound-through-heel-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1556.5076635042392, 'std': 35.5332533241774}, annotation_units=1, channel_map={'4103_Speed-of-sound-through-heel-left_1_0': 0, })
ukb_4103_0 = TensorMap('4103_Speed-of-sound-through-heel-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1555.2454639231773, 'std': 37.30059670558073}, annotation_units=1, channel_map={'4103_Speed-of-sound-through-heel-left_0_0': 0, })
ukb_4104_2 = TensorMap('4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.11810498954831, 'std': 22.175384189548094}, annotation_units=1, channel_map={'4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_2_0': 0, })
ukb_4104_1 = TensorMap('4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.32014558942268, 'std': 22.018449527299175}, annotation_units=1, channel_map={'4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_1_0': 0, })
ukb_4104_0 = TensorMap('4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 98.59814059883446, 'std': 23.20942289761857}, annotation_units=1, channel_map={'4104_Heel-quantitative-ultrasound-index-QUI-direct-entry-left_0_0': 0, })
ukb_4105_0 = TensorMap('4105_Heel-bone-mineral-density-BMD-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.547471017627668, 'std': 0.14610377587001142}, annotation_units=1, channel_map={'4105_Heel-bone-mineral-density-BMD-left_0_0': 0, })
ukb_4105_2 = TensorMap('4105_Heel-bone-mineral-density-BMD-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5508511518486517, 'std': 0.1393418206640907}, annotation_units=1, channel_map={'4105_Heel-bone-mineral-density-BMD-left_2_0': 0, })
ukb_4105_1 = TensorMap('4105_Heel-bone-mineral-density-BMD-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5522859307906035, 'std': 0.13805145667428556}, annotation_units=1, channel_map={'4105_Heel-bone-mineral-density-BMD-left_1_0': 0, })
ukb_4106_1 = TensorMap('4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.26177664513524423, 'std': 1.243063833941224}, annotation_units=1, channel_map={'4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_1_0': 0, })
ukb_4106_0 = TensorMap('4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.30253547397073416, 'std': 1.3102957402769355}, annotation_units=1, channel_map={'4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_0_0': 0, })
ukb_4106_2 = TensorMap('4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.2731805053629783, 'std': 1.2519193447591126}, annotation_units=1, channel_map={'4106_Heel-bone-mineral-density-BMD-Tscore-automated-left_2_0': 0, })
ukb_4119_2 = TensorMap('4119_Ankle-spacing-width-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.059126038605875, 'std': 4.665801852053487}, annotation_units=1, channel_map={'4119_Ankle-spacing-width-right_2_0': 0, })
ukb_4119_0 = TensorMap('4119_Ankle-spacing-width-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.03751425161515, 'std': 4.577033018195549}, annotation_units=1, channel_map={'4119_Ankle-spacing-width-right_0_0': 0, })
ukb_4119_1 = TensorMap('4119_Ankle-spacing-width-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.048066288354356, 'std': 4.442768064331982}, annotation_units=1, channel_map={'4119_Ankle-spacing-width-right_1_0': 0, })
ukb_4120_0 = TensorMap('4120_Heel-broadband-ultrasound-attenuation-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.1111864994064, 'std': 20.4853761942473}, annotation_units=1, channel_map={'4120_Heel-broadband-ultrasound-attenuation-right_0_0': 0, })
ukb_4120_2 = TensorMap('4120_Heel-broadband-ultrasound-attenuation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.6839604303798, 'std': 19.946955888825908}, annotation_units=1, channel_map={'4120_Heel-broadband-ultrasound-attenuation-right_2_0': 0, })
ukb_4120_1 = TensorMap('4120_Heel-broadband-ultrasound-attenuation-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 80.9842621534753, 'std': 19.75594727940456}, annotation_units=1, channel_map={'4120_Heel-broadband-ultrasound-attenuation-right_1_0': 0, })
ukb_4122_2 = TensorMap('4122_Speed-of-sound-through-heel-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1557.0496799421364, 'std': 36.24962197279599}, annotation_units=1, channel_map={'4122_Speed-of-sound-through-heel-right_2_0': 0, })
ukb_4122_0 = TensorMap('4122_Speed-of-sound-through-heel-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1555.8457418694866, 'std': 37.17015218498807}, annotation_units=1, channel_map={'4122_Speed-of-sound-through-heel-right_0_0': 0, })
ukb_4122_1 = TensorMap('4122_Speed-of-sound-through-heel-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1557.9249409933216, 'std': 36.08331890669707}, annotation_units=1, channel_map={'4122_Speed-of-sound-through-heel-right_1_0': 0, })
ukb_4123_0 = TensorMap('4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.32985275259257, 'std': 23.02203573498749}, annotation_units=1, channel_map={'4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_0_0': 0, })
ukb_4123_2 = TensorMap('4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 100.05915932281538, 'std': 22.439337241747854}, annotation_units=1, channel_map={'4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_2_0': 0, })
ukb_4123_1 = TensorMap('4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 100.95097780344497, 'std': 22.27703353631703}, annotation_units=1, channel_map={'4123_Heel-quantitative-ultrasound-index-QUI-direct-entry-right_1_0': 0, })
ukb_4124_0 = TensorMap('4124_Heel-bone-mineral-density-BMD-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5521481253541807, 'std': 0.14481589050864077}, annotation_units=1, channel_map={'4124_Heel-bone-mineral-density-BMD-right_0_0': 0, })
ukb_4124_2 = TensorMap('4124_Heel-bone-mineral-density-BMD-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5567434434090294, 'std': 0.14114169383237624}, annotation_units=1, channel_map={'4124_Heel-bone-mineral-density-BMD-right_2_0': 0, })
ukb_4124_1 = TensorMap('4124_Heel-bone-mineral-density-BMD-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5623149432217865, 'std': 0.140214279642488}, annotation_units=1, channel_map={'4124_Heel-bone-mineral-density-BMD-right_1_0': 0, })
ukb_4125_2 = TensorMap('4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.22005297238861263, 'std': 1.2668207235566284}, annotation_units=1, channel_map={'4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_2_0': 0, })
ukb_4125_0 = TensorMap('4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.2612268506396633, 'std': 1.2997182814485158}, annotation_units=1, channel_map={'4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_0_0': 0, })
ukb_4125_1 = TensorMap('4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.16970466529402892, 'std': 1.2576590401441707}, annotation_units=1, channel_map={'4125_Heel-bone-mineral-density-BMD-Tscore-automated-right_1_0': 0, })
ukb_4138_2 = TensorMap('4138_Heel-bone-mineral-density-BMD-Tscore-manual-entry-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7788732394366197, 'std': 0.776136182566386}, annotation_units=1, channel_map={'4138_Heel-bone-mineral-density-BMD-Tscore-manual-entry-left_2_0': 0, })
ukb_4138_1 = TensorMap('4138_Heel-bone-mineral-density-BMD-Tscore-manual-entry-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6492682926829269, 'std': 1.0759227463426735}, annotation_units=1, channel_map={'4138_Heel-bone-mineral-density-BMD-Tscore-manual-entry-left_1_0': 0, })
ukb_4138_0 = TensorMap('4138_Heel-bone-mineral-density-BMD-Tscore-manual-entry-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6429804560260586, 'std': 1.0076706181075972}, annotation_units=1, channel_map={'4138_Heel-bone-mineral-density-BMD-Tscore-manual-entry-left_0_0': 0, })
ukb_4139_2 = TensorMap('4139_Heel-quantitative-ultrasound-index-QUI-manual-entry-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 98.75845070422534, 'std': 19.77679076346332}, annotation_units=1, channel_map={'4139_Heel-quantitative-ultrasound-index-QUI-manual-entry-left_2_0': 0, })
ukb_4139_1 = TensorMap('4139_Heel-quantitative-ultrasound-index-QUI-manual-entry-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.07658536585366, 'std': 20.86020082547617}, annotation_units=1, channel_map={'4139_Heel-quantitative-ultrasound-index-QUI-manual-entry-left_1_0': 0, })
ukb_4139_0 = TensorMap('4139_Heel-quantitative-ultrasound-index-QUI-manual-entry-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 97.7030944625407, 'std': 20.444443087821778}, annotation_units=1, channel_map={'4139_Heel-quantitative-ultrasound-index-QUI-manual-entry-left_0_0': 0, })
ukb_4140_0 = TensorMap('4140_Heel-bone-mineral-density-BMD-manual-entry-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5419315960912052, 'std': 0.12838987287660628}, annotation_units=1, channel_map={'4140_Heel-bone-mineral-density-BMD-manual-entry-left_0_0': 0, })
ukb_4140_2 = TensorMap('4140_Heel-bone-mineral-density-BMD-manual-entry-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5518077464788732, 'std': 0.11564688371338036}, annotation_units=1, channel_map={'4140_Heel-bone-mineral-density-BMD-manual-entry-left_2_0': 0, })
ukb_4140_1 = TensorMap('4140_Heel-bone-mineral-density-BMD-manual-entry-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5504829268292682, 'std': 0.13190878440481657}, annotation_units=1, channel_map={'4140_Heel-bone-mineral-density-BMD-manual-entry-left_1_0': 0, })
ukb_4141_0 = TensorMap('4141_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.35365635179151, 'std': 24.762803899282975}, annotation_units=1, channel_map={'4141_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-left_0_0': 0, })
ukb_4141_2 = TensorMap('4141_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.81549295774649, 'std': 18.16511072434213}, annotation_units=1, channel_map={'4141_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-left_2_0': 0, })
ukb_4141_1 = TensorMap('4141_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.6029268292683, 'std': 19.690417476815636}, annotation_units=1, channel_map={'4141_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-left_1_0': 0, })
ukb_4142_0 = TensorMap('4142_Speed-of-sound-through-heel-manual-entry-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1551.8999185667753, 'std': 69.66185558734954}, annotation_units=1, channel_map={'4142_Speed-of-sound-through-heel-manual-entry-left_0_0': 0, })
ukb_4142_2 = TensorMap('4142_Speed-of-sound-through-heel-manual-entry-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1548.7866197183098, 'std': 109.4657394693457}, annotation_units=1, channel_map={'4142_Speed-of-sound-through-heel-manual-entry-left_2_0': 0, })
ukb_4142_1 = TensorMap('4142_Speed-of-sound-through-heel-manual-entry-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1556.7346341463415, 'std': 32.773320098552155}, annotation_units=1, channel_map={'4142_Speed-of-sound-through-heel-manual-entry-left_1_0': 0, })
ukb_4143_0 = TensorMap('4143_Heel-bone-mineral-density-BMD-Tscore-manual-entry-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6763915094339622, 'std': 0.9576533150241747}, annotation_units=1, channel_map={'4143_Heel-bone-mineral-density-BMD-Tscore-manual-entry-right_0_0': 0, })
ukb_4143_2 = TensorMap('4143_Heel-bone-mineral-density-BMD-Tscore-manual-entry-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7347222222222222, 'std': 0.7301996512540462}, annotation_units=1, channel_map={'4143_Heel-bone-mineral-density-BMD-Tscore-manual-entry-right_2_0': 0, })
ukb_4143_1 = TensorMap('4143_Heel-bone-mineral-density-BMD-Tscore-manual-entry-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6608465608465608, 'std': 1.054247398571618}, annotation_units=1, channel_map={'4143_Heel-bone-mineral-density-BMD-Tscore-manual-entry-right_1_0': 0, })
ukb_4144_2 = TensorMap('4144_Heel-quantitative-ultrasound-index-QUI-manual-entry-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 101.65347222222223, 'std': 17.794754130716434}, annotation_units=1, channel_map={'4144_Heel-quantitative-ultrasound-index-QUI-manual-entry-right_2_0': 0, })
ukb_4144_0 = TensorMap('4144_Heel-quantitative-ultrasound-index-QUI-manual-entry-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 98.52704402515724, 'std': 20.42349870900061}, annotation_units=1, channel_map={'4144_Heel-quantitative-ultrasound-index-QUI-manual-entry-right_0_0': 0, })
ukb_4144_1 = TensorMap('4144_Heel-quantitative-ultrasound-index-QUI-manual-entry-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 100.59894179894181, 'std': 20.891235537385278}, annotation_units=1, channel_map={'4144_Heel-quantitative-ultrasound-index-QUI-manual-entry-right_1_0': 0, })
ukb_4145_1 = TensorMap('4145_Heel-bone-mineral-density-BMD-manual-entry-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5597830687830688, 'std': 0.13205568026606232}, annotation_units=1, channel_map={'4145_Heel-bone-mineral-density-BMD-manual-entry-right_1_0': 0, })
ukb_4145_0 = TensorMap('4145_Heel-bone-mineral-density-BMD-manual-entry-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5485754716981133, 'std': 0.12684714143138237}, annotation_units=1, channel_map={'4145_Heel-bone-mineral-density-BMD-manual-entry-right_0_0': 0, })
ukb_4145_2 = TensorMap('4145_Heel-bone-mineral-density-BMD-manual-entry-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5664374999999999, 'std': 0.11254790549569783}, annotation_units=1, channel_map={'4145_Heel-bone-mineral-density-BMD-manual-entry-right_2_0': 0, })
ukb_4146_0 = TensorMap('4146_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.37633647798742, 'std': 20.01929945554501}, annotation_units=1, channel_map={'4146_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-right_0_0': 0, })
ukb_4146_1 = TensorMap('4146_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.74391534391535, 'std': 19.870758680684204}, annotation_units=1, channel_map={'4146_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-right_1_0': 0, })
ukb_4146_2 = TensorMap('4146_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 80.64097222222222, 'std': 16.278646243582372}, annotation_units=1, channel_map={'4146_Heel-broadband-ultrasound-attenuation-BUA-manual-entry-right_2_0': 0, })
ukb_4147_0 = TensorMap('4147_Speed-of-sound-through-heel-manual-entry-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1554.7818396226412, 'std': 47.47448667160834}, annotation_units=1, channel_map={'4147_Speed-of-sound-through-heel-manual-entry-right_0_0': 0, })
ukb_4147_2 = TensorMap('4147_Speed-of-sound-through-heel-manual-entry-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1559.975, 'std': 28.441888983523032}, annotation_units=1, channel_map={'4147_Speed-of-sound-through-heel-manual-entry-right_2_0': 0, })
ukb_4147_1 = TensorMap('4147_Speed-of-sound-through-heel-manual-entry-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1558.3084656084657, 'std': 32.74356575214486}, annotation_units=1, channel_map={'4147_Speed-of-sound-through-heel-manual-entry-right_1_0': 0, })
ukb_4194_1 = TensorMap('4194_Pulse-rate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 66.64900400375664, 'std': 10.987568982350094}, annotation_units=1, channel_map={'4194_Pulse-rate_1_0': 0, })
ukb_4194_0 = TensorMap('4194_Pulse-rate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 68.99360502228286, 'std': 11.565213398676876}, annotation_units=1, channel_map={'4194_Pulse-rate_0_0': 0, })
ukb_4194_2 = TensorMap('4194_Pulse-rate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 70.80985105158855, 'std': 12.322764670818012}, annotation_units=1, channel_map={'4194_Pulse-rate_2_0': 0, })
ukb_4195_1 = TensorMap('4195_Pulse-wave-reflection-index_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 71.85087242350846, 'std': 29.40419657428217}, annotation_units=1, channel_map={'4195_Pulse-wave-reflection-index_1_0': 0, })
ukb_4195_2 = TensorMap('4195_Pulse-wave-reflection-index_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 71.78840375886978, 'std': 80.32894518966847}, annotation_units=1, channel_map={'4195_Pulse-wave-reflection-index_2_0': 0, })
ukb_4195_0 = TensorMap('4195_Pulse-wave-reflection-index_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 67.72835399598262, 'std': 32.180795506207915}, annotation_units=1, channel_map={'4195_Pulse-wave-reflection-index_0_0': 0, })
ukb_4196_0 = TensorMap('4196_Pulse-wave-peak-to-peak-time_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 201.85429179142517, 'std': 73.04525271657313}, annotation_units=1, channel_map={'4196_Pulse-wave-peak-to-peak-time_0_0': 0, })
ukb_4196_2 = TensorMap('4196_Pulse-wave-peak-to-peak-time_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 195.29893147354275, 'std': 150.8309726185657}, annotation_units=1, channel_map={'4196_Pulse-wave-peak-to-peak-time_2_0': 0, })
ukb_4196_1 = TensorMap('4196_Pulse-wave-peak-to-peak-time_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 181.82781129462958, 'std': 65.27634456016912}, annotation_units=1, channel_map={'4196_Pulse-wave-peak-to-peak-time_1_0': 0, })
ukb_4198_0 = TensorMap('4198_Position-of-the-pulse-wave-peak_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 22.28466435151295, 'std': 5.413583747596621}, annotation_units=1, channel_map={'4198_Position-of-the-pulse-wave-peak_0_0': 0, })
ukb_4198_2 = TensorMap('4198_Position-of-the-pulse-wave-peak_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.594547081761803, 'std': 4.592681428316416}, annotation_units=1, channel_map={'4198_Position-of-the-pulse-wave-peak_2_0': 0, })
ukb_4198_1 = TensorMap('4198_Position-of-the-pulse-wave-peak_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.110918886856805, 'std': 4.523892184303231}, annotation_units=1, channel_map={'4198_Position-of-the-pulse-wave-peak_1_0': 0, })
ukb_4199_2 = TensorMap('4199_Position-of-pulse-wave-notch_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.24681966374737, 'std': 6.647584572420587}, annotation_units=1, channel_map={'4199_Position-of-pulse-wave-notch_2_0': 0, })
ukb_4199_0 = TensorMap('4199_Position-of-pulse-wave-notch_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.55981236713729, 'std': 6.35301197646072}, annotation_units=1, channel_map={'4199_Position-of-pulse-wave-notch_0_0': 0, })
ukb_4199_1 = TensorMap('4199_Position-of-pulse-wave-notch_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.04947852305869, 'std': 5.882123827464215}, annotation_units=1, channel_map={'4199_Position-of-pulse-wave-notch_1_0': 0, })
ukb_4200_2 = TensorMap('4200_Position-of-the-shoulder-on-the-pulse-waveform_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 22.66266061497156, 'std': 4.9746415286765755}, annotation_units=1, channel_map={'4200_Position-of-the-shoulder-on-the-pulse-waveform_2_0': 0, })
ukb_4200_1 = TensorMap('4200_Position-of-the-shoulder-on-the-pulse-waveform_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 22.289259057881473, 'std': 4.855475559226108}, annotation_units=1, channel_map={'4200_Position-of-the-shoulder-on-the-pulse-waveform_1_0': 0, })
ukb_4200_0 = TensorMap('4200_Position-of-the-shoulder-on-the-pulse-waveform_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.145409612377684, 'std': 5.606065558195835}, annotation_units=1, channel_map={'4200_Position-of-the-shoulder-on-the-pulse-waveform_0_0': 0, })
ukb_4207_2 = TensorMap('4207_Pulse-wave-velocity-manual-entry_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 11.618529411764706, 'std': 7.05376331539199}, annotation_units=1, channel_map={'4207_Pulse-wave-velocity-manual-entry_2_0': 0, })
ukb_4207_1 = TensorMap('4207_Pulse-wave-velocity-manual-entry_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.98, 'std': 3.7924442426136378}, annotation_units=1, channel_map={'4207_Pulse-wave-velocity-manual-entry_1_0': 0, })
ukb_4207_0 = TensorMap('4207_Pulse-wave-velocity-manual-entry_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.84577639751553, 'std': 6.273794170969385}, annotation_units=1, channel_map={'4207_Pulse-wave-velocity-manual-entry_0_0': 0, })
ukb_4230_2 = TensorMap('4230_Signaltonoiseratio-SNR-of-triplet-left_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -4.057530078827272, 'std': 3.4459987654748874}, annotation_units=16, channel_map={'4230_Signaltonoiseratio-SNR-of-triplet-left_2_0': 0, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_1': 1, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_2': 2, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_3': 3, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_4': 4, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_5': 5, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_6': 6, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_7': 7, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_8': 8, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_9': 9, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_10': 10, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_11': 11, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_12': 12, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_13': 13, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_14': 14, '4230_Signaltonoiseratio-SNR-of-triplet-left_2_15': 15, })
ukb_4230_1 = TensorMap('4230_Signaltonoiseratio-SNR-of-triplet-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -5.567006368145184, 'std': 3.5961125511068013}, annotation_units=16, channel_map={'4230_Signaltonoiseratio-SNR-of-triplet-left_1_0': 0, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_1': 1, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_2': 2, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_3': 3, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_4': 4, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_5': 5, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_6': 6, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_7': 7, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_8': 8, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_9': 9, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_10': 10, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_11': 11, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_12': 12, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_13': 13, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_14': 14, '4230_Signaltonoiseratio-SNR-of-triplet-left_1_15': 15, })
ukb_4230_0 = TensorMap('4230_Signaltonoiseratio-SNR-of-triplet-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -5.540003979162151, 'std': 3.5632233749909186}, annotation_units=16, channel_map={'4230_Signaltonoiseratio-SNR-of-triplet-left_0_0': 0, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_1': 1, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_2': 2, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_3': 3, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_4': 4, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_5': 5, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_6': 6, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_7': 7, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_8': 8, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_9': 9, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_10': 10, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_11': 11, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_12': 12, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_13': 13, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_14': 14, '4230_Signaltonoiseratio-SNR-of-triplet-left_0_15': 15, })
ukb_4233_2 = TensorMap('4233_Mean-signaltonoise-ratio-SNR-left_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -46768.65862064095, 'std': 49892.459119437335}, annotation_units=16, channel_map={'4233_Mean-signaltonoise-ratio-SNR-left_2_0': 0, '4233_Mean-signaltonoise-ratio-SNR-left_2_1': 1, '4233_Mean-signaltonoise-ratio-SNR-left_2_2': 2, '4233_Mean-signaltonoise-ratio-SNR-left_2_3': 3, '4233_Mean-signaltonoise-ratio-SNR-left_2_4': 4, '4233_Mean-signaltonoise-ratio-SNR-left_2_5': 5, '4233_Mean-signaltonoise-ratio-SNR-left_2_6': 6, '4233_Mean-signaltonoise-ratio-SNR-left_2_7': 7, '4233_Mean-signaltonoise-ratio-SNR-left_2_8': 8, '4233_Mean-signaltonoise-ratio-SNR-left_2_9': 9, '4233_Mean-signaltonoise-ratio-SNR-left_2_10': 10, '4233_Mean-signaltonoise-ratio-SNR-left_2_11': 11, '4233_Mean-signaltonoise-ratio-SNR-left_2_12': 12, '4233_Mean-signaltonoise-ratio-SNR-left_2_13': 13, '4233_Mean-signaltonoise-ratio-SNR-left_2_14': 14, '4233_Mean-signaltonoise-ratio-SNR-left_2_15': 15, })
ukb_4233_0 = TensorMap('4233_Mean-signaltonoise-ratio-SNR-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -46872.081917352734, 'std': 49898.156016575645}, annotation_units=16, channel_map={'4233_Mean-signaltonoise-ratio-SNR-left_0_0': 0, '4233_Mean-signaltonoise-ratio-SNR-left_0_1': 1, '4233_Mean-signaltonoise-ratio-SNR-left_0_2': 2, '4233_Mean-signaltonoise-ratio-SNR-left_0_3': 3, '4233_Mean-signaltonoise-ratio-SNR-left_0_4': 4, '4233_Mean-signaltonoise-ratio-SNR-left_0_5': 5, '4233_Mean-signaltonoise-ratio-SNR-left_0_6': 6, '4233_Mean-signaltonoise-ratio-SNR-left_0_7': 7, '4233_Mean-signaltonoise-ratio-SNR-left_0_8': 8, '4233_Mean-signaltonoise-ratio-SNR-left_0_9': 9, '4233_Mean-signaltonoise-ratio-SNR-left_0_10': 10, '4233_Mean-signaltonoise-ratio-SNR-left_0_11': 11, '4233_Mean-signaltonoise-ratio-SNR-left_0_12': 12, '4233_Mean-signaltonoise-ratio-SNR-left_0_13': 13, '4233_Mean-signaltonoise-ratio-SNR-left_0_14': 14, '4233_Mean-signaltonoise-ratio-SNR-left_0_15': 15, })
ukb_4233_1 = TensorMap('4233_Mean-signaltonoise-ratio-SNR-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -46756.082029442616, 'std': 49890.74184658031}, annotation_units=16, channel_map={'4233_Mean-signaltonoise-ratio-SNR-left_1_0': 0, '4233_Mean-signaltonoise-ratio-SNR-left_1_1': 1, '4233_Mean-signaltonoise-ratio-SNR-left_1_2': 2, '4233_Mean-signaltonoise-ratio-SNR-left_1_3': 3, '4233_Mean-signaltonoise-ratio-SNR-left_1_4': 4, '4233_Mean-signaltonoise-ratio-SNR-left_1_5': 5, '4233_Mean-signaltonoise-ratio-SNR-left_1_6': 6, '4233_Mean-signaltonoise-ratio-SNR-left_1_7': 7, '4233_Mean-signaltonoise-ratio-SNR-left_1_8': 8, '4233_Mean-signaltonoise-ratio-SNR-left_1_9': 9, '4233_Mean-signaltonoise-ratio-SNR-left_1_10': 10, '4233_Mean-signaltonoise-ratio-SNR-left_1_11': 11, '4233_Mean-signaltonoise-ratio-SNR-left_1_12': 12, '4233_Mean-signaltonoise-ratio-SNR-left_1_13': 13, '4233_Mean-signaltonoise-ratio-SNR-left_1_14': 14, '4233_Mean-signaltonoise-ratio-SNR-left_1_15': 15, })
ukb_4234_0 = TensorMap('4234_Time-to-press-first-digit-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.86120598007727, 'std': 67.55919547209987}, annotation_units=16, channel_map={'4234_Time-to-press-first-digit-left_0_0': 0, '4234_Time-to-press-first-digit-left_0_1': 1, '4234_Time-to-press-first-digit-left_0_2': 2, '4234_Time-to-press-first-digit-left_0_3': 3, '4234_Time-to-press-first-digit-left_0_4': 4, '4234_Time-to-press-first-digit-left_0_5': 5, '4234_Time-to-press-first-digit-left_0_6': 6, '4234_Time-to-press-first-digit-left_0_7': 7, '4234_Time-to-press-first-digit-left_0_8': 8, '4234_Time-to-press-first-digit-left_0_9': 9, '4234_Time-to-press-first-digit-left_0_10': 10, '4234_Time-to-press-first-digit-left_0_11': 11, '4234_Time-to-press-first-digit-left_0_12': 12, '4234_Time-to-press-first-digit-left_0_13': 13, '4234_Time-to-press-first-digit-left_0_14': 14, '4234_Time-to-press-first-digit-left_0_15': 15, })
ukb_4234_1 = TensorMap('4234_Time-to-press-first-digit-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.553788669043627, 'std': 55.87485045902546}, annotation_units=16, channel_map={'4234_Time-to-press-first-digit-left_1_0': 0, '4234_Time-to-press-first-digit-left_1_1': 1, '4234_Time-to-press-first-digit-left_1_2': 2, '4234_Time-to-press-first-digit-left_1_3': 3, '4234_Time-to-press-first-digit-left_1_4': 4, '4234_Time-to-press-first-digit-left_1_5': 5, '4234_Time-to-press-first-digit-left_1_6': 6, '4234_Time-to-press-first-digit-left_1_7': 7, '4234_Time-to-press-first-digit-left_1_8': 8, '4234_Time-to-press-first-digit-left_1_9': 9, '4234_Time-to-press-first-digit-left_1_10': 10, '4234_Time-to-press-first-digit-left_1_11': 11, '4234_Time-to-press-first-digit-left_1_12': 12, '4234_Time-to-press-first-digit-left_1_13': 13, '4234_Time-to-press-first-digit-left_1_14': 14, '4234_Time-to-press-first-digit-left_1_15': 15, })
ukb_4234_2 = TensorMap('4234_Time-to-press-first-digit-left_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.940269111234258, 'std': 64.24147269189058}, annotation_units=16, channel_map={'4234_Time-to-press-first-digit-left_2_0': 0, '4234_Time-to-press-first-digit-left_2_1': 1, '4234_Time-to-press-first-digit-left_2_2': 2, '4234_Time-to-press-first-digit-left_2_3': 3, '4234_Time-to-press-first-digit-left_2_4': 4, '4234_Time-to-press-first-digit-left_2_5': 5, '4234_Time-to-press-first-digit-left_2_6': 6, '4234_Time-to-press-first-digit-left_2_7': 7, '4234_Time-to-press-first-digit-left_2_8': 8, '4234_Time-to-press-first-digit-left_2_9': 9, '4234_Time-to-press-first-digit-left_2_10': 10, '4234_Time-to-press-first-digit-left_2_11': 11, '4234_Time-to-press-first-digit-left_2_12': 12, '4234_Time-to-press-first-digit-left_2_13': 13, '4234_Time-to-press-first-digit-left_2_14': 14, '4234_Time-to-press-first-digit-left_2_15': 15, })
ukb_4235_1 = TensorMap('4235_Time-to-press-last-digit-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.755245036711116, 'std': 80.58161705611097}, annotation_units=16, channel_map={'4235_Time-to-press-last-digit-left_1_0': 0, '4235_Time-to-press-last-digit-left_1_1': 1, '4235_Time-to-press-last-digit-left_1_2': 2, '4235_Time-to-press-last-digit-left_1_3': 3, '4235_Time-to-press-last-digit-left_1_4': 4, '4235_Time-to-press-last-digit-left_1_5': 5, '4235_Time-to-press-last-digit-left_1_6': 6, '4235_Time-to-press-last-digit-left_1_7': 7, '4235_Time-to-press-last-digit-left_1_8': 8, '4235_Time-to-press-last-digit-left_1_9': 9, '4235_Time-to-press-last-digit-left_1_10': 10, '4235_Time-to-press-last-digit-left_1_11': 11, '4235_Time-to-press-last-digit-left_1_12': 12, '4235_Time-to-press-last-digit-left_1_13': 13, '4235_Time-to-press-last-digit-left_1_14': 14, '4235_Time-to-press-last-digit-left_1_15': 15, })
ukb_4235_0 = TensorMap('4235_Time-to-press-last-digit-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.78676793307917, 'std': 99.78521409195568}, annotation_units=16, channel_map={'4235_Time-to-press-last-digit-left_0_0': 0, '4235_Time-to-press-last-digit-left_0_1': 1, '4235_Time-to-press-last-digit-left_0_2': 2, '4235_Time-to-press-last-digit-left_0_3': 3, '4235_Time-to-press-last-digit-left_0_4': 4, '4235_Time-to-press-last-digit-left_0_5': 5, '4235_Time-to-press-last-digit-left_0_6': 6, '4235_Time-to-press-last-digit-left_0_7': 7, '4235_Time-to-press-last-digit-left_0_8': 8, '4235_Time-to-press-last-digit-left_0_9': 9, '4235_Time-to-press-last-digit-left_0_10': 10, '4235_Time-to-press-last-digit-left_0_11': 11, '4235_Time-to-press-last-digit-left_0_12': 12, '4235_Time-to-press-last-digit-left_0_13': 13, '4235_Time-to-press-last-digit-left_0_14': 14, '4235_Time-to-press-last-digit-left_0_15': 15, })
ukb_4235_2 = TensorMap('4235_Time-to-press-last-digit-left_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.48053942708468, 'std': 79.81213796118806}, annotation_units=16, channel_map={'4235_Time-to-press-last-digit-left_2_0': 0, '4235_Time-to-press-last-digit-left_2_1': 1, '4235_Time-to-press-last-digit-left_2_2': 2, '4235_Time-to-press-last-digit-left_2_3': 3, '4235_Time-to-press-last-digit-left_2_4': 4, '4235_Time-to-press-last-digit-left_2_5': 5, '4235_Time-to-press-last-digit-left_2_6': 6, '4235_Time-to-press-last-digit-left_2_7': 7, '4235_Time-to-press-last-digit-left_2_8': 8, '4235_Time-to-press-last-digit-left_2_9': 9, '4235_Time-to-press-last-digit-left_2_10': 10, '4235_Time-to-press-last-digit-left_2_11': 11, '4235_Time-to-press-last-digit-left_2_12': 12, '4235_Time-to-press-last-digit-left_2_13': 13, '4235_Time-to-press-last-digit-left_2_14': 14, '4235_Time-to-press-last-digit-left_2_15': 15, })
ukb_4237_0 = TensorMap('4237_Time-to-press-next-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.17278591817637, 'std': 103.73516223231547}, annotation_units=16, channel_map={'4237_Time-to-press-next-left_0_0': 0, '4237_Time-to-press-next-left_0_1': 1, '4237_Time-to-press-next-left_0_2': 2, '4237_Time-to-press-next-left_0_3': 3, '4237_Time-to-press-next-left_0_4': 4, '4237_Time-to-press-next-left_0_5': 5, '4237_Time-to-press-next-left_0_6': 6, '4237_Time-to-press-next-left_0_7': 7, '4237_Time-to-press-next-left_0_8': 8, '4237_Time-to-press-next-left_0_9': 9, '4237_Time-to-press-next-left_0_10': 10, '4237_Time-to-press-next-left_0_11': 11, '4237_Time-to-press-next-left_0_12': 12, '4237_Time-to-press-next-left_0_13': 13, '4237_Time-to-press-next-left_0_14': 14, '4237_Time-to-press-next-left_0_15': 15, })
ukb_4237_2 = TensorMap('4237_Time-to-press-next-left_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 48.466538776591854, 'std': 81.83454047965638}, annotation_units=16, channel_map={'4237_Time-to-press-next-left_2_0': 0, '4237_Time-to-press-next-left_2_1': 1, '4237_Time-to-press-next-left_2_2': 2, '4237_Time-to-press-next-left_2_3': 3, '4237_Time-to-press-next-left_2_4': 4, '4237_Time-to-press-next-left_2_5': 5, '4237_Time-to-press-next-left_2_6': 6, '4237_Time-to-press-next-left_2_7': 7, '4237_Time-to-press-next-left_2_8': 8, '4237_Time-to-press-next-left_2_9': 9, '4237_Time-to-press-next-left_2_10': 10, '4237_Time-to-press-next-left_2_11': 11, '4237_Time-to-press-next-left_2_12': 12, '4237_Time-to-press-next-left_2_13': 13, '4237_Time-to-press-next-left_2_14': 14, '4237_Time-to-press-next-left_2_15': 15, })
ukb_4237_1 = TensorMap('4237_Time-to-press-next-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 51.92030898339365, 'std': 82.89267421446205}, annotation_units=16, channel_map={'4237_Time-to-press-next-left_1_0': 0, '4237_Time-to-press-next-left_1_1': 1, '4237_Time-to-press-next-left_1_2': 2, '4237_Time-to-press-next-left_1_3': 3, '4237_Time-to-press-next-left_1_4': 4, '4237_Time-to-press-next-left_1_5': 5, '4237_Time-to-press-next-left_1_6': 6, '4237_Time-to-press-next-left_1_7': 7, '4237_Time-to-press-next-left_1_8': 8, '4237_Time-to-press-next-left_1_9': 9, '4237_Time-to-press-next-left_1_10': 10, '4237_Time-to-press-next-left_1_11': 11, '4237_Time-to-press-next-left_1_12': 12, '4237_Time-to-press-next-left_1_13': 13, '4237_Time-to-press-next-left_1_14': 14, '4237_Time-to-press-next-left_1_15': 15, })
ukb_4239_1 = TensorMap('4239_Number-of-times-clear-was-pressed-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07538488881734932, 'std': 0.4292264470174809}, annotation_units=16, channel_map={'4239_Number-of-times-clear-was-pressed-left_1_0': 0, '4239_Number-of-times-clear-was-pressed-left_1_1': 1, '4239_Number-of-times-clear-was-pressed-left_1_2': 2, '4239_Number-of-times-clear-was-pressed-left_1_3': 3, '4239_Number-of-times-clear-was-pressed-left_1_4': 4, '4239_Number-of-times-clear-was-pressed-left_1_5': 5, '4239_Number-of-times-clear-was-pressed-left_1_6': 6, '4239_Number-of-times-clear-was-pressed-left_1_7': 7, '4239_Number-of-times-clear-was-pressed-left_1_8': 8, '4239_Number-of-times-clear-was-pressed-left_1_9': 9, '4239_Number-of-times-clear-was-pressed-left_1_10': 10, '4239_Number-of-times-clear-was-pressed-left_1_11': 11, '4239_Number-of-times-clear-was-pressed-left_1_12': 12, '4239_Number-of-times-clear-was-pressed-left_1_13': 13, '4239_Number-of-times-clear-was-pressed-left_1_14': 14, '4239_Number-of-times-clear-was-pressed-left_1_15': 15, })
ukb_4239_2 = TensorMap('4239_Number-of-times-clear-was-pressed-left_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09715115475038014, 'std': 0.4177337632796932}, annotation_units=16, channel_map={'4239_Number-of-times-clear-was-pressed-left_2_0': 0, '4239_Number-of-times-clear-was-pressed-left_2_1': 1, '4239_Number-of-times-clear-was-pressed-left_2_2': 2, '4239_Number-of-times-clear-was-pressed-left_2_3': 3, '4239_Number-of-times-clear-was-pressed-left_2_4': 4, '4239_Number-of-times-clear-was-pressed-left_2_5': 5, '4239_Number-of-times-clear-was-pressed-left_2_6': 6, '4239_Number-of-times-clear-was-pressed-left_2_7': 7, '4239_Number-of-times-clear-was-pressed-left_2_8': 8, '4239_Number-of-times-clear-was-pressed-left_2_9': 9, '4239_Number-of-times-clear-was-pressed-left_2_10': 10, '4239_Number-of-times-clear-was-pressed-left_2_11': 11, '4239_Number-of-times-clear-was-pressed-left_2_12': 12, '4239_Number-of-times-clear-was-pressed-left_2_13': 13, '4239_Number-of-times-clear-was-pressed-left_2_14': 14, '4239_Number-of-times-clear-was-pressed-left_2_15': 15, })
ukb_4239_0 = TensorMap('4239_Number-of-times-clear-was-pressed-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08774661669705094, 'std': 0.49309733708179393}, annotation_units=16, channel_map={'4239_Number-of-times-clear-was-pressed-left_0_0': 0, '4239_Number-of-times-clear-was-pressed-left_0_1': 1, '4239_Number-of-times-clear-was-pressed-left_0_2': 2, '4239_Number-of-times-clear-was-pressed-left_0_3': 3, '4239_Number-of-times-clear-was-pressed-left_0_4': 4, '4239_Number-of-times-clear-was-pressed-left_0_5': 5, '4239_Number-of-times-clear-was-pressed-left_0_6': 6, '4239_Number-of-times-clear-was-pressed-left_0_7': 7, '4239_Number-of-times-clear-was-pressed-left_0_8': 8, '4239_Number-of-times-clear-was-pressed-left_0_9': 9, '4239_Number-of-times-clear-was-pressed-left_0_10': 10, '4239_Number-of-times-clear-was-pressed-left_0_11': 11, '4239_Number-of-times-clear-was-pressed-left_0_12': 12, '4239_Number-of-times-clear-was-pressed-left_0_13': 13, '4239_Number-of-times-clear-was-pressed-left_0_14': 14, '4239_Number-of-times-clear-was-pressed-left_0_15': 15, })
ukb_4241_2 = TensorMap('4241_Signaltonoiseratio-SNR-of-triplet-right_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -4.066626254381335, 'std': 3.4293067311207555}, annotation_units=16, channel_map={'4241_Signaltonoiseratio-SNR-of-triplet-right_2_0': 0, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_1': 1, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_2': 2, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_3': 3, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_4': 4, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_5': 5, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_6': 6, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_7': 7, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_8': 8, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_9': 9, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_10': 10, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_11': 11, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_12': 12, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_13': 13, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_14': 14, '4241_Signaltonoiseratio-SNR-of-triplet-right_2_15': 15, })
ukb_4241_1 = TensorMap('4241_Signaltonoiseratio-SNR-of-triplet-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -5.5532600207134415, 'std': 3.623575182088199}, annotation_units=16, channel_map={'4241_Signaltonoiseratio-SNR-of-triplet-right_1_0': 0, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_1': 1, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_2': 2, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_3': 3, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_4': 4, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_5': 5, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_6': 6, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_7': 7, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_8': 8, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_9': 9, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_10': 10, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_11': 11, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_12': 12, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_13': 13, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_14': 14, '4241_Signaltonoiseratio-SNR-of-triplet-right_1_15': 15, })
ukb_4241_0 = TensorMap('4241_Signaltonoiseratio-SNR-of-triplet-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -5.49925056354944, 'std': 3.5660704758105073}, annotation_units=16, channel_map={'4241_Signaltonoiseratio-SNR-of-triplet-right_0_0': 0, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_1': 1, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_2': 2, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_3': 3, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_4': 4, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_5': 5, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_6': 6, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_7': 7, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_8': 8, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_9': 9, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_10': 10, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_11': 11, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_12': 12, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_13': 13, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_14': 14, '4241_Signaltonoiseratio-SNR-of-triplet-right_0_15': 15, })
ukb_4244_1 = TensorMap('4244_Mean-signaltonoise-ratio-SNR-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -46755.43225730901, 'std': 49890.70074665193}, annotation_units=16, channel_map={'4244_Mean-signaltonoise-ratio-SNR-right_1_0': 0, '4244_Mean-signaltonoise-ratio-SNR-right_1_1': 1, '4244_Mean-signaltonoise-ratio-SNR-right_1_2': 2, '4244_Mean-signaltonoise-ratio-SNR-right_1_3': 3, '4244_Mean-signaltonoise-ratio-SNR-right_1_4': 4, '4244_Mean-signaltonoise-ratio-SNR-right_1_5': 5, '4244_Mean-signaltonoise-ratio-SNR-right_1_6': 6, '4244_Mean-signaltonoise-ratio-SNR-right_1_7': 7, '4244_Mean-signaltonoise-ratio-SNR-right_1_8': 8, '4244_Mean-signaltonoise-ratio-SNR-right_1_9': 9, '4244_Mean-signaltonoise-ratio-SNR-right_1_10': 10, '4244_Mean-signaltonoise-ratio-SNR-right_1_11': 11, '4244_Mean-signaltonoise-ratio-SNR-right_1_12': 12, '4244_Mean-signaltonoise-ratio-SNR-right_1_13': 13, '4244_Mean-signaltonoise-ratio-SNR-right_1_14': 14, '4244_Mean-signaltonoise-ratio-SNR-right_1_15': 15, })
ukb_4244_2 = TensorMap('4244_Mean-signaltonoise-ratio-SNR-right_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -46758.21567698974, 'std': 49891.77675188698}, annotation_units=16, channel_map={'4244_Mean-signaltonoise-ratio-SNR-right_2_0': 0, '4244_Mean-signaltonoise-ratio-SNR-right_2_1': 1, '4244_Mean-signaltonoise-ratio-SNR-right_2_2': 2, '4244_Mean-signaltonoise-ratio-SNR-right_2_3': 3, '4244_Mean-signaltonoise-ratio-SNR-right_2_4': 4, '4244_Mean-signaltonoise-ratio-SNR-right_2_5': 5, '4244_Mean-signaltonoise-ratio-SNR-right_2_6': 6, '4244_Mean-signaltonoise-ratio-SNR-right_2_7': 7, '4244_Mean-signaltonoise-ratio-SNR-right_2_8': 8, '4244_Mean-signaltonoise-ratio-SNR-right_2_9': 9, '4244_Mean-signaltonoise-ratio-SNR-right_2_10': 10, '4244_Mean-signaltonoise-ratio-SNR-right_2_11': 11, '4244_Mean-signaltonoise-ratio-SNR-right_2_12': 12, '4244_Mean-signaltonoise-ratio-SNR-right_2_13': 13, '4244_Mean-signaltonoise-ratio-SNR-right_2_14': 14, '4244_Mean-signaltonoise-ratio-SNR-right_2_15': 15, })
ukb_4244_0 = TensorMap('4244_Mean-signaltonoise-ratio-SNR-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': -46879.64721595625, 'std': 49898.65251149256}, annotation_units=16, channel_map={'4244_Mean-signaltonoise-ratio-SNR-right_0_0': 0, '4244_Mean-signaltonoise-ratio-SNR-right_0_1': 1, '4244_Mean-signaltonoise-ratio-SNR-right_0_2': 2, '4244_Mean-signaltonoise-ratio-SNR-right_0_3': 3, '4244_Mean-signaltonoise-ratio-SNR-right_0_4': 4, '4244_Mean-signaltonoise-ratio-SNR-right_0_5': 5, '4244_Mean-signaltonoise-ratio-SNR-right_0_6': 6, '4244_Mean-signaltonoise-ratio-SNR-right_0_7': 7, '4244_Mean-signaltonoise-ratio-SNR-right_0_8': 8, '4244_Mean-signaltonoise-ratio-SNR-right_0_9': 9, '4244_Mean-signaltonoise-ratio-SNR-right_0_10': 10, '4244_Mean-signaltonoise-ratio-SNR-right_0_11': 11, '4244_Mean-signaltonoise-ratio-SNR-right_0_12': 12, '4244_Mean-signaltonoise-ratio-SNR-right_0_13': 13, '4244_Mean-signaltonoise-ratio-SNR-right_0_14': 14, '4244_Mean-signaltonoise-ratio-SNR-right_0_15': 15, })
ukb_4245_2 = TensorMap('4245_Time-to-press-first-digit-right_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.821255733703687, 'std': 41.54314558371766}, annotation_units=16, channel_map={'4245_Time-to-press-first-digit-right_2_0': 0, '4245_Time-to-press-first-digit-right_2_1': 1, '4245_Time-to-press-first-digit-right_2_2': 2, '4245_Time-to-press-first-digit-right_2_3': 3, '4245_Time-to-press-first-digit-right_2_4': 4, '4245_Time-to-press-first-digit-right_2_5': 5, '4245_Time-to-press-first-digit-right_2_6': 6, '4245_Time-to-press-first-digit-right_2_7': 7, '4245_Time-to-press-first-digit-right_2_8': 8, '4245_Time-to-press-first-digit-right_2_9': 9, '4245_Time-to-press-first-digit-right_2_10': 10, '4245_Time-to-press-first-digit-right_2_11': 11, '4245_Time-to-press-first-digit-right_2_12': 12, '4245_Time-to-press-first-digit-right_2_13': 13, '4245_Time-to-press-first-digit-right_2_14': 14, '4245_Time-to-press-first-digit-right_2_15': 15, })
ukb_4245_0 = TensorMap('4245_Time-to-press-first-digit-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.944958341987178, 'std': 70.05160077672136}, annotation_units=16, channel_map={'4245_Time-to-press-first-digit-right_0_0': 0, '4245_Time-to-press-first-digit-right_0_1': 1, '4245_Time-to-press-first-digit-right_0_2': 2, '4245_Time-to-press-first-digit-right_0_3': 3, '4245_Time-to-press-first-digit-right_0_4': 4, '4245_Time-to-press-first-digit-right_0_5': 5, '4245_Time-to-press-first-digit-right_0_6': 6, '4245_Time-to-press-first-digit-right_0_7': 7, '4245_Time-to-press-first-digit-right_0_8': 8, '4245_Time-to-press-first-digit-right_0_9': 9, '4245_Time-to-press-first-digit-right_0_10': 10, '4245_Time-to-press-first-digit-right_0_11': 11, '4245_Time-to-press-first-digit-right_0_12': 12, '4245_Time-to-press-first-digit-right_0_13': 13, '4245_Time-to-press-first-digit-right_0_14': 14, '4245_Time-to-press-first-digit-right_0_15': 15, })
ukb_4245_1 = TensorMap('4245_Time-to-press-first-digit-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.61442531721956, 'std': 54.2815034724928}, annotation_units=16, channel_map={'4245_Time-to-press-first-digit-right_1_0': 0, '4245_Time-to-press-first-digit-right_1_1': 1, '4245_Time-to-press-first-digit-right_1_2': 2, '4245_Time-to-press-first-digit-right_1_3': 3, '4245_Time-to-press-first-digit-right_1_4': 4, '4245_Time-to-press-first-digit-right_1_5': 5, '4245_Time-to-press-first-digit-right_1_6': 6, '4245_Time-to-press-first-digit-right_1_7': 7, '4245_Time-to-press-first-digit-right_1_8': 8, '4245_Time-to-press-first-digit-right_1_9': 9, '4245_Time-to-press-first-digit-right_1_10': 10, '4245_Time-to-press-first-digit-right_1_11': 11, '4245_Time-to-press-first-digit-right_1_12': 12, '4245_Time-to-press-first-digit-right_1_13': 13, '4245_Time-to-press-first-digit-right_1_14': 14, '4245_Time-to-press-first-digit-right_1_15': 15, })
ukb_4246_2 = TensorMap('4246_Time-to-press-last-digit-right_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.12209557475308, 'std': 61.514140278161555}, annotation_units=16, channel_map={'4246_Time-to-press-last-digit-right_2_0': 0, '4246_Time-to-press-last-digit-right_2_1': 1, '4246_Time-to-press-last-digit-right_2_2': 2, '4246_Time-to-press-last-digit-right_2_3': 3, '4246_Time-to-press-last-digit-right_2_4': 4, '4246_Time-to-press-last-digit-right_2_5': 5, '4246_Time-to-press-last-digit-right_2_6': 6, '4246_Time-to-press-last-digit-right_2_7': 7, '4246_Time-to-press-last-digit-right_2_8': 8, '4246_Time-to-press-last-digit-right_2_9': 9, '4246_Time-to-press-last-digit-right_2_10': 10, '4246_Time-to-press-last-digit-right_2_11': 11, '4246_Time-to-press-last-digit-right_2_12': 12, '4246_Time-to-press-last-digit-right_2_13': 13, '4246_Time-to-press-last-digit-right_2_14': 14, '4246_Time-to-press-last-digit-right_2_15': 15, })
ukb_4246_1 = TensorMap('4246_Time-to-press-last-digit-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.704222330243475, 'std': 80.13609007128268}, annotation_units=16, channel_map={'4246_Time-to-press-last-digit-right_1_0': 0, '4246_Time-to-press-last-digit-right_1_1': 1, '4246_Time-to-press-last-digit-right_1_2': 2, '4246_Time-to-press-last-digit-right_1_3': 3, '4246_Time-to-press-last-digit-right_1_4': 4, '4246_Time-to-press-last-digit-right_1_5': 5, '4246_Time-to-press-last-digit-right_1_6': 6, '4246_Time-to-press-last-digit-right_1_7': 7, '4246_Time-to-press-last-digit-right_1_8': 8, '4246_Time-to-press-last-digit-right_1_9': 9, '4246_Time-to-press-last-digit-right_1_10': 10, '4246_Time-to-press-last-digit-right_1_11': 11, '4246_Time-to-press-last-digit-right_1_12': 12, '4246_Time-to-press-last-digit-right_1_13': 13, '4246_Time-to-press-last-digit-right_1_14': 14, '4246_Time-to-press-last-digit-right_1_15': 15, })
ukb_4246_0 = TensorMap('4246_Time-to-press-last-digit-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.800946678800166, 'std': 101.12143365743562}, annotation_units=16, channel_map={'4246_Time-to-press-last-digit-right_0_0': 0, '4246_Time-to-press-last-digit-right_0_1': 1, '4246_Time-to-press-last-digit-right_0_2': 2, '4246_Time-to-press-last-digit-right_0_3': 3, '4246_Time-to-press-last-digit-right_0_4': 4, '4246_Time-to-press-last-digit-right_0_5': 5, '4246_Time-to-press-last-digit-right_0_6': 6, '4246_Time-to-press-last-digit-right_0_7': 7, '4246_Time-to-press-last-digit-right_0_8': 8, '4246_Time-to-press-last-digit-right_0_9': 9, '4246_Time-to-press-last-digit-right_0_10': 10, '4246_Time-to-press-last-digit-right_0_11': 11, '4246_Time-to-press-last-digit-right_0_12': 12, '4246_Time-to-press-last-digit-right_0_13': 13, '4246_Time-to-press-last-digit-right_0_14': 14, '4246_Time-to-press-last-digit-right_0_15': 15, })
ukb_4247_0 = TensorMap('4247_Time-to-press-next-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.08769581745342, 'std': 104.03369540215628}, annotation_units=16, channel_map={'4247_Time-to-press-next-right_0_0': 0, '4247_Time-to-press-next-right_0_1': 1, '4247_Time-to-press-next-right_0_2': 2, '4247_Time-to-press-next-right_0_3': 3, '4247_Time-to-press-next-right_0_4': 4, '4247_Time-to-press-next-right_0_5': 5, '4247_Time-to-press-next-right_0_6': 6, '4247_Time-to-press-next-right_0_7': 7, '4247_Time-to-press-next-right_0_8': 8, '4247_Time-to-press-next-right_0_9': 9, '4247_Time-to-press-next-right_0_10': 10, '4247_Time-to-press-next-right_0_11': 11, '4247_Time-to-press-next-right_0_12': 12, '4247_Time-to-press-next-right_0_13': 13, '4247_Time-to-press-next-right_0_14': 14, '4247_Time-to-press-next-right_0_15': 15, })
ukb_4247_2 = TensorMap('4247_Time-to-press-next-right_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 48.017211631528006, 'std': 63.80431922701709}, annotation_units=16, channel_map={'4247_Time-to-press-next-right_2_0': 0, '4247_Time-to-press-next-right_2_1': 1, '4247_Time-to-press-next-right_2_2': 2, '4247_Time-to-press-next-right_2_3': 3, '4247_Time-to-press-next-right_2_4': 4, '4247_Time-to-press-next-right_2_5': 5, '4247_Time-to-press-next-right_2_6': 6, '4247_Time-to-press-next-right_2_7': 7, '4247_Time-to-press-next-right_2_8': 8, '4247_Time-to-press-next-right_2_9': 9, '4247_Time-to-press-next-right_2_10': 10, '4247_Time-to-press-next-right_2_11': 11, '4247_Time-to-press-next-right_2_12': 12, '4247_Time-to-press-next-right_2_13': 13, '4247_Time-to-press-next-right_2_14': 14, '4247_Time-to-press-next-right_2_15': 15, })
ukb_4247_1 = TensorMap('4247_Time-to-press-next-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 51.81792520826701, 'std': 82.8842536130109}, annotation_units=16, channel_map={'4247_Time-to-press-next-right_1_0': 0, '4247_Time-to-press-next-right_1_1': 1, '4247_Time-to-press-next-right_1_2': 2, '4247_Time-to-press-next-right_1_3': 3, '4247_Time-to-press-next-right_1_4': 4, '4247_Time-to-press-next-right_1_5': 5, '4247_Time-to-press-next-right_1_6': 6, '4247_Time-to-press-next-right_1_7': 7, '4247_Time-to-press-next-right_1_8': 8, '4247_Time-to-press-next-right_1_9': 9, '4247_Time-to-press-next-right_1_10': 10, '4247_Time-to-press-next-right_1_11': 11, '4247_Time-to-press-next-right_1_12': 12, '4247_Time-to-press-next-right_1_13': 13, '4247_Time-to-press-next-right_1_14': 14, '4247_Time-to-press-next-right_1_15': 15, })
ukb_4249_1 = TensorMap('4249_Number-of-times-clear-was-pressed-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07285902459134534, 'std': 0.43378002862275217}, annotation_units=16, channel_map={'4249_Number-of-times-clear-was-pressed-right_1_0': 0, '4249_Number-of-times-clear-was-pressed-right_1_1': 1, '4249_Number-of-times-clear-was-pressed-right_1_2': 2, '4249_Number-of-times-clear-was-pressed-right_1_3': 3, '4249_Number-of-times-clear-was-pressed-right_1_4': 4, '4249_Number-of-times-clear-was-pressed-right_1_5': 5, '4249_Number-of-times-clear-was-pressed-right_1_6': 6, '4249_Number-of-times-clear-was-pressed-right_1_7': 7, '4249_Number-of-times-clear-was-pressed-right_1_8': 8, '4249_Number-of-times-clear-was-pressed-right_1_9': 9, '4249_Number-of-times-clear-was-pressed-right_1_10': 10, '4249_Number-of-times-clear-was-pressed-right_1_11': 11, '4249_Number-of-times-clear-was-pressed-right_1_12': 12, '4249_Number-of-times-clear-was-pressed-right_1_13': 13, '4249_Number-of-times-clear-was-pressed-right_1_14': 14, '4249_Number-of-times-clear-was-pressed-right_1_15': 15, })
ukb_4249_0 = TensorMap('4249_Number-of-times-clear-was-pressed-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08729198620760759, 'std': 0.49111663097303243}, annotation_units=16, channel_map={'4249_Number-of-times-clear-was-pressed-right_0_0': 0, '4249_Number-of-times-clear-was-pressed-right_0_1': 1, '4249_Number-of-times-clear-was-pressed-right_0_2': 2, '4249_Number-of-times-clear-was-pressed-right_0_3': 3, '4249_Number-of-times-clear-was-pressed-right_0_4': 4, '4249_Number-of-times-clear-was-pressed-right_0_5': 5, '4249_Number-of-times-clear-was-pressed-right_0_6': 6, '4249_Number-of-times-clear-was-pressed-right_0_7': 7, '4249_Number-of-times-clear-was-pressed-right_0_8': 8, '4249_Number-of-times-clear-was-pressed-right_0_9': 9, '4249_Number-of-times-clear-was-pressed-right_0_10': 10, '4249_Number-of-times-clear-was-pressed-right_0_11': 11, '4249_Number-of-times-clear-was-pressed-right_0_12': 12, '4249_Number-of-times-clear-was-pressed-right_0_13': 13, '4249_Number-of-times-clear-was-pressed-right_0_14': 14, '4249_Number-of-times-clear-was-pressed-right_0_15': 15, })
ukb_4249_2 = TensorMap('4249_Number-of-times-clear-was-pressed-right_2_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09451373217458124, 'std': 0.3857519319630944}, annotation_units=16, channel_map={'4249_Number-of-times-clear-was-pressed-right_2_0': 0, '4249_Number-of-times-clear-was-pressed-right_2_1': 1, '4249_Number-of-times-clear-was-pressed-right_2_2': 2, '4249_Number-of-times-clear-was-pressed-right_2_3': 3, '4249_Number-of-times-clear-was-pressed-right_2_4': 4, '4249_Number-of-times-clear-was-pressed-right_2_5': 5, '4249_Number-of-times-clear-was-pressed-right_2_6': 6, '4249_Number-of-times-clear-was-pressed-right_2_7': 7, '4249_Number-of-times-clear-was-pressed-right_2_8': 8, '4249_Number-of-times-clear-was-pressed-right_2_9': 9, '4249_Number-of-times-clear-was-pressed-right_2_10': 10, '4249_Number-of-times-clear-was-pressed-right_2_11': 11, '4249_Number-of-times-clear-was-pressed-right_2_12': 12, '4249_Number-of-times-clear-was-pressed-right_2_13': 13, '4249_Number-of-times-clear-was-pressed-right_2_14': 14, '4249_Number-of-times-clear-was-pressed-right_2_15': 15, })
ukb_4269_2 = TensorMap('4269_Number-of-triplets-attempted-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.451946472019474, 'std': 2.699255638802792}, annotation_units=1, channel_map={'4269_Number-of-triplets-attempted-left_2_0': 0, })
ukb_4269_1 = TensorMap('4269_Number-of-triplets-attempted-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.871484141232795, 'std': 1.2724629145214108}, annotation_units=1, channel_map={'4269_Number-of-triplets-attempted-left_1_0': 0, })
ukb_4269_0 = TensorMap('4269_Number-of-triplets-attempted-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.663892064399361, 'std': 2.0772264814190566}, annotation_units=1, channel_map={'4269_Number-of-triplets-attempted-left_0_0': 0, })
ukb_4272_1 = TensorMap('4272_Duration-of-hearing-test-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1667.4916716537011, 'std': 512.7172409905635}, annotation_units=1, channel_map={'4272_Duration-of-hearing-test-left_1_0': 0, })
ukb_4272_2 = TensorMap('4272_Duration-of-hearing-test-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1573.0773375043445, 'std': 543.9639176486238}, annotation_units=1, channel_map={'4272_Duration-of-hearing-test-left_2_0': 0, })
ukb_4272_0 = TensorMap('4272_Duration-of-hearing-test-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1723.7206106506612, 'std': 601.5799891774069}, annotation_units=1, channel_map={'4272_Duration-of-hearing-test-left_0_0': 0, })
ukb_4276_1 = TensorMap('4276_Number-of-triplets-attempted-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.86952404684774, 'std': 1.2854311461108745}, annotation_units=1, channel_map={'4276_Number-of-triplets-attempted-right_1_0': 0, })
ukb_4276_2 = TensorMap('4276_Number-of-triplets-attempted-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.467589979594495, 'std': 2.6682043941793636}, annotation_units=1, channel_map={'4276_Number-of-triplets-attempted-right_2_0': 0, })
ukb_4276_0 = TensorMap('4276_Number-of-triplets-attempted-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.652431840630257, 'std': 2.1115365192707105}, annotation_units=1, channel_map={'4276_Number-of-triplets-attempted-right_0_0': 0, })
ukb_4279_2 = TensorMap('4279_Duration-of-hearing-test-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1562.7181001172228, 'std': 498.18505656316495}, annotation_units=1, channel_map={'4279_Duration-of-hearing-test-right_2_0': 0, })
ukb_4279_0 = TensorMap('4279_Duration-of-hearing-test-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1719.6134819292888, 'std': 611.073307616882}, annotation_units=1, channel_map={'4279_Duration-of-hearing-test-right_0_0': 0, })
ukb_4279_1 = TensorMap('4279_Duration-of-hearing-test-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1665.0501370545726, 'std': 524.8591813670189}, annotation_units=1, channel_map={'4279_Duration-of-hearing-test-right_1_0': 0, })
ukb_4288_0 = TensorMap('4288_Time-to-answer_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 462.3668914626113, 'std': 196.6502921727969}, annotation_units=1, channel_map={'4288_Time-to-answer_0_0': 0, })
ukb_4288_2 = TensorMap('4288_Time-to-answer_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1288.9954721081867, 'std': 861.2957459784974}, annotation_units=1, channel_map={'4288_Time-to-answer_2_0': 0, })
ukb_4288_1 = TensorMap('4288_Time-to-answer_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 389.62895927601807, 'std': 58.538996152743906}, annotation_units=1, channel_map={'4288_Time-to-answer_1_0': 0, })
ukb_4290_2 = TensorMap('4290_Duration-screen-displayed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 96.87798840859698, 'std': 156.20940856069802}, annotation_units=1, channel_map={'4290_Duration-screen-displayed_2_0': 0, })
ukb_4290_1 = TensorMap('4290_Duration-screen-displayed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 100.52857564430452, 'std': 181.808713115601}, annotation_units=1, channel_map={'4290_Duration-screen-displayed_1_0': 0, })
ukb_4290_0 = TensorMap('4290_Duration-screen-displayed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 145.98503756637507, 'std': 343.02637798290317}, annotation_units=1, channel_map={'4290_Duration-screen-displayed_0_0': 0, })
ukb_4291_0 = TensorMap('4291_Number-of-attempts_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.4689705822350978, 'std': 1.408918886739235}, annotation_units=1, channel_map={'4291_Number-of-attempts_0_0': 0, })
ukb_4291_2 = TensorMap('4291_Number-of-attempts_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2927131127746931, 'std': 1.1270937114604693}, annotation_units=1, channel_map={'4291_Number-of-attempts_2_0': 0, })
ukb_4291_1 = TensorMap('4291_Number-of-attempts_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2770153952092862, 'std': 1.116478752784683}, annotation_units=1, channel_map={'4291_Number-of-attempts_1_0': 0, })
ukb_4407_1 = TensorMap('4407_Average-monthly-red-wine-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0268441236183172, 'std': 1.7964946281152334}, annotation_units=1, channel_map={'4407_Average-monthly-red-wine-intake_1_0': 0, })
ukb_4407_0 = TensorMap('4407_Average-monthly-red-wine-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9920382955503746, 'std': 1.8401968606027042}, annotation_units=1, channel_map={'4407_Average-monthly-red-wine-intake_0_0': 0, })
ukb_4407_2 = TensorMap('4407_Average-monthly-red-wine-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9679018789144052, 'std': 1.7139152247566793}, annotation_units=1, channel_map={'4407_Average-monthly-red-wine-intake_2_0': 0, })
ukb_4418_2 = TensorMap('4418_Average-monthly-champagne-plus-white-wine-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9707838789617841, 'std': 1.7440389169739223}, annotation_units=1, channel_map={'4418_Average-monthly-champagne-plus-white-wine-intake_2_0': 0, })
ukb_4418_0 = TensorMap('4418_Average-monthly-champagne-plus-white-wine-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9912825352672361, 'std': 1.6260730175053386}, annotation_units=1, channel_map={'4418_Average-monthly-champagne-plus-white-wine-intake_0_0': 0, })
ukb_4418_1 = TensorMap('4418_Average-monthly-champagne-plus-white-wine-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9166103222898354, 'std': 1.4835395172962027}, annotation_units=1, channel_map={'4418_Average-monthly-champagne-plus-white-wine-intake_1_0': 0, })
ukb_4429_0 = TensorMap('4429_Average-monthly-beer-plus-cider-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0122382376937722, 'std': 2.3720428127631483}, annotation_units=1, channel_map={'4429_Average-monthly-beer-plus-cider-intake_0_0': 0, })
ukb_4429_2 = TensorMap('4429_Average-monthly-beer-plus-cider-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0350260416666668, 'std': 2.181370471002315}, annotation_units=1, channel_map={'4429_Average-monthly-beer-plus-cider-intake_2_0': 0, })
ukb_4429_1 = TensorMap('4429_Average-monthly-beer-plus-cider-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9340832395950508, 'std': 2.059624636416694}, annotation_units=1, channel_map={'4429_Average-monthly-beer-plus-cider-intake_1_0': 0, })
ukb_4440_2 = TensorMap('4440_Average-monthly-spirits-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7863938485598854, 'std': 1.8475845742038393}, annotation_units=1, channel_map={'4440_Average-monthly-spirits-intake_2_0': 0, })
ukb_4440_1 = TensorMap('4440_Average-monthly-spirits-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6053285165951685, 'std': 1.6156217285568688}, annotation_units=1, channel_map={'4440_Average-monthly-spirits-intake_1_0': 0, })
ukb_4440_0 = TensorMap('4440_Average-monthly-spirits-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7451297087764637, 'std': 2.229589789958396}, annotation_units=1, channel_map={'4440_Average-monthly-spirits-intake_0_0': 0, })
ukb_4451_0 = TensorMap('4451_Average-monthly-fortified-wine-intake_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.16545126443032654, 'std': 0.8525999041972709}, annotation_units=1, channel_map={'4451_Average-monthly-fortified-wine-intake_0_0': 0, })
ukb_4451_2 = TensorMap('4451_Average-monthly-fortified-wine-intake_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1399219765929779, 'std': 0.72660493698432}, annotation_units=1, channel_map={'4451_Average-monthly-fortified-wine-intake_2_0': 0, })
ukb_4451_1 = TensorMap('4451_Average-monthly-fortified-wine-intake_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.17742298178547333, 'std': 0.7990960462213873}, annotation_units=1, channel_map={'4451_Average-monthly-fortified-wine-intake_1_0': 0, })
ukb_4462_1 = TensorMap('4462_Average-monthly-intake-of-other-alcoholic-drinks_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.051437556154537295, 'std': 0.3759869210837249}, annotation_units=1, channel_map={'4462_Average-monthly-intake-of-other-alcoholic-drinks_1_0': 0, })
ukb_4462_2 = TensorMap('4462_Average-monthly-intake-of-other-alcoholic-drinks_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05863996879469511, 'std': 0.4364047127971609}, annotation_units=1, channel_map={'4462_Average-monthly-intake-of-other-alcoholic-drinks_2_0': 0, })
ukb_4462_0 = TensorMap('4462_Average-monthly-intake-of-other-alcoholic-drinks_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10946219566398344, 'std': 0.6432363263177207}, annotation_units=1, channel_map={'4462_Average-monthly-intake-of-other-alcoholic-drinks_0_0': 0, })
ukb_4689_1 = TensorMap('4689_Age-glaucoma-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.011881188118814, 'std': 10.137539140596479}, annotation_units=1, channel_map={'4689_Age-glaucoma-diagnosed_1_0': 0, })
ukb_4689_0 = TensorMap('4689_Age-glaucoma-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 52.86446342369106, 'std': 11.284868847905614}, annotation_units=1, channel_map={'4689_Age-glaucoma-diagnosed_0_0': 0, })
ukb_4689_2 = TensorMap('4689_Age-glaucoma-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.38461538461537, 'std': 11.116181631999659}, annotation_units=1, channel_map={'4689_Age-glaucoma-diagnosed_2_0': 0, })
ukb_4700_0 = TensorMap('4700_Age-cataract-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.1457837694951, 'std': 11.461952927161272}, annotation_units=1, channel_map={'4700_Age-cataract-diagnosed_0_0': 0, })
ukb_4700_1 = TensorMap('4700_Age-cataract-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 61.26470588235293, 'std': 10.24620064392814}, annotation_units=1, channel_map={'4700_Age-cataract-diagnosed_1_0': 0, })
ukb_4700_2 = TensorMap('4700_Age-cataract-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 63.65046354297168, 'std': 9.705299131851017}, annotation_units=1, channel_map={'4700_Age-cataract-diagnosed_2_0': 0, })
ukb_5057_2 = TensorMap('5057_Number-of-older-siblings_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.037547818818915, 'std': 1.2451686888035636}, annotation_units=1, channel_map={'5057_Number-of-older-siblings_2_0': 0, })
ukb_5057_0 = TensorMap('5057_Number-of-older-siblings_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.20091181779571, 'std': 1.491901005831908}, annotation_units=1, channel_map={'5057_Number-of-older-siblings_0_0': 0, })
ukb_5057_1 = TensorMap('5057_Number-of-older-siblings_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9773210082103689, 'std': 1.229985421184087}, annotation_units=1, channel_map={'5057_Number-of-older-siblings_1_0': 0, })
ukb_5074_0 = TensorMap('5074_Number-of-letters-shown-in-round-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.982229886781138, 'std': 0.20385247864802822}, annotation_units=16, channel_map={'5074_Number-of-letters-shown-in-round-left_0_0': 0, '5074_Number-of-letters-shown-in-round-left_0_1': 1, '5074_Number-of-letters-shown-in-round-left_0_2': 2, '5074_Number-of-letters-shown-in-round-left_0_3': 3, '5074_Number-of-letters-shown-in-round-left_0_4': 4, '5074_Number-of-letters-shown-in-round-left_0_5': 5, '5074_Number-of-letters-shown-in-round-left_0_6': 6, '5074_Number-of-letters-shown-in-round-left_0_7': 7, '5074_Number-of-letters-shown-in-round-left_0_8': 8, '5074_Number-of-letters-shown-in-round-left_0_9': 9, '5074_Number-of-letters-shown-in-round-left_0_10': 10, '5074_Number-of-letters-shown-in-round-left_0_11': 11, '5074_Number-of-letters-shown-in-round-left_0_12': 12, '5074_Number-of-letters-shown-in-round-left_0_13': 13, '5074_Number-of-letters-shown-in-round-left_0_14': 14, '5074_Number-of-letters-shown-in-round-left_0_15': 15, })
ukb_5074_1 = TensorMap('5074_Number-of-letters-shown-in-round-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.982593827477131, 'std': 0.20284046294809138}, annotation_units=16, channel_map={'5074_Number-of-letters-shown-in-round-left_1_0': 0, '5074_Number-of-letters-shown-in-round-left_1_1': 1, '5074_Number-of-letters-shown-in-round-left_1_2': 2, '5074_Number-of-letters-shown-in-round-left_1_3': 3, '5074_Number-of-letters-shown-in-round-left_1_4': 4, '5074_Number-of-letters-shown-in-round-left_1_5': 5, '5074_Number-of-letters-shown-in-round-left_1_6': 6, '5074_Number-of-letters-shown-in-round-left_1_7': 7, '5074_Number-of-letters-shown-in-round-left_1_8': 8, '5074_Number-of-letters-shown-in-round-left_1_9': 9, '5074_Number-of-letters-shown-in-round-left_1_10': 10, '5074_Number-of-letters-shown-in-round-left_1_11': 11, '5074_Number-of-letters-shown-in-round-left_1_12': 12, '5074_Number-of-letters-shown-in-round-left_1_13': 13, '5074_Number-of-letters-shown-in-round-left_1_14': 14, '5074_Number-of-letters-shown-in-round-left_1_15': 15, })
ukb_5075_1 = TensorMap('5075_Number-of-letters-shown-in-round-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.9848447554414745, 'std': 0.18728205349529364}, annotation_units=16, channel_map={'5075_Number-of-letters-shown-in-round-right_1_0': 0, '5075_Number-of-letters-shown-in-round-right_1_1': 1, '5075_Number-of-letters-shown-in-round-right_1_2': 2, '5075_Number-of-letters-shown-in-round-right_1_3': 3, '5075_Number-of-letters-shown-in-round-right_1_4': 4, '5075_Number-of-letters-shown-in-round-right_1_5': 5, '5075_Number-of-letters-shown-in-round-right_1_6': 6, '5075_Number-of-letters-shown-in-round-right_1_7': 7, '5075_Number-of-letters-shown-in-round-right_1_8': 8, '5075_Number-of-letters-shown-in-round-right_1_9': 9, '5075_Number-of-letters-shown-in-round-right_1_10': 10, '5075_Number-of-letters-shown-in-round-right_1_11': 11, '5075_Number-of-letters-shown-in-round-right_1_12': 12, '5075_Number-of-letters-shown-in-round-right_1_13': 13, '5075_Number-of-letters-shown-in-round-right_1_14': 14, '5075_Number-of-letters-shown-in-round-right_1_15': 15, })
ukb_5075_0 = TensorMap('5075_Number-of-letters-shown-in-round-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.984711453020578, 'std': 0.1896981176486569}, annotation_units=16, channel_map={'5075_Number-of-letters-shown-in-round-right_0_0': 0, '5075_Number-of-letters-shown-in-round-right_0_1': 1, '5075_Number-of-letters-shown-in-round-right_0_2': 2, '5075_Number-of-letters-shown-in-round-right_0_3': 3, '5075_Number-of-letters-shown-in-round-right_0_4': 4, '5075_Number-of-letters-shown-in-round-right_0_5': 5, '5075_Number-of-letters-shown-in-round-right_0_6': 6, '5075_Number-of-letters-shown-in-round-right_0_7': 7, '5075_Number-of-letters-shown-in-round-right_0_8': 8, '5075_Number-of-letters-shown-in-round-right_0_9': 9, '5075_Number-of-letters-shown-in-round-right_0_10': 10, '5075_Number-of-letters-shown-in-round-right_0_11': 11, '5075_Number-of-letters-shown-in-round-right_0_12': 12, '5075_Number-of-letters-shown-in-round-right_0_13': 13, '5075_Number-of-letters-shown-in-round-right_0_14': 14, '5075_Number-of-letters-shown-in-round-right_0_15': 15, })
ukb_5076_0 = TensorMap('5076_Number-of-letters-correct-in-round-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.519436295692189, 'std': 1.7524776629924468}, annotation_units=16, channel_map={'5076_Number-of-letters-correct-in-round-right_0_0': 0, '5076_Number-of-letters-correct-in-round-right_0_1': 1, '5076_Number-of-letters-correct-in-round-right_0_2': 2, '5076_Number-of-letters-correct-in-round-right_0_3': 3, '5076_Number-of-letters-correct-in-round-right_0_4': 4, '5076_Number-of-letters-correct-in-round-right_0_5': 5, '5076_Number-of-letters-correct-in-round-right_0_6': 6, '5076_Number-of-letters-correct-in-round-right_0_7': 7, '5076_Number-of-letters-correct-in-round-right_0_8': 8, '5076_Number-of-letters-correct-in-round-right_0_9': 9, '5076_Number-of-letters-correct-in-round-right_0_10': 10, '5076_Number-of-letters-correct-in-round-right_0_11': 11, '5076_Number-of-letters-correct-in-round-right_0_12': 12, '5076_Number-of-letters-correct-in-round-right_0_13': 13, '5076_Number-of-letters-correct-in-round-right_0_14': 14, '5076_Number-of-letters-correct-in-round-right_0_15': 15, })
ukb_5076_1 = TensorMap('5076_Number-of-letters-correct-in-round-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5128872323218103, 'std': 1.7678237854446686}, annotation_units=16, channel_map={'5076_Number-of-letters-correct-in-round-right_1_0': 0, '5076_Number-of-letters-correct-in-round-right_1_1': 1, '5076_Number-of-letters-correct-in-round-right_1_2': 2, '5076_Number-of-letters-correct-in-round-right_1_3': 3, '5076_Number-of-letters-correct-in-round-right_1_4': 4, '5076_Number-of-letters-correct-in-round-right_1_5': 5, '5076_Number-of-letters-correct-in-round-right_1_6': 6, '5076_Number-of-letters-correct-in-round-right_1_7': 7, '5076_Number-of-letters-correct-in-round-right_1_8': 8, '5076_Number-of-letters-correct-in-round-right_1_9': 9, '5076_Number-of-letters-correct-in-round-right_1_10': 10, '5076_Number-of-letters-correct-in-round-right_1_11': 11, '5076_Number-of-letters-correct-in-round-right_1_12': 12, '5076_Number-of-letters-correct-in-round-right_1_13': 13, '5076_Number-of-letters-correct-in-round-right_1_14': 14, '5076_Number-of-letters-correct-in-round-right_1_15': 15, })
ukb_5077_1 = TensorMap('5077_Number-of-letters-correct-in-round-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.53379499016842, 'std': 1.8048016257895025}, annotation_units=16, channel_map={'5077_Number-of-letters-correct-in-round-left_1_0': 0, '5077_Number-of-letters-correct-in-round-left_1_1': 1, '5077_Number-of-letters-correct-in-round-left_1_2': 2, '5077_Number-of-letters-correct-in-round-left_1_3': 3, '5077_Number-of-letters-correct-in-round-left_1_4': 4, '5077_Number-of-letters-correct-in-round-left_1_5': 5, '5077_Number-of-letters-correct-in-round-left_1_6': 6, '5077_Number-of-letters-correct-in-round-left_1_7': 7, '5077_Number-of-letters-correct-in-round-left_1_8': 8, '5077_Number-of-letters-correct-in-round-left_1_9': 9, '5077_Number-of-letters-correct-in-round-left_1_10': 10, '5077_Number-of-letters-correct-in-round-left_1_11': 11, '5077_Number-of-letters-correct-in-round-left_1_12': 12, '5077_Number-of-letters-correct-in-round-left_1_13': 13, '5077_Number-of-letters-correct-in-round-left_1_14': 14, '5077_Number-of-letters-correct-in-round-left_1_15': 15, })
ukb_5077_0 = TensorMap('5077_Number-of-letters-correct-in-round-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5207587913323852, 'std': 1.7938301076973655}, annotation_units=16, channel_map={'5077_Number-of-letters-correct-in-round-left_0_0': 0, '5077_Number-of-letters-correct-in-round-left_0_1': 1, '5077_Number-of-letters-correct-in-round-left_0_2': 2, '5077_Number-of-letters-correct-in-round-left_0_3': 3, '5077_Number-of-letters-correct-in-round-left_0_4': 4, '5077_Number-of-letters-correct-in-round-left_0_5': 5, '5077_Number-of-letters-correct-in-round-left_0_6': 6, '5077_Number-of-letters-correct-in-round-left_0_7': 7, '5077_Number-of-letters-correct-in-round-left_0_8': 8, '5077_Number-of-letters-correct-in-round-left_0_9': 9, '5077_Number-of-letters-correct-in-round-left_0_10': 10, '5077_Number-of-letters-correct-in-round-left_0_11': 11, '5077_Number-of-letters-correct-in-round-left_0_12': 12, '5077_Number-of-letters-correct-in-round-left_0_13': 13, '5077_Number-of-letters-correct-in-round-left_0_14': 14, '5077_Number-of-letters-correct-in-round-left_0_15': 15, })
ukb_5078_1 = TensorMap('5078_logMAR-in-round-left_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03930238522698128, 'std': 0.25558060867275095}, annotation_units=16, channel_map={'5078_logMAR-in-round-left_1_0': 0, '5078_logMAR-in-round-left_1_1': 1, '5078_logMAR-in-round-left_1_2': 2, '5078_logMAR-in-round-left_1_3': 3, '5078_logMAR-in-round-left_1_4': 4, '5078_logMAR-in-round-left_1_5': 5, '5078_logMAR-in-round-left_1_6': 6, '5078_logMAR-in-round-left_1_7': 7, '5078_logMAR-in-round-left_1_8': 8, '5078_logMAR-in-round-left_1_9': 9, '5078_logMAR-in-round-left_1_10': 10, '5078_logMAR-in-round-left_1_11': 11, '5078_logMAR-in-round-left_1_12': 12, '5078_logMAR-in-round-left_1_13': 13, '5078_logMAR-in-round-left_1_14': 14, '5078_logMAR-in-round-left_1_15': 15, })
ukb_5078_0 = TensorMap('5078_logMAR-in-round-left_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06420994226587719, 'std': 0.251582737043921}, annotation_units=16, channel_map={'5078_logMAR-in-round-left_0_0': 0, '5078_logMAR-in-round-left_0_1': 1, '5078_logMAR-in-round-left_0_2': 2, '5078_logMAR-in-round-left_0_3': 3, '5078_logMAR-in-round-left_0_4': 4, '5078_logMAR-in-round-left_0_5': 5, '5078_logMAR-in-round-left_0_6': 6, '5078_logMAR-in-round-left_0_7': 7, '5078_logMAR-in-round-left_0_8': 8, '5078_logMAR-in-round-left_0_9': 9, '5078_logMAR-in-round-left_0_10': 10, '5078_logMAR-in-round-left_0_11': 11, '5078_logMAR-in-round-left_0_12': 12, '5078_logMAR-in-round-left_0_13': 13, '5078_logMAR-in-round-left_0_14': 14, '5078_logMAR-in-round-left_0_15': 15, })
ukb_5079_1 = TensorMap('5079_logMAR-in-round-right_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03127395478040714, 'std': 0.24564171924463868}, annotation_units=16, channel_map={'5079_logMAR-in-round-right_1_0': 0, '5079_logMAR-in-round-right_1_1': 1, '5079_logMAR-in-round-right_1_2': 2, '5079_logMAR-in-round-right_1_3': 3, '5079_logMAR-in-round-right_1_4': 4, '5079_logMAR-in-round-right_1_5': 5, '5079_logMAR-in-round-right_1_6': 6, '5079_logMAR-in-round-right_1_7': 7, '5079_logMAR-in-round-right_1_8': 8, '5079_logMAR-in-round-right_1_9': 9, '5079_logMAR-in-round-right_1_10': 10, '5079_logMAR-in-round-right_1_11': 11, '5079_logMAR-in-round-right_1_12': 12, '5079_logMAR-in-round-right_1_13': 13, '5079_logMAR-in-round-right_1_14': 14, '5079_logMAR-in-round-right_1_15': 15, })
ukb_5079_0 = TensorMap('5079_logMAR-in-round-right_0_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06616830791328378, 'std': 0.24566479978416353}, annotation_units=16, channel_map={'5079_logMAR-in-round-right_0_0': 0, '5079_logMAR-in-round-right_0_1': 1, '5079_logMAR-in-round-right_0_2': 2, '5079_logMAR-in-round-right_0_3': 3, '5079_logMAR-in-round-right_0_4': 4, '5079_logMAR-in-round-right_0_5': 5, '5079_logMAR-in-round-right_0_6': 6, '5079_logMAR-in-round-right_0_7': 7, '5079_logMAR-in-round-right_0_8': 8, '5079_logMAR-in-round-right_0_9': 9, '5079_logMAR-in-round-right_0_10': 10, '5079_logMAR-in-round-right_0_11': 11, '5079_logMAR-in-round-right_0_12': 12, '5079_logMAR-in-round-right_0_13': 13, '5079_logMAR-in-round-right_0_14': 14, '5079_logMAR-in-round-right_0_15': 15, })
ukb_5084_1 = TensorMap('5084_Spherical-power-right_1_7', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8194113603162327, 'std': 2.9347619402140452}, annotation_units=8, channel_map={'5084_Spherical-power-right_1_0': 0, '5084_Spherical-power-right_1_1': 1, '5084_Spherical-power-right_1_2': 2, '5084_Spherical-power-right_1_3': 3, '5084_Spherical-power-right_1_4': 4, '5084_Spherical-power-right_1_5': 5, '5084_Spherical-power-right_1_6': 6, '5084_Spherical-power-right_1_7': 7, })
ukb_5084_0 = TensorMap('5084_Spherical-power-right_0_9', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.7995140702990166, 'std': 2.8265271735844455}, annotation_units=10, channel_map={'5084_Spherical-power-right_0_0': 0, '5084_Spherical-power-right_0_1': 1, '5084_Spherical-power-right_0_2': 2, '5084_Spherical-power-right_0_3': 3, '5084_Spherical-power-right_0_4': 4, '5084_Spherical-power-right_0_5': 5, '5084_Spherical-power-right_0_6': 6, '5084_Spherical-power-right_0_7': 7, '5084_Spherical-power-right_0_8': 8, '5084_Spherical-power-right_0_9': 9, })
ukb_5085_0 = TensorMap('5085_Spherical-power-left_0_9', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.7414680505335891, 'std': 2.8397422506629497}, annotation_units=10, channel_map={'5085_Spherical-power-left_0_0': 0, '5085_Spherical-power-left_0_1': 1, '5085_Spherical-power-left_0_2': 2, '5085_Spherical-power-left_0_3': 3, '5085_Spherical-power-left_0_4': 4, '5085_Spherical-power-left_0_5': 5, '5085_Spherical-power-left_0_6': 6, '5085_Spherical-power-left_0_7': 7, '5085_Spherical-power-left_0_8': 8, '5085_Spherical-power-left_0_9': 9, })
ukb_5085_1 = TensorMap('5085_Spherical-power-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.779588979301491, 'std': 2.9745336292705526}, annotation_units=6, channel_map={'5085_Spherical-power-left_1_0': 0, '5085_Spherical-power-left_1_1': 1, '5085_Spherical-power-left_1_2': 2, '5085_Spherical-power-left_1_3': 3, '5085_Spherical-power-left_1_4': 4, '5085_Spherical-power-left_1_5': 5, })
ukb_5086_1 = TensorMap('5086_Cylindrical-power-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9513624928431389, 'std': 0.7548492904582892}, annotation_units=6, channel_map={'5086_Cylindrical-power-left_1_0': 0, '5086_Cylindrical-power-left_1_1': 1, '5086_Cylindrical-power-left_1_2': 2, '5086_Cylindrical-power-left_1_3': 3, '5086_Cylindrical-power-left_1_4': 4, '5086_Cylindrical-power-left_1_5': 5, })
ukb_5086_0 = TensorMap('5086_Cylindrical-power-left_0_9', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8734884279252244, 'std': 0.7384053630445866}, annotation_units=10, channel_map={'5086_Cylindrical-power-left_0_0': 0, '5086_Cylindrical-power-left_0_1': 1, '5086_Cylindrical-power-left_0_2': 2, '5086_Cylindrical-power-left_0_3': 3, '5086_Cylindrical-power-left_0_4': 4, '5086_Cylindrical-power-left_0_5': 5, '5086_Cylindrical-power-left_0_6': 6, '5086_Cylindrical-power-left_0_7': 7, '5086_Cylindrical-power-left_0_8': 8, '5086_Cylindrical-power-left_0_9': 9, })
ukb_5087_0 = TensorMap('5087_Cylindrical-power-right_0_9', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8870669578299205, 'std': 0.7420501538536189}, annotation_units=10, channel_map={'5087_Cylindrical-power-right_0_0': 0, '5087_Cylindrical-power-right_0_1': 1, '5087_Cylindrical-power-right_0_2': 2, '5087_Cylindrical-power-right_0_3': 3, '5087_Cylindrical-power-right_0_4': 4, '5087_Cylindrical-power-right_0_5': 5, '5087_Cylindrical-power-right_0_6': 6, '5087_Cylindrical-power-right_0_7': 7, '5087_Cylindrical-power-right_0_8': 8, '5087_Cylindrical-power-right_0_9': 9, })
ukb_5087_1 = TensorMap('5087_Cylindrical-power-right_1_7', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9660501847555205, 'std': 0.7584826527408577}, annotation_units=8, channel_map={'5087_Cylindrical-power-right_1_0': 0, '5087_Cylindrical-power-right_1_1': 1, '5087_Cylindrical-power-right_1_2': 2, '5087_Cylindrical-power-right_1_3': 3, '5087_Cylindrical-power-right_1_4': 4, '5087_Cylindrical-power-right_1_5': 5, '5087_Cylindrical-power-right_1_6': 6, '5087_Cylindrical-power-right_1_7': 7, })
ukb_5088_0 = TensorMap('5088_Astigmatism-angle-right_0_9', loss='logcosh', path_prefix='continuous', normalization={'mean': 94.00180810732, 'std': 57.11136528637008}, annotation_units=10, channel_map={'5088_Astigmatism-angle-right_0_0': 0, '5088_Astigmatism-angle-right_0_1': 1, '5088_Astigmatism-angle-right_0_2': 2, '5088_Astigmatism-angle-right_0_3': 3, '5088_Astigmatism-angle-right_0_4': 4, '5088_Astigmatism-angle-right_0_5': 5, '5088_Astigmatism-angle-right_0_6': 6, '5088_Astigmatism-angle-right_0_7': 7, '5088_Astigmatism-angle-right_0_8': 8, '5088_Astigmatism-angle-right_0_9': 9, })
ukb_5088_1 = TensorMap('5088_Astigmatism-angle-right_1_7', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.58807653298032, 'std': 59.446167118752015}, annotation_units=8, channel_map={'5088_Astigmatism-angle-right_1_0': 0, '5088_Astigmatism-angle-right_1_1': 1, '5088_Astigmatism-angle-right_1_2': 2, '5088_Astigmatism-angle-right_1_3': 3, '5088_Astigmatism-angle-right_1_4': 4, '5088_Astigmatism-angle-right_1_5': 5, '5088_Astigmatism-angle-right_1_6': 6, '5088_Astigmatism-angle-right_1_7': 7, })
ukb_5089_0 = TensorMap('5089_Astigmatism-angle-left_0_9', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.46927127260784, 'std': 56.27046457482316}, annotation_units=10, channel_map={'5089_Astigmatism-angle-left_0_0': 0, '5089_Astigmatism-angle-left_0_1': 1, '5089_Astigmatism-angle-left_0_2': 2, '5089_Astigmatism-angle-left_0_3': 3, '5089_Astigmatism-angle-left_0_4': 4, '5089_Astigmatism-angle-left_0_5': 5, '5089_Astigmatism-angle-left_0_6': 6, '5089_Astigmatism-angle-left_0_7': 7, '5089_Astigmatism-angle-left_0_8': 8, '5089_Astigmatism-angle-left_0_9': 9, })
ukb_5089_1 = TensorMap('5089_Astigmatism-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 93.02731642977395, 'std': 58.444200672298294}, annotation_units=6, channel_map={'5089_Astigmatism-angle-left_1_0': 0, '5089_Astigmatism-angle-left_1_1': 1, '5089_Astigmatism-angle-left_1_2': 2, '5089_Astigmatism-angle-left_1_3': 3, '5089_Astigmatism-angle-left_1_4': 4, '5089_Astigmatism-angle-left_1_5': 5, })
ukb_5096_1 = TensorMap('5096_3mm-weak-meridian-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.10357498698785, 'std': 1.5059108909428265}, annotation_units=6, channel_map={'5096_3mm-weak-meridian-left_1_0': 0, '5096_3mm-weak-meridian-left_1_1': 1, '5096_3mm-weak-meridian-left_1_2': 2, '5096_3mm-weak-meridian-left_1_3': 3, '5096_3mm-weak-meridian-left_1_4': 4, '5096_3mm-weak-meridian-left_1_5': 5, })
ukb_5096_0 = TensorMap('5096_3mm-weak-meridian-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.12724202952481, 'std': 1.528157468013433}, annotation_units=6, channel_map={'5096_3mm-weak-meridian-left_0_0': 0, '5096_3mm-weak-meridian-left_0_1': 1, '5096_3mm-weak-meridian-left_0_2': 2, '5096_3mm-weak-meridian-left_0_3': 3, '5096_3mm-weak-meridian-left_0_4': 4, '5096_3mm-weak-meridian-left_0_5': 5, })
ukb_5097_0 = TensorMap('5097_6mm-weak-meridian-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.63592044522244, 'std': 1.5381520009454699}, annotation_units=6, channel_map={'5097_6mm-weak-meridian-left_0_0': 0, '5097_6mm-weak-meridian-left_0_1': 1, '5097_6mm-weak-meridian-left_0_2': 2, '5097_6mm-weak-meridian-left_0_3': 3, '5097_6mm-weak-meridian-left_0_4': 4, '5097_6mm-weak-meridian-left_0_5': 5, })
ukb_5097_1 = TensorMap('5097_6mm-weak-meridian-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.584022826546416, 'std': 1.518018113442274}, annotation_units=6, channel_map={'5097_6mm-weak-meridian-left_1_0': 0, '5097_6mm-weak-meridian-left_1_1': 1, '5097_6mm-weak-meridian-left_1_2': 2, '5097_6mm-weak-meridian-left_1_3': 3, '5097_6mm-weak-meridian-left_1_4': 4, '5097_6mm-weak-meridian-left_1_5': 5, })
ukb_5098_1 = TensorMap('5098_6mm-weak-meridian-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.5238056746948, 'std': 1.523396374730744}, annotation_units=6, channel_map={'5098_6mm-weak-meridian-right_1_0': 0, '5098_6mm-weak-meridian-right_1_1': 1, '5098_6mm-weak-meridian-right_1_2': 2, '5098_6mm-weak-meridian-right_1_3': 3, '5098_6mm-weak-meridian-right_1_4': 4, '5098_6mm-weak-meridian-right_1_5': 5, })
ukb_5098_0 = TensorMap('5098_6mm-weak-meridian-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.59086038597763, 'std': 1.5354686330880993}, annotation_units=6, channel_map={'5098_6mm-weak-meridian-right_0_0': 0, '5098_6mm-weak-meridian-right_0_1': 1, '5098_6mm-weak-meridian-right_0_2': 2, '5098_6mm-weak-meridian-right_0_3': 3, '5098_6mm-weak-meridian-right_0_4': 4, '5098_6mm-weak-meridian-right_0_5': 5, })
ukb_5099_0 = TensorMap('5099_3mm-weak-meridian-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.092410993287594, 'std': 1.52464192184014}, annotation_units=6, channel_map={'5099_3mm-weak-meridian-right_0_0': 0, '5099_3mm-weak-meridian-right_0_1': 1, '5099_3mm-weak-meridian-right_0_2': 2, '5099_3mm-weak-meridian-right_0_3': 3, '5099_3mm-weak-meridian-right_0_4': 4, '5099_3mm-weak-meridian-right_0_5': 5, })
ukb_5099_1 = TensorMap('5099_3mm-weak-meridian-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.05472852573863, 'std': 1.510993217438951}, annotation_units=6, channel_map={'5099_3mm-weak-meridian-right_1_0': 0, '5099_3mm-weak-meridian-right_1_1': 1, '5099_3mm-weak-meridian-right_1_2': 2, '5099_3mm-weak-meridian-right_1_3': 3, '5099_3mm-weak-meridian-right_1_4': 4, '5099_3mm-weak-meridian-right_1_5': 5, })
ukb_5100_1 = TensorMap('5100_3mm-weak-meridian-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.58403137374351, 'std': 64.57797126950123}, annotation_units=6, channel_map={'5100_3mm-weak-meridian-angle-right_1_0': 0, '5100_3mm-weak-meridian-angle-right_1_1': 1, '5100_3mm-weak-meridian-angle-right_1_2': 2, '5100_3mm-weak-meridian-angle-right_1_3': 3, '5100_3mm-weak-meridian-angle-right_1_4': 4, '5100_3mm-weak-meridian-angle-right_1_5': 5, })
ukb_5100_0 = TensorMap('5100_3mm-weak-meridian-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.9387712414986, 'std': 68.06384277305852}, annotation_units=6, channel_map={'5100_3mm-weak-meridian-angle-right_0_0': 0, '5100_3mm-weak-meridian-angle-right_0_1': 1, '5100_3mm-weak-meridian-angle-right_0_2': 2, '5100_3mm-weak-meridian-angle-right_0_3': 3, '5100_3mm-weak-meridian-angle-right_0_4': 4, '5100_3mm-weak-meridian-angle-right_0_5': 5, })
ukb_5101_1 = TensorMap('5101_6mm-weak-meridian-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 87.20804072206256, 'std': 57.64984326558321}, annotation_units=6, channel_map={'5101_6mm-weak-meridian-angle-right_1_0': 0, '5101_6mm-weak-meridian-angle-right_1_1': 1, '5101_6mm-weak-meridian-angle-right_1_2': 2, '5101_6mm-weak-meridian-angle-right_1_3': 3, '5101_6mm-weak-meridian-angle-right_1_4': 4, '5101_6mm-weak-meridian-angle-right_1_5': 5, })
ukb_5101_0 = TensorMap('5101_6mm-weak-meridian-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 83.45198814750806, 'std': 62.0498737775849}, annotation_units=6, channel_map={'5101_6mm-weak-meridian-angle-right_0_0': 0, '5101_6mm-weak-meridian-angle-right_0_1': 1, '5101_6mm-weak-meridian-angle-right_0_2': 2, '5101_6mm-weak-meridian-angle-right_0_3': 3, '5101_6mm-weak-meridian-angle-right_0_4': 4, '5101_6mm-weak-meridian-angle-right_0_5': 5, })
ukb_5102_0 = TensorMap('5102_6mm-weak-meridian-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 98.51123453196068, 'std': 62.56834297588253}, annotation_units=6, channel_map={'5102_6mm-weak-meridian-angle-left_0_0': 0, '5102_6mm-weak-meridian-angle-left_0_1': 1, '5102_6mm-weak-meridian-angle-left_0_2': 2, '5102_6mm-weak-meridian-angle-left_0_3': 3, '5102_6mm-weak-meridian-angle-left_0_4': 4, '5102_6mm-weak-meridian-angle-left_0_5': 5, })
ukb_5102_1 = TensorMap('5102_6mm-weak-meridian-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 100.70705932083972, 'std': 57.74231635668554}, annotation_units=6, channel_map={'5102_6mm-weak-meridian-angle-left_1_0': 0, '5102_6mm-weak-meridian-angle-left_1_1': 1, '5102_6mm-weak-meridian-angle-left_1_2': 2, '5102_6mm-weak-meridian-angle-left_1_3': 3, '5102_6mm-weak-meridian-angle-left_1_4': 4, '5102_6mm-weak-meridian-angle-left_1_5': 5, })
ukb_5103_1 = TensorMap('5103_3mm-weak-meridian-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.85089969514463, 'std': 65.803409408443}, annotation_units=6, channel_map={'5103_3mm-weak-meridian-angle-left_1_0': 0, '5103_3mm-weak-meridian-angle-left_1_1': 1, '5103_3mm-weak-meridian-angle-left_1_2': 2, '5103_3mm-weak-meridian-angle-left_1_3': 3, '5103_3mm-weak-meridian-angle-left_1_4': 4, '5103_3mm-weak-meridian-angle-left_1_5': 5, })
ukb_5103_0 = TensorMap('5103_3mm-weak-meridian-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 86.27880846048856, 'std': 67.96516785733142}, annotation_units=6, channel_map={'5103_3mm-weak-meridian-angle-left_0_0': 0, '5103_3mm-weak-meridian-angle-left_0_1': 1, '5103_3mm-weak-meridian-angle-left_0_2': 2, '5103_3mm-weak-meridian-angle-left_0_3': 3, '5103_3mm-weak-meridian-angle-left_0_4': 4, '5103_3mm-weak-meridian-angle-left_0_5': 5, })
ukb_5104_1 = TensorMap('5104_3mm-strong-meridian-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.94124098445975, 'std': 40.78480753146676}, annotation_units=6, channel_map={'5104_3mm-strong-meridian-angle-left_1_0': 0, '5104_3mm-strong-meridian-angle-left_1_1': 1, '5104_3mm-strong-meridian-angle-left_1_2': 2, '5104_3mm-strong-meridian-angle-left_1_3': 3, '5104_3mm-strong-meridian-angle-left_1_4': 4, '5104_3mm-strong-meridian-angle-left_1_5': 5, })
ukb_5104_0 = TensorMap('5104_3mm-strong-meridian-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 94.36608051110272, 'std': 37.6552118153151}, annotation_units=6, channel_map={'5104_3mm-strong-meridian-angle-left_0_0': 0, '5104_3mm-strong-meridian-angle-left_0_1': 1, '5104_3mm-strong-meridian-angle-left_0_2': 2, '5104_3mm-strong-meridian-angle-left_0_3': 3, '5104_3mm-strong-meridian-angle-left_0_4': 4, '5104_3mm-strong-meridian-angle-left_0_5': 5, })
ukb_5105_0 = TensorMap('5105_6mm-strong-meridian-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 86.84853136330482, 'std': 44.53524516032851}, annotation_units=6, channel_map={'5105_6mm-strong-meridian-angle-left_0_0': 0, '5105_6mm-strong-meridian-angle-left_0_1': 1, '5105_6mm-strong-meridian-angle-left_0_2': 2, '5105_6mm-strong-meridian-angle-left_0_3': 3, '5105_6mm-strong-meridian-angle-left_0_4': 4, '5105_6mm-strong-meridian-angle-left_0_5': 5, })
ukb_5105_1 = TensorMap('5105_6mm-strong-meridian-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 84.99901366774692, 'std': 49.40503392095581}, annotation_units=6, channel_map={'5105_6mm-strong-meridian-angle-left_1_0': 0, '5105_6mm-strong-meridian-angle-left_1_1': 1, '5105_6mm-strong-meridian-angle-left_1_2': 2, '5105_6mm-strong-meridian-angle-left_1_3': 3, '5105_6mm-strong-meridian-angle-left_1_4': 4, '5105_6mm-strong-meridian-angle-left_1_5': 5, })
ukb_5106_0 = TensorMap('5106_6mm-strong-meridian-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 93.05320117598951, 'std': 45.7408559807841}, annotation_units=6, channel_map={'5106_6mm-strong-meridian-angle-right_0_0': 0, '5106_6mm-strong-meridian-angle-right_0_1': 1, '5106_6mm-strong-meridian-angle-right_0_2': 2, '5106_6mm-strong-meridian-angle-right_0_3': 3, '5106_6mm-strong-meridian-angle-right_0_4': 4, '5106_6mm-strong-meridian-angle-right_0_5': 5, })
ukb_5106_1 = TensorMap('5106_6mm-strong-meridian-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.18159966064947, 'std': 51.29410499368635}, annotation_units=6, channel_map={'5106_6mm-strong-meridian-angle-right_1_0': 0, '5106_6mm-strong-meridian-angle-right_1_1': 1, '5106_6mm-strong-meridian-angle-right_1_2': 2, '5106_6mm-strong-meridian-angle-right_1_3': 3, '5106_6mm-strong-meridian-angle-right_1_4': 4, '5106_6mm-strong-meridian-angle-right_1_5': 5, })
ukb_5107_1 = TensorMap('5107_3mm-strong-meridian-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 83.03285866585436, 'std': 40.98014484034505}, annotation_units=6, channel_map={'5107_3mm-strong-meridian-angle-right_1_0': 0, '5107_3mm-strong-meridian-angle-right_1_1': 1, '5107_3mm-strong-meridian-angle-right_1_2': 2, '5107_3mm-strong-meridian-angle-right_1_3': 3, '5107_3mm-strong-meridian-angle-right_1_4': 4, '5107_3mm-strong-meridian-angle-right_1_5': 5, })
ukb_5107_0 = TensorMap('5107_3mm-strong-meridian-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 87.54627334901022, 'std': 38.07773873439575}, annotation_units=6, channel_map={'5107_3mm-strong-meridian-angle-right_0_0': 0, '5107_3mm-strong-meridian-angle-right_0_1': 1, '5107_3mm-strong-meridian-angle-right_0_2': 2, '5107_3mm-strong-meridian-angle-right_0_3': 3, '5107_3mm-strong-meridian-angle-right_0_4': 4, '5107_3mm-strong-meridian-angle-right_0_5': 5, })
ukb_5108_1 = TensorMap('5108_3mm-asymmetry-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 181.9657950474077, 'std': 128.40772355101151}, annotation_units=6, channel_map={'5108_3mm-asymmetry-angle-right_1_0': 0, '5108_3mm-asymmetry-angle-right_1_1': 1, '5108_3mm-asymmetry-angle-right_1_2': 2, '5108_3mm-asymmetry-angle-right_1_3': 3, '5108_3mm-asymmetry-angle-right_1_4': 4, '5108_3mm-asymmetry-angle-right_1_5': 5, })
ukb_5108_0 = TensorMap('5108_3mm-asymmetry-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 172.96171592590056, 'std': 125.62527821935153}, annotation_units=6, channel_map={'5108_3mm-asymmetry-angle-right_0_0': 0, '5108_3mm-asymmetry-angle-right_0_1': 1, '5108_3mm-asymmetry-angle-right_0_2': 2, '5108_3mm-asymmetry-angle-right_0_3': 3, '5108_3mm-asymmetry-angle-right_0_4': 4, '5108_3mm-asymmetry-angle-right_0_5': 5, })
ukb_5109_0 = TensorMap('5109_6mm-asymmetry-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 143.18429828411914, 'std': 130.42863315195004}, annotation_units=6, channel_map={'5109_6mm-asymmetry-angle-right_0_0': 0, '5109_6mm-asymmetry-angle-right_0_1': 1, '5109_6mm-asymmetry-angle-right_0_2': 2, '5109_6mm-asymmetry-angle-right_0_3': 3, '5109_6mm-asymmetry-angle-right_0_4': 4, '5109_6mm-asymmetry-angle-right_0_5': 5, })
ukb_5109_1 = TensorMap('5109_6mm-asymmetry-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 142.78700880758797, 'std': 131.24416164654872}, annotation_units=6, channel_map={'5109_6mm-asymmetry-angle-right_1_0': 0, '5109_6mm-asymmetry-angle-right_1_1': 1, '5109_6mm-asymmetry-angle-right_1_2': 2, '5109_6mm-asymmetry-angle-right_1_3': 3, '5109_6mm-asymmetry-angle-right_1_4': 4, '5109_6mm-asymmetry-angle-right_1_5': 5, })
ukb_5110_1 = TensorMap('5110_6mm-asymmetry-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 152.20844282444688, 'std': 66.06552520727078}, annotation_units=6, channel_map={'5110_6mm-asymmetry-angle-left_1_0': 0, '5110_6mm-asymmetry-angle-left_1_1': 1, '5110_6mm-asymmetry-angle-left_1_2': 2, '5110_6mm-asymmetry-angle-left_1_3': 3, '5110_6mm-asymmetry-angle-left_1_4': 4, '5110_6mm-asymmetry-angle-left_1_5': 5, })
ukb_5110_0 = TensorMap('5110_6mm-asymmetry-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 154.43447613559982, 'std': 64.7641091705269}, annotation_units=6, channel_map={'5110_6mm-asymmetry-angle-left_0_0': 0, '5110_6mm-asymmetry-angle-left_0_1': 1, '5110_6mm-asymmetry-angle-left_0_2': 2, '5110_6mm-asymmetry-angle-left_0_3': 3, '5110_6mm-asymmetry-angle-left_0_4': 4, '5110_6mm-asymmetry-angle-left_0_5': 5, })
ukb_5111_0 = TensorMap('5111_3mm-asymmetry-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 171.03095739681632, 'std': 88.06431006909182}, annotation_units=6, channel_map={'5111_3mm-asymmetry-angle-left_0_0': 0, '5111_3mm-asymmetry-angle-left_0_1': 1, '5111_3mm-asymmetry-angle-left_0_2': 2, '5111_3mm-asymmetry-angle-left_0_3': 3, '5111_3mm-asymmetry-angle-left_0_4': 4, '5111_3mm-asymmetry-angle-left_0_5': 5, })
ukb_5111_1 = TensorMap('5111_3mm-asymmetry-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 175.92833561789303, 'std': 92.25522241823276}, annotation_units=6, channel_map={'5111_3mm-asymmetry-angle-left_1_0': 0, '5111_3mm-asymmetry-angle-left_1_1': 1, '5111_3mm-asymmetry-angle-left_1_2': 2, '5111_3mm-asymmetry-angle-left_1_3': 3, '5111_3mm-asymmetry-angle-left_1_4': 4, '5111_3mm-asymmetry-angle-left_1_5': 5, })
ukb_5112_1 = TensorMap('5112_3mm-cylindrical-power-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.8629488898823, 'std': 65.83330536140146}, annotation_units=6, channel_map={'5112_3mm-cylindrical-power-angle-left_1_0': 0, '5112_3mm-cylindrical-power-angle-left_1_1': 1, '5112_3mm-cylindrical-power-angle-left_1_2': 2, '5112_3mm-cylindrical-power-angle-left_1_3': 3, '5112_3mm-cylindrical-power-angle-left_1_4': 4, '5112_3mm-cylindrical-power-angle-left_1_5': 5, })
ukb_5112_0 = TensorMap('5112_3mm-cylindrical-power-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 86.27697359020405, 'std': 67.98828464032387}, annotation_units=6, channel_map={'5112_3mm-cylindrical-power-angle-left_0_0': 0, '5112_3mm-cylindrical-power-angle-left_0_1': 1, '5112_3mm-cylindrical-power-angle-left_0_2': 2, '5112_3mm-cylindrical-power-angle-left_0_3': 3, '5112_3mm-cylindrical-power-angle-left_0_4': 4, '5112_3mm-cylindrical-power-angle-left_0_5': 5, })
ukb_5113_0 = TensorMap('5113_6mm-cylindrical-power-angle-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 98.54271765971295, 'std': 62.58221515291403}, annotation_units=6, channel_map={'5113_6mm-cylindrical-power-angle-left_0_0': 0, '5113_6mm-cylindrical-power-angle-left_0_1': 1, '5113_6mm-cylindrical-power-angle-left_0_2': 2, '5113_6mm-cylindrical-power-angle-left_0_3': 3, '5113_6mm-cylindrical-power-angle-left_0_4': 4, '5113_6mm-cylindrical-power-angle-left_0_5': 5, })
ukb_5113_1 = TensorMap('5113_6mm-cylindrical-power-angle-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 100.71760000000005, 'std': 57.744474257564484}, annotation_units=6, channel_map={'5113_6mm-cylindrical-power-angle-left_1_0': 0, '5113_6mm-cylindrical-power-angle-left_1_1': 1, '5113_6mm-cylindrical-power-angle-left_1_2': 2, '5113_6mm-cylindrical-power-angle-left_1_3': 3, '5113_6mm-cylindrical-power-angle-left_1_4': 4, '5113_6mm-cylindrical-power-angle-left_1_5': 5, })
ukb_5114_0 = TensorMap('5114_6mm-cylindrical-power-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 83.45155380333955, 'std': 62.06636373512518}, annotation_units=6, channel_map={'5114_6mm-cylindrical-power-angle-right_0_0': 0, '5114_6mm-cylindrical-power-angle-right_0_1': 1, '5114_6mm-cylindrical-power-angle-right_0_2': 2, '5114_6mm-cylindrical-power-angle-right_0_3': 3, '5114_6mm-cylindrical-power-angle-right_0_4': 4, '5114_6mm-cylindrical-power-angle-right_0_5': 5, })
ukb_5114_1 = TensorMap('5114_6mm-cylindrical-power-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 87.22007694668018, 'std': 57.64947751307708}, annotation_units=6, channel_map={'5114_6mm-cylindrical-power-angle-right_1_0': 0, '5114_6mm-cylindrical-power-angle-right_1_1': 1, '5114_6mm-cylindrical-power-angle-right_1_2': 2, '5114_6mm-cylindrical-power-angle-right_1_3': 3, '5114_6mm-cylindrical-power-angle-right_1_4': 4, '5114_6mm-cylindrical-power-angle-right_1_5': 5, })
ukb_5115_0 = TensorMap('5115_3mm-cylindrical-power-angle-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.94881139838482, 'std': 68.09054638103878}, annotation_units=6, channel_map={'5115_3mm-cylindrical-power-angle-right_0_0': 0, '5115_3mm-cylindrical-power-angle-right_0_1': 1, '5115_3mm-cylindrical-power-angle-right_0_2': 2, '5115_3mm-cylindrical-power-angle-right_0_3': 3, '5115_3mm-cylindrical-power-angle-right_0_4': 4, '5115_3mm-cylindrical-power-angle-right_0_5': 5, })
ukb_5115_1 = TensorMap('5115_3mm-cylindrical-power-angle-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.61179993895921, 'std': 64.60068773073615}, annotation_units=6, channel_map={'5115_3mm-cylindrical-power-angle-right_1_0': 0, '5115_3mm-cylindrical-power-angle-right_1_1': 1, '5115_3mm-cylindrical-power-angle-right_1_2': 2, '5115_3mm-cylindrical-power-angle-right_1_3': 3, '5115_3mm-cylindrical-power-angle-right_1_4': 4, '5115_3mm-cylindrical-power-angle-right_1_5': 5, })
ukb_5116_0 = TensorMap('5116_3mm-cylindrical-power-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8662533831107498, 'std': 0.6599557419512332}, annotation_units=6, channel_map={'5116_3mm-cylindrical-power-right_0_0': 0, '5116_3mm-cylindrical-power-right_0_1': 1, '5116_3mm-cylindrical-power-right_0_2': 2, '5116_3mm-cylindrical-power-right_0_3': 3, '5116_3mm-cylindrical-power-right_0_4': 4, '5116_3mm-cylindrical-power-right_0_5': 5, })
ukb_5116_1 = TensorMap('5116_3mm-cylindrical-power-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8574697304294854, 'std': 0.6397741626160982}, annotation_units=6, channel_map={'5116_3mm-cylindrical-power-right_1_0': 0, '5116_3mm-cylindrical-power-right_1_1': 1, '5116_3mm-cylindrical-power-right_1_2': 2, '5116_3mm-cylindrical-power-right_1_3': 3, '5116_3mm-cylindrical-power-right_1_4': 4, '5116_3mm-cylindrical-power-right_1_5': 5, })
ukb_5117_0 = TensorMap('5117_6mm-cylindrical-power-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8422039770666638, 'std': 0.6740590879663511}, annotation_units=6, channel_map={'5117_6mm-cylindrical-power-right_0_0': 0, '5117_6mm-cylindrical-power-right_0_1': 1, '5117_6mm-cylindrical-power-right_0_2': 2, '5117_6mm-cylindrical-power-right_0_3': 3, '5117_6mm-cylindrical-power-right_0_4': 4, '5117_6mm-cylindrical-power-right_0_5': 5, })
ukb_5117_1 = TensorMap('5117_6mm-cylindrical-power-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8356063533958616, 'std': 0.6511670770003918}, annotation_units=6, channel_map={'5117_6mm-cylindrical-power-right_1_0': 0, '5117_6mm-cylindrical-power-right_1_1': 1, '5117_6mm-cylindrical-power-right_1_2': 2, '5117_6mm-cylindrical-power-right_1_3': 3, '5117_6mm-cylindrical-power-right_1_4': 4, '5117_6mm-cylindrical-power-right_1_5': 5, })
ukb_5118_0 = TensorMap('5118_6mm-cylindrical-power-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8553726885371101, 'std': 0.6666154827636736}, annotation_units=6, channel_map={'5118_6mm-cylindrical-power-left_0_0': 0, '5118_6mm-cylindrical-power-left_0_1': 1, '5118_6mm-cylindrical-power-left_0_2': 2, '5118_6mm-cylindrical-power-left_0_3': 3, '5118_6mm-cylindrical-power-left_0_4': 4, '5118_6mm-cylindrical-power-left_0_5': 5, })
ukb_5118_1 = TensorMap('5118_6mm-cylindrical-power-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8538654830679625, 'std': 0.6566722210484399}, annotation_units=6, channel_map={'5118_6mm-cylindrical-power-left_1_0': 0, '5118_6mm-cylindrical-power-left_1_1': 1, '5118_6mm-cylindrical-power-left_1_2': 2, '5118_6mm-cylindrical-power-left_1_3': 3, '5118_6mm-cylindrical-power-left_1_4': 4, '5118_6mm-cylindrical-power-left_1_5': 5, })
ukb_5119_0 = TensorMap('5119_3mm-cylindrical-power-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8757534735144528, 'std': 0.6661926436810551}, annotation_units=6, channel_map={'5119_3mm-cylindrical-power-left_0_0': 0, '5119_3mm-cylindrical-power-left_0_1': 1, '5119_3mm-cylindrical-power-left_0_2': 2, '5119_3mm-cylindrical-power-left_0_3': 3, '5119_3mm-cylindrical-power-left_0_4': 4, '5119_3mm-cylindrical-power-left_0_5': 5, })
ukb_5119_1 = TensorMap('5119_3mm-cylindrical-power-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.8682249609636407, 'std': 0.658712751589956}, annotation_units=6, channel_map={'5119_3mm-cylindrical-power-left_1_0': 0, '5119_3mm-cylindrical-power-left_1_1': 1, '5119_3mm-cylindrical-power-left_1_2': 2, '5119_3mm-cylindrical-power-left_1_3': 3, '5119_3mm-cylindrical-power-left_1_4': 4, '5119_3mm-cylindrical-power-left_1_5': 5, })
ukb_5132_0 = TensorMap('5132_3mm-strong-meridian-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.95866437639845, 'std': 1.6046630931358528}, annotation_units=6, channel_map={'5132_3mm-strong-meridian-right_0_0': 0, '5132_3mm-strong-meridian-right_0_1': 1, '5132_3mm-strong-meridian-right_0_2': 2, '5132_3mm-strong-meridian-right_0_3': 3, '5132_3mm-strong-meridian-right_0_4': 4, '5132_3mm-strong-meridian-right_0_5': 5, })
ukb_5132_1 = TensorMap('5132_3mm-strong-meridian-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.912198256168146, 'std': 1.5729281924922818}, annotation_units=6, channel_map={'5132_3mm-strong-meridian-right_1_0': 0, '5132_3mm-strong-meridian-right_1_1': 1, '5132_3mm-strong-meridian-right_1_2': 2, '5132_3mm-strong-meridian-right_1_3': 3, '5132_3mm-strong-meridian-right_1_4': 4, '5132_3mm-strong-meridian-right_1_5': 5, })
ukb_5133_1 = TensorMap('5133_6mm-strong-meridian-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.35941202809067, 'std': 1.546012371650306}, annotation_units=6, channel_map={'5133_6mm-strong-meridian-right_1_0': 0, '5133_6mm-strong-meridian-right_1_1': 1, '5133_6mm-strong-meridian-right_1_2': 2, '5133_6mm-strong-meridian-right_1_3': 3, '5133_6mm-strong-meridian-right_1_4': 4, '5133_6mm-strong-meridian-right_1_5': 5, })
ukb_5133_0 = TensorMap('5133_6mm-strong-meridian-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.4330643630443, 'std': 1.5733846084011525}, annotation_units=6, channel_map={'5133_6mm-strong-meridian-right_0_0': 0, '5133_6mm-strong-meridian-right_0_1': 1, '5133_6mm-strong-meridian-right_0_2': 2, '5133_6mm-strong-meridian-right_0_3': 3, '5133_6mm-strong-meridian-right_0_4': 4, '5133_6mm-strong-meridian-right_0_5': 5, })
ukb_5134_0 = TensorMap('5134_6mm-strong-meridian-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.49129313375953, 'std': 1.5778281893149044}, annotation_units=6, channel_map={'5134_6mm-strong-meridian-left_0_0': 0, '5134_6mm-strong-meridian-left_0_1': 1, '5134_6mm-strong-meridian-left_0_2': 2, '5134_6mm-strong-meridian-left_0_3': 3, '5134_6mm-strong-meridian-left_0_4': 4, '5134_6mm-strong-meridian-left_0_5': 5, })
ukb_5134_1 = TensorMap('5134_6mm-strong-meridian-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.43788830961441, 'std': 1.551476726780896}, annotation_units=6, channel_map={'5134_6mm-strong-meridian-left_1_0': 0, '5134_6mm-strong-meridian-left_1_1': 1, '5134_6mm-strong-meridian-left_1_2': 2, '5134_6mm-strong-meridian-left_1_3': 3, '5134_6mm-strong-meridian-left_1_4': 4, '5134_6mm-strong-meridian-left_1_5': 5, })
ukb_5135_1 = TensorMap('5135_3mm-strong-meridian-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.971799947951524, 'std': 1.5838369949475675}, annotation_units=6, channel_map={'5135_3mm-strong-meridian-left_1_0': 0, '5135_3mm-strong-meridian-left_1_1': 1, '5135_3mm-strong-meridian-left_1_2': 2, '5135_3mm-strong-meridian-left_1_3': 3, '5135_3mm-strong-meridian-left_1_4': 4, '5135_3mm-strong-meridian-left_1_5': 5, })
ukb_5135_0 = TensorMap('5135_3mm-strong-meridian-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.002995503039436, 'std': 1.6117596794707862}, annotation_units=6, channel_map={'5135_3mm-strong-meridian-left_0_0': 0, '5135_3mm-strong-meridian-left_0_1': 1, '5135_3mm-strong-meridian-left_0_2': 2, '5135_3mm-strong-meridian-left_0_3': 3, '5135_3mm-strong-meridian-left_0_4': 4, '5135_3mm-strong-meridian-left_0_5': 5, })
ukb_5156_1 = TensorMap('5156_3mm-asymmetry-index-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.591561613158547, 'std': 7.322924880390097}, annotation_units=6, channel_map={'5156_3mm-asymmetry-index-left_1_0': 0, '5156_3mm-asymmetry-index-left_1_1': 1, '5156_3mm-asymmetry-index-left_1_2': 2, '5156_3mm-asymmetry-index-left_1_3': 3, '5156_3mm-asymmetry-index-left_1_4': 4, '5156_3mm-asymmetry-index-left_1_5': 5, })
ukb_5156_0 = TensorMap('5156_3mm-asymmetry-index-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.501806372008415, 'std': 7.308955928298859}, annotation_units=6, channel_map={'5156_3mm-asymmetry-index-left_0_0': 0, '5156_3mm-asymmetry-index-left_0_1': 1, '5156_3mm-asymmetry-index-left_0_2': 2, '5156_3mm-asymmetry-index-left_0_3': 3, '5156_3mm-asymmetry-index-left_0_4': 4, '5156_3mm-asymmetry-index-left_0_5': 5, })
ukb_5157_0 = TensorMap('5157_6mm-asymmetry-index-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.36216204103844, 'std': 10.177172264130801}, annotation_units=6, channel_map={'5157_6mm-asymmetry-index-left_0_0': 0, '5157_6mm-asymmetry-index-left_0_1': 1, '5157_6mm-asymmetry-index-left_0_2': 2, '5157_6mm-asymmetry-index-left_0_3': 3, '5157_6mm-asymmetry-index-left_0_4': 4, '5157_6mm-asymmetry-index-left_0_5': 5, })
ukb_5157_1 = TensorMap('5157_6mm-asymmetry-index-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.212257353564471, 'std': 10.097861543264063}, annotation_units=6, channel_map={'5157_6mm-asymmetry-index-left_1_0': 0, '5157_6mm-asymmetry-index-left_1_1': 1, '5157_6mm-asymmetry-index-left_1_2': 2, '5157_6mm-asymmetry-index-left_1_3': 3, '5157_6mm-asymmetry-index-left_1_4': 4, '5157_6mm-asymmetry-index-left_1_5': 5, })
ukb_5158_0 = TensorMap('5158_6mm-asymmetry-index-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.087403772447638, 'std': 10.207964620098544}, annotation_units=6, channel_map={'5158_6mm-asymmetry-index-right_0_0': 0, '5158_6mm-asymmetry-index-right_0_1': 1, '5158_6mm-asymmetry-index-right_0_2': 2, '5158_6mm-asymmetry-index-right_0_3': 3, '5158_6mm-asymmetry-index-right_0_4': 4, '5158_6mm-asymmetry-index-right_0_5': 5, })
ukb_5158_1 = TensorMap('5158_6mm-asymmetry-index-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.366209349593491, 'std': 9.524671853532102}, annotation_units=6, channel_map={'5158_6mm-asymmetry-index-right_1_0': 0, '5158_6mm-asymmetry-index-right_1_1': 1, '5158_6mm-asymmetry-index-right_1_2': 2, '5158_6mm-asymmetry-index-right_1_3': 3, '5158_6mm-asymmetry-index-right_1_4': 4, '5158_6mm-asymmetry-index-right_1_5': 5, })
ukb_5159_0 = TensorMap('5159_3mm-asymmetry-index-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.132884128720479, 'std': 7.381779218319648}, annotation_units=6, channel_map={'5159_3mm-asymmetry-index-right_0_0': 0, '5159_3mm-asymmetry-index-right_0_1': 1, '5159_3mm-asymmetry-index-right_0_2': 2, '5159_3mm-asymmetry-index-right_0_3': 3, '5159_3mm-asymmetry-index-right_0_4': 4, '5159_3mm-asymmetry-index-right_0_5': 5, })
ukb_5159_1 = TensorMap('5159_3mm-asymmetry-index-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.595124992577652, 'std': 6.870326665441686}, annotation_units=6, channel_map={'5159_3mm-asymmetry-index-right_1_0': 0, '5159_3mm-asymmetry-index-right_1_1': 1, '5159_3mm-asymmetry-index-right_1_2': 2, '5159_3mm-asymmetry-index-right_1_3': 3, '5159_3mm-asymmetry-index-right_1_4': 4, '5159_3mm-asymmetry-index-right_1_5': 5, })
ukb_5160_1 = TensorMap('5160_3mm-regularity-index-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.180368656303152, 'std': 1.4879505482664677}, annotation_units=6, channel_map={'5160_3mm-regularity-index-right_1_0': 0, '5160_3mm-regularity-index-right_1_1': 1, '5160_3mm-regularity-index-right_1_2': 2, '5160_3mm-regularity-index-right_1_3': 3, '5160_3mm-regularity-index-right_1_4': 4, '5160_3mm-regularity-index-right_1_5': 5, })
ukb_5160_0 = TensorMap('5160_3mm-regularity-index-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.080791996502064, 'std': 1.493267604888403}, annotation_units=6, channel_map={'5160_3mm-regularity-index-right_0_0': 0, '5160_3mm-regularity-index-right_0_1': 1, '5160_3mm-regularity-index-right_0_2': 2, '5160_3mm-regularity-index-right_0_3': 3, '5160_3mm-regularity-index-right_0_4': 4, '5160_3mm-regularity-index-right_0_5': 5, })
ukb_5161_1 = TensorMap('5161_6mm-regularity-index-right_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.75332263784514, 'std': 1.6796130507812574}, annotation_units=6, channel_map={'5161_6mm-regularity-index-right_1_0': 0, '5161_6mm-regularity-index-right_1_1': 1, '5161_6mm-regularity-index-right_1_2': 2, '5161_6mm-regularity-index-right_1_3': 3, '5161_6mm-regularity-index-right_1_4': 4, '5161_6mm-regularity-index-right_1_5': 5, })
ukb_5161_0 = TensorMap('5161_6mm-regularity-index-right_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5848690357187163, 'std': 1.7574062763765321}, annotation_units=6, channel_map={'5161_6mm-regularity-index-right_0_0': 0, '5161_6mm-regularity-index-right_0_1': 1, '5161_6mm-regularity-index-right_0_2': 2, '5161_6mm-regularity-index-right_0_3': 3, '5161_6mm-regularity-index-right_0_4': 4, '5161_6mm-regularity-index-right_0_5': 5, })
ukb_5162_0 = TensorMap('5162_6mm-regularity-index-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5843893486129437, 'std': 1.676692464550955}, annotation_units=6, channel_map={'5162_6mm-regularity-index-left_0_0': 0, '5162_6mm-regularity-index-left_0_1': 1, '5162_6mm-regularity-index-left_0_2': 2, '5162_6mm-regularity-index-left_0_3': 3, '5162_6mm-regularity-index-left_0_4': 4, '5162_6mm-regularity-index-left_0_5': 5, })
ukb_5162_1 = TensorMap('5162_6mm-regularity-index-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7130276472331594, 'std': 1.765787111651511}, annotation_units=6, channel_map={'5162_6mm-regularity-index-left_1_0': 0, '5162_6mm-regularity-index-left_1_1': 1, '5162_6mm-regularity-index-left_1_2': 2, '5162_6mm-regularity-index-left_1_3': 3, '5162_6mm-regularity-index-left_1_4': 4, '5162_6mm-regularity-index-left_1_5': 5, })
ukb_5163_0 = TensorMap('5163_3mm-regularity-index-left_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.0747083278743013, 'std': 1.4867992409125146}, annotation_units=6, channel_map={'5163_3mm-regularity-index-left_0_0': 0, '5163_3mm-regularity-index-left_0_1': 1, '5163_3mm-regularity-index-left_0_2': 2, '5163_3mm-regularity-index-left_0_3': 3, '5163_3mm-regularity-index-left_0_4': 4, '5163_3mm-regularity-index-left_0_5': 5, })
ukb_5163_1 = TensorMap('5163_3mm-regularity-index-left_1_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.202095796827358, 'std': 1.5962640710217482}, annotation_units=6, channel_map={'5163_3mm-regularity-index-left_1_0': 0, '5163_3mm-regularity-index-left_1_1': 1, '5163_3mm-regularity-index-left_1_2': 2, '5163_3mm-regularity-index-left_1_3': 3, '5163_3mm-regularity-index-left_1_4': 4, '5163_3mm-regularity-index-left_1_5': 5, })
ukb_5186_1 = TensorMap('5186_Duration-visualacuity-screen-displayed-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.539610549667238, 'std': 24.787687924345708}, annotation_units=1, channel_map={'5186_Duration-visualacuity-screen-displayed-right_1_0': 0, })
ukb_5186_0 = TensorMap('5186_Duration-visualacuity-screen-displayed-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.123006427041176, 'std': 21.512350140468634}, annotation_units=1, channel_map={'5186_Duration-visualacuity-screen-displayed-right_0_0': 0, })
ukb_5188_0 = TensorMap('5188_Duration-visualacuity-screen-displayed-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.05816385525777, 'std': 43.04048322254273}, annotation_units=1, channel_map={'5188_Duration-visualacuity-screen-displayed-left_0_0': 0, })
ukb_5188_1 = TensorMap('5188_Duration-visualacuity-screen-displayed-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 85.72695017986501, 'std': 44.10888608439217}, annotation_units=1, channel_map={'5188_Duration-visualacuity-screen-displayed-left_1_0': 0, })
ukb_5190_1 = TensorMap('5190_Duration-at-which-refractometer-first-shown-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 127.13887222555492, 'std': 51.90455662269104}, annotation_units=1, channel_map={'5190_Duration-at-which-refractometer-first-shown-right_1_0': 0, })
ukb_5190_0 = TensorMap('5190_Duration-at-which-refractometer-first-shown-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 135.90780062583843, 'std': 55.56543146059977}, annotation_units=1, channel_map={'5190_Duration-at-which-refractometer-first-shown-right_0_0': 0, })
ukb_5193_0 = TensorMap('5193_Duration-at-which-refractometer-first-shown-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 207.66262004453554, 'std': 81.13389201354755}, annotation_units=1, channel_map={'5193_Duration-at-which-refractometer-first-shown-left_0_0': 0, })
ukb_5193_1 = TensorMap('5193_Duration-at-which-refractometer-first-shown-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 192.20022986208278, 'std': 74.92628552385878}, annotation_units=1, channel_map={'5193_Duration-at-which-refractometer-first-shown-left_1_0': 0, })
ukb_5199_0 = TensorMap('5199_logMAR-initial-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03792344130072977, 'std': 0.24625813899727247}, annotation_units=1, channel_map={'5199_logMAR-initial-right_0_0': 0, })
ukb_5199_1 = TensorMap('5199_logMAR-initial-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.009058060684056823, 'std': 0.24093323889849458}, annotation_units=1, channel_map={'5199_logMAR-initial-right_1_0': 0, })
ukb_5200_0 = TensorMap('5200_Final-number-of-letters-displayed-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.985627107071228, 'std': 0.18950917702981687}, annotation_units=1, channel_map={'5200_Final-number-of-letters-displayed-right_0_0': 0, })
ukb_5200_1 = TensorMap('5200_Final-number-of-letters-displayed-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.986388160174228, 'std': 0.17871832802097412}, annotation_units=1, channel_map={'5200_Final-number-of-letters-displayed-right_1_0': 0, })
ukb_5201_1 = TensorMap('5201_logMAR-final-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.276785714285792e-05, 'std': 0.21072544173335506}, annotation_units=1, channel_map={'5201_logMAR-final-right_1_0': 0, })
ukb_5201_0 = TensorMap('5201_logMAR-final-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.027592779285683496, 'std': 0.20804208413111092}, annotation_units=1, channel_map={'5201_logMAR-final-right_0_0': 0, })
ukb_5202_0 = TensorMap('5202_Number-of-rounds-to-result-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8046942346263806, 'std': 1.3398411998093893}, annotation_units=1, channel_map={'5202_Number-of-rounds-to-result-right_0_0': 0, })
ukb_5202_1 = TensorMap('5202_Number-of-rounds-to-result-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.815324456763847, 'std': 1.2819518060503037}, annotation_units=1, channel_map={'5202_Number-of-rounds-to-result-right_1_0': 0, })
ukb_5204_0 = TensorMap('5204_Distance-of-viewer-to-screen-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.9997055434643407, 'std': 0.017587173508789494}, annotation_units=1, channel_map={'5204_Distance-of-viewer-to-screen-right_0_0': 0, })
ukb_5204_1 = TensorMap('5204_Distance-of-viewer-to-screen-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.986710884522102, 'std': 0.08042873997190708}, annotation_units=1, channel_map={'5204_Distance-of-viewer-to-screen-right_1_0': 0, })
ukb_5206_0 = TensorMap('5206_logMAR-initial-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03740222700413291, 'std': 0.2523146285723492}, annotation_units=1, channel_map={'5206_logMAR-initial-left_0_0': 0, })
ukb_5206_1 = TensorMap('5206_logMAR-initial-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.01770034153343561, 'std': 0.25213771294250026}, annotation_units=1, channel_map={'5206_logMAR-initial-left_1_0': 0, })
ukb_5207_0 = TensorMap('5207_Final-number-of-letters-displayed-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.983434095023399, 'std': 0.20249458598246797}, annotation_units=1, channel_map={'5207_Final-number-of-letters-displayed-left_0_0': 0, })
ukb_5207_1 = TensorMap('5207_Final-number-of-letters-displayed-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.983962777805274, 'std': 0.19684723462440312}, annotation_units=1, channel_map={'5207_Final-number-of-letters-displayed-left_1_0': 0, })
ukb_5208_0 = TensorMap('5208_logMAR-final-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.024298824725568272, 'std': 0.2127143307841083}, annotation_units=1, channel_map={'5208_logMAR-final-left_0_0': 0, })
ukb_5208_1 = TensorMap('5208_logMAR-final-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001989230769230767, 'std': 0.2172452201199142}, annotation_units=1, channel_map={'5208_logMAR-final-left_1_0': 0, })
ukb_5209_0 = TensorMap('5209_Number-of-rounds-to-result-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.847158178775149, 'std': 1.3276310294688691}, annotation_units=1, channel_map={'5209_Number-of-rounds-to-result-left_0_0': 0, })
ukb_5209_1 = TensorMap('5209_Number-of-rounds-to-result-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.895010395010396, 'std': 1.331650947217899}, annotation_units=1, channel_map={'5209_Number-of-rounds-to-result-left_1_0': 0, })
ukb_5211_0 = TensorMap('5211_Distance-of-viewer-to-screen-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.999705400143452, 'std': 0.017591451928771422}, annotation_units=1, channel_map={'5211_Distance-of-viewer-to-screen-left_0_0': 0, })
ukb_5211_1 = TensorMap('5211_Distance-of-viewer-to-screen-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.986661386922736, 'std': 0.08057427731191333}, annotation_units=1, channel_map={'5211_Distance-of-viewer-to-screen-left_1_0': 0, })
ukb_5215_1 = TensorMap('5215_Vertex-distance-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.0, 'std': 0.0}, annotation_units=1, channel_map={'5215_Vertex-distance-right_1_0': 0, })
ukb_5215_0 = TensorMap('5215_Vertex-distance-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.0, 'std': 0.0}, annotation_units=1, channel_map={'5215_Vertex-distance-right_0_0': 0, })
ukb_5221_0 = TensorMap('5221_Index-of-best-refractometry-result-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0269484924073338, 'std': 0.8334014421417066}, annotation_units=1, channel_map={'5221_Index-of-best-refractometry-result-right_0_0': 0, })
ukb_5221_1 = TensorMap('5221_Index-of-best-refractometry-result-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.996718453571246, 'std': 0.8311995716902664}, annotation_units=1, channel_map={'5221_Index-of-best-refractometry-result-right_1_0': 0, })
ukb_5237_0 = TensorMap('5237_3mm-index-of-best-keratometry-results-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.6441612313402285, 'std': 0.6504831779573323}, annotation_units=1, channel_map={'5237_3mm-index-of-best-keratometry-results-right_0_0': 0, })
ukb_5237_1 = TensorMap('5237_3mm-index-of-best-keratometry-results-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.6278994146975936, 'std': 0.6391368198981704}, annotation_units=1, channel_map={'5237_3mm-index-of-best-keratometry-results-right_1_0': 0, })
ukb_5251_1 = TensorMap('5251_6mm-index-of-best-keratometry-results-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.4739474517252291, 'std': 0.7218627403487591}, annotation_units=1, channel_map={'5251_6mm-index-of-best-keratometry-results-right_1_0': 0, })
ukb_5251_0 = TensorMap('5251_6mm-index-of-best-keratometry-results-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.4736809241336246, 'std': 0.7158207375462096}, annotation_units=1, channel_map={'5251_6mm-index-of-best-keratometry-results-right_0_0': 0, })
ukb_5254_1 = TensorMap('5254_Intraocular-pressure-cornealcompensated-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.524160815866345, 'std': 5.051064836257994}, annotation_units=1, channel_map={'5254_Intraocular-pressure-cornealcompensated-right_1_0': 0, })
ukb_5254_0 = TensorMap('5254_Intraocular-pressure-cornealcompensated-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.079121757020324, 'std': 4.3476889312070846}, annotation_units=1, channel_map={'5254_Intraocular-pressure-cornealcompensated-right_0_0': 0, })
ukb_5255_0 = TensorMap('5255_Intraocular-pressure-Goldmanncorrelated-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.884651187645282, 'std': 3.968793940440774}, annotation_units=1, channel_map={'5255_Intraocular-pressure-Goldmanncorrelated-right_0_0': 0, })
ukb_5255_1 = TensorMap('5255_Intraocular-pressure-Goldmanncorrelated-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.1859739660739, 'std': 4.2801812955354395}, annotation_units=1, channel_map={'5255_Intraocular-pressure-Goldmanncorrelated-right_1_0': 0, })
ukb_5256_0 = TensorMap('5256_Corneal-hysteresis-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.65828938110633, 'std': 2.3627045187390396}, annotation_units=1, channel_map={'5256_Corneal-hysteresis-right_0_0': 0, })
ukb_5256_1 = TensorMap('5256_Corneal-hysteresis-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.509265105314403, 'std': 2.425839924005888}, annotation_units=1, channel_map={'5256_Corneal-hysteresis-right_1_0': 0, })
ukb_5257_1 = TensorMap('5257_Corneal-resistance-factor-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.693469481883874, 'std': 2.471804721076224}, annotation_units=1, channel_map={'5257_Corneal-resistance-factor-right_1_0': 0, })
ukb_5257_0 = TensorMap('5257_Corneal-resistance-factor-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.73805051611582, 'std': 2.4458486957160432}, annotation_units=1, channel_map={'5257_Corneal-resistance-factor-right_0_0': 0, })
ukb_5262_1 = TensorMap('5262_Intraocular-pressure-cornealcompensated-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.996934936786918, 'std': 4.843786558284194}, annotation_units=1, channel_map={'5262_Intraocular-pressure-cornealcompensated-left_1_0': 0, })
ukb_5262_0 = TensorMap('5262_Intraocular-pressure-cornealcompensated-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.006954506069718, 'std': 4.395220730341303}, annotation_units=1, channel_map={'5262_Intraocular-pressure-cornealcompensated-left_0_0': 0, })
ukb_5263_0 = TensorMap('5263_Intraocular-pressure-Goldmanncorrelated-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.74138277251248, 'std': 4.020427502684884}, annotation_units=1, channel_map={'5263_Intraocular-pressure-Goldmanncorrelated-left_0_0': 0, })
ukb_5263_1 = TensorMap('5263_Intraocular-pressure-Goldmanncorrelated-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.05286308973173, 'std': 4.304117679285243}, annotation_units=1, channel_map={'5263_Intraocular-pressure-Goldmanncorrelated-left_1_0': 0, })
ukb_5264_0 = TensorMap('5264_Corneal-hysteresis-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.614780123091771, 'std': 2.4973445204977587}, annotation_units=1, channel_map={'5264_Corneal-hysteresis-left_0_0': 0, })
ukb_5264_1 = TensorMap('5264_Corneal-hysteresis-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.88103813341556, 'std': 2.615214926123047}, annotation_units=1, channel_map={'5264_Corneal-hysteresis-left_1_0': 0, })
ukb_5265_1 = TensorMap('5265_Corneal-resistance-factor-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.971450817144623, 'std': 2.7036353421819865}, annotation_units=1, channel_map={'5265_Corneal-resistance-factor-left_1_0': 0, })
ukb_5265_0 = TensorMap('5265_Corneal-resistance-factor-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.657931468793832, 'std': 2.580144784385491}, annotation_units=1, channel_map={'5265_Corneal-resistance-factor-left_0_0': 0, })
ukb_5274_1 = TensorMap('5274_Vertex-distance-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.0, 'std': 0.0}, annotation_units=1, channel_map={'5274_Vertex-distance-left_1_0': 0, })
ukb_5274_0 = TensorMap('5274_Vertex-distance-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.000026343056847, 'std': 0.006286035180549466}, annotation_units=1, channel_map={'5274_Vertex-distance-left_0_0': 0, })
ukb_5276_0 = TensorMap('5276_Index-of-best-refractometry-result-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0177815633726137, 'std': 0.8297585981359478}, annotation_units=1, channel_map={'5276_Index-of-best-refractometry-result-left_0_0': 0, })
ukb_5276_1 = TensorMap('5276_Index-of-best-refractometry-result-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0160482997058677, 'std': 0.828584469488342}, annotation_units=1, channel_map={'5276_Index-of-best-refractometry-result-left_1_0': 0, })
ukb_5292_1 = TensorMap('5292_3mm-index-of-best-keratometry-results-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7029713726981373, 'std': 0.595317427348043}, annotation_units=1, channel_map={'5292_3mm-index-of-best-keratometry-results-left_1_0': 0, })
ukb_5292_0 = TensorMap('5292_3mm-index-of-best-keratometry-results-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.706202205915783, 'std': 0.5957777064183039}, annotation_units=1, channel_map={'5292_3mm-index-of-best-keratometry-results-left_0_0': 0, })
ukb_5306_0 = TensorMap('5306_6mm-index-of-best-keratometry-results-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5475421186245082, 'std': 0.6728018700501439}, annotation_units=1, channel_map={'5306_6mm-index-of-best-keratometry-results-left_0_0': 0, })
ukb_5306_1 = TensorMap('5306_6mm-index-of-best-keratometry-results-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.559044324042039, 'std': 0.6746653147384141}, annotation_units=1, channel_map={'5306_6mm-index-of-best-keratometry-results-left_1_0': 0, })
ukb_5364_2 = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.024101630520816847, 'std': 0.42446281668569763}, annotation_units=1, channel_map={'5364_Average-weekly-intake-of-other-alcoholic-drinks_2_0': 0, })
ukb_5364_1 = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.021452714354694074, 'std': 0.3127982247891923}, annotation_units=1, channel_map={'5364_Average-weekly-intake-of-other-alcoholic-drinks_1_0': 0, })
ukb_5364_0 = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.038529032927324756, 'std': 0.5126326168922742}, annotation_units=1, channel_map={'5364_Average-weekly-intake-of-other-alcoholic-drinks_0_0': 0, })
ukb_5430_0 = TensorMap('5430_Age-when-loss-of-vision-due-to-injury-or-trauma-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.992546583850924, 'std': 19.249796641670024}, annotation_units=1, channel_map={'5430_Age-when-loss-of-vision-due-to-injury-or-trauma-diagnosed_0_0': 0, })
ukb_5430_2 = TensorMap('5430_Age-when-loss-of-vision-due-to-injury-or-trauma-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 43.34857142857142, 'std': 21.83845300282624}, annotation_units=1, channel_map={'5430_Age-when-loss-of-vision-due-to-injury-or-trauma-diagnosed_2_0': 0, })
ukb_5430_1 = TensorMap('5430_Age-when-loss-of-vision-due-to-injury-or-trauma-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 35.50000000000001, 'std': 20.677233857554544}, annotation_units=1, channel_map={'5430_Age-when-loss-of-vision-due-to-injury-or-trauma-diagnosed_1_0': 0, })
ukb_5901_0 = TensorMap('5901_Age-when-diabetesrelated-eye-disease-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.89264531435351, 'std': 13.567756523947297}, annotation_units=1, channel_map={'5901_Age-when-diabetesrelated-eye-disease-diagnosed_0_0': 0, })
ukb_5901_1 = TensorMap('5901_Age-when-diabetesrelated-eye-disease-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.01415094339624, 'std': 10.477906692635344}, annotation_units=1, channel_map={'5901_Age-when-diabetesrelated-eye-disease-diagnosed_1_0': 0, })
ukb_5901_2 = TensorMap('5901_Age-when-diabetesrelated-eye-disease-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.16117216117217, 'std': 12.641101907844757}, annotation_units=1, channel_map={'5901_Age-when-diabetesrelated-eye-disease-diagnosed_2_0': 0, })
ukb_5923_1 = TensorMap('5923_Age-macular-degeneration-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 60.412568306010925, 'std': 9.428127864510966}, annotation_units=1, channel_map={'5923_Age-macular-degeneration-diagnosed_1_0': 0, })
ukb_5923_0 = TensorMap('5923_Age-macular-degeneration-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.24879614767254, 'std': 10.779978213172935}, annotation_units=1, channel_map={'5923_Age-macular-degeneration-diagnosed_0_0': 0, })
ukb_5923_2 = TensorMap('5923_Age-macular-degeneration-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 62.336858006042306, 'std': 10.492838899767731}, annotation_units=1, channel_map={'5923_Age-macular-degeneration-diagnosed_2_0': 0, })
ukb_5945_0 = TensorMap('5945_Age-other-serious-eye-condition-diagnosed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.571974522293, 'std': 17.76893787108477}, annotation_units=1, channel_map={'5945_Age-other-serious-eye-condition-diagnosed_0_0': 0, })
ukb_5945_2 = TensorMap('5945_Age-other-serious-eye-condition-diagnosed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.785347043701805, 'std': 17.634613827196805}, annotation_units=1, channel_map={'5945_Age-other-serious-eye-condition-diagnosed_2_0': 0, })
ukb_5945_1 = TensorMap('5945_Age-other-serious-eye-condition-diagnosed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.41719077568133, 'std': 17.091939905661334}, annotation_units=1, channel_map={'5945_Age-other-serious-eye-condition-diagnosed_1_0': 0, })
ukb_5983_1 = TensorMap('5983_ECG-heart-rate_1_109', loss='logcosh', path_prefix='continuous', normalization={'mean': 93.24813919817296, 'std': 16.407335191112274}, annotation_units=110, channel_map={'5983_ECG-heart-rate_1_0': 0, '5983_ECG-heart-rate_1_1': 1, '5983_ECG-heart-rate_1_2': 2, '5983_ECG-heart-rate_1_3': 3, '5983_ECG-heart-rate_1_4': 4, '5983_ECG-heart-rate_1_5': 5, '5983_ECG-heart-rate_1_6': 6, '5983_ECG-heart-rate_1_7': 7, '5983_ECG-heart-rate_1_8': 8, '5983_ECG-heart-rate_1_9': 9, '5983_ECG-heart-rate_1_10': 10, '5983_ECG-heart-rate_1_11': 11, '5983_ECG-heart-rate_1_12': 12, '5983_ECG-heart-rate_1_13': 13, '5983_ECG-heart-rate_1_14': 14, '5983_ECG-heart-rate_1_15': 15, '5983_ECG-heart-rate_1_16': 16, '5983_ECG-heart-rate_1_17': 17, '5983_ECG-heart-rate_1_18': 18, '5983_ECG-heart-rate_1_19': 19, '5983_ECG-heart-rate_1_20': 20, '5983_ECG-heart-rate_1_21': 21, '5983_ECG-heart-rate_1_22': 22, '5983_ECG-heart-rate_1_23': 23, '5983_ECG-heart-rate_1_24': 24, '5983_ECG-heart-rate_1_25': 25, '5983_ECG-heart-rate_1_26': 26, '5983_ECG-heart-rate_1_27': 27, '5983_ECG-heart-rate_1_28': 28, '5983_ECG-heart-rate_1_29': 29, '5983_ECG-heart-rate_1_30': 30, '5983_ECG-heart-rate_1_31': 31, '5983_ECG-heart-rate_1_32': 32, '5983_ECG-heart-rate_1_33': 33, '5983_ECG-heart-rate_1_34': 34, '5983_ECG-heart-rate_1_35': 35, '5983_ECG-heart-rate_1_36': 36, '5983_ECG-heart-rate_1_37': 37, '5983_ECG-heart-rate_1_38': 38, '5983_ECG-heart-rate_1_39': 39, '5983_ECG-heart-rate_1_40': 40, '5983_ECG-heart-rate_1_41': 41, '5983_ECG-heart-rate_1_42': 42, '5983_ECG-heart-rate_1_43': 43, '5983_ECG-heart-rate_1_44': 44, '5983_ECG-heart-rate_1_45': 45, '5983_ECG-heart-rate_1_46': 46, '5983_ECG-heart-rate_1_47': 47, '5983_ECG-heart-rate_1_48': 48, '5983_ECG-heart-rate_1_49': 49, '5983_ECG-heart-rate_1_50': 50, '5983_ECG-heart-rate_1_51': 51, '5983_ECG-heart-rate_1_52': 52, '5983_ECG-heart-rate_1_53': 53, '5983_ECG-heart-rate_1_54': 54, '5983_ECG-heart-rate_1_55': 55, '5983_ECG-heart-rate_1_56': 56, '5983_ECG-heart-rate_1_57': 57, '5983_ECG-heart-rate_1_58': 58, '5983_ECG-heart-rate_1_59': 59, '5983_ECG-heart-rate_1_60': 60, '5983_ECG-heart-rate_1_61': 61, '5983_ECG-heart-rate_1_62': 62, '5983_ECG-heart-rate_1_63': 63, '5983_ECG-heart-rate_1_64': 64, '5983_ECG-heart-rate_1_65': 65, '5983_ECG-heart-rate_1_66': 66, '5983_ECG-heart-rate_1_67': 67, '5983_ECG-heart-rate_1_68': 68, '5983_ECG-heart-rate_1_69': 69, '5983_ECG-heart-rate_1_70': 70, '5983_ECG-heart-rate_1_71': 71, '5983_ECG-heart-rate_1_72': 72, '5983_ECG-heart-rate_1_73': 73, '5983_ECG-heart-rate_1_74': 74, '5983_ECG-heart-rate_1_75': 75, '5983_ECG-heart-rate_1_76': 76, '5983_ECG-heart-rate_1_77': 77, '5983_ECG-heart-rate_1_78': 78, '5983_ECG-heart-rate_1_79': 79, '5983_ECG-heart-rate_1_80': 80, '5983_ECG-heart-rate_1_81': 81, '5983_ECG-heart-rate_1_82': 82, '5983_ECG-heart-rate_1_83': 83, '5983_ECG-heart-rate_1_84': 84, '5983_ECG-heart-rate_1_85': 85, '5983_ECG-heart-rate_1_86': 86, '5983_ECG-heart-rate_1_87': 87, '5983_ECG-heart-rate_1_88': 88, '5983_ECG-heart-rate_1_89': 89, '5983_ECG-heart-rate_1_90': 90, '5983_ECG-heart-rate_1_91': 91, '5983_ECG-heart-rate_1_92': 92, '5983_ECG-heart-rate_1_93': 93, '5983_ECG-heart-rate_1_94': 94, '5983_ECG-heart-rate_1_95': 95, '5983_ECG-heart-rate_1_96': 96, '5983_ECG-heart-rate_1_97': 97, '5983_ECG-heart-rate_1_98': 98, '5983_ECG-heart-rate_1_99': 99, '5983_ECG-heart-rate_1_100': 100, '5983_ECG-heart-rate_1_101': 101, '5983_ECG-heart-rate_1_102': 102, '5983_ECG-heart-rate_1_103': 103, '5983_ECG-heart-rate_1_104': 104, '5983_ECG-heart-rate_1_105': 105, '5983_ECG-heart-rate_1_106': 106, '5983_ECG-heart-rate_1_107': 107, '5983_ECG-heart-rate_1_108': 108, '5983_ECG-heart-rate_1_109': 109, })
ukb_5983_0 = TensorMap('5983_ECG-heart-rate_0_111', loss='logcosh', path_prefix='continuous', normalization={'mean': 95.64040545394423, 'std': 16.575427118260908}, annotation_units=112, channel_map={'5983_ECG-heart-rate_0_0': 0, '5983_ECG-heart-rate_0_1': 1, '5983_ECG-heart-rate_0_2': 2, '5983_ECG-heart-rate_0_3': 3, '5983_ECG-heart-rate_0_4': 4, '5983_ECG-heart-rate_0_5': 5, '5983_ECG-heart-rate_0_6': 6, '5983_ECG-heart-rate_0_7': 7, '5983_ECG-heart-rate_0_8': 8, '5983_ECG-heart-rate_0_9': 9, '5983_ECG-heart-rate_0_10': 10, '5983_ECG-heart-rate_0_11': 11, '5983_ECG-heart-rate_0_12': 12, '5983_ECG-heart-rate_0_13': 13, '5983_ECG-heart-rate_0_14': 14, '5983_ECG-heart-rate_0_15': 15, '5983_ECG-heart-rate_0_16': 16, '5983_ECG-heart-rate_0_17': 17, '5983_ECG-heart-rate_0_18': 18, '5983_ECG-heart-rate_0_19': 19, '5983_ECG-heart-rate_0_20': 20, '5983_ECG-heart-rate_0_21': 21, '5983_ECG-heart-rate_0_22': 22, '5983_ECG-heart-rate_0_23': 23, '5983_ECG-heart-rate_0_24': 24, '5983_ECG-heart-rate_0_25': 25, '5983_ECG-heart-rate_0_26': 26, '5983_ECG-heart-rate_0_27': 27, '5983_ECG-heart-rate_0_28': 28, '5983_ECG-heart-rate_0_29': 29, '5983_ECG-heart-rate_0_30': 30, '5983_ECG-heart-rate_0_31': 31, '5983_ECG-heart-rate_0_32': 32, '5983_ECG-heart-rate_0_33': 33, '5983_ECG-heart-rate_0_34': 34, '5983_ECG-heart-rate_0_35': 35, '5983_ECG-heart-rate_0_36': 36, '5983_ECG-heart-rate_0_37': 37, '5983_ECG-heart-rate_0_38': 38, '5983_ECG-heart-rate_0_39': 39, '5983_ECG-heart-rate_0_40': 40, '5983_ECG-heart-rate_0_41': 41, '5983_ECG-heart-rate_0_42': 42, '5983_ECG-heart-rate_0_43': 43, '5983_ECG-heart-rate_0_44': 44, '5983_ECG-heart-rate_0_45': 45, '5983_ECG-heart-rate_0_46': 46, '5983_ECG-heart-rate_0_47': 47, '5983_ECG-heart-rate_0_48': 48, '5983_ECG-heart-rate_0_49': 49, '5983_ECG-heart-rate_0_50': 50, '5983_ECG-heart-rate_0_51': 51, '5983_ECG-heart-rate_0_52': 52, '5983_ECG-heart-rate_0_53': 53, '5983_ECG-heart-rate_0_54': 54, '5983_ECG-heart-rate_0_55': 55, '5983_ECG-heart-rate_0_56': 56, '5983_ECG-heart-rate_0_57': 57, '5983_ECG-heart-rate_0_58': 58, '5983_ECG-heart-rate_0_59': 59, '5983_ECG-heart-rate_0_60': 60, '5983_ECG-heart-rate_0_61': 61, '5983_ECG-heart-rate_0_62': 62, '5983_ECG-heart-rate_0_63': 63, '5983_ECG-heart-rate_0_64': 64, '5983_ECG-heart-rate_0_65': 65, '5983_ECG-heart-rate_0_66': 66, '5983_ECG-heart-rate_0_67': 67, '5983_ECG-heart-rate_0_68': 68, '5983_ECG-heart-rate_0_69': 69, '5983_ECG-heart-rate_0_70': 70, '5983_ECG-heart-rate_0_71': 71, '5983_ECG-heart-rate_0_72': 72, '5983_ECG-heart-rate_0_73': 73, '5983_ECG-heart-rate_0_74': 74, '5983_ECG-heart-rate_0_75': 75, '5983_ECG-heart-rate_0_76': 76, '5983_ECG-heart-rate_0_77': 77, '5983_ECG-heart-rate_0_78': 78, '5983_ECG-heart-rate_0_79': 79, '5983_ECG-heart-rate_0_80': 80, '5983_ECG-heart-rate_0_81': 81, '5983_ECG-heart-rate_0_82': 82, '5983_ECG-heart-rate_0_83': 83, '5983_ECG-heart-rate_0_84': 84, '5983_ECG-heart-rate_0_85': 85, '5983_ECG-heart-rate_0_86': 86, '5983_ECG-heart-rate_0_87': 87, '5983_ECG-heart-rate_0_88': 88, '5983_ECG-heart-rate_0_89': 89, '5983_ECG-heart-rate_0_90': 90, '5983_ECG-heart-rate_0_91': 91, '5983_ECG-heart-rate_0_92': 92, '5983_ECG-heart-rate_0_93': 93, '5983_ECG-heart-rate_0_94': 94, '5983_ECG-heart-rate_0_95': 95, '5983_ECG-heart-rate_0_96': 96, '5983_ECG-heart-rate_0_97': 97, '5983_ECG-heart-rate_0_98': 98, '5983_ECG-heart-rate_0_99': 99, '5983_ECG-heart-rate_0_100': 100, '5983_ECG-heart-rate_0_101': 101, '5983_ECG-heart-rate_0_102': 102, '5983_ECG-heart-rate_0_103': 103, '5983_ECG-heart-rate_0_104': 104, '5983_ECG-heart-rate_0_105': 105, '5983_ECG-heart-rate_0_106': 106, '5983_ECG-heart-rate_0_107': 107, '5983_ECG-heart-rate_0_108': 108, '5983_ECG-heart-rate_0_109': 109, '5983_ECG-heart-rate_0_110': 110, '5983_ECG-heart-rate_0_111': 111, })
ukb_5984_1 = TensorMap('5984_ECG-load_1_109', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.868254139305364, 'std': 27.877069728521047}, annotation_units=110, channel_map={'5984_ECG-load_1_0': 0, '5984_ECG-load_1_1': 1, '5984_ECG-load_1_2': 2, '5984_ECG-load_1_3': 3, '5984_ECG-load_1_4': 4, '5984_ECG-load_1_5': 5, '5984_ECG-load_1_6': 6, '5984_ECG-load_1_7': 7, '5984_ECG-load_1_8': 8, '5984_ECG-load_1_9': 9, '5984_ECG-load_1_10': 10, '5984_ECG-load_1_11': 11, '5984_ECG-load_1_12': 12, '5984_ECG-load_1_13': 13, '5984_ECG-load_1_14': 14, '5984_ECG-load_1_15': 15, '5984_ECG-load_1_16': 16, '5984_ECG-load_1_17': 17, '5984_ECG-load_1_18': 18, '5984_ECG-load_1_19': 19, '5984_ECG-load_1_20': 20, '5984_ECG-load_1_21': 21, '5984_ECG-load_1_22': 22, '5984_ECG-load_1_23': 23, '5984_ECG-load_1_24': 24, '5984_ECG-load_1_25': 25, '5984_ECG-load_1_26': 26, '5984_ECG-load_1_27': 27, '5984_ECG-load_1_28': 28, '5984_ECG-load_1_29': 29, '5984_ECG-load_1_30': 30, '5984_ECG-load_1_31': 31, '5984_ECG-load_1_32': 32, '5984_ECG-load_1_33': 33, '5984_ECG-load_1_34': 34, '5984_ECG-load_1_35': 35, '5984_ECG-load_1_36': 36, '5984_ECG-load_1_37': 37, '5984_ECG-load_1_38': 38, '5984_ECG-load_1_39': 39, '5984_ECG-load_1_40': 40, '5984_ECG-load_1_41': 41, '5984_ECG-load_1_42': 42, '5984_ECG-load_1_43': 43, '5984_ECG-load_1_44': 44, '5984_ECG-load_1_45': 45, '5984_ECG-load_1_46': 46, '5984_ECG-load_1_47': 47, '5984_ECG-load_1_48': 48, '5984_ECG-load_1_49': 49, '5984_ECG-load_1_50': 50, '5984_ECG-load_1_51': 51, '5984_ECG-load_1_52': 52, '5984_ECG-load_1_53': 53, '5984_ECG-load_1_54': 54, '5984_ECG-load_1_55': 55, '5984_ECG-load_1_56': 56, '5984_ECG-load_1_57': 57, '5984_ECG-load_1_58': 58, '5984_ECG-load_1_59': 59, '5984_ECG-load_1_60': 60, '5984_ECG-load_1_61': 61, '5984_ECG-load_1_62': 62, '5984_ECG-load_1_63': 63, '5984_ECG-load_1_64': 64, '5984_ECG-load_1_65': 65, '5984_ECG-load_1_66': 66, '5984_ECG-load_1_67': 67, '5984_ECG-load_1_68': 68, '5984_ECG-load_1_69': 69, '5984_ECG-load_1_70': 70, '5984_ECG-load_1_71': 71, '5984_ECG-load_1_72': 72, '5984_ECG-load_1_73': 73, '5984_ECG-load_1_74': 74, '5984_ECG-load_1_75': 75, '5984_ECG-load_1_76': 76, '5984_ECG-load_1_77': 77, '5984_ECG-load_1_78': 78, '5984_ECG-load_1_79': 79, '5984_ECG-load_1_80': 80, '5984_ECG-load_1_81': 81, '5984_ECG-load_1_82': 82, '5984_ECG-load_1_83': 83, '5984_ECG-load_1_84': 84, '5984_ECG-load_1_85': 85, '5984_ECG-load_1_86': 86, '5984_ECG-load_1_87': 87, '5984_ECG-load_1_88': 88, '5984_ECG-load_1_89': 89, '5984_ECG-load_1_90': 90, '5984_ECG-load_1_91': 91, '5984_ECG-load_1_92': 92, '5984_ECG-load_1_93': 93, '5984_ECG-load_1_94': 94, '5984_ECG-load_1_95': 95, '5984_ECG-load_1_96': 96, '5984_ECG-load_1_97': 97, '5984_ECG-load_1_98': 98, '5984_ECG-load_1_99': 99, '5984_ECG-load_1_100': 100, '5984_ECG-load_1_101': 101, '5984_ECG-load_1_102': 102, '5984_ECG-load_1_103': 103, '5984_ECG-load_1_104': 104, '5984_ECG-load_1_105': 105, '5984_ECG-load_1_106': 106, '5984_ECG-load_1_107': 107, '5984_ECG-load_1_108': 108, '5984_ECG-load_1_109': 109, })
ukb_5984_0 = TensorMap('5984_ECG-load_0_113', loss='logcosh', path_prefix='continuous', normalization={'mean': 48.350435331279996, 'std': 28.925085323783993}, annotation_units=114, channel_map={'5984_ECG-load_0_0': 0, '5984_ECG-load_0_1': 1, '5984_ECG-load_0_2': 2, '5984_ECG-load_0_3': 3, '5984_ECG-load_0_4': 4, '5984_ECG-load_0_5': 5, '5984_ECG-load_0_6': 6, '5984_ECG-load_0_7': 7, '5984_ECG-load_0_8': 8, '5984_ECG-load_0_9': 9, '5984_ECG-load_0_10': 10, '5984_ECG-load_0_11': 11, '5984_ECG-load_0_12': 12, '5984_ECG-load_0_13': 13, '5984_ECG-load_0_14': 14, '5984_ECG-load_0_15': 15, '5984_ECG-load_0_16': 16, '5984_ECG-load_0_17': 17, '5984_ECG-load_0_18': 18, '5984_ECG-load_0_19': 19, '5984_ECG-load_0_20': 20, '5984_ECG-load_0_21': 21, '5984_ECG-load_0_22': 22, '5984_ECG-load_0_23': 23, '5984_ECG-load_0_24': 24, '5984_ECG-load_0_25': 25, '5984_ECG-load_0_26': 26, '5984_ECG-load_0_27': 27, '5984_ECG-load_0_28': 28, '5984_ECG-load_0_29': 29, '5984_ECG-load_0_30': 30, '5984_ECG-load_0_31': 31, '5984_ECG-load_0_32': 32, '5984_ECG-load_0_33': 33, '5984_ECG-load_0_34': 34, '5984_ECG-load_0_35': 35, '5984_ECG-load_0_36': 36, '5984_ECG-load_0_37': 37, '5984_ECG-load_0_38': 38, '5984_ECG-load_0_39': 39, '5984_ECG-load_0_40': 40, '5984_ECG-load_0_41': 41, '5984_ECG-load_0_42': 42, '5984_ECG-load_0_43': 43, '5984_ECG-load_0_44': 44, '5984_ECG-load_0_45': 45, '5984_ECG-load_0_46': 46, '5984_ECG-load_0_47': 47, '5984_ECG-load_0_48': 48, '5984_ECG-load_0_49': 49, '5984_ECG-load_0_50': 50, '5984_ECG-load_0_51': 51, '5984_ECG-load_0_52': 52, '5984_ECG-load_0_53': 53, '5984_ECG-load_0_54': 54, '5984_ECG-load_0_55': 55, '5984_ECG-load_0_56': 56, '5984_ECG-load_0_57': 57, '5984_ECG-load_0_58': 58, '5984_ECG-load_0_59': 59, '5984_ECG-load_0_60': 60, '5984_ECG-load_0_61': 61, '5984_ECG-load_0_62': 62, '5984_ECG-load_0_63': 63, '5984_ECG-load_0_64': 64, '5984_ECG-load_0_65': 65, '5984_ECG-load_0_66': 66, '5984_ECG-load_0_67': 67, '5984_ECG-load_0_68': 68, '5984_ECG-load_0_69': 69, '5984_ECG-load_0_70': 70, '5984_ECG-load_0_71': 71, '5984_ECG-load_0_72': 72, '5984_ECG-load_0_73': 73, '5984_ECG-load_0_74': 74, '5984_ECG-load_0_75': 75, '5984_ECG-load_0_76': 76, '5984_ECG-load_0_77': 77, '5984_ECG-load_0_78': 78, '5984_ECG-load_0_79': 79, '5984_ECG-load_0_80': 80, '5984_ECG-load_0_81': 81, '5984_ECG-load_0_82': 82, '5984_ECG-load_0_83': 83, '5984_ECG-load_0_84': 84, '5984_ECG-load_0_85': 85, '5984_ECG-load_0_86': 86, '5984_ECG-load_0_87': 87, '5984_ECG-load_0_88': 88, '5984_ECG-load_0_89': 89, '5984_ECG-load_0_90': 90, '5984_ECG-load_0_91': 91, '5984_ECG-load_0_92': 92, '5984_ECG-load_0_93': 93, '5984_ECG-load_0_94': 94, '5984_ECG-load_0_95': 95, '5984_ECG-load_0_96': 96, '5984_ECG-load_0_97': 97, '5984_ECG-load_0_98': 98, '5984_ECG-load_0_99': 99, '5984_ECG-load_0_100': 100, '5984_ECG-load_0_101': 101, '5984_ECG-load_0_102': 102, '5984_ECG-load_0_103': 103, '5984_ECG-load_0_104': 104, '5984_ECG-load_0_105': 105, '5984_ECG-load_0_106': 106, '5984_ECG-load_0_107': 107, '5984_ECG-load_0_108': 108, '5984_ECG-load_0_109': 109, '5984_ECG-load_0_110': 110, '5984_ECG-load_0_111': 111, '5984_ECG-load_0_112': 112, '5984_ECG-load_0_113': 113, })
ukb_5986_1 = TensorMap('5986_ECG-phase-time_1_109', loss='logcosh', path_prefix='continuous', normalization={'mean': 177.64182131027354, 'std': 110.88179832863214}, annotation_units=110, channel_map={'5986_ECG-phase-time_1_0': 0, '5986_ECG-phase-time_1_1': 1, '5986_ECG-phase-time_1_2': 2, '5986_ECG-phase-time_1_3': 3, '5986_ECG-phase-time_1_4': 4, '5986_ECG-phase-time_1_5': 5, '5986_ECG-phase-time_1_6': 6, '5986_ECG-phase-time_1_7': 7, '5986_ECG-phase-time_1_8': 8, '5986_ECG-phase-time_1_9': 9, '5986_ECG-phase-time_1_10': 10, '5986_ECG-phase-time_1_11': 11, '5986_ECG-phase-time_1_12': 12, '5986_ECG-phase-time_1_13': 13, '5986_ECG-phase-time_1_14': 14, '5986_ECG-phase-time_1_15': 15, '5986_ECG-phase-time_1_16': 16, '5986_ECG-phase-time_1_17': 17, '5986_ECG-phase-time_1_18': 18, '5986_ECG-phase-time_1_19': 19, '5986_ECG-phase-time_1_20': 20, '5986_ECG-phase-time_1_21': 21, '5986_ECG-phase-time_1_22': 22, '5986_ECG-phase-time_1_23': 23, '5986_ECG-phase-time_1_24': 24, '5986_ECG-phase-time_1_25': 25, '5986_ECG-phase-time_1_26': 26, '5986_ECG-phase-time_1_27': 27, '5986_ECG-phase-time_1_28': 28, '5986_ECG-phase-time_1_29': 29, '5986_ECG-phase-time_1_30': 30, '5986_ECG-phase-time_1_31': 31, '5986_ECG-phase-time_1_32': 32, '5986_ECG-phase-time_1_33': 33, '5986_ECG-phase-time_1_34': 34, '5986_ECG-phase-time_1_35': 35, '5986_ECG-phase-time_1_36': 36, '5986_ECG-phase-time_1_37': 37, '5986_ECG-phase-time_1_38': 38, '5986_ECG-phase-time_1_39': 39, '5986_ECG-phase-time_1_40': 40, '5986_ECG-phase-time_1_41': 41, '5986_ECG-phase-time_1_42': 42, '5986_ECG-phase-time_1_43': 43, '5986_ECG-phase-time_1_44': 44, '5986_ECG-phase-time_1_45': 45, '5986_ECG-phase-time_1_46': 46, '5986_ECG-phase-time_1_47': 47, '5986_ECG-phase-time_1_48': 48, '5986_ECG-phase-time_1_49': 49, '5986_ECG-phase-time_1_50': 50, '5986_ECG-phase-time_1_51': 51, '5986_ECG-phase-time_1_52': 52, '5986_ECG-phase-time_1_53': 53, '5986_ECG-phase-time_1_54': 54, '5986_ECG-phase-time_1_55': 55, '5986_ECG-phase-time_1_56': 56, '5986_ECG-phase-time_1_57': 57, '5986_ECG-phase-time_1_58': 58, '5986_ECG-phase-time_1_59': 59, '5986_ECG-phase-time_1_60': 60, '5986_ECG-phase-time_1_61': 61, '5986_ECG-phase-time_1_62': 62, '5986_ECG-phase-time_1_63': 63, '5986_ECG-phase-time_1_64': 64, '5986_ECG-phase-time_1_65': 65, '5986_ECG-phase-time_1_66': 66, '5986_ECG-phase-time_1_67': 67, '5986_ECG-phase-time_1_68': 68, '5986_ECG-phase-time_1_69': 69, '5986_ECG-phase-time_1_70': 70, '5986_ECG-phase-time_1_71': 71, '5986_ECG-phase-time_1_72': 72, '5986_ECG-phase-time_1_73': 73, '5986_ECG-phase-time_1_74': 74, '5986_ECG-phase-time_1_75': 75, '5986_ECG-phase-time_1_76': 76, '5986_ECG-phase-time_1_77': 77, '5986_ECG-phase-time_1_78': 78, '5986_ECG-phase-time_1_79': 79, '5986_ECG-phase-time_1_80': 80, '5986_ECG-phase-time_1_81': 81, '5986_ECG-phase-time_1_82': 82, '5986_ECG-phase-time_1_83': 83, '5986_ECG-phase-time_1_84': 84, '5986_ECG-phase-time_1_85': 85, '5986_ECG-phase-time_1_86': 86, '5986_ECG-phase-time_1_87': 87, '5986_ECG-phase-time_1_88': 88, '5986_ECG-phase-time_1_89': 89, '5986_ECG-phase-time_1_90': 90, '5986_ECG-phase-time_1_91': 91, '5986_ECG-phase-time_1_92': 92, '5986_ECG-phase-time_1_93': 93, '5986_ECG-phase-time_1_94': 94, '5986_ECG-phase-time_1_95': 95, '5986_ECG-phase-time_1_96': 96, '5986_ECG-phase-time_1_97': 97, '5986_ECG-phase-time_1_98': 98, '5986_ECG-phase-time_1_99': 99, '5986_ECG-phase-time_1_100': 100, '5986_ECG-phase-time_1_101': 101, '5986_ECG-phase-time_1_102': 102, '5986_ECG-phase-time_1_103': 103, '5986_ECG-phase-time_1_104': 104, '5986_ECG-phase-time_1_105': 105, '5986_ECG-phase-time_1_106': 106, '5986_ECG-phase-time_1_107': 107, '5986_ECG-phase-time_1_108': 108, '5986_ECG-phase-time_1_109': 109, })
ukb_5986_0 = TensorMap('5986_ECG-phase-time_0_113', loss='logcosh', path_prefix='continuous', normalization={'mean': 179.18541495338116, 'std': 110.51368871605243}, annotation_units=114, channel_map={'5986_ECG-phase-time_0_0': 0, '5986_ECG-phase-time_0_1': 1, '5986_ECG-phase-time_0_2': 2, '5986_ECG-phase-time_0_3': 3, '5986_ECG-phase-time_0_4': 4, '5986_ECG-phase-time_0_5': 5, '5986_ECG-phase-time_0_6': 6, '5986_ECG-phase-time_0_7': 7, '5986_ECG-phase-time_0_8': 8, '5986_ECG-phase-time_0_9': 9, '5986_ECG-phase-time_0_10': 10, '5986_ECG-phase-time_0_11': 11, '5986_ECG-phase-time_0_12': 12, '5986_ECG-phase-time_0_13': 13, '5986_ECG-phase-time_0_14': 14, '5986_ECG-phase-time_0_15': 15, '5986_ECG-phase-time_0_16': 16, '5986_ECG-phase-time_0_17': 17, '5986_ECG-phase-time_0_18': 18, '5986_ECG-phase-time_0_19': 19, '5986_ECG-phase-time_0_20': 20, '5986_ECG-phase-time_0_21': 21, '5986_ECG-phase-time_0_22': 22, '5986_ECG-phase-time_0_23': 23, '5986_ECG-phase-time_0_24': 24, '5986_ECG-phase-time_0_25': 25, '5986_ECG-phase-time_0_26': 26, '5986_ECG-phase-time_0_27': 27, '5986_ECG-phase-time_0_28': 28, '5986_ECG-phase-time_0_29': 29, '5986_ECG-phase-time_0_30': 30, '5986_ECG-phase-time_0_31': 31, '5986_ECG-phase-time_0_32': 32, '5986_ECG-phase-time_0_33': 33, '5986_ECG-phase-time_0_34': 34, '5986_ECG-phase-time_0_35': 35, '5986_ECG-phase-time_0_36': 36, '5986_ECG-phase-time_0_37': 37, '5986_ECG-phase-time_0_38': 38, '5986_ECG-phase-time_0_39': 39, '5986_ECG-phase-time_0_40': 40, '5986_ECG-phase-time_0_41': 41, '5986_ECG-phase-time_0_42': 42, '5986_ECG-phase-time_0_43': 43, '5986_ECG-phase-time_0_44': 44, '5986_ECG-phase-time_0_45': 45, '5986_ECG-phase-time_0_46': 46, '5986_ECG-phase-time_0_47': 47, '5986_ECG-phase-time_0_48': 48, '5986_ECG-phase-time_0_49': 49, '5986_ECG-phase-time_0_50': 50, '5986_ECG-phase-time_0_51': 51, '5986_ECG-phase-time_0_52': 52, '5986_ECG-phase-time_0_53': 53, '5986_ECG-phase-time_0_54': 54, '5986_ECG-phase-time_0_55': 55, '5986_ECG-phase-time_0_56': 56, '5986_ECG-phase-time_0_57': 57, '5986_ECG-phase-time_0_58': 58, '5986_ECG-phase-time_0_59': 59, '5986_ECG-phase-time_0_60': 60, '5986_ECG-phase-time_0_61': 61, '5986_ECG-phase-time_0_62': 62, '5986_ECG-phase-time_0_63': 63, '5986_ECG-phase-time_0_64': 64, '5986_ECG-phase-time_0_65': 65, '5986_ECG-phase-time_0_66': 66, '5986_ECG-phase-time_0_67': 67, '5986_ECG-phase-time_0_68': 68, '5986_ECG-phase-time_0_69': 69, '5986_ECG-phase-time_0_70': 70, '5986_ECG-phase-time_0_71': 71, '5986_ECG-phase-time_0_72': 72, '5986_ECG-phase-time_0_73': 73, '5986_ECG-phase-time_0_74': 74, '5986_ECG-phase-time_0_75': 75, '5986_ECG-phase-time_0_76': 76, '5986_ECG-phase-time_0_77': 77, '5986_ECG-phase-time_0_78': 78, '5986_ECG-phase-time_0_79': 79, '5986_ECG-phase-time_0_80': 80, '5986_ECG-phase-time_0_81': 81, '5986_ECG-phase-time_0_82': 82, '5986_ECG-phase-time_0_83': 83, '5986_ECG-phase-time_0_84': 84, '5986_ECG-phase-time_0_85': 85, '5986_ECG-phase-time_0_86': 86, '5986_ECG-phase-time_0_87': 87, '5986_ECG-phase-time_0_88': 88, '5986_ECG-phase-time_0_89': 89, '5986_ECG-phase-time_0_90': 90, '5986_ECG-phase-time_0_91': 91, '5986_ECG-phase-time_0_92': 92, '5986_ECG-phase-time_0_93': 93, '5986_ECG-phase-time_0_94': 94, '5986_ECG-phase-time_0_95': 95, '5986_ECG-phase-time_0_96': 96, '5986_ECG-phase-time_0_97': 97, '5986_ECG-phase-time_0_98': 98, '5986_ECG-phase-time_0_99': 99, '5986_ECG-phase-time_0_100': 100, '5986_ECG-phase-time_0_101': 101, '5986_ECG-phase-time_0_102': 102, '5986_ECG-phase-time_0_103': 103, '5986_ECG-phase-time_0_104': 104, '5986_ECG-phase-time_0_105': 105, '5986_ECG-phase-time_0_106': 106, '5986_ECG-phase-time_0_107': 107, '5986_ECG-phase-time_0_108': 108, '5986_ECG-phase-time_0_109': 109, '5986_ECG-phase-time_0_110': 110, '5986_ECG-phase-time_0_111': 111, '5986_ECG-phase-time_0_112': 112, '5986_ECG-phase-time_0_113': 113, })
ukb_5992_1 = TensorMap('5992_ECG-phase-duration_1_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 131.13946695874407, 'std': 147.0804538523044}, annotation_units=3, channel_map={'5992_ECG-phase-duration_1_0': 0, '5992_ECG-phase-duration_1_1': 1, '5992_ECG-phase-duration_1_2': 2, })
ukb_5992_0 = TensorMap('5992_ECG-phase-duration_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 128.69241195888281, 'std': 145.77693042814843}, annotation_units=3, channel_map={'5992_ECG-phase-duration_0_0': 0, '5992_ECG-phase-duration_0_1': 1, '5992_ECG-phase-duration_0_2': 2, })
ukb_5993_1 = TensorMap('5993_ECG-number-of-stages-in-a-phase_1_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.709333200570877, 'std': 8.581536168729823}, annotation_units=3, channel_map={'5993_ECG-number-of-stages-in-a-phase_1_0': 0, '5993_ECG-number-of-stages-in-a-phase_1_1': 1, '5993_ECG-number-of-stages-in-a-phase_1_2': 2, })
ukb_5993_0 = TensorMap('5993_ECG-number-of-stages-in-a-phase_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.942142576154065, 'std': 9.066003869976084}, annotation_units=3, channel_map={'5993_ECG-number-of-stages-in-a-phase_0_0': 0, '5993_ECG-number-of-stages-in-a-phase_0_1': 1, '5993_ECG-number-of-stages-in-a-phase_0_2': 2, })
ukb_6032_1 = TensorMap('6032_Maximum-workload-during-fitness-test_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 70.27695907597332, 'std': 32.530845764395586}, annotation_units=1, channel_map={'6032_Maximum-workload-during-fitness-test_1_0': 0, })
ukb_6032_0 = TensorMap('6032_Maximum-workload-during-fitness-test_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 71.65201718940419, 'std': 35.67292137350584}, annotation_units=1, channel_map={'6032_Maximum-workload-during-fitness-test_0_0': 0, })
ukb_6033_1 = TensorMap('6033_Maximum-heart-rate-during-fitness-test_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 109.18266932270916, 'std': 19.63776497679266}, annotation_units=1, channel_map={'6033_Maximum-heart-rate-during-fitness-test_1_0': 0, })
ukb_6033_0 = TensorMap('6033_Maximum-heart-rate-during-fitness-test_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 110.24494797213205, 'std': 20.14191123730601}, annotation_units=1, channel_map={'6033_Maximum-heart-rate-during-fitness-test_0_0': 0, })
ukb_6038_0 = TensorMap('6038_Number-of-trend-entries_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.75908204628251, 'std': 25.778401312237246}, annotation_units=1, channel_map={'6038_Number-of-trend-entries_0_0': 0, })
ukb_6038_1 = TensorMap('6038_Number-of-trend-entries_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.678880812506215, 'std': 23.69787090539729}, annotation_units=1, channel_map={'6038_Number-of-trend-entries_1_0': 0, })
ukb_6039_0 = TensorMap('6039_Duration-of-fitness-test_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 386.3728641797937, 'std': 107.8361824926779}, annotation_units=1, channel_map={'6039_Duration-of-fitness-test_0_0': 0, })
ukb_6039_1 = TensorMap('6039_Duration-of-fitness-test_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 393.7038235586976, 'std': 99.89144295544399}, annotation_units=1, channel_map={'6039_Duration-of-fitness-test_1_0': 0, })
ukb_6071_0 = TensorMap('6071_Duration-at-which-OCT-screen-shown-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 306.49513701322576, 'std': 106.94475390518122}, annotation_units=1, channel_map={'6071_Duration-at-which-OCT-screen-shown-right_0_0': 0, })
ukb_6071_1 = TensorMap('6071_Duration-at-which-OCT-screen-shown-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 295.5291441711657, 'std': 103.45635030284373}, annotation_units=1, channel_map={'6071_Duration-at-which-OCT-screen-shown-right_1_0': 0, })
ukb_6073_1 = TensorMap('6073_Duration-at-which-OCT-screen-shown-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 493.80301818908663, 'std': 139.49730839843235}, annotation_units=1, channel_map={'6073_Duration-at-which-OCT-screen-shown-left_1_0': 0, })
ukb_6073_0 = TensorMap('6073_Duration-at-which-OCT-screen-shown-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 489.35185584397465, 'std': 165.82155361810305}, annotation_units=1, channel_map={'6073_Duration-at-which-OCT-screen-shown-left_0_0': 0, })
ukb_6183_2 = TensorMap('6183_Number-of-cigarettes-previously-smoked-daily-current-cigarpipe-smokers_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.8030303030303, 'std': 7.453962236978601}, annotation_units=1, channel_map={'6183_Number-of-cigarettes-previously-smoked-daily-current-cigarpipe-smokers_2_0': 0, })
ukb_6183_1 = TensorMap('6183_Number-of-cigarettes-previously-smoked-daily-current-cigarpipe-smokers_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.646341463414636, 'std': 8.288730227271715}, annotation_units=1, channel_map={'6183_Number-of-cigarettes-previously-smoked-daily-current-cigarpipe-smokers_1_0': 0, })
ukb_6183_0 = TensorMap('6183_Number-of-cigarettes-previously-smoked-daily-current-cigarpipe-smokers_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.75287356321839, 'std': 10.459531687591946}, annotation_units=1, channel_map={'6183_Number-of-cigarettes-previously-smoked-daily-current-cigarpipe-smokers_0_0': 0, })
ukb_6194_2 = TensorMap('6194_Age-stopped-smoking-cigarettes-current-cigarpipe-or-previous-cigarette-smoker_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 36.6875, 'std': 11.42277718900607}, annotation_units=1, channel_map={'6194_Age-stopped-smoking-cigarettes-current-cigarpipe-or-previous-cigarette-smoker_2_0': 0, })
ukb_6194_0 = TensorMap('6194_Age-stopped-smoking-cigarettes-current-cigarpipe-or-previous-cigarette-smoker_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 37.70771950500883, 'std': 11.806893295410129}, annotation_units=1, channel_map={'6194_Age-stopped-smoking-cigarettes-current-cigarpipe-or-previous-cigarette-smoker_0_0': 0, })
ukb_6194_1 = TensorMap('6194_Age-stopped-smoking-cigarettes-current-cigarpipe-or-previous-cigarette-smoker_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 36.73170731707318, 'std': 12.27400584618547}, annotation_units=1, channel_map={'6194_Age-stopped-smoking-cigarettes-current-cigarpipe-or-previous-cigarette-smoker_1_0': 0, })
ukb_10694_0 = TensorMap('10694_Forced-vital-capacity-FVC-pilot_0_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.6323018053375193, 'std': 1.0726680201026209}, annotation_units=5, channel_map={'10694_Forced-vital-capacity-FVC-pilot_0_0': 0, '10694_Forced-vital-capacity-FVC-pilot_0_1': 1, '10694_Forced-vital-capacity-FVC-pilot_0_2': 2, '10694_Forced-vital-capacity-FVC-pilot_0_3': 3, '10694_Forced-vital-capacity-FVC-pilot_0_4': 4, })
ukb_10695_0 = TensorMap('10695_Forced-expiratory-volume-in-1second-FEV1-pilot_0_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.772009419152276, 'std': 0.804651649484746}, annotation_units=5, channel_map={'10695_Forced-expiratory-volume-in-1second-FEV1-pilot_0_0': 0, '10695_Forced-expiratory-volume-in-1second-FEV1-pilot_0_1': 1, '10695_Forced-expiratory-volume-in-1second-FEV1-pilot_0_2': 2, '10695_Forced-expiratory-volume-in-1second-FEV1-pilot_0_3': 3, '10695_Forced-expiratory-volume-in-1second-FEV1-pilot_0_4': 4, })
ukb_10696_0 = TensorMap('10696_Peak-expiratory-flow-PEF-pilot_0_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 389.86371664050233, 'std': 140.95580121591217}, annotation_units=5, channel_map={'10696_Peak-expiratory-flow-PEF-pilot_0_0': 0, '10696_Peak-expiratory-flow-PEF-pilot_0_1': 1, '10696_Peak-expiratory-flow-PEF-pilot_0_2': 2, '10696_Peak-expiratory-flow-PEF-pilot_0_3': 3, '10696_Peak-expiratory-flow-PEF-pilot_0_4': 4, })
ukb_12336_2 = TensorMap('12336_Ventricular-rate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 62.02991341709527, 'std': 10.946793190781303}, annotation_units=1, channel_map={'12336_Ventricular-rate_2_0': 0, })
ukb_12338_2 = TensorMap('12338_P-duration_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 97.4182534471437, 'std': 16.466513802515898}, annotation_units=1, channel_map={'12338_P-duration_2_0': 0, })
ukb_12340_2 = TensorMap('12340_QRS-duration_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 88.85757072630798, 'std': 14.069689940451143}, annotation_units=1, channel_map={'12340_QRS-duration_2_0': 0, })
ukb_12651_2 = TensorMap('12651_Duration-of-eprime-test_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 279.97305646130087, 'std': 36.467867072289266}, annotation_units=1, channel_map={'12651_Duration-of-eprime-test_2_0': 0, })
ukb_12654_2 = TensorMap('12654_Number-of-automated-diagnostic-comments-recorded-during-12lead-ECG_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.675522563000584, 'std': 0.9198367396514107}, annotation_units=1, channel_map={'12654_Number-of-automated-diagnostic-comments-recorded-during-12lead-ECG_2_0': 0, })
ukb_12657_2 = TensorMap('12657_Suspicious-flag-for-12lead-ECG_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'12657_Suspicious-flag-for-12lead-ECG_2_0': 0, })
ukb_12673_2 = TensorMap('12673_Heart-rate-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 62.689834686524954, 'std': 10.979421753580818}, annotation_units=4, channel_map={'12673_Heart-rate-during-PWA_2_0': 0, '12673_Heart-rate-during-PWA_2_1': 1, '12673_Heart-rate-during-PWA_2_2': 2, '12673_Heart-rate-during-PWA_2_3': 3, })
ukb_12674_2 = TensorMap('12674_Systolic-brachial-blood-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 136.91613264698668, 'std': 21.64587510385275}, annotation_units=4, channel_map={'12674_Systolic-brachial-blood-pressure-during-PWA_2_0': 0, '12674_Systolic-brachial-blood-pressure-during-PWA_2_1': 1, '12674_Systolic-brachial-blood-pressure-during-PWA_2_2': 2, '12674_Systolic-brachial-blood-pressure-during-PWA_2_3': 3, })
ukb_12675_2 = TensorMap('12675_Diastolic-brachial-blood-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 69.19338515950552, 'std': 12.114468037906546}, annotation_units=4, channel_map={'12675_Diastolic-brachial-blood-pressure-during-PWA_2_0': 0, '12675_Diastolic-brachial-blood-pressure-during-PWA_2_1': 1, '12675_Diastolic-brachial-blood-pressure-during-PWA_2_2': 2, '12675_Diastolic-brachial-blood-pressure-during-PWA_2_3': 3, })
ukb_12676_2 = TensorMap('12676_Peripheral-pulse-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 67.73457341530482, 'std': 19.42484386177738}, annotation_units=4, channel_map={'12676_Peripheral-pulse-pressure-during-PWA_2_0': 0, '12676_Peripheral-pulse-pressure-during-PWA_2_1': 1, '12676_Peripheral-pulse-pressure-during-PWA_2_2': 2, '12676_Peripheral-pulse-pressure-during-PWA_2_3': 3, })
ukb_12677_2 = TensorMap('12677_Central-systolic-blood-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 132.94330998248688, 'std': 21.432793530295836}, annotation_units=4, channel_map={'12677_Central-systolic-blood-pressure-during-PWA_2_0': 0, '12677_Central-systolic-blood-pressure-during-PWA_2_1': 1, '12677_Central-systolic-blood-pressure-during-PWA_2_2': 2, '12677_Central-systolic-blood-pressure-during-PWA_2_3': 3, })
ukb_12678_2 = TensorMap('12678_Central-pulse-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 63.75643248734521, 'std': 19.058851094818294}, annotation_units=4, channel_map={'12678_Central-pulse-pressure-during-PWA_2_0': 0, '12678_Central-pulse-pressure-during-PWA_2_1': 1, '12678_Central-pulse-pressure-during-PWA_2_2': 2, '12678_Central-pulse-pressure-during-PWA_2_3': 3, })
ukb_12679_2 = TensorMap('12679_Number-of-beats-in-waveform-average-for-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.978147440202507, 'std': 2.630467750003038}, annotation_units=4, channel_map={'12679_Number-of-beats-in-waveform-average-for-PWA_2_0': 0, '12679_Number-of-beats-in-waveform-average-for-PWA_2_1': 1, '12679_Number-of-beats-in-waveform-average-for-PWA_2_2': 2, '12679_Number-of-beats-in-waveform-average-for-PWA_2_3': 3, })
ukb_12680_2 = TensorMap('12680_Central-augmentation-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.318947368421048, 'std': 8.247159588102935}, annotation_units=4, channel_map={'12680_Central-augmentation-pressure-during-PWA_2_0': 0, '12680_Central-augmentation-pressure-during-PWA_2_1': 1, '12680_Central-augmentation-pressure-during-PWA_2_2': 2, '12680_Central-augmentation-pressure-during-PWA_2_3': 3, })
ukb_12681_2 = TensorMap('12681_Augmentation-index-for-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.270845551454098, 'std': 9.801668705840774}, annotation_units=4, channel_map={'12681_Augmentation-index-for-PWA_2_0': 0, '12681_Augmentation-index-for-PWA_2_1': 1, '12681_Augmentation-index-for-PWA_2_2': 2, '12681_Augmentation-index-for-PWA_2_3': 3, })
ukb_12682_2 = TensorMap('12682_Cardiac-output-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.120456946164492, 'std': 2.71286922367419}, annotation_units=4, channel_map={'12682_Cardiac-output-during-PWA_2_0': 0, '12682_Cardiac-output-during-PWA_2_1': 1, '12682_Cardiac-output-during-PWA_2_2': 2, '12682_Cardiac-output-during-PWA_2_3': 3, })
ukb_12683_2 = TensorMap('12683_End-systolic-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 112.92685967099142, 'std': 19.550972246229563}, annotation_units=4, channel_map={'12683_End-systolic-pressure-during-PWA_2_0': 0, '12683_End-systolic-pressure-during-PWA_2_1': 1, '12683_End-systolic-pressure-during-PWA_2_2': 2, '12683_End-systolic-pressure-during-PWA_2_3': 3, })
ukb_12684_2 = TensorMap('12684_End-systolic-pressure-index-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7096449057634525, 'std': 0.10138785227297838}, annotation_units=4, channel_map={'12684_End-systolic-pressure-index-during-PWA_2_0': 0, '12684_End-systolic-pressure-index-during-PWA_2_1': 1, '12684_End-systolic-pressure-index-during-PWA_2_2': 2, '12684_End-systolic-pressure-index-during-PWA_2_3': 3, })
ukb_12685_2 = TensorMap('12685_Total-peripheral-resistance-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': -2105.720112947707, 'std': 123278.68593021583}, annotation_units=4, channel_map={'12685_Total-peripheral-resistance-during-PWA_2_0': 0, '12685_Total-peripheral-resistance-during-PWA_2_1': 1, '12685_Total-peripheral-resistance-during-PWA_2_2': 2, '12685_Total-peripheral-resistance-during-PWA_2_3': 3, })
ukb_12686_2 = TensorMap('12686_Stroke-volume-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 114.51456699361272, 'std': 37.655545943304034}, annotation_units=4, channel_map={'12686_Stroke-volume-during-PWA_2_0': 0, '12686_Stroke-volume-during-PWA_2_1': 1, '12686_Stroke-volume-during-PWA_2_2': 2, '12686_Stroke-volume-during-PWA_2_3': 3, })
ukb_12687_2 = TensorMap('12687_Mean-arterial-pressure-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 95.55761821366022, 'std': 14.297013172366999}, annotation_units=4, channel_map={'12687_Mean-arterial-pressure-during-PWA_2_0': 0, '12687_Mean-arterial-pressure-during-PWA_2_1': 1, '12687_Mean-arterial-pressure-during-PWA_2_2': 2, '12687_Mean-arterial-pressure-during-PWA_2_3': 3, })
ukb_12697_2 = TensorMap('12697_Systolic-brachial-blood-pressure_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 136.5654958607297, 'std': 22.327476981043972}, annotation_units=1, channel_map={'12697_Systolic-brachial-blood-pressure_2_0': 0, })
ukb_12698_2 = TensorMap('12698_Diastolic-brachial-blood-pressure_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 69.16583205586778, 'std': 12.312767175713873}, annotation_units=1, channel_map={'12698_Diastolic-brachial-blood-pressure_2_0': 0, })
ukb_12699_2 = TensorMap('12699_Number-of-PWA-tests-performed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.97192422596415, 'std': 0.2826855073653946}, annotation_units=1, channel_map={'12699_Number-of-PWA-tests-performed_2_0': 0, })
ukb_12702_2 = TensorMap('12702_Cardiac-index-during-PWA_2_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.8947176664628156, 'std': 1.7975829971510535}, annotation_units=4, channel_map={'12702_Cardiac-index-during-PWA_2_0': 0, '12702_Cardiac-index-during-PWA_2_1': 1, '12702_Cardiac-index-during-PWA_2_2': 2, '12702_Cardiac-index-during-PWA_2_3': 3, })
ukb_20006_0 = TensorMap('20006_Interpolated-Year-when-cancer-first-diagnosed_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 1995.3927593749997, 'std': 94.83441466298387}, annotation_units=6, channel_map={'20006_Interpolated-Year-when-cancer-first-diagnosed_0_0': 0, '20006_Interpolated-Year-when-cancer-first-diagnosed_0_1': 1, '20006_Interpolated-Year-when-cancer-first-diagnosed_0_2': 2, '20006_Interpolated-Year-when-cancer-first-diagnosed_0_3': 3, '20006_Interpolated-Year-when-cancer-first-diagnosed_0_4': 4, '20006_Interpolated-Year-when-cancer-first-diagnosed_0_5': 5, })
ukb_20006_2 = TensorMap('20006_Interpolated-Year-when-cancer-first-diagnosed_2_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 2006.489133539633, 'std': 32.44738901098238}, annotation_units=5, channel_map={'20006_Interpolated-Year-when-cancer-first-diagnosed_2_0': 0, '20006_Interpolated-Year-when-cancer-first-diagnosed_2_1': 1, '20006_Interpolated-Year-when-cancer-first-diagnosed_2_2': 2, '20006_Interpolated-Year-when-cancer-first-diagnosed_2_3': 3, '20006_Interpolated-Year-when-cancer-first-diagnosed_2_4': 4, })
ukb_20006_1 = TensorMap('20006_Interpolated-Year-when-cancer-first-diagnosed_1_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 2000.9511796136458, 'std': 70.91861630674187}, annotation_units=4, channel_map={'20006_Interpolated-Year-when-cancer-first-diagnosed_1_0': 0, '20006_Interpolated-Year-when-cancer-first-diagnosed_1_1': 1, '20006_Interpolated-Year-when-cancer-first-diagnosed_1_2': 2, '20006_Interpolated-Year-when-cancer-first-diagnosed_1_3': 3, })
ukb_20007_2 = TensorMap('20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_2_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.10189141156869, 'std': 12.131009864821909}, annotation_units=5, channel_map={'20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_2_0': 0, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_2_1': 1, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_2_2': 2, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_2_3': 3, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_2_4': 4, })
ukb_20007_0 = TensorMap('20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_5', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.96990048178571, 'std': 11.462360746646969}, annotation_units=6, channel_map={'20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_0': 0, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_1': 1, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_2': 2, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_3': 3, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_4': 4, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_0_5': 5, })
ukb_20007_1 = TensorMap('20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_1_3', loss='logcosh', path_prefix='continuous', normalization={'mean': 54.65085833127826, 'std': 11.493902156910064}, annotation_units=4, channel_map={'20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_1_0': 0, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_1_1': 1, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_1_2': 2, '20007_Interpolated-Age-of-participant-when-cancer-first-diagnosed_1_3': 3, })
ukb_20008_2 = TensorMap('20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_33', loss='logcosh', path_prefix='continuous', normalization={'mean': 1931.4335749156755, 'std': 351.4888153513326}, annotation_units=34, channel_map={'20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_0': 0, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_1': 1, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_2': 2, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_3': 3, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_4': 4, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_5': 5, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_6': 6, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_7': 7, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_8': 8, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_9': 9, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_10': 10, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_11': 11, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_12': 12, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_13': 13, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_14': 14, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_15': 15, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_16': 16, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_17': 17, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_18': 18, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_19': 19, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_20': 20, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_21': 21, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_22': 22, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_23': 23, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_24': 24, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_25': 25, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_26': 26, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_27': 27, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_28': 28, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_29': 29, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_30': 30, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_31': 31, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_32': 32, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_2_33': 33, })
ukb_20008_0 = TensorMap('20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_28', loss='logcosh', path_prefix='continuous', normalization={'mean': 1924.7027612473553, 'std': 365.80733532433004}, annotation_units=29, channel_map={'20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_0': 0, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_1': 1, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_2': 2, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_3': 3, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_4': 4, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_5': 5, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_6': 6, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_7': 7, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_8': 8, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_9': 9, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_10': 10, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_11': 11, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_12': 12, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_13': 13, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_14': 14, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_15': 15, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_16': 16, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_17': 17, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_18': 18, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_19': 19, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_20': 20, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_21': 21, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_22': 22, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_23': 23, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_24': 24, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_25': 25, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_26': 26, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_27': 27, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_0_28': 28, })
ukb_20008_1 = TensorMap('20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 1964.277166630762, 'std': 241.49266719407777}, annotation_units=16, channel_map={'20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_0': 0, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_1': 1, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_2': 2, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_3': 3, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_4': 4, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_5': 5, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_6': 6, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_7': 7, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_8': 8, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_9': 9, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_10': 10, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_11': 11, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_12': 12, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_13': 13, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_14': 14, '20008_Interpolated-Year-when-noncancer-illness-first-diagnosed_1_15': 15, })
ukb_20009_1 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_15', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.93715207946843, 'std': 20.147053994405734}, annotation_units=16, channel_map={'20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_0': 0, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_1': 1, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_2': 2, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_3': 3, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_4': 4, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_5': 5, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_6': 6, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_7': 7, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_8': 8, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_9': 9, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_10': 10, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_11': 11, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_12': 12, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_13': 13, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_14': 14, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_1_15': 15, })
ukb_20009_2 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_33', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.11278982993964, 'std': 22.53695261417262}, annotation_units=34, channel_map={'20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_0': 0, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_1': 1, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_2': 2, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_3': 3, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_4': 4, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_5': 5, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_6': 6, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_7': 7, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_8': 8, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_9': 9, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_10': 10, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_11': 11, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_12': 12, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_13': 13, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_14': 14, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_15': 15, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_16': 16, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_17': 17, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_18': 18, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_19': 19, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_20': 20, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_21': 21, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_22': 22, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_23': 23, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_24': 24, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_25': 25, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_26': 26, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_27': 27, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_28': 28, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_29': 29, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_30': 30, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_31': 31, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_32': 32, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_2_33': 33, })
ukb_20009_0 = TensorMap('20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_28', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.3039380472219, 'std': 18.4542979266316}, annotation_units=29, channel_map={'20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_0': 0, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_1': 1, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_2': 2, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_3': 3, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_4': 4, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_5': 5, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_6': 6, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_7': 7, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_8': 8, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_9': 9, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_10': 10, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_11': 11, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_12': 12, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_13': 13, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_14': 14, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_15': 15, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_16': 16, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_17': 17, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_18': 18, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_19': 19, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_20': 20, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_21': 21, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_22': 22, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_23': 23, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_24': 24, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_25': 25, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_26': 26, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_27': 27, '20009_Interpolated-Age-of-participant-when-noncancer-illness-first-diagnosed_0_28': 28, })
ukb_20010_2 = TensorMap('20010_Interpolated-Year-when-operation-took-place_2_17', loss='logcosh', path_prefix='continuous', normalization={'mean': 1981.4264742984299, 'std': 142.23972595250473}, annotation_units=18, channel_map={'20010_Interpolated-Year-when-operation-took-place_2_0': 0, '20010_Interpolated-Year-when-operation-took-place_2_1': 1, '20010_Interpolated-Year-when-operation-took-place_2_2': 2, '20010_Interpolated-Year-when-operation-took-place_2_3': 3, '20010_Interpolated-Year-when-operation-took-place_2_4': 4, '20010_Interpolated-Year-when-operation-took-place_2_5': 5, '20010_Interpolated-Year-when-operation-took-place_2_6': 6, '20010_Interpolated-Year-when-operation-took-place_2_7': 7, '20010_Interpolated-Year-when-operation-took-place_2_8': 8, '20010_Interpolated-Year-when-operation-took-place_2_9': 9, '20010_Interpolated-Year-when-operation-took-place_2_10': 10, '20010_Interpolated-Year-when-operation-took-place_2_11': 11, '20010_Interpolated-Year-when-operation-took-place_2_12': 12, '20010_Interpolated-Year-when-operation-took-place_2_13': 13, '20010_Interpolated-Year-when-operation-took-place_2_14': 14, '20010_Interpolated-Year-when-operation-took-place_2_15': 15, '20010_Interpolated-Year-when-operation-took-place_2_16': 16, '20010_Interpolated-Year-when-operation-took-place_2_17': 17, })
ukb_20010_1 = TensorMap('20010_Interpolated-Year-when-operation-took-place_1_14', loss='logcosh', path_prefix='continuous', normalization={'mean': 1982.3291184471518, 'std': 109.48591496119252}, annotation_units=15, channel_map={'20010_Interpolated-Year-when-operation-took-place_1_0': 0, '20010_Interpolated-Year-when-operation-took-place_1_1': 1, '20010_Interpolated-Year-when-operation-took-place_1_2': 2, '20010_Interpolated-Year-when-operation-took-place_1_3': 3, '20010_Interpolated-Year-when-operation-took-place_1_4': 4, '20010_Interpolated-Year-when-operation-took-place_1_5': 5, '20010_Interpolated-Year-when-operation-took-place_1_6': 6, '20010_Interpolated-Year-when-operation-took-place_1_7': 7, '20010_Interpolated-Year-when-operation-took-place_1_8': 8, '20010_Interpolated-Year-when-operation-took-place_1_9': 9, '20010_Interpolated-Year-when-operation-took-place_1_10': 10, '20010_Interpolated-Year-when-operation-took-place_1_11': 11, '20010_Interpolated-Year-when-operation-took-place_1_12': 12, '20010_Interpolated-Year-when-operation-took-place_1_13': 13, '20010_Interpolated-Year-when-operation-took-place_1_14': 14, })
ukb_20010_0 = TensorMap('20010_Interpolated-Year-when-operation-took-place_0_31', loss='logcosh', path_prefix='continuous', normalization={'mean': 1966.4213097327615, 'std': 197.69272637040814}, annotation_units=32, channel_map={'20010_Interpolated-Year-when-operation-took-place_0_0': 0, '20010_Interpolated-Year-when-operation-took-place_0_1': 1, '20010_Interpolated-Year-when-operation-took-place_0_2': 2, '20010_Interpolated-Year-when-operation-took-place_0_3': 3, '20010_Interpolated-Year-when-operation-took-place_0_4': 4, '20010_Interpolated-Year-when-operation-took-place_0_5': 5, '20010_Interpolated-Year-when-operation-took-place_0_6': 6, '20010_Interpolated-Year-when-operation-took-place_0_7': 7, '20010_Interpolated-Year-when-operation-took-place_0_8': 8, '20010_Interpolated-Year-when-operation-took-place_0_9': 9, '20010_Interpolated-Year-when-operation-took-place_0_10': 10, '20010_Interpolated-Year-when-operation-took-place_0_11': 11, '20010_Interpolated-Year-when-operation-took-place_0_12': 12, '20010_Interpolated-Year-when-operation-took-place_0_13': 13, '20010_Interpolated-Year-when-operation-took-place_0_14': 14, '20010_Interpolated-Year-when-operation-took-place_0_15': 15, '20010_Interpolated-Year-when-operation-took-place_0_16': 16, '20010_Interpolated-Year-when-operation-took-place_0_17': 17, '20010_Interpolated-Year-when-operation-took-place_0_18': 18, '20010_Interpolated-Year-when-operation-took-place_0_19': 19, '20010_Interpolated-Year-when-operation-took-place_0_20': 20, '20010_Interpolated-Year-when-operation-took-place_0_21': 21, '20010_Interpolated-Year-when-operation-took-place_0_22': 22, '20010_Interpolated-Year-when-operation-took-place_0_23': 23, '20010_Interpolated-Year-when-operation-took-place_0_24': 24, '20010_Interpolated-Year-when-operation-took-place_0_25': 25, '20010_Interpolated-Year-when-operation-took-place_0_26': 26, '20010_Interpolated-Year-when-operation-took-place_0_27': 27, '20010_Interpolated-Year-when-operation-took-place_0_28': 28, '20010_Interpolated-Year-when-operation-took-place_0_29': 29, '20010_Interpolated-Year-when-operation-took-place_0_30': 30, '20010_Interpolated-Year-when-operation-took-place_0_31': 31, })
ukb_20011_1 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_1_14', loss='logcosh', path_prefix='continuous', normalization={'mean': 37.95305212968631, 'std': 19.622457449458512}, annotation_units=15, channel_map={'20011_Interpolated-Age-of-participant-when-operation-took-place_1_0': 0, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_1': 1, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_2': 2, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_3': 3, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_4': 4, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_5': 5, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_6': 6, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_7': 7, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_8': 8, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_9': 9, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_10': 10, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_11': 11, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_12': 12, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_13': 13, '20011_Interpolated-Age-of-participant-when-operation-took-place_1_14': 14, })
ukb_20011_2 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_2_17', loss='logcosh', path_prefix='continuous', normalization={'mean': 38.32191541024701, 'std': 19.843448392970622}, annotation_units=18, channel_map={'20011_Interpolated-Age-of-participant-when-operation-took-place_2_0': 0, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_1': 1, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_2': 2, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_3': 3, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_4': 4, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_5': 5, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_6': 6, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_7': 7, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_8': 8, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_9': 9, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_10': 10, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_11': 11, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_12': 12, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_13': 13, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_14': 14, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_15': 15, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_16': 16, '20011_Interpolated-Age-of-participant-when-operation-took-place_2_17': 17, })
ukb_20011_0 = TensorMap('20011_Interpolated-Age-of-participant-when-operation-took-place_0_31', loss='logcosh', path_prefix='continuous', normalization={'mean': 35.098792258036305, 'std': 18.04442651398569}, annotation_units=32, channel_map={'20011_Interpolated-Age-of-participant-when-operation-took-place_0_0': 0, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_1': 1, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_2': 2, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_3': 3, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_4': 4, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_5': 5, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_6': 6, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_7': 7, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_8': 8, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_9': 9, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_10': 10, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_11': 11, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_12': 12, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_13': 13, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_14': 14, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_15': 15, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_16': 16, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_17': 17, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_18': 18, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_19': 19, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_20': 20, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_21': 21, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_22': 22, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_23': 23, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_24': 24, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_25': 25, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_26': 26, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_27': 27, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_28': 28, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_29': 29, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_30': 30, '20011_Interpolated-Age-of-participant-when-operation-took-place_0_31': 31, })
ukb_20015_2 = TensorMap('20015_Sitting-height_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 89.22356237183372, 'std': 4.773724966456604}, annotation_units=1, channel_map={'20015_Sitting-height_2_0': 0, })
ukb_20015_1 = TensorMap('20015_Sitting-height_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 88.7645353416027, 'std': 4.771544055241578}, annotation_units=1, channel_map={'20015_Sitting-height_1_0': 0, })
ukb_20015_0 = TensorMap('20015_Sitting-height_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 89.14320146537861, 'std': 4.882706194277553}, annotation_units=1, channel_map={'20015_Sitting-height_0_0': 0, })
ukb_20019_1 = TensorMap('20019_Speechreceptionthreshold-SRT-estimate-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -6.705844548616361, 'std': 1.9725063548492705}, annotation_units=1, channel_map={'20019_Speechreceptionthreshold-SRT-estimate-left_1_0': 0, })
ukb_20019_2 = TensorMap('20019_Speechreceptionthreshold-SRT-estimate-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -5.023380887681158, 'std': 2.0998865203997945}, annotation_units=1, channel_map={'20019_Speechreceptionthreshold-SRT-estimate-left_2_0': 0, })
ukb_20019_0 = TensorMap('20019_Speechreceptionthreshold-SRT-estimate-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -6.587135061642197, 'std': 2.0968723587332954}, annotation_units=1, channel_map={'20019_Speechreceptionthreshold-SRT-estimate-left_0_0': 0, })
ukb_20021_2 = TensorMap('20021_Speechreceptionthreshold-SRT-estimate-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -5.038486738059738, 'std': 2.067804341648226}, annotation_units=1, channel_map={'20021_Speechreceptionthreshold-SRT-estimate-right_2_0': 0, })
ukb_20021_1 = TensorMap('20021_Speechreceptionthreshold-SRT-estimate-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -6.696689509221003, 'std': 2.0100775648259375}, annotation_units=1, channel_map={'20021_Speechreceptionthreshold-SRT-estimate-right_1_0': 0, })
ukb_20021_0 = TensorMap('20021_Speechreceptionthreshold-SRT-estimate-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -6.54686408694907, 'std': 2.108436901790483}, annotation_units=1, channel_map={'20021_Speechreceptionthreshold-SRT-estimate-right_0_0': 0, })
ukb_20022_1 = TensorMap('20022_Birth-weight_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.333021937392163, 'std': 0.6445969645957392}, annotation_units=1, channel_map={'20022_Birth-weight_1_0': 0, })
ukb_20022_0 = TensorMap('20022_Birth-weight_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.319047215072434, 'std': 0.6682703698213345}, annotation_units=1, channel_map={'20022_Birth-weight_0_0': 0, })
ukb_20022_2 = TensorMap('20022_Birth-weight_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.3579985955056175, 'std': 0.6967101317797486}, annotation_units=1, channel_map={'20022_Birth-weight_2_0': 0, })
ukb_20077_0 = TensorMap('20077_Number-of-diet-questionnaires-completed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.1733421159879045, 'std': 1.18059959185329}, annotation_units=1, channel_map={'20077_Number-of-diet-questionnaires-completed_0_0': 0, })
ukb_20081_4 = TensorMap('20081_Hourofday-questionnaire-completed_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.954133748304198, 'std': 4.966464197613874}, annotation_units=1, channel_map={'20081_Hourofday-questionnaire-completed_4_0': 0, })
ukb_20081_0 = TensorMap('20081_Hourofday-questionnaire-completed_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.106683259326312, 'std': 2.77766114427824}, annotation_units=1, channel_map={'20081_Hourofday-questionnaire-completed_0_0': 0, })
ukb_20081_3 = TensorMap('20081_Hourofday-questionnaire-completed_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.133691745912468, 'std': 4.771582721701741}, annotation_units=1, channel_map={'20081_Hourofday-questionnaire-completed_3_0': 0, })
ukb_20081_1 = TensorMap('20081_Hourofday-questionnaire-completed_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.004463265042395, 'std': 4.810778656731283}, annotation_units=1, channel_map={'20081_Hourofday-questionnaire-completed_1_0': 0, })
ukb_20081_2 = TensorMap('20081_Hourofday-questionnaire-completed_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.754963069717185, 'std': 4.973644884939068}, annotation_units=1, channel_map={'20081_Hourofday-questionnaire-completed_2_0': 0, })
ukb_20082_0 = TensorMap('20082_Duration-of-questionnaire_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.230562547727455, 'std': 6.762106821946561}, annotation_units=1, channel_map={'20082_Duration-of-questionnaire_0_0': 0, })
ukb_20082_2 = TensorMap('20082_Duration-of-questionnaire_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 18.418639284213114, 'std': 31.594234911886296}, annotation_units=1, channel_map={'20082_Duration-of-questionnaire_2_0': 0, })
ukb_20082_4 = TensorMap('20082_Duration-of-questionnaire_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.580250977575556, 'std': 25.69806178627702}, annotation_units=1, channel_map={'20082_Duration-of-questionnaire_4_0': 0, })
ukb_20082_3 = TensorMap('20082_Duration-of-questionnaire_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 18.11517376265764, 'std': 33.95605106770433}, annotation_units=1, channel_map={'20082_Duration-of-questionnaire_3_0': 0, })
ukb_20082_1 = TensorMap('20082_Duration-of-questionnaire_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.374178669768, 'std': 31.2800735694486}, annotation_units=1, channel_map={'20082_Duration-of-questionnaire_1_0': 0, })
ukb_20083_2 = TensorMap('20083_Delay-between-questionnaire-request-and-completion_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.47978142076502717, 'std': 0.7125566513834837}, annotation_units=1, channel_map={'20083_Delay-between-questionnaire-request-and-completion_2_0': 0, })
ukb_20083_3 = TensorMap('20083_Delay-between-questionnaire-request-and-completion_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5060458035860516, 'std': 2.471812294541296}, annotation_units=1, channel_map={'20083_Delay-between-questionnaire-request-and-completion_3_0': 0, })
ukb_20083_0 = TensorMap('20083_Delay-between-questionnaire-request-and-completion_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'20083_Delay-between-questionnaire-request-and-completion_0_0': 0, })
ukb_20083_1 = TensorMap('20083_Delay-between-questionnaire-request-and-completion_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4680066402250518, 'std': 0.7065286132079722}, annotation_units=1, channel_map={'20083_Delay-between-questionnaire-request-and-completion_1_0': 0, })
ukb_20083_4 = TensorMap('20083_Delay-between-questionnaire-request-and-completion_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5971989466123977, 'std': 2.5530744226897983}, annotation_units=1, channel_map={'20083_Delay-between-questionnaire-request-and-completion_4_0': 0, })
ukb_20129_0 = TensorMap('20129_Number-of-columns-displayed-in-round_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.529685057489506, 'std': 0.49911900960793165}, annotation_units=3, channel_map={'20129_Number-of-columns-displayed-in-round_0_0': 0, '20129_Number-of-columns-displayed-in-round_0_1': 1, '20129_Number-of-columns-displayed-in-round_0_2': 2, })
ukb_20130_0 = TensorMap('20130_Number-of-rows-displayed-in-round_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5890551724685187, 'std': 0.6006753180732118}, annotation_units=3, channel_map={'20130_Number-of-rows-displayed-in-round_0_0': 0, '20130_Number-of-rows-displayed-in-round_0_1': 1, '20130_Number-of-rows-displayed-in-round_0_2': 2, })
ukb_20131_0 = TensorMap('20131_Number-of-correct-matches-in-round_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.584250493957451, 'std': 1.8198698323431461}, annotation_units=3, channel_map={'20131_Number-of-correct-matches-in-round_0_0': 0, '20131_Number-of-correct-matches-in-round_0_1': 1, '20131_Number-of-correct-matches-in-round_0_2': 2, })
ukb_20132_0 = TensorMap('20132_Number-of-incorrect-matches-in-round_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7142347031891, 'std': 3.154258741225523}, annotation_units=3, channel_map={'20132_Number-of-incorrect-matches-in-round_0_0': 0, '20132_Number-of-incorrect-matches-in-round_0_1': 1, '20132_Number-of-incorrect-matches-in-round_0_2': 2, })
ukb_20133_0 = TensorMap('20133_Time-to-complete-round_0_2', loss='logcosh', path_prefix='continuous', normalization={'mean': 22915.969608722204, 'std': 16223.147679056185}, annotation_units=3, channel_map={'20133_Time-to-complete-round_0_0': 0, '20133_Time-to-complete-round_0_1': 1, '20133_Time-to-complete-round_0_2': 2, })
ukb_20147_0 = TensorMap('20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_25', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9382255108275066, 'std': 3.3531821392536605}, annotation_units=26, channel_map={'20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_0': 0, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_1': 1, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_2': 2, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_3': 3, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_4': 4, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_5': 5, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_6': 6, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_7': 7, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_8': 8, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_9': 9, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_10': 10, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_11': 11, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_12': 12, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_13': 13, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_14': 14, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_15': 15, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_16': 16, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_17': 17, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_18': 18, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_19': 19, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_20': 20, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_21': 21, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_22': 22, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_23': 23, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_24': 24, '20147_Errors-before-selecting-correct-item-in-numeric-path-trail-1_0_25': 25, })
ukb_20148_0 = TensorMap('20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_25', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9077937699614254, 'std': 3.0175222154830945}, annotation_units=26, channel_map={'20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_0': 0, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_1': 1, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_2': 2, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_3': 3, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_4': 4, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_5': 5, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_6': 6, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_7': 7, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_8': 8, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_9': 9, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_10': 10, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_11': 11, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_12': 12, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_13': 13, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_14': 14, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_15': 15, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_16': 16, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_17': 17, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_18': 18, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_19': 19, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_20': 20, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_21': 21, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_22': 22, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_23': 23, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_24': 24, '20148_Errors-before-selecting-correct-item-in-alphanumeric-path-trail-2_0_25': 25, })
ukb_20149_0 = TensorMap('20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_25', loss='logcosh', path_prefix='continuous', normalization={'mean': 1623.1255128003022, 'std': 2275.8786609627978}, annotation_units=26, channel_map={'20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_0': 0, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_1': 1, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_2': 2, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_3': 3, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_4': 4, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_5': 5, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_6': 6, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_7': 7, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_8': 8, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_9': 9, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_10': 10, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_11': 11, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_12': 12, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_13': 13, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_14': 14, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_15': 15, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_16': 16, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_17': 17, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_18': 18, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_19': 19, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_20': 20, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_21': 21, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_22': 22, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_23': 23, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_24': 24, '20149_Interval-between-previous-point-and-current-one-in-numeric-path-trail-1_0_25': 25, })
ukb_20155_0 = TensorMap('20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_25', loss='logcosh', path_prefix='continuous', normalization={'mean': 2732.5354530531304, 'std': 3144.96750649209}, annotation_units=26, channel_map={'20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_0': 0, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_1': 1, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_2': 2, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_3': 3, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_4': 4, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_5': 5, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_6': 6, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_7': 7, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_8': 8, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_9': 9, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_10': 10, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_11': 11, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_12': 12, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_13': 13, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_14': 14, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_15': 15, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_16': 16, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_17': 17, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_18': 18, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_19': 19, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_20': 20, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_21': 21, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_22': 22, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_23': 23, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_24': 24, '20155_Interval-between-previous-point-and-current-one-in-alphanumeric-path-trail-2_0_25': 25, })
ukb_20156_0 = TensorMap('20156_Duration-to-complete-numeric-path-trail-1_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.19247662098, 'std': 14.99853529075132}, annotation_units=1, channel_map={'20156_Duration-to-complete-numeric-path-trail-1_0_0': 0, })
ukb_20157_0 = TensorMap('20157_Duration-to-complete-alphanumeric-path-trail-2_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 66.80345922914259, 'std': 25.752353412671365}, annotation_units=1, channel_map={'20157_Duration-to-complete-alphanumeric-path-trail-2_0_0': 0, })
ukb_20159_0 = TensorMap('20159_Number-of-symbol-digit-matches-made-correctly_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.676524953789286, 'std': 5.237203203595653}, annotation_units=1, channel_map={'20159_Number-of-symbol-digit-matches-made-correctly_0_0': 0, })
ukb_20161_0 = TensorMap('20161_Pack-years-of-smoking_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.381888870486776, 'std': 18.803128176701886}, annotation_units=1, channel_map={'20161_Pack-years-of-smoking_0_0': 0, })
ukb_20161_1 = TensorMap('20161_Pack-years-of-smoking_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.009768769893284, 'std': 17.368628259048467}, annotation_units=1, channel_map={'20161_Pack-years-of-smoking_1_0': 0, })
ukb_20161_2 = TensorMap('20161_Pack-years-of-smoking_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 19.532445805843548, 'std': 16.18498535111979}, annotation_units=1, channel_map={'20161_Pack-years-of-smoking_2_0': 0, })
ukb_20162_2 = TensorMap('20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3970285399870023, 'std': 0.3224213120068275}, annotation_units=1, channel_map={'20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_2_0': 0, })
ukb_20162_0 = TensorMap('20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5507673284808219, 'std': 0.41873210298871866}, annotation_units=1, channel_map={'20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_0_0': 0, })
ukb_20162_1 = TensorMap('20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4428679234094737, 'std': 0.35700561829060395}, annotation_units=1, channel_map={'20162_Pack-years-adult-smoking-as-proportion-of-life-span-exposed-to-smoking_1_0': 0, })
ukb_20191_0 = TensorMap('20191_Fluid-intelligence-score_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.434275224015786, 'std': 2.0388517058485514}, annotation_units=1, channel_map={'20191_Fluid-intelligence-score_0_0': 0, })
ukb_20195_0 = TensorMap('20195_Number-of-symbol-digit-matches-attempted_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.26419027844597, 'std': 4.862302855855112}, annotation_units=1, channel_map={'20195_Number-of-symbol-digit-matches-attempted_0_0': 0, })
ukb_20200_0 = TensorMap('20200_Values-wanted_0_116', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.493356376699333, 'std': 2.2875311522945077}, annotation_units=117, channel_map={'20200_Values-wanted_0_0': 0, '20200_Values-wanted_0_1': 1, '20200_Values-wanted_0_2': 2, '20200_Values-wanted_0_3': 3, '20200_Values-wanted_0_4': 4, '20200_Values-wanted_0_5': 5, '20200_Values-wanted_0_6': 6, '20200_Values-wanted_0_7': 7, '20200_Values-wanted_0_8': 8, '20200_Values-wanted_0_9': 9, '20200_Values-wanted_0_10': 10, '20200_Values-wanted_0_11': 11, '20200_Values-wanted_0_12': 12, '20200_Values-wanted_0_13': 13, '20200_Values-wanted_0_14': 14, '20200_Values-wanted_0_15': 15, '20200_Values-wanted_0_16': 16, '20200_Values-wanted_0_17': 17, '20200_Values-wanted_0_18': 18, '20200_Values-wanted_0_19': 19, '20200_Values-wanted_0_20': 20, '20200_Values-wanted_0_21': 21, '20200_Values-wanted_0_22': 22, '20200_Values-wanted_0_23': 23, '20200_Values-wanted_0_24': 24, '20200_Values-wanted_0_25': 25, '20200_Values-wanted_0_26': 26, '20200_Values-wanted_0_27': 27, '20200_Values-wanted_0_28': 28, '20200_Values-wanted_0_29': 29, '20200_Values-wanted_0_30': 30, '20200_Values-wanted_0_31': 31, '20200_Values-wanted_0_32': 32, '20200_Values-wanted_0_33': 33, '20200_Values-wanted_0_34': 34, '20200_Values-wanted_0_35': 35, '20200_Values-wanted_0_36': 36, '20200_Values-wanted_0_37': 37, '20200_Values-wanted_0_38': 38, '20200_Values-wanted_0_39': 39, '20200_Values-wanted_0_40': 40, '20200_Values-wanted_0_41': 41, '20200_Values-wanted_0_42': 42, '20200_Values-wanted_0_43': 43, '20200_Values-wanted_0_44': 44, '20200_Values-wanted_0_45': 45, '20200_Values-wanted_0_46': 46, '20200_Values-wanted_0_47': 47, '20200_Values-wanted_0_48': 48, '20200_Values-wanted_0_49': 49, '20200_Values-wanted_0_50': 50, '20200_Values-wanted_0_51': 51, '20200_Values-wanted_0_52': 52, '20200_Values-wanted_0_53': 53, '20200_Values-wanted_0_54': 54, '20200_Values-wanted_0_55': 55, '20200_Values-wanted_0_56': 56, '20200_Values-wanted_0_57': 57, '20200_Values-wanted_0_58': 58, '20200_Values-wanted_0_59': 59, '20200_Values-wanted_0_60': 60, '20200_Values-wanted_0_61': 61, '20200_Values-wanted_0_62': 62, '20200_Values-wanted_0_63': 63, '20200_Values-wanted_0_64': 64, '20200_Values-wanted_0_65': 65, '20200_Values-wanted_0_66': 66, '20200_Values-wanted_0_67': 67, '20200_Values-wanted_0_68': 68, '20200_Values-wanted_0_69': 69, '20200_Values-wanted_0_70': 70, '20200_Values-wanted_0_71': 71, '20200_Values-wanted_0_72': 72, '20200_Values-wanted_0_73': 73, '20200_Values-wanted_0_74': 74, '20200_Values-wanted_0_75': 75, '20200_Values-wanted_0_76': 76, '20200_Values-wanted_0_77': 77, '20200_Values-wanted_0_78': 78, '20200_Values-wanted_0_79': 79, '20200_Values-wanted_0_80': 80, '20200_Values-wanted_0_81': 81, '20200_Values-wanted_0_82': 82, '20200_Values-wanted_0_83': 83, '20200_Values-wanted_0_84': 84, '20200_Values-wanted_0_85': 85, '20200_Values-wanted_0_86': 86, '20200_Values-wanted_0_87': 87, '20200_Values-wanted_0_88': 88, '20200_Values-wanted_0_89': 89, '20200_Values-wanted_0_90': 90, '20200_Values-wanted_0_91': 91, '20200_Values-wanted_0_92': 92, '20200_Values-wanted_0_93': 93, '20200_Values-wanted_0_94': 94, '20200_Values-wanted_0_95': 95, '20200_Values-wanted_0_96': 96, '20200_Values-wanted_0_97': 97, '20200_Values-wanted_0_98': 98, '20200_Values-wanted_0_99': 99, '20200_Values-wanted_0_100': 100, '20200_Values-wanted_0_101': 101, '20200_Values-wanted_0_102': 102, '20200_Values-wanted_0_103': 103, '20200_Values-wanted_0_104': 104, '20200_Values-wanted_0_105': 105, '20200_Values-wanted_0_106': 106, '20200_Values-wanted_0_107': 107, '20200_Values-wanted_0_108': 108, '20200_Values-wanted_0_109': 109, '20200_Values-wanted_0_110': 110, '20200_Values-wanted_0_111': 111, '20200_Values-wanted_0_112': 112, '20200_Values-wanted_0_113': 113, '20200_Values-wanted_0_114': 114, '20200_Values-wanted_0_115': 115, '20200_Values-wanted_0_116': 116, })
ukb_20229_0 = TensorMap('20229_Values-entered_0_116', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.4857287837609965, 'std': 2.289528372599998}, annotation_units=117, channel_map={'20229_Values-entered_0_0': 0, '20229_Values-entered_0_1': 1, '20229_Values-entered_0_2': 2, '20229_Values-entered_0_3': 3, '20229_Values-entered_0_4': 4, '20229_Values-entered_0_5': 5, '20229_Values-entered_0_6': 6, '20229_Values-entered_0_7': 7, '20229_Values-entered_0_8': 8, '20229_Values-entered_0_9': 9, '20229_Values-entered_0_10': 10, '20229_Values-entered_0_11': 11, '20229_Values-entered_0_12': 12, '20229_Values-entered_0_13': 13, '20229_Values-entered_0_14': 14, '20229_Values-entered_0_15': 15, '20229_Values-entered_0_16': 16, '20229_Values-entered_0_17': 17, '20229_Values-entered_0_18': 18, '20229_Values-entered_0_19': 19, '20229_Values-entered_0_20': 20, '20229_Values-entered_0_21': 21, '20229_Values-entered_0_22': 22, '20229_Values-entered_0_23': 23, '20229_Values-entered_0_24': 24, '20229_Values-entered_0_25': 25, '20229_Values-entered_0_26': 26, '20229_Values-entered_0_27': 27, '20229_Values-entered_0_28': 28, '20229_Values-entered_0_29': 29, '20229_Values-entered_0_30': 30, '20229_Values-entered_0_31': 31, '20229_Values-entered_0_32': 32, '20229_Values-entered_0_33': 33, '20229_Values-entered_0_34': 34, '20229_Values-entered_0_35': 35, '20229_Values-entered_0_36': 36, '20229_Values-entered_0_37': 37, '20229_Values-entered_0_38': 38, '20229_Values-entered_0_39': 39, '20229_Values-entered_0_40': 40, '20229_Values-entered_0_41': 41, '20229_Values-entered_0_42': 42, '20229_Values-entered_0_43': 43, '20229_Values-entered_0_44': 44, '20229_Values-entered_0_45': 45, '20229_Values-entered_0_46': 46, '20229_Values-entered_0_47': 47, '20229_Values-entered_0_48': 48, '20229_Values-entered_0_49': 49, '20229_Values-entered_0_50': 50, '20229_Values-entered_0_51': 51, '20229_Values-entered_0_52': 52, '20229_Values-entered_0_53': 53, '20229_Values-entered_0_54': 54, '20229_Values-entered_0_55': 55, '20229_Values-entered_0_56': 56, '20229_Values-entered_0_57': 57, '20229_Values-entered_0_58': 58, '20229_Values-entered_0_59': 59, '20229_Values-entered_0_60': 60, '20229_Values-entered_0_61': 61, '20229_Values-entered_0_62': 62, '20229_Values-entered_0_63': 63, '20229_Values-entered_0_64': 64, '20229_Values-entered_0_65': 65, '20229_Values-entered_0_66': 66, '20229_Values-entered_0_67': 67, '20229_Values-entered_0_68': 68, '20229_Values-entered_0_69': 69, '20229_Values-entered_0_70': 70, '20229_Values-entered_0_71': 71, '20229_Values-entered_0_72': 72, '20229_Values-entered_0_73': 73, '20229_Values-entered_0_74': 74, '20229_Values-entered_0_75': 75, '20229_Values-entered_0_76': 76, '20229_Values-entered_0_77': 77, '20229_Values-entered_0_78': 78, '20229_Values-entered_0_79': 79, '20229_Values-entered_0_80': 80, '20229_Values-entered_0_81': 81, '20229_Values-entered_0_82': 82, '20229_Values-entered_0_83': 83, '20229_Values-entered_0_84': 84, '20229_Values-entered_0_85': 85, '20229_Values-entered_0_86': 86, '20229_Values-entered_0_87': 87, '20229_Values-entered_0_88': 88, '20229_Values-entered_0_89': 89, '20229_Values-entered_0_90': 90, '20229_Values-entered_0_91': 91, '20229_Values-entered_0_92': 92, '20229_Values-entered_0_93': 93, '20229_Values-entered_0_94': 94, '20229_Values-entered_0_95': 95, '20229_Values-entered_0_96': 96, '20229_Values-entered_0_97': 97, '20229_Values-entered_0_98': 98, '20229_Values-entered_0_99': 99, '20229_Values-entered_0_100': 100, '20229_Values-entered_0_101': 101, '20229_Values-entered_0_102': 102, '20229_Values-entered_0_103': 103, '20229_Values-entered_0_104': 104, '20229_Values-entered_0_105': 105, '20229_Values-entered_0_106': 106, '20229_Values-entered_0_107': 107, '20229_Values-entered_0_108': 108, '20229_Values-entered_0_109': 109, '20229_Values-entered_0_110': 110, '20229_Values-entered_0_111': 111, '20229_Values-entered_0_112': 112, '20229_Values-entered_0_113': 113, '20229_Values-entered_0_114': 114, '20229_Values-entered_0_115': 115, '20229_Values-entered_0_116': 116, })
ukb_20230_0 = TensorMap('20230_Duration-to-entering-value_0_116', loss='logcosh', path_prefix='continuous', normalization={'mean': 3076.7662062470804, 'std': 2461.816026167483}, annotation_units=117, channel_map={'20230_Duration-to-entering-value_0_0': 0, '20230_Duration-to-entering-value_0_1': 1, '20230_Duration-to-entering-value_0_2': 2, '20230_Duration-to-entering-value_0_3': 3, '20230_Duration-to-entering-value_0_4': 4, '20230_Duration-to-entering-value_0_5': 5, '20230_Duration-to-entering-value_0_6': 6, '20230_Duration-to-entering-value_0_7': 7, '20230_Duration-to-entering-value_0_8': 8, '20230_Duration-to-entering-value_0_9': 9, '20230_Duration-to-entering-value_0_10': 10, '20230_Duration-to-entering-value_0_11': 11, '20230_Duration-to-entering-value_0_12': 12, '20230_Duration-to-entering-value_0_13': 13, '20230_Duration-to-entering-value_0_14': 14, '20230_Duration-to-entering-value_0_15': 15, '20230_Duration-to-entering-value_0_16': 16, '20230_Duration-to-entering-value_0_17': 17, '20230_Duration-to-entering-value_0_18': 18, '20230_Duration-to-entering-value_0_19': 19, '20230_Duration-to-entering-value_0_20': 20, '20230_Duration-to-entering-value_0_21': 21, '20230_Duration-to-entering-value_0_22': 22, '20230_Duration-to-entering-value_0_23': 23, '20230_Duration-to-entering-value_0_24': 24, '20230_Duration-to-entering-value_0_25': 25, '20230_Duration-to-entering-value_0_26': 26, '20230_Duration-to-entering-value_0_27': 27, '20230_Duration-to-entering-value_0_28': 28, '20230_Duration-to-entering-value_0_29': 29, '20230_Duration-to-entering-value_0_30': 30, '20230_Duration-to-entering-value_0_31': 31, '20230_Duration-to-entering-value_0_32': 32, '20230_Duration-to-entering-value_0_33': 33, '20230_Duration-to-entering-value_0_34': 34, '20230_Duration-to-entering-value_0_35': 35, '20230_Duration-to-entering-value_0_36': 36, '20230_Duration-to-entering-value_0_37': 37, '20230_Duration-to-entering-value_0_38': 38, '20230_Duration-to-entering-value_0_39': 39, '20230_Duration-to-entering-value_0_40': 40, '20230_Duration-to-entering-value_0_41': 41, '20230_Duration-to-entering-value_0_42': 42, '20230_Duration-to-entering-value_0_43': 43, '20230_Duration-to-entering-value_0_44': 44, '20230_Duration-to-entering-value_0_45': 45, '20230_Duration-to-entering-value_0_46': 46, '20230_Duration-to-entering-value_0_47': 47, '20230_Duration-to-entering-value_0_48': 48, '20230_Duration-to-entering-value_0_49': 49, '20230_Duration-to-entering-value_0_50': 50, '20230_Duration-to-entering-value_0_51': 51, '20230_Duration-to-entering-value_0_52': 52, '20230_Duration-to-entering-value_0_53': 53, '20230_Duration-to-entering-value_0_54': 54, '20230_Duration-to-entering-value_0_55': 55, '20230_Duration-to-entering-value_0_56': 56, '20230_Duration-to-entering-value_0_57': 57, '20230_Duration-to-entering-value_0_58': 58, '20230_Duration-to-entering-value_0_59': 59, '20230_Duration-to-entering-value_0_60': 60, '20230_Duration-to-entering-value_0_61': 61, '20230_Duration-to-entering-value_0_62': 62, '20230_Duration-to-entering-value_0_63': 63, '20230_Duration-to-entering-value_0_64': 64, '20230_Duration-to-entering-value_0_65': 65, '20230_Duration-to-entering-value_0_66': 66, '20230_Duration-to-entering-value_0_67': 67, '20230_Duration-to-entering-value_0_68': 68, '20230_Duration-to-entering-value_0_69': 69, '20230_Duration-to-entering-value_0_70': 70, '20230_Duration-to-entering-value_0_71': 71, '20230_Duration-to-entering-value_0_72': 72, '20230_Duration-to-entering-value_0_73': 73, '20230_Duration-to-entering-value_0_74': 74, '20230_Duration-to-entering-value_0_75': 75, '20230_Duration-to-entering-value_0_76': 76, '20230_Duration-to-entering-value_0_77': 77, '20230_Duration-to-entering-value_0_78': 78, '20230_Duration-to-entering-value_0_79': 79, '20230_Duration-to-entering-value_0_80': 80, '20230_Duration-to-entering-value_0_81': 81, '20230_Duration-to-entering-value_0_82': 82, '20230_Duration-to-entering-value_0_83': 83, '20230_Duration-to-entering-value_0_84': 84, '20230_Duration-to-entering-value_0_85': 85, '20230_Duration-to-entering-value_0_86': 86, '20230_Duration-to-entering-value_0_87': 87, '20230_Duration-to-entering-value_0_88': 88, '20230_Duration-to-entering-value_0_89': 89, '20230_Duration-to-entering-value_0_90': 90, '20230_Duration-to-entering-value_0_91': 91, '20230_Duration-to-entering-value_0_92': 92, '20230_Duration-to-entering-value_0_93': 93, '20230_Duration-to-entering-value_0_94': 94, '20230_Duration-to-entering-value_0_95': 95, '20230_Duration-to-entering-value_0_96': 96, '20230_Duration-to-entering-value_0_97': 97, '20230_Duration-to-entering-value_0_98': 98, '20230_Duration-to-entering-value_0_99': 99, '20230_Duration-to-entering-value_0_100': 100, '20230_Duration-to-entering-value_0_101': 101, '20230_Duration-to-entering-value_0_102': 102, '20230_Duration-to-entering-value_0_103': 103, '20230_Duration-to-entering-value_0_104': 104, '20230_Duration-to-entering-value_0_105': 105, '20230_Duration-to-entering-value_0_106': 106, '20230_Duration-to-entering-value_0_107': 107, '20230_Duration-to-entering-value_0_108': 108, '20230_Duration-to-entering-value_0_109': 109, '20230_Duration-to-entering-value_0_110': 110, '20230_Duration-to-entering-value_0_111': 111, '20230_Duration-to-entering-value_0_112': 112, '20230_Duration-to-entering-value_0_113': 113, '20230_Duration-to-entering-value_0_114': 114, '20230_Duration-to-entering-value_0_115': 115, '20230_Duration-to-entering-value_0_116': 116, })
ukb_20240_0 = TensorMap('20240_Maximum-digits-remembered-correctly_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.924194681023461, 'std': 1.4852872307187126}, annotation_units=1, channel_map={'20240_Maximum-digits-remembered-correctly_0_0': 0, })
ukb_20247_0 = TensorMap('20247_Total-errors-traversing-numeric-path-trail-1_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.2489758761948115, 'std': 6.700618919918851}, annotation_units=1, channel_map={'20247_Total-errors-traversing-numeric-path-trail-1_0_0': 0, })
ukb_20248_0 = TensorMap('20248_Total-errors-traversing-alphanumeric-path-trail-2_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.2258328462887187, 'std': 5.642607994971585}, annotation_units=1, channel_map={'20248_Total-errors-traversing-alphanumeric-path-trail-2_0_0': 0, })
ukb_20410_0 = TensorMap('20410_Age-when-known-person-last-commented-about-drinking-habits_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.58124301006551, 'std': 13.129766558596943}, annotation_units=1, channel_map={'20410_Age-when-known-person-last-commented-about-drinking-habits_0_0': 0, })
ukb_20455_0 = TensorMap('20455_Age-when-last-took-cannabis_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.22470976064193, 'std': 13.125617474389651}, annotation_units=1, channel_map={'20455_Age-when-last-took-cannabis_0_0': 0, })
ukb_21001_2 = TensorMap('21001_Body-mass-index-BMI_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.654992387063203, 'std': 4.512142395474697}, annotation_units=1, channel_map={'21001_Body-mass-index-BMI_2_0': 0, })
ukb_21001_1 = TensorMap('21001_Body-mass-index-BMI_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.955179334975366, 'std': 4.58661977093225}, annotation_units=1, channel_map={'21001_Body-mass-index-BMI_1_0': 0, })
ukb_21001_0 = TensorMap('21001_Body-mass-index-BMI_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.43260556333107, 'std': 4.803103191651008}, annotation_units=1, channel_map={'21001_Body-mass-index-BMI_0_0': 0, })
ukb_21002_0 = TensorMap('21002_Weight_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.05452345316367, 'std': 15.945823011839185}, annotation_units=1, channel_map={'21002_Weight_0_0': 0, })
ukb_21002_2 = TensorMap('21002_Weight_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.54221851446981, 'std': 15.468377584726051}, annotation_units=1, channel_map={'21002_Weight_2_0': 0, })
ukb_21002_1 = TensorMap('21002_Weight_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.94104609929077, 'std': 15.432186528043976}, annotation_units=1, channel_map={'21002_Weight_1_0': 0, })
ukb_21003_0 = TensorMap('21003_Age-when-attended-assessment-centre_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.528471592084955, 'std': 8.095287610193825}, annotation_units=1, channel_map={'21003_Age-when-attended-assessment-centre_0_0': 0, })
ukb_21003_1 = TensorMap('21003_Age-when-attended-assessment-centre_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 61.44765555883222, 'std': 7.3992113757847005}, annotation_units=1, channel_map={'21003_Age-when-attended-assessment-centre_1_0': 0, })
ukb_21003_2 = TensorMap('21003_Age-when-attended-assessment-centre_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 63.357988914835545, 'std': 7.554638350423904}, annotation_units=1, channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0, })
ukb_21021_2 = TensorMap('21021_Pulse-wave-Arterial-Stiffness-index_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.72545038614405, 'std': 5.105012489854747}, annotation_units=1, channel_map={'21021_Pulse-wave-Arterial-Stiffness-index_2_0': 0, })
ukb_21021_0 = TensorMap('21021_Pulse-wave-Arterial-Stiffness-index_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.337679791931922, 'std': 4.045878710918761}, annotation_units=1, channel_map={'21021_Pulse-wave-Arterial-Stiffness-index_0_0': 0, })
ukb_21021_1 = TensorMap('21021_Pulse-wave-Arterial-Stiffness-index_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.198329560374278, 'std': 3.036472032052337}, annotation_units=1, channel_map={'21021_Pulse-wave-Arterial-Stiffness-index_1_0': 0, })
ukb_21022_0 = TensorMap('21022_Age-at-recruitment_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.52836413709631, 'std': 8.095272747107957}, annotation_units=1, channel_map={'21022_Age-at-recruitment_0_0': 0, })
ukb_22003_0 = TensorMap('22003_Heterozygosity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1881170596901828, 'std': 0.004555422619636816}, annotation_units=1, channel_map={'22003_Heterozygosity_0_0': 0, })
ukb_22004_0 = TensorMap('22004_Heterozygosity-PCA-corrected_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.190295677307245, 'std': 0.00166976887080989}, annotation_units=1, channel_map={'22004_Heterozygosity-PCA-corrected_0_0': 0, })
ukb_22005_0 = TensorMap('22005_Missingness_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.004749983787436921, 'std': 0.004848366188829004}, annotation_units=1, channel_map={'22005_Missingness_0_0': 0, })
ukb_22009_0 = TensorMap('22009_Genetic-principal-components_0_40', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.014422761536727894, 'std': 10.57799283718005}, annotation_units=41, channel_map={'22009_Genetic-principal-components_0_0': 0, '22009_Genetic-principal-components_0_1': 1, '22009_Genetic-principal-components_0_2': 2, '22009_Genetic-principal-components_0_3': 3, '22009_Genetic-principal-components_0_4': 4, '22009_Genetic-principal-components_0_5': 5, '22009_Genetic-principal-components_0_6': 6, '22009_Genetic-principal-components_0_7': 7, '22009_Genetic-principal-components_0_8': 8, '22009_Genetic-principal-components_0_9': 9, '22009_Genetic-principal-components_0_10': 10, '22009_Genetic-principal-components_0_11': 11, '22009_Genetic-principal-components_0_12': 12, '22009_Genetic-principal-components_0_13': 13, '22009_Genetic-principal-components_0_14': 14, '22009_Genetic-principal-components_0_15': 15, '22009_Genetic-principal-components_0_16': 16, '22009_Genetic-principal-components_0_17': 17, '22009_Genetic-principal-components_0_18': 18, '22009_Genetic-principal-components_0_19': 19, '22009_Genetic-principal-components_0_20': 20, '22009_Genetic-principal-components_0_21': 21, '22009_Genetic-principal-components_0_22': 22, '22009_Genetic-principal-components_0_23': 23, '22009_Genetic-principal-components_0_24': 24, '22009_Genetic-principal-components_0_25': 25, '22009_Genetic-principal-components_0_26': 26, '22009_Genetic-principal-components_0_27': 27, '22009_Genetic-principal-components_0_28': 28, '22009_Genetic-principal-components_0_29': 29, '22009_Genetic-principal-components_0_30': 30, '22009_Genetic-principal-components_0_31': 31, '22009_Genetic-principal-components_0_32': 32, '22009_Genetic-principal-components_0_33': 33, '22009_Genetic-principal-components_0_34': 34, '22009_Genetic-principal-components_0_35': 35, '22009_Genetic-principal-components_0_36': 36, '22009_Genetic-principal-components_0_37': 37, '22009_Genetic-principal-components_0_38': 38, '22009_Genetic-principal-components_0_39': 39, '22009_Genetic-principal-components_0_40': 40, })
ukb_22011_0 = TensorMap('22011_Genetic-relatedness-pairing_0_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 4847.150884978584, 'std': 2797.7387362788168}, annotation_units=5, channel_map={'22011_Genetic-relatedness-pairing_0_0': 0, '22011_Genetic-relatedness-pairing_0_1': 1, '22011_Genetic-relatedness-pairing_0_2': 2, '22011_Genetic-relatedness-pairing_0_3': 3, '22011_Genetic-relatedness-pairing_0_4': 4, })
ukb_22012_0 = TensorMap('22012_Genetic-relatedness-factor_0_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12354977036998814, 'std': 0.08572920073419561}, annotation_units=5, channel_map={'22012_Genetic-relatedness-factor_0_0': 0, '22012_Genetic-relatedness-factor_0_1': 1, '22012_Genetic-relatedness-factor_0_2': 2, '22012_Genetic-relatedness-factor_0_3': 3, '22012_Genetic-relatedness-factor_0_4': 4, })
ukb_22013_0 = TensorMap('22013_Genetic-relatedness-IBS0_0_4', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.01806950823055885, 'std': 0.008310056717435408}, annotation_units=5, channel_map={'22013_Genetic-relatedness-IBS0_0_0': 0, '22013_Genetic-relatedness-IBS0_0_1': 1, '22013_Genetic-relatedness-IBS0_0_2': 2, '22013_Genetic-relatedness-IBS0_0_3': 3, '22013_Genetic-relatedness-IBS0_0_4': 4, })
ukb_22033_0 = TensorMap('22033_Summed-days-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.61687678114278, 'std': 4.8437819580014985}, annotation_units=1, channel_map={'22033_Summed-days-activity_0_0': 0, })
ukb_22034_0 = TensorMap('22034_Summed-minutes-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 127.09579274353938, 'std': 101.90442571639983}, annotation_units=1, channel_map={'22034_Summed-minutes-activity_0_0': 0, })
ukb_22037_0 = TensorMap('22037_MET-minutes-per-week-for-walking_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1039.5601282306268, 'std': 1086.1500250153515}, annotation_units=1, channel_map={'22037_MET-minutes-per-week-for-walking_0_0': 0, })
ukb_22038_0 = TensorMap('22038_MET-minutes-per-week-for-moderate-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 933.9493836978008, 'std': 1226.501777985143}, annotation_units=1, channel_map={'22038_MET-minutes-per-week-for-moderate-activity_0_0': 0, })
ukb_22039_0 = TensorMap('22039_MET-minutes-per-week-for-vigorous-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 676.7597017892753, 'std': 1207.0031041595707}, annotation_units=1, channel_map={'22039_MET-minutes-per-week-for-vigorous-activity_0_0': 0, })
ukb_22040_0 = TensorMap('22040_Summed-MET-minutes-per-week-for-all-activity_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2650.269213717759, 'std': 2713.1806848809524}, annotation_units=1, channel_map={'22040_Summed-MET-minutes-per-week-for-all-activity_0_0': 0, })
ukb_22200_0 = TensorMap('22200_Year-of-birth_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1952.0636527790373, 'std': 7.656383767945112}, annotation_units=1, channel_map={'22200_Year-of-birth_0_0': 0, })
ukb_22330_2 = TensorMap('22330_PQ-interval_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 163.45217281135191, 'std': 26.837963274628994}, annotation_units=1, channel_map={'22330_PQ-interval_2_0': 0, })
ukb_22331_2 = TensorMap('22331_QT-interval_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 418.03586904188705, 'std': 32.41293321965725}, annotation_units=1, channel_map={'22331_QT-interval_2_0': 0, })
ukb_22332_2 = TensorMap('22332_QTC-interval_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 420.4707912052643, 'std': 26.60052481125549}, annotation_units=1, channel_map={'22332_QTC-interval_2_0': 0, })
ukb_22333_2 = TensorMap('22333_RR-interval_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 995.2002888781902, 'std': 167.32327071641998}, annotation_units=1, channel_map={'22333_RR-interval_2_0': 0, })
ukb_22334_2 = TensorMap('22334_PP-interval_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 990.771344888462, 'std': 180.8745640164416}, annotation_units=1, channel_map={'22334_PP-interval_2_0': 0, })
ukb_22335_2 = TensorMap('22335_P-axis_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.67882134727259, 'std': 23.379567742879424}, annotation_units=1, channel_map={'22335_P-axis_2_0': 0, })
ukb_22336_2 = TensorMap('22336_R-axis_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.437409725565733, 'std': 37.42771463154406}, annotation_units=1, channel_map={'22336_R-axis_2_0': 0, })
ukb_22337_2 = TensorMap('22337_T-axis_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 40.73350987000487, 'std': 34.943148565746945}, annotation_units=1, channel_map={'22337_T-axis_2_0': 0, })
ukb_22338_2 = TensorMap('22338_QRS-num_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.055528807575023, 'std': 1.841829067775441}, annotation_units=1, channel_map={'22338_QRS-num_2_0': 0, })
ukb_22400_0 = TensorMap('22400_Liver-iron-Fe_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.346558291366906, 'std': 0.3540960977285698}, annotation_units=1, channel_map={'22400_Liver-iron-Fe_0_0': 0, })
ukb_22401_0 = TensorMap('22401_Liver-inflammation-factor-LIF_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8966719726027398, 'std': 0.3122547550918289}, annotation_units=1, channel_map={'22401_Liver-inflammation-factor-LIF_0_0': 0, })
ukb_22402_2 = TensorMap('22402_Liver-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.910121290881519, 'std': 4.644873852230679}, annotation_units=1, channel_map={'22402_Liver-fat-percentage_2_0': 0, })
ukb_22403_2 = TensorMap('22403_Anterior-thigh-lean-muscle-volume-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7215095992131373, 'std': 0.4803665407453433}, annotation_units=1, channel_map={'22403_Anterior-thigh-lean-muscle-volume-right_2_0': 0, })
ukb_22404_2 = TensorMap('22404_Posterior-thigh-lean-muscle-volume-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.4241736828600753, 'std': 0.8155806070126518}, annotation_units=1, channel_map={'22404_Posterior-thigh-lean-muscle-volume-right_2_0': 0, })
ukb_22405_2 = TensorMap('22405_Anterior-thigh-lean-muscle-volume-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7013542580700847, 'std': 0.4698424875749462}, annotation_units=1, channel_map={'22405_Anterior-thigh-lean-muscle-volume-left_2_0': 0, })
ukb_22406_2 = TensorMap('22406_Posterior-thigh-lean-muscle-volume-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.3885738632832725, 'std': 0.8041351474364459}, annotation_units=1, channel_map={'22406_Posterior-thigh-lean-muscle-volume-left_2_0': 0, })
ukb_22407_2 = TensorMap('22407_Visceral-adipose-tissue-volume-VAT_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.712806088740617, 'std': 2.244466135149923}, annotation_units=1, channel_map={'22407_Visceral-adipose-tissue-volume-VAT_2_0': 0, })
ukb_22408_2 = TensorMap('22408_Abdominal-subcutaneous-adipose-tissue-volume-ASAT_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.033651218015013, 'std': 3.1852199542825397}, annotation_units=1, channel_map={'22408_Abdominal-subcutaneous-adipose-tissue-volume-ASAT_2_0': 0, })
ukb_22409_2 = TensorMap('22409_Total-thigh-muscle-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.196864727399408, 'std': 2.497286129837571}, annotation_units=1, channel_map={'22409_Total-thigh-muscle-volume_2_0': 0, })
ukb_22410_2 = TensorMap('22410_Total-trunk-fat-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.746456949958299, 'std': 4.541506694278862}, annotation_units=1, channel_map={'22410_Total-trunk-fat-volume_2_0': 0, })
ukb_22420_2 = TensorMap('22420_LV-ejection-fraction_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 55.873063043939716, 'std': 7.01844841794735}, annotation_units=1, channel_map={'22420_LV-ejection-fraction_2_0': 0, })
ukb_22421_2 = TensorMap('22421_LV-end-diastolic-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 141.71598386754408, 'std': 59.034996253733226}, annotation_units=1, channel_map={'22421_LV-end-diastolic-volume_2_0': 0, })
ukb_22422_2 = TensorMap('22422_LV-end-systolic-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 64.14943748673319, 'std': 47.11061383020241}, annotation_units=1, channel_map={'22422_LV-end-systolic-volume_2_0': 0, })
ukb_22423_2 = TensorMap('22423_LV-stroke-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.56039057524943, 'std': 19.859766158351505}, annotation_units=1, channel_map={'22423_LV-stroke-volume_2_0': 0, })
ukb_22424_2 = TensorMap('22424_Cardiac-output_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.749819571216302, 'std': 1.2098970273107719}, annotation_units=1, channel_map={'22424_Cardiac-output_2_0': 0, })
ukb_22425_2 = TensorMap('22425_Cardiac-index_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.560517936743791, 'std': 0.5723116488455566}, annotation_units=1, channel_map={'22425_Cardiac-index_2_0': 0, })
ukb_22426_2 = TensorMap('22426_Average-heart-rate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 62.13033326257697, 'std': 10.116518364374457}, annotation_units=1, channel_map={'22426_Average-heart-rate_2_0': 0, })
ukb_22427_2 = TensorMap('22427_Body-surface-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.85485035024411, 'std': 0.20631406268645755}, annotation_units=1, channel_map={'22427_Body-surface-area_2_0': 0, })
ukb_22599_0 = TensorMap('22599_Number-of-jobs-held_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.3231889186593193, 'std': 2.1908846910597255}, annotation_units=1, channel_map={'22599_Number-of-jobs-held_0_0': 0, })
ukb_22602_0 = TensorMap('22602_Year-job-started_0_39', loss='logcosh', path_prefix='continuous', normalization={'mean': 1982.127364921072, 'std': 13.80770390480482}, annotation_units=40, channel_map={'22602_Year-job-started_0_0': 0, '22602_Year-job-started_0_1': 1, '22602_Year-job-started_0_2': 2, '22602_Year-job-started_0_3': 3, '22602_Year-job-started_0_4': 4, '22602_Year-job-started_0_5': 5, '22602_Year-job-started_0_6': 6, '22602_Year-job-started_0_7': 7, '22602_Year-job-started_0_8': 8, '22602_Year-job-started_0_9': 9, '22602_Year-job-started_0_10': 10, '22602_Year-job-started_0_11': 11, '22602_Year-job-started_0_12': 12, '22602_Year-job-started_0_13': 13, '22602_Year-job-started_0_14': 14, '22602_Year-job-started_0_15': 15, '22602_Year-job-started_0_16': 16, '22602_Year-job-started_0_17': 17, '22602_Year-job-started_0_18': 18, '22602_Year-job-started_0_19': 19, '22602_Year-job-started_0_20': 20, '22602_Year-job-started_0_21': 21, '22602_Year-job-started_0_22': 22, '22602_Year-job-started_0_23': 23, '22602_Year-job-started_0_24': 24, '22602_Year-job-started_0_25': 25, '22602_Year-job-started_0_26': 26, '22602_Year-job-started_0_27': 27, '22602_Year-job-started_0_28': 28, '22602_Year-job-started_0_29': 29, '22602_Year-job-started_0_30': 30, '22602_Year-job-started_0_31': 31, '22602_Year-job-started_0_32': 32, '22602_Year-job-started_0_33': 33, '22602_Year-job-started_0_34': 34, '22602_Year-job-started_0_35': 35, '22602_Year-job-started_0_36': 36, '22602_Year-job-started_0_37': 37, '22602_Year-job-started_0_38': 38, '22602_Year-job-started_0_39': 39, })
ukb_22603_0 = TensorMap('22603_Year-job-ended_0_39', loss='logcosh', path_prefix='continuous', normalization={'mean': 1989.6405042745812, 'std': 14.927501935765907}, annotation_units=40, channel_map={'22603_Year-job-ended_0_0': 0, '22603_Year-job-ended_0_1': 1, '22603_Year-job-ended_0_2': 2, '22603_Year-job-ended_0_3': 3, '22603_Year-job-ended_0_4': 4, '22603_Year-job-ended_0_5': 5, '22603_Year-job-ended_0_6': 6, '22603_Year-job-ended_0_7': 7, '22603_Year-job-ended_0_8': 8, '22603_Year-job-ended_0_9': 9, '22603_Year-job-ended_0_10': 10, '22603_Year-job-ended_0_11': 11, '22603_Year-job-ended_0_12': 12, '22603_Year-job-ended_0_13': 13, '22603_Year-job-ended_0_14': 14, '22603_Year-job-ended_0_15': 15, '22603_Year-job-ended_0_16': 16, '22603_Year-job-ended_0_17': 17, '22603_Year-job-ended_0_18': 18, '22603_Year-job-ended_0_19': 19, '22603_Year-job-ended_0_20': 20, '22603_Year-job-ended_0_21': 21, '22603_Year-job-ended_0_22': 22, '22603_Year-job-ended_0_23': 23, '22603_Year-job-ended_0_24': 24, '22603_Year-job-ended_0_25': 25, '22603_Year-job-ended_0_26': 26, '22603_Year-job-ended_0_27': 27, '22603_Year-job-ended_0_28': 28, '22603_Year-job-ended_0_29': 29, '22603_Year-job-ended_0_30': 30, '22603_Year-job-ended_0_31': 31, '22603_Year-job-ended_0_32': 32, '22603_Year-job-ended_0_33': 33, '22603_Year-job-ended_0_34': 34, '22603_Year-job-ended_0_35': 35, '22603_Year-job-ended_0_36': 36, '22603_Year-job-ended_0_37': 37, '22603_Year-job-ended_0_38': 38, '22603_Year-job-ended_0_39': 39, })
ukb_22605_0 = TensorMap('22605_Work-hours-per-week-exact-value_0_34', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.377421316445606, 'std': 10.330619868423486}, annotation_units=35, channel_map={'22605_Work-hours-per-week-exact-value_0_0': 0, '22605_Work-hours-per-week-exact-value_0_1': 1, '22605_Work-hours-per-week-exact-value_0_2': 2, '22605_Work-hours-per-week-exact-value_0_3': 3, '22605_Work-hours-per-week-exact-value_0_4': 4, '22605_Work-hours-per-week-exact-value_0_5': 5, '22605_Work-hours-per-week-exact-value_0_6': 6, '22605_Work-hours-per-week-exact-value_0_7': 7, '22605_Work-hours-per-week-exact-value_0_8': 8, '22605_Work-hours-per-week-exact-value_0_9': 9, '22605_Work-hours-per-week-exact-value_0_10': 10, '22605_Work-hours-per-week-exact-value_0_11': 11, '22605_Work-hours-per-week-exact-value_0_12': 12, '22605_Work-hours-per-week-exact-value_0_13': 13, '22605_Work-hours-per-week-exact-value_0_14': 14, '22605_Work-hours-per-week-exact-value_0_15': 15, '22605_Work-hours-per-week-exact-value_0_16': 16, '22605_Work-hours-per-week-exact-value_0_17': 17, '22605_Work-hours-per-week-exact-value_0_18': 18, '22605_Work-hours-per-week-exact-value_0_19': 19, '22605_Work-hours-per-week-exact-value_0_20': 20, '22605_Work-hours-per-week-exact-value_0_21': 21, '22605_Work-hours-per-week-exact-value_0_22': 22, '22605_Work-hours-per-week-exact-value_0_23': 23, '22605_Work-hours-per-week-exact-value_0_24': 24, '22605_Work-hours-per-week-exact-value_0_25': 25, '22605_Work-hours-per-week-exact-value_0_26': 26, '22605_Work-hours-per-week-exact-value_0_27': 27, '22605_Work-hours-per-week-exact-value_0_28': 28, '22605_Work-hours-per-week-exact-value_0_29': 29, '22605_Work-hours-per-week-exact-value_0_30': 30, '22605_Work-hours-per-week-exact-value_0_31': 31, '22605_Work-hours-per-week-exact-value_0_32': 32, '22605_Work-hours-per-week-exact-value_0_33': 33, '22605_Work-hours-per-week-exact-value_0_34': 34, })
ukb_22631_0 = TensorMap('22631_Period-spent-working-day-shifts_0_14', loss='logcosh', path_prefix='continuous', normalization={'mean': -279.3364485981308, 'std': 455.1359954540528}, annotation_units=15, channel_map={'22631_Period-spent-working-day-shifts_0_0': 0, '22631_Period-spent-working-day-shifts_0_1': 1, '22631_Period-spent-working-day-shifts_0_2': 2, '22631_Period-spent-working-day-shifts_0_3': 3, '22631_Period-spent-working-day-shifts_0_4': 4, '22631_Period-spent-working-day-shifts_0_5': 5, '22631_Period-spent-working-day-shifts_0_6': 6, '22631_Period-spent-working-day-shifts_0_7': 7, '22631_Period-spent-working-day-shifts_0_8': 8, '22631_Period-spent-working-day-shifts_0_9': 9, '22631_Period-spent-working-day-shifts_0_10': 10, '22631_Period-spent-working-day-shifts_0_11': 11, '22631_Period-spent-working-day-shifts_0_12': 12, '22631_Period-spent-working-day-shifts_0_13': 13, '22631_Period-spent-working-day-shifts_0_14': 14, })
ukb_22641_0 = TensorMap('22641_Period-spent-working-mix-of-day-and-night-shifts_0_13', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.502556967737084, 'std': 6.807737162854637}, annotation_units=14, channel_map={'22641_Period-spent-working-mix-of-day-and-night-shifts_0_0': 0, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_1': 1, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_2': 2, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_3': 3, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_4': 4, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_5': 5, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_6': 6, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_7': 7, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_8': 8, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_9': 9, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_10': 10, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_11': 11, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_12': 12, '22641_Period-spent-working-mix-of-day-and-night-shifts_0_13': 13, })
ukb_22642_0 = TensorMap('22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_19', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.994711361363995, 'std': 3.089320388052217}, annotation_units=20, channel_map={'22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_0': 0, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_1': 1, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_2': 2, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_3': 3, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_4': 4, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_5': 5, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_6': 6, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_7': 7, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_8': 8, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_9': 9, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_10': 10, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_11': 11, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_12': 12, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_13': 13, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_14': 14, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_15': 15, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_16': 16, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_17': 17, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_18': 18, '22642_Usual-length-of-each-night-shift-during-mixed-shift-periods_0_19': 19, })
ukb_22643_0 = TensorMap('22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_19', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.263142661684995, 'std': 5.986238901275712}, annotation_units=20, channel_map={'22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_0': 0, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_1': 1, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_2': 2, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_3': 3, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_4': 4, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_5': 5, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_6': 6, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_7': 7, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_8': 8, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_9': 9, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_10': 10, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_11': 11, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_12': 12, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_13': 13, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_14': 14, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_15': 15, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_16': 16, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_17': 17, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_18': 18, '22643_Number-of-night-shifts-worked-monthly-during-mixed-shift-periods_0_19': 19, })
ukb_22644_0 = TensorMap('22644_Consecutive-night-shifts-during-mixed-shift-periods_0_19', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.0698310793032695, 'std': 3.34691790276513}, annotation_units=20, channel_map={'22644_Consecutive-night-shifts-during-mixed-shift-periods_0_0': 0, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_1': 1, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_2': 2, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_3': 3, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_4': 4, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_5': 5, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_6': 6, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_7': 7, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_8': 8, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_9': 9, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_10': 10, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_11': 11, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_12': 12, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_13': 13, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_14': 14, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_15': 15, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_16': 16, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_17': 17, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_18': 18, '22644_Consecutive-night-shifts-during-mixed-shift-periods_0_19': 19, })
ukb_22645_0 = TensorMap('22645_Rest-days-during-mixed-shift-periods_0_19', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8234570558750267, 'std': 2.5939951270035366}, annotation_units=20, channel_map={'22645_Rest-days-during-mixed-shift-periods_0_0': 0, '22645_Rest-days-during-mixed-shift-periods_0_1': 1, '22645_Rest-days-during-mixed-shift-periods_0_2': 2, '22645_Rest-days-during-mixed-shift-periods_0_3': 3, '22645_Rest-days-during-mixed-shift-periods_0_4': 4, '22645_Rest-days-during-mixed-shift-periods_0_5': 5, '22645_Rest-days-during-mixed-shift-periods_0_6': 6, '22645_Rest-days-during-mixed-shift-periods_0_7': 7, '22645_Rest-days-during-mixed-shift-periods_0_8': 8, '22645_Rest-days-during-mixed-shift-periods_0_9': 9, '22645_Rest-days-during-mixed-shift-periods_0_10': 10, '22645_Rest-days-during-mixed-shift-periods_0_11': 11, '22645_Rest-days-during-mixed-shift-periods_0_12': 12, '22645_Rest-days-during-mixed-shift-periods_0_13': 13, '22645_Rest-days-during-mixed-shift-periods_0_14': 14, '22645_Rest-days-during-mixed-shift-periods_0_15': 15, '22645_Rest-days-during-mixed-shift-periods_0_16': 16, '22645_Rest-days-during-mixed-shift-periods_0_17': 17, '22645_Rest-days-during-mixed-shift-periods_0_18': 18, '22645_Rest-days-during-mixed-shift-periods_0_19': 19, })
ukb_22651_0 = TensorMap('22651_Period-spent-working-night-shifts_0_32', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.139737991266377, 'std': 7.250671654528354}, annotation_units=33, channel_map={'22651_Period-spent-working-night-shifts_0_0': 0, '22651_Period-spent-working-night-shifts_0_1': 1, '22651_Period-spent-working-night-shifts_0_2': 2, '22651_Period-spent-working-night-shifts_0_3': 3, '22651_Period-spent-working-night-shifts_0_4': 4, '22651_Period-spent-working-night-shifts_0_5': 5, '22651_Period-spent-working-night-shifts_0_6': 6, '22651_Period-spent-working-night-shifts_0_7': 7, '22651_Period-spent-working-night-shifts_0_8': 8, '22651_Period-spent-working-night-shifts_0_9': 9, '22651_Period-spent-working-night-shifts_0_10': 10, '22651_Period-spent-working-night-shifts_0_11': 11, '22651_Period-spent-working-night-shifts_0_12': 12, '22651_Period-spent-working-night-shifts_0_13': 13, '22651_Period-spent-working-night-shifts_0_14': 14, '22651_Period-spent-working-night-shifts_0_15': 15, '22651_Period-spent-working-night-shifts_0_16': 16, '22651_Period-spent-working-night-shifts_0_17': 17, '22651_Period-spent-working-night-shifts_0_18': 18, '22651_Period-spent-working-night-shifts_0_19': 19, '22651_Period-spent-working-night-shifts_0_20': 20, '22651_Period-spent-working-night-shifts_0_21': 21, '22651_Period-spent-working-night-shifts_0_22': 22, '22651_Period-spent-working-night-shifts_0_23': 23, '22651_Period-spent-working-night-shifts_0_24': 24, '22651_Period-spent-working-night-shifts_0_25': 25, '22651_Period-spent-working-night-shifts_0_26': 26, '22651_Period-spent-working-night-shifts_0_27': 27, '22651_Period-spent-working-night-shifts_0_28': 28, '22651_Period-spent-working-night-shifts_0_29': 29, '22651_Period-spent-working-night-shifts_0_30': 30, '22651_Period-spent-working-night-shifts_0_31': 31, '22651_Period-spent-working-night-shifts_0_32': 32, })
ukb_22652_0 = TensorMap('22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_32', loss='logcosh', path_prefix='continuous', normalization={'mean': 10.077380952380953, 'std': 2.175321464464055}, annotation_units=33, channel_map={'22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_0': 0, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_1': 1, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_2': 2, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_3': 3, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_4': 4, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_5': 5, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_6': 6, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_7': 7, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_8': 8, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_9': 9, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_10': 10, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_11': 11, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_12': 12, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_13': 13, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_14': 14, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_15': 15, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_16': 16, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_17': 17, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_18': 18, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_19': 19, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_20': 20, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_21': 21, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_22': 22, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_23': 23, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_24': 24, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_25': 25, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_26': 26, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_27': 27, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_28': 28, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_29': 29, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_30': 30, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_31': 31, '22652_Usual-length-of-each-night-shift-during-night-shift-periods_0_32': 32, })
ukb_22653_0 = TensorMap('22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_32', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.656641604010023, 'std': 6.750318984488951}, annotation_units=33, channel_map={'22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_0': 0, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_1': 1, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_2': 2, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_3': 3, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_4': 4, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_5': 5, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_6': 6, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_7': 7, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_8': 8, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_9': 9, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_10': 10, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_11': 11, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_12': 12, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_13': 13, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_14': 14, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_15': 15, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_16': 16, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_17': 17, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_18': 18, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_19': 19, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_20': 20, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_21': 21, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_22': 22, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_23': 23, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_24': 24, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_25': 25, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_26': 26, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_27': 27, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_28': 28, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_29': 29, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_30': 30, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_31': 31, '22653_Number-of-night-shifts-worked-monthly-during-night-shift-periods_0_32': 32, })
ukb_22654_0 = TensorMap('22654_Consecutive-night-shifts-during-night-shift-periods_0_32', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.8223684210526314, 'std': 2.9717294774846077}, annotation_units=33, channel_map={'22654_Consecutive-night-shifts-during-night-shift-periods_0_0': 0, '22654_Consecutive-night-shifts-during-night-shift-periods_0_1': 1, '22654_Consecutive-night-shifts-during-night-shift-periods_0_2': 2, '22654_Consecutive-night-shifts-during-night-shift-periods_0_3': 3, '22654_Consecutive-night-shifts-during-night-shift-periods_0_4': 4, '22654_Consecutive-night-shifts-during-night-shift-periods_0_5': 5, '22654_Consecutive-night-shifts-during-night-shift-periods_0_6': 6, '22654_Consecutive-night-shifts-during-night-shift-periods_0_7': 7, '22654_Consecutive-night-shifts-during-night-shift-periods_0_8': 8, '22654_Consecutive-night-shifts-during-night-shift-periods_0_9': 9, '22654_Consecutive-night-shifts-during-night-shift-periods_0_10': 10, '22654_Consecutive-night-shifts-during-night-shift-periods_0_11': 11, '22654_Consecutive-night-shifts-during-night-shift-periods_0_12': 12, '22654_Consecutive-night-shifts-during-night-shift-periods_0_13': 13, '22654_Consecutive-night-shifts-during-night-shift-periods_0_14': 14, '22654_Consecutive-night-shifts-during-night-shift-periods_0_15': 15, '22654_Consecutive-night-shifts-during-night-shift-periods_0_16': 16, '22654_Consecutive-night-shifts-during-night-shift-periods_0_17': 17, '22654_Consecutive-night-shifts-during-night-shift-periods_0_18': 18, '22654_Consecutive-night-shifts-during-night-shift-periods_0_19': 19, '22654_Consecutive-night-shifts-during-night-shift-periods_0_20': 20, '22654_Consecutive-night-shifts-during-night-shift-periods_0_21': 21, '22654_Consecutive-night-shifts-during-night-shift-periods_0_22': 22, '22654_Consecutive-night-shifts-during-night-shift-periods_0_23': 23, '22654_Consecutive-night-shifts-during-night-shift-periods_0_24': 24, '22654_Consecutive-night-shifts-during-night-shift-periods_0_25': 25, '22654_Consecutive-night-shifts-during-night-shift-periods_0_26': 26, '22654_Consecutive-night-shifts-during-night-shift-periods_0_27': 27, '22654_Consecutive-night-shifts-during-night-shift-periods_0_28': 28, '22654_Consecutive-night-shifts-during-night-shift-periods_0_29': 29, '22654_Consecutive-night-shifts-during-night-shift-periods_0_30': 30, '22654_Consecutive-night-shifts-during-night-shift-periods_0_31': 31, '22654_Consecutive-night-shifts-during-night-shift-periods_0_32': 32, })
ukb_22655_0 = TensorMap('22655_Rest-days-during-night-shift-periods_0_16', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.340453499835688, 'std': 2.01533324901676}, annotation_units=17, channel_map={'22655_Rest-days-during-night-shift-periods_0_0': 0, '22655_Rest-days-during-night-shift-periods_0_1': 1, '22655_Rest-days-during-night-shift-periods_0_2': 2, '22655_Rest-days-during-night-shift-periods_0_3': 3, '22655_Rest-days-during-night-shift-periods_0_4': 4, '22655_Rest-days-during-night-shift-periods_0_5': 5, '22655_Rest-days-during-night-shift-periods_0_6': 6, '22655_Rest-days-during-night-shift-periods_0_7': 7, '22655_Rest-days-during-night-shift-periods_0_8': 8, '22655_Rest-days-during-night-shift-periods_0_9': 9, '22655_Rest-days-during-night-shift-periods_0_10': 10, '22655_Rest-days-during-night-shift-periods_0_11': 11, '22655_Rest-days-during-night-shift-periods_0_12': 12, '22655_Rest-days-during-night-shift-periods_0_13': 13, '22655_Rest-days-during-night-shift-periods_0_14': 14, '22655_Rest-days-during-night-shift-periods_0_15': 15, '22655_Rest-days-during-night-shift-periods_0_16': 16, })
ukb_22661_0 = TensorMap('22661_Number-of-gap-periods_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.483561666919435, 'std': 1.351709066932489}, annotation_units=1, channel_map={'22661_Number-of-gap-periods_0_0': 0, })
ukb_22663_0 = TensorMap('22663_Year-gap-started_0_32', loss='logcosh', path_prefix='continuous', normalization={'mean': 1993.5931295198686, 'std': 15.790408772806646}, annotation_units=33, channel_map={'22663_Year-gap-started_0_0': 0, '22663_Year-gap-started_0_1': 1, '22663_Year-gap-started_0_2': 2, '22663_Year-gap-started_0_3': 3, '22663_Year-gap-started_0_4': 4, '22663_Year-gap-started_0_5': 5, '22663_Year-gap-started_0_6': 6, '22663_Year-gap-started_0_7': 7, '22663_Year-gap-started_0_8': 8, '22663_Year-gap-started_0_9': 9, '22663_Year-gap-started_0_10': 10, '22663_Year-gap-started_0_11': 11, '22663_Year-gap-started_0_12': 12, '22663_Year-gap-started_0_13': 13, '22663_Year-gap-started_0_14': 14, '22663_Year-gap-started_0_15': 15, '22663_Year-gap-started_0_16': 16, '22663_Year-gap-started_0_17': 17, '22663_Year-gap-started_0_18': 18, '22663_Year-gap-started_0_19': 19, '22663_Year-gap-started_0_20': 20, '22663_Year-gap-started_0_21': 21, '22663_Year-gap-started_0_22': 22, '22663_Year-gap-started_0_23': 23, '22663_Year-gap-started_0_24': 24, '22663_Year-gap-started_0_25': 25, '22663_Year-gap-started_0_26': 26, '22663_Year-gap-started_0_27': 27, '22663_Year-gap-started_0_28': 28, '22663_Year-gap-started_0_29': 29, '22663_Year-gap-started_0_30': 30, '22663_Year-gap-started_0_31': 31, '22663_Year-gap-started_0_32': 32, })
ukb_22664_0 = TensorMap('22664_Year-gap-ended_0_31', loss='logcosh', path_prefix='continuous', normalization={'mean': 1989.0204133883462, 'std': 12.980833002400056}, annotation_units=32, channel_map={'22664_Year-gap-ended_0_0': 0, '22664_Year-gap-ended_0_1': 1, '22664_Year-gap-ended_0_2': 2, '22664_Year-gap-ended_0_3': 3, '22664_Year-gap-ended_0_4': 4, '22664_Year-gap-ended_0_5': 5, '22664_Year-gap-ended_0_6': 6, '22664_Year-gap-ended_0_7': 7, '22664_Year-gap-ended_0_8': 8, '22664_Year-gap-ended_0_9': 9, '22664_Year-gap-ended_0_10': 10, '22664_Year-gap-ended_0_11': 11, '22664_Year-gap-ended_0_12': 12, '22664_Year-gap-ended_0_13': 13, '22664_Year-gap-ended_0_14': 14, '22664_Year-gap-ended_0_15': 15, '22664_Year-gap-ended_0_16': 16, '22664_Year-gap-ended_0_17': 17, '22664_Year-gap-ended_0_18': 18, '22664_Year-gap-ended_0_19': 19, '22664_Year-gap-ended_0_20': 20, '22664_Year-gap-ended_0_21': 21, '22664_Year-gap-ended_0_22': 22, '22664_Year-gap-ended_0_23': 23, '22664_Year-gap-ended_0_24': 24, '22664_Year-gap-ended_0_25': 25, '22664_Year-gap-ended_0_26': 26, '22664_Year-gap-ended_0_27': 27, '22664_Year-gap-ended_0_28': 28, '22664_Year-gap-ended_0_29': 29, '22664_Year-gap-ended_0_30': 30, '22664_Year-gap-ended_0_31': 31, })
ukb_23098_1 = TensorMap('23098_Weight_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.89022200049888, 'std': 15.400899817937916}, annotation_units=1, channel_map={'23098_Weight_1_0': 0, })
ukb_23098_0 = TensorMap('23098_Weight_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.03287707120329, 'std': 15.89979106473436}, annotation_units=1, channel_map={'23098_Weight_0_0': 0, })
ukb_23099_0 = TensorMap('23099_Body-fat-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.451786835511985, 'std': 8.547574421386416}, annotation_units=1, channel_map={'23099_Body-fat-percentage_0_0': 0, })
ukb_23099_1 = TensorMap('23099_Body-fat-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.264657151412308, 'std': 8.292322694561557}, annotation_units=1, channel_map={'23099_Body-fat-percentage_1_0': 0, })
ukb_23099_2 = TensorMap('23099_Body-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.264657151412308, 'std': 8.292322694561557}, annotation_units=1, channel_map={'23099_Body-fat-percentage_2_0': 0, })
ukb_23100_1 = TensorMap('23100_Whole-body-fat-mass_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.313408909308837, 'std': 9.148610869548651}, annotation_units=1, channel_map={'23100_Whole-body-fat-mass_1_0': 0, })
ukb_23100_2 = TensorMap('23100_Whole-body-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.85684055488536, 'std': 9.565426323411884}, annotation_units=1, channel_map={'23100_Whole-body-fat-mass_2_0': 0, })
ukb_23101_2 = TensorMap('23101_Whole-body-fatfree-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 52.606156455797255, 'std': 11.126052649633138}, annotation_units=1, channel_map={'23101_Whole-body-fatfree-mass_2_0': 0, })
ukb_23101_1 = TensorMap('23101_Whole-body-fatfree-mass_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 52.606156455797255, 'std': 11.126052649633138}, annotation_units=1, channel_map={'23101_Whole-body-fatfree-mass_1_0': 0, })
ukb_23101_0 = TensorMap('23101_Whole-body-fatfree-mass_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.21773908875434, 'std': 11.496421819167042}, annotation_units=1, channel_map={'23101_Whole-body-fatfree-mass_0_0': 0, })
ukb_23102_0 = TensorMap('23102_Whole-body-water-mass_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 38.947460814537706, 'std': 8.414129150401765}, annotation_units=1, channel_map={'23102_Whole-body-water-mass_0_0': 0, })
ukb_23102_1 = TensorMap('23102_Whole-body-water-mass_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 38.51306061361937, 'std': 8.148777477201547}, annotation_units=1, channel_map={'23102_Whole-body-water-mass_1_0': 0, })
ukb_23104_0 = TensorMap('23104_Body-mass-index-BMI_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.431847749487705, 'std': 4.785169842092807}, annotation_units=1, channel_map={'23104_Body-mass-index-BMI_0_0': 0, })
ukb_23104_1 = TensorMap('23104_Body-mass-index-BMI_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.959885258169116, 'std': 4.5750386374357355}, annotation_units=1, channel_map={'23104_Body-mass-index-BMI_1_0': 0, })
ukb_23105_0 = TensorMap('23105_Basal-metabolic-rate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6616.133658821905, 'std': 1364.2964637736743}, annotation_units=1, channel_map={'23105_Basal-metabolic-rate_0_0': 0, })
ukb_23105_1 = TensorMap('23105_Basal-metabolic-rate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6518.47827388376, 'std': 1312.3174057551125}, annotation_units=1, channel_map={'23105_Basal-metabolic-rate_1_0': 0, })
ukb_23106_0 = TensorMap('23106_Impedance-of-whole-body_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 600.1263049712803, 'std': 88.96912814227957}, annotation_units=1, channel_map={'23106_Impedance-of-whole-body_0_0': 0, })
ukb_23106_1 = TensorMap('23106_Impedance-of-whole-body_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 604.970861191498, 'std': 89.22813483454858}, annotation_units=1, channel_map={'23106_Impedance-of-whole-body_1_0': 0, })
ukb_23106_2 = TensorMap('23106_Impedance-of-whole-body_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 605.3075723699694, 'std': 89.3136920125616}, annotation_units=1, channel_map={'23106_Impedance-of-whole-body_2_0': 0, })
ukb_23107_1 = TensorMap('23107_Impedance-of-leg-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 248.25612372162638, 'std': 35.78734884701803}, annotation_units=1, channel_map={'23107_Impedance-of-leg-right_1_0': 0, })
ukb_23107_2 = TensorMap('23107_Impedance-of-leg-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 255.85326390727502, 'std': 38.59815624280374}, annotation_units=1, channel_map={'23107_Impedance-of-leg-right_2_0': 0, })
ukb_23107_0 = TensorMap('23107_Impedance-of-leg-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 246.97691606228565, 'std': 36.023204333919224}, annotation_units=1, channel_map={'23107_Impedance-of-leg-right_0_0': 0, })
ukb_23108_2 = TensorMap('23108_Impedance-of-leg-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 256.6206866119495, 'std': 38.189538396513306}, annotation_units=1, channel_map={'23108_Impedance-of-leg-left_2_0': 0, })
ukb_23108_0 = TensorMap('23108_Impedance-of-leg-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 247.7682438648905, 'std': 35.70859686372287}, annotation_units=1, channel_map={'23108_Impedance-of-leg-left_0_0': 0, })
ukb_23108_1 = TensorMap('23108_Impedance-of-leg-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 249.13739087054128, 'std': 35.64318057314908}, annotation_units=1, channel_map={'23108_Impedance-of-leg-left_1_0': 0, })
ukb_23109_1 = TensorMap('23109_Impedance-of-arm-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 330.0137690197055, 'std': 55.98571896818939}, annotation_units=1, channel_map={'23109_Impedance-of-arm-right_1_0': 0, })
ukb_23109_0 = TensorMap('23109_Impedance-of-arm-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 325.5549915810041, 'std': 55.414078193098426}, annotation_units=1, channel_map={'23109_Impedance-of-arm-right_0_0': 0, })
ukb_23109_2 = TensorMap('23109_Impedance-of-arm-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 321.1530020886654, 'std': 53.38985006117276}, annotation_units=1, channel_map={'23109_Impedance-of-arm-right_2_0': 0, })
ukb_23110_1 = TensorMap('23110_Impedance-of-arm-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 337.08795210775764, 'std': 57.39185814366964}, annotation_units=1, channel_map={'23110_Impedance-of-arm-left_1_0': 0, })
ukb_23110_2 = TensorMap('23110_Impedance-of-arm-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 326.5779425175772, 'std': 54.696785486803726}, annotation_units=1, channel_map={'23110_Impedance-of-arm-left_2_0': 0, })
ukb_23110_0 = TensorMap('23110_Impedance-of-arm-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 332.26580591961624, 'std': 56.83465434478377}, annotation_units=1, channel_map={'23110_Impedance-of-arm-left_0_0': 0, })
ukb_23111_2 = TensorMap('23111_Leg-fat-percentage-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.525313045647284, 'std': 10.627681452664147}, annotation_units=1, channel_map={'23111_Leg-fat-percentage-right_2_0': 0, })
ukb_23111_1 = TensorMap('23111_Leg-fat-percentage-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.525313045647284, 'std': 10.627681452664147}, annotation_units=1, channel_map={'23111_Leg-fat-percentage-right_1_0': 0, })
ukb_23111_0 = TensorMap('23111_Leg-fat-percentage-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.04635516524227, 'std': 10.694715982950983}, annotation_units=1, channel_map={'23111_Leg-fat-percentage-right_0_0': 0, })
ukb_23112_2 = TensorMap('23112_Leg-fat-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.126630082314794, 'std': 1.8301929159686312}, annotation_units=1, channel_map={'23112_Leg-fat-mass-right_2_0': 0, })
ukb_23112_1 = TensorMap('23112_Leg-fat-mass-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.126630082314794, 'std': 1.8301929159686312}, annotation_units=1, channel_map={'23112_Leg-fat-mass-right_1_0': 0, })
ukb_23112_0 = TensorMap('23112_Leg-fat-mass-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.308648770510948, 'std': 1.898437368686469}, annotation_units=1, channel_map={'23112_Leg-fat-mass-right_0_0': 0, })
ukb_23113_2 = TensorMap('23113_Leg-fatfree-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.792466949363929, 'std': 1.9185583701790712}, annotation_units=1, channel_map={'23113_Leg-fatfree-mass-right_2_0': 0, })
ukb_23113_1 = TensorMap('23113_Leg-fatfree-mass-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.792466949363929, 'std': 1.9185583701790712}, annotation_units=1, channel_map={'23113_Leg-fatfree-mass-right_1_0': 0, })
ukb_23113_0 = TensorMap('23113_Leg-fatfree-mass-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.96966001620808, 'std': 2.0246336838262913}, annotation_units=1, channel_map={'23113_Leg-fatfree-mass-right_0_0': 0, })
ukb_23114_0 = TensorMap('23114_Leg-predicted-mass-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.48647434933969, 'std': 1.9273038036404626}, annotation_units=1, channel_map={'23114_Leg-predicted-mass-right_0_0': 0, })
ukb_23114_1 = TensorMap('23114_Leg-predicted-mass-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.322803691693691, 'std': 1.828108029602454}, annotation_units=1, channel_map={'23114_Leg-predicted-mass-right_1_0': 0, })
ukb_23115_2 = TensorMap('23115_Leg-fat-percentage-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.41810426540285, 'std': 10.597272488375483}, annotation_units=1, channel_map={'23115_Leg-fat-percentage-left_2_0': 0, })
ukb_23115_1 = TensorMap('23115_Leg-fat-percentage-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.41810426540285, 'std': 10.597272488375483}, annotation_units=1, channel_map={'23115_Leg-fat-percentage-left_1_0': 0, })
ukb_23115_0 = TensorMap('23115_Leg-fat-percentage-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.951824770075714, 'std': 10.648351373024937}, annotation_units=1, channel_map={'23115_Leg-fat-percentage-left_0_0': 0, })
ukb_23116_0 = TensorMap('23116_Leg-fat-mass-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.242730601973062, 'std': 1.871902668963224}, annotation_units=1, channel_map={'23116_Leg-fat-mass-left_0_0': 0, })
ukb_23116_1 = TensorMap('23116_Leg-fat-mass-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.064025941631329, 'std': 1.8044984851208248}, annotation_units=1, channel_map={'23116_Leg-fat-mass-left_1_0': 0, })
ukb_23116_2 = TensorMap('23116_Leg-fat-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.064025941631329, 'std': 1.8044984851208248}, annotation_units=1, channel_map={'23116_Leg-fat-mass-left_2_0': 0, })
ukb_23117_2 = TensorMap('23117_Leg-fatfree-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.697755051134946, 'std': 1.9100971996154277}, annotation_units=1, channel_map={'23117_Leg-fatfree-mass-left_2_0': 0, })
ukb_23117_1 = TensorMap('23117_Leg-fatfree-mass-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.697755051134946, 'std': 1.9100971996154277}, annotation_units=1, channel_map={'23117_Leg-fatfree-mass-left_1_0': 0, })
ukb_23117_0 = TensorMap('23117_Leg-fatfree-mass-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.866400981491417, 'std': 2.01067768245527}, annotation_units=1, channel_map={'23117_Leg-fatfree-mass-left_0_0': 0, })
ukb_23118_1 = TensorMap('23118_Leg-predicted-mass-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.23430281865802, 'std': 1.819746251807764}, annotation_units=1, channel_map={'23118_Leg-predicted-mass-left_1_0': 0, })
ukb_23118_0 = TensorMap('23118_Leg-predicted-mass-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.389605651769429, 'std': 1.914070754426116}, annotation_units=1, channel_map={'23118_Leg-predicted-mass-left_0_0': 0, })
ukb_23119_0 = TensorMap('23119_Arm-fat-percentage-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.531533189516836, 'std': 10.173278005428738}, annotation_units=1, channel_map={'23119_Arm-fat-percentage-right_0_0': 0, })
ukb_23119_1 = TensorMap('23119_Arm-fat-percentage-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.08498378648042, 'std': 9.660940777028525}, annotation_units=1, channel_map={'23119_Arm-fat-percentage-right_1_0': 0, })
ukb_23119_2 = TensorMap('23119_Arm-fat-percentage-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.08498378648042, 'std': 9.660940777028525}, annotation_units=1, channel_map={'23119_Arm-fat-percentage-right_2_0': 0, })
ukb_23120_2 = TensorMap('23120_Arm-fat-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.190551259665752, 'std': 0.5962184768943481}, annotation_units=1, channel_map={'23120_Arm-fat-mass-right_2_0': 0, })
ukb_23120_1 = TensorMap('23120_Arm-fat-mass-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.190551259665752, 'std': 0.5962184768943481}, annotation_units=1, channel_map={'23120_Arm-fat-mass-right_1_0': 0, })
ukb_23120_0 = TensorMap('23120_Arm-fat-mass-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.239808434567098, 'std': 0.6384610930020305}, annotation_units=1, channel_map={'23120_Arm-fat-mass-right_0_0': 0, })
ukb_23121_2 = TensorMap('23121_Arm-fatfree-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.838897480668496, 'std': 0.7716166268521579}, annotation_units=1, channel_map={'23121_Arm-fatfree-mass-right_2_0': 0, })
ukb_23121_1 = TensorMap('23121_Arm-fatfree-mass-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.838897480668496, 'std': 0.7716166268521579}, annotation_units=1, channel_map={'23121_Arm-fatfree-mass-right_1_0': 0, })
ukb_23121_0 = TensorMap('23121_Arm-fatfree-mass-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8931570158616644, 'std': 0.8215411754827264}, annotation_units=1, channel_map={'23121_Arm-fatfree-mass-right_0_0': 0, })
ukb_23122_0 = TensorMap('23122_Arm-predicted-mass-right_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7094604752076186, 'std': 0.7833267524137338}, annotation_units=1, channel_map={'23122_Arm-predicted-mass-right_0_0': 0, })
ukb_23122_1 = TensorMap('23122_Arm-predicted-mass-right_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.6591219755550015, 'std': 0.7367273170117283}, annotation_units=1, channel_map={'23122_Arm-predicted-mass-right_1_0': 0, })
ukb_23123_2 = TensorMap('23123_Arm-fat-percentage-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.945781990521326, 'std': 9.82011062734792}, annotation_units=1, channel_map={'23123_Arm-fat-percentage-left_2_0': 0, })
ukb_23123_1 = TensorMap('23123_Arm-fat-percentage-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.945781990521326, 'std': 9.82011062734792}, annotation_units=1, channel_map={'23123_Arm-fat-percentage-left_1_0': 0, })
ukb_23123_0 = TensorMap('23123_Arm-fat-percentage-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.42513586266444, 'std': 10.267361725229156}, annotation_units=1, channel_map={'23123_Arm-fat-percentage-left_0_0': 0, })
ukb_23124_0 = TensorMap('23124_Arm-fat-mass-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.3197648029289917, 'std': 0.7131871500892241}, annotation_units=1, channel_map={'23124_Arm-fat-mass-left_0_0': 0, })
ukb_23124_1 = TensorMap('23124_Arm-fat-mass-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2664488126122535, 'std': 0.6681665072338635}, annotation_units=1, channel_map={'23124_Arm-fat-mass-left_1_0': 0, })
ukb_23124_2 = TensorMap('23124_Arm-fat-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2664488126122535, 'std': 0.6681665072338635}, annotation_units=1, channel_map={'23124_Arm-fat-mass-left_2_0': 0, })
ukb_23125_2 = TensorMap('23125_Arm-fatfree-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8760987777500615, 'std': 0.7893732349099633}, annotation_units=1, channel_map={'23125_Arm-fatfree-mass-left_2_0': 0, })
ukb_23125_1 = TensorMap('23125_Arm-fatfree-mass-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.8760987777500615, 'std': 0.7893732349099633}, annotation_units=1, channel_map={'23125_Arm-fatfree-mass-left_1_0': 0, })
ukb_23125_0 = TensorMap('23125_Arm-fatfree-mass-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.9256461729453624, 'std': 0.8372290130379668}, annotation_units=1, channel_map={'23125_Arm-fatfree-mass-left_0_0': 0, })
ukb_23126_0 = TensorMap('23126_Arm-predicted-mass-left_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7400753431894462, 'std': 0.7977945107270243}, annotation_units=1, channel_map={'23126_Arm-predicted-mass-left_0_0': 0, })
ukb_23126_1 = TensorMap('23126_Arm-predicted-mass-left_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.6941082564230485, 'std': 0.7532883471284786}, annotation_units=1, channel_map={'23126_Arm-predicted-mass-left_1_0': 0, })
ukb_23127_2 = TensorMap('23127_Trunk-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.227587927163885, 'std': 7.733697629204932}, annotation_units=1, channel_map={'23127_Trunk-fat-percentage_2_0': 0, })
ukb_23127_1 = TensorMap('23127_Trunk-fat-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.227587927163885, 'std': 7.733697629204932}, annotation_units=1, channel_map={'23127_Trunk-fat-percentage_1_0': 0, })
ukb_23127_0 = TensorMap('23127_Trunk-fat-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.17342071305921, 'std': 8.008510881672985}, annotation_units=1, channel_map={'23127_Trunk-fat-percentage_0_0': 0, })
ukb_23128_2 = TensorMap('23128_Trunk-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.669508605637319, 'std': 5.019106497864643}, annotation_units=1, channel_map={'23128_Trunk-fat-mass_2_0': 0, })
ukb_23128_1 = TensorMap('23128_Trunk-fat-mass_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.669508605637319, 'std': 5.019106497864643}, annotation_units=1, channel_map={'23128_Trunk-fat-mass_1_0': 0, })
ukb_23128_0 = TensorMap('23128_Trunk-fat-mass_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.735808957116548, 'std': 5.1711708664366345}, annotation_units=1, channel_map={'23128_Trunk-fat-mass_0_0': 0, })
ukb_23129_0 = TensorMap('23129_Trunk-fatfree-mass_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.585953690492165, 'std': 5.981176227799396}, annotation_units=1, channel_map={'23129_Trunk-fatfree-mass_0_0': 0, })
ukb_23129_1 = TensorMap('23129_Trunk-fatfree-mass_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.42263906211025, 'std': 5.885068008414232}, annotation_units=1, channel_map={'23129_Trunk-fatfree-mass_1_0': 0, })
ukb_23129_2 = TensorMap('23129_Trunk-fatfree-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.42263906211025, 'std': 5.885068008414232}, annotation_units=1, channel_map={'23129_Trunk-fatfree-mass_2_0': 0, })
ukb_23130_0 = TensorMap('23130_Trunk-predicted-mass_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.36869182530567, 'std': 5.804851285767162}, annotation_units=1, channel_map={'23130_Trunk-predicted-mass_0_0': 0, })
ukb_23130_1 = TensorMap('23130_Trunk-predicted-mass_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.2279920179596, 'std': 5.708378220103142}, annotation_units=1, channel_map={'23130_Trunk-predicted-mass_1_0': 0, })
ukb_23200_2 = TensorMap('23200_L1L4-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.829805862939246, 'std': 8.280381408666097}, annotation_units=1, channel_map={'23200_L1L4-area_2_0': 0, })
ukb_23201_2 = TensorMap('23201_L1L4-average-height_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.747485925063092, 'std': 0.9437726341635638}, annotation_units=1, channel_map={'23201_L1L4-average-height_2_0': 0, })
ukb_23202_2 = TensorMap('23202_L1L4-average-width_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.345530964861191, 'std': 0.6502154568692372}, annotation_units=1, channel_map={'23202_L1L4-average-width_2_0': 0, })
ukb_23203_2 = TensorMap('23203_L1L4-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 72.06511939429237, 'std': 18.674986125215845}, annotation_units=1, channel_map={'23203_L1L4-BMC-bone-mineral-content_2_0': 0, })
ukb_23204_2 = TensorMap('23204_L1L4-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.192932828576975, 'std': 0.19841542511613613}, annotation_units=1, channel_map={'23204_L1L4-BMD-bone-mineral-density_2_0': 0, })
ukb_23205_2 = TensorMap('23205_L1L4-BMD-bone-mineral-density-Tscore_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.049648437500000024, 'std': 1.6092135195285258}, annotation_units=1, channel_map={'23205_L1L4-BMD-bone-mineral-density-Tscore_2_0': 0, })
ukb_23206_2 = TensorMap('23206_Femur-lower-neck-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1261894197952222, 'std': 0.16620142649462438}, annotation_units=1, channel_map={'23206_Femur-lower-neck-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23208_2 = TensorMap('23208_Femur-neck-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9528873239436619, 'std': 0.1511936465486992}, annotation_units=1, channel_map={'23208_Femur-neck-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23209_2 = TensorMap('23209_Femur-neck-BMD-bone-mineral-density-Tscore-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.5446709376820035, 'std': 1.123089637315089}, annotation_units=1, channel_map={'23209_Femur-neck-BMD-bone-mineral-density-Tscore-right_2_0': 0, })
ukb_23210_2 = TensorMap('23210_Femur-shaft-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.190519543343653, 'std': 0.19402535554951122}, annotation_units=1, channel_map={'23210_Femur-shaft-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23212_2 = TensorMap('23212_Femur-total-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0087637383900931, 'std': 0.1638536053074267}, annotation_units=1, channel_map={'23212_Femur-total-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23213_2 = TensorMap('23213_Femur-total-BMD-bone-mineral-density-Tscore-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.26550515865291024, 'std': 1.182821577241909}, annotation_units=1, channel_map={'23213_Femur-total-BMD-bone-mineral-density-Tscore-right_2_0': 0, })
ukb_23214_2 = TensorMap('23214_Femur-troch-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.841719853366776, 'std': 0.1673365117026208}, annotation_units=1, channel_map={'23214_Femur-troch-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23215_2 = TensorMap('23215_Femur-troch-BMD-bone-mineral-density-Tscore-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.12081149291399731, 'std': 1.2799733771472557}, annotation_units=1, channel_map={'23215_Femur-troch-BMD-bone-mineral-density-Tscore-right_2_0': 0, })
ukb_23216_2 = TensorMap('23216_Femur-upper-neck-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7751139078498294, 'std': 0.15622054595504034}, annotation_units=1, channel_map={'23216_Femur-upper-neck-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23217_2 = TensorMap('23217_Femur-upper-neck-BMD-bone-mineral-density-Tscore-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.68983704974271, 'std': 1.1531519370816836}, annotation_units=1, channel_map={'23217_Femur-upper-neck-BMD-bone-mineral-density-Tscore-right_2_0': 0, })
ukb_23218_2 = TensorMap('23218_Femur-wards-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7461217441636118, 'std': 0.15861385048627122}, annotation_units=1, channel_map={'23218_Femur-wards-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23219_2 = TensorMap('23219_Femur-wards-BMD-bone-mineral-density-Tscore-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -1.430601941747573, 'std': 1.1712866584220247}, annotation_units=1, channel_map={'23219_Femur-wards-BMD-bone-mineral-density-Tscore-right_2_0': 0, })
ukb_23220_2 = TensorMap('23220_Arm-BMC-bone-mineral-content-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 178.61241379310349, 'std': 49.96474850607681}, annotation_units=1, channel_map={'23220_Arm-BMC-bone-mineral-content-left_2_0': 0, })
ukb_23221_2 = TensorMap('23221_Arm-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9289678160919542, 'std': 0.14689063282597334}, annotation_units=1, channel_map={'23221_Arm-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23222_2 = TensorMap('23222_Arm-BMC-bone-mineral-content-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 185.07885057471267, 'std': 50.77053992811699}, annotation_units=1, channel_map={'23222_Arm-BMC-bone-mineral-content-right_2_0': 0, })
ukb_23223_2 = TensorMap('23223_Arm-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9394977011494252, 'std': 0.14775684713923068}, annotation_units=1, channel_map={'23223_Arm-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23224_2 = TensorMap('23224_Arms-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 363.09864603481634, 'std': 99.98375345924187}, annotation_units=1, channel_map={'23224_Arms-BMC-bone-mineral-content_2_0': 0, })
ukb_23225_2 = TensorMap('23225_Arms-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9342775628626693, 'std': 0.14588362570702634}, annotation_units=1, channel_map={'23225_Arms-BMD-bone-mineral-density_2_0': 0, })
ukb_23226_2 = TensorMap('23226_Head-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.3210990328820107, 'std': 0.3192063514023758}, annotation_units=1, channel_map={'23226_Head-BMD-bone-mineral-density_2_0': 0, })
ukb_23227_2 = TensorMap('23227_Leg-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2391468965517245, 'std': 0.18568648562418075}, annotation_units=1, channel_map={'23227_Leg-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23228_2 = TensorMap('23228_Leg-BMC-bone-mineral-content-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 492.62597701149423, 'std': 118.88160951486836}, annotation_units=1, channel_map={'23228_Leg-BMC-bone-mineral-content-right_2_0': 0, })
ukb_23229_2 = TensorMap('23229_Leg-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2497655172413793, 'std': 0.18596518267425682}, annotation_units=1, channel_map={'23229_Leg-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23230_2 = TensorMap('23230_Legs-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 981.4075435203096, 'std': 236.2089922220141}, annotation_units=1, channel_map={'23230_Legs-BMC-bone-mineral-content_2_0': 0, })
ukb_23231_2 = TensorMap('23231_Legs-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2426439071566726, 'std': 0.1839992006221099}, annotation_units=1, channel_map={'23231_Legs-BMD-bone-mineral-density_2_0': 0, })
ukb_23232_2 = TensorMap('23232_Pelvis-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.016658413926499, 'std': 0.15418581069970355}, annotation_units=1, channel_map={'23232_Pelvis-BMD-bone-mineral-density_2_0': 0, })
ukb_23233_2 = TensorMap('23233_Ribs-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.862888007736944, 'std': 0.14372852933365288}, annotation_units=1, channel_map={'23233_Ribs-BMD-bone-mineral-density_2_0': 0, })
ukb_23234_2 = TensorMap('23234_Spine-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1008007736943914, 'std': 0.17834454276161832}, annotation_units=1, channel_map={'23234_Spine-BMD-bone-mineral-density_2_0': 0, })
ukb_23235_2 = TensorMap('23235_Total-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2636.6630560928434, 'std': 569.1671210240542}, annotation_units=1, channel_map={'23235_Total-BMC-bone-mineral-content_2_0': 0, })
ukb_23236_2 = TensorMap('23236_Total-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2090688588007736, 'std': 0.15015346987862968}, annotation_units=1, channel_map={'23236_Total-BMD-bone-mineral-density_2_0': 0, })
ukb_23237_2 = TensorMap('23237_Total-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1955160919540233, 'std': 0.1505259319544245}, annotation_units=1, channel_map={'23237_Total-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23238_2 = TensorMap('23238_Total-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2236802298850573, 'std': 0.15243824724378388}, annotation_units=1, channel_map={'23238_Total-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23239_2 = TensorMap('23239_Total-BMD-bone-mineral-density-Tscore_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7110527339949405, 'std': 1.2584190750511246}, annotation_units=1, channel_map={'23239_Total-BMD-bone-mineral-density-Tscore_2_0': 0, })
ukb_23240_2 = TensorMap('23240_Trunk-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 776.7123791102514, 'std': 201.55227524673637}, annotation_units=1, channel_map={'23240_Trunk-BMC-bone-mineral-content_2_0': 0, })
ukb_23241_2 = TensorMap('23241_Trunk-BMD-bone-mineral-density_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9888796905222439, 'std': 0.14621254055773872}, annotation_units=1, channel_map={'23241_Trunk-BMD-bone-mineral-density_2_0': 0, })
ukb_23242_2 = TensorMap('23242_Trunk-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.976147356321839, 'std': 0.14391214533440977}, annotation_units=1, channel_map={'23242_Trunk-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23243_2 = TensorMap('23243_Trunk-BMD-bone-mineral-density-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0027404597701146, 'std': 0.149235440920196}, annotation_units=1, channel_map={'23243_Trunk-BMD-bone-mineral-density-right_2_0': 0, })
ukb_23244_2 = TensorMap('23244_Android-bone-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49.02437137330753, 'std': 13.129273234710652}, annotation_units=1, channel_map={'23244_Android-bone-mass_2_0': 0, })
ukb_23245_2 = TensorMap('23245_Android-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2454.00754352031, 'std': 1223.6535984426241}, annotation_units=1, channel_map={'23245_Android-fat-mass_2_0': 0, })
ukb_23246_2 = TensorMap('23246_Android-lean-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3474.6061895551265, 'std': 765.0217147682314}, annotation_units=1, channel_map={'23246_Android-lean-mass_2_0': 0, })
ukb_23247_2 = TensorMap('23247_Android-tissue-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.490367504835596, 'std': 10.987487037844383}, annotation_units=1, channel_map={'23247_Android-tissue-fat-percentage_2_0': 0, })
ukb_23248_2 = TensorMap('23248_Android-total-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.9763249516440995, 'std': 1.711538469872076}, annotation_units=1, channel_map={'23248_Android-total-mass_2_0': 0, })
ukb_23249_2 = TensorMap('23249_Arm-fat-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1319.1310344827587, 'std': 491.2709514020777}, annotation_units=1, channel_map={'23249_Arm-fat-mass-left_2_0': 0, })
ukb_23250_2 = TensorMap('23250_Arm-lean-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2625.320689655172, 'std': 797.8234085467117}, annotation_units=1, channel_map={'23250_Arm-lean-mass-left_2_0': 0, })
ukb_23251_2 = TensorMap('23251_Arm-tissue-fat-percentage-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.81149425287358, 'std': 9.968969391135532}, annotation_units=1, channel_map={'23251_Arm-tissue-fat-percentage-left_2_0': 0, })
ukb_23252_2 = TensorMap('23252_Arm-total-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.123540229885059, 'std': 0.9863647871464949}, annotation_units=1, channel_map={'23252_Arm-total-mass-left_2_0': 0, })
ukb_23253_2 = TensorMap('23253_Arm-fat-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1349.1085057471262, 'std': 492.1359868640303}, annotation_units=1, channel_map={'23253_Arm-fat-mass-right_2_0': 0, })
ukb_23254_2 = TensorMap('23254_Arm-lean-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2758.77448275862, 'std': 813.385262430146}, annotation_units=1, channel_map={'23254_Arm-lean-mass-right_2_0': 0, })
ukb_23255_2 = TensorMap('23255_Arm-tissue-fat-percentage-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.20367816091954, 'std': 9.716473640275426}, annotation_units=1, channel_map={'23255_Arm-tissue-fat-percentage-right_2_0': 0, })
ukb_23256_2 = TensorMap('23256_Arm-total-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.293655172413794, 'std': 0.9931178311848479}, annotation_units=1, channel_map={'23256_Arm-total-mass-right_2_0': 0, })
ukb_23257_2 = TensorMap('23257_Arms-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2665.6292069632495, 'std': 980.5539900148582}, annotation_units=1, channel_map={'23257_Arms-fat-mass_2_0': 0, })
ukb_23258_2 = TensorMap('23258_Arms-lean-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5374.296324951645, 'std': 1598.1323374754575}, annotation_units=1, channel_map={'23258_Arms-lean-mass_2_0': 0, })
ukb_23259_2 = TensorMap('23259_Arms-tissue-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.51253384912961, 'std': 9.824731092215913}, annotation_units=1, channel_map={'23259_Arms-tissue-fat-percentage_2_0': 0, })
ukb_23260_2 = TensorMap('23260_Arms-total-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.402823984526115, 'std': 1.961728955103614}, annotation_units=1, channel_map={'23260_Arms-total-mass_2_0': 0, })
ukb_23261_2 = TensorMap('23261_Gynoid-bone-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 273.7969052224372, 'std': 67.39087214151851}, annotation_units=1, channel_map={'23261_Gynoid-bone-mass_2_0': 0, })
ukb_23262_2 = TensorMap('23262_Gynoid-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4169.788781431334, 'std': 1540.6770344188394}, annotation_units=1, channel_map={'23262_Gynoid-fat-mass_2_0': 0, })
ukb_23263_2 = TensorMap('23263_Gynoid-lean-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7313.69497098646, 'std': 1601.461852253468}, annotation_units=1, channel_map={'23263_Gynoid-lean-mass_2_0': 0, })
ukb_23264_2 = TensorMap('23264_Gynoid-tissue-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 35.961992263056075, 'std': 9.296591460521656}, annotation_units=1, channel_map={'23264_Gynoid-tissue-fat-percentage_2_0': 0, })
ukb_23265_2 = TensorMap('23265_Gynoid-total-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 11.757369439071566, 'std': 2.292236437699385}, annotation_units=1, channel_map={'23265_Gynoid-total-mass_2_0': 0, })
ukb_23266_2 = TensorMap('23266_Leg-fat-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3843.314712643678, 'std': 1539.0427882636404}, annotation_units=1, channel_map={'23266_Leg-fat-mass-left_2_0': 0, })
ukb_23267_2 = TensorMap('23267_Leg-lean-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7935.57356321839, 'std': 1848.8141159427614}, annotation_units=1, channel_map={'23267_Leg-lean-mass-left_2_0': 0, })
ukb_23268_2 = TensorMap('23268_Leg-tissue-fat-percentage-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.413517241379324, 'std': 9.62576046198015}, annotation_units=1, channel_map={'23268_Leg-tissue-fat-percentage-left_2_0': 0, })
ukb_23269_2 = TensorMap('23269_Leg-total-mass-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.268942528735634, 'std': 2.4518473303319612}, annotation_units=1, channel_map={'23269_Leg-total-mass-left_2_0': 0, })
ukb_23270_2 = TensorMap('23270_Leg-fat-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3913.9066666666668, 'std': 1548.6971689635848}, annotation_units=1, channel_map={'23270_Leg-fat-mass-right_2_0': 0, })
ukb_23271_2 = TensorMap('23271_Leg-lean-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8036.686896551724, 'std': 1872.447406921876}, annotation_units=1, channel_map={'23271_Leg-lean-mass-right_2_0': 0, })
ukb_23272_2 = TensorMap('23272_Leg-tissue-fat-percentage-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.549770114942525, 'std': 9.580209908097942}, annotation_units=1, channel_map={'23272_Leg-tissue-fat-percentage-right_2_0': 0, })
ukb_23273_2 = TensorMap('23273_Leg-total-mass-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.443103448275862, 'std': 2.4738470214708066}, annotation_units=1, channel_map={'23273_Leg-total-mass-right_2_0': 0, })
ukb_23274_2 = TensorMap('23274_Legs-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7771.539458413927, 'std': 3102.6891511009517}, annotation_units=1, channel_map={'23274_Legs-fat-mass_2_0': 0, })
ukb_23275_2 = TensorMap('23275_Legs-lean-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15965.519342359767, 'std': 3709.9909367460996}, annotation_units=1, channel_map={'23275_Legs-lean-mass_2_0': 0, })
ukb_23276_2 = TensorMap('23276_Legs-tissue-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.51330754352033, 'std': 9.604290897562407}, annotation_units=1, channel_map={'23276_Legs-tissue-fat-percentage_2_0': 0, })
ukb_23277_2 = TensorMap('23277_Legs-total-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.718607350096715, 'std': 4.947861451291104}, annotation_units=1, channel_map={'23277_Legs-total-mass_2_0': 0, })
ukb_23278_2 = TensorMap('23278_Total-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25978.565570599625, 'std': 9367.811087276115}, annotation_units=1, channel_map={'23278_Total-fat-mass_2_0': 0, })
ukb_23279_2 = TensorMap('23279_Total-fatfree-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 49891.71199226306, 'std': 10312.50128707526}, annotation_units=1, channel_map={'23279_Total-fatfree-mass_2_0': 0, })
ukb_23280_2 = TensorMap('23280_Total-lean-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47255.056479690524, 'std': 9812.656733986454}, annotation_units=1, channel_map={'23280_Total-lean-mass_2_0': 0, })
ukb_23281_2 = TensorMap('23281_Total-tissue-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 35.017098646034825, 'std': 8.24891474104609}, annotation_units=1, channel_map={'23281_Total-tissue-fat-percentage_2_0': 0, })
ukb_23282_2 = TensorMap('23282_Total-tissue-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 73233.62417794972, 'std': 14992.089234201609}, annotation_units=1, channel_map={'23282_Total-tissue-mass_2_0': 0, })
ukb_23283_2 = TensorMap('23283_Total-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 75.87054158607349, 'std': 15.380383897614125}, annotation_units=1, channel_map={'23283_Total-mass_2_0': 0, })
ukb_23284_2 = TensorMap('23284_Trunk-fat-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14658.281044487429, 'std': 6147.278394669261}, annotation_units=1, channel_map={'23284_Trunk-fat-mass_2_0': 0, })
ukb_23285_2 = TensorMap('23285_Trunk-lean-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 22778.582591876206, 'std': 4430.941670622264}, annotation_units=1, channel_map={'23285_Trunk-lean-mass_2_0': 0, })
ukb_23286_2 = TensorMap('23286_Trunk-tissue-fat-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 38.01613152804643, 'std': 9.510811624154739}, annotation_units=1, channel_map={'23286_Trunk-tissue-fat-percentage_2_0': 0, })
ukb_23287_2 = TensorMap('23287_Trunk-total-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 38.213017408123804, 'std': 8.776005535157296}, annotation_units=1, channel_map={'23287_Trunk-total-mass_2_0': 0, })
ukb_23288_2 = TensorMap('23288_VAT-visceral-adipose-tissue-mass_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1217.2716774319827, 'std': 904.1393932713576}, annotation_units=1, channel_map={'23288_VAT-visceral-adipose-tissue-mass_2_0': 0, })
ukb_23289_2 = TensorMap('23289_VAT-visceral-adipose-tissue-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1290.307692307692, 'std': 958.3865604226946}, annotation_units=1, channel_map={'23289_VAT-visceral-adipose-tissue-volume_2_0': 0, })
ukb_23290_2 = TensorMap('23290_Femur-shaft-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1897502420135528, 'std': 0.19312754091249473}, annotation_units=1, channel_map={'23290_Femur-shaft-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23291_2 = TensorMap('23291_Femur-total-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0077095837366892, 'std': 0.16325008162622526}, annotation_units=1, channel_map={'23291_Femur-total-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23292_2 = TensorMap('23292_Femur-upper-neck-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7654942430703626, 'std': 0.15649949864747772}, annotation_units=1, channel_map={'23292_Femur-upper-neck-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23293_2 = TensorMap('23293_Femur-total-BMD-bone-mineral-density-Tscore-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.27341254382547714, 'std': 1.1783273113714956}, annotation_units=1, channel_map={'23293_Femur-total-BMD-bone-mineral-density-Tscore-left_2_0': 0, })
ukb_23295_2 = TensorMap('23295_Femur-troch-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8440169785838317, 'std': 0.1668881640577414}, annotation_units=1, channel_map={'23295_Femur-troch-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23296_2 = TensorMap('23296_Femur-upper-neck-BMD-bone-mineral-density-Tscore-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.7632119914346893, 'std': 1.1549138544355584}, annotation_units=1, channel_map={'23296_Femur-upper-neck-BMD-bone-mineral-density-Tscore-left_2_0': 0, })
ukb_23297_2 = TensorMap('23297_Femur-wards-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7391385298089906, 'std': 0.15680427842400219}, annotation_units=1, channel_map={'23297_Femur-wards-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23298_2 = TensorMap('23298_Femur-troch-BMD-bone-mineral-density-Tscore-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.09753445932828579, 'std': 1.2647789390201845}, annotation_units=1, channel_map={'23298_Femur-troch-BMD-bone-mineral-density-Tscore-left_2_0': 0, })
ukb_23299_2 = TensorMap('23299_Femur-neck-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9477680879799345, 'std': 0.15011018418204117}, annotation_units=1, channel_map={'23299_Femur-neck-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23300_2 = TensorMap('23300_Femur-neck-BMD-bone-mineral-density-Tscore-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.5846855590062113, 'std': 1.1184169493740577}, annotation_units=1, channel_map={'23300_Femur-neck-BMD-bone-mineral-density-Tscore-left_2_0': 0, })
ukb_23301_2 = TensorMap('23301_Femur-wards-BMD-bone-mineral-density-Tscore-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -1.4851873422636386, 'std': 1.157095442456807}, annotation_units=1, channel_map={'23301_Femur-wards-BMD-bone-mineral-density-Tscore-left_2_0': 0, })
ukb_23302_2 = TensorMap('23302_Femur-lower-neck-BMD-bone-mineral-density-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.121624307036247, 'std': 0.16656900098186583}, annotation_units=1, channel_map={'23302_Femur-lower-neck-BMD-bone-mineral-density-left_2_0': 0, })
ukb_23304_2 = TensorMap('23304_Trunk-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 774.9088974854934, 'std': 104.63513894636239}, annotation_units=1, channel_map={'23304_Trunk-bone-area_2_0': 0, })
ukb_23305_2 = TensorMap('23305_Head-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 221.93907156673117, 'std': 20.29418865488097}, annotation_units=1, channel_map={'23305_Head-bone-area_2_0': 0, })
ukb_23306_2 = TensorMap('23306_Head-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 515.4665377176015, 'std': 85.10823479041316}, annotation_units=1, channel_map={'23306_Head-BMC-bone-mineral-content_2_0': 0, })
ukb_23307_2 = TensorMap('23307_Pelvis-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 326.5013539651836, 'std': 43.57673465460346}, annotation_units=1, channel_map={'23307_Pelvis-bone-area_2_0': 0, })
ukb_23308_2 = TensorMap('23308_Pelvis-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 334.877756286267, 'std': 81.39791504102728}, annotation_units=1, channel_map={'23308_Pelvis-BMC-bone-mineral-content_2_0': 0, })
ukb_23309_2 = TensorMap('23309_Ribs-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 251.99516441005795, 'std': 47.70315300439073}, annotation_units=1, channel_map={'23309_Ribs-bone-area_2_0': 0, })
ukb_23310_2 = TensorMap('23310_Ribs-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 222.46421663442945, 'std': 73.16900177449878}, annotation_units=1, channel_map={'23310_Ribs-BMC-bone-mineral-content_2_0': 0, })
ukb_23311_2 = TensorMap('23311_Spine-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 196.41257253384913, 'std': 27.373080665712642}, annotation_units=1, channel_map={'23311_Spine-bone-area_2_0': 0, })
ukb_23312_2 = TensorMap('23312_Spine-BMC-bone-mineral-content_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 219.35338491295934, 'std': 60.40750467740856}, annotation_units=1, channel_map={'23312_Spine-BMC-bone-mineral-content_2_0': 0, })
ukb_23313_2 = TensorMap('23313_Arm-bone-area-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 189.76827586206903, 'std': 31.154110817807545}, annotation_units=1, channel_map={'23313_Arm-bone-area-left_2_0': 0, })
ukb_23314_2 = TensorMap('23314_Arm-bone-area-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 194.56758620689658, 'std': 31.14574448175661}, annotation_units=1, channel_map={'23314_Arm-bone-area-right_2_0': 0, })
ukb_23315_2 = TensorMap('23315_Leg-bone-area-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 391.1229885057471, 'std': 47.98438851176848}, annotation_units=1, channel_map={'23315_Leg-bone-area-left_2_0': 0, })
ukb_23316_2 = TensorMap('23316_Leg-bone-area-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 390.0793103448275, 'std': 47.756249629675615}, annotation_units=1, channel_map={'23316_Leg-bone-area-right_2_0': 0, })
ukb_23317_2 = TensorMap('23317_Arms-combined-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 383.98607350096717, 'std': 69.32704044534736}, annotation_units=1, channel_map={'23317_Arms-combined-bone-area_2_0': 0, })
ukb_23318_2 = TensorMap('23318_Legs-combined-bone-area_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 781.9789168278533, 'std': 104.08771125891757}, annotation_units=1, channel_map={'23318_Legs-combined-bone-area_2_0': 0, })
ukb_23320_2 = TensorMap('23320_Leg-BMC-bone-mineral-content-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 489.74942528735625, 'std': 118.5906420044152}, annotation_units=1, channel_map={'23320_Leg-BMC-bone-mineral-content-left_2_0': 0, })
ukb_25000_2 = TensorMap('25000_Volumetric-scaling-from-T1-head-image-to-standard-space_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2986842811229975, 'std': 0.12257002052536026}, annotation_units=1, channel_map={'25000_Volumetric-scaling-from-T1-head-image-to-standard-space_2_0': 0, })
ukb_25001_2 = TensorMap('25001_Volume-of-peripheral-cortical-grey-matter-normalised-for-head-size_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 621318.6602980333, 'std': 40623.72277718259}, annotation_units=1, channel_map={'25001_Volume-of-peripheral-cortical-grey-matter-normalised-for-head-size_2_0': 0, })
ukb_25002_2 = TensorMap('25002_Volume-of-peripheral-cortical-grey-matter_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 481727.0592329612, 'std': 46063.006550528386}, annotation_units=1, channel_map={'25002_Volume-of-peripheral-cortical-grey-matter_2_0': 0, })
ukb_25003_2 = TensorMap('25003_Volume-of-ventricular-cerebrospinal-fluid-normalised-for-head-size_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45149.886049890214, 'std': 19335.83352116443}, annotation_units=1, channel_map={'25003_Volume-of-ventricular-cerebrospinal-fluid-normalised-for-head-size_2_0': 0, })
ukb_25004_2 = TensorMap('25004_Volume-of-ventricular-cerebrospinal-fluid_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 35453.93290839445, 'std': 16672.586716324793}, annotation_units=1, channel_map={'25004_Volume-of-ventricular-cerebrospinal-fluid_2_0': 0, })
ukb_25005_2 = TensorMap('25005_Volume-of-grey-matter-normalised-for-head-size_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 795072.9180174709, 'std': 48150.86599016548}, annotation_units=1, channel_map={'25005_Volume-of-grey-matter-normalised-for-head-size_2_0': 0, })
ukb_25006_2 = TensorMap('25006_Volume-of-grey-matter_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 616329.1279955154, 'std': 55685.18586763382}, annotation_units=1, channel_map={'25006_Volume-of-grey-matter_2_0': 0, })
ukb_25007_2 = TensorMap('25007_Volume-of-white-matter-normalised-for-head-size_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 707281.8634558789, 'std': 40955.823132033365}, annotation_units=1, channel_map={'25007_Volume-of-white-matter-normalised-for-head-size_2_0': 0, })
ukb_25008_2 = TensorMap('25008_Volume-of-white-matter_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 549610.8523380202, 'std': 62098.74161127222}, annotation_units=1, channel_map={'25008_Volume-of-white-matter_2_0': 0, })
ukb_25009_2 = TensorMap('25009_Volume-of-brain-greywhite-matter-normalised-for-head-size_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1502354.7988975572, 'std': 72718.63914765933}, annotation_units=1, channel_map={'25009_Volume-of-brain-greywhite-matter-normalised-for-head-size_2_0': 0, })
ukb_25010_2 = TensorMap('25010_Volume-of-brain-greywhite-matter_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1165939.9963096187, 'std': 111513.20334568707}, annotation_units=1, channel_map={'25010_Volume-of-brain-greywhite-matter_2_0': 0, })
ukb_25011_2 = TensorMap('25011_Volume-of-thalamus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7775.053903693313, 'std': 754.828448750019}, annotation_units=1, channel_map={'25011_Volume-of-thalamus-left_2_0': 0, })
ukb_25012_2 = TensorMap('25012_Volume-of-thalamus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7581.622907900888, 'std': 731.3288233641989}, annotation_units=1, channel_map={'25012_Volume-of-thalamus-right_2_0': 0, })
ukb_25013_2 = TensorMap('25013_Volume-of-caudate-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3384.333146330061, 'std': 420.64184685657324}, annotation_units=1, channel_map={'25013_Volume-of-caudate-left_2_0': 0, })
ukb_25014_2 = TensorMap('25014_Volume-of-caudate-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3563.699906498364, 'std': 441.8958089478498}, annotation_units=1, channel_map={'25014_Volume-of-caudate-right_2_0': 0, })
ukb_25015_2 = TensorMap('25015_Volume-of-putamen-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4795.005142589998, 'std': 599.1196603785951}, annotation_units=1, channel_map={'25015_Volume-of-putamen-left_2_0': 0, })
ukb_25016_2 = TensorMap('25016_Volume-of-putamen-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4844.9104721832655, 'std': 587.67124630837}, annotation_units=1, channel_map={'25016_Volume-of-putamen-right_2_0': 0, })
ukb_25017_2 = TensorMap('25017_Volume-of-pallidum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1758.6637681159411, 'std': 242.0310197718099}, annotation_units=1, channel_map={'25017_Volume-of-pallidum-left_2_0': 0, })
ukb_25018_2 = TensorMap('25018_Volume-of-pallidum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1802.9363253856943, 'std': 241.69443010608802}, annotation_units=1, channel_map={'25018_Volume-of-pallidum-right_2_0': 0, })
ukb_25019_2 = TensorMap('25019_Volume-of-hippocampus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3796.4228611500703, 'std': 477.08247659799434}, annotation_units=1, channel_map={'25019_Volume-of-hippocampus-left_2_0': 0, })
ukb_25020_2 = TensorMap('25020_Volume-of-hippocampus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3904.775315568023, 'std': 488.8204795486511}, annotation_units=1, channel_map={'25020_Volume-of-hippocampus-right_2_0': 0, })
ukb_25021_2 = TensorMap('25021_Volume-of-amygdala-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1271.9443665264143, 'std': 246.77470893176226}, annotation_units=1, channel_map={'25021_Volume-of-amygdala-left_2_0': 0, })
ukb_25022_2 = TensorMap('25022_Volume-of-amygdala-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1238.120383356709, 'std': 273.5315360114857}, annotation_units=1, channel_map={'25022_Volume-of-amygdala-right_2_0': 0, })
ukb_25023_2 = TensorMap('25023_Volume-of-accumbens-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 502.2672276764844, 'std': 119.66718550102077}, annotation_units=1, channel_map={'25023_Volume-of-accumbens-left_2_0': 0, })
ukb_25024_2 = TensorMap('25024_Volume-of-accumbens-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 395.50000000000006, 'std': 110.56412910404944}, annotation_units=1, channel_map={'25024_Volume-of-accumbens-right_2_0': 0, })
ukb_25025_2 = TensorMap('25025_Volume-of-brain-stem-4th-ventricle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 22847.279943899015, 'std': 2764.2637993861836}, annotation_units=1, channel_map={'25025_Volume-of-brain-stem-4th-ventricle_2_0': 0, })
ukb_25026_2 = TensorMap('25026_Median-T2star-in-thalamus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.18778422955656, 'std': 3.6527169842902225}, annotation_units=1, channel_map={'25026_Median-T2star-in-thalamus-left_2_0': 0, })
ukb_25027_2 = TensorMap('25027_Median-T2star-in-thalamus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.010170962164366, 'std': 3.486705845794037}, annotation_units=1, channel_map={'25027_Median-T2star-in-thalamus-right_2_0': 0, })
ukb_25028_2 = TensorMap('25028_Median-T2star-in-caudate-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.1243198026851, 'std': 4.191032941658035}, annotation_units=1, channel_map={'25028_Median-T2star-in-caudate-left_2_0': 0, })
ukb_25029_2 = TensorMap('25029_Median-T2star-in-caudate-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.39002971928398, 'std': 4.110437900475554}, annotation_units=1, channel_map={'25029_Median-T2star-in-caudate-right_2_0': 0, })
ukb_25030_2 = TensorMap('25030_Median-T2star-in-putamen-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 37.11477493388933, 'std': 4.718739348451725}, annotation_units=1, channel_map={'25030_Median-T2star-in-putamen-left_2_0': 0, })
ukb_25031_2 = TensorMap('25031_Median-T2star-in-putamen-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 36.82571717351503, 'std': 4.459799742053833}, annotation_units=1, channel_map={'25031_Median-T2star-in-putamen-right_2_0': 0, })
ukb_25032_2 = TensorMap('25032_Median-T2star-in-pallidum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.30007659784377, 'std': 3.62604260351589}, annotation_units=1, channel_map={'25032_Median-T2star-in-pallidum-left_2_0': 0, })
ukb_25033_2 = TensorMap('25033_Median-T2star-in-pallidum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.42881057262002, 'std': 3.5248322910314296}, annotation_units=1, channel_map={'25033_Median-T2star-in-pallidum-right_2_0': 0, })
ukb_25034_2 = TensorMap('25034_Median-T2star-in-hippocampus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.34125369711147, 'std': 4.035087300743622}, annotation_units=1, channel_map={'25034_Median-T2star-in-hippocampus-left_2_0': 0, })
ukb_25035_2 = TensorMap('25035_Median-T2star-in-hippocampus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.431263440805516, 'std': 3.9144069954038843}, annotation_units=1, channel_map={'25035_Median-T2star-in-hippocampus-right_2_0': 0, })
ukb_25036_2 = TensorMap('25036_Median-T2star-in-amygdala-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.00824789971522, 'std': 5.009710829280863}, annotation_units=1, channel_map={'25036_Median-T2star-in-amygdala-left_2_0': 0, })
ukb_25037_2 = TensorMap('25037_Median-T2star-in-amygdala-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.81853516069976, 'std': 5.182339078221385}, annotation_units=1, channel_map={'25037_Median-T2star-in-amygdala-right_2_0': 0, })
ukb_25038_2 = TensorMap('25038_Median-T2star-in-accumbens-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 40.90814786411716, 'std': 6.158312201701319}, annotation_units=1, channel_map={'25038_Median-T2star-in-accumbens-left_2_0': 0, })
ukb_25039_2 = TensorMap('25039_Median-T2star-in-accumbens-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.27336789564688, 'std': 6.285723657671166}, annotation_units=1, channel_map={'25039_Median-T2star-in-accumbens-right_2_0': 0, })
ukb_25040_2 = TensorMap('25040_Median-BOLD-effect-in-groupdefined-mask-for-shapes-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6223457643465679, 'std': 0.3346037845359786}, annotation_units=1, channel_map={'25040_Median-BOLD-effect-in-groupdefined-mask-for-shapes-activation_2_0': 0, })
ukb_25042_2 = TensorMap('25042_Median-zstatistic-in-groupdefined-mask-for-shapes-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.6571508815838825, 'std': 1.4236215829729055}, annotation_units=1, channel_map={'25042_Median-zstatistic-in-groupdefined-mask-for-shapes-activation_2_0': 0, })
ukb_25044_2 = TensorMap('25044_Median-BOLD-effect-in-groupdefined-mask-for-faces-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6920879815678078, 'std': 0.3636483942228372}, annotation_units=1, channel_map={'25044_Median-BOLD-effect-in-groupdefined-mask-for-faces-activation_2_0': 0, })
ukb_25046_2 = TensorMap('25046_Median-zstatistic-in-groupdefined-mask-for-faces-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.843345633660184, 'std': 1.361236980178268}, annotation_units=1, channel_map={'25046_Median-zstatistic-in-groupdefined-mask-for-faces-activation_2_0': 0, })
ukb_25048_2 = TensorMap('25048_Median-BOLD-effect-in-groupdefined-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5204846487702943, 'std': 0.30005678322967605}, annotation_units=1, channel_map={'25048_Median-BOLD-effect-in-groupdefined-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25050_2 = TensorMap('25050_Median-zstatistic-in-groupdefined-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5707538469163587, 'std': 1.4061036113540806}, annotation_units=1, channel_map={'25050_Median-zstatistic-in-groupdefined-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25052_2 = TensorMap('25052_Median-BOLD-effect-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1396969404704496, 'std': 0.11676160645155512}, annotation_units=1, channel_map={'25052_Median-BOLD-effect-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25054_2 = TensorMap('25054_Median-zstatistic-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.209816693993463, 'std': 0.90238179906095}, annotation_units=1, channel_map={'25054_Median-zstatistic-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25056_2 = TensorMap('25056_Mean-FA-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5462210045888114, 'std': 0.023414387373035916}, annotation_units=1, channel_map={'25056_Mean-FA-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25057_2 = TensorMap('25057_Mean-FA-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4107096006702758, 'std': 0.03366132229724678}, annotation_units=1, channel_map={'25057_Mean-FA-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25058_2 = TensorMap('25058_Mean-FA-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7205092672338231, 'std': 0.03512334129480641}, annotation_units=1, channel_map={'25058_Mean-FA-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25059_2 = TensorMap('25059_Mean-FA-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7140345976282546, 'std': 0.029750195674061303}, annotation_units=1, channel_map={'25059_Mean-FA-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25060_2 = TensorMap('25060_Mean-FA-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7908209603506058, 'std': 0.02164254376057029}, annotation_units=1, channel_map={'25060_Mean-FA-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25061_2 = TensorMap('25061_Mean-FA-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4422183114204692, 'std': 0.08609871425065362}, annotation_units=1, channel_map={'25061_Mean-FA-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25062_2 = TensorMap('25062_Mean-FA-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5217352404743489, 'std': 0.037161194590578026}, annotation_units=1, channel_map={'25062_Mean-FA-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25063_2 = TensorMap('25063_Mean-FA-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.541600495282289, 'std': 0.033975007711245404}, annotation_units=1, channel_map={'25063_Mean-FA-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25064_2 = TensorMap('25064_Mean-FA-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5930021679814386, 'std': 0.028142871106043982}, annotation_units=1, channel_map={'25064_Mean-FA-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25065_2 = TensorMap('25065_Mean-FA-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5859665797370455, 'std': 0.028621206263301606}, annotation_units=1, channel_map={'25065_Mean-FA-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25066_2 = TensorMap('25066_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5532455374581078, 'std': 0.030313089151443273}, annotation_units=1, channel_map={'25066_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25067_2 = TensorMap('25067_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5480877288992007, 'std': 0.031013220828694323}, annotation_units=1, channel_map={'25067_Mean-FA-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25068_2 = TensorMap('25068_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6974787304459911, 'std': 0.025832524986613564}, annotation_units=1, channel_map={'25068_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25069_2 = TensorMap('25069_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7076919827790669, 'std': 0.026116735753257885}, annotation_units=1, channel_map={'25069_Mean-FA-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25070_2 = TensorMap('25070_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7119579190513018, 'std': 0.0247901660930988}, annotation_units=1, channel_map={'25070_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25071_2 = TensorMap('25071_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7159478244392883, 'std': 0.02478919390963223}, annotation_units=1, channel_map={'25071_Mean-FA-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25072_2 = TensorMap('25072_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6007321301366333, 'std': 0.025316255882892685}, annotation_units=1, channel_map={'25072_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25073_2 = TensorMap('25073_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5903566466615109, 'std': 0.024569661197694066}, annotation_units=1, channel_map={'25073_Mean-FA-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25074_2 = TensorMap('25074_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6839250670275845, 'std': 0.024139520600770948}, annotation_units=1, channel_map={'25074_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25075_2 = TensorMap('25075_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6867448731116267, 'std': 0.024451232685528238}, annotation_units=1, channel_map={'25075_Mean-FA-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25076_2 = TensorMap('25076_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5983789422531579, 'std': 0.028279789710634707}, annotation_units=1, channel_map={'25076_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25077_2 = TensorMap('25077_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6152762362464552, 'std': 0.02727136973652318}, annotation_units=1, channel_map={'25077_Mean-FA-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25078_2 = TensorMap('25078_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46265153730342873, 'std': 0.03126308583410255}, annotation_units=1, channel_map={'25078_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25079_2 = TensorMap('25079_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4535338121680845, 'std': 0.03088016935894359}, annotation_units=1, channel_map={'25079_Mean-FA-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25080_2 = TensorMap('25080_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.48453128136117557, 'std': 0.024740610989841077}, annotation_units=1, channel_map={'25080_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25081_2 = TensorMap('25081_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4901271390564577, 'std': 0.02486343912194385}, annotation_units=1, channel_map={'25081_Mean-FA-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25082_2 = TensorMap('25082_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.49131109878834756, 'std': 0.026982000055673562}, annotation_units=1, channel_map={'25082_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25083_2 = TensorMap('25083_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4804470461459139, 'std': 0.025301595814430686}, annotation_units=1, channel_map={'25083_Mean-FA-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25084_2 = TensorMap('25084_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5942921648878575, 'std': 0.03349306707612423}, annotation_units=1, channel_map={'25084_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25085_2 = TensorMap('25085_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6018214504253672, 'std': 0.034167420657018735}, annotation_units=1, channel_map={'25085_Mean-FA-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25086_2 = TensorMap('25086_Mean-FA-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5607059307553494, 'std': 0.030678620557348123}, annotation_units=1, channel_map={'25086_Mean-FA-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25087_2 = TensorMap('25087_Mean-FA-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5661505457592165, 'std': 0.03206984544729702}, annotation_units=1, channel_map={'25087_Mean-FA-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25088_2 = TensorMap('25088_Mean-FA-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.460532789378706, 'std': 0.022919229407068416}, annotation_units=1, channel_map={'25088_Mean-FA-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25089_2 = TensorMap('25089_Mean-FA-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46893503542150045, 'std': 0.02499436489697856}, annotation_units=1, channel_map={'25089_Mean-FA-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25090_2 = TensorMap('25090_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5815898662026299, 'std': 0.03353848902349903}, annotation_units=1, channel_map={'25090_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25091_2 = TensorMap('25091_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.623389054395463, 'std': 0.03256709800587815}, annotation_units=1, channel_map={'25091_Mean-FA-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25092_2 = TensorMap('25092_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46092420247486465, 'std': 0.035866967969688426}, annotation_units=1, channel_map={'25092_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25093_2 = TensorMap('25093_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4611374575921628, 'std': 0.03543226379312232}, annotation_units=1, channel_map={'25093_Mean-FA-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25094_2 = TensorMap('25094_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5235408987367878, 'std': 0.0361049683739947}, annotation_units=1, channel_map={'25094_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25095_2 = TensorMap('25095_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5229278309358084, 'std': 0.03790623462985725}, annotation_units=1, channel_map={'25095_Mean-FA-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25096_2 = TensorMap('25096_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5218666096416601, 'std': 0.025989901292689698}, annotation_units=1, channel_map={'25096_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25097_2 = TensorMap('25097_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5284978607373035, 'std': 0.026424628198946438}, annotation_units=1, channel_map={'25097_Mean-FA-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25098_2 = TensorMap('25098_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46075076829079664, 'std': 0.04265160278241941}, annotation_units=1, channel_map={'25098_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25099_2 = TensorMap('25099_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4624648329981955, 'std': 0.04399535844637362}, annotation_units=1, channel_map={'25099_Mean-FA-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25100_2 = TensorMap('25100_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.527276962103635, 'std': 0.03766476428388945}, annotation_units=1, channel_map={'25100_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25101_2 = TensorMap('25101_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5130611882959525, 'std': 0.038583793316197254}, annotation_units=1, channel_map={'25101_Mean-FA-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25102_2 = TensorMap('25102_Mean-FA-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5515943303944314, 'std': 0.06640781642357084}, annotation_units=1, channel_map={'25102_Mean-FA-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25103_2 = TensorMap('25103_Mean-FA-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5828673031193606, 'std': 0.06447256268633707}, annotation_units=1, channel_map={'25103_Mean-FA-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25104_2 = TensorMap('25104_Mean-MD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007105226604795049, 'std': 3.0446812540399537e-05}, annotation_units=1, channel_map={'25104_Mean-MD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25105_2 = TensorMap('25105_Mean-MD-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007946389275586492, 'std': 7.911087795044624e-05}, annotation_units=1, channel_map={'25105_Mean-MD-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25106_2 = TensorMap('25106_Mean-MD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007734786285125033, 'std': 4.875284762099488e-05}, annotation_units=1, channel_map={'25106_Mean-MD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25107_2 = TensorMap('25107_Mean-MD-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007914060324825983, 'std': 4.156858273937262e-05}, annotation_units=1, channel_map={'25107_Mean-MD-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25108_2 = TensorMap('25108_Mean-MD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000735309667440062, 'std': 3.2165613560896705e-05}, annotation_units=1, channel_map={'25108_Mean-MD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25109_2 = TensorMap('25109_Mean-MD-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0015548519205980922, 'std': 0.00030554251784522087}, annotation_units=1, channel_map={'25109_Mean-MD-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25110_2 = TensorMap('25110_Mean-MD-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007885098736787831, 'std': 7.390583703064105e-05}, annotation_units=1, channel_map={'25110_Mean-MD-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25111_2 = TensorMap('25111_Mean-MD-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007858498582108793, 'std': 6.70720926112742e-05}, annotation_units=1, channel_map={'25111_Mean-MD-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25112_2 = TensorMap('25112_Mean-MD-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007597696313482855, 'std': 4.546477904364971e-05}, annotation_units=1, channel_map={'25112_Mean-MD-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25113_2 = TensorMap('25113_Mean-MD-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007649336426914155, 'std': 4.793629843607117e-05}, annotation_units=1, channel_map={'25113_Mean-MD-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25114_2 = TensorMap('25114_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007275122969837588, 'std': 2.812514738624599e-05}, annotation_units=1, channel_map={'25114_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25115_2 = TensorMap('25115_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007385215777262179, 'std': 3.0059464389564907e-05}, annotation_units=1, channel_map={'25115_Mean-MD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25116_2 = TensorMap('25116_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008081415828821859, 'std': 3.985991584632281e-05}, annotation_units=1, channel_map={'25116_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25117_2 = TensorMap('25117_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008007053879865946, 'std': 3.946732735024455e-05}, annotation_units=1, channel_map={'25117_Mean-MD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25118_2 = TensorMap('25118_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007232379994844032, 'std': 3.3412759052250644e-05}, annotation_units=1, channel_map={'25118_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25119_2 = TensorMap('25119_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000711267800979634, 'std': 3.130859735958997e-05}, annotation_units=1, channel_map={'25119_Mean-MD-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25120_2 = TensorMap('25120_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000722490899716422, 'std': 3.47302553025666e-05}, annotation_units=1, channel_map={'25120_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25121_2 = TensorMap('25121_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007192620778551176, 'std': 3.445561166070029e-05}, annotation_units=1, channel_map={'25121_Mean-MD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25122_2 = TensorMap('25122_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007089748904356791, 'std': 2.2638068549593372e-05}, annotation_units=1, channel_map={'25122_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25123_2 = TensorMap('25123_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000700556380510441, 'std': 2.2384820403244303e-05}, annotation_units=1, channel_map={'25123_Mean-MD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25124_2 = TensorMap('25124_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007714178912090746, 'std': 3.562832061220104e-05}, annotation_units=1, channel_map={'25124_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25125_2 = TensorMap('25125_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007655411704047436, 'std': 3.3354140740221465e-05}, annotation_units=1, channel_map={'25125_Mean-MD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25126_2 = TensorMap('25126_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007626026811033773, 'std': 4.1482409016855995e-05}, annotation_units=1, channel_map={'25126_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25127_2 = TensorMap('25127_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007719716421758184, 'std': 4.3289818362851346e-05}, annotation_units=1, channel_map={'25127_Mean-MD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25128_2 = TensorMap('25128_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000727426707914411, 'std': 3.454423161069724e-05}, annotation_units=1, channel_map={'25128_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25129_2 = TensorMap('25129_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007318056715648363, 'std': 3.516773884436978e-05}, annotation_units=1, channel_map={'25129_Mean-MD-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25130_2 = TensorMap('25130_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008260991492652744, 'std': 5.643201537500201e-05}, annotation_units=1, channel_map={'25130_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25131_2 = TensorMap('25131_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000807911265790152, 'std': 5.239210212506524e-05}, annotation_units=1, channel_map={'25131_Mean-MD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25132_2 = TensorMap('25132_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008426874452178395, 'std': 4.7296604739538525e-05}, annotation_units=1, channel_map={'25132_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25133_2 = TensorMap('25133_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008471511729827274, 'std': 5.1549200140296985e-05}, annotation_units=1, channel_map={'25133_Mean-MD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25134_2 = TensorMap('25134_Mean-MD-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008203412735241039, 'std': 4.236452308060806e-05}, annotation_units=1, channel_map={'25134_Mean-MD-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25135_2 = TensorMap('25135_Mean-MD-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000803174065480794, 'std': 4.112119757890872e-05}, annotation_units=1, channel_map={'25135_Mean-MD-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25136_2 = TensorMap('25136_Mean-MD-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007641459654550143, 'std': 3.531659414199056e-05}, annotation_units=1, channel_map={'25136_Mean-MD-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25137_2 = TensorMap('25137_Mean-MD-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007546492394947151, 'std': 3.69121287360018e-05}, annotation_units=1, channel_map={'25137_Mean-MD-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25138_2 = TensorMap('25138_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007566087651456562, 'std': 2.9537889551327075e-05}, annotation_units=1, channel_map={'25138_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25139_2 = TensorMap('25139_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007561767981438515, 'std': 2.9102221542506396e-05}, annotation_units=1, channel_map={'25139_Mean-MD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25140_2 = TensorMap('25140_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007624871358597576, 'std': 3.513501818641231e-05}, annotation_units=1, channel_map={'25140_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25141_2 = TensorMap('25141_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007605444702242846, 'std': 3.600963112551396e-05}, annotation_units=1, channel_map={'25141_Mean-MD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25142_2 = TensorMap('25142_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000869343129672596, 'std': 6.222948157776833e-05}, annotation_units=1, channel_map={'25142_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25143_2 = TensorMap('25143_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008494158288218615, 'std': 5.9929853697149005e-05}, annotation_units=1, channel_map={'25143_Mean-MD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25144_2 = TensorMap('25144_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007407705594225315, 'std': 3.313718395042791e-05}, annotation_units=1, channel_map={'25144_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25145_2 = TensorMap('25145_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007239692704305232, 'std': 3.1791263300678984e-05}, annotation_units=1, channel_map={'25145_Mean-MD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25146_2 = TensorMap('25146_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007441228151585462, 'std': 6.709237506996024e-05}, annotation_units=1, channel_map={'25146_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25147_2 = TensorMap('25147_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007530260891982468, 'std': 8.639459788802517e-05}, annotation_units=1, channel_map={'25147_Mean-MD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25148_2 = TensorMap('25148_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007490411446249034, 'std': 3.394371191183398e-05}, annotation_units=1, channel_map={'25148_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25149_2 = TensorMap('25149_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007309253415828821, 'std': 4.233243748245232e-05}, annotation_units=1, channel_map={'25149_Mean-MD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25150_2 = TensorMap('25150_Mean-MD-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0010340231502964682, 'std': 0.00013538728086349188}, annotation_units=1, channel_map={'25150_Mean-MD-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25151_2 = TensorMap('25151_Mean-MD-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0009587816447538023, 'std': 0.0001354870577610467}, annotation_units=1, channel_map={'25151_Mean-MD-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25152_2 = TensorMap('25152_Mean-MO-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5624609455014176, 'std': 0.0518558596873961}, annotation_units=1, channel_map={'25152_Mean-MO-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25153_2 = TensorMap('25153_Mean-MO-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.26455716988914674, 'std': 0.17739750974590057}, annotation_units=1, channel_map={'25153_Mean-MO-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25154_2 = TensorMap('25154_Mean-MO-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9037268359886571, 'std': 0.03157272780141393}, annotation_units=1, channel_map={'25154_Mean-MO-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25155_2 = TensorMap('25155_Mean-MO-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8988152441866464, 'std': 0.03767176142394583}, annotation_units=1, channel_map={'25155_Mean-MO-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25156_2 = TensorMap('25156_Mean-MO-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9065881806651197, 'std': 0.030461016238928874}, annotation_units=1, channel_map={'25156_Mean-MO-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25157_2 = TensorMap('25157_Mean-MO-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.810637246300593, 'std': 0.11108728671188703}, annotation_units=1, channel_map={'25157_Mean-MO-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25158_2 = TensorMap('25158_Mean-MO-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4712001390048981, 'std': 0.1397470707362097}, annotation_units=1, channel_map={'25158_Mean-MO-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25159_2 = TensorMap('25159_Mean-MO-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5461051461201338, 'std': 0.11774309688921121}, annotation_units=1, channel_map={'25159_Mean-MO-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25160_2 = TensorMap('25160_Mean-MO-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7729073236916731, 'std': 0.05737609955981621}, annotation_units=1, channel_map={'25160_Mean-MO-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25161_2 = TensorMap('25161_Mean-MO-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7835322609951019, 'std': 0.05350977867327742}, annotation_units=1, channel_map={'25161_Mean-MO-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25162_2 = TensorMap('25162_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6096148617169377, 'std': 0.0907616449083623}, annotation_units=1, channel_map={'25162_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25163_2 = TensorMap('25163_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6114382490332559, 'std': 0.08878721349344425}, annotation_units=1, channel_map={'25163_Mean-MO-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25164_2 = TensorMap('25164_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8507901537509669, 'std': 0.03105160646903311}, annotation_units=1, channel_map={'25164_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25165_2 = TensorMap('25165_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.862198972106213, 'std': 0.030365755481444035}, annotation_units=1, channel_map={'25165_Mean-MO-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25166_2 = TensorMap('25166_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8481737048208301, 'std': 0.03175176206148621}, annotation_units=1, channel_map={'25166_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25167_2 = TensorMap('25167_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8265997735498841, 'std': 0.035077495905279706}, annotation_units=1, channel_map={'25167_Mean-MO-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25168_2 = TensorMap('25168_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8087259251353442, 'std': 0.053400402597919124}, annotation_units=1, channel_map={'25168_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25169_2 = TensorMap('25169_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7885032208816704, 'std': 0.05182096132987702}, annotation_units=1, channel_map={'25169_Mean-MO-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25170_2 = TensorMap('25170_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8389646259860789, 'std': 0.0453472933920434}, annotation_units=1, channel_map={'25170_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25171_2 = TensorMap('25171_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8100403856148489, 'std': 0.05325055584847589}, annotation_units=1, channel_map={'25171_Mean-MO-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25172_2 = TensorMap('25172_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6717684894044857, 'std': 0.07215962635275072}, annotation_units=1, channel_map={'25172_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25173_2 = TensorMap('25173_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.744640193864398, 'std': 0.057577607001359894}, annotation_units=1, channel_map={'25173_Mean-MO-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25174_2 = TensorMap('25174_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6492947369425109, 'std': 0.07433418554530281}, annotation_units=1, channel_map={'25174_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25175_2 = TensorMap('25175_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5790100476927046, 'std': 0.08379200958858561}, annotation_units=1, channel_map={'25175_Mean-MO-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25176_2 = TensorMap('25176_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5835545456045371, 'std': 0.09347445120519747}, annotation_units=1, channel_map={'25176_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25177_2 = TensorMap('25177_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5480559237947923, 'std': 0.10342089078649416}, annotation_units=1, channel_map={'25177_Mean-MO-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25178_2 = TensorMap('25178_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5255624961072441, 'std': 0.10324572731776388}, annotation_units=1, channel_map={'25178_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25179_2 = TensorMap('25179_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5347958610466615, 'std': 0.08785411296557642}, annotation_units=1, channel_map={'25179_Mean-MO-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25180_2 = TensorMap('25180_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7624144427429752, 'std': 0.0687811206599589}, annotation_units=1, channel_map={'25180_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25181_2 = TensorMap('25181_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7734773303944313, 'std': 0.06546913829128283}, annotation_units=1, channel_map={'25181_Mean-MO-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25182_2 = TensorMap('25182_Mean-MO-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6199068692962104, 'std': 0.07910786836897021}, annotation_units=1, channel_map={'25182_Mean-MO-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25183_2 = TensorMap('25183_Mean-MO-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.604672369837587, 'std': 0.07496251018751224}, annotation_units=1, channel_map={'25183_Mean-MO-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25184_2 = TensorMap('25184_Mean-MO-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.45627444155710223, 'std': 0.06767834214499742}, annotation_units=1, channel_map={'25184_Mean-MO-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25185_2 = TensorMap('25185_Mean-MO-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.48905961412735255, 'std': 0.06666742760236313}, annotation_units=1, channel_map={'25185_Mean-MO-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25186_2 = TensorMap('25186_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6885621862851251, 'std': 0.06798279547927595}, annotation_units=1, channel_map={'25186_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25187_2 = TensorMap('25187_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7399877462748131, 'std': 0.05379022578661981}, annotation_units=1, channel_map={'25187_Mean-MO-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25188_2 = TensorMap('25188_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6276432808971383, 'std': 0.08329198845035772}, annotation_units=1, channel_map={'25188_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25189_2 = TensorMap('25189_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6344650253673626, 'std': 0.09492227121386286}, annotation_units=1, channel_map={'25189_Mean-MO-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25190_2 = TensorMap('25190_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5838272329981953, 'std': 0.09500601216608659}, annotation_units=1, channel_map={'25190_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25191_2 = TensorMap('25191_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5455303063676205, 'std': 0.10426936523631934}, annotation_units=1, channel_map={'25191_Mean-MO-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25192_2 = TensorMap('25192_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5106826826501678, 'std': 0.08304391805826579}, annotation_units=1, channel_map={'25192_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25193_2 = TensorMap('25193_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5025446954369683, 'std': 0.08946618466003331}, annotation_units=1, channel_map={'25193_Mean-MO-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25194_2 = TensorMap('25194_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.776969126785254, 'std': 0.11849543068980917}, annotation_units=1, channel_map={'25194_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25195_2 = TensorMap('25195_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.731801420314514, 'std': 0.14739603128691633}, annotation_units=1, channel_map={'25195_Mean-MO-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25196_2 = TensorMap('25196_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7467207389017789, 'std': 0.08431493049935739}, annotation_units=1, channel_map={'25196_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25197_2 = TensorMap('25197_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7065362514050015, 'std': 0.0931263556551804}, annotation_units=1, channel_map={'25197_Mean-MO-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25198_2 = TensorMap('25198_Mean-MO-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8590522768239238, 'std': 0.08567687924395909}, annotation_units=1, channel_map={'25198_Mean-MO-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25199_2 = TensorMap('25199_Mean-MO-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8540758697086878, 'std': 0.10284350338994121}, annotation_units=1, channel_map={'25199_Mean-MO-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25200_2 = TensorMap('25200_Mean-L1-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011949464294921375, 'std': 4.68205217076586e-05}, annotation_units=1, channel_map={'25200_Mean-L1-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25201_2 = TensorMap('25201_Mean-L1-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011472379479247224, 'std': 9.834031951503224e-05}, annotation_units=1, channel_map={'25201_Mean-L1-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25202_2 = TensorMap('25202_Mean-L1-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0015752037122969838, 'std': 5.735258829355501e-05}, annotation_units=1, channel_map={'25202_Mean-L1-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25203_2 = TensorMap('25203_Mean-L1-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0016076131992781646, 'std': 5.201709219899409e-05}, annotation_units=1, channel_map={'25203_Mean-L1-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25204_2 = TensorMap('25204_Mean-L1-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0016212961072441354, 'std': 5.789653966117855e-05}, annotation_units=1, channel_map={'25204_Mean-L1-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25205_2 = TensorMap('25205_Mean-L1-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.002305857643722609, 'std': 0.00024265187459554334}, annotation_units=1, channel_map={'25205_Mean-L1-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25206_2 = TensorMap('25206_Mean-L1-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00127311090487239, 'std': 0.00010127315414369404}, annotation_units=1, channel_map={'25206_Mean-L1-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25207_2 = TensorMap('25207_Mean-L1-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012995518948182522, 'std': 9.805640783315845e-05}, annotation_units=1, channel_map={'25207_Mean-L1-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25208_2 = TensorMap('25208_Mean-L1-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013502275844289766, 'std': 6.997911172098018e-05}, annotation_units=1, channel_map={'25208_Mean-L1-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25209_2 = TensorMap('25209_Mean-L1-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013507628770301622, 'std': 7.372000466588152e-05}, annotation_units=1, channel_map={'25209_Mean-L1-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25210_2 = TensorMap('25210_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001231962619231761, 'std': 4.997498079770312e-05}, annotation_units=1, channel_map={'25210_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25211_2 = TensorMap('25211_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012439981954111888, 'std': 4.723899537640904e-05}, annotation_units=1, channel_map={'25211_Mean-L1-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25212_2 = TensorMap('25212_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0015921211652487755, 'std': 7.225676849301225e-05}, annotation_units=1, channel_map={'25212_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25213_2 = TensorMap('25213_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0015962510956432072, 'std': 6.893505363220449e-05}, annotation_units=1, channel_map={'25213_Mean-L1-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25214_2 = TensorMap('25214_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0014515406548079397, 'std': 6.0208539271269076e-05}, annotation_units=1, channel_map={'25214_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25215_2 = TensorMap('25215_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0014331954111884509, 'std': 5.67062256658523e-05}, annotation_units=1, channel_map={'25215_Mean-L1-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25216_2 = TensorMap('25216_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012940755349316832, 'std': 4.957952953028207e-05}, annotation_units=1, channel_map={'25216_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25217_2 = TensorMap('25217_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012753117813869557, 'std': 4.774987340643609e-05}, annotation_units=1, channel_map={'25217_Mean-L1-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25218_2 = TensorMap('25218_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013777211136890952, 'std': 4.764473804544218e-05}, annotation_units=1, channel_map={'25218_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25219_2 = TensorMap('25219_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001363507398814127, 'std': 4.842416916415225e-05}, annotation_units=1, channel_map={'25219_Mean-L1-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25220_2 = TensorMap('25220_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013619320959010054, 'std': 5.190951741106763e-05}, annotation_units=1, channel_map={'25220_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25221_2 = TensorMap('25221_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013808601701469452, 'std': 4.8863110071984835e-05}, annotation_units=1, channel_map={'25221_Mean-L1-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25222_2 = TensorMap('25222_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011869726733694254, 'std': 4.679665766982745e-05}, annotation_units=1, channel_map={'25222_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25223_2 = TensorMap('25223_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011871431812322765, 'std': 4.6720858063669224e-05}, annotation_units=1, channel_map={'25223_Mean-L1-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25224_2 = TensorMap('25224_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011546259860788861, 'std': 4.8650742607545684e-05}, annotation_units=1, channel_map={'25224_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25225_2 = TensorMap('25225_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001164154421242588, 'std': 5.0345028235367264e-05}, annotation_units=1, channel_map={'25225_Mean-L1-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25226_2 = TensorMap('25226_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013127521010569735, 'std': 7.187604202539507e-05}, annotation_units=1, channel_map={'25226_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25227_2 = TensorMap('25227_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001272265893271462, 'std': 6.474934249258906e-05}, annotation_units=1, channel_map={'25227_Mean-L1-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25228_2 = TensorMap('25228_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0014970761020881667, 'std': 6.034386343148041e-05}, annotation_units=1, channel_map={'25228_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25229_2 = TensorMap('25229_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0015157666924465072, 'std': 6.325708306971018e-05}, annotation_units=1, channel_map={'25229_Mean-L1-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25230_2 = TensorMap('25230_Mean-L1-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001397255117298273, 'std': 5.7160114848413015e-05}, annotation_units=1, channel_map={'25230_Mean-L1-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25231_2 = TensorMap('25231_Mean-L1-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001376287445217839, 'std': 5.654250058248917e-05}, annotation_units=1, channel_map={'25231_Mean-L1-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25232_2 = TensorMap('25232_Mean-L1-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011810050528486722, 'std': 4.041147766809327e-05}, annotation_units=1, channel_map={'25232_Mean-L1-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25233_2 = TensorMap('25233_Mean-L1-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001176959680329982, 'std': 4.208714721515857e-05}, annotation_units=1, channel_map={'25233_Mean-L1-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25234_2 = TensorMap('25234_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001331061562258314, 'std': 5.83840999721569e-05}, annotation_units=1, channel_map={'25234_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25235_2 = TensorMap('25235_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013936686774941997, 'std': 5.856470730581609e-05}, annotation_units=1, channel_map={'25235_Mean-L1-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25236_2 = TensorMap('25236_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011878629543696831, 'std': 5.583680483787712e-05}, annotation_units=1, channel_map={'25236_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25237_2 = TensorMap('25237_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011838248002062383, 'std': 5.5773823899347336e-05}, annotation_units=1, channel_map={'25237_Mean-L1-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25238_2 = TensorMap('25238_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0014156426914153132, 'std': 7.253758572820502e-05}, annotation_units=1, channel_map={'25238_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25239_2 = TensorMap('25239_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013819418406805877, 'std': 7.164988278150399e-05}, annotation_units=1, channel_map={'25239_Mean-L1-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25240_2 = TensorMap('25240_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012123745810775975, 'std': 4.201492321470444e-05}, annotation_units=1, channel_map={'25240_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25241_2 = TensorMap('25241_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001191505594225316, 'std': 4.0093502235049593e-05}, annotation_units=1, channel_map={'25241_Mean-L1-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25242_2 = TensorMap('25242_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011595504511472028, 'std': 7.72145851615899e-05}, annotation_units=1, channel_map={'25242_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25243_2 = TensorMap('25243_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011670591905130186, 'std': 8.924042215695256e-05}, annotation_units=1, channel_map={'25243_Mean-L1-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25244_2 = TensorMap('25244_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00124768904356793, 'std': 5.634525398634305e-05}, annotation_units=1, channel_map={'25244_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25245_2 = TensorMap('25245_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011997113689095128, 'std': 6.568659524313329e-05}, annotation_units=1, channel_map={'25245_Mean-L1-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25246_2 = TensorMap('25246_Mean-L1-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0017326550141789116, 'std': 0.0001502087253404274}, annotation_units=1, channel_map={'25246_Mean-L1-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25247_2 = TensorMap('25247_Mean-L1-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0016578075792730083, 'std': 0.00016690380820845655}, annotation_units=1, channel_map={'25247_Mean-L1-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25248_2 = TensorMap('25248_Mean-L2-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005656051044083527, 'std': 3.3234651326773197e-05}, annotation_units=1, channel_map={'25248_Mean-L2-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25249_2 = TensorMap('25249_Mean-L2-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007439736530033512, 'std': 8.507678578231215e-05}, annotation_units=1, channel_map={'25249_Mean-L2-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25250_2 = TensorMap('25250_Mean-L2-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004371425109564321, 'std': 5.41974741189315e-05}, annotation_units=1, channel_map={'25250_Mean-L2-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25251_2 = TensorMap('25251_Mean-L2-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004431817478731631, 'std': 4.8662994715642005e-05}, annotation_units=1, channel_map={'25251_Mean-L2-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25252_2 = TensorMap('25252_Mean-L2-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003683125547821605, 'std': 3.44244560273766e-05}, annotation_units=1, channel_map={'25252_Mean-L2-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25253_2 = TensorMap('25253_Mean-L2-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012694705852023716, 'std': 0.00035240396610251683}, annotation_units=1, channel_map={'25253_Mean-L2-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25254_2 = TensorMap('25254_Mean-L2-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006732603248259858, 'std': 7.607928954327307e-05}, annotation_units=1, channel_map={'25254_Mean-L2-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25255_2 = TensorMap('25255_Mean-L2-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006474544985821086, 'std': 6.589426117282102e-05}, annotation_units=1, channel_map={'25255_Mean-L2-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25256_2 = TensorMap('25256_Mean-L2-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005497278164475381, 'std': 4.377686781943826e-05}, annotation_units=1, channel_map={'25256_Mean-L2-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25257_2 = TensorMap('25257_Mean-L2-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005557510698633671, 'std': 4.55408664263806e-05}, annotation_units=1, channel_map={'25257_Mean-L2-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25258_2 = TensorMap('25258_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005823225573601446, 'std': 3.835776376029597e-05}, annotation_units=1, channel_map={'25258_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25259_2 = TensorMap('25259_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005926888888888888, 'std': 4.206357781390601e-05}, annotation_units=1, channel_map={'25259_Mean-L2-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25260_2 = TensorMap('25260_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004938977571539058, 'std': 4.070788256873193e-05}, annotation_units=1, channel_map={'25260_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25261_2 = TensorMap('25261_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004771244650683165, 'std': 4.179962190407171e-05}, annotation_units=1, channel_map={'25261_Mean-L2-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25262_2 = TensorMap('25262_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00044945877803557614, 'std': 3.4674310613070164e-05}, annotation_units=1, channel_map={'25262_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25263_2 = TensorMap('25263_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004486540861046661, 'std': 3.436206309286074e-05}, annotation_units=1, channel_map={'25263_Mean-L2-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25264_2 = TensorMap('25264_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005260758442897654, 'std': 3.6617602187870775e-05}, annotation_units=1, channel_map={'25264_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25265_2 = TensorMap('25265_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005300704305233307, 'std': 3.5810106255197204e-05}, annotation_units=1, channel_map={'25265_Mean-L2-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25266_2 = TensorMap('25266_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00047131698891466874, 'std': 3.0276658363667534e-05}, annotation_units=1, channel_map={'25266_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25267_2 = TensorMap('25267_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00047527754575921627, 'std': 3.0984591482074304e-05}, annotation_units=1, channel_map={'25267_Mean-L2-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25268_2 = TensorMap('25268_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005979336426914154, 'std': 4.204821478512074e-05}, annotation_units=1, channel_map={'25268_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25269_2 = TensorMap('25269_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005655100283578241, 'std': 3.915113306072191e-05}, annotation_units=1, channel_map={'25269_Mean-L2-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25270_2 = TensorMap('25270_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006361532869296212, 'std': 4.6128657146117595e-05}, annotation_units=1, channel_map={'25270_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25271_2 = TensorMap('25271_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006594444960041247, 'std': 4.8009594470076633e-05}, annotation_units=1, channel_map={'25271_Mean-L2-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25272_2 = TensorMap('25272_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006133441093065222, 'std': 3.5410807778741304e-05}, annotation_units=1, channel_map={'25272_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25273_2 = TensorMap('25273_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006229146687290537, 'std': 3.586405543358517e-05}, annotation_units=1, channel_map={'25273_Mean-L2-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25274_2 = TensorMap('25274_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007024981696313483, 'std': 5.4055324157431295e-05}, annotation_units=1, channel_map={'25274_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25275_2 = TensorMap('25275_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006893936581593194, 'std': 4.979071877348096e-05}, annotation_units=1, channel_map={'25275_Mean-L2-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25276_2 = TensorMap('25276_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006113236401134313, 'std': 5.336951490043589e-05}, annotation_units=1, channel_map={'25276_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25277_2 = TensorMap('25277_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006081255478216033, 'std': 5.717465416820743e-05}, annotation_units=1, channel_map={'25277_Mean-L2-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25278_2 = TensorMap('25278_Mean-L2-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006531591131734982, 'std': 4.774601352582063e-05}, annotation_units=1, channel_map={'25278_Mean-L2-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25279_2 = TensorMap('25279_Mean-L2-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006404411446249034, 'std': 4.7247663781761175e-05}, annotation_units=1, channel_map={'25279_Mean-L2-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25280_2 = TensorMap('25280_Mean-L2-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006665009538540863, 'std': 3.958759945937163e-05}, annotation_units=1, channel_map={'25280_Mean-L2-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25281_2 = TensorMap('25281_Mean-L2-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006518356792987879, 'std': 4.1253067491937386e-05}, annotation_units=1, channel_map={'25281_Mean-L2-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25282_2 = TensorMap('25282_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000571831090487239, 'std': 3.8589078468210255e-05}, annotation_units=1, channel_map={'25282_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25283_2 = TensorMap('25283_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005359419953596284, 'std': 3.822581778783964e-05}, annotation_units=1, channel_map={'25283_Mean-L2-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25284_2 = TensorMap('25284_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006370406290281, 'std': 4.1297987474331276e-05}, annotation_units=1, channel_map={'25284_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25285_2 = TensorMap('25285_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006353741170404745, 'std': 4.3080202613395585e-05}, annotation_units=1, channel_map={'25285_Mean-L2-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25286_2 = TensorMap('25286_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007155120907450373, 'std': 7.376937223690392e-05}, annotation_units=1, channel_map={'25286_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25287_2 = TensorMap('25287_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007075828821861307, 'std': 7.275335037342405e-05}, annotation_units=1, channel_map={'25287_Mean-L2-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25288_2 = TensorMap('25288_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006240189224026811, 'std': 3.866837029765611e-05}, annotation_units=1, channel_map={'25288_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25289_2 = TensorMap('25289_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006127341067285383, 'std': 3.8329666559714486e-05}, annotation_units=1, channel_map={'25289_Mean-L2-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25290_2 = TensorMap('25290_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006036827532869296, 'std': 7.558537327579957e-05}, annotation_units=1, channel_map={'25290_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25291_2 = TensorMap('25291_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006192511987625677, 'std': 9.912160050195586e-05}, annotation_units=1, channel_map={'25291_Mean-L2-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25292_2 = TensorMap('25292_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005854515081206496, 'std': 4.50641603399859e-05}, annotation_units=1, channel_map={'25292_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25293_2 = TensorMap('25293_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005859764887857698, 'std': 5.1315408803985346e-05}, annotation_units=1, channel_map={'25293_Mean-L2-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25294_2 = TensorMap('25294_Mean-L2-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007687502449084816, 'std': 0.00014150821615390523}, annotation_units=1, channel_map={'25294_Mean-L2-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25295_2 = TensorMap('25295_Mean-L2-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006994627481309613, 'std': 0.00013276603923190625}, annotation_units=1, channel_map={'25295_Mean-L2-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25296_2 = TensorMap('25296_Mean-L3-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003710079917504511, 'std': 2.8633684554494065e-05}, annotation_units=1, channel_map={'25296_Mean-L3-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25297_2 = TensorMap('25297_Mean-L3-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004926933230213973, 'std': 7.01153904944374e-05}, annotation_units=1, channel_map={'25297_Mean-L3-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25298_2 = TensorMap('25298_Mean-L3-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00030808971384377413, 'std': 5.649273490034053e-05}, annotation_units=1, channel_map={'25298_Mean-L3-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25299_2 = TensorMap('25299_Mean-L3-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003234211394689352, 'std': 4.608266877237891e-05}, annotation_units=1, channel_map={'25299_Mean-L3-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25300_2 = TensorMap('25300_Mean-L3-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00021631956689868525, 'std': 3.3917899562217766e-05}, annotation_units=1, channel_map={'25300_Mean-L3-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25301_2 = TensorMap('25301_Mean-L3-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0010892159319412223, 'std': 0.0003300043621198059}, annotation_units=1, channel_map={'25301_Mean-L3-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25302_2 = TensorMap('25302_Mean-L3-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00041915998968806376, 'std': 6.930920824784075e-05}, annotation_units=1, channel_map={'25302_Mean-L3-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25303_2 = TensorMap('25303_Mean-L3-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004105510183036865, 'std': 6.173813164829262e-05}, annotation_units=1, channel_map={'25303_Mean-L3-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25304_2 = TensorMap('25304_Mean-L3-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00037935395720546536, 'std': 4.295108093397834e-05}, annotation_units=1, channel_map={'25304_Mean-L3-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25305_2 = TensorMap('25305_Mean-L3-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00038827847383346225, 'std': 4.424169846418709e-05}, annotation_units=1, channel_map={'25305_Mean-L3-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25306_2 = TensorMap('25306_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00036825125032224806, 'std': 3.097982645477827e-05}, annotation_units=1, channel_map={'25306_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25307_2 = TensorMap('25307_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003788801237432327, 'std': 3.335575185317172e-05}, annotation_units=1, channel_map={'25307_Mean-L3-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25308_2 = TensorMap('25308_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000338406548079402, 'std': 3.5986212575573827e-05}, annotation_units=1, channel_map={'25308_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25309_2 = TensorMap('25309_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00032874539829853047, 'std': 3.696270775116646e-05}, annotation_units=1, channel_map={'25309_Mean-L3-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25310_2 = TensorMap('25310_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0002687099252384636, 'std': 3.0864796492844134e-05}, annotation_units=1, channel_map={'25310_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25311_2 = TensorMap('25311_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0002519624645527197, 'std': 3.064043189195348e-05}, annotation_units=1, channel_map={'25311_Mean-L3-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25312_2 = TensorMap('25312_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00034732199020366076, 'std': 3.7422247034679546e-05}, annotation_units=1, channel_map={'25312_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25313_2 = TensorMap('25313_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00035239938128383603, 'std': 3.697656961724181e-05}, annotation_units=1, channel_map={'25313_Mean-L3-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25314_2 = TensorMap('25314_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0002778735756638309, 'std': 2.4446056399405565e-05}, annotation_units=1, channel_map={'25314_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25315_2 = TensorMap('25315_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00026289321990203657, 'std': 2.4459905737303918e-05}, annotation_units=1, channel_map={'25315_Mean-L3-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25316_2 = TensorMap('25316_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003543772106212941, 'std': 3.720570147558196e-05}, annotation_units=1, channel_map={'25316_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25317_2 = TensorMap('25317_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003502572828048465, 'std': 3.594445326718289e-05}, annotation_units=1, channel_map={'25317_Mean-L3-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25318_2 = TensorMap('25318_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004646825470482082, 'std': 4.643299716636984e-05}, annotation_units=1, channel_map={'25318_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25319_2 = TensorMap('25319_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00046933091002835774, 'std': 4.8496918770345514e-05}, annotation_units=1, channel_map={'25319_Mean-L3-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25320_2 = TensorMap('25320_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00041430899716421747, 'std': 3.761974877077602e-05}, annotation_units=1, channel_map={'25320_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25321_2 = TensorMap('25321_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00040835137922144894, 'std': 3.8468284703768805e-05}, annotation_units=1, channel_map={'25321_Mean-L3-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25322_2 = TensorMap('25322_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004630494457334364, 'std': 5.880316828950514e-05}, annotation_units=1, channel_map={'25322_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25323_2 = TensorMap('25323_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004620703789636504, 'std': 5.4653308837549596e-05}, annotation_units=1, channel_map={'25323_Mean-L3-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25324_2 = TensorMap('25324_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004196779066769786, 'std': 5.230098224988641e-05}, annotation_units=1, channel_map={'25324_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25325_2 = TensorMap('25325_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00041756328950760504, 'std': 5.6467431762540285e-05}, annotation_units=1, channel_map={'25325_Mean-L3-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25326_2 = TensorMap('25326_Mean-L3-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004106141789120908, 'std': 4.480631372074689e-05}, annotation_units=1, channel_map={'25326_Mean-L3-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25327_2 = TensorMap('25327_Mean-L3-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00039279381283836026, 'std': 4.465321816648295e-05}, annotation_units=1, channel_map={'25327_Mean-L3-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25328_2 = TensorMap('25328_Mean-L3-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004449262180974478, 'std': 3.605459993470109e-05}, annotation_units=1, channel_map={'25328_Mean-L3-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25329_2 = TensorMap('25329_Mean-L3-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004351589069347771, 'std': 3.8065988800951893e-05}, annotation_units=1, channel_map={'25329_Mean-L3-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25330_2 = TensorMap('25330_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00036692962103634973, 'std': 3.284348944687288e-05}, annotation_units=1, channel_map={'25330_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25331_2 = TensorMap('25331_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00033891848414539837, 'std': 3.2974029020487713e-05}, annotation_units=1, channel_map={'25331_Mean-L3-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25332_2 = TensorMap('25332_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004625660737303429, 'std': 3.897174026081075e-05}, annotation_units=1, channel_map={'25332_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25333_2 = TensorMap('25333_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004624208301108532, 'std': 3.982435861720428e-05}, annotation_units=1, channel_map={'25333_Mean-L3-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25334_2 = TensorMap('25334_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00047688058778035586, 'std': 6.466441378607682e-05}, annotation_units=1, channel_map={'25334_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25335_2 = TensorMap('25335_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00045872482598607883, 'std': 6.389582067374513e-05}, annotation_units=1, channel_map={'25335_Mean-L3-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25336_2 = TensorMap('25336_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0003859275070894561, 'std': 3.575873204329132e-05}, annotation_units=1, channel_map={'25336_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25337_2 = TensorMap('25337_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00036767955658674906, 'std': 3.505758124121823e-05}, annotation_units=1, channel_map={'25337_Mean-L3-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25338_2 = TensorMap('25338_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004691374581077598, 'std': 6.905896941686052e-05}, annotation_units=1, channel_map={'25338_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25339_2 = TensorMap('25339_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004727559164733179, 'std': 8.65907855307709e-05}, annotation_units=1, channel_map={'25339_Mean-L3-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25340_2 = TensorMap('25340_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004139774684196958, 'std': 3.806568599307937e-05}, annotation_units=1, channel_map={'25340_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25341_2 = TensorMap('25341_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004070772364011344, 'std': 4.2328011163093296e-05}, annotation_units=1, channel_map={'25341_Mean-L3-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25342_2 = TensorMap('25342_Mean-L3-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006006643464810519, 'std': 0.00014488347081099723}, annotation_units=1, channel_map={'25342_Mean-L3-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25343_2 = TensorMap('25343_Mean-L3-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005190773910801751, 'std': 0.0001413119750512089}, annotation_units=1, channel_map={'25343_Mean-L3-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25344_2 = TensorMap('25344_Mean-ICVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7481068312793278, 'std': 0.02681829845208484}, annotation_units=1, channel_map={'25344_Mean-ICVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25345_2 = TensorMap('25345_Mean-ICVF-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6855349943278501, 'std': 0.04170849758458551}, annotation_units=1, channel_map={'25345_Mean-ICVF-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25346_2 = TensorMap('25346_Mean-ICVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6234766743670399, 'std': 0.04032671182875544}, annotation_units=1, channel_map={'25346_Mean-ICVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25347_2 = TensorMap('25347_Mean-ICVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6593653633269738, 'std': 0.033048970504767926}, annotation_units=1, channel_map={'25347_Mean-ICVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25348_2 = TensorMap('25348_Mean-ICVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.698666089516836, 'std': 0.03136751161569296}, annotation_units=1, channel_map={'25348_Mean-ICVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25349_2 = TensorMap('25349_Mean-ICVF-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5618417133501776, 'std': 0.0670875654146576}, annotation_units=1, channel_map={'25349_Mean-ICVF-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25350_2 = TensorMap('25350_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7088014646521941, 'std': 0.044092042949029256}, annotation_units=1, channel_map={'25350_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25351_2 = TensorMap('25351_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7046006306914865, 'std': 0.04321804112434849}, annotation_units=1, channel_map={'25351_Mean-ICVF-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25352_2 = TensorMap('25352_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6167492400350642, 'std': 0.029647694249411747}, annotation_units=1, channel_map={'25352_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25353_2 = TensorMap('25353_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6195315852627238, 'std': 0.03162744428807357}, annotation_units=1, channel_map={'25353_Mean-ICVF-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25354_2 = TensorMap('25354_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.671117019749394, 'std': 0.02953666645592142}, annotation_units=1, channel_map={'25354_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25355_2 = TensorMap('25355_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6683452364770792, 'std': 0.029674190532234815}, annotation_units=1, channel_map={'25355_Mean-ICVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25356_2 = TensorMap('25356_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6964817204661475, 'std': 0.02434525992559885}, annotation_units=1, channel_map={'25356_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25357_2 = TensorMap('25357_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6967080066518846, 'std': 0.023915178058592033}, annotation_units=1, channel_map={'25357_Mean-ICVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25358_2 = TensorMap('25358_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7383904793997831, 'std': 0.030621533698744194}, annotation_units=1, channel_map={'25358_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25359_2 = TensorMap('25359_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7519810964265455, 'std': 0.02997966108968678}, annotation_units=1, channel_map={'25359_Mean-ICVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25360_2 = TensorMap('25360_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6578665561285, 'std': 0.031892658713024144}, annotation_units=1, channel_map={'25360_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25361_2 = TensorMap('25361_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6524151534058682, 'std': 0.03183692268973134}, annotation_units=1, channel_map={'25361_Mean-ICVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25362_2 = TensorMap('25362_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7133243628113236, 'std': 0.02514090439917014}, annotation_units=1, channel_map={'25362_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25363_2 = TensorMap('25363_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7322115230753363, 'std': 0.025448659402555352}, annotation_units=1, channel_map={'25363_Mean-ICVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25364_2 = TensorMap('25364_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5962624048883618, 'std': 0.035297936304253694}, annotation_units=1, channel_map={'25364_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25365_2 = TensorMap('25365_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6087925650492445, 'std': 0.0334606309889725}, annotation_units=1, channel_map={'25365_Mean-ICVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25366_2 = TensorMap('25366_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5620322591656783, 'std': 0.04435995069031044}, annotation_units=1, channel_map={'25366_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25367_2 = TensorMap('25367_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5528993596658588, 'std': 0.044235934321152096}, annotation_units=1, channel_map={'25367_Mean-ICVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25368_2 = TensorMap('25368_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6523950603310473, 'std': 0.036772698989151614}, annotation_units=1, channel_map={'25368_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25369_2 = TensorMap('25369_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6434211173619346, 'std': 0.037835731654482056}, annotation_units=1, channel_map={'25369_Mean-ICVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25370_2 = TensorMap('25370_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5412059385860875, 'std': 0.04687458331192564}, annotation_units=1, channel_map={'25370_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25371_2 = TensorMap('25371_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5546787842004846, 'std': 0.04503057247155445}, annotation_units=1, channel_map={'25371_Mean-ICVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25372_2 = TensorMap('25372_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5417456631774352, 'std': 0.0432410266915723}, annotation_units=1, channel_map={'25372_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25373_2 = TensorMap('25373_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5304279534368073, 'std': 0.04247593971845362}, annotation_units=1, channel_map={'25373_Mean-ICVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25374_2 = TensorMap('25374_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5308181692878875, 'std': 0.04286880222222924}, annotation_units=1, channel_map={'25374_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25375_2 = TensorMap('25375_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5331598335481879, 'std': 0.040144884428623996}, annotation_units=1, channel_map={'25375_Mean-ICVF-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25376_2 = TensorMap('25376_Mean-ICVF-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5124511785180219, 'std': 0.02762133716250724}, annotation_units=1, channel_map={'25376_Mean-ICVF-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25377_2 = TensorMap('25377_Mean-ICVF-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5139056967462486, 'std': 0.02787183592379542}, annotation_units=1, channel_map={'25377_Mean-ICVF-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25378_2 = TensorMap('25378_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5893529452895374, 'std': 0.03648317409458542}, annotation_units=1, channel_map={'25378_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25379_2 = TensorMap('25379_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6114728111174135, 'std': 0.037225447920900434}, annotation_units=1, channel_map={'25379_Mean-ICVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25380_2 = TensorMap('25380_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.49766142303924094, 'std': 0.032532823156771716}, annotation_units=1, channel_map={'25380_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25381_2 = TensorMap('25381_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.49646148553601815, 'std': 0.03135282032144341}, annotation_units=1, channel_map={'25381_Mean-ICVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25382_2 = TensorMap('25382_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5766239615840768, 'std': 0.034435485574000876}, annotation_units=1, channel_map={'25382_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25383_2 = TensorMap('25383_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5790222075491158, 'std': 0.03428173705544477}, annotation_units=1, channel_map={'25383_Mean-ICVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25384_2 = TensorMap('25384_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6416578999123393, 'std': 0.038265493926288065}, annotation_units=1, channel_map={'25384_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25385_2 = TensorMap('25385_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6543150087660495, 'std': 0.03840570365729213}, annotation_units=1, channel_map={'25385_Mean-ICVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25386_2 = TensorMap('25386_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.609676176867942, 'std': 0.05554521035086826}, annotation_units=1, channel_map={'25386_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25387_2 = TensorMap('25387_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6127446541019953, 'std': 0.06196780871695057}, annotation_units=1, channel_map={'25387_Mean-ICVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25388_2 = TensorMap('25388_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5063758997060795, 'std': 0.03062867080220147}, annotation_units=1, channel_map={'25388_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25389_2 = TensorMap('25389_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5070844518125093, 'std': 0.0314160488964558}, annotation_units=1, channel_map={'25389_Mean-ICVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25390_2 = TensorMap('25390_Mean-ICVF-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.47742412478729435, 'std': 0.06502709145220313}, annotation_units=1, channel_map={'25390_Mean-ICVF-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25391_2 = TensorMap('25391_Mean-ICVF-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5120490937967307, 'std': 0.0623002573602955}, annotation_units=1, channel_map={'25391_Mean-ICVF-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25392_2 = TensorMap('25392_Mean-OD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.19477689656061462, 'std': 0.016039870936557832}, annotation_units=1, channel_map={'25392_Mean-OD-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25393_2 = TensorMap('25393_Mean-OD-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.21188390666735432, 'std': 0.03605655823715568}, annotation_units=1, channel_map={'25393_Mean-OD-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25394_2 = TensorMap('25394_Mean-OD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06169824173670912, 'std': 0.010827282320478023}, annotation_units=1, channel_map={'25394_Mean-OD-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25395_2 = TensorMap('25395_Mean-OD-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06181817418656217, 'std': 0.012147800278580126}, annotation_units=1, channel_map={'25395_Mean-OD-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25396_2 = TensorMap('25396_Mean-OD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05290952106430153, 'std': 0.00948157449084136}, annotation_units=1, channel_map={'25396_Mean-OD-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25397_2 = TensorMap('25397_Mean-OD-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15055565111122565, 'std': 0.14178483239868903}, annotation_units=1, channel_map={'25397_Mean-OD-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25398_2 = TensorMap('25398_Mean-OD-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15500283292940742, 'std': 0.031283697562918655}, annotation_units=1, channel_map={'25398_Mean-OD-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25399_2 = TensorMap('25399_Mean-OD-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14548477558913012, 'std': 0.027968453569442785}, annotation_units=1, channel_map={'25399_Mean-OD-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25400_2 = TensorMap('25400_Mean-OD-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10297104939926777, 'std': 0.015185752562302994}, annotation_units=1, channel_map={'25400_Mean-OD-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25401_2 = TensorMap('25401_Mean-OD-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10609525437013358, 'std': 0.015291677476833493}, annotation_units=1, channel_map={'25401_Mean-OD-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25402_2 = TensorMap('25402_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15438273165575206, 'std': 0.02220936094830354}, annotation_units=1, channel_map={'25402_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25403_2 = TensorMap('25403_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15574285484453154, 'std': 0.021681718765491013}, annotation_units=1, channel_map={'25403_Mean-OD-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25404_2 = TensorMap('25404_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08091166307430517, 'std': 0.00995599964828363}, annotation_units=1, channel_map={'25404_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25405_2 = TensorMap('25405_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07897978791316454, 'std': 0.010137633711340108}, annotation_units=1, channel_map={'25405_Mean-OD-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25406_2 = TensorMap('25406_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10035389893260455, 'std': 0.013385836744490037}, annotation_units=1, channel_map={'25406_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25407_2 = TensorMap('25407_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1059949443613675, 'std': 0.013127446073340604}, annotation_units=1, channel_map={'25407_Mean-OD-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25408_2 = TensorMap('25408_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10038318620120662, 'std': 0.013046950202169838}, annotation_units=1, channel_map={'25408_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25409_2 = TensorMap('25409_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10862363754963132, 'std': 0.013660762697532936}, annotation_units=1, channel_map={'25409_Mean-OD-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25410_2 = TensorMap('25410_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09624926035167329, 'std': 0.013502319250505003}, annotation_units=1, channel_map={'25410_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25411_2 = TensorMap('25411_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10228201990408911, 'std': 0.013647964314776516}, annotation_units=1, channel_map={'25411_Mean-OD-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25412_2 = TensorMap('25412_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10313756406950961, 'std': 0.015388621405140718}, annotation_units=1, channel_map={'25412_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25413_2 = TensorMap('25413_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09635350229464243, 'std': 0.013447574714342403}, annotation_units=1, channel_map={'25413_Mean-OD-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25414_2 = TensorMap('25414_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1763186840612592, 'std': 0.018674699123807873}, annotation_units=1, channel_map={'25414_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25415_2 = TensorMap('25415_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18139603671427837, 'std': 0.018671510731686908}, annotation_units=1, channel_map={'25415_Mean-OD-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25416_2 = TensorMap('25416_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18857361455164237, 'std': 0.01979801774508456}, annotation_units=1, channel_map={'25416_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25417_2 = TensorMap('25417_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18374425679368847, 'std': 0.019850754014831835}, annotation_units=1, channel_map={'25417_Mean-OD-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25418_2 = TensorMap('25418_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1378032085804156, 'std': 0.0184939884017953}, annotation_units=1, channel_map={'25418_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25419_2 = TensorMap('25419_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15276610380034036, 'std': 0.017140545272070817}, annotation_units=1, channel_map={'25419_Mean-OD-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25420_2 = TensorMap('25420_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0803023249626154, 'std': 0.012107001824017962}, annotation_units=1, channel_map={'25420_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25421_2 = TensorMap('25421_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07439495457123703, 'std': 0.011776828449000666}, annotation_units=1, channel_map={'25421_Mean-OD-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25422_2 = TensorMap('25422_Mean-OD-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09787221703707522, 'std': 0.014074391182644542}, annotation_units=1, channel_map={'25422_Mean-OD-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25423_2 = TensorMap('25423_Mean-OD-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10038272376630743, 'std': 0.015337202894014633}, annotation_units=1, channel_map={'25423_Mean-OD-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25424_2 = TensorMap('25424_Mean-OD-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.17220051874387668, 'std': 0.01517830722808373}, annotation_units=1, channel_map={'25424_Mean-OD-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25425_2 = TensorMap('25425_Mean-OD-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.16981031459805085, 'std': 0.016489873159278023}, annotation_units=1, channel_map={'25425_Mean-OD-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25426_2 = TensorMap('25426_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08959805538080753, 'std': 0.017109168879605226}, annotation_units=1, channel_map={'25426_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25427_2 = TensorMap('25427_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07727519460630129, 'std': 0.014107956454346498}, annotation_units=1, channel_map={'25427_Mean-OD-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25428_2 = TensorMap('25428_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.16949292873717317, 'std': 0.027802559737129424}, annotation_units=1, channel_map={'25428_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25429_2 = TensorMap('25429_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.17234454483576547, 'std': 0.02918316375926478}, annotation_units=1, channel_map={'25429_Mean-OD-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25430_2 = TensorMap('25430_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.13432280797194862, 'std': 0.022636045008225768}, annotation_units=1, channel_map={'25430_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25431_2 = TensorMap('25431_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.13768959170834832, 'std': 0.024318914302282214}, annotation_units=1, channel_map={'25431_Mean-OD-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25432_2 = TensorMap('25432_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.139586067240757, 'std': 0.01578343424269306}, annotation_units=1, channel_map={'25432_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25433_2 = TensorMap('25433_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14272715356056312, 'std': 0.016650377948438298}, annotation_units=1, channel_map={'25433_Mean-OD-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25434_2 = TensorMap('25434_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1830753433713196, 'std': 0.030762896127372118}, annotation_units=1, channel_map={'25434_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25435_2 = TensorMap('25435_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1852713691538184, 'std': 0.029737551717517588}, annotation_units=1, channel_map={'25435_Mean-OD-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25436_2 = TensorMap('25436_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.11752799113082038, 'std': 0.023317558977997078}, annotation_units=1, channel_map={'25436_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25437_2 = TensorMap('25437_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1442115241582014, 'std': 0.024053984665982273}, annotation_units=1, channel_map={'25437_Mean-OD-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25438_2 = TensorMap('25438_Mean-OD-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0695937017996184, 'std': 0.020544197503427696}, annotation_units=1, channel_map={'25438_Mean-OD-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25439_2 = TensorMap('25439_Mean-OD-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06537176821533543, 'std': 0.024061884073445707}, annotation_units=1, channel_map={'25439_Mean-OD-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25440_2 = TensorMap('25440_Mean-ISOVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10698402011034909, 'std': 0.01773097017561393}, annotation_units=1, channel_map={'25440_Mean-ISOVF-in-middle-cerebellar-peduncle-on-FA-skeleton_2_0': 0, })
ukb_25441_2 = TensorMap('25441_Mean-ISOVF-in-pontine-crossing-tract-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15438005991852732, 'std': 0.057579338901531114}, annotation_units=1, channel_map={'25441_Mean-ISOVF-in-pontine-crossing-tract-on-FA-skeleton_2_0': 0, })
ukb_25442_2 = TensorMap('25442_Mean-ISOVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07657186324962614, 'std': 0.026162938619335562}, annotation_units=1, channel_map={'25442_Mean-ISOVF-in-genu-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25443_2 = TensorMap('25443_Mean-ISOVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10488392105398857, 'std': 0.021633856827710105}, annotation_units=1, channel_map={'25443_Mean-ISOVF-in-body-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25444_2 = TensorMap('25444_Mean-ISOVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07678576537926055, 'std': 0.014417864525933235}, annotation_units=1, channel_map={'25444_Mean-ISOVF-in-splenium-of-corpus-callosum-on-FA-skeleton_2_0': 0, })
ukb_25445_2 = TensorMap('25445_Mean-ISOVF-in-fornix-on-FA-skeleton_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5211547159799929, 'std': 0.136847357344079}, annotation_units=1, channel_map={'25445_Mean-ISOVF-in-fornix-on-FA-skeleton_2_0': 0, })
ukb_25446_2 = TensorMap('25446_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15335651905326667, 'std': 0.05059254048559402}, annotation_units=1, channel_map={'25446_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-right_2_0': 0, })
ukb_25447_2 = TensorMap('25447_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14739420063940595, 'std': 0.046079317514275214}, annotation_units=1, channel_map={'25447_Mean-ISOVF-in-corticospinal-tract-on-FA-skeleton-left_2_0': 0, })
ukb_25448_2 = TensorMap('25448_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0769796756561646, 'std': 0.03185971475080036}, annotation_units=1, channel_map={'25448_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-right_2_0': 0, })
ukb_25449_2 = TensorMap('25449_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08270397416593617, 'std': 0.0355525569682087}, annotation_units=1, channel_map={'25449_Mean-ISOVF-in-medial-lemniscus-on-FA-skeleton-left_2_0': 0, })
ukb_25450_2 = TensorMap('25450_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0855988069406487, 'std': 0.018582383020316784}, annotation_units=1, channel_map={'25450_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25451_2 = TensorMap('25451_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09295244232455008, 'std': 0.021322059792521962}, annotation_units=1, channel_map={'25451_Mean-ISOVF-in-inferior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25452_2 = TensorMap('25452_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1353281396895787, 'std': 0.023835587682274045}, annotation_units=1, channel_map={'25452_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25453_2 = TensorMap('25453_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12885126566286803, 'std': 0.02315939752352389}, annotation_units=1, channel_map={'25453_Mean-ISOVF-in-superior-cerebellar-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25454_2 = TensorMap('25454_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10303650105708245, 'std': 0.021587334285376427}, annotation_units=1, channel_map={'25454_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-right_2_0': 0, })
ukb_25455_2 = TensorMap('25455_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.10114516799876243, 'std': 0.019281584480603388}, annotation_units=1, channel_map={'25455_Mean-ISOVF-in-cerebral-peduncle-on-FA-skeleton-left_2_0': 0, })
ukb_25456_2 = TensorMap('25456_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0737033240860104, 'std': 0.0204504871586523}, annotation_units=1, channel_map={'25456_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25457_2 = TensorMap('25457_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06970067606868459, 'std': 0.01873541934966964}, annotation_units=1, channel_map={'25457_Mean-ISOVF-in-anterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25458_2 = TensorMap('25458_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07757493776104778, 'std': 0.014520955438902606}, annotation_units=1, channel_map={'25458_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25459_2 = TensorMap('25459_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08216944789356984, 'std': 0.015081267615285157}, annotation_units=1, channel_map={'25459_Mean-ISOVF-in-posterior-limb-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25460_2 = TensorMap('25460_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07044725529830348, 'std': 0.017934752495424748}, annotation_units=1, channel_map={'25460_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25461_2 = TensorMap('25461_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07153694606301242, 'std': 0.01759697627087865}, annotation_units=1, channel_map={'25461_Mean-ISOVF-in-retrolenticular-part-of-internal-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25462_2 = TensorMap('25462_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.051786008456659614, 'std': 0.01417312743952105}, annotation_units=1, channel_map={'25462_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25463_2 = TensorMap('25463_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.053213915639663795, 'std': 0.014859608120490022}, annotation_units=1, channel_map={'25463_Mean-ISOVF-in-anterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25464_2 = TensorMap('25464_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08186501706801422, 'std': 0.014837403218613696}, annotation_units=1, channel_map={'25464_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25465_2 = TensorMap('25465_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07924905749497242, 'std': 0.014565035176854755}, annotation_units=1, channel_map={'25465_Mean-ISOVF-in-superior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25466_2 = TensorMap('25466_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08457295106481717, 'std': 0.02093349439674238}, annotation_units=1, channel_map={'25466_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-right_2_0': 0, })
ukb_25467_2 = TensorMap('25467_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08070702604032379, 'std': 0.018969008241735065}, annotation_units=1, channel_map={'25467_Mean-ISOVF-in-posterior-corona-radiata-on-FA-skeleton-left_2_0': 0, })
ukb_25468_2 = TensorMap('25468_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08896963038209664, 'std': 0.017893082449300563}, annotation_units=1, channel_map={'25468_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-right_2_0': 0, })
ukb_25469_2 = TensorMap('25469_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08009975862424586, 'std': 0.018664484351796518}, annotation_units=1, channel_map={'25469_Mean-ISOVF-in-posterior-thalamic-radiation-on-FA-skeleton-left_2_0': 0, })
ukb_25470_2 = TensorMap('25470_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06902422626720982, 'std': 0.017138786000679886}, annotation_units=1, channel_map={'25470_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-right_2_0': 0, })
ukb_25471_2 = TensorMap('25471_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.055099579435878926, 'std': 0.01644345700977556}, annotation_units=1, channel_map={'25471_Mean-ISOVF-in-sagittal-stratum-on-FA-skeleton-left_2_0': 0, })
ukb_25472_2 = TensorMap('25472_Mean-ISOVF-in-external-capsule-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03152925143092868, 'std': 0.013942915042069218}, annotation_units=1, channel_map={'25472_Mean-ISOVF-in-external-capsule-on-FA-skeleton-right_2_0': 0, })
ukb_25473_2 = TensorMap('25473_Mean-ISOVF-in-external-capsule-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.029410372608673234, 'std': 0.013161206160798113}, annotation_units=1, channel_map={'25473_Mean-ISOVF-in-external-capsule-on-FA-skeleton-left_2_0': 0, })
ukb_25474_2 = TensorMap('25474_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.057677709018718104, 'std': 0.017024295396865792}, annotation_units=1, channel_map={'25474_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-right_2_0': 0, })
ukb_25475_2 = TensorMap('25475_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06564833347083997, 'std': 0.01758122253718271}, annotation_units=1, channel_map={'25475_Mean-ISOVF-in-cingulum-cingulate-gyrus-on-FA-skeleton-left_2_0': 0, })
ukb_25476_2 = TensorMap('25476_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.030317177177332028, 'std': 0.01360981671222079}, annotation_units=1, channel_map={'25476_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-right_2_0': 0, })
ukb_25477_2 = TensorMap('25477_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03142494580518744, 'std': 0.015246041944447293}, annotation_units=1, channel_map={'25477_Mean-ISOVF-in-cingulum-hippocampus-on-FA-skeleton-left_2_0': 0, })
ukb_25478_2 = TensorMap('25478_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.13913323678646936, 'std': 0.04499240748789466}, annotation_units=1, channel_map={'25478_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-right_2_0': 0, })
ukb_25479_2 = TensorMap('25479_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12483551910483165, 'std': 0.043648885017233685}, annotation_units=1, channel_map={'25479_Mean-ISOVF-in-fornix-cresstria-terminalis-on-FA-skeleton-left_2_0': 0, })
ukb_25480_2 = TensorMap('25480_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08048866683854998, 'std': 0.014695081475721605}, annotation_units=1, channel_map={'25480_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25481_2 = TensorMap('25481_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07349104279894811, 'std': 0.013101655013735147}, annotation_units=1, channel_map={'25481_Mean-ISOVF-in-superior-longitudinal-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25482_2 = TensorMap('25482_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06816360805445262, 'std': 0.028148141958239586}, annotation_units=1, channel_map={'25482_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25483_2 = TensorMap('25483_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07562154385603054, 'std': 0.033112918566038}, annotation_units=1, channel_map={'25483_Mean-ISOVF-in-superior-frontooccipital-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25484_2 = TensorMap('25484_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.018655110813180022, 'std': 0.013334566127985583}, annotation_units=1, channel_map={'25484_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-right_2_0': 0, })
ukb_25485_2 = TensorMap('25485_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.020850734388696952, 'std': 0.014943710681234595}, annotation_units=1, channel_map={'25485_Mean-ISOVF-in-uncinate-fasciculus-on-FA-skeleton-left_2_0': 0, })
ukb_25486_2 = TensorMap('25486_Mean-ISOVF-in-tapetum-on-FA-skeleton-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.17718389671530965, 'std': 0.07452034550090023}, annotation_units=1, channel_map={'25486_Mean-ISOVF-in-tapetum-on-FA-skeleton-right_2_0': 0, })
ukb_25487_2 = TensorMap('25487_Mean-ISOVF-in-tapetum-on-FA-skeleton-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14511192249780844, 'std': 0.07899431549101103}, annotation_units=1, channel_map={'25487_Mean-ISOVF-in-tapetum-on-FA-skeleton-left_2_0': 0, })
ukb_25488_2 = TensorMap('25488_Weightedmean-FA-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.42162731196823594, 'std': 0.024394649781991212}, annotation_units=1, channel_map={'25488_Weightedmean-FA-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25489_2 = TensorMap('25489_Weightedmean-FA-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4119231913577065, 'std': 0.02360338953913445}, annotation_units=1, channel_map={'25489_Weightedmean-FA-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25490_2 = TensorMap('25490_Weightedmean-FA-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.398801685814469, 'std': 0.020188517572455465}, annotation_units=1, channel_map={'25490_Weightedmean-FA-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25491_2 = TensorMap('25491_Weightedmean-FA-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3912875507657403, 'std': 0.019757856381831713}, annotation_units=1, channel_map={'25491_Weightedmean-FA-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25492_2 = TensorMap('25492_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5334738380343425, 'std': 0.0352614634322073}, annotation_units=1, channel_map={'25492_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25493_2 = TensorMap('25493_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4961802892796371, 'std': 0.03485187185906978}, annotation_units=1, channel_map={'25493_Weightedmean-FA-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25494_2 = TensorMap('25494_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3105439489506522, 'std': 0.0314814707367384}, annotation_units=1, channel_map={'25494_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25495_2 = TensorMap('25495_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3088009859227555, 'std': 0.033348592419987534}, annotation_units=1, channel_map={'25495_Weightedmean-FA-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25496_2 = TensorMap('25496_Weightedmean-FA-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5447855636054247, 'std': 0.026092574295239805}, annotation_units=1, channel_map={'25496_Weightedmean-FA-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25497_2 = TensorMap('25497_Weightedmean-FA-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5379864079822616, 'std': 0.02685776741775756}, annotation_units=1, channel_map={'25497_Weightedmean-FA-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25498_2 = TensorMap('25498_Weightedmean-FA-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5815279512195123, 'std': 0.02937310961936468}, annotation_units=1, channel_map={'25498_Weightedmean-FA-in-tract-forceps-major_2_0': 0, })
ukb_25499_2 = TensorMap('25499_Weightedmean-FA-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4630739470427475, 'std': 0.02242947074167559}, annotation_units=1, channel_map={'25499_Weightedmean-FA-in-tract-forceps-minor_2_0': 0, })
ukb_25500_2 = TensorMap('25500_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.47623995374619704, 'std': 0.024237071265860187}, annotation_units=1, channel_map={'25500_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25501_2 = TensorMap('25501_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46416634821842934, 'std': 0.022886610981506773}, annotation_units=1, channel_map={'25501_Weightedmean-FA-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25502_2 = TensorMap('25502_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.46072070468725834, 'std': 0.022097695304961813}, annotation_units=1, channel_map={'25502_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25503_2 = TensorMap('25503_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4504131259217243, 'std': 0.020332599437893843}, annotation_units=1, channel_map={'25503_Weightedmean-FA-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25504_2 = TensorMap('25504_Weightedmean-FA-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4801722974269067, 'std': 0.031134842699146563}, annotation_units=1, channel_map={'25504_Weightedmean-FA-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25505_2 = TensorMap('25505_Weightedmean-FA-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4192665318929511, 'std': 0.02460691066158585}, annotation_units=1, channel_map={'25505_Weightedmean-FA-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25506_2 = TensorMap('25506_Weightedmean-FA-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4217943451760943, 'std': 0.025192821230324302}, annotation_units=1, channel_map={'25506_Weightedmean-FA-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25507_2 = TensorMap('25507_Weightedmean-FA-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4575505770123239, 'std': 0.02337073509417849}, annotation_units=1, channel_map={'25507_Weightedmean-FA-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25508_2 = TensorMap('25508_Weightedmean-FA-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.45350572412726253, 'std': 0.02292296784572249}, annotation_units=1, channel_map={'25508_Weightedmean-FA-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25509_2 = TensorMap('25509_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.44245008126643637, 'std': 0.021683278621781495}, annotation_units=1, channel_map={'25509_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25510_2 = TensorMap('25510_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.42475595637601204, 'std': 0.02108071704158007}, annotation_units=1, channel_map={'25510_Weightedmean-FA-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25511_2 = TensorMap('25511_Weightedmean-FA-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4224994293817357, 'std': 0.019229385125723714}, annotation_units=1, channel_map={'25511_Weightedmean-FA-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25512_2 = TensorMap('25512_Weightedmean-FA-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.42100600835352964, 'std': 0.019803843674420143}, annotation_units=1, channel_map={'25512_Weightedmean-FA-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25513_2 = TensorMap('25513_Weightedmean-FA-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.38870587021090097, 'std': 0.024431506861994107}, annotation_units=1, channel_map={'25513_Weightedmean-FA-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25514_2 = TensorMap('25514_Weightedmean-FA-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.38839686278554125, 'std': 0.02196908663674339}, annotation_units=1, channel_map={'25514_Weightedmean-FA-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25515_2 = TensorMap('25515_Weightedmean-MD-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007869030282060537, 'std': 4.3084909814231626e-05}, annotation_units=1, channel_map={'25515_Weightedmean-MD-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25516_2 = TensorMap('25516_Weightedmean-MD-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007905745437529003, 'std': 4.3469946141717226e-05}, annotation_units=1, channel_map={'25516_Weightedmean-MD-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25517_2 = TensorMap('25517_Weightedmean-MD-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007831724827515082, 'std': 3.9738280255526846e-05}, annotation_units=1, channel_map={'25517_Weightedmean-MD-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25518_2 = TensorMap('25518_Weightedmean-MD-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007811141574795028, 'std': 3.948309769992209e-05}, annotation_units=1, channel_map={'25518_Weightedmean-MD-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25519_2 = TensorMap('25519_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007587803560563089, 'std': 3.156040541586139e-05}, annotation_units=1, channel_map={'25519_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25520_2 = TensorMap('25520_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007614496131078223, 'std': 3.17772113128607e-05}, annotation_units=1, channel_map={'25520_Weightedmean-MD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25521_2 = TensorMap('25521_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008807903554375291, 'std': 6.945548497095596e-05}, annotation_units=1, channel_map={'25521_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25522_2 = TensorMap('25522_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008938067207755376, 'std': 6.834920047889979e-05}, annotation_units=1, channel_map={'25522_Weightedmean-MD-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25523_2 = TensorMap('25523_Weightedmean-MD-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007741282360129942, 'std': 3.126916228285023e-05}, annotation_units=1, channel_map={'25523_Weightedmean-MD-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25524_2 = TensorMap('25524_Weightedmean-MD-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007774557617181458, 'std': 3.151923752754262e-05}, annotation_units=1, channel_map={'25524_Weightedmean-MD-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25525_2 = TensorMap('25525_Weightedmean-MD-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008984488731501056, 'std': 5.591892034177489e-05}, annotation_units=1, channel_map={'25525_Weightedmean-MD-in-tract-forceps-major_2_0': 0, })
ukb_25526_2 = TensorMap('25526_Weightedmean-MD-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008353120969421958, 'std': 3.7847879034828514e-05}, annotation_units=1, channel_map={'25526_Weightedmean-MD-in-tract-forceps-minor_2_0': 0, })
ukb_25527_2 = TensorMap('25527_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000805302299644202, 'std': 3.7500160537494495e-05}, annotation_units=1, channel_map={'25527_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25528_2 = TensorMap('25528_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008135866021244778, 'std': 3.780330935182627e-05}, annotation_units=1, channel_map={'25528_Weightedmean-MD-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25529_2 = TensorMap('25529_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008138452035786107, 'std': 3.7020952454305964e-05}, annotation_units=1, channel_map={'25529_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25530_2 = TensorMap('25530_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000823214381735678, 'std': 3.5627461666418666e-05}, annotation_units=1, channel_map={'25530_Weightedmean-MD-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25531_2 = TensorMap('25531_Weightedmean-MD-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007636320447584182, 'std': 6.563146968349835e-05}, annotation_units=1, channel_map={'25531_Weightedmean-MD-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25532_2 = TensorMap('25532_Weightedmean-MD-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0009031885684009695, 'std': 4.436746316091865e-05}, annotation_units=1, channel_map={'25532_Weightedmean-MD-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25533_2 = TensorMap('25533_Weightedmean-MD-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0009031237405249317, 'std': 4.4826159465692625e-05}, annotation_units=1, channel_map={'25533_Weightedmean-MD-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25534_2 = TensorMap('25534_Weightedmean-MD-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000842272579693704, 'std': 4.891610348987932e-05}, annotation_units=1, channel_map={'25534_Weightedmean-MD-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25535_2 = TensorMap('25535_Weightedmean-MD-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008443117707935854, 'std': 4.732692811860795e-05}, annotation_units=1, channel_map={'25535_Weightedmean-MD-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25536_2 = TensorMap('25536_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007425505761872841, 'std': 3.2841134673408404e-05}, annotation_units=1, channel_map={'25536_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25537_2 = TensorMap('25537_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007522032833496624, 'std': 3.3358966798692036e-05}, annotation_units=1, channel_map={'25537_Weightedmean-MD-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25538_2 = TensorMap('25538_Weightedmean-MD-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000759284841695457, 'std': 3.099410043052588e-05}, annotation_units=1, channel_map={'25538_Weightedmean-MD-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25539_2 = TensorMap('25539_Weightedmean-MD-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007548022818027124, 'std': 3.226096974016518e-05}, annotation_units=1, channel_map={'25539_Weightedmean-MD-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25540_2 = TensorMap('25540_Weightedmean-MD-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008003234097870366, 'std': 4.2293884426181756e-05}, annotation_units=1, channel_map={'25540_Weightedmean-MD-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25541_2 = TensorMap('25541_Weightedmean-MD-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008014866571443306, 'std': 3.7014979297540434e-05}, annotation_units=1, channel_map={'25541_Weightedmean-MD-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25542_2 = TensorMap('25542_Weightedmean-MO-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3723602302377147, 'std': 0.06343002132052704}, annotation_units=1, channel_map={'25542_Weightedmean-MO-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25543_2 = TensorMap('25543_Weightedmean-MO-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3359465264528438, 'std': 0.06121759561551163}, annotation_units=1, channel_map={'25543_Weightedmean-MO-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25544_2 = TensorMap('25544_Weightedmean-MO-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.40753672051771245, 'std': 0.04344058967648053}, annotation_units=1, channel_map={'25544_Weightedmean-MO-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25545_2 = TensorMap('25545_Weightedmean-MO-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.411174270252153, 'std': 0.04283216746118492}, annotation_units=1, channel_map={'25545_Weightedmean-MO-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25546_2 = TensorMap('25546_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6399078433455369, 'std': 0.06291288270814267}, annotation_units=1, channel_map={'25546_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25547_2 = TensorMap('25547_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5666190229979889, 'std': 0.07118829535741229}, annotation_units=1, channel_map={'25547_Weightedmean-MO-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25548_2 = TensorMap('25548_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.42142478641262326, 'std': 0.06665914359134581}, annotation_units=1, channel_map={'25548_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25549_2 = TensorMap('25549_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.43607815651008086, 'std': 0.06142206542913853}, annotation_units=1, channel_map={'25549_Weightedmean-MO-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25550_2 = TensorMap('25550_Weightedmean-MO-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.629911979734956, 'std': 0.049756135353166704}, annotation_units=1, channel_map={'25550_Weightedmean-MO-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25551_2 = TensorMap('25551_Weightedmean-MO-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6347086356932916, 'std': 0.04872839957944134}, annotation_units=1, channel_map={'25551_Weightedmean-MO-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25552_2 = TensorMap('25552_Weightedmean-MO-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.727311957562007, 'std': 0.03721642333809005}, annotation_units=1, channel_map={'25552_Weightedmean-MO-in-tract-forceps-major_2_0': 0, })
ukb_25553_2 = TensorMap('25553_Weightedmean-MO-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.49274900711596975, 'std': 0.038652042785986815}, annotation_units=1, channel_map={'25553_Weightedmean-MO-in-tract-forceps-minor_2_0': 0, })
ukb_25554_2 = TensorMap('25554_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5420935222502961, 'std': 0.03807206978641983}, annotation_units=1, channel_map={'25554_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25555_2 = TensorMap('25555_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5349394000412521, 'std': 0.03696540400460718}, annotation_units=1, channel_map={'25555_Weightedmean-MO-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25556_2 = TensorMap('25556_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.49921609642654574, 'std': 0.036456518722492316}, annotation_units=1, channel_map={'25556_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25557_2 = TensorMap('25557_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4795984077244366, 'std': 0.03427846662633617}, annotation_units=1, channel_map={'25557_Weightedmean-MO-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25558_2 = TensorMap('25558_Weightedmean-MO-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.49747569050172735, 'std': 0.07162082206084824}, annotation_units=1, channel_map={'25558_Weightedmean-MO-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25559_2 = TensorMap('25559_Weightedmean-MO-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6682906909709692, 'std': 0.044214581757383534}, annotation_units=1, channel_map={'25559_Weightedmean-MO-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25560_2 = TensorMap('25560_Weightedmean-MO-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6716288589955136, 'std': 0.044757298747290036}, annotation_units=1, channel_map={'25560_Weightedmean-MO-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25561_2 = TensorMap('25561_Weightedmean-MO-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.48495937405249323, 'std': 0.047113795794956725}, annotation_units=1, channel_map={'25561_Weightedmean-MO-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25562_2 = TensorMap('25562_Weightedmean-MO-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.48110872118805753, 'std': 0.046647154321018784}, annotation_units=1, channel_map={'25562_Weightedmean-MO-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25563_2 = TensorMap('25563_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.2990255153663693, 'std': 0.049384809897619245}, annotation_units=1, channel_map={'25563_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25564_2 = TensorMap('25564_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3086045419589542, 'std': 0.04488263709961174}, annotation_units=1, channel_map={'25564_Weightedmean-MO-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25565_2 = TensorMap('25565_Weightedmean-MO-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.39366036986541525, 'std': 0.058838487794720166}, annotation_units=1, channel_map={'25565_Weightedmean-MO-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25566_2 = TensorMap('25566_Weightedmean-MO-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4161323384726448, 'std': 0.05873463279563613}, annotation_units=1, channel_map={'25566_Weightedmean-MO-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25567_2 = TensorMap('25567_Weightedmean-MO-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4792618499680298, 'std': 0.04287813027371429}, annotation_units=1, channel_map={'25567_Weightedmean-MO-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25568_2 = TensorMap('25568_Weightedmean-MO-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5035559040272264, 'std': 0.04423289524073138}, annotation_units=1, channel_map={'25568_Weightedmean-MO-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25569_2 = TensorMap('25569_Weightedmean-L1-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001160300154901253, 'std': 5.523033658496576e-05}, annotation_units=1, channel_map={'25569_Weightedmean-L1-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25570_2 = TensorMap('25570_Weightedmean-L1-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011511015739699897, 'std': 5.34608867927525e-05}, annotation_units=1, channel_map={'25570_Weightedmean-L1-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25571_2 = TensorMap('25571_Weightedmean-L1-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011343078196256382, 'std': 4.608247687997705e-05}, annotation_units=1, channel_map={'25571_Weightedmean-L1-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25572_2 = TensorMap('25572_Weightedmean-L1-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011256360114474295, 'std': 4.665547873788362e-05}, annotation_units=1, channel_map={'25572_Weightedmean-L1-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25573_2 = TensorMap('25573_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001284373860877636, 'std': 6.049086943621804e-05}, annotation_units=1, channel_map={'25573_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25574_2 = TensorMap('25574_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001233968592636518, 'std': 5.8572778207711583e-05}, annotation_units=1, channel_map={'25574_Weightedmean-L1-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25575_2 = TensorMap('25575_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011749437494972415, 'std': 7.04336940603974e-05}, annotation_units=1, channel_map={'25575_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25576_2 = TensorMap('25576_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011900531577373277, 'std': 7.024728746625796e-05}, annotation_units=1, channel_map={'25576_Weightedmean-L1-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25577_2 = TensorMap('25577_Weightedmean-L1-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001309760491930078, 'std': 5.079245695438066e-05}, annotation_units=1, channel_map={'25577_Weightedmean-L1-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25578_2 = TensorMap('25578_Weightedmean-L1-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001306003479090393, 'std': 5.1430129243528776e-05}, annotation_units=1, channel_map={'25578_Weightedmean-L1-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25579_2 = TensorMap('25579_Weightedmean-L1-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0015699223049553959, 'std': 6.813094733848717e-05}, annotation_units=1, channel_map={'25579_Weightedmean-L1-in-tract-forceps-major_2_0': 0, })
ukb_25580_2 = TensorMap('25580_Weightedmean-L1-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013003536312071367, 'std': 4.666779588558039e-05}, annotation_units=1, channel_map={'25580_Weightedmean-L1-in-tract-forceps-minor_2_0': 0, })
ukb_25581_2 = TensorMap('25581_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001269560530088176, 'std': 4.7266768198446516e-05}, annotation_units=1, channel_map={'25581_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25582_2 = TensorMap('25582_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001267374112824215, 'std': 4.7535814100451844e-05}, annotation_units=1, channel_map={'25582_Weightedmean-L1-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25583_2 = TensorMap('25583_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012643147501675857, 'std': 4.3136957399530045e-05}, annotation_units=1, channel_map={'25583_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25584_2 = TensorMap('25584_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012648308214304135, 'std': 4.0859940282448314e-05}, annotation_units=1, channel_map={'25584_Weightedmean-L1-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25585_2 = TensorMap('25585_Weightedmean-L1-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011919913877687825, 'std': 9.972581890669109e-05}, annotation_units=1, channel_map={'25585_Weightedmean-L1-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25586_2 = TensorMap('25586_Weightedmean-L1-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013939196369824158, 'std': 5.701800935053113e-05}, annotation_units=1, channel_map={'25586_Weightedmean-L1-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25587_2 = TensorMap('25587_Weightedmean-L1-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013927123912236374, 'std': 5.6755816847230096e-05}, annotation_units=1, channel_map={'25587_Weightedmean-L1-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25588_2 = TensorMap('25588_Weightedmean-L1-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0013036436322384362, 'std': 5.7031257123307634e-05}, annotation_units=1, channel_map={'25588_Weightedmean-L1-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25589_2 = TensorMap('25589_Weightedmean-L1-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0012996039395658228, 'std': 5.706440378378183e-05}, annotation_units=1, channel_map={'25589_Weightedmean-L1-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25590_2 = TensorMap('25590_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011144643957097925, 'std': 3.6315952186549145e-05}, annotation_units=1, channel_map={'25590_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25591_2 = TensorMap('25591_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001111387938276698, 'std': 3.723812965057725e-05}, annotation_units=1, channel_map={'25591_Weightedmean-L1-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25592_2 = TensorMap('25592_Weightedmean-L1-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001121314615067292, 'std': 4.2414256244690855e-05}, annotation_units=1, channel_map={'25592_Weightedmean-L1-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25593_2 = TensorMap('25593_Weightedmean-L1-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011152959492084773, 'std': 4.3879232947486283e-05}, annotation_units=1, channel_map={'25593_Weightedmean-L1-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25594_2 = TensorMap('25594_Weightedmean-L1-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011554275707729587, 'std': 4.562599219413665e-05}, annotation_units=1, channel_map={'25594_Weightedmean-L1-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25595_2 = TensorMap('25595_Weightedmean-L1-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0011609301984736762, 'std': 4.221534064619681e-05}, annotation_units=1, channel_map={'25595_Weightedmean-L1-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25596_2 = TensorMap('25596_Weightedmean-L2-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007102206953539937, 'std': 4.34087445847932e-05}, annotation_units=1, channel_map={'25596_Weightedmean-L2-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25597_2 = TensorMap('25597_Weightedmean-L2-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007235375597380496, 'std': 4.3785109596356453e-05}, annotation_units=1, channel_map={'25597_Weightedmean-L2-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25598_2 = TensorMap('25598_Weightedmean-L2-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007063118435002318, 'std': 3.945149451225539e-05}, annotation_units=1, channel_map={'25598_Weightedmean-L2-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25599_2 = TensorMap('25599_Weightedmean-L2-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007044746939617387, 'std': 3.887479751579304e-05}, annotation_units=1, channel_map={'25599_Weightedmean-L2-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25600_2 = TensorMap('25600_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005894011401536634, 'std': 3.765395509095012e-05}, annotation_units=1, channel_map={'25600_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25601_2 = TensorMap('25601_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006240759934512455, 'std': 3.744511676473526e-05}, annotation_units=1, channel_map={'25601_Weightedmean-L2-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25602_2 = TensorMap('25602_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008125015006961271, 'std': 7.38622573106632e-05}, annotation_units=1, channel_map={'25602_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25603_2 = TensorMap('25603_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0008228520331047286, 'std': 7.200775819873211e-05}, annotation_units=1, channel_map={'25603_Weightedmean-L2-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25604_2 = TensorMap('25604_Weightedmean-L2-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006113771945031712, 'std': 3.180780703687487e-05}, annotation_units=1, channel_map={'25604_Weightedmean-L2-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25605_2 = TensorMap('25605_Weightedmean-L2-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006151738767596553, 'std': 3.180385667142979e-05}, annotation_units=1, channel_map={'25605_Weightedmean-L2-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25606_2 = TensorMap('25606_Weightedmean-L2-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006445785606146546, 'std': 5.8579443526464155e-05}, annotation_units=1, channel_map={'25606_Weightedmean-L2-in-tract-forceps-major_2_0': 0, })
ukb_25607_2 = TensorMap('25607_Weightedmean-L2-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006985027987418138, 'std': 3.7994247568881644e-05}, annotation_units=1, channel_map={'25607_Weightedmean-L2-in-tract-forceps-minor_2_0': 0, })
ukb_25608_2 = TensorMap('25608_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000678761891300985, 'std': 3.8352522774597145e-05}, annotation_units=1, channel_map={'25608_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25609_2 = TensorMap('25609_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000690968733099572, 'std': 3.7814763344068e-05}, annotation_units=1, channel_map={'25609_Weightedmean-L2-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25610_2 = TensorMap('25610_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000696131311710411, 'std': 3.849070155453533e-05}, annotation_units=1, channel_map={'25610_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25611_2 = TensorMap('25611_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007111307221162276, 'std': 3.66541757265504e-05}, annotation_units=1, channel_map={'25611_Weightedmean-L2-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25612_2 = TensorMap('25612_Weightedmean-L2-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006468601731036974, 'std': 5.6288491153471694e-05}, annotation_units=1, channel_map={'25612_Weightedmean-L2-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25613_2 = TensorMap('25613_Weightedmean-L2-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007440684308255556, 'std': 4.527072912764642e-05}, annotation_units=1, channel_map={'25613_Weightedmean-L2-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25614_2 = TensorMap('25614_Weightedmean-L2-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007435297715670602, 'std': 4.5700919767002754e-05}, annotation_units=1, channel_map={'25614_Weightedmean-L2-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25615_2 = TensorMap('25615_Weightedmean-L2-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007169772422007942, 'std': 5.019571967787343e-05}, annotation_units=1, channel_map={'25615_Weightedmean-L2-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25616_2 = TensorMap('25616_Weightedmean-L2-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007230264622802043, 'std': 4.74539533059934e-05}, annotation_units=1, channel_map={'25616_Weightedmean-L2-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25617_2 = TensorMap('25617_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00068149288279276, 'std': 3.429711033486968e-05}, annotation_units=1, channel_map={'25617_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25618_2 = TensorMap('25618_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006914052067756403, 'std': 3.432637323549065e-05}, annotation_units=1, channel_map={'25618_Weightedmean-L2-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25619_2 = TensorMap('25619_Weightedmean-L2-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006829259738049809, 'std': 3.089357707357006e-05}, annotation_units=1, channel_map={'25619_Weightedmean-L2-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25620_2 = TensorMap('25620_Weightedmean-L2-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006758934721291188, 'std': 3.204204166038166e-05}, annotation_units=1, channel_map={'25620_Weightedmean-L2-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25621_2 = TensorMap('25621_Weightedmean-L2-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007115792162120352, 'std': 4.4333818462185704e-05}, annotation_units=1, channel_map={'25621_Weightedmean-L2-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25622_2 = TensorMap('25622_Weightedmean-L2-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0007073313640488837, 'std': 3.889499450577756e-05}, annotation_units=1, channel_map={'25622_Weightedmean-L2-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25623_2 = TensorMap('25623_Weightedmean-L3-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004901882160573403, 'std': 4.056474348148887e-05}, annotation_units=1, channel_map={'25623_Weightedmean-L3-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25624_2 = TensorMap('25624_Weightedmean-L3-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004970844661475792, 'std': 4.142442047067991e-05}, annotation_units=1, channel_map={'25624_Weightedmean-L3-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25625_2 = TensorMap('25625_Weightedmean-L3-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005088977910070646, 'std': 3.871858893767657e-05}, annotation_units=1, channel_map={'25625_Weightedmean-L3-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25626_2 = TensorMap('25626_Weightedmean-L3-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005132317678028155, 'std': 3.801188084269072e-05}, annotation_units=1, channel_map={'25626_Weightedmean-L3-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25627_2 = TensorMap('25627_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004025661043159903, 'std': 3.3672488006517885e-05}, annotation_units=1, channel_map={'25627_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25628_2 = TensorMap('25628_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004263042629814881, 'std': 3.331434939074652e-05}, annotation_units=1, channel_map={'25628_Weightedmean-L3-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25629_2 = TensorMap('25629_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006549257625947507, 'std': 7.198790606799525e-05}, annotation_units=1, channel_map={'25629_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25630_2 = TensorMap('25630_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0006685150187696591, 'std': 7.225237465586918e-05}, annotation_units=1, channel_map={'25630_Weightedmean-L3-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25631_2 = TensorMap('25631_Weightedmean-L3-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004012470198525241, 'std': 2.8073056430446206e-05}, annotation_units=1, channel_map={'25631_Weightedmean-L3-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25632_2 = TensorMap('25632_Weightedmean-L3-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004111899181147836, 'std': 2.8362386115899962e-05}, annotation_units=1, channel_map={'25632_Weightedmean-L3-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25633_2 = TensorMap('25633_Weightedmean-L3-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00048084577950806984, 'std': 5.625349829206131e-05}, annotation_units=1, channel_map={'25633_Weightedmean-L3-in-tract-forceps-major_2_0': 0, })
ukb_25634_2 = TensorMap('25634_Weightedmean-L3-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005070798322074977, 'std': 3.760515192153008e-05}, annotation_units=1, channel_map={'25634_Weightedmean-L3-in-tract-forceps-minor_2_0': 0, })
ukb_25635_2 = TensorMap('25635_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004675844617645543, 'std': 3.6685680944106954e-05}, annotation_units=1, channel_map={'25635_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25636_2 = TensorMap('25636_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004824169880369204, 'std': 3.656775249486129e-05}, annotation_units=1, channel_map={'25636_Weightedmean-L3-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25637_2 = TensorMap('25637_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004810895143092869, 'std': 3.700856198365724e-05}, annotation_units=1, channel_map={'25637_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25638_2 = TensorMap('25638_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004936816047027276, 'std': 3.6109077242992956e-05}, annotation_units=1, channel_map={'25638_Weightedmean-L3-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25639_2 = TensorMap('25639_Weightedmean-L3-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00045204458459237865, 'std': 5.365052618647152e-05}, annotation_units=1, channel_map={'25639_Weightedmean-L3-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25640_2 = TensorMap('25640_Weightedmean-L3-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005715775982571032, 'std': 4.275367683340966e-05}, annotation_units=1, channel_map={'25640_Weightedmean-L3-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25641_2 = TensorMap('25641_Weightedmean-L3-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005731290993657507, 'std': 4.324825524806439e-05}, annotation_units=1, channel_map={'25641_Weightedmean-L3-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25642_2 = TensorMap('25642_Weightedmean-L3-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005061968642293611, 'std': 4.772005479675345e-05}, annotation_units=1, channel_map={'25642_Weightedmean-L3-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25643_2 = TensorMap('25643_Weightedmean-L3-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005103049279636981, 'std': 4.5857687798029426e-05}, annotation_units=1, channel_map={'25643_Weightedmean-L3-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25644_2 = TensorMap('25644_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00043169444309802506, 'std': 3.489234285906859e-05}, annotation_units=1, channel_map={'25644_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25645_2 = TensorMap('25645_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004538167004589284, 'std': 3.510983020656342e-05}, annotation_units=1, channel_map={'25645_Weightedmean-L3-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25646_2 = TensorMap('25646_Weightedmean-L3-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.00047361394018460265, 'std': 2.9507552061374636e-05}, annotation_units=1, channel_map={'25646_Weightedmean-L3-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25647_2 = TensorMap('25647_Weightedmean-L3-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0004732174051977517, 'std': 3.0063816019897e-05}, annotation_units=1, channel_map={'25647_Weightedmean-L3-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25648_2 = TensorMap('25648_Weightedmean-L3-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0005339634880111379, 'std': 4.459654112298566e-05}, annotation_units=1, channel_map={'25648_Weightedmean-L3-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25649_2 = TensorMap('25649_Weightedmean-L3-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.000536198376527613, 'std': 3.7947343235996294e-05}, annotation_units=1, channel_map={'25649_Weightedmean-L3-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25650_2 = TensorMap('25650_Weightedmean-ICVF-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.54651073195132, 'std': 0.03059970171415634}, annotation_units=1, channel_map={'25650_Weightedmean-ICVF-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25651_2 = TensorMap('25651_Weightedmean-ICVF-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5451292063737624, 'std': 0.030464022808030327}, annotation_units=1, channel_map={'25651_Weightedmean-ICVF-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25652_2 = TensorMap('25652_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5390192206064357, 'std': 0.032227631353049124}, annotation_units=1, channel_map={'25652_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25653_2 = TensorMap('25653_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5364947394286302, 'std': 0.03202270319318773}, annotation_units=1, channel_map={'25653_Weightedmean-ICVF-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25654_2 = TensorMap('25654_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5682723030631187, 'std': 0.03747946571161812}, annotation_units=1, channel_map={'25654_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25655_2 = TensorMap('25655_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5437084729785479, 'std': 0.03508218503321819}, annotation_units=1, channel_map={'25655_Weightedmean-ICVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25656_2 = TensorMap('25656_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.44831432889851497, 'std': 0.02643646514498657}, annotation_units=1, channel_map={'25656_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25657_2 = TensorMap('25657_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4497087966171617, 'std': 0.027487843347138805}, annotation_units=1, channel_map={'25657_Weightedmean-ICVF-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25658_2 = TensorMap('25658_Weightedmean-ICVF-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6849965903981023, 'std': 0.028728744698191443}, annotation_units=1, channel_map={'25658_Weightedmean-ICVF-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25659_2 = TensorMap('25659_Weightedmean-ICVF-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6773358372524749, 'std': 0.02892090486350828}, annotation_units=1, channel_map={'25659_Weightedmean-ICVF-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25660_2 = TensorMap('25660_Weightedmean-ICVF-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5970407517533004, 'std': 0.03337445895248408}, annotation_units=1, channel_map={'25660_Weightedmean-ICVF-in-tract-forceps-major_2_0': 0, })
ukb_25661_2 = TensorMap('25661_Weightedmean-ICVF-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5447953855713693, 'std': 0.03702417170929035}, annotation_units=1, channel_map={'25661_Weightedmean-ICVF-in-tract-forceps-minor_2_0': 0, })
ukb_25662_2 = TensorMap('25662_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5236321237108086, 'std': 0.03477184539520496}, annotation_units=1, channel_map={'25662_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25663_2 = TensorMap('25663_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.524498696833746, 'std': 0.03516830534922071}, annotation_units=1, channel_map={'25663_Weightedmean-ICVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25664_2 = TensorMap('25664_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5177463779393564, 'std': 0.032688864114933665}, annotation_units=1, channel_map={'25664_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25665_2 = TensorMap('25665_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5166769641604786, 'std': 0.0325459195060416}, annotation_units=1, channel_map={'25665_Weightedmean-ICVF-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25666_2 = TensorMap('25666_Weightedmean-ICVF-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7313375890057756, 'std': 0.03482855863661249}, annotation_units=1, channel_map={'25666_Weightedmean-ICVF-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25667_2 = TensorMap('25667_Weightedmean-ICVF-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6053301196885315, 'std': 0.024060599632510046}, annotation_units=1, channel_map={'25667_Weightedmean-ICVF-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25668_2 = TensorMap('25668_Weightedmean-ICVF-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6027035077867162, 'std': 0.02470023251754849}, annotation_units=1, channel_map={'25668_Weightedmean-ICVF-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25669_2 = TensorMap('25669_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5144741102000825, 'std': 0.032338192192043994}, annotation_units=1, channel_map={'25669_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25670_2 = TensorMap('25670_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5193375863242573, 'std': 0.033265863677969035}, annotation_units=1, channel_map={'25670_Weightedmean-ICVF-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25671_2 = TensorMap('25671_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6197472975453794, 'std': 0.03577729073228715}, annotation_units=1, channel_map={'25671_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25672_2 = TensorMap('25672_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6131296345915841, 'std': 0.03542386304480216}, annotation_units=1, channel_map={'25672_Weightedmean-ICVF-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25673_2 = TensorMap('25673_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6034666844059408, 'std': 0.02861046304700328}, annotation_units=1, channel_map={'25673_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25674_2 = TensorMap('25674_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6060331222153464, 'std': 0.029136010041383047}, annotation_units=1, channel_map={'25674_Weightedmean-ICVF-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25675_2 = TensorMap('25675_Weightedmean-ICVF-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4716619689562706, 'std': 0.029572584166344954}, annotation_units=1, channel_map={'25675_Weightedmean-ICVF-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25676_2 = TensorMap('25676_Weightedmean-ICVF-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4729911404187293, 'std': 0.02836741514362607}, annotation_units=1, channel_map={'25676_Weightedmean-ICVF-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25677_2 = TensorMap('25677_Weightedmean-OD-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.2437835132013202, 'std': 0.017881174157980875}, annotation_units=1, channel_map={'25677_Weightedmean-OD-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25678_2 = TensorMap('25678_Weightedmean-OD-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.25309392032797035, 'std': 0.016645319094354134}, annotation_units=1, channel_map={'25678_Weightedmean-OD-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25679_2 = TensorMap('25679_Weightedmean-OD-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23678127212252476, 'std': 0.012293502789229237}, annotation_units=1, channel_map={'25679_Weightedmean-OD-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25680_2 = TensorMap('25680_Weightedmean-OD-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.2419313592718647, 'std': 0.013075875467276576}, annotation_units=1, channel_map={'25680_Weightedmean-OD-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25681_2 = TensorMap('25681_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.13126137966687296, 'std': 0.017700563753055707}, annotation_units=1, channel_map={'25681_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25682_2 = TensorMap('25682_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14489086099938123, 'std': 0.019756385757917515}, annotation_units=1, channel_map={'25682_Weightedmean-OD-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25683_2 = TensorMap('25683_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.28998730234117165, 'std': 0.028219766215045572}, annotation_units=1, channel_map={'25683_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25684_2 = TensorMap('25684_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.28847783854166675, 'std': 0.029480690112875664}, annotation_units=1, channel_map={'25684_Weightedmean-OD-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25685_2 = TensorMap('25685_Weightedmean-OD-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.15722159096534663, 'std': 0.014721129003917945}, annotation_units=1, channel_map={'25685_Weightedmean-OD-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25686_2 = TensorMap('25686_Weightedmean-OD-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1578228140985974, 'std': 0.014473586284200401}, annotation_units=1, channel_map={'25686_Weightedmean-OD-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25687_2 = TensorMap('25687_Weightedmean-OD-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12814347409240925, 'std': 0.011127165860738775}, annotation_units=1, channel_map={'25687_Weightedmean-OD-in-tract-forceps-major_2_0': 0, })
ukb_25688_2 = TensorMap('25688_Weightedmean-OD-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1960019630259901, 'std': 0.010832566639580554}, annotation_units=1, channel_map={'25688_Weightedmean-OD-in-tract-forceps-minor_2_0': 0, })
ukb_25689_2 = TensorMap('25689_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1644629071266502, 'std': 0.011116376416460749}, annotation_units=1, channel_map={'25689_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25690_2 = TensorMap('25690_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1696632888304455, 'std': 0.010831572197077168}, annotation_units=1, channel_map={'25690_Weightedmean-OD-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25691_2 = TensorMap('25691_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18248786602722772, 'std': 0.009411027861576719}, annotation_units=1, channel_map={'25691_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25692_2 = TensorMap('25692_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18639721312912544, 'std': 0.009150288695076548}, annotation_units=1, channel_map={'25692_Weightedmean-OD-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25693_2 = TensorMap('25693_Weightedmean-OD-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.2318885318688119, 'std': 0.023822949344690116}, annotation_units=1, channel_map={'25693_Weightedmean-OD-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25694_2 = TensorMap('25694_Weightedmean-OD-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18374499355404297, 'std': 0.014213831364610494}, annotation_units=1, channel_map={'25694_Weightedmean-OD-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25695_2 = TensorMap('25695_Weightedmean-OD-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.182905619740099, 'std': 0.014742341728834417}, annotation_units=1, channel_map={'25695_Weightedmean-OD-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25696_2 = TensorMap('25696_Weightedmean-OD-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1836718114170792, 'std': 0.010793449056657937}, annotation_units=1, channel_map={'25696_Weightedmean-OD-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25697_2 = TensorMap('25697_Weightedmean-OD-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.18566617600041252, 'std': 0.011079175596073207}, annotation_units=1, channel_map={'25697_Weightedmean-OD-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25698_2 = TensorMap('25698_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.22283072921823432, 'std': 0.011994605137269253}, annotation_units=1, channel_map={'25698_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25699_2 = TensorMap('25699_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23290199412128712, 'std': 0.012346531929245093}, annotation_units=1, channel_map={'25699_Weightedmean-OD-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25700_2 = TensorMap('25700_Weightedmean-OD-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.2482352115820957, 'std': 0.01618241922923502}, annotation_units=1, channel_map={'25700_Weightedmean-OD-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25701_2 = TensorMap('25701_Weightedmean-OD-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.24719744554455444, 'std': 0.016510454346072935}, annotation_units=1, channel_map={'25701_Weightedmean-OD-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25702_2 = TensorMap('25702_Weightedmean-OD-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23419575871493395, 'std': 0.017232289192297516}, annotation_units=1, channel_map={'25702_Weightedmean-OD-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25703_2 = TensorMap('25703_Weightedmean-OD-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.22851342868193072, 'std': 0.016125523981200336}, annotation_units=1, channel_map={'25703_Weightedmean-OD-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25704_2 = TensorMap('25704_Weightedmean-ISOVF-in-tract-acoustic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06989263308580859, 'std': 0.02144667695421762}, annotation_units=1, channel_map={'25704_Weightedmean-ISOVF-in-tract-acoustic-radiation-left_2_0': 0, })
ukb_25705_2 = TensorMap('25705_Weightedmean-ISOVF-in-tract-acoustic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07246584117161717, 'std': 0.020938999616310353}, annotation_units=1, channel_map={'25705_Weightedmean-ISOVF-in-tract-acoustic-radiation-right_2_0': 0, })
ukb_25706_2 = TensorMap('25706_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06050436574360561, 'std': 0.013900495384338638}, annotation_units=1, channel_map={'25706_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-left_2_0': 0, })
ukb_25707_2 = TensorMap('25707_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05679792092615511, 'std': 0.013222091366710056}, annotation_units=1, channel_map={'25707_Weightedmean-ISOVF-in-tract-anterior-thalamic-radiation-right_2_0': 0, })
ukb_25708_2 = TensorMap('25708_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05215354155528052, 'std': 0.016178321340961553}, annotation_units=1, channel_map={'25708_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-left_2_0': 0, })
ukb_25709_2 = TensorMap('25709_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.04174198411716172, 'std': 0.014321629768299508}, annotation_units=1, channel_map={'25709_Weightedmean-ISOVF-in-tract-cingulate-gyrus-part-of-cingulum-right_2_0': 0, })
ukb_25710_2 = TensorMap('25710_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07808335353805694, 'std': 0.040029414121596434}, annotation_units=1, channel_map={'25710_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-left_2_0': 0, })
ukb_25711_2 = TensorMap('25711_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08611521444255361, 'std': 0.03854093152612957}, annotation_units=1, channel_map={'25711_Weightedmean-ISOVF-in-tract-parahippocampal-part-of-cingulum-right_2_0': 0, })
ukb_25712_2 = TensorMap('25712_Weightedmean-ISOVF-in-tract-corticospinal-tract-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12027915617264852, 'std': 0.014288135361819698}, annotation_units=1, channel_map={'25712_Weightedmean-ISOVF-in-tract-corticospinal-tract-left_2_0': 0, })
ukb_25713_2 = TensorMap('25713_Weightedmean-ISOVF-in-tract-corticospinal-tract-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.11899185840552806, 'std': 0.014435450005608776}, annotation_units=1, channel_map={'25713_Weightedmean-ISOVF-in-tract-corticospinal-tract-right_2_0': 0, })
ukb_25714_2 = TensorMap('25714_Weightedmean-ISOVF-in-tract-forceps-major_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14505657856332504, 'std': 0.02704681105091364}, annotation_units=1, channel_map={'25714_Weightedmean-ISOVF-in-tract-forceps-major_2_0': 0, })
ukb_25715_2 = TensorMap('25715_Weightedmean-ISOVF-in-tract-forceps-minor_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.09095256360354786, 'std': 0.016699541304723963}, annotation_units=1, channel_map={'25715_Weightedmean-ISOVF-in-tract-forceps-minor_2_0': 0, })
ukb_25716_2 = TensorMap('25716_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05875598565903466, 'std': 0.012330135738652225}, annotation_units=1, channel_map={'25716_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-left_2_0': 0, })
ukb_25717_2 = TensorMap('25717_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06641701176773927, 'std': 0.013020546901233712}, annotation_units=1, channel_map={'25717_Weightedmean-ISOVF-in-tract-inferior-frontooccipital-fasciculus-right_2_0': 0, })
ukb_25718_2 = TensorMap('25718_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06426033659240923, 'std': 0.013316064813723797}, annotation_units=1, channel_map={'25718_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25719_2 = TensorMap('25719_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07168946311881186, 'std': 0.013782837394086958}, annotation_units=1, channel_map={'25719_Weightedmean-ISOVF-in-tract-inferior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25720_2 = TensorMap('25720_Weightedmean-ISOVF-in-tract-middle-cerebellar-peduncle_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.13319593108498354, 'std': 0.03528633163098292}, annotation_units=1, channel_map={'25720_Weightedmean-ISOVF-in-tract-middle-cerebellar-peduncle_2_0': 0, })
ukb_25721_2 = TensorMap('25721_Weightedmean-ISOVF-in-tract-medial-lemniscus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.16532780739995873, 'std': 0.027262537769140987}, annotation_units=1, channel_map={'25721_Weightedmean-ISOVF-in-tract-medial-lemniscus-left_2_0': 0, })
ukb_25722_2 = TensorMap('25722_Weightedmean-ISOVF-in-tract-medial-lemniscus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.16382138715449676, 'std': 0.02703791994798627}, annotation_units=1, channel_map={'25722_Weightedmean-ISOVF-in-tract-medial-lemniscus-right_2_0': 0, })
ukb_25723_2 = TensorMap('25723_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08286110106745051, 'std': 0.023214383871807985}, annotation_units=1, channel_map={'25723_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-left_2_0': 0, })
ukb_25724_2 = TensorMap('25724_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08913989245565176, 'std': 0.022187735582556115}, annotation_units=1, channel_map={'25724_Weightedmean-ISOVF-in-tract-posterior-thalamic-radiation-right_2_0': 0, })
ukb_25725_2 = TensorMap('25725_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07619628581373758, 'std': 0.014100588444302212}, annotation_units=1, channel_map={'25725_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-left_2_0': 0, })
ukb_25726_2 = TensorMap('25726_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.08160502266914194, 'std': 0.01518637575144453}, annotation_units=1, channel_map={'25726_Weightedmean-ISOVF-in-tract-superior-longitudinal-fasciculus-right_2_0': 0, })
ukb_25727_2 = TensorMap('25727_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07919584999999998, 'std': 0.013756385079391639}, annotation_units=1, channel_map={'25727_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-left_2_0': 0, })
ukb_25728_2 = TensorMap('25728_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0777009232260726, 'std': 0.014052374530156855}, annotation_units=1, channel_map={'25728_Weightedmean-ISOVF-in-tract-superior-thalamic-radiation-right_2_0': 0, })
ukb_25729_2 = TensorMap('25729_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03390422213232261, 'std': 0.014351557853941625}, annotation_units=1, channel_map={'25729_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-left_2_0': 0, })
ukb_25730_2 = TensorMap('25730_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03321514487726898, 'std': 0.012564433444449567}, annotation_units=1, channel_map={'25730_Weightedmean-ISOVF-in-tract-uncinate-fasciculus-right_2_0': 0, })
ukb_25731_2 = TensorMap('25731_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-linearlyaligned_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6666008612136218, 'std': 0.040536820525704884}, annotation_units=1, channel_map={'25731_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-linearlyaligned_2_0': 0, })
ukb_25732_2 = TensorMap('25732_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-nonlinearlyaligned_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4985629469799598, 'std': 0.032655723571612325}, annotation_units=1, channel_map={'25732_Discrepancy-between-T1-brain-image-and-standardspace-brain-template-nonlinearlyaligned_2_0': 0, })
ukb_25733_2 = TensorMap('25733_Amount-of-warping-applied-to-nonlinearly-align-T1-brain-image-to-standardspace_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.07624585303872565, 'std': 0.021313962845083995}, annotation_units=1, channel_map={'25733_Amount-of-warping-applied-to-nonlinearly-align-T1-brain-image-to-standardspace_2_0': 0, })
ukb_25734_2 = TensorMap('25734_Inverted-signaltonoise-ratio-in-T1_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.015924975830335874, 'std': 0.0016921273142691483}, annotation_units=1, channel_map={'25734_Inverted-signaltonoise-ratio-in-T1_2_0': 0, })
ukb_25735_2 = TensorMap('25735_Inverted-contrasttonoise-ratio-in-T1_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.037151274858691075, 'std': 0.0037563906689053856}, annotation_units=1, channel_map={'25735_Inverted-contrasttonoise-ratio-in-T1_2_0': 0, })
ukb_25736_2 = TensorMap('25736_Discrepancy-between-T2-FLAIR-brain-image-and-T1-brain-image_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.35882551373422433, 'std': 0.05262301216306896}, annotation_units=1, channel_map={'25736_Discrepancy-between-T2-FLAIR-brain-image-and-T1-brain-image_2_0': 0, })
ukb_25737_2 = TensorMap('25737_Discrepancy-between-dMRI-brain-image-and-T1-brain-image_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5935836941991169, 'std': 0.07582857298692376}, annotation_units=1, channel_map={'25737_Discrepancy-between-dMRI-brain-image-and-T1-brain-image_2_0': 0, })
ukb_25738_2 = TensorMap('25738_Discrepancy-between-SWI-brain-image-and-T1-brain-image_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4144582681228742, 'std': 0.051931737564547595}, annotation_units=1, channel_map={'25738_Discrepancy-between-SWI-brain-image-and-T1-brain-image_2_0': 0, })
ukb_25739_2 = TensorMap('25739_Discrepancy-between-rfMRI-brain-image-and-T1-brain-image_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7444824269176566, 'std': 0.07223138802423033}, annotation_units=1, channel_map={'25739_Discrepancy-between-rfMRI-brain-image-and-T1-brain-image_2_0': 0, })
ukb_25740_2 = TensorMap('25740_Discrepancy-between-tfMRI-brain-image-and-T1-brain-image_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7560109574274394, 'std': 0.06999287565291917}, annotation_units=1, channel_map={'25740_Discrepancy-between-tfMRI-brain-image-and-T1-brain-image_2_0': 0, })
ukb_25741_2 = TensorMap('25741_Mean-rfMRI-head-motion-averaged-across-space-and-time-points_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12143296971008656, 'std': 0.058762274505188734}, annotation_units=1, channel_map={'25741_Mean-rfMRI-head-motion-averaged-across-space-and-time-points_2_0': 0, })
ukb_25742_2 = TensorMap('25742_Mean-tfMRI-head-motion-averaged-across-space-and-time-points_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.14410601563671419, 'std': 0.06547364749003134}, annotation_units=1, channel_map={'25742_Mean-tfMRI-head-motion-averaged-across-space-and-time-points_2_0': 0, })
ukb_25743_2 = TensorMap('25743_Inverted-temporal-signaltonoise-ratio-in-preprocessed-rfMRI_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.028483716498892697, 'std': 0.003994642133767273}, annotation_units=1, channel_map={'25743_Inverted-temporal-signaltonoise-ratio-in-preprocessed-rfMRI_2_0': 0, })
ukb_25744_2 = TensorMap('25744_Inverted-temporal-signaltonoise-ratio-in-artefactcleaned-preprocessed-rfMRI_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.020804039695962957, 'std': 0.0019625862852633356}, annotation_units=1, channel_map={'25744_Inverted-temporal-signaltonoise-ratio-in-artefactcleaned-preprocessed-rfMRI_2_0': 0, })
ukb_25745_2 = TensorMap('25745_Inverted-temporal-signaltonoise-ratio-in-preprocessed-tfMRI_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.008281219639605868, 'std': 0.0018873877498348313}, annotation_units=1, channel_map={'25745_Inverted-temporal-signaltonoise-ratio-in-preprocessed-tfMRI_2_0': 0, })
ukb_25746_2 = TensorMap('25746_Number-of-dMRI-outlier-slices-detected-and-corrected_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.905520334003405, 'std': 35.00262421588656}, annotation_units=1, channel_map={'25746_Number-of-dMRI-outlier-slices-detected-and-corrected_2_0': 0, })
ukb_25761_2 = TensorMap('25761_90th-percentile-of-BOLD-effect-in-groupdefined-mask-for-shapes-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9181375448748863, 'std': 0.6615635143049833}, annotation_units=1, channel_map={'25761_90th-percentile-of-BOLD-effect-in-groupdefined-mask-for-shapes-activation_2_0': 0, })
ukb_25762_2 = TensorMap('25762_90th-percentile-of-zstatistic-in-groupdefined-mask-for-shapes-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.738075962063978, 'std': 2.1012646075589143}, annotation_units=1, channel_map={'25762_90th-percentile-of-zstatistic-in-groupdefined-mask-for-shapes-activation_2_0': 0, })
ukb_25763_2 = TensorMap('25763_90th-percentile-of-BOLD-effect-in-groupdefined-mask-for-faces-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.251493811284359, 'std': 0.7136883033758361}, annotation_units=1, channel_map={'25763_90th-percentile-of-BOLD-effect-in-groupdefined-mask-for-faces-activation_2_0': 0, })
ukb_25764_2 = TensorMap('25764_90th-percentile-of-zstatistic-in-groupdefined-mask-for-faces-activation_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.32185568043723, 'std': 1.815524234696942}, annotation_units=1, channel_map={'25764_90th-percentile-of-zstatistic-in-groupdefined-mask-for-faces-activation_2_0': 0, })
ukb_25765_2 = TensorMap('25765_90th-percentile-of-BOLD-effect-in-groupdefined-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5004619300219684, 'std': 0.5607068949933574}, annotation_units=1, channel_map={'25765_90th-percentile-of-BOLD-effect-in-groupdefined-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25766_2 = TensorMap('25766_90th-percentile-of-zstatistic-in-groupdefined-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.139088458929431, 'std': 1.940085730017117}, annotation_units=1, channel_map={'25766_90th-percentile-of-zstatistic-in-groupdefined-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25767_2 = TensorMap('25767_90th-percentile-of-BOLD-effect-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3962289021057707, 'std': 0.1794998123231883}, annotation_units=1, channel_map={'25767_90th-percentile-of-BOLD-effect-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25768_2 = TensorMap('25768_90th-percentile-of-zstatistic-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.9786837287681514, 'std': 1.0393824363796085}, annotation_units=1, channel_map={'25768_90th-percentile-of-zstatistic-in-groupdefined-amygdala-activation-mask-for-facesshapes-contrast_2_0': 0, })
ukb_25781_2 = TensorMap('25781_Total-volume-of-white-matter-hyperintensities-from-T1-and-T2FLAIR-images_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4607.239653465345, 'std': 6021.463014232158}, annotation_units=1, channel_map={'25781_Total-volume-of-white-matter-hyperintensities-from-T1-and-T2FLAIR-images_2_0': 0, })
ukb_25782_2 = TensorMap('25782_Volume-of-grey-matter-in-Frontal-Pole-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 23225.041804504257, 'std': 2753.8979558200635}, annotation_units=1, channel_map={'25782_Volume-of-grey-matter-in-Frontal-Pole-left_2_0': 0, })
ukb_25783_2 = TensorMap('25783_Volume-of-grey-matter-in-Frontal-Pole-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26218.0797635735, 'std': 3064.796268681824}, annotation_units=1, channel_map={'25783_Volume-of-grey-matter-in-Frontal-Pole-right_2_0': 0, })
ukb_25784_2 = TensorMap('25784_Volume-of-grey-matter-in-Insular-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6288.064134193065, 'std': 677.648355952458}, annotation_units=1, channel_map={'25784_Volume-of-grey-matter-in-Insular-Cortex-left_2_0': 0, })
ukb_25785_2 = TensorMap('25785_Volume-of-grey-matter-in-Insular-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6274.860182693205, 'std': 670.6589038394169}, annotation_units=1, channel_map={'25785_Volume-of-grey-matter-in-Insular-Cortex-right_2_0': 0, })
ukb_25786_2 = TensorMap('25786_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 11045.975897579669, 'std': 1765.6306568117689}, annotation_units=1, channel_map={'25786_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-left_2_0': 0, })
ukb_25787_2 = TensorMap('25787_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9599.685957854406, 'std': 1611.420311459197}, annotation_units=1, channel_map={'25787_Volume-of-grey-matter-in-Superior-Frontal-Gyrus-right_2_0': 0, })
ukb_25788_2 = TensorMap('25788_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10076.390610690589, 'std': 1785.313563358117}, annotation_units=1, channel_map={'25788_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-left_2_0': 0, })
ukb_25789_2 = TensorMap('25789_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9553.201795159332, 'std': 1707.8596622948214}, annotation_units=1, channel_map={'25789_Volume-of-grey-matter-in-Middle-Frontal-Gyrus-right_2_0': 0, })
ukb_25790_2 = TensorMap('25790_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2444.873028128213, 'std': 582.6848603985181}, annotation_units=1, channel_map={'25790_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-left_2_0': 0, })
ukb_25791_2 = TensorMap('25791_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2258.9134368750583, 'std': 503.5983192507444}, annotation_units=1, channel_map={'25791_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-triangularis-right_2_0': 0, })
ukb_25792_2 = TensorMap('25792_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2652.1187262872627, 'std': 537.7627915252457}, annotation_units=1, channel_map={'25792_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-left_2_0': 0, })
ukb_25793_2 = TensorMap('25793_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2473.6742645547147, 'std': 528.1760781965238}, annotation_units=1, channel_map={'25793_Volume-of-grey-matter-in-Inferior-Frontal-Gyrus-pars-opercularis-right_2_0': 0, })
ukb_25794_2 = TensorMap('25794_Volume-of-grey-matter-in-Precentral-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14054.645076161107, 'std': 1759.9566227429991}, annotation_units=1, channel_map={'25794_Volume-of-grey-matter-in-Precentral-Gyrus-left_2_0': 0, })
ukb_25795_2 = TensorMap('25795_Volume-of-grey-matter-in-Precentral-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13699.323417904863, 'std': 1731.154455220328}, annotation_units=1, channel_map={'25795_Volume-of-grey-matter-in-Precentral-Gyrus-right_2_0': 0, })
ukb_25796_2 = TensorMap('25796_Volume-of-grey-matter-in-Temporal-Pole-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9514.904122511916, 'std': 1279.4968573969265}, annotation_units=1, channel_map={'25796_Volume-of-grey-matter-in-Temporal-Pole-left_2_0': 0, })
ukb_25797_2 = TensorMap('25797_Volume-of-grey-matter-in-Temporal-Pole-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9468.977994579947, 'std': 1237.9691398183397}, annotation_units=1, channel_map={'25797_Volume-of-grey-matter-in-Temporal-Pole-right_2_0': 0, })
ukb_25798_2 = TensorMap('25798_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1469.7367589010378, 'std': 299.5580750133219}, annotation_units=1, channel_map={'25798_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-left_2_0': 0, })
ukb_25799_2 = TensorMap('25799_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1472.4572772638073, 'std': 302.68858013035117}, annotation_units=1, channel_map={'25799_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-anterior-division-right_2_0': 0, })
ukb_25800_2 = TensorMap('25800_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2590.4702915615367, 'std': 490.97419345459474}, annotation_units=1, channel_map={'25800_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-left_2_0': 0, })
ukb_25801_2 = TensorMap('25801_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3122.437807214278, 'std': 541.636010542984}, annotation_units=1, channel_map={'25801_Volume-of-grey-matter-in-Superior-Temporal-Gyrus-posterior-division-right_2_0': 0, })
ukb_25802_2 = TensorMap('25802_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1835.7073999158956, 'std': 394.35919072750625}, annotation_units=1, channel_map={'25802_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-left_2_0': 0, })
ukb_25803_2 = TensorMap('25803_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1620.6961884403324, 'std': 317.51218863047967}, annotation_units=1, channel_map={'25803_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-anterior-division-right_2_0': 0, })
ukb_25804_2 = TensorMap('25804_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5327.056167180638, 'std': 880.6620937812055}, annotation_units=1, channel_map={'25804_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-left_2_0': 0, })
ukb_25805_2 = TensorMap('25805_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5512.1732791327895, 'std': 867.7281870828903}, annotation_units=1, channel_map={'25805_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-posterior-division-right_2_0': 0, })
ukb_25806_2 = TensorMap('25806_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3493.3138496869446, 'std': 768.3271026975603}, annotation_units=1, channel_map={'25806_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-left_2_0': 0, })
ukb_25807_2 = TensorMap('25807_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4719.370520979346, 'std': 947.7155731059905}, annotation_units=1, channel_map={'25807_Volume-of-grey-matter-in-Middle-Temporal-Gyrus-temporooccipital-part-right_2_0': 0, })
ukb_25808_2 = TensorMap('25808_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1443.3246640500888, 'std': 314.8178131490452}, annotation_units=1, channel_map={'25808_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-left_2_0': 0, })
ukb_25809_2 = TensorMap('25809_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1506.02491355948, 'std': 286.18526151338153}, annotation_units=1, channel_map={'25809_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-anterior-division-right_2_0': 0, })
ukb_25810_2 = TensorMap('25810_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4149.731123306234, 'std': 785.8889518402577}, annotation_units=1, channel_map={'25810_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-left_2_0': 0, })
ukb_25811_2 = TensorMap('25811_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4196.051064386505, 'std': 755.2030890770037}, annotation_units=1, channel_map={'25811_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-posterior-division-right_2_0': 0, })
ukb_25812_2 = TensorMap('25812_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3022.581731520419, 'std': 606.4751646522122}, annotation_units=1, channel_map={'25812_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-left_2_0': 0, })
ukb_25813_2 = TensorMap('25813_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3860.4732431548464, 'std': 761.9734202156617}, annotation_units=1, channel_map={'25813_Volume-of-grey-matter-in-Inferior-Temporal-Gyrus-temporooccipital-part-right_2_0': 0, })
ukb_25814_2 = TensorMap('25814_Volume-of-grey-matter-in-Postcentral-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 11301.11938089898, 'std': 1494.7211200412921}, annotation_units=1, channel_map={'25814_Volume-of-grey-matter-in-Postcentral-Gyrus-left_2_0': 0, })
ukb_25815_2 = TensorMap('25815_Volume-of-grey-matter-in-Postcentral-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10505.791742827772, 'std': 1443.3476300289256}, annotation_units=1, channel_map={'25815_Volume-of-grey-matter-in-Postcentral-Gyrus-right_2_0': 0, })
ukb_25816_2 = TensorMap('25816_Volume-of-grey-matter-in-Superior-Parietal-Lobule-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5136.128181945612, 'std': 933.2605965877225}, annotation_units=1, channel_map={'25816_Volume-of-grey-matter-in-Superior-Parietal-Lobule-left_2_0': 0, })
ukb_25817_2 = TensorMap('25817_Volume-of-grey-matter-in-Superior-Parietal-Lobule-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4648.10990655079, 'std': 893.4764209845091}, annotation_units=1, channel_map={'25817_Volume-of-grey-matter-in-Superior-Parietal-Lobule-right_2_0': 0, })
ukb_25818_2 = TensorMap('25818_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3200.500196243342, 'std': 655.2593836652613}, annotation_units=1, channel_map={'25818_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-left_2_0': 0, })
ukb_25819_2 = TensorMap('25819_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3140.376449397252, 'std': 615.0541842560269}, annotation_units=1, channel_map={'25819_Volume-of-grey-matter-in-Supramarginal-Gyrus-anterior-division-right_2_0': 0, })
ukb_25820_2 = TensorMap('25820_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4646.28251284927, 'std': 854.7823805088657}, annotation_units=1, channel_map={'25820_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-left_2_0': 0, })
ukb_25821_2 = TensorMap('25821_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5471.84393748248, 'std': 1036.486623209911}, annotation_units=1, channel_map={'25821_Volume-of-grey-matter-in-Supramarginal-Gyrus-posterior-division-right_2_0': 0, })
ukb_25822_2 = TensorMap('25822_Volume-of-grey-matter-in-Angular-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4076.5691897953457, 'std': 779.5040916942896}, annotation_units=1, channel_map={'25822_Volume-of-grey-matter-in-Angular-Gyrus-left_2_0': 0, })
ukb_25823_2 = TensorMap('25823_Volume-of-grey-matter-in-Angular-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5520.821724605176, 'std': 1122.6806823477625}, annotation_units=1, channel_map={'25823_Volume-of-grey-matter-in-Angular-Gyrus-right_2_0': 0, })
ukb_25824_2 = TensorMap('25824_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16333.799209419678, 'std': 2273.6126950527137}, annotation_units=1, channel_map={'25824_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-left_2_0': 0, })
ukb_25825_2 = TensorMap('25825_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15647.313951032615, 'std': 2315.315797665822}, annotation_units=1, channel_map={'25825_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-superior-division-right_2_0': 0, })
ukb_25826_2 = TensorMap('25826_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7229.486316232125, 'std': 1191.482155793171}, annotation_units=1, channel_map={'25826_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-left_2_0': 0, })
ukb_25827_2 = TensorMap('25827_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7749.844171105505, 'std': 1346.4270413331033}, annotation_units=1, channel_map={'25827_Volume-of-grey-matter-in-Lateral-Occipital-Cortex-inferior-division-right_2_0': 0, })
ukb_25828_2 = TensorMap('25828_Volume-of-grey-matter-in-Intracalcarine-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2614.5531739557046, 'std': 620.6734402941912}, annotation_units=1, channel_map={'25828_Volume-of-grey-matter-in-Intracalcarine-Cortex-left_2_0': 0, })
ukb_25829_2 = TensorMap('25829_Volume-of-grey-matter-in-Intracalcarine-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2723.6597383422113, 'std': 593.7992612335527}, annotation_units=1, channel_map={'25829_Volume-of-grey-matter-in-Intracalcarine-Cortex-right_2_0': 0, })
ukb_25830_2 = TensorMap('25830_Volume-of-grey-matter-in-Frontal-Medial-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1894.4696429305668, 'std': 325.2490811410566}, annotation_units=1, channel_map={'25830_Volume-of-grey-matter-in-Frontal-Medial-Cortex-left_2_0': 0, })
ukb_25831_2 = TensorMap('25831_Volume-of-grey-matter-in-Frontal-Medial-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1894.0895491075598, 'std': 327.382922199766}, annotation_units=1, channel_map={'25831_Volume-of-grey-matter-in-Frontal-Medial-Cortex-right_2_0': 0, })
ukb_25832_2 = TensorMap('25832_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2872.150175684515, 'std': 565.1398169495145}, annotation_units=1, channel_map={'25832_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-left_2_0': 0, })
ukb_25833_2 = TensorMap('25833_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2779.071966638632, 'std': 564.1756767853253}, annotation_units=1, channel_map={'25833_Volume-of-grey-matter-in-Juxtapositional-Lobule-Cortex-formerly-Supplementary-Motor-Cortex-right_2_0': 0, })
ukb_25834_2 = TensorMap('25834_Volume-of-grey-matter-in-Subcallosal-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2939.409000560695, 'std': 420.4083615608552}, annotation_units=1, channel_map={'25834_Volume-of-grey-matter-in-Subcallosal-Cortex-left_2_0': 0, })
ukb_25835_2 = TensorMap('25835_Volume-of-grey-matter-in-Subcallosal-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2722.8408022614713, 'std': 374.1705113258144}, annotation_units=1, channel_map={'25835_Volume-of-grey-matter-in-Subcallosal-Cortex-right_2_0': 0, })
ukb_25836_2 = TensorMap('25836_Volume-of-grey-matter-in-Paracingulate-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5744.066317633867, 'std': 857.6799287324944}, annotation_units=1, channel_map={'25836_Volume-of-grey-matter-in-Paracingulate-Gyrus-left_2_0': 0, })
ukb_25837_2 = TensorMap('25837_Volume-of-grey-matter-in-Paracingulate-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5693.302755349966, 'std': 853.7781865706716}, annotation_units=1, channel_map={'25837_Volume-of-grey-matter-in-Paracingulate-Gyrus-right_2_0': 0, })
ukb_25838_2 = TensorMap('25838_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5099.016205962063, 'std': 1056.5097412289504}, annotation_units=1, channel_map={'25838_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-left_2_0': 0, })
ukb_25839_2 = TensorMap('25839_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5630.916994206151, 'std': 1170.3462461012346}, annotation_units=1, channel_map={'25839_Volume-of-grey-matter-in-Cingulate-Gyrus-anterior-division-right_2_0': 0, })
ukb_25840_2 = TensorMap('25840_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5209.213884216429, 'std': 772.4283488705786}, annotation_units=1, channel_map={'25840_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-left_2_0': 0, })
ukb_25841_2 = TensorMap('25841_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5449.975124754694, 'std': 788.2641027113689}, annotation_units=1, channel_map={'25841_Volume-of-grey-matter-in-Cingulate-Gyrus-posterior-division-right_2_0': 0, })
ukb_25842_2 = TensorMap('25842_Volume-of-grey-matter-in-Precuneous-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10114.767209139334, 'std': 1396.5350066856874}, annotation_units=1, channel_map={'25842_Volume-of-grey-matter-in-Precuneous-Cortex-left_2_0': 0, })
ukb_25843_2 = TensorMap('25843_Volume-of-grey-matter-in-Precuneous-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10564.796545649939, 'std': 1492.668580414912}, annotation_units=1, channel_map={'25843_Volume-of-grey-matter-in-Precuneous-Cortex-right_2_0': 0, })
ukb_25844_2 = TensorMap('25844_Volume-of-grey-matter-in-Cuneal-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1959.3610479394451, 'std': 399.73012577905416}, annotation_units=1, channel_map={'25844_Volume-of-grey-matter-in-Cuneal-Cortex-left_2_0': 0, })
ukb_25845_2 = TensorMap('25845_Volume-of-grey-matter-in-Cuneal-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2330.375844734137, 'std': 474.1509357994705}, annotation_units=1, channel_map={'25845_Volume-of-grey-matter-in-Cuneal-Cortex-right_2_0': 0, })
ukb_25846_2 = TensorMap('25846_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6625.776614802356, 'std': 844.3620811179205}, annotation_units=1, channel_map={'25846_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-left_2_0': 0, })
ukb_25847_2 = TensorMap('25847_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6028.00409354266, 'std': 757.270810241121}, annotation_units=1, channel_map={'25847_Volume-of-grey-matter-in-Frontal-Orbital-Cortex-right_2_0': 0, })
ukb_25848_2 = TensorMap('25848_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2940.9474212690398, 'std': 433.05449059487927}, annotation_units=1, channel_map={'25848_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-left_2_0': 0, })
ukb_25849_2 = TensorMap('25849_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3005.041942809084, 'std': 465.3212564151695}, annotation_units=1, channel_map={'25849_Volume-of-grey-matter-in-Parahippocampal-Gyrus-anterior-division-right_2_0': 0, })
ukb_25850_2 = TensorMap('25850_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1703.131791841884, 'std': 215.40699173572003}, annotation_units=1, channel_map={'25850_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-left_2_0': 0, })
ukb_25851_2 = TensorMap('25851_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1380.2256962433416, 'std': 178.54047858468394}, annotation_units=1, channel_map={'25851_Volume-of-grey-matter-in-Parahippocampal-Gyrus-posterior-division-right_2_0': 0, })
ukb_25852_2 = TensorMap('25852_Volume-of-grey-matter-in-Lingual-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6525.733035230354, 'std': 895.3386109206342}, annotation_units=1, channel_map={'25852_Volume-of-grey-matter-in-Lingual-Gyrus-left_2_0': 0, })
ukb_25853_2 = TensorMap('25853_Volume-of-grey-matter-in-Lingual-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6984.320880291562, 'std': 932.9258304683041}, annotation_units=1, channel_map={'25853_Volume-of-grey-matter-in-Lingual-Gyrus-right_2_0': 0, })
ukb_25854_2 = TensorMap('25854_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1508.383016727409, 'std': 252.6707183069591}, annotation_units=1, channel_map={'25854_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-left_2_0': 0, })
ukb_25855_2 = TensorMap('25855_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1377.3071166713394, 'std': 233.17490502023122}, annotation_units=1, channel_map={'25855_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-anterior-division-right_2_0': 0, })
ukb_25856_2 = TensorMap('25856_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3971.642253527708, 'std': 569.3631329744059}, annotation_units=1, channel_map={'25856_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-left_2_0': 0, })
ukb_25857_2 = TensorMap('25857_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3247.0995093916463, 'std': 474.744492504012}, annotation_units=1, channel_map={'25857_Volume-of-grey-matter-in-Temporal-Fusiform-Cortex-posterior-division-right_2_0': 0, })
ukb_25858_2 = TensorMap('25858_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2625.1988874871513, 'std': 474.4825587570503}, annotation_units=1, channel_map={'25858_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-left_2_0': 0, })
ukb_25859_2 = TensorMap('25859_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3349.117181571816, 'std': 597.642106359717}, annotation_units=1, channel_map={'25859_Volume-of-grey-matter-in-Temporal-Occipital-Fusiform-Cortex-right_2_0': 0, })
ukb_25860_2 = TensorMap('25860_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3921.8338645920935, 'std': 612.6792020037699}, annotation_units=1, channel_map={'25860_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-left_2_0': 0, })
ukb_25861_2 = TensorMap('25861_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3723.381976918045, 'std': 602.6915576638589}, annotation_units=1, channel_map={'25861_Volume-of-grey-matter-in-Occipital-Fusiform-Gyrus-right_2_0': 0, })
ukb_25862_2 = TensorMap('25862_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1507.7120705541536, 'std': 258.0684943095674}, annotation_units=1, channel_map={'25862_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-left_2_0': 0, })
ukb_25863_2 = TensorMap('25863_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1340.1218730959731, 'std': 245.8491144479392}, annotation_units=1, channel_map={'25863_Volume-of-grey-matter-in-Frontal-Operculum-Cortex-right_2_0': 0, })
ukb_25864_2 = TensorMap('25864_Volume-of-grey-matter-in-Central-Opercular-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3773.6542000747595, 'std': 585.9454468182374}, annotation_units=1, channel_map={'25864_Volume-of-grey-matter-in-Central-Opercular-Cortex-left_2_0': 0, })
ukb_25865_2 = TensorMap('25865_Volume-of-grey-matter-in-Central-Opercular-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3737.121397065695, 'std': 606.7553898394947}, annotation_units=1, channel_map={'25865_Volume-of-grey-matter-in-Central-Opercular-Cortex-right_2_0': 0, })
ukb_25866_2 = TensorMap('25866_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2061.2263723483793, 'std': 458.9473351666637}, annotation_units=1, channel_map={'25866_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-left_2_0': 0, })
ukb_25867_2 = TensorMap('25867_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2041.379370993365, 'std': 430.05750683471405}, annotation_units=1, channel_map={'25867_Volume-of-grey-matter-in-Parietal-Operculum-Cortex-right_2_0': 0, })
ukb_25868_2 = TensorMap('25868_Volume-of-grey-matter-in-Planum-Polare-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1379.8110634987381, 'std': 216.32007715145397}, annotation_units=1, channel_map={'25868_Volume-of-grey-matter-in-Planum-Polare-left_2_0': 0, })
ukb_25869_2 = TensorMap('25869_Volume-of-grey-matter-in-Planum-Polare-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1458.80124198673, 'std': 225.29005946405636}, annotation_units=1, channel_map={'25869_Volume-of-grey-matter-in-Planum-Polare-right_2_0': 0, })
ukb_25870_2 = TensorMap('25870_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1191.603291982058, 'std': 270.185974740721}, annotation_units=1, channel_map={'25870_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-left_2_0': 0, })
ukb_25871_2 = TensorMap('25871_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1024.7466579758902, 'std': 205.47980917758048}, annotation_units=1, channel_map={'25871_Volume-of-grey-matter-in-Heschls-Gyrus-includes-H1-and-H2-right_2_0': 0, })
ukb_25872_2 = TensorMap('25872_Volume-of-grey-matter-in-Planum-Temporale-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2050.569680170078, 'std': 481.8515174136458}, annotation_units=1, channel_map={'25872_Volume-of-grey-matter-in-Planum-Temporale-left_2_0': 0, })
ukb_25873_2 = TensorMap('25873_Volume-of-grey-matter-in-Planum-Temporale-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1605.6542190916743, 'std': 318.1157815993862}, annotation_units=1, channel_map={'25873_Volume-of-grey-matter-in-Planum-Temporale-right_2_0': 0, })
ukb_25874_2 = TensorMap('25874_Volume-of-grey-matter-in-Supracalcarine-Cortex-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 506.6534898140361, 'std': 117.90758677687415}, annotation_units=1, channel_map={'25874_Volume-of-grey-matter-in-Supracalcarine-Cortex-left_2_0': 0, })
ukb_25875_2 = TensorMap('25875_Volume-of-grey-matter-in-Supracalcarine-Cortex-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 722.9406312026915, 'std': 163.1895588584789}, annotation_units=1, channel_map={'25875_Volume-of-grey-matter-in-Supracalcarine-Cortex-right_2_0': 0, })
ukb_25876_2 = TensorMap('25876_Volume-of-grey-matter-in-Occipital-Pole-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8533.018132884781, 'std': 1405.4599882766547}, annotation_units=1, channel_map={'25876_Volume-of-grey-matter-in-Occipital-Pole-left_2_0': 0, })
ukb_25877_2 = TensorMap('25877_Volume-of-grey-matter-in-Occipital-Pole-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8171.7009144005215, 'std': 1370.8419490177882}, annotation_units=1, channel_map={'25877_Volume-of-grey-matter-in-Occipital-Pole-right_2_0': 0, })
ukb_25878_2 = TensorMap('25878_Volume-of-grey-matter-in-Thalamus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2693.7454074385564, 'std': 287.4017812166893}, annotation_units=1, channel_map={'25878_Volume-of-grey-matter-in-Thalamus-left_2_0': 0, })
ukb_25879_2 = TensorMap('25879_Volume-of-grey-matter-in-Thalamus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2820.78573731427, 'std': 306.8179661761049}, annotation_units=1, channel_map={'25879_Volume-of-grey-matter-in-Thalamus-right_2_0': 0, })
ukb_25880_2 = TensorMap('25880_Volume-of-grey-matter-in-Caudate-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3014.074642089524, 'std': 663.5538379222165}, annotation_units=1, channel_map={'25880_Volume-of-grey-matter-in-Caudate-left_2_0': 0, })
ukb_25881_2 = TensorMap('25881_Volume-of-grey-matter-in-Caudate-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3216.403425380805, 'std': 704.606279153414}, annotation_units=1, channel_map={'25881_Volume-of-grey-matter-in-Caudate-right_2_0': 0, })
ukb_25882_2 = TensorMap('25882_Volume-of-grey-matter-in-Putamen-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1731.9727092841795, 'std': 408.866478231961}, annotation_units=1, channel_map={'25882_Volume-of-grey-matter-in-Putamen-left_2_0': 0, })
ukb_25883_2 = TensorMap('25883_Volume-of-grey-matter-in-Putamen-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2136.8382952527795, 'std': 444.10063193353494}, annotation_units=1, channel_map={'25883_Volume-of-grey-matter-in-Putamen-right_2_0': 0, })
ukb_25884_2 = TensorMap('25884_Volume-of-grey-matter-in-Pallidum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.401947968040375, 'std': 38.09451437617279}, annotation_units=1, channel_map={'25884_Volume-of-grey-matter-in-Pallidum-left_2_0': 0, })
ukb_25885_2 = TensorMap('25885_Volume-of-grey-matter-in-Pallidum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.488568621764315, 'std': 42.158160143257504}, annotation_units=1, channel_map={'25885_Volume-of-grey-matter-in-Pallidum-right_2_0': 0, })
ukb_25886_2 = TensorMap('25886_Volume-of-grey-matter-in-Hippocampus-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4204.271571348473, 'std': 417.6015690795573}, annotation_units=1, channel_map={'25886_Volume-of-grey-matter-in-Hippocampus-left_2_0': 0, })
ukb_25887_2 = TensorMap('25887_Volume-of-grey-matter-in-Hippocampus-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4366.076320904589, 'std': 431.26127725500584}, annotation_units=1, channel_map={'25887_Volume-of-grey-matter-in-Hippocampus-right_2_0': 0, })
ukb_25888_2 = TensorMap('25888_Volume-of-grey-matter-in-Amygdala-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1818.5390607419863, 'std': 237.00928563328193}, annotation_units=1, channel_map={'25888_Volume-of-grey-matter-in-Amygdala-left_2_0': 0, })
ukb_25889_2 = TensorMap('25889_Volume-of-grey-matter-in-Amygdala-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2088.9461120456026, 'std': 265.07624313378403}, annotation_units=1, channel_map={'25889_Volume-of-grey-matter-in-Amygdala-right_2_0': 0, })
ukb_25890_2 = TensorMap('25890_Volume-of-grey-matter-in-Ventral-Striatum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 555.302782123166, 'std': 115.51926131258652}, annotation_units=1, channel_map={'25890_Volume-of-grey-matter-in-Ventral-Striatum-left_2_0': 0, })
ukb_25891_2 = TensorMap('25891_Volume-of-grey-matter-in-Ventral-Striatum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 578.8000690122419, 'std': 114.98938074126018}, annotation_units=1, channel_map={'25891_Volume-of-grey-matter-in-Ventral-Striatum-right_2_0': 0, })
ukb_25892_2 = TensorMap('25892_Volume-of-grey-matter-in-BrainStem_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4926.986440052331, 'std': 832.053987333486}, annotation_units=1, channel_map={'25892_Volume-of-grey-matter-in-BrainStem_2_0': 0, })
ukb_25893_2 = TensorMap('25893_Volume-of-grey-matter-in-IIV-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1850.589530090646, 'std': 284.80683606022586}, annotation_units=1, channel_map={'25893_Volume-of-grey-matter-in-IIV-Cerebellum-left_2_0': 0, })
ukb_25894_2 = TensorMap('25894_Volume-of-grey-matter-in-IIV-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2101.2206088216053, 'std': 308.74274187363477}, annotation_units=1, channel_map={'25894_Volume-of-grey-matter-in-IIV-Cerebellum-right_2_0': 0, })
ukb_25895_2 = TensorMap('25895_Volume-of-grey-matter-in-V-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2608.5774137931026, 'std': 383.50451435356257}, annotation_units=1, channel_map={'25895_Volume-of-grey-matter-in-V-Cerebellum-left_2_0': 0, })
ukb_25896_2 = TensorMap('25896_Volume-of-grey-matter-in-V-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2601.2581613867865, 'std': 379.62264934402236}, annotation_units=1, channel_map={'25896_Volume-of-grey-matter-in-V-Cerebellum-right_2_0': 0, })
ukb_25897_2 = TensorMap('25897_Volume-of-grey-matter-in-VI-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6731.5115461171845, 'std': 1082.2020656614404}, annotation_units=1, channel_map={'25897_Volume-of-grey-matter-in-VI-Cerebellum-left_2_0': 0, })
ukb_25898_2 = TensorMap('25898_Volume-of-grey-matter-in-Vermis-VI-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1504.15352537146, 'std': 237.2285802963334}, annotation_units=1, channel_map={'25898_Volume-of-grey-matter-in-Vermis-VI-Cerebellum_2_0': 0, })
ukb_25899_2 = TensorMap('25899_Volume-of-grey-matter-in-VI-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6583.816492851133, 'std': 1051.657457096091}, annotation_units=1, channel_map={'25899_Volume-of-grey-matter-in-VI-Cerebellum-right_2_0': 0, })
ukb_25900_2 = TensorMap('25900_Volume-of-grey-matter-in-Crus-I-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 10599.137187178767, 'std': 1586.9682994994514}, annotation_units=1, channel_map={'25900_Volume-of-grey-matter-in-Crus-I-Cerebellum-left_2_0': 0, })
ukb_25901_2 = TensorMap('25901_Volume-of-grey-matter-in-Vermis-Crus-I-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.0349224657508636, 'std': 1.5326181063892008}, annotation_units=1, channel_map={'25901_Volume-of-grey-matter-in-Vermis-Crus-I-Cerebellum_2_0': 0, })
ukb_25902_2 = TensorMap('25902_Volume-of-grey-matter-in-Crus-I-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 11457.855776562934, 'std': 1764.2765635932872}, annotation_units=1, channel_map={'25902_Volume-of-grey-matter-in-Crus-I-Cerebellum-right_2_0': 0, })
ukb_25903_2 = TensorMap('25903_Volume-of-grey-matter-in-Crus-II-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7942.692540416781, 'std': 1196.1534084850525}, annotation_units=1, channel_map={'25903_Volume-of-grey-matter-in-Crus-II-Cerebellum-left_2_0': 0, })
ukb_25904_2 = TensorMap('25904_Volume-of-grey-matter-in-Vermis-Crus-II-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 398.76648350621446, 'std': 75.35994493238857}, annotation_units=1, channel_map={'25904_Volume-of-grey-matter-in-Vermis-Crus-II-Cerebellum_2_0': 0, })
ukb_25905_2 = TensorMap('25905_Volume-of-grey-matter-in-Crus-II-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7764.2724950939155, 'std': 1165.9146982296538}, annotation_units=1, channel_map={'25905_Volume-of-grey-matter-in-Crus-II-Cerebellum-right_2_0': 0, })
ukb_25906_2 = TensorMap('25906_Volume-of-grey-matter-in-VIIb-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3825.7000836370426, 'std': 672.5853069635776}, annotation_units=1, channel_map={'25906_Volume-of-grey-matter-in-VIIb-Cerebellum-left_2_0': 0, })
ukb_25907_2 = TensorMap('25907_Volume-of-grey-matter-in-Vermis-VIIb-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 137.93647446033077, 'std': 28.197450905432245}, annotation_units=1, channel_map={'25907_Volume-of-grey-matter-in-Vermis-VIIb-Cerebellum_2_0': 0, })
ukb_25908_2 = TensorMap('25908_Volume-of-grey-matter-in-VIIb-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4072.3234697691805, 'std': 715.1573561074416}, annotation_units=1, channel_map={'25908_Volume-of-grey-matter-in-VIIb-Cerebellum-right_2_0': 0, })
ukb_25909_2 = TensorMap('25909_Volume-of-grey-matter-in-VIIIa-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3750.8491514811703, 'std': 696.5351853088838}, annotation_units=1, channel_map={'25909_Volume-of-grey-matter-in-VIIIa-Cerebellum-left_2_0': 0, })
ukb_25910_2 = TensorMap('25910_Volume-of-grey-matter-in-Vermis-VIIIa-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 890.8067704887393, 'std': 171.3196493666788}, annotation_units=1, channel_map={'25910_Volume-of-grey-matter-in-Vermis-VIIIa-Cerebellum_2_0': 0, })
ukb_25911_2 = TensorMap('25911_Volume-of-grey-matter-in-VIIIa-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3797.3036833006254, 'std': 742.2634080149609}, annotation_units=1, channel_map={'25911_Volume-of-grey-matter-in-VIIIa-Cerebellum-right_2_0': 0, })
ukb_25912_2 = TensorMap('25912_Volume-of-grey-matter-in-VIIIb-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2637.155436454537, 'std': 508.35275853471296}, annotation_units=1, channel_map={'25912_Volume-of-grey-matter-in-VIIIb-Cerebellum-left_2_0': 0, })
ukb_25913_2 = TensorMap('25913_Volume-of-grey-matter-in-Vermis-VIIIb-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 440.26137977759106, 'std': 86.2149164167236}, annotation_units=1, channel_map={'25913_Volume-of-grey-matter-in-Vermis-VIIIb-Cerebellum_2_0': 0, })
ukb_25914_2 = TensorMap('25914_Volume-of-grey-matter-in-VIIIb-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2665.4887148864586, 'std': 544.0051680395945}, annotation_units=1, channel_map={'25914_Volume-of-grey-matter-in-VIIIb-Cerebellum-right_2_0': 0, })
ukb_25915_2 = TensorMap('25915_Volume-of-grey-matter-in-IX-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1872.9248617418937, 'std': 387.2851609175637}, annotation_units=1, channel_map={'25915_Volume-of-grey-matter-in-IX-Cerebellum-left_2_0': 0, })
ukb_25916_2 = TensorMap('25916_Volume-of-grey-matter-in-Vermis-IX-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 427.65095070554145, 'std': 90.91506799708905}, annotation_units=1, channel_map={'25916_Volume-of-grey-matter-in-Vermis-IX-Cerebellum_2_0': 0, })
ukb_25917_2 = TensorMap('25917_Volume-of-grey-matter-in-IX-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2102.1202614708905, 'std': 425.56211809522324}, annotation_units=1, channel_map={'25917_Volume-of-grey-matter-in-IX-Cerebellum-right_2_0': 0, })
ukb_25918_2 = TensorMap('25918_Volume-of-grey-matter-in-X-Cerebellum-left_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 496.48001420427994, 'std': 80.04424940657563}, annotation_units=1, channel_map={'25918_Volume-of-grey-matter-in-X-Cerebellum-left_2_0': 0, })
ukb_25919_2 = TensorMap('25919_Volume-of-grey-matter-in-Vermis-X-Cerebellum_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 230.8770009298197, 'std': 42.79710533256611}, annotation_units=1, channel_map={'25919_Volume-of-grey-matter-in-Vermis-X-Cerebellum_2_0': 0, })
ukb_25920_2 = TensorMap('25920_Volume-of-grey-matter-in-X-Cerebellum-right_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 457.0044774787403, 'std': 79.41110275879713}, annotation_units=1, channel_map={'25920_Volume-of-grey-matter-in-X-Cerebellum-right_2_0': 0, })
ukb_30000_0 = TensorMap('30000_White-blood-cell-leukocyte-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.885049530003448, 'std': 2.122710204449377}, annotation_units=1, channel_map={'30000_White-blood-cell-leukocyte-count_0_0': 0, })
ukb_30000_1 = TensorMap('30000_White-blood-cell-leukocyte-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.806443437049247, 'std': 2.3880105716871545}, annotation_units=1, channel_map={'30000_White-blood-cell-leukocyte-count_1_0': 0, })
ukb_30000_2 = TensorMap('30000_White-blood-cell-leukocyte-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.812533674339302, 'std': 1.9645225281411858}, annotation_units=1, channel_map={'30000_White-blood-cell-leukocyte-count_2_0': 0, })
ukb_30001_0 = TensorMap('30001_White-blood-cell-leukocyte-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30001_White-blood-cell-leukocyte-count-freezethaw-cycles_0_0': 0, })
ukb_30010_1 = TensorMap('30010_Red-blood-cell-erythrocyte-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.537960074184741, 'std': 0.4000851145956889}, annotation_units=1, channel_map={'30010_Red-blood-cell-erythrocyte-count_1_0': 0, })
ukb_30010_0 = TensorMap('30010_Red-blood-cell-erythrocyte-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.516898456712669, 'std': 0.4169165557365781}, annotation_units=1, channel_map={'30010_Red-blood-cell-erythrocyte-count_0_0': 0, })
ukb_30010_2 = TensorMap('30010_Red-blood-cell-erythrocyte-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.607987723785165, 'std': 0.4217830842704462}, annotation_units=1, channel_map={'30010_Red-blood-cell-erythrocyte-count_2_0': 0, })
ukb_30011_0 = TensorMap('30011_Red-blood-cell-erythrocyte-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30011_Red-blood-cell-erythrocyte-count-freezethaw-cycles_0_0': 0, })
ukb_30020_0 = TensorMap('30020_Haemoglobin-concentration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.174758887494772, 'std': 1.246999741921853}, annotation_units=1, channel_map={'30020_Haemoglobin-concentration_0_0': 0, })
ukb_30020_1 = TensorMap('30020_Haemoglobin-concentration_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.321980116416833, 'std': 1.220970660253692}, annotation_units=1, channel_map={'30020_Haemoglobin-concentration_1_0': 0, })
ukb_30020_2 = TensorMap('30020_Haemoglobin-concentration_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.195028132992327, 'std': 1.1911445093759974}, annotation_units=1, channel_map={'30020_Haemoglobin-concentration_2_0': 0, })
ukb_30021_0 = TensorMap('30021_Haemoglobin-concentration-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30021_Haemoglobin-concentration-freezethaw-cycles_0_0': 0, })
ukb_30030_1 = TensorMap('30030_Haematocrit-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.13178746200998, 'std': 3.372949271837765}, annotation_units=1, channel_map={'30030_Haematocrit-percentage_1_0': 0, })
ukb_30030_2 = TensorMap('30030_Haematocrit-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.75153964194374, 'std': 3.4841556293740257}, annotation_units=1, channel_map={'30030_Haematocrit-percentage_2_0': 0, })
ukb_30030_0 = TensorMap('30030_Haematocrit-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.08495568799665, 'std': 3.556031330918497}, annotation_units=1, channel_map={'30030_Haematocrit-percentage_0_0': 0, })
ukb_30031_0 = TensorMap('30031_Haematocrit-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30031_Haematocrit-freezethaw-cycles_0_0': 0, })
ukb_30040_1 = TensorMap('30040_Mean-corpuscular-volume_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.79351465512796, 'std': 4.276123105316493}, annotation_units=1, channel_map={'30040_Mean-corpuscular-volume_1_0': 0, })
ukb_30040_2 = TensorMap('30040_Mean-corpuscular-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 90.77227791986358, 'std': 4.227289255382017}, annotation_units=1, channel_map={'30040_Mean-corpuscular-volume_2_0': 0, })
ukb_30040_0 = TensorMap('30040_Mean-corpuscular-volume_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.1200520077457, 'std': 4.606194293568709}, annotation_units=1, channel_map={'30040_Mean-corpuscular-volume_0_0': 0, })
ukb_30041_0 = TensorMap('30041_Mean-corpuscular-volume-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30041_Mean-corpuscular-volume-freezethaw-cycles_0_0': 0, })
ukb_30050_0 = TensorMap('30050_Mean-corpuscular-haemoglobin_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.450949838664815, 'std': 1.9215704123050523}, annotation_units=1, channel_map={'30050_Mean-corpuscular-haemoglobin_0_0': 0, })
ukb_30050_1 = TensorMap('30050_Mean-corpuscular-haemoglobin_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.62269214918607, 'std': 1.896862118084125}, annotation_units=1, channel_map={'30050_Mean-corpuscular-haemoglobin_1_0': 0, })
ukb_30050_2 = TensorMap('30050_Mean-corpuscular-haemoglobin_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.862143953607365, 'std': 1.7747781227346195}, annotation_units=1, channel_map={'30050_Mean-corpuscular-haemoglobin_2_0': 0, })
ukb_30051_0 = TensorMap('30051_Mean-corpuscular-haemoglobin-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30051_Mean-corpuscular-haemoglobin-freezethaw-cycles_0_0': 0, })
ukb_30060_1 = TensorMap('30060_Mean-corpuscular-haemoglobin-concentration_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 34.833202658149595, 'std': 1.3383639379990926}, annotation_units=1, channel_map={'30060_Mean-corpuscular-haemoglobin-concentration_1_0': 0, })
ukb_30060_0 = TensorMap('30060_Mean-corpuscular-haemoglobin-concentration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 34.51283943930591, 'std': 1.076032227105021}, annotation_units=1, channel_map={'30060_Mean-corpuscular-haemoglobin-concentration_0_0': 0, })
ukb_30060_2 = TensorMap('30060_Mean-corpuscular-haemoglobin-concentration_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 33.99354596622888, 'std': 0.8858010289410925}, annotation_units=1, channel_map={'30060_Mean-corpuscular-haemoglobin-concentration_2_0': 0, })
ukb_30061_0 = TensorMap('30061_Mean-corpuscular-haemoglobin-concentration-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30061_Mean-corpuscular-haemoglobin-concentration-freezethaw-cycles_0_0': 0, })
ukb_30070_1 = TensorMap('30070_Red-blood-cell-erythrocyte-distribution-width_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.481004996651729, 'std': 0.9034491139458598}, annotation_units=1, channel_map={'30070_Red-blood-cell-erythrocyte-distribution-width_1_0': 0, })
ukb_30070_2 = TensorMap('30070_Red-blood-cell-erythrocyte-distribution-width_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.582770673486786, 'std': 0.9659156387805118}, annotation_units=1, channel_map={'30070_Red-blood-cell-erythrocyte-distribution-width_2_0': 0, })
ukb_30070_0 = TensorMap('30070_Red-blood-cell-erythrocyte-distribution-width_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.493247420524549, 'std': 0.9877284383592139}, annotation_units=1, channel_map={'30070_Red-blood-cell-erythrocyte-distribution-width_0_0': 0, })
ukb_30071_0 = TensorMap('30071_Red-blood-cell-erythrocyte-distribution-width-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30071_Red-blood-cell-erythrocyte-distribution-width-freezethaw-cycles_0_0': 0, })
ukb_30080_0 = TensorMap('30080_Platelet-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 252.9890089231007, 'std': 60.048374290196065}, annotation_units=1, channel_map={'30080_Platelet-count_0_0': 0, })
ukb_30080_2 = TensorMap('30080_Platelet-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 231.58323955669223, 'std': 55.54623291338252}, annotation_units=1, channel_map={'30080_Platelet-count_2_0': 0, })
ukb_30080_1 = TensorMap('30080_Platelet-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 234.413547622727, 'std': 57.81102570286743}, annotation_units=1, channel_map={'30080_Platelet-count_1_0': 0, })
ukb_30081_0 = TensorMap('30081_Platelet-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30081_Platelet-count-freezethaw-cycles_0_0': 0, })
ukb_30090_0 = TensorMap('30090_Platelet-crit_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.23272601230047285, 'std': 0.0490126877229711}, annotation_units=1, channel_map={'30090_Platelet-crit_0_0': 0, })
ukb_30090_2 = TensorMap('30090_Platelet-crit_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.21047655583972716, 'std': 0.04502567674761242}, annotation_units=1, channel_map={'30090_Platelet-crit_2_0': 0, })
ukb_30090_1 = TensorMap('30090_Platelet-crit_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.20907895502838011, 'std': 0.04482859457559635}, annotation_units=1, channel_map={'30090_Platelet-crit_1_0': 0, })
ukb_30091_0 = TensorMap('30091_Platelet-crit-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30091_Platelet-crit-freezethaw-cycles_0_0': 0, })
ukb_30100_2 = TensorMap('30100_Mean-platelet-thrombocyte-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.22571526001705, 'std': 1.098198498807845}, annotation_units=1, channel_map={'30100_Mean-platelet-thrombocyte-volume_2_0': 0, })
ukb_30100_0 = TensorMap('30100_Mean-platelet-thrombocyte-volume_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.333080812727937, 'std': 1.0849852535700208}, annotation_units=1, channel_map={'30100_Mean-platelet-thrombocyte-volume_0_0': 0, })
ukb_30100_1 = TensorMap('30100_Mean-platelet-thrombocyte-volume_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.106210911338932, 'std': 1.065262587267326}, annotation_units=1, channel_map={'30100_Mean-platelet-thrombocyte-volume_1_0': 0, })
ukb_30101_0 = TensorMap('30101_Mean-platelet-thrombocyte-volume-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30101_Mean-platelet-thrombocyte-volume-freezethaw-cycles_0_0': 0, })
ukb_30110_1 = TensorMap('30110_Platelet-distribution-width_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.639384369087473, 'std': 0.5301429965676511}, annotation_units=1, channel_map={'30110_Platelet-distribution-width_1_0': 0, })
ukb_30110_2 = TensorMap('30110_Platelet-distribution-width_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.58054049445866, 'std': 0.5422564124286726}, annotation_units=1, channel_map={'30110_Platelet-distribution-width_2_0': 0, })
ukb_30110_0 = TensorMap('30110_Platelet-distribution-width_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.490355254793062, 'std': 0.5224214443834251}, annotation_units=1, channel_map={'30110_Platelet-distribution-width_0_0': 0, })
ukb_30111_0 = TensorMap('30111_Platelet-distribution-width-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30111_Platelet-distribution-width-freezethaw-cycles_0_0': 0, })
ukb_30120_0 = TensorMap('30120_Lymphocyte-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9659849029249274, 'std': 1.1723467154746938}, annotation_units=1, channel_map={'30120_Lymphocyte-count_0_0': 0, })
ukb_30120_1 = TensorMap('30120_Lymphocyte-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9266769754136386, 'std': 1.6174605090003007}, annotation_units=1, channel_map={'30120_Lymphocyte-count_1_0': 0, })
ukb_30120_2 = TensorMap('30120_Lymphocyte-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8465739160122907, 'std': 1.0008152164983248}, annotation_units=1, channel_map={'30120_Lymphocyte-count_2_0': 0, })
ukb_30121_0 = TensorMap('30121_Lymphocyte-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30121_Lymphocyte-count-freezethaw-cycles_0_0': 0, })
ukb_30130_1 = TensorMap('30130_Monocyte-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.44614762125663615, 'std': 0.24445028516690565}, annotation_units=1, channel_map={'30130_Monocyte-count_1_0': 0, })
ukb_30130_0 = TensorMap('30130_Monocyte-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4759545663100157, 'std': 0.2729186969343378}, annotation_units=1, channel_map={'30130_Monocyte-count_0_0': 0, })
ukb_30130_2 = TensorMap('30130_Monocyte-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.3927381358825538, 'std': 0.22669643145303373}, annotation_units=1, channel_map={'30130_Monocyte-count_2_0': 0, })
ukb_30131_0 = TensorMap('30131_Monocyte-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30131_Monocyte-count-freezethaw-cycles_0_0': 0, })
ukb_30140_2 = TensorMap('30140_Neutrophill-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.3699044042335276, 'std': 1.473735185867849}, annotation_units=1, channel_map={'30140_Neutrophill-count_2_0': 0, })
ukb_30140_0 = TensorMap('30140_Neutrophill-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.227884419173243, 'std': 1.4218134581914312}, annotation_units=1, channel_map={'30140_Neutrophill-count_0_0': 0, })
ukb_30140_1 = TensorMap('30140_Neutrophill-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.229779908252154, 'std': 1.3826807288556813}, annotation_units=1, channel_map={'30140_Neutrophill-count_1_0': 0, })
ukb_30141_0 = TensorMap('30141_Neutrophill-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30141_Neutrophill-count-freezethaw-cycles_0_0': 0, })
ukb_30150_0 = TensorMap('30150_Eosinophill-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.17487985820565613, 'std': 0.13857791589761187}, annotation_units=1, channel_map={'30150_Eosinophill-count_0_0': 0, })
ukb_30150_2 = TensorMap('30150_Eosinophill-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.1739279617616934, 'std': 0.14020736222479818}, annotation_units=1, channel_map={'30150_Eosinophill-count_2_0': 0, })
ukb_30150_1 = TensorMap('30150_Eosinophill-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.16508839750528326, 'std': 0.12736166406787428}, annotation_units=1, channel_map={'30150_Eosinophill-count_1_0': 0, })
ukb_30151_0 = TensorMap('30151_Eosinophill-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30151_Eosinophill-count-freezethaw-cycles_0_0': 0, })
ukb_30160_1 = TensorMap('30160_Basophill-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.022648317097056853, 'std': 0.04931015321075054}, annotation_units=1, channel_map={'30160_Basophill-count_1_0': 0, })
ukb_30160_0 = TensorMap('30160_Basophill-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03405815076543385, 'std': 0.05160802741611583}, annotation_units=1, channel_map={'30160_Basophill-count_0_0': 0, })
ukb_30160_2 = TensorMap('30160_Basophill-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.02968419255718675, 'std': 0.03517879344605347}, annotation_units=1, channel_map={'30160_Basophill-count_2_0': 0, })
ukb_30161_0 = TensorMap('30161_Basophill-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30161_Basophill-count-freezethaw-cycles_0_0': 0, })
ukb_30170_2 = TensorMap('30170_Nucleated-red-blood-cell-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0034152585765489006, 'std': 0.04063586577830749}, annotation_units=1, channel_map={'30170_Nucleated-red-blood-cell-count_2_0': 0, })
ukb_30170_1 = TensorMap('30170_Nucleated-red-blood-cell-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.006932215743440233, 'std': 0.045461709553699874}, annotation_units=1, channel_map={'30170_Nucleated-red-blood-cell-count_1_0': 0, })
ukb_30170_0 = TensorMap('30170_Nucleated-red-blood-cell-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.002217473287240727, 'std': 0.0331917060804216}, annotation_units=1, channel_map={'30170_Nucleated-red-blood-cell-count_0_0': 0, })
ukb_30171_0 = TensorMap('30171_Nucleated-red-blood-cell-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30171_Nucleated-red-blood-cell-count-freezethaw-cycles_0_0': 0, })
ukb_30180_0 = TensorMap('30180_Lymphocyte-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.91320742818714, 'std': 7.499219296319767}, annotation_units=1, channel_map={'30180_Lymphocyte-percentage_0_0': 0, })
ukb_30180_2 = TensorMap('30180_Lymphocyte-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.47120006828269, 'std': 7.566856288879453}, annotation_units=1, channel_map={'30180_Lymphocyte-percentage_2_0': 0, })
ukb_30180_1 = TensorMap('30180_Lymphocyte-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 28.442778064117107, 'std': 7.653464130265143}, annotation_units=1, channel_map={'30180_Lymphocyte-percentage_1_0': 0, })
ukb_30181_0 = TensorMap('30181_Lymphocyte-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30181_Lymphocyte-percentage-freezethaw-cycles_0_0': 0, })
ukb_30190_1 = TensorMap('30190_Monocyte-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.679890732914133, 'std': 2.999576464169819}, annotation_units=1, channel_map={'30190_Monocyte-percentage_1_0': 0, })
ukb_30190_0 = TensorMap('30190_Monocyte-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.064322787581421, 'std': 2.698148749114867}, annotation_units=1, channel_map={'30190_Monocyte-percentage_0_0': 0, })
ukb_30190_2 = TensorMap('30190_Monocyte-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.917794469102084, 'std': 3.0685951954228448}, annotation_units=1, channel_map={'30190_Monocyte-percentage_2_0': 0, })
ukb_30191_0 = TensorMap('30191_Monocyte-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30191_Monocyte-percentage-freezethaw-cycles_0_0': 0, })
ukb_30200_2 = TensorMap('30200_Neutrophill-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 63.58398088084669, 'std': 8.800997013517138}, annotation_units=1, channel_map={'30200_Neutrophill-percentage_2_0': 0, })
ukb_30200_1 = TensorMap('30200_Neutrophill-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 61.89233378002266, 'std': 8.71822241388449}, annotation_units=1, channel_map={'30200_Neutrophill-percentage_1_0': 0, })
ukb_30200_0 = TensorMap('30200_Neutrophill-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 60.88044484064049, 'std': 8.525766569129084}, annotation_units=1, channel_map={'30200_Neutrophill-percentage_0_0': 0, })
ukb_30201_0 = TensorMap('30201_Neutrophill-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30201_Neutrophill-percentage-freezethaw-cycles_0_0': 0, })
ukb_30210_0 = TensorMap('30210_Eosinophill-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5724457959804488, 'std': 1.8756692623102362}, annotation_units=1, channel_map={'30210_Eosinophill-percentage_0_0': 0, })
ukb_30210_1 = TensorMap('30210_Eosinophill-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.455085042779095, 'std': 1.7383476491589944}, annotation_units=1, channel_map={'30210_Eosinophill-percentage_1_0': 0, })
ukb_30210_2 = TensorMap('30210_Eosinophill-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5809730283373162, 'std': 1.8512379540171102}, annotation_units=1, channel_map={'30210_Eosinophill-percentage_2_0': 0, })
ukb_30211_0 = TensorMap('30211_Eosinophill-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30211_Eosinophill-percentage-freezethaw-cycles_0_0': 0, })
ukb_30220_1 = TensorMap('30220_Basophill-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.529863931553448, 'std': 0.5072258165824638}, annotation_units=1, channel_map={'30220_Basophill-percentage_1_0': 0, })
ukb_30220_0 = TensorMap('30220_Basophill-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5696815742996796, 'std': 0.6104582668162517}, annotation_units=1, channel_map={'30220_Basophill-percentage_0_0': 0, })
ukb_30220_2 = TensorMap('30220_Basophill-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.4477569136223967, 'std': 0.4301342699160611}, annotation_units=1, channel_map={'30220_Basophill-percentage_2_0': 0, })
ukb_30221_0 = TensorMap('30221_Basophill-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30221_Basophill-percentage-freezethaw-cycles_0_0': 0, })
ukb_30230_0 = TensorMap('30230_Nucleated-red-blood-cell-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.03519063641849084, 'std': 0.44716185942014464}, annotation_units=1, channel_map={'30230_Nucleated-red-blood-cell-percentage_0_0': 0, })
ukb_30230_2 = TensorMap('30230_Nucleated-red-blood-cell-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05682539682539682, 'std': 0.6343693761184723}, annotation_units=1, channel_map={'30230_Nucleated-red-blood-cell-percentage_2_0': 0, })
ukb_30230_1 = TensorMap('30230_Nucleated-red-blood-cell-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.11536734693877551, 'std': 0.753556920094557}, annotation_units=1, channel_map={'30230_Nucleated-red-blood-cell-percentage_1_0': 0, })
ukb_30231_0 = TensorMap('30231_Nucleated-red-blood-cell-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30231_Nucleated-red-blood-cell-percentage-freezethaw-cycles_0_0': 0, })
ukb_30240_2 = TensorMap('30240_Reticulocyte-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2338956568751127, 'std': 0.48703594269648104}, annotation_units=1, channel_map={'30240_Reticulocyte-percentage_2_0': 0, })
ukb_30240_1 = TensorMap('30240_Reticulocyte-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2152385416127192, 'std': 0.5148993736572332}, annotation_units=1, channel_map={'30240_Reticulocyte-percentage_1_0': 0, })
ukb_30240_0 = TensorMap('30240_Reticulocyte-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.3502999076355988, 'std': 0.8905014830344062}, annotation_units=1, channel_map={'30240_Reticulocyte-percentage_0_0': 0, })
ukb_30241_0 = TensorMap('30241_Reticulocyte-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30241_Reticulocyte-percentage-freezethaw-cycles_0_0': 0, })
ukb_30250_2 = TensorMap('30250_Reticulocyte-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05688826815642458, 'std': 0.02294692603050867}, annotation_units=1, channel_map={'30250_Reticulocyte-count_2_0': 0, })
ukb_30250_0 = TensorMap('30250_Reticulocyte-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0610076849735463, 'std': 0.038791227146065387}, annotation_units=1, channel_map={'30250_Reticulocyte-count_0_0': 0, })
ukb_30250_1 = TensorMap('30250_Reticulocyte-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.05514464008285861, 'std': 0.02355164636757402}, annotation_units=1, channel_map={'30250_Reticulocyte-count_1_0': 0, })
ukb_30251_0 = TensorMap('30251_Reticulocyte-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30251_Reticulocyte-count-freezethaw-cycles_0_0': 0, })
ukb_30260_1 = TensorMap('30260_Mean-reticulocyte-volume_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 111.29789083941793, 'std': 7.245879722759731}, annotation_units=1, channel_map={'30260_Mean-reticulocyte-volume_1_0': 0, })
ukb_30260_0 = TensorMap('30260_Mean-reticulocyte-volume_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 105.91970732766238, 'std': 7.83416220909367}, annotation_units=1, channel_map={'30260_Mean-reticulocyte-volume_0_0': 0, })
ukb_30260_2 = TensorMap('30260_Mean-reticulocyte-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 114.05980897459003, 'std': 7.436665250636225}, annotation_units=1, channel_map={'30260_Mean-reticulocyte-volume_2_0': 0, })
ukb_30261_0 = TensorMap('30261_Mean-reticulocyte-volume-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30261_Mean-reticulocyte-volume-freezethaw-cycles_0_0': 0, })
ukb_30270_2 = TensorMap('30270_Mean-sphered-cell-volume_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 91.25678500630745, 'std': 5.680833205230017}, annotation_units=1, channel_map={'30270_Mean-sphered-cell-volume_2_0': 0, })
ukb_30270_0 = TensorMap('30270_Mean-sphered-cell-volume_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.86684850780733, 'std': 5.315066434406766}, annotation_units=1, channel_map={'30270_Mean-sphered-cell-volume_0_0': 0, })
ukb_30270_1 = TensorMap('30270_Mean-sphered-cell-volume_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 87.30648373244048, 'std': 4.8295262580817795}, annotation_units=1, channel_map={'30270_Mean-sphered-cell-volume_1_0': 0, })
ukb_30271_0 = TensorMap('30271_Mean-sphered-cell-volume-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30271_Mean-sphered-cell-volume-freezethaw-cycles_0_0': 0, })
ukb_30280_1 = TensorMap('30280_Immature-reticulocyte-fraction_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.310937550040032, 'std': 0.05898770720206083}, annotation_units=1, channel_map={'30280_Immature-reticulocyte-fraction_1_0': 0, })
ukb_30280_2 = TensorMap('30280_Immature-reticulocyte-fraction_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.32253937646422787, 'std': 0.06069062790474906}, annotation_units=1, channel_map={'30280_Immature-reticulocyte-fraction_2_0': 0, })
ukb_30280_0 = TensorMap('30280_Immature-reticulocyte-fraction_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.29079298624959304, 'std': 0.06104801809509128}, annotation_units=1, channel_map={'30280_Immature-reticulocyte-fraction_0_0': 0, })
ukb_30281_0 = TensorMap('30281_Immature-reticulocyte-fraction-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30281_Immature-reticulocyte-fraction-freezethaw-cycles_0_0': 0, })
ukb_30290_2 = TensorMap('30290_High-light-scatter-reticulocyte-percentage_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.41064173724995495, 'std': 0.21191613617085456}, annotation_units=1, channel_map={'30290_High-light-scatter-reticulocyte-percentage_2_0': 0, })
ukb_30290_0 = TensorMap('30290_High-light-scatter-reticulocyte-percentage_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.40206229050457704, 'std': 0.33171775634018935}, annotation_units=1, channel_map={'30290_High-light-scatter-reticulocyte-percentage_0_0': 0, })
ukb_30290_1 = TensorMap('30290_High-light-scatter-reticulocyte-percentage_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.39794832229419896, 'std': 0.2157504373657863}, annotation_units=1, channel_map={'30290_High-light-scatter-reticulocyte-percentage_1_0': 0, })
ukb_30291_0 = TensorMap('30291_High-light-scatter-reticulocyte-percentage-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30291_High-light-scatter-reticulocyte-percentage-freezethaw-cycles_0_0': 0, })
ukb_30300_0 = TensorMap('30300_High-light-scatter-reticulocyte-count_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.018121005878121554, 'std': 0.010238246080133753}, annotation_units=1, channel_map={'30300_High-light-scatter-reticulocyte-count_0_0': 0, })
ukb_30300_1 = TensorMap('30300_High-light-scatter-reticulocyte-count_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.01804068709513065, 'std': 0.009616470620763947}, annotation_units=1, channel_map={'30300_High-light-scatter-reticulocyte-count_1_0': 0, })
ukb_30300_2 = TensorMap('30300_High-light-scatter-reticulocyte-count_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.018913678140205448, 'std': 0.009774140541703702}, annotation_units=1, channel_map={'30300_High-light-scatter-reticulocyte-count_2_0': 0, })
ukb_30301_0 = TensorMap('30301_High-light-scatter-reticulocyte-count-freezethaw-cycles_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.0, 'std': 0.0}, annotation_units=1, channel_map={'30301_High-light-scatter-reticulocyte-count-freezethaw-cycles_0_0': 0, })
ukb_30314_0 = TensorMap('30314_Volume-of-EDTA1-plasma-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3196.4802821483136, 'std': 511.1004065680865}, annotation_units=1, channel_map={'30314_Volume-of-EDTA1-plasma-held-by-UKB_0_0': 0, })
ukb_30314_1 = TensorMap('30314_Volume-of-EDTA1-plasma-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2732.784238377221, 'std': 1243.3770949577754}, annotation_units=1, channel_map={'30314_Volume-of-EDTA1-plasma-held-by-UKB_1_0': 0, })
ukb_30324_0 = TensorMap('30324_Volume-of-EDTA1-red-cells-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 834.5843524257385, 'std': 110.04775521812479}, annotation_units=1, channel_map={'30324_Volume-of-EDTA1-red-cells-held-by-UKB_0_0': 0, })
ukb_30324_1 = TensorMap('30324_Volume-of-EDTA1-red-cells-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 707.5700869784445, 'std': 317.5354074534916}, annotation_units=1, channel_map={'30324_Volume-of-EDTA1-red-cells-held-by-UKB_1_0': 0, })
ukb_30334_0 = TensorMap('30334_Volume-of-EDTA1-buffy-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 464.9637793718353, 'std': 420.31808614438194}, annotation_units=1, channel_map={'30334_Volume-of-EDTA1-buffy-held-by-UKB_0_0': 0, })
ukb_30334_1 = TensorMap('30334_Volume-of-EDTA1-buffy-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 695.2849672379032, 'std': 325.0732810073748}, annotation_units=1, channel_map={'30334_Volume-of-EDTA1-buffy-held-by-UKB_1_0': 0, })
ukb_30344_1 = TensorMap('30344_Volume-of-EDTA2-plasma-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2710.304729431079, 'std': 1259.6258660875385}, annotation_units=1, channel_map={'30344_Volume-of-EDTA2-plasma-held-by-UKB_1_0': 0, })
ukb_30344_0 = TensorMap('30344_Volume-of-EDTA2-plasma-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3175.606143632216, 'std': 516.7827685337743}, annotation_units=1, channel_map={'30344_Volume-of-EDTA2-plasma-held-by-UKB_0_0': 0, })
ukb_30354_1 = TensorMap('30354_Volume-of-EDTA2-buffy-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 702.6986778009745, 'std': 318.72922239544437}, annotation_units=1, channel_map={'30354_Volume-of-EDTA2-buffy-held-by-UKB_1_0': 0, })
ukb_30354_0 = TensorMap('30354_Volume-of-EDTA2-buffy-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 830.8630044408109, 'std': 115.85718425156944}, annotation_units=1, channel_map={'30354_Volume-of-EDTA2-buffy-held-by-UKB_0_0': 0, })
ukb_30364_1 = TensorMap('30364_Volume-of-EDTA2-red-cells-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 704.5972160708636, 'std': 319.92448507449996}, annotation_units=1, channel_map={'30364_Volume-of-EDTA2-red-cells-held-by-UKB_1_0': 0, })
ukb_30364_0 = TensorMap('30364_Volume-of-EDTA2-red-cells-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 835.0722763644254, 'std': 106.87264953988611}, annotation_units=1, channel_map={'30364_Volume-of-EDTA2-red-cells-held-by-UKB_0_0': 0, })
ukb_30374_1 = TensorMap('30374_Volume-of-LiHep-plasma-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2153.8984758625083, 'std': 1152.5040845132582}, annotation_units=1, channel_map={'30374_Volume-of-LiHep-plasma-held-by-UKB_1_0': 0, })
ukb_30374_0 = TensorMap('30374_Volume-of-LiHep-plasma-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2621.252433023846, 'std': 582.5761642840705}, annotation_units=1, channel_map={'30374_Volume-of-LiHep-plasma-held-by-UKB_0_0': 0, })
ukb_30384_0 = TensorMap('30384_Volume-of-serum-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2262.705916292833, 'std': 585.7574239974638}, annotation_units=1, channel_map={'30384_Volume-of-serum-held-by-UKB_0_0': 0, })
ukb_30384_1 = TensorMap('30384_Volume-of-serum-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1768.9905534774618, 'std': 1153.0201873654507}, annotation_units=1, channel_map={'30384_Volume-of-serum-held-by-UKB_1_0': 0, })
ukb_30404_0 = TensorMap('30404_Volume-of-ACD-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1394.9801762068926, 'std': 78.47424530021985}, annotation_units=1, channel_map={'30404_Volume-of-ACD-held-by-UKB_0_0': 0, })
ukb_30404_1 = TensorMap('30404_Volume-of-ACD-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1069.4503117444963, 'std': 594.6344409696749}, annotation_units=1, channel_map={'30404_Volume-of-ACD-held-by-UKB_1_0': 0, })
ukb_30414_0 = TensorMap('30414_Volume-of-RNA-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2606.8100669047035, 'std': 2548.692803060345}, annotation_units=1, channel_map={'30414_Volume-of-RNA-held-by-UKB_0_0': 0, })
ukb_30414_1 = TensorMap('30414_Volume-of-RNA-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3888.2675438596493, 'std': 2169.379087842377}, annotation_units=1, channel_map={'30414_Volume-of-RNA-held-by-UKB_1_0': 0, })
ukb_30500_1 = TensorMap('30500_Microalbumin-in-urine_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.65019215987702, 'std': 103.07030537890607}, annotation_units=1, channel_map={'30500_Microalbumin-in-urine_1_0': 0, })
ukb_30500_0 = TensorMap('30500_Microalbumin-in-urine_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 31.210916582191565, 'std': 129.41440770156902}, annotation_units=1, channel_map={'30500_Microalbumin-in-urine_0_0': 0, })
ukb_30510_1 = TensorMap('30510_Creatinine-enzymatic-in-urine_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9030.527592448925, 'std': 5643.6518059119935}, annotation_units=1, channel_map={'30510_Creatinine-enzymatic-in-urine_1_0': 0, })
ukb_30510_0 = TensorMap('30510_Creatinine-enzymatic-in-urine_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8881.438460331421, 'std': 5821.1123370478745}, annotation_units=1, channel_map={'30510_Creatinine-enzymatic-in-urine_0_0': 0, })
ukb_30520_1 = TensorMap('30520_Potassium-in-urine_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 65.54439782439785, 'std': 33.07899851568515}, annotation_units=1, channel_map={'30520_Potassium-in-urine_1_0': 0, })
ukb_30520_0 = TensorMap('30520_Potassium-in-urine_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 63.061827003451164, 'std': 33.84208704773539}, annotation_units=1, channel_map={'30520_Potassium-in-urine_0_0': 0, })
ukb_30530_0 = TensorMap('30530_Sodium-in-urine_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.45323967267048, 'std': 44.44123684846377}, annotation_units=1, channel_map={'30530_Sodium-in-urine_0_0': 0, })
ukb_30530_1 = TensorMap('30530_Sodium-in-urine_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 74.91479427919991, 'std': 42.32910571904886}, annotation_units=1, channel_map={'30530_Sodium-in-urine_1_0': 0, })
ukb_30600_1 = TensorMap('30600_Albumin_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.63048805352024, 'std': 2.603231807651387}, annotation_units=1, channel_map={'30600_Albumin_1_0': 0, })
ukb_30600_0 = TensorMap('30600_Albumin_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.210079284255926, 'std': 2.629164605833256}, annotation_units=1, channel_map={'30600_Albumin_0_0': 0, })
ukb_30610_0 = TensorMap('30610_Alkaline-phosphatase_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 83.66915643871293, 'std': 26.459966244953588}, annotation_units=1, channel_map={'30610_Alkaline-phosphatase_0_0': 0, })
ukb_30610_1 = TensorMap('30610_Alkaline-phosphatase_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 84.5702590789547, 'std': 24.05524945728438}, annotation_units=1, channel_map={'30610_Alkaline-phosphatase_1_0': 0, })
ukb_30620_0 = TensorMap('30620_Alanine-aminotransferase_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.54631834129685, 'std': 14.180177116607988}, annotation_units=1, channel_map={'30620_Alanine-aminotransferase_0_0': 0, })
ukb_30620_1 = TensorMap('30620_Alanine-aminotransferase_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 22.99633921074727, 'std': 12.150454473798915}, annotation_units=1, channel_map={'30620_Alanine-aminotransferase_1_0': 0, })
ukb_30630_0 = TensorMap('30630_Apolipoprotein-A_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5376508308150054, 'std': 0.27065607852670565}, annotation_units=1, channel_map={'30630_Apolipoprotein-A_0_0': 0, })
ukb_30630_1 = TensorMap('30630_Apolipoprotein-A_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5836051769542039, 'std': 0.2736422536674678}, annotation_units=1, channel_map={'30630_Apolipoprotein-A_1_0': 0, })
ukb_30640_0 = TensorMap('30640_Apolipoprotein-B_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.0320267939695962, 'std': 0.23825112625325112}, annotation_units=1, channel_map={'30640_Apolipoprotein-B_0_0': 0, })
ukb_30640_1 = TensorMap('30640_Apolipoprotein-B_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.022535975267004, 'std': 0.23722369478859215}, annotation_units=1, channel_map={'30640_Apolipoprotein-B_1_0': 0, })
ukb_30650_1 = TensorMap('30650_Aspartate-aminotransferase_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.546372009435025, 'std': 10.282026098615185}, annotation_units=1, channel_map={'30650_Aspartate-aminotransferase_1_0': 0, })
ukb_30650_0 = TensorMap('30650_Aspartate-aminotransferase_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.23016083040135, 'std': 10.661580442803643}, annotation_units=1, channel_map={'30650_Aspartate-aminotransferase_0_0': 0, })
ukb_30660_1 = TensorMap('30660_Direct-bilirubin_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8043401993355488, 'std': 0.7923055740545305}, annotation_units=1, channel_map={'30660_Direct-bilirubin_1_0': 0, })
ukb_30660_0 = TensorMap('30660_Direct-bilirubin_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.8336304228941829, 'std': 0.8506083585971784}, annotation_units=1, channel_map={'30660_Direct-bilirubin_0_0': 0, })
ukb_30670_1 = TensorMap('30670_Urea_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.606288371050862, 'std': 1.382369287565926}, annotation_units=1, channel_map={'30670_Urea_1_0': 0, })
ukb_30670_0 = TensorMap('30670_Urea_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.403444556432281, 'std': 1.401322152174818}, annotation_units=1, channel_map={'30670_Urea_0_0': 0, })
ukb_30680_0 = TensorMap('30680_Calcium_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.3798800595411223, 'std': 0.09418845422721601}, annotation_units=1, channel_map={'30680_Calcium_0_0': 0, })
ukb_30680_1 = TensorMap('30680_Calcium_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.4031776630330843, 'std': 0.09739641131326769}, annotation_units=1, channel_map={'30680_Calcium_1_0': 0, })
ukb_30690_0 = TensorMap('30690_Cholesterol_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.692381214399044, 'std': 1.1449409331668705}, annotation_units=1, channel_map={'30690_Cholesterol_0_0': 0, })
ukb_30690_1 = TensorMap('30690_Cholesterol_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.697317037699968, 'std': 1.1874487969407612}, annotation_units=1, channel_map={'30690_Cholesterol_1_0': 0, })
ukb_30700_1 = TensorMap('30700_Creatinine_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 75.04581932302172, 'std': 17.61833737041235}, annotation_units=1, channel_map={'30700_Creatinine_1_0': 0, })
ukb_30700_0 = TensorMap('30700_Creatinine_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 72.30598402583821, 'std': 18.549736297431203}, annotation_units=1, channel_map={'30700_Creatinine_0_0': 0, })
ukb_30710_0 = TensorMap('30710_Creactive-protein_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5987742480697564, 'std': 4.361150498539324}, annotation_units=1, channel_map={'30710_Creactive-protein_0_0': 0, })
ukb_30710_1 = TensorMap('30710_Creactive-protein_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.3319688305863893, 'std': 4.056273724938837}, annotation_units=1, channel_map={'30710_Creactive-protein_1_0': 0, })
ukb_30720_1 = TensorMap('30720_Cystatin-C_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9180014553596415, 'std': 0.17043527852871537}, annotation_units=1, channel_map={'30720_Cystatin-C_1_0': 0, })
ukb_30720_0 = TensorMap('30720_Cystatin-C_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9075203776108212, 'std': 0.17630752583057654}, annotation_units=1, channel_map={'30720_Cystatin-C_0_0': 0, })
ukb_30730_0 = TensorMap('30730_Gamma-glutamyltransferase_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 37.38983782447888, 'std': 42.08532805492437}, annotation_units=1, channel_map={'30730_Gamma-glutamyltransferase_0_0': 0, })
ukb_30730_1 = TensorMap('30730_Gamma-glutamyltransferase_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 35.09419502910879, 'std': 35.71700777901077}, annotation_units=1, channel_map={'30730_Gamma-glutamyltransferase_1_0': 0, })
ukb_30740_1 = TensorMap('30740_Glucose_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.07379931148795, 'std': 1.0640032559502723}, annotation_units=1, channel_map={'30740_Glucose_1_0': 0, })
ukb_30740_0 = TensorMap('30740_Glucose_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.12498858213645, 'std': 1.2434621168913167}, annotation_units=1, channel_map={'30740_Glucose_0_0': 0, })
ukb_30750_0 = TensorMap('30750_Glycated-haemoglobin-HbA1c_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 36.131166055773356, 'std': 6.775811313289605}, annotation_units=1, channel_map={'30750_Glycated-haemoglobin-HbA1c_0_0': 0, })
ukb_30750_1 = TensorMap('30750_Glycated-haemoglobin-HbA1c_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 36.493216062793465, 'std': 5.756369829686542}, annotation_units=1, channel_map={'30750_Glycated-haemoglobin-HbA1c_1_0': 0, })
ukb_30760_0 = TensorMap('30760_HDL-cholesterol_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.4480129055069357, 'std': 0.3823115953478376}, annotation_units=1, channel_map={'30760_HDL-cholesterol_0_0': 0, })
ukb_30760_1 = TensorMap('30760_HDL-cholesterol_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5190205837369364, 'std': 0.39690801402650105}, annotation_units=1, channel_map={'30760_HDL-cholesterol_1_0': 0, })
ukb_30770_1 = TensorMap('30770_IGF1_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.956882090395492, 'std': 5.53021668661157}, annotation_units=1, channel_map={'30770_IGF1_1_0': 0, })
ukb_30770_0 = TensorMap('30770_IGF1_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.39891163561463, 'std': 5.699765525677172}, annotation_units=1, channel_map={'30770_IGF1_0_0': 0, })
ukb_30780_1 = TensorMap('30780_LDL-direct_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5434994393990364, 'std': 0.8920277942847082}, annotation_units=1, channel_map={'30780_LDL-direct_1_0': 0, })
ukb_30780_0 = TensorMap('30780_LDL-direct_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5563651660223763, 'std': 0.8699758052966666}, annotation_units=1, channel_map={'30780_LDL-direct_0_0': 0, })
ukb_30790_1 = TensorMap('30790_Lipoprotein-A_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.43141190526389, 'std': 49.31803100446482}, annotation_units=1, channel_map={'30790_Lipoprotein-A_1_0': 0, })
ukb_30790_0 = TensorMap('30790_Lipoprotein-A_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 44.64786510991498, 'std': 49.21244270456523}, annotation_units=1, channel_map={'30790_Lipoprotein-A_0_0': 0, })
ukb_30800_1 = TensorMap('30800_Oestradiol_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 399.74236743438684, 'std': 374.2447259161634}, annotation_units=1, channel_map={'30800_Oestradiol_1_0': 0, })
ukb_30800_0 = TensorMap('30800_Oestradiol_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 461.17321908339187, 'std': 431.1611967874333}, annotation_units=1, channel_map={'30800_Oestradiol_0_0': 0, })
ukb_30810_1 = TensorMap('30810_Phosphate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.2274558466453676, 'std': 0.15766116882586884}, annotation_units=1, channel_map={'30810_Phosphate_1_0': 0, })
ukb_30810_0 = TensorMap('30810_Phosphate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1593289926713424, 'std': 0.16110980316809714}, annotation_units=1, channel_map={'30810_Phosphate_0_0': 0, })
ukb_30820_0 = TensorMap('30820_Rheumatoid-factor_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.556292961564534, 'std': 19.858152125238586}, annotation_units=1, channel_map={'30820_Rheumatoid-factor_0_0': 0, })
ukb_30820_1 = TensorMap('30820_Rheumatoid-factor_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.783372921615204, 'std': 19.520473802392154}, annotation_units=1, channel_map={'30820_Rheumatoid-factor_1_0': 0, })
ukb_30830_0 = TensorMap('30830_SHBG_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 51.62937739069994, 'std': 27.781222505868797}, annotation_units=1, channel_map={'30830_SHBG_0_0': 0, })
ukb_30830_1 = TensorMap('30830_SHBG_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.78426575236123, 'std': 26.115366059983458}, annotation_units=1, channel_map={'30830_SHBG_1_0': 0, })
ukb_30840_0 = TensorMap('30840_Total-bilirubin_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.125066682920398, 'std': 4.4266920697166166}, annotation_units=1, channel_map={'30840_Total-bilirubin_0_0': 0, })
ukb_30840_1 = TensorMap('30840_Total-bilirubin_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.964484135913603, 'std': 4.380548776990894}, annotation_units=1, channel_map={'30840_Total-bilirubin_1_0': 0, })
ukb_30850_1 = TensorMap('30850_Testosterone_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.204103858861665, 'std': 6.18526590396918}, annotation_units=1, channel_map={'30850_Testosterone_1_0': 0, })
ukb_30850_0 = TensorMap('30850_Testosterone_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.557780011852343, 'std': 6.054145820936543}, annotation_units=1, channel_map={'30850_Testosterone_0_0': 0, })
ukb_30860_0 = TensorMap('30860_Total-protein_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 72.5102489368684, 'std': 4.115743513768043}, annotation_units=1, channel_map={'30860_Total-protein_0_0': 0, })
ukb_30860_1 = TensorMap('30860_Total-protein_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 73.28005164169588, 'std': 4.180186943556039}, annotation_units=1, channel_map={'30860_Total-protein_1_0': 0, })
ukb_30870_1 = TensorMap('30870_Triglycerides_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.698357579153824, 'std': 0.934018248006202}, annotation_units=1, channel_map={'30870_Triglycerides_1_0': 0, })
ukb_30870_0 = TensorMap('30870_Triglycerides_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.7475781220697324, 'std': 1.02768393353082}, annotation_units=1, channel_map={'30870_Triglycerides_0_0': 0, })
ukb_30880_0 = TensorMap('30880_Urate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 309.2054548439295, 'std': 80.43205542125857}, annotation_units=1, channel_map={'30880_Urate_0_0': 0, })
ukb_30880_1 = TensorMap('30880_Urate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 314.3456421761075, 'std': 79.18969882248577}, annotation_units=1, channel_map={'30880_Urate_1_0': 0, })
ukb_30890_0 = TensorMap('30890_Vitamin-D_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 48.61285706639076, 'std': 21.11087406647955}, annotation_units=1, channel_map={'30890_Vitamin-D_0_0': 0, })
ukb_30890_1 = TensorMap('30890_Vitamin-D_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.71268266917073, 'std': 21.962877938755963}, annotation_units=1, channel_map={'30890_Vitamin-D_1_0': 0, })
ukb_30897_0 = TensorMap('30897_Estimated-sample-dilution-factor_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9970687105513253, 'std': 0.010569893704978957}, annotation_units=1, channel_map={'30897_Estimated-sample-dilution-factor_0_0': 0, })
ukb_30897_1 = TensorMap('30897_Estimated-sample-dilution-factor_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9951866081126381, 'std': 0.013259919105324118}, annotation_units=1, channel_map={'30897_Estimated-sample-dilution-factor_1_0': 0, })
ukb_40007_0 = TensorMap('40007_Age-at-death_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 67.2703209396039, 'std': 7.095180722295174}, annotation_units=1, channel_map={'40007_Age-at-death_0_0': 0, })
ukb_40007_2 = TensorMap('40007_Age-at-death_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 64.42408050139271, 'std': 6.743856891804327}, annotation_units=1, channel_map={'40007_Age-at-death_2_0': 0, })
ukb_40007_1 = TensorMap('40007_Age-at-death_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 65.71046171761282, 'std': 6.8254349587598435}, annotation_units=1, channel_map={'40007_Age-at-death_1_0': 0, })
ukb_40008_6 = TensorMap('40008_Age-at-cancer-diagnosis_6_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.89033470769232, 'std': 10.228621609394228}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_6_0': 0, })
ukb_40008_11 = TensorMap('40008_Age-at-cancer-diagnosis_11_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.71769323308269, 'std': 10.939378666495141}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_11_0': 0, })
ukb_40008_13 = TensorMap('40008_Age-at-cancer-diagnosis_13_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.67445476190478, 'std': 9.545504097107894}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_13_0': 0, })
ukb_40008_17 = TensorMap('40008_Age-at-cancer-diagnosis_17_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.44071818181818, 'std': 9.169381990059984}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_17_0': 0, })
ukb_40008_26 = TensorMap('40008_Age-at-cancer-diagnosis_26_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 51.59035, 'std': 0.5072076941451117}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_26_0': 0, })
ukb_40008_0 = TensorMap('40008_Age-at-cancer-diagnosis_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.59601672041155, 'std': 11.456438149908758}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_0_0': 0, })
ukb_40008_2 = TensorMap('40008_Age-at-cancer-diagnosis_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 55.79798074494419, 'std': 11.11756300372685}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_2_0': 0, })
ukb_40008_15 = TensorMap('40008_Age-at-cancer-diagnosis_15_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.48068124999999, 'std': 9.847364257588168}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_15_0': 0, })
ukb_40008_21 = TensorMap('40008_Age-at-cancer-diagnosis_21_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 62.476299999999995, 'std': 6.605650128488489}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_21_0': 0, })
ukb_40008_4 = TensorMap('40008_Age-at-cancer-diagnosis_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.49157407778574, 'std': 10.332953188266154}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_4_0': 0, })
ukb_40008_9 = TensorMap('40008_Age-at-cancer-diagnosis_9_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 60.20864264150944, 'std': 10.254704918594934}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_9_0': 0, })
ukb_40008_28 = TensorMap('40008_Age-at-cancer-diagnosis_28_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.3947, 'std': None}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_28_0': 0, })
ukb_40008_3 = TensorMap('40008_Age-at-cancer-diagnosis_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.617202984820075, 'std': 10.470985153031211}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_3_0': 0, })
ukb_40008_8 = TensorMap('40008_Age-at-cancer-diagnosis_8_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.483288337468984, 'std': 10.163049906291755}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_8_0': 0, })
ukb_40008_7 = TensorMap('40008_Age-at-cancer-diagnosis_7_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.39773298337709, 'std': 10.914294120960472}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_7_0': 0, })
ukb_40008_25 = TensorMap('40008_Age-at-cancer-diagnosis_25_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 50.5746, 'std': 0.9292797318353605}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_25_0': 0, })
ukb_40008_5 = TensorMap('40008_Age-at-cancer-diagnosis_5_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.77415359343237, 'std': 10.708318050750744}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_5_0': 0, })
ukb_40008_12 = TensorMap('40008_Age-at-cancer-diagnosis_12_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 60.5177641509434, 'std': 8.86330516540169}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_12_0': 0, })
ukb_40008_16 = TensorMap('40008_Age-at-cancer-diagnosis_16_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.612675, 'std': 11.180292402024767}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_16_0': 0, })
ukb_40008_18 = TensorMap('40008_Age-at-cancer-diagnosis_18_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 60.502250000000004, 'std': 8.34125690477332}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_18_0': 0, })
ukb_40008_19 = TensorMap('40008_Age-at-cancer-diagnosis_19_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 55.502871428571424, 'std': 13.23028705933401}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_19_0': 0, })
ukb_40008_23 = TensorMap('40008_Age-at-cancer-diagnosis_23_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.298950000000005, 'std': 13.000228882792795}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_23_0': 0, })
ukb_40008_30 = TensorMap('40008_Age-at-cancer-diagnosis_30_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 74.3177, 'std': None}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_30_0': 0, })
ukb_40008_31 = TensorMap('40008_Age-at-cancer-diagnosis_31_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 74.3177, 'std': None}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_31_0': 0, })
ukb_40008_1 = TensorMap('40008_Age-at-cancer-diagnosis_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.647423485294546, 'std': 11.515449301952044}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_1_0': 0, })
ukb_40008_10 = TensorMap('40008_Age-at-cancer-diagnosis_10_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.32004245810055, 'std': 10.654634185548094}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_10_0': 0, })
ukb_40008_22 = TensorMap('40008_Age-at-cancer-diagnosis_22_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 56.2476, 'std': 15.414362144441785}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_22_0': 0, })
ukb_40008_24 = TensorMap('40008_Age-at-cancer-diagnosis_24_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.3825, 'std': 12.146397444510038}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_24_0': 0, })
ukb_40008_29 = TensorMap('40008_Age-at-cancer-diagnosis_29_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 38.1664, 'std': None}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_29_0': 0, })
ukb_40008_14 = TensorMap('40008_Age-at-cancer-diagnosis_14_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 59.974486666666664, 'std': 9.451299833383013}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_14_0': 0, })
ukb_40008_20 = TensorMap('40008_Age-at-cancer-diagnosis_20_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 67.93130000000001, 'std': 6.190336962153196}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_20_0': 0, })
ukb_40008_27 = TensorMap('40008_Age-at-cancer-diagnosis_27_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 58.55015, 'std': 12.010845074556574}, annotation_units=1, channel_map={'40008_Age-at-cancer-diagnosis_27_0': 0, })
ukb_40009_0 = TensorMap('40009_Reported-occurrences-of-cancer_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.322494402299235, 'std': 0.638438643694705}, annotation_units=1, channel_map={'40009_Reported-occurrences-of-cancer_0_0': 0, })
ukb_40425_0 = TensorMap('40425_Volume-of-DNA-held-by-UKB_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 835.59599561966, 'std': 49.91770948969756}, annotation_units=1, channel_map={'40425_Volume-of-DNA-held-by-UKB_0_0': 0, })
ukb_40425_1 = TensorMap('40425_Volume-of-DNA-held-by-UKB_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 845.0927419354841, 'std': 55.10794952593842}, annotation_units=1, channel_map={'40425_Volume-of-DNA-held-by-UKB_1_0': 0, })
ukb_41076_0 = TensorMap('41076_Episodes-containing-Diagnoses-secondary-ICD9-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.313970841464814, 'std': 3.1620287662073925}, annotation_units=1, channel_map={'41076_Episodes-containing-Diagnoses-secondary-ICD9-data_0_0': 0, })
ukb_41078_0 = TensorMap('41078_Episodes-containing-Diagnoses-secondary-ICD10-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.851323793623393, 'std': 42.31535192479268}, annotation_units=1, channel_map={'41078_Episodes-containing-Diagnoses-secondary-ICD10-data_0_0': 0, })
ukb_41079_0 = TensorMap('41079_Episodes-containing-Diagnoses-secondary-ICD10-addendum-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.9076533839342202, 'std': 10.227108478165237}, annotation_units=1, channel_map={'41079_Episodes-containing-Diagnoses-secondary-ICD10-addendum-data_0_0': 0, })
ukb_41080_0 = TensorMap('41080_Episodes-containing-Operative-procedures-secondary-OPCS4-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.69780814909158, 'std': 12.681988097428164}, annotation_units=1, channel_map={'41080_Episodes-containing-Operative-procedures-secondary-OPCS4-data_0_0': 0, })
ukb_41082_0 = TensorMap('41082_Episodes-containing-Dates-of-operations-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.7053936868865165, 'std': 12.844337680824585}, annotation_units=1, channel_map={'41082_Episodes-containing-Dates-of-operations-data_0_0': 0, })
ukb_41083_0 = TensorMap('41083_Episodes-containing-Episode-start-date-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.704735980721438, 'std': 15.79404559278517}, annotation_units=1, channel_map={'41083_Episodes-containing-Episode-start-date-data_0_0': 0, })
ukb_41084_0 = TensorMap('41084_Episodes-containing-Episode-end-date-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.7039982048079905, 'std': 15.793277411726411}, annotation_units=1, channel_map={'41084_Episodes-containing-Episode-end-date-data_0_0': 0, })
ukb_41085_0 = TensorMap('41085_Episodes-containing-Duration-of-episode-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.782855498903626, 'std': 15.821459717550045}, annotation_units=1, channel_map={'41085_Episodes-containing-Duration-of-episode-data_0_0': 0, })
ukb_41086_0 = TensorMap('41086_Episodes-containing-Episode-order-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.711730812466996, 'std': 16.03313567386901}, annotation_units=1, channel_map={'41086_Episodes-containing-Episode-order-data_0_0': 0, })
ukb_41087_0 = TensorMap('41087_Episodes-containing-Episode-status-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.653981042524684, 'std': 16.137775500554614}, annotation_units=1, channel_map={'41087_Episodes-containing-Episode-status-data_0_0': 0, })
ukb_41088_0 = TensorMap('41088_Episodes-containing-Duration-of-episode-within-HES-data-year-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.653806053658669, 'std': 16.13782806078228}, annotation_units=1, channel_map={'41088_Episodes-containing-Duration-of-episode-within-HES-data-year-data_0_0': 0, })
ukb_41089_0 = TensorMap('41089_Episodes-containing-Beginning-of-spell-indicator-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.654014427210276, 'std': 16.13793801756679}, annotation_units=1, channel_map={'41089_Episodes-containing-Beginning-of-spell-indicator-data_0_0': 0, })
ukb_41090_0 = TensorMap('41090_Episodes-containing-End-of-spell-indicator-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.653981042524684, 'std': 16.137775500554614}, annotation_units=1, channel_map={'41090_Episodes-containing-End-of-spell-indicator-data_0_0': 0, })
ukb_41091_0 = TensorMap('41091_Episodes-containing-Duration-of-spell-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.284067998713457, 'std': 15.468551649881961}, annotation_units=1, channel_map={'41091_Episodes-containing-Duration-of-spell-data_0_0': 0, })
ukb_41092_0 = TensorMap('41092_Episodes-containing-Current-PCT-responsible-for-patient-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.470361569183401, 'std': 14.456057430461586}, annotation_units=1, channel_map={'41092_Episodes-containing-Current-PCT-responsible-for-patient-data_0_0': 0, })
ukb_41093_0 = TensorMap('41093_Episodes-containing-PCT-where-patients-GP-was-registered-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.653970125464318, 'std': 16.137768670664578}, annotation_units=1, channel_map={'41093_Episodes-containing-PCT-where-patients-GP-was-registered-data_0_0': 0, })
ukb_41094_0 = TensorMap('41094_Episodes-containing-Administrative-and-legal-status-of-patient-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.2281050017832262, 'std': 3.632830811250619}, annotation_units=1, channel_map={'41094_Episodes-containing-Administrative-and-legal-status-of-patient-data_0_0': 0, })
ukb_41095_0 = TensorMap('41095_Episodes-containing-Date-of-decision-to-admit-to-hospital-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.043723843006867, 'std': 13.762791499857013}, annotation_units=1, channel_map={'41095_Episodes-containing-Date-of-decision-to-admit-to-hospital-data_0_0': 0, })
ukb_41096_0 = TensorMap('41096_Episodes-containing-Date-of-admission-to-hospital-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.511724623943542, 'std': 15.664253231461565}, annotation_units=1, channel_map={'41096_Episodes-containing-Date-of-admission-to-hospital-data_0_0': 0, })
ukb_41097_0 = TensorMap('41097_Episodes-containing-Waiting-time-for-hospital-admission-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.617979766799029, 'std': 9.896810076158262}, annotation_units=1, channel_map={'41097_Episodes-containing-Waiting-time-for-hospital-admission-data_0_0': 0, })
ukb_41098_0 = TensorMap('41098_Episodes-containing-Method-of-admission-to-hospital-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.783512108091641, 'std': 15.822041903378432}, annotation_units=1, channel_map={'41098_Episodes-containing-Method-of-admission-to-hospital-data-polymorphic_0_0': 0, })
ukb_41099_0 = TensorMap('41099_Episodes-containing-Source-of-admission-to-hospital-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.783473427954618, 'std': 15.822042138052467}, annotation_units=1, channel_map={'41099_Episodes-containing-Source-of-admission-to-hospital-data-polymorphic_0_0': 0, })
ukb_41100_0 = TensorMap('41100_Episodes-containing-First-regular-day-or-night-of-admission-to-hosp-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.7123678882552875, 'std': 14.701539208428704}, annotation_units=1, channel_map={'41100_Episodes-containing-First-regular-day-or-night-of-admission-to-hosp-data_0_0': 0, })
ukb_41101_0 = TensorMap('41101_Episodes-containing-Date-of-discharge-from-hospital-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.142465719426689, 'std': 15.170703160611824}, annotation_units=1, channel_map={'41101_Episodes-containing-Date-of-discharge-from-hospital-data_0_0': 0, })
ukb_41102_0 = TensorMap('41102_Episodes-containing-Method-of-discharge-from-hospital-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.7834747985601185, 'std': 15.822013108027429}, annotation_units=1, channel_map={'41102_Episodes-containing-Method-of-discharge-from-hospital-data-polymorphic_0_0': 0, })
ukb_41103_0 = TensorMap('41103_Episodes-containing-Destination-on-discharge-from-hospital-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.704589080354638, 'std': 15.793837358474429}, annotation_units=1, channel_map={'41103_Episodes-containing-Destination-on-discharge-from-hospital-data-polymorphic_0_0': 0, })
ukb_41104_0 = TensorMap('41104_Episodes-containing-External-cause-ICD10-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.597188335076101, 'std': 1.6128953009115339}, annotation_units=1, channel_map={'41104_Episodes-containing-External-cause-ICD10-data_0_0': 0, })
ukb_41105_0 = TensorMap('41105_Episodes-containing-External-cause-ICD10-addendum-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.1807909604519768, 'std': 0.5012183410175326}, annotation_units=1, channel_map={'41105_Episodes-containing-External-cause-ICD10-addendum-data_0_0': 0, })
ukb_41106_0 = TensorMap('41106_Episodes-containing-Operation-status-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.495564409854112, 'std': 14.267147028675097}, annotation_units=1, channel_map={'41106_Episodes-containing-Operation-status-data_0_0': 0, })
ukb_41107_0 = TensorMap('41107_Episodes-containing-Duration-of-postoperative-stay-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.284227340943201, 'std': 13.95827867859391}, annotation_units=1, channel_map={'41107_Episodes-containing-Duration-of-postoperative-stay-data_0_0': 0, })
ukb_41108_0 = TensorMap('41108_Episodes-containing-Duration-of-preoperative-stay-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.284736156981339, 'std': 13.960304410797844}, annotation_units=1, channel_map={'41108_Episodes-containing-Duration-of-preoperative-stay-data_0_0': 0, })
ukb_41109_0 = TensorMap('41109_Episodes-containing-Patient-classification-on-admission-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.704775168538228, 'std': 15.793942180331861}, annotation_units=1, channel_map={'41109_Episodes-containing-Patient-classification-on-admission-data-polymorphic_0_0': 0, })
ukb_41110_0 = TensorMap('41110_Episodes-containing-Intended-management-of-patient-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.711635543215977, 'std': 16.03317641452055}, annotation_units=1, channel_map={'41110_Episodes-containing-Intended-management-of-patient-data-polymorphic_0_0': 0, })
ukb_41111_0 = TensorMap('41111_Episodes-containing-Main-speciality-of-consultant-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.711742330165618, 'std': 16.033265618616817}, annotation_units=1, channel_map={'41111_Episodes-containing-Main-speciality-of-consultant-data-polymorphic_0_0': 0, })
ukb_41112_0 = TensorMap('41112_Episodes-containing-Treatment-speciality-of-consultant-data-polymorphic_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.783491721201893, 'std': 15.822069033321933}, annotation_units=1, channel_map={'41112_Episodes-containing-Treatment-speciality-of-consultant-data-polymorphic_0_0': 0, })
ukb_41132_0 = TensorMap('41132_Episodes-containing-Episode-type-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.653981042524682, 'std': 16.137775500554614}, annotation_units=1, channel_map={'41132_Episodes-containing-Episode-type-data_0_0': 0, })
ukb_41142_0 = TensorMap('41142_Episodes-containing-Diagnoses-main-ICD10-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.693771950990641, 'std': 15.775197738988686}, annotation_units=1, channel_map={'41142_Episodes-containing-Diagnoses-main-ICD10-data_0_0': 0, })
ukb_41143_0 = TensorMap('41143_Episodes-containing-Diagnoses-main-ICD10-addendum-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.9050652107976949, 'std': 3.3783148684541815}, annotation_units=1, channel_map={'41143_Episodes-containing-Diagnoses-main-ICD10-addendum-data_0_0': 0, })
ukb_41144_0 = TensorMap('41144_Episodes-containing-Diagnoses-main-ICD9-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5641594010147277, 'std': 3.1467841304346864}, annotation_units=1, channel_map={'41144_Episodes-containing-Diagnoses-main-ICD9-data_0_0': 0, })
ukb_41146_0 = TensorMap('41146_Episodes-containing-Operative-procedure-main-OPCS4-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.5398056217106975, 'std': 15.62292252188927}, annotation_units=1, channel_map={'41146_Episodes-containing-Operative-procedure-main-OPCS4-data_0_0': 0, })
ukb_41148_0 = TensorMap('41148_Episodes-containing-Date-of-operation-data_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.339758636044593, 'std': 13.586225369016747}, annotation_units=1, channel_map={'41148_Episodes-containing-Date-of-operation-data_0_0': 0, })
ukb_90012_0 = TensorMap('90012_Overall-acceleration-average_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.42347162351126, 'std': 13.995464309516002}, annotation_units=1, channel_map={'90012_Overall-acceleration-average_0_0': 0, })
ukb_90013_0 = TensorMap('90013_Standard-deviation-of-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 53.71698992236849, 'std': 28.682196402063585}, annotation_units=1, channel_map={'90013_Standard-deviation-of-acceleration_0_0': 0, })
ukb_90019_0 = TensorMap('90019_Monday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.969843676165677, 'std': 17.154154150610005}, annotation_units=1, channel_map={'90019_Monday-average-acceleration_0_0': 0, })
ukb_90020_0 = TensorMap('90020_Tuesday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.229667582814987, 'std': 14.185801237575442}, annotation_units=1, channel_map={'90020_Tuesday-average-acceleration_0_0': 0, })
ukb_90021_0 = TensorMap('90021_Wednesday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.07040059790733, 'std': 15.968964215997316}, annotation_units=1, channel_map={'90021_Wednesday-average-acceleration_0_0': 0, })
ukb_90022_0 = TensorMap('90022_Thursday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.076211582043495, 'std': 13.197469717161153}, annotation_units=1, channel_map={'90022_Thursday-average-acceleration_0_0': 0, })
ukb_90023_0 = TensorMap('90023_Friday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.135146149766136, 'std': 12.890072457932929}, annotation_units=1, channel_map={'90023_Friday-average-acceleration_0_0': 0, })
ukb_90024_0 = TensorMap('90024_Saturday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.196507353295722, 'std': 16.948090391736685}, annotation_units=1, channel_map={'90024_Saturday-average-acceleration_0_0': 0, })
ukb_90025_0 = TensorMap('90025_Sunday-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.36648054390279, 'std': 40.20427495270532}, annotation_units=1, channel_map={'90025_Sunday-average-acceleration_0_0': 0, })
ukb_90027_0 = TensorMap('90027_Average-acceleration-0000-0059_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7.106457881286464, 'std': 12.689775430639592}, annotation_units=1, channel_map={'90027_Average-acceleration-0000-0059_0_0': 0, })
ukb_90028_0 = TensorMap('90028_Average-acceleration-0100-0159_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 4.532738511982257, 'std': 11.430339221271472}, annotation_units=1, channel_map={'90028_Average-acceleration-0100-0159_0_0': 0, })
ukb_90029_0 = TensorMap('90029_Average-acceleration-0200-0259_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.7740854428853843, 'std': 12.794841854591494}, annotation_units=1, channel_map={'90029_Average-acceleration-0200-0259_0_0': 0, })
ukb_90030_0 = TensorMap('90030_Average-acceleration-0300-0359_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.5407789189449828, 'std': 10.602845467400767}, annotation_units=1, channel_map={'90030_Average-acceleration-0300-0359_0_0': 0, })
ukb_90031_0 = TensorMap('90031_Average-acceleration-0400-0459_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.7459988427600175, 'std': 12.657702371897654}, annotation_units=1, channel_map={'90031_Average-acceleration-0400-0459_0_0': 0, })
ukb_90032_0 = TensorMap('90032_Average-acceleration-0500-0559_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.277517238053909, 'std': 13.172113120629314}, annotation_units=1, channel_map={'90032_Average-acceleration-0500-0559_0_0': 0, })
ukb_90033_0 = TensorMap('90033_Average-acceleration-0600-0659_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 12.450125560538115, 'std': 20.904985053301118}, annotation_units=1, channel_map={'90033_Average-acceleration-0600-0659_0_0': 0, })
ukb_90034_0 = TensorMap('90034_Average-acceleration-0700-0759_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 26.661664400405034, 'std': 27.91514099760778}, annotation_units=1, channel_map={'90034_Average-acceleration-0700-0759_0_0': 0, })
ukb_90035_0 = TensorMap('90035_Average-acceleration-0800-0859_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.65167510487486, 'std': 27.94983269121508}, annotation_units=1, channel_map={'90035_Average-acceleration-0800-0859_0_0': 0, })
ukb_90036_0 = TensorMap('90036_Average-acceleration-0900-0959_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.857465837311366, 'std': 29.639430751174306}, annotation_units=1, channel_map={'90036_Average-acceleration-0900-0959_0_0': 0, })
ukb_90037_0 = TensorMap('90037_Average-acceleration-1000-1059_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 48.647600270022664, 'std': 26.850548452754857}, annotation_units=1, channel_map={'90037_Average-acceleration-1000-1059_0_0': 0, })
ukb_90038_0 = TensorMap('90038_Average-acceleration-1100-1159_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 47.493301219923836, 'std': 25.308289932762253}, annotation_units=1, channel_map={'90038_Average-acceleration-1100-1159_0_0': 0, })
ukb_90039_0 = TensorMap('90039_Average-acceleration-1200-1259_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 46.324744683928834, 'std': 24.465856886866412}, annotation_units=1, channel_map={'90039_Average-acceleration-1200-1259_0_0': 0, })
ukb_90040_0 = TensorMap('90040_Average-acceleration-1300-1359_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.95905376344084, 'std': 22.344195193642026}, annotation_units=1, channel_map={'90040_Average-acceleration-1300-1359_0_0': 0, })
ukb_90041_0 = TensorMap('90041_Average-acceleration-1400-1459_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 42.27841101306717, 'std': 23.11589589343296}, annotation_units=1, channel_map={'90041_Average-acceleration-1400-1459_0_0': 0, })
ukb_90042_0 = TensorMap('90042_Average-acceleration-1500-1559_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 41.38379179323976, 'std': 24.118297411291845}, annotation_units=1, channel_map={'90042_Average-acceleration-1500-1559_0_0': 0, })
ukb_90043_0 = TensorMap('90043_Average-acceleration-1600-1659_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.977998167703376, 'std': 24.230618393642825}, annotation_units=1, channel_map={'90043_Average-acceleration-1600-1659_0_0': 0, })
ukb_90044_0 = TensorMap('90044_Average-acceleration-1700-1759_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 39.299199961425344, 'std': 24.07987078886049}, annotation_units=1, channel_map={'90044_Average-acceleration-1700-1759_0_0': 0, })
ukb_90045_0 = TensorMap('90045_Average-acceleration-1800-1859_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 37.70637388495105, 'std': 24.526134967495604}, annotation_units=1, channel_map={'90045_Average-acceleration-1800-1859_0_0': 0, })
ukb_90046_0 = TensorMap('90046_Average-acceleration-1900-1959_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 32.83279251651477, 'std': 24.44931196678211}, annotation_units=1, channel_map={'90046_Average-acceleration-1900-1959_0_0': 0, })
ukb_90047_0 = TensorMap('90047_Average-acceleration-2000-2059_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25.389408843242204, 'std': 19.687511042542315}, annotation_units=1, channel_map={'90047_Average-acceleration-2000-2059_0_0': 0, })
ukb_90048_0 = TensorMap('90048_Average-acceleration-2100-2159_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.541821013549352, 'std': 17.39582659438696}, annotation_units=1, channel_map={'90048_Average-acceleration-2100-2159_0_0': 0, })
ukb_90049_0 = TensorMap('90049_Average-acceleration-2200-2259_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.07180317276628, 'std': 16.71302784678877}, annotation_units=1, channel_map={'90049_Average-acceleration-2200-2259_0_0': 0, })
ukb_90050_0 = TensorMap('90050_Average-acceleration-2300-2359_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.353390616712476, 'std': 13.591936624266381}, annotation_units=1, channel_map={'90050_Average-acceleration-2300-2359_0_0': 0, })
ukb_90051_0 = TensorMap('90051_Wear-duration-overall_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.29415979555427, 'std': 1.5294148761434958}, annotation_units=1, channel_map={'90051_Wear-duration-overall_0_0': 0, })
ukb_90052_0 = TensorMap('90052_Non-wear-duration-overall_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5593253290901201, 'std': 1.3194473341425854}, annotation_units=1, channel_map={'90052_Non-wear-duration-overall_0_0': 0, })
ukb_90053_0 = TensorMap('90053_Wear-duration-during-Monday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.585114036356618, 'std': 5.938389363274724}, annotation_units=1, channel_map={'90053_Wear-duration-during-Monday_0_0': 0, })
ukb_90054_0 = TensorMap('90054_Wear-duration-during-Tuesday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.970947297362457, 'std': 5.839235323860917}, annotation_units=1, channel_map={'90054_Wear-duration-during-Tuesday_0_0': 0, })
ukb_90055_0 = TensorMap('90055_Wear-duration-during-Wednesday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.764196537923713, 'std': 5.849517437390719}, annotation_units=1, channel_map={'90055_Wear-duration-during-Wednesday_0_0': 0, })
ukb_90056_0 = TensorMap('90056_Wear-duration-during-Thursday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.634715849365932, 'std': 5.9951926261779365}, annotation_units=1, channel_map={'90056_Wear-duration-during-Thursday_0_0': 0, })
ukb_90057_0 = TensorMap('90057_Wear-duration-during-Friday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.490339553498245, 'std': 6.00413178592472}, annotation_units=1, channel_map={'90057_Wear-duration-during-Friday_0_0': 0, })
ukb_90058_0 = TensorMap('90058_Wear-duration-during-Saturday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.302158831187626, 'std': 6.222934862981404}, annotation_units=1, channel_map={'90058_Wear-duration-during-Saturday_0_0': 0, })
ukb_90059_0 = TensorMap('90059_Wear-duration-during-Sunday_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 21.295139881382898, 'std': 6.777737389457544}, annotation_units=1, channel_map={'90059_Wear-duration-during-Sunday_0_0': 0, })
ukb_90060_0 = TensorMap('90060_Wear-duration-during-0000-0059_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.133015767394764, 'std': 1.669471265351599}, annotation_units=1, channel_map={'90060_Wear-duration-during-0000-0059_0_0': 0, })
ukb_90061_0 = TensorMap('90061_Wear-duration-during-0100-0159_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.123179709725638, 'std': 1.682555042465214}, annotation_units=1, channel_map={'90061_Wear-duration-during-0100-0159_0_0': 0, })
ukb_90062_0 = TensorMap('90062_Wear-duration-during-0200-0259_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.157569892473119, 'std': 1.6878175385530005}, annotation_units=1, channel_map={'90062_Wear-duration-during-0200-0259_0_0': 0, })
ukb_90063_0 = TensorMap('90063_Wear-duration-during-0300-0359_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.174573894594726, 'std': 1.69128563285464}, annotation_units=1, channel_map={'90063_Wear-duration-during-0300-0359_0_0': 0, })
ukb_90064_0 = TensorMap('90064_Wear-duration-during-0400-0459_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.187488499927674, 'std': 1.693640774697293}, annotation_units=1, channel_map={'90064_Wear-duration-during-0400-0459_0_0': 0, })
ukb_90065_0 = TensorMap('90065_Wear-duration-during-0500-0559_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.2065316553353576, 'std': 1.687248785574505}, annotation_units=1, channel_map={'90065_Wear-duration-during-0500-0559_0_0': 0, })
ukb_90066_0 = TensorMap('90066_Wear-duration-during-0600-0659_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.231038526447756, 'std': 1.6555288367090255}, annotation_units=1, channel_map={'90066_Wear-duration-during-0600-0659_0_0': 0, })
ukb_90067_0 = TensorMap('90067_Wear-duration-during-0700-0759_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.251128694729738, 'std': 1.5970561147391025}, annotation_units=1, channel_map={'90067_Wear-duration-during-0700-0759_0_0': 0, })
ukb_90068_0 = TensorMap('90068_Wear-duration-during-0800-0859_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.236876898596845, 'std': 1.5573282783522373}, annotation_units=1, channel_map={'90068_Wear-duration-during-0800-0859_0_0': 0, })
ukb_90069_0 = TensorMap('90069_Wear-duration-during-0900-0959_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.240367520131153, 'std': 1.5347165401031984}, annotation_units=1, channel_map={'90069_Wear-duration-during-0900-0959_0_0': 0, })
ukb_90070_0 = TensorMap('90070_Wear-duration-during-1000-1059_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.406913158782969, 'std': 1.5052166238840226}, annotation_units=1, channel_map={'90070_Wear-duration-during-1000-1059_0_0': 0, })
ukb_90071_0 = TensorMap('90071_Wear-duration-during-1100-1159_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.3789960943150605, 'std': 1.518017034790592}, annotation_units=1, channel_map={'90071_Wear-duration-during-1100-1159_0_0': 0, })
ukb_90072_0 = TensorMap('90072_Wear-duration-during-1200-1259_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.373508365880707, 'std': 1.528598180240083}, annotation_units=1, channel_map={'90072_Wear-duration-during-1200-1259_0_0': 0, })
ukb_90073_0 = TensorMap('90073_Wear-duration-during-1300-1359_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.368636192680459, 'std': 1.5356541955859178}, annotation_units=1, channel_map={'90073_Wear-duration-during-1300-1359_0_0': 0, })
ukb_90074_0 = TensorMap('90074_Wear-duration-during-1400-1459_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.368456724046485, 'std': 1.5369328079538038}, annotation_units=1, channel_map={'90074_Wear-duration-during-1400-1459_0_0': 0, })
ukb_90075_0 = TensorMap('90075_Wear-duration-during-1500-1559_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.370889242489996, 'std': 1.5392009728803167}, annotation_units=1, channel_map={'90075_Wear-duration-during-1500-1559_0_0': 0, })
ukb_90076_0 = TensorMap('90076_Wear-duration-during-1600-1659_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.3810393943777415, 'std': 1.5343787540493092}, annotation_units=1, channel_map={'90076_Wear-duration-during-1600-1659_0_0': 0, })
ukb_90077_0 = TensorMap('90077_Wear-duration-during-1700-1759_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.393261295144414, 'std': 1.522967702780382}, annotation_units=1, channel_map={'90077_Wear-duration-during-1700-1759_0_0': 0, })
ukb_90078_0 = TensorMap('90078_Wear-duration-during-1800-1859_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.397466223058007, 'std': 1.5182796325322336}, annotation_units=1, channel_map={'90078_Wear-duration-during-1800-1859_0_0': 0, })
ukb_90079_0 = TensorMap('90079_Wear-duration-during-1900-1959_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.386784030088239, 'std': 1.5313735219461648}, annotation_units=1, channel_map={'90079_Wear-duration-during-1900-1959_0_0': 0, })
ukb_90080_0 = TensorMap('90080_Wear-duration-during-2000-2059_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.376575437581366, 'std': 1.5428021577386761}, annotation_units=1, channel_map={'90080_Wear-duration-during-2000-2059_0_0': 0, })
ukb_90081_0 = TensorMap('90081_Wear-duration-during-2100-2159_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.360730411302377, 'std': 1.5580538235743924}, annotation_units=1, channel_map={'90081_Wear-duration-during-2100-2159_0_0': 0, })
ukb_90082_0 = TensorMap('90082_Wear-duration-during-2200-2259_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.317136602536286, 'std': 1.5791490404943695}, annotation_units=1, channel_map={'90082_Wear-duration-during-2200-2259_0_0': 0, })
ukb_90083_0 = TensorMap('90083_Wear-duration-during-2300-2359_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.209585611649548, 'std': 1.6306264000655888}, annotation_units=1, channel_map={'90083_Wear-duration-during-2300-2359_0_0': 0, })
ukb_90084_0 = TensorMap('90084_Unique-hours-of-wear-in-a-24-hour-cycle-scattered-over-multiple-days_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 23.517768455566802, 'std': 2.6777486507372092}, annotation_units=1, channel_map={'90084_Unique-hours-of-wear-in-a-24-hour-cycle-scattered-over-multiple-days_0_0': 0, })
ukb_90085_0 = TensorMap('90085_Unique-minutes-of-wear-in-a-24-hour-cycle-scattered-over-multiple-days_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1399.4772457688414, 'std': 211.1790901048031}, annotation_units=1, channel_map={'90085_Unique-minutes-of-wear-in-a-24-hour-cycle-scattered-over-multiple-days_0_0': 0, })
ukb_90086_0 = TensorMap('90086_Non-wear-episodes-over-1-hour-duration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.519803269202952, 'std': 3.8033126137708777}, annotation_units=1, channel_map={'90086_Non-wear-episodes-over-1-hour-duration_0_0': 0, })
ukb_90087_0 = TensorMap('90087_Nowear-time-bias-adjusted-average-acceleration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 27.558812768214477, 'std': 14.108189524346061}, annotation_units=1, channel_map={'90087_Nowear-time-bias-adjusted-average-acceleration_0_0': 0, })
ukb_90088_0 = TensorMap('90088_Nowear-time-bias-adjusted-acceleration-standard-deviation_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 54.832495684459204, 'std': 29.17821840388565}, annotation_units=1, channel_map={'90088_Nowear-time-bias-adjusted-acceleration-standard-deviation_0_0': 0, })
ukb_90089_0 = TensorMap('90089_Nowear-time-bias-adjusted-acceleration-median_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 5.529378176382659, 'std': 5.919172559006304}, annotation_units=1, channel_map={'90089_Nowear-time-bias-adjusted-acceleration-median_0_0': 0, })
ukb_90090_0 = TensorMap('90090_Nowear-time-bias-adjusted-acceleration-minimum_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.06461314431746948, 'std': 2.310496983099447}, annotation_units=1, channel_map={'90090_Nowear-time-bias-adjusted-acceleration-minimum_0_0': 0, })
ukb_90091_0 = TensorMap('90091_Nowear-time-bias-adjusted-acceleration-maximum_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1575.5054467428517, 'std': 786.3941609034839}, annotation_units=1, channel_map={'90091_Nowear-time-bias-adjusted-acceleration-maximum_0_0': 0, })
ukb_90092_0 = TensorMap('90092_Fraction-acceleration-1-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.19363048362987595, 'std': 0.08378851179642582}, annotation_units=1, channel_map={'90092_Fraction-acceleration-1-milligravities_0_0': 0, })
ukb_90093_0 = TensorMap('90093_Fraction-acceleration-2-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.30898027870196243, 'std': 0.08800661058556765}, annotation_units=1, channel_map={'90093_Fraction-acceleration-2-milligravities_0_0': 0, })
ukb_90094_0 = TensorMap('90094_Fraction-acceleration-3-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.39714024784222973, 'std': 0.09338600035939268}, annotation_units=1, channel_map={'90094_Fraction-acceleration-3-milligravities_0_0': 0, })
ukb_90095_0 = TensorMap('90095_Fraction-acceleration-4-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.463467949274314, 'std': 0.09662230798179156}, annotation_units=1, channel_map={'90095_Fraction-acceleration-4-milligravities_0_0': 0, })
ukb_90096_0 = TensorMap('90096_Fraction-acceleration-5-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5120027966632913, 'std': 0.0979423196999406}, annotation_units=1, channel_map={'90096_Fraction-acceleration-5-milligravities_0_0': 0, })
ukb_90097_0 = TensorMap('90097_Fraction-acceleration-6-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.546760036645933, 'std': 0.09760498520981384}, annotation_units=1, channel_map={'90097_Fraction-acceleration-6-milligravities_0_0': 0, })
ukb_90098_0 = TensorMap('90098_Fraction-acceleration-7-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5718037128116109, 'std': 0.09658138066480366}, annotation_units=1, channel_map={'90098_Fraction-acceleration-7-milligravities_0_0': 0, })
ukb_90099_0 = TensorMap('90099_Fraction-acceleration-8-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.5905090987993639, 'std': 0.09527670148745399}, annotation_units=1, channel_map={'90099_Fraction-acceleration-8-milligravities_0_0': 0, })
ukb_90100_0 = TensorMap('90100_Fraction-acceleration-9-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6052221129273347, 'std': 0.09409262987835801}, annotation_units=1, channel_map={'90100_Fraction-acceleration-9-milligravities_0_0': 0, })
ukb_90101_0 = TensorMap('90101_Fraction-acceleration-10-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6173177009499011, 'std': 0.09295317946174637}, annotation_units=1, channel_map={'90101_Fraction-acceleration-10-milligravities_0_0': 0, })
ukb_90102_0 = TensorMap('90102_Fraction-acceleration-11-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6277743574907179, 'std': 0.09177480667842472}, annotation_units=1, channel_map={'90102_Fraction-acceleration-11-milligravities_0_0': 0, })
ukb_90103_0 = TensorMap('90103_Fraction-acceleration-12-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6370736101065624, 'std': 0.0907544385318843}, annotation_units=1, channel_map={'90103_Fraction-acceleration-12-milligravities_0_0': 0, })
ukb_90104_0 = TensorMap('90104_Fraction-acceleration-13-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6455888037031677, 'std': 0.08972766566548142}, annotation_units=1, channel_map={'90104_Fraction-acceleration-13-milligravities_0_0': 0, })
ukb_90105_0 = TensorMap('90105_Fraction-acceleration-14-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.653550653358407, 'std': 0.08873358582581779}, annotation_units=1, channel_map={'90105_Fraction-acceleration-14-milligravities_0_0': 0, })
ukb_90106_0 = TensorMap('90106_Fraction-acceleration-15-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6611008245334878, 'std': 0.08772987877780972}, annotation_units=1, channel_map={'90106_Fraction-acceleration-15-milligravities_0_0': 0, })
ukb_90107_0 = TensorMap('90107_Fraction-acceleration-16-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6683853705578862, 'std': 0.0867126863188436}, annotation_units=1, channel_map={'90107_Fraction-acceleration-16-milligravities_0_0': 0, })
ukb_90108_0 = TensorMap('90108_Fraction-acceleration-17-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6754435797290131, 'std': 0.08572298347738518}, annotation_units=1, channel_map={'90108_Fraction-acceleration-17-milligravities_0_0': 0, })
ukb_90109_0 = TensorMap('90109_Fraction-acceleration-18-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6822708713052704, 'std': 0.0848315599265804}, annotation_units=1, channel_map={'90109_Fraction-acceleration-18-milligravities_0_0': 0, })
ukb_90110_0 = TensorMap('90110_Fraction-acceleration-19-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6888709773856024, 'std': 0.08391288170980778}, annotation_units=1, channel_map={'90110_Fraction-acceleration-19-milligravities_0_0': 0, })
ukb_90111_0 = TensorMap('90111_Fraction-acceleration-20-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.6953570085346447, 'std': 0.08294789744247397}, annotation_units=1, channel_map={'90111_Fraction-acceleration-20-milligravities_0_0': 0, })
ukb_90112_0 = TensorMap('90112_Fraction-acceleration-25-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7252989633058489, 'std': 0.0788696380170397}, annotation_units=1, channel_map={'90112_Fraction-acceleration-25-milligravities_0_0': 0, })
ukb_90113_0 = TensorMap('90113_Fraction-acceleration-30-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7515875403828535, 'std': 0.07505707262659325}, annotation_units=1, channel_map={'90113_Fraction-acceleration-30-milligravities_0_0': 0, })
ukb_90114_0 = TensorMap('90114_Fraction-acceleration-35-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.7745163894112541, 'std': 0.0713324070987888}, annotation_units=1, channel_map={'90114_Fraction-acceleration-35-milligravities_0_0': 0, })
ukb_90115_0 = TensorMap('90115_Fraction-acceleration-40-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.794562090746902, 'std': 0.06806537892803469}, annotation_units=1, channel_map={'90115_Fraction-acceleration-40-milligravities_0_0': 0, })
ukb_90116_0 = TensorMap('90116_Fraction-acceleration-45-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8122575437581373, 'std': 0.0650123811140275}, annotation_units=1, channel_map={'90116_Fraction-acceleration-45-milligravities_0_0': 0, })
ukb_90117_0 = TensorMap('90117_Fraction-acceleration-50-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8281616567819083, 'std': 0.061961534132324504}, annotation_units=1, channel_map={'90117_Fraction-acceleration-50-milligravities_0_0': 0, })
ukb_90118_0 = TensorMap('90118_Fraction-acceleration-55-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8425518202420558, 'std': 0.05886666618434548}, annotation_units=1, channel_map={'90118_Fraction-acceleration-55-milligravities_0_0': 0, })
ukb_90119_0 = TensorMap('90119_Fraction-acceleration-60-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8555916968031243, 'std': 0.05605786106696202}, annotation_units=1, channel_map={'90119_Fraction-acceleration-60-milligravities_0_0': 0, })
ukb_90120_0 = TensorMap('90120_Fraction-acceleration-65-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8675027629104587, 'std': 0.0532510533854409}, annotation_units=1, channel_map={'90120_Fraction-acceleration-65-milligravities_0_0': 0, })
ukb_90121_0 = TensorMap('90121_Fraction-acceleration-70-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8783745985823813, 'std': 0.05056520200794273}, annotation_units=1, channel_map={'90121_Fraction-acceleration-70-milligravities_0_0': 0, })
ukb_90122_0 = TensorMap('90122_Fraction-acceleration-75-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8882932831862671, 'std': 0.048040170612471324}, annotation_units=1, channel_map={'90122_Fraction-acceleration-75-milligravities_0_0': 0, })
ukb_90123_0 = TensorMap('90123_Fraction-acceleration-80-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8973672308211582, 'std': 0.04554887887696288}, annotation_units=1, channel_map={'90123_Fraction-acceleration-80-milligravities_0_0': 0, })
ukb_90124_0 = TensorMap('90124_Fraction-acceleration-85-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9056765996431843, 'std': 0.04289798097062886}, annotation_units=1, channel_map={'90124_Fraction-acceleration-85-milligravities_0_0': 0, })
ukb_90125_0 = TensorMap('90125_Fraction-acceleration-90-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.913222537248662, 'std': 0.04063906775147445}, annotation_units=1, channel_map={'90125_Fraction-acceleration-90-milligravities_0_0': 0, })
ukb_90126_0 = TensorMap('90126_Fraction-acceleration-95-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9201158590095952, 'std': 0.038505290499785585}, annotation_units=1, channel_map={'90126_Fraction-acceleration-95-milligravities_0_0': 0, })
ukb_90127_0 = TensorMap('90127_Fraction-acceleration-100-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9264224408119967, 'std': 0.03641961072561103}, annotation_units=1, channel_map={'90127_Fraction-acceleration-100-milligravities_0_0': 0, })
ukb_90128_0 = TensorMap('90128_Fraction-acceleration-125-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9507029557837892, 'std': 0.02784310358608152}, annotation_units=1, channel_map={'90128_Fraction-acceleration-125-milligravities_0_0': 0, })
ukb_90129_0 = TensorMap('90129_Fraction-acceleration-150-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9664515646848932, 'std': 0.0215438068876052}, annotation_units=1, channel_map={'90129_Fraction-acceleration-150-milligravities_0_0': 0, })
ukb_90130_0 = TensorMap('90130_Fraction-acceleration-175-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9767494768310914, 'std': 0.016886568885961222}, annotation_units=1, channel_map={'90130_Fraction-acceleration-175-milligravities_0_0': 0, })
ukb_90131_0 = TensorMap('90131_Fraction-acceleration-200-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9834226529726603, 'std': 0.01350380120921056}, annotation_units=1, channel_map={'90131_Fraction-acceleration-200-milligravities_0_0': 0, })
ukb_90132_0 = TensorMap('90132_Fraction-acceleration-225-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9877418004725398, 'std': 0.011152215531305685}, annotation_units=1, channel_map={'90132_Fraction-acceleration-225-milligravities_0_0': 0, })
ukb_90133_0 = TensorMap('90133_Fraction-acceleration-250-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9905989488403495, 'std': 0.009523163978600962}, annotation_units=1, channel_map={'90133_Fraction-acceleration-250-milligravities_0_0': 0, })
ukb_90134_0 = TensorMap('90134_Fraction-acceleration-275-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9925612131732481, 'std': 0.008376066923715682}, annotation_units=1, channel_map={'90134_Fraction-acceleration-275-milligravities_0_0': 0, })
ukb_90135_0 = TensorMap('90135_Fraction-acceleration-300-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9939693234967932, 'std': 0.007534428867424408}, annotation_units=1, channel_map={'90135_Fraction-acceleration-300-milligravities_0_0': 0, })
ukb_90136_0 = TensorMap('90136_Fraction-acceleration-325-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9950144944307827, 'std': 0.00688573756573124}, annotation_units=1, channel_map={'90136_Fraction-acceleration-325-milligravities_0_0': 0, })
ukb_90137_0 = TensorMap('90137_Fraction-acceleration-350-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9958143979941174, 'std': 0.006392347972267872}, annotation_units=1, channel_map={'90137_Fraction-acceleration-350-milligravities_0_0': 0, })
ukb_90138_0 = TensorMap('90138_Fraction-acceleration-375-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9964408987897199, 'std': 0.005995999357561531}, annotation_units=1, channel_map={'90138_Fraction-acceleration-375-milligravities_0_0': 0, })
ukb_90139_0 = TensorMap('90139_Fraction-acceleration-400-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9969382130285936, 'std': 0.005691974016055797}, annotation_units=1, channel_map={'90139_Fraction-acceleration-400-milligravities_0_0': 0, })
ukb_90140_0 = TensorMap('90140_Fraction-acceleration-425-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9973430637928541, 'std': 0.00546682919665486}, annotation_units=1, channel_map={'90140_Fraction-acceleration-425-milligravities_0_0': 0, })
ukb_90141_0 = TensorMap('90141_Fraction-acceleration-450-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9976744394618836, 'std': 0.0052852627167081605}, annotation_units=1, channel_map={'90141_Fraction-acceleration-450-milligravities_0_0': 0, })
ukb_90142_0 = TensorMap('90142_Fraction-acceleration-475-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9979510583924006, 'std': 0.005138435576908388}, annotation_units=1, channel_map={'90142_Fraction-acceleration-475-milligravities_0_0': 0, })
ukb_90143_0 = TensorMap('90143_Fraction-acceleration-500-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9981833164569169, 'std': 0.005016929252235901}, annotation_units=1, channel_map={'90143_Fraction-acceleration-500-milligravities_0_0': 0, })
ukb_90144_0 = TensorMap('90144_Fraction-acceleration-600-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9988182361733932, 'std': 0.004660288361213305}, annotation_units=1, channel_map={'90144_Fraction-acceleration-600-milligravities_0_0': 0, })
ukb_90145_0 = TensorMap('90145_Fraction-acceleration-700-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9991858913158781, 'std': 0.0043855159117143615}, annotation_units=1, channel_map={'90145_Fraction-acceleration-700-milligravities_0_0': 0, })
ukb_90146_0 = TensorMap('90146_Fraction-acceleration-800-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9994224022373307, 'std': 0.0040383398929516754}, annotation_units=1, channel_map={'90146_Fraction-acceleration-800-milligravities_0_0': 0, })
ukb_90147_0 = TensorMap('90147_Fraction-acceleration-900-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9995828632045903, 'std': 0.0037803780755459636}, annotation_units=1, channel_map={'90147_Fraction-acceleration-900-milligravities_0_0': 0, })
ukb_90148_0 = TensorMap('90148_Fraction-acceleration-1000-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.999698538984522, 'std': 0.003524523347059043}, annotation_units=1, channel_map={'90148_Fraction-acceleration-1000-milligravities_0_0': 0, })
ukb_90149_0 = TensorMap('90149_Fraction-acceleration-1100-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.999783075365254, 'std': 0.003246607516580006}, annotation_units=1, channel_map={'90149_Fraction-acceleration-1100-milligravities_0_0': 0, })
ukb_90150_0 = TensorMap('90150_Fraction-acceleration-1200-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9998441776363373, 'std': 0.0030152885490644273}, annotation_units=1, channel_map={'90150_Fraction-acceleration-1200-milligravities_0_0': 0, })
ukb_90151_0 = TensorMap('90151_Fraction-acceleration-1300-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9998890013983315, 'std': 0.002757928944855472}, annotation_units=1, channel_map={'90151_Fraction-acceleration-1300-milligravities_0_0': 0, })
ukb_90152_0 = TensorMap('90152_Fraction-acceleration-1400-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.999920285452529, 'std': 0.0025236075170088587}, annotation_units=1, channel_map={'90152_Fraction-acceleration-1400-milligravities_0_0': 0, })
ukb_90153_0 = TensorMap('90153_Fraction-acceleration-1500-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9999418486908725, 'std': 0.0022951564321133386}, annotation_units=1, channel_map={'90153_Fraction-acceleration-1500-milligravities_0_0': 0, })
ukb_90154_0 = TensorMap('90154_Fraction-acceleration-1600-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.999956198466657, 'std': 0.0021084202911035806}, annotation_units=1, channel_map={'90154_Fraction-acceleration-1600-milligravities_0_0': 0, })
ukb_90155_0 = TensorMap('90155_Fraction-acceleration-1700-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.999965986788177, 'std': 0.001956991803023523}, annotation_units=1, channel_map={'90155_Fraction-acceleration-1700-milligravities_0_0': 0, })
ukb_90156_0 = TensorMap('90156_Fraction-acceleration-1800-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9999728820097402, 'std': 0.0017957441716811974}, annotation_units=1, channel_map={'90156_Fraction-acceleration-1800-milligravities_0_0': 0, })
ukb_90157_0 = TensorMap('90157_Fraction-acceleration-1900-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9999779256473311, 'std': 0.0016428644276848337}, annotation_units=1, channel_map={'90157_Fraction-acceleration-1900-milligravities_0_0': 0, })
ukb_90158_0 = TensorMap('90158_Fraction-acceleration-2000-milligravities_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9999824003085971, 'std': 0.0014220166230694642}, annotation_units=1, channel_map={'90158_Fraction-acceleration-2000-milligravities_0_0': 0, })
ukb_90159_0 = TensorMap('90159_Error-tolerance-before-calibration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.42395264959738, 'std': 38.2372423263187}, annotation_units=1, channel_map={'90159_Error-tolerance-before-calibration_0_0': 0, })
ukb_90160_0 = TensorMap('90160_Error-tolerance-after-calibration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.5632515550412265, 'std': 0.7807888919170612}, annotation_units=1, channel_map={'90160_Error-tolerance-after-calibration_0_0': 0, })
ukb_90161_0 = TensorMap('90161_Calibration-coefficients-x-offset_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.004632296639182217, 'std': 0.022627992536624612}, annotation_units=1, channel_map={'90161_Calibration-coefficients-x-offset_0_0': 0, })
ukb_90162_0 = TensorMap('90162_Calibration-coefficients-y-offset_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.013208814311201123, 'std': 0.025080515617090206}, annotation_units=1, channel_map={'90162_Calibration-coefficients-y-offset_0_0': 0, })
ukb_90163_0 = TensorMap('90163_Calibration-coefficients-z-offset_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.12664468875066293, 'std': 0.07037598553692073}, annotation_units=1, channel_map={'90163_Calibration-coefficients-z-offset_0_0': 0, })
ukb_90164_0 = TensorMap('90164_Calibration-coefficients-x-slope_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9877194464535416, 'std': 0.010874759864400715}, annotation_units=1, channel_map={'90164_Calibration-coefficients-x-slope_0_0': 0, })
ukb_90165_0 = TensorMap('90165_Calibration-coefficients-y-slope_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9865097159940207, 'std': 0.013083087702843672}, annotation_units=1, channel_map={'90165_Calibration-coefficients-y-slope_0_0': 0, })
ukb_90166_0 = TensorMap('90166_Calibration-coefficients-z-slope_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.00844197888037, 'std': 0.008938183729208531}, annotation_units=1, channel_map={'90166_Calibration-coefficients-z-slope_0_0': 0, })
ukb_90167_0 = TensorMap('90167_Calibration-coefficients-x-temp_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -4.48719803269203e-05, 'std': 0.0011886910403996016}, annotation_units=1, channel_map={'90167_Calibration-coefficients-x-temp_0_0': 0, })
ukb_90168_0 = TensorMap('90168_Calibration-coefficients-y-temp_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001052336178214957, 'std': 0.0012639903963996546}, annotation_units=1, channel_map={'90168_Calibration-coefficients-y-temp_0_0': 0, })
ukb_90169_0 = TensorMap('90169_Calibration-coefficients-z-temp_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.001756092386325281, 'std': 0.0044200366843543774}, annotation_units=1, channel_map={'90169_Calibration-coefficients-z-temp_0_0': 0, })
ukb_90170_0 = TensorMap('90170_Calibration-coefficients-mean-temperature_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.89597097256377, 'std': 1.6886012877960286}, annotation_units=1, channel_map={'90170_Calibration-coefficients-mean-temperature_0_0': 0, })
ukb_90171_0 = TensorMap('90171_Calibration-number-of-static-points_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 25900.544847871155, 'std': 7809.333180639968}, annotation_units=1, channel_map={'90171_Calibration-number-of-static-points_0_0': 0, })
ukb_90172_0 = TensorMap('90172_Calibration-minimum-x-stationary-value_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.976558754038285, 'std': 0.17685980211984123}, annotation_units=1, channel_map={'90172_Calibration-minimum-x-stationary-value_0_0': 0, })
ukb_90173_0 = TensorMap('90173_Calibration-maximum-x-stationary-value_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9643294276483918, 'std': 0.1755304898544724}, annotation_units=1, channel_map={'90173_Calibration-maximum-x-stationary-value_0_0': 0, })
ukb_90174_0 = TensorMap('90174_Calibration-minimum-y-stationary-value_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -0.9894458749216452, 'std': 0.18071984845447228}, annotation_units=1, channel_map={'90174_Calibration-minimum-y-stationary-value_0_0': 0, })
ukb_90175_0 = TensorMap('90175_Calibration-maximum-y-stationary-value_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.9632147162351126, 'std': 0.17389544271220922}, annotation_units=1, channel_map={'90175_Calibration-maximum-y-stationary-value_0_0': 0, })
ukb_90176_0 = TensorMap('90176_Calibration-minimum-z-stationary-value_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': -1.0817622836202323, 'std': 0.20687482049366254}, annotation_units=1, channel_map={'90176_Calibration-minimum-z-stationary-value_0_0': 0, })
ukb_90177_0 = TensorMap('90177_Calibration-maximum-z-stationary-value_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.849756111673658, 'std': 0.16670333138533874}, annotation_units=1, channel_map={'90177_Calibration-maximum-z-stationary-value_0_0': 0, })
ukb_90179_0 = TensorMap('90179_Device-ID_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 7122.602738801292, 'std': 5264.016557806373}, annotation_units=1, channel_map={'90179_Device-ID_0_0': 0, })
ukb_90180_0 = TensorMap('90180_Interrupted-recording-periods_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.04432229133516563, 'std': 1.6731041903517416}, annotation_units=1, channel_map={'90180_Interrupted-recording-periods_0_0': 0, })
ukb_90181_0 = TensorMap('90181_Duration-of-interrupted-recording-periods_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 0.8079965282800524, 'std': 46.99727836703531}, annotation_units=1, channel_map={'90181_Duration-of-interrupted-recording-periods_0_0': 0, })
ukb_90182_0 = TensorMap('90182_Data-recording-errors_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 232.43163122619217, 'std': 1096.9020106007322}, annotation_units=1, channel_map={'90182_Data-recording-errors_0_0': 0, })
ukb_90183_0 = TensorMap('90183_Readings-exceeding-8-gravities-before-calibration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 425.532253242683, 'std': 6632.992953339034}, annotation_units=1, channel_map={'90183_Readings-exceeding-8-gravities-before-calibration_0_0': 0, })
ukb_90184_0 = TensorMap('90184_Maximum-readings-exceeding-8-gravities-before-calibration-in-a-5-second-epoch_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.386566372534837, 'std': 18.692966630166143}, annotation_units=1, channel_map={'90184_Maximum-readings-exceeding-8-gravities-before-calibration-in-a-5-second-epoch_0_0': 0, })
ukb_90185_0 = TensorMap('90185_Readings-exceeding-8-gravities-after-calibration_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 522.3285115000723, 'std': 27295.393171089177}, annotation_units=1, channel_map={'90185_Readings-exceeding-8-gravities-after-calibration_0_0': 0, })
ukb_90186_0 = TensorMap('90186_Maximum-readings-exceeding-8-gravities-after-calibration-in-a-5-second-epoch_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.61730073774049, 'std': 18.944154910931644}, annotation_units=1, channel_map={'90186_Maximum-readings-exceeding-8-gravities-after-calibration-in-a-5-second-epoch_0_0': 0, })
ukb_90187_0 = TensorMap('90187_Total-data-readings_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 54066561.1910314, 'std': 13123664.817067225}, annotation_units=1, channel_map={'90187_Total-data-readings_0_0': 0, })
ukb_90188_0 = TensorMap('90188_Sample-rate-average_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 99.98452866579875, 'std': 74.59443675687426}, annotation_units=1, channel_map={'90188_Sample-rate-average_0_0': 0, })
ukb_90189_0 = TensorMap('90189_Sample-rate-standard-deviation_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 45.92016779979749, 'std': 1300.4627411111455}, annotation_units=1, channel_map={'90189_Sample-rate-standard-deviation_0_0': 0, })
ukb_90190_0 = TensorMap('90190_Sample-rate-minimum_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 57.281677997974825, 'std': 31.634722185781893}, annotation_units=1, channel_map={'90190_Sample-rate-minimum_0_0': 0, })
ukb_90191_0 = TensorMap('90191_Sample-rate-maximum_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8184.358912194418, 'std': 220940.12536187496}, annotation_units=1, channel_map={'90191_Sample-rate-maximum_0_0': 0, })
ukb_90192_0 = TensorMap('90192_Temperature-average_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 20.348456531173152, 'std': 1.5328010781114556}, annotation_units=1, channel_map={'90192_Temperature-average_0_0': 0, })
ukb_90193_0 = TensorMap('90193_Temperature-standard-deviation_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 1.5691094073966914, 'std': 0.4871367060327803}, annotation_units=1, channel_map={'90193_Temperature-standard-deviation_0_0': 0, })
ukb_90194_0 = TensorMap('90194_Temperature-minimum_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8.30701480302811, 'std': 3.4470480575742024}, annotation_units=1, channel_map={'90194_Temperature-minimum_0_0': 0, })
ukb_90195_0 = TensorMap('90195_Temperature-maximum_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 24.13212883938473, 'std': 2.024626480148313}, annotation_units=1, channel_map={'90195_Temperature-maximum_0_0': 0, })
ukb_100001_0 = TensorMap('100001_Food-weight_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3263.2873074476574, 'std': 939.1711925662723}, annotation_units=1, channel_map={'100001_Food-weight_0_0': 0, })
ukb_100001_1 = TensorMap('100001_Food-weight_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3184.0043906003034, 'std': 843.8812188783728}, annotation_units=1, channel_map={'100001_Food-weight_1_0': 0, })
ukb_100001_4 = TensorMap('100001_Food-weight_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3180.2928506611343, 'std': 849.8447851374056}, annotation_units=1, channel_map={'100001_Food-weight_4_0': 0, })
ukb_100001_3 = TensorMap('100001_Food-weight_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3119.3004597701165, 'std': 829.8747298336872}, annotation_units=1, channel_map={'100001_Food-weight_3_0': 0, })
ukb_100001_2 = TensorMap('100001_Food-weight_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3232.7114414219623, 'std': 868.9299727668377}, annotation_units=1, channel_map={'100001_Food-weight_2_0': 0, })
ukb_100002_0 = TensorMap('100002_Energy_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8825.824255162745, 'std': 3229.866916357447}, annotation_units=1, channel_map={'100002_Energy_0_0': 0, })
ukb_100002_4 = TensorMap('100002_Energy_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8881.947846074047, 'std': 3032.3527502998813}, annotation_units=1, channel_map={'100002_Energy_4_0': 0, })
ukb_100002_2 = TensorMap('100002_Energy_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8864.672170539843, 'std': 3023.516248262591}, annotation_units=1, channel_map={'100002_Energy_2_0': 0, })
ukb_100002_3 = TensorMap('100002_Energy_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8782.807566262945, 'std': 2951.073678156437}, annotation_units=1, channel_map={'100002_Energy_3_0': 0, })
ukb_100002_1 = TensorMap('100002_Energy_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 8868.971860058307, 'std': 2978.922159133099}, annotation_units=1, channel_map={'100002_Energy_1_0': 0, })
ukb_100003_0 = TensorMap('100003_Protein_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 81.80167003547854, 'std': 31.330517846881715}, annotation_units=1, channel_map={'100003_Protein_0_0': 0, })
ukb_100003_3 = TensorMap('100003_Protein_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.06538283666204, 'std': 28.855564888740236}, annotation_units=1, channel_map={'100003_Protein_3_0': 0, })
ukb_100003_1 = TensorMap('100003_Protein_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.87431505283372, 'std': 29.262260867430726}, annotation_units=1, channel_map={'100003_Protein_1_0': 0, })
ukb_100003_4 = TensorMap('100003_Protein_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.65940926587002, 'std': 29.671199909468076}, annotation_units=1, channel_map={'100003_Protein_4_0': 0, })
ukb_100003_2 = TensorMap('100003_Protein_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 82.49425088572622, 'std': 29.98023378065929}, annotation_units=1, channel_map={'100003_Protein_2_0': 0, })
ukb_100004_1 = TensorMap('100004_Fat_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.3367746200261, 'std': 34.658044544818495}, annotation_units=1, channel_map={'100004_Fat_1_0': 0, })
ukb_100004_2 = TensorMap('100004_Fat_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 78.13620945175046, 'std': 34.972992000143506}, annotation_units=1, channel_map={'100004_Fat_2_0': 0, })
ukb_100004_3 = TensorMap('100004_Fat_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 77.86341118208698, 'std': 34.21845492958865}, annotation_units=1, channel_map={'100004_Fat_3_0': 0, })
ukb_100004_4 = TensorMap('100004_Fat_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 79.38589946351281, 'std': 35.24871981010938}, annotation_units=1, channel_map={'100004_Fat_4_0': 0, })
ukb_100004_0 = TensorMap('100004_Fat_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 76.75582300309543, 'std': 36.27309870587293}, annotation_units=1, channel_map={'100004_Fat_0_0': 0, })
ukb_100005_4 = TensorMap('100005_Carbohydrate_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 254.13792096288475, 'std': 100.94090840812653}, annotation_units=1, channel_map={'100005_Carbohydrate_4_0': 0, })
ukb_100005_3 = TensorMap('100005_Carbohydrate_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 252.3601259261402, 'std': 98.78530366558122}, annotation_units=1, channel_map={'100005_Carbohydrate_3_0': 0, })
ukb_100005_0 = TensorMap('100005_Carbohydrate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 258.13639009427897, 'std': 108.62251937217879}, annotation_units=1, channel_map={'100005_Carbohydrate_0_0': 0, })
ukb_100005_1 = TensorMap('100005_Carbohydrate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 255.57288998896558, 'std': 99.49372178780574}, annotation_units=1, channel_map={'100005_Carbohydrate_1_0': 0, })
ukb_100005_2 = TensorMap('100005_Carbohydrate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 254.11853167597374, 'std': 100.83178290990924}, annotation_units=1, channel_map={'100005_Carbohydrate_2_0': 0, })
ukb_100006_1 = TensorMap('100006_Saturated-fat_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.06014632352211, 'std': 15.085404246067869}, annotation_units=1, channel_map={'100006_Saturated-fat_1_0': 0, })
ukb_100006_2 = TensorMap('100006_Saturated-fat_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.987648471746855, 'std': 15.348406205897625}, annotation_units=1, channel_map={'100006_Saturated-fat_2_0': 0, })
ukb_100006_3 = TensorMap('100006_Saturated-fat_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.122623252497764, 'std': 15.030865867921447}, annotation_units=1, channel_map={'100006_Saturated-fat_3_0': 0, })
ukb_100006_4 = TensorMap('100006_Saturated-fat_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 30.524596737201158, 'std': 15.431147922587424}, annotation_units=1, channel_map={'100006_Saturated-fat_4_0': 0, })
ukb_100006_0 = TensorMap('100006_Saturated-fat_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 29.519929749671345, 'std': 15.821915288796678}, annotation_units=1, channel_map={'100006_Saturated-fat_0_0': 0, })
ukb_100007_2 = TensorMap('100007_Polyunsaturated-fat_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.373419443944046, 'std': 8.505740790816308}, annotation_units=1, channel_map={'100007_Polyunsaturated-fat_2_0': 0, })
ukb_100007_4 = TensorMap('100007_Polyunsaturated-fat_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.604646297441237, 'std': 8.638209201072604}, annotation_units=1, channel_map={'100007_Polyunsaturated-fat_4_0': 0, })
ukb_100007_1 = TensorMap('100007_Polyunsaturated-fat_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.450855575105154, 'std': 8.561629456759748}, annotation_units=1, channel_map={'100007_Polyunsaturated-fat_1_0': 0, })
ukb_100007_3 = TensorMap('100007_Polyunsaturated-fat_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.254366370879922, 'std': 8.532226210287696}, annotation_units=1, channel_map={'100007_Polyunsaturated-fat_3_0': 0, })
ukb_100007_0 = TensorMap('100007_Polyunsaturated-fat_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 14.232844784937871, 'std': 8.825575758319898}, annotation_units=1, channel_map={'100007_Polyunsaturated-fat_0_0': 0, })
ukb_100008_1 = TensorMap('100008_Total-sugars_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 121.15616208908614, 'std': 56.237269522080226}, annotation_units=1, channel_map={'100008_Total-sugars_1_0': 0, })
ukb_100008_3 = TensorMap('100008_Total-sugars_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 118.51599377595355, 'std': 55.45124186024215}, annotation_units=1, channel_map={'100008_Total-sugars_3_0': 0, })
ukb_100008_0 = TensorMap('100008_Total-sugars_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 123.24034361881073, 'std': 61.98506769921744}, annotation_units=1, channel_map={'100008_Total-sugars_0_0': 0, })
ukb_100008_4 = TensorMap('100008_Total-sugars_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 120.16173201571584, 'std': 56.4110885905138}, annotation_units=1, channel_map={'100008_Total-sugars_4_0': 0, })
ukb_100008_2 = TensorMap('100008_Total-sugars_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 120.93616957905489, 'std': 56.76787562654187}, annotation_units=1, channel_map={'100008_Total-sugars_2_0': 0, })
ukb_100009_2 = TensorMap('100009_Englyst-dietary-fibre_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.377987029364075, 'std': 7.331710078565476}, annotation_units=1, channel_map={'100009_Englyst-dietary-fibre_2_0': 0, })
ukb_100009_1 = TensorMap('100009_Englyst-dietary-fibre_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.566871141860247, 'std': 7.299273313419655}, annotation_units=1, channel_map={'100009_Englyst-dietary-fibre_1_0': 0, })
ukb_100009_0 = TensorMap('100009_Englyst-dietary-fibre_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.643010021626434, 'std': 7.874280707474585}, annotation_units=1, channel_map={'100009_Englyst-dietary-fibre_0_0': 0, })
ukb_100009_4 = TensorMap('100009_Englyst-dietary-fibre_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.32643515286891, 'std': 7.194617777885266}, annotation_units=1, channel_map={'100009_Englyst-dietary-fibre_4_0': 0, })
ukb_100009_3 = TensorMap('100009_Englyst-dietary-fibre_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.415944638745152, 'std': 7.213252151919217}, annotation_units=1, channel_map={'100009_Englyst-dietary-fibre_3_0': 0, })
ukb_100011_3 = TensorMap('100011_Iron_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.637650759699763, 'std': 4.908695253160161}, annotation_units=1, channel_map={'100011_Iron_3_0': 0, })
ukb_100011_1 = TensorMap('100011_Iron_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.756281473970933, 'std': 4.953505319903377}, annotation_units=1, channel_map={'100011_Iron_1_0': 0, })
ukb_100011_2 = TensorMap('100011_Iron_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.879047138653664, 'std': 5.032768623365607}, annotation_units=1, channel_map={'100011_Iron_2_0': 0, })
ukb_100011_0 = TensorMap('100011_Iron_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.648492798281188, 'std': 5.320383273608107}, annotation_units=1, channel_map={'100011_Iron_0_0': 0, })
ukb_100011_4 = TensorMap('100011_Iron_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 13.761195329171692, 'std': 4.975550594928983}, annotation_units=1, channel_map={'100011_Iron_4_0': 0, })
ukb_100012_0 = TensorMap('100012_Vitamin-B6_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.1851879231628226, 'std': 0.8770346918366498}, annotation_units=1, channel_map={'100012_Vitamin-B6_0_0': 0, })
ukb_100012_4 = TensorMap('100012_Vitamin-B6_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.1835743204164215, 'std': 0.8272890221569056}, annotation_units=1, channel_map={'100012_Vitamin-B6_4_0': 0, })
ukb_100012_2 = TensorMap('100012_Vitamin-B6_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.189867531375727, 'std': 0.8369495308468423}, annotation_units=1, channel_map={'100012_Vitamin-B6_2_0': 0, })
ukb_100012_3 = TensorMap('100012_Vitamin-B6_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.146791629332015, 'std': 0.806491743055921}, annotation_units=1, channel_map={'100012_Vitamin-B6_3_0': 0, })
ukb_100012_1 = TensorMap('100012_Vitamin-B6_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.176045288720556, 'std': 0.8156806079274149}, annotation_units=1, channel_map={'100012_Vitamin-B6_1_0': 0, })
ukb_100013_1 = TensorMap('100013_Vitamin-B12_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.632391276255207, 'std': 5.587200469751595}, annotation_units=1, channel_map={'100013_Vitamin-B12_1_0': 0, })
ukb_100013_3 = TensorMap('100013_Vitamin-B12_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.449235290150414, 'std': 5.452402635403458}, annotation_units=1, channel_map={'100013_Vitamin-B12_3_0': 0, })
ukb_100013_0 = TensorMap('100013_Vitamin-B12_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.353503046065566, 'std': 5.466821683842562}, annotation_units=1, channel_map={'100013_Vitamin-B12_0_0': 0, })
ukb_100013_2 = TensorMap('100013_Vitamin-B12_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.586702576112411, 'std': 5.707475292987482}, annotation_units=1, channel_map={'100013_Vitamin-B12_2_0': 0, })
ukb_100013_4 = TensorMap('100013_Vitamin-B12_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 6.683704652878881, 'std': 5.673911722897342}, annotation_units=1, channel_map={'100013_Vitamin-B12_4_0': 0, })
ukb_100014_1 = TensorMap('100014_Folate_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 305.8258464795867, 'std': 125.59889008317676}, annotation_units=1, channel_map={'100014_Folate_1_0': 0, })
ukb_100014_4 = TensorMap('100014_Folate_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 302.2448234977364, 'std': 122.85426887994477}, annotation_units=1, channel_map={'100014_Folate_4_0': 0, })
ukb_100014_2 = TensorMap('100014_Folate_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 300.8046193478662, 'std': 123.44103045115011}, annotation_units=1, channel_map={'100014_Folate_2_0': 0, })
ukb_100014_3 = TensorMap('100014_Folate_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 300.900947095606, 'std': 124.30395972674758}, annotation_units=1, channel_map={'100014_Folate_3_0': 0, })
ukb_100014_0 = TensorMap('100014_Folate_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 307.6532973836351, 'std': 135.19408933325468}, annotation_units=1, channel_map={'100014_Folate_0_0': 0, })
ukb_100015_2 = TensorMap('100015_Vitamin-C_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 164.0119361076083, 'std': 121.63779641726717}, annotation_units=1, channel_map={'100015_Vitamin-C_2_0': 0, })
ukb_100015_4 = TensorMap('100015_Vitamin-C_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 151.56502762210582, 'std': 114.23808257500417}, annotation_units=1, channel_map={'100015_Vitamin-C_4_0': 0, })
ukb_100015_0 = TensorMap('100015_Vitamin-C_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 154.23202566893315, 'std': 129.2278644203107}, annotation_units=1, channel_map={'100015_Vitamin-C_0_0': 0, })
ukb_100015_1 = TensorMap('100015_Vitamin-C_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 150.18312806290302, 'std': 114.6724241350799}, annotation_units=1, channel_map={'100015_Vitamin-C_1_0': 0, })
ukb_100015_3 = TensorMap('100015_Vitamin-C_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 142.0416125675637, 'std': 108.68220814088339}, annotation_units=1, channel_map={'100015_Vitamin-C_3_0': 0, })
ukb_100016_1 = TensorMap('100016_Potassium_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3798.061823775578, 'std': 1444.2931270507665}, annotation_units=1, channel_map={'100016_Potassium_1_0': 0, })
ukb_100016_4 = TensorMap('100016_Potassium_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3778.0538044713903, 'std': 1450.8005107398185}, annotation_units=1, channel_map={'100016_Potassium_4_0': 0, })
ukb_100016_2 = TensorMap('100016_Potassium_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3802.304902179792, 'std': 1467.1970877100725}, annotation_units=1, channel_map={'100016_Potassium_2_0': 0, })
ukb_100016_0 = TensorMap('100016_Potassium_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3779.994103354204, 'std': 1559.9782812695232}, annotation_units=1, channel_map={'100016_Potassium_0_0': 0, })
ukb_100016_3 = TensorMap('100016_Potassium_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3744.117824377835, 'std': 1422.9870125181951}, annotation_units=1, channel_map={'100016_Potassium_3_0': 0, })
ukb_100017_0 = TensorMap('100017_Magnesium_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 349.6639003774008, 'std': 129.84604714346878}, annotation_units=1, channel_map={'100017_Magnesium_0_0': 0, })
ukb_100017_1 = TensorMap('100017_Magnesium_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 351.1433965546371, 'std': 120.39694622954579}, annotation_units=1, channel_map={'100017_Magnesium_1_0': 0, })
ukb_100017_4 = TensorMap('100017_Magnesium_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 351.1619144013881, 'std': 121.84036171278738}, annotation_units=1, channel_map={'100017_Magnesium_4_0': 0, })
ukb_100017_2 = TensorMap('100017_Magnesium_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 349.2335239296218, 'std': 121.46561958628556}, annotation_units=1, channel_map={'100017_Magnesium_2_0': 0, })
ukb_100017_3 = TensorMap('100017_Magnesium_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 347.4513896195235, 'std': 119.32574493357643}, annotation_units=1, channel_map={'100017_Magnesium_3_0': 0, })
ukb_100018_2 = TensorMap('100018_Retinol_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 323.03031284971814, 'std': 200.08480831545492}, annotation_units=1, channel_map={'100018_Retinol_2_0': 0, })
ukb_100018_4 = TensorMap('100018_Retinol_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 330.2788852347116, 'std': 201.25250974972818}, annotation_units=1, channel_map={'100018_Retinol_4_0': 0, })
ukb_100018_1 = TensorMap('100018_Retinol_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 327.24587060521674, 'std': 199.073798085362}, annotation_units=1, channel_map={'100018_Retinol_1_0': 0, })
ukb_100018_0 = TensorMap('100018_Retinol_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 314.2649757391184, 'std': 200.19802174665372}, annotation_units=1, channel_map={'100018_Retinol_0_0': 0, })
ukb_100018_3 = TensorMap('100018_Retinol_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 329.9855860558613, 'std': 199.6004935410949}, annotation_units=1, channel_map={'100018_Retinol_3_0': 0, })
ukb_100019_1 = TensorMap('100019_Carotene_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3191.959398999989, 'std': 3056.498423025295}, annotation_units=1, channel_map={'100019_Carotene_1_0': 0, })
ukb_100019_2 = TensorMap('100019_Carotene_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2983.205142977243, 'std': 2865.778477977818}, annotation_units=1, channel_map={'100019_Carotene_2_0': 0, })
ukb_100019_3 = TensorMap('100019_Carotene_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3117.200206183588, 'std': 2947.4077166625775}, annotation_units=1, channel_map={'100019_Carotene_3_0': 0, })
ukb_100019_0 = TensorMap('100019_Carotene_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3211.757054716104, 'std': 3273.2902655356083}, annotation_units=1, channel_map={'100019_Carotene_0_0': 0, })
ukb_100019_4 = TensorMap('100019_Carotene_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2999.491419596738, 'std': 2879.6710502115293}, annotation_units=1, channel_map={'100019_Carotene_4_0': 0, })
ukb_100021_1 = TensorMap('100021_Vitamin-D_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.871529041044142, 'std': 3.320970541110873}, annotation_units=1, channel_map={'100021_Vitamin-D_1_0': 0, })
ukb_100021_4 = TensorMap('100021_Vitamin-D_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 3.0253114217905495, 'std': 3.465652683391171}, annotation_units=1, channel_map={'100021_Vitamin-D_4_0': 0, })
ukb_100021_2 = TensorMap('100021_Vitamin-D_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.9850030625112582, 'std': 3.457097144235845}, annotation_units=1, channel_map={'100021_Vitamin-D_2_0': 0, })
ukb_100021_3 = TensorMap('100021_Vitamin-D_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7739618078638784, 'std': 3.192748906139195}, annotation_units=1, channel_map={'100021_Vitamin-D_3_0': 0, })
ukb_100021_0 = TensorMap('100021_Vitamin-D_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 2.7866783043803993, 'std': 3.338445090473445}, annotation_units=1, channel_map={'100021_Vitamin-D_0_0': 0, })
ukb_100022_1 = TensorMap('100022_Alcohol_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.075506217755645, 'std': 23.191554372155892}, annotation_units=1, channel_map={'100022_Alcohol_1_0': 0, })
ukb_100022_0 = TensorMap('100022_Alcohol_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.87385910356625, 'std': 24.128454819959085}, annotation_units=1, channel_map={'100022_Alcohol_0_0': 0, })
ukb_100022_2 = TensorMap('100022_Alcohol_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 17.258375427850837, 'std': 24.09995558240885}, annotation_units=1, channel_map={'100022_Alcohol_2_0': 0, })
ukb_100022_3 = TensorMap('100022_Alcohol_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 15.948475494021634, 'std': 23.06622063364617}, annotation_units=1, channel_map={'100022_Alcohol_3_0': 0, })
ukb_100022_4 = TensorMap('100022_Alcohol_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 16.129553359526152, 'std': 23.256385523870556}, annotation_units=1, channel_map={'100022_Alcohol_4_0': 0, })
ukb_100023_2 = TensorMap('100023_Starch_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 120.73261706599396, 'std': 52.394631269316115}, annotation_units=1, channel_map={'100023_Starch_2_0': 0, })
ukb_100023_3 = TensorMap('100023_Starch_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 122.77632617471676, 'std': 52.293392219643785}, annotation_units=1, channel_map={'100023_Starch_3_0': 0, })
ukb_100023_4 = TensorMap('100023_Starch_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 122.05941824056146, 'std': 52.96447249391592}, annotation_units=1, channel_map={'100023_Starch_4_0': 0, })
ukb_100023_0 = TensorMap('100023_Starch_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 123.36970754943658, 'std': 56.41255225846996}, annotation_units=1, channel_map={'100023_Starch_0_0': 0, })
ukb_100023_1 = TensorMap('100023_Starch_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 123.05099175936168, 'std': 52.48096849530664}, annotation_units=1, channel_map={'100023_Starch_1_0': 0, })
ukb_100024_1 = TensorMap('100024_Calcium_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 992.2979304963279, 'std': 445.04861600122393}, annotation_units=1, channel_map={'100024_Calcium_1_0': 0, })
ukb_100024_2 = TensorMap('100024_Calcium_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 968.0836015132412, 'std': 441.97935479697077}, annotation_units=1, channel_map={'100024_Calcium_2_0': 0, })
ukb_100024_4 = TensorMap('100024_Calcium_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 985.3634699148411, 'std': 446.8720587603772}, annotation_units=1, channel_map={'100024_Calcium_4_0': 0, })
ukb_100024_0 = TensorMap('100024_Calcium_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 987.0313623192508, 'std': 470.33844826989474}, annotation_units=1, channel_map={'100024_Calcium_0_0': 0, })
ukb_100024_3 = TensorMap('100024_Calcium_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 976.3927741326331, 'std': 439.750722149824}, annotation_units=1, channel_map={'100024_Calcium_3_0': 0, })
ukb_100025_4 = TensorMap('100025_Vitamin-E_4_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.500682275981749, 'std': 5.424533805765152}, annotation_units=1, channel_map={'100025_Vitamin-E_4_0': 0, })
ukb_100025_0 = TensorMap('100025_Vitamin-E_0_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.228437530920038, 'std': 5.565657210717742}, annotation_units=1, channel_map={'100025_Vitamin-E_0_0': 0, })
ukb_100025_1 = TensorMap('100025_Vitamin-E_1_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.42730514219824, 'std': 5.3821109971739505}, annotation_units=1, channel_map={'100025_Vitamin-E_1_0': 0, })
ukb_100025_3 = TensorMap('100025_Vitamin-E_3_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.175370215143888, 'std': 5.253523156045647}, annotation_units=1, channel_map={'100025_Vitamin-E_3_0': 0, })
ukb_100025_2 = TensorMap('100025_Vitamin-E_2_0', loss='logcosh', path_prefix='continuous', normalization={'mean': 9.485245541343879, 'std': 5.431265583618386}, annotation_units=1, channel_map={'100025_Vitamin-E_2_0': 0, })
